/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
*
*  版权所有(c)，2020-2099，中船第709研究所
*
*******************************************************************************
*
*  文件名称: gp200.c
*  版 本 号: V1.0
*  作    者: 胡艳明
*  生成日期: 2022年4月21日
*
*  功能描述: gp200
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: huym
*            修改内容: 创建文件
*
******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/
#include "ljmicro_device.h"
#include "ljmicro_mode.h"
#include "gp200.h"
#include "ljmicro_hdmi.h"
#include "ljmicro_compat.h"
#include "ljmicro_ddc.h"
#include "ljmicro_vgav2.h"
#include "ljmicro_clk.h"
#include "ljmicro_zoomvideo.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

u32 g_devpll_val = 0x0;
u32 g_dev2pll_val = 0x0;
u32 g_gpupll_val = 0x0;
u32 g_gpu2pll_val = 0x0;
u32 g_ckmux_val = 0x0;

/*****************************************************************************
 *  函 数 名: ljm_pre_get_clock
 *  功能描述: 获取当前的时钟设置
 *  输入参数: void *mmio_base
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2023年2月23日
 *            作    者: huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_pre_get_clock(void *mmio_base)
{
    g_devpll_val  = LJM_CRU_R32_BASE(mmio_base, CRU_DEVPLL);  /* get DEVPLL value */
    g_dev2pll_val = LJM_CRU_R32_BASE(mmio_base, CRU_DEV2PLL); /* get DEV2PLL value */
    g_gpupll_val  = LJM_CRU_R32_BASE(mmio_base, CRU_GPUPLL);  /* get GPUPLL value */
    g_gpu2pll_val = LJM_CRU_R32_BASE(mmio_base, CRU_GPU2PLL); /* get GPU2PLL value */
    g_ckmux_val   = LJM_CRU_R32_BASE(mmio_base, CRU_CKMUX);   /* get CKMUX */
}

//void ljm_ddrv2_set_urgent_qos(void *mmio_base, int channel_mask);

/**
 * start physical address of high memio is 0x8000000000
 */
#define LJM_HIGH_MEMIO_BANKS    32

struct ljm_high_memio_bank {
    unsigned long src_start;        /* cpu physical address start */
    unsigned long dst_start;    /* gpu physical address start */
    unsigned long size;
};

struct ljm_high_memio_info {
    int nr_banks;               /* count of banks */
    struct ljm_high_memio_bank banks[LJM_HIGH_MEMIO_BANKS];
};

#define HIGN_MEMIO_BLOCK_START  0x8000000000UL  /* start of high memio */
#define HIGN_MEMIO_BLOCK_SIZE   0x100000000UL   /* 4GB block size */

struct ljm_high_memio_info g_high_memio_info = {0};

/* get gtt memio info */
void gp200_high_memio_init(struct ljm_high_memio_info *high_memio_info)
{
    struct resource *iomem = iomem_resource.child;
    int i;
    unsigned long cur_size, total_size;
    struct ljm_high_memio_bank *bank;
    unsigned long cur_gap_size, cur_gap_start, max_gap_size, max_gap_start;
    unsigned long former_iomem_start, former_iomem_size;
    unsigned long src_start;

    memset((void*)high_memio_info, 0, sizeof(struct ljm_high_memio_info));

    while (iomem) {
        //printk(KERN_NOTICE "----start=0x%llx end=0x%llx flags=0x%lx name=%s\n",
        //        iomem->start, iomem->end, iomem->flags, iomem->name);
        if ((iomem->flags & IORESOURCE_MEM) && (iomem->name && strcmp("System RAM", iomem->name) == 0)) {
            //printk("----index %d: start=0x%llx end=0x%llx flags=0x%lx name=%s\n",
            //    high_memio_info->nr_banks, iomem->start, iomem->end, iomem->flags, iomem->name);
            //if (high_memio_info->nr_banks > 0) {
            //    printk("----rank %d: start=0x%lx end=0x%lx left_memio=0x%lx\n",
            //        high_memio_info->nr_banks-1, high_memio_info->bank[high_memio_info->nr_banks - 1].start,
            //        high_memio_info->bank[high_memio_info->nr_banks - 1].end,
            //        left_memio);
            //}

            if (iomem->start < HIGN_MEMIO_BLOCK_START) {
                iomem = iomem->sibling;
                continue;
            }

            if (high_memio_info->nr_banks == 0) {   /* bank 0 */
                bank = &high_memio_info->banks[0];
                bank->dst_start = iomem->start;
                bank->size = (iomem->end - iomem->start + HIGN_MEMIO_BLOCK_SIZE) / HIGN_MEMIO_BLOCK_SIZE * HIGN_MEMIO_BLOCK_SIZE;
            }
            else {
                bank = &high_memio_info->banks[high_memio_info->nr_banks - 1];
                if (iomem->start < bank->dst_start + bank->size) { /* check and merge into the former bank */
                    //printk("----merge into rank %d\n", high_memio_info->nr_banks - 1);
                    cur_size = (iomem->end - bank->dst_start + HIGN_MEMIO_BLOCK_SIZE) / HIGN_MEMIO_BLOCK_SIZE * HIGN_MEMIO_BLOCK_SIZE;
                    if (bank->size < cur_size)
                        bank->size = cur_size;
                } else {    /* add a new bank */
                    bank = &high_memio_info->banks[high_memio_info->nr_banks];
                    bank->dst_start = iomem->start;
                    bank->size = (iomem->end - iomem->start + HIGN_MEMIO_BLOCK_SIZE) / HIGN_MEMIO_BLOCK_SIZE * HIGN_MEMIO_BLOCK_SIZE;
                }
            }
            high_memio_info->nr_banks++;
        }

        // skip children
        //if (iomem->child)
        //    iomem = iomem->child;
        //while (!iomem->sibling && iomem->parent)
        //    iomem = iomem->parent;
        iomem = iomem->sibling;
    }

    /* calculate total size */
    for (i=0, total_size = 0; i<high_memio_info->nr_banks; i++) {
        //printk(KERN_NOTICE "memio: rank %d dst_start=0x%lx size=0x%lx\n",
        //    i, high_memio_info->banks[i].dst_start, high_memio_info->banks[i].size);
        total_size += high_memio_info->banks[i].size;
    }
    //printk(KERN_NOTICE "high memio (>=0x%lx): total_size=0x%lx\n", HIGN_MEMIO_BLOCK_START, total_size);

    /* find a low memio gap for all high memio banks */
    cur_gap_size = cur_gap_start = max_gap_size = max_gap_start = 0;
    former_iomem_start = former_iomem_size = 0;
    iomem = iomem_resource.child;
    while (iomem) {
        if (iomem->start < cur_gap_start + cur_gap_size) {
            printk(KERN_ERR "gap (%lx@%lx) over iomem(%llx-%llx)\n", cur_gap_start, cur_gap_size, iomem->start, iomem->end);
        }
        cur_gap_start = former_iomem_start + former_iomem_size;
        cur_gap_size = iomem->start -cur_gap_start;
        former_iomem_start = iomem->start;
        former_iomem_size = iomem->end - iomem->start + 1;

        if (cur_gap_size == 0) {
            iomem = iomem->sibling;
            continue;
        }

        /* make sure gap is under HIGN_MEMIO_BLOCK_START */
        if (cur_gap_start + cur_gap_size > HIGN_MEMIO_BLOCK_START) {
            if (cur_gap_start < HIGN_MEMIO_BLOCK_START)
                cur_gap_size = HIGN_MEMIO_BLOCK_START - cur_gap_start;
            else
                cur_gap_size = 0;
        }

        //printk(KERN_NOTICE "----cur_gap (%lx-%lx)----\n", cur_gap_start, cur_gap_start + cur_gap_size);
        if (cur_gap_size > max_gap_size) {
            max_gap_start = cur_gap_start;
            max_gap_size = cur_gap_size;
            //printk(KERN_NOTICE "----max_gap (%lx-%lx)----\n", max_gap_start, max_gap_start + max_gap_size);
        }

        if (iomem->start >= HIGN_MEMIO_BLOCK_START)
            break;

        iomem = iomem->sibling;
    }

    /* set src_start */
    if (max_gap_size < total_size) {
        printk(KERN_ERR "max iomem gap size 0x%lx is less than total high iomem size 0x%lx\n", max_gap_size, total_size);
        src_start = max_gap_start;
    } else {
        src_start = max_gap_size - total_size + max_gap_start;  /* src_start is at top of max gap to distinguish */
    }
    for (i=0; i<high_memio_info->nr_banks; i++) {
        bank = &high_memio_info->banks[i];
        bank->src_start = src_start;
        src_start += bank->size;
        //printk(KERN_NOTICE "memio: bank %d dst_start=0x%lx src_start=0x%lx size=0x%lx\n",
        //    i, bank->dst_start, bank->src_start, bank->size);
    }
}

void gp200_iatu_init(void *mmio_5_base)
{
    int i, j;

    for (i=0, j=0; i<g_high_memio_info.nr_banks; i++) {
        struct ljm_high_memio_bank *bank = &g_high_memio_info.banks[i];
        unsigned long left_size = bank->size;
        unsigned long cur_size;
        unsigned long offset = 0;
        while (left_size > 0) {
            if (left_size < HIGN_MEMIO_BLOCK_SIZE)
                cur_size = left_size;
            else
                cur_size = HIGN_MEMIO_BLOCK_SIZE;
            left_size -= cur_size;
            WREG32_BASE(mmio_5_base, IATU_LWR_BASE_ADDR_OFF_OUTBOUND(j),   (bank->src_start + offset) & 0xffffffff); /* low base */
            WREG32_BASE(mmio_5_base, IATU_UPPER_BASE_ADDR_OFF_OUTBOUND(j), (bank->src_start + offset) >> 32);        /* upper base */
            WREG32_BASE(mmio_5_base, IATU_LIMIT_ADDR_OFF_OUTBOUND(j), 0xffffffff);  /* limit 4GB */
            WREG32_BASE(mmio_5_base, IATU_LWR_TARGET_BASE_ADDR_OFF_OUTBOUND(j),   (bank->dst_start + offset) & 0xffffffff); /* low target */
            WREG32_BASE(mmio_5_base, IATU_UPPER_TARGET_BASE_ADDR_OFF_OUTBOUND(j), (bank->dst_start + offset) >> 32);        /* upper target */
            WREG32_BASE(mmio_5_base, IATU_REGION_CTRL_2_OFF_OUTBOUND(j), 0x80000000);   /* enable the region */
            //printk("----IATU %d: src=%lx dst=%lx----\n", j, (bank->src_start + offset), (bank->dst_start + offset));
            offset += cur_size;
            j++;
            if (j >= 32) {
                printk(KERN_ERR "IATU region is out of range\n");
                g_high_memio_info.nr_banks = i + 1;
                bank->size -= left_size;
                return;
            }
        }
    }
}

int gp200_get_gpu_phys(u64 cpu_physical, u64 *gpu_physical)
{
    int i;

    if (cpu_physical < g_high_memio_info.banks[0].dst_start) {
        *gpu_physical = cpu_physical;
        return 0;
    }

    for (i=0; i<g_high_memio_info.nr_banks; i++) {
        if (cpu_physical >= g_high_memio_info.banks[i].dst_start
            && cpu_physical < g_high_memio_info.banks[i].dst_start + g_high_memio_info.banks[i].size) {
            *gpu_physical = cpu_physical - g_high_memio_info.banks[i].dst_start + g_high_memio_info.banks[i].src_start;
            //printk("----cpu %llx -> gpu %llx----\n", cpu_physical, *gpu_physical);
            return 0;
        }
    }
    *gpu_physical = cpu_physical;
    return -1;
}

int gp200_get_cpu_phys(u64 gpu_physical, u64 *cpu_physical)
{
    int i;

    if (gpu_physical < g_high_memio_info.banks[0].src_start) {
        *cpu_physical = gpu_physical;
        return 0;
    }

    for (i=0; i<g_high_memio_info.nr_banks; i++) {
        if (gpu_physical >= g_high_memio_info.banks[i].src_start
            && gpu_physical < g_high_memio_info.banks[i].src_start + g_high_memio_info.banks[i].size) {
            *cpu_physical = gpu_physical - g_high_memio_info.banks[i].src_start + g_high_memio_info.banks[i].dst_start;
            //printk("----gpu %llx -> cpu %llx----\n", gpu_physical, *cpu_physical);
            return 0;
        }
    }
    *cpu_physical = gpu_physical;
    return -1;
}

/* pre-initalize GP201 */
int gp200_pre_init(int card, struct pci_dev *pdev, unsigned long flags, unsigned long *pddr_size)
{
    resource_size_t     mmio_start; /* start of physical address for GP register memory. PCIE bar 2 */
    size_t              mmio_len;   /* memory size for GP register memory */
    void __iomem *      mmio_base;  /* mapped address for GP register memory */

    HEADER_ARG("card=%d pdev=%p flags=%lx", card, pdev, flags);

    mmio_start = pci_resource_start(pdev, 2) + MMIO_OFFSET;
    mmio_len = MMIO_SIZE;
    if (mmio_len > pci_resource_len(pdev, 2))
        mmio_len = pci_resource_len(pdev, 2);

    if (g_mmio_bases[card] == NULL) {
        /* ioremap GP private register memory */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
        mmio_base = ioremap(mmio_start, mmio_len);
#else
        mmio_base = ioremap_nocache(mmio_start, mmio_len);
#endif
        if (!mmio_base) {
            DRM_INFO("%s: Cannot map register area.\n", __FUNCTION__);
            FOOTER();
            return -ENOMEM;
        }
        TRACE("map 0x%x bytes @ 0x%08X to 0x%p(%lx)", (unsigned int)mmio_len, (unsigned int)mmio_start, mmio_base, (unsigned long)mmio_base);
        //printk(KERN_INFO "map 0x%x bytes @ 0x%08X to 0x%p", (unsigned int)mmio_len, (unsigned int)mmio_start, mmio_base);

        /* assign to global variable to support RREG8/16/32 and WREG8/16/32 */
        g_mmio_bases[card] = mmio_base;
    }
    else {
        mmio_base = g_mmio_bases[card];
    }
    if (0 == card)
        g_mmio_base = mmio_base;

    /* now RREG8/16/32 and WREG8/16/32 are available */

    switch (pdev->device) {
        case CHIP_GP201:
            gp201_pre_init(card, mmio_base, pddr_size, flags);
            break;
        case CHIP_GP202:
            gp202_pre_init(card, mmio_base, pddr_size, flags);
            break;
        default:
            ERR_PRINT("device id (0x%x) is not supported!\n", pdev->device);
            break;
    }
    if (flags & LJMICRO_IS_FPGA) {
        ddcEnable = 0;  /**< disable ddc for FPGA board */
        REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, 0);  /**< force VGA disabled for FPGA board */
    }

    /* get clock */
    ljm_pre_get_clock(mmio_base);

    ljm_gop_info_get(pdev);   //在ddr初始化之后，进行gop_info的读取操作

    gp200_high_memio_init(&g_high_memio_info);

    mmio_start = pci_resource_start(pdev, 5);
    mmio_len = pci_resource_len(pdev, 5);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    mmio_base = ioremap(mmio_start, mmio_len);
#else
    mmio_base = ioremap_nocache(mmio_start, mmio_len);
#endif
    if (!mmio_base) {
        DRM_INFO("%s: Cannot map bar5 register area.\n", __FUNCTION__);
        FOOTER();
        return -ENOMEM;
    }
    TRACE("map 0x%x bytes @ 0x%llx to 0x%px", (unsigned int)mmio_len, mmio_start, mmio_base);

    gp200_iatu_init(mmio_base);

    iounmap(mmio_base);

    FOOTER();
    return 0;
}

int gp200_suspend(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
//    int i;

    HEADER_ARG("drm=%p", ddev);

    /* backup register HDMI_DAC_DVO_OUT_MUX */
    if(ljm_dev->once_edp)
        ljm_dev->once_edp_vga_mux = REG_R32(ljm_dev->cur_card, HDMI_DAC_DVO_OUT_MUX);

    /* backup register MISC_INTR_ENPCIE and mask all pcie interrupts except EDMA */
    ljm_dev->misc_intr_maskpcie = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN);
    LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, ljm_dev->misc_intr_maskpcie & m_INTR_EN_EDMA);

    /* backup clock registers */
    ljm_pre_get_clock(g_mmio_bases[ljm_dev->cur_card]);

#if 0
    /* turn off the display ? */
    for (i=0; i<ljm_dev->num_encoders; i++) {
        encoder_dpms(ljm_dev, ljm_dev->encoders[i], DRM_MODE_DPMS_OFF);
    }
    for (i=0; i<ljm_dev->num_crtcs; i++) {
        struct ljm_crtc *ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[i]);
        ljm_crtc_mode_enable(ljm_dev, ljm_crtc, false);
    }
#endif

    FOOTER();
    return 0;
}

void gp200_fini(struct ljm_device *ljm_dev)
{
    int i;
    unsigned long irqflags;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    for (i=0; i<ljm_dev->num_crtcs; i++) {
        ljm_dev->irq.crtc_vblank_int[i] = false;
    }
    ljm_irq_vblank_set(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

#if USE_CLOCK_REGISTER
    ljm_clk_remove(ljm_dev);
#endif

    FOOTER();
}

int gp200_irq_vblank_set(struct ljm_device *ljm_dev)
{
    uint8_t data8;
    uint i;

    //HEADER_ARG("ljm_dev=%p", ljm_dev);

    if (!ljm_dev->irq.installed) {
        WARN(1, "Can't enable IRQ/MSI because no handler is installed\n");
        //FOOTER();
        return -EINVAL;
    }

    for (i = 0; i < ljm_dev->num_crtcs; i++) {
        switch (i) {
        case 0:
            data8 = CR_RD(ljm_dev->cur_card, CRTC_DISP_CTRL_ADR);
            if (ljm_dev->irq.crtc_vblank_int[i]){
                data8 |= 0x04;
            }
            else {
                data8 &= 0xfb;
            }
            CR_WR(ljm_dev->cur_card, CRTC_DISP_CTRL_ADR, data8);
            break;
        case 1:
        case 2:
        case 3:
            data8 = REG_R8(ljm_dev->cur_card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE*(i-1));
            if (ljm_dev->irq.crtc_vblank_int[i])
                data8 |= 0x20;
            else
                data8 &= 0xdf;
            REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE*(i-1), data8);
            break;
        default:
            break;
        }
    }

    //FOOTER();
    return 0;
}

/*****************************************************************************
*  函 数 名: gp200_irq_set
*  功能描述: 中断设置
*  输入参数: struct ljm_device *ljm_dev: ljm_gp设备指针
*  输出参数: 无
*  返 回 值: int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
int gp200_irq_set(struct ljm_device *ljm_dev)
{
    uint i;
    int r;
    struct ljm_port *port;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    /* set crtc vblank interrupt enable */
    r = gp200_irq_vblank_set(ljm_dev);
    if (r) {
        FOOTER_ARG("r=%d", r);
        return r;
    }

    /* set hpd interrupt enable */
    for (i=0; i<ljm_dev->num_ports; i++)
    {
        port = ljm_dev->ports[i];
        switch(port->type)
        {
            case LJM_PORT_VGA:
                ljm_dac_hpd_irq_enable(ljm_dev->cur_card, 0, port->hpd_en);
                break;
            case LJM_PORT_HDMI:
                ljm_hdmiv2_hpd_enable(&port->vdev.hdmi, port->hpd_en);
                break;
            case LJM_PORT_DVO:
                ljm_dvo_hpd_enable(ljm_dev->cur_card, port->id, port->hpd_en);
                break;
#if USE_DP_CONNECTOR
            case LJM_PORT_DP:
                ljm_dp_hpd_enable(&port->vdev.dp, port->hpd_en);
                /* one DP HPD should connect to a DVO HPD separately */
                ljm_dvo_hpd_enable(ljm_dev->cur_card, 5, port->hpd_en);
                ljm_dvo_hpd_enable(ljm_dev->cur_card, 6, port->hpd_en);
                break;
#endif
            default:
                ERR_PRINT("[%s]Unknown port type.\n", __FUNCTION__);
                break;
        }
    }

    FOOTER();
    return 0;
}

/* bandwidth error interrupt enable */
void gp200_bwe_enable(struct ljm_device *ljm_dev, bool enable)
{
    int i;
    u8 data;
    u16 data16;

    HEADER_ARG("ljm_dev=%lx enable=%d", (unsigned long)ljm_dev, enable);

    /* set crtc bandwidth error interrupt enable */
    for (i = 0; i < ljm_dev->num_crtcs; i++) {
        switch (i) {
        case 0:
            data = CR_RD(ljm_dev->cur_card, CRTC_DISP_CTRL_ADR);
            if (enable){
                data |= 0x20u;
            }
            else{
                data &= ~0x20u;
            }
            CR_WR(ljm_dev->cur_card, CRTC_DISP_CTRL_ADR, data);
            break;
        case 1:
        case 2:
        case 3:
            data = REG_R8(ljm_dev->cur_card, CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE*(i-1));
            if (enable)
                data |= 0x02u;
            else
                data &= ~0x02u;
            REG_W8(ljm_dev->cur_card, CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE*(i-1), data);
            break;
        default:
            break;
        }
    }

    /* set layer bandwidth error interrupt enable */
    for (i=0; i<4; i++) {
        data16 = REG_R16(ljm_dev->cur_card, LAYER1_CTRL_ADDR + LAYER_REG_SIZE * i);
        if (enable)
            data16 |= LAYER_CTRL_BWE_INT_EN_MASK;
        else
            data16 &= ~LAYER_CTRL_BWE_INT_EN_MASK;
        REG_W16(ljm_dev->cur_card, LAYER1_CTRL_ADDR + LAYER_REG_SIZE * i, data16);
    }

    FOOTER();
}

void gp200_irq_postinstall(struct ljm_device *ljm_dev)
{
    int i;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    /* enable/disable bandwidth enable */
    if (bweIntEnable)
        gp200_bwe_enable(ljm_dev, true);
    else
        gp200_bwe_enable(ljm_dev, false);

    /* enabe frame counter */
    CR_WR(ljm_dev->cur_card, CRTC_DISP_CTRL_ADR, CR_RD(ljm_dev->cur_card, CRTC_DISP_CTRL_ADR) | 0x80);
    for (i=1; i<ljm_dev->num_crtcs; i++)
        REG_W8(ljm_dev->cur_card,
            CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE * (i-1),
            REG_R8(ljm_dev->cur_card, CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE * (i-1)) | 0x08
            );

    /* enable VGA interrupt to pcie */
    DRM_DEBUG_KMS("enable VGA interrupt to pcie\n");
    LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN) | m_INTR_EN_VGA);

    /* enable external 0 interrupt */
    if (ljm_dev->dp_irq_is_extra)
    {
        DRM_DEBUG_KMS("enable external irq0 interrupt to pcie\n");
        LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN) | m_INTR_EN_EIRQ0);
    }

    FOOTER();
}

bool gp200_hpd_sense(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector);

/*****************************************************************************
*  函 数 名: gp200_hpd_init
*  功能描述: gp200热拔插初始化
*  输入参数: struct ljm_device *ljm_dev:  ljm_gp设备指针
*  输出参数: 无
*  返 回 值: void
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
void gp200_hpd_init(struct ljm_device *ljm_dev)
{
    unsigned long irqflags;
    int i;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

#if 0
    for (i=0; i<ljm_dev->num_connectors; i++) {
        if (i == 0) ljm_dev->irq.hpd[i] = false;   /* disable vga hpd interrupt */
        else ljm_dev->irq.hpd[i] = true;
    }
#endif

    for(i = 0; i < ljm_dev->num_ports; i++) {
        if (ddcEnable == 0)
            ljm_dev->ports[i]->hpd_en = false;
        else {
            ljm_dev->ports[i]->hpd_en = ljm_dev->ports[i]->info->hpd;

            /* get the initial status of hpd */
            if (ljm_dev->ports[i]->vdev_connector->hpd_type != LJM_HPD_TYPE_NONE) {
                ljm_dev->ports[i]->vdev_connector->hpd.plugged_state
                    = gp200_hpd_sense(ljm_dev, ljm_dev->ports[i]->vdev_connector) ? LJM_HPD_STATE_IN : LJM_HPD_STATE_OUT;
            }
        }
    }

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    gp200_irq_set(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    FOOTER();
}

void gp200_hpd_fini(struct ljm_device *ljm_dev)
{
    int i;
    unsigned long irqflags;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    for (i=0; i<ljm_dev->num_ports; i++)
        ljm_dev->ports[i]->hpd_en = false;

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    gp200_irq_set(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    FOOTER();
}

/* hpd for digital panel detect/disconnect */
/**
 * gp200_hpd_sense - hpd sense callback.
 *
 *
 * Checks if a digital monitor is connected.
 * Returns true if connected, false if not connected.
 */
bool gp200_hpd_sense(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector)
{
    bool connected = false;
    struct ljm_port *ljm_port = to_ljm_port(ljm_connector->encoder);

    //HEADER_ARG("connector%d(%d)=%lx", ljm_connector->base.index, ljm_connector->id, (unsigned long)ljm_connector);

    switch (ljm_connector->hpd_type) {
    case LJM_HPD_TYPE_DAC:
        //connected = ljm_dac_is_connected(ljm_dev->cur_card, ljm_connector->id);
        if(ljm_connector->ddc_type == LJM_DDC_TYPE_OCI2C && ljm_port->vdev.vga.i2c_dev)
            connected = ljm_i2c_oc_detect(ljm_port->vdev.vga.i2c_dev);
        else if (ljm_connector->ddc_type == LJM_DDC_TYPE_DWI2C && ljm_port->vdev.vga.i2c_dev)
            connected = ljm_i2c_dw_detect(ljm_port->vdev.vga.i2c_dev, DDC_EDID_ADDR);
        if(!connected)
            ljm_connector->edid_dirty = true;
        ljm_port->vdev.vga.hw_connected = connected; //保存vga硬件连接状态，用于gp200_encoder_dpms();
        // connected &&= !(CR_RD(ljm_dev->cur_card, CRTC_HPD_ST_ADR) & CRTC_HPD_ST_MASK); //当vga和dvo0同时接入时，告知应用层vga关闭，但硬件上不关闭encoder_dpms。
        break;
    case LJM_HPD_TYPE_HDMIV2:
        /* HDMI 0, 1, 2, 3 */
        connected = ljm_hdmiv2_hpd_sense(&ljm_port->vdev.hdmi);
        break;
    case LJM_HPD_TYPE_CRTC:
        if (ljm_connector->id == 8) {
            /* DVO 0 */
            if (CR_RD(ljm_dev->cur_card, CRTC_HPD_ST_ADR) & CRTC_HPD_ST_MASK) {
                connected = true;
            }
        } else {
            /* DVO 1, 2, 3 */
            if (REG_R8(ljm_dev->cur_card, CRTC1_HPD_ST_ADR + (ljm_connector->id - 9) * CRTCx_REG_SIZE) & CRTC_HPD_ST_MASK) {
                connected = true;
            }
        }
        break;
#if USE_DP_CONNECTOR
    case LJM_HPD_TYPE_DP:
    case LJM_HPD_TYPE_EIRQ:
        /* DP 0, 1, 2, 3 */
        connected = ljm_dp_hpd_sense(&ljm_port->vdev.dp);
        if (!connected)
        {
            ljm_connector->edid_dirty = true;
            if(ljm_dev->once_edp && ljm_port->vdev.dp.is_edp)
                ljm_port->base.once_edp_crtc = -1;
        }
        break;
#endif
    default:
        WARN_PRINT("connector%d: hpd type is invalid (%d)", ljm_connector->base.index, ljm_connector->hpd_type);
        break;
    }

    //FOOTER_ARG("connector%d (%d)  is %s",
    //    ljm_connector->base.index, ljm_connector->id, connected ? "connected" : "disconnected");
    return connected;
}

/**
 * gp200_hpd_set_polarity - hpd set polarity callback.
 *
 * @ljm_dev: radeon_device pointer
 * @hpd: hpd (hotplug detect) index
 *
 * Set the polarity of the hpd pin.
 */
void gp200_hpd_set_polarity(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector)
{
    struct ljm_port *ljm_port = to_ljm_port(ljm_connector->encoder);

    //HEADER_ARG("ljm_dev=%p ljm_connectoer%d=%p", ljm_dev, ljm_connector->id, ljm_connector);

    if (ljm_connector->hpd_type == LJM_HPD_TYPE_HDMIV2) {
        ljm_hdmiv2_hpd_set_polarity(&ljm_port->vdev.hdmi);
    }
    else if (ljm_connector->hpd_type == LJM_HPD_TYPE_DP) {
        //ljm_dp_hpd_set_polarity( ljm_connector->dp_dev);
    }
#if 0
    switch (ljm_connector->id) {
    case LJM_HPD_4:
    case LJM_HPD_5:
    case LJM_HPD_6:
    case LJM_HPD_7:
        ljm_hdmiv2_hpd_set_polarity(ljm_connector->hdmi_dev);
        break;
    case LJM_HPD_12:
    case LJM_HPD_13:
    case LJM_HPD_14:
    case LJM_HPD_15:
        /* DP connectors */
        break;
    default:
        /*
         * LJM_HPD_0, LJM_HPD_1, LJM_HPD_2, LJM_HPD_3 are for DAC, no need to set polarity
         * LJM_HPD_8, LJM_HPD_9, LJM_HPD_10, LJM_HPD_11 are for DVO, no need to set polarity
         */
        break;
    }
#endif

    //FOOTER();
}

int gp200_connector_detect(struct drm_connector *connector)
{
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);
    int r = connector_status_connected;

    if (ddcEnable == 0) {
        if (ljm_dev->flags & LJMICRO_IS_FPGA) {
            if (ljm_connector->id == 4 || ljm_connector->id == 8 ||
                    ljm_connector->id == 9 || ljm_connector->id == 12) {
                r = connector_status_connected;
                INFO_PRINT("connect%d id=%d force to be connected\n", connector->index, ljm_connector->id);
            }
            else
                r = connector_status_disconnected;
        }
        else {
#if USE_IN_SDT
            if (ljm_connector->id == 4) /* huym: only force hdmi 0 on */
#else
            if (ljm_connector->id >= 4 && ljm_connector->id < 4 + ljm_dev->num_crtcs)   /* force all hdmis on */
#endif
            {
                r = connector_status_connected;
                INFO_PRINT("connect%d id=%d force to be connected\n", connector->index, ljm_connector->id);
            }
            else
                r = connector_status_disconnected;
        }
    }
    else {
        if (gp200_hpd_sense(ljm_dev, ljm_connector))
            r = connector_status_connected;
        else
            r = connector_status_disconnected;
    }
    return r;
}
/*****************************************************************************
*  函 数 名: gp200_connector_get_edid
*  功能描述: gp200获取edid
*  输入参数: IN struct drm_connector *connector: 对应connector指针，依赖drm的数据类型
*            OUT int *size
*  输出参数: 无
*  返 回 值: struct edid *
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
struct edid *gp200_connector_get_edid(IN struct drm_connector *connector, OUT int *psize)
{
    struct edid *edid = NULL;
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    struct ljm_port *ljm_port = to_ljm_port(ljm_connector->encoder);
    struct hdmi_dev *hdmi;

    HEADER_ARG("connector%d=0x%lx %d", connector->index, (unsigned long)connector, connector->base.id);

    switch(ljm_port->type)
    {
        case LJM_PORT_HDMI:
            hdmi = &ljm_port->vdev.hdmi;
            edid = ljm_hdmiv2_get_edid(hdmi, psize);
            memset(&hdmi->edid, 0, sizeof(struct hdmi_edid));
            if (edid)
            {
                ljm_hdmiv2_parse_edid(hdmi, edid);
                /* hdmi scdc support */
                /*GPU只连接一个hdmi时，该hdmi拔出不会使dpms_state变化。根据该参数判断是否单口插拔。*/
                if(hdmi->need_scdc && hdmi->dpms_state == DRM_MODE_DPMS_ON)
                    ljm_hdmiv2_scrambling_set(hdmi);
            }
            break;
#if USE_DP_CONNECTOR
        case LJM_PORT_DP:
            edid = ljm_dp_get_edid(ljm_connector, psize);
            break;
#endif
        case LJM_PORT_VGA:
            if(ljm_connector->ddc_type == LJM_DDC_TYPE_OCI2C && ljm_port->vdev.vga.i2c_dev)
                edid = ljm_ddc_get_edid(ljm_port->vdev.vga.i2c_dev, ljm_i2c_oc_read_req, ljm_i2c_oc_raw_write_req, psize);
            else if (ljm_connector->ddc_type == LJM_DDC_TYPE_DWI2C && ljm_port->vdev.vga.i2c_dev)
                edid = ljm_ddc_get_edid(ljm_port->vdev.vga.i2c_dev, ljm_i2c_dw_read_req, ljm_i2c_dw_raw_write_req, psize);
            break;

        case LJM_PORT_DVO:
             if(ljm_connector->ddc_type == LJM_DDC_TYPE_OCI2C && ljm_port->vdev.dvo.i2c_dev)
                edid = ljm_ddc_get_edid(ljm_port->vdev.dvo.i2c_dev, ljm_i2c_oc_read_req, ljm_i2c_oc_raw_write_req, psize);
            else if (ljm_connector->ddc_type == LJM_DDC_TYPE_DWI2C && ljm_port->vdev.dvo.i2c_dev)
                edid = ljm_ddc_get_edid(ljm_port->vdev.dvo.i2c_dev, ljm_i2c_dw_read_req, ljm_i2c_dw_raw_write_req, psize);
            break;

        default:
            WARN_PRINT("Unknown port type!\n");
    }
    FOOTER();
    return edid;
}

/* resolution blacklist */
typedef struct _resolution_list {
    unsigned short w;   // width
    unsigned short h;   // hight
    unsigned int clock; // pixel clock in kHz. 0 represent any pixel clock
} resolution_list;

static const resolution_list ResolutionBlacklist[] ={
//    {2560, 1440, 0},  // adjust 2560x1440 to CVT timings @312.25MHz, so as to avoid display bandwith error
    {4096, 2160, 0}
};
static unsigned int ResolutionBlacklistCount = sizeof(ResolutionBlacklist) / sizeof(resolution_list);

static const resolution_list ResolutionWhitelist[] ={
    {3840, 2160, 262750},
};
static unsigned int ResolutionWhitelistCount = sizeof(ResolutionWhitelist) / sizeof(resolution_list);

#if 0
static bool _check_vfresh(int pclk, int vtotal, int htotal, int vfresh)
{
    int freq = vfresh * htotal * vtotal;

    if (freq > pclk * 1000) {
        if (freq - pclk * 1000 < (htotal * vtotal)) return true;    // < 1/vfresh hz
        else return false;
    }
    else if (freq < pclk * 1000) {
        if (pclk * 1000 - freq <= (htotal * vtotal)) return true;   // <= 1/vfresh hz
        else return false;
    }
    else return true;
}
#endif

int gp200_connector_mode_valid(struct drm_display_mode *mode, struct ljm_connector *ljm_connector)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ljm_connector->base.dev);
    struct ljm_port *ljm_port = to_ljm_port(ljm_connector->encoder);
    unsigned int i;

    //HEADER_ARG("mode=0x%p ljm_connector=0x%p", mode, ljm_connector);

    /* check black list */
    for (i=0; i < ResolutionBlacklistCount; i++) {
        if (mode->hdisplay == ResolutionBlacklist[i].w
            && mode->vdisplay == ResolutionBlacklist[i].h) {
            if (ResolutionBlacklist[i].clock == 0
                || mode->clock == ResolutionBlacklist[i].clock) {
                return MODE_PANEL;
            }
        }
    }

    if(ljm_dev->flags & LJMICRO_IS_FPGA){
        if (!ljm_vgav2_crtc_mode_valid(mode, true)) {
            return MODE_BAD;
        }
    }

    /* check white list */
    for (i=0; i < ResolutionWhitelistCount; i++) {
        if (mode->hdisplay == ResolutionWhitelist[i].w
            && mode->vdisplay == ResolutionWhitelist[i].h) {
            if (ResolutionWhitelist[i].clock == 0
                || mode->clock == ResolutionWhitelist[i].clock) {
                return MODE_OK;
            }
        }
    }

    //若分辨率超过1920x1080，则不支持60Hz以上的mode.
    if (drm_mode_vrefresh(mode) > 75)
        return MODE_PANEL;

    if(!ddcEnable)
        return MODE_OK;

    switch (ljm_port->type)
    {
        case LJM_PORT_VGA:
            if(!ljm_dac_mode_valid(mode))
                return MODE_BAD;
            break;
        case LJM_PORT_HDMI:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                if(!ljm_hdmiv2_mode_valid(&ljm_port->vdev.hdmi, mode))
                    return MODE_BAD;
            } else {
                if(!ljm_hdmiv2_mode_valid_ex(&ljm_port->vdev.hdmi, mode))
                    return MODE_BAD;
            }
            break;
        case LJM_PORT_DVO:
            if(mode->clock > 165000)
                return MODE_BAD;
            break;
#if USE_DP_CONNECTOR
        case LJM_PORT_DP:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                if(!ljm_dp_mode_valid(&ljm_port->vdev.dp, mode))
                    return MODE_BAD;
            } else {
                if(!ljm_dp_mode_valid_ex(&ljm_port->vdev.dp, mode))
                    return MODE_BAD;
            }
            break;
#endif
        default:
            ERR_PRINT("Unexpected port type =%d\n", ljm_port->type);
            return MODE_BAD;
    }

    return MODE_OK;
}

/*****************************************************************************
*  函 数 名: gp200_encoder_dpms
*  功能描述: 控制encoder dpms电源
*  输入参数: struct drm_encoder *encoder  : encoder指针，依赖drm
*            int mode: dpms状态
*  输出参数: 无
*  返 回 值: void
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*         2. 日    期: 2023年2月22日
*            作    者: 翟一夫
*            修改内容: vga与dvo0共用同一crtc。vga与dvo0同时插入时，
*               vga不应关闭,此修改是为了适配gp200_hpd_sense()中的改动。
*
*****************************************************************************/
void gp200_encoder_dpms(struct drm_encoder *encoder, int mode)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(encoder->dev);
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    struct ljm_port *ljm_port = to_ljm_port(ljm_encoder);

    HEADER_ARG("encoder%d=0x%lx %d mode=%d", encoder->index, (unsigned long)encoder, encoder->base.id, mode);

    if (ljm_dev->flags & LJMICRO_IS_FPGA) {
        FOOTER();
        return;
    }

    if (ljm_encoder->hard_type == LJM_ENCODER_DAC) {
#if USE_DAC_CONNECTOR
        if((ljm_port->vdev.vga.hw_connected) && ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201)) //仅gp201需要vga0与dvo0绑定
            ljm_dac_dpms(&ljm_port->vdev.vga, ljm_dev->cur_card, ljm_encoder->id, DRM_MODE_DPMS_ON);
        else
            ljm_dac_dpms(&ljm_port->vdev.vga, ljm_dev->cur_card, ljm_encoder->id, mode);
#endif
    } else if (ljm_encoder->hard_type == LJM_ENCODER_HDMIV2) {
#if USE_HDMI_CONNECTOR
        if (ddcEnable == 0 && mode != DRM_MODE_DPMS_ON) {
#if !USE_IN_SDT
            ljm_hdmiv2_dpms(&ljm_port->vdev.hdmi, mode);
#endif
        } else
            ljm_hdmiv2_dpms(&ljm_port->vdev.hdmi, mode);
#endif
    } else if (ljm_encoder->hard_type == LJM_ENCODER_DP) {
#if USE_DP_CONNECTOR
        ljm_dp_dpms(&ljm_port->vdev.dp, mode);
#endif
    } else if (ljm_encoder->hard_type == LJM_ENCODER_DVO) {
        ljm_dvo_dpms(ljm_dev->cur_card, ljm_port->info->sub_id, mode);
    }
    FOOTER();
}

void gp200_encoder_mux_set(struct ljm_device *ljm_dev, int encoder_id, int crtc_id)
{
    u32 reg_out_mux, reg_out_mux_old;
    unsigned int crtc_bits;

    reg_out_mux = REG_R32(ljm_dev->cur_card, HDMI_DAC_DVO_OUT_MUX);
    reg_out_mux_old = reg_out_mux;
    switch (encoder_id) {
    case 0: /* DAC0 */
        reg_out_mux &= 0xfffffffc;
        reg_out_mux |= crtc_id;
        break;
    case 1: /* DAC1 */
        reg_out_mux &= 0xfffffcff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 8;
        break;
    case 2: /* DAC2 */
        reg_out_mux &= 0xfffcffff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x2;
        else if (crtc_id == 2) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 16;
        break;
    case 3: /* DAC3 */
        reg_out_mux &= 0xfcffffff;
        if (crtc_id == 3) crtc_bits = 0x0;
        else crtc_bits = crtc_id + 1;
        reg_out_mux |= crtc_bits << 24;
        break;
    case 4: /* HDMI0 */
        reg_out_mux &= 0xfffffff3;
        reg_out_mux |= crtc_id << 2;
        break;
    case 5: /* HDMI1 */
        reg_out_mux &= 0xfffff3ff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 10;
        break;
    case 6: /* HDMI2 */
        reg_out_mux &= 0xfff3ffff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x2;
        else if (crtc_id == 2) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 18;
        break;
    case 7: /* HDMI3 */
        reg_out_mux &= 0xf3ffffff;
        if (crtc_id == 3) crtc_bits = 0x0;
        else crtc_bits = crtc_id + 1;
        reg_out_mux |= crtc_bits << 26;
        break;
    case 8: /* DVO0 */
        reg_out_mux &= 0xffffffcf;
        reg_out_mux |= crtc_id << 4;
        break;
    case 9: /* DVO1 */
        reg_out_mux &= 0xffffcfff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 12;
        break;
    case 10:    /* DVO2 */
        reg_out_mux &= 0xffcfffff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x2;
        else if (crtc_id == 2) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 20;
        break;
    case 11:    /* DVO3 */
        reg_out_mux &= 0xcfffffff;
        if (crtc_id == 3) crtc_bits = 0x0;
        else crtc_bits = crtc_id + 1;
        reg_out_mux |= crtc_bits << 28;
        break;
    case 12:    /* DP0 */
        reg_out_mux &= 0xffffff3f;
        reg_out_mux |= crtc_id << 6;
        break;
    case 13:    /* DP1 */
        reg_out_mux &= 0xffff3fff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 14;
        break;
    case 14:    /* DP2 */
        reg_out_mux &= 0xff3fffff;
        if (crtc_id == 0) crtc_bits = 0x1;
        else if (crtc_id == 1) crtc_bits = 0x2;
        else if (crtc_id == 2) crtc_bits = 0x0;
        else crtc_bits = crtc_id;
        reg_out_mux |= crtc_bits << 22;
        break;
    case 15:    /* DP3 */
        reg_out_mux &= 0x3fffffff;
        if (crtc_id == 3) crtc_bits = 0x0;
        else crtc_bits = crtc_id + 1;
        reg_out_mux |= crtc_bits << 30;
        break;
    default:
        break;
    }
#if USE_IN_SDT
    if (ddcEnable == 0)
        reg_out_mux = 0x00ffffff;   /* crtc3 mux to all connectors */
#endif

    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201
        && (encoder_id == 0 || encoder_id == 8)) {  /* DAC0, DVO0, share hsync and vsync, must use same crtc for GP201 */
        reg_out_mux &= 0xffffffcc;
        reg_out_mux |= crtc_id;
        reg_out_mux |= crtc_id << 4;
    }

    REG_W32(ljm_dev->cur_card, HDMI_DAC_DVO_OUT_MUX, reg_out_mux);

    if (reg_out_mux != reg_out_mux_old)
        DEBUG_PRINT("OUT_MUX is changed to 0x%x from 0x%x for encoder id(%d) and crtc id(%d)\n",
            reg_out_mux, reg_out_mux_old, encoder_id, crtc_id);
}

void gp200_encoder_mode_set(struct drm_encoder *encoder, struct drm_display_mode *mode, struct drm_display_mode *adjustedMode)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(encoder->dev);
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(encoder->crtc);
    struct ljm_port *ljm_port = to_ljm_port(ljm_encoder);

    HEADER_ARG("encoder%d=0x%lx %d mode=%lx adjustedMode=%lx",
        encoder->index, (unsigned long)encoder, ljm_encoder->id,
        (unsigned long)mode, (unsigned long)adjustedMode);

    if(ljm_dev->once_edp)
    {
        if (ljm_encoder->once_edp_crtc >= 0) {
            if (!ljm_encoder->once_edp_mode) { /* 保留EDP屏的显示模式，以便S3恢复时进行设置 */
                ljm_encoder->once_edp_mode = kmalloc(sizeof(struct drm_display_mode), GFP_KERNEL);
                memcpy(ljm_encoder->once_edp_mode, adjustedMode, sizeof(struct drm_display_mode));
            }
            if(ljm_encoder->once_edp_modeset) {
                FOOTER();
                return; /* no change from VBIOS */
            }
        }
    }

    gp200_encoder_mux_set(ljm_dev, ljm_encoder->id, ljm_crtc->id);

    if (ljm_encoder->hard_type == LJM_ENCODER_HDMIV2) {
        struct hdmi_dev *hdmi_dev;

#if USE_IN_SDT
        if (ddcEnable == 0) {
            int i;
            for (i=0; i<ljm_dev->num_ports; i++) {
                if (ljm_dev->ports[i]->type != LJM_PORT_HDMI)
                    continue;

                hdmi_dev = &ljm_dev->ports[i]->vdev.hdmi;

                hdmi_displaymode_to_videomode(adjustedMode, &hdmi_dev->video.mode);
                hdmi_dev->video.vic = hdmi_videomode_to_vic(&hdmi_dev->video.mode);
                if (hdmi_dev->video.vic == 0) {
                    /* look for a CEA-861 mode */
                    hdmi_dev->video.vic = drm_match_cea_mode(adjustedMode);
                }
                hdmi_dev->video.color_input = HDMI_COLOR_RGB_0_255;
                hdmi_dev->video.color_output = HDMI_COLOR_RGB_0_255;
                hdmi_dev->video.color_output_depth = 8;
                hdmi_dev->video.colorimetry = 0;
                hdmi_dev->video.sink_hdmi = hdmi_dev->edid.sink_hdmi;
                hdmi_dev->video.format_3d = -1;
                hdmi_dev->video.eotf = 0;

                if (ljm_hdmiv2_mode_set(hdmi_dev, &hdmi_dev->video)) {
                    DRM_ERROR("HDMI %d mode set failed\n", hdmi_dev->index);
                }

                gp200_encoder_dpms(&ljm_dev->ports[i]->base.base, DRM_MODE_DPMS_ON);
            }
            FOOTER();
            return;
        }
#endif

        hdmi_dev = &ljm_port->vdev.hdmi;

        hdmi_displaymode_to_videomode(adjustedMode, &hdmi_dev->video.mode);
        hdmi_dev->video.vic = hdmi_videomode_to_vic(&hdmi_dev->video.mode);
        if (hdmi_dev->video.vic == 0) {
            /* look for a CEA-861 mode */
            hdmi_dev->video.vic = drm_match_cea_mode(adjustedMode);
        }
        hdmi_dev->video.color_input = HDMI_COLOR_RGB_0_255;
        hdmi_dev->video.color_output = HDMI_COLOR_RGB_0_255;
        hdmi_dev->video.color_output_depth = 8;
        hdmi_dev->video.colorimetry = 0;
        hdmi_dev->video.sink_hdmi = hdmi_dev->edid.sink_hdmi;
        hdmi_dev->video.format_3d = -1;
        hdmi_dev->video.eotf = 0;

        if (ljm_hdmiv2_mode_set(hdmi_dev, &hdmi_dev->video)) {
            DRM_ERROR("HDMI %d mode set failed\n", hdmi_dev->index);
        }
    }
#if USE_DP_CONNECTOR
    else if (ljm_encoder->hard_type == LJM_ENCODER_DP) {
        ljm_dp_mode_set(&ljm_port->vdev.dp, adjustedMode);
        if(ljm_dev->once_edp && ljm_encoder->once_edp_crtc >= 0)
            ljm_encoder->once_edp_modeset = true;
    }
#endif
    FOOTER();
}

void gp200_crtc_cursor_hide(struct ljm_device *ljm_dev, struct drm_crtc *crtc);
void gp200_crtc_cursor_show(struct ljm_device *ljm_dev, struct drm_crtc *crtc);

/*****************************************************************************
 *  函 数 名: gp200_crtc_cursor_move
 *  功能描述: GP200系列硬光标移动
 *  输入参数: struct drm_crtc *crtc
 *            int x
 *            int y
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年12月6日
 *            作    者: 胡艳明
 *            修改内容: 优化光标移动逻辑
 *         2. 日    期: 2023年1月4日
 *            作    者: 胡艳明
 *            修改内容: 解决跨屏光标显示问题
 *
 *****************************************************************************/
void gp200_crtc_cursor_move(struct drm_crtc *crtc, int x, int y)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);

#if 0   /* updated and removed by huym on 20221206 */
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct ljm_crtc *ljm_crtc_tmp;
    u16 cursor_sel;

    cursor_sel = REG_R16(ljm_dev->cur_card, CURSOR_CTRL) & 0xffe7;
    cursor_sel |= ljm_crtc->id << 3;

    switch (crtc->index) {
    case 0:
        break;
    case 1:
        if(ljm_dev->curor_en[0]) {
            ljm_crtc_tmp = to_ljm_crtc(ljm_dev->crtcs[0]);
            cursor_sel |= 1 << (7 + ljm_crtc_tmp->id);
        }
        break;
    case 2:
        if(ljm_dev->curor_en[0]) {
            ljm_crtc_tmp = to_ljm_crtc(ljm_dev->crtcs[0]);
            cursor_sel |= 1 << (7 + ljm_crtc_tmp->id);
        }
        if(ljm_dev->curor_en[1]) {
            ljm_crtc_tmp = to_ljm_crtc(ljm_dev->crtcs[1]);
            cursor_sel |= 1 << (7 + ljm_crtc_tmp->id);
        }
        break;
    case 3:
        if(ljm_dev->curor_en[0]) {
            ljm_crtc_tmp = to_ljm_crtc(ljm_dev->crtcs[0]);
            cursor_sel |= 1 << (7 + ljm_crtc_tmp->id);
        }
        if(ljm_dev->curor_en[1]) {
            ljm_crtc_tmp = to_ljm_crtc(ljm_dev->crtcs[1]);
            cursor_sel |= 1 << (7 + ljm_crtc_tmp->id);
        }
        if(ljm_dev->curor_en[2]) {
            ljm_crtc_tmp = to_ljm_crtc(ljm_dev->crtcs[2]);
            cursor_sel |= 1 << (7 + ljm_crtc_tmp->id);
        }
        break;
    default:
        break;
    }
    REG_W16(ljm_dev->cur_card, CURSOR_CTRL, cursor_sel);
#endif
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

    if (ljm_crtc->cursor_hw_show) {
        u16 cursor_reg = REG_R16(ljm_dev->cur_card, CURSOR_CTRL);

        if (cursor_reg & 0x700) {   /* in cursor clone mode */
            if (x < 0 || y < 0) {   /* 跨屏光标不显示 */
                gp200_crtc_cursor_hide(ljm_dev, crtc);
                return;
            }
        } else {    /* not in cursor clone mode */
            /* 边缘越界不隐藏光标 */
            if (x < 0) x = 0;
            if (y < 0) y = 0;
        }
    } else {
        if (x > 0 && y > 0) {   /* 光标离开跨屏模式 */
            gp200_crtc_cursor_show(ljm_dev, crtc);
        } else {
            return;
        }
    }

    REG_W32(ljm_dev->cur_card, CURSOR_H, (y<<16) | x);
}

/*****************************************************************************
 *  函 数 名: gp200_crtc_cursor_hide
 *  功能描述: GP200系列硬光标隐藏
 *  输入参数: struct ljm_device *ljm_dev
 *            struct drm_crtc *crtc
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年12月6日
 *            作    者: 胡艳明
 *            修改内容: 优化光标隐藏逻辑，适应CRTC顺序的调整
 *         2. 日    期: 2023年1月4日
 *            作    者: 胡艳明
 *            修改内容: 解决跨屏光标显示问题
 *
 *****************************************************************************/
void gp200_crtc_cursor_hide(struct ljm_device *ljm_dev, struct drm_crtc *crtc)
{
    u16 cursor_reg;
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    //printk("crtc %d: id=%d hide\n", crtc->index, ljm_crtc->id);

    ljm_crtc->cursor_hw_show = false;

    cursor_reg = REG_R16(ljm_dev->cur_card, CURSOR_CTRL);

    /* hardware mouse is disabled */
    if ((cursor_reg & 0x1) == 0) {
        FOOTER();
        return;
    }

    if(ljm_crtc->id == 0) {
        if ((cursor_reg & 0x18) == 0x00) {  /* mouse in CRTC 0 */
            if (cursor_reg & 0x100) {       /* CRTC 1 is cloned */
                cursor_reg = (cursor_reg & 0xfee7) | 0x08;
            }
            else if (cursor_reg & 0x200) {  /* CRTC 2 is cloned */
                cursor_reg = (cursor_reg & 0xfce7) | 0x10;
            }
            else if (cursor_reg & 0x400) {  /* CRTC 3 is cloned */
                cursor_reg = (cursor_reg & 0xf8e7) | 0x18;
            }
            else {
                cursor_reg &= 0xfffe;       /* no mouse */
            }
        }
        /* else do nothing */
    }
    else if(ljm_crtc->id == 1) {
        if ((cursor_reg & 0x18) == 0x08) {  /* mouse in CRTC 1 */
            if (cursor_reg & 0x200) {       /* CRTC 2 is cloned */
                cursor_reg = (cursor_reg & 0xfce7) | 0x10;
            }
            else if (cursor_reg & 0x400) {  /* CRTC 3 is cloned */
                cursor_reg = (cursor_reg & 0xf8e7) | 0x18;
            }
            else {
                cursor_reg &= 0xfffe;       /* no mouse */
            }
        }
        else {
            cursor_reg &= 0xfeff;   /* CRTC 1: disable cloning */
        }
    }
    else if(ljm_crtc->id == 2) {
        if ((cursor_reg & 0x18) == 0x10) {  /* mouse in CRTC 2 */
            if (cursor_reg & 0x400) {       /* CRTC 3 is cloned */
                cursor_reg = (cursor_reg & 0xf8e7) | 0x18;
            }
            else {
                cursor_reg &= 0xfffe;       /* no mouse */
            }
        }
        else {
            cursor_reg &= 0xfdff;   /* CRTC 2: disable cloning */
        }
    }
    else if(ljm_crtc->id == 3) {
        if ((cursor_reg & 0x18) == 0x18) {  /* mouse in CRTC 3 */
            cursor_reg &= 0xfffe;           /* no mouse */
        }
        else {
            cursor_reg &= 0xfbff;   /* CRTC 3: disable cloning */
        }
    }

    REG_W16(ljm_dev->cur_card, CURSOR_CTRL, cursor_reg);

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: gp200_crtc_cursor_show
 *  功能描述: GP200系列硬光标显示
 *  输入参数: struct ljm_device *ljm_dev
 *            struct drm_crtc *crtc
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年12月6日
 *            作    者: 胡艳明
 *            修改内容: 优化光标显示逻辑，适应CRTC顺序的调整
 *         2. 日    期: 2023年1月4日
 *            作    者: 胡艳明
 *            修改内容: 解决跨屏光标显示问题
 *
 *****************************************************************************/
void gp200_crtc_cursor_show(struct ljm_device *ljm_dev, struct drm_crtc *crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    u16 cursor_reg = 0;

    HEADER_ARG("ljm_dev=%p, crtc=%p", ljm_dev, crtc);

    //printk("crtc %d: id=%d show\n", crtc->index, ljm_crtc->id);

    ljm_crtc->cursor_hw_show = true;

    REG_W32(ljm_dev->cur_card, CURSOR_STARTADDR, (ljm_crtc->cursor_addr / 4));

    cursor_reg = REG_R16(ljm_dev->cur_card, CURSOR_CTRL);

    /* cursor size */
    if (ljm_crtc->cursor_size == 64)
        cursor_reg |= 1u<<2;
    else
        cursor_reg &= ~(1u<<2);

    /* 32-bit color */
    cursor_reg |= 1u<<1;

    if ((cursor_reg & 0x1) == 0) {   /* mouse is not enable */
        cursor_reg = (cursor_reg & 0xf8e7) | (ljm_crtc->id << 3);   /* just set crtc id */
    }
    else {  /* mouse has been enabled */
        switch (ljm_crtc->id) {
        case 0:
            if ((cursor_reg & 0x18) == 0x00) {  /* mouse in CRTC 0 */
                /* do nothing */
            }
            else {
                cursor_reg |= 1 << (7 + ((cursor_reg & 0x18)>>3));  /* clone mouse */
                cursor_reg &= 0xffe7;                               /* set mouse in crtc 0 */
            }
            break;
        case 1:
            if ((cursor_reg & 0x18) == 0x00) {  /* mouse in CRTC 0 */
                cursor_reg |= 0x100;            /* clone mouse in the current crtc */
            }
            else if ((cursor_reg & 0x18) == 0x08) { /* mouse in CRTC 1 */
                /* do nothing */
            }
            else {
                cursor_reg |= 1 << (7 + ((cursor_reg & 0x18)>>3));  /* clone mouse in the current crtc */
                cursor_reg = (cursor_reg & 0xffe7) | 0x08;          /* set mouse in crtc 1 */
            }
            break;
        case 2:
            if ((cursor_reg & 0x18) == 0x00         /* mouse in CRTC 0 */
                || (cursor_reg & 0x18) == 0x08) {   /* mouse in CRTC 1 */
                cursor_reg |= 0x200;                /* clone mouse in the current crtc */
            }
            else if ((cursor_reg & 0x18) == 0x10) { /* mouse in CRTC 2 */
                /* do nothing */
            }
            else {
                cursor_reg |= 0x400;                        /* clone mouse in crtc 3 */
                cursor_reg = (cursor_reg & 0xffe7) | 0x10;  /* set mouse in crtc 2 */
            }
            break;
        case 3:
            if ((cursor_reg & 0x18) == 0x00         /* mouse in CRTC 0 */
                || (cursor_reg & 0x18) == 0x08      /* mouse in CRTC 1 */
                || (cursor_reg & 0x18) == 0x10) {   /* mouse in CRTC 2 */
                cursor_reg |= 0x400;                /* clone mouse in crtc 3 */
            }
            break;
        default:
            break;
        }
    }

    cursor_reg |= 1<<0;

    REG_W16(ljm_dev->cur_card, CURSOR_CTRL, cursor_reg);

    FOOTER();
}

/*mode needn't to fixup in GP201*/
bool gp200_crtc_mode_fixup(struct ljm_device *ljm_dev, const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
    HEADER_ARG("ljm_dev=0x%p mode=0x%p adjustedmode=0x%p", ljm_dev, mode, adjusted_mode);

    FOOTER();
    return true;
}

int gp200_crtc_mode_set(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, struct drm_display_mode *adjusted_mode)
{
    int i;
    if(ljm_dev->once_edp)
    {
        for (i=0; i<ljm_dev->num_ports; i++) {
            if (ljm_crtc->id == ljm_dev->ports[i]->base.once_edp_crtc) {
                if (!ljm_crtc->once_edp_mode) {    /* 保留EDP屏的显示模式，以便S3恢复时进行设置 */
                    ljm_crtc->once_edp_mode = kmalloc(sizeof(struct drm_display_mode), GFP_KERNEL);
                    memcpy(ljm_crtc->once_edp_mode, adjusted_mode, sizeof(struct drm_display_mode));
                }
                if(ljm_dev->ports[i]->base.once_edp_modeset)
                    return 0;   /* no change from VBIOS */
            }
        }
    }
    return ljm_vgav2_crtc_mode_set(ljm_dev->cur_card, ljm_crtc->id, adjusted_mode, ljm_dev->flags & LJMICRO_IS_FPGA);
}

void gp200_crtc_mode_enable(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, bool enable)
{
    int i;
    HEADER_ARG("ljm_crtc=%p enable=%d", ljm_crtc, (int)enable);

    if (enable) {
        ljm_vgav2_crtc_enable(ljm_dev->cur_card, ljm_crtc->id);
    }
    else {
        if(ljm_dev->once_edp)
        {
            for (i=0; i<ljm_dev->num_ports; i++) {
                if (ljm_crtc->id == ljm_dev->ports[i]->base.once_edp_crtc &&
                    ljm_dev->ports[i]->base.once_edp_modeset) {
                    FOOTER();
                    return; /* no change from VBIOS */
                }
            }
        }
        ljm_vgav2_crtc_disable(ljm_dev->cur_card, ljm_crtc->id);
    }

    FOOTER();
}

/**
 * gp200_crtc_set_base - set crtc base only.
 *
 * @crtc_id: crtc to cleanup pageflip on
 * @crtc_base: new address of the crtc (GPU VRAM address)
 *
 * During vblank we take the crtc lock and wait for the update_pending
 * bit to go high, when it does, we release the lock, and allow the
 * double buffered update to take place.
 */
int gp200_crtc_set_base(struct ljm_device *ljm_dev, int crtc_id, u64 crtc_base)
{
    return ljm_vgav2_crtc_set_base(ljm_dev->cur_card, crtc_id, crtc_base);
}

int gp200_mode_set_base(struct drm_crtc *crtc, int bpp, int pitches, u64 base)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);

    return ljm_vgav2_crtc_mode_set_base(ljm_dev->cur_card, ljm_crtc->id, bpp, pitches, base);
}

/*****************************************************************************
 *  函 数 名：gp200_crtc_prog_reset
 *  功能描述：设置CRTC延时计数
 *            4路CRTC显示，当4路同时读取数据，对总线的瞬时带宽需求是最高的，设置延时计数以均匀错开读取数据时间点
 *  输入参数：struct ljm_device *ljm_dev
 *            int crtc_used
 *            long min_hscan_time
 *  输出参数：无
 *  返 回 值：int
 *
 *  修改历史：
 *         1. 日    期：2022年8月11日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
int gp200_crtc_prog_reset(struct ljm_device *ljm_dev,  int crtc_used, long min_hscan_time)
{
    int reg_val;
    HEADER_ARG("ljm_dev=%p crtc_used=%d min_hscan_time=%ld ", ljm_dev, crtc_used, min_hscan_time);

    /* AHB clock of GP200 is 400MHz */
    reg_val = (int)((min_hscan_time / crtc_used) * 400000000 / 1000000000);
    ljm_vgav2_crtc_prog_reset(ljm_dev->cur_card, reg_val);

    FOOTER();
    return 0;
}

u32 gp200_get_vblank_counter(struct ljm_device *ljm_dev, int crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[crtc]);

    return ljm_vgav2_get_vblank_counter(ljm_dev->cur_card, ljm_crtc->id);
}

/**
 * gp200_page_flip - pageflip callback.
 *
 * @ljm_dev: ljm_device pointer
 * @crtc_id: crtc to cleanup pageflip on
 * @crtc_base: new address of the crtc (GPU MC address)
 *
 * Update double buffer address to registers.
 * It will be taken active after negedge of vsync.
 */
void gp200_page_flip(struct ljm_device *ljm_dev, int crtc_id, u64 crtc_base, bool async)
{
    //HEADER_ARG("ljm_dev=%p crtc_id=%d crtc_base=%llx async=%d", ljm_dev, crtc_id, crtc_base, async);

    gp200_crtc_set_base(ljm_dev, crtc_id, crtc_base);

    //FOOTER();
}

/**
 * gp200_page_flip_pending - check if page flip is still pending
 *
 * @ljm_dev: ljm_device pointer
 * @crtc_id: crtc to check
 *
 * Check if the last pagefilp is still pending.
 * Returns the current update pending status.
 */
bool gp200_page_flip_pending(struct ljm_device *ljm_dev, int crtc_id)
{
    return ljm_vgav2_page_flip_pending(ljm_dev->cur_card, crtc_id);
}

/*
 * Retrieve current video scanout position of crtc on a given gpu.
 *
 * \param ljm_dev ljm_device to query.
 * \param pipe Crtc index to query.
 * \param *vpos Location where vertical scanout position should be stored.
 * \param *hpos Location where horizontal scanout position should go.
 *
 * Returns vpos as a positive number from vgate start until end of vblank.
 * Returns hpos as a positive number from hgate start until end of hblank.
 *
 * \return Flags, or'ed together as follows:
 *
 * DRM_SCANOUTPOS_VALID = Query successful.
 * DRM_SCANOUTPOS_INVBL = Inside vblank.
 * DRM_SCANOUTPOS_ACCURATE = Returned position is accurate.
 */
int gp200_get_crtc_scanout_pos(struct ljm_device *ljm_dev, unsigned int pipe, int *vpos, int *hpos)
{
    return ljm_vgav2_get_crtc_scanout_pos(ljm_dev->cur_card, pipe, vpos, hpos);
}


#define CRTC_CODE0_TO_INDEX(CODE) return (CODE)

#define CRTC_CODE1_TO_INDEX(CODE) \
    if ((CODE) == 0) \
        return 1; \
    else if ((CODE) == 1) \
        return 0; \
    else \
        return (CODE)

#define CRTC_CODE2_TO_INDEX(CODE) \
    if ((CODE) == 0) \
        return 2; \
    else if ((CODE) == 1) \
        return 0; \
    else if ((CODE) == 2) \
        return 1; \
    else \
        return 3;

#define CRTC_CODE3_TO_INDEX(CODE) \
    if ((CODE) == 0) \
        return 3; \
    else if ((CODE) == 1) \
        return 0; \
    else if ((CODE) == 2) \
        return 1; \
    else \
        return 2;

int gp200_port_get_crtc_id(int card, uint32_t port_id)
{
    int      code;
    uint32_t vga_mux = REG_R32(card, HDMI_DAC_DVO_OUT_MUX);

    switch (port_id)
    {
        case 0: /* DAC0 */
            code = (vga_mux >> 0) & 0x3;
            CRTC_CODE0_TO_INDEX(code);
        case 1: /* DAC1 */
            code = (vga_mux >> 8) & 0x3;
            CRTC_CODE1_TO_INDEX(code);
        case 2: /* DAC2 */
            code = (vga_mux >> 16) & 0x3;
            CRTC_CODE2_TO_INDEX(code);
        case 3: /* DAC3 */
            code = (vga_mux >> 24) & 0x3;
            CRTC_CODE3_TO_INDEX(code);
        case 4: /* HDMI0 */
            code = (vga_mux >> 2) & 0x3;
            CRTC_CODE0_TO_INDEX(code);
        case 5: /* HDMI1 */
            code = (vga_mux >> 10) & 0x3;
            CRTC_CODE1_TO_INDEX(code);
        case 6: /* HDMI2 */
            code = (vga_mux >> 18) & 0x3;
            CRTC_CODE2_TO_INDEX(code);
        case 7: /* HDMI3 */
            code = (vga_mux >> 26) & 0x3;
            CRTC_CODE3_TO_INDEX(code);
        case 8: /* DVO0 */
            code = (vga_mux >> 4) & 0x3;
            CRTC_CODE0_TO_INDEX(code);
        case 9: /* DVO1 */
            code = (vga_mux >> 12) & 0x3;
            CRTC_CODE1_TO_INDEX(code);
        case 10: /* DVO2 */
            code = (vga_mux >> 20) & 0x3;
            CRTC_CODE2_TO_INDEX(code);
        case 11: /* DVO3 */
            code = (vga_mux >> 28) & 0x3;
            CRTC_CODE3_TO_INDEX(code);
        case 12: /* DP0 */
            code = (vga_mux >> 6) & 0x3;
            CRTC_CODE0_TO_INDEX(code);
        case 13: /* DP1 */
            code = (vga_mux >> 14) & 0x3;
            CRTC_CODE1_TO_INDEX(code);
        case 14: /* DP2 */
            code = (vga_mux >> 22) & 0x3;
            CRTC_CODE2_TO_INDEX(code);
        case 15: /* DP3 */
            code = (vga_mux >> 30) & 0x3;
            CRTC_CODE3_TO_INDEX(code);
        default:
            ERR_PRINT("Invalid port id for output mux\n");
            return 0;
    }
}

#if 0
#ifdef LJMICRO_DRV_ONLY
int gpu_config_power_management(int core, bool enable)
{
    return 0;
}
#else
int gpu_config_power_management(int core, bool enable);
#endif

int gp200_power_set_state(struct ljm_device *ljm_dev, int power_state)
{
    int r = 0;
    uint32_t val;

    //printk("++%s(%d): ljm_dev=%px power_state=%d\n", __FUNCTION__, __LINE__, ljm_dev, power_state);

    switch (power_state)
    {
        case LJM_POWER_STATE_PERFORMANCE:
            //printk("power performance...\n");
            val = REG_R32(ljm_dev->cur_card, 0x308000); //TODO: 0x308000偏移与GP202区分。
            if ((val & 0x3000)
                || (REG_R32(ljm_dev->cur_card, 0x0030800c) != g_gpupll_val)
                || (REG_R32(ljm_dev->cur_card, 0x00308010) != g_gpu2pll_val)) {
                /* bypass 3D, 2D clock */
                val |= 0x3000;
                REG_W32(ljm_dev->cur_card, 0x308000, val);

                /* restore 3D, 2D clock */
                REG_W32(ljm_dev->cur_card, 0x0030800c, g_gpupll_val );
                REG_W32(ljm_dev->cur_card, 0x00308010, g_gpu2pll_val);

                val &= 0xffffcfff;
                REG_W32(ljm_dev->cur_card, 0x308000, val);
            }

            /* disable power management for cores */
            gpu_config_power_management(-1, false);
            break;
        case LJM_POWER_STATE_SAVE:
            //printk("power save...\n");
            val = REG_R32(ljm_dev->cur_card, 0x308000);
            if ((val & 0x3000)
                || (REG_R32(ljm_dev->cur_card, 0x0030800c) != g_gpupll_val)
                || (REG_R32(ljm_dev->cur_card, 0x00308010) != g_gpu2pll_val)) {
                /* bypass 3D, 2D clock */
                val |= 0x3000;
                REG_W32(ljm_dev->cur_card, 0x308000, val);

                /* restore 3D, 2D clock */
                REG_W32(ljm_dev->cur_card, 0x0030800c, g_gpupll_val );
                REG_W32(ljm_dev->cur_card, 0x00308010, g_gpu2pll_val);

                val &= 0xffffcfff;
                REG_W32(ljm_dev->cur_card, 0x308000, val);
            }

            /* enable power management for cores */
            gpu_config_power_management(-1, true);
            break;
        case LJM_POWER_STATE_STANDBY:
            //printk("power standby...\n");
            val = REG_R32(ljm_dev->cur_card, 0x308000);
            if ((val & 0x3000)
                || (REG_R32(ljm_dev->cur_card, 0x0030800c) != 0xb10231)
                || (REG_R32(ljm_dev->cur_card, 0x00308010) != 0xb10241)) {
                /* bypass 3D, 2D clock */
                val |= 0x3000;
                REG_W32(ljm_dev->cur_card, 0x308000, val);

                /* restore 3D, 2D clock */
                REG_W32(ljm_dev->cur_card, 0x0030800c, 0xb10231 );  /* gpupll (2D): 400M */
                REG_W32(ljm_dev->cur_card, 0x00308010, 0xb10241);   /* gpu2pll (3D): 200M */

                val &= 0xffffcfff;
                REG_W32(ljm_dev->cur_card, 0x308000, val);
            }

            /* enable power management for cores */
            gpu_config_power_management(-1, true);
            break;
        default:
            r = -EINVAL;
    }

    //printk("--%s(%d): r=%d\n", __FUNCTION__, __LINE__, r);

    return r;
}
#endif
inline u32 ljm_get_device_id(int card)
{
    return LJM_MISC_R32(card, MISC_CHIP_R01);
}

/**
 * 根据PLL寄存器值计算PLL的输出频率，输出频率单位为KHz
 * parameter clk_in_khz 输入时钟, 单位为KHz, 0代表默认输入时钟27000KHz
 * return 时钟频率, 单位为KHz
 * by huym on 20240203
 */
u32 gp200_pll_code_to_freq(u32 clk_in_khz, u32 code)
{
    u64 PLLOUT;
    u32 DIVQ;
    int      i, x;

    if (clk_in_khz == 0)
        clk_in_khz = GP200_CLOCK_IN_KHZ;

    x     = 1;
    DIVQ  = (code & 0x70) >> 4;
    for (i = 0; i < DIVQ; i++)
    {
        x *= 2;
    }
    PLLOUT = (clk_in_khz * 2 * (((code & 0x1ff0000) >> 16) + 1) + (((code & 0x1f00) >> 8) + 1) * x / 2)
             / ((((code & 0x1f00) >> 8) + 1) * x);
    return (u32)PLLOUT;
}


/**
 * gp200_crtc_commit_for_dp - 在crtc_commit后，对dp/edp复位以提高dp/edp稳定性。
 * @ljm_dev: GPU设备指针，该函数内主要依赖此变量进行ljm_port遍历。
 *
 * 返回值：无
 *
 * 实际调试中发现，edp屏在多接口情况下，存在概率点不亮的问题。在ljm_crtc_commit()函数末尾复位
 * crtc和dptx可以解决该问题，但具体原因尚不明确。
 * 该函数原设计：判断此drm_crtc是否用于edp屏，若是，则进行crtc和dptx复位。但在测试过程中，接
 * 入vga后，vga对另一crtc进行配置，并未对edp映射的crtc进行操作，结果同样导致edp无法点亮，原因
 * 不明。
 * 由于上述原因，更改此函数设计：在任一crtc配置的ljm_crtc_commit()函数末尾，遍历所有ljm_port，
 * 所有被用于edp的crtc都会被进行复位操作。
*/
#if USE_DP_CONNECTOR
void gp200_crtc_commit_for_dp(struct ljm_device *ljm_dev)
{
    int i, crtc_id;
    struct ljm_port *port, **p;
    struct dp_dev *dp;

    for(i=0, p=&ljm_dev->ports[0]; i<ljm_dev->num_ports; i++, p++)
    {
        port = *p;

        /*只对edp接口生效*/
        if(port->type == LJM_PORT_DP)
        {
            dp = &port->vdev.dp;
            crtc_id = gp200_port_get_crtc_id(ljm_dev->cur_card, port->id);
            if(dp->is_edp)
            {
                if(crtc_id == 0)
                {
                    CR_WR(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL, (CR_RD(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL) & 0x7f));
                    msleep(5);
                    ljm_dptx_soft_reset(dp);
                    msleep(5);
                    CR_WR(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL, (CR_RD(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL) | BIT(7)));
                }
                else
                {
                    REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1), (RREG8(CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1)) & 0xfe));
                    msleep(5);
                    ljm_dptx_soft_reset(dp);
                    msleep(5);
                    REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1), (RREG8(CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1)) | 0x1));
                }
            }
        }
    }

    return;
}
#else
void gp200_crtc_commit_for_dp(struct ljm_device *ljm_dev)
{
    return;
}
#endif
