/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __GP200_H__
#define __GP200_H__

#include "ljmicro_family.h"

/**
 * DDR
 */
#define REG_DDR_CTRL_SIZE 0x8000
#define REG_DDR_PHY_SIZE  0x40000

#define LJM_NOC_R32(CARD, ADDR)            readl(g_mmio_bases[CARD] + ljm_gpu_hardware.reg_noc_offset + (ADDR))
#define LJM_NOC_W32(CARD, ADDR, VAL)       writel((u32)(VAL), g_mmio_bases[CARD] + ljm_gpu_hardware.reg_noc_offset + (ADDR))
#define LJM_NOC_R32_BASE(BASE, ADDR)       readl((void __iomem *)(BASE) + ljm_gpu_hardware.reg_noc_offset + (ADDR))
#define LJM_NOC_W32_BASE(BASE, ADDR, VAL)  writel((u32)(VAL), (void __iomem *)(BASE) + ljm_gpu_hardware.reg_noc_offset + (ADDR))

/**
 * CRU registers
 */
#define CRU_PLLEN            0x0
#define m_CFG_DEVPLL_BYPASS      (1U << 15)
#define m_CFG_DEV2PLL_BYPASS     (1U << 14)
#define m_CFG_GPUPLL_BYPASS      (1U << 13)
#define m_CFG_GPU2PLL_BYPASS     (1U << 12)
#define m_CFG_DDRPLL0_BYPASS     (1U << 11)
#define m_CFG_DDRPLL1_BYPASS     (1U << 10)
#define m_CFG_DDRPLL2_BYPASS     (1U << 9)
#define m_CFG_DDRPLL3_BYPASS     (1U << 8)
#define m_CFG_PXLPLL0_BYPASS     (1U << 7)
#define m_CFG_PXLPLL1_BYPASS     (1U << 6)
#define m_CFG_PXLPLL2_BYPASS     (1U << 5)
#define m_CFG_PXLPLL3_BYPASS     (1U << 4)

#define CRU_DEVPLL           0x4
#define CRU_DEV2PLL          0x8
#define CRU_GPUPLL           0xc
#define CRU_GPU2PLL          0x10
#define CRU_DDRPLL           0x14
#define m_CFG_PLL_BP         (1U << 31) /* GP202: PLL bypass or DDRPLL0 bypass */
#define m_CFG_DDRPLL_BP1     (1U << 30) /* GP202: DDRPLL1 bypass */
#define CRU_PXLPLL0          0x18
#define CRU_PXLPLL1          0x1c
#define CRU_PXLPLL2          0x20
#define CRU_PXLPLL3          0x24
#define v_DEVPLL_666M        0x01270231
#define v_DEVPLL_952M        0x008c0122
#define v_DEVPLL_923M        0x00cc0221
#define PLL_FREQ_800         0xb10221
#define PLL_FREQ_999         0xdd0221
#define PLL_FREQ_500         0x490033
#define PLL_FREQ_600         0x580033
#define PLL_FREQ_200         0xb10241
#define PLL_FREQ_400         0xb10231

#define CRU_CKMUX            0x30
#define m_CFG_GPU2D_CORE_CKSEL  (1U << 10) /**< GP201 0: DEVPLL/OUT3, 1: GPUPLL/OUT0; GP202 0: DEV2PLL/OUT3; 1: GPU2PLL/OUT0 */
#define m_CFG_GPU3D_SH_CKSEL    (1U << 9)  /**< 0: GPUPLL/OUT0; 1: GPU2PLL/OUT0 */
#define m_CFG_GPU3D_CORE_CKSEL  (1U << 8)  /**< 0: GPUPLL/OUT0; 1: GPU2PLL/OUT0 */
#define m_CFG_BUS_CKSEL         (1U << 0)  /**< GP201 0: DEVPLL or DEV2PLL, 1: HOSC; GP202 0: HOSC, 1: PLL */

#define CRU_CKDIV            0x34
#define CRU_CKEN0            0x38
#define m_CFG_VGA_PLX3_CKEN  (1U << 31)
#define m_CFG_VGA_PLX2_CKEN  (1U << 30)
#define m_CFG_VGA_PLX1_CKEN  (1U << 29)
#define m_CFG_VGA_PLX0_CKEN  (1U << 28)
#define m_CFG_HDMI3_BUS_CKEN (1U << 27)
#define m_CFG_HDMI2_BUS_CKEN (1U << 26)
#define m_CFG_HDMI1_BUS_CKEN (1U << 25)
#define m_CFG_HDMI0_BUS_CKEN (1U << 24)
#define m_CFG_DP3_BUS_CKEN   (1U << 23)
#define m_CFG_DP2_BUS_CKEN   (1U << 22)
#define m_CFG_DP1_BUS_CKEN   (1U << 21)
#define m_CFG_DP0_BUS_CKEN   (1U << 20)
#define m_CFG_VGA_BUS_CKEN   (1U << 3)
#define m_CFG_VPU_BUS_CKEN   (1U << 2)
#define m_CFG_GPU2D_BUS_CKEN (1U << 1)
#define m_CFG_GPU3D_BUS_CKEN (1U << 0)

#define CRU_CKEN1            0x3c
#define m_CFG_QSPI_CKEN       (1U << 31)
#define m_CFG_DMA_CKEN        (1U << 29)
#define m_CFG_MCU_CKEN        (1U << 28)
#define m_CFG_I2S0_CKEN       (1U << 27)
#define m_CFG_I2S1_CKEN       (1U << 26)

#define CRU_CNTCLK           0x40
#define CRU_STATUS           0x50
#define m_STA_DEVPLL_LOCK    (0x1 << 31)
#define m_STA_DEV2PLL_LOCK   (0x1 << 30)
#define m_STA_GPUPLL_LOCK    (0x1 << 29)
#define m_STA_GPU2PLL_LOCK   (0x1 << 28)
#define m_STA_DDR0_LOCK      (0x1 << 27)
#define m_STA_DDR1_LOCK      (0x1 << 26)
#define m_STA_DDR2_LOCK      (0x1 << 25)
#define m_STA_DDR3_LOCK      (0x1 << 24)
#define m_STA_PXLPLL0_LOCK   (0x1 << 23)
#define m_STA_PXLPLL1_LOCK   (0x1 << 22)
#define m_STA_PXLPLL2_LOCK   (0x1 << 21)
#define m_STA_PXLPLL3_LOCK   (0x1 << 20)

#define CRU_SOFTRST0         0x60
#define m_CFG_RESET_PXL3     (1U << 31)
#define m_CFG_RESET_PXL2     (1U << 30)
#define m_CFG_RESET_PXL1     (1U << 29)
#define m_CFG_RESET_PXL0     (1U << 28)
#define m_CFG_RESET_HDMI3    (1U << 27)
#define m_CFG_RESET_HDMI2    (1U << 26)
#define m_CFG_RESET_HDMI1    (1U << 25)
#define m_CFG_RESET_HDMI0    (1U << 24)
#define m_CFG_RESET_DDRC     (0xf << 16)
#define v_CFG_RESET_DDRC(n)  ((n) << 16)
#define m_CFG_RESET_DDRC_BUS (0xf << 8)
#define m_CFG_RESET_DP3      (1U << 7)
#define m_CFG_RESET_DP2      (1U << 6)
#define m_CFG_RESET_DP1      (1U << 5)
#define m_CFG_RESET_DP0      (1U << 4)
#define m_CFG_RESET_VGA      (1U << 3)
#define m_CFG_RESET_VPU      (1U << 2)
#define m_CFG_RESET_GPU2D    (1U << 1)
#define m_CFG_RESET_GPU3D    (1U << 0)

#define CRU_SOFTRST1         0x64
#define m_CFG_RESET_QSPI     (1U << 31)
#define m_CFG_RESET_MISC     (1U << 30)
#define m_CFG_RESET_I2C      (1U << 29)
#define m_CFG_RESET_DMAC     (1U << 28)
#define m_CFG_RESET_IOMUX    (1U << 27)
#define m_CFG_RESET_I2S0     (1U << 26)
#define m_CFG_RESET_I2S1     (1U << 25)

#define CRU_HWRST            0x68
#define CRU_DEBUG            0x6c

#define LJM_PLL_CODE_400M   0x00b10231
#define LJM_PLL_CODE_450M   0x00c70231
#define LJM_PLL_CODE_500M   0x00dd0231
#define LJM_PLL_CODE_550M   0x00a20132
#define LJM_PLL_CODE_600M   0x00580033
#define LJM_PLL_CODE_650M   0x01200231
#define LJM_PLL_CODE_700M   0x01360231
#define LJM_PLL_CODE_750M   0x014c0231
#define LJM_PLL_CODE_800M   0x00b10221
#define LJM_PLL_CODE_850M   0x003e0023
#define LJM_PLL_CODE_900M   0x00c70221
#define LJM_PLL_CODE_910M   0x00860122
#define LJM_PLL_CODE_920M   0x00cc0221
#define LJM_PLL_CODE_930M   0x00440023
#define LJM_PLL_CODE_940M   0x00d00221
#define LJM_PLL_CODE_950M   0x008c0122
#define LJM_PLL_CODE_1000M  0x00490023
#define LJM_PLL_CODE_1050M  0x00e80221
#define LJM_PLL_CODE_1100M  0x00a20122
#define LJM_PLL_CODE_1150M  0x00ff0221
#define LJM_PLL_CODE_1200M  0x01090221
#define LJM_PLL_CODE_1250M  0x01150221

/**
 * PVT
 */
#define PVT_SAMP_CTL    0x170
#define PVT_DATA        0x174
#define m_PVT_DATA_MASK 0x3ff
#define PVT_TEMP_HIGH   0x178

#define LJM_PVT_R32(CARD, ADDR)           readl(g_mmio_bases[CARD] + ljm_gpu_hardware.reg_misc_offset + (ADDR))
#define LJM_PVT_W32(CARD, ADDR, VAL)      writel((u32)(VAL), g_mmio_bases[CARD] + ljm_gpu_hardware.reg_misc_offset + (ADDR))
#define LJM_PVT_R32_BASE(BASE, ADDR)      readl((void __iomem *)(BASE) + ljm_gpu_hardware.reg_misc_offset + (ADDR))
#define LJM_PVT_W32_BASE(BASE, ADDR, VAL) writel((u32)(VAL), (void __iomem *)(BASE) + ljm_gpu_hardware.reg_misc_offset + (ADDR))

#define LJM_PVT_TEMP_MINUS_62 0x0
#define LJM_PVT_TEMP_MINUS_60 0x8
#define LJM_PVT_TEMP_MINUS_55 0x17
#define LJM_PVT_TEMP_MINUS_50 0x27
#define LJM_PVT_TEMP_MINUS_45 0x36
#define LJM_PVT_TEMP_MINUS_40 0x46
#define LJM_PVT_TEMP_MINUS_35 0x56
#define LJM_PVT_TEMP_MINUS_30 0x67
#define LJM_PVT_TEMP_MINUS_25 0x77
#define LJM_PVT_TEMP_MINUS_20 0x88
#define LJM_PVT_TEMP_MINUS_15 0x99
#define LJM_PVT_TEMP_MINUS_10 0xab
#define LJM_PVT_TEMP_MINUS_05 0xbd
#define LJM_PVT_TEMP_0        0xcf
#define LJM_PVT_TEMP_5        0xe2
#define LJM_PVT_TEMP_10       0xf5
#define LJM_PVT_TEMP_15       0x108
#define LJM_PVT_TEMP_20       0x11b
#define LJM_PVT_TEMP_25       0x12f
#define LJM_PVT_TEMP_30       0x143
#define LJM_PVT_TEMP_35       0x158
#define LJM_PVT_TEMP_40       0x16d
#define LJM_PVT_TEMP_45       0x183
#define LJM_PVT_TEMP_50       0x199
#define LJM_PVT_TEMP_55       0x1b0
#define LJM_PVT_TEMP_60       0x1c7
#define LJM_PVT_TEMP_65       0x1df
#define LJM_PVT_TEMP_70       0x1f7
#define LJM_PVT_TEMP_75       0x210
#define LJM_PVT_TEMP_80       0x22a
#define LJM_PVT_TEMP_85       0x244
#define LJM_PVT_TEMP_90       0x25f
#define LJM_PVT_TEMP_95       0x27b
#define LJM_PVT_TEMP_100      0x298
#define LJM_PVT_TEMP_105      0x2b5
#define LJM_PVT_TEMP_110      0x2d3
#define LJM_PVT_TEMP_115      0x2f2
#define LJM_PVT_TEMP_120      0x311
#define LJM_PVT_TEMP_125      0x331
#define LJM_PVT_TEMP_130      0x352
#define LJM_PVT_TEMP_135      0x374
#define LJM_PVT_TEMP_140      0x396
#define LJM_PVT_TEMP_145      0x3b8
#define LJM_PVT_TEMP_150      0x3db
#define LJM_PVT_TEMP_155      0x3fe


struct ljm_temp_to_freq {
    uint32_t temp_code;     /* PVT传感器温度码值档位, 不同的温度码值档位对应不同的频率.
                             * 在温度上升过程中, 当PVT传感器温度码值 >= temp_code时, freq_code将被设置;
                             * 在温度下降过程中, 当PVT传感器温度码值 <= temp_code - delt时, freq_code将被设置.
                             */
    int delt;               /* temp_code迟滞量, 避免在温度码值档位附近小幅摆动时出现频率频繁设置情况 */
    uint32_t freq_code;     /* freqency code for AXI/2D/3D */
};

/* OUTBOUND ATU */
#define IATU_LWR_BASE_ADDR_OFF_OUTBOUND(i)          (0x4008 + 0x200*(i))
#define IATU_UPPER_BASE_ADDR_OFF_OUTBOUND(i)        (0x400c + 0x200*(i))
#define IATU_LIMIT_ADDR_OFF_OUTBOUND(i)             (0x4010 + 0x200*(i))
#define IATU_LWR_TARGET_BASE_ADDR_OFF_OUTBOUND(i)   (0x4014 + 0x200*(i))
#define IATU_UPPER_TARGET_BASE_ADDR_OFF_OUTBOUND(i) (0x4018 + 0x200*(i))
#define IATU_REGION_CTRL_2_OFF_OUTBOUND(i)          (0x4004 + 0x200*(i))

/**
 * global variables
 */
extern uint ddcEnable;
extern int bweIntEnable;
extern ulong bweIntCount;

/**
 * chip pre-initialize
 */
int gp201_pre_init(int card, void *mmio_base, unsigned long *pddr_size, unsigned long flags);
int gp201_ddrv2_noc_init(void *mmio_base, uint8_t channel_mask);
unsigned char gp201_ddrv2_get_channel_mask(int card);
int gp202_pre_init(int card, void *mmio_base, unsigned long *pddr_size, unsigned long flags);
int gp202_ddrv2_noc_init(void *mmio_base, uint8_t channel_mask);
unsigned char gp202_ddrv2_get_channel_mask(int card);
void ljm_pre_get_clock(void *mmio_base);

#if 0
void ljm_ddrv2_set_urgent_qos(void *mmio_base, int channel_mask);
#endif

/** DDR初始化 */
void ljm_ddr_init(int card, void *mmio_base, unsigned long *pddr_size, unsigned long flags);

/** PLL code to frequency */
#define GP200_CLOCK_IN_KHZ 27000   /* 27000 KHz */
u32 gp200_pll_code_to_freq(u32 clk_in_khz, u32 code);

/** get PVT temperature */
uint32_t gp201_pvt_get_temp(int card, int index);
uint32_t gp202_pvt_get_temp(int card, int index);
/** get PVT voltage */
uint32_t gp201_pvt_get_volt(int card, int index);
uint32_t gp202_pvt_get_volt(int card, int index);
/** get PVT process **/
uint32_t gp201_pvt_get_svt_process(int card, int index);
uint32_t gp201_pvt_get_lvt_process(int card, int index);
uint32_t gp201_pvt_get_ulvt_process(int card, int index);
uint32_t gp202_pvt_get_svt_process(int card, int index);
uint32_t gp202_pvt_get_lvt_process(int card, int index);
uint32_t gp202_pvt_get_ulvt_process(int card, int index);

/**
 * global variable
 */
extern u32 g_devpll_val;
extern u32 g_dev2pll_val;
extern u32 g_gpupll_val;
extern u32 g_gpu2pll_val;
extern u32 g_ckmux_val;

#ifdef LJMICRO_DRV_ONLY
static inline int gpu_config_power_management(int core, bool enable)
{
    return 0;
}
#else
int gpu_config_power_management(int core, bool enable);
#endif

#endif
