/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
*
*  版权所有(c)，2020-2099，中船第709研究所
*
*******************************************************************************
*
*  文件名称: gp200.c
*  版 本 号: V1.0
*  作    者: 胡艳明
*  生成日期: 2022年4月21日
*
*  功能描述: gp200
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: huym
*            修改内容: 创建文件
*
******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/
#include "ljmicro_device.h"
#include "ljmicro_mcu.h"
#include "ljmicro_info.h"
#include "ljmicro_vgav2.h"
#include "ljmicro_mode.h"
#include "ljmicro_gpio.h"
#include <linux/kthread.h>
#include "gp201.h"
#include "ljmicro_pm.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

#define LJM_MONITOR_R32_BASE(BASE, ADDR)      readl((void __iomem *)(BASE) + REG_MONITOR_OFFSET + (ADDR))
#define LJM_MONITOR_W32_BASE(BASE, ADDR, VAL) writel((u32)(VAL), (void __iomem *)(BASE) + REG_MONITOR_OFFSET + (ADDR))

extern int ddrChannelMask;

/*****************************************************************************
 *  函 数 名: gp201_pre_set_clock
 *  功能描述: 预设置时钟频率
 *  输入参数: void *mmio_base
 *  输出参数: 无
 *  返 回 值: static void
 *
 *  修改历史:
 *         1. 日    期: 2023年2月23日
 *            作    者: huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static void gp201_pre_set_clock(void *mmio_base)
{
    LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                m_CFG_GPUPLL_BYPASS | m_CFG_GPU2PLL_BYPASS
                | m_CFG_DDRPLL0_BYPASS | m_CFG_DDRPLL1_BYPASS | m_CFG_DDRPLL2_BYPASS | m_CFG_DDRPLL3_BYPASS
                | m_CFG_GPUPLL_CKEN1 | m_CFG_GPUPLL_CKEN0 | m_CFG_DEV2PLL_CKEN1 | m_CFG_DEVPLL_CKEN3);

    if (g_devpll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_DEVPLL, g_devpll_val);   /* restore DEVPLL for AXI bus */
    else
        LJM_CRU_W32_BASE(mmio_base, CRU_DEVPLL, v_DEVPLL_952M);   /* set DEVPLL to 950M for AXI bus */ /* 950MHz解决2560x1440分辨率闪首行白线问题 */

    if (g_gpupll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_GPUPLL, g_gpupll_val);   /* restore GPUPLL */
    else
        LJM_CRU_W32_BASE(mmio_base, CRU_GPUPLL, v_DEVPLL_923M);   /* set GPUPLL to 920M */

    if (g_gpu2pll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_GPU2PLL, g_gpu2pll_val);   /* restore GPU2PLL */
    else
        LJM_CRU_W32_BASE(mmio_base, CRU_GPU2PLL, v_DEVPLL_923M);   /* set GPU2PLL to 920M */

    LJM_CRU_W32_BASE(mmio_base, CRU_DDRPLL, v_DEVPLL_666M);      /* set DDRPLL default to 666M */

    if (g_ckmux_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_CKMUX, g_ckmux_val);    /* restore CKMUX */
    else
        LJM_CRU_W32_BASE(mmio_base, CRU_CKMUX,
                    m_CFG_GPU2D_CORE_CKSEL | m_CFG_GPU3D_SH_CKSEL | m_CFG_GPU3D_CORE_CKSEL);   /*2D to GPUPLL, 3D to GPU2PLL */

    LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN, m_CFG_GPUPLL_CKEN1 | m_CFG_GPUPLL_CKEN0 | m_CFG_DEV2PLL_CKEN1 | m_CFG_DEVPLL_CKEN3);
}

/*****************************************************************************
 *  函 数 名: gp201_ddrv2_noc_init
 *  功能描述: initialize ddrv2 controller and phy
 *  输入参数: void *mmio_base   MMIO基地址
 *            uint8_t channel_mask  DDR通道mask位
 *            int ddr4_freq     DDR4频率枚举
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年3月18日
 *            作    者: huym
 *            修改内容: 新生成函数
 *         2. 日    期: 2024年6月22日
 *            作    者: zhaiyf
 *            修改内容: 函数更名，与gp202做区分
 *****************************************************************************/
int gp201_ddrv2_noc_init(void *mmio_base, uint8_t channel_mask)
{
    u32 val;
    int i;

    HEADER_ARG("mmio_base=%p channel_mask=0x%x", mmio_base, channel_mask);

    switch (channel_mask) {
        case 0x1:
            val = v_REMAP_DDR0;
            break;
        case 0x2:
            val = v_REMAP_DDR1;
            break;
        case 0x3:
            val = v_REMAP_DDR01;
            break;
        case 0xc:
            val = v_REMAP_DDR23;
            break;
        case 0xf:
            val = v_REMAP_DDR0123;
            break;
        default:
            val = v_REMAP_DDR0;
            break;
    }

    for (i = 0; i < NOC_REMAP_BLK_CNT; i++) {
        LJM_NOC_W32_BASE(mmio_base, NOC_REMAP_BLK0 + i * NOC_REMAP_BLK_SIZE, val);
    }

    FOOTER();
    return 0;
}

unsigned char gp201_ddrv2_get_channel_mask(int card)
{
    uint32_t tmp;
    tmp = LJM_NOC_R32(card, NOC_REMAP_BLK0);

    switch (tmp) {
    case v_REMAP_DDR0:
        return (unsigned char)0x1;
    case v_REMAP_DDR1:
        return (unsigned char)0x2;
    case v_REMAP_DDR01:
        return (unsigned char)0x3;
    case v_REMAP_DDR23:
        return (unsigned char)0xc;
    case v_REMAP_DDR0123:
        return (unsigned char)0xf;
    default:
        return (unsigned char)0x1;
    }
}

/*****************************************************************************
 *  函 数 名: gp201_ddr_set_qos
 *  功能描述: 设置DDR通道的优先级
 *  输入参数: void *mmio_base
 *  输出参数: 无
 *  返 回 值: static void
 *
 *  修改历史:
 *         1. 日    期: 2024年6月18日
 *            作    者: huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static void gp201_ddr_set_qos(void *mmio_base)
{
    LJM_MONITOR_W32_BASE(mmio_base, 0x38, 0x0fffffff);
    LJM_MONITOR_W32_BASE(mmio_base, 0x20, 0x55555555);
    LJM_MONITOR_W32_BASE(mmio_base, 0x24, 0x55555555);
}

void gp201_bypass_unused_pll(void *mmio_base)
{
    uint32_t tmp;

    /** bypass unused DDR PLL */
    tmp = LJM_NOC_R32_BASE(mmio_base, NOC_REMAP_BLK0);
    switch (tmp) {
    case v_REMAP_DDR0:
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         (LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & 0xfffff0ff) | 0x700); /* bypass DDRPLL 1, 2, 3 */
        break;
    case v_REMAP_DDR1:
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         (LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & 0xfffff0ff) | 0xb00); /* bypass DDRPLL 0, 2, 3 */
        break;
    case v_REMAP_DDR01:
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         (LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & 0xfffff0ff) | 0x300); /* bypass DDRPLL 2, 3 */
        break;
    case v_REMAP_DDR23:
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         (LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & 0xfffff0ff) | 0xc00); /* bypass DDRPLL 0, 1 */
        break;
    case v_REMAP_DDR0123:
        break;
    default:
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         (LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & 0xfffff0ff) | 0x300); /* bypass DDRPLL 2, 3 */
        break;
    }

    /** bypass unused GPUPLL */
    tmp = LJM_CRU_R32_BASE(mmio_base, CRU_CKMUX);  /* CRU_CKMUX */
    if ((tmp & m_CFG_GPU2D_CORE_CKSEL)             /* check GPU2D selecting GPUPLL */
        || ((tmp & m_CFG_GPU3D_SH_CKSEL) == 0)    /* check GPU3D SH selecting GPUPLL */
        || ((tmp & m_CFG_GPU3D_CORE_CKSEL) == 0)) /* check GPU3D CORE selecting GPUPLL */
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & ~m_CFG_GPUPLL_BYPASS);    /* un-bypass GPUPLL */
    else
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) | m_CFG_GPUPLL_BYPASS);     /* bypass GPUPLL */

    /** bypass unused GPU2PLL */
    if (tmp & (m_CFG_GPU3D_SH_CKSEL | m_CFG_GPU3D_CORE_CKSEL)) /* check GPU3D selecting GPU2PLL */
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) & ~m_CFG_GPU2PLL_BYPASS);   /* un-bypass GPU2PLL */
    else
        LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN,
                         LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN) | m_CFG_GPU2PLL_BYPASS);    /* bypass GPU2PLL */
}

/**
 * @brief GP201 special pre-initialization
 * @param  card card index
 * @param  mmio_base kernel virtual base address for regsters
 * @param  pddr_size    pointer to return DDR size
 * @param  flags    flags of PCI device
 * @return int  if succeed return 0, otherwise return error code
 *
 * history:
 *  1. date: 2023.10.31
 *     author: huym
 *     description: create
 *  2. date: 2024.5.20
 *     author: huym
 *     description: add flags parameter and check FPGA board here
 */
int gp201_pre_init(int card, void *mmio_base, unsigned long *pddr_size, unsigned long flags)
{
    int i;

    /** init gpu hardware */
    ljm_gpu_hardware.max_ddr_channels = 4;
    for (i = 0; i < ljm_gpu_hardware.max_ddr_channels; i++) {
        ljm_gpu_hardware.reg_ddr_ctr_offset[i] = REG_DDR_CTRL_OFFSET + REG_DDR_CTRL_SIZE * i;
        ljm_gpu_hardware.reg_ddr_phy_offset[i] = REG_DDR_PHY_OFFSET + REG_DDR_PHY_SIZE * i;
    }
    ljm_gpu_hardware.reg_misc_offset = REG_MISC_OFFSET;
    ljm_gpu_hardware.reg_spi_offset = REG_SPI_OFFSET;
    ljm_gpu_hardware.reg_cru_offset = REG_CRU_OFFSET;
    ljm_gpu_hardware.reg_noc_offset = REG_NOC_OFFSET;
    ljm_gpu_hardware.reg_iomux_offset = REG_IOMUX_OFFSET;

    if (flags & LJMICRO_IS_FPGA) {  /* FPGA board */
        if (pddr_size)
            *pddr_size = (flags & LJMICRO_VRAM_2G) ? 0x80000000UL : 0x40000000UL;
    }
    else {  /* GP201 chip */
#if USE_DP_CONNECTOR
        ljm_mcu_pre_init(card, MCU_FLASH_OFFSET, flags);
#endif
        ljm_fan_pre_init(card, 25);/* set fan speed: 25% ,called here only once, then kernel adjust speed */
        if (RREG32_BASE(mmio_base, ljm_gpu_hardware.reg_cru_offset + CRU_SOFTRST0) & m_CFG_RESET_DDRC) {
            gp201_pre_set_clock(mmio_base);   /* set default clock */

            gp201_ddr_set_qos(mmio_base);
            ljm_ddr_init(card, mmio_base, pddr_size, flags);
            gp201_ddrv2_noc_init(mmio_base, ddrChannelMask);

            /* sleep more than 120 millioseconds to ensure ddr port work */
            /* FIXME: wait until ddr read/write check OK, should be put in noc */
            msleep(150);
        }
        else {
            INFO_PRINT("ddr initialized in vbios\n");
            ddrChannelMask = gp201_ddrv2_get_channel_mask(0);
            if (pddr_size)
                *pddr_size = ljm_ddrv2_get_ddr_size(card, mmio_base);
        }

        gp201_bypass_unused_pll(mmio_base);
        //ljm_ddrv2_set_urgent_qos(mmio_base, 0x3);
    }

    return 0;
}

static struct ljm_temp_to_freq ljm_3d_temp_high_table[] = {
    {LJM_PVT_TEMP_140, 5, LJM_PLL_CODE_400M}, /** >= 140 degree Celsius: frequency 400MHz */
    {LJM_PVT_TEMP_125, 5, LJM_PLL_CODE_600M}, /** >= 125 degree Celsius: frequency 600MHz */
    {LJM_PVT_TEMP_110, 5, LJM_PLL_CODE_800M}, /** >= 110 degree Celsius: frequency 800MHz */
    {LJM_PVT_TEMP_MINUS_62, 0, LJM_PLL_CODE_920M}, /** >= -62 degree Celsius: frequency 920MHz */
};

static const int ljm_3d_temp_high_table_count = sizeof(ljm_3d_temp_high_table)/sizeof(ljm_3d_temp_high_table[0]);

bool gp201_adjust_3d_freq(struct ljm_device *ljm_dev, u32 temp_code)
{
    u32 freq_code;
    u32 tmp;
    int i;
    u32 timeout = 1000;
    unsigned long flags = 0;

    tmp = LJM_CRU_R32(ljm_dev->cur_card, CRU_CKMUX);
    if (!(tmp & 0x100))
        freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL);
    else
        freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPU2PLL);

    /** high temperature to reduce frequency */
    for (i=0; i < ljm_3d_temp_high_table_count; i++) {
        /* skip low temperature items */
        if (temp_code < ljm_3d_temp_high_table[i].temp_code)
            continue;

        /* current frequency fit the temperature, just return */
        if (freq_code == ljm_3d_temp_high_table[i].freq_code)
            return false;

        /* 温度进入低温一档，但是未超过delt值，频率保持不变 */
        if (i > 0 &&
            (temp_code > (ljm_3d_temp_high_table[i-1].temp_code - ljm_3d_temp_high_table[i-1].delt) &&
            freq_code == ljm_3d_temp_high_table[i-1].freq_code))
            return false;

        /* adjust 3d frequency */
        DEBUG_PRINT("3D: %dKhz -> %dKhz (temp_code=0x%x)\n",
            gp200_pll_code_to_freq(0, freq_code),
            gp200_pll_code_to_freq(0, ljm_3d_temp_high_table[i].freq_code),
            temp_code);

        spin_lock_irqsave(&ljm_dev->cru_lock, flags);
        if (tmp & m_CFG_GPU3D_CORE_CKSEL) {   /* GPU2PLL for 3D */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_CKMUX, tmp & ~(m_CFG_GPU3D_CORE_CKSEL | m_CFG_GPU3D_SH_CKSEL));   /* switch to GPUPLL for core and shader of 3D */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, ljm_3d_temp_high_table[i].freq_code);  /* configure GPU2PLL */
            /* wait GPU2PLL locked */
            while (!(LJM_CRU_R32(ljm_dev->cur_card, CRU_STATUS) & m_STA_GPU2PLL_LOCK)) {  /* wait DEVPLL locked */
                if (--timeout == 0) {
                    ERR_PRINT("wait CRU_GPU2PLL lock timeout\n");
                    break;
                }
            }
            //INFO_PRINT("timeout=%d\n", timeout);
            LJM_CRU_W32(ljm_dev->cur_card, CRU_CKMUX, tmp | (m_CFG_GPU3D_CORE_CKSEL | m_CFG_GPU3D_SH_CKSEL));   /* switch back to GPU2PLL for core and shader of 3D */
        } else {    /* GPUPLL for 3D，通过CRU_PLLEN关闭GPUPLL cken0和cken1设置3d频率会卡死 */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_CKMUX, tmp | (m_CFG_GPU3D_CORE_CKSEL | m_CFG_GPU3D_SH_CKSEL));   /* switch to GPU2PLL for core and shader of 3D */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, ljm_3d_temp_high_table[i].freq_code);   /* configure GPUPLL */
            /* wait GPUPLL locked */
            while (!(LJM_CRU_R32(ljm_dev->cur_card, CRU_STATUS) & m_STA_GPUPLL_LOCK)) {  /* wait DEVPLL locked */
                if (--timeout == 0) {
                    ERR_PRINT("wait CRU_GPUPLL lock timeout\n");
                    break;
                }
            }
            LJM_CRU_W32(ljm_dev->cur_card, CRU_CKMUX, tmp & ~(m_CFG_GPU3D_CORE_CKSEL | m_CFG_GPU3D_SH_CKSEL));   /* switch back to GPUPLL for core and shader of 3D */
        }
        spin_unlock_irqrestore(&ljm_dev->cru_lock, flags);
        return true;
    }

    return false;
}

static struct ljm_temp_to_freq ljm_2d_temp_high_table[] = {
    {LJM_PVT_TEMP_140, 5, LJM_PLL_CODE_400M}, /** >= 140 degree Celsius: frequency 400MHz */
    {LJM_PVT_TEMP_125, 5, LJM_PLL_CODE_600M}, /** >= 125 degree Celsius: frequency 600MHz */
    {LJM_PVT_TEMP_110, 5, LJM_PLL_CODE_800M}, /** >= 110 degree Celsius: frequency 800MHz */
    {LJM_PVT_TEMP_MINUS_62, 0, LJM_PLL_CODE_920M}, /** >= -62 degree Celsius: frequency 920MHz */
};

static const int ljm_2d_temp_high_table_count = sizeof(ljm_2d_temp_high_table)/sizeof(ljm_2d_temp_high_table[0]);

bool gp201_adjust_2d_freq(struct ljm_device *ljm_dev, u32 temp_code)
{
    u32 freq_code;
    u32 tmp, tmp1;
    int i;
    u32 timeout = 1000;
    unsigned long flags = 0;

    tmp = LJM_CRU_R32(ljm_dev->cur_card, CRU_CKMUX);
    if (!(tmp & 0x400))
        freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_DEVPLL);
    else
        freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL);

    /** high temperature to reduce frequency */
    for (i=0; i < ljm_2d_temp_high_table_count; i++) {
        /* skip low temperature items */
        if (temp_code < ljm_2d_temp_high_table[i].temp_code)
            continue;

        /* current frequency fit the temperature, just return */
        if (freq_code == ljm_2d_temp_high_table[i].freq_code)
            return false;

        /* 温度进入低温一档，但是未超过delt值，频率保持不变 */
        if (i > 0 &&
            (temp_code > (ljm_2d_temp_high_table[i-1].temp_code - ljm_2d_temp_high_table[i-1].delt) &&
            freq_code == ljm_2d_temp_high_table[i-1].freq_code))
            return false;

        /* adjust 3d frequency */
        DEBUG_PRINT("2D: %dKhz -> %dKhz\n",
            gp200_pll_code_to_freq(0, freq_code),
            gp200_pll_code_to_freq(0, ljm_2d_temp_high_table[i].freq_code));
        spin_lock_irqsave(&ljm_dev->cru_lock, flags);
        if (tmp & m_CFG_GPU2D_CORE_CKSEL) {   /* GPUPLL for 2D */
            tmp1 = LJM_CRU_R32(ljm_dev->cur_card, CRU_PLLEN);
            LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, tmp1 & ~(m_CFG_GPUPLL_CKEN0 | m_CFG_GPUPLL_CKEN1));   /* disable GPUPLL cken0, cken1 */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, ljm_2d_temp_high_table[i].freq_code);   /* configure GPUPLL */
            /* wait GPUPLL locked */
            while (!(LJM_CRU_R32(ljm_dev->cur_card, CRU_STATUS) & m_STA_GPUPLL_LOCK)) {  /* wait DEVPLL locked */
                if (--timeout == 0) {
                    ERR_PRINT("wait CRU_GPUPLL lock timeout\n");
                    break;
                }
            }
            //INFO_PRINT("timeout=%d\n", timeout);
            LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, tmp1 | (m_CFG_GPUPLL_CKEN0 | m_CFG_GPUPLL_CKEN1));    /* enable GPUPLL cken0, cken1 */
        } else {    /* DEVPLL for 2D */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_CKMUX, tmp & 0x1);   /* AXI selects HOSC */
            LJM_CRU_W32(ljm_dev->cur_card, CRU_DEVPLL, ljm_2d_temp_high_table[i].freq_code);    /* set DEVPLL */
            while (!(LJM_CRU_R32(ljm_dev->cur_card, CRU_STATUS) & m_STA_DEVPLL_LOCK)) {  /* wait DEVPLL locked */
                if (--timeout == 0) {
                    ERR_PRINT("wait DEVPLL lock timeout\n");
                    break;
                }
            }
            //INFO_PRINT("timeout=%d\n", timeout);
            LJM_CRU_W32(ljm_dev->cur_card, CRU_CKMUX, tmp);         /* AXI selects DEVPLL */
        }
        spin_unlock_irqrestore(&ljm_dev->cru_lock, flags);
        return true;
    }

    return false;
}

#ifdef TEST_AUTO_ADJUST_FREQ
uint testTempCode = 0x199;
module_param(testTempCode, uint, 0644);
MODULE_PARM_DESC(testTempCode, "Virtual PVT temperature code from module parameter");
#endif

static int gp201_monitor(void *data)
{
    struct ljm_device *ljm_dev = (struct ljm_device *)data;
    u32 tmp;

    DEBUG_PRINT("3D: high %px(%d)\n",
        (void*)ljm_3d_temp_high_table, ljm_3d_temp_high_table_count);
    DEBUG_PRINT("2D: high %px(%d)\n",
        (void*)ljm_2d_temp_high_table, ljm_2d_temp_high_table_count);

    while (1) {
        if (unlikely(ljm_dev->killMonitorThread)) {
            /* The daemon exits. */
            while (!kthread_should_stop())
                msleep(1);

            return 0;
        }

        mutex_lock(&ljm_dev->hw_mutex);
        /* adjust frequency for stability */
        tmp = LJM_PVT_R32(ljm_dev->cur_card, PVT_SAMP_CTL);
        if ((tmp & m_SAMP_CTL_MODE_MASK) != v_SAMP_CTL_MODE_TEMP0) {
            /* switch to TEMP0 mode */
            LJM_PVT_W32(ljm_dev->cur_card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_TEMP0);
            msleep(1100);   /* sleep 1.1s to get valid data */
        }

        tmp = LJM_PVT_R32(ljm_dev->cur_card, PVT_DATA) & m_PVT_DATA_MASK;
        mutex_unlock(&ljm_dev->hw_mutex);
#ifdef TEST_AUTO_ADJUST_FREQ
        /* use module parameter in test mode */
        tmp = testTempCode & m_PVT_DATA_MASK;
#endif
        if (!ljm_dev->suspend)
        {
            gp201_adjust_3d_freq(ljm_dev, tmp);
            gp201_adjust_2d_freq(ljm_dev, tmp);
        }

        ssleep(5);
    }

    return 0;
}

uint32_t ljm_gp201_get_freq(int card, int type);

/*****************************************************************************
*  函 数 名: gp201_init
*  功能描述: gp201初始化
*  输入参数: struct ljm_device *ljm_dev: gp设备数据指针
*  输出参数: 无
*  返 回 值: int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: huym
*            修改内容: 新生成函数
*
*****************************************************************************/
int gp201_init(struct ljm_device *ljm_dev)
{
    struct ljm_drm_info *info = ljm_dev->drm_info;
    struct ljm_port *port;
    int i, num = 0;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    board_info_image_handle(ljm_dev);
    ljm_dev->num_crtcs = GP201_MAX_CRTCS;
    ljm_dev->num_planes = GP201_MAX_PLANES;
    ljm_vgav2_init(ljm_dev->cur_card, ljm_dev->num_crtcs);
    ljm_pwm_fan_init(ljm_dev);
    ljm_pm_init(ljm_dev);
    if (!CHECK_GP201_MISC(g_mmio_bases[ljm_dev->cur_card])) {
        ERR_PRINT("GP201: MISC_INTR registers are not supported!\n");
        return -1;
    }

    if (ljm_dev->flags & LJMICRO_IS_FPGA) {
        ljm_dev->i2c_dev_cnt = 2;
        ljm_dev->i2c_devs[0] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_dw_init(ljm_dev->i2c_devs[0], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 0, 0);
        ljm_dev->i2c_devs[1] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_dw_init(ljm_dev->i2c_devs[1], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 1, 1);

        for(i=0; i<LJM_MAX_PORTS; i++)
        {
            if(info->ports[i].type <= 0)
                continue;

            port = kzalloc(sizeof(struct ljm_port), GFP_KERNEL);
            port->type = info->ports[i].type;
            switch(info->ports[i].type)
            {
                case LJM_PORT_DVO:
                    port->vdev.dvo.i2c_dev = ljm_dev->i2c_devs[info->ports[i].sub_id];
                    break;
                case LJM_PORT_HDMI:
                    ljm_hdmiv2_init(&port->vdev.hdmi, ljm_dev->mmio_base, REG_HDMI_OFFSET + REG_HDMI_SIZE * info->ports[i].sub_id, &info->ports[i]);
                    break;
#if USE_DP_CONNECTOR
                case LJM_PORT_DP:
                    port->vdev.dp.ddev = ljm_dev->ddev;
                    ljm_dp_init(&port->vdev.dp, ljm_dev->mmio_base, REG_DP_OFFSET + REG_DP_SIZE * info->ports[i].sub_id, &info->ports[i]);
                    break;
#endif
                default:
                    break;
            }
            port->id = info->ports[i].index;
            port->info = &info->ports[i];
            ljm_dev->ports[num] = port;
            ljm_dev->num_ports = ++num;
        }
    }
    else {
        /* only a DAC in CR register space, need not to initialize */
        ljm_dev->mcu = ljm_mcu_init(ljm_dev->cur_card, ljm_dev->mmio_base);

        /* i2c init */
        ljm_dev->i2c_dev_cnt = 6;
        ljm_dev->i2c_devs[0] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_oc_init(ljm_dev->i2c_devs[0], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 0, 0);
        ljm_dev->i2c_devs[1] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_oc_init(ljm_dev->i2c_devs[1], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 1, 1);
        ljm_dev->i2c_devs[2] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_oc_init(ljm_dev->i2c_devs[2], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 2, 2);
        ljm_dev->i2c_devs[3] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_oc_init(ljm_dev->i2c_devs[3], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 3, 3);
        ljm_dev->i2c_devs[4] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_dw_init(ljm_dev->i2c_devs[4], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 4, 4);
        ljm_dev->i2c_devs[5] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_dw_init(ljm_dev->i2c_devs[5], ljm_dev->mmio_base, REG_I2C_OC_OFFSET + REG_I2C_OC_SIZE * 5, 5);

        /*
         * VGA[0]   ==>  i2c_devs[0]
         *
         * HDMI[0]  ==>  REG_HDMI_OFFSET + REG_HDMI_SIZE * 0
         * HDMI[1]  ==>  REG_HDMI_OFFSET + REG_HDMI_SIZE * 1
         * HDMI[2]  ==>  REG_HDMI_OFFSET + REG_HDMI_SIZE * 2
         * HDMI[3]  ==>  REG_HDMI_OFFSET + REG_HDMI_SIZE * 3
         *
         * DVO[0]   ==>  i2c_devs[2]
         * DVO[0]   ==>  i2c_devs[3]
         *
         * DP[0]    ==>  REG_DP_OFFSET + REG_DP_SIZE * 0
         * DP[1]    ==>  REG_DP_OFFSET + REG_DP_SIZE * 1
         * DP[2]    ==>  REG_DP_OFFSET + REG_DP_SIZE * 2
         * DP[3]    ==>  REG_DP_OFFSET + REG_DP_SIZE * 3
        */
        for(i=0; i< LJM_MAX_PORTS; i++)
        {
            if(info->ports[i].type <= 0)
                continue;

#if !USE_DP_CONNECTOR
            if (info->ports[i].type == LJM_PORT_DP)
                continue;
#endif

            port = kzalloc(sizeof(struct ljm_port), GFP_KERNEL);
            port->type = info->ports[i].type;

            switch (info->ports[i].type)
            {
                case LJM_PORT_VGA:
                    port->vdev.vga.i2c_dev = ljm_dev->i2c_devs[info->ports[i].sub_id];
                    ljm_dac_ch_gain_set(&port->vdev.vga, 0x06, 0x06, 0x06);
                    break;
                case LJM_PORT_HDMI:
                    ljm_hdmiv2_init(&port->vdev.hdmi, ljm_dev->mmio_base, REG_HDMI_OFFSET + REG_HDMI_SIZE * info->ports[i].sub_id, &info->ports[i]);
                    break;
                case LJM_PORT_DVO:
                    port->vdev.dvo.i2c_dev = ljm_dev->i2c_devs[info->ports[i].sub_id + 2];
                    break;
#if USE_DP_CONNECTOR
                case LJM_PORT_DP:
                    port->vdev.dp.ddev = ljm_dev->ddev;
                    ljm_dp_init(&port->vdev.dp, ljm_dev->mmio_base, REG_DP_OFFSET + REG_DP_SIZE * info->ports[i].sub_id,  &info->ports[i]);
                    break;
#endif
                default:
                    break;
            }

            port->id = info->ports[i].index;
            port->info = &info->ports[i];
            ljm_dev->ports[num] = port;
            ljm_dev->num_ports = ++num;
        }

        ljm_dev->dp_irq_is_extra = ljm_i2c_dw_detect(ljm_dev->i2c_devs[5], DP_HPD_EXTRA_I2C_TARGET);
        if (ljm_dev->dp_irq_is_extra)
        {
            INFO_PRINT("DP uses extra IRQ\n");
        }

        /* enable bandwidth error interrupt */
        bweIntEnable = 1;
    }

    /* 98.B=10 */
    pci_write_config_byte(ljm_dev->pdev, 0x98, 0x10);   /* timeout configure for GP201 only */

    /* backlight */
    //ljm_register_backlight_device(ljm_dev, 0);
    //ljm_dev->backlight_dev_cnt = 1;

#if USE_CLOCK_REGISTER
    /* gp200 clk register */
    ljm_clk_init(ljm_dev);
#endif

    /** 根据温度变化自动调整频率功能 */
    {
        /* 设置ljm_axi/3d/2d_temp_table中的最高频率不超过系统刚加载时的频率 */
        u32 tmp, freq_code;
        int i;

        /* 3D temperature table adjust */
        tmp = LJM_CRU_R32(ljm_dev->cur_card, CRU_CKMUX);
        if (!(tmp & 0x100))
            freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL);
        else
            freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPU2PLL);

        for (i=0; i < ljm_3d_temp_high_table_count; i++) {
            if (gp200_pll_code_to_freq(0, ljm_3d_temp_high_table[i].freq_code) > gp200_pll_code_to_freq(0, freq_code)) {
                DEBUG_PRINT("%d: 3D freq_code from %08x to %08x (%d to %d)\n",
                    i, ljm_3d_temp_high_table[i].freq_code, freq_code,
                    gp200_pll_code_to_freq(0, ljm_3d_temp_high_table[i].freq_code),
                    gp200_pll_code_to_freq(0, freq_code));
                ljm_3d_temp_high_table[i].freq_code = freq_code;
            }
        }

        /* 确保3D核频率不超过1.25GHz */
        if (gp200_pll_code_to_freq(0, freq_code) > gp200_pll_code_to_freq(0, LJM_PLL_CODE_1250M))
            freq_code = LJM_PLL_CODE_1250M;
        ljm_3d_temp_high_table[ljm_3d_temp_high_table_count - 1].freq_code = freq_code;

        /* 2D temperature table adjust */
        tmp = LJM_CRU_R32(ljm_dev->cur_card, CRU_CKMUX);
        if (!(tmp & 0x400))
            freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_DEVPLL);
        else
            freq_code = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL);

        for (i=0; i < ljm_2d_temp_high_table_count; i++) {
            if (gp200_pll_code_to_freq(0, ljm_2d_temp_high_table[i].freq_code) > gp200_pll_code_to_freq(0, freq_code)) {
                DEBUG_PRINT("%d: 2D freq_code from %08x to %08x (%d to %d)\n",
                    i, ljm_2d_temp_high_table[i].freq_code, freq_code,
                    gp200_pll_code_to_freq(0, ljm_2d_temp_high_table[i].freq_code),
                    gp200_pll_code_to_freq(0, freq_code));
                ljm_2d_temp_high_table[i].freq_code = freq_code;
            }
        }

        /* 确保2D核频率不超过1.25GHz */
        if (gp200_pll_code_to_freq(0, freq_code) > gp200_pll_code_to_freq(0, LJM_PLL_CODE_1250M))
            freq_code = LJM_PLL_CODE_1250M;
        ljm_2d_temp_high_table[ljm_2d_temp_high_table_count - 1].freq_code = freq_code;

        /* start temperature monitor thread */
        ljm_dev->killMonitorThread = false;
        ljm_dev->monitorTask = kthread_run(gp201_monitor, (void*)ljm_dev, "ljmicro monitor");
        if (!ljm_dev->monitorTask) {
            ERR_PRINT("monitor thread failed\n");
            ljm_dev->killMonitorThread = true;
        }
    }

    /* i2s init */
    ljm_dev->axi_bus_clk = ljm_gp201_get_freq(ljm_dev->cur_card, freq_axi);
    for (i = 0; i < LJM_MAX_I2SS; i++)
    {
        ljm_dev->i2s_devs[i] = devm_kzalloc(ljm_dev->dev, sizeof(struct ljm_i2s_dev), GFP_KERNEL);
        if (!ljm_dev->i2s_devs[i]) {
            return -ENOMEM;
        }

        ljm_dev->i2s_devs[i]->i2s_base = ljm_dev->mmio_base + REG_I2S_OFFSET + REG_I2S_SIZE * i;
        if (IS_ERR(ljm_dev->i2s_devs[i]->i2s_base)) {
            return PTR_ERR(ljm_dev->i2s_devs[i]->i2s_base);
        }
    }

    FOOTER();
    return 0;
}

int gp201_irq_process(struct ljm_device *ljm_dev)
{
    irqreturn_t r = IRQ_NONE;
    int i;
    bool queue_hotplug = false;
    u32 misc_intr_src;
    u32 misc_intr_mask;
    u32 misc_intr_vgapd;

    misc_intr_src = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_SRC);
    misc_intr_mask = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN);

    if ((misc_intr_src & m_INTR_SRC_VGA) && (misc_intr_mask & m_INTR_EN_VGA)) {   /* DC interrupt */
        //DRM_DEBUG_KMS("misc_intr_src=0x%x misc_intr_mask=0x%x\n", misc_intr_src, misc_intr_mask);
        r = IRQ_HANDLED;
    } else if ((misc_intr_src & misc_intr_mask) == 0x0) {   /* extra duplicated interrupt, which has been handled off */
        /* GP200内部中断源到PCIE中断有屏蔽位进行控制，如果中断源与中断屏蔽寄存器与运算后为0，又由于GP200采用MSI中断模式，
         * OS理应收不到GP200的中断。但是GP200中断采取重复发送模式，避免中断丢失，可能存在中断正在被处理时，
         * PCIE刚好又重复发出该中断事件（在途中断），对于此类中断，服务例程应该返回OS为IRQ_HANDLED，标示中断已被处理。
         */
        r = IRQ_HANDLED;
        /* FIXME: 频繁发生，是否正常？ */
        //printk(KERN_INFO "ignore duplicated interrupt\n");
        //DRM_DEBUG_KMS("ignore duplicated interrupt\n");
    }

#if USE_DP_CONNECTOR
    /*meet with DP intterupt*/
    if ((misc_intr_src & m_INTR_SRC_EIRQ0) && (misc_intr_mask & m_INTR_EN_EIRQ0))
    {
        queue_hotplug = true;
        misc_intr_mask &= ~m_INTR_EN_EIRQ0;
        LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, misc_intr_mask);
        //TRACE("DP got hpd interrupt\n");
        printk("DP got hpd interrupt\n");
    }
#endif

    /* check DC (display controller) interrupt */
    if (misc_intr_src & m_INTR_SRC_VGA) {   /* meet with DC intterupt */
        misc_intr_vgapd = LJM_MISC_R32(ljm_dev->cur_card,  MISC_INTR_VGAPD);

        if (ljm_dev->mode_config_initialized && !(ljm_dev->flags & LJMICRO_IS_FPGA)) {
            /* check hdmi interrupt */
            for (i=0; i<ljm_dev->num_ports; i++) {
#if USE_HDMI_CONNECTOR
                if(ljm_dev->ports[i]->type == LJM_PORT_HDMI) {
                    if (misc_intr_vgapd & (m_INTR_VGAPD_HDMI0 << ljm_dev->ports[i]->vdev.hdmi.index)) {
                        if (ljm_hdmiv2_irq_process(&ljm_dev->ports[i]->vdev.hdmi) & LJM_IRQ_TYPE_HPD) {
                            queue_hotplug = true;
                            //TRACE("HDMI%d got hpd interrupt\n", ljm_dev->ports[i]->vdev.hdmi.index);
                            printk("HDMI%d got hpd interrupt\n", ljm_dev->ports[i]->vdev.hdmi.index);
                        } else {
                            ERR_PRINT("HDMI%d: mismatch with misc_intr\n", ljm_dev->ports[i]->vdev.hdmi.index);
                        }
                    }
                }
#endif
            }

#if USE_DAC_CONNECTOR
            /* check DAC interrupt */
            if (misc_intr_vgapd & m_INTR_VGAPD_DAC0) {
                if (ljm_dac_irq_process(ljm_dev->cur_card, 0) & LJM_IRQ_TYPE_HPD) {
                    queue_hotplug = true;
                    //TRACE("DAC0 got hpd interrupt\n");
                    printk("DAC0 got hpd interrupt\n");
                } else {
                    ERR_PRINT("DAC0: mismatch with misc_intr\n");
                }
            }
#endif
        }

        /* check crtc interrupt */
        for (i=0; i<GP201_MAX_CRTCS; i++) {
            /* check DVO HPD */
            if (misc_intr_vgapd & (m_INTR_VGAPD_CRTC0_HPD >> i * 3)) {
                if (i == 0) {
                    CR_WR(ljm_dev->cur_card, CRTC_HPD_ST_ADR, 0x02);    // clear hpd interrupt
                } else {
                    REG_W8(ljm_dev->cur_card, CRTC1_HPD_ST_ADR + CRTCx_REG_SIZE * (i - 1), 0x02);   // clear hpd interrupt
                }

                queue_hotplug = true;
                //TRACE("DVO%d got hpd interrupt\n", i);
                printk("DVO%d got hpd interrupt\n", i);
            }

            /* check vblank interrupt */
            if (misc_intr_vgapd & (m_INTR_VGAPD_CRTC0_VS >> i * 3)) {
                if (i == 0) {
                    CR_WR(ljm_dev->cur_card, CRTC_STATUS_ADR, 0x08);    // clear vsync interrupt
                } else {
                    REG_W8(ljm_dev->cur_card, CRTC1_HV_ST_ADR + CRTCx_REG_SIZE * (i - 1), 0x08);    // clear vsync interrupt
                }

#if USE_KMS_VBLANK
                if (ljm_dev->mode_config_initialized) {
                    /* handle vblank */
                    if (ljm_dev->irq.crtc_vblank_int[i]) {
                        drm_handle_vblank(ljm_dev->ddev, i);
                        //ljm_dev->pm.vblank_sync = true;    /* FIXME: pm.vblank_syn */
                        wake_up(&ljm_dev->irq.vblank_queue);
                    }
                    if (atomic_read(&ljm_dev->irq.pflip[i]))
                        ljm_crtc_handle_vblank(ljm_dev, i);
                    DRM_DEBUG_VBL("crtc %d vblank\n", i);
                }
#endif
            }

            /* check CRTC bandwidth error interrupt */
            if (misc_intr_vgapd & (m_INTR_VGAPD_CRTC0_BW >> i * 3)) {
                if (i == 0) {
                    CR_WR(ljm_dev->cur_card, CRTC_BW_ERR_ST_ADR, 0x02); // clear bandwidth error interrupt
                } else {
                    REG_W8(ljm_dev->cur_card, CRTC1_BW_ERR_ST_ADR + CRTCx_REG_SIZE*(i - 1), 0x02);    // clear bandwidth error interrupt
                }
#if defined(_DEBUG)
                //WARN_PRINT("CRTC%d got bandwidth error interrupt\n", i);
#endif
                bweIntCount++;
            }
        }

        /* check layer bandwidth error interrupt */
        for (i=0; i<GP201_MAX_PLANES; i++) {
            if (misc_intr_vgapd & (m_INTR_VGAPD_L1_BW >> i)) {
                if (REG_R8(ljm_dev->cur_card, LAYER1_BW_ERR_ST_ADDR + LAYER_REG_SIZE * i) & 0x2) {
#if defined(_DEBUG)
                    //WARN_PRINT("LAYER%d got bandwidth error interrupt\n", i);
#endif
                    bweIntCount++;
                }
                else {
                    ERR_PRINT("LAYER%d: mismatch with misc_intr\n", i);
                }

                REG_W8(ljm_dev->cur_card, LAYER1_BW_ERR_ST_ADDR + LAYER_REG_SIZE * i, 0x02);  // clear bandwidth error interrupt
            }
        }

        /* check zoom video interrupt */
        for (i=0; i<GP201_MAX_ZOOMVIDEO; i++) {
            if (misc_intr_vgapd & (m_INTR_VGAPD_ZV0 >> i)) {
                if (REG_R8(ljm_dev->cur_card, ZV0_INT_ST_ADDR + ZV_REG_SIZE * i) & 0x1) {
                    //TRACE( "ZV%d got frame interrupt\n", i);
                    ljm_zoomvideo_irq_handle(i);
                }
                else
                    ERR_PRINT("ZV%d: mismatch with misc_intr\n", i);

                REG_W8(ljm_dev->cur_card, ZV0_INT_ST_ADDR + ZV_REG_SIZE * i, 0x01);   // clear zoom video frame interrupt
            }
        }
    }

    if (queue_hotplug)
        schedule_delayed_work(&ljm_dev->hotplug_work, msecs_to_jiffies(500));

    return r;
}

uint32_t ljm_gp201_get_freq(int card, int type)
{
    uint64_t PLLOUT;
    uint32_t value;
    uint32_t DIVQ_CKDIV; /*从CRU_CKDIV读出来的值*/
    uint32_t DIVQx;      /*将DIVQ分解成AXI、ahb和apb三部分*/
    uint32_t DIVQ;
    int      i, x;
    switch (type)
    {
        case freq_axi:
            value      = REG_R32(card, 0x308004);
            DIVQ_CKDIV = REG_R32(card, 0x308034);
            x          = 1;
            DIVQ       = (value & 0x70) >> 4;
            DIVQx      = (DIVQ_CKDIV & 0xf00) >> 8;
            for (i = 0; i < DIVQ; i++)
            {
                x *= 2;
            }
            PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x * (DIVQx + 1) / 2)
                     / ((((value & 0x1f00) >> 8) + 1) * x * (DIVQx + 1));
            return PLLOUT;

        case freq_ahb:
            value      = REG_R32(card, 0x308004);
            DIVQ_CKDIV = REG_R32(card, 0x308034);
            x          = 1;
            DIVQ       = (value & 0x70) >> 4;
            DIVQx      = (DIVQ_CKDIV & 0xf0) >> 4;
            for (i = 0; i < DIVQ; i++)
            {
                x *= 2;
            }
            PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x * (DIVQx + 1) / 2)
                     / ((((value & 0x1f00) >> 8) + 1) * x * (DIVQx + 1));
            return PLLOUT;

        case freq_apb:
            value      = REG_R32(card, 0x308004);
            DIVQ_CKDIV = REG_R32(card, 0x308034);
            x          = 1;
            DIVQ       = (value & 0x70) >> 4;
            DIVQx      = DIVQ_CKDIV & 0xf;
            for (i = 0; i < DIVQ; i++)
            {
                x *= 2;
            }
            PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x * (DIVQx + 1) / 2)
                     / ((((value & 0x1f00) >> 8) + 1) * x * (DIVQx + 1));
            return PLLOUT;

        case freq_pcie:
            value = REG_R32(card, 0x308008);
            x     = 1;
            DIVQ  = (value & 0x70) >> 4;
            for (i = 0; i < DIVQ; i++)
            {
                x *= 2;
            }
            PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                     / ((((value & 0x1f00) >> 8) + 1) * x);
            return PLLOUT;

        case freq_vpu:
            value = REG_R32(card, 0x308008);
            x     = 1;
            DIVQ  = (value & 0x70) >> 4;
            for (i = 0; i < DIVQ; i++)
            {
                x *= 2;
            }
            PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                     / ((((value & 0x1f00) >> 8) + 1) * x);
            return PLLOUT;

        case freq_3d:
            value = REG_R32(card, 0x308030);
            if (!(value & 0x100))
            {
                value = REG_R32(card, 0x30800c);
                x     = 1;
                DIVQ  = (value & 0x70) >> 4;
                for (i = 0; i < DIVQ; i++)
                {
                    x *= 2;
                }
                PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                         / ((((value & 0x1f00) >> 8) + 1) * x);
                return PLLOUT;
            }
            else
            {
                value = REG_R32(card, 0x308010);
                x     = 1;
                DIVQ  = (value & 0x70) >> 4;
                for (i = 0; i < DIVQ; i++)
                {
                    x *= 2;
                }
                PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                         / ((((value & 0x1f00) >> 8) + 1) * x);
                return PLLOUT;
            }
            break;

        case freq_2d:
            value = REG_R32(card, 0x308030);
            if (!(value & 0x400))
            {
                value = REG_R32(card, 0x308004);
                x     = 1;
                DIVQ  = (value & 0x70) >> 4;
                for (i = 0; i < DIVQ; i++)
                {
                    x *= 2;
                }
                PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                         / ((((value & 0x1f00) >> 8) + 1) * x);
                return PLLOUT;
            }
            else
            {
                value = REG_R32(card, 0x30800c);
                x     = 1;
                DIVQ  = (value & 0x70) >> 4;
                for (i = 0; i < DIVQ; i++)
                {
                    x *= 2;
                }
                PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                         / ((((value & 0x1f00) >> 8) + 1) * x);
                return PLLOUT;
            }
            break;
        case freq_ddr:
            value = REG_R32(card, 0x308014);
            x     = 1;
            DIVQ  = (value & 0x70) >> 4;
            for (i = 0; i < DIVQ; i++)
            {
                x *= 2;
            }
            PLLOUT = (27 * 2 * (((value & 0x1ff0000) >> 16) + 1) + (((value & 0x1f00) >> 8) + 1) * x / 2)
                     / ((((value & 0x1f00) >> 8) + 1) * x);
            return PLLOUT;
        default:
            printk(KERN_INFO "%s: command error\n", __FUNCTION__);
            return -EPERM;
    }
}

static void gp201_mode_adjust(struct drm_display_mode *mode, bool force_sync)
{
    int hsync;

    /* adjust 4K@30@262.75 to vesa 4K@30 */
    if (mode->hdisplay == 3840 && mode->vdisplay == 2160 && mode->clock == 262750) {
        DRM_DEBUG_KMS("adjust 4K@30@262.75MHz to vesa 4K@30\n");
        mode->clock = 297000;
        mode->hsync_start = 4016;
        mode->hsync_end = 4104;
        mode->htotal = 4400;
        mode->vsync_start = 2168;
        mode->vsync_end = 2178;
        mode->vtotal = 2250;
        mode->flags = DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC;
    }

    if (mode->hdisplay == 3840 && mode->vdisplay == 2160 && mode->clock > 300000) {
        if (drm_mode_vrefresh(mode) < 55) { /* adjust 4K@30@50 to vesa 4K@50 */
            DRM_DEBUG_KMS("adjust 4K@30@50 to vesa 4K@50\n");
            mode->clock = 594000;
            mode->hsync_start = 4896;
            mode->hsync_end = 4984;
            mode->htotal = 5280;
            mode->vsync_start = 2168;
            mode->vsync_end = 2178;
            mode->vtotal = 2250;
            mode->flags = DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC;
        } else {    /* adjust 4K@30@60 to vesa 4K@60 */
            DRM_DEBUG_KMS("adjust 4K@30@60 to vesa 4K@60\n");
            mode->clock = 594000;
            mode->hsync_start = 4016;
            mode->hsync_end = 4104;
            mode->htotal = 4400;
            mode->vsync_start = 2168;
            mode->vsync_end = 2178;
            mode->vtotal = 2250;
            mode->flags = DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC;
        }
    }

    if (mode->hdisplay == 2560 && mode->vdisplay == 1440) {
        DRM_DEBUG_KMS("adjust 2560x1440@%dKHz to 146250 kHz\n", mode->clock);
        mode->clock = 146250;
        mode->hsync_start = 2680;
        mode->hsync_end = 2944;
        mode->htotal = 3328;
        mode->vsync_start = 1443;
        mode->vsync_end = 1448;
        mode->vtotal = 1468;
        mode->flags = DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_PVSYNC;
    }

    if(force_sync)
    {
        hsync = mode->hsync_end - mode->hsync_start;
        mode->hsync_start = mode->hdisplay + 1;
        mode->hsync_end = mode->hsync_start + hsync;
    }

    return;
}

/**
 * gp201_encoder_mode_fixup - 修改mode使之能被硬件支持.
 *
 * @ljm_dev: ljm_device指针.
 * @mode: 源mode指针.
 * @adjustedMode: 修改后的mode，mode_set时真正使用的是此mode.
 *
 * 针对一些特殊mode做修改，调整参数使之能够被驱动支持。该函数会在encoder与crtc进行modeset之前被drm调用。
 * 具体调用过程可参考内核drm源码drm_crtc_helper_set_mode()函数。
 *
 * Returns:
 * True if the mode was set successfully, false otherwise.
 */
bool gp201_encoder_mode_fixup(struct drm_encoder *encoder, const struct drm_display_mode *mode, struct drm_display_mode *adjustedMode)
{
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    struct ljm_port *ljm_port = to_ljm_port(ljm_encoder);

    /*预留，用于设置crtc_timing, 驱动中暂未使用*/
    //drm_mode_set_crtcinfo(adjustedMode, 0);

    if(ljm_port->type == LJM_PORT_HDMI && ljm_port->vdev.hdmi.edid.vendor == TCD0709)
        gp201_mode_adjust(adjustedMode, true);
    else
        gp201_mode_adjust(adjustedMode, false);

    return true;
}

uint32_t gp201_pvt_get_temp(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if (((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_TEMP0) && ((tmp & 0x1) == index)) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_TEMP0 + index);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp201_pvt_get_volt(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if (((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_VOLT0) && ((tmp & 0x1) == index)) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_VOLT0 + index);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp201_pvt_get_svt_process(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if (((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_SVT0_PROCESS) && ((tmp & 0x1) == index)) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_SVT0_PROCESS + index);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp201_pvt_get_lvt_process(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if (((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_LVT0_PROCESS) && ((tmp & 0x1) == index)) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_LVT0_PROCESS + index);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp201_pvt_get_ulvt_process(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if (((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_ULVT0_PROCESS) && ((tmp & 0x1) == index)) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_ULVT0_PROCESS + index);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
    }

    return tmp & m_PVT_DATA_MASK;
}

int gp201_crtc_load_lut(struct ljm_crtc *ljm_crtc, u16 *red, u16 *green, u16 *blue, uint32_t gamma_size)
{
    return 0;
}

int gp201_resume(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    int i;
#if USE_DP_CONNECTOR
    struct drm_display_mode *mode;
#endif
    struct ljm_port *ljm_port;

    HEADER_ARG("drm=%p", ddev);

    DEBUG_PRINT("gp200_resume...");

    /* vga init */
    ljm_vgav2_init(ljm_dev->cur_card, ljm_dev->num_crtcs);

#if USE_DP_CONNECTOR
    if(ljm_dev->once_edp)
    {
        struct ljm_crtc *ljm_crtc;

        for (i=0; i<ljm_dev->num_crtcs; i++) {
            ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[i]);
            if (ljm_crtc->once_edp_mode) {
                DEBUG_PRINT("crtc%d: resume\n", ljm_crtc->id);
                mode = ljm_crtc->once_edp_mode;
                DEBUG_PRINT("crtc %d mode: %d %d %d %d %d %d %d %d %d\n", ljm_crtc->id,
                    mode->hdisplay, mode->hsync_start, mode->hsync_end, mode->htotal,
                    mode->vdisplay, mode->vsync_start, mode->vsync_end, mode->vtotal,
                    mode->clock);
                ljm_vgav2_crtc_mode_set(ljm_dev->cur_card, ljm_crtc->id, ljm_crtc->once_edp_mode, ljm_dev->flags & LJMICRO_IS_FPGA);
            }
        }
        REG_W32(ljm_dev->cur_card, HDMI_DAC_DVO_OUT_MUX, ljm_dev->once_edp_vga_mux);
    }

    /* restore register HDMI_DAC_DVO_OUT_MUX */

    DEBUG_PRINT("dptx init\n");
    ljm_resume_dp(ddev);

    /* DP mode set */
    for (i=0; i<ljm_dev->num_ports; i++) {
        ljm_port = ljm_dev->ports[i];

        if (ljm_dev->once_edp && ljm_port->base.once_edp_mode) {
            DEBUG_PRINT("port%d: mode set\n", ljm_port->id);
            mode = ljm_port->base.once_edp_mode;
            DEBUG_PRINT("DP%d mode: %d %d %d %d %d %d %d %d %d\n", ljm_port->vdev.dp.index,
                mode->hdisplay, mode->hsync_start, mode->hsync_end, mode->htotal,
                mode->vdisplay, mode->vsync_start, mode->vsync_end, mode->vtotal,
                mode->clock);
            ljm_dp_mode_set(&ljm_port->vdev.dp, ljm_port->base.once_edp_mode);
        }
    }
#endif

    /* HDMI irq set */
    for (i=0; i<ljm_dev->num_ports; i++) {
        ljm_port = ljm_dev->ports[i];
        if(ljm_port->type == LJM_PORT_HDMI)
            ljm_hdmiv2_irq_init(&ljm_port->vdev.hdmi);
    }
    /* restore register MISC_INTR_ENPCIE */
    LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, ljm_dev->misc_intr_maskpcie);

    FOOTER();
    return 0;
}

int gp201_power_set_state(struct ljm_device *ljm_dev, int power_state)
{
    int r = 0;
    uint32_t val;

    ljm_dev->killMonitorThread = true;/* 关闭GP201的自动调频线程 */
    val = LJM_CRU_R32(ljm_dev->cur_card, CRU_PLLEN);
    switch (power_state)
    {
        case LJM_POWER_STATE_PERFORMANCE:
            //printk("power performance...\n");
            if ((val & 0x3000)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL) != g_gpupll_val)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPU2PLL) != g_gpu2pll_val))
            {
                /* bypass 3D, 2D clock */
                val |= 0x3000;
                LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, val);

                /* restore 3D, 2D clock */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, g_gpupll_val );
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, g_gpu2pll_val);

                val &= 0xffffcfff;
                LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, val);
            }

            /* disable power management for cores */
            gpu_config_power_management(-1, false);
            break;
        case LJM_POWER_STATE_SAVE:
            if ((val & 0x3000)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL) != PLL_FREQ_500)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPU2PLL) != PLL_FREQ_600))
            {
                /* bypass 3D, 2D clock */
                val |= 0x3000;
                LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, val);

                /* restore 3D, 2D clock */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, PLL_FREQ_500);
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, PLL_FREQ_600);

                val &= 0xffffcfff;
                LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, val);
            }

            /* enable power management for cores */
            gpu_config_power_management(-1, true);
            break;
        case LJM_POWER_STATE_STANDBY:
            //printk("power standby...\n");
            if ((val & 0x3000)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL) != PLL_FREQ_400)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPU2PLL) != PLL_FREQ_200))
            {
                /* bypass 3D, 2D clock */
                val |= 0x3000;
                LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, val);

                /* restore 3D, 2D clock */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, PLL_FREQ_400 );  /* gpupll (2D): 400M */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, PLL_FREQ_200);   /* gpu2pll (3D): 200M */

                val &= 0xffffcfff;
                LJM_CRU_W32(ljm_dev->cur_card, CRU_PLLEN, val);
            }

            /* enable power management for cores */
            gpu_config_power_management(-1, true);
            break;
        default:
            r = -EINVAL;
    }

    return r;
}