/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __GP201_H__
#define __GP201_H__

#include "gp200.h"

#define GP201_MAX_CRTCS         4	/* max crts */
#define GP201_MAX_PLANES        4	/* max layers */
#define GP201_MAX_ZOOMVIDEO     4	/* max zoom video */

/*******************************************************************************
 * Partion of GP201's visual video memory (256MB)
 * Region           Size     Remark
 * VRAM_NORMAL      bar0     used for 2D, 3D, framebuffer, etc.
 * VRAM_FIX         4MB      fix region for misc use, such as DMA table
 *******************************************************************************/

/********************************************************************************
 * GP201 local address
 * PCIE SPACE                   AMBA Space
 * BAR          Size            Range                           Remark
 * Legacy IO    3B0~3BB,3C0~3DF 0x0000_0000~0x0000_03FF         VGA legacy registers
 * Legacy MEM   A0000~BFFFF     0x000A_0000~0x000B_FFFF         VGA legacy VRAM
 * ROM (BAR6)   256KB           0x0020_0000~0x0023_FFFF         Flash(ROM)
 * IO (BAR4)    256B            0x0010_0008~0x0010_000F         SPI Flash Controller
 * MEM_IO(BAR2) 16MB            0x0100_0000~0x01FF_FFFF         Registers
 * MEM (BAR0)   256MB           0x80_0000_0000~0x80_0FFF_FFFF   VRAM
 * MEM          32GB            0x80_0000_0000~0x87_FFFF_FFFF   VRAM
 * HOST_MEM     512GB           0x00_0000_0000~0x7F_FFFF_FFFF   Host data
********************************************************************************/

/*****************************************************************************
 * GP201 PCIE address mapping by host (CPU)
 * BAR          Size    Offset                  Remark
 * MEMIO(BAR2)  16MB
 *                      0x01_0000~01_0FFF       VGA Registers (32bit AHB, 4K)

 *                      0x03_0000~03_7FFF       SRAM0 (32bit AHB, 256K)
 *                      0x03_8000~03_FFFF       SRAM1 (32bit AHB, 256K)

 *                      0x04_0000~07_FFFF       3D GPU Registers (32bit AHB, 256K)
 *                      0x08_0000~0B_FFFF       3D1 GPU Registers (32bit AHB, 256K)

 *                      0x0C_0000~0F_FFFF       2D GPU Registers (32bit AHB, 256K)

 *                      0x10_0000~10_03FF       DDC0 (8bit APB, 1K)
 *                      0x10_0400~10_07FF       DDC1 (8bit APB, 1K)
 *                      0x10_0800~10_0BFF       DDC2 (8bit APB, 1K)
 *                      0x10_0C00~10_0FFF       DDC3 (8bit APB, 1K)
 *                      0x10_1000~10_13FF       DDC4 (8bit APB, 1K)
 *                      0x10_1400~10_17FF       DDC5 (8bit APB, 1K)
 *                      0x10_1800~10_1BFF       DDC6 (8bit APB, 1K)

 *                      0x10_2000~10_23FF       DAC0 (8bit APB, 1K)
 *                      0x10_2400~10_27FF       DAC1 (8bit APB, 1K)
 *                      0x10_2800~10_2BFF       DAC2 (8bit APB, 1K)
 *                      0x10_2C00~10_2FFF       DAC3 (8bit APB, 1K)

 *                      0x11_0000~11_FFFF       HDMI0 (8bit APB, 64K)
 *                      0x12_0000~12_FFFF       HDMI1 (8bit APB, 64K)
 *                      0x13_0000~13_FFFF       HDMI2 (8bit APB, 64K)
 *                      0x14_0000~14_FFFF       HDMI3 (8bit APB, 64K)

 *                      0x20_0000~27_FFFF       PCIE Config (32bit APB, 512K)

 *                      0x30_8000~30_83FF       Clock Reset Registers (32bit APB, 1K)
 *                      0x30_8400~30_87FF       Misc (gart_base) (32bit APB, 1K)

 *                      0x30_8800~30_8BFF       I2S0 Registers (32bit APB, 1K)
 *                      0x30_8C00~30_8FFF       I2S1 Registers (32bit APB, 1K)

 *                      0x30_9000~30_9FFF       Monitor Registers (32bit APB, 4K)
 *                      0x30_A000~30_A3FF       IO Mux Registers (32bit APB, 1K)

 *                      0x30_A400~30_ABFF       PCIE PHY (32bit APB, 2K)

 *                      0x30_AC00~30_AFFF       Flash (MEM_IO) (32bit APB, 1K)

 *                      0x31_0000~31_FFFF       VPU0 Registers (32bit APB, 64K)
 *                      0x32_0000~32_FFFF       VPU1 Registers (32bit APB, 64K)

 *                      0x33_0000~33_0FFF       DMA0 Registers (32bit APB, 4K)
 *                      0x33_1000~33_1FFF       DMA1 Registers (32bit APB, 4K)
 *                      0x33_2000~33_23FF       DMA Configuration Registers (32bit APB, 1K)

 *                      0x3C_0000~3C_7FFF       DDR4 Ctrl0 Registers (32bit APB, 32K)
 *                      0x3C_8000~3C_FFFF       DDR4 Ctrl1 Registers (32bit APB, 32K)
 *                      0x3D_0000~3D_7FFF       DDR4 Ctrl2 Registers (32bit APB, 32K)
 *                      0x3D_8000~3D_FFFF       DDR4 Ctrl3 Registers (32bit APB, 32K)

 *                      0x40_0000~43_FFFF       DDR4 PHY0 Registers (32bit APB, 256K)
 *                      0x44_0000~47_FFFF       DDR4 PHY1 Registers (32bit APB, 256K)
 *                      0x48_0000~4B_FFFF       DDR4 PHY2 Registers (32bit APB, 256K)
 *                      0x4C_0000~4F_FFFF       DDR4 PHY3 Registers (32bit APB, 256K)

 *                      0x50_0000~51_FFFF       DP0 (32bit APB, 128K)
 *                      0x52_0000~53_FFFF       DP1 (32bit APB, 128K)
 *                      0x54_0000~55_FFFF       DP2 (32bit APB, 128K)
 *                      0x56_0000~57_FFFF       DP3 (32bit APB, 128K)

 *                      0x80_0000~87_FFFF       NOC Registers (64bit AXI, 512K)

 *                      0xC0_0000~FF_FFFF       PCIE_DBI (32bit AXI, 4M)
 *
 * smem (BAR0)  256MB   0x000_0000~0xFFF_FFFF   VRAM
 *****************************************************************************/
#define REG_I2C_OC_OFFSET   0x100000    /* I2C (OpenCore) for I2C 0~3 */
#define REG_I2C_OC_SIZE     0x400       /* 1K */
#define REG_I2C_DW_OFFSET   0x101000    /* I2C (DW) for I2C 4~5 */
#define REG_I2C_DW_SIZE     0x400       /* 1K */
#define REG_MONITOR_OFFSET  0x309000    /* Monitor */
#define REG_MONITOR_SIZE    0x1000
#define REG_HDMI_OFFSET     0x110000
#define REG_HDMI_SIZE       0x10000     /* 64K */
#define REG_DP_OFFSET       0x500000
#define REG_DP_SIZE         0x20000     /* 128K */

#define REG_CRU_OFFSET      0x308000
#define REG_CRU_SIZE        0x400

/* #define CRU_PLLEN            0x0
 * #define m_CFG_DEVPLL_BYPASS      (1U << 15)
 * #define m_CFG_DEV2PLL_BYPASS     (1U << 14)
 * #define m_CFG_GPUPLL_BYPASS      (1U << 13)
 * #define m_CFG_GPU2PLL_BYPASS     (1U << 12)
 * #define m_CFG_DDRPLL0_BYPASS     (1U << 11)
 * #define m_CFG_DDRPLL1_BYPASS     (1U << 10)
 * #define m_CFG_DDRPLL2_BYPASS     (1U << 9)
 * #define m_CFG_DDRPLL3_BYPASS     (1U << 8)
 * #define m_CFG_PXLPLL0_BYPASS     (1U << 7)
 * #define m_CFG_PXLPLL1_BYPASS     (1U << 6)
 * #define m_CFG_PXLPLL2_BYPASS     (1U << 5)
 * #define m_CFG_PXLPLL3_BYPASS     (1U << 4)
 */
#define m_CFG_GPUPLL_CKEN1       (1U << 3)
#define m_CFG_GPUPLL_CKEN0       (1U << 2)
#define m_CFG_DEV2PLL_CKEN1      (1U << 1)
#define m_CFG_DEVPLL_CKEN3       (1U << 0)

/* #define CRU_CKEN1            0x3c
 * #define m_CFG_QSPI_CKEN       (1U << 31)
 * #define m_CFG_DMA_CKEN        (1U << 29)
 * #define m_CFG_MCU_CKEN        (1U << 28)
 * #define m_CFG_I2S0_CKEN       (1U << 27)
 * #define m_CFG_I2S1_CKEN       (1U << 26)
 */
#define m_CFG_DDRC0_CORE_CKEN (1U << 11)
#define m_CFG_DDRC1_CORE_CKEN (1U << 10)
#define m_CFG_DDRC2_CORE_CKEN (1U << 9)
#define m_CFG_DDRC3_CORE_CKEN (1U << 8)
#define m_CFG_DDRC0_BUS_CKEN  (1U << 3)
#define m_CFG_DDRC1_BUS_CKEN  (1U << 2)
#define m_CFG_DDRC2_BUS_CKEN  (1U << 1)
#define m_CFG_DDRC3_BUS_CKEN  (1U << 0)

#define REG_MISC_OFFSET     0x308400
#define REG_MISC_SIZE       0x400
#define REG_I2S_OFFSET      0x308800
#define REG_I2S_SIZE        0x400

#define REG_SPI_OFFSET      0x30AC00

#define REG_DDR_CTRL_OFFSET 0x3C0000
#define REG_DDR_PHY_OFFSET  0x400000

/**
 * NOC
 */
#define REG_NOC_OFFSET      0x800000

#define NOC_REMAP_BLK0     0x30
#define NOC_REMAP_BLK_SIZE 0x400
#define NOC_REMAP_BLK_CNT  32
#define v_REMAP_DDR0       0
#define v_REMAP_DDR1       1
#define v_REMAP_DDR01      2
#define v_REMAP_DDR23      4
#define v_REMAP_DDR0123    8

/**
 * PVT
 */
#define m_SAMP_CTL_MODE_MASK     0x1ffffe
#define v_SAMP_CTL_MODE_TEMP0    0xf0000
#define v_SAMP_CTL_MODE_TEMP1    0xf0001
#define v_SAMP_CTL_MODE_VOLT0    0xf0100
#define v_SAMP_CTL_MODE_VOLT1    0xf0101
#define v_SAMP_CTL_MODE_SVT0_PROCESS    0xf0600
#define v_SAMP_CTL_MODE_SVT1_PROCESS    0xf0601
#define v_SAMP_CTL_MODE_LVT0_PROCESS    0xf0200
#define v_SAMP_CTL_MODE_LVT1_PROCESS    0xf0201
#define v_SAMP_CTL_MODE_ULVT0_PROCESS   0xf0400
#define v_SAMP_CTL_MODE_ULVT1_PROCESS   0xf0401

/**
 * IO MUX
 */
#define REG_IOMUX_OFFSET    0x30A000

/**
 * MCU
*/
#define MCU_FLASH_OFFSET 0x220000

/** chip value for GP201 */
#define MISC_CHIP_R00_DEF_VALUE 0x3c3c3c3c /* 0x5a5a5a5a */
#define MISC_CHIP_R01_DEF_VALUE 0x02010000

#define DP_HPD_EXTRA_I2C_TARGET     0x38

#define CHECK_GP201_MISC(base) \
    (/*readl((volatile void __iomem *)(base) + REG_MISC_OFFSET + MISC_CHIP_R00) == MISC_CHIP_R00_DEF_VALUE*/ \
     /*&&*/ readl((volatile void __iomem *)(base) + REG_MISC_OFFSET + MISC_CHIP_R01) == MISC_CHIP_R01_DEF_VALUE)

#endif  /* ~__GP201_H__ */
