/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
*
*  版权所有(c)，2020-2099，中船第709研究所
*
*******************************************************************************
*
*  文件名称: gp202.c
*  版 本 号: V1.0
*  作    者: huym
*  生成日期: 2022年4月21日
*
*  功能描述: gp200
*
*  修改历史:
*         1. 日    期: 2023年10月31日
*            作    者: huym
*            修改内容: 创建文件
*
******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/
#include "ljmicro_device.h"
#include "ljmicro_vgav2.h"
#include "ljmicro_mode.h"
#include "gp202.h"
#include "ljmicro_mcu.h"
#include "ljmicro_gpio.h"
#include "ljmicro_pm.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

extern int ddrChannelMask;
int _gp202_crtc_load_lut(struct ljm_crtc *ljm_crtc);

/*****************************************************************************
 *  函 数 名: gp202_pre_set_clock
 *  功能描述: 预设置时钟频率
 *  输入参数: void *mmio_base  kernel virtual base address for registers
 *  输出参数: 无
 *  返 回 值: static void
 *
 *  修改历史:
 *         1. 日    期: 2024年5月20日
 *            作    者: huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static void gp202_pre_set_clock(void *mmio_base)
{
    u32 data;

    HEADER_ARG("mmio_base=%p", mmio_base);

    /** CRU_DEVPLL default 801MHz without bypass */
    if (g_devpll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_DEVPLL, g_devpll_val);   /* restore DEVPLL */

    /** CRU_DEV2PLL default 999MHz without bypass */
    if (g_dev2pll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_DEV2PLL, g_dev2pll_val);   /* restore DEV2PLL */

    /** CRU_GPUPLL default 801MHz */
    if (g_gpupll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_GPUPLL, g_gpupll_val);   /* restore GPUPLL */
    else
    {
        data = LJM_CRU_R32_BASE(mmio_base, CRU_GPUPLL);
        data &= ~m_CFG_PLL_BP;
        LJM_CRU_W32_BASE(mmio_base, CRU_GPUPLL, data);
    }

    /** CRU_GPU2PLL default 499.5MHz with bypass */
    if (g_gpu2pll_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_GPU2PLL, g_gpu2pll_val);   /* restore GPU2PLL */
    else
    {
        data = LJM_CRU_R32_BASE(mmio_base, CRU_GPU2PLL);
        data &= ~m_CFG_PLL_BP;
        LJM_CRU_W32_BASE(mmio_base, CRU_GPU2PLL, data);
    }

    if (g_ckmux_val != 0)
        LJM_CRU_W32_BASE(mmio_base, CRU_CKMUX, g_ckmux_val);    /* restore CKMUX */
    else
    {
        /** default: 3D core -> GPUPLL/OUT0, 3D shader -> GPUPLL/OUT0,
         ** 2D core -> DEV2PLL/OUT0,
         ** VPU core -> DEV2PLL/OUT1 */
        data = LJM_CRU_R32_BASE(mmio_base, CRU_CKMUX);
        data |= m_CFG_BUS_CKSEL;    /**< BUS clock select PLL */
        LJM_CRU_W32_BASE(mmio_base, CRU_CKMUX, data);
    }

    /** enable clock */
    LJM_CRU_W32_BASE(mmio_base, CRU_CKEN0, m_CFG_VGA_PLX1_CKEN | m_CFG_VGA_PLX0_CKEN | m_CFG_HDMI1_BUS_CKEN | m_CFG_HDMI0_BUS_CKEN | m_CFG_DP0_BUS_CKEN
        | m_CFG_VGA_BUS_CKEN | m_CFG_VPU_BUS_CKEN | m_CFG_GPU2D_BUS_CKEN | m_CFG_GPU3D_BUS_CKEN);

    LJM_CRU_W32_BASE(mmio_base, CRU_CKEN1, m_CFG_QSPI_CKEN | m_CFG_DMA_CKEN | m_CFG_MCU_CKEN | m_CFG_I2S0_CKEN | m_CFG_I2S1_CKEN
        | m_CFG_DDRC1_CORE_CKEN | m_CFG_DDRC0_CORE_CKEN | m_CFG_DDRC1_BUS_CKEN | m_CFG_DDRC0_BUS_CKEN);

    /** release reset */
    data = LJM_CRU_R32_BASE(mmio_base, CRU_SOFTRST0);
    data &= ~(m_CFG_RESET_PXL1 | m_CFG_RESET_PXL0 | m_CFG_RESET_HDMI1 | m_CFG_RESET_HDMI0 | m_CFG_RESET_DP0);
    LJM_CRU_W32_BASE(mmio_base, CRU_SOFTRST0, data);

    data = LJM_CRU_R32_BASE(mmio_base, CRU_SOFTRST1);
    data &= ~(m_CFG_RESET_I2S0 | m_CFG_RESET_I2S1);
    LJM_CRU_W32_BASE(mmio_base, CRU_SOFTRST1, data);

    /* un-bypass for PXLPLL0, PXLPLL1 */
    data = LJM_CRU_R32_BASE(mmio_base, CRU_PLLEN);
    data &= ~(m_CFG_PXLPLL0_BYPASS | m_CFG_PXLPLL1_BYPASS);
    LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN, data);

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: gp202_ddrv2_noc_init
 *  功能描述: initialize ddrv2 controller and phy
 *  输入参数: void *mmio_base   MMIO基地址
 *            uint8_t channel_mask  DDR通道mask位
 *            int ddr4_freq     DDR4频率枚举
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2024年6月22日
 *            作    者: zhaiyf
 *            修改内容: 新生成函数
 *****************************************************************************/
int gp202_ddrv2_noc_init(void *mmio_base, uint8_t channel_mask)
{
    u32 val;
    int i;

    HEADER_ARG("mmio_base=%p channel_mask=0x%x", mmio_base, channel_mask);

    switch (channel_mask) {
        case 0x1:
            val = v_REMAP_DDR0;
            break;
        case 0x2:
            val = v_REMAP_DDR1;
            break;
        case 0x3:
            val = v_REMAP_DDR01;
            break;
        default:
            val = v_REMAP_DDR0;
            break;
    }

    for (i = 0; i < NOC_REMAP_BLK_CNT; i++) {
        LJM_NOC_W32_BASE(mmio_base, NOC_REMAP_BLK0 + i * NOC_REMAP_BLK_SIZE, val);
    }

    FOOTER();
    return 0;
}

unsigned char gp202_ddrv2_get_channel_mask(int card)
{
    uint32_t tmp;
    tmp = LJM_NOC_R32(card, NOC_REMAP_BLK0);

    switch (tmp) {
    case v_REMAP_DDR0:
        return (unsigned char)0x1;
    case v_REMAP_DDR1:
        return (unsigned char)0x2;
    case v_REMAP_DDR01:
    case v_REMAP_DDR01_128B:
    case v_REMAP_DDR01_512B:
        return (unsigned char)0x3;
    default:
        return (unsigned char)0x1;
    }
}

/**
 * @brief GP202 special pre-initialization
 * @param  card card index
 * @param  mmio_base  kernel virtual base address for registers
 * @param  pddr_size    pointer to return DDR size
 * @param  flags    flags of PCI device
 * @return int  if succeed return 0, otherwise return error code
 *
 * history:
 *  1. date: 2023.10.31
 *     author: huym
 *     description: create
 *  2. date: 2024.5.20
 *     author: huym
 *     description: add flags parameter and check FPGA board here
 */
int gp202_pre_init(int card, void *mmio_base, unsigned long *pddr_size, unsigned long flags)
{
    HEADER_ARG("card=%d mmio_base=%p flags=%lx", card, mmio_base, flags);

    /** init gpu hardware */
    ljm_gpu_hardware.max_ddr_channels = 2;
    ljm_gpu_hardware.reg_ddr_ctr_offset[0] = REG_DDR_CTRL0_OFFSET;
    ljm_gpu_hardware.reg_ddr_phy_offset[0] = REG_DDR_PHY0_OFFSET;
    ljm_gpu_hardware.reg_ddr_ctr_offset[1] = REG_DDR_CTRL1_OFFSET;
    ljm_gpu_hardware.reg_ddr_phy_offset[1] = REG_DDR_PHY1_OFFSET;
    ljm_gpu_hardware.reg_misc_offset = REG_MISC_OFFSET;
    ljm_gpu_hardware.reg_spi_offset = REG_SPI_OFFSET;
    ljm_gpu_hardware.reg_cru_offset = REG_CRU_OFFSET;
    ljm_gpu_hardware.reg_noc_offset = REG_NOC_OFFSET;
    ljm_gpu_hardware.reg_iomux_offset = REG_IOMUX_OFFSET;

    gp202_pre_set_clock(mmio_base);   /* set default clock */

    if (flags & LJMICRO_IS_FPGA) {  /* FPGA board */
        if (pddr_size)
            *pddr_size = (flags & LJMICRO_VRAM_2G) ? 0x80000000UL : 0x40000000UL;
    }
    else {  /* GP202 chip */
        // ljm_mcu_pre_init(card, MCU_FLASH_OFFSET, flags);
        ljm_fan_pre_init(card, 30);/* set fan speed: 30% ,called here only once, then kernel adjust speed */
        if (RREG32_BASE(mmio_base, ljm_gpu_hardware.reg_cru_offset + CRU_SOFTRST0) & m_CFG_RESET_DDRC) {
            ljm_ddr_init(card, mmio_base, pddr_size, flags);
            gp202_ddrv2_noc_init(mmio_base, ddrChannelMask);

            /* sleep more than 120 millioseconds to ensure ddr port work */
            /* FIXME: wait until ddr read/write check OK, should be put in noc */
            msleep(150);
        }
        else {
            ddrChannelMask = gp202_ddrv2_get_channel_mask(0);
            if (pddr_size)
                *pddr_size = ljm_ddrv2_get_ddr_size(card, mmio_base);
        }

        //ljm_bypass_unused_pll(mmio_base);
        //ljm_ddrv2_set_urgent_qos(mmio_base, 0x3);
        if (ddrChannelMask == 0x1 || ddrChannelMask == 0x2) { /* only a ddr channel */
            /* set CRTC port bus best qos and urgent */
            WREG32_BASE(mmio_base, REG_MONITOR_OFFSET + REG_MONITOR_QOS2, 0xffffff00);
            WREG32_BASE(mmio_base, REG_MONITOR_OFFSET + REG_MONITOR_URGENT1, 0x3fff);
        }
    }

    FOOTER();
    return 0;
}

uint32_t ljm_gp202_get_freq(int card, int type);

/*****************************************************************************
*  函 数 名: gp200_init
*  功能描述: gp200初始化
*  输入参数: struct ljm_device *ljm_dev: gp设备数据指针
*  输出参数: 无
*  返 回 值: int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: huym
*            修改内容: 新生成函数
*
*****************************************************************************/
int gp202_init(struct ljm_device *ljm_dev)
{
    struct ljm_drm_info *info = ljm_dev->drm_info;
    struct ljm_port *port;
    int i, num = 0;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    ljm_dev->num_crtcs = GP202_MAX_CRTCS;
    ljm_dev->num_planes = GP202_MAX_PLANES;
    ljm_vgav2_init(ljm_dev->cur_card, ljm_dev->num_crtcs);
    ljm_pwm_fan_init(ljm_dev);
    ljm_pm_init(ljm_dev);
    /* enable DVO0, crtc0 vsync_start lock, enable double line buffer */
    CR_WR(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2,
            ( m_CRTC_CRT_MODE_CTRL2_DVO0
            | m_CRTC_CRT_MODE_CTRL2_ADDRLATCH
            | m_CRTC_CRT_MODE_CTRL2_2BUFFER));
    REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL2,
                ( m_CRTC1_DISP_CTRL2_DVO0
                | m_CRTC1_DISP_CTRL2_ADDRLATCH
                | m_CRTC1_DISP_CTRL2_2BUFFER));
    /*防止gp202在vbios使用双buffer时出现显示异常*/
    REG_W8(ljm_dev->cur_card, CRTC_FRAME_BUFF_SEL, 0x0);

    if (ljm_dev->flags & LJMICRO_IS_FPGA) {
        int i2c_index = 0;

        ljm_dev->i2c_dev_cnt = 2;
        ljm_dev->i2c_devs[0] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_dw_init(ljm_dev->i2c_devs[0], ljm_dev->mmio_base, REG_I2C_OFFSET + REG_I2C_SIZE * 0, 0);
        ljm_dev->i2c_devs[1] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
        ljm_i2c_dw_init(ljm_dev->i2c_devs[1], ljm_dev->mmio_base, REG_I2C_OFFSET + REG_I2C_SIZE * 1, 1);

        for(i=0; i<LJM_MAX_PORTS; i++) {
            if(info->ports[i].type != LJM_PORT_DVO)
                continue;

            port = kzalloc(sizeof(struct ljm_port), GFP_KERNEL);
            port->type = info->ports[i].type;
            port->vdev.dvo.i2c_dev = ljm_dev->i2c_devs[i2c_index];
            port->id = info->ports[i].index;
            port->info = &info->ports[i];
            ljm_dev->ports[num] = port;
            ljm_dev->num_ports = ++num;

            if (i2c_index < ljm_dev->i2c_dev_cnt - 1)
                i2c_index++;
        }
    }
    else {
        /* only a DAC in CR register space, need not to initialize */

        /* i2c init */
        ljm_dev->i2c_dev_cnt = 4;
        for (i=0; i<ljm_dev->i2c_dev_cnt; i++) {
            ljm_dev->i2c_devs[i] = kzalloc(sizeof(struct i2c_dev), GFP_KERNEL);
            ljm_i2c_dw_init(ljm_dev->i2c_devs[i], ljm_dev->mmio_base, REG_I2C_OFFSET + REG_I2C_SIZE * i, i);
        }

        /*
         * VGA[0]   ==>  i2c_devs[0]
         *
         * HDMI[0]  ==>  REG_HDMI_OFFSET + REG_HDMI_SIZE * 0
         * HDMI[1]  ==>  REG_HDMI_OFFSET + REG_HDMI_SIZE * 1
         *
         * DVO[0]   ==>  i2c_devs[1]
         * DVO[1]   ==>  i2c_devs[2]
         *
         * DP[0]    ==>  REG_DP_OFFSET + REG_DP_SIZE * 0
         */
        for(i=0; i<LJM_MAX_PORTS; i++)
        {
            if(info->ports[i].type <= 0)
                continue;

#if !USE_DP_CONNECTOR
            if (info->ports[i].type == LJM_PORT_DP)
                continue;
#endif

            port = kzalloc(sizeof(struct ljm_port), GFP_KERNEL);
            port->type = info->ports[i].type;

            switch (info->ports[i].type)
            {
                case LJM_PORT_VGA:
                    port->vdev.vga.i2c_dev = ljm_dev->i2c_devs[info->ports[i].sub_id];
                    ljm_dac_ch_gain_set(&port->vdev.vga, 0x06, 0x06, 0x06);
                    break;
                case LJM_PORT_HDMI:
                    ljm_hdmiv2_init(&port->vdev.hdmi, ljm_dev->mmio_base, REG_HDMI_OFFSET + REG_HDMI_SIZE * info->ports[i].sub_id, &info->ports[i]);
                    break;
                case LJM_PORT_DVO:
                    port->vdev.dvo.i2c_dev = ljm_dev->i2c_devs[info->ports[i].sub_id + 2];
                    break;
#if USE_DP_CONNECTOR
                case LJM_PORT_DP:
                    port->vdev.dp.ddev = ljm_dev->ddev;
                    ljm_dp_init(&port->vdev.dp, ljm_dev->mmio_base, REG_DP_OFFSET + REG_DP_SIZE * info->ports[i].sub_id,  &info->ports[i]);
                    break;
#endif
                default:
                    break;
            }

            port->id = info->ports[i].index;
            port->info = &info->ports[i];
            ljm_dev->ports[num] = port;
            ljm_dev->num_ports = ++num;
        }

        /* enable bandwidth error interrupt */
        bweIntEnable = 1;
    }

    /* backlight */
    //ljm_register_backlight_device(ljm_dev, 0);
    //ljm_dev->backlight_dev_cnt = 1;

#if USE_CLOCK_REGISTER
    /* gp200 clk register */
    ljm_clk_init(ljm_dev);
#endif

    /* I2s init */
    ljm_dev->axi_bus_clk = ljm_gp202_get_freq(ljm_dev->cur_card, freq_axi);
    for (i = 0; i < LJM_MAX_I2SS; i++)
    {
        ljm_dev->i2s_devs[i] = devm_kzalloc(ljm_dev->dev, sizeof(struct ljm_i2s_dev), GFP_KERNEL);
        if (!ljm_dev->i2s_devs[i]) {
            return -ENOMEM;
        }

        ljm_dev->i2s_devs[i]->i2s_base = ljm_dev->mmio_base + REG_I2S_OFFSET + REG_I2S_SIZE * i;
        if (IS_ERR(ljm_dev->i2s_devs[i]->i2s_base)) {
            return PTR_ERR(ljm_dev->i2s_devs[i]->i2s_base);
        }
    }

    FOOTER();
    return 0;
}

int gp202_irq_process(struct ljm_device *ljm_dev)
{
    irqreturn_t r = IRQ_NONE;
    int i;
    bool queue_hotplug = false;
    u32 misc_intr_src;
    u32 misc_intr_mask;
    u32 misc_intr_vgapd;
    u32 i2s_intr_mask_reg;
    struct ljm_crtc *ljm_crtc = NULL;

    misc_intr_src = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_SRC);
    misc_intr_mask = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN);

    if ((misc_intr_src & m_INTR_SRC_VGA) && (misc_intr_mask & m_INTR_EN_VGA)) {   /* DC interrupt */
        //DRM_DEBUG_KMS("misc_intr_src=0x%x misc_intr_mask=0x%x\n", misc_intr_src, misc_intr_mask);
        r = IRQ_HANDLED;
    } else if ((misc_intr_src & misc_intr_mask) == 0x0) {   /* extra duplicated interrupt, which has been handled off */
        /* GP200内部中断源到PCIE中断有屏蔽位进行控制，如果中断源与中断屏蔽寄存器与运算后为0，又由于GP200采用MSI中断模式，
         * OS理应收不到GP200的中断。但是GP200中断采取重复发送模式，避免中断丢失，可能存在中断正在被处理时，
         * PCIE刚好又重复发出该中断事件（在途中断），对于此类中断，服务例程应该返回OS为IRQ_HANDLED，标示中断已被处理。
         */
        r = IRQ_HANDLED;
        /* FIXME: 频繁发生，是否正常？ */
        //printk(KERN_INFO "ignore duplicated interrupt\n");
        //DRM_DEBUG_KMS("ignore duplicated interrupt\n");
    }

    /* check DC (display controller) interrupt */
    if (misc_intr_src & m_INTR_SRC_VGA) {   /* meet with DC intterupt */
        misc_intr_vgapd = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_VGAPD);

        if (ljm_dev->mode_config_initialized && !(ljm_dev->flags & LJMICRO_IS_FPGA)) {
            /* check hdmi interrupt */
            for (i=0; i<ljm_dev->num_ports; i++) {
                if(ljm_dev->ports[i]->type == LJM_PORT_HDMI)
                {
                    if (misc_intr_vgapd & (m_INTR_VGAPD_HDMI0 << ljm_dev->ports[i]->vdev.hdmi.index)) {
                        if (ljm_hdmiv2_irq_process(&ljm_dev->ports[i]->vdev.hdmi) & LJM_IRQ_TYPE_HPD) {
                            queue_hotplug = true;
                            //TRACE("HDMI%d got hpd interrupt\n", ljm_dev->ports[i]->vdev.hdmi.index);
                            printk("HDMI%d got hpd interrupt\n", ljm_dev->ports[i]->vdev.hdmi.index);
                        } else {
                            ERR_PRINT("HDMI%d: mismatch with misc_intr\n", ljm_dev->ports[i]->vdev.hdmi.index);
                        }
                    }
                }
#if USE_DP_CONNECTOR
                if(ljm_dev->ports[i]->type == LJM_PORT_DP) {
                    if (misc_intr_vgapd & m_INTR_VGAPD_DP0_CTRL) {
                        if (ljm_dp_dc_irq_process(&ljm_dev->ports[i]->vdev.dp) & LJM_IRQ_TYPE_HPD) {
                            queue_hotplug = true;
                            printk("DP%d got hpd interrupt\n", ljm_dev->ports[i]->vdev.dp.index);
                        } else {
                            DEBUG_PRINT("DP%d: mismatch with misc_intr\n", ljm_dev->ports[i]->vdev.dp.index);
                        }
                    }
                    else if (misc_intr_vgapd & m_INTR_VGAPD_DP0_PHY)
                    {
                        ljm_dp_phy_irq_process(&ljm_dev->ports[i]->vdev.dp);
                        printk("DP%d PHY: got hpd interrupt\n", ljm_dev->ports[i]->vdev.dp.index);
                    }
                }
#endif
            }

#if USE_DAC_CONNECTOR
            /* check DAC interrupt */
            if (misc_intr_vgapd & m_INTR_VGAPD_DAC0) {
                if (ljm_dac_irq_process(ljm_dev->cur_card, 0) & LJM_IRQ_TYPE_HPD) {
                    queue_hotplug = true;
                    //TRACE("DAC0 got hpd interrupt\n");
                    printk("DAC0 got hpd interrupt\n");
                } else {
                    ERR_PRINT("DAC0: mismatch with misc_intr\n");
                }
            }
#endif
        }

        /* check crtc interrupt */
        for (i=0; i<GP202_MAX_CRTCS; i++) {
            /* check DVO HPD */
            if (misc_intr_vgapd & (m_INTR_VGAPD_CRTC0_HPD >> (i * 3))) {
                if (i == 0) {
                    CR_WR(ljm_dev->cur_card, CRTC_HPD_ST_ADR, 0x02);    // clear hpd interrupt
                } else {
                    REG_W8(ljm_dev->cur_card, CRTC1_HPD_ST_ADR + CRTCx_REG_SIZE * (i - 1), 0x02);   // clear hpd interrupt
                }

                queue_hotplug = true;
                //TRACE("DVO%d got hpd interrupt\n", i);
                printk("DVO%d got hpd interrupt\n", i);
            }

            /* check vblank interrupt */
            if (misc_intr_vgapd & (m_INTR_VGAPD_CRTC0_VS >> (i * 3))) {
                if (i == 0) {
                    CR_WR(ljm_dev->cur_card, CRTC_STATUS_ADR, 0x08);    // clear vsync interrupt
                } else {
                    REG_W8(ljm_dev->cur_card, CRTC1_HV_ST_ADR + CRTCx_REG_SIZE * (i - 1), 0x08);    // clear vsync interrupt
                }
                /* 若有挂起的lut更新，则执行。 */
                ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[i]);
                _gp202_crtc_load_lut(ljm_crtc);
#if USE_KMS_VBLANK
                if (ljm_dev->mode_config_initialized) {
                    /* handle vblank */
                    if (ljm_dev->irq.crtc_vblank_int[i]) {
                        drm_handle_vblank(ljm_dev->ddev, i);
                        //ljm_dev->pm.vblank_sync = true;    /* FIXME: pm.vblank_syn */
                        wake_up(&ljm_dev->irq.vblank_queue);
                    }
                    if (atomic_read(&ljm_dev->irq.pflip[i]))
                        ljm_crtc_handle_vblank(ljm_dev, i);
                    DRM_DEBUG_VBL("crtc %d vblank\n", i);
                }
#endif
            }

            /* check CRTC bandwidth error interrupt */
            if (misc_intr_vgapd & (m_INTR_VGAPD_CRTC0_BW >> (i * 3))) {
                if (i == 0) {
                    CR_WR(ljm_dev->cur_card, CRTC_BW_ERR_ST_ADR, 0x02); // clear bandwidth error interrupt
                } else {
                    REG_W8(ljm_dev->cur_card, CRTC1_BW_ERR_ST_ADR + CRTCx_REG_SIZE * (i - 1), 0x02);    // clear bandwidth error interrupt
                }
#if defined(_DEBUG)
                //WARN_PRINT("CRTC%d got bandwidth error interrupt\n", i);
#endif
                bweIntCount++;
            }
        }

        /* check layer bandwidth error interrupt */
        for (i=0; i<GP202_MAX_PLANES; i++) {
            if (misc_intr_vgapd & (m_INTR_VGAPD_L1_BW >> i)) {
                if (REG_R8(ljm_dev->cur_card, LAYER1_BW_ERR_ST_ADDR + LAYER_REG_SIZE * i) & 0x2) {
#if defined(_DEBUG)
                    //WARN_PRINT("LAYER%d got bandwidth error interrupt\n", i);
#endif
                    bweIntCount++;
                }
                else {
                    ERR_PRINT("LAYER%d: mismatch with misc_intr\n", i);
                }

                REG_W8(ljm_dev->cur_card, LAYER1_BW_ERR_ST_ADDR + LAYER_REG_SIZE * i, 0x02);  // clear bandwidth error interrupt
            }
        }

        /* check zoom video interrupt */
        for (i=0; i<GP202_MAX_ZOOMVIDEO; i++) {
            if (misc_intr_vgapd & (m_INTR_VGAPD_ZV0 >> i)) {
                if (REG_R8(ljm_dev->cur_card, ZV0_INT_ST_ADDR + ZV_REG_SIZE * i) & 0x1) {
                    //TRACE( "ZV%d got frame interrupt\n", i);
                    ljm_zoomvideo_irq_handle(i);
                }
                else
                    ERR_PRINT("ZV%d: mismatch with misc_intr\n", i);

                REG_W8(ljm_dev->cur_card, ZV0_INT_ST_ADDR + ZV_REG_SIZE * i, 0x01);   // clear zoom video frame interrupt
            }
        }

        /* check i2s interrupt */
        if (misc_intr_vgapd & m_INTR_VGAPD_I2S0)
        {
            printk("I2S0 get interrupt, now disable the interrupt\n");
            for (i = 0; i < 4; i++)  /* I2S has 4 stereo channels*/
            {
                i2s_intr_mask_reg = REG_R32(ljm_dev->cur_card, REG_I2S_OFFSET + IMR(i));
                REG_W32(ljm_dev->cur_card, REG_I2S_OFFSET + IMR(i), i2s_intr_mask_reg | 0x33);
            }
        }
        if (misc_intr_vgapd & m_INTR_VGAPD_I2S1)
        {
            printk("I2S1 get interrupt, now disable the interrupt\n");
            for (i = 0; i < 4; i++)  /* I2S has 4 stereo channels*/
            {
                i2s_intr_mask_reg = REG_R32(ljm_dev->cur_card, REG_I2S_OFFSET + REG_I2S_SIZE + IMR(i));
                REG_W32(ljm_dev->cur_card, REG_I2S_OFFSET + REG_I2S_SIZE + IMR(i), i2s_intr_mask_reg | 0x33);
            }
        }

    }

    if (queue_hotplug)
        schedule_delayed_work(&ljm_dev->hotplug_work, msecs_to_jiffies(500));

    return r;
}


inline uint64_t calc_pllout(uint32_t pll_val, uint32_t divqx)
{
    int i, x;
    uint32_t DIVQ = (pll_val & 0x70) >> 4;;

    for (i=0, x=1; i<DIVQ; i++)
        x *= 2;

    return (27ULL * 2 * (((pll_val & 0x1ff0000) >> 16) + 1) + (((pll_val & 0x1f00) >> 8) + 1) * x * (divqx + 1) / 2)
        / ((((pll_val & 0x1f00) >> 8) + 1) * x * (divqx + 1));
}

uint32_t ljm_gp202_get_freq(int card, int type)
{
    uint32_t value;
    uint32_t DIVQ_CKDIV; /*从CRU_CKDIV读出来的值*/
    uint32_t DIVQx;      /*将DIVQ分解成AXI、ahb和apb三部分*/

    switch (type) {
        case freq_axi:
        case freq_ahb:
        case freq_apb:
            value = LJM_CRU_R32(card, CRU_CKMUX);
            if (!(value & m_CFG_BUS_CKSEL))
                return 27;  /* HOSC 27MHz */

            value = LJM_CRU_R32(card, CRU_DEVPLL);
            DIVQ_CKDIV = LJM_CRU_R32(card, CRU_CKDIV);
            if (type == freq_axi)
                DIVQx = (DIVQ_CKDIV & 0xf) >> 0;
            else if (type == freq_ahb)
                DIVQx = (DIVQ_CKDIV & 0xf0) >> 4;
            else
                DIVQx = (DIVQ_CKDIV & 0xf00) >> 8;
            return calc_pllout(value, DIVQx);
        case freq_pcie:
            value = LJM_CRU_R32(card, CRU_CKMUX);
            if (!(value & m_CFG_BUS_CKSEL))
                return 27;  /* HOSC 27MHz */

            value      = LJM_CRU_R32(card, CRU_DEV2PLL);
            DIVQ_CKDIV = LJM_CRU_R32(card, CRU_CKDIV);
            DIVQx      = (DIVQ_CKDIV & 0xf0000) >> 16;
            return calc_pllout(value, DIVQx);
        case freq_vpu:
            value = LJM_CRU_R32(card, CRU_CKMUX);
            if (!(value & m_CFG_BUS_CKSEL))
                return 27;  /* HOSC 27MHz */

            value      = LJM_CRU_R32(card, CRU_DEV2PLL);
            DIVQ_CKDIV = LJM_CRU_R32(card, CRU_CKDIV);
            DIVQx      = (DIVQ_CKDIV & 0xf00000) >> 20;
            return calc_pllout(value, DIVQx);
        case freq_3d:
            value = LJM_CRU_R32(card, CRU_CKMUX);
            if (!(value & 0x100)) {
                value = LJM_CRU_R32(card, CRU_GPUPLL);
                if (value & m_CFG_PLL_BP)
                    return 27;  /* HOSC 27MHz */

                return calc_pllout(value, 0);
            }
            else {
                value = LJM_CRU_R32(card, CRU_GPU2PLL);
                if (value & m_CFG_PLL_BP)
                    return 27;  /* HOSC 27MHz */

                return calc_pllout(value, 0);
            }
        case freq_2d:
            value = LJM_CRU_R32(card, CRU_CKMUX);
            if (!(value & 0x400)) {
                value = LJM_CRU_R32(card, CRU_DEV2PLL);
                if (value & m_CFG_PLL_BP)
                    return 27;  /* HOSC 27MHz */

                DIVQ_CKDIV = LJM_CRU_R32(card, CRU_CKDIV);
                DIVQx      = (DIVQ_CKDIV & 0xf0000000) >> 28;
                return calc_pllout(value, DIVQx);
            }
            else {
                value = LJM_CRU_R32(card, CRU_GPU2PLL);
                if (value & m_CFG_PLL_BP)
                    return 27;  /* HOSC 27MHz */

                return calc_pllout(value, 0);
            }
        case freq_ddr:
            value = LJM_CRU_R32(card, CRU_DDRPLL);
            if ((value & (m_CFG_PLL_BP | m_CFG_DDRPLL_BP1)) == (m_CFG_PLL_BP | m_CFG_DDRPLL_BP1))
                return 27;  /* HOSC 27MHz */

            return calc_pllout(value, 0);
        default:
            printk(KERN_INFO "%s: command error\n", __FUNCTION__);
            return -EPERM;
    }
}

/**
 * gp202_encoder_mode_fixup - mode fixup for GP202 chip
 *
 * @ljm_dev: pointer to ljm_device
 * @mode: pointer of source mode
 * @adjustedMode: pointer of adjusted mode, would be used in mode_set of crtc object
 *
 * mode fixup for some chips or monitors
 *
 * Returns:
 * True if the mode was set successfully, false otherwise.
 */
bool gp202_encoder_mode_fixup(struct drm_encoder *encoder, const struct drm_display_mode *mode, struct drm_display_mode *adjustedMode)
{
    return true;
}

uint32_t gp202_pvt_get_temp(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if ((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_TEMP) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_TEMP);
        msleep(1100);
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp202_pvt_get_volt(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if ((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_VOLT) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_VOLT);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp202_pvt_get_svt_process(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if ((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_SVT_PROCESS) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_SVT_PROCESS);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp202_pvt_get_lvt_process(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if ((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_LVT_PROCESS) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_LVT_PROCESS);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    }

    return tmp & m_PVT_DATA_MASK;
}

uint32_t gp202_pvt_get_ulvt_process(int card, int index)
{
    uint32_t tmp;

    tmp = LJM_PVT_R32(card, PVT_SAMP_CTL);
    if ((tmp & m_SAMP_CTL_MODE_MASK) == v_SAMP_CTL_MODE_ULVT_PROCESS) {
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    } else {
        LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_ULVT_PROCESS);
        msleep(1100); /* sleep 1.1s to get valid data */
        tmp = LJM_PVT_R32(card, PVT_DATA);
        tmp = (tmp >> (16 * index));
    }

    return tmp & m_PVT_DATA_MASK;
}

void gp202_crtc_gamma_enable(struct ljm_crtc *ljm_crtc)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ljm_crtc->base.dev);

    if (ljm_crtc->id == 0)
        CR_WR(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2, (CR_RD(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2) | m_CRTC_CRT_MODE_CTRL2_GAMMA));
    else if (ljm_crtc->id == 1)
        REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL2, (REG_R8(ljm_dev->cur_card, CRTC1_DISP_CTRL2) | m_CRTC1_DISP_CTRL2_GAMMA));

    return;
}

void gp202_crtc_gamma_disable(struct ljm_crtc *ljm_crtc)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ljm_crtc->base.dev);

    if (ljm_crtc->id == 0)
        CR_WR(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2, (CR_RD(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2) & ~m_CRTC_CRT_MODE_CTRL2_GAMMA));
    else if (ljm_crtc->id == 1)
        REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL2, (REG_R8(ljm_dev->cur_card, CRTC1_DISP_CTRL2) & ~m_CRTC1_DISP_CTRL2_GAMMA));

    return;
}

bool gp202_crtc_gamma_en_check(struct ljm_crtc *ljm_crtc)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ljm_crtc->base.dev);

    if (ljm_crtc->id == 0) {
        if (CR_RD(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2) & m_CRTC_CRT_MODE_CTRL2_GAMMA) {
            return true;
        }
        else {
            return false;
        }
    }
    else if (ljm_crtc->id == 1) {
        if (REG_R8(ljm_dev->cur_card, CRTC1_DISP_CTRL2) & m_CRTC1_DISP_CTRL2_GAMMA) {
            return true;
        }
        else {
            return false;
        }
    }
    else {
        printk(KERN_ERR "%s: error crtc num %d, exceed GP202 max crtc num.\n", __FUNCTION__, ljm_crtc->id);
        return false;
    }
}

/**
 * gp202_crtc_load_lut - 颜色查找表更新函数
 * 
 * GP202在更新相关寄存器时，若CRTC同时正在取数据，可能导致显示异常。
 * 该函数只保存颜色查找表，将更新需求挂起，等待vblank中断上半区处理。
 * 
 * @ljm_crtc: 更新lut的crtc设备指针；
 * @red: 颜色查找表红色分量；
 * @green: 颜色查找表红色分量；
 * @blue: 颜色查找表红色分量；
 * @gamma_size: 颜色查找表各色彩分量长度。
 * 
 * 返回值：
 * 成功返回0，失败返回错误码。
*/
int gp202_crtc_load_lut(struct ljm_crtc *ljm_crtc, u16 *red, u16 *green, u16 *blue, uint32_t gamma_size)
{
    struct ljm_lut tmp;

    if(gamma_size != ljm_crtc->base.gamma_size)
        return -EINVAL;

    tmp = ljm_crtc->lut;

    memcpy(tmp.red, red, gamma_size * sizeof(u16));
    memcpy(tmp.green, green, gamma_size * sizeof(u16));
    memcpy(tmp.blue, blue, gamma_size * sizeof(u16));
    tmp.gamma_size = gamma_size;
    tmp.lut_changed = true;

    ljm_crtc->lut = tmp;
    DRM_DEBUG_KMS("ljm_crtc%d gamma set pending.\n", ljm_crtc->id);
    return 0;
}

int _gp202_crtc_load_lut(struct ljm_crtc *ljm_crtc)
{
    int i;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ljm_crtc->base.dev);
    u16 *red, *green, *blue;
    uint32_t gamma_size;

    if(!ljm_crtc->lut.lut_changed)
        return 0;

    red = ljm_crtc->lut.red;
    green = ljm_crtc->lut.green;
    blue = ljm_crtc->lut.blue;
    gamma_size = ljm_crtc->lut.gamma_size;

    for(i=0; i<gamma_size; i++)
    {
        REG_W32(ljm_dev->cur_card, CRTC_GAMMA_LUT_RWCTL, 0x1); //禁止读lut，允许写lut，每次写值前都要更新读写控制位。
        REG_W16(ljm_dev->cur_card, CRTCX_GAMMA_LUTR_ADDR(ljm_crtc->id), i);
        REG_W16(ljm_dev->cur_card, CRTCX_GAMMA_LUTR_DATA(ljm_crtc->id), red[i]);

        REG_W32(ljm_dev->cur_card, CRTC_GAMMA_LUT_RWCTL, 0x1);
        REG_W16(ljm_dev->cur_card, CRTCX_GAMMA_LUTG_ADDR(ljm_crtc->id), i);
        REG_W16(ljm_dev->cur_card, CRTCX_GAMMA_LUTG_DATA(ljm_crtc->id), green[i]);

        REG_W32(ljm_dev->cur_card, CRTC_GAMMA_LUT_RWCTL, 0x1);
        REG_W16(ljm_dev->cur_card, CRTCX_GAMMA_LUTB_ADDR(ljm_crtc->id), i);
        REG_W16(ljm_dev->cur_card, CRTCX_GAMMA_LUTB_DATA(ljm_crtc->id), blue[i]);
    }

    if (!gp202_crtc_gamma_en_check(ljm_crtc))
    {
        gp202_crtc_gamma_enable(ljm_crtc);
    }

    ljm_crtc->lut.lut_changed = false;
    DRM_DEBUG_KMS("ljm_crtc%d gamma set.\n", ljm_crtc->id);

    return 0;
}

int gp202_resume(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    int i;
    struct ljm_port *ljm_port;

    HEADER_ARG("drm=%p", ddev);

    DEBUG_PRINT("gp200_resume...");

    /* vga init */
    ljm_vgav2_init(ljm_dev->cur_card, ljm_dev->num_crtcs);
    /* enable DVO0, crtc0 vsync_start lock, enable double line buffer */
    CR_WR(ljm_dev->cur_card, CRTC_CRT_MODE_CTRL2,
            ( m_CRTC_CRT_MODE_CTRL2_DVO0
            | m_CRTC_CRT_MODE_CTRL2_ADDRLATCH
            | m_CRTC_CRT_MODE_CTRL2_2BUFFER));
    REG_W8(ljm_dev->cur_card, CRTC1_DISP_CTRL2,
                ( m_CRTC1_DISP_CTRL2_DVO0
                | m_CRTC1_DISP_CTRL2_ADDRLATCH
                | m_CRTC1_DISP_CTRL2_2BUFFER));

#if USE_DP_CONNECTOR
    ljm_resume_dp(ddev);
#endif

    /* HDMI irq set */
    for (i=0; i<ljm_dev->num_ports; i++) {
        ljm_port = ljm_dev->ports[i];
        if(ljm_port->type == LJM_PORT_HDMI)
            ljm_hdmiv2_irq_init(&ljm_port->vdev.hdmi);
    }
    /* restore register MISC_INTR_ENPCIE */
    LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, ljm_dev->misc_intr_maskpcie);

    FOOTER();
    return 0;
}

int gp202_power_set_state(struct ljm_device *ljm_dev, int power_state)
{
    uint32_t gp202_gpupll_val, gp202_dev2pll_val, gp202_gpu2pll_val, ckmux_val;

    int r = 0;
    gp202_gpupll_val = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL);
    gp202_dev2pll_val = LJM_CRU_R32(ljm_dev->cur_card, CRU_DEV2PLL);
    gp202_gpu2pll_val = LJM_CRU_R32(ljm_dev->cur_card, CRU_GPU2PLL);
    ckmux_val = LJM_CRU_R32(ljm_dev->cur_card, CRU_CKMUX); /* 驱动运行中 2D 使用的时钟源可能是 DEV2PLL  */
    switch (power_state)
    {
        case LJM_POWER_STATE_PERFORMANCE:
            //printk("power performance...\n");
            if (!(gp202_gpupll_val & 0x80000000) || !(gp202_dev2pll_val & 0x80000000) || !(gp202_gpu2pll_val & 0x80000000)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL) != g_gpupll_val) //800MHZ
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_DEV2PLL) != g_dev2pll_val)) //999MHZ
            {
                if (ckmux_val & (0x1U << 10))
                {
                    /* bypass */
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, gp202_gpu2pll_val & (0x1U << 31));
                    /* write 2D clock */
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, g_gpu2pll_val);
                }
                else
                {
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_DEV2PLL, gp202_dev2pll_val & (0x1U << 31));
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_DEV2PLL, g_dev2pll_val);
                }
                /* write 3D clock */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, gp202_gpupll_val & (0x1U << 31));
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, g_gpupll_val);
            }

            /* disable power management for cores */
            gpu_config_power_management(-1, false);
            break;
        case LJM_POWER_STATE_SAVE:/* 省电模式 */
            //printk("power save...\n");
            if (!(gp202_gpupll_val & 0x80000000) || !(gp202_dev2pll_val & 0x80000000)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL) != PLL_FREQ_600)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_DEV2PLL) != PLL_FREQ_500))
            {
                if (ckmux_val & (0x1U << 10))
                {
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, gp202_gpu2pll_val & (0x1U << 31));
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, PLL_FREQ_500);//2d, 500MHZ
                }
                else
                {
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_DEV2PLL, gp202_dev2pll_val & (0x1U << 31));
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_DEV2PLL, PLL_FREQ_500);//2d, 500MHZ
                }
                /* write 3D, 2D clock */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, gp202_gpupll_val & (0x1U << 31));
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, PLL_FREQ_600);//3d, 600MHZ
            }
            /* enable power management for cores */
            gpu_config_power_management(-1, true);
            break;
        case LJM_POWER_STATE_STANDBY:/* 备用模式 */
            //printk("power standby...\n");

            if (!(gp202_gpupll_val & 0x80000000) || !(gp202_dev2pll_val & 0x80000000)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_GPUPLL) != PLL_FREQ_400)
                || (LJM_CRU_R32(ljm_dev->cur_card, CRU_DEV2PLL) !=  PLL_FREQ_200))
            {
                if (ckmux_val & (0x1U << 10))
                {
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, gp202_gpu2pll_val & (0x1U << 31));
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_GPU2PLL, PLL_FREQ_400);//2d, 400MHZ
                }
                else
                {
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_DEV2PLL, gp202_dev2pll_val & (0x1U << 31));
                    LJM_CRU_W32(ljm_dev->cur_card, CRU_DEV2PLL, PLL_FREQ_400);//2d, 400MHZ
                }
                /* write 3D, 2D clock */
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, gp202_gpupll_val & (0x1U << 31));
                LJM_CRU_W32(ljm_dev->cur_card, CRU_GPUPLL, PLL_FREQ_200);//3d, 200MHZ
            }
            /* enable power management for cores */
            gpu_config_power_management(-1, true);
            break;
        default:
            r = -EINVAL;
    }

    return r;
}