/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __GP202_H__
#define __GP202_H__

#include "gp200.h"

#define GP202_MAX_CRTCS     2 /* max crts */
#define GP202_MAX_PLANES    2 /* max layers */
#define GP202_MAX_ZOOMVIDEO 2 /* max zoom video */

/*******************************************************************************
 * Partion of visual video memory of GP202
 * |Region           |Size            |Remark                                    |
 * |VRAM_NORMAL      |bar0 - VRAM_FIX |used for 2D, 3D, framebuffer, etc.        |
 * |VRAM_FIX         |4MB (top addr)  |fix region for misc use, such as DMA table|
 *******************************************************************************/

/********************************************************************************
 * GP202 local address
 * PCIE SPACE                   AMBA Space
 * BAR          Size            Range                           Remark
 * Legacy IO    3B0~3BB,3C0~3DF 0x0000_0000~0x0000_03FF         VGA legacy registers
 * Legacy MEM   A0000~BFFFF     0x000A_0000~0x000B_FFFF         VGA legacy VRAM
 * ROM (BAR6)   512KB           0x0020_0000~0x0027_FFFF         Flash(ROM)
 * IO (BAR4)    256B            0x0010_0000~0x0010_00FF         SPI Flash Controller
 * MEM_IO(BAR2) 8MB             0x0100_0000~0x017F_FFFF         Registers
 * MEM (BAR0)   256MB           0x80_0000_0000~0x80_0FFF_FFFF   VRAM
 * MEM          32GB            0x80_0000_0000~0x87_FFFF_FFFF   VRAM
 * HOST_MEM     512GB           0x00_0000_0000~0x7F_FFFF_FFFF   Host data
 ********************************************************************************/

/*****************************************************************************
 * GP202 PCIE address mapping by host (CPU)
 * BAR          Size    Offset                  Remark
 * MEMIO(BAR2)  8MB
 *                      0x01_0000~01_0FFF       VGA Registers (32bit AHB, 4K)
 *                      0x03_0000~03_7FFF       SRAM0 (MCU) (32bit AHB, 32K)
 *                      0x03_8000~03_FFFF       SRAM1 (MCU) (32bit AHB, 32K)
 *                      0x04_0000~07_FFFF       3D GPU Registers (32bit AHB, 256K)
 *                      0x0C_0000~0F_FFFF       2D GPU Registers (32bit AHB, 256K)

 *                      0x10_0000~10_03FF       DDC0 (8bit APB, 1K)
 *                      0x10_0400~10_07FF       DDC1 (8bit APB, 1K)
 *                      0x10_0800~10_0BFF       DDC2 (8bit APB, 1K)
 *                      0x10_0C00~10_0FFF       DDC3 (8bit APB, 1K)
 *                      0x11_0000~11_FFFF       HDMI0 (8bit APB, 64K)
 *                      0x12_0000~12_FFFF       HDMI1 (8bit APB, 64K)

 *                      0x13_0000~13_7FFF       DDR4 Ctrl1 Registers (32bit APB, 32K)
 *                      0x18_8800~18_8BFF       I2S0 Registers (32bit APB, 1K)
 *                      0x18_8C00~18_8FFF       I2S1 Registers (32bit APB, 1K)
 *                      0x19_0000~19_3FFF       DMA Registers (32bit APB, 16K)
 *                      0x1C_0000~1F_FFFF       DDR4 PHY1 Registers (32bit APB, 256K)
 *                      0x20_0000~27_FFFF       PCIE Config (32bit APB, 512K)
 *                      0x28_0000~28_7FFF       DDR4 Ctrl0 Registers (32bit APB, 32K)
 *                      0x28_8000~28_83FF       Clock Reset Registers (32bit APB, 1K)
 *                      0x28_8400~28_87FF       Misc (gart_base) (32bit APB, 1K)
 *                      0x28_8800~28_8FFF       Monitor Registers (32bit APB, 4K)
 *                      0x28_9000~28_93FF       UART Registers (32bit APB, 1K)
 *                      0x28_A000~28_A3FF       IO Mux Registers (32bit APB, 1K)
 *                      0x28_AC00~28_AFFF       Flash (MEM_IO) (32bit APB, 1K)
 *                      0x29_0000~29_FFFF       VPU0 Registers (32bit APB, 64K)
 *                      0x2A_0000~2B_FFFF       DP0 (32bit APB, 128K)
 *                      0x2C_0000~2F_FFFF       DDR4 PHY0 Registers (32bit APB, 256K)
 *                      0x30_0000~37_FFFF       NOC Registers (64bit AXI, 512K)
 *                      0x40_0000~7F_FFFF       PCIE_DBI (32bit AXI, 4M)
 *****************************************************************************/

#define REG_I2C_OFFSET       0x100000 /* I2C (DW) for I2C 0~3 */
#define REG_I2C_SIZE         0x400    /* 1K */
#define REG_HDMI_OFFSET      0x110000 /* HDMI */
#define REG_HDMI_SIZE        0x10000  /* 64K */
#define REG_DDR_CTRL0_OFFSET 0x280000 /* register offset for DDR4 controller 0 */
#define REG_DDR_PHY0_OFFSET  0x2C0000 /* register offset for DDR4 PHY 0 */
#define REG_DDR_CTRL1_OFFSET 0x130000 /* register offset for DDR4 controller 1 */
#define REG_DDR_PHY1_OFFSET  0x1C0000 /* register offset for DDR4 PHY 1 */
#define REG_DP_OFFSET        0x2A0000 /* DP */
#define REG_DP_SIZE          0x20000  /* 128K */

#define REG_CRU_OFFSET      0x288000
#define REG_CRU_SIZE        0x400

/* #define CRU_PLLEN            0x0
 * #define m_CFG_PXLPLL0_BYPASS     (1U << 7)
 * #define m_CFG_PXLPLL1_BYPASS     (1U << 6)
 */
#define m_CFG_DEV2PLL_CKEN3      (1U << 3)
#define m_CFG_DEV2PLL_CKEN1      (1U << 2)
#define m_CFG_GPU2PLL_CKEN1      (1U << 1)
#define m_CFG_GPU2PLL_CKEN0      (1U << 0)

/* #define CRU_CKEN1            0x3c
 * #define m_CFG_QSPI_CKEN       (1U << 31)
 * #define m_CFG_DMA_CKEN        (1U << 29)
 * #define m_CFG_MCU_CKEN        (1U << 28)
 * #define m_CFG_I2S0_CKEN       (1U << 27)
 * #define m_CFG_I2S1_CKEN       (1U << 26)
 */
#define m_CFG_DDRC1_CORE_CKEN (1U << 9)
#define m_CFG_DDRC0_CORE_CKEN (1U << 8)
#define m_CFG_DDRC1_BUS_CKEN  (1U << 1)
#define m_CFG_DDRC0_BUS_CKEN  (1U << 0)

#define REG_MISC_OFFSET     0x288400
#define REG_MISC_SIZE       0x400

/**
 * Monitor
 */
#define REG_MONITOR_OFFSET  0x288800
#define REG_MONITOR_SIZE    0x800
#define REG_MONITOR_CTL         0x0
#define REG_MONITOR_TIMEOUT     0x4
#define REG_MONITOR_LOG_CYCLES  0x8
#define REG_MONITOR_QOS0        0xC
#define REG_MONITOR_QOS1        0x10
#define REG_MONITOR_QOS2        0x14
#define REG_MONITOR_QOS3        0x18
#define REG_MONITOR_URGENT0     0x1C
#define REG_MONITOR_URGENT1     0x20

/**
 * SPI
 */
#define REG_SPI_OFFSET      0x28AC00
#define REG_SPI_SIZE        0x400

/**
 * I2S
 */
#define REG_I2S_OFFSET      0x188800
#define REG_I2S_SIZE        0x400

/**
 * NOC
 */
#define REG_NOC_OFFSET      0x300000

#define NOC_REMAP_BLK0     0x30
#define NOC_REMAP_BLK_SIZE 0x400
#define NOC_REMAP_BLK_CNT  25
#define v_REMAP_DDR0       0x1    /* interleave size: 256bytes*/
#define v_REMAP_DDR1       0x2    /* interleave size: 256bytes*/
#define v_REMAP_DDR01      0x0    /* interleave size: 256bytes*/
#define v_REMAP_DDR01_128B      0x4    /* interleave size: 128bytes*/
#define v_REMAP_DDR01_512B      0x8    /* interleave size: 512bytes*/

/**
 * MCU
*/
#define MCU_FLASH_OFFSET 0x280000

/**
 * PVT
 */
#define m_SAMP_CTL_MODE_MASK     0xffffffff
#define v_SAMP_CTL_MODE_TEMP     0x01e201e2
#define v_SAMP_CTL_MODE_VOLT     0x01e601e6
#define v_SAMP_CTL_MODE_SVT_PROCESS    0x01fa01fa
#define v_SAMP_CTL_MODE_LVT_PROCESS    0x01ea01ea
#define v_SAMP_CTL_MODE_ULVT_PROCESS   0x01f201f2

/**
 * IO MUX
 */
#define REG_IOMUX_OFFSET    0x28A000


#endif /* ~__GP202_H__ */
