/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_h_
#define __lj_hal_h_

#include "lj_hal_types.h"
#include "lj_hal_enum.h"
#include "lj_hal_base.h"
#include "lj_hal_profiler.h"
#include "lj_hal_driver.h"
#if ljdENABLE_3D
#include "lj_hal_statistics.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct _ljkVIDMEM          *ljkVIDMEM;
typedef struct _ljkKERNEL          *ljkKERNEL;
typedef struct _ljkCOMMAND         *ljkCOMMAND;
typedef struct _ljkEVENT           *ljkEVENT;
typedef struct _ljkDB              *ljkDB;
typedef struct _ljkDVFS            *ljkDVFS;
typedef struct _ljkMMU             *ljkMMU;
typedef struct _ljsDEVICE          *ljkDEVICE;

/******************************************************************************
 ****************************** Alignment Macros ******************************
 ******************************************************************************/

/* Alignment with a non-power of two value. */
#define ljmALIGN_NP2(n, align) (((n) + (align) - 1) - (((n) + (align) - 1) % (align)))

#define ljmALIGN_NP2_SAFE(n, align)                                         \
(\
    (ljmALIGN_NP2((n) & ~0ULL, (align) & ~0ULL) ^ ljmALIGN_NP2(n, align)) ? \
        (n) : ljmALIGN_NP2(n, align)                                        \
)

/* Alignment with a power of two value. */
#define ljmALIGN(n, align) (((n) + ((align) - 1)) & ~((align) - 1))

#define ljmALIGN_SAFE(n, align)                                        \
(\
    (ljmALIGN((n) & ~0ULL, (align) & ~0ULL) ^ ljmALIGN(n, align)) ?    \
         (n) : ljmALIGN(n, align)                                      \
)

#define ljmALIGN_CHECK_OVERFLOW(n, align)                              \
(\
    (ljmALIGN((n) & ~0ULL, (align) & ~0ULL) ^ ljmALIGN(n, align)) ?    \
         ljvSTATUS_RESLUT_OVERFLOW : ljvSTATUS_OK                      \
)

#define ljmALIGN_BASE(n, align) \
(\
    ((n) & ~((align) - 1)) \
)

/******************************************************************************
 **************************** Element Count Macro *****************************
 ******************************************************************************/

#define ljmSIZEOF(a)    ((ljtSIZE_T)(sizeof(a)))

#define ljmCOUNTOF(a)   (sizeof(a) / sizeof(a[0]))

/******************************************************************************
 ******************************** Cast Macro **********************************
 ******************************************************************************/
#define ljmNAME_TO_PTR(na) \
    ljkKERNEL_QueryPointerFromName(kernel, ljmALL_TO_UINT32(na))

#define ljmPTR_TO_NAME(ptr) \
    ljkKERNEL_AllocateNameFromPointer(kernel, ptr)

#define ljmRELEASE_NAME(na) \
    ljkKERNEL_DeleteName(kernel, ljmALL_TO_UINT32(na))

#define ljmALL_TO_UINT32(t) ((ljtUINT32)(ljtUINTPTR_T)(t))

#define ljmPTR_TO_UINT64(p) ((ljtUINT64)(ljtUINTPTR_T)(p))

#define ljmUINT64_TO_PTR(u) ((ljtPOINTER)(ljtUINTPTR_T)(u))

#define ljmUINT64_TO_TYPE(u, t) ((t)(ljtUINTPTR_T)(u))

/******************************************************************************
 ******************************* Useful Macro *********************************
 ******************************************************************************/

#define ljvINVALID_ADDRESS              ~0ULL
#define ljvINVALID_VALUE                0xCCCCCCCC

#define ljvINVALID_PHYSICAL_ADDRESS     ~0ULL

#define ljmGET_PRE_ROTATION(rotate) \
    ((rotate) & (~(ljvSURF_POST_FLIP_X | ljvSURF_POST_FLIP_Y)))

#define ljmGET_POST_ROTATION(rotate) \
    ((rotate) & (ljvSURF_POST_FLIP_X | ljvSURF_POST_FLIP_Y))

typedef struct _ljkHARDWARE         *ljkHARDWARE;

#define ljdMAX_GPU_COUNT            ljvCORE_COUNT

#define ljdMAX_SURF_LAYERS          4

#define ljdMAX_DRAW_BUFFERS         16

#define ljdMAX_3DGPU_COUNT          8

#define ljdMAX_VERTEX_STREAM_COUNT  4
/*******************************************************************************
 **
 **  ljmVERIFY_OBJECT
 **
 **      Assert if an object is invalid or is not of the specified type.  If the
 **      object is invalid or not of the specified type, ljvSTATUS_INVALID_OBJECT
 **      will be returned from the current function.  In retail mode this macro
 **      does nothing.
 **
 **  ARGUMENTS:
 **
 **      obj     Object to test.
 **      t       Expected type of the object.
 */
#if ljmIS_DEBUG(ljdDEBUG_TRACE)
#define _ljmVERIFY_OBJECT(prefix, obj, t) \
    if ((obj) == ljvNULL) \
    { \
        prefix##TRACE(ljvLEVEL_ERROR, \
                      #prefix "VERIFY_OBJECT failed: NULL"); \
        prefix##TRACE(ljvLEVEL_ERROR, "  expected: %c%c%c%c", \
                      ljmCC_PRINT(t)); \
        prefix##ASSERT((obj) != ljvNULL); \
        prefix##FOOTER_ARG("status=%d", ljvSTATUS_INVALID_OBJECT); \
        return ljvSTATUS_INVALID_OBJECT; \
    } \
    else if (((ljsOBJECT*) (obj))->type != t) \
    { \
        prefix##TRACE(ljvLEVEL_ERROR, \
                      #prefix "VERIFY_OBJECT failed: %c%c%c%c", \
                      ljmCC_PRINT(((ljsOBJECT*) (obj))->type)); \
        prefix##TRACE(ljvLEVEL_ERROR, "  expected: %c%c%c%c", \
                      ljmCC_PRINT(t)); \
        prefix##ASSERT(((ljsOBJECT*)(obj))->type == t); \
        prefix##FOOTER_ARG("status=%d", ljvSTATUS_INVALID_OBJECT); \
        return ljvSTATUS_INVALID_OBJECT; \
    }

#define ljmVERIFY_OBJECT(obj, t)        _ljmVERIFY_OBJECT(ljm, obj, t)
#define ljmkVERIFY_OBJECT(obj, t)       _ljmVERIFY_OBJECT(ljmk, obj, t)
#else
#define ljmVERIFY_OBJECT(obj, t)        do {} while (ljvFALSE)
#define ljmkVERIFY_OBJECT(obj, t)       do {} while (ljvFALSE)
#endif

/******************************************************************************/
/*VERIFY_OBJECT if special return expected*/
/******************************************************************************/
#ifndef EGL_API_ANDROID
# define _ljmVERIFY_OBJECT_RETURN(prefix, obj, t, retVal)                    \
    do {                                                                     \
        if ((obj) == ljvNULL) {                                              \
            prefix##PRINT_VERSION();                                         \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "VERIFY_OBJECT_RETURN failed: NULL");      \
            prefix##TRACE(ljvLEVEL_ERROR, "  expected: %c%c%c%c", \
                          ljmCC_PRINT(t));                                   \
            prefix##ASSERT((obj) != ljvNULL);                                \
            prefix##FOOTER_ARG("retVal=%d", retVal);                         \
            return retVal;                                                   \
        } else if (((ljsOBJECT *)(obj))->type != t) {                        \
            prefix##PRINT_VERSION();                                         \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "VERIFY_OBJECT_RETURN failed: %c%c%c%c", \
                          ljmCC_PRINT(((ljsOBJECT *)(obj))->type));          \
            prefix##TRACE(ljvLEVEL_ERROR, "  expected: %c%c%c%c", \
                          ljmCC_PRINT(t));                                   \
            prefix##ASSERT(((ljsOBJECT *)(obj))->type == t);                 \
            prefix##FOOTER_ARG("retVal=%d", retVal);                         \
            return retVal;                                                   \
        }                                                                    \
    } while (ljvFALSE)
# define ljmVERIFY_OBJECT_RETURN(obj, t, retVal)  \
    _ljmVERIFY_OBJECT_RETURN(ljm, obj, t, retVal)
# define ljmkVERIFY_OBJECT_RETURN(obj, t, retVal) \
    _ljmVERIFY_OBJECT_RETURN(ljmk, obj, t, retVal)
#else
# define ljmVERIFY_OBJECT_RETURN(obj, t)      do {} while (ljvFALSE)
# define ljmkVERIFY_OBJECT_RETURN(obj, t)     do {} while (ljvFALSE)
#endif

/******************************************************************************
 ********************************* ljkOS Object *******************************
 ******************************************************************************/

/* Construct a new ljkOS object. */
ljeSTATUS
ljkOS_Construct(IN ljtPOINTER Context, OUT ljkOS *Os);

/* Destroy an ljkOS object. */
ljeSTATUS
ljkOS_Destroy(IN ljkOS Os);

/* Query the video memory. */
ljeSTATUS
ljkOS_QueryVideoMemory(IN ljkOS Os,
                       OUT ljtPHYS_ADDR *InternalAddress,
                       OUT ljtSIZE_T *InternalSize,
                       OUT ljtPHYS_ADDR *ExternalAddress,
                       OUT ljtSIZE_T *ExternalSize,
                       OUT ljtPHYS_ADDR *ContiguousAddress,
                       OUT ljtSIZE_T *ContiguousSize);

/* Allocate memory from the heap. */
ljeSTATUS
ljkOS_Allocate(IN ljkOS Os, IN ljtSIZE_T Bytes, OUT ljtPOINTER *Memory);

/* Free allocated memory. */
ljeSTATUS
ljkOS_Free(IN ljkOS Os, IN ljtPOINTER Memory);

/* Wrapper for allocation memory.. */
ljeSTATUS
ljkOS_AllocateMemory(IN ljkOS Os, IN ljtSIZE_T Bytes, OUT ljtPOINTER *Memory);

/* Wrapper for freeing memory. */
ljeSTATUS
ljkOS_FreeMemory(IN ljkOS Os, IN ljtPOINTER Memory);

/* Allocate paged memory. */
ljeSTATUS
ljkOS_AllocatePagedMemory(IN ljkOS Os,
                          IN ljkKERNEL Kernel,
                          IN ljtUINT32 Flag,
                          IN ljeVIDMEM_TYPE Type,
                          IN OUT ljtSIZE_T *Bytes,
                          OUT ljtUINT32 *Gid,
                          OUT ljtPHYS_ADDR *Physical);

/* Lock pages. */
ljeSTATUS
ljkOS_LockPages(IN ljkOS Os,
                IN ljtPHYS_ADDR Physical,
                IN ljtSIZE_T Bytes,
                IN ljtBOOL Cacheable,
                OUT ljtPOINTER *Logical);

/* Map pages. */
ljeSTATUS
ljkOS_MapPagesEx(IN ljkOS Os,
                 IN ljkKERNEL Kernel,
                 IN ljkMMU Mmu,
                 IN ljtPHYS_ADDR Physical,
                 IN ljtSIZE_T Offset,
                 IN ljtSIZE_T PageCount,
                 IN ljtADDRESS Address,
                 IN ljtPOINTER PageTable,
                 IN ljtBOOL Writable,
                 IN ljeVIDMEM_TYPE Type);

/* Map 1M pages. */
ljeSTATUS
ljkOS_Map1MPages(IN ljkOS Os,
                 IN ljkKERNEL Kernel,
                 IN ljkMMU Mmu,
                 IN ljtPHYS_ADDR Physical,
                 IN ljtSIZE_T PageCount,
                 IN ljtADDRESS Address,
                 IN ljtPOINTER PageTable,
                 IN ljtBOOL Writable,
                 IN ljeVIDMEM_TYPE Type);

ljeSTATUS
ljkOS_UnmapPages(IN ljkOS Os, IN ljtSIZE_T PageCount, IN ljtADDRESS Address);

/* Unlock pages. */
ljeSTATUS
ljkOS_UnlockPages(IN ljkOS Os, IN ljtPHYS_ADDR Physical,
                  IN ljtSIZE_T Bytes, IN ljtPOINTER Logical);

/* Free paged memory. */
ljeSTATUS
ljkOS_FreePagedMemory(IN ljkOS Os, IN ljtPHYS_ADDR Physical, IN ljtSIZE_T Bytes);

/* Allocate non-paged memory. */
ljeSTATUS
ljkOS_AllocateNonPagedMemory(IN ljkOS Os,
                             IN ljkKERNEL Kernel,
                             IN ljtBOOL InUserSpace,
                             IN ljtUINT32 Flag,
                             IN OUT ljtSIZE_T *Bytes,
                             OUT ljtPHYS_ADDR *Physical,
                             OUT ljtPOINTER *Logical);

/* Free non-paged memory. */
ljeSTATUS
ljkOS_FreeNonPagedMemory(IN ljkOS Os,
                         IN ljtPHYS_ADDR Physical,
                         IN ljtPOINTER Logical,
                         IN ljtSIZE_T Bytes);

/* Reserved memory. */
ljeSTATUS
ljkOS_RequestReservedMemory(ljkOS Os,
                            ljtPHYS_ADDR_T Start,
                            ljtSIZE_T Size,
                            const char *Name,
                            ljtBOOL Requested,
                            ljtPOINTER *MemoryHandle);

void
ljkOS_ReleaseReservedMemory(ljkOS Os, ljtPOINTER MemoryHandle);

/* Reserved memory sub area */
ljeSTATUS
ljkOS_RequestReservedMemoryArea(IN ljkOS Os,
                                IN ljtPOINTER MemoryHandle,
                                IN ljtSIZE_T Offset,
                                IN ljtSIZE_T Size,
                                OUT ljtPOINTER *MemoryAreaHandle);

void
ljkOS_ReleaseReservedMemoryArea(ljtPOINTER MemoryAreaHandle);

/* Get the number fo bytes per page. */
ljeSTATUS
ljkOS_GetPageSize(IN ljkOS Os, OUT ljtSIZE_T *PageSize);

/* Get the physical address of a corresponding logical address. */
ljeSTATUS
ljkOS_GetPhysicalAddress(IN ljkOS Os, IN ljtPOINTER Logical,
                         OUT ljtPHYS_ADDR_T *Address);

/* Get real physical address from handle. */
ljeSTATUS
ljkOS_GetPhysicalFromHandle(IN ljkOS Os,
                            IN ljtPHYS_ADDR Physical,
                            IN ljtSIZE_T Offset,
                            OUT ljtPHYS_ADDR_T *PhysicalAddress);

/* Get DMA address (IOVA) from handle. */
ljeSTATUS
ljkOS_GetDmaAddressFromHandle(IN ljkOS Os,
                            IN ljtPHYS_ADDR Physical,
                            IN ljtSIZE_T Offset,
                            OUT ljtPHYS_ADDR_T *DmaAddress);

/* Get the physical address of a corresponding user logical address. */
ljeSTATUS
ljkOS_UserLogicalToPhysical(IN ljkOS Os, IN ljtPOINTER Logical,
                            OUT ljtPHYS_ADDR_T *Address);

/* Map physical memory. */
ljeSTATUS
ljkOS_MapPhysical(IN ljkOS Os,
                  IN ljtPHYS_ADDR_T Physical,
                  IN ljtSIZE_T Bytes,
                  OUT ljtPOINTER *Logical);

/* Unmap previously mapped physical memory. */
ljeSTATUS
ljkOS_UnmapPhysical(IN ljkOS Os, IN ljtPOINTER Logical, IN ljtSIZE_T Bytes);

/* Read data from a hardware register. */
ljeSTATUS
ljkOS_ReadRegister(IN ljkOS Os, IN ljtUINT32 Address, OUT ljtUINT32 *Data);

/* Read data from a hardware register. */
ljeSTATUS
ljkOS_ReadRegisterEx(IN ljkOS Os,
                     IN ljkKERNEL Kernel,
                     IN ljtUINT32 Address,
                     OUT ljtUINT32 *Data);

/* Write data to a hardware register. */
ljeSTATUS
ljkOS_WriteRegister(IN ljkOS Os, IN ljtUINT32 Address, IN ljtUINT32 Data);

/* Write data to a hardware register. */
ljeSTATUS
ljkOS_WriteRegisterEx(IN ljkOS Os,
                      IN ljkKERNEL Kernel,
                      IN ljtUINT32 Address,
                      IN ljtUINT32 Data);

/* Write data to a hardware register without dump. */
ljeSTATUS
ljkOS_WriteRegisterEx_NoDump(IN ljkOS Os,
                             IN ljkKERNEL Kernel,
                             IN ljtUINT32 Address,
                             IN ljtUINT32 Data);

#ifdef __QNXNTO__
static ljmINLINE ljeSTATUS
ljkOS_WriteMemory(IN ljkOS Os, IN ljtPOINTER Address, IN ljtUINT32 Data)
{
    /* Write memory. */
    *(ljtUINT32 *)Address = Data;
    return ljvSTATUS_OK;
}

#else
/* Write data to a 32-bit memory location. */
ljeSTATUS
ljkOS_WriteMemory(IN ljkOS Os, IN ljtPOINTER Address, IN ljtUINT32 Data);
#endif

/* Map physical memory into the process space. */
ljeSTATUS
ljkOS_MapMemory(IN ljkOS Os, IN ljtPHYS_ADDR Physical,
                IN ljtSIZE_T Bytes, OUT ljtPOINTER *Logical);

/* Unmap physical memory from the specified process space. */
ljeSTATUS
ljkOS_UnmapMemoryEx(IN ljkOS Os,
                    IN ljtPHYS_ADDR Physical,
                    IN ljtSIZE_T Bytes,
                    IN ljtPOINTER Logical,
                    IN ljtUINT32 PID);

/* Unmap physical memory from the process space. */
ljeSTATUS
ljkOS_UnmapMemory(IN ljkOS Os, IN ljtPHYS_ADDR Physical,
                  IN ljtSIZE_T Bytes, IN ljtPOINTER Logical);

/* Delete a mutex. */
ljeSTATUS
ljkOS_DeleteMutex(IN ljkOS Os, IN ljtPOINTER Mutex);

/* Acquire a mutex. */
ljeSTATUS
ljkOS_AcquireMutex(IN ljkOS Os, IN ljtPOINTER Mutex, IN ljtUINT32 Timeout);

/* Release a mutex. */
ljeSTATUS
ljkOS_ReleaseMutex(IN ljkOS Os, IN ljtPOINTER Mutex);

/* Atomically exchange a pair of 32-bit values. */
ljeSTATUS
ljkOS_AtomicExchange(IN ljkOS Os,
                     IN OUT ljtUINT32_PTR Target,
                     IN ljtUINT32 NewValue,
                     OUT ljtUINT32_PTR OldValue);

/* Atomically exchange a pair of pointers. */
ljeSTATUS
ljkOS_AtomicExchangePtr(IN ljkOS Os,
                        IN OUT ljtPOINTER *Target,
                        IN ljtPOINTER NewValue,
                        OUT ljtPOINTER *OldValue);

ljeSTATUS
ljkOS_AtomSetMask(IN ljtPOINTER Atom, IN ljtUINT32 Mask);

ljeSTATUS
ljkOS_AtomClearMask(IN ljtPOINTER Atom, IN ljtUINT32 Mask);

ljeSTATUS
ljkOS_DumpCallStack(IN ljkOS Os);

ljeSTATUS
ljkOS_GetProcessNameByPid(IN ljtINT Pid, IN ljtSIZE_T Length, OUT ljtUINT8_PTR String);

ljeSTATUS
ljkOS_QueryCPUFrequency(IN ljkOS Os, IN ljtUINT32 CPUId, OUT ljtUINT32 *Frequency);

ljeSTATUS
ljkOS_TraceGpuMemory(IN ljkOS Os, IN ljtINT32 ProcessID, IN ljtINT64 Delta);

void
ljkOS_NodeIdAssign(ljkOS Os, ljuVIDMEM_NODE_PTR Node);

#if ljdENABLE_CLEAR_FENCE
ljeSTATUS
ljkOS_ClearAllFence(ljkDEVICE Device);

ljtUINT64
ljkOS_AllocFenceRecordId(IN ljkDEVICE Device, IN ljsUSER_FENCE_INFO_PTR fence_info);

ljsUSER_FENCE_INFO_PTR
ljkOS_ReleaseFenceRecordId(IN ljkDEVICE Device, IN ljtUINT64 recordId);

void
ljkOS_PreLoadFenceRecId(IN ljkDEVICE Device);

void
ljkOS_PreLoadEndFenceRecId(IN ljkDEVICE Device);

#endif
/*******************************************************************************
 **
 **  ljkOS_AtomConstruct
 **
 **  Create an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Atom
 **          Pointer to a variable receiving the constructed atom.
 */
ljeSTATUS
ljkOS_AtomConstruct(IN ljkOS Os, OUT ljtPOINTER *Atom);

/*******************************************************************************
 **
 **  ljkOS_AtomDestroy
 **
 **  Destroy an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom to destroy.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomDestroy(IN ljkOS Os, OUT ljtPOINTER Atom);

/*******************************************************************************
 **
 **  ljkOS_AtomGet
 **
 **  Get the 32-bit value protected by an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **  OUTPUT:
 **
 **      ljtINT32_PTR Value
 **          Pointer to a variable the receives the value of the atom.
 */
ljeSTATUS
ljkOS_AtomGet(IN ljkOS Os, IN ljtPOINTER Atom, OUT ljtINT32_PTR Value);

/*******************************************************************************
 **
 **  ljkOS_AtomSet
 **
 **  Set the 32-bit value protected by an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **      ljtINT32 Value
 **          The value of the atom.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomSet(IN ljkOS Os, IN ljtPOINTER Atom, IN ljtINT32 Value);

/*******************************************************************************
 **
 **  ljkOS_AtomIncrement
 **
 **  Atomically increment the 32-bit integer value inside an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **  OUTPUT:
 **
 **      ljtINT32_PTR Value
 **          Pointer to a variable the receives the original value of the atom.
 */
ljeSTATUS
ljkOS_AtomIncrement(IN ljkOS Os, IN ljtPOINTER Atom, OUT ljtINT32_PTR Value);

/*******************************************************************************
 **
 **  ljkOS_AtomDecrement
 **
 **  Atomically decrement the 32-bit integer value inside an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **  OUTPUT:
 **
 **      ljtINT32_PTR Value
 **          Pointer to a variable the receives the original value of the atom.
 */
ljeSTATUS
ljkOS_AtomDecrement(IN ljkOS Os, IN ljtPOINTER Atom, OUT ljtINT32_PTR Value);

/* Delay a number of milliseconds. */
ljeSTATUS
ljkOS_Delay(IN ljkOS Os, IN ljtUINT32 Delay);

/* Delay a number of milliseconds. */
ljeSTATUS
ljkOS_Udelay(IN ljkOS Os, IN ljtUINT32 Delay);

/* Get time in milliseconds. */
ljeSTATUS
ljkOS_GetTicks(OUT ljtUINT32_PTR Time);

/* Compare time value. */
ljeSTATUS
ljkOS_TicksAfter(IN ljtUINT32 Time1, IN ljtUINT32 Time2, OUT ljtBOOL_PTR IsAfter);

/* Get time in microseconds. */
ljeSTATUS
ljkOS_GetTime(OUT ljtUINT64_PTR Time);

/* Memory barrier. */
ljeSTATUS
ljkOS_MemoryBarrier(IN ljkOS Os, IN ljtPOINTER Address);

/* Map user pointer. */
ljeSTATUS
ljkOS_MapUserPointer(IN ljkOS Os,
                     IN ljtPOINTER Pointer,
                     IN ljtSIZE_T Size,
                     OUT ljtPOINTER *KernelPointer);

/* Unmap user pointer. */
ljeSTATUS
ljkOS_UnmapUserPointer(IN ljkOS Os,
                       IN ljtPOINTER Pointer,
                       IN ljtSIZE_T Size,
                       IN ljtPOINTER KernelPointer);

/*******************************************************************************
 **
 **  ljkOS_QueryNeedCopy
 **
 **  Query whether the memory can be accessed or mapped directly or it has to be
 **  copied.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtUINT32 ProcessID
 **          Process ID of the current process.
 **
 **  OUTPUT:
 **
 **      ljtBOOL_PTR NeedCopy
 **          Pointer to a boolean receiving ljvTRUE if the memory needs a copy or
 **          ljvFALSE if the memory can be accessed or mapped dircetly.
 */
ljeSTATUS
ljkOS_QueryNeedCopy(IN ljkOS Os, IN ljtUINT32 ProcessID, OUT ljtBOOL_PTR NeedCopy);

/*******************************************************************************
 **
 **  ljkOS_CopyFromUserData
 **
 **  Copy data from user to kernel memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER KernelPointer
 **          Pointer to kernel memory.
 **
 **      ljtPOINTER Pointer
 **          Pointer to user memory.
 **
 **      ljtSIZE_T Size
 **          Number of bytes to copy.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_CopyFromUserData(IN ljkOS Os,
                       IN ljtPOINTER KernelPointer,
                       IN ljtPOINTER Pointer,
                       IN ljtSIZE_T Size);

/*******************************************************************************
 **
 **  ljkOS_CopyToUserData
 **
 **  Copy data from kernel to user memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER KernelPointer
 **          Pointer to kernel memory.
 **
 **      ljtPOINTER Pointer
 **          Pointer to user memory.
 **
 **      ljtSIZE_T Size
 **          Number of bytes to copy.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_CopyToUserData(IN ljkOS Os,
                     IN ljtPOINTER KernelPointer,
                     IN ljtPOINTER Pointer,
                     IN ljtSIZE_T Size);

ljeSTATUS
ljkOS_SuspendInterrupt(IN ljkOS Os);

ljeSTATUS
ljkOS_SuspendInterruptEx(IN ljkOS Os, IN ljeCORE Core);

ljeSTATUS
ljkOS_ResumeInterrupt(IN ljkOS Os);

ljeSTATUS
ljkOS_ResumeInterruptEx(IN ljkOS Os, IN ljeCORE Core);

/* Get the base address for the physical memory. */
ljeSTATUS
ljkOS_GetBaseAddress(IN ljkOS Os, OUT ljtUINT32_PTR BaseAddress);

/* Perform a memory copy. */
ljeSTATUS
ljkOS_MemCopy(IN ljtPOINTER Destination,
              IN ljtCONST_POINTER Source,
              IN ljtSIZE_T Bytes);

/* Zero memory. */
ljeSTATUS
ljkOS_ZeroMemory(IN ljtPOINTER Memory, IN ljtSIZE_T Bytes);

/*******************************************************************************
 **
 **  ljkOS_GetProcessID
 **
 **  Get current process ID.
 **
 **  INPUT:
 **
 **      Nothing.
 **
 **  OUTPUT:
 **
 **      ljtUINT32_PTR ProcessID
 **          Pointer to the variable that receives the process ID.
 */
ljeSTATUS
ljkOS_GetProcessID(OUT ljtUINT32_PTR ProcessID);

ljeSTATUS
ljkOS_GetCurrentProcessID(OUT ljtUINT32_PTR ProcessID);

/*******************************************************************************
 **
 **  ljkOS_GetThreadID
 **
 **  Get current thread ID.
 **
 **  INPUT:
 **
 **      Nothing.
 **
 **  OUTPUT:
 **
 **      ljtUINT32_PTR ThreadID
 **          Pointer to the variable that receives the thread ID.
 */
ljeSTATUS
ljkOS_GetThreadID(OUT ljtUINT32_PTR ThreadID);

/******************************************************************************
 ********************************* Signal Object ******************************
 ******************************************************************************/

/* Create a signal. */
ljeSTATUS
ljkOS_CreateSignal(IN ljkOS Os, IN ljtBOOL ManualReset, OUT ljtSIGNAL *Signal);

/* Destroy a signal. */
ljeSTATUS
ljkOS_DestroySignal(IN ljkOS Os, IN ljtSIGNAL Signal);

/* Signal a signal. */
ljeSTATUS
ljkOS_Signal(IN ljkOS Os, IN ljtSIGNAL Signal, IN ljtBOOL State);

/* Wait for a signal. */
ljeSTATUS
ljkOS_WaitSignal(IN ljkOS Os, IN ljtSIGNAL Signal,
                 IN ljtBOOL Interruptable, IN ljtUINT32 Wait);

#ifdef __QNXNTO__
ljeSTATUS
ljkOS_SignalPulse(IN ljkOS Os, IN ljtSIGNAL Signal);

ljeSTATUS
ljkOS_SignalPending(IN ljkOS Os, IN ljtSIGNAL Signal);
#endif

/* Map a user signal to the kernel space. */
ljeSTATUS
ljkOS_MapSignal(IN ljkOS Os, IN ljtSIGNAL Signal,
                IN ljtHANDLE Process, OUT ljtSIGNAL *MappedSignal);

/* Unmap a user signal */
ljeSTATUS
ljkOS_UnmapSignal(IN ljkOS Os, IN ljtSIGNAL Signal);

/* Get scatter-gather table from memory. */
ljeSTATUS
ljkOS_MemoryGetSGT(IN ljkOS Os,
                   IN ljtPHYS_ADDR Physical,
                   IN ljtSIZE_T Offset,
                   IN ljtSIZE_T Bytes,
                   OUT ljtPOINTER *SGT);

/* Map a page range of memory to user space. */
ljeSTATUS
ljkOS_MemoryMmap(IN ljkOS Os,
                 IN ljtPHYS_ADDR Physical,
                 IN ljtSIZE_T skipPages,
                 IN ljtSIZE_T numPages,
                 INOUT ljtPOINTER Vma);

/* Wrap a user memory to ljtPHYS_ADDR. */
ljeSTATUS
ljkOS_WrapMemory(IN ljkOS Os,
                 IN ljkKERNEL Kernel,
                 IN ljsUSER_MEMORY_DESC_PTR Desc,
                 OUT ljtSIZE_T *Bytes,
                 OUT ljtPHYS_ADDR *Physical,
                 OUT ljtBOOL *Contiguous,
                 OUT ljtSIZE_T *PageCountCpu);

ljeSTATUS
ljkOS_GetPolicyID(IN ljkOS Os,
                  IN ljeVIDMEM_TYPE Type,
                  OUT ljtUINT32_PTR PolicyID,
                  OUT ljtUINT32_PTR AXIConfig);

#if ljdENABLE_MP_SWITCH
ljeSTATUS
ljkOS_SwitchCoreCount(IN ljkOS Os, OUT ljtUINT32 *Count);
#endif

/******************************************************************************
 ************************* Android Native Fence Sync **************************
 ******************************************************************************/
ljeSTATUS
ljkOS_CreateSyncTimeline(IN ljkOS Os, IN ljeCORE Core, OUT ljtHANDLE *Timeline);

ljeSTATUS
ljkOS_DestroySyncTimeline(IN ljkOS Os, IN ljtHANDLE Timeline);

ljeSTATUS
ljkOS_CreateNativeFence(IN ljkOS Os,
                        IN ljtHANDLE Timeline,
                        IN ljtSIGNAL Signal,
                        OUT ljtINT *FenceFD);

ljeSTATUS
ljkOS_WaitNativeFence(IN ljkOS Os, IN ljtHANDLE Timeline,
                      IN ljtINT FenceFD, IN ljtUINT32 Timeout);

#if !USE_NEW_LINUX_SIGNAL
/* Create signal to be used in the user space. */
ljeSTATUS
ljkOS_CreateUserSignal(IN ljkOS Os, IN ljtBOOL ManualReset, OUT ljtINT *SignalID);

/* Destroy signal used in the user space. */
ljeSTATUS
ljkOS_DestroyUserSignal(IN ljkOS Os, IN ljtINT SignalID);

/* Wait for signal used in the user space. */
ljeSTATUS
ljkOS_WaitUserSignal(IN ljkOS Os,
                     IN ljtINT SignalID,
                     IN ljtUINT32 Wait,
                     OUT ljeSIGNAL_STATUS *SignalStatus);

/* Signal a signal used in the user space. */
ljeSTATUS
ljkOS_SignalUserSignal(IN ljkOS Os, IN ljtINT SignalID, IN ljtBOOL State);
#endif /* USE_NEW_LINUX_SIGNAL */

/* Set a signal owned by a process. */
#if defined(__QNXNTO__)
ljeSTATUS
ljkOS_UserSignal(IN ljkOS Os, IN ljtSIGNAL Signal,
                 IN ljtINT Rcvid, IN const struct sigevent *Event);
#else
ljeSTATUS
ljkOS_UserSignal(IN ljkOS Os, IN ljtSIGNAL Signal, IN ljtHANDLE Handle);
#endif

/******************************************************************************\
 ** Cache Support
 */

ljeSTATUS
ljkOS_CacheClean(ljkOS Os,
                 ljtUINT32 ProcessID,
                 ljtPHYS_ADDR Handle,
                 ljtSIZE_T Offset,
                 ljtPOINTER Logical,
                 ljtSIZE_T Bytes);

ljeSTATUS
ljkOS_CacheFlush(ljkOS Os,
                 ljtUINT32 ProcessID,
                 ljtPHYS_ADDR Handle,
                 ljtSIZE_T Offset,
                 ljtPOINTER Logical,
                 ljtSIZE_T Bytes);

ljeSTATUS
ljkOS_CacheInvalidate(ljkOS Os,
                      ljtUINT32 ProcessID,
                      ljtPHYS_ADDR Handle,
                      ljtSIZE_T Offset,
                      ljtPOINTER Logical,
                      ljtSIZE_T Bytes);

ljeSTATUS
ljkOS_CPUPhysicalToGPUPhysical(IN ljkOS Os,
                               IN ljtPHYS_ADDR_T CPUPhysical,
                               IN ljtPHYS_ADDR_T *GPUPhysical);

ljeSTATUS
ljkOS_GPUPhysicalToCPUPhysical(IN ljkOS Os,
                               IN ljtPHYS_ADDR_T GPUPhysical,
                               IN ljtPHYS_ADDR_T *CPUPhysical);

ljeSTATUS
ljkOS_QueryOption(IN ljkOS Os, IN ljtCONST_STRING Option, OUT ljtUINT64 *Value);

/******************************************************************************\
 ** Debug Support
 */

void
ljkOS_SetDebugLevel(IN ljtUINT32 Level);

void
ljkOS_SetDebugZone(IN ljtUINT32 Zone);

void
ljkOS_SetDebugLevelZone(IN ljtUINT32 Level, IN ljtUINT32 Zone);

void
ljkOS_SetDebugZones(IN ljtUINT32 Zones, IN ljtBOOL Enable);

void
ljkOS_SetDebugFile(IN ljtCONST_STRING FileName);

ljeSTATUS
ljkOS_Broadcast(IN ljkOS Os, IN ljkHARDWARE Hardware, IN ljeBROADCAST Reason);

ljeSTATUS
ljkOS_BroadcastHurry(IN ljkOS Os, IN ljkHARDWARE Hardware, IN ljtUINT Urgency);

ljeSTATUS
ljkOS_BroadcastCalibrateSpeed(IN ljkOS Os,
                              IN ljkHARDWARE Hardware,
                              IN ljtUINT Idle,
                              IN ljtUINT Time);

/*******************************************************************************
 **
 **  ljkOS_SetGPUPower
 **
 **  Set the power of the GPU on or off.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkKERNEL Kernel
**          Core whose power is set.
 **
 **      ljtBOOL Clock
 **          ljvTRUE to turn on the clock, or ljvFALSE to turn off the clock.
 **
 **      ljtBOOL Power
 **          ljvTRUE to turn on the power, or ljvFALSE to turn off the power.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_SetGPUPower(IN ljkOS Os,
                  IN ljkKERNEL Kernel,
                  IN ljtBOOL Clock,
                  IN ljtBOOL Power);

ljeSTATUS
ljkOS_SetClockState(IN ljkOS Os, IN ljkKERNEL Kernel, IN ljtBOOL Clock);

ljeSTATUS
ljkOS_GetClockState(IN ljkOS Os, IN ljkKERNEL Kernel, IN ljtBOOL *Clock);

ljeSTATUS
ljkOS_ResetGPU(IN ljkOS Os, IN ljkKERNEL Kernel);

ljeSTATUS
ljkOS_PrepareGPUFrequency(IN ljkOS Os, IN ljeCORE Core);

ljeSTATUS
ljkOS_FinishGPUFrequency(IN ljkOS Os, IN ljeCORE Core);

ljeSTATUS
ljkOS_QueryGPUFrequency(IN ljkOS Os,
                        IN ljkHARDWARE Hardware,
                        OUT ljtUINT32 *Frequency,
                        OUT ljtUINT8 *Scale);

ljeSTATUS
ljkOS_SetGPUFrequency(IN ljkOS Os, IN ljkHARDWARE Hardware, IN ljtUINT8 Scale);

/*******************************************************************************
 ** Semaphores.
 */

/* Create a new semaphore. */
ljeSTATUS
ljkOS_CreateSemaphore(IN ljkOS Os, OUT ljtPOINTER *Semaphore);

ljeSTATUS
ljkOS_CreateSemaphoreEx(IN ljkOS Os, OUT ljtPOINTER *Semaphore);

#if ljdENABLE_VG
ljeSTATUS
ljkOS_CreateSemaphoreVG(IN ljkOS Os, OUT ljtPOINTER *Semaphore);
#endif

/* Delete a semaphore. */
ljeSTATUS
ljkOS_DestroySemaphore(IN ljkOS Os, IN ljtPOINTER Semaphore);

/* Acquire a semaphore. */
ljeSTATUS
ljkOS_AcquireSemaphore(IN ljkOS Os, IN ljtPOINTER Semaphore);

/* Try to acquire a semaphore. */
ljeSTATUS
ljkOS_TryAcquireSemaphore(IN ljkOS Os, IN ljtPOINTER Semaphore);

/* Release a semaphore. */
ljeSTATUS
ljkOS_ReleaseSemaphore(IN ljkOS Os, IN ljtPOINTER Semaphore);

/* Release a semaphore. */
ljeSTATUS
ljkOS_ReleaseSemaphoreEx(IN ljkOS Os, IN ljtPOINTER Semaphore);

/*******************************************************************************
 ** Timer API.
 */

typedef void (*ljtTIMERFUNCTION)(ljtPOINTER);

/* Create a timer. */
ljeSTATUS
ljkOS_CreateTimer(IN ljkOS Os,
                  IN ljtTIMERFUNCTION Function,
                  IN ljtPOINTER Data,
                  OUT ljtPOINTER *Timer);

/* Destroy a timer. */
ljeSTATUS
ljkOS_DestroyTimer(IN ljkOS Os, IN ljtPOINTER Timer);

/* Start a timer. */
ljeSTATUS
ljkOS_StartTimer(IN ljkOS Os, IN ljtPOINTER Timer, IN ljtUINT32 Delay);

/* Stop a timer. */
ljeSTATUS
ljkOS_StopTimer(IN ljkOS Os, IN ljtPOINTER Timer);

/******************************************************************************
 ******************************** ljkHEAP Object ******************************
 ******************************************************************************/

typedef struct _ljkHEAP *ljkHEAP;

/* Construct a new ljkHEAP object. */
ljeSTATUS
ljkHEAP_Construct(IN ljkOS Os, IN ljtSIZE_T AllocationSize, OUT ljkHEAP *Heap);

/* Destroy an ljkHEAP object. */
ljeSTATUS
ljkHEAP_Destroy(IN ljkHEAP Heap);

/* Allocate memory. */
ljeSTATUS
ljkHEAP_Allocate(IN ljkHEAP Heap, IN ljtSIZE_T Bytes, OUT ljtPOINTER *Node);

/* Free memory. */
ljeSTATUS
ljkHEAP_Free(IN ljkHEAP Heap, IN ljtPOINTER Node);

/* Profile the heap. */
ljeSTATUS
ljkHEAP_ProfileStart(IN ljkHEAP Heap);

ljeSTATUS
ljkHEAP_ProfileEnd(IN ljkHEAP Heap, IN ljtCONST_STRING Title);

/******************************************************************************
 ******************************* ljkKERNEL Object *****************************
 ******************************************************************************/

struct _ljsHAL_INTERFACE;

/* Construct a new ljkKERNEL object. */
ljeSTATUS
ljkKERNEL_Construct(IN ljkOS Os,
                    IN ljeCORE Core,
                    IN ljtUINT ChipID,
                    IN ljtPOINTER Context,
                    IN ljkDEVICE Device,
                    IN ljkDB SharedDB,
                    OUT ljkKERNEL *Kernel);

/* Destroy an ljkKERNEL object. */
ljeSTATUS
ljkKERNEL_Destroy(IN ljkKERNEL Kernel);

/* Dispatch a user-level command. */
ljeSTATUS
ljkKERNEL_Dispatch(IN ljkKERNEL Kernel,
                   IN ljkDEVICE Device,
                   IN OUT struct _ljsHAL_INTERFACE *Interface);

/* Query Database requirements. */
ljeSTATUS
ljkKERNEL_QueryDatabase(IN ljkKERNEL Kernel,
                        IN ljtUINT32 ProcessID,
                        IN OUT ljsHAL_INTERFACE *Interface);

/* Query the video memory. */
ljeSTATUS
ljkKERNEL_QueryVideoMemory(IN ljkKERNEL Kernel,
                           OUT struct _ljsHAL_INTERFACE *Interface);

/* Lookup the ljkVIDMEM object for a pool. */
ljeSTATUS
ljkKERNEL_GetVideoMemoryPool(IN ljkKERNEL Kernel, IN ljePOOL Pool,
                             OUT ljkVIDMEM *VideoMemory);

/* Map dedicated video memory node. */
ljeSTATUS
ljkKERNEL_MapVideoMemory(IN ljkKERNEL Kernel,
                         IN ljtBOOL InUserSpace,
                         IN ljePOOL Pool,
                         IN ljtPHYS_ADDR Physical,
                         IN ljtSIZE_T Offset,
                         IN ljtSIZE_T Bytes,
                         OUT ljtPOINTER *Logical);

/* Unmap dedicated video memory. */
ljeSTATUS
ljkKERNEL_UnmapVideoMemory(IN ljkKERNEL Kernel,
                           IN ljePOOL Pool,
                           IN ljtPHYS_ADDR Physical,
                           IN ljtPOINTER Logical,
                           IN ljtUINT32 Pid,
                           IN ljtSIZE_T Bytes);

/* Map memory. */
ljeSTATUS
ljkKERNEL_MapMemory(IN ljkKERNEL Kernel,
                    IN ljtPHYS_ADDR Physical,
                    IN ljtSIZE_T Bytes,
                    OUT ljtPOINTER *Logical);

/* Unmap memory. */
ljeSTATUS
ljkKERNEL_UnmapMemory(IN ljkKERNEL Kernel,
                      IN ljtPHYS_ADDR Physical,
                      IN ljtSIZE_T Bytes,
                      IN ljtPOINTER Logical,
                      IN ljtUINT32 ProcessID);
/* Destroy reserved mem when destroy process*/
ljeSTATUS
ljkKERNEL_DestroyProcessReservedUserMap(IN ljkKERNEL Kernel, IN ljtUINT32 Pid);

/* Notification of events. */
ljeSTATUS
ljkKERNEL_Notify(IN ljkKERNEL Kernel, IN ljeNOTIFY Notifcation);

#if ljdENABLE_VIDEO_MEMORY_MIRROR
ljeSTATUS
ljkKERNEL_SyncVideoMemoryMirror(ljkKERNEL Kernel,
                                ljkVIDMEM_NODE Node,
                                ljtSIZE_T Offset,
                                ljtSIZE_T Bytes,
                                ljtUINT32 Reason);
#endif
/*******************************************************************************
 **
 **  ljkKERNEL_Recovery
 **
 **  Try to recover the GPU from a fatal error.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkKERNEL_Recovery(IN ljkKERNEL Kernel);

/* Get access to the user data. */
ljeSTATUS
ljkKERNEL_OpenUserData(IN ljkKERNEL Kernel,
                       IN ljtBOOL NeedCopy,
                       IN ljtPOINTER StaticStorage,
                       IN ljtPOINTER UserPointer,
                       IN ljtSIZE_T Size,
                       OUT ljtPOINTER *KernelPointer);

/* Release resources associated with the user data connection. */
ljeSTATUS
ljkKERNEL_CloseUserData(IN ljkKERNEL Kernel,
                        IN ljtBOOL NeedCopy,
                        IN ljtBOOL FlushData,
                        IN ljtPOINTER UserPointer,
                        IN ljtSIZE_T Size,
                        OUT ljtPOINTER *KernelPointer);

/* Query kernel by core index */
ljeSTATUS
ljkOS_QueryKernel(IN ljkKERNEL Kernel, IN ljtINT index, OUT ljkKERNEL *KernelOut);

ljeSTATUS
ljkDVFS_Construct(IN ljkHARDWARE Hardware, OUT ljkDVFS *Frequency);

ljeSTATUS
ljkDVFS_Destroy(IN ljkDVFS Dvfs);

ljeSTATUS
ljkDVFS_Start(IN ljkDVFS Dvfs);

ljeSTATUS
ljkDVFS_Stop(IN ljkDVFS Dvfs);

/* Dma operation of CPU Mem <=>GPU DDR and GPU DDR <=> GPU DDR*/
ljeSTATUS
ljkKERNEL_DmaOperation(
    IN ljkKERNEL Kernel,
    IN ljkVIDMEM_NODE Node_Src,
    IN ljkVIDMEM_NODE Node_Dst,
    IN ljtUINT32 Dir,
    IN ljtUINT64 TransLen
    );

ljeSTATUS
ljoOS_GetVideoMemoryPool(
    IN ljtUINT32 Node,
    OUT ljePOOL *Pool
    );

/******************************************************************************\
******************************* ljkHARDWARE Object *****************************
\******************************************************************************/

/* Construct a new ljkHARDWARE object. */
ljeSTATUS
ljkHARDWARE_Construct(IN ljkOS Os,
                      IN ljkKERNEL Kernel,
                      OUT ljkHARDWARE *Hardware);

/* Post hardware resource allocation after ljkHARDWARE object constructed. */
ljeSTATUS
ljkHARDWARE_PostConstruct(IN ljkHARDWARE Hardware);

/* Pre-destroy hardwre resource before destroying an ljkHARDWARE object. */
ljeSTATUS
ljkHARDWARE_PreDestroy(IN ljkHARDWARE Hardware);

/* Destroy an ljkHARDWARE object. */
ljeSTATUS
ljkHARDWARE_Destroy(IN ljkHARDWARE Hardware);

/* Get hardware type. */
ljeSTATUS
ljkHARDWARE_GetType(IN ljkHARDWARE Hardware, OUT ljeHARDWARE_TYPE *Type);

/* Query system memory requirements. */
ljeSTATUS
ljkHARDWARE_QuerySystemMemory(IN ljkHARDWARE Hardware,
                              OUT ljtSIZE_T *SystemSize,
                              OUT ljtUINT32 *SystemBaseAddress);

/* Build virtual address. */
ljeSTATUS
ljkHARDWARE_BuildVirtualAddress(IN ljkHARDWARE Hardware,
                                IN ljtUINT32 Index,
                                IN ljtUINT32 Offset,
                                OUT ljtUINT32 *Address);

/* Query command buffer requirements. */
ljeSTATUS
ljkHARDWARE_QueryCommandBuffer(IN ljkHARDWARE Hardware,
                               IN ljeENGINE Engine,
                               OUT ljtUINT32 *Alignment,
                               OUT ljtUINT32 *ReservedHead,
                               OUT ljtUINT32 *ReservedTail);

/* Add a PIPESELECT command in the command queue. */
ljeSTATUS
ljkHARDWARE_PipeSelect(IN ljkHARDWARE Hardware,
                       IN ljtPOINTER Logical,
                       IN ljePIPE_SELECT Pipe,
                       IN OUT ljtUINT32  *Bytes);

/* Query the available memory. */
ljeSTATUS
ljkHARDWARE_QueryMemory(IN ljkHARDWARE Hardware,
                        OUT ljtSIZE_T *InternalSize,
                        OUT ljtADDRESS *InternalBaseAddress,
                        OUT ljtUINT32 *InternalAlignment,
                        OUT ljtSIZE_T *ExternalSize,
                        OUT ljtADDRESS *ExternalBaseAddress,
                        OUT ljtUINT32 *ExternalAlignment,
                        OUT ljtUINT32 *HorizontalTileSize,
                        OUT ljtUINT32 *VerticalTileSize);

/* Query the identity of the hardware. */
ljeSTATUS
ljkHARDWARE_QueryChipIdentity(IN ljkHARDWARE Hardware,
                              OUT ljsHAL_QUERY_CHIP_IDENTITY_PTR Identity);

ljeSTATUS
ljkHARDWARE_QueryChipOptions(IN ljkHARDWARE Hardware,
                             OUT ljsHAL_QUERY_CHIP_OPTIONS_PTR Options);

/* Split a harwdare specific address into API stuff. */
ljeSTATUS
ljkHARDWARE_SplitMemory(IN ljkHARDWARE Hardware,
                        IN ljtUINT32 Address,
                        OUT ljePOOL *Pool,
                        OUT ljtUINT32 *Offset);

/* Update command queue tail pointer. */
ljeSTATUS
ljkHARDWARE_UpdateQueueTail(IN ljkHARDWARE Hardware,
                            IN ljtPOINTER Logical,
                            IN ljtUINT32 Offset);

/* Interrupt manager. */
ljeSTATUS
ljkHARDWARE_Interrupt(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_Notify(IN ljkHARDWARE Hardware);

/* Program MMU. */
ljeSTATUS
ljkHARDWARE_SetMMU(IN ljkHARDWARE Hardware, IN ljkMMU Mmu);

/* Flush the MMU. */
ljeSTATUS
ljkHARDWARE_FlushMMU(IN ljkHARDWARE Hardware,
                     IN ljtPOINTER Logical,
                     IN ljtADDRESS Address,
                     IN ljtUINT32 SubsequentBytes,
                     IN OUT ljtUINT32 *Bytes);

ljeSTATUS
ljkHARDWARE_FlushAsyncMMU(IN ljkHARDWARE Hardware,
                          IN ljtPOINTER Logical,
                          IN OUT ljtUINT32 *Bytes);

ljeSTATUS
ljkHARDWARE_FlushMcfeMMU(IN ljkHARDWARE Hardware,
                         IN ljtPOINTER Logical,
                         IN OUT ljtUINT32 *Bytes);

/* Get idle register. */
ljeSTATUS
ljkHARDWARE_GetIdle(IN ljkHARDWARE Hardware, IN ljtBOOL Wait, OUT ljtUINT32 *Data);

/* Flush the caches. */
ljeSTATUS
ljkHARDWARE_Flush(IN ljkHARDWARE Hardware,
                  IN ljeKERNEL_FLUSH Flush,
                  IN ljtPOINTER Logical,
                  IN OUT ljtUINT32 *Bytes);

/* Enable/disable fast clear. */
ljeSTATUS
ljkHARDWARE_SetFastClear(IN ljkHARDWARE Hardware,
                         IN ljtINT Enable,
                         IN ljtINT Compression);

ljeSTATUS
ljkHARDWARE_ReadInterrupt(IN ljkHARDWARE Hardware, OUT ljtUINT32_PTR IDs);

/*
 * State timer helper.
 */
ljeSTATUS
ljkHARDWARE_StartTimerReset(IN ljkHARDWARE Hardware);

/* Power management. */
ljeSTATUS
ljkHARDWARE_SetPowerState(IN ljkHARDWARE Hardware, IN ljeCHIPPOWERSTATE State);

ljeSTATUS
ljkHARDWARE_QueryPowerStateUnlocked(IN ljkHARDWARE Hardware, OUT ljeCHIPPOWERSTATE *State);

ljeSTATUS
ljkHARDWARE_QueryPowerState(IN ljkHARDWARE Hardware, OUT ljeCHIPPOWERSTATE *State);

ljeSTATUS
ljkHARDWARE_EnablePowerManagement(IN ljkHARDWARE Hardware, IN ljtBOOL Enable);

ljeSTATUS
ljkHARDWARE_QueryPowerManagement(IN ljkHARDWARE Hardware, OUT ljtBOOL *Enable);

ljeSTATUS
ljkHARDWARE_SetGpuProfiler(IN ljkHARDWARE Hardware, IN ljtBOOL GpuProfiler);

#if ljdENABLE_FSCALE_VAL_ADJUST
ljeSTATUS
ljkHARDWARE_SetFscaleValue(IN ljkHARDWARE Hardware,
                           IN ljtUINT32 FscaleValue,
                           IN ljtUINT32 ShaderFscaleValue);

ljeSTATUS
ljkHARDWARE_GetFscaleValue(IN ljkHARDWARE Hardware,
                           IN ljtUINT *FscaleValue,
                           IN ljtUINT *MinFscaleValue,
                           IN ljtUINT *MaxFscaleValue);

ljeSTATUS
ljkHARDWARE_SetMinFscaleValue(IN ljkHARDWARE Hardware, IN ljtUINT MinFscaleValue);
#endif

ljeSTATUS
ljkHARDWARE_InitializeHardware(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_Reset(IN ljkHARDWARE Hardware);

/* Check for Hardware features. */
ljeSTATUS
ljkHARDWARE_IsFeatureAvailable(IN ljkHARDWARE Hardware, IN ljeFEATURE Feature);

ljeSTATUS
ljkHARDWARE_DumpMMUException(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_DumpGPUState(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_InitDVFS(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_QueryLoad(IN ljkHARDWARE Hardware, OUT ljtUINT32 *Load);

ljeSTATUS
ljkHARDWARE_SetDVFSPeroid(IN ljkHARDWARE Hardware, IN ljtUINT32 Frequency);

ljeSTATUS
ljkHARDWARE_QueryStateTimer(IN ljkHARDWARE Hardware,
                            OUT ljtUINT64_PTR On,
                            OUT ljtUINT64_PTR Off,
                            OUT ljtUINT64_PTR Idle,
                            OUT ljtUINT64_PTR Suspend);

ljeSTATUS
ljkHARDWARE_Fence(IN ljkHARDWARE Hardware,
                  IN ljeENGINE Engine,
                  IN ljtPOINTER Logical,
                  IN ljtADDRESS FenceAddress,
                  IN ljtUINT64 FenceData,
                  IN OUT ljtUINT32 *Bytes);

#if !ljdENABLE_VG
/******************************************************************************
 **************************** ljkINTERRUPT Object *****************************
 ******************************************************************************/

typedef struct _ljkINTERRUPT *ljkINTERRUPT;

typedef ljeSTATUS (*ljtINTERRUPT_HANDLER)(IN ljkKERNEL Kernel);

ljeSTATUS
ljkINTERRUPT_Construct(IN ljkKERNEL Kernel, OUT ljkINTERRUPT *Interrupt);

ljeSTATUS
ljkINTERRUPT_Destroy(IN ljkINTERRUPT Interrupt);

ljeSTATUS
ljkINTERRUPT_SetHandler(IN ljkINTERRUPT Interrupt,
                        IN OUT ljtINT32_PTR Id,
                        IN ljtINTERRUPT_HANDLER Handler);

ljeSTATUS
ljkINTERRUPT_Notify(IN ljkINTERRUPT Interrupt, IN ljtBOOL Valid);
#endif

/******************************************************************************
 ******************************** ljkMMU Object *******************************
 ******************************************************************************/

/* Construct a new ljkMMU object. */
ljeSTATUS
ljkMMU_Construct(IN ljkKERNEL Kernel, IN ljtSIZE_T MmuSize, OUT ljkMMU *Mmu);

/* Destroy an ljkMMU object. */
ljeSTATUS
ljkMMU_Destroy(IN ljkMMU Mmu);

/* Allocate pages inside the MMU. */
ljeSTATUS
ljkMMU_AllocatePages(IN ljkMMU Mmu,
                     IN ljtSIZE_T PageCount,
                     IN ljePAGE_TYPE PageType,
                     OUT ljtPOINTER *PageTable,
                     OUT ljtADDRESS *Address);

ljeSTATUS
ljkMMU_AllocatePagesEx(IN ljkMMU Mmu,
                       IN ljtSIZE_T PageCount,
                       IN ljeVIDMEM_TYPE Type,
                       IN ljePAGE_TYPE PageType,
                       IN ljtBOOL LowVA,
                       IN ljtBOOL Secure,
                       OUT ljtPOINTER *PageTable,
                       OUT ljtADDRESS *Address);

/* Remove a page table from the MMU. */
ljeSTATUS
ljkMMU_FreePages(IN ljkMMU Mmu,
                 IN ljtBOOL Secure,
                 IN ljePAGE_TYPE PageType,
                 IN ljtBOOL LowVA,
                 IN ljtADDRESS Address,
                 IN ljtPOINTER PageTable,
                 IN ljtSIZE_T PageCount);

/* Set the MMU page with info. */
ljeSTATUS
ljkMMU_SetPage(IN ljkMMU Mmu,
               IN ljtPHYS_ADDR_T PageAddress,
               IN ljePAGE_TYPE PageType,
               IN ljtBOOL LowVA,
               IN ljtBOOL Writable,
               IN ljtUINT32 *PageEntry);

ljeSTATUS
ljkMMU_Flush(IN ljkMMU Mmu, IN ljeVIDMEM_TYPE Type);

ljeSTATUS
ljkMMU_DumpPageTableEntry(IN ljkMMU Mmu, IN ljeAREA_TYPE AreaType, IN ljtADDRESS Address);

ljeSTATUS
ljkMMU_FillFlatMapping(IN ljkMMU Mmu,
                       IN ljtUINT64 PhysBase,
                       IN ljtSIZE_T Size,
                       IN ljtBOOL Reserved,
                       IN ljtBOOL AbleToShift,
                       OUT ljtADDRESS *GpuBaseAddress);

ljeSTATUS
ljkMMU_IsFlatMapped(IN ljkMMU Mmu,
                    IN ljtUINT64 Physical,
                    IN ljtSIZE_T Bytes,
                    OUT ljtBOOL *In,
                    INOUT ljtADDRESS *Address);

ljeSTATUS
ljkMMU_GetAreaType(IN ljkMMU Mmu, IN ljtADDRESS GpuAddress, OUT ljeAREA_TYPE *AreaType);

ljeSTATUS
ljkHARDWARE_QueryContextProfile(IN ljkHARDWARE Hardware,
                                IN ljtBOOL Reset,
                                OUT ljsPROFILER_COUNTERS_PART1 *Counters_part1,
                                OUT ljsPROFILER_COUNTERS_PART2 *Counters_part2);

ljeSTATUS
ljkHARDWARE_UpdateContextProfile(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_InitProfiler(IN ljkHARDWARE Hardware);

ljeSTATUS
ljkOS_DetectProcessByName(IN ljtCONST_POINTER Name);

void
ljkOS_DumpParam(void);

ljeSTATUS
lj_mmuinfo_show(void);

#ifdef __cplusplus
}
#endif

#if ljdENABLE_VG
#include "lj_hal_vg.h"
#endif

#endif /* __lj_hal_h_ */


