/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_base_h_
#define __lj_hal_base_h_

#include "lj_hal_enum.h"
#include "lj_hal_types.h"
#include "lj_hal_debug_zones.h"
#include "lj_hal_base_shared.h"


#ifdef __QNXNTO__
# define CHECK_PRINTF_FORMAT(string_index, first_to_check) \
        __attribute__((__format__(__printf__, (string_index), (first_to_check))))
#else
# define CHECK_PRINTF_FORMAT(string_index, first_to_check)
#endif

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************
 ****************************** Object Declarations ***************************
 ******************************************************************************/

#if ljdENABLE_LJMLAYER_GL
typedef struct _ljsLJMLAYERWIN {
    ljtBOOL useLjmLayer;
    ljtINT  width;
    ljtINT  height;
    void* logicalAddr;
    uint64_t physicalAddr;
} ljsLJMLAYERWIN;
#endif

typedef struct _ljkOS                  *ljkOS;
typedef struct _ljoHAL                 *ljoHAL;
typedef struct _ljoOS                  *ljoOS;
typedef struct _ljo2D                  *ljo2D;
typedef struct ljsATOM                 *ljsATOM_PTR;

typedef struct _ljo3D                  *ljo3D;
typedef struct _ljoCL                  *ljoCL;
typedef struct _ljoVX                  *ljoVX;
typedef struct _ljsFAST_FLUSH          *ljsFAST_FLUSH_PTR;

typedef struct _ljoSURF                *ljoSURF;
typedef struct _ljsSURF_NODE           *ljsSURF_NODE_PTR;
typedef struct _ljsSURF_FORMAT_INFO    *ljsSURF_FORMAT_INFO_PTR;
typedef struct _ljsPOINT               *ljsPOINT_PTR;
typedef struct _ljsSIZE                *ljsSIZE_PTR;
typedef struct _ljsRECT                *ljsRECT_PTR;
typedef struct _ljsBOUNDARY            *ljsBOUNDARY_PTR;
typedef struct _ljoHARDWARE            *ljoHARDWARE;
typedef struct _ljoDEVICE              *ljoDEVICE;
#if ljdENABLE_MULTI_DEVICE_MANAGEMENT
typedef struct _ljsHARDWARE_PROFILE     ljsHARDWARE_PROFILE;
#endif
typedef union  _ljuVIDMEM_NODE         *ljuVIDMEM_NODE_PTR;
typedef struct _ljsVIDMEM_NODE         *ljkVIDMEM_NODE;
typedef struct _ljsVIDMEM_BLOCK        *ljkVIDMEM_BLOCK;

#if ljdENABLE_VG
typedef struct _ljoVG                  *ljoVG;
typedef struct _ljsCOMPLETION_SIGNAL   *ljsCOMPLETION_SIGNAL_PTR;
typedef struct _ljsCONTEXT_MAP         *ljsCONTEXT_MAP_PTR;
#else
typedef void                           *ljoVG;
#endif

typedef struct _ljoFENCE               *ljoFENCE;
typedef struct _ljsSYNC_CONTEXT        *ljsSYNC_CONTEXT_PTR;

typedef struct _ljsUSER_MEMORY_DESC    *ljsUSER_MEMORY_DESC_PTR;

#if ljdENABLE_CLEAR_FENCE
typedef struct _ljsUSER_FENCE_INFO *ljsUSER_FENCE_INFO_PTR;
#endif

/* Immuatable features from database */
typedef struct _ljsNN_FIXED_FEATURE {
    ljtUINT  hwChipVersion;
    ljtUINT  vipCoreCount;
    ljtUINT  vipRingCount;
    ljtUINT  nnMadPerCore;
    ljtUINT  nnInputBufferDepth;
    ljtUINT  nnAccumBufferDepth;
    ljtUINT  nnFCNonPrunAccel;
    ljtUINT  nnInImageOffsetBits;
    ljtUINT  tpCoreCount; /* full-function core count */
    ljtUINT  tpPwlLUTCount;
    ljtUINT  tpPwlLUTSize;
    ljtUINT  vip7Version;
    ljtUINT  vipBrickMode;
    ljtUINT  tpReorderInImageSize;
    ljtUINT  tpliteCoreCount; /* fc-only core count */
    ljtUINT  nnFP16XYDPX;
    ljtUINT  nnFP16XYDPY;
    ljtUINT  nnFP16ZDP;
    ljtUINT  zrlBits;
    ljtUINT  uscCacheControllers;
    ljtUINT  uscBanks;
    ljtUINT  nnLanesPerOutCycle;
    ljtUINT  maxOTNumber;
    ljtUINT  physicalVipSramWidthInByte;
    ljtUINT  equivalentVipsramWidthInByte;
    ljtUINT  shaderCoreCount;
    ljtUINT  latencyHidingAtFullAxiBw;
    ljtUINT  axiBusWidth;
    ljtUINT  nnMaxKXSize;
    ljtUINT  nnMaxKYSize;
    ljtUINT  nnMaxKZSize;
    ljtUINT  nnClusterNumForPowerControl;
    ljtUINT  vipMinAxiBurstSize;
    ljtUINT  nnInLinesPerCycle;
    ljtUINT  nnPreprocessorMaxSegmentPerCycle;

    /* stream processor info */
    ljtUINT  streamProcessorExecCount;
    ljtUINT  streamProcessorVectorSize;

    /* add related information for check in/out size */
    ljtUINT  outImageXStrideBits;
    ljtUINT  outImageYStrideBits;
    ljtUINT  inImageXStrideBits;
    ljtUINT  inImageYStrideBits;
    ljtUINT  outImageXSizeBits;
    ljtUINT  outImageYSizeBits;
    ljtUINT  inImageXSizeBits;
    ljtUINT  inImageYSizeBits;
    ljtUINT  smallAccumBits;
    ljtUINT  coefDecompressPerfX;
} ljsNN_FIXED_FEATURE;

/* Features can be customized from outside */
typedef struct _ljsNN_CUSTOMIZED_FEATURE {
    ljtUINT  nnActiveCoreCount;
    ljtUINT  nnCoreCount;         /* total nn core count */
    ljtUINT  nnCoreCountInt8;     /* total nn core count supporting int8 */
    ljtUINT  nnCoreCountInt16;    /* total nn core count supporting int16 */
    ljtUINT  nnCoreCountUint16;    /* total nn core count supporting uint16 */
    ljtUINT  nnCoreCountFloat16;  /* total nn core count supporting float16 */
    ljtUINT  nnCoreCountBFloat16; /* total nn core count supporting Bfloat16 */
    ljtUINT  vipSRAMSize;
    ljtUINT  axiSRAMSize;
    ljtFLOAT ddrReadBWLimit;
    ljtFLOAT ddrWriteBWLimit;
    ljtFLOAT ddrTotalBWLimit;
    ljtFLOAT axiSramReadBWLimit;
    ljtFLOAT axiSramWriteBWLimit;
    ljtFLOAT axiSramTotalBWLimit;
    ljtFLOAT axiBusReadBWLimit;
    ljtFLOAT axiBusWriteBWLimit;
    ljtFLOAT axiBusTotalBWLimit;
    ljtUINT  vipSWTiling;
    ljtFLOAT ddrLatency;
    ljtUINT  freqInMHZ;
    ljtUINT  axiClockFreqInMHZ;
    ljtUINT  maxSocOTNumber; /*max SOC outstanding transfer number*/
    ljtUINT  nnWriteWithoutUSC;
    ljtUINT  depthWiseSupport;
    ljtUINT  vipVectorPrune;
    ljtUINT  ddrKernelBurstSize;
    ljtFLOAT  axiSRAMLatency;
} ljsNN_CUSTOMIZED_FEATURE;

/* Features are unified (hardcoded) for hardwares */
typedef struct _ljsNN_UNIFIED_FEATURE {
    ljtUINT  nnUSCCacheSize;
    ljtUINT  nnCmdSizeInBytes;
    ljtUINT  tpCmdSizeInBytes;
    ljtUINT  vipCoefDecodePerf;
    ljtUINT  vipCachedReadFromSram;
    ljtUINT  vipImagePartialCache;
    ljtUINT  lanesPerConv;
    ljtUINT  maxTileSize;
    ljtUINT  fullCacheKernelHeadFix : 1;
    ljtUINT  conv1x1HalfPerformance : 1;
    ljtUINT  per3DTileBubbleFix : 1;
    ljtUINT  cacheLineModeDisabled : 1;
    ljtUINT  tpReOrderFix : 1;
    ljtUINT  zdp3NoCompressFix : 1;
    ljtUINT  asyncCopyPerfFix : 1;
    ljtUINT  accurateTileBW : 1;
    ljtUINT  zxdp3KernelReadConflictFix : 1;
    ljtUINT  axiSramSlowedDownByAddr : 1;
    ljtUINT  slowNNReqArbitrationFix : 1;
    ljtUINT  singlePortAccBuffer : 1;
    ljtUINT  convOutFifoDepthFix : 1;
    ljtUINT  smallBatchEnable : 1;
    ljtUINT  axiSramOnlySWTiling : 1;
    ljtUINT  imageNotPackedInSram : 1;
    ljtUINT  coefDeltaCordOverFlowZRL8BitFix : 1;
    ljtUINT  lowEfficiencyOfIDWriteImgBufFix : 1;
    ljtUINT  xyOffsetLimitationFix : 1;
    ljtUINT  kernelPerCoreLTOneThirdCoefFix : 1;
    ljtUINT  diffConditionForCachelineModePreFix : 1;
} ljsNN_UNIFIED_FEATURE;

/* Features are derived from above ones */
typedef struct _ljsNN_DERIVIED_FEATURE {
    ljtUINT  nnDPAmount;
    ljtUINT  nnXYDPX;
    ljtUINT  nnXYDPY;
    ljtUINT  nnZDP;
    ljtFLOAT totalLatency;
    ljtFLOAT internalLatency;
    ljtFLOAT ddrReadBWInBytePerCycle;
    ljtFLOAT ddrWriteBWInBytePerCycle;
    ljtFLOAT totalAxiSRAMLatency;
} ljsNN_DERIVED_FEATURE;

/******************************************************************************
 ********************* Share obj lock/unlock macros. **************************
 ******************************************************************************/
#define ljmLOCK_SHARE_OBJ(Obj)                                             \
    {                                                                      \
        if (Obj->sharedLock != ljvNULL) {                                  \
            (ljoOS_AcquireMutex(ljvNULL, Obj->sharedLock, ljvINFINITE));   \
        }                                                                  \
    }

#define ljmUNLOCK_SHARE_OBJ(Obj)                                           \
    {                                                                      \
        if (Obj->sharedLock != ljvNULL) {                                  \
            (ljoOS_ReleaseMutex(ljvNULL, Obj->sharedLock));                \
        }                                                                  \
    }

typedef struct _ljsSystemInfo {
    /* memory latency number for SH data fetch, in SH cycle*/
    ljtUINT32 memoryLatencySH;
} ljsSystemInfo;

#define ljPLS_INITIALIZER                                    \
{                                                            \
    ljvNULL, /* ljoOS object.      */               \
    ljvNULL, /* ljoHAL object.     */               \
    0, /* internalSize       */               \
    0, /* internalPhysName   */               \
    ljvNULL, /* internalLogical    */               \
    0, /* externalSize       */               \
    0, /* externalPhysName   */               \
    ljvNULL, /* externalLogical    */               \
    0, /* contiguousSize     */               \
    0, /* contiguousPhysName */               \
    ljvNULL, /* contiguousLogical  */               \
    ljvNULL, /* eglDisplayInfo     */               \
    ljvNULL, /* eglSurfaceInfo     */               \
    ljvSURF_A8R8G8B8, /* eglConfigFormat    */               \
    ljvNULL, /* reference          */               \
    0, /* processID          */               \
    0, /* threadID           */               \
    ljvFALSE, /* exiting            */               \
    ljvFALSE, /* Special flag for NP2 texture. */    \
    ljvFALSE, /* device open.       */               \
    {ljvNULL}, /* destructor         */               \
    ljvNULL, /* accessLock         */               \
    ljvNULL, /* GL FE compiler lock*/               \
    ljvNULL, /* CL FE compiler lock*/               \
    ljvNULL, /* VX context lock    */               \
    ljvPATCH_NOTINIT, /* global patchID     */               \
    ljvNULL, /* global fenceID*/                    \
    ljvNULL, /* mainThreadHandle */                 \
    ljvFALSE, /* memory profile flag */              \
    ljvNULL, /* profileLock;        */              \
    0, /* allocCount;         */              \
    0, /* allocSize;          */              \
    0, /* maxAllocSize;       */              \
    0, /* freeCount;          */              \
    0, /* freeSize;           */              \
    0, /* currentSize;        */              \
    0, /* video_allocCount;   */              \
    0, /* video_allocSize;    */              \
    0, /* video_maxAllocSize; */              \
    0, /* video_freeCount;    */              \
    0, /* video_freeSize;     */              \
    0, /* video_currentSize;  */              \
}

/******************************************************************************
 ******************************* Thread local storage *************************
 ******************************************************************************/

typedef struct _ljsDRIVER_TLS *ljsDRIVER_TLS_PTR;

typedef struct _ljsDRIVER_TLS {
    void (*destructor)(ljsDRIVER_TLS_PTR Tls);
} ljsDRIVER_TLS;

typedef struct _ljsTLS *ljsTLS_PTR;

typedef struct _ljsTLS {
    ljeHARDWARE_TYPE            currentType;
    ljeHARDWARE_TYPE            targetType;

    /* To which core device control is called,
     * it is index in a hardware type.
     */
    ljtUINT32                   currentCoreIndex;

    /* Current device index of this thread. */
    ljtUINT32                   currentDevIndex;

    /* Current 3D hardwre of this thread */
    ljoHARDWARE                 currentHardware;

    /* Default 3D hardware of this thread */
    ljoHARDWARE                 defaultHardware;

    /* Only for separated 3D and 2D */
    ljoHARDWARE                 hardware2D;
#if ljdENABLE_VG
    ljoVGHARDWARE               vg;
    ljoVG                       engineVG;
#endif /* ljdENABLE_VG */
#if ljdENABLE_3D
    ljo3D                       engine3D;
#endif
#if ljdENABLE_2D
    ljo2D                       engine2D;
#if ljdDUMP_2D
    ljtUINT32                   newDump2DFlag;
# endif

#endif
    ljoVX                       engineVX;

    ljtBOOL                     copied;

    /* libLJM.so handle */
    ljtHANDLE                   handle;

    ljtHANDLE                   graph;

    /* If true, do not releas 2d engine and hardware in hal layer */
    ljtBOOL                     release2DUpper;

    /* Driver tls. */
    ljsDRIVER_TLS_PTR           driverTLS[ljvTLS_KEY_COUNT];

    ljtINT32                    cmd_buf_size;

#if ljdENABLE_SW_PREEMPTION
    /* PriorityID. */
    ljtUINT                     priorityID;
#endif
} ljsTLS;

typedef struct _ljsSURF_VIEW {
    ljoSURF surf;
    ljtUINT firstSlice;
    ljtUINT numSlices;
} ljsSURF_VIEW;

/* ljsHAL_Limits*/
typedef struct _ljsHAL_LIMITS {
    /* chip info */
    ljeCHIPMODEL    chipModel;
    ljtUINT32       chipRevision;
    ljtUINT32       featureCount;
    ljtUINT32      *chipFeatures;

    /* target caps */
    ljtUINT32       maxWidth;
    ljtUINT32       maxHeight;
    ljtUINT32       multiTargetCount;
    ljtUINT32       maxSamples;

} ljsHAL_LIMITS;

typedef struct _ljsHAL_CHIPIDENTITY {
    ljeCHIPMODEL    chipModel;
    ljtUINT32       chipRevision;
    ljtUINT32       productID;
    ljtUINT32       customerID;
    ljtUINT32       ecoID;
    ljeCHIP_FLAG    chipFlags;
    ljtUINT64       platformFlagBits;
} ljsHAL_CHIPIDENTITY;

/******************************************************************************
 ******************************** ljoHAL Object *******************************
 ******************************************************************************/

/* Construct a new ljoHAL object. */
ljeSTATUS
ljoHAL_ConstructEx(IN ljtPOINTER Context, IN ljoOS Os, OUT ljoHAL *Hal);

/* Destroy an ljoHAL object. */
ljeSTATUS
ljoHAL_DestroyEx(IN ljoHAL Hal);

/* Empty function for compatibility. */
ljeSTATUS
ljoHAL_Construct(IN ljtPOINTER Context, IN ljoOS Os, OUT ljoHAL *Hal);

/* Empty function for compatibility. */
ljeSTATUS
ljoHAL_Destroy(IN ljoHAL Hal);

/* Get HAL options */
ljeSTATUS
ljoHAL_GetOption(IN ljoHAL Hal, IN ljeOPTION Option);

ljeSTATUS
ljoHAL_FrameInfoOps(IN ljoHAL Hal,
                    IN ljeFRAMEINFO FrameInfo,
                    IN ljeFRAMEINFO_OP Op,
                    IN OUT ljtUINT *Val);

/* Set HAL options */
ljeSTATUS
ljoHAL_SetOption(IN ljoHAL Hal, IN ljeOPTION Option, IN ljtBOOL Value);

ljeSTATUS
ljoHAL_GetHardware(IN ljoHAL Hal, OUT ljoHARDWARE *Hw);

#if ljdENABLE_2D
/* Get pointer to ljo2D object. */
ljeSTATUS
ljoHAL_Get2DEngine(IN ljoHAL Hal, OUT ljo2D *Engine);
#endif

#if ljdENABLE_3D
ljeSTATUS
ljoHAL_GetSpecialHintData(IN ljoHAL Hal, OUT ljtINT *Hint);
/*
 ** Deprecated(Don't use it), keep it here for external library(libgcu.so)
 */
ljeSTATUS
ljoHAL_Get3DEngine(IN ljoHAL Hal, OUT ljo3D *Engine);
#endif /* ljdENABLE_3D */

ljeSTATUS
ljoHAL_GetProductName(IN ljoHAL Hal,
                      OUT ljtSTRING *ProductName,
                      OUT ljtUINT *PID );
ljeSTATUS
ljoHAL_GetProductNameWithHardware(IN ljoHARDWARE Hardware,
                                  OUT ljtSTRING *ProductName,
                                  OUT ljtUINT *PID);

ljeSTATUS
ljoHAL_SetFscaleValue(IN ljoHAL Hal,
                      IN ljtUINT CoreIndex,
                      IN ljtUINT FscaleValue,
                      IN ljtUINT ShaderFscaleValue);

ljeSTATUS
ljoHAL_CancelJob(ljoHAL Hal);

ljeSTATUS
ljoHAL_GetFscaleValue(OUT ljtUINT *FscaleValue,
                      OUT ljtUINT *MinFscaleValue,
                      OUT ljtUINT *MaxFscaleValue);

ljeSTATUS
ljoHAL_SetBltNP2Texture(ljtBOOL enable);

ljeSTATUS
ljoHAL_ExportVideoMemory(IN ljtUINT32 Handle,
                         IN ljtUINT32 Flags,
                         OUT ljtINT32 *FD);

ljeSTATUS
ljoHAL_NameVideoMemory(IN ljtUINT32 Handle, OUT ljtUINT32 *Name);

ljeSTATUS
ljoHAL_ImportVideoMemory(IN ljtUINT32 Name, OUT ljtUINT32 *Handle);

ljeSTATUS
ljoHAL_GetVideoMemoryFd(IN ljtUINT32 Handle, OUT ljtINT *Fd);

ljeSTATUS
ljoHAL_GetExportedVideoMemoryFd(IN ljtUINT32 Handle, OUT ljtINT *Fd);

/* Verify whether the specified feature is available in hardware. */
ljeSTATUS
ljoHAL_IsFeatureAvailable(IN ljoHAL Hal, IN ljeFEATURE Feature);

ljeSTATUS
ljoHAL_IsFeatureAvailableWithHardware(IN ljoHARDWARE Hardware, IN ljeFEATURE Feature);

ljeSTATUS
ljoHAL_IsFeatureAvailable1(IN ljoHAL Hal, IN ljeFEATURE Feature);

/* Query the identity of the hardware. */
ljeSTATUS
ljoHAL_QueryChipIdentity(IN ljoHAL Hal,
                         OUT ljeCHIPMODEL *ChipModel,
                         OUT ljtUINT32 *ChipRevision,
                         OUT ljtUINT32 *ChipFeatures,
                         OUT ljtUINT32 *ChipMinorFeatures);

ljeSTATUS
ljoHAL_QueryChipIdentityWithHardware(IN ljoHARDWARE Hardware,
                                     OUT ljeCHIPMODEL *ChipModel,
                                     OUT ljtUINT32 *ChipRevision);

ljeSTATUS
ljoHAL_QueryChipIdentityEx(IN ljoHAL Hal,
                           IN ljtUINT32 SizeOfParam,
                           OUT ljsHAL_CHIPIDENTITY *ChipIdentity);

ljeSTATUS
ljoHAL_QuerySuperTileMode(OUT ljtUINT32_PTR SuperTileMode);

ljeSTATUS
ljoHAL_QueryChipAxiBusWidth(OUT ljtBOOL *AXI128Bits);

ljeSTATUS
ljoHAL_QueryMultiGPUAffinityConfig(IN ljeHARDWARE_TYPE Type,
                                   OUT ljeMULTI_PROCESSOR_MODE *Mode,
                                   OUT ljtUINT32_PTR CoreIndex);

ljeSTATUS
ljoHAL_QueryHwDeviceIdByEnv(IN ljoHAL Hal,
                            OUT ljtUINT32 *DeviceID,
                            OUT ljtBOOL *HasEnv);

ljeSTATUS
ljoHAL_QueryMultiGPUAffinityConfigfor3D(
    IN ljeHARDWARE_TYPE Type,
    IN ljtINT32 patchId,
    OUT ljeMULTI_PROCESSOR_MODE *Mode,
    OUT ljtUINT32_PTR CoreIndex
    );

ljeSTATUS
ljoHAL_QuerySRAM(IN ljoHAL Hal,
                 IN ljePOOL Type,
                 OUT ljtUINT32 *Size,
                 OUT ljtADDRESS *GPUVirtAddr,
                 OUT ljtPHYS_ADDR_T *GPUPhysAddr,
                 OUT ljtUINT32 *GPUPhysName,
                 OUT ljtPHYS_ADDR_T *CPUPhysAddr);

#ifdef LINUX
ljtINT32
ljoOS_EndRecordAllocation(void);
void
ljoOS_RecordAllocation(void);
void
ljoOS_AddRecordAllocation(ljtSIZE_T Size);
#endif

/* Query the amount of video memory. */
ljeSTATUS
ljoHAL_QueryVideoMemory(IN ljoHAL Hal,
                        OUT ljtUINT32 *InternalPhysName,
                        OUT ljtSIZE_T *InternalSize,
                        OUT ljtUINT32 *ExternalPhysName,
                        OUT ljtSIZE_T *ExternalSize,
                        OUT ljtUINT32 *ContiguousPhysName,
                        OUT ljtSIZE_T *ContiguousSize);

/* Map video memory. */
ljeSTATUS
ljoHAL_MapMemory(IN ljoHAL Hal, IN ljtUINT32 PhysName,
                 IN ljtSIZE_T NumberOfBytes, OUT ljtPOINTER *Logical);

/* Unmap video memory. */
ljeSTATUS
ljoHAL_UnmapMemory(IN ljoHAL Hal, IN ljtUINT32 PhysName,
                   IN ljtSIZE_T NumberOfBytes, IN ljtPOINTER Logical);

/* Schedule an unmap of a buffer mapped through its physical address. */
ljeSTATUS
ljoHAL_ScheduleUnmapMemory(IN ljoHAL Hal, IN ljtUINT32 PhysName,
                           IN ljtSIZE_T NumberOfBytes, IN ljtPOINTER Logical);

/* Allocate video memory. */
ljeSTATUS
ljoOS_AllocateVideoMemory(IN ljoOS Os,
                          IN ljtBOOL InUserSpace,
                          IN ljtBOOL InCacheable,
                          IN OUT ljtSIZE_T *Bytes,
                          OUT ljtUINT32 *Address,
                          OUT ljtPOINTER *Logical,
                          OUT ljtPOINTER *Handle);

/* Free video memory. */
ljeSTATUS
ljoOS_FreeVideoMemory(IN ljoOS Os, IN ljtPOINTER Handle);

/* Lock video memory. */
ljeSTATUS
ljoOS_LockVideoMemory(IN ljoOS Os,
                      IN ljtPOINTER Handle,
                      IN ljtBOOL InUserSpace,
                      IN ljtBOOL InCacheable,
                      OUT ljtUINT32 *Address,
                      OUT ljtPOINTER *Logical);

/* Commit the current command buffer. */
ljeSTATUS
ljoHAL_Commit(IN ljoHAL Hal, IN ljtBOOL Stall);

#if ljdENABLE_3D
/* Sencd fence command. */
ljeSTATUS
ljoHAL_SendFence(IN ljoHAL Hal);

/* Send fence command for GL_TIME_ELAPSED. */
ljeSTATUS
ljoHAL_TimeQuery_SendFence(IN ljoHAL Hal, IN ljtADDRESS physical);
/* Wait fence result for GL_TIME_ELAPSED. */
ljeSTATUS
ljoHAL_TimeQuery_WaitFence(IN ljoHAL Hal,
                           IN ljsSURF_NODE_PTR node,
                           IN ljtPOINTER nodeHeaderLocked,
                           IN ljtPOINTER logical);
#endif /* ljdENABLE_3D */

/* Query the tile capabilities. */
ljeSTATUS
ljoHAL_QueryTiled(IN ljoHAL Hal,
                  OUT ljtINT32 *TileWidth2D,
                  OUT ljtINT32 *TileHeight2D,
                  OUT ljtINT32 *TileWidth3D,
                  OUT ljtINT32 *TileHeight3D);

ljeSTATUS
ljoHAL_Compact(IN ljoHAL Hal);

#if LJMICRO_PROFILER_SYSTEM_MEMORY
ljeSTATUS
ljoHAL_ProfileStart(IN ljoHAL Hal);

ljeSTATUS
ljoHAL_ProfileEnd(IN ljoHAL Hal, IN ljtCONST_STRING Title);
#endif

/* Power Management */
ljeSTATUS
ljoHAL_SetPowerManagementState(IN ljoHAL Hal,
                               IN ljeCHIPPOWERSTATE State);

ljeSTATUS
ljoHAL_QueryPowerManagementState(IN ljoHAL Hal,
                                 OUT ljeCHIPPOWERSTATE *State);

/* Set the filter type for filter blit. */
ljeSTATUS
ljoHAL_SetFilterType(IN ljoHAL Hal,
                     IN ljeFILTER_TYPE FilterType);

/* Call the kernel HAL layer. */
ljeSTATUS
ljoHAL_Call(IN ljoHAL Hal,
            IN OUT ljsHAL_INTERFACE_PTR Interface);

/* Schedule an event. */
ljeSTATUS
ljoHAL_ScheduleEvent(IN ljoHAL Hal,
                     IN OUT ljsHAL_INTERFACE_PTR Interface);

/* Request a start/stop timestamp. */
ljeSTATUS
ljoHAL_SetTimer(IN ljoHAL Hal, IN ljtUINT32 Index, IN ljtBOOL Start);

/* Get Time delta from a Timer in microseconds. */
ljeSTATUS
ljoHAL_GetTimerTime(IN ljoHAL Hal, IN ljtUINT32 Timer,
                    OUT ljtINT32_PTR TimeDelta);

/* set timeout value. */
ljeSTATUS
ljoHAL_SetTimeOut(IN ljoHAL Hal, IN ljtUINT32 timeOut);

ljeSTATUS
ljoHAL_SetHardwareType(IN ljoHAL Hal,
                       IN ljeHARDWARE_TYPE HardwardType);

ljeSTATUS
ljoHAL_GetHardwareType(IN ljoHAL Hal,
                       OUT ljeHARDWARE_TYPE *HardwardType);

ljeSTATUS
ljoHAL_QueryChipCount(IN ljoHAL Hal, OUT ljtINT32 *Count);

ljeSTATUS
ljoHAL_Query3DCoreCount(IN ljoHAL Hal, OUT ljtUINT32 *Count);

ljeSTATUS
ljoHAL_Query2DCoreCount(IN ljoHAL Hal, OUT ljtUINT32 *Count);

ljeSTATUS
ljoHAL_QueryCluster(IN ljoHAL Hal,
                    OUT ljtINT32 *ClusterMinID,
                    OUT ljtINT32 *ClusterMaxID,
                    OUT ljtUINT32 *ClusterCount,
                    OUT ljtUINT32 *ClusterIDWidth);

ljeSTATUS
ljoHAL_QueryUscAttribCacheRatio(IN ljoHAL Hal,
                                OUT ljtUINT32 *UscAttribCacheRatio);

ljeSTATUS
ljoHAL_QueryCoreCount(IN ljoHAL Hal,
                      IN ljeHARDWARE_TYPE Type,
                      OUT ljtUINT *Count,
                      OUT ljtUINT_PTR ChipIDs);

ljeSTATUS
ljoHAL_QuerySeparated2D(IN ljoHAL Hal);

ljeSTATUS
ljoHAL_QueryHybrid2D(IN ljoHAL Hal);

ljeSTATUS
ljoHAL_Is3DAvailable(IN ljoHAL Hal);

/* Get pointer to ljoVG object. */
ljeSTATUS
ljoHAL_GetVGEngine(IN ljoHAL Hal, OUT ljoVG *Engine);

ljeSTATUS
ljoHAL_QueryChipLimits(IN ljoHAL Hal,
                       IN ljtINT32 Chip,
                       OUT ljsHAL_LIMITS *Limits);

ljeSTATUS
ljoHAL_QueryChipFeature(IN ljoHAL Hal, IN ljtINT32 Chip, IN ljeFEATURE Feature);

ljeSTATUS
ljoHAL_SetDeviceIndex(IN ljoHAL Hal, IN ljtUINT32 DeviceIndex);

ljeSTATUS
ljoHAL_GetCurrentDeviceIndex(IN ljoHAL Hal, OUT ljtUINT32 *DeviceIndex);

ljeSTATUS
ljoHAL_SetCoreIndex(IN ljoHAL Hal, IN ljtUINT32 Core);

ljeSTATUS
ljoHAL_GetCurrentCoreIndex(IN ljoHAL Hal, OUT ljtUINT32 *Core);

ljeSTATUS
ljoHAL_InitCoreIndexByType(IN ljoHAL Hal,
                           IN ljeHARDWARE_TYPE Type,
                           IN ljtBOOL Init,
                           OUT ljtUINT32 *CoreIndex);

ljeSTATUS
ljoHAL_ConvertCoreIndexGlobal(IN ljoHAL Hal,
                              IN ljeHARDWARE_TYPE Type,
                              IN ljtUINT32 CoreCount,
                              IN ljtUINT32 *LocalCoreIndexs,
                              OUT ljtUINT32 *GlobalCoreIndexs);

ljeSTATUS
ljoHAL_ConvertCoreIndexLocal(IN ljoHAL Hal,
                             IN ljeHARDWARE_TYPE Type,
                             IN ljtUINT32 CoreCount,
                             IN ljtUINT32 *GlobalCoreIndexs,
                             OUT ljtUINT32 *LocalCoreIndexs);

ljeSTATUS
ljoHAL_SelectChannel(IN ljoHAL Hal, IN ljtBOOL Priority, IN ljtUINT32 ChannelId);

ljeSTATUS
ljoHAL_MCFESemaphore(IN ljtUINT32 SemaHandle, IN ljtBOOL SendSema);

ljeSTATUS
ljoHAL_AllocateMCFESemaphore(OUT ljtUINT32 *SemaHandle);

ljeSTATUS
ljoHAL_FreeMCFESemaphore(IN ljtUINT32 SemaHandle);

/*----------------------------------------------------------------------------*/
/*----- Shared Buffer --------------------------------------------------------*/

/* Create shared buffer. */
ljeSTATUS
ljoHAL_CreateShBuffer(IN ljtUINT32 Size, OUT ljtSHBUF *ShBuf);

/* Destroy shared buffer. */
ljeSTATUS
ljoHAL_DestroyShBuffer(IN ljtSHBUF ShBuf);

/* Map shared buffer to current process. */
ljeSTATUS
ljoHAL_MapShBuffer(IN ljtSHBUF ShBuf);

/* Write user data to shared buffer. */
ljeSTATUS
ljoHAL_WriteShBuffer(IN ljtSHBUF ShBuf, IN ljtCONST_POINTER Data, IN ljtUINT32 ByteCount);

/* Read user data from shared buffer. */
ljeSTATUS
ljoHAL_ReadShBuffer(IN ljtSHBUF ShBuf,
                    IN ljtPOINTER Data,
                    IN ljtUINT32 BytesCount,
                    OUT ljtUINT32 *BytesRead);

/* Config power management to be enabled or disabled. */
ljeSTATUS
ljoHAL_ConfigPowerManagement(IN ljtBOOL Enable, OUT ljtBOOL *OldValue);

ljeSTATUS
ljoHAL_AllocateVideoMemory(IN ljtUINT Alignment,
                           IN ljeVIDMEM_TYPE Type,
                           IN ljtUINT32 Flag,
                           IN OUT ljePOOL *Pool,
                           IN OUT ljtSIZE_T *Bytes,
                           OUT ljtUINT32_PTR Node);

ljeSTATUS
ljoHAL_LockVideoMemory(IN ljtUINT32 Node,
                       IN ljtBOOL Cacheable,
                       IN ljeENGINE engine,
                       OUT ljtADDRESS *Address,
                       OUT ljtPOINTER *Logical);

ljeSTATUS
ljoHAL_LockVideoMemoryEx(IN ljtUINT32 Node,
                         IN ljtBOOL Cacheable,
                         IN ljeENGINE engine,
                         IN ljeLOCK_VIDEO_MEMORY_OP Op,
                         OUT ljtADDRESS *Address,
                         OUT ljtPOINTER *Logical);

ljeSTATUS
ljoHAL_UnlockVideoMemory(IN ljtUINT32 Node, IN ljeVIDMEM_TYPE Type, IN ljeENGINE engine);

ljeSTATUS
ljoHAL_UnlockVideoMemoryEX(IN ljtUINT32 Node,
                           IN ljeVIDMEM_TYPE Type,
                           IN ljeENGINE Engine,
                           IN ljtBOOL Sync,
                           IN ljeLOCK_VIDEO_MEMORY_OP Op);

ljeSTATUS
ljoHAL_ReleaseVideoMemory(IN ljtUINT32 Node);

#if ljdENABLE_3D || ljdENABLE_VG
/* Query the target capabilities. */
ljeSTATUS
ljoHAL_QueryTargetCaps(IN ljoHAL Hal,
                       OUT ljtUINT *MaxWidth,
                       OUT ljtUINT *MaxHeight,
                       OUT ljtUINT *MultiTargetCount,
                       OUT ljtUINT *MaxSamples);
#endif

ljeSTATUS
ljoHAL_PrepareVideoMemory(IN ljtUINT32 Node);

ljeSTATUS
ljoHAL_FinishVideoMemory(IN ljtUINT32 Node);

ljeSTATUS
ljoHAL_WrapUserMemory(IN ljsUSER_MEMORY_DESC_PTR UserMemoryDesc,
                      IN ljeVIDMEM_TYPE Type,
                      OUT ljtUINT32_PTR Node);

ljeSTATUS
ljoHAL_QueryResetTimeStamp(OUT ljtUINT64_PTR ResetTimeStamp,
                           OUT ljtUINT64_PTR ContextID);

ljeSTATUS
ljoHAL_WaitFence(IN ljtUINT32 Handle, IN ljtUINT32 TimeOut);

ljeSTATUS
ljoHAL_ScheduleSignal(IN ljtSIGNAL Signal,
                      IN ljtSIGNAL AuxSignal,
                      IN ljtINT ProcessID,
                      IN ljeKERNEL_WHERE FromWhere);

ljeSTATUS
ljoHAL_GetGraphicBufferFd(IN ljtUINT32 Node[3],
                          IN ljtSHBUF ShBuf,
                          IN ljtSIGNAL Signal,
                          OUT ljtINT32 *Fd);

ljeSTATUS
ljoHAL_AlignToTile(IN OUT ljtUINT32 *Width,
                   IN OUT ljtUINT32 *Height,
                   IN ljeSURF_TYPE Type,
                   IN ljeSURF_FORMAT Format);

ljeSTATUS
ljoHAL_GetLastCommitStatus(IN ljoHAL Hal, OUT ljtBOOL *Pending);

ljeSTATUS
ljoHAL_SetLastCommitStatus(IN ljoHAL Hal, IN ljtBOOL Pending);

ljeSTATUS
ljoHAL_CommitDone(IN ljoHAL Hal);

ljeSTATUS
ljoHAL_IsFlatMapped(IN ljtPHYS_ADDR_T PhysicalAddress,
                    OUT ljtADDRESS *Address);

ljeSTATUS
ljoHAL_QueryMCFESemaphoreCapacity(IN ljoHAL Hal,
                                  OUT ljtUINT32 *Capacity);

#if ljdENABLE_MP_SWITCH
ljeSTATUS
ljoHAL_SwitchMpMode(ljoHAL Hal);
#endif

ljeSTATUS
ljoHAL_CommandBufferAutoCommit(ljoHAL Hal, ljtBOOL AutoCommit);

ljeSTATUS
ljoHAL_CommandBufferAutoSync(ljoHAL Hal, ljtBOOL AutoSync);


#if ljdENABLE_MULTI_DEVICE_MANAGEMENT
ljeSTATUS
ljoHAL_SwitchContext(IN ljoHAL Hal,
                     IN ljoHARDWARE Hardware,
                     OUT ljoHARDWARE *SavedHardware,
                     OUT ljeHARDWARE_TYPE *SavedType,
                     OUT ljtUINT32 *SavedHwDeviceIndex,
                     OUT ljtUINT32 *SavedCoreIndex);

ljeSTATUS
ljoHAL_RestoreContext(IN ljoHAL Hal,
                      IN ljoHARDWARE Hardware,
                      IN ljeHARDWARE_TYPE Type,
                      IN ljtUINT32 HwDeviceIndex,
                      IN ljtUINT32 CoreIndex);

ljeSTATUS
ljoHAL_ShowDeviceInfo(ljoHAL Hal);
#endif

/******************************************************************************
 ********************************** ljoOS Object ******************************
 ******************************************************************************/
/* Lock PLS access */
ljeSTATUS
ljoOS_LockPLS(void);

/* Unlock PLS access */
ljeSTATUS
ljoOS_UnLockPLS(void);

/* Get PLS value for given key */
ljtPOINTER
ljoOS_GetPLSValue(IN ljePLS_VALUE key);

/* Set PLS value of a given key */
void
ljoOS_SetPLSValue(IN ljePLS_VALUE key, OUT ljtPOINTER value);

/* Lock GL FE compiler access */
ljeSTATUS
ljoOS_LockGLFECompiler(void);

/* Unlock GL FE compiler access */
ljeSTATUS
ljoOS_UnLockGLFECompiler(void);

/* Lock CL FE compiler access */
ljeSTATUS
ljoOS_LockCLFECompiler(void);

/* Unlock CL FE compiler access */
ljeSTATUS
ljoOS_UnLockCLFECompiler(void);

/* Get access to the thread local storage. */
ljeSTATUS
ljoOS_GetTLS(OUT ljsTLS_PTR *TLS);

/* Copy the TLS from a source thread. */
ljeSTATUS
ljoOS_CopyTLS(IN ljsTLS_PTR Source);

/* Query the thread local storage. */
ljeSTATUS
ljoOS_QueryTLS(OUT ljsTLS_PTR *TLS);

/* Get access to driver tls. */
ljeSTATUS
ljoOS_GetDriverTLS(IN ljeTLS_KEY Key,
                   OUT ljsDRIVER_TLS_PTR *TLS);

/*
 * Set driver tls.
 * May cause memory leak if 'destructor' not set.
 */
ljeSTATUS
ljoOS_SetDriverTLS(IN ljeTLS_KEY Key, IN ljsDRIVER_TLS *TLS);

/* Destroy the objects associated with the current thread. */
void
ljoOS_FreeThreadData(void);

/* Empty function for compatibility. */
ljeSTATUS
ljoOS_Construct(IN ljtPOINTER Context, OUT ljoOS *Os);

/* Empty function for compatibility. */
ljeSTATUS
ljoOS_Destroy(IN ljoOS Os);

/* Deprecated API: please use ljoHAL_GetBaseAddr() instead.
 **                This API was kept only for legacy BSP usage.
 **
 ** Get the base address for the physical memory.
 */
ljeSTATUS
ljoOS_GetBaseAddress(IN ljoOS Os, OUT ljtUINT32_PTR BaseAddress);

/* Allocate memory from the heap. */
ljeSTATUS
ljoOS_Allocate(IN ljoOS Os, IN ljtSIZE_T Bytes,
               OUT ljtPOINTER *Memory);

ljeSTATUS
ljoOS_Realloc(IN ljoOS Os,
              IN ljtSIZE_T Bytes,
              IN ljtSIZE_T OrgBytes,
              OUT ljtPOINTER *Memory);

/* Get allocated memory size. */
ljeSTATUS
ljoOS_GetMemorySize(IN ljoOS Os, IN ljtPOINTER Memory,
                    OUT ljtSIZE_T_PTR MemorySize);

/* Free allocated memory. */
ljeSTATUS
ljoOS_Free(IN ljoOS Os, IN ljtPOINTER Memory);

/* Allocate memory. */
ljeSTATUS
ljoOS_AllocateSharedMemory(IN ljoOS Os, IN ljtSIZE_T Bytes,
                           OUT ljtPOINTER *Memory);

/* Free memory. */
ljeSTATUS
ljoOS_FreeSharedMemory(IN ljoOS Os, IN ljtPOINTER Memory);

/* Allocate memory. */
ljeSTATUS
ljoOS_AllocateMemory(IN ljoOS Os, IN ljtSIZE_T Bytes,
                     OUT ljtPOINTER *Memory);

/* Realloc memory. */
ljeSTATUS
ljoOS_ReallocMemory(IN ljoOS Os, IN ljtSIZE_T Bytes,
                    IN ljtSIZE_T OrgBytes, OUT ljtPOINTER *Memory);

/* Free memory. */
ljeSTATUS
ljoOS_FreeMemory(IN ljoOS Os, IN ljtPOINTER Memory);

/* Device I/O Control call to the kernel HAL layer. */
ljeSTATUS
ljoOS_DeviceControl(IN ljoOS Os,
                    IN ljtUINT32 IoControlCode,
                    IN ljtPOINTER InputBuffer,
                    IN ljtSIZE_T InputBufferSize,
                    IN ljtPOINTER OutputBuffer,
                    IN ljtSIZE_T OutputBufferSize);

ljeSTATUS
ljoOS_DeviceControlWithDRM(IN ljoOS Os,
                    IN ljsHAL_INTERFACE_PTR InputBuffer,
                    IN ljsHAL_INTERFACE_PTR OutputBuffer);


#define ljmOS_SAFE_FREE(os, mem) \
    ljoOS_Free(os, mem); \
    mem = ljvNULL

#define ljmOS_SAFE_FREE_SHARED_MEMORY(os, mem) \
    ljoOS_FreeSharedMemory(os, mem); \
    mem = ljvNULL

#define ljmkOS_SAFE_FREE(os, mem) \
    ljkOS_Free(os, mem); \
    mem = ljvNULL

#define ljdMAX_PATH 512

#define ljdMAX_ARGUMENT_SIZE 1024
#define ljdMAX_ARGUMENT_COUNT 64

/* Open a file. */
ljeSTATUS
ljoOS_Open(IN ljoOS Os,
           IN ljtCONST_STRING FileName,
           IN ljeFILE_MODE Mode,
           OUT ljtFILE *File);

/* Close a file. */
ljeSTATUS
ljoOS_Close(IN ljoOS Os, IN ljtFILE File);

/* Remove a file. */
ljeSTATUS
ljoOS_Remove(IN ljoOS Os, IN ljtCONST_STRING FileName);

/* Read data from a file. */
ljeSTATUS
ljoOS_Read(IN ljoOS Os,
           IN ljtFILE File,
           IN ljtSIZE_T ByteCount,
           IN ljtPOINTER Data,
           OUT ljtSIZE_T *ByteRead);

/* Write data to a file. */
ljeSTATUS
ljoOS_Write(IN ljoOS Os,
            IN ljtFILE File,
            IN ljtSIZE_T ByteCount,
            IN ljtCONST_POINTER Data);

/* Flush data to a file. */
ljeSTATUS
ljoOS_Flush(IN ljoOS Os, IN ljtFILE File);

/* Close a file descriptor. */
ljeSTATUS
ljoOS_CloseFD(IN ljoOS Os, IN ljtINT FD);

/* Scan a file. */
ljeSTATUS
ljoOS_FscanfI(IN ljoOS Os,
              IN ljtFILE File,
              IN ljtCONST_STRING Format,
              OUT ljtUINT *result);

/* Dup file descriptor to another. */
ljeSTATUS
ljoOS_DupFD(IN ljoOS Os,
            IN ljtINT FD,
            OUT ljtINT *FD2);

/* Lock a file. */
ljeSTATUS
ljoOS_LockFile(IN ljoOS Os,
               IN ljtFILE File,
               IN ljtBOOL Shared,
               IN ljtBOOL Block);

/* Unlock a file. */
ljeSTATUS
ljoOS_UnlockFile(IN ljoOS Os, IN ljtFILE File);

/* Create an endpoint for communication. */
ljeSTATUS
ljoOS_Socket(IN ljoOS Os,
             IN ljtINT Domain,
             IN ljtINT Type,
             IN ljtINT Protocol,
             OUT ljtINT *SockFd);

/* Close a socket. */
ljeSTATUS
ljoOS_CloseSocket(IN ljoOS Os, IN ljtINT SockFd);

/* Initiate a connection on a socket. */
ljeSTATUS
ljoOS_Connect(IN ljoOS Os,
              IN ljtINT SockFd,
              IN ljtCONST_POINTER HostName,
              IN ljtUINT Port);

/* Shut down part of connection on a socket. */
ljeSTATUS
ljoOS_Shutdown(IN ljoOS Os, IN ljtINT SockFd, IN ljtINT How);

/* Send a message on a socket. */
ljeSTATUS
ljoOS_Send(IN ljoOS Os,
           IN ljtINT SockFd,
           IN ljtSIZE_T ByteCount,
           IN ljtCONST_POINTER Data,
           IN ljtINT Flags);

/* Initiate a connection on a socket. */
ljeSTATUS
ljoOS_WaitForSend(IN ljoOS Os, IN ljtINT SockFd,
                  IN ljtINT Seconds, IN ljtINT MicroSeconds);

/* Get environment variable value. */
ljeSTATUS
ljoOS_GetEnv(IN ljoOS Os, IN ljtCONST_STRING VarName, OUT ljtSTRING *Value);

/* Set environment variable value. */
ljeSTATUS
ljoOS_SetEnv(IN ljoOS Os, IN ljtCONST_STRING VarName, IN ljtSTRING Value);

/* Get current working directory. */
ljeSTATUS
ljoOS_GetCwd(IN ljoOS Os, IN ljtINT SizeInBytes, OUT ljtSTRING Buffer);

/* Get file status info. */
ljeSTATUS
ljoOS_Stat(IN ljoOS Os, IN ljtCONST_STRING FileName, OUT ljtPOINTER Buffer);

/* Set the current position of a file. */
ljeSTATUS
ljoOS_Seek(IN ljoOS Os, IN ljtFILE File, IN ljtUINT32 Offset, IN ljeFILE_WHENCE Whence);

/* Set the current position of a file. */
ljeSTATUS
ljoOS_SetPos(IN ljoOS Os, IN ljtFILE File, IN ljtUINT32 Position);

/* Get the current position of a file. */
ljeSTATUS
ljoOS_GetPos(IN ljoOS Os, IN ljtFILE File, OUT ljtUINT32 *Position);

/* Same as strstr. */
ljeSTATUS
ljoOS_StrStr(IN ljtCONST_STRING String,
             IN ljtCONST_STRING SubString,
             OUT ljtSTRING *Output);

/* Find the last occurrence of a character inside a string. */
ljeSTATUS
ljoOS_StrFindReverse(IN ljtCONST_STRING String,
                     IN ljtINT8 Character,
                     OUT ljtSTRING *Output);

ljeSTATUS
ljoOS_StrDup(IN ljoOS Os, IN ljtCONST_STRING String, OUT ljtSTRING *Target);

/* Copy a string. */
ljeSTATUS
ljoOS_StrCopySafe(IN ljtSTRING Destination,
                  IN ljtSIZE_T DestinationSize,
                  IN ljtCONST_STRING Source);

/* Append a string. */
ljeSTATUS
ljoOS_StrCatSafe(IN ljtSTRING Destination,
                 IN ljtSIZE_T DestinationSize,
                 IN ljtCONST_STRING Source);

/* Compare two strings. */
ljeSTATUS
ljoOS_StrCmp(IN ljtCONST_STRING String1, IN ljtCONST_STRING String2);

/* Compare characters of two strings. */
ljeSTATUS
ljoOS_StrNCmp(IN ljtCONST_STRING String1,
              IN ljtCONST_STRING String2,
              IN ljtSIZE_T Count);

/* Convert string to float. */
ljeSTATUS
ljoOS_StrToFloat(IN ljtCONST_STRING String, OUT ljtFLOAT *Float);

/* Convert string to double. */
ljeSTATUS
ljoOS_StrToDouble(IN ljtCONST_STRING String, OUT ljtDOUBLE* Double);

/* Convert hex string to integer. */
ljeSTATUS
ljoOS_HexStrToInt(IN ljtCONST_STRING String, OUT ljtINT *Int);

/* Convert hex string to float. */
ljeSTATUS
ljoOS_HexStrToFloat(IN ljtCONST_STRING String, OUT ljtFLOAT *Float);

/* Convert string to integer. */
ljeSTATUS
ljoOS_StrToInt(IN ljtCONST_STRING String, OUT ljtINT *Int);

ljeSTATUS
ljoOS_MemCmp(IN ljtCONST_POINTER Memory1,
             IN ljtCONST_POINTER Memory2,
             IN ljtSIZE_T Bytes);

ljeSTATUS
ljoOS_PrintStrSafe(OUT ljtSTRING String,
                   IN ljtSIZE_T StringSize,
                   IN OUT ljtUINT *Offset,
                   IN ljtCONST_STRING Format,
                   ...)
CHECK_PRINTF_FORMAT(4, 5);

ljeSTATUS
ljoOS_LoadLibrary(IN ljoOS Os, IN ljtCONST_STRING Library, OUT ljtHANDLE *Handle);

ljeSTATUS
ljoOS_FreeLibrary(IN ljoOS Os, IN ljtHANDLE Handle);

ljeSTATUS
ljoOS_GetProcAddress(IN ljoOS Os,
                     IN ljtHANDLE Handle,
                     IN ljtCONST_STRING Name,
                     OUT ljtPOINTER *Function);

ljeSTATUS
ljoOS_Compact(IN ljoOS Os);

ljeSTATUS
ljoOS_AddSignalHandler(IN ljeSignalHandlerType SignalHandlerType);

#if LJMICRO_PROFILER_SYSTEM_MEMORY
ljeSTATUS
ljoOS_ProfileStart(IN ljoOS Os);

ljeSTATUS
ljoOS_ProfileEnd(IN ljoOS Os, IN ljtCONST_STRING Title);

ljeSTATUS
ljoOS_SetProfileSetting(IN ljoOS Os,
                        IN ljtBOOL Enable,
                        IN ljeProfilerMode ProfileMode,
                        IN ljtCONST_STRING FileName);
#endif

/* Get the amount of physical system memory */
ljeSTATUS
ljoOS_GetPhysicalSystemMemorySize(OUT ljtUINT64 *PhysicalSystemMemorySize);

/* Query the video memory. */
ljeSTATUS
ljoOS_QueryVideoMemory(IN ljoOS Os,
                       OUT ljtUINT32 *InternalPhysName,
                       OUT ljtSIZE_T *InternalSize,
                       OUT ljtUINT32 *ExternalPhysName,
                       OUT ljtSIZE_T *ExternalSize,
                       OUT ljtUINT32 *ContiguousPhysName,
                       OUT ljtSIZE_T *ContiguousSize);

ljeSTATUS
ljoOS_QueryCurrentProcessName(OUT ljtSTRING Name, IN ljtSIZE_T Size);

ljeSTATUS
ljoOS_QueryCurrentProcessArguments(OUT ljtCHAR Argv[ljdMAX_ARGUMENT_COUNT][ljdMAX_ARGUMENT_SIZE],
                                   OUT ljtUINT32 *Argc,
                                   IN  ljtUINT32 MaxArgc,
                                   IN  ljtUINT32 MaxSizePerArg);

/*----------------------------------------------------------------------------*/
/*----- Atoms ----------------------------------------------------------------*/

/* Construct an atom. */
ljeSTATUS
ljoOS_AtomConstruct(IN ljoOS Os, OUT ljsATOM_PTR *Atom);

/* Destroy an atom. */
ljeSTATUS
ljoOS_AtomDestroy(IN ljoOS Os, IN ljsATOM_PTR Atom);

/* Get the 32-bit value protected by an atom. */
ljeSTATUS
ljoOS_AtomGet(IN ljoOS Os, IN ljsATOM_PTR Atom, OUT ljtINT32_PTR Value);

/* Set the 32-bit value protected by an atom. */
ljeSTATUS
ljoOS_AtomSet(IN ljoOS Os, IN ljsATOM_PTR Atom, IN ljtINT32 Value);

/* Increment an atom. */
ljeSTATUS
ljoOS_AtomIncrement(IN ljoOS Os, IN ljsATOM_PTR Atom, OUT ljtINT32_PTR OldValue);

/* Decrement an atom. */
ljeSTATUS
ljoOS_AtomDecrement(IN ljoOS Os, IN ljsATOM_PTR Atom, OUT ljtINT32_PTR OldValue);

ljtHANDLE
ljoOS_GetCurrentProcessID(void);

ljtHANDLE
ljoOS_GetCurrentThreadID(void);

/*----------------------------------------------------------------------------*/
/*----- Time -----------------------------------------------------------------*/

/* Get the number of milliseconds since the system started. */
ljtUINT32
ljoOS_GetTicks(void);

/* Get time in microseconds. */
ljeSTATUS
ljoOS_GetTime(ljtUINT64_PTR Time);

/* Get CPU usage in microseconds. */
ljeSTATUS
ljoOS_GetCPUTime(ljtUINT64_PTR CPUTime);

/* Get memory usage. */
ljeSTATUS
ljoOS_GetMemoryUsage(ljtUINT32_PTR MaxRSS,
                     ljtUINT32_PTR IxRSS,
                     ljtUINT32_PTR IdRSS,
                     ljtUINT32_PTR IsRSS);

/* Delay a number of milliseconds. */
ljeSTATUS
ljoOS_Delay(IN ljoOS Os, IN ljtUINT32 Delay);


/* Delay a number of microseconds. */
ljeSTATUS
ljoOS_DelayUs(IN ljoOS Os, IN ljtUINT32 Delay);

/*----------------------------------------------------------------------------*/
/*----- Threads --------------------------------------------------------------*/

#ifdef _WIN32
/* Cannot include windows.h here because "near" and "far"
 * which are used in ljsDEPTH_INFO, are defined to nothing in WinDef.h.
 * So, use the real value of DWORD and WINAPI, instead.
 * DWORD is unsigned long, and WINAPI is __stdcall.
 * If these two are change in WinDef.h, the following two typdefs
 * need to be changed, too.
 */
typedef unsigned long ljtTHREAD_RETURN;
typedef unsigned long(__stdcall *ljTHREAD_ROUTINE)(void *Argument);
#else
typedef void *ljtTHREAD_RETURN;
typedef void *(*ljTHREAD_ROUTINE)(void *);
#endif

/* Create a new thread. */
ljeSTATUS
ljoOS_CreateThread(IN ljoOS Os,
                   IN ljTHREAD_ROUTINE Worker,
                   IN ljtPOINTER Argument,
                   OUT ljtPOINTER *Thread);

/* Close a thread. */
ljeSTATUS
ljoOS_CloseThread(IN ljoOS Os, IN ljtPOINTER Thread);

/*----------------------------------------------------------------------------*/
/*----- Mutexes --------------------------------------------------------------*/

/* Create a new mutex. */
ljeSTATUS
ljoOS_CreateMutex(IN ljoOS Os, OUT ljtPOINTER *Mutex);

/* Delete a mutex. */
ljeSTATUS
ljoOS_DeleteMutex(IN ljoOS Os, IN ljtPOINTER Mutex);

/* Acquire a mutex. */
ljeSTATUS
ljoOS_AcquireMutex(IN ljoOS Os, IN ljtPOINTER Mutex, IN ljtUINT32 Timeout);

/* Release a mutex. */
ljeSTATUS
ljoOS_ReleaseMutex(IN ljoOS Os, IN ljtPOINTER Mutex);

/*----------------------------------------------------------------------------*/
/*----- Signals --------------------------------------------------------------*/

/* Create a signal. */
ljeSTATUS
ljoOS_CreateSignal(IN ljoOS Os, IN ljtBOOL ManualReset, OUT ljtSIGNAL *Signal);

/* Destroy a signal. */
ljeSTATUS
ljoOS_DestroySignal(IN ljoOS Os, IN ljtSIGNAL Signal);

/* Signal a signal. */
ljeSTATUS
ljoOS_Signal(IN ljoOS Os, IN ljtSIGNAL Signal, IN ljtBOOL State);

/* Wait for a signal. */
ljeSTATUS
ljoOS_WaitSignal(IN ljoOS Os, IN ljtSIGNAL Signal, IN ljtUINT32 Wait);

/* Map a signal from another process */
ljeSTATUS
ljoOS_MapSignal(IN ljtSIGNAL RemoteSignal, OUT ljtSIGNAL *LocalSignal);

/* Unmap a signal mapped from another process */
ljeSTATUS
ljoOS_UnmapSignal(IN ljtSIGNAL Signal);

/*----------------------------------------------------------------------------*/
/*----- Android Native Fence -------------------------------------------------*/

/* Create native fence. */
ljeSTATUS
ljoOS_CreateNativeFence(IN ljoOS Os, IN ljtSIGNAL Signal, OUT ljtINT *FenceFD);

/* (CPU) Wait on native fence. */
ljeSTATUS
ljoOS_ClientWaitNativeFence(IN ljoOS Os, IN ljtINT FenceFD, IN ljtUINT32 Timeout);

/* (GPU) Wait on native fence. */
ljeSTATUS
ljoOS_WaitNativeFence(IN ljoOS Os, IN ljtINT FenceFD, IN ljtUINT32 Timeout);

/*----------------------------------------------------------------------------*/
/*----- Memory Access and Cache ----------------------------------------------*/

/* Write a register. */
ljeSTATUS
ljoOS_WriteRegister(IN ljoOS Os, IN ljtUINT32 Address, IN ljtUINT32 Data);

/* Read a register. */
ljeSTATUS
ljoOS_ReadRegister(IN ljoOS Os, IN ljtUINT32 Address, OUT ljtUINT32 *Data);

ljeSTATUS
ljoOS_CacheClean(IN ljoOS Os, IN ljtUINT32 Node,
                 IN ljtPOINTER Logical, IN ljtSIZE_T Bytes);

ljeSTATUS
ljoOS_CacheFlush(IN ljoOS Os, IN ljtUINT32 Node,
                 IN ljtPOINTER Logical, IN ljtSIZE_T Bytes);

ljeSTATUS
ljoOS_CacheInvalidate(IN ljoOS Os, IN ljtUINT32 Node,
                      IN ljtPOINTER Logical, IN ljtSIZE_T Bytes);

ljeSTATUS
ljoOS_CacheCleanEx(IN ljoOS Os, IN ljtUINT32 Node,
                   IN ljtPOINTER Logical, IN ljtSIZE_T Offset, IN ljtSIZE_T Bytes);

ljeSTATUS
ljoOS_CacheFlushEx(IN ljoOS Os, IN ljtUINT32 Node,
                   IN ljtPOINTER Logical, IN ljtSIZE_T Offset, IN ljtSIZE_T Bytes);

ljeSTATUS
ljoOS_CacheInvalidateEx(IN ljoOS Os, IN ljtUINT32 Node,
                        IN ljtPOINTER Logical, IN ljtSIZE_T Offset, IN ljtSIZE_T Bytes);


ljeSTATUS
ljoOS_MemoryBarrier(IN ljoOS Os, IN ljtPOINTER Logical);

ljeSTATUS
ljoOS_CPUPhysicalToGPUPhysical(IN ljtPHYS_ADDR_T CPUPhysical,
                               OUT ljtPHYS_ADDR_T *GPUPhysical);

ljeSTATUS
ljoHAL_QueryCPUFrequency(IN ljtUINT32 CPUId, OUT ljtUINT32_PTR CPUFrequency);

ljeSTATUS
ljoOS_QuerySystemInfo(IN ljoOS Os, OUT ljsSystemInfo *Info);


/*----------------------------------------------------------------------------*/
/*----- Profile --------------------------------------------------------------*/

ljeSTATUS
ljkOS_GetProfileTick(OUT ljtUINT64_PTR Tick);

ljeSTATUS
ljkOS_QueryProfileTickRate(OUT ljtUINT64_PTR TickRate);

ljtUINT32
ljkOS_ProfileToMS(IN ljtUINT64 Ticks);

ljeSTATUS
ljoOS_GetProfileTick(OUT ljtUINT64_PTR Tick);

ljeSTATUS
ljoOS_QueryProfileTickRate(OUT ljtUINT64_PTR TickRate);

ljeSTATUS
ljoOS_MonitorEnableDisable(
    IN ljtBOOL Enable
    );

ljeSTATUS
ljoOS_MonitorClear(
    void
    );

ljeSTATUS
ljoOS_MonitorRead(
    IN ljtUINT32 MornitorOffet,
    OUT ljtUINT32 * Data
    );

ljeSTATUS
ljoOS_UintToStr(
    IN ljtUINT32 Uint,
    OUT ljtSTRING String,
    IN ljtUINT32 Radix
    );

#if ljdSTATIC_LINK
void ljoOS_ModuleConstructor(void);

void ljoOS_ModuleDestructor(void);
#endif

#define _ljmPROFILE_INIT(prefix, freq, start)                               \
    do {                                                                    \
        prefix##OS_QueryProfileTickRate(&(freq));                           \
        prefix##OS_GetProfileTick(&(start));                                \
    } while (ljvFALSE)

#define _ljmPROFILE_QUERY(prefix, start, ticks)                             \
    do {                                                                    \
        prefix##OS_GetProfileTick(&(ticks));                                \
        (ticks) = ((ticks) > (start)) ? ((ticks) - (start))                 \
                                      : (~0ull - (start) + (ticks) + 1);    \
    } while (ljvFALSE)

#if ljdENABLE_PROFILING
# define ljmkPROFILE_INIT(freq, start)    _ljmPROFILE_INIT(ljk, freq, start)
# define ljmkPROFILE_QUERY(start, ticks)  _ljmPROFILE_QUERY(ljk, start, ticks)
# define ljmPROFILE_INIT(freq, start)     _ljmPROFILE_INIT(ljo, freq, start)
# define ljmPROFILE_QUERY(start, ticks)   _ljmPROFILE_QUERY(ljo, start, ticks)
# define ljmPROFILE_ONLY(x)               x
# define ljmPROFILE_ELSE(x)               do { } while (ljvFALSE)
# define ljmPROFILE_DECLARE_ONLY(x)       x
#else
# define ljmkPROFILE_INIT(start, freq)    do { } while (ljvFALSE)
# define ljmkPROFILE_QUERY(start, ticks)  do { } while (ljvFALSE)
# define ljmPROFILE_INIT(start, freq)     do { } while (ljvFALSE)
# define ljmPROFILE_QUERY(start, ticks)   do { } while (ljvFALSE)
# define ljmPROFILE_ONLY(x)               do { } while (ljvFALSE)
# define ljmPROFILE_ELSE(x)               x
# define ljmPROFILE_DECLARE_ONLY(x)       do { } while (ljvFALSE)
#endif

/*******************************************************************************
 **  ljoMATH object
 */

#define ljdPI       3.14159265358979323846f

/* User. */
ljtUINT32
ljoMATH_Log2in5dot5(IN ljtINT X);

ljtFLOAT
ljoMATH_UIntAsFloat(IN ljtUINT32 X);

ljtUINT32
ljoMATH_FloatAsUInt(IN ljtFLOAT X);

ljtBOOL
ljoMATH_CompareEqualF(IN ljtFLOAT X, IN ljtFLOAT Y);

ljtUINT16
ljoMATH_UInt8AsFloat16(IN ljtUINT8 X);

ljtUINT32
ljoMATH_Float16ToFloat(IN ljtUINT16 In);

ljtUINT16
ljoMATH_FloatToFloat16(IN ljtUINT32 In);

ljtUINT32
ljoMATH_Float11ToFloat(IN ljtUINT32 In);

ljtUINT16
ljoMATH_FloatToFloat11(IN ljtUINT32 In);

ljtUINT32
ljoMATH_Float10ToFloat(IN ljtUINT32 In);

ljtUINT16
ljoMATH_FloatToFloat10(IN ljtUINT32 In);

ljtUINT32
ljoMATH_Float14ToFloat(IN ljtUINT16 In);

/******************************************************************************
 **************************** Coordinate Structures ***************************
 ******************************************************************************/

typedef struct _ljsPOINT {
    ljtINT32        x;
    ljtINT32        y;
} ljsPOINT;

typedef struct _ljsSIZE {
    ljtINT32        width;
    ljtINT32        height;
} ljsSIZE;

typedef struct _ljsRECT {
    ljtINT32        left;
    ljtINT32        top;
    ljtINT32        right;
    ljtINT32        bottom;
} ljsRECT;

typedef struct _ljs2D_RGBU32
{
    ljtUINT32       R;
    ljtUINT32       G;
    ljtUINT32       B;
} ljs2D_RGBU32;

typedef struct _ljsPIXEL {
    union {
        struct {
            ljtFLOAT r, g, b, a;
        } f;
        struct {
            ljtINT32 r, g, b, a;
        } i;
        struct {
            ljtUINT32 r, g, b, a;
        } ui;
    } color;

    ljtFLOAT  d;
    ljtUINT32 s;

} ljsPIXEL;

/******************************************************************************
 ******************************** ljoSURF Object ******************************
 ******************************************************************************/

/*----------------------------------------------------------------------------*/
/*------------------------------- ljoSURF Common ------------------------------*/

/* Color format component parameters. */
typedef struct _ljsFORMAT_COMPONENT {
    ljtUINT8                    start;
    ljtUINT8                    width;
} ljsFORMAT_COMPONENT;

/* RGBA color format class. */
typedef struct _ljsFORMAT_CLASS_TYPE_RGBA {
    ljsFORMAT_COMPONENT         alpha;
    ljsFORMAT_COMPONENT         red;
    ljsFORMAT_COMPONENT         green;
    ljsFORMAT_COMPONENT         blue;
} ljsFORMAT_CLASS_TYPE_RGBA;

/* YUV color format class. */
typedef struct _ljsFORMAT_CLASS_TYPE_YUV {
    ljsFORMAT_COMPONENT         y;
    ljsFORMAT_COMPONENT         u;
    ljsFORMAT_COMPONENT         v;
} ljsFORMAT_CLASS_TYPE_YUV;

/* Index color format class. */
typedef struct _ljsFORMAT_CLASS_TYPE_INDEX {
    ljsFORMAT_COMPONENT         value;
} ljsFORMAT_CLASS_TYPE_INDEX;

/* Luminance color format class. */
typedef struct _ljsFORMAT_CLASS_TYPE_LUMINANCE {
    ljsFORMAT_COMPONENT         alpha;
    ljsFORMAT_COMPONENT         value;
} ljsFORMAT_CLASS_TYPE_LUMINANCE;

/* Bump map color format class. */
typedef struct _ljsFORMAT_CLASS_TYPE_BUMP {
    ljsFORMAT_COMPONENT         alpha;
    ljsFORMAT_COMPONENT         l;
    ljsFORMAT_COMPONENT         v;
    ljsFORMAT_COMPONENT         u;
    ljsFORMAT_COMPONENT         q;
    ljsFORMAT_COMPONENT         w;
} ljsFORMAT_CLASS_TYPE_BUMP;

/* Depth and stencil format class. */
typedef struct _ljsFORMAT_CLASS_TYPE_DEPTH {
    ljsFORMAT_COMPONENT         depth;
    ljsFORMAT_COMPONENT         stencil;
} ljsFORMAT_CLASS_TYPE_DEPTH;

/* Intensity format class. */
typedef struct _ljsFORMAT_CLASs_TYPE_INTENSITY {
    ljsFORMAT_COMPONENT         value;
} ljsFORMAT_CLASs_TYPE_INTENSITY;

typedef union _ljuPIXEL_FORMAT_CLASS {
    ljsFORMAT_CLASS_TYPE_BUMP       bump;
    ljsFORMAT_CLASS_TYPE_RGBA       rgba;
    ljsFORMAT_CLASS_TYPE_YUV        yuv;
    ljsFORMAT_CLASS_TYPE_LUMINANCE  lum;
    ljsFORMAT_CLASS_TYPE_INDEX      index;
    ljsFORMAT_CLASS_TYPE_DEPTH      depth;
    ljsFORMAT_CLASs_TYPE_INTENSITY  intensity;
} ljuPIXEL_FORMAT_CLASS;

/* Format parameters. */
typedef struct _ljsSURF_FORMAT_INFO {
    /* Name of the format */
    ljtCONST_STRING             formatName;

    /* Format code and class. */
    ljeSURF_FORMAT              format;
    ljeFORMAT_CLASS             fmtClass;

    /* Format data type */
    ljeFORMAT_DATATYPE          fmtDataType;

    /* The size of one pixel in bits. */
    ljtUINT8                    bitsPerPixel;

    /* Pixel block dimensions. */
    ljtUINT                     blockWidth;
    ljtUINT                     blockHeight;

    /* Pixel block size in bits. */
    ljtUINT                     blockSize;

    /* Some formats are larger than what the GPU can support.      */
    /* These formats are read in the number of layers specified.   */
    ljtUINT8                    layers;

    /* The format is faked and software will interpret it differently
     * with HW. Most of them can't be blendable(PE) or filterable(TX).
     */
    ljtBOOL                     fakedFormat;

    /* Some formats have two neighbour pixels interleaved together. */
    /* To describe such format, set the flag to 1 and add another   */
    /* like this one describing the odd pixel format.               */
    ljtBOOL                     interleaved;

    /* sRGB format. */
    ljtBOOL                     sRGB;

    /* How GPU read from big-endian host memory */
    ljeENDIAN_HINT              endian;

    /* Format components. */
    ljuPIXEL_FORMAT_CLASS       u;

    /* Format components. */
    ljuPIXEL_FORMAT_CLASS       uOdd;

    /* Render format. */
    ljeSURF_FORMAT              closestRenderFormat;
    /*ljtCLOSEST_FORMAT           dynamicClosestRenderFormat;*/
    ljtUINT                     renderFormat;
    const ljeTEXTURE_SWIZZLE   *pixelSwizzle;

    /* Texture format. */
    ljeSURF_FORMAT              closestTXFormat;
    ljtUINT                     txFormat;
    const ljeTEXTURE_SWIZZLE   *txSwizzle;
    ljtBOOL                     txIntFilter;
} ljsSURF_FORMAT_INFO;

/* Frame buffer information. */
typedef struct _ljsSURF_FRAMEBUFFER {
    ljtPOINTER                  logical;
    ljtUINT                     width, height;
    ljtINT                      stride;
    ljeSURF_FORMAT              format;
} ljsSURF_FRAMEBUFFER;

typedef union _lju2D_STATE_VALUE
{
    ljs2D_RGBU32                    minValue;
    ljs2D_RGBU32                    maxMinReciprocal;
    ljs2D_RGBU32                    stdReciprocal;
    ljs2D_RGBU32                    meanValue;
    ljtBOOL                         enable;
    ljtUINT32                       stepReciprocal;
    lje2D_NORMALIZATION_MODE        normalizationMode;
    ljtBOOL                         byPassQuantization;
    lje2D_U8ToU10_CONVERSION_MODE   u8Tu10_Mode;
    lje2D_MULTICORE_MODE            multicoreMode;
} lju2D_STATE_VALUE;

/* Set 2D state */
typedef struct _ljs2D_STATE_CONFIG
{
    lje2D_STATE_KEY             state;
    lju2D_STATE_VALUE           value;
} ljs2D_STATE_CONFIG;

/* Generic pixel component descriptors. */
extern ljsFORMAT_COMPONENT ljvPIXEL_COMP_XXX8;
extern ljsFORMAT_COMPONENT ljvPIXEL_COMP_XX8X;
extern ljsFORMAT_COMPONENT ljvPIXEL_COMP_X8XX;
extern ljsFORMAT_COMPONENT ljvPIXEL_COMP_8XXX;

/* Construct a new ljoSURF object. */
ljeSTATUS
ljoSURF_Construct(IN ljoHAL Hal,
                  IN ljtUINT Width,
                  IN ljtUINT Height,
                  IN ljtUINT Depth,
                  IN ljeSURF_TYPE Type,
                  IN ljeSURF_FORMAT Format,
                  IN ljePOOL Pool,
                  OUT ljoSURF *Surface);

ljeSTATUS
ljoSURF_ConstructWithUserPool(IN ljoHAL Hal,
                              IN ljtUINT Width,
                              IN ljtUINT Height,
                              IN ljtUINT Depth,
                              IN ljeSURF_TYPE Type,
                              IN ljeSURF_FORMAT Format,
                              IN ljtPOINTER TileStatusLogical,
                              IN ljtPHYS_ADDR_T TileStatusPhysical,
                              IN ljtPOINTER Logical,
                              IN ljtPHYS_ADDR_T Physical,
                              OUT ljoSURF *Surface);

/* Destroy an ljoSURF object. */
ljeSTATUS
ljoSURF_Destroy(IN ljoSURF Surface);

ljeSTATUS
ljoSURF_DestroyForAllHWType(IN ljoSURF Surface);

/* Map user-allocated surface. */
ljeSTATUS
ljoSURF_MapUserSurface(IN ljoSURF Surface,
                       IN ljtUINT Alignment,
                       IN ljtPOINTER Logical,
                       IN ljtPHYS_ADDR_T Physical);

/* Wrapp surface with known logical/GPU address */
ljeSTATUS
ljoSURF_WrapSurface(IN ljoSURF Surface,
                    IN ljtUINT Alignment,
                    IN ljtPOINTER Logical,
                    IN ljtADDRESS Address);

/* Query vid mem node info. */
ljeSTATUS
ljoSURF_QueryVidMemNode(IN ljoSURF Surface,
                        OUT ljtUINT32 *Node,
                        OUT ljePOOL *Pool,
                        OUT ljtSIZE_T_PTR Bytes,
                        OUT ljtUINT32 *TsNode,
                        OUT ljePOOL *TsPool,
                        OUT ljtSIZE_T_PTR TsBytes);

/* Query vid mem Multi node info. */
ljeSTATUS
ljoSURF_QueryVidMemMultiNode(IN ljoSURF Surface,
                             OUT ljtUINT32 *Node,
                             OUT ljePOOL *Pool,
                             OUT ljtSIZE_T_PTR Bytes,
                             OUT ljtUINT32 *Node2,
                             OUT ljePOOL *Pool2,
                             OUT ljtSIZE_T_PTR Bytes2,
                             OUT ljtUINT32 *Node3,
                             OUT ljePOOL *Pool3,
                             OUT ljtSIZE_T_PTR Bytes3);

/* Set the color type of the surface. */
ljeSTATUS
ljoSURF_SetColorType(IN ljoSURF Surface, IN ljeSURF_COLOR_TYPE ColorType);

/* Get the color type of the surface. */
ljeSTATUS
ljoSURF_GetColorType(IN ljoSURF Surface, OUT ljeSURF_COLOR_TYPE *ColorType);

/* Set the color space of the surface. */
ljeSTATUS
ljoSURF_SetColorSpace(IN ljoSURF Surface, IN ljeSURF_COLOR_SPACE ColorSpace);

/* Get the color space of the surface. */
ljeSTATUS
ljoSURF_GetColorSpace(IN ljoSURF Surface, OUT ljeSURF_COLOR_SPACE *ColorSpace);

/* Set the surface ration angle. */
ljeSTATUS
ljoSURF_SetRotation(IN ljoSURF Surface, IN ljeSURF_ROTATION Rotation);

ljeSTATUS
ljoSURF_IsValid(IN ljoSURF Surface);

#if ljdENABLE_3D
/* Verify and return the state of the tile status mechanism. */
ljeSTATUS
ljoSURF_IsTileStatusSupported(IN ljoSURF Surface);

/* Verify if surface has tile status enabled. */
ljeSTATUS
ljoSURF_IsTileStatusEnabled(IN ljsSURF_VIEW *SurfView);

/* Verify if surface is compressed. */
ljeSTATUS
ljoSURF_IsCompressed(IN ljsSURF_VIEW *SurfView);

/* Enable tile status for the specified surface on zero slot. */
ljeSTATUS
ljoSURF_EnableTileStatus(IN ljsSURF_VIEW *Surface);

/* Enable tile status for the specified surface on specified slot. */
ljeSTATUS
ljoSURF_EnableTileStatusEx(IN ljsSURF_VIEW *surfView, IN ljtUINT RtIndex);

/* Disable tile status for the specified surface. */
ljeSTATUS
ljoSURF_DisableTileStatus(IN ljsSURF_VIEW *SurfView, IN ljtBOOL Decompress);

/* Flush tile status cache for the specified surface. */
ljeSTATUS
ljoSURF_FlushTileStatus(IN ljsSURF_VIEW *SurfView, IN ljtBOOL Decompress);
#endif /* ljdENABLE_3D */

/* Get surface size. */
ljeSTATUS
ljoSURF_GetSize(IN ljoSURF Surface,
                OUT ljtUINT *Width,
                OUT ljtUINT *Height,
                OUT ljtUINT *Depth);

/* Get surface information */
ljeSTATUS
ljoSURF_GetInfo(IN ljoSURF Surface,
                IN ljeSURF_INFO_TYPE InfoType,
                IN OUT ljtINT32 *Value);

/* Get surface aligned sizes. */
ljeSTATUS
ljoSURF_GetAlignedSize(IN ljoSURF Surface,
                       OUT ljtUINT *Width,
                       OUT ljtUINT *Height,
                       OUT ljtINT *Stride);

/* Get alignments. */
ljeSTATUS
ljoSURF_GetAlignment(IN ljeSURF_TYPE Type,
                     IN ljeSURF_FORMAT Format,
                     OUT ljtUINT *AddressAlignment,
                     OUT ljtUINT *XAlignment,
                     OUT ljtUINT *YAlignment);

ljeSTATUS
ljoSURF_AlignResolveRect(IN ljoSURF Surf,
                         IN ljsPOINT_PTR RectOrigin,
                         IN ljsPOINT_PTR RectSize,
                         OUT ljsPOINT_PTR AlignedOrigin,
                         OUT ljsPOINT_PTR AlignedSize);

/* Get surface type and format. */
ljeSTATUS
ljoSURF_GetFormat(IN ljoSURF Surface,
                  OUT OPTIONAL ljeSURF_TYPE *Type,
                  OUT OPTIONAL ljeSURF_FORMAT *Format);

/* Get surface information */
ljeSTATUS
ljoSURF_GetFormatInfo(IN ljoSURF Surface,
                      OUT ljsSURF_FORMAT_INFO_PTR *formatInfo);

/* Get Surface pack format */
ljeSTATUS
ljoSURF_GetPackedFormat(IN ljoSURF Surface,
                        OUT ljeSURF_FORMAT *Format);

/* Get surface tiling. */
ljeSTATUS
ljoSURF_GetTiling(IN ljoSURF Surface, OUT ljeTILING *Tiling);

/* Get bottom buffer offset bytes. */
ljeSTATUS
ljoSURF_GetBottomBufferOffset(IN ljoSURF Surface,
                              OUT ljtUINT_PTR BottomBufferOffset);

/* Lock the surface. */
ljeSTATUS
ljoSURF_Lock(IN ljoSURF Surface,
             IN OUT ljtADDRESS *Address,
             IN OUT ljtPOINTER *Memory);

/* Unlock the surface. */
ljeSTATUS
ljoSURF_Unlock(IN ljoSURF Surface, IN ljtPOINTER Memory);

/*. Query surface flags.*/
ljeSTATUS
ljoSURF_QueryFlags(IN ljoSURF Surface, IN ljeSURF_FLAG Flag);

ljeSTATUS
ljoSURF_QueryHints(IN ljoSURF Surface, IN ljeSURF_TYPE Hints);

/* Return pixel format parameters; Info is required to be a pointer to an
 * array of at least two items because some formats have up to two records
 * of description.
 */
ljeSTATUS
ljoSURF_QueryFormat(IN ljeSURF_FORMAT Format,
                    OUT ljsSURF_FORMAT_INFO_PTR *Info);

/* Compute the color pixel mask. */
ljeSTATUS
ljoSURF_ComputeColorMask(IN ljsSURF_FORMAT_INFO_PTR Format,
                         OUT ljtUINT32_PTR ColorMask);

/* Flush the surface. */
ljeSTATUS
ljoSURF_Flush(IN ljoSURF Surface);

ljeSTATUS
ljoSURF_3DBlitClearTileStatus(IN ljsSURF_VIEW *SurfView,
                              IN ljtBOOL ClearAsDirty);

/* Fill surface from it's tile status buffer. */
ljeSTATUS
ljoSURF_FillFromTile(IN ljsSURF_VIEW *SurView);

/* Fill surface with a value. */
ljeSTATUS
ljoSURF_Fill(IN ljoSURF Surface,
             IN ljsPOINT_PTR Origin,
             IN ljsSIZE_PTR Size,
             IN ljtUINT32 Value,
             IN ljtUINT32 Mask);

/* Alpha blend two surfaces together. */
ljeSTATUS
ljoSURF_Blend(IN ljoSURF SrcSurf,
              IN ljoSURF DstSurf,
              IN ljsPOINT_PTR SrcOrigin,
              IN ljsPOINT_PTR DstOrigin,
              IN ljsSIZE_PTR Size,
              IN ljeSURF_BLEND_MODE Mode);

/* Create a new ljoSURF wrapper object. */
ljeSTATUS
ljoSURF_ConstructWrapper(IN ljoHAL Hal, OUT ljoSURF *Surface);

/* Set surface flags.*/
ljeSTATUS
ljoSURF_SetFlags(IN ljoSURF Surface, IN ljeSURF_FLAG Flag, IN ljtBOOL Value);

/* Set the underlying buffer for the surface wrapper. */
ljeSTATUS
ljoSURF_SetBuffer(IN ljoSURF Surface,
                  IN ljeSURF_TYPE Type,
                  IN ljeSURF_FORMAT Format,
                  IN ljtUINT Stride,
                  IN ljtPOINTER Logical,
                  IN ljtUINT64 Physical);

/* Set the size of the surface in pixels and map the underlying buffer. */
ljeSTATUS
ljoSURF_SetWindow(IN ljoSURF Surface,
                  IN ljtUINT X,
                  IN ljtUINT Y,
                  IN ljtUINT Width,
                  IN ljtUINT Height);

/* Set the size of the surface in pixels and map the underlying buffer. */
ljeSTATUS
ljoSURF_SetImage(IN ljoSURF Surface,
                 IN ljtUINT X,
                 IN ljtUINT Y,
                 IN ljtUINT Width,
                 IN ljtUINT Height,
                 IN ljtUINT Depth);

/* Set width/height alignment of the surface directly and calculate stride/size.
 * This is only for dri backend now. Please be careful before use.
 */
ljeSTATUS
ljoSURF_SetAlignment(IN ljoSURF Surface, IN ljtUINT Width, IN ljtUINT Height);

/* Increase reference count of the surface. */
ljeSTATUS
ljoSURF_ReferenceSurface(IN ljoSURF Surface);

/* Get surface reference count. */
ljeSTATUS
ljoSURF_QueryReferenceCount(IN ljoSURF Surface, OUT ljtINT32 *ReferenceCount);

/* Set surface orientation. */
ljeSTATUS
ljoSURF_SetOrientation(IN ljoSURF Surface, IN ljeORIENTATION Orientation);

/* Query surface orientation. */
ljeSTATUS
ljoSURF_QueryOrientation(IN ljoSURF Surface, OUT ljeORIENTATION *Orientation);

ljeSTATUS
ljoSURF_NODE_Cache(IN ljsSURF_NODE_PTR Node,
                   IN ljtPOINTER Logical,
                   IN ljtSIZE_T Bytes,
                   IN ljeCACHEOPERATION Operation);

ljeSTATUS
ljoSURF_NODE_CacheEx(IN ljsSURF_NODE_PTR Node,
                   IN ljtPOINTER Logical,
                   IN ljtSIZE_T Offset,
                   IN ljtSIZE_T Bytes,
                   IN ljeCACHEOPERATION Operation);

ljeSTATUS
ljsSURF_NODE_SetHardwareAddress(IN ljsSURF_NODE_PTR Node, IN ljtADDRESS Address);

ljeSTATUS
ljsSURF_NODE_GetHardwareAddress(IN ljsSURF_NODE_PTR Node,
                                OUT ljtADDRESS *Physical,
                                OUT ljtADDRESS *Physical2,
                                OUT ljtADDRESS *Physical3,
                                OUT ljtADDRESS *PhysicalBottom);

ljtADDRESS
ljsSURF_NODE_GetHWAddress(IN ljsSURF_NODE_PTR Node);

/* Lock and unlock surface node */
ljeSTATUS
ljoSURF_LockNode(IN ljsSURF_NODE_PTR Node,
                 OUT ljtADDRESS *Address,
                 OUT ljtPOINTER *Memory);

ljeSTATUS
ljoSURF_UnLockNode(IN ljsSURF_NODE_PTR Node, IN ljeSURF_TYPE Type);

/* Perform CPU cache operation on surface node */
ljeSTATUS
ljoSURF_NODE_CPUCacheOperation(IN ljsSURF_NODE_PTR Node,
                               IN ljeSURF_TYPE Type,
                               IN ljtSIZE_T Offset,
                               IN ljtSIZE_T Length,
                               IN ljeCACHEOPERATION Operation);

/* Perform CPU cache operation on surface */
ljeSTATUS
ljoSURF_CPUCacheOperation(IN ljoSURF Surface,
                          IN ljeCACHEOPERATION Operation);

ljeSTATUS
ljoSURF_Swap(IN ljoSURF Surface1, IN ljoSURF Surface2);

ljeSTATUS
ljoSURF_ResetSurWH(IN ljoSURF Surface,
                   IN ljtUINT oriw,
                   IN ljtUINT orih,
                   IN ljtUINT alignw,
                   IN ljtUINT alignh,
                   IN ljeSURF_FORMAT fmt);

/* Update surface timestamp. */
ljeSTATUS
ljoSURF_UpdateTimeStamp(IN ljoSURF Surface);

/* Query surface current timestamp. */
ljeSTATUS
ljoSURF_QueryTimeStamp(IN ljoSURF Surface, OUT ljtUINT64 *TimeStamp);

/*
 * Allocate shared buffer for this surface, so that
 * surface states can be shared across processes.
 */
ljeSTATUS
ljoSURF_AllocShBuffer(IN ljoSURF Surface, OUT ljtSHBUF *ShBuf);

/* Bind shared buffer to this surface */
ljeSTATUS
ljoSURF_BindShBuffer(IN ljoSURF Surface, IN ljtSHBUF ShBuf);

/* Push surface shared states to shared buffer. */
ljeSTATUS
ljoSURF_PushSharedInfo(IN ljoSURF Surface);

/* Pop shared states from shared buffer. */
ljeSTATUS
ljoSURF_PopSharedInfo(IN ljoSURF Surface);

#if (ljdENABLE_3D || ljdENABLE_VG)
/* Copy surface. */
ljeSTATUS
ljoSURF_Copy(IN ljoSURF Surface, IN ljoSURF Source);

/* Set number of samples for a ljoSURF object. */
ljeSTATUS
ljoSURF_SetSamples(IN ljoSURF Surface, IN ljtUINT Samples);

/* Get the number of samples per pixel. */
ljeSTATUS
ljoSURF_GetSamples(IN ljoSURF Surface, OUT ljtUINT_PTR Samples);

/* Append tile status buffer to user pool surface. */
ljeSTATUS
ljoSURF_AppendTileStatus(IN ljoSURF Surface);
#endif

ljeSTATUS
ljoSURF_WrapUserMemory(IN ljoHAL Hal,
                       IN ljtUINT Width,
                       IN ljtUINT Height,
                       IN ljtUINT Stride,
                       IN ljtUINT Depth,
                       IN ljeSURF_TYPE Type,
                       IN ljeSURF_FORMAT Format,
                       IN ljtUINT32 Handle,
                       IN ljtUINT32 Flag,
                       OUT ljoSURF *Surface);

#ifdef EMULATOR
/* Wrap user memory by external address. */
ljeSTATUS
ljoSURF_WrapUserMemory_2(IN ljoHAL Hal,
                       IN ljtUINT Width,
                       IN ljtUINT Height,
                       IN ljtUINT Stride,
                       IN ljtUINT Depth,
                       IN ljeSURF_TYPE Type,
                       IN ljeSURF_FORMAT Format,
                       IN ljtUINT32_PTR Address,
                       IN ljtUINT32 Flag,
                       OUT ljoSURF * Surface);
#endif

ljeSTATUS
ljoSURF_WrapUserMultiBuffer(IN ljoHAL Hal,
                            IN ljtUINT Width,
                            IN ljtUINT Height,
                            IN ljeSURF_TYPE Type,
                            IN ljeSURF_FORMAT Format,
                            IN ljtUINT Stride[3],
                            IN ljtUINT32 Handle[3],
                            IN ljtUINT BufferOffset[3],
                            IN ljtUINT32 Flag,
                            OUT ljoSURF *Surface);

ljeSTATUS
ljoSURF_UpdateMetadata(IN ljoSURF Surface, IN ljtINT TsFD);

#define MAX_SURF_MIX_SRC_NUM 64
ljeSTATUS
ljoSURF_MixSurfacesCPU(IN ljoSURF TargetSurface,
                       IN ljtUINT TargetSliceIndex,
                       IN ljoSURF *SourceSurface,
                       IN ljtUINT *SourceSliceIndices,
                       IN ljtFLOAT *Weights,
                       IN ljtINT Count);

ljeSTATUS
ljsSURF_NODE_Update(
    IN OUT ljsSURF_NODE_PTR Node
    );

/******************************************************************************
 ****************************** Hash Structure ********************************
 ******************************************************************************/

typedef struct _ljsHASH_MD5CTX {
    ljtBOOL   bigEndian;
    ljtSIZE_T bytes; /* Number of bytes processed */
    ljtUINT32 states[4];
    ljtUINT8  buffer[64];
} ljsHASH_MD5CTX;

void
ljsHASH_MD5Init(ljsHASH_MD5CTX *ctx);
void
ljsHASH_MD5Update(ljsHASH_MD5CTX *ctx, const void *data, ljtSIZE_T bytes);
void
ljsHASH_MD5Final(ljsHASH_MD5CTX *ctx, ljtUINT8 digest[16]);

/******************************************************************************
 ******************************* ljsRECT Structure ****************************
 ******************************************************************************/

/* Initialize rectangle structure. */
ljeSTATUS
ljsRECT_Set(OUT ljsRECT_PTR Rect,
            IN ljtINT32 Left,
            IN ljtINT32 Top,
            IN ljtINT32 Right,
            IN ljtINT32 Bottom);

/* Return the width of the rectangle. */
ljeSTATUS
ljsRECT_Width(IN ljsRECT_PTR Rect, OUT ljtINT32 *Width);

/* Return the height of the rectangle. */
ljeSTATUS
ljsRECT_Height(IN ljsRECT_PTR Rect, OUT ljtINT32 *Height);

/* Ensure that top left corner is to the left and above the right bottom. */
ljeSTATUS
ljsRECT_Normalize(IN OUT ljsRECT_PTR Rect);

/* Compare two rectangles. */
ljeSTATUS
ljsRECT_IsEqual(IN ljsRECT_PTR Rect1, IN ljsRECT_PTR Rect2, OUT ljtBOOL *Equal);

/* Compare the sizes of two rectangles. */
ljeSTATUS
ljsRECT_IsOfEqualSize(IN ljsRECT_PTR Rect1, IN ljsRECT_PTR Rect2, OUT ljtBOOL *EqualSize);

ljeSTATUS
ljsRECT_RelativeRotation(IN ljeSURF_ROTATION Orientation,
                         IN OUT ljeSURF_ROTATION *Relation);

ljeSTATUS
ljsRECT_Rotate(IN OUT ljsRECT_PTR Rect,
               IN ljeSURF_ROTATION Rotation,
               IN ljeSURF_ROTATION toRotation,
               IN ljtINT32 SurfaceWidth,
               IN ljtINT32 SurfaceHeight);

/******************************************************************************
 **************************** ljsBOUNDARY Structure ***************************
 ******************************************************************************/

typedef struct _ljsBOUNDARY {
    ljtINT x;
    ljtINT y;
    ljtINT width;
    ljtINT height;
} ljsBOUNDARY;

/******************************************************************************
 ********************************* ljoHEAP Object *****************************
 ******************************************************************************/

typedef struct _ljoHEAP *ljoHEAP;

/* Construct a new ljoHEAP object. */
ljeSTATUS
ljoHEAP_Construct(IN ljoOS Os, IN ljtSIZE_T AllocationSize, OUT ljoHEAP *Heap);

/* Destroy an ljoHEAP object. */
ljeSTATUS
ljoHEAP_Destroy(IN ljoHEAP Heap);

/* Allocate memory. */
ljeSTATUS
ljoHEAP_Allocate(IN ljoHEAP Heap, IN ljtSIZE_T Bytes, OUT ljtPOINTER *Node);

ljeSTATUS
ljoHEAP_GetMemorySize(IN ljoHEAP Heap, IN ljtPOINTER Memory, OUT ljtSIZE_T_PTR MemorySize);

/* Free memory. */
ljeSTATUS
ljoHEAP_Free(IN ljoHEAP Heap, IN ljtPOINTER Node);

#if (LJMICRO_PROFILER_SYSTEM_MEMORY  || ljdDEBUG)
/* Profile the heap. */
ljeSTATUS
ljoHEAP_ProfileStart(IN ljoHEAP Heap);

ljeSTATUS
ljoHEAP_ProfileEnd(IN ljoHEAP Heap, IN ljtCONST_STRING Title);
#endif


/******************************************************************************
 ******************************* Debugging Macros *****************************
 ******************************************************************************/

void
ljoOS_SetDebugLevel(IN ljtUINT32 Level);

void
ljoOS_GetDebugLevel(OUT ljtUINT32_PTR DebugLevel);

void
ljoOS_GetDebugZone(IN ljtUINT32 Zone, OUT ljtUINT32_PTR DebugZone);

void
ljoOS_SetDebugZone(IN ljtUINT32 Zone);

void
ljoOS_SetDebugFile(IN ljtCONST_STRING FileName);

void
ljoOS_EnableDebugDump(IN ljtBOOL Enable);

ljtFILE
ljoOS_ReplaceDebugFile(IN ljtFILE fp);

/*******************************************************************************
 **
 **  ljmFATAL
 **
 **      Print a message to the debugger and execute a break point.
 **
 **  ARGUMENTS:
 **
 **      message .
 **      ...     Optional arguments.
 */

void
ljkOS_DebugFatal(IN ljtCONST_STRING Message, ...);

void
ljoOS_DebugFatal(IN ljtCONST_STRING Message, ...);

#if ljmIS_DEBUG(ljdDEBUG_FATAL)
# define ljmFATAL        ljoOS_DebugFatal
# define ljmkFATAL       ljkOS_DebugFatal
#elif ljdHAS_ELLIPSIS
# define ljmFATAL(...)
# define ljmkFATAL(...)
#else
ljmINLINE static void
__dummy_fatal(IN ljtCONST_STRING Message, ...)
{
}

# define ljmFATAL        __dummy_fatal
# define ljmkFATAL       __dummy_fatal
#endif

/*******************************************************************************
 **
 **  ljmTRACE
 **
 **      Print a message to the debugfer if the correct level has been set.  In
 **      retail mode this macro does nothing.
 **
 **  ARGUMENTS:
 **
 **      level   Level of message.
 **      message .
 **      ...     Optional arguments.
 */
#define ljvLEVEL_NONE       -1
#define ljvLEVEL_ERROR      0
#define ljvLEVEL_WARNING    1
#define ljvLEVEL_INFO       2
#define ljvLEVEL_VERBOSE    3

void
ljkOS_DebugTrace(IN ljtUINT32 Level, IN ljtCONST_STRING Message, ...)
CHECK_PRINTF_FORMAT(2, 3);

void
ljoOS_DebugTrace(IN ljtUINT32 Level, IN ljtCONST_STRING Message, ...)
CHECK_PRINTF_FORMAT(2, 3);

#if ljmIS_DEBUG(ljdDEBUG_TRACE)
# define ljmTRACE            ljoOS_DebugTrace
# define ljmkTRACE           ljkOS_DebugTrace
# define ljmkTRACE_N(Level, ArgumentSize, ...) \
        ljkOS_DebugTrace(Level, __VA_ARGS__)
#elif ljdHAS_ELLIPSIS
# define ljmTRACE(...)
# define ljmkTRACE(...)
# define ljmkTRACE_N(...)
#else
    ljmINLINE static void
    __dummy_trace(IN ljtUINT32 Level, IN ljtCONST_STRING Message, ...)
    {
    }

    ljmINLINE static void
    __dummy_trace_n(IN ljtUINT32 Level, IN ljtUINT ArgumentSize, IN ljtCONST_STRING Message, ...)
    {
    }

# define ljmTRACE            __dummy_trace
# define ljmkTRACE           __dummy_trace
# define ljmkTRACE_N         __dummy_trace_n
#endif

/*******************************************************************************
 **
 **  ljmTRACE_ZONE
 **
 **      Print a message to the debugger if the correct level and zone has been
 **      set.  In retail mode this macro does nothing.
 **
 **  ARGUMENTS:
 **
 **      Level   Level of message.
 **      Zone    Zone of message.
 **      Message debuger message.
 **      ...     Optional arguments.
 */

void
ljkOS_DebugTraceZone(IN ljtUINT32 Level, IN ljtUINT32 Zone, IN ljtCONST_STRING Message, ...);

void
ljoOS_DebugTraceZone(IN ljtUINT32 Level, IN ljtUINT32 Zone, IN ljtCONST_STRING Message, ...);

#if ljmIS_DEBUG(ljdDEBUG_TRACE)
# define ljmTRACE_ZONE      ljoOS_DebugTraceZone
# define ljmkTRACE_ZONE     ljkOS_DebugTraceZone
# define ljmkTRACE_ZONE_N(Level, Zone, ArgumentSize, ...) \
        ljkOS_DebugTraceZone(Level, Zone, __VA_ARGS__)
#elif ljdHAS_ELLIPSIS
# define ljmTRACE_ZONE(...)
# define ljmkTRACE_ZONE(...)
# define ljmkTRACE_ZONE_N(...)
#else
    ljmINLINE static void
    __dummy_trace_zone(IN ljtUINT32 Level,
                       IN ljtUINT32 Zone,
                       IN ljtCONST_STRING Message, ...)
    {
    }

    ljmINLINE static void
    __dummy_trace_zone_n(IN ljtUINT32 Level,
                         IN ljtUINT32 Zone,
                         IN ljtUINT ArgumentSize,
                         IN ljtCONST_STRING Message,
                         ...)
    {
    }

# define ljmTRACE_ZONE           __dummy_trace_zone
# define ljmkTRACE_ZONE          __dummy_trace_zone
# define ljmkTRACE_ZONE_N        __dummy_trace_zone_n
#endif

/*******************************************************************************
 **
 **  ljmDEBUG_ONLY
 **
 **      Execute a statement or function only in DEBUG mode.
 **
 **  ARGUMENTS:
 **
 **      f       Statement or function to execute.
 */
#if ljmIS_DEBUG(ljdDEBUG_CODE)
# define ljmDEBUG_ONLY(f)        f
#else
# define ljmDEBUG_ONLY(f)
#endif

/*******************************************************************************
 **
 **  ljmSTACK_PUSH
 **  ljmSTACK_POP
 **  ljmSTACK_DUMP
 **  ljmSTACK_REMOVE
 **
 **      Push or pop a function with entry arguments on the trace stack.
 **
 **  ARGUMENTS:
 **
 **      Function    Name of function.
 **      Line        Line number.
 **      Text        Optional text.
 **      ...         Optional arguments for text.
 **
 **      Thread      Thread id.
 */
void
ljoOS_StackPush(IN ljtINT8_PTR Identity,
                IN ljtCONST_STRING Function,
                IN ljtINT Line,
                IN ljtCONST_STRING Text,
                ...);

void
ljoOS_StackPop(IN ljtINT8_PTR Identity, IN ljtCONST_STRING Function);

void
ljoOS_StackDump(void);

void
ljoOS_StackRemove(IN ljtHANDLE Thread);

#if ljmIS_DEBUG(ljdDEBUG_STACK)
# define ljmSTACK_PUSH           ljoOS_StackPush
# define ljmSTACK_POP            ljoOS_StackPop
# define ljmSTACK_DUMP           ljoOS_StackDump
# define ljmSTACK_REMOVE         ljoOS_StackRemove
#elif ljdHAS_ELLIPSIS
# define ljmSTACK_PUSH(...)
# define ljmSTACK_POP(...)
# define ljmSTACK_DUMP()
# define ljmSTACK_REMOVE(...)
#else
    ljmINLINE static void
    __dummy_stack_push(IN ljtCONST_STRING Function,
                       IN ljtINT Line,
                       IN ljtCONST_STRING Text, ...)
    {
    }

    ljmINLINE static void
    __dummy_stack_pop(IN ljtINT8_PTR Identity, IN ljtCONST_STRING Function);

    ljmINLINE static void
    __dummy_stack_remove(IN ljtHANDLE Thread);

# define ljmSTACK_PUSH               __dummy_stack_push
# define ljmSTACK_POP(a, b)          __dummy_stack_pop
# define ljmSTACK_DUMP()
# define ljmSTACK_REMOVE(a)          __dummy_stack_remove
#endif

/*******************************************************************************
 **
 **  ljmBINARY_TRACE
 **
 **      Push or pop a function with entry arguments on the trace stack.
 **
 **  ARGUMENTS:
 **
 **      Function    Name of function
 **      Line        Line number
 **      Text        Optional text
 **      ...         Optional arguments for text.
 */
void
ljoOS_BinaryTrace(IN ljtCONST_STRING Function,
                  IN ljtINT Line,
                  IN ljtCONST_STRING Text OPTIONAL,
                  ...);

void
ljkOS_BinaryTrace(IN ljtCONST_STRING Function,
                  IN ljtINT Line,
                  IN ljtCONST_STRING Text OPTIONAL,
                  ...);

#if ljdBINARY_TRACE
# define ljmBINARY_TRACE         ljoOS_BinaryTrace
# define ljmkBINARY_TRACE        ljkOS_BinaryTrace
#elif ljdHAS_ELLIPSIS
# define ljmBINARY_TRACE(Function, Line, Text, ...)
# define ljmkBINARY_TRACE(Function, Line, Text, ...)
#else
    ljmINLINE static void
    __dummy_binary_trace(IN ljtCONST_STRING Function,
                         IN ljtINT Line,
                         IN ljtCONST_STRING Text, ...)
    {
    }

# define ljmBINARY_TRACE         __dummy_binary_trace
# define ljmkBINARY_TRACE        __dummy_binary_trace
#endif

/*******************************************************************************
 **
 **  ljmSYSTRACE_BEGIN
 **  ljmSYSTRACE_END
 **
 **      Systrace is a performance tuning tool on linux.
 **
 **  ARGUMENTS:
 **
 **      FuncName Function name
 **      Zone     Systrace zone. Only specified zones are traced.
 */

void
ljoOS_SysTraceBegin(IN ljtUINT32 Zone, IN ljtCONST_STRING FuncName);

void
ljoOS_SysTraceEnd(IN ljtUINT32 Zone);

#if defined(LINUX) && ljdSYSTRACE
# define ljmSYSTRACE_BEGIN       ljoOS_SysTraceBegin
# define ljmSYSTRACE_END         ljoOS_SysTraceEnd
#elif ljdHAS_ELLIPSIS
# define ljmSYSTRACE_BEGIN(...)
# define ljmSYSTRACE_END(...)
#else
    ljmINLINE static void
    __dummy_systrace_begin(IN ljtUINT32 Zone,
                           IN ljtCONST_STRING FuncName)
    {
    }

    ljmINLINE static void
    __dummy_systrace_end(IN ljtUINT32 Zone)
    {
    }

# define ljmSYSTRACE_BEGIN       __dummy_systrace_begin
# define ljmSYSTRACE_END         __dummy_systrace_end
#endif

/******************************************************************************
 ******************************** Logging Macros ******************************
 ******************************************************************************/

#define ljdHEADER_LEVEL ljvLEVEL_VERBOSE

/* Always enable header/footer when systrace build is on */
#if defined(LINUX) && ljdSYSTRACE
# undef ljdEMPTY_HEADER_FOOTER
#endif

#ifndef ljdEMPTY_HEADER_FOOTER
# define ljdEMPTY_HEADER_FOOTER 0
#endif

#if ljdENABLE_PROFILING
void
ljoOS_ProfileDB(IN ljtCONST_STRING Function, IN OUT ljtBOOL_PTR Initialized);

#define ljmHEADER() \
    ljtINT8 __user__ = 1; \
    static ljtBOOL __profile__initialized__ = ljvFALSE; \
    ljmSTACK_PUSH(&__user__, __FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
    ljoOS_ProfileDB(__FUNCTION__, &__profile__initialized__)

#define ljmHEADER_ARG(...) \
    ljtINT8 __user__ = 1; \
    static ljtBOOL __profile__initialized__ = ljvFALSE; \
    ljmSTACK_PUSH(&__user__, __FUNCTION__, __LINE__, Text, __VA_ARGS__); \
    ljoOS_ProfileDB(__FUNCTION__, &__profile__initialized__)

#define ljmFOOTER() \
    ljmSTACK_POP(&__user__, __FUNCTION__); \
    ljoOS_ProfileDB(__FUNCTION__, ljvNULL)

#define ljmFOOTER_NO() \
    ljmSTACK_POP(&__user__, __FUNCTION__); \
    ljoOS_ProfileDB(__FUNCTION__, ljvNULL)

#define ljmFOOTER_ARG(...) \
    ljmSTACK_POP(&__user__, __FUNCTION__); \
    ljoOS_ProfileDB(__FUNCTION__, ljvNULL)

#define ljmFOOTER_KILL() \
    ljmSTACK_POP(&__user__, __FUNCTION__); \
    ljoOS_ProfileDB(ljvNULL, ljvNULL)

#else /* !ljdENABLE_PROFILING */

#if ljdEMPTY_HEADER_FOOTER
#   define ljmHEADER()
#elif ljdHAS_ELLIPSIS
#define ljmHEADER() \
    ljtINT8 __user__ = 1; \
    ljtINT8_PTR __user_ptr__ = &__user__; \
    ljmSTACK_PUSH(__user_ptr__, __FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
    ljmSYSTRACE_BEGIN(_LJ_OBJ_ZONE, __FUNCTION__); \
    ljmBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
    ljmTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                  "++%s(%d)", __FUNCTION__, __LINE__)
#else
    ljmINLINE static void
    __dummy_header(void)
    {
    }
# define ljmHEADER                   __dummy_header
#endif

#if ljdHAS_ELLIPSIS
#if ljdEMPTY_HEADER_FOOTER
# define ljmHEADER_ARG(Text, ...)
#else
# define ljmHEADER_ARG(Text, ...) \
    ljtINT8 __user__ = 1; \
    ljtINT8_PTR __user_ptr__ = &__user__; \
    ljmSTACK_PUSH(__user_ptr__, __FUNCTION__, __LINE__, Text, __VA_ARGS__); \
    ljmSYSTRACE_BEGIN(_LJ_OBJ_ZONE, __FUNCTION__); \
    ljmBINARY_TRACE(__FUNCTION__, __LINE__, Text, __VA_ARGS__); \
    ljmTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                  "++%s(%d): " Text, __FUNCTION__, __LINE__, __VA_ARGS__)
#endif
#else
    ljmINLINE static void
    __dummy_header_arg(
        IN ljtCONST_STRING Text,
        ...
        )
    {
    }
# define ljmHEADER_ARG                __dummy_header_arg
#endif

#if ljdEMPTY_HEADER_FOOTER
# define ljmFOOTER()
#elif ljdHAS_ELLIPSIS
# define ljmFOOTER() \
    ljmSTACK_POP(__user_ptr__, __FUNCTION__); \
    ljmSYSTRACE_END(_LJ_OBJ_ZONE); \
    ljmBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
    ljmTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                  "--%s(%d): status=%d(%s)", \
                  __FUNCTION__, __LINE__, \
                  status, ljoOS_DebugStatus2Name(status)); \
    *__user_ptr__ -= 1
#else
    ljmINLINE static void
    __dummy_footer(void)
    {
    }
# define ljmFOOTER                    __dummy_footer
#endif

#if ljdEMPTY_HEADER_FOOTER
# define ljmFOOTER_NO()
#elif ljdHAS_ELLIPSIS
#define ljmFOOTER_NO() \
    ljmSTACK_POP(__user_ptr__, __FUNCTION__); \
    ljmSYSTRACE_END(_LJ_OBJ_ZONE); \
    ljmBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
    ljmTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                  "--%s(%d)", __FUNCTION__, __LINE__); \
    *__user_ptr__ -= 1
#else
    ljmINLINE static void
    __dummy_footer_no(void)
    {
    }
# define ljmFOOTER_NO                 __dummy_footer_no
#endif

#if ljdEMPTY_HEADER_FOOTER
# define ljmFOOTER_KILL()
#elif ljdHAS_ELLIPSIS
#define ljmFOOTER_KILL() \
    ljmSTACK_POP(__user_ptr__, __FUNCTION__); \
    ljmSYSTRACE_END(_LJ_OBJ_ZONE); \
    ljmBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
    ljmTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                  "--%s(%d)", __FUNCTION__, __LINE__); \
    *__user_ptr__ -= 1
#else
    ljmINLINE static void
    __dummy_footer_kill(void)
    {
    }
# define ljmFOOTER_KILL               __dummy_footer_kill
#endif

#if ljdHAS_ELLIPSIS
#if ljdEMPTY_HEADER_FOOTER
# define ljmFOOTER_ARG(Text, ...)
#else
# define ljmFOOTER_ARG(Text, ...) \
    ljmSTACK_POP(__user_ptr__, __FUNCTION__); \
    ljmSYSTRACE_END(_LJ_OBJ_ZONE); \
    ljmBINARY_TRACE(__FUNCTION__, __LINE__, Text, __VA_ARGS__); \
    ljmTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                  "--%s(%d): " Text, __FUNCTION__, __LINE__, __VA_ARGS__); \
    *__user_ptr__ -= 1
#endif
#else
    ljmINLINE static void
    __dummy_footer_arg(
        IN ljtCONST_STRING Text,
        ...
        )
    {
    }
# define ljmFOOTER_ARG                __dummy_footer_arg
#endif

#endif /* ljdENABLE_PROFILING */

#if ljdHAS_ELLIPSIS
# define ljmkHEADER()                                               \
    do {                                                            \
        ljmkBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
        ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                       "++%s(%d)", __FUNCTION__, __LINE__);         \
    } while (0)
#else
    ljmINLINE static void
    __dummy_kheader(void)
    {
    }
# define ljmkHEADER              __dummy_kheader
#endif

#if ljdHAS_ELLIPSIS
# define ljmkHEADER_ARG(Text, ...)                                   \
    do {                                                             \
        ljmkBINARY_TRACE(__FUNCTION__, __LINE__, Text, __VA_ARGS__); \
        ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                       "++%s(%d): " Text, \
                       __FUNCTION__, __LINE__, __VA_ARGS__);         \
    } while (0)
#else
    ljmINLINE static void
    __dummy_kheader_arg(IN ljtCONST_STRING Text, ...)
    {
    }
# define ljmkHEADER_ARG          __dummy_kheader_arg
#endif

#if ljdHAS_ELLIPSIS
# define ljmkFOOTER()                                              \
    do {                                                           \
        ljmkBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, status); \
        ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                       "--%s(%d): status=%d(%s)", \
                       __FUNCTION__, __LINE__, status, \
                       ljkOS_DebugStatus2Name(status));            \
    } while (0)
#else
    ljmINLINE static void
    __dummy_kfooter(void)
    {
    }
# define ljmkFOOTER              __dummy_kfooter
#endif

#if ljdHAS_ELLIPSIS
# define ljmkFOOTER_NO()                                            \
    do {                                                            \
        ljmkBINARY_TRACE(__FUNCTION__, __LINE__, ljvNULL, ljvNULL); \
        ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                       "--%s(%d)", __FUNCTION__, __LINE__);         \
    } while (0)
#else
    ljmINLINE static void
    __dummy_kfooter_no(void)
    {
    }
# define ljmkFOOTER_NO           __dummy_kfooter_no
#endif

#if ljdHAS_ELLIPSIS
# define ljmkFOOTER_ARG(Text, ...)                                   \
    do {                                                             \
        ljmkBINARY_TRACE(__FUNCTION__, __LINE__, Text, __VA_ARGS__); \
        ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, \
                       "--%s(%d): " Text, \
                       __FUNCTION__, __LINE__, __VA_ARGS__);         \
    } while (0)
#else
    ljmINLINE static void
    __dummy_kfooter_arg(IN ljtCONST_STRING Text, ...)
    {
    }
# define ljmkFOOTER_ARG          __dummy_kfooter_arg
#endif

#define ljmOPT_VALUE(ptr)              (((ptr) == ljvNULL) ? 0 : *(ptr))
#define ljmOPT_VALUE_INDEX(ptr, index) (((ptr) == ljvNULL) ? 0 : ptr[index])
#define ljmOPT_POINTER(ptr)            (((ptr) == ljvNULL) ? ljvNULL : *(ptr))
#define ljmOPT_STRING(ptr)             (((ptr) == ljvNULL) ? "(nil)" : (ptr))

#if ljdENABLE_DRM_DEBUG
# define ljmkHEADER_DRM()                                               \
        do {                                                            \
            ljkOS_DebugTraceZone(ljvLEVEL_VERBOSE, _LJ_OBJ_ZONE, \
                           "++%s(%d)", __FUNCTION__, __LINE__);         \
        } while (0)

# define ljmkHEADER_DRM_ARG(Text, ...)                               \
    do {                                                             \
        ljkOS_DebugTraceZone(ljvLEVEL_VERBOSE, _LJ_OBJ_ZONE, \
                       "++%s(%d): " Text, \
                       __FUNCTION__, __LINE__, __VA_ARGS__);         \
    } while (0)

# define ljmkFOOTER_DRM()                                              \
        do {                                                           \
            ljkOS_DebugTraceZone(ljvLEVEL_VERBOSE, _LJ_OBJ_ZONE, \
                           "--%s(%d): status=%d(%s)", \
                           __FUNCTION__, __LINE__, status, \
                           ljkOS_DebugStatus2Name(status));            \
        } while (0)

# define ljmkFOOTER_DRM_ARG(Text, ...)                               \
    do {                                                             \
        ljkOS_DebugTraceZone(ljvLEVEL_VERBOSE, _LJ_OBJ_ZONE, \
                       "--%s(%d): " Text, \
                       __FUNCTION__, __LINE__, __VA_ARGS__);         \
    } while (0)
#else
# define ljmkHEADER_DRM()                                            \
        do {                                                         \
        } while (0)

# define ljmkHEADER_DRM_ARG(Text, ...)                               \
    do {                                                             \
    } while (0)

# define ljmkFOOTER_DRM()                                            \
        do {                                                         \
        } while (0)

# define ljmkFOOTER_DRM_ARG(Text, ...)                               \
    do {                                                             \
    } while (0)

#endif

void
ljkOS_Print(IN ljtCONST_STRING Message, ...) CHECK_PRINTF_FORMAT(1, 2);

void
ljoOS_Print(IN ljtCONST_STRING Message, ...) CHECK_PRINTF_FORMAT(1, 2);

#define ljmPRINT                       ljoOS_Print
#define ljmkPRINT                      ljkOS_Print
#define ljmkPRINT_N(ArgumentSize, ...) ljkOS_Print(__VA_ARGS__)

#if ljdHAS_ELLIPSIS
# define ljmPRINT_ONCE(Text, ...)          \
    {                                      \
        static ljtBOOL _once = ljvFALSE;   \
        if (!_once) {                      \
            ljmPRINT(Text, __VA_ARGS__);   \
            _once = ljvTRUE;               \
        }                                  \
    }

#else
    ljmINLINE static void
    __dummy_printonce_arg(IN ljtCONST_STRING Text, ...)
    {
    }
# define ljmPRINT_ONCE           __dummy_printonce_arg
#endif

#if ljdFEATURE_SANITYCHECK
# define ljmFEATURE_CHECKED      ljmPRINT_ONCE
#else
# define ljmFEATURE_CHECKED(Text, ...)
#endif

#if ljdPRINT_VERSION
# define ljmPRINT_VERSION()          \
    do {                             \
        _ljmPRINT_VERSION(ljm);      \
        ljmSTACK_DUMP();             \
    } while (0)
# define ljmkPRINT_VERSION()     _ljmPRINT_VERSION(ljmk)
# define _ljmPRINT_VERSION(prefix)    \
        prefix##TRACE(ljvLEVEL_ERROR, "Ljmicro HAL version %s", ljvVERSION_STRING)
#else
# define ljmPRINT_VERSION()          \
    do {                             \
        ljmSTACK_DUMP();             \
    } while (ljvFALSE)
# define ljmkPRINT_VERSION()         \
    do {                             \
    } while (ljvFALSE)
#endif

void
ljkOS_Dump(IN ljkOS Os, IN ljtCONST_STRING Format, ...);

void
ljkOS_DumpBuffer(IN ljkOS Os,
                 IN ljeDUMP_BUFFER_TYPE Type,
                 IN ljtPOINTER Buffer,
                 IN ljtUINT64 Address,
                 IN ljtSIZE_T Size);

#if ljdDUMP_IN_KERNEL
# define ljmkDUMP            ljkOS_Dump

# define ljmkDUMP_BUFFER     ljkOS_DumpBuffer
#else
# define ljmkDUMP(...)               \
    do {                             \
    } while (0)
# define ljmkDUMP_BUFFER(...)        \
    do {                             \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmDUMP_FRAMERATE
 **
 **      Print average frame rate
 **
 */
ljeSTATUS
ljoOS_DumpFrameRate(void);
#define ljmDUMP_FRAMERATE       ljoOS_DumpFrameRate

/*******************************************************************************
 **
 **  ljoOS_SetDumpFlag
 **
 **      Dump print switch.
 **
 **  ARGUMENTS:
 **
 **      DumpState
 **          True to enable dump prints.
 */

ljeSTATUS
ljoOS_SetDumpFlag(IN ljtBOOL DumpState);

/*******************************************************************************
 **
 **  ljmDUMP
 **
 **      Print a dump message.
 **
 **  ARGUMENTS:
 **
 **      ljtSTRING   Message.
 **
 **      ...         Optional arguments.
 */

#if ljdDUMP
ljeSTATUS
ljoOS_Dump(IN ljoOS Os, IN ljtCONST_STRING String, ...);
# define ljmDUMP     ljoOS_Dump
#else
# define ljmDUMP(...)              \
    do {                           \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmDUMP_BUFFER
 **
 **      Print a buffer to the dump.
 **
 **  ARGUMENTS:
 **
 **      ljtSTRING Tag
 **          Tag for dump.
 **
 **      ljtADDRESS Address
 **          GPU address of buffer.
 **
 **      ljtPOINTER Logical
 **          Logical address of buffer.
 **
 **      ljtUINT32 Offset
 **          Offset into buffer.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes.
 */

#if ljdDUMP
ljeSTATUS
ljoOS_DumpBuffer(IN ljoOS Os,
                 IN ljeDUMP_BUFFER_TYPE Type,
                 IN ljtADDRESS Address,
                 IN ljtPOINTER Logical,
                 IN ljtSIZE_T Offset,
                 IN ljtSIZE_T Bytes);
# define ljmDUMP_BUFFER      ljoOS_DumpBuffer
#else
# define ljmDUMP_BUFFER(...)     \
    do {                         \
    } while (0)
#endif

#if ljdDUMP
void
ljoOS_DumpLock(void);
# define ljmDUMP_LOCK        ljoOS_DumpLock
#else
# define ljmDUMP_LOCK(...)          \
    do {                            \
    } while (0)
#endif

#if ljdDUMP
void
ljoOS_DumpUnlock(void);
# define ljmDUMP_UNLOCK      ljoOS_DumpUnlock
#else
# define ljmDUMP_UNLOCK(...)        \
    do {                            \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmDUMP_API
 **
 **      Print a dump message for a high level API prefixed by the function name.
 **
 **  ARGUMENTS:
 **
 **      ljtSTRING   Message.
 **
 **      ...         Optional arguments.
 */
ljeSTATUS
ljoOS_DumpApi(IN ljtCONST_STRING String, ...);
#if ljdDUMP_API
# define ljmDUMP_API         ljoOS_DumpApi
#else
# define ljmDUMP_API(...)           \
    do {                            \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmDUMP_API_ARRAY
 **
 **      Print an array of data.
 **
 **  ARGUMENTS:
 **
 **      ljtUINT32_PTR   Pointer to array.
 **      ljtUINT32       Size.
 */
ljeSTATUS
ljoOS_DumpArray(IN ljtCONST_POINTER Data, IN ljtUINT32 Size);
#if ljdDUMP_API
# define ljmDUMP_API_ARRAY       ljoOS_DumpArray
#else
# define ljmDUMP_API_ARRAY(...)     \
    do {                            \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmDUMP_API_ARRAY_TOKEN
 **
 **      Print an array of data terminated by a token.
 **
 **  ARGUMENTS:
 **
 **      ljtUINT32_PTR   Pointer to array.
 **      ljtUINT32       Termination.
 */
ljeSTATUS
ljoOS_DumpArrayToken(IN ljtCONST_POINTER Data, IN ljtUINT32 Termination);
#if ljdDUMP_API
# define ljmDUMP_API_ARRAY_TOKEN         ljoOS_DumpArrayToken
#else
# define ljmDUMP_API_ARRAY_TOKEN(...) \
    do {                              \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmDUMP_API_DATA
 **
 **      Print an array of bytes.
 **
 **  ARGUMENTS:
 **
 **      ljtCONST_POINTER    Pointer to array.
 **      ljtSIZE_T           Size.
 */
ljeSTATUS
ljoOS_DumpApiData(IN ljtCONST_POINTER Data, IN ljtSIZE_T Size);
#if ljdDUMP_API
# define ljmDUMP_API_DATA        ljoOS_DumpApiData
#else
# define ljmDUMP_API_DATA(...)       \
    do {                             \
    } while (0)
#endif

/*******************************************************************************
 ** ljmDUMP_2D_COMMAND
 **
 **      Print the 2D command buffer.
 **
 **  ARGUMENTS:
 **
 **      ljtUINT32_PTR       Pointer to the command buffer.
 **      ljtUINT32           Command buffer size.
 */
ljeSTATUS
ljoOS_Dump2DCommand(IN ljtUINT32_PTR Command, IN ljtUINT32 Size);
#if ljdDUMP_2D
# define ljmDUMP_2D_COMMAND(cmd, size)      \
    do {                                    \
        if (Hardware->newDump2DLevel > 1)   \
            ljoOS_Dump2DCommand(cmd, size); \
    } while (0)
#else
# define ljmDUMP_2D_COMMAND(...)        \
    do {                                \
    } while (0)
#endif

/*******************************************************************************
 ** ljmDUMP_2D_SURFACE
 **
 **      Print the 2D surface memory.
 **
 **  ARGUMENTS:
 **
 **      ljtBOOL             Src.
 **      ljtADDRESS          Address.
 */
ljeSTATUS
ljoOS_Dump2DSurface(IN ljtBOOL Src, IN ljtADDRESS Address);
#if ljdDUMP_2D
# define ljmDUMP_2D_SURFACE(src, addr)      \
    do {                                    \
        if (Hardware->newDump2DLevel > 2)   \
            ljoOS_Dump2DSurface(src, addr); \
    } while (0)
#else
# define ljmDUMP_2D_SURFACE(...)        \
    do {                                \
    } while (0)
#endif

/*******************************************************************************
 ** ljmDUMP_ADD_MEMORY_INFO
 **
 **      Record the memory info.
 **
 **  ARGUMENTS:
 **
 **      ljtADDRESS          Address.
 **      ljtSIZE_T           Size.
 */
ljeSTATUS
ljfAddMemoryInfo(IN ljtADDRESS GPUAddress,
                 IN ljtPOINTER Logical,
                 IN ljtUINT64 Physical,
                 IN ljtUINT32 Size);
#if ljdDUMP_2D
# define ljmDUMP_ADD_MEMORY_INFO         ljfAddMemoryInfo
#else
# define ljmDUMP_ADD_MEMORY_INFO(...)  \
    do {                               \
    } while (0)
#endif

/*******************************************************************************
 ** ljmDUMP_DEL_MEMORY_INFO
 **
 **      Record the memory info.
 **
 **  ARGUMENTS:
 **
 **      ljtADDRESS Address.
 */
ljeSTATUS
ljfDelMemoryInfo(IN ljtADDRESS Address);
#if ljdDUMP_2D
# define ljmDUMP_DEL_MEMORY_INFO         ljfDelMemoryInfo
#else
# define ljmDUMP_DEL_MEMORY_INFO(...) \
    do {                              \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmTRACE_RELEASE
 **
 **      Print a message to the shader debugger.
 **
 **  ARGUMENTS:
 **
 **      ljtCONST_STRING  Message.
 **      ...              Optional arguments.
 */

#define ljmTRACE_RELEASE        ljoOS_DebugShaderTrace

void
ljoOS_DebugShaderTrace(IN ljtCONST_STRING Message, ...);

void
ljoOS_SetDebugShaderFiles(IN ljtCONST_STRING VSFileName, IN ljtCONST_STRING FSFileName);

void
ljoOS_SetDebugShaderFileType(IN ljtUINT32 ShaderType);

void
ljoOS_EnableDebugBuffer(IN ljtBOOL Enable);

/*******************************************************************************
 **
 **  ljmBREAK
 **
 **      Break into the debugger.  In retail mode this macro does nothing.
 **
 **  ARGUMENTS:
 **
 **      None.
 */

void
ljoOS_DebugBreak(void);

void
ljkOS_DebugBreak(void);

#if ljmIS_DEBUG(ljdDEBUG_BREAK)
# define ljmBREAK        ljoOS_DebugBreak
# define ljmkBREAK       ljkOS_DebugBreak
#else
# define ljmBREAK()
# define ljmkBREAK()
#endif

/*******************************************************************************
 **
 **  ljmASSERT
 **
 **      Evaluate an expression and break into the debugger if the expression
 **      evaluates to false.  In retail mode this macro does nothing.
 **
 **  ARGUMENTS:
 **
 **      exp     Expression to evaluate.
 */
#if ljmIS_DEBUG(ljdDEBUG_ASSERT)
# define _ljmASSERT(prefix, exp) \
    do \
    { \
        if (!(exp)) \
        { \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "ASSERT at %s(%d)", \
                          __FUNCTION__, __LINE__); \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          "(%s)", #exp); \
            prefix##BREAK(); \
        } \
    } \
    while (ljvFALSE)
# define ljmASSERT(exp)          _ljmASSERT(ljm, exp)
# define ljmkASSERT(exp)         _ljmASSERT(ljmk, exp)
#else
# define ljmASSERT(exp)
# define ljmkASSERT(exp)
#endif

/*******************************************************************************
 **
 **  ljmSTATIC_ASSERT
 **
 **      Tests a software assertion at compile time. If the specific constant
 **      expression is false, the compiler displays the specified message and
 **      the compilation fails with an error, otherwise the declaration has
 **      no effect.
 **      Static assert is suitable for both user and kernel space.
 **
 **  ARGUMENTS:
 **
 **      exp     Constant expression.
 **      message Error message displayed on assertion failure.
 */
#if defined(__GNUC__) && ((__GNUC__ == 4 && __GNUC_MINOR__ >= 6) || (__GNUC__ > 4))
# define ljmSTATIC_ASSERT(constExp, message) \
    do {                                     \
        _Static_assert((constExp), message); \
    }                                        \
    while (0)

#elif defined(_MSC_VER) && (_MSC_VER >= 1600)
# define ljmSTATIC_ASSERT(constExp, message) \
        static_assert((constExp), message)

#else
# define ljmSTATIC_ASSERT(constExp, message) \
    do {                                     \
    } while (0)
#endif

/*******************************************************************************
 **
 **  ljmVERIFY
 **
 **      Verify if an expression returns true.  If the expression does not
 **      evaluates to true, an assertion will happen in debug mode.
 **
 **  ARGUMENTS:
 **
 **      exp     Expression to evaluate.
 */
#if ljmIS_DEBUG(ljdDEBUG_ASSERT)
# define ljmVERIFY(exp)          ljmASSERT(exp)
# define ljmkVERIFY(exp)         ljmkASSERT(exp)
#else
# define ljmVERIFY(exp)          ((void)exp)
# define ljmkVERIFY(exp)         ((void)exp)
#endif

/*******************************************************************************
 **
 **  ljmVERIFY_OK
 **
 **      Verify a function returns ljvSTATUS_OK.  If the function does not return
 **      ljvSTATUS_OK, an assertion will happen in debug mode.
 **
 **  ARGUMENTS:
 **
 **      func    Function to evaluate.
 */

void
ljoOS_Verify(IN ljeSTATUS status);

void
ljkOS_Verify(IN ljeSTATUS status);

#if ljmIS_DEBUG(ljdDEBUG_ASSERT)
# define ljmVERIFY_OK(func)                                     \
    do {                                                        \
        ljeSTATUS verifyStatus = func;                          \
        ljoOS_Verify(verifyStatus);                             \
        if (verifyStatus != ljvSTATUS_OK) {                     \
            ljmTRACE(ljvLEVEL_ERROR, \
                     "ljmVERIFY_OK(%d): function returned %d", \
                     __LINE__, verifyStatus);                   \
        }                                                       \
        ljmASSERT(verifyStatus == ljvSTATUS_OK);                \
    } while (ljvFALSE)

# define ljmkVERIFY_OK(func)                                    \
    do {                                                        \
        ljeSTATUS verifyStatus = func;                          \
        if (verifyStatus != ljvSTATUS_OK) {                     \
            ljmkTRACE(ljvLEVEL_ERROR, \
                      "ljmkVERIFY_OK(%d): function returned %d",\
                      __LINE__, verifyStatus);                  \
        }                                                       \
        ljkOS_Verify(verifyStatus);                             \
        ljmkASSERT(verifyStatus == ljvSTATUS_OK);               \
    } while (ljvFALSE)
#else
# define ljmVERIFY_OK(func)          func
# define ljmkVERIFY_OK(func)         func
#endif

ljtCONST_STRING
ljoOS_DebugStatus2Name(ljeSTATUS status);

ljtCONST_STRING
ljkOS_DebugStatus2Name(ljeSTATUS status);

/*******************************************************************************
 **
 **  ljmERR_BREAK
 **
 **      Executes a break statement on error.
 **
 **  ASSUMPTIONS:
 **
 **      'status' variable of ljeSTATUS type must be defined.
 **
 **  ARGUMENTS:
 **
 **      func    Function to evaluate.
 */
#define _ljmERR_BREAK(prefix, func) {                              \
    status = func;                                                 \
    if (ljmIS_ERROR(status)) {                                     \
        prefix##PRINT_VERSION();                                   \
        prefix##TRACE(ljvLEVEL_ERROR, \
                      #prefix "ERR_BREAK: status=%d(%s) @ %s(%d)", \
                      status, ljoOS_DebugStatus2Name(status), \
                      __FUNCTION__, __LINE__);                     \
        break;                                                     \
    }                                                              \
    do { } while (ljvFALSE);                                       \
}

#define _ljmkERR_BREAK(prefix, func) {                             \
    status = func;                                                 \
    if (ljmIS_ERROR(status)) {                                     \
        prefix##PRINT_VERSION();                                   \
        prefix##TRACE(ljvLEVEL_ERROR, \
                      #prefix "ERR_BREAK: status=%d(%s) @ %s(%d)", \
                      status, ljkOS_DebugStatus2Name(status), \
                      __FUNCTION__, __LINE__);                     \
        break;                                                     \
    }                                                              \
    do { } while (ljvFALSE);                                       \
}

#define ljmERR_BREAK(func)          _ljmERR_BREAK(ljm, func)
#define ljmkERR_BREAK(func)         _ljmkERR_BREAK(ljmk, func)

/*******************************************************************************
 **
 **  ljmERR_RETURN
 **
 **      Executes a return on error.
 **
 **  ASSUMPTIONS:
 **
 **      'status' variable of ljeSTATUS type must be defined.
 **
 **  ARGUMENTS:
 **
 **      func    Function to evaluate.
 */
#define _ljmERR_RETURN(prefix, func)                                    \
    do {                                                                \
        status = func;                                                  \
        if (ljmIS_ERROR(status)) {                                      \
            prefix##PRINT_VERSION();                                    \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "ERR_RETURN: status=%d(%s) @ %s(%d)", \
                          status, ljoOS_DebugStatus2Name(status), \
                          __FUNCTION__, __LINE__);                      \
            prefix##FOOTER();                                           \
            return status;                                              \
        }                                                               \
    } while (ljvFALSE)
#define _ljmkERR_RETURN(prefix, func)                                   \
    do {                                                                \
        status = func;                                                  \
        if (ljmIS_ERROR(status)) {                                      \
            prefix##PRINT_VERSION();                                    \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "ERR_RETURN: status=%d(%s) @ %s(%d)", \
                          status, ljkOS_DebugStatus2Name(status), \
                          __FUNCTION__, __LINE__);                      \
            prefix##FOOTER();                                           \
            return status;                                              \
        }                                                               \
    } while (ljvFALSE)
#define ljmERR_RETURN(func)         _ljmERR_RETURN(ljm, func)
#define ljmkERR_RETURN(func)        _ljmkERR_RETURN(ljmk, func)

/*******************************************************************************
 **
 **  ljmONERROR
 **
 **      Jump to the error handler in case there is an error.
 **
 **  ASSUMPTIONS:
 **
 **      'status' variable of ljeSTATUS type must be defined.
 **
 **  ARGUMENTS:
 **
 **      func    Function to evaluate.
 */
#define _ljmONERROR(prefix, func)                                        \
    do {                                                                 \
        status = func;                                                   \
        if (ljmIS_ERROR(status)) {                                       \
            prefix##PRINT_VERSION();                                     \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "ONERROR: status=%d(%s) @ %s(%d)", \
                          status, ljoOS_DebugStatus2Name(status), \
                          __FUNCTION__, __LINE__);                       \
            goto OnError;                                                \
        }                                                                \
    } while (ljvFALSE)
#define _ljmkONERROR(prefix, func)                                       \
    do {                                                                 \
        status = func;                                                   \
        if (ljmIS_ERROR(status)) {                                       \
            prefix##PRINT_VERSION();                                     \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "ONERROR: status=%d(%s) @ %s(%d)", \
                          status, ljkOS_DebugStatus2Name(status), \
                          __FUNCTION__, __LINE__);                       \
            goto OnError;                                                \
        }                                                                \
    } while (ljvFALSE)

/* Ignore the debug info when the specific error occurs. */
#define _ljmkONERROR_EX(prefix, func, error)                             \
    do {                                                                 \
        status = func;                                                   \
        if (ljmIS_ERROR(status)) {                                       \
            if (status != (error)) {                                     \
                prefix##PRINT_VERSION();                                 \
                prefix##TRACE(ljvLEVEL_ERROR, \
                              #prefix "ONERROR: status=%d(%s) @ %s(%d)", \
                              status, ljkOS_DebugStatus2Name(status), \
                              __FUNCTION__, __LINE__);                   \
            }                                                            \
            goto OnError;                                                \
        }                                                                \
    } while (ljvFALSE)

#define ljmONERROR(func)                    _ljmONERROR(ljm, func)
#define ljmkONERROR(func)                   _ljmkONERROR(ljmk, func)
#define ljmkONERROR_EX(func, error)         _ljmkONERROR_EX(ljmk, func, error)

#define ljmGET_INDEX_SIZE(type, size) \
    switch (type) \
    { \
    case ljvINDEX_8: \
        size = 1; \
        break; \
    case ljvINDEX_16: \
        size = 2; \
        break; \
    case ljvINDEX_32: \
        size = 4; \
        break; \
    default: \
        ljmONERROR(ljvSTATUS_INVALID_ARGUMENT); \
    } \

/*******************************************************************************
 **
 **  ljmkSAFECASTSIZET
 **
 **      Check whether value of a ljtSIZE_T variable beyond the capability
 **      of 32bits GPU hardware.
 **
 **  ASSUMPTIONS:
 **
 **
 **
 **  ARGUMENTS:
 **
 **      x   A ljtUINT32 variable
 **      y   A ljtSIZE_T variable
 */
#define ljmkSAFECASTSIZET(x, y)                                   \
    do {                                                          \
        ljtUINT32 tmp = (ljtUINT32)(y);                           \
        if (ljmSIZEOF(ljtSIZE_T) > ljmSIZEOF(ljtUINT32)) {        \
            ljmkASSERT(tmp <= ljvMAXUINT32);                      \
        }                                                         \
        (x) = tmp;                                                \
    } while (ljvFALSE)

#define ljmSAFECASTSIZET(x, y)                                    \
    do {                                                          \
        ljtUINT32 tmp = (ljtUINT32)(y);                           \
        if (ljmSIZEOF(ljtSIZE_T) > ljmSIZEOF(ljtUINT32)) {        \
            ljmASSERT(tmp <= ljvMAXUINT32);                       \
        }                                                         \
        (x) = tmp;                                                \
    } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmkSAFECASTPHYSADDRT
 **
 **      Check whether value of a ljtPHYS_ADDR_T variable beyond the capability
 **      of 32bits GPU hardware.
 **
 **  ASSUMPTIONS:
 **
 **
 **
 **  ARGUMENTS:
 **
 **      x   A ljtUINT32 variable
 **      y   A ljtPHYS_ADDR_T variable
 */
#define ljmkSAFECASTPHYSADDRT(x, y)                               \
    do {                                                          \
        ljtUINT32 tmp = (ljtUINT32)(y);                           \
        if (ljmSIZEOF(ljtPHYS_ADDR_T) > ljmSIZEOF(ljtUINT32)) {   \
            ljmkASSERT(tmp <= ljvMAXUINT32);                      \
        }                                                         \
        (x) = tmp;                                                \
    } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmSAFECASTPHYSADDRT
 **
 **      Check whether value of a ljtPHYS_ADDR_T variable beyond the capability
 **      of 32bits GPU hardware.
 **
 **  ASSUMPTIONS:
 **
 **
 **
 **  ARGUMENTS:
 **
 **      x   A ljtUINT32 variable
 **      y   A ljtPHYS_ADDR_T variable
 */
#define ljmSAFECASTPHYSADDRT(x, y)                                \
    do {                                                          \
        ljtUINT32 tmp = (ljtUINT32)(y);                           \
        if (ljmSIZEOF(ljtPHYS_ADDR_T) > ljmSIZEOF(ljtUINT32)) {   \
            ljmASSERT(tmp <= ljvMAXUINT32);                       \
        }                                                         \
        (x) = tmp;                                                \
    } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmkSAFECASTVA
 **
 **      Check whether value of a ljtADDRESS variable beyond the capability
 **      of 32bits GPU hardware.
 **
 **  ASSUMPTIONS:
 **
 **
 **
 **  ARGUMENTS:
 **
 **      x   A ljtUINT32 variable
 **      y   A ljtADDRESS variable
 */
#define ljmkSAFECASTVA(x, y)                                     \
    do {                                                         \
        ljtUINT32 tmp = (ljtUINT32)(y);                          \
        if (ljmSIZEOF(ljtADDRESS) > ljmSIZEOF(ljtUINT32)) {      \
            ljmkASSERT(tmp <= ljvMAXUINT32);                     \
        }                                                        \
        (x) = tmp;                                               \
    } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmSAFECASTVA
 **
 **      Check whether value of a ljtADDRESS variable beyond the capability
 **      of 32bits GPU hardware.
 **
 **  ASSUMPTIONS:
 **
 **
 **
 **  ARGUMENTS:
 **
 **      x   A ljtUINT32 variable
 **      y   A ljtADDRESS variable
 */
#define ljmSAFECASTVA(x, y)                                      \
    do {                                                         \
        ljtUINT32 tmp = (ljtUINT32)(y);                          \
        if (ljmSIZEOF(ljtADDRESS) > ljmSIZEOF(ljtUINT32)) {      \
            ljmASSERT(tmp <= ljvMAXUINT32);                      \
        }                                                        \
        (x) = tmp;                                               \
    } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmVERIFY_LOCK
 **
 **      Verifies whether the surface is locked.
 **
 **  ARGUMENTS:
 **
 **      surfaceInfo Pointer to the surface iniformational structure.
 */
#define ljmVERIFY_LOCK(surfaceInfo) \
    if (!surfaceInfo->node.valid) \
    { \
        ljmONERROR(ljvSTATUS_MEMORY_UNLOCKED); \
    } \

/*******************************************************************************
 **
 **  ljmVERIFY_NODE_LOCK
 **
 **      Verifies whether the surface node is locked.
 **
 **  ARGUMENTS:
 **
 **      surfaceInfo Pointer to the surface iniformational structure.
 */
#define ljmVERIFY_NODE_LOCK(surfaceNode) \
    if (!(surfaceNode)->valid) \
    { \
        status = ljvSTATUS_MEMORY_UNLOCKED; \
        break; \
    } \
    do { } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmBADOBJECT_BREAK
 **
 **      Executes a break statement on bad object.
 **
 **  ARGUMENTS:
 **
 **      obj     Object to test.
 **      t       Expected type of the object.
 */
#define ljmBADOBJECT_BREAK(obj, t) \
    if ((obj == ljvNULL) \
    ||  (((ljsOBJECT *)(obj))->type != t) \
    ) \
    { \
        status = ljvSTATUS_INVALID_OBJECT; \
        break; \
    } \
    do { } while (ljvFALSE)

/*******************************************************************************
 **
 **  ljmCHECK_STATUS
 **
 **      Executes a break statement on error.
 **
 **  ASSUMPTIONS:
 **
 **      'status' variable of ljeSTATUS type must be defined.
 **
 **  ARGUMENTS:
 **
 **      func    Function to evaluate.
 */
#define _ljmCHECK_STATUS(prefix, func)                                    \
    do {                                                                  \
        last = func;                                                      \
        if (ljmIS_ERROR(last)) {                                          \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "CHECK_STATUS: status=%d(%s) @ %s(%d)", \
                          last, ljoOS_DebugStatus2Name(last), \
                          __FUNCTION__, __LINE__);                        \
            status = last;                                                \
        }                                                                 \
    } while (ljvFALSE)
#define _ljmkCHECK_STATUS(prefix, func)                                   \
    do {                                                                  \
        last = func;                                                      \
        if (ljmIS_ERROR(last)) {                                          \
            prefix##TRACE(ljvLEVEL_ERROR, \
                          #prefix "CHECK_STATUS: status=%d(%s) @ %s(%d)", \
                          last, ljkOS_DebugStatus2Name(last), \
                          __FUNCTION__, __LINE__);                        \
            status = last;                                                \
        }                                                                 \
    } while (ljvFALSE)
#define ljmCHECK_STATUS(func)       _ljmCHECK_STATUS(ljm, func)
#define ljmkCHECK_STATUS(func)      _ljmkCHECK_STATUS(ljmk, func)

/*******************************************************************************
 **
 **  ljmVERIFY_ARGUMENT
 **
 **      Assert if an argument does not apply to the specified expression.  If
 **      the argument evaluates to false, ljvSTATUS_INVALID_ARGUMENT will be
 **      returned from the current function.  In retail mode this macro does
 **      nothing.
 **
 **  ARGUMENTS:
 **
 **      arg     Argument to evaluate.
 */
#define _ljmVERIFY_ARGUMENT(prefix, arg) \
       do \
       { \
           if (!(arg)) \
           { \
               prefix##TRACE(ljvLEVEL_ERROR, #prefix "VERIFY_ARGUMENT failed:"); \
               prefix##ASSERT(arg); \
               prefix##FOOTER_ARG("status=%d", ljvSTATUS_INVALID_ARGUMENT); \
               return ljvSTATUS_INVALID_ARGUMENT; \
           } \
       } \
       while (ljvFALSE)
#define ljmVERIFY_ARGUMENT(arg)             _ljmVERIFY_ARGUMENT(ljm, arg)
#define ljmkVERIFY_ARGUMENT(arg)            _ljmVERIFY_ARGUMENT(ljmk, arg)

/*******************************************************************************
 **
 **  ljmDEBUG_VERIFY_ARGUMENT
 **
 **      Works just like ljmVERIFY_ARGUMENT, but is only valid in debug mode.
 **      Use this to verify arguments inside non-public API functions.
 */
#if ljdDEBUG
# define ljmDEBUG_VERIFY_ARGUMENT(arg)   _ljmVERIFY_ARGUMENT(ljm, arg)
# define ljmkDEBUG_VERIFY_ARGUMENT(arg)  _ljmkVERIFY_ARGUMENT(ljm, arg)
#else
# define ljmDEBUG_VERIFY_ARGUMENT(arg)
# define ljmkDEBUG_VERIFY_ARGUMENT(arg)
#endif

/*******************************************************************************
 **
 **  _ljmCHECK_ADD_OVERFLOW
 */
#define _ljmCHECK_ADD_OVERFLOW(x, y)                                                            \
(\
    ((x) > 0 && (y) > 0 && ljvMAXSIZE_T - (x) < (y)) ? ljvSTATUS_RESLUT_OVERFLOW : ljvSTATUS_OK \
)

#define ljmCHECK_ADD_OVERFLOW(x, y)                 _ljmCHECK_ADD_OVERFLOW(x, y)
#define ljmkCHECK_ADD_OVERFLOW(x, y)                _ljmCHECK_ADD_OVERFLOW(x, y)

#define MAX_LOOP_COUNT 0x7FFFFFFF

/******************************************************************************
 ****************************** User Debug Option *****************************
 ******************************************************************************/

typedef struct _ljsUSER_DEBUG_OPTION {
    ljeDEBUG_MSG debugMsg;
} ljsUSER_DEBUG_OPTION;

ljsUSER_DEBUG_OPTION *
ljoHAL_GetUserDebugOption(void);

#if ljdHAS_ELLIPSIS
# define ljmUSER_DEBUG_MSG(level, ...) \
    do \
    { \
        if (level <= ljoHAL_GetUserDebugOption()->debugMsg) \
        { \
            ljoOS_Print(__VA_ARGS__); \
        } \
    } while (ljvFALSE)

# define ljmUSER_DEBUG_ERROR_MSG(...)   ljmUSER_DEBUG_MSG(ljvDEBUG_MSG_ERROR, "Error: " __VA_ARGS__)
# define ljmUSER_DEBUG_WARNING_MSG(...) ljmUSER_DEBUG_MSG(ljvDEBUG_MSG_WARNING, "Warring: " __VA_ARGS__)
#else
# define ljmUSER_DEBUG_MSG
# define ljmUSER_DEBUG_ERROR_MSG
# define ljmUSER_DEBUG_WARNING_MSG
#endif

/*******************************************************************************
 **
 **  A set of macros to aid state loading.
 **
 **  ARGUMENTS:
 **
 **      CommandBuffer   Pointer to a ljoCMDBUF object.
 **      StateDelta      Pointer to a ljsSTATE_DELTA state delta structure.
 **      Memory          Destination memory pointer of ljtUINT32_PTR type.
 **      PartOfContext   Whether or not the state is a part of the context.
 **      FixedPoint      Whether or not the state is of the fixed point format.
 **      Count           Number of consecutive states to be loaded.
 **      Address         State address.
 **      Data            Data to be set to the state.
 */

/*----------------------------------------------------------------------------*/

#if ljmIS_DEBUG(ljdDEBUG_CODE)

# define ljmSTORELOADSTATE(CommandBuffer, Memory, Address, Count) \
        CommandBuffer->lastLoadStatePtr = ljmPTR_TO_UINT64(Memory); \
        CommandBuffer->lastLoadStateAddress = Address; \
        CommandBuffer->lastLoadStateCount = Count

# define ljmVERIFYLOADSTATE(CommandBuffer, Memory, Address) \
        ljmASSERT(\
            (ljtUINT) (Memory  - ljmUINT64_TO_TYPE(CommandBuffer->lastLoadStatePtr, ljtUINT32_PTR) - 1) \
            == \
            (ljtUINT) (Address - CommandBuffer->lastLoadStateAddress) \
            ); \
        \
        ljmASSERT(CommandBuffer->lastLoadStateCount > 0); \
        \
        CommandBuffer->lastLoadStateCount -= 1

# define ljmVERIFYLOADSTATEDONE(CommandBuffer) \
        ljmASSERT(CommandBuffer->lastLoadStateCount == 0);

# define ljmDEFINELOADSTATEBASE() \
        ljtUINT32_PTR LoadStateBase;

# define ljmSETLOADSTATEBASE(CommandBuffer, OutSide) \
        if (OutSide) \
        {\
            LoadStateBase = (ljtUINT32_PTR)*OutSide; \
        }\
        else\
        {\
            LoadStateBase = (ljtUINT_PTR)CommandBuffer->buffer;\
        }


# define ljmVERIFYLOADSTATEALIGNED(CommandBuffer, Memory) \
        ljmASSERT(((Memory - LoadStateBase) & 1) == 0);

# define ljmUNSETLOADSTATEBASE() \
        LoadStateBase = LoadStateBase;

#else

# define ljmSTORELOADSTATE(CommandBuffer, Memory, Address, Count)
# define ljmVERIFYLOADSTATE(CommandBuffer, Memory, Address)
# define ljmVERIFYLOADSTATEDONE(CommandBuffer)

# define ljmDEFINELOADSTATEBASE()
# define ljmSETLOADSTATEBASE(CommandBuffer, OutSide)
# define ljmVERIFYLOADSTATEALIGNED(CommandBuffer, Memory)
# define ljmUNSETLOADSTATEBASE()

#endif

/*----------------------------------------------------------------------------*/

#define ljmDUMPSTATEDATA(StateDelta, FixedPoint, Address, Data)

#define ljmDEFINESTATEBUFFER(CommandBuffer, StateDelta, Memory, ReserveSize) \
    ljtSIZE_T ReserveSize;                                                   \
    ljoCMDBUF CommandBuffer;                                                 \
    ljtUINT32_PTR Memory;                                                    \
    ljsSTATE_DELTA_PTR StateDelta;                                           \
    ljeENGINE CurrentEngine = ljvENGINE_RENDER

#define ljmBEGINSTATEBUFFER(Hardware, CommandBuffer, StateDelta, Memory, ReserveSize) \
    {                                                                                 \
        ljmONERROR(ljoBUFFER_Reserve(Hardware->engine[CurrentEngine].buffer, \
                                     ReserveSize, ljvTRUE, \
                                     ljvCOMMAND_3D, &CommandBuffer));                 \
                                                                                      \
        Memory = (ljtUINT32_PTR)ljmUINT64_TO_PTR(CommandBuffer->lastReserve);         \
                                                                                      \
        StateDelta = Hardware->delta;                                                 \
    }

#define ljmENDSTATEBUFFER(Hardware, CommandBuffer, Memory, ReserveSize)                     \
    {                                                                                       \
        ljmASSERT(ljmUINT64_TO_TYPE(CommandBuffer->lastReserve, ljtUINT8_PTR) + ReserveSize \
                  == (ljtUINT8_PTR)Memory);                                                 \
    }

/*----------------------------------------------------------------------------*/

#define ljmBEGINSTATEBATCH(CommandBuffer, Memory, FixedPoint, Address, Count)                   \
    {                                                                                           \
        ljmASSERT(((Memory - ljmUINT64_TO_TYPE(CommandBuffer->lastReserve, ljtUINT32_PTR)) & 1) \
                  == 0);                                                                        \
        ljmASSERT((ljtUINT32)Count <= 1024);                                                    \
                                                                                                \
        ljmVERIFYLOADSTATEDONE(CommandBuffer);                                                  \
                                                                                                \
        ljmSTORELOADSTATE(CommandBuffer, Memory, Address, Count);                               \
                                                                                                \
        *Memory++ = ljmSETFIELDVALUE(0, AQ_COMMAND_LOAD_STATE_COMMAND, OPCODE, LOAD_STATE) |    \
                         ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, FLOAT, FixedPoint) |          \
                         ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, COUNT, Count) |               \
                         ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, ADDRESS, Address);            \
    }

#define ljmENDSTATEBATCH(CommandBuffer, Memory)                                                 \
    {                                                                                           \
        ljmVERIFYLOADSTATEDONE(CommandBuffer);                                                  \
                                                                                                \
        ljmASSERT(((Memory - ljmUINT64_TO_TYPE(CommandBuffer->lastReserve, ljtUINT32_PTR)) & 1) \
                  == 0);                                                                        \
    }

/*----------------------------------------------------------------------------*/

#define ljmSETSTATEDATA(StateDelta, CommandBuffer, Memory, \
                        FixedPoint, Address, Data)             \
    {                                                          \
        ljtUINT32 __temp_data32__;                             \
                                                               \
        ljmVERIFYLOADSTATE(CommandBuffer, Memory, Address);    \
                                                               \
        ljmSAFECASTSIZET(__temp_data32__, Data);               \
                                                               \
        *Memory++ = __temp_data32__;                           \
                                                               \
        ljoHARDWARE_UpdateDelta(StateDelta, Address, \
                                0, __temp_data32__);           \
                                                               \
        ljmDUMPSTATEDATA(StateDelta, FixedPoint, \
                         Address, __temp_data32__);            \
    }

#define ljmSETSTATEDATAWITHMASK(StateDelta, CommandBuffer, Memory, \
                                FixedPoint, Address, Mask, Data)    \
    {                                                               \
        ljtUINT32 __temp_data32__;                                  \
                                                                    \
        ljmVERIFYLOADSTATE(CommandBuffer, Memory, Address);         \
                                                                    \
        ljmSAFECASTSIZET(__temp_data32__, Data);                    \
                                                                    \
        *Memory++ = __temp_data32__;                                \
                                                                    \
        ljoHARDWARE_UpdateDelta(StateDelta, Address, \
                                Mask, __temp_data32__);             \
                                                                    \
        ljmDUMPSTATEDATA(StateDelta, FixedPoint, \
                         Address, __temp_data32__);                 \
    }

#define ljmSETCTRLSTATE(StateDelta, CommandBuffer, \
                        Memory, Address, Data)                      \
    {                                                               \
        ljtUINT32 __temp_data32__;                                  \
                                                                    \
        ljmVERIFYLOADSTATE(CommandBuffer, Memory, Address);         \
                                                                    \
        ljmSAFECASTSIZET(__temp_data32__, Data);                    \
                                                                    \
        *Memory++ = __temp_data32__;                                \
                                                                    \
        ljmDUMPSTATEDATA(StateDelta, ljvFALSE, \
                         Address, __temp_data32__);                 \
    }

#define ljmSETFILLER(CommandBuffer, Memory)                         \
    {                                                               \
        ljmVERIFYLOADSTATEDONE(CommandBuffer);                      \
                                                                    \
        *(ljtUINT32_PTR)Memory = 0x18000000;                        \
        Memory += 1;                                                \
    }

/*----------------------------------------------------------------------------*/

#define ljmSETSINGLESTATE(StateDelta, CommandBuffer, \
                          Memory, FixedPoint, Address, Data)        \
    {                                                               \
        ljmBEGINSTATEBATCH(CommandBuffer, Memory, \
                           FixedPoint, Address, 1);                 \
        ljmSETSTATEDATA(StateDelta, CommandBuffer, Memory, \
                        FixedPoint, Address, Data);                 \
        ljmENDSTATEBATCH(CommandBuffer, Memory);                    \
    }

#define ljmSETSINGLESTATEWITHMASK(StateDelta, CommandBuffer, Memory,\
                                  FixedPoint, Address, Mask, Data)  \
                                                                    \
    {                                                               \
        ljmBEGINSTATEBATCH(CommandBuffer, Memory, \
                           FixedPoint, Address, 1);                 \
        ljmSETSTATEDATAWITHMASK(StateDelta, CommandBuffer, Memory, \
                                FixedPoint, Address, Mask, Data);   \
                                                                    \
        ljmENDSTATEBATCH(CommandBuffer, Memory);                    \
    }

#define ljmSETSINGLECTRLSTATE(StateDelta, CommandBuffer, \
                              Memory, FixedPoint, Address, Data)    \
    {                                                               \
        ljmBEGINSTATEBATCH(CommandBuffer, Memory, FixedPoint, Address, 1); \
        ljmSETCTRLSTATE(StateDelta, CommandBuffer, Memory, Address, Data); \
        ljmENDSTATEBATCH(CommandBuffer, Memory);                    \
    }

#define ljmSETSEMASTALLPIPE(StateDelta, CommandBuffer, Memory, Data)\
    {                                                               \
        ljmSETSINGLECTRLSTATE(StateDelta, CommandBuffer, Memory, \
                              ljvFALSE, 0x0E02, Data); \
                                                                    \
        *Memory++ = ljmSETFIELDVALUE(0, STALL_COMMAND, OPCODE, STALL); \
                                                                    \
        *Memory++ = Data;                                           \
                                                                    \
        ljmDUMP(ljvNULL, "#[stall 0x%08X 0x%08X]", \
                ljmSETFIELDVALUE(0, AQ_SEMAPHORE, SOURCE, FRONT_END), \
                ljmSETFIELDVALUE(0, AQ_SEMAPHORE, DESTINATION, PIXEL_ENGINE)); \
    }

/*******************************************************************************
 **
 **  ljmSETSTARTDECOMMAND
 **
 **      Form a START_DE command.
 **
 **  ARGUMENTS:
 **
 **      Memory          Destination memory pointer of ljtUINT32_PTR type.
 **      Count           Number of the rectangles.
 */

#define ljmSETSTARTDECOMMAND(Memory, Count)                                              \
    {                                                                                    \
        *Memory++ = ljmSETFIELDVALUE(0, AQ_COMMAND_START_DE_COMMAND, OPCODE, START_DE) | \
                         ljmSETFIELD(0, AQ_COMMAND_START_DE_COMMAND, COUNT, Count) |    \
                         ljmSETFIELD(0, AQ_COMMAND_START_DE_COMMAND, DATA_COUNT, 0);     \
                                                                                         \
        *Memory++ = 0xDEADDEED;                                                          \
    }

/*****************************************
 ** Temp command buffer macro
 */
#define ljmDEFINESTATEBUFFER_NEW(CommandBuffer, StateDelta, Memory) \
    ljmDEFINELOADSTATEBASE()                                        \
    ljsTEMPCMDBUF CommandBuffer = ljvNULL;                          \
    ljtUINT32_PTR        Memory;                                    \
    ljsSTATE_DELTA_PTR   StateDelta;                                \
    ljeENGINE            CurrentEngine = ljvENGINE_RENDER

#define ljmBEGINSTATEBUFFER_NEW(Hardware, CommandBuffer, \
                                StateDelta, Memory, OutSide)        \
    {                                                               \
        if (OutSide) {                                              \
            Memory = (ljtUINT32_PTR)*OutSide;                       \
        } else {                                                    \
            ljmONERROR(ljoBUFFER_StartTEMPCMDBUF(Hardware->engine[CurrentEngine].buffer, \
                                                 Hardware->engine[CurrentEngine].queue, \
                                                 &CommandBuffer));                       \
                                                                                         \
            Memory = (ljtUINT32_PTR)(CommandBuffer->buffer); \
        }                                                    \
        StateDelta = Hardware->tempDelta;                    \
                                                             \
        ljmSETLOADSTATEBASE(CommandBuffer, OutSide);         \
    }

#define ljmENDSTATEBUFFER_NEW(Hardware, CommandBuffer, Memory, OutSide)                  \
    {                                                                                    \
        if (OutSide) {                                                                   \
            *OutSide = Memory;                                                           \
        } else {                                                                         \
            CommandBuffer->currentByteSize =                                             \
                (ljtUINT32)((ljtUINT8_PTR)Memory - (ljtUINT8_PTR)CommandBuffer->buffer); \
                                                                                         \
            ljmONERROR(ljoBUFFER_EndTEMPCMDBUF(\
                Hardware->engine[CurrentEngine].buffer, ljvFALSE));                      \
            if (Hardware->constructType != ljvHARDWARE_2D) {                             \
                ljoHARDWARE_UpdateTempDelta(Hardware);                                   \
            }                                                                            \
        }                                                                                \
        ljmUNSETLOADSTATEBASE()                                                          \
    }

#define ljmDEFINECTRLSTATEBUFFER(CommandBuffer, Memory)                         \
    ljmDEFINELOADSTATEBASE()                                                    \
    ljsTEMPCMDBUF CommandBuffer = ljvNULL;                                      \
    ljtUINT32_PTR Memory;                                                       \
    ljeENGINE CurrentEngine = ljvENGINE_RENDER

#define ljmBEGINCTRLSTATEBUFFER(Hardware, CommandBuffer, Memory, OutSide)       \
{                                                                               \
    if (OutSide)                                                                \
    {                                                                           \
        Memory = (ljtUINT32_PTR)*OutSide;                                       \
    }                                                                           \
    else                                                                        \
    {                                                                           \
        ljmONERROR(ljoBUFFER_StartTEMPCMDBUF(\
            Hardware->engine[CurrentEngine].buffer, \
            Hardware->engine[CurrentEngine].queue, &CommandBuffer               \
            ));                                                                 \
                                                                                \
        Memory = (ljtUINT32_PTR)(CommandBuffer->buffer);                        \
    }                                                                           \
    ljmSETLOADSTATEBASE(CommandBuffer,OutSide);                                 \
}

/*----------------------------------------------------------------------------*/

#define ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, Count)                \
    {                                                                     \
        ljmVERIFYLOADSTATEALIGNED(CommandBuffer, Memory);                 \
        ljmASSERT((ljtUINT32)Count <= 1024);                              \
                                                                          \
        *Memory++ = ljmSETFIELDVALUE(0, AQ_COMMAND_LOAD_STATE_COMMAND, OPCODE, LOAD_STATE) |  \
                         ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, FLOAT, FixedPoint) |  \
                         ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, COUNT, Count) |       \
                         ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, ADDRESS, Address);      \
    }

#define ljmENDSTATEBATCH_NEW(CommandBuffer, Memory)  \
    ljmVERIFYLOADSTATEALIGNED(CommandBuffer, Memory)

/*----------------------------------------------------------------------------*/

#define ljmSETSTATEDATA_NEW(StateDelta, CommandBuffer, Memory, \
                            FixedPoint, Address, Data)                      \
    {                                                                       \
        ljtUINT32 __temp_data32__;                                          \
                                                                            \
        ljmSAFECASTSIZET(__temp_data32__, Data);                            \
                                                                            \
        *Memory++ = __temp_data32__;                                        \
                                                                            \
        ljoHARDWARE_UpdateDelta(StateDelta, Address, 0, __temp_data32__);   \
                                                                            \
        ljmDUMPSTATEDATA(StateDelta, FixedPoint, Address, __temp_data32__); \
    }

#define ljmSETSTATEDATAWITHMASK_NEW(StateDelta, CommandBuffer, Memory, \
                                    FixedPoint, Address, Mask, Data)        \
    {                                                                       \
        ljtUINT32 __temp_data32__;                                          \
                                                                            \
        ljmSAFECASTSIZET(__temp_data32__, Data);                            \
                                                                            \
        *Memory++ = __temp_data32__;                                        \
                                                                            \
        ljoHARDWARE_UpdateDelta(StateDelta, Address, Mask, __temp_data32__);\
                                                                            \
        ljmDUMPSTATEDATA(StateDelta, FixedPoint, Address, __temp_data32__); \
    }

#define ljmSETCTRLSTATE_NEW(StateDelta, CommandBuffer, Memory, Address, Data) \
    {                                                                         \
        ljtUINT32 __temp_data32__;                                            \
                                                                              \
        ljmSAFECASTSIZET(__temp_data32__, Data);                              \
                                                                              \
        *Memory++ = __temp_data32__;                                          \
                                                                              \
        ljmDUMPSTATEDATA(StateDelta, ljvFALSE, Address, __temp_data32__);     \
    }

#define ljmSETFILLER_NEW(CommandBuffer, Memory)      \
    {                                                \
        *(ljtUINT32_PTR)Memory = 0x18000000;         \
        Memory += 1;                                 \
    }

/*----------------------------------------------------------------------------*/

#define ljmSETSINGLESTATE_NEW(StateDelta, CommandBuffer, Memory, \
                              FixedPoint, Address, Data)            \
    {                                                               \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, 1);             \
        ljmSETSTATEDATA_NEW(StateDelta, CommandBuffer, Memory, \
                            FixedPoint, Address, Data);             \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                \
    }

#define ljmSETSINGLESTATEWITHMASK_NEW(StateDelta, CommandBuffer, Memory, \
                                      FixedPoint, Address, Mask, Data)      \
    {                                                                       \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, 1);                     \
        ljmSETSTATEDATAWITHMASK_NEW(StateDelta, CommandBuffer, Memory, \
                                    FixedPoint, Address, Mask, Data);       \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                        \
    }

#define ljmSETSINGLECTRLSTATE_NEW(StateDelta, CommandBuffer, Memory, \
                                  FixedPoint, Address, Data)            \
    {                                                                   \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, 1);                 \
        ljmSETCTRLSTATE_NEW(StateDelta, CommandBuffer, \
                            Memory, Address, Data);                     \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                    \
    }

#define ljmSETBLOCKCTRLSTATE_NEW(StateDelta, CommandBuffer, Memory, \
                                 FixedPoint, Address, Data, Count)      \
    {                                                                   \
        ljtUINT32 c;                                                    \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, Count);             \
        for (c = 0; c < Count; c++) {                                   \
            ljmSETCTRLSTATE_NEW(StateDelta, CommandBuffer, \
                                Memory, Address, Data);                 \
        }                                                               \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                    \
    }

#define ljmSETCTRLSTATES_NEW(StateDelta, CommandBuffer, Memory, \
                             FixedPoint, Address, Data, Count)          \
    {                                                                   \
        ljtUINT32 c;                                                    \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, Count);             \
        for (c = 0; c < Count; c++) {                                   \
            ljmSETCTRLSTATE_NEW(StateDelta, CommandBuffer, \
                                Memory, Address, Data[c]);              \
        }                                                               \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                    \
    }

#define ljmSETSEMASTALLPIPE_NEW(StateDelta, CommandBuffer, Memory, Data)\
    {                                                                   \
        ljmSETSINGLECTRLSTATE_NEW(StateDelta, CommandBuffer, Memory, \
                                  ljvFALSE, 0x0E02, Data); \
                                                                        \
        *Memory++ = ljmSETFIELDVALUE(0, STALL_COMMAND, OPCODE, STALL);  \
                                                                        \
        *Memory++ = Data;                                               \
                                                                        \
        ljmDUMP(ljvNULL, "#[stall 0x%08X 0x%08X]", \
                ljmSETFIELDVALUE(0, AQ_SEMAPHORE, SOURCE, FRONT_END), \
                ljmSETFIELDVALUE(0, AQ_SEMAPHORE, DESTINATION, PIXEL_ENGINE)); \
    }

#define ljmSETSTATEDATA_NEW_FAST(StateDelta, CommandBuffer, Memory, \
                                 FixedPoint, Address, Data)                 \
    {                                                                       \
        ljtUINT32 __temp_data32__;                                          \
                                                                            \
        ljmSAFECASTSIZET(__temp_data32__, Data);                            \
                                                                            \
        *Memory++ = __temp_data32__;                                        \
                                                                            \
        ljmDUMPSTATEDATA(StateDelta, FixedPoint, Address, __temp_data32__); \
    }

#define ljmSETSTATEDATAWITHMASK_NEW_FAST(StateDelta, CommandBuffer, Memory, \
                                         FixedPoint, Address, Mask, Data)   \
    {                                                                       \
        ljtUINT32 __temp_data32__;                                          \
                                                                            \
        __temp_data32__ = Data;                                             \
                                                                            \
        *Memory++ = __temp_data32__;                                        \
                                                                            \
        ljmDUMPSTATEDATA(StateDelta, FixedPoint, Address, __temp_data32__); \
    }

#define ljmSETSINGLESTATE_NEW_FAST(StateDelta, CommandBuffer, Memory, \
                                   FixedPoint, Address, Data)               \
    {                                                                       \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, \
                               FixedPoint, Address, 1);                     \
        ljmSETSTATEDATA_NEW_FAST(StateDelta, CommandBuffer, Memory, \
                                 FixedPoint, Address, Data);                \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                        \
    }

#define ljmSETSINGLESTATEWITHMASK_NEW_FAST(StateDelta, CommandBuffer, Memory, \
                                           FixedPoint, Address, Mask, Data)    \
    {                                                                          \
        ljmBEGINSTATEBATCH_NEW(CommandBuffer, Memory, FixedPoint, Address, 1); \
        ljmSETSTATEDATAWITHMASK_NEW_FAST(StateDelta, CommandBuffer, Memory, \
                                         FixedPoint, Address, Mask, Data);     \
        ljmENDSTATEBATCH_NEW(CommandBuffer, Memory);                           \
    }

#define ljmDEFINESTATEBUFFER_NEW_FAST(CommandBuffer, Memory)                \
    ljmDEFINELOADSTATEBASE()                                                \
    ljsTEMPCMDBUF CommandBuffer = ljvNULL;                                  \
    ljtUINT32_PTR Memory;

#define ljmBEGINSTATEBUFFER_NEW_FAST(Hardware, CommandBuffer, Memory, OutSide)      \
{                                                                                   \
    if (OutSide) {                                                                  \
        Memory = (ljtUINT32_PTR)*OutSide;                                           \
    } else {                                                                        \
        ljmONERROR(ljoBUFFER_StartTEMPCMDBUF(Hardware->engine[ljvENGINE_RENDER].buffer, \
                                             Hardware->engine[ljvENGINE_RENDER].queue, \
                                             &CommandBuffer));                      \
                                                                                    \
        Memory = (ljtUINT32_PTR)(CommandBuffer->buffer);                            \
                                                                                    \
    }                                                                               \
                                                                                    \
    ljmSETLOADSTATEBASE(CommandBuffer,OutSide);                                     \
}

#define ljmENDSTATEBUFFER_NEW_FAST(Hardware, CommandBuffer, Memory, OutSide)             \
    {                                                                                    \
        if (OutSide) {                                                                   \
            *OutSide = Memory;                                                           \
        } else {                                                                         \
            CommandBuffer->currentByteSize =                                             \
                (ljtUINT32)((ljtUINT8_PTR)Memory - (ljtUINT8_PTR)CommandBuffer->buffer); \
                                                                                         \
            ljmONERROR(ljoBUFFER_EndTEMPCMDBUF(\
                Hardware->engine[ljvENGINE_RENDER].buffer, ljvFALSE));                   \
        }                                                                                \
        ljmUNSETLOADSTATEBASE()                                                          \
    }

/*******************************************************************************
 **
 **  ljmCONFIGUREUNIFORMS
 **
 **      Configure uniforms according to chip and numConstants.
 */
#if !ljdENABLE_UNIFIED_CONSTANT
# define ljmCONFIGUREUNIFORMS(ChipModel, ChipRevision, NumConstants, \
                              UnifiedConst, VsConstBase, PsConstBase, \
                              VsConstMax, PsConstMax, ConstMax)              \
        {                                                                    \
            if (ChipModel == ljv2000 &&                                      \
                (ChipRevision == 0x5118 || ChipRevision == 0x5140)) {        \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 256;                                            \
                PsConstMax = 64;                                             \
                ConstMax = 320;                                              \
            } else if (NumConstants == 320) {                                \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 256;                                            \
                PsConstMax = 64;                                             \
                ConstMax = 320;                                              \
            } /* All LJ1000 series chips can only support 64 uniforms for ps on non-unified const mode. */ \
            else if (NumConstants > 256 && ChipModel == ljv1000) {           \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 256;                                            \
                PsConstMax = 64;                                             \
                ConstMax = 320;                                              \
            } else if (NumConstants > 256) {                                 \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 256;                                            \
                PsConstMax = 256;                                            \
                ConstMax = 512;                                              \
            } else if (NumConstants == 256) {                                \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 256;                                            \
                PsConstMax = 256;                                            \
                ConstMax = 512;                                              \
            } else {                                                         \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 168;                                            \
                PsConstMax = 64;                                             \
                ConstMax = 232;                                              \
            }                                                                \
        }
#else
# define ljmCONFIGUREUNIFORMS(ChipModel, ChipRevision, Halti5Avail, \
                              SmallBatch, ComputeOnly, NumConstants, \
                              UnifiedConst, VsConstBase, PsConstBase, \
                              VsConstMax, PsConstMax, ConstMax)              \
        {                                                                    \
            if (NumConstants > 256) {                                        \
                UnifiedConst = ljvTRUE;                                      \
                if (SmallBatch) {                                            \
                    VsConstBase = 0xD000;                \
                    PsConstBase = 0xD000;                \
                } else if (Halti5Avail) {                                    \
                    VsConstBase = 0xD000;                \
                    PsConstBase = 0xD800;                \
                } else {                                                     \
                    VsConstBase = 0xC000;                   \
                    PsConstBase = 0xC000;                   \
                }                                                            \
                if ((ChipModel == ljv880) &&                                 \
                    ((ChipRevision & 0xfff0) == 0x5120)) {                   \
                    VsConstMax = 512;                                        \
                    PsConstMax = 64;                                         \
                    ConstMax = 576;                                          \
                } else {                                                     \
                    VsConstMax = ljmMIN(512, NumConstants - 64);             \
                    PsConstMax = ljmMIN(512, NumConstants - 64);             \
                    ConstMax = NumConstants;                                 \
                }                                                            \
            } else if (NumConstants == 256) {                                \
                if (ChipModel == ljv2000 &&                                  \
                    (ChipRevision == 0x5118 || ChipRevision == 0x5140)) {    \
                    UnifiedConst = ljvFALSE;                                 \
                    VsConstBase = 0x1400;               \
                    PsConstBase = 0x1C00;                \
                    VsConstMax = 256;                                        \
                    PsConstMax = 64;                                         \
                    ConstMax = 320;                                          \
                } else {                                                     \
                    UnifiedConst = ljvFALSE;                                 \
                    VsConstBase = 0x1400;               \
                    PsConstBase = 0x1C00;                \
                    VsConstMax = 256;                                        \
                    PsConstMax = 256;                                        \
                    ConstMax = 512;                                          \
                }                                                            \
            } else if (NumConstants == 160 && ComputeOnly) {                 \
                UnifiedConst = ljvTRUE;                                      \
                VsConstBase = 0xD000;                    \
                PsConstBase = 0xD800;                    \
                VsConstMax = 0;                                              \
                PsConstMax = 160;                                            \
                ConstMax = 160;                                              \
            } else {                                                         \
                UnifiedConst = ljvFALSE;                                     \
                VsConstBase = 0x1400;                   \
                PsConstBase = 0x1C00;                    \
                VsConstMax = 168;                                            \
                PsConstMax = 64;                                             \
                ConstMax = 232;                                              \
            }                                                                \
        }
#endif

/*******************************************************************************
 **
 **  ljmCONFIGUREUNIFORMS2
 **  only fix clang build error
 **
 **      Configure uniforms according to chip and numConstants.
 */
#if !ljdENABLE_UNIFIED_CONSTANT
#define ljmCONFIGUREUNIFORMS2(ChipModel, ChipRevision, NumConstants, \
             UnifiedConst, VsConstMax, PsConstMax) \
{ \
    if (ChipModel == ljv2000 && (ChipRevision == 0x5118 || ChipRevision == 0x5140)) \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 256; \
        PsConstMax   = 64; \
    } \
    else if (NumConstants == 320) \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 256; \
        PsConstMax   = 64; \
    } \
    /* All LJ1000 series chips can only support 64 uniforms for ps on non-unified const mode. */ \
    else if (NumConstants > 256 && ChipModel == ljv1000) \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 256; \
        PsConstMax   = 64; \
    } \
    else if (NumConstants > 256) \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 256; \
        PsConstMax   = 256; \
    } \
    else if (NumConstants == 256) \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 256; \
        PsConstMax   = 256; \
    } \
    else \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 168; \
        PsConstMax   = 64; \
    } \
}
#else
#define ljmCONFIGUREUNIFORMS2(ChipModel, ChipRevision, Halti5Avail, SmallBatch, ComputeOnly, NumConstants, \
             UnifiedConst, VsConstMax, PsConstMax) \
{ \
    if (NumConstants > 256) \
    { \
        UnifiedConst = ljvTRUE; \
        if ((ChipModel == ljv880) && ((ChipRevision & 0xfff0) == 0x5120)) \
        { \
            VsConstMax   = 512; \
            PsConstMax   = 64; \
        } \
        else \
        { \
            VsConstMax   = ljmMIN(512, NumConstants - 64); \
            PsConstMax   = ljmMIN(512, NumConstants - 64); \
        } \
    } \
    else if (NumConstants == 256) \
    { \
        if (ChipModel == ljv2000 && (ChipRevision == 0x5118 || ChipRevision == 0x5140)) \
        { \
            UnifiedConst = ljvFALSE; \
            VsConstMax   = 256; \
            PsConstMax   = 64; \
        } \
        else \
        { \
            UnifiedConst = ljvFALSE; \
            VsConstMax   = 256; \
            PsConstMax   = 256; \
        } \
    } \
    else if (NumConstants == 160 && ComputeOnly) \
    { \
        UnifiedConst = ljvTRUE; \
        VsConstMax   = 0; \
        PsConstMax   = 160; \
    } \
    else \
    { \
        UnifiedConst = ljvFALSE; \
        VsConstMax   = 168; \
        PsConstMax   = 64; \
    } \
}
#endif

#define ljmAnyTileStatusEnableForFullMultiSlice(SurfView, anyTsEnableForMultiSlice) \
    {                                                                               \
        ljtUINT i = 0;                                                              \
        for (; i < (SurfView->surf->requestD); i++) {                               \
            if ((SurfView->surf->tileStatusNode.pool != ljvPOOL_UNKNOWN) &&         \
                (SurfView->surf->tileStatusDisabled[i] == ljvFALSE)) {              \
                *anyTsEnableForMultiSlice = ljvTRUE;                                \
                break;                                                              \
            }                                                                       \
        }                                                                           \
    }

#define ljmAnyTileStatusEnableForMultiSlice(SurfView, anyTsEnableForMultiSlice) \
    {                                                                           \
        ljtUINT i = SurfView->firstSlice;                                       \
        for (; i < (SurfView->firstSlice + SurfView->numSlices); i++) {         \
            if ((SurfView->surf->tileStatusNode.pool != ljvPOOL_UNKNOWN) &&     \
                (SurfView->surf->tileStatusDisabled[i] == ljvFALSE)) {          \
                *anyTsEnableForMultiSlice = ljvTRUE;                            \
                break;                                                          \
            }                                                                   \
        }                                                                       \
    }

#define ljmCanTileStatusEnabledForMultiSlice(SurfView, canTsEnabled)                               \
    {                                                                                              \
        if (SurfView->numSlices > 1) {                                                             \
            if (SurfView->surf->tileStatusNode.pool != ljvPOOL_UNKNOWN) {                          \
                ljtUINT i = 0;                                                                     \
                for (; i < SurfView->numSlices; i++) {                                             \
                    if (SurfView->surf->tileStatusDisabled[i] == ljvTRUE) {                        \
                        *canTsEnabled = ljvFALSE;                                                  \
                        break;                                                                     \
                    }                                                                              \
                    if (SurfView->surf->fcValue[i] != SurfView->surf->fcValue[0]) {                \
                        *canTsEnabled = ljvFALSE;                                                  \
                        break;                                                                     \
                    }                                                                              \
                                                                                                   \
                    if (SurfView->surf->fcValueUpper[i] != SurfView->surf->fcValueUpper[0]) {      \
                        *canTsEnabled = ljvFALSE;                                                  \
                        break;                                                                     \
                    }                                                                              \
                }                                                                                  \
            } else {                                                                               \
                *canTsEnabled = ljvFALSE;                                                          \
            }                                                                                      \
        } else {                                                                                   \
            if ((SurfView->surf->tileStatusNode.pool == ljvPOOL_UNKNOWN) ||                        \
                (SurfView->surf->tileStatusDisabled[SurfView->firstSlice] == ljvTRUE)) {           \
                *canTsEnabled = ljvFALSE;                                                          \
            }                                                                                      \
        }                                                                                          \
    }

#define ljmCONFIGUSC(prefix, featureUSC, featureSeparateLS, \
                     featureComputeOnly, featureTS, featureGS, \
                     featureUSCFullCacheFix, featureL1CacheSize, \
                     featureUSCMaxPages, attribCacheRatio, L1CacheRatio)        \
    {                                                                           \
        attribCacheRatio = 0x2;        \
                                                                                \
        if (featureUSC) {                                                       \
            if (featureSeparateLS) {                                            \
                L1CacheRatio = 0x0;              \
            } else {                                                            \
                ljtUINT L1cacheSize;                                            \
                                                                                \
                if (featureComputeOnly) {                                       \
                    L1cacheSize = featureL1CacheSize;                           \
                } else {                                                        \
                    ljtUINT attribBufSizeInKB;                                  \
                    if (featureTS) {                                            \
                        /* GS/TS must be bundled. */                            \
                        prefix##ASSERT(featureGS);                              \
                        featureGS         = featureGS;                          \
                        attribBufSizeInKB = 42;                                 \
                    } else {                                                    \
                        prefix##ASSERT(!featureGS);                             \
                        attribBufSizeInKB = 8;                                  \
                    }                                                           \
                    if (attribBufSizeInKB < featureUSCMaxPages) {               \
                        L1cacheSize = featureUSCMaxPages - attribBufSizeInKB;   \
                    } else {                                                    \
                        attribBufSizeInKB -= 2;                                 \
                        L1cacheSize = 2;                                        \
                    }                                                           \
                }                                                               \
                prefix##ASSERT(L1cacheSize);                                    \
                if (L1cacheSize >= featureL1CacheSize) {                        \
                    L1CacheRatio = 0x0;          \
                    prefix##ASSERT(featureUSCFullCacheFix);                     \
                    featureUSCFullCacheFix = featureUSCFullCacheFix;            \
                } else {                                                        \
                    static const ljtINT s_uscCacheRatio[] = {                   \
                        100000, /* 1.0f */                                      \
                        50000, /* 0.5f */                                      \
                        25000, /* 0.25f */                                     \
                        12500, /* 0.125f */                                    \
                        62500, /* 0.0625f */                                   \
                        3125, /* 0.03125f */                                  \
                        75000, /* 0.75f */                                     \
                        0, /* 0.0f */                                      \
                    };                                                          \
                    ljtINT maxL1cacheSize = L1cacheSize * 100000;               \
                    ljtINT delta = 2147483647; /* start with very big delta */  \
                    ljtINT i = 0;                                               \
                    ljtINT curIndex = -1;                                       \
                    for (; i < ljmCOUNTOF(s_uscCacheRatio); ++i) {              \
                        ljtINT curL1cacheSize = featureL1CacheSize * s_uscCacheRatio[i]; \
                                                                                \
                        if ((maxL1cacheSize >= curL1cacheSize) &&               \
                            ((maxL1cacheSize - curL1cacheSize) < delta)) {      \
                            curIndex = i;                                       \
                            delta = maxL1cacheSize - curL1cacheSize;            \
                        }                                                       \
                    }                                                           \
                    prefix##ASSERT(-1 != curIndex);                             \
                    L1CacheRatio = curIndex;                                    \
                }                                                               \
            }                                                                   \
        }                                                                       \
    }

#define ljmCONFIGUSC2(prefix, Hardware, featureUSC, featureSeparateLS, featureComputeOnly, \
    featureTS, featureL1CacheSize, featureUSCMaxPages, \
    attribCacheRatio, L1CacheRatio) \
{ \
    attribCacheRatio = 0x2; \
    \
    if (featureUSC) \
    { \
        if (featureSeparateLS) \
        { \
            L1CacheRatio = 0x0; \
        } \
        else \
        { \
            ljtUINT L1cacheSize; \
            \
            if (featureComputeOnly) \
            { \
                L1cacheSize = featureL1CacheSize; \
                attribCacheRatio = 0x7; \
            } \
            else \
            { \
                ljtUINT attribBufSizeInKB; \
                if (featureTS) \
                { \
                    /* GS/TS must be bundled. */ \
                    attribBufSizeInKB = 42; \
                    attribCacheRatio = (Hardware->identity.chipModel == ljv8800)? \
                                        0x3 \
                                       : 0x4; \
                } \
                else \
                { \
                    attribBufSizeInKB = 8; \
                } \
                if (attribBufSizeInKB < featureUSCMaxPages) \
                { \
                    L1cacheSize = featureUSCMaxPages - attribBufSizeInKB; \
                } \
                else \
                { \
                    attribBufSizeInKB -= 2; \
                    L1cacheSize = 2; \
                } \
            } \
            prefix##ASSERT(L1cacheSize); \
            if (L1cacheSize >= featureL1CacheSize) \
            { \
                L1CacheRatio = 0x0; \
            } \
            else \
            { \
                static const ljtINT s_uscCacheRatio[] = \
                { \
                    100000,/* 1.0f */     \
                    50000, /* 0.5f */     \
                    25000, /* 0.25f */    \
                    12500, /* 0.125f */   \
                    62500, /* 0.0625f */  \
                    3125, /* 0.03125f */ \
                    75000, /* 0.75f */    \
                    0, /*0.0f */      \
                }; \
                ljtINT maxL1cacheSize = L1cacheSize * 100000; \
                ljtINT delta = 2147483647; /* start with very big delta */ \
                ljtINT i = 0; \
                ljtINT curIndex = -1; \
                for (; i < ljmCOUNTOF(s_uscCacheRatio); ++i) \
                { \
                    ljtINT curL1cacheSize = featureL1CacheSize * s_uscCacheRatio[i]; \
                  \
                    if ((maxL1cacheSize >= curL1cacheSize) && \
                        ((maxL1cacheSize - curL1cacheSize) < delta)) \
                    { \
                        curIndex = i; \
                        delta = maxL1cacheSize - curL1cacheSize; \
                    } \
                } \
                prefix##ASSERT(-1 != curIndex); \
                L1CacheRatio = curIndex; \
            } \
        } \
    } \
} \

#if LJMICRO_PROFILER_SYSTEM_MEMORY
typedef struct _memory_profile_info {
    struct {
        ljtUINT64 currentSize;
        ljtUINT64 peakSize;
        ljtUINT64 total_allocate;
        ljtUINT64 total_free;
        ljtUINT32 total_allocateCount;
        ljtUINT32 total_freeCount;
    } system_memory, gpu_memory;
} memory_profile_info;

ljeSTATUS
ljoOS_GetMemoryProfileInfo(size_t size, struct _memory_profile_info *info);

ljeSTATUS
ljoOS_DumpMemoryProfile(void);
ljeSTATUS
ljoOS_InitMemoryProfile(void);
ljeSTATUS
ljoOS_DeInitMemoryProfile(void);
#endif

void
ljd_2D_printf(
    const char *format, ...
    );

#if ljd_2D_PRINT_TIME_STAMP
void
ljd_2D_print_time_stamp(
    IN const char* function_name,
    IN char* order
    );

#define ljmPRINT_2D_TIME_STAMP(function_name, order) \
    ljd_2D_print_time_stamp(function_name, order)

#else
#define ljmPRINT_2D_TIME_STAMP(function_name, order)

#endif

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_base_h_ */


