/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_debug_zones_h_
#define __lj_hal_debug_zones_h_

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************\
 ************************ Debug Zone Pattern Summary ***************************
 * A debug zone is an unsigned integer of 32 bit (Bit 31- Bit 0).              *
 * Bit 31 to 28 defines API, which is 0 for HAL API and has value of 1 - 14    *
 * for Khronos API. Value 15 (0xF) is reserved for ljdZONE_NONE.               *
 * Bit 27 to 0 defines subzones of each API. Value 0xFFFFFFF is resevered for  *
 * ljdZONE_ALL.                                                                *
 *                                                                             *
\******************************************************************************/

/* Retrieve API bits 31 to 28 */
#define ljmZONE_GET_API(zone) ((zone) >> 28)

/* Retrieve Subzone bits 27 to 0 */
#define ljmZONE_GET_SUBZONES(zone) ((zone) << 4)

/******************************************************************************
 ******************************** HAL Zone ************************************
 ******************************************************************************/

#define ljdZONE_API_HAL ((ljtUINT32)0 << 28)

/******************************************************************************
 ******************************** HAL Subzones ********************************
 ******************************************************************************/

/* Subzones Kernel and User have in common */
#define ljvZONE_OS       (1 << 0)
#define ljvZONE_HARDWARE (1 << 1)
#define ljvZONE_HEAP     (1 << 2)
#define ljvZONE_SIGNAL   (1 << 3)

/* Subzones of HAL Kernel */
#define ljvZONE_KERNEL        (1 << 4)
#define ljvZONE_VIDMEM        (1 << 5)
#define ljvZONE_COMMAND       (1 << 6)
#define ljvZONE_DRIVER        (1 << 7)
#define ljvZONE_CMODEL        (1 << 8)
#define ljvZONE_MMU           (1 << 9)
#define ljvZONE_EVENT         (1 << 10)
#define ljvZONE_DEVICE        (1 << 11)
#define ljvZONE_DATABASE      (1 << 12)
#define ljvZONE_INTERRUPT     (1 << 13)
#define ljvZONE_POWER         (1 << 14)
#define ljvZONE_ASYNC_COMMAND (1 << 15)
#define ljvZONE_ALLOCATOR     (1 << 16)
#define ljvZONE_KMS           (1 << 17)   /* add by hym */

/* Subzones of HAL User */
#define ljdZONE_HAL_API     (1 << 4)
#define ljdZONE_BUFFER      (1 << 5)
#define ljdZONE_VGBUFFER    (1 << 6)
#define ljdZONE_SURFACE     (1 << 7)
#define ljdZONE_INDEX       (1 << 8)
#define ljdZONE_STREAM      (1 << 9)
#define ljdZONE_TEXTURE     (1 << 10)
#define ljdZONE_2D          (1 << 11)
#define ljdZONE_3D          (1 << 12)
#define ljdZONE_COMPILER    (1 << 13)
#define ljdZONE_MEM         (1 << 14)
#define ljdZONE_VERTEXARRAY (1 << 15)
#define ljdZONE_CL          (1 << 16)
#define ljdZONE_VG          (1 << 17)
#define ljdZONE_VX          (1 << 18)
#define ljdZONE_UTILITY     (1 << 19)
#define ljdZONE_RECT        (1 << 20)
#define ljdZONE_BUFOBJ      (1 << 21)
#define ljdZONE_PROFILER    (1 << 22)
#define ljdZONE_SHADER      (1 << 23)

/******************************************************************************
 ******************************** Khronos API Zones ***************************
 ******************************************************************************/

#define ljdZONE_API_EGL  ((ljtUINT32)1 << 28)
#define ljdZONE_API_ES11 ((ljtUINT32)2 << 28)
#define ljdZONE_API_ES30 ((ljtUINT32)3 << 28)
#define ljdZONE_API_GL40 ((ljtUINT32)4 << 28)
#define ljdZONE_API_VG3D ((ljtUINT32)5 << 28)
#define ljdZONE_API_CL   ((ljtUINT32)6 << 28)
#define ljdZONE_API_VX   ((ljtUINT32)7 << 28)
#define ljdZONE_API_VG   ((ljtUINT32)8 << 28)

/******************************************************************************
 ************************* Subzones of Khronos API Zones **********************
 ******************************************************************************/

/* Subzones of EGL API */
#define ljdZONE_EGL_API           (ljdZONE_API_EGL | (1 << 0))
#define ljdZONE_EGL_SURFACE       (ljdZONE_API_EGL | (1 << 1))
#define ljdZONE_EGL_CONTEXT       (ljdZONE_API_EGL | (1 << 2))
#define ljdZONE_EGL_CONFIG        (ljdZONE_API_EGL | (1 << 3))
#define ljdZONE_EGL_OS            (ljdZONE_API_EGL | (1 << 4)) /* unused */
#define ljdZONE_EGL_IMAGE         (ljdZONE_API_EGL | (1 << 5))
#define ljdZONE_EGL_SWAP          (ljdZONE_API_EGL | (1 << 6))
#define ljdZONE_EGL_INIT          (ljdZONE_API_EGL | (1 << 7))
#define ljdZONE_EGL_SYNC          (ljdZONE_API_EGL | (1 << 8))
#define ljdZONE_EGL_COMPOSE       (ljdZONE_API_EGL | (1 << 9))  /* unused */
#define ljdZONE_EGL_RENDER_THREAD (ljdZONE_API_EGL | (1 << 10)) /* unused */

/* Subzones of ES11 API */
#define ljdZONE_ES11_BUFFER    (ljdZONE_API_ES11 | (1 << 0))
#define ljdZONE_ES11_CLEAR     (ljdZONE_API_ES11 | (1 << 1))
#define ljdZONE_ES11_CLIP      (ljdZONE_API_ES11 | (1 << 2))
#define ljdZONE_ES11_CONTEXT   (ljdZONE_API_ES11 | (1 << 3))
#define ljdZONE_ES11_DRAW      (ljdZONE_API_ES11 | (1 << 4))
#define ljdZONE_ES11_ENABLE    (ljdZONE_API_ES11 | (1 << 5))
#define ljdZONE_ES11_EXTENTION (ljdZONE_API_ES11 | (1 << 6))
#define ljdZONE_ES11_FOG       (ljdZONE_API_ES11 | (1 << 7))
#define ljdZONE_ES11_FRAGMENT  (ljdZONE_API_ES11 | (1 << 8))
#define ljdZONE_ES11_LIGHT     (ljdZONE_API_ES11 | (1 << 9))
#define ljdZONE_ES11_MATRIX    (ljdZONE_API_ES11 | (1 << 10))
#define ljdZONE_ES11_PIXEL     (ljdZONE_API_ES11 | (1 << 11))
#define ljdZONE_ES11_POLIGON   (ljdZONE_API_ES11 | (1 << 12))
#define ljdZONE_ES11_LINE      (ljdZONE_API_ES11 | (1 << 13)) /* unused */
#define ljdZONE_ES11_QUERY     (ljdZONE_API_ES11 | (1 << 14))
#define ljdZONE_ES11_TEXTURE   (ljdZONE_API_ES11 | (1 << 15))
#define ljdZONE_ES11_STATES    (ljdZONE_API_ES11 | (1 << 16))
#define ljdZONE_ES11_STREAM    (ljdZONE_API_ES11 | (1 << 17))
#define ljdZONE_ES11_VIEWPORT  (ljdZONE_API_ES11 | (1 << 18))
#define ljdZONE_ES11_SHADER    (ljdZONE_API_ES11 | (1 << 19))
#define ljdZONE_ES11_HASH      (ljdZONE_API_ES11 | (1 << 20))
#define ljdZONE_ES11_TRACE     (ljdZONE_API_ES11 | (1 << 21))

/* Subzones of ES30 API */
#define ljdZONE_ES30_TRACE    (ljdZONE_API_ES30 | (1 << 0))
#define ljdZONE_ES30_BUFFER   (ljdZONE_API_ES30 | (1 << 1))
#define ljdZONE_ES30_CLEAR    (ljdZONE_API_ES30 | (1 << 2))
#define ljdZONE_ES30_CODEC    (ljdZONE_API_ES30 | (1 << 3))
#define ljdZONE_ES30_CONTEXT  (ljdZONE_API_ES30 | (1 << 4))
#define ljdZONE_ES30_DEPTH    (ljdZONE_API_ES30 | (1 << 5))
#define ljdZONE_ES30_DEVICE   (ljdZONE_API_ES30 | (1 << 6))
#define ljdZONE_ES30_DRAW     (ljdZONE_API_ES30 | (1 << 7))
#define ljdZONE_ES30_FBO      (ljdZONE_API_ES30 | (1 << 8))
#define ljdZONE_ES30_PIXEL    (ljdZONE_API_ES30 | (1 << 9))
#define ljdZONE_ES30_SHADER   (ljdZONE_API_ES30 | (1 << 10))
#define ljdZONE_ES30_STATE    (ljdZONE_API_ES30 | (1 << 11))
#define ljdZONE_ES30_TEXTURE  (ljdZONE_API_ES30 | (1 << 12))
#define ljdZONE_ES30_UTILS    (ljdZONE_API_ES30 | (1 << 13))
#define ljdZONE_ES30_PROFILER (ljdZONE_API_ES30 | (1 << 14))
#define ljdZONE_ES30_CORE     (ljdZONE_API_ES30 | (1 << 15))

/* Subzones of GL40 API */
#define ljdZONE_GL40_TRACE     (ljdZONE_API_GL40 | (1 << 0))
#define ljdZONE_GL40_BUFFER    (ljdZONE_API_GL40 | (1 << 1))
#define ljdZONE_GL40_CLEAR     (ljdZONE_API_GL40 | (1 << 2)) /* unused */
#define ljdZONE_GL40_CODEC     (ljdZONE_API_GL40 | (1 << 3))
#define ljdZONE_GL40_CONTEXT   (ljdZONE_API_GL40 | (1 << 4))
#define ljdZONE_GL40_DEPTH     (ljdZONE_API_GL40 | (1 << 5))
#define ljdZONE_GL40_DEVICE    (ljdZONE_API_GL40 | (1 << 6))
#define ljdZONE_GL40_DRAW      (ljdZONE_API_GL40 | (1 << 7))
#define ljdZONE_GL40_FBO       (ljdZONE_API_GL40 | (1 << 8))
#define ljdZONE_GL40_PIXEL     (ljdZONE_API_GL40 | (1 << 9))
#define ljdZONE_GL40_SHADER    (ljdZONE_API_GL40 | (1 << 10))
#define ljdZONE_GL40_STATE     (ljdZONE_API_GL40 | (1 << 11))
#define ljdZONE_GL40_TEXTURE   (ljdZONE_API_GL40 | (1 << 12))
#define ljdZONE_GL40_UTILS     (ljdZONE_API_GL40 | (1 << 13))
#define ljdZONE_GL40_PROFILER  (ljdZONE_API_GL40 | (1 << 14))
#define ljdZONE_GL40_CORE      (ljdZONE_API_GL40 | (1 << 15))
#define ljdZONE_GL40_FIXVERTEX (ljdZONE_API_GL40 | (1 << 16))
#define ljdZONE_GL40_FIXFRAG   (ljdZONE_API_GL40 | (1 << 17))
#define ljdZONE_GL40_HASH      (ljdZONE_API_GL40 | (1 << 18))

/* Subzones of VG3D API  */
#define ljdZONE_VG3D_CONTEXT     (ljdZONE_API_VG3D | (1 << 0))
#define ljdZONE_VG3D_DUMP        (ljdZONE_API_VG3D | (1 << 1))
#define ljdZONE_VG3D_EGL         (ljdZONE_API_VG3D | (1 << 2))
#define ljdZONE_VG3D_FONT        (ljdZONE_API_VG3D | (1 << 3))
#define ljdZONE_VG3D_HARDWARE    (ljdZONE_API_VG3D | (1 << 4))
#define ljdZONE_VG3D_IMAGE       (ljdZONE_API_VG3D | (1 << 5))
#define ljdZONE_VG3D_MASK        (ljdZONE_API_VG3D | (1 << 6))
#define ljdZONE_VG3D_MATRIX      (ljdZONE_API_VG3D | (1 << 7))
#define ljdZONE_VG3D_OBJECT      (ljdZONE_API_VG3D | (1 << 8))
#define ljdZONE_VG3D_PAINT       (ljdZONE_API_VG3D | (1 << 9))
#define ljdZONE_VG3D_PATH        (ljdZONE_API_VG3D | (1 << 10))
#define ljdZONE_VG3D_PROFILER    (ljdZONE_API_VG3D | (1 << 11))
#define ljdZONE_VG3D_SCANLINE    (ljdZONE_API_VG3D | (1 << 12))
#define ljdZONE_VG3D_SHADER      (ljdZONE_API_VG3D | (1 << 13))
#define ljdZONE_VG3D_TESSELLATOR (ljdZONE_API_VG3D | (1 << 14))
#define ljdZONE_VG3D_VGU         (ljdZONE_API_VG3D | (1 << 15))

/* Subzones of VG11 API  */
#define ljdZONE_VG_ARC       (ljdZONE_API_VG | (1 << 0))
#define ljdZONE_VG_CONTEXT   (ljdZONE_API_VG | (1 << 1))
#define ljdZONE_VG_DEBUG     (ljdZONE_API_VG | (1 << 2))
#define ljdZONE_VG_FILTER    (ljdZONE_API_VG | (1 << 3))
#define ljdZONE_VG_FORMAT    (ljdZONE_API_VG | (1 << 4))
#define ljdZONE_VG_IMAGE     (ljdZONE_API_VG | (1 << 5))
#define ljdZONE_VG_MAIN      (ljdZONE_API_VG | (1 << 6))
#define ljdZONE_VG_MASK      (ljdZONE_API_VG | (1 << 7))
#define ljdZONE_VG_MATRIX    (ljdZONE_API_VG | (1 << 8))
#define ljdZONE_VG_MEMORYMGR (ljdZONE_API_VG | (1 << 9))
#define ljdZONE_VG_OBJECT    (ljdZONE_API_VG | (1 << 10))
#define ljdZONE_VG_PAINT     (ljdZONE_API_VG | (1 << 11))
#define ljdZONE_VG_PATH      (ljdZONE_API_VG | (1 << 12))
#define ljdZONE_VG_STATE     (ljdZONE_API_VG | (1 << 13))
#define ljdZONE_VG_STROKE    (ljdZONE_API_VG | (1 << 14))
#define ljdZONE_VG_TEXT      (ljdZONE_API_VG | (1 << 15))
#define ljdZONE_VG_VGU       (ljdZONE_API_VG | (1 << 16))

/* Subzones of CL API  */
#define ljdZONE_CL_COMMAND        (ljdZONE_API_CL | (1 << 0))
#define ljdZONE_CL_CONTEXT        (ljdZONE_API_CL | (1 << 1))
#define ljdZONE_CL_DEVICE         (ljdZONE_API_CL | (1 << 2))
#define ljdZONE_CL_ENQUEUE        (ljdZONE_API_CL | (1 << 3))
#define ljdZONE_CL_EVENT          (ljdZONE_API_CL | (1 << 4))
#define ljdZONE_CL_EXT            (ljdZONE_API_CL | (1 << 5))
#define ljdZONE_CL_GL             (ljdZONE_API_CL | (1 << 6))
#define ljdZONE_CL_KERNEL         (ljdZONE_API_CL | (1 << 7))
#define ljdZONE_CL_MEM            (ljdZONE_API_CL | (1 << 8))
#define ljdZONE_CL_PLATFORM       (ljdZONE_API_CL | (1 << 9))
#define ljdZONE_CL_PROFILER       (ljdZONE_API_CL | (1 << 10))
#define ljdZONE_CL_PROGRAM        (ljdZONE_API_CL | (1 << 11))
#define ljdZONE_CL_SAMPLER        (ljdZONE_API_CL | (1 << 12))
#define ljdZONE_CL_COMMAND_BUFFER (ljdZONE_API_CL | (1 << 13))

/* Subzones of VX API  */
#define ljdZONE_VX_ARRAY     (ljdZONE_API_VX | (1 << 0))
#define ljdZONE_VX_BINARY    (ljdZONE_API_VX | (1 << 1))
#define ljdZONE_VX_CONTEXT   (ljdZONE_API_VX | (1 << 2))
#define ljdZONE_VX_CONV      (ljdZONE_API_VX | (1 << 3))
#define ljdZONE_VX_DELAY     (ljdZONE_API_VX | (1 << 4))
#define ljdZONE_VX_DIST      (ljdZONE_API_VX | (1 << 5))
#define ljdZONE_VX_GPULAYER  (ljdZONE_API_VX | (1 << 6))
#define ljdZONE_VX_GRAPH     (ljdZONE_API_VX | (1 << 7))
#define ljdZONE_VX_IMAGE     (ljdZONE_API_VX | (1 << 8))
#define ljdZONE_VX_INTERFACE (ljdZONE_API_VX | (1 << 9))
#define ljdZONE_VX_KERNEL    (ljdZONE_API_VX | (1 << 10))
#define ljdZONE_VX_LAYER     (ljdZONE_API_VX | (1 << 11))
#define ljdZONE_VX_LUT       (ljdZONE_API_VX | (1 << 12))
#define ljdZONE_VX_MATRIX    (ljdZONE_API_VX | (1 << 13))
#define ljdZONE_VX_MEMORY    (ljdZONE_API_VX | (1 << 14))
#define ljdZONE_VX_METAFMT   (ljdZONE_API_VX | (1 << 15))
#define ljdZONE_VX_NODE      (ljdZONE_API_VX | (1 << 16))
#define ljdZONE_VX_OBJARRAY  (ljdZONE_API_VX | (1 << 17))
#define ljdZONE_VX_PARAM     (ljdZONE_API_VX | (1 << 18))
#define ljdZONE_VX_PROGRAM   (ljdZONE_API_VX | (1 << 19))
#define ljdZONE_VX_PYRAMID   (ljdZONE_API_VX | (1 << 20))
#define ljdZONE_VX_REF       (ljdZONE_API_VX | (1 << 21))
#define ljdZONE_VX_REMAP     (ljdZONE_API_VX | (1 << 22))
#define ljdZONE_VX_SCALAR    (ljdZONE_API_VX | (1 << 23))
#define ljdZONE_VX_TARGET    (ljdZONE_API_VX | (1 << 24))
#define ljdZONE_VX_TENSOR    (ljdZONE_API_VX | (1 << 25))
#define ljdZONE_VX_THRESHOLD (ljdZONE_API_VX | (1 << 26))
#define ljdZONE_VX_SPINST    (ljdZONE_API_VX | (1 << 27))
#define ljdZONE_VX_SP        (ljdZONE_API_VX | (1 << 28))
#define ljdZONE_VX_OTHERS    (ljdZONE_API_VX | (1 << 29))

/******************************************************************************
 ******************************** Utility Zones *******************************
 ******************************************************************************/

/* Value for Disabling All Subzones */
#define ljdZONE_NONE 0xF0000000

/* Value for Enabling All Subzones */
#define ljdZONE_ALL 0x0FFFFFFF

/******************************************************************************
 ********************************** END ***************************************
 ******************************************************************************/

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_debug_zones_h_ */


