/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_driver_shared_h_
#define __lj_hal_driver_shared_h_

#include "lj_hal_enum_shared.h"
#include "lj_hal_types_shared.h"

#if ljdENABLE_VG
#include "lj_hal_driver_vg.h"
#endif

#if defined(__QNXNTO__)
#include <sys/siginfo.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* the DC instruction will cause segfault when disable writebufer and cache.*/
#ifndef ljdSKIP_ARM_DC_INSTRUCTION
#define ljdSKIP_ARM_DC_INSTRUCTION 0
#endif

/* The number of context buffers per user. */
#if ljdCAPTURE_ONLY_MODE
#define ljdCONTEXT_BUFFER_COUNT 1
#else
#define ljdCONTEXT_BUFFER_COUNT 2
#endif

#define ljdRENDER_FENCE_LENGTH              (6 * ljmSIZEOF(ljtUINT32))
#define ljdBLT_FENCE_LENGTH                 (10 * ljmSIZEOF(ljtUINT32))

/******************************************************************************
 ****************************** I/O Control Codes *****************************
 ******************************************************************************/

#define ljvHAL_CLASS                        "ljmcore"
#define IOCTL_LJHAL_INTERFACE               30000
#define IOCTL_LJHAL_PROFILER_INTERFACE      30001
#define IOCTL_LJHAL_TERMINATE               30002

/******************************************************************************
 ***************************** Interface Structure ****************************
 ******************************************************************************/

#define ljdMAX_PROFILE_FILE_NAME            128
#define ljdMAX_FLAT_MAPPING_COUNT           8

/* ljvHAL_CHIP_INFO */
typedef struct _ljsHAL_CHIP_INFO {
    /* Chip count. */
    OUT ljtUINT16           count;

    /* Chip types. */
    OUT ljtUINT8            types[ljdGLOBAL_CORE_COUNT];

    /* Chip IDs. */
    OUT ljtUINT8            ids[ljdGLOBAL_CORE_COUNT];

    OUT ljtUINT16           coreIndexs[ljdGLOBAL_CORE_COUNT];

    /* Hardware device ID. */
    OUT ljtUINT8            hwDevIDs[ljdGLOBAL_CORE_COUNT];
} ljsHAL_CHIP_INFO;

/* ljvHAL_VERSION */
typedef struct _ljsHAL_VERSION {
    /* version: <major>.<minor>.<patch>. */
    OUT ljtINT32            major;
    OUT ljtINT32            minor;
    OUT ljtINT32            patch;

    /* Build version. */
    OUT ljtUINT32           build;
} ljsHAL_VERSION;

/* ljvHAL_SET_TIMEOUT. */
typedef struct _ljsHAL_SET_TIMEOUT {
    ljtUINT32               timeOut;
} ljsHAL_SET_TIMEOUT;

/* ljvHAL_QUERY_VIDEO_MEMORY */
typedef struct _ljsHAL_QUERY_VIDEO_MEMORY {
    /* Physical memory address of internal memory. Just a name. */
    OUT ljtUINT32           internalPhysName;
    /* Size in bytes of internal memory. */
    OUT ljtUINT64           internalSize;

    /* Physical memory address of external memory. Just a name. */
    OUT ljtUINT32           externalPhysName;
    /* Size in bytes of external memory.*/
    OUT ljtUINT64           externalSize;

    /* Physical memory address of contiguous memory. Just a name. */
    OUT ljtUINT32           contiguousPhysName;
    /* Size in bytes of contiguous memory.*/
    OUT ljtUINT64           contiguousSize;

    /* Physical memory address of exclusive memory. Just a name. */
    OUT ljtUINT32           exclusivePhysName;
    /* Size in bytes of exclusive memory.*/
    OUT ljtUINT64           exclusiveSize;

    /* If the virtual pool can be an available video memory pool. */
    OUT ljtBOOL             virtualPoolEnabled;
} ljsHAL_QUERY_VIDEO_MEMORY;

/* ljvHAL_QUERY_CHIP_IDENTITY */
typedef struct _ljsHAL_QUERY_CHIP_IDENTITY *ljsHAL_QUERY_CHIP_IDENTITY_PTR;
typedef struct _ljsHAL_QUERY_CHIP_IDENTITY {
    /* Chip model. */
    ljeCHIPMODEL                chipModel;

    /* Revision value.*/
    ljtUINT32                   chipRevision;

    /* Chip date. */
    ljtUINT32                   chipDate;

    /* Supported feature fields. */
    ljtUINT32                   chipFeatures;

    /* Supported minor feature fields. */
    ljtUINT32                   chipMinorFeatures;

    /* Supported minor feature 1 fields. */
    ljtUINT32                   chipMinorFeatures1;

    /* Supported minor feature 2 fields. */
    ljtUINT32                   chipMinorFeatures2;

    /* Supported minor feature 3 fields. */
    ljtUINT32                   chipMinorFeatures3;

    /* Supported minor feature 4 fields. */
    ljtUINT32                   chipMinorFeatures4;

    /* Supported minor feature 5 fields. */
    ljtUINT32                   chipMinorFeatures5;

    /* Supported minor feature 6 fields. */
    ljtUINT32                   chipMinorFeatures6;

    /* Number of streams supported. */
    ljtUINT32                   streamCount;

    /* Number of pixel pipes. */
    ljtUINT32                   pixelPipes;

    /* Number of resolve pipes. */
    ljtUINT32                   resolvePipes;

    /* Number of instructions. */
    ljtUINT32                   instructionCount;

    /* Number of PS instructions. */
    ljtUINT32                   PSInstructionCount;

    /* Number of constants. */
    ljtUINT32                   numConstants;

    /* Number of varyings */
    ljtUINT32                   varyingsCount;

    /* Number of 3D GPUs */
    ljtUINT32                   gpuCoreCount;

    /* Physical mask of all AVAILABLE clusters in core.*/
    ljtUINT32                   clusterAvailMask;

    /* Product ID */
    ljtUINT32                   productID;

    /* Special chip flag bits */
    ljeCHIP_FLAG                chipFlags;

    /* ECO ID. */
    ljtUINT32                   ecoID;

    /* Customer ID. */
    ljtUINT32                   customerID;

    /* CPU view physical address and size of SRAMs. */
    ljtUINT64                   sRAMBases[ljvSRAM_INTER_COUNT];
    ljtUINT32                   sRAMSizes[ljvSRAM_INTER_COUNT];

    ljtUINT64                   platformFlagBits;

    /* APB register offset. */
    ljtUINT64                   registerAPB;

    /* Enabled NN clusters number. */
    ljtUINT32                   nnClusterNum;

    /* Virtual address bits. */
    ljtUINT32                   virtualAddressBits;

    ljtUINT32                   chipConfig;
} ljsHAL_QUERY_CHIP_IDENTITY;

/* ljvHAL_QUERY_CHIP_OPTION. */
typedef struct _ljsHAL_QUERY_CHIP_OPTIONS *ljsHAL_QUERY_CHIP_OPTIONS_PTR;
typedef struct _ljsHAL_QUERY_CHIP_OPTIONS {
    ljtBOOL                     gpuProfiler;
    ljtBOOL                     allowFastClear;
    ljtBOOL                     powerManagement;
    /*
     * Whether use new MMU. It is meaningless
     * for old MMU since old MMU is always enabled.
     */
    ljtBOOL                     enableMMU;
    ljeCOMPRESSION_OPTION       allowCompression;
    ljtBOOL                     smallBatch;
    ljtUINT32                   uscL1CacheRatio;
    ljtUINT32                   uscAttribCacheRatio;
    ljtUINT32                   userClusterMask;
    ljtUINT32                   userClusterMasks[ljdMAX_MAJOR_CORE_COUNT];

    /* Internal SRAM. */
    ljtADDRESS                  sRAMGPUVirtAddrs[ljvSRAM_INTER_COUNT];
    ljtUINT32                   sRAMSizes[ljvSRAM_INTER_COUNT];
    ljtUINT32                   sRAMCount;

    /* External SRAM. */
    ljtPHYS_ADDR_T              extSRAMCPUPhysAddrs[ljvSRAM_EXT_COUNT];
    ljtPHYS_ADDR_T              extSRAMGPUPhysAddrs[ljvSRAM_EXT_COUNT];
    ljtADDRESS                  extSRAMGPUVirtAddrs[ljvSRAM_EXT_COUNT];
    ljtUINT32                   extSRAMGPUPhysNames[ljvSRAM_EXT_COUNT];
    ljtUINT32                   extSRAMSizes[ljvSRAM_EXT_COUNT];
    ljtUINT32                   extSRAMCount;

    ljeSECURE_MODE              secureMode;

    ljtBOOL                     hasShader;

    /* NN clusters power control. */
    ljtUINT32                   enableNNClusters;
    ljtUINT32                   configNNPowerControl;
    /* Active NN core count. */
    ljtUINT32                   activeNNCoreCount;

    /* Only represents system reserved memory pool currently. */
    ljtUINT32                   vidMemCount;

    ljtBOOL                     processPageTable;
    ljtBOOL                     sharedPageTable;
} ljsHAL_QUERY_CHIP_OPTIONS;

/* ljvHAL_QUERY_CHIP_FREQUENCY. */
typedef struct _ljsHAL_QUERY_CHIP_FREQUENCY *ljsHAL_QUERY_CHIP_FREQUENCY_PTR;
typedef struct _ljsHAL_QUERY_CHIP_FREQUENCY {
    OUT ljtUINT64               mcClk;
    OUT ljtUINT64               shClk;
} ljsHAL_QUERY_CHIP_FREQUENCY;

/* Obsolete for userpace. */
/* ljvHAL_ALLOCATE_NON_PAGED_MEMORY */
typedef struct _ljsHAL_ALLOCATE_NON_PAGED_MEMORY {
    /* Allocation flags. */
    IN ljtUINT32                flags;

    /* Number of bytes to allocate. */
    IN OUT ljtUINT64            bytes;

    /* Physical address of allocation. Just a name. */
    OUT ljtUINT32               physName;

    /* Logical address of allocation. */
    OUT ljtUINT64               logical;
} ljsHAL_ALLOCATE_NON_PAGED_MEMORY;

/* Obsolete for userpace. */
/* ljvHAL_FREE_NON_PAGED_MEMORY */
typedef struct _ljsHAL_FREE_NON_PAGED_MEMORY {
    /* Number of bytes allocated. */
    IN ljtUINT64                bytes;

    /* Physical address of allocation. Just a name. */
    IN ljtUINT32                physName;

    /* Logical address of allocation. */
    IN ljtUINT64                logical;
} ljsHAL_FREE_NON_PAGED_MEMORY;

/* Video memory allocation. */
/* ljvHAL_ALLOCATE_LINEAR_VIDEO_MEMORY */
typedef struct _ljsHAL_ALLOCATE_LINEAR_VIDEO_MEMORY {
    /* kernel child process, added by hym on 20221007 */
    IN ljtUINT32                pid;

    /* Number of bytes to allocate. */
    IN OUT ljtUINT64            bytes;

    /* Buffer alignment. */
    IN ljtUINT32                alignment;

    /* Type of allocation, see ljeVIDMEM_TYPE. */
    IN ljtUINT32                type;

    /* Flag of allocation. */
    IN ljtUINT32                flag;

    /* Memory pool to allocate from. */
    IN OUT ljtUINT32            pool;

    /* Internal SRAM index. */
    IN ljtINT32                 sRAMIndex;

    /* External SRAM index. */
    IN ljtINT32                 extSRAMIndex;

    /* Allocated video memory. */
    OUT ljtUINT32               node;

    /* Video memory index, only represents system reserved memroy pool currently. */
    IN ljtINT32                 vidMemIndex;
} ljsHAL_ALLOCATE_LINEAR_VIDEO_MEMORY;

typedef struct _ljsUSER_MEMORY_DESC {
    /* Import flag. */
    ljtUINT32                   flag;

    /* ljvALLOC_FLAG_DMABUF */
    ljtUINT32                   handle;
    ljtUINT64                   dmabuf;

    /* ljvALLOC_FLAG_USERMEMORY */
    ljtUINT64                   logical;
    ljtUINT64                   physical;
    ljtUINT64                   size;
} ljsUSER_MEMORY_DESC;

/* ljvHAL_WRAP_USER_MEMORY. */
typedef struct _ljsHAL_WRAP_USER_MEMORY {
    /* Description of user memory. */
    IN ljsUSER_MEMORY_DESC      desc;

    /* Video memory allocation type. */
    IN ljtUINT32                type;

    /* Output video mmory node. */
    OUT ljtUINT32               node;

    /* size of the node in bytes */
    OUT ljtUINT64               bytes;
} ljsHAL_WRAP_USER_MEMORY;

/* ljvHAL_RELEASE_VIDEO_MEMORY */
typedef struct _ljsHAL_RELEASE_VIDEO_MEMORY {
    /* kernel child process, added by hym on 20221007 */
    IN ljtUINT32                pid;

    /* Allocated video memory. */
    IN ljtUINT32                node;

#ifdef __QNXNTO__
    /* Mapped logical address to unmap in user space. */
    OUT ljtUINT64               memory;

    /* Number of bytes to allocated. */
    OUT ljtUINT64               bytes;
#endif
} ljsHAL_RELEASE_VIDEO_MEMORY;

/* ljvHAL_LOCK_VIDEO_MEMORY */
typedef struct _ljsHAL_LOCK_VIDEO_MEMORY {
    /* Process id of creator, by hym on 20221008 */
    IN ljtUINT32                pid;

    /* Allocated video memory. */
    IN ljtUINT32                node;

    /* Cache configuration. */
    /* Only ljvPOOL_VIRTUAL can be configured */
    IN ljtBOOL                  cacheable;

    /* Hardware specific address. */
    OUT ljtADDRESS              address;

    /* Mapped logical address. */
    OUT ljtUINT64               memory;

    /* Customer priviate handle*/
    OUT ljtUINT32               gid;

    /* Bus address of a contiguous video node. */
    OUT ljtUINT64               physicalAddress;

#if ljdCAPTURE_ONLY_MODE
    IN ljtBOOL                  queryCapSize;
    IN ljtPOINTER               captureLogical;
    OUT ljtUINT64               captureSize;
#endif

    IN ljeLOCK_VIDEO_MEMORY_OP  op;
} ljsHAL_LOCK_VIDEO_MEMORY;

/* ljvHAL_UNLOCK_VIDEO_MEMORY */
typedef struct _ljsHAL_UNLOCK_VIDEO_MEMORY {
    /* Process id of creator, by hym on 20221008 */
    IN ljtUINT32                pid;

    /* Allocated video memory. */
    IN ljtUINT64                node;

    /* Video memory allocation type. */
    IN ljtUINT32                type;

    /* Pool of the unlock node */
    OUT ljtUINT32               pool;

    /* Bytes of the unlock node */
    OUT ljtUINT64               bytes;

    /* Flag to unlock surface asynchroneously. */
    IN OUT ljtBOOL              asynchroneous;

#if ljdCAPTURE_ONLY_MODE
    OUT ljtPOINTER              captureLogical;
#endif

    IN ljeLOCK_VIDEO_MEMORY_OP  op;

    IN ljtUINT64                mmu;
} ljsHAL_UNLOCK_VIDEO_MEMORY;

/* ljvHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY: */
typedef struct _ljsHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY {
    /* Process id of creator, by hym on 20221008 */
    IN ljtUINT32                pid;

    /* Allocated video memory. */
    IN ljtUINT32                node;

    /* Video memory allocation type. */
    IN ljtUINT32                type;
} ljsHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY;

/* ljvHAL_EXPORT_VIDEO_MEMORY. */
typedef struct _ljsHAL_EXPORT_VIDEO_MEMORY {
    /* Allocated video memory. */
    IN ljtUINT32                node;

    /* Export flags */
    IN ljtUINT32                flags;

    /* Exported dma_buf fd */
    OUT ljtINT32                fd;
} ljsHAL_EXPORT_VIDEO_MEMORY;

/* ljvHAL_NAME_VIDEO_MEMORY. */
typedef struct _ljsHAL_NAME_VIDEO_MEMORY {
    IN ljtUINT32                handle;
    OUT ljtUINT32               name;
} ljsHAL_NAME_VIDEO_MEMORY;

/* ljvHAL_IMPORT_VIDEO_MEMORY. */
typedef struct _ljsHAL_IMPORT_VIDEO_MEMORY {
    IN ljtUINT32                name;
    OUT ljtUINT32               handle;
} ljsHAL_IMPORT_VIDEO_MEMORY;

/* ljvHAL_MAP_MEMORY */
typedef struct _ljsHAL_MAP_MEMORY {
    /* Physical memory address to map. Just a name on Linux/Qnx. */
    IN ljtUINT32                physName;

    /* Number of bytes in physical memory to map. */
    IN ljtUINT64                bytes;

    /* Address of mapped memory. */
    OUT ljtUINT64               logical;
} ljsHAL_MAP_MEMORY;

/* ljvHAL_UNMAP_MEMORY */
typedef struct _ljsHAL_UNMAP_MEMORY {
    /* Physical memory address to unmap. Just a name on Linux/Qnx. */
    IN ljtUINT32                physName;

    /* Number of bytes in physical memory to unmap. */
    IN ljtUINT64                bytes;

    /* Address of mapped memory to unmap. */
    IN ljtUINT64                logical;
} ljsHAL_UNMAP_MEMORY;

/* ljvHAL_CACHE */
typedef struct _ljsHAL_CACHE {
    IN ljeCACHEOPERATION        operation;
    IN ljtUINT64                process;
    IN ljtUINT64                logical;
    IN ljtUINT64                bytes;
    IN ljtUINT32                node;
    IN ljtUINT64                offset;
} ljsHAL_CACHE;

/* ljvHAL_ATTACH */
typedef struct _ljsHAL_ATTACH {
    /* Handle of context buffer object. */
    OUT ljtUINT32               context;

    /* Maximum state in the buffer. */
    OUT ljtUINT64               maxState;

    /* Number of states in the buffer. */
    OUT ljtUINT32               numStates;

    /* Map context buffer to user or not. */
    IN ljtBOOL                  map;

    /* Physical of context buffer. */
    OUT ljtUINT64               logicals[2];

    /* Bytes of context buffer. */
    OUT ljtUINT32               bytes;

    /* The context is for multi-core or not. */
    IN ljtBOOL                  shared;

#if ljdCAPTURE_ONLY_MODE
    IN ljtBOOL                  queryCapSize;
    IN ljtPOINTER               contextLogical[ljdCONTEXT_BUFFER_COUNT];
    OUT ljtUINT64               captureSize;
#endif
} ljsHAL_ATTACH;

/* ljvHAL_DETACH */
typedef struct _ljsHAL_DETACH {
    /* Context buffer object ljkCONTEXT. Just a name. */
    IN ljtUINT32                context;
} ljsHAL_DETACH;

/* ljvHAL_EVENT_COMMIT. */
typedef struct _ljsHAL_EVENT_COMMIT {
    /* Event queue in ljsQUEUE. */
    IN ljtUINT64                queue;

    /* Brother cores in user device of current commit process. */
    IN ljtUINT32                broCoreMask;

    IN ljtBOOL                  shared;
#if ljdENABLE_SW_PREEMPTION
    /* Priority ID. */
    IN ljtUINT32                priorityID;

    /* Does it require top priority? */
    IN ljtBOOL                  topPriority;
#endif
} ljsHAL_EVENT_COMMIT;

typedef struct _ljsHAL_COMMAND_LOCATION {
    ljtUINT32                   priority;
    ljtUINT32                   channelId;

    ljtUINT32                   videoMemNode;

    ljtADDRESS                  address;
    ljtUINT64                   logical;
    ljtUINT32                   startOffset;
    /* size includes reservedHead and reservedTail. */
    ljtUINT32                   size;

    ljtUINT32                   reservedHead;
    ljtUINT32                   reservedTail;

    /* Pointer to patch list. */
    ljtUINT64                   patchHead;

    /*
     * Location index of exit commands, ie where to put the chipEnable/link back
     * commands in the reservedTail area.
     * It's used in fully shared command buffer for multiple cores.
     */
    ljtUINT32                   exitIndex;
    ljtUINT32                   entryPipe;
    ljtUINT32                   exitPipe;

    /* struct _ljsHAL_COMMAND_LOCATION * next; */
    ljtUINT64                   next;
#if ljdCAPTURE_ONLY_MODE
    ljtPOINTER                  contextLogical[ljdCONTEXT_BUFFER_COUNT];
#endif
} ljsHAL_COMMAND_LOCATION;

typedef struct _ljsHAL_SUBCOMMIT {
    ljtUINT32                   coreId;

    /* user ljsSTATE_DELTA_PTR. */
    ljtUINT64                   delta;

    /* Kernel ljkCONTEXT. */
    ljtUINT64                   context;

    /* Event queue in user ljsQUEUE *. */
    ljtUINT64                   queue;

    /* Locate the commands. */
    ljsHAL_COMMAND_LOCATION     commandBuffer;

    /* struct _ljsHAL_SUBCOMMIT * next; */
    ljtUINT64                   next;

#if ljdENABLE_SW_PREEMPTION
    /* Process ID. */
    ljtUINT32                   pid;

    /* Engine type. */
    ljeENGINE                   engine;

    /* Is it multi-core and shared mode?. */
    ljtBOOL                     shared;

    /* Priority ID. */
    ljtUINT32                   priorityID;

    /* Does it require top priority. */
    ljtBOOL                     topPriority;
#endif
} ljsHAL_SUBCOMMIT, *ljsHAL_SUBCOMMIT_PTR;

/* ljvHAL_COMMIT */
typedef struct _ljsHAL_COMMIT {
    ljsHAL_SUBCOMMIT            subCommit;

    /* The command buffer is linked to multiple command queue. */
    ljtBOOL                     shared;

    ljtBOOL                     contextSwitched;

    /* Commit stamp of this commit. */
    OUT ljtUINT64               commitStamp;

    /* Brother cores in user device of current commit process. */
    ljtUINT32                   broCoreMask;

#if ljdENABLE_MP_SWITCH
    /* Multi-processor mode. */
    ljtUINT32                   mpMode;

    /* Switch multi-processor mode. */
    ljtUINT32                   switchMpMode;
#endif

#if ljdENABLE_SW_PREEMPTION
    /* If user need to merge the delta. */
    ljtBOOL                     needMerge;

    /* If this commit is pending. */
    ljtBOOL                     pending;
#endif
} ljsHAL_COMMIT;

#if ljdENABLE_VG
/* ljvHAL_COMMIT */
typedef struct _ljsHAL_VGCOMMIT {
    /* Context buffer. ljsVGCONTEXT_PTR */
    IN ljtUINT64                context;

    /* Command queue. ljsVGCMDQUEUE_PTR */
    IN ljtUINT64                queue;

    /* Number of entries in the queue. */
    IN ljtUINT32                entryCount;

    /* Task table. ljsTASK_MASTER_TABLE_PTR */
    IN ljtUINT64                taskTable;
} ljsHAL_VGCOMMIT;
#endif

typedef struct _ljsHAL_COMMIT_DONE {
    IN ljtUINT64                context;

#if ljdENABLE_SW_PREEMPTION
    /* Priority ID. */
    IN ljtUINT32                priorityID;
#endif
} ljsHAL_COMMIT_DONE;

/* ljvHAL_USER_SIGNAL  */
typedef struct _ljsHAL_USER_SIGNAL {
    /* Command. */
    ljeUSER_SIGNAL_COMMAND_CODES command;

    /* Signal ID. */
    IN OUT ljtINT32             id;

    /* Reset mode. */
    IN ljtBOOL                  manualReset;

    /* Wait timedout. */
    IN ljtUINT32                wait;

    /* State. */
    IN ljtBOOL                  state;

    /* Return status */
    IN ljeSIGNAL_STATUS         status;
} ljsHAL_USER_SIGNAL;

/* ljvHAL_SIGNAL. */
typedef struct _ljsHAL_SIGNAL {
    /* Signal handle to signal ljtSIGNAL. */
    IN ljtUINT64                signal;

    /* Reserved ljtSIGNAL. */
    IN ljtUINT64                auxSignal;

    /* Process owning the signal ljtHANDLE. */
    IN ljtUINT64                process;

#if defined(__QNXNTO__)
    /* Client pulse event. */
    IN struct sigevent          event;

    /* Set by server. */
    IN ljtINT32                 rcvid;
#endif
    /* Event generated from where of pipeline */
    IN ljeKERNEL_WHERE          fromWhere;

#if ljdENABLE_SW_PREEMPTION
    /* If it is a fence signal. */
    IN ljtBOOL                  fenceSignal;
#endif
} ljsHAL_SIGNAL;

/* ljvHAL_WRITE_DATA. */
typedef struct _ljsHAL_WRITE_DATA {
    /* Address to write data to. */
    IN ljtUINT32                address;

    /* Data to write. */
    IN ljtUINT32                data;
} ljsHAL_WRITE_DATA;

/* ljvHAL_READ_REGISTER */
typedef struct _ljsHAL_READ_REGISTER {
    /* Logical address of memory to write data to. */
    IN ljtUINT32                address;

    /* Data read. */
    OUT ljtUINT32               data;
} ljsHAL_READ_REGISTER;

/* ljvHAL_WRITE_REGISTER */
typedef struct _ljsHAL_WRITE_REGISTER {
    /* Logical address of memory to write data to. */
    IN ljtUINT32                address;

    /* Data read. */
    IN ljtUINT32                data;
} ljsHAL_WRITE_REGISTER;

/* ljvHAL_READ_REGISTER_EX */
typedef struct _ljsHAL_READ_REGISTER_EX {
    /* Logical address of memory to write data to. */
    IN ljtUINT32                address;

    IN ljtUINT32                coreSelect;

    /* Data read. */
    OUT ljtUINT32               data[4];
} ljsHAL_READ_REGISTER_EX;

/* ljvHAL_WRITE_REGISTER_EX */
typedef struct _ljsHAL_WRITE_REGISTER_EX {
    /* Logical address of memory to write data to. */
    IN ljtUINT32                address;

    IN ljtUINT32                coreSelect;

    /* Data read. */
    IN ljtUINT32                data[4];
} ljsHAL_WRITE_REGISTER_EX;

/* ljvHAL_APB_AXIFEE_ACCESS */
typedef struct _ljsHAL_APB_AXIFE_ACCESS {
    /* Logical address of memory to write data to. */
    IN ljtUINT32                address;

    IN ljtUINT32                coreSelect;

    IN ljtBOOL                  isRead;

    /* Data read. */
    IN ljtUINT32                data;
} ljsHAL_APB_AXIFE_ACCESS;

#if LJMICRO_PROFILER
/* ljvHAL_GET_PROFILE_SETTING */
typedef struct _ljsHAL_GET_PROFILE_SETTING {
    /* Enable profiling */
    OUT ljtBOOL                 enable;
    /* Profile mode */
    OUT ljeProfilerMode         profileMode;
    /* Probe mode */
    OUT ljeProbeMode            probeMode;
} ljsHAL_GET_PROFILE_SETTING;

/* ljvHAL_SET_PROFILE_SETTING */
typedef struct _ljsHAL_SET_PROFILE_SETTING {
    /* Enable profiling */
    IN ljtBOOL                  enable;
    /* Profile mode */
    IN ljeProfilerMode          profileMode;
    /* Probe mode */
    IN ljeProbeMode             probeMode;
} ljsHAL_SET_PROFILE_SETTING;

/* ljvHAL_READ_PROFILER_REGISTER_SETTING */
typedef struct _ljsHAL_READ_PROFILER_REGISTER_SETTING {
    /*Should Clear Register*/
    IN ljtBOOL                  bclear;
} ljsHAL_READ_PROFILER_REGISTER_SETTING;

typedef struct _ljsHAL_READ_ALL_PROFILE_REGISTERS_PART1 {
    /* Context buffer object ljkCONTEXT. Just a name. */
    IN ljtUINT32                    context;

    /* Data read. */
    OUT ljsPROFILER_COUNTERS_PART1  Counters;
} ljsHAL_READ_ALL_PROFILE_REGISTERS_PART1;

typedef struct _ljsHAL_READ_ALL_PROFILE_REGISTERS_PART2 {
    /* Context buffer object ljkCONTEXT. Just a name. */
    IN ljtUINT32                    context;

    /* Data read. */
    OUT ljsPROFILER_COUNTERS_PART2  Counters;
} ljsHAL_READ_ALL_PROFILE_REGISTERS_PART2;

/* ljvHAL_PROFILE_REGISTERS_2D */
typedef struct _ljsHAL_PROFILE_REGISTERS_2D {
    /* Data read in ljs2D_PROFILE. */
    OUT ljtUINT64               hwProfile2D;
} ljsHAL_PROFILE_REGISTERS_2D;
#endif

/* ljvHAL_SET_POWER_MANAGEMENT_STATE */
typedef struct _ljsHAL_SET_POWER_MANAGEMENT {
    /* Data read. */
    IN ljeCHIPPOWERSTATE        state;
} ljsHAL_SET_POWER_MANAGEMENT;

/* ljvHAL_QUERY_POWER_MANAGEMENT_STATE */
typedef struct _ljsHAL_QUERY_POWER_MANAGEMENT {
    /* Data read. */
    OUT ljeCHIPPOWERSTATE       state;

    /* Idle query. */
    OUT ljtBOOL                 isIdle;
} ljsHAL_QUERY_POWER_MANAGEMENT;

/* ljvHAL_CONFIG_POWER_MANAGEMENT. */
typedef struct _ljsHAL_CONFIG_POWER_MANAGEMENT {
    IN ljtBOOL                  enable;
    OUT ljtBOOL                 oldValue;
} ljsHAL_CONFIG_POWER_MANAGEMENT;

typedef struct _ljsFLAT_MAPPING_RANGE {
    ljtUINT64                   start;
    ljtUINT64                   end;
    ljtUINT64                   size;
    ljeFLATMAP_FLAG             flag;

    /* Corresponding virtual start. */
    ljtUINT64                   vStart;
} ljsFLAT_MAPPING_RANGE;

/* ljvHAL_GET_BASE_ADDRESS */
typedef struct _ljsHAL_GET_BASE_ADDRESS {
    /* Physical memory address of internal memory. */
    OUT ljtUINT32               baseAddress;

    OUT ljtUINT32               flatMappingRangeCount;

    OUT ljsFLAT_MAPPING_RANGE   flatMappingRanges[ljdMAX_FLAT_MAPPING_COUNT];
} ljsHAL_GET_BASE_ADDRESS;

typedef struct _ljsHAL_SET_DEBUG_LEVEL_ZONE {
    IN ljtUINT32                level;
    IN ljtUINT32                zones;
    IN ljtBOOL                  enable;
} ljsHAL_SET_DEBUG_LEVEL_ZONE;

typedef struct _ljsHAL_QUERY_CPU_FREQUENCY
{
    IN  ljtUINT32               CPUId;
    OUT ljtUINT32               CPUFrequency;
} ljsHAL_QUERY_CPU_FREQUENCY;

/* ljvHAL_DEBUG_DUMP. */
typedef struct _ljsHAL_DEBUG_DUMP {
    /* ljeDUMP_BUFFER_TYPE      type. */
    IN ljtUINT32                type;

    IN ljtUINT64                ptr;
    IN ljtADDRESS               address;
    IN ljtUINT32                size;
} ljsHAL_DEBUG_DUMP;


/* ljvHAL_TIMESTAMP */
typedef struct _ljsHAL_TIMESTAMP {
    /* Timer select. */
    IN ljtUINT32                timer;

    /* Timer request type (0-stop, 1-start, 2-send delta). */
    IN ljtUINT32                request;

    /* Result of delta time in microseconds. */
    OUT ljtINT32                timeDelta;
} ljsHAL_TIMESTAMP;

/* ljvHAL_DATABASE */
typedef struct _ljsHAL_DATABASE {
    /*
     * Set to ljvTRUE if you want to query a particular process ID.
     * Set to ljvFALSE to query the last detached process.
     */
    IN ljtBOOL                  validProcessID;

    /* Process ID to query. */
    IN ljtUINT32                processID;

    /* Information. */
    OUT ljuDATABASE_INFO        vidMem;
    OUT ljuDATABASE_INFO        nonPaged;
    OUT ljuDATABASE_INFO        contiguous;
    OUT ljuDATABASE_INFO        gpuIdle;

    /* Detail information about video memory. */
    OUT ljuDATABASE_INFO        vidMemPool[3];
} ljsHAL_DATABASE;

/* ljvHAL_GET_FRAME_INFO. */
typedef struct _ljsHAL_GET_FRAME_INFO {
    /* ljsHAL_FRAME_INFO* */
    OUT ljtUINT64               frameInfo;
} ljsHAL_GET_FRAME_INFO;

#if ljdENABLE_VG
/* ljvHAL_QUERY_COMMAND_BUFFER */
typedef struct _ljsHAL_QUERY_COMMAND_BUFFER {
    /* Command buffer attributes. */
    OUT ljsCOMMAND_BUFFER_INFO  information;
} ljsHAL_QUERY_COMMAND_BUFFER;
#endif

typedef struct _ljsHAL_SET_FSCALE_VALUE {
    IN ljtUINT32                value;
    IN ljtUINT32                shValue;
} ljsHAL_SET_FSCALE_VALUE;

typedef struct _ljsHAL_GET_FSCALE_VALUE {
    OUT ljtUINT32               value;
    OUT ljtUINT32               minValue;
    OUT ljtUINT32               maxValue;
} ljsHAL_GET_FSCALE_VALUE;

/* ljvHAL_QUERY_RESET_TIME_STAMP. */
typedef struct _ljsHAL_QUERY_RESET_TIME_STAMP {
    OUT ljtUINT64               timeStamp;
    OUT ljtUINT64               contextID;
} ljsHAL_QUERY_RESET_TIME_STAMP;

/* ljvHAL_CREATE_NATIVE_FENCE. */
typedef struct _ljsHAL_CREATE_NATIVE_FENCE {
    /* Signal id. */
    IN ljtUINT64                signal;

    /* Native fence file descriptor. */
    OUT ljtINT32                fenceFD;

} ljsHAL_CREATE_NATIVE_FENCE;

/* ljvHAL_WAIT_NATIVE_FENCE. */
typedef struct _ljsHAL_WAIT_NATIVE_FENCE {
    /* Native fence file descriptor. */
    IN ljtINT32                 fenceFD;

    /* Wait timeout. */
    IN ljtUINT32                timeout;
} ljsHAL_WAIT_NATIVE_FENCE;

/* ljvHAL_SHBUF. */
typedef struct _ljsHAL_SHBUF {
    ljeSHBUF_COMMAND_CODES      command;

    /* Shared buffer. */
    IN OUT ljtUINT64            id;

    /* User data to be shared. */
    IN ljtUINT64                data;

    /* Data size. */
    IN OUT ljtUINT32            bytes;
} ljsHAL_SHBUF;

/* ljvHAL_GET_GRAPHIC_BUFFER_FD. */
/*
 * Fd representation of android graphic buffer contents.
 * Currently, it is only to reference video nodes, signal, etc to avoid being
 * destroyed when trasfering across processes.
 */
typedef struct _ljsHAL_GET_GRAPHIC_BUFFER_FD {
    /* Max 3 video nodes, node handle here. */
    IN ljtUINT32                node[3];

    /* A shBuf. */
    IN ljtUINT64                shBuf;

    /* A signal. */
    IN ljtUINT64                signal;

    OUT ljtINT32                fd;
} ljsHAL_GET_GRAPHIC_BUFFER_FD;

typedef struct _ljsHAL_VIDEO_MEMORY_METADATA {
    /* Allocated video memory. */
    IN ljtUINT32                node;

    IN ljtUINT32                readback;

    INOUT ljtINT32              ts_fd;
    INOUT ljtUINT32             fc_enabled;
    INOUT ljtUINT32             fc_value;
    INOUT ljtUINT32             fc_value_upper;

    INOUT ljtUINT32             compressed;
    INOUT ljtUINT32             compress_format;
} ljsHAL_VIDEO_MEMORY_METADATA;

/* ljvHAL_GET_VIDEO_MEMORY_FD. */
typedef struct _ljsHAL_GET_VIDEO_MEMORY_FD {
    IN ljtUINT32                handle;
    IN ljtBOOL                  exported;
    OUT ljtINT32                fd;
} ljsHAL_GET_VIDEO_MEMORY_FD;

/* ljvHAL_DESTROY_MMU. */
typedef struct _ljsHAL_DESTROY_MMU {
    /* Mmu object. */
    IN ljtUINT64                mmu;
    IN ljtUINT64                database;
    IN ljtUINT32                pid;
} ljsHAL_DESTROY_MMU;

/* ljvHAL_WAIT_FENCE. */
typedef struct _ljsHAL_WAIT_FENCE {
    IN ljtUINT32                handle;
    IN ljtUINT32                timeOut;
} ljsHAL_WAIT_FENCE;

/* ljvHAL_DEVICE_MUTEX: */
typedef struct _ljsHAL_DEVICE_MUTEX {
    /* Lock or Release device mutex. */
    ljtBOOL                     isMutexLocked;
} ljsHAL_DEVICE_MUTEX;


#if ljdDEC_ENABLE_AHB
/* ljvHAL_DEC300_READ. */
typedef struct _ljsHAL_DEC300_READ {
    ljtUINT32                   enable;
    ljtUINT32                   readId;
    ljtUINT32                   format;
    ljtUINT32                   strides[3];
    ljtUINT32                   is3D;
    ljtUINT32                   isMSAA;
    ljtUINT32                   clearValue;
    ljtUINT32                   isTPC;
    ljtUINT32                   isTPCCompressed;
    ljtUINT32                   surfAddrs[3];
    ljtUINT32                   tileAddrs[3];
} DEC300Read;

/* ljvHAL_DEC300_WRITE. */
typedef struct _ljsHAL_DEC300_WRITE {
    ljtUINT32                   enable;
    ljtUINT32                   readId;
    ljtUINT32                   writeId;
    ljtUINT32                   format;
    ljtUINT32                   surfAddr;
    ljtUINT32                   tileAddr;
} DEC300Write;

/* ljvHAL_DEC300_FLUSH. */
typedef struct _ljsHAL_DEC300_FLUSH {
    IN ljtUINT8                 useless;
} DEC300Flush;

/* ljvHAL_DEC300_FLUSH_WAIT. */
typedef struct _ljsHAL_DEC300_FLUSH_WAIT {
    IN ljtUINT32                done;
} DEC300FlushWait;
#endif

#if ljdENABLE_VIDEO_MEMORY_MIRROR
typedef struct _ljsHAL_SYNC_VIDEO_MEMORY {
    IN ljtUINT64                node;
    IN ljeSYNC_MEMORY_DIRECTION dir;
} ljsHAL_SYNC_VIDEO_MEMORY;
#endif

#if ljdENABLE_CLEAR_FENCE
typedef struct _ljsHAL_STORE_CLEAR_FENCE {
    ljtADDRESS                  address;
    ljtUINT64                   fenceValue;
    ljtUINT64                   recordId;
    ljtBOOL                     use64BitFence;
    ljtBOOL                     isClear;
} ljsHAL_STORE_CLEAR_FENCE;
#endif

typedef struct _ljsHAL_DMA_OPERATION {
    IN ljtUINT32 node_src;
    IN ljtUINT32 node_dst;
    IN ljeDMA_OPERATION_DIR dir;
    IN ljtUINT64 len;
}
ljsHAL_DMA_OPERATION;

typedef struct _ljsHAL_GET_VIDEO_MEMORY_POOL {
    IN ljtUINT32 node;
    OUT ljePOOL pool;
}
ljsHAL_GET_VIDEO_MEMORY_POOL;

typedef struct _ljsHAL_GET_PT_INFO_3D {
    OUT ljtUINT ptCnt;
}
ljsHAL_GET_PT_INFO_3D;

typedef struct _ljsHAL_PHYS_ADDRESS_CONVERT {
    IN ljtPHYS_ADDR_T physAddrSrc;
    OUT ljtPHYS_ADDR_T physAddrDst;
}
ljsHAL_PHYS_ADDRESS_CONVERT;

typedef struct _ljsHAL_INTERFACE {
    /* Command code. */
    ljeHAL_COMMAND_CODES        command;

    /* Hardware type. */
    ljeHARDWARE_TYPE            hardwareType;

    /* Core index for current hardware type. */
    ljtUINT32                   coreIndex;

    /* Status value. */
    ljeSTATUS                   status;

    /* Engine */
    ljeENGINE                   engine;

    /* Ignore information from TSL when doing IO control */
    ljtBOOL                     ignoreTLS;

    /* The mutext already acquired */
    IN ljtBOOL                  commitMutex;

    /* Union of command structures. */
    union _u {
        ljsHAL_CHIP_INFO                    ChipInfo;
        ljsHAL_VERSION                      Version;
        ljsHAL_SET_TIMEOUT                  SetTimeOut;

        ljsHAL_QUERY_VIDEO_MEMORY           QueryVideoMemory;
        ljsHAL_QUERY_CHIP_IDENTITY          QueryChipIdentity;
        ljsHAL_QUERY_CHIP_OPTIONS           QueryChipOptions;
        ljsHAL_QUERY_CHIP_FREQUENCY         QueryChipFrequency;

        ljsHAL_ALLOCATE_NON_PAGED_MEMORY    AllocateNonPagedMemory;
        ljsHAL_FREE_NON_PAGED_MEMORY        FreeNonPagedMemory;

        ljsHAL_ALLOCATE_LINEAR_VIDEO_MEMORY AllocateLinearVideoMemory;
        ljsHAL_WRAP_USER_MEMORY             WrapUserMemory;
        ljsHAL_RELEASE_VIDEO_MEMORY         ReleaseVideoMemory;

        ljsHAL_LOCK_VIDEO_MEMORY               LockVideoMemory;
        ljsHAL_UNLOCK_VIDEO_MEMORY             UnlockVideoMemory;
        ljsHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY BottomHalfUnlockVideoMemory;

        ljsHAL_EXPORT_VIDEO_MEMORY          ExportVideoMemory;
        ljsHAL_NAME_VIDEO_MEMORY            NameVideoMemory;
        ljsHAL_IMPORT_VIDEO_MEMORY          ImportVideoMemory;

        ljsHAL_MAP_MEMORY                   MapMemory;
        ljsHAL_UNMAP_MEMORY                 UnmapMemory;

        ljsHAL_CACHE                        Cache;

        ljsHAL_ATTACH                       Attach;
        ljsHAL_DETACH                       Detach;

        ljsHAL_EVENT_COMMIT                 Event;
        ljsHAL_COMMIT                       Commit;
#if ljdENABLE_VG
        ljsHAL_VGCOMMIT                     VGCommit;
#endif
        ljsHAL_COMMIT_DONE                  CommitDone;

        ljsHAL_USER_SIGNAL                  UserSignal;
        ljsHAL_SIGNAL                       Signal;

        ljsHAL_WRITE_DATA                   WriteData;
        ljsHAL_READ_REGISTER                ReadRegisterData;
        ljsHAL_WRITE_REGISTER               WriteRegisterData;
        ljsHAL_APB_AXIFE_ACCESS             APBAXIFEAccess;
        ljsHAL_READ_REGISTER_EX             ReadRegisterDataEx;
        ljsHAL_WRITE_REGISTER_EX            WriteRegisterDataEx;
        ljsHAL_SET_POWER_MANAGEMENT         SetPowerManagement;
        ljsHAL_QUERY_POWER_MANAGEMENT       QueryPowerManagement;
        ljsHAL_CONFIG_POWER_MANAGEMENT      ConfigPowerManagement;

        ljsHAL_GET_BASE_ADDRESS             GetBaseAddress;

        ljsHAL_SET_DEBUG_LEVEL_ZONE         DebugLevelZone;

        ljsHAL_QUERY_CPU_FREQUENCY          QueryCPUFrequency;

        ljsHAL_DEBUG_DUMP                   DebugDump;

        ljsHAL_TIMESTAMP                    TimeStamp;
        ljsHAL_DATABASE                     Database;

        ljsHAL_GET_FRAME_INFO               GetFrameInfo;

#if ljdENABLE_VG
        ljsHAL_QUERY_COMMAND_BUFFER         QueryCommandBuffer;
#endif

        /* ljsHAL_DUMP_GPU_STATE */
        /* ljsHAL_DUMP_EVENT */

        ljsHAL_SET_FSCALE_VALUE             SetFscaleValue;
        ljsHAL_GET_FSCALE_VALUE             GetFscaleValue;

        ljsHAL_QUERY_RESET_TIME_STAMP       QueryResetTimeStamp;

        ljsHAL_CREATE_NATIVE_FENCE          CreateNativeFence;
        ljsHAL_WAIT_NATIVE_FENCE            WaitNativeFence;
        ljsHAL_SHBUF                        ShBuf;
        ljsHAL_GET_GRAPHIC_BUFFER_FD        GetGraphicBufferFd;
        ljsHAL_VIDEO_MEMORY_METADATA        SetVidMemMetadata;
        ljsHAL_GET_VIDEO_MEMORY_FD          GetVideoMemoryFd;

        ljsHAL_DESTROY_MMU                  DestroyMmu;

        ljsHAL_WAIT_FENCE                   WaitFence;

        /* ljvHAL_DEVICE_MUTEX: */
        ljsHAL_DEVICE_MUTEX                 DeviceMutex;


#if ljdDEC_ENABLE_AHB
        ljsHAL_DEC300_READ                  DEC300Read;
        ljsHAL_DEC300_WRITE                 DEC300Write;
        ljsHAL_DEC300_FLUSH                 DEC300Flush;
        ljsHAL_DEC300_FLUSH_WAIT            DEC300FlushWait;
#endif
#if ljdENABLE_VIDEO_MEMORY_MIRROR
        ljsHAL_SYNC_VIDEO_MEMORY            SyncVideoMemory;
#endif
#if ljdENABLE_CLEAR_FENCE
        ljsHAL_STORE_CLEAR_FENCE            UserFence;
#endif
        ljsHAL_DMA_OPERATION                DmaOperation;
        ljsHAL_GET_VIDEO_MEMORY_POOL        GetVideoMemoryPool;

        ljsHAL_GET_PT_INFO_3D               GetPTInfo3D;

        ljsHAL_PHYS_ADDRESS_CONVERT         PhysAddress;
    } u;

#if ljdENABLE_MULTI_DEVICE_MANAGEMENT
    /* Device index. */
    ljtUINT32                   devIndex;
#endif

    /* O/S specific device context. -- Needed for Windows WDDM device callbacks and kernel mode thunks. */
    ljtUINT64                   devCtxt;

    /* API type. -- Needed for Windows WDDM device kernel mode thunks to set ClientHint when a context is created. */
    ljeAPI                      api;

} ljsHAL_INTERFACE;

#if LJMICRO_PROFILER
typedef struct _ljsHAL_PROFILER_INTERFACE {
    /* Command code. */
    ljeHAL_COMMAND_CODES        command;

    /* Hardware type. */
    ljeHARDWARE_TYPE            hardwareType;

    /* Core index for current hardware type. */
    ljtUINT32                   coreIndex;

    /* Status value. */
    ljeSTATUS                   status;

    /* Ignore information from TSL when doing IO control */
    ljtBOOL                     ignoreTLS;

    /* Union of command structures. */
    union profiler_u {
        ljsHAL_GET_PROFILE_SETTING              GetProfileSetting;
        ljsHAL_SET_PROFILE_SETTING              SetProfileSetting;
        ljsHAL_READ_PROFILER_REGISTER_SETTING   SetProfilerRegisterClear;
        ljsHAL_READ_ALL_PROFILE_REGISTERS_PART1 RegisterProfileData_part1;
        ljsHAL_READ_ALL_PROFILE_REGISTERS_PART2 RegisterProfileData_part2;
        ljsHAL_PROFILE_REGISTERS_2D             RegisterProfileData2D;
    } u;
#if ljdENABLE_MULTI_DEVICE_MANAGEMENT
    /* Device index. */
    ljtUINT32                   devIndex;
#endif

} ljsHAL_PROFILER_INTERFACE;
#endif

/* State delta record. */
typedef struct _ljsSTATE_DELTA_RECORD *ljsSTATE_DELTA_RECORD_PTR;
typedef struct _ljsSTATE_DELTA_RECORD {
    /* State address. */
    ljtUINT                     address;

    /* State mask. */
    ljtUINT32                   mask;

    /* State data. */
    ljtUINT32                   data;
} ljsSTATE_DELTA_RECORD;

/* State delta. */
typedef struct _ljsSTATE_DELTA {
    /* For debugging: the number of delta in the order of creation. */
    ljtUINT                     num;

    /*
     * Main state delta ID. Every time state delta structure gets reinitialized,
     * main ID is incremented. If main state ID overflows, all map entry IDs get
     * reinitialized to make sure there is no potential erroneous match after
     * the overflow.
     */
    ljtUINT                     id;

    /* The number of contexts pending modification by the delta. */
    ljtINT                      refCount;

    /* Vertex element count for the delta buffer. */
    ljtUINT                     elementCount;

    /* Number of states currently stored in the record array. */
    ljtUINT                     recordCount;

    /* Record array; holds all modified states in ljsSTATE_DELTA_RECORD. */
    ljtUINT64                   recordArray;

    /*
     * Map entry ID is used for map entry validation. If map entry ID does not
     * match the main state delta ID, the entry and the corresponding state are
     * considered not in use.
     */
    ljtUINT64                   mapEntryID;
    ljtUINT                     mapEntryIDSize;

    /*
     * If the map entry ID matches the main state delta ID, index points to
     * the state record in the record array.
     */
    ljtUINT64                   mapEntryIndex;

    /* Previous and next state deltas in ljsSTATE_DELTA. */
    ljtUINT64                   prev;
    ljtUINT64                   next;
} ljsSTATE_DELTA;

typedef struct _ljsQUEUE {
    /* Pointer to next ljsQUEUE structure in ljsQUEUE. */
    ljtUINT64                   next;

    /* Event information. */
    ljsHAL_INTERFACE            iface;
} ljsQUEUE;

/* A record chunk include multiple records to save allocation. */
typedef struct _ljsQUEUE_CHUNK {
    struct _ljsQUEUE_CHUNK      *next;

    ljsQUEUE                    record[16];
} ljsQUEUE_CHUNK;

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_driver_shared_h_ */


