/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_drm_h_
#define __lj_hal_drm_h_

#ifdef __KERNEL__
#include <uapi/drm/drm.h>
#else
#include <drm.h>
#endif

#if defined(__cplusplus)
extern "C" {
#endif

/*
 * Kernel modesetting world below.
 */
#define LJM_GPU_GEM_ALIGN           256


#define LJM_GEM_DOMAIN_UNKNOWN              (1u << 0)
#define LJM_GEM_DOMAIN_GTT                  (1u << 1)   //virtual
#define LJM_GEM_DOMAIN_VRAM                 (1u << 2)   //externel->exclusive
#define LJM_GEM_DOMAIN_CPU                  (1u << 3)
#define LJM_GEM_DOMAIN_INVISIBLE_VRAM       (1u << 4)   //exclusive->externel
#define LJM_GEM_DOMAIN_DEFAULT              (1u << 5)   //external->exclusive->virtual
#define LJM_GEM_DOMAIN_VRAM_NOEVICT         (1u << 6)   //externel
#define LJM_GEM_DOMAIN_INVISIBLE_NOEVICT    (1u << 7)   //exclusive
#define LJM_GEM_DOMAIN_INVISIBLE_FIRST      (1u << 8)   //exclusive->externel->virtual


/* creation flag bits. */
#define DRM_VIV_GEM_CONTIGUOUS                  (1u << 0)
#define DRM_VIV_GEM_CACHED                      (1u << 1)
#define DRM_VIV_GEM_SECURE                      (1u << 2)
#define DRM_VIV_GEM_CMA_LIMIT                   (1u << 3)
#define DRM_VIV_GEM_EXCLUSIVE_NONEED_VIRTUAL    (1u << 4)
#define DRM_VIV_GEM_CHANGE_BANK                 (1u << 5)

#define DRM_VIV_GEM_PAGEFLIP_SCANOUT            (1u << 16)
#define DRM_VIV_GEM_PAGEFLIP_SCANOUT_BACK       (1u << 17)
#define DRM_VIV_GEM_CRTC_INDEX_MASK             (3u << 18)
#define DRM_VIV_GEM_CRTC_INDEX_SHIFT            18


struct drm_ljm_gem_create {
    __u64 size;
    __u32 alignment; /*type;*/
    __u32 domains;	/*pool;*/
    __u32 flags;
    __u32 handle;
    __u32 node;
    __u32 type;
    __u32 domain;
};

struct drm_ljm_gem_lock {
    __u32 handle;
    __u32 cacheable;
    __u64 logical;
    __u32 node;
    __u64 priv;
    __u32 priv_size;

    __u64 physical;
    __u64 gpuBusAddr;
};

struct drm_ljm_gem_unlock {
    __u32 handle;
    __u32 node;
    __u32 async;
    __u64 priv;
    __u32 priv_size;
};

struct drm_ljm_gem_release {
    __u32 handle;
    __u32 node;
};

struct drm_ljm_gem_move {
    __u32 handle;
    __u32 node;
    __u32 domain;
};

struct drm_ljm_gem_wrap {
    __u32 handle;
    __u64 priv;
    __u32 priv_size;
};

struct drm_ljm_gem_pf {
    __u32 handle;
    __u32 node;
};

#define DRM_VIV_GEM_CLEAN_CACHE         0x01
#define DRM_VIV_GEM_INVALIDATE_CACHE    0x02
#define DRM_VIV_GEM_FLUSH_CACHE         0x03
#define DRM_VIV_GEM_MEMORY_BARRIER      0x04

struct drm_ljm_gem_cache {
    __u32 handle;
    __u32 op;
    __u64 logical;
    __u64 bytes;
};

#define DRM_VIV_GEM_PARAM_POOL      0x00
#define DRM_VIV_GEM_PARAM_SIZE      0x01
#define DRM_VIV_GEM_PARAM_OFFSET    0x02
#define DRM_VIV_GEM_PARAM_GPU_ADDR  0x03
#define DRM_VIV_GEM_PARAM_CPU_ADDR  0x04
#define DRM_VIV_GEM_PARAM_CPU_TYPE  0x05
#define DRM_VIV_GEM_PARAM_SWCURSOR  0x06
#define DRM_VIV_GEM_PARAM_HWCURSOR  0x07
#define DRM_VIV_GEM_PARAM_VERSION   0x08
#define DRM_VIV_GEM_PARAM_REF       0x09
#define DRM_VIV_GEM_IOMMU           0x0A

struct drm_ljm_gem_query {
    __u32 handle;
    __u32 param;
    __u64 value;
};

struct drm_ljm_gem_timestamp {
    __u32 handle;
    /* inc count, 0 for query current. */
    __u32 inc;
    /* output inc'ed timestamp. */
    __u64 timestamp;
};

/* basic tiling mode. */
#define DRM_VIV_GEM_TILING_LINEAR     0x01
#define DRM_VIV_GEM_TILING_TILED      0x02
#define DRM_VIV_GEM_TILING_SUPERTILED 0x04
#define DRM_VIV_GEM_TILING_MINORTILED 0x08

/* tiling mode modifiers. */
#define DRM_VIV_GEM_TILING_SPLIT    0x10
#define DRM_VIV_GEM_TILING_X_MAJOR  0x20
#define DRM_VIV_GEM_TILING_Y_MAJOR  0x40
#define DRM_VIV_GEM_TILING_SWAP     0x80

/* ts mode. */
#define DRM_VIV_GEM_TS_NONE         0x00
#define DRM_VIV_GEM_TS_DISABLED     0x01
#define DRM_VIV_GEM_TS_NORMAL       0x02
#define DRM_VIV_GEM_TS_COMPRESSED   0x03

/* ts cache mode. */
#define DRM_VIV_GEM_TS_CACHE_MODE_64B     0x00
#define DRM_VIV_GEM_TS_CACHE_MODE_128B    0x01
#define DRM_VIV_GEM_TS_CACHE_MODE_256B    0x02

struct drm_ljm_gem_set_tiling {
    __u32 handle;
    __u32 tiling_mode;

    __u32 ts_mode;
    __u32 ts_cache_mode;
    __u64 clear_value;
};

struct drm_ljm_gem_get_tiling {
    __u32 handle;
    __u32 tiling_mode;

    __u32 ts_mode;
    __u32 ts_cache_mode;
    __u64 clear_value;
};

struct drm_ljm_gem_attach_aux {
    __u32 handle;
    __u32 ts_handle;
};

struct drm_ljm_gem_ref_node {
    __u32 handle;

    /* output. */
    __u32 node;
    __u32 ts_node;
};

struct drm_ljm_gem_deref_node {
    __u32 handle;
    __u32 node;
    __u32 ts_node;
};


struct drm_ljm_gem_lock_gpu {
    __u32 handle;
    __u32 cacheable;
    __u64 physical;
    __u64 gpuBusAddr;
};

struct drm_ljm_gem_unlock_gpu {
    __u32 handle;
};

struct drm_ljm_gpu_info {
    uint32_t infoFlag;
    uint8_t  pdata[64];
};

/*
 * LJM layer&plane ioctl
*/
#define LJM_MAX_LAYERS 4

enum ljm_plane_prop{
    PLANE_PROP_VISIBLE=0,
    PLANE_PROP_IDLE,
    PLANE_PROP_COLOR_DEPTH,
    PLANE_PROP_Z_POS,
    PLANE_PROP_X_POS,
    PLANE_PROP_Y_POS,
    PLANE_PROP_WIDTH,
    PLANE_PROP_HEIGHT,
    PLANE_PROP_ALPHA,
    PLANE_PROP_TRANSPARENT_ENALBE,
    PLANE_PROP_TRANSPARENT,
    PLANE_PROP_VRAM0,
    PLANE_PROP_VRAM1,
    PLANE_PROP_PITCHES,
    PLANE_PROP_RESIZE,
    PLANE_PROP_PORT,
};

struct ljm_plane_property {
    uint32_t plane_id;
    enum ljm_plane_prop name;

    /**
     * @num_values: size of the @values array.
     */
    uint32_t num_values;

    /**
     * @values:
     *
     * Array with limits and values for the property.*/
    uint64_t *values;

};

/*
 * LJM zoomvideo ioctl
*/
#define LJM_MAX_ZVIDEOS 4
enum ljm_zv_prop{
    ZV_TRYLOCK=0,
    ZV_UNLOCK,
    ZV_PROP_ENABLE,
    ZV_PROP_CFG,
    ZV_PROP_BUFF0,
    ZV_PROP_BUFF1,
    ZV_PROP_MAPLAYER,
    ZV_PROP_WHITH,
    ZV_PROP_HEIGHT,
};

struct ljm_zvideo_property {
    uint32_t zv_id;
    enum ljm_zv_prop name;

    /**
     * @num_values: size of the @values array.
     */
    uint32_t num_values;

    /**
     * @values:
     *
     * Array with limits and values for the property.*/
    uint64_t *values;

};

#define LJM_ZOOMVIDEO_DEVICE_NAME "ljmzoomvideo"

#define DRM_VIV_GEM_CREATE          0x00
#define DRM_VIV_GEM_LOCK            0x01
#define DRM_VIV_GEM_UNLOCK          0x02
#define DRM_VIV_GEM_CACHE           0x03
#define DRM_VIV_GEM_QUERY           0x04
#define DRM_VIV_GEM_TIMESTAMP       0x05
#define DRM_VIV_GEM_SET_TILING      0x06
#define DRM_VIV_GEM_GET_TILING      0x07
#define DRM_VIV_GEM_ATTACH_AUX      0x08
#define DRM_VIV_GEM_REF_NODE        0x09
#define DRM_VIV_GEM_RELEASE         0x0A /* design for ttm */
#define DRM_VIV_GEM_WRAP            0x0B /* design for ttm */
#define DRM_VIV_GEM_MOVE_BUFFER     0x0C /* design for ttm */
#define DRM_VIV_GEM_PF              0x0D /* design for ttm */
#define DRM_VIV_GEM_DEREF_NODE      0x0E
#define DRM_VIV_GEM_LOCK_GPU        0x0F
#define DRM_VIV_GEM_UNLOCK_GPU      0x10
#define DRM_VIV_GPU_INFO            0x11
#define LJM_REQUEST_PLANE           0x12
#define LJM_FREE_PLANE              0x13
#define LJM_SET_PROPPLANE           0x14
#define LJM_GET_PROPPLANE           0x15
#define LJM_SET_PROPZVIDEO          0x16
#define LJM_GET_PROPZVIDEO          0x17
#define DRM_VIV_NUM_IOCTLS          0x18

#define DRM_IOCTL_VIV_GEM_CREATE        DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_CREATE, struct drm_ljm_gem_create)
#define DRM_IOCTL_VIV_GEM_LOCK          DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_LOCK, struct drm_ljm_gem_lock)
#define DRM_IOCTL_VIV_GEM_UNLOCK        DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_UNLOCK, struct drm_ljm_gem_unlock)
#define DRM_IOCTL_VIV_GEM_CACHE         DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_CACHE, struct drm_ljm_gem_cache)
#define DRM_IOCTL_VIV_GEM_QUERY         DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_QUERY, struct drm_ljm_gem_query)
#define DRM_IOCTL_VIV_GEM_TIMESTAMP     DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_TIMESTAMP, struct drm_ljm_gem_timestamp)
#define DRM_IOCTL_VIV_GEM_SET_TILING    DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_SET_TILING, struct drm_ljm_gem_set_tiling)
#define DRM_IOCTL_VIV_GEM_GET_TILING    DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_GET_TILING, struct drm_ljm_gem_get_tiling)
#define DRM_IOCTL_VIV_GEM_ATTACH_AUX    DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_ATTACH_AUX, struct drm_ljm_gem_attach_aux)
#define DRM_IOCTL_VIV_GEM_REF_NODE      DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_REF_NODE, struct drm_ljm_gem_ref_node)
#if ljdENABLE_TTM
#define DRM_IOCTL_VIV_GEM_RELEASE       DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_RELEASE, struct drm_ljm_gem_release)
#define DRM_IOCTL_VIV_GEM_WRAP          DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_WRAP, struct drm_ljm_gem_wrap)
#define DRM_IOCTL_VIV_GEM_MOVE_BUFFER   DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_MOVE_BUFFER,struct drm_ljm_gem_move)
#define DRM_IOCTL_VIV_GEM_PF            DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_PF, struct drm_ljm_gem_pf)
#endif
#define DRM_IOCTL_VIV_GEM_DEREF_NODE    DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_DEREF_NODE, struct drm_ljm_gem_deref_node)
#define DRM_IOCTL_VIV_GEM_LOCK_GPU      DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_LOCK_GPU, struct drm_ljm_gem_lock_gpu)
#define DRM_IOCTL_VIV_GEM_UNLOCK_GPU    DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GEM_UNLOCK_GPU, struct drm_ljm_gem_unlock_gpu)
#define DRM_IOCTL_VIV_GPU_INFO          DRM_IOWR(DRM_COMMAND_BASE + DRM_VIV_GPU_INFO, struct drm_ljm_gpu_info)
#define DRM_IOCTL_VIV_PLANE_REQUEST     DRM_IOWR(DRM_COMMAND_BASE + LJM_REQUEST_PLANE, int)
#define DRM_IOCTL_VIV_PLANE_FREE        DRM_IOWR(DRM_COMMAND_BASE + LJM_FREE_PLANE, int)
#define DRM_IOCTL_VIV_PROPPLANE_SET     DRM_IOWR(DRM_COMMAND_BASE + LJM_SET_PROPPLANE, struct ljm_plane_property)
#define DRM_IOCTL_VIV_PROPPLANE_GET     DRM_IOWR(DRM_COMMAND_BASE + LJM_GET_PROPPLANE, struct ljm_plane_property)
#define DRM_IOCTL_VIV_PROPZVIDEO_SET    DRM_IOWR(DRM_COMMAND_BASE + LJM_SET_PROPZVIDEO, struct ljm_zvideo_property)
#define DRM_IOCTL_VIV_PROPZVIDEO_GET    DRM_IOWR(DRM_COMMAND_BASE + LJM_GET_PROPZVIDEO, struct ljm_zvideo_property)

#ifdef __KERNEL__
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
#  define drm_gem_object_unreference_unlocked drm_gem_object_put
#  define drm_dev_unref                       drm_dev_put
# elif LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)
#  define drm_gem_object_unreference_unlocked drm_gem_object_put_unlocked
#  define drm_dev_unref                       drm_dev_put
# endif
#endif

#if defined(__cplusplus)
}
#endif

#endif /* __lj_hal_drm_h_ */


