/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_dump_h_
#define __lj_hal_dump_h_

/*
 *  ljdDUMP_KEY
 *
 *      Set this to a string that appears in 'cat /proc/<pid>/cmdline'. E.g. 'camera'.
 *      HAL will create dumps for the processes matching this key.
 */
#ifndef ljdDUMP_KEY
#define ljdDUMP_KEY             "process"
#endif

/*
 *  ljdDUMP_PATH
 *
 *      The dump file location. Some processes cannot write to the sdcard.
 *      Try apps' data dir, e.g. /data/data/com.android.launcher
 */
#ifndef ljdDUMP_PATH
#if defined(ANDROID)
#  define ljdDUMP_PATH        "/mnt/sdcard/"
# else
#  define ljdDUMP_PATH        "./"
# endif
#endif

/*
 *  ljdDUMP_FILE_IN_KERNEL
 *
 *      Default dump file for ljdDUMP_IN_KERNEL.
 *      The file will be writen globally in kernel side.
 *      Can be overwritten in runtime by debugfs:/lj/dump/dump_file
 *
 *      2 pseudo files:
 *      [dmesg]:    means dump to kernel debug message.
 *      [ignored]:  means dump ignored, nothing will be dumpped.
 */
#ifndef ljdDUMP_FILE_IN_KERNEL
#define ljdDUMP_FILE_IN_KERNEL  "[dmesg]"
#endif

/*
 *  ljdDUMP_VERIFY_PER_DRAW
 *
 *      Sub feature of ljdDUMP.
 *      When set to 1, verify RT and images(if used) for every single draw
 *      to ease simulation debug.
 *      Only valid for ES3 driver for now.
 */
#ifndef ljdDUMP_VERIFY_PER_DRAW
#define ljdDUMP_VERIFY_PER_DRAW     0
#endif

/* Standalone dump features below. */

/*
 *  ljdDUMP_FRAMERATE
 *      When set to a value other than zero, averaqe frame rate will be dumped.
 *      The value set is the starting frame that the average will be calculated.
 *      This is needed because sometimes first few frames are too slow to be included
 *      in the average. Frame count starts from 1.
 */
#ifndef ljdDUMP_FRAMERATE
#define ljdDUMP_FRAMERATE           0
#endif

/*
 *  ljdDUMP_FRAME_TGA
 *
 *      When set to a value other than 0, a dump of the frame specified by the value,
 *      will be done into frame.tga. Frame count starts from 1.
 */
#ifndef ljdDUMP_FRAME_TGA
#define ljdDUMP_FRAME_TGA           0
#endif

/*
 *  ljdDUMP_AHB_ACCESS
 *
 *      When set to 1, a dump of all AHB register access will be printed to kernel
 *      message.
 */
#ifndef ljdDUMP_AHB_ACCESS
#define ljdDUMP_AHB_ACCESS          0
#endif

#endif /* __lj_hal_dump_h_ */



