/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_allocator_h_
#define __lj_hal_kernel_allocator_h_

#include "lj_hal_kernel_linux.h"
#include <linux/slab.h>
#include <linux/mm_types.h>

typedef struct _ljsALLOCATOR  *ljkALLOCATOR;
typedef union _ljsATTACH_DESC *ljsATTACH_DESC_PTR;

typedef struct _ljsALLOCATOR_OPERATIONS {
    /**************************************************************************
     **
     ** Alloc
     **
     ** Allocate memory, request size is page aligned.
     **
     ** INPUT:
     **
     **    ljkALLOCATOR Allocator
     **        Pointer to an ljkALLOCATOER object.
     **
     **    PLINUX_Mdl
     **        Pointer to Mdl whichs stores information
     **        about allocated memory.
     **
     **    ljtSIZE_T NumPages
     **        Number of pages need to allocate.
     **
     **    ljtUINT32 Flag
     **        Allocation option.
     **
     ** OUTPUT:
     **
     **      Nothing.
     **
     */
    ljeSTATUS
    (*Alloc)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T NumPages, ljtUINT32 Flag);

    /**************************************************************************
     **
     ** Free
     **
     ** Free memory.
     **
     ** INPUT:
     **
     **     ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **     PLINUX_MDL Mdl
     **          Mdl which stores information.
     **
     ** OUTPUT:
     **
     **      Nothing.
     **
     */
    void (*Free)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl);

    /**************************************************************************
     **
     ** Mmap
     **
     ** Map a page range of the memory to user space.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      PLINUX_MDL Mdl
     **          Pointer to a Mdl.
     **
     **      ljtSIZE_T skipPages
     **          Number of page to be skipped from beginning of this memory.
     **
     **      ljtSIZE_T numPages
     **          Number of pages to be mapped from skipPages.
     **
     ** INOUT:
     **
     **      struct vm_area_struct *vma
     **          Pointer to VMM memory area.
     **
     */
    ljeSTATUS
    (*Mmap)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtBOOL Cacheable,
            ljtSIZE_T skipPages, ljtSIZE_T numPages, struct vm_area_struct *vma);

    /**************************************************************************
     **
     ** MapUser
     **
     ** Map memory to user space.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      PLINUX_MDL Mdl
     **          Pointer to a Mdl.
     **
     **      ljtBOOL Cacheable
     **          Whether this mapping is cacheable.
     **
     ** OUTPUT:
     **
     **      ljtPOINTER *UserLogical
     **          Pointer to user logical address.
     **
     **      Nothing.
     **
     */
    ljeSTATUS
    (*MapUser)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
               PLINUX_MDL_MAP MdlMap, ljtBOOL Cacheable);

    /**************************************************************************
     **
     ** UnmapUser
     **
     ** Unmap address from user address space.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      ljtPOINTER Logical
     **          Address to be unmap
     **
     **      ljtUINT32 Size
     **          Size of address space
     **
     ** OUTPUT:
     **
     **      Nothing.
     **
     */
    void
    (*UnmapUser)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtUINT32 Size);

    /**************************************************************************
     **
     ** MapKernel
     **
     ** Map memory to kernel space.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      PLINUX_MDL Mdl
     **          Pointer to a Mdl object.
     **
     ** OUTPUT:
     **      ljtPOINTER *Logical
     **          Mapped kernel address.
     */
    ljeSTATUS
    (*MapKernel)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                 ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER * Logical);

    /**************************************************************************
     **
     ** UnmapKernel
     **
     ** Unmap memory from kernel space.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      PLINUX_MDL Mdl
     **          Pointer to a Mdl object.
     **
     **      ljtPOINTER Logical
     **          Mapped kernel address.
     **
     ** OUTPUT:
     **
     **      Nothing.
     **
     */
    ljeSTATUS
    (*UnmapKernel)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtPOINTER Logical);

    /**************************************************************************
     **
     ** Cache
     **
     ** Maintain cache coherency.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      PLINUX_MDL Mdl
     **          Pointer to a Mdl object.
     **
     **      ljtPOINTER Logical
     **          Logical address, could be user address or kernel address
     **
     **      ljtSIZE_T Offset
     **          Physical address.
     **
     **      ljtUINT32 Bytes
     **          Size of memory region.
     **
     **      ljeCACHEOPERATION Opertaion
     **          Cache operation.
     **
     ** OUTPUT:
     **
     **      Nothing.
     **
     */
    ljeSTATUS
    (*Cache)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T Offset,
             ljtPOINTER Logical, ljtSIZE_T Bytes, ljeCACHEOPERATION Operation);

    /**************************************************************************
     **
     ** Physical
     **
     ** Get physical address from a offset in memory region.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      PLINUX_MDL Mdl
     **          Pointer to a Mdl object.
     **
     **      unsigned long Offset
     **          Offset in this memory region.
     **
     ** OUTPUT:
     **      ljtUINT32_PTR Physical
     **          Physical address.
     **
     */
    ljeSTATUS
    (*Physical)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, unsigned long Offset, ljtPHYS_ADDR_T * Physical);

    /**************************************************************************
     **
     ** Attach
     **
     ** Import memory allocated by an external allocator.
     **
     ** INPUT:
     **      ljkALLOCATOR Allocator
     **          Pointer to an ljkALLOCATOER object.
     **
     **      ljtUINT32 Handle
     **          Handle of the memory.
     **
     ** OUTPUT:
     **      None.
     **
     */
    ljeSTATUS
    (*Attach)(ljkALLOCATOR Allocator, ljsATTACH_DESC_PTR Desc, PLINUX_MDL Mdl);

    /**************************************************************************
     *
     * GetSGT
     *
     * Get scatter-gather table from a range of the memory.
     *
     * INPUT:
     *      ljkALLOCATOR Allocator
     *          Pointer to an ljkALLOCATOER object.
     *
     *      ljtUINT32 Handle
     *          The Handle of the memory.
     *
     *      ljtSIZE_T Offset
     *          The offset to the beginning of this mdl.
     *
     *      ljtSIZE_T Bytes
     *          Total bytes form Offset.
     *
     * OUTPUT:
     *      ljtPOINTER *SGT
     *          scatter-gather table
     *
     */
    ljeSTATUS
    (*GetSGT)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
              ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER * SGT);

    /**************************************************************************
     *
     * GetDmaAddress
     *
     * Get DMA address of the memory.
     *
     * INPUT:
     *      PLINUX_MDL Mdl
     *          Pointer to a Mdl object.
     *
     * OUTPUT:
     *      ljtPHYS_ADDR_T
     *          DMAaddress
     *
     */
    ljeSTATUS
    (*GetDmaAddress)(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, unsigned long Offset, ljtPHYS_ADDR_T * DmaAddress);
} ljsALLOCATOR_OPERATIONS;

typedef struct _ljsALLOCATOR {
    /* Pointer to ljkOS Object. */
    ljkOS                   os;

    /* Name. */
    ljtSTRING               name;

    /* Operations. */
    ljsALLOCATOR_OPERATIONS *ops;

    /* Capability of this allocator. */
    ljtUINT32               capability;

    /* Debugfs entry of this allocator. */
    ljsDEBUGFS_DIR          debugfsDir;

    /* Private data used by customer allocator. */
    void                    *privateData;

    /* Allocator destructor. */
    void                    (*destructor)(struct _ljsALLOCATOR *Allocator);

    struct list_head        link;
#if ljdANON_FILE_FOR_ALLOCATOR
    /*Anonymous file for map to user. */
    struct file             *anon_file;
#endif
} ljsALLOCATOR;

typedef struct _ljsALLOCATOR_DESC {
    /* Name of a allocator. */
    char    *name;

    /* Entry function to construct a allocator. */
    ljeSTATUS
    (*construct)(ljkOS, ljsDEBUGFS_DIR *, ljkALLOCATOR *);
} ljsALLOCATOR_DESC;

typedef union _ljsATTACH_DESC {
    /* ljvALLOC_FLAG_DMABUF */
    struct {
        ljtPOINTER          dmabuf;
    } dmaBuf;

    /* ljvALLOC_FLAG_USERMEMORY */
    struct {
        ljtPOINTER          memory;
        ljtPHYS_ADDR_T      physical;
        ljtSIZE_T           size;
    } userMem;

    /* Reserved memory. */
    struct {
        unsigned long       start;
        unsigned long       size;
        const char          *name;
        int                 requested;
        ljtBOOL             root;
    } reservedMem;
} ljsATTACH_DESC;

/*
 * Helpers
 */

/* Fill a ljsALLOCATOR_DESC structure. */
#define ljmkDEFINE_ALLOCATOR_DESC(Name, Construct)  \
    {                                               \
        .name = Name,                               \
        .construct = Construct,                     \
    }

/* Construct a allocator. */
static inline ljeSTATUS
ljkALLOCATOR_Construct(ljkOS Os,
                       ljsALLOCATOR_OPERATIONS *Operations,
                       ljkALLOCATOR *Allocator)
{
    ljeSTATUS status;
    ljkALLOCATOR allocator;

    ljmkASSERT(Allocator != ljvNULL);
    ljmkASSERT(Operations &&
               (Operations->Alloc || Operations->Attach) &&
               (Operations->Free) &&
               Operations->MapUser &&
               Operations->UnmapUser &&
               Operations->MapKernel &&
               Operations->UnmapKernel &&
               Operations->Cache &&
               Operations->Physical);

    allocator = kzalloc(sizeof(ljsALLOCATOR), GFP_KERNEL | ljdNOWARN);
    if (unlikely(!allocator))
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Record os. */
    allocator->os = Os;

    /* Set operations. */
    allocator->ops = Operations;

    *Allocator = allocator;

    return ljvSTATUS_OK;

OnError:
    return status;
}

#if !ljdUSE_LINUX_SG_TABLE_API
int
alloc_sg_list_from_pages(struct scatterlist **sgl, struct page **pages, unsigned int n_pages,
                         unsigned long offset, unsigned long size, unsigned int *nents);
#endif

/*
 *   How to implement customer allocator
 *
 *   Build in customer alloctor
 *
 *       It is recommended that customer allocator is implmented in independent
 *       source file(s) which is specified by CUSOMTER_ALLOCATOR_OBJS in Kbuld.
 *
 *   Register ljsALLOCATOR
 *
 *       For each customer specified allocator, a desciption entry must be added
 *       to allocatorArray defined in lj_hal_kernel_allocator_array.h.
 *
 *       An entry in allocatorArray is a ljsALLOCATOR_DESC structure which describes
 *       name and constructor of a ljkALLOCATOR object.
 *
 *   Implement ljsALLOCATOR_DESC.init()
 *
 *       In ljsALLOCATOR_DESC.init(), ljkALLOCATOR_Construct should be called
 *       to create a ljkALLOCATOR object, customer specified private data can
 *       be put in ljsALLOCATOR.privateData.
 *
 *   Implement ljsALLOCATOR_OPERATIONS
 *
 *       When call ljkALLOCATOR_Construct to create a ljkALLOCATOR object, a
 *       ljsALLOCATOR_OPERATIONS structure must be provided whose all members
 *       implemented.
 *
 */
#endif /* __lj_hal_kernel_allocator_h_ */
