/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_debug_h_
#define __lj_hal_kernel_debug_h_

#include "lj_hal_kernel_linux.h"
#include <linux/spinlock.h>
#include <linux/time.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 16, 0)
# include <linux/stdarg.h>
#else
# include <stdarg.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
# include <linux/nmi.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 ***************************** OS-dependent Macros *****************************
 *******************************************************************************/

typedef va_list ljtARGUMENTS;

#define ljmkARGUMENTS_START(Arguments, Pointer)      \
    va_start(Arguments, Pointer)

#define ljmkARGUMENTS_END(Arguments)                 \
    va_end(Arguments)

#define ljmkARGUMENTS_ARG(Arguments, Type)           \
    va_arg(Arguments, Type)

#define ljmkDECLARE_MUTEX(__mutex__)                 \
    /* Declare mutex */                              \
    DEFINE_MUTEX((__mutex__))

#define ljmkMUTEX_LOCK(__mutex__)                    \
    mutex_lock(&(__mutex__))

#define ljmkMUTEX_UNLOCK(__mutex__)                  \
    mutex_unlock(&(__mutex__))

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
# define ljmkGETPROCESSID()                          \
    task_tgid_nr(current)
#else
# define ljmkGETPROCESSID()                          \
    current->tgid
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
# define ljmkGETTHREADID()                           \
    task_pid_vnr(current)
#else
# define ljmkGETTHREADID()                           \
    current->pid
#endif

#define ljmkOUTPUT_STRING(String)                    \
    pr_err("%s", String)

#if ljdDUMP_IN_KERNEL
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
#  define ljmkDUMP_STRING(Os, String) \
    do { \
        mutex_lock(&Os->dumpFilpMutex); \
        if (Os->dumpTarget == 0) \
            pr_warn("%s", String); \
        else if (Os->dumpFilp && Os->dumpTarget == 1) \
            kernel_write(Os->dumpFilp, String, strlen(String), &Os->dumpFilp->f_pos); \
        mutex_unlock(&Os->dumpFilpMutex); \
    } while (0)
# else
#  define ljmkDUMP_STRING(Os, String) \
    do { \
        mutex_lock(&Os->dumpFilpMutex); \
        if (Os->dumpTarget == 0) { \
            pr_warn("%s", String); \
        } else if (Os->dumpFilp && Os->dumpTarget == 1) { \
            mm_segment_t oldFs; \
            oldFs = get_fs(); \
            set_fs(KERNEL_DS); \
            vfs_write(Os->dumpFilp, String, strlen(String), &Os->dumpFilp->f_pos); \
            set_fs(oldFs); \
        } \
        mutex_unlock(&Os->dumpFilpMutex); \
    } while (0)
# endif
#endif

#define ljmkSPRINTF(Destination, Size, ...)                  \
    snprintf(Destination, Size, __VA_ARGS__)

#define ljmkVSPRINTF(Destination, Size, Message, Arguments)  \
    vsnprintf(Destination, Size, Message, *((va_list *)Arguments))

#define ljmkSTRCATSAFE(Destination, Size, String)            \
    strncat(Destination, String, (Size) - 1)

#define ljmkMEMCPY(Destination, Source, Size)                \
    memcpy(Destination, Source, Size)

#define ljmkSTRLEN(String)                                   \
    strlen(String)

/*
 * If not zero, forces data alignment in the variable argument list
 * by its individual size.
 */
#define ljdALIGNBYSIZE 1

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_kernel_debug_h_ */
