/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifdef MODULE
# include <linux/module.h>
#endif
#include <linux/init.h>
#include <linux/debugfs.h>
#include <linux/slab.h>
#ifdef MODVERSIONS
# include <linux/modversions.h>
#endif
#include <linux/stddef.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/vmalloc.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/uaccess.h>
#include <linux/completion.h>
#include <linux/seq_file.h>
#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel.h"
#include "lj_hal_kernel_debug.h"

#define _LJ_OBJ_ZONE ljvZONE_KERNEL

static int
lj_debugfs_open(struct inode *inode, struct file *file)
{
    ljsINFO_NODE *node = inode->i_private;

    return single_open(file, node->info->show, node);
}

static ssize_t
lj_debugfs_write(struct file *file, const char __user *buf, size_t count, loff_t *pos)
{
    struct seq_file *s = file->private_data;
    ljsINFO_NODE *node = s->private;
    ljsINFO *info = node->info;

    if (info->write)
        info->write(buf, count, node);

    return count;
}

static const struct file_operations lj_debugfs_operations = {
    .owner   = THIS_MODULE,
    .open    = lj_debugfs_open,
    .write   = lj_debugfs_write,
    .read    = seq_read,
    .llseek  = seq_lseek,
    .release = single_release,
};

ljeSTATUS
ljkDEBUGFS_DIR_Init(ljkDEBUGFS_DIR Dir, struct dentry *root, ljtCONST_STRING Name)
{
    Dir->root = debugfs_create_dir(Name, root);

    if (!Dir->root)
        return ljvSTATUS_NOT_SUPPORTED;

    INIT_LIST_HEAD(&Dir->nodeList);

    return ljvSTATUS_OK;
}

ljeSTATUS
ljkDEBUGFS_DIR_CreateFiles(ljkDEBUGFS_DIR Dir, ljsINFO *List, int count, ljtPOINTER Data)
{
    int i;
    ljsINFO_NODE *node;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Dir=%p List=%p count=%d Data=%p", Dir, List, count, Data);

    for (i = 0; i < count; i++) {
        umode_t mode = 0;

        /* Create a node. */
        node = kzalloc(sizeof(ljsINFO_NODE), GFP_KERNEL);

        if (!node)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        node->info   = &List[i];
        node->device = Data;

        mode |= List[i].show ? 0444 : 0;
        mode |= List[i].write ? 0200 : 0;

        node->entry = debugfs_create_file(List[i].name, mode,
                                          Dir->root, node, &lj_debugfs_operations);

        if (!node->entry)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        list_add(&node->head, &Dir->nodeList);
    }

OnError:
    if (ljmIS_ERROR(status))
        ljmkVERIFY_OK(ljkDEBUGFS_DIR_RemoveFiles(Dir, List, count));

    ljmkFOOTER();
    return status;
}

ljeSTATUS
ljkDEBUGFS_DIR_RemoveFiles(ljkDEBUGFS_DIR Dir, ljsINFO *List, int count)
{
    int i;
    ljsINFO_NODE *node;
    ljsINFO_NODE *temp;

    ljmkHEADER_ARG("Dir=%p List=%p count=%d", Dir, List, count);

    for (i = 0; i < count; i++) {
        list_for_each_entry_safe(node, temp, &Dir->nodeList, head) {
            if (node->info == &List[i]) {
                debugfs_remove(node->entry);
                list_del(&node->head);
                kfree(node);
            }
        }
    }

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

void
ljkDEBUGFS_DIR_Deinit(ljkDEBUGFS_DIR Dir)
{
    debugfs_remove(Dir->root);
    Dir->root = NULL;
}
