/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_allocator.h"
#include "ljmicro_family.h"
#include "ljmicro_ioctl.h"
#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/io.h>

#define _LJ_OBJ_ZONE    ljvZONE_DEVICE

static ljkGALDEVICE     galDevice;

ljkGALDEVICE get_galdevice(void)
{
    return galDevice;
}

#ifdef CONFIG_DEBUG_FS
#if defined(CONFIG_CPU_CSKYV2) && LINUX_VERSION_CODE <= KERNEL_VERSION(3, 0, 8)
static void
seq_vprintf(struct seq_file *m, const char *f, va_list args)
{
    int len;

    if (m->count < m->size) {
        len = vsnprintf(m->buf + m->count, m->size - m->count, f, args);
        if (m->count + len < m->size) {
            m->count += len;
            return;
        }
    }
    m->count = m->size;
}
# endif

static int
debugfs_printf(void *obj, const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    seq_vprintf((struct seq_file *)obj, fmt, args);
    va_end(args);

    return 0;
}
#else
static int
sys_printf(void *obj, const char *fmt, ...)
{
    int len = 0;
    va_list args;

    va_start(args, fmt);
    len = vsprintf((char *)obj, fmt, args);
    va_end(args);

    return len;
}
#endif

#ifdef CONFIG_DEBUG_FS
# define fs_printf debugfs_printf
#else
# define fs_printf sys_printf
#endif

/*******************************************************************************
 ******************************* Debugfs Support *******************************
 *******************************************************************************/

/*******************************************************************************
 **************************** DEBUG SHOW FUNCTIONS *****************************
 *******************************************************************************/

int
lj_info_show(void *m, void *data)
{
    ljkGALDEVICE gal_device = galDevice;
    ljkDEVICE device = ljvNULL;
    int i = 0;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif
    ljeCHIPMODEL chipModel = 0;
    ljtUINT32 chipRevision = 0;
    ljtUINT32 productID = 0;
    ljtUINT32 ecoID = 0;

    if (!gal_device)
        return -ENXIO;

    device = gal_device->devices[0];

    for (i = 0; i < ljdMAX_GPU_COUNT; i++) {
        if (device->kernels[i]) {
            if (i == ljvCORE_VG) {
#if ljdENABLE_VG
                chipModel = device->kernels[i]->vg->hardware->chipModel;
                chipRevision = device->kernels[i]->vg->hardware->chipRevision;
#endif
            } else {
                chipModel = device->kernels[i]->hardware->identity.chipModel;
                chipRevision = device->kernels[i]->hardware->identity.chipRevision;
                productID = device->kernels[i]->hardware->identity.productID;
                ecoID = device->kernels[i]->hardware->identity.ecoID;
            }

            len = fs_printf(ptr, "gpu      : %d\n", i);
            len += fs_printf(ptr + len, "model    : %4x\n", chipModel);
            len += fs_printf(ptr + len, "revision : %4x\n", chipRevision);
            len += fs_printf(ptr + len, "product  : %4x\n", productID);
            len += fs_printf(ptr + len, "eco      : %4x\n", ecoID);
            len += fs_printf(ptr + len, "\n");
        }
    }
    return len;
}

int
lj_clients_show(void *m, void *data)
{
    ljkGALDEVICE device = galDevice;

    ljkKERNEL kernel = _GetValidKernel(0, device);

    ljsDATABASE_PTR database;
    ljeSTATUS status;
    ljtINT i, pid;
    char name[24];
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!kernel)
        return -ENXIO;

    len = fs_printf(ptr, "%-8s%s\n", "PID", "NAME");
    len += fs_printf(ptr + len, "------------------------\n");

    /* Acquire the database mutex. */
    ljmkONERROR(ljkOS_AcquireMutex(kernel->os, kernel->db->dbMutex, ljvINFINITE));

    /* Walk the databases. */
    for (i = 0; i < ljmCOUNTOF(kernel->db->db); ++i) {
        for (database = kernel->db->db[i];
             database != ljvNULL;
             database = database->next) {
            pid = database->processID;

            ljmkVERIFY_OK(ljkOS_GetProcessNameByPid(pid, ljmSIZEOF(name), name));

            len += fs_printf(ptr + len, "%-8d%s\n", pid, name);
        }
    }

    /* Release the database mutex. */
    ljmkVERIFY_OK(ljkOS_ReleaseMutex(kernel->os, kernel->db->dbMutex));

    /* Success. */
    return len;

OnError:
    if (status == ljvSTATUS_INVALID_ARGUMENT)
        return -EINVAL;
    return status;
}

int
lj_meminfo_show(void *m, void *data)
{
    ljkGALDEVICE device = galDevice;
    ljkKERNEL kernel = _GetValidKernel(0, device);
    ljkVIDMEM memory;
    ljeSTATUS status;
    ljsDATABASE_PTR database;
    ljtUINT32 i;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    ljtUINT64 free = 0, used = 0, total = 0, minFree = 0, maxUsed = 0;

    ljsDATABASE_COUNTERS virtualCounter  = { 0, 0, 0 };
    ljsDATABASE_COUNTERS nonPagedCounter = { 0, 0, 0 };

    if (!kernel)
        return -ENXIO;

    len = fs_printf(ptr, "VIDEO MEMORY:\n");

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++)
    {
        if (!kernel->device->contiguousSizes[i])
            break;

        kernel->device->memIndex = i;

        status = ljkKERNEL_GetVideoMemoryPool(kernel, ljvPOOL_SYSTEM, &memory);

        if (ljmIS_SUCCESS(status)) {
            ljmkONERROR(ljkOS_AcquireMutex(memory->os, memory->mutex, ljvINFINITE));

            free = memory->freeBytes;
            minFree = memory->minFreeBytes;
            used = memory->bytes - memory->freeBytes;
            maxUsed = memory->bytes - memory->minFreeBytes;
            total = memory->bytes;

            ljmkVERIFY_OK(ljkOS_ReleaseMutex(memory->os, memory->mutex));
        }

        status = ljkKERNEL_GetVideoMemoryPool(kernel, ljvPOOL_SYSTEM_32BIT_VA, &memory);

        if (!i && ljmIS_SUCCESS(status)) {
            ljmkONERROR(ljkOS_AcquireMutex(memory->os, memory->mutex, ljvINFINITE));

            free += memory->freeBytes;
            minFree += memory->minFreeBytes;
            used += memory->bytes - memory->freeBytes;
            maxUsed += memory->bytes - memory->minFreeBytes;
            total += memory->bytes;

            ljmkVERIFY_OK(ljkOS_ReleaseMutex(memory->os, memory->mutex));
        }

        len += fs_printf(ptr + len, "  POOL SYSTEM%d:\n", i);
        len += fs_printf(ptr + len, "    Free :    %13llu B\n", free);
        len += fs_printf(ptr + len, "    Used :    %13llu B\n", used);
        len += fs_printf(ptr + len, "    MinFree : %13llu B\n", minFree);
        len += fs_printf(ptr + len, "    MaxUsed : %13llu B\n", maxUsed);
        len += fs_printf(ptr + len, "    Total :   %13llu B\n", total);
    }

    kernel->device->memIndex = 0;

    status = ljkKERNEL_GetVideoMemoryPool(kernel, ljvPOOL_LOCAL_EXTERNAL, &memory);

    if (ljmIS_SUCCESS(status))
    {
        ljmkVERIFY_OK(
            ljkOS_AcquireMutex(memory->os, memory->mutex, ljvINFINITE));

        free    = memory->freeBytes;
        minFree = memory->minFreeBytes;
        used    = memory->bytes - memory->freeBytes;
        maxUsed = memory->bytes - memory->minFreeBytes;
        total   = memory->bytes;

        ljmkVERIFY_OK(ljkOS_ReleaseMutex(memory->os, memory->mutex));
    }

    //len = fs_printf(ptr, "VIDEO MEMORY:\n");
    len += fs_printf(ptr + len, "  POOL EXTERNAL:\n");
    len += fs_printf(ptr + len, "    Free :    %10llu B\n", free);
    len += fs_printf(ptr + len, "    Used :    %10llu B\n", used);
    len += fs_printf(ptr + len, "    MinFree : %10llu B\n", minFree);
    len += fs_printf(ptr + len, "    MaxUsed : %10llu B\n", maxUsed);
    len += fs_printf(ptr + len, "    Total :   %10llu B\n", total);

#if 1
    status = ljkKERNEL_GetVideoMemoryPool(kernel, ljvPOOL_LOCAL_EXCLUSIVE, &memory);

    if (ljmIS_SUCCESS(status))
    {
        ljmkVERIFY_OK(
            ljkOS_AcquireMutex(memory->os, memory->mutex, ljvINFINITE));

        free    = memory->freeBytes;
        minFree = memory->minFreeBytes;
        used    = memory->bytes - memory->freeBytes;
        maxUsed = memory->bytes - memory->minFreeBytes;
        total   = memory->bytes;

        ljmkVERIFY_OK(ljkOS_ReleaseMutex(memory->os, memory->mutex));
    }

    //len  = fs_printf(ptr, "VIDEO MEMORY:\n");
    len += fs_printf(ptr + len, "  POOL EXCLUSIVE:\n");
    len += fs_printf(ptr + len, "    Free :    %10u B\n", free);
    len += fs_printf(ptr + len, "    Used :    %10u B\n", used);
    len += fs_printf(ptr + len, "    MinFree : %10u B\n", minFree);
    len += fs_printf(ptr + len, "    MaxUsed : %10u B\n", maxUsed);
    len += fs_printf(ptr + len, "    Total :   %10u B\n", total);
#endif

    /* Acquire the database mutex. */
    ljmkONERROR(ljkOS_AcquireMutex(kernel->os, kernel->db->dbMutex, ljvINFINITE));

    /* Walk the databases. */
    for (i = 0; i < ljmCOUNTOF(kernel->db->db); ++i) {
        for (database = kernel->db->db[i];
             database != ljvNULL;
             database = database->next) {
            ljsDATABASE_COUNTERS *counter;

            counter = &database->vidMemPool[ljvPOOL_VIRTUAL];
            virtualCounter.bytes += counter->bytes;
            virtualCounter.maxBytes += counter->maxBytes;

            counter = &database->nonPaged;
            nonPagedCounter.bytes += counter->bytes;
            nonPagedCounter.bytes += counter->maxBytes;
        }
    }

    /* Release the database mutex. */
    ljmkVERIFY_OK(ljkOS_ReleaseMutex(kernel->os, kernel->db->dbMutex));

    len += fs_printf(ptr + len, "  POOL VIRTUAL:\n");
    len += fs_printf(ptr + len, "    Used :    %10llu B\n", virtualCounter.bytes);
    len += fs_printf(ptr + len, "    MaxUsed : %10llu B\n", virtualCounter.bytes);

    return len;

OnError:
    if (status == ljvSTATUS_INVALID_ARGUMENT)
        return -EINVAL;
    return status;
}

ljeSTATUS
lj_mmuinfo_show(void)
{
    ljkGALDEVICE device = galDevice;
    ljkKERNEL kernel = _GetValidKernel(0, device);
    ljkMMU mmu = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtUINT32 mStart = 0;
    ljtUINT32 mEnd = ljdMMU_MTLB_ENTRY_NUM;
    ljtUINT32 sStart = 0;
    ljtUINT32 sEnd = 0;

    ljtUINT32 mtlbEntry = 0;
    ljtUINT32 stlbEntry = 0;

    ljtUINT32 stlbSize = 0;
    ljtUINT32 stlbEntryNum = 0;

    ljtUINT32_PTR stlbLogical = ljvNULL;
    ljtPHYS_ADDR_T stlbPhysical = 0;
    ljtUINT32 extAddr = 0;

    ljtUINT32 i;

    if (!kernel)
        ljmkONERROR(ljvSTATUS_INVALID_OBJECT);

    mmu = kernel->mmu;
    if (!mmu)
        ljmkONERROR(ljvSTATUS_INVALID_OBJECT);

    ljmkPRINT("MMU PAGE TABLE:\n");

    ljmkPRINT("  BASE ADDRESS:\n");
    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++)
        ljmkPRINT("    contiguousBaseAddresses[%d]: 0x%010X\n", i, mmu->contiguousBaseAddresses[i]);

    ljmkPRINT("    externalBaseAddress:        0x%010X\n", mmu->externalBaseAddress);
    ljmkPRINT("    internalBaseAddress:        0x%010X\n", mmu->internalBaseAddress);
    ljmkPRINT("    exclusiveBaseAddress:       0x%010X\n", mmu->exclusiveBaseAddress);

    /* get extAddress */
    extAddr = (ljtUINT32)(mmu->mtlbPhysical >> 32);
    /* more than 40bit physical address */
    if (extAddr & 0xFFFFFF00)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    ljmkPRINT("  Extensional 8 bit for 40bit address: 0x%02X\n", extAddr);

    ljmkPRINT("  MTLB ENTRY:\n");

    /* first mtlb entry is reserved for SRAM */
    ljmkPRINT("    MTLB entry[0] is reserved for SRAM.\n");
    ++mStart;

    while (mStart < mEnd) {
        mtlbEntry = *(mmu->mtlbLogical + mStart);
        if (mtlbEntry) {
            ljtUINT64 dataMtlb = mtlbEntry & ~0X1F;
            ljtUINT8 maskMtlb = mtlbEntry & 0X1F;

            ljmkASSERT(dataMtlb);

            /* stlb physical address */
            stlbPhysical = ((ljtUINT64)extAddr << 32) | (ljtUINT64)dataMtlb;

            /* get stlb entry info */
            switch ((mtlbEntry & 0X0C) >> 2) {
            case 0:
                sEnd = ljdMMU_STLB_4K_ENTRY_NUM;
                stlbSize = ljdMMU_STLB_4K_SIZE;
                ljmkPRINT("    MTLB entry[%d]:   STLB physical = 0x%010lX, PAGE_SIZE = 4K, MASK = 0x%02X\n", mStart, stlbPhysical, maskMtlb);
                break;
            case 1:
                sEnd = ljdMMU_STLB_64K_ENTRY_NUM;
                stlbSize = ljdMMU_STLB_64K_SIZE;
                ljmkPRINT("    MTLB entry[%d]:   STLB physical = 0x%010lX, PAGE_SIZE = 64K, MASK = 0x%02X\n", mStart, stlbPhysical, maskMtlb);
                break;
            case 2:
                sEnd = ljdMMU_STLB_1M_ENTRY_NUM;
                stlbSize = ljdMMU_STLB_1M_SIZE;
                ljmkPRINT("    MTLB entry[%d]:   STLB physical = 0x%010lX, PAGE_SIZE = 1M, MASK = 0x%02X\n", mStart, stlbPhysical, maskMtlb);
                break;
            case 3:
                sEnd = ljdMMU_STLB_16M_ENTRY_NUM;
                stlbSize = ljdMMU_STLB_16M_SIZE;
                ljmkPRINT("    MTLB entry[%d]:   STLB physical = 0x%010lX, PAGE_SIZE = 16M, MASK = 0x%02X\n", mStart, stlbPhysical, maskMtlb);
                break;
            default:
                pr_warn("\nERROR: unknown PAGE SIZE...\n");
                ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
                break;
            }

            /* Convert GPU physical to CPU physical. */
            ljkOS_GPUPhysicalToCPUPhysical(kernel->os, stlbPhysical, &stlbPhysical);

            ljmkONERROR(ljkOS_MapPhysical(kernel->os, stlbPhysical, stlbSize, (ljtPOINTER *)&stlbLogical));

            /* print stlb */
            ljmkPRINT("      STLB entry : \n");
            stlbEntryNum = stlbSize >> 2;
            for (sStart = 0; sStart < stlbEntryNum; ++sStart) {
                stlbEntry = _ReadPageEntry(stlbLogical + sStart);
                if (stlbEntry && stlbEntry & ~0XF) {
                    ljtUINT64 pagePhysical = ((ljtUINT64)stlbEntry & 0XFF0) << 28 | ((ljtUINT64)stlbEntry & ~0XFFF);
                    ljtUINT32 maskStlb = stlbEntry & 0XF;

                    ljmkPRINT("        stlb entry[%04d]:      Page Physical = 0x%010lX  mask = 0x%01X\n", sStart, pagePhysical, maskStlb);
                }
            }

            ljmkONERROR(ljkOS_UnmapPhysical(kernel->os, stlbLogical, stlbSize));
        } else {
            ljmkPRINT("      STLB is empty\n");
        }
        mStart++;
    }

    return status;
OnError:
    ljmkPRINT("ERROR...\n");
    return status;
}

int
lj_load_show(void *m, void *data)
{
    int len = 0;
    ljtUINT32 i, j;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_device = galDevice;
    ljkDEVICE device = ljvNULL;
    ljeCHIPPOWERSTATE statesStored = ljvPOWER_INVALID;
    ljeCHIPPOWERSTATE state = ljvPOWER_INVALID;
    ljtUINT32 load[ljdDEVICE_COUNT][ljvCORE_3D_MAX + 1] = { {0}, {0} };
    ljtUINT32 hi_total_cycle_count[ljdDEVICE_COUNT][ljvCORE_3D_MAX + 1] = { {0}, {0} };
    ljtUINT32 hi_total_idle_cycle_count[ljdDEVICE_COUNT][ljvCORE_3D_MAX + 1] = { {0}, {0} };

    ljtBOOL powerManagement[ljdDEVICE_COUNT][ljvCORE_3D_MAX + 1] = { {ljvFALSE}, {ljvFALSE} };

#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!gal_device)
        return -ENXIO;

    for (j = 0; j < ljdDEVICE_COUNT; j++) {
        if (gal_device->devices[j])
            device = gal_device->devices[j];

        for (i = 0; i <= ljvCORE_3D_MAX; i++) {
            if (device->kernels[i]) {
                if (device->kernels[i]->hardware) {
                    ljkHARDWARE Hardware = device->kernels[i]->hardware;

                    powerManagement[j][i] = Hardware->options.powerManagement;

                    if (powerManagement[j][i])
                        ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvFALSE));

                    ljmkONERROR(ljkHARDWARE_QueryPowerState(Hardware, &statesStored));

                    ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_ON_AUTO));

                    ljmkONERROR(ljkHARDWARE_SetGpuProfiler(Hardware, ljvTRUE));

                    Hardware->waitCount = 200 * 100;
                }
            }
        }
    }

    for (j = 0; j < ljdDEVICE_COUNT; j++) {
        if (gal_device->devices[j])
            device = gal_device->devices[j];

        for (i = 0; i <= ljvCORE_3D_MAX; i++) {
            if (device->kernels[i]) {
                if (device->kernels[i]->hardware)
                    ljmkONERROR(ljkHARDWARE_CleanCycleCount(device->kernels[i]->hardware));
            }
        }
    }

    ljkOS_Delay(ljvNULL, 1);

    for (j = 0; j < ljdDEVICE_COUNT; j++) {
        if (gal_device->devices[j])
            device = gal_device->devices[j];

        for (i = 0; i <= ljvCORE_3D_MAX; i++) {
            if (device->kernels[i]) {
                if (device->kernels[i]->hardware)
                    ljmkONERROR(ljkHARDWARE_QueryCycleCount(device->kernels[i]->hardware,
                                                            &hi_total_cycle_count[j][i],
                                                            &hi_total_idle_cycle_count[j][i]));
            }
        }
    }

    for (j = 0; j < ljdDEVICE_COUNT; j++) {
        if (gal_device->devices[j]) {
            len += fs_printf(ptr + len, "device        : %d\n", j);
            device = gal_device->devices[j];
        }

        for (i = 0; i <= ljvCORE_3D_MAX; i++) {
            if (device->kernels[i]) {
                if (device->kernels[i]->hardware) {
                    ljkHARDWARE Hardware = device->kernels[i]->hardware;

                    switch (statesStored) {
                    case ljvPOWER_OFF:
                        state = ljvPOWER_OFF_BROADCAST;
                        break;
                    case ljvPOWER_IDLE:
                        state = ljvPOWER_IDLE_BROADCAST;
                        break;
                    case ljvPOWER_SUSPEND:
                        state = ljvPOWER_SUSPEND_BROADCAST;
                        break;
                    case ljvPOWER_ON:
                        state = ljvPOWER_ON_AUTO;
                        break;
                    default:
                        state = statesStored;
                        break;
                    }

                    Hardware->waitCount = 200;

                    ljmkONERROR(ljkHARDWARE_SetGpuProfiler(Hardware, ljvFALSE));

                    if (powerManagement[j][i])
                        ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvTRUE));

                    ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, state));

                    if (hi_total_cycle_count[j][i] == 0) {
                        len += fs_printf(ptr, "The current HW doesn't support use AHB register to read cycle counter.\n");
                        goto OnError;
                    } else
                        load[j][i] = div_u64((ljtUINT64)(hi_total_cycle_count[j][i] - hi_total_idle_cycle_count[j][i]) * 100, hi_total_cycle_count[j][i]);

                    len += fs_printf(ptr + len, "    core      : %d\n", i);
                    len += fs_printf(ptr + len, "    load      : %d%%\n", load[j][i]);
                    len += fs_printf(ptr + len, "\n");
                }
            }
        }
    }

OnError:
    return len;
}

/* return load (%) */
int
lj_load_get(int core)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_device = galDevice;
    ljkDEVICE device = ljvNULL;
    ljeCHIPPOWERSTATE statesStored = ljvPOWER_INVALID;
    ljeCHIPPOWERSTATE state = ljvPOWER_INVALID;
    ljtUINT32 load = 0;
    ljtUINT32 hi_total_cycle_count = 0;
    ljtUINT32 hi_total_idle_cycle_count = 0;

    static ljtBOOL profilerEnable = ljvFALSE;
    ljtBOOL powerManagement = ljvFALSE;

    if (!gal_device)
        return -1;

    device = gal_device->devices[0];

    if (core < 0 || core > ljvCORE_2D_MAX || device->kernels[core] == ljvNULL)
        return -1;

    if (device->kernels[core]->hardware) {
        ljkHARDWARE Hardware = device->kernels[core]->hardware;
        powerManagement = Hardware->options.powerManagement;

        if (powerManagement)
            ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvFALSE));

        ljmkONERROR(ljkHARDWARE_QueryPowerState(Hardware, &statesStored));

        ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_ON_AUTO));

        if (!profilerEnable) {
            ljmkONERROR(ljkHARDWARE_SetGpuProfiler(Hardware, ljvTRUE));

            ljmkONERROR(ljkHARDWARE_InitProfiler(Hardware));

            profilerEnable = ljvTRUE;
        }

        Hardware->waitCount = 200 * 100;
    }

    if (device->kernels[core]) {
        if (device->kernels[core]->hardware)
            ljmkONERROR(ljkHARDWARE_CleanCycleCount(device->kernels[core]->hardware));
    }

    ljkOS_Delay(ljvNULL, 1);

    if (device->kernels[core]) {
        if (device->kernels[core]->hardware)
            ljmkONERROR(ljkHARDWARE_QueryCycleCount(device->kernels[core]->hardware,
                                                    &hi_total_cycle_count,
                                                    &hi_total_idle_cycle_count));
    }

    if (device->kernels[core]) {
        if (device->kernels[core]->hardware) {
            ljkHARDWARE Hardware = device->kernels[core]->hardware;

            switch (statesStored) {
            case ljvPOWER_OFF:
                state = ljvPOWER_OFF_BROADCAST;
                break;
            case ljvPOWER_IDLE:
                state = ljvPOWER_IDLE_BROADCAST;
                break;
            case ljvPOWER_SUSPEND:
                state = ljvPOWER_SUSPEND_BROADCAST;
                break;
            case ljvPOWER_ON:
                state = ljvPOWER_ON_AUTO;
                break;
            default:
                state = statesStored;
                break;
            }

            Hardware->waitCount = 200;

            if (powerManagement)
                ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvTRUE));

            ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, state));

            if (hi_total_cycle_count == 0) {
                pr_warn("Core[%d]: The current HW doesn't support use AHB register to read cycle counter.\n", core);
                goto OnError;
            }
            else
                load = (hi_total_cycle_count - hi_total_idle_cycle_count) * 100 / hi_total_cycle_count;
        }
    }

OnError:
    return load;
}

static const char *vidmemTypeStr[ljvVIDMEM_TYPE_COUNT] = {
    "Generic",
    "Index",
    "Vertex",
    "Texture",
    "RenderTarget",
    "Depth",
    "Bitmap",
    "TileStatus",
    "Image",
    "Mask",
    "Scissor",
    "HZ",
    "ICache",
    "TxDesc",
    "Fence",
    "TFBHeader",
    "Command",
};

static const char *poolStr[ljvPOOL_NUMBER_OF_POOLS] = {
    "Unknown",
    "Default",
    "Local",
    "Internal",
    "External",
    "Unified",
    "System",
    "Virtual",
    "User",
    "Insram",
    "Exsram",
    "Exclusive",
    "Exclu-Exter",
    "Exter-Exclu",
    "Exclu-Exter-Vir",
    "System-32bit",
};

static int
_ShowCounters(void *File, ljsDATABASE_PTR Database)
{
    ljtUINT i = 0;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = File;
#else
    char *ptr = (char *)File;
#endif

    static const char * const otherCounterNames[] = {
        "AllocNonPaged",
        "AllocContiguous",
        "MapMemory",
    };

    ljsDATABASE_COUNTERS *otherCounters[] = {
        &Database->nonPaged,
        &Database->contiguous,
        &Database->mapMemory,
    };

    len = fs_printf(ptr, "%-16s %16s %16s %16s\n", "", "Current", "Maximum", "Total");

    /* Print surface type counters. */
    len += fs_printf(ptr + len,
                     "%-16s %16lld %16lld %16lld\n",
                     "All-Types",
                     Database->vidMem.bytes,
                     Database->vidMem.maxBytes,
                     Database->vidMem.totalBytes);

    for (i = 1; i < ljvVIDMEM_TYPE_COUNT; i++) {
        len += fs_printf(ptr + len,
                         "%-16s %16lld %16lld %16lld\n",
                         vidmemTypeStr[i],
                         Database->vidMemType[i].bytes,
                         Database->vidMemType[i].maxBytes,
                         Database->vidMemType[i].totalBytes);
    }
    /*seq_puts(File, "\n");*/
    len += fs_printf(ptr + len, "\n");

    /* Print surface pool counters. */
    len += fs_printf(ptr + len,
                     "%-16s %16lld %16lld %16lld\n",
                     "All-Pools",
                     Database->vidMem.bytes,
                     Database->vidMem.maxBytes,
                     Database->vidMem.totalBytes);

    for (i = 1; i < ljvPOOL_NUMBER_OF_POOLS; i++) {
        len += fs_printf(ptr + len,
                         "%-16s %16lld %16lld %16lld\n",
                         poolStr[i],
                         Database->vidMemPool[i].bytes,
                         Database->vidMemPool[i].maxBytes,
                         Database->vidMemPool[i].totalBytes);
    }
    /*seq_puts(File, "\n");*/
    len += fs_printf(ptr + len, "\n");

    /* Print other counters. */
    for (i = 0; i < ljmCOUNTOF(otherCounterNames); i++) {
        len += fs_printf(ptr + len,
                         "%-16s %16lld %16lld %16lld\n",
                         otherCounterNames[i],
                         otherCounters[i]->bytes,
                         otherCounters[i]->maxBytes,
                         otherCounters[i]->totalBytes);
    }
    /*seq_puts(File, "\n");*/
    len += fs_printf(ptr + len, "\n");
    return len;
}

static int
_ShowRecord(void *File, ljsDATABASE_PTR Database, ljsDATABASE_RECORD_PTR Record)
{
    ljtUINT32 handle;
    ljkVIDMEM_NODE nodeObject;
    ljtPHYS_ADDR_T physical;
    ljeSTATUS status = ljvSTATUS_OK;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = File;
#else
    char *ptr = (char *)File;
#endif

    static const char *recordTypes[ljvDB_NUM_TYPES] = {
        "Unknown",
        "VideoMemory",
        "CommandBuffer",
        "NonPaged",
        "Contiguous",
        "Signal",
        "VidMemLock",
        "Context",
        "Idel",
        "MapMemory",
        "ShBuf",
    };

    handle = ljmPTR2INT32(Record->data);

    if (Record->type == ljvDB_VIDEO_MEMORY ||
        Record->type == ljvDB_VIDEO_MEMORY_LOCKED) {
        status = ljkVIDMEM_HANDLE_Lookup2(Record->kernel, Database, handle, &nodeObject);

        if (ljmIS_ERROR(status)) {
            len += fs_printf(ptr + len, "%6u Invalid Node\n", handle);
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
        ljmkONERROR(ljkVIDMEM_NODE_GetCPUPhysical(Record->kernel, nodeObject, 0, &physical));
    } else {
        physical = (ljtUINT64)(ljtUINTPTR_T)Record->physical;
    }

    len += fs_printf(ptr + len,
                     "%-14s %3d %16x %16zx %16zu\n",
                     recordTypes[Record->type],
                     Record->kernel->core,
                     ljmPTR2INT32(Record->data),
                     (size_t)physical,
                     Record->bytes);

OnError:
    return len;
}

static int
_ShowDataBaseOldFormat(void *File, ljsDATABASE_PTR Database)
{
    ljtINT  pid;
    ljtUINT i;
    char name[24];
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = File;
#else
    char *ptr = (char *)File;
#endif

    /* Process ID and name */
    pid = Database->processID;
    ljmkVERIFY_OK(ljkOS_GetProcessNameByPid(pid, ljmSIZEOF(name), name));

    len = fs_printf(ptr, "--------------------------------------------------------------------------------\n");
    len += fs_printf(ptr + len, "Process: %-8d %s\n", pid, name);

    len += fs_printf(ptr + len, "Records:\n");

    len += fs_printf(ptr + len,
                     "%14s %3s %16s %16s %16s\n",
                     "Type", "GPU", "Data/Node", "Physical/Node", "Bytes");

    for (i = 0; i < ljmCOUNTOF(Database->list); i++) {
        ljsDATABASE_RECORD_PTR record = Database->list[i];

        while (record) {
            len += _ShowRecord(ptr + len, Database, record);
            record = record->next;
        }
    }

    len += fs_printf(ptr + len, "Counters:\n");

    len += _ShowCounters(ptr + len, Database);
    return len;
}

static int
lj_db_old_show(void *m, void *data, ljtBOOL all)
{
    ljsDATABASE_PTR  database;
    ljtINT i;

    static ljtUINT64 idleTime;
    ljeSTATUS status;
    ljkGALDEVICE device = galDevice;
    ljkKERNEL kernel = _GetValidKernel(0, device);
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!kernel)
        return -ENXIO;

    /* Acquire the database mutex. */
    ljmkONERROR(ljkOS_AcquireMutex(kernel->os, kernel->db->dbMutex, ljvINFINITE));

    if (kernel->db->idleTime) {
        /* Record idle time if DB upated. */
        idleTime = kernel->db->idleTime;
        kernel->db->idleTime = 0;
    }

    /* Idle time since last call */
    len = fs_printf(ptr, "GPU Idle: %llu ns\n", idleTime);

    if (all) {
        /* Walk the databases. */
        for (i = 0; i < ljmCOUNTOF(kernel->db->db); ++i) {
            for (database = kernel->db->db[i];
                 database != ljvNULL;
                 database = database->next)
                len += _ShowDataBaseOldFormat(ptr + len, database);
        }
    }

    /* Release the database mutex. */
    ljmkVERIFY_OK(ljkOS_ReleaseMutex(kernel->os, kernel->db->dbMutex));

    return len;

OnError:
    if (status == ljvSTATUS_INVALID_ARGUMENT)
        return -EINVAL;
    return status;
}

static int
lj_db_show(void *m, void *data, ljtBOOL all)
{
    return 0;
}

static int
lj_version_show(void *m, void *data)
{
    ljkGALDEVICE device = galDevice;
    ljsPLATFORM *platform = ljvNULL;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!device)
        return -ENXIO;

    platform = device->platform;
    if (!platform)
        return -ENXIO;

#ifdef CONFIG_DEBUG_FS
    len = fs_printf(ptr, "%s built at %s\n", ljvVERSION_STRING, HOST);

    if (platform->name)
        len += fs_printf(ptr + len, "Platform path: %s\n", platform->name);
    else
        len += fs_printf(ptr + len, "Code path: %s\n", __FILE__);
#else
    len = fs_printf(ptr, "%s\n", ljvVERSION_STRING);
#endif

    return len;
}

static void
print_ull(char dest[32], unsigned long long u)
{
    unsigned int t[7];
    int i;

    if (u < 1000) {
        sprintf(dest, "%27llu", u);
        return;
    }

    for (i = 0; i < 7 && u; i++)
        t[i] = do_div(u, 1000);

    dest += sprintf(dest, "%*s", (7 - i) * 4, "");
    dest += sprintf(dest, "%3u", t[--i]);

    for (i--; i >= 0; i--)
        dest += sprintf(dest, ",%03u", t[i]);
}

/*******************************************************************************
 **
 ** Show PM state timer.
 **
 ** Entry is called as 'idle' for compatible reason, it shows more information
 ** than idle actually.
 **
 **  Start: Start time of this counting period.
 **  End: End time of this counting period.
 **  On: Time GPU stays in ljvPOWER_0N.
 **  Off: Time GPU stays in ljvPOWER_0FF.
 **  Idle: Time GPU stays in ljvPOWER_IDLE.
 **  Suspend: Time GPU stays in ljvPOWER_SUSPEND.
 */
static int
lj_idle_show(void *m, void *data)
{
    ljkGALDEVICE device = galDevice;
    ljkKERNEL kernel = _GetValidKernel(0, device);
    char str[32];

    ljtUINT64 on;
    ljtUINT64 off;
    ljtUINT64 idle;
    ljtUINT64 suspend;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!kernel)
        return -ENXIO;

    ljkHARDWARE_QueryStateTimer(kernel->hardware, &on, &off, &idle, &suspend);

    /* Idle time since last call */
    print_ull(str, on);
    len = fs_printf(ptr, "On:      %s ns\n", str);
    print_ull(str, off);
    len += fs_printf(ptr + len, "Off:     %s ns\n", str);
    print_ull(str, idle);
    len += fs_printf(ptr + len, "Idle:    %s ns\n", str);
    print_ull(str, suspend);
    len += fs_printf(ptr + len, "Suspend: %s ns\n", str);

    return len;
}

extern void
_DumpState(ljkKERNEL Kernel);

static ljtUINT dumpDevice = 0;
static ljtUINT dumpCore   = 0;

static int
lj_dump_trigger_show(void *m, void *data)
{
    int len = 0;

#if ljdENABLE_3D || ljdENABLE_2D
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
# else
    char *ptr = (char *)m;
# endif

    ljkGALDEVICE gal_device = galDevice;
    ljkDEVICE device = ljvNULL;
    ljkKERNEL kernel = ljvNULL;
    ljkHARDWARE Hardware = ljvNULL;
    ljtBOOL powerManagement = ljvFALSE;
    ljeSTATUS status = ljvSTATUS_OK;
    ljeCHIPPOWERSTATE statesStored, state;

    ljtINT i = 0;

    len += fs_printf(ptr + len, "Dump one device: For example, dump cores on device 0: echo 0 > /sys/kernel/debug/lj/dump_trigger; cat /sys/kernel/debug/lj/dump_trigger\n");
    len += fs_printf(ptr + len, "The dump will be in [dmesg].\n");

    device = gal_device->devices[dumpDevice];

    for (i = 0; i < ljvCORE_COUNT; ++i) {
        if (!device->kernels[i])
            continue;

        kernel = device->kernels[i];
        Hardware = kernel->hardware;
        powerManagement = Hardware->options.powerManagement;

        if (powerManagement)
            ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvFALSE));

        ljmkONERROR(ljkHARDWARE_QueryPowerState(Hardware, &statesStored));

        ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_ON_AUTO));

        _DumpState(kernel);

        switch (statesStored) {
        case ljvPOWER_OFF:
            state = ljvPOWER_OFF_BROADCAST;
            break;
        case ljvPOWER_IDLE:
            state = ljvPOWER_IDLE_BROADCAST;
            break;
        case ljvPOWER_SUSPEND:
            state = ljvPOWER_SUSPEND_BROADCAST;
            break;
        case ljvPOWER_ON:
            state = ljvPOWER_ON_AUTO;
            break;
        default:
            state = statesStored;
            break;
        }

        if (powerManagement)
            ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvTRUE));

        ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, state));
    }

OnError:
#endif
    return len;
}

static int dumpProcess;

static int
_ShowVideoMemoryOldFormat(void *File, ljsDATABASE_PTR Database, ljtBOOL All)
{
    ljtUINT i = 0;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = File;
#else
    char *ptr = (char *)File;
#endif

    static const char * const otherCounterNames[] = {
        "AllocNonPaged",
        "AllocContiguous",
        "MapMemory",
    };

    ljsDATABASE_COUNTERS *otherCounters[] = {
        &Database->nonPaged,
        &Database->contiguous,
        &Database->mapMemory,
    };

    len = fs_printf(ptr, "%-16s %16s %16s %16s\n", "", "Current", "Maximum", "Total");

    /* Print surface type counters. */
    len += fs_printf(ptr + len,
                     "%-16s %16llu %16llu %16llu\n",
                     "All-Types",
                     Database->vidMem.bytes,
                     Database->vidMem.maxBytes,
                     Database->vidMem.totalBytes);

    if (All) {
        for (i = 1; i < ljvVIDMEM_TYPE_COUNT; i++) {
            len += fs_printf(ptr + len,
                             "%-16s %16llu %16llu %16llu\n",
                             vidmemTypeStr[i],
                             Database->vidMemType[i].bytes,
                             Database->vidMemType[i].maxBytes,
                             Database->vidMemType[i].totalBytes);
        }
        /*seq_puts(File, "\n");*/
        len += fs_printf(ptr + len, "\n");

        /* Print surface pool counters. */
        len += fs_printf(ptr + len,
                         "%-16s %16llu %16llu %16llu\n",
                         "All-Pools",
                         Database->vidMem.bytes,
                         Database->vidMem.maxBytes,
                         Database->vidMem.totalBytes);

        for (i = 1; i < ljvPOOL_NUMBER_OF_POOLS; i++) {
            len += fs_printf(ptr + len,
                             "%-16s %16llu %16llu %16llu\n",
                             poolStr[i],
                             Database->vidMemPool[i].bytes,
                             Database->vidMemPool[i].maxBytes,
                             Database->vidMemPool[i].totalBytes);
        }
        /*seq_puts(File, "\n");*/
        len += fs_printf(ptr + len, "\n");

        /* Print other counters. */
        for (i = 0; i < ljmCOUNTOF(otherCounterNames); i++) {
            len += fs_printf(ptr + len,
                             "%-16s %16llu %16llu %16llu\n",
                             otherCounterNames[i],
                             otherCounters[i]->bytes,
                             otherCounters[i]->maxBytes,
                             otherCounters[i]->totalBytes);
        }
        /*seq_puts(File, "\n");*/
        len += fs_printf(ptr + len, "\n");
    }

    return len;
}

static int
lj_vidmem_old_show(void *m, void *unused, ljtBOOL all)
{
    ljeSTATUS status;
    ljsDATABASE_PTR database;
    ljkGALDEVICE gal_device = galDevice;
    ljkKERNEL kernel = ljvNULL;
    char name[64];
    int i;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    kernel = _GetValidKernel(0, gal_device);

    if (!kernel)
        return -ENXIO;

    if (dumpProcess == 0) {
        /* Acquire the database mutex. */
        ljmkONERROR(ljkOS_AcquireMutex(kernel->os, kernel->db->dbMutex, ljvINFINITE));

        for (i = 0; i < ljmCOUNTOF(kernel->db->db); i++) {
            for (database = kernel->db->db[i];
                 database != ljvNULL;
                 database = database->next) {
                ljkOS_GetProcessNameByPid(database->processID, ljmSIZEOF(name), name);
                len += fs_printf(ptr + len,
                                 "VidMem Usage (Process %u: %s):\n",
                                 database->processID, name);
                len += _ShowVideoMemoryOldFormat(ptr + len, database, all);
                /*seq_puts(m, "\n");*/
                len += fs_printf(ptr + len, "\n");
            }
        }

        /* Release the database mutex. */
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(kernel->os, kernel->db->dbMutex));
    } else {
        /* Find the database. */
        status = ljkKERNEL_FindDatabase(kernel, dumpProcess, ljvFALSE, &database);

        if (ljmIS_ERROR(status)) {
            len += fs_printf(ptr + len, "ERROR: process %d not found\n", dumpProcess);
            return len;
        }

        ljkOS_GetProcessNameByPid(dumpProcess, ljmSIZEOF(name), name);
        len += fs_printf(ptr + len, "VidMem Usage (Process %d: %s):\n", dumpProcess, name);
        len += _ShowVideoMemoryOldFormat(ptr + len, database, all);
    }

    return len;

OnError:
    if (status == ljvSTATUS_INVALID_ARGUMENT)
        return -EINVAL;
    return status;
}

static int
lj_vidmem_show(void *m, void *unused, ljtBOOL all)
{
    return 0;
}

#ifdef CONFIG_DEBUG_FS
static int
lj_reserved_mem_usage_show(void *m)
{
    int len = 0;
    ljkGALDEVICE gal_device = galDevice;
    ljkVIDMEM vidMem;
    int usage;
    int i = 0;
    size_t t_freeBytes = 0;
    size_t t_bytes = 0;
    int t_usage = 0;
    void *ptr = m;

    len = fs_printf(ptr, "%-25s %16s %16s %16s\n", "", "FreeBytes", "TotalBytes", "Usage(%)");


    /* Internal POOL */
    if (gal_device->internalVidMem) {
        vidMem = gal_device->internalVidMem;
        usage = (vidMem->bytes - vidMem->freeBytes) * 100 / vidMem->bytes;
        len += fs_printf(ptr + len, "%-25s", "Local Internel Pool:");
        len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", "", vidMem->freeBytes, vidMem->bytes, usage);
    }

    /* LOCAL EXTERNAL POOL */
    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (gal_device->externalVidMem[i]) {
            vidMem = gal_device->externalVidMem[i];
            t_freeBytes += vidMem->freeBytes;
            t_bytes += vidMem->bytes;
        } else {
            break;
        }
    }

    if (t_bytes > 0) {
        t_usage = (t_bytes - t_freeBytes) * 100 / t_bytes;
        len += fs_printf(ptr + len, "%-25s", "Local External Pool:");
        len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", t_freeBytes, t_bytes, t_usage);
    }

    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (gal_device->externalVidMem[i]) {
            vidMem = gal_device->externalVidMem[i];
            usage = (vidMem->bytes - vidMem->freeBytes) * 100 / vidMem->bytes;
            len += fs_printf(ptr + len, "%4s%s %d %s:%9s", "", "Card", i, "Pool", "");
            len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", vidMem->freeBytes, vidMem->bytes, usage);
        } else {
            break;
        }
    }

    t_freeBytes = 0;
    t_bytes = 0;
    t_usage = 0;

    /* LOCAL EXCLUSIVE POOL */
    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (gal_device->exclusiveVidMem[i]) {
            vidMem = gal_device->exclusiveVidMem[i];
            t_freeBytes += vidMem->freeBytes;
            t_bytes += vidMem->bytes;
        } else {
            break;
        }
    }

    if (t_bytes > 0) {
        t_usage = (t_bytes - t_freeBytes) * 100 / t_bytes;
        len += fs_printf(ptr + len, "%-25s", "Local Exclusive Pool:");
        len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", t_freeBytes, t_bytes, t_usage);
    }

    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (gal_device->exclusiveVidMem[i]) {
            vidMem = gal_device->exclusiveVidMem[i];
            usage = (vidMem->bytes - vidMem->freeBytes) * 100 / vidMem->bytes;
            len += fs_printf(ptr + len, "%4s%s %d %s:%9s", "", "Card", i, "Pool", "");
            len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", vidMem->freeBytes, vidMem->bytes, usage);
        } else {
            break;
        }
    }

    t_freeBytes = 0;
    t_bytes = 0;
    t_usage = 0;

    /* RESERVED POOL */
    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        if (gal_device->contiguousVidMems[i]) {
            vidMem = gal_device->contiguousVidMems[i];
            t_freeBytes += vidMem->freeBytes;
            t_bytes += vidMem->bytes;
        } else {
            break;
        }
    }

    if (t_bytes > 0) {
        t_usage = (t_bytes - t_freeBytes) * 100 / t_bytes;
        len += fs_printf(ptr + len, "%-25s", "System Reserved Pool:");
        len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", t_freeBytes, t_bytes, t_usage);
    }

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        if (gal_device->contiguousVidMems[i]) {
            vidMem = gal_device->contiguousVidMems[i];
            usage = (vidMem->bytes - vidMem->freeBytes) * 100 / vidMem->bytes;
            len += fs_printf(ptr + len, "%4s%s %d:%14s", "", "Pool", i, "");
            len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", vidMem->freeBytes, vidMem->bytes, usage);
        } else {
            break;
        }
    }

    t_freeBytes = 0;
    t_bytes = 0;
    t_usage = 0;

    /* SRAM POOL */
    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        if (gal_device->extSRAMVidMems[i]) {
            vidMem = gal_device->extSRAMVidMems[i];
            t_freeBytes += vidMem->freeBytes;
            t_bytes += vidMem->bytes;
        } else {
            break;
        }
    }

    if (t_bytes > 0) {
        t_usage = (t_bytes - t_freeBytes) * 100 / t_bytes;
        len += fs_printf(ptr + len, "%-25s", "SRAM Pool:");
        len += fs_printf(ptr + len, "%16lld %16lld %16d%%\n", t_freeBytes, t_bytes, t_usage);
    }

    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        if (gal_device->extSRAMVidMems[i]) {
            vidMem = gal_device->extSRAMVidMems[i];
            usage = (vidMem->bytes - vidMem->freeBytes) * 100 / vidMem->bytes;
            len += fs_printf(ptr + len, "SRAM Pool %d: ", i);
            len += fs_printf(ptr + len, "%-16s %16lld %16lld %16d%%\n", "", vidMem->freeBytes, vidMem->bytes, usage);
        } else {
            break;
        }
        len += fs_printf(ptr + len, "\n");
    }
    return 0;
}

static inline int
strtoint_from_user(const char __user *s, size_t count, int *res)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 0, 0)
    int ret = kstrtoint_from_user(s, count, 10, res);

    return ret < 0 ? ret : count;
# else
    /* sign, base 2 representation, newline, terminator */
    char buf[1 + sizeof(long) * 8 + 1 + 1];

    size_t len = min(count, sizeof(buf) - 1);

    if (copy_from_user(buf, s, len))
        return -EFAULT;
    buf[len] = '\0';

    if (kstrtol(buf, 0, res))
        return ljvSTATUS_INVALID_DATA;

    return count;
# endif
}

static int
lj_vidmem_write(const char __user *buf, size_t count, void *data)
{
    return strtoint_from_user(buf, count, &dumpProcess);
}

static int
lj_dump_trigger_write(const char __user *buf, size_t count, void *data)
{
    return strtoint_from_user(buf, count, &dumpDevice);
}

#if ljdENABLE_MP_SWITCH
static int
lj_switch_core_count(void *m, void *data)
{
    return 0;
}

static int
lj_switch_core_count_write(const char __user *buf, size_t count, void *data)
{
    ljkGALDEVICE device = galDevice;
    int coreCount = 0;
    int ret;

    ret = strtoint_from_user(buf, count, &coreCount);

    if (ret && coreCount)
        device->platform->coreCount = coreCount;

    return ret;
}
# endif
#endif

static int
lj_clk_show(void *m, void *data)
{
    ljkGALDEVICE gal_device = galDevice;
    ljkDEVICE device = ljvNULL;
    ljtUINT i;
    ljeSTATUS status;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!gal_device)
        return -ENXIO;

    device = gal_device->devices[0];

    for (i = ljvCORE_MAJOR; i < ljvCORE_COUNT; i++) {
        if (device->kernels[i]) {
            ljkHARDWARE hardware = device->kernels[i]->hardware;

            if (i == ljvCORE_VG)
                continue;

            status = ljkHARDWARE_QueryFrequency(hardware);
            if (ljmIS_ERROR(status)) {
                len += fs_printf(ptr + len, "query gpu%d clock fail.\n", i);
                continue;
            }

            if (hardware->mcClk)
                len += fs_printf(ptr + len, "gpu%d mc clock: %llu HZ.\n", i, hardware->mcClk);

            if (hardware->shClk)
                len += fs_printf(ptr + len, "gpu%d sh clock: %llu HZ.\n", i, hardware->shClk);
        }
    }

    return len;
}

static ljtINT clkScale[2] = { 0, 0 };

static int
set_clk(const char *buf)
{
    ljkHARDWARE hardware;
    ljtINT n, j, k;
    ljtBOOL isSpace = ljvFALSE;
    char data[20];
    ljkGALDEVICE gal_device = galDevice;
    ljkDEVICE device = ljvNULL;
    ljtUINT32 devIndex = 0;

    if (!gal_device)
        return -ENXIO;

    memset(data, 0, 20);
    n = 0; j = 0; k = 0;

    while (ljvTRUE) {
        if ((buf[k] >= '0') && (buf[k] <= '9')) {
            if (isSpace) {
                data[n++] = ' ';
                isSpace = ljvFALSE;
            }
            data[n++] = buf[k];
        } else if (buf[k] == ' ') {
            if (n > 0)
                isSpace = ljvTRUE;
        } else if (buf[k] == '\n') {
            break;
        } else {
            pr_err("Error: command format must be this: echo \"0 32 32\" > /sys/kernel/debug/lj/clk\n");
            return 0;
        }

        k++;

        if (k >= 20)
            break;
    }

    if (4 == sscanf(data, "%u %u %d %d", &devIndex, &dumpCore, &clkScale[0], &clkScale[1])) {
        pr_warn("Change device:%d core:%d MC scale:%d SH scale:%d\n",
                devIndex, dumpCore, clkScale[0], clkScale[1]);
    } else {
        pr_warn("usage: echo \"0 0 32 32\" > clk\n");
        return 0;
    }

    /* Currently only support to scale one device. */
    device = gal_device->devices[devIndex];

    if (dumpCore < ljvCORE_COUNT && device->kernels[dumpCore]) {
        hardware = device->kernels[dumpCore]->hardware;

        ljkHARDWARE_SetClock(hardware, clkScale[0], clkScale[1]);
    } else {
        pr_err("Error: invalid core\n");
    }

    return 0;
}

static int
lj_poweroff_timeout_show(void *m, void *data)
{
    ljkGALDEVICE gal_device = galDevice;
    ljkHARDWARE hardware;
    int len = 0;
#ifdef CONFIG_DEBUG_FS
    void *ptr = m;
#else
    char *ptr = (char *)m;
#endif

    if (!gal_device)
        return -ENXIO;

    hardware = gal_device->devices[0]->kernels[0]->hardware;

#ifdef CONFIG_DEBUG_FS
    len += fs_printf(ptr + len, "power off timeout: %d ms.\n", hardware->powerOffTimeout);
#else
    len += sprintf(ptr + len, "power off timeout: %d ms.\n", hardware->powerOffTimeout);
#endif

    return len;
}

static int
poweroff_timeout_set(const char *buf)
{
    ljkGALDEVICE gal_device = galDevice;
    ljtINT i, ret;
    ljkDEVICE device = ljvNULL;
    ljtUINT32 devIndex;

    if (!gal_device)
        return -ENXIO;

    for (devIndex = 0; devIndex < gal_device->args.devCount; devIndex++) {
        device = gal_device->devices[devIndex];

        for (i = ljvCORE_MAJOR; i < ljvCORE_COUNT; i++) {
            if (device->kernels[i]) {
                ljkHARDWARE hardware = device->kernels[i]->hardware;

                if (i == ljvCORE_VG) {
                    continue;
                }

                ret = kstrtouint(buf, 0, &hardware->powerOffTimeout);
                if (ret < 0)
                    return ret;
            }
        }
    }

    return 0;
}

#ifdef CONFIG_DEBUG_FS
static int
debugfs_copy_from_user(char *k_buf, const char __user *buf, size_t count)
{
    int ret;

    ret = copy_from_user(k_buf, buf, count);
    if (ret != 0) {
        pr_err("Error: lost data: %d\n", (int)ret);
        return -1;
    }

    k_buf[count] = 0;

    return count;
}

static int
lj_clk_write(const char __user *buf, size_t count, void *data)
{
    size_t ret, _count;
    char k_buf[30];

    _count = min_t(size_t, count, (sizeof(k_buf) - 1));

    ret = debugfs_copy_from_user(k_buf, buf, _count);
    if (ret == -1)
        return ret;

    set_clk(k_buf);

    return ret;
}

static int
lj_poweroff_timeout_write(const char __user *buf, size_t count, void *data)
{
    size_t ret;
    char k_buf[30];

    ret = debugfs_copy_from_user(k_buf, buf, count);
    if (ret == -1)
        return ret;

    poweroff_timeout_set(k_buf);

    return ret;
}

int
lj_info_show_debugfs(struct seq_file *m, void *data)
{
    return lj_info_show((void *)m, data);
}

int
lj_clients_show_debugfs(struct seq_file *m, void *data)
{
    return lj_clients_show((void *)m, data);
}

int
lj_meminfo_show_debugfs(struct seq_file *m, void *data)
{
    return lj_meminfo_show((void *)m, data);
}

int
lj_mmuinfo_show_debugfs(struct seq_file *m, void *data)
{
    return lj_mmuinfo_show();
}

int
lj_idle_show_debugfs(struct seq_file *m, void *data)
{
    return lj_idle_show((void *)m, data);
}

int
lj_db_old_show_debugfs(struct seq_file *m, void *data)
{
    return lj_db_old_show((void *)m, data, ljvTRUE);
}

int
lj_db_show_debugfs(struct seq_file *m, void *data)
{
    return lj_db_show((void *)m, data, ljvTRUE);
}

int
lj_version_show_debugfs(struct seq_file *m, void *data)
{
    return lj_version_show((void *)m, data);
}

int
lj_vidmem_old_show_debugfs(struct seq_file *m, void *data)
{
    return lj_vidmem_old_show((void *)m, data, ljvTRUE);
}

int
lj_vidmem_show_debugfs(struct seq_file *m, void *data)
{
    return lj_vidmem_show((void *)m, data, ljvTRUE);
}

int
lj_reserved_mem_usage_show_debugfs(struct seq_file *m, void *data)
{
    return lj_reserved_mem_usage_show((void *)m);
}

int
lj_dump_trigger_show_debugfs(struct seq_file *m, void *data)
{
    return lj_dump_trigger_show((void *)m, data);
}

int
lj_clk_show_debugfs(struct seq_file *m, void *data)
{
    return lj_clk_show((void *)m, data);
}

int
lj_poweroff_timeout_show_debugfs(struct seq_file *m, void *data)
{
    return lj_poweroff_timeout_show((void *)m, data);
}

#if ljdENABLE_MP_SWITCH
int
lj_switch_core_count_debugfs(struct seq_file *m, void *data)
{
    return lj_switch_core_count((void *)m, data);
}
# endif

#if LJMICRO_PROFILER
int
lj_load_show_debugfs(struct seq_file *m, void *data)
{
    return lj_load_show((void *)m, data);
}
# endif

static ljsINFO InfoList[] = {
    { "info", lj_info_show_debugfs },
    { "clients", lj_clients_show_debugfs },
    { "meminfo", lj_meminfo_show_debugfs },
    { "mmu", lj_mmuinfo_show_debugfs },
    { "idle", lj_idle_show_debugfs },
    { "database", lj_db_old_show_debugfs },
    { "database64x", lj_db_show_debugfs },
    { "version", lj_version_show_debugfs },
    { "vidmem", lj_vidmem_old_show_debugfs, lj_vidmem_write },
    { "vidmem64x", lj_vidmem_show_debugfs, lj_vidmem_write },
    { "reserved_mem_usage", lj_reserved_mem_usage_show_debugfs, lj_vidmem_write },
    { "dump_trigger", lj_dump_trigger_show_debugfs, lj_dump_trigger_write },
    { "clk", lj_clk_show_debugfs, lj_clk_write },
    { "poweroff_timeout", lj_poweroff_timeout_show_debugfs, lj_poweroff_timeout_write },
#if ljdENABLE_MP_SWITCH
    { "core_count", lj_switch_core_count_debugfs, lj_switch_core_count_write },
# endif
#if LJMICRO_PROFILER
    { "load", lj_load_show_debugfs },
# endif
};

#else
static ssize_t
info_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_info_show((void *)buf, NULL);
}
DEVICE_ATTR_RO(info);

static ssize_t
clients_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_clients_show((void *)buf, NULL);
}
DEVICE_ATTR_RO(clients);

static ssize_t
meminfo_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_meminfo_show((void *)buf, NULL);
}
DEVICE_ATTR_RO(meminfo);

static ssize_t
mmu_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_mmuinfo_show();
}
DEVICE_ATTR_RO(mmu);

static ssize_t
idle_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_idle_show((void *)buf, NULL);
}
DEVICE_ATTR_RO(idle);

static ssize_t
database_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_db_old_show((void *)buf, NULL, ljvFALSE);
}
DEVICE_ATTR_RO(database);

static ssize_t
database64x_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_db_show((void *)buf, NULL, ljvFALSE);
}
DEVICE_ATTR_RO(database64x);

static ssize_t
version_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_version_show((void *)buf, NULL);
}
DEVICE_ATTR_RO(version);

static ssize_t
load_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_load_show((void *)buf, NULL);
}
DEVICE_ATTR_RO(load);

static ssize_t
vidmem_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_vidmem_old_show((void *)buf, NULL, ljvFALSE);
}

static ssize_t
vidmem_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    if (kstrtoint(buf, 0, &dumpProcess))
        return ljvSTATUS_INVALID_DATA;
    return count;
}
DEVICE_ATTR_RW(vidmem);

static ssize_t
vidmem64x_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_vidmem_show((void *)buf, NULL, ljvFALSE);
}

static ssize_t
vidmem64x_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    if (kstrtoint(buf, 0, &dumpProcess))
        return ljvSTATUS_INVALID_DATA;
    return count;
}
DEVICE_ATTR_RW(vidmem64x);

static ssize_t
dump_trigger_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_dump_trigger_show((void *)buf, NULL);
}

static ssize_t
dump_trigger_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    if (kstrtoint(buf, 0, &dumpDevice))
        return ljvSTATUS_INVALID_DATA;
    return count;
}
DEVICE_ATTR_RW(dump_trigger);

static ssize_t
clk_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_clk_show((void *)buf, NULL);
}

static ssize_t
clk_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    set_clk(buf);
    return count;
}
DEVICE_ATTR_RW(clk);

static ssize_t
poweroff_timeout_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return lj_poweroff_timeout_show((void *)buf, NULL);
}

static ssize_t
poweroff_timeout_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    poweroff_timeout_set(buf);
    return count;
}
DEVICE_ATTR_RW(poweroff_timeout);

static struct attribute *Info_attrs[] = {
    &dev_attr_info.attr,
    &dev_attr_clients.attr,
    &dev_attr_meminfo.attr,
    &dev_attr_mmu.attr,
    &dev_attr_idle.attr,
    &dev_attr_database.attr,
    &dev_attr_database64x.attr,
    &dev_attr_version.attr,
    &dev_attr_vidmem.attr,
    &dev_attr_vidmem64x.attr,
    &dev_attr_dump_trigger.attr,
    &dev_attr_clk.attr,
    &dev_attr_poweroff_timeout.attr,
    NULL,
};
ATTRIBUTE_GROUPS(Info);
#endif

static ljeSTATUS
_DebugfsInit(ljkGALDEVICE Device)
{
    ljeSTATUS status = ljvSTATUS_OK;

#ifdef CONFIG_DEBUG_FS
    ljkDEBUGFS_DIR dir = &Device->debugfsDir;

    ljmkONERROR(ljkDEBUGFS_DIR_Init(dir, ljvNULL, "lj"));
    ljmkONERROR(ljkDEBUGFS_DIR_CreateFiles(dir, InfoList, ljmCOUNTOF(InfoList), Device));
#else
    int ret;
    /* TODO. */
    struct device *dev = (struct device *)Device->devices[0]->dev;

    ret = sysfs_create_groups(&dev->kobj, Info_groups);
    if (ret < 0)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
#endif
    galDevice = Device;

OnError:
    return status;
}

static void
_DebugfsCleanup(ljkGALDEVICE Device)
{
#ifdef CONFIG_DEBUG_FS
    ljkDEBUGFS_DIR dir = &Device->debugfsDir;

    if (Device->debugfsDir.root) {
        ljmkVERIFY_OK(ljkDEBUGFS_DIR_RemoveFiles(dir, InfoList, ljmCOUNTOF(InfoList)));

        ljkDEBUGFS_DIR_Deinit(dir);
    }
#else
    /* TODO. */
    struct device *dev = (struct device *)Device->devices[0]->dev;

    sysfs_remove_groups(&dev->kobj, Info_groups);
#endif
}

/*******************************************************************************
 ************************** Memory Allocation Wrappers *************************
 *******************************************************************************/

static ljeSTATUS
_AllocateMemory(ljkGALDEVICE Device, ljtSIZE_T Bytes,
                ljtPOINTER *Logical, ljtPHYS_ADDR *Physical, ljtUINT64 *PhysAddr)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtPHYS_ADDR_T physAddr;

    ljmkHEADER_ARG("Device=%p Bytes=0x%zx", Device, Bytes);

    ljmkVERIFY_ARGUMENT(Device);
    ljmkVERIFY_ARGUMENT(Logical);
    ljmkVERIFY_ARGUMENT(Physical);
    ljmkVERIFY_ARGUMENT(PhysAddr);

    ljmkONERROR(ljkOS_AllocateNonPagedMemory(Device->os,
                                             Device->devices[0]->kernels[0],
                                             ljvFALSE,
                                             ljvALLOC_FLAG_CONTIGUOUS,
                                             &Bytes, Physical, Logical));

    ljmkONERROR(ljkOS_GetDmaAddressFromHandle(Device->os, *Physical, 0, &physAddr));

    *PhysAddr = physAddr;

OnError:
    ljmkFOOTER_ARG("*Logical=%p *Physical=%p *PhysAddr=0x%llx",
                   ljmOPT_POINTER(Logical), ljmOPT_POINTER(Physical), ljmOPT_VALUE(PhysAddr));

    return status;
}

static ljeSTATUS
_FreeMemory(ljkGALDEVICE Device, ljtPOINTER Logical, ljtPHYS_ADDR Physical)
{
    ljeSTATUS status;

    ljmkHEADER_ARG("Device=%p Logical=%p Physical=%p",
                   Device, Logical, Physical);

    ljmkVERIFY_ARGUMENT(Device);

    status = ljkOS_FreeNonPagedMemory(Device->os, Physical, Logical,
                                      ((PLINUX_MDL)Physical)->numPages * PAGE_SIZE);

    ljmkFOOTER();
    return status;
}

static ljeSTATUS
_SetupContiguousVidMem(ljkGALDEVICE Device, ljtUINT Index, const ljsMODULE_PARAMETERS *Args)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtUINT64 physAddr = ~0ULL;
    ljkGALDEVICE device = Device;
    ljtUINT i = Index;

    ljmkHEADER_ARG("Device=%p Args=%p", Device, Args);

    if (i >= ljdSYSTEM_RESERVE_COUNT)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    /* set up the contiguous memory */
    device->contiguousBases[i] = Args->contiguousBases[i];
    device->contiguousSizes[i] = Args->contiguousSizes[i];
    device->requestedContiguousBases[i] = 0;
    device->requestedContiguousSizes[i] = 0;

    if (!device->contiguousSizes[i]) {
        ljmkFOOTER_NO();
        return ljvSTATUS_OK;
    }

    if (!device->contiguousBases[i]) {
        while (device->contiguousSizes[i] > 0) {
            /* Allocate contiguous memory. */
            status = _AllocateMemory(device,
                                     device->contiguousSizes[i],
                                     &device->contiguousLogicals[i],
                                     &device->contiguousPhysicals[i],
                                     &physAddr);

            if (ljmIS_SUCCESS(status)) {
                status = ljkVIDMEM_Construct(device->os,
                                             physAddr,
                                             device->contiguousSizes[i],
                                             64,
                                             Args->bankSize,
                                             &device->contiguousVidMems[i]);

                if (ljmIS_SUCCESS(status)) {
                    ljkALLOCATOR allocator = ((PLINUX_MDL)device->contiguousPhysicals[i])->allocator;
                    device->contiguousVidMems[i]->capability = allocator->capability | ljvALLOC_FLAG_MEMLIMIT;
                    device->contiguousVidMems[i]->physical = device->contiguousPhysicals[i];
                    device->contiguousBases[i] = physAddr;
                    if (device->contiguousBases[i] > 0xFFFFFFFFULL)
                        device->contiguousVidMems[i]->capability &= ~ljvALLOC_FLAG_4GB_ADDR;
                    break;
                }

                ljmkONERROR(_FreeMemory(device,
                                        device->contiguousLogicals[i],
                                        device->contiguousPhysicals[i]));

                device->contiguousLogicals[i]  = ljvNULL;
                device->contiguousPhysicals[i] = ljvNULL;
            }

            if (device->contiguousSizes[i] <= (4 << 20))
                device->contiguousSizes[i] = 0;
            else
                device->contiguousSizes[i] -= (4 << 20);
        }
    } else if (device->os->iommu) {
        /* Disable contiguous memory pool. */
        device->contiguousVidMems[i] = ljvNULL;
        device->contiguousSizes[i] = 0;
    } else {
        ljtPHYS_ADDR_T contiguousBases = device->contiguousBases[i];
        ljtSIZE_T contiguousSizes = device->contiguousSizes[i];

#if ljdENABLE_40BIT_VA
        if (!device->lowContiguousSize) {
            ljtPHYS_ADDR_T gpuBase;
            ljtPHYS_ADDR_T lowContiguousBase = contiguousBases;
            ljtSIZE_T lowContiguousSize = contiguousSizes;

            ljkOS_CPUPhysicalToGPUPhysical(device->os, contiguousBases, &gpuBase);

            if (gpuBase < ljd4G_SIZE) {
                if (gpuBase + contiguousSizes > ljd4G_SIZE) {
                    ljkOS_GPUPhysicalToCPUPhysical(device->os, ljd4G_SIZE, &contiguousBases);

                    /* Adjust low VA range. */
                    if (gpuBase < ljdVA_RESERVED_SIZE)
                        contiguousBases -= ljdVA_RESERVED_SIZE;
                    else if (gpuBase > ljd4G_SIZE - ljd4G_VA_FM_SIZE)
                        contiguousBases += gpuBase - (ljd4G_SIZE - ljd4G_VA_FM_SIZE);

                    lowContiguousSize = contiguousBases - lowContiguousBase;
                    contiguousSizes = (lowContiguousSize >= ljd4G_VA_FM_SIZE) ? (contiguousSizes - lowContiguousSize) : 0;
                } else {
                    contiguousSizes = 0;
                }
            } else {
                if (contiguousSizes > ljd4G_VA_FM_SIZE) {
                    /* At least reserve ljd4G_VA_FM_SIZE. */
                    lowContiguousSize = ljd4G_VA_FM_SIZE;
                    contiguousBases += ljd4G_VA_FM_SIZE;
                    contiguousSizes -= ljd4G_VA_FM_SIZE;
                } else {
                    contiguousSizes = 0;
                }
            }

            /* Construct reserve memory pool below 4G. */
            {
                /* Create the contiguous memory heap. */
                status = ljkVIDMEM_Construct(device->os,
                                             lowContiguousBase,
                                             lowContiguousSize,
                                             64,
                                             Args->bankSize,
                                             &device->lowContiguousVidMem);

                if (ljmIS_ERROR(status)) {
                    /* Error, disable contiguous memory pool. */
                    device->lowContiguousVidMem = ljvNULL;
                    device->lowContiguousSize   = 0;
                } else {
                    ljkALLOCATOR allocator;
                    ljtBOOL contiguousRequested = Args->contiguousRequested;

#if ljdCAPTURE_ONLY_MODE
                    contiguousRequested = ljvTRUE;
#endif

                    ljmkONERROR(ljkOS_RequestReservedMemory(device->os,
                                                            lowContiguousBase,
                                                            lowContiguousSize,
                                                            "ljLowContMem",
                                                            contiguousRequested,
                                                            &device->lowContiguousPhysical));

                    allocator = ((PLINUX_MDL)device->lowContiguousPhysical)->allocator;

                    device->lowContiguousVidMem->capability = allocator->capability | ljvALLOC_FLAG_MEMLIMIT;
                    device->lowContiguousVidMem->physical = device->lowContiguousPhysical;
                    device->lowContiguousBase = lowContiguousBase;
                    device->lowContiguousSize = lowContiguousSize;
                }
            }

            if (contiguousSizes == 0) {
                device->contiguousVidMems[i] = ljvNULL;
                device->contiguousSizes[i] = 0;

                return ljvSTATUS_OK;
            }
        }
#endif
        /* Create the contiguous memory heap. */
        status = ljkVIDMEM_Construct(device->os,
                                     contiguousBases,
                                     contiguousSizes,
                                     64,
                                     Args->bankSize,
                                     &device->contiguousVidMems[i]);

        if (ljmIS_ERROR(status)) {
            /* Error, disable contiguous memory pool. */
            device->contiguousVidMems[i] = ljvNULL;
            device->contiguousSizes[i] = 0;
        } else {
            ljkALLOCATOR allocator;
            ljtBOOL contiguousRequested = Args->contiguousRequested;

#if ljdCAPTURE_ONLY_MODE
            contiguousRequested = ljvTRUE;
#endif

            ljmkONERROR(ljkOS_RequestReservedMemory(device->os,
                                                    contiguousBases,
                                                    contiguousSizes,
                                                    "ljContMem",
                                                    contiguousRequested,
                                                    &device->contiguousPhysicals[i]));

            allocator = ((PLINUX_MDL)device->contiguousPhysicals[i])->allocator;

            device->contiguousVidMems[i]->capability = allocator->capability | ljvALLOC_FLAG_MEMLIMIT;
            device->contiguousVidMems[i]->physical = device->contiguousPhysicals[i];
            device->requestedContiguousBases[i] = contiguousBases;
            device->requestedContiguousSizes[i] = contiguousSizes;
            device->contiguousBases[i] = contiguousBases;
            device->contiguousSizes[i] = contiguousSizes;
        }
    }

    if (Args->showArgs) {
        ljmkPRINT("Ljmcore Info: ContiguousBase=0x%llx ContiguousSize=0x%zx\n",
                  device->contiguousBases[i], device->contiguousSizes[i]);
    }

OnError:
    ljmkFOOTER();
    return status;
}

static ljeSTATUS
_SetupExternalSRAMVidMem(ljkGALDEVICE galDevice, const ljsMODULE_PARAMETERS *Args)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkDEVICE device;
    ljtUINT32 i, j, index, devIndex;

    ljmkHEADER_ARG("Device=%p", galDevice);

    /* Setup external SRAM memory region. */
    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        if (!galDevice->extSRAMSizes[i] && galDevice->devices[0]->extSRAMSizes[i])
            galDevice->extSRAMSizes[i] = galDevice->devices[0]->extSRAMSizes[i];

        if (galDevice->extSRAMSizes[i] > 0) {
            /* create the external SRAM memory heap */
            status = ljkVIDMEM_Construct(galDevice->os,
                                         galDevice->extSRAMBases[i],
                                         galDevice->extSRAMSizes[i],
                                         64,
                                         0,
                                         &galDevice->extSRAMVidMems[i]);

            if (ljmIS_ERROR(status)) {
                /* Error, disable external SRAM heap. */
                galDevice->extSRAMSizes[i] = 0;
            } else {
                char sRAMName[40];
                snprintf(sRAMName, ljmSIZEOF(sRAMName) - 1, "Ljmcore external sram%d", i);

#if ljdCAPTURE_ONLY_MODE
                galDevice->args.sRAMRequested = ljvTRUE;
#endif
                /* Map external SRAM memory. */
                ljmkONERROR(ljkOS_RequestReservedMemory(galDevice->os,
                                                        galDevice->extSRAMBases[i],
                                                        galDevice->extSRAMSizes[i],
                                                        sRAMName,
                                                        galDevice->args.sRAMRequested,
                                                        &galDevice->extSRAMPhysicals[i]));

                galDevice->extSRAMVidMems[i]->physical = galDevice->extSRAMPhysicals[i];
            }
        }
    }

    for (devIndex = 0; devIndex < galDevice->args.devCount; devIndex++) {
        device = galDevice->devices[devIndex];
        index = 0;

        for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
            if (Args->devSRAMIDs[devIndex] & (1 << i)) {
                device->extSRAMVidMems[index] = galDevice->extSRAMVidMems[i];
                device->extSRAMPhysicals[index] = galDevice->extSRAMPhysicals[i];

                for (j = 0; j < ljvCORE_COUNT; j++) {
                    if (device->irqLines[j] != -1 &&
                        device->kernels[j] &&
                        device->extSRAMPhysicals[index]) {
                        device->kernels[j]->hardware->options.extSRAMGPUPhysNames[index]
                            = ljkKERNEL_AllocateNameFromPointer(device->kernels[j], device->extSRAMPhysicals[index]);
                    }
                }

                index++;
            }
        }
    }

OnError:
    ljmkFOOTER();
    return status;
}

/*******************************************************************************
 ****************************** Interrupt Handler ******************************
 *******************************************************************************/
extern    struct semaphore    sema_edma;

static irqreturn_t
isrRoutine(int irq, void *ctxt)
{
#if !LEGACY_INT
    ljeSTATUS status;
    ljkKERNEL kernel;// = (ljkKERNEL)ctxt;
    ljkDEVICE device = (ljkDEVICE)ctxt;
    ljtUINT32 misc_intr_src;
    irqreturn_t ret = IRQ_NONE;

    /* check misk interrupt src */
    spin_lock(&device->os->registerAccessLock);
    misc_intr_src = LJM_MISC_R32(device->platformIndex, MISC_INTR_SRC);
    spin_unlock(&device->os->registerAccessLock);

    /* handle 3D0 interrupt */
    if (misc_intr_src & m_INTR_SRC_G3D0) {
        kernel = device->kernels[ljvCORE_MAJOR];

        if (kernel) {
            /* Call kernel interrupt notification. */
            status = ljkHARDWARE_Interrupt(kernel->hardware);

            if (ljmIS_SUCCESS(status)) {
                up(kernel->sema);
                ret = IRQ_HANDLED;
            }
        }
    }

    /* handle 3D1 interrupt */
    if (misc_intr_src & m_INTR_SRC_G3D1) {
        kernel = device->kernels[ljvCORE_3D1];

        if (kernel) {
            /* Call kernel interrupt notification. */
            status = ljkHARDWARE_Interrupt(kernel->hardware);

            if (ljmIS_SUCCESS(status)) {
                up(kernel->sema);
                ret = IRQ_HANDLED;
            }
        }
    }

    /* handle 2D interrupt */
    if (misc_intr_src & m_INTR_SRC_G2D) {
        kernel = device->kernels[ljvCORE_2D];

        if (kernel) {
            /* Call kernel interrupt notification. */
            status = ljkHARDWARE_Interrupt(kernel->hardware);

            if (ljmIS_SUCCESS(status)) {
                up(kernel->sema);
                ret = IRQ_HANDLED;
            }
        }
    }
#if 1
    /* handle edma interrupt */
    if (misc_intr_src & m_INTR_SRC_EDMA) {

        /* mask all read interrupt */
        //writel(0x00FF00FF, (ljtUINT8 *)pcie_platform->pcie_info[device->platformIndex].bar[5].logical + DMA_READ_INT_MASK_REG);
        /* mask all write interrupt */
        //writel(0x00FF00FF, (ljtUINT8 *)pcie_platform->pcie_info[device->platformIndex].bar[5].logical + DMA_WRITE_INT_MASK_REG);
        /* mask all edma interrupt in misc */
        //edma_write(edma_dev->reg_base_2, REG_MISC_INTR_EN_ADDR, 0xFFFF0000 & edma_read(edma_dev->reg_base_2, REG_MISC_INTR_EN_ADDR));
        spin_lock(&device->os->registerAccessLock);
        LJM_MISC_W32(device->platformIndex, MISC_INTR_EN, LJM_MISC_R32(device->platformIndex, MISC_INTR_EN) & ~(m_INTR_EN_EDMA));
        spin_unlock(&device->os->registerAccessLock);

        up(&sema_edma);
        ret = IRQ_HANDLED;
    }
#endif
    return ret;
#else
    ljeSTATUS status;
    ljkKERNEL kernel = (ljkKERNEL)ctxt;


    /* Call kernel interrupt notification. */
    status = ljkHARDWARE_Interrupt(kernel->hardware);

    if (ljmIS_SUCCESS(status)) {
        up(kernel->sema);
        return IRQ_HANDLED;
    }

    return IRQ_NONE;
#endif
}

static irqreturn_t
isrRoutineVG(int irq, void *ctxt)
{
#if ljdENABLE_VG
    ljeSTATUS status;
    ljkKERNEL kernel = (ljkKERNEL)ctxt;

    /* Serve the interrupt. */
    status = ljkVGINTERRUPT_Enque(kernel->vg->interrupt);

    /* Determine the return value. */
    return (status == ljvSTATUS_NOT_OUR_INTERRUPT) ? IRQ_RETVAL(0) : IRQ_RETVAL(1);
#else
    return IRQ_NONE;
#endif
}

static const char * const isrNames[] = {
    "ljmcore:0",
    "ljmcore:3d-1",
    "ljmcore:3d-2",
    "ljmcore:3d-3",
    "ljmcore:3d-4",
    "ljmcore:3d-5",
    "ljmcore:3d-6",
    "ljmcore:3d-7",
    "ljmcore:3d-8",
    "ljmcore:3d-9",
    "ljmcore:3d-10",
    "ljmcore:3d-11",
    "ljmcore:3d-12",
    "ljmcore:3d-13",
    "ljmcore:3d-14",
    "ljmcore:3d-15",
    "ljmcore:2d",
    "ljmcore:2d1",
    "ljmcore:2d2",
    "ljmcore:2d3",
    "ljmcore:vg",
#if ljdDEC_ENABLE_AHB
    "ljmcore:dec"
#endif
};

static int
isrRoutinePoll(void *ctxt)
{
    ljkKERNEL kernel = (ljkKERNEL)ctxt;

    while (1) {
        if (unlikely(kernel->killThread)) {
            /* The daemon exits. */
            while (!kthread_should_stop())
                ljkOS_Delay(kernel->os, 1);

            return 0;
        }

#if ljdENABLE_VG
        if (kernel->core == ljvCORE_VG)
            isrRoutineVG(-1, ljvNULL);
        else
#endif
            isrRoutine(-1, kernel);

        ljkOS_Delay(kernel->os, 10);
    }

    return 0;
}

static ljeSTATUS
_SetupIsr(ljkDEVICE Device, ljeCORE Core)
{
    ljtINT ret = 0;
    ljeSTATUS status = ljvSTATUS_OK;
    irq_handler_t handler;

    ljmkHEADER_ARG("Device=%p Core=%d", Device, Core);

    ljmkVERIFY_ARGUMENT(Device);

    ljmSTATIC_ASSERT(ljvCORE_COUNT == ljmCOUNTOF(isrNames),
                     "isrNames array does not match core types");

    if (Device->irqLines[Core] == -1) {
        ljtUINT64 isrPolling = -1;

        if (Device->isrThread[Core])
            return status;

        ljkOS_QueryOption(Device->os, "isrPoll", &isrPolling);

        /* use kthread to poll int stat */
        if (ljmBITTEST(isrPolling, Core) != 0) {
            struct task_struct *task;

            Device->killIsrThread = ljvFALSE;

            task = kthread_run(isrRoutinePoll,
                               (ljtPOINTER)Device->kernels[Core],
                               "%s_poll", isrNames[Core]);

            if (IS_ERR(task)) {
                ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                               "%s(%d): Could not start the intr poll thread.\n",
                               __func__, __LINE__);

                ljmkONERROR(ljvSTATUS_GENERIC_IO);
            }

            ljmkPRINT("ljmcore: polling core%d int state\n", Core);

            Device->isrThread[Core] = task;
            Device->isrInitializeds[Core] = ljvTRUE;

            return status;
        }
        /* it should not run to here */
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    handler = (Core == ljvCORE_VG) ? isrRoutineVG : isrRoutine;

    /*
     * Hook up the isr based on the irq line.
     * For shared irq, device-id can not be 0, but CORE_MAJOR value is.
     * Add by 1 here and subtract by 1 in isr to fix the issue.
     */
    ret = request_irq(Device->irqLines[Core], handler, ljdIRQF_FLAG,
                      isrNames[Core], (void *)Device->kernels[Core]);

    if (ret != 0) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): Could not register irq line %d (error=%d)\n",
                       __func__, __LINE__, Device->irqLines[Core], ret);

        ljmkONERROR(ljvSTATUS_GENERIC_IO);
    }

    /* Mark ISR as initialized. */
    Device->isrInitializeds[Core] = ljvTRUE;

OnError:
    ljmkFOOTER();
    return status;
}

#if !LEGACY_INT
/* only request a isr for all 3D and 2D cores
 * add by hym on 20220915
 */
static ljeSTATUS
_SetupIsrEx(IN ljkDEVICE Device)
{
    ljtINT        ret    = 0;
    ljeSTATUS     status = ljvSTATUS_OK;
    irq_handler_t handler;
    ljtINT        irqLine = -1;
    ljtUINT       i;

    ljmkHEADER_ARG("Device=%p", Device);

    ljmkVERIFY_ARGUMENT(Device);

    ljmSTATIC_ASSERT(ljvCORE_COUNT == ljmCOUNTOF(isrNames),
                     "isrNames array does not match core types");

    for (i = 0; i < ljvCORE_COUNT; i++) {
        if (Device->irqLines[i] > 0) {
            irqLine = Device->irqLines[i];
            break;
        }
    }

    if (irqLine == -1) {
        /* it should not run to here */
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    handler = isrRoutine;

    /*
     * Hook up the isr based on the irq line.
     * For shared irq, device-id can not be 0, but CORE_MAJOR value is.
     * Add by 1 here and subtract by 1 in isr to fix the issue.
     */
    ret = request_irq(irqLine, handler, ljdIRQF_FLAG,
                      "ljmcore", (void *)Device);

    if (ret != 0) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): Could not register irq line %d (error=%d)\n",
                       __func__, __LINE__, irqLine, ret);

        ljmkONERROR(ljvSTATUS_GENERIC_IO);
    }

    for (i = 0; i < ljvCORE_COUNT; i++) {
        if (Device->kernels[i] && i != ljvCORE_VG) {
            /* Mark ISR as initialized. */
            Device->isrInitializeds[i] = ljvTRUE;
        }
    }

OnError:
    ljmkFOOTER();
    return status;
}
#endif

static ljeSTATUS
_ReleaseIsr(ljkDEVICE Device, ljeCORE Core)
{
    ljmkHEADER_ARG("Device=%p Core=%d", Device, Core);

    ljmkVERIFY_ARGUMENT(Device);

    /* release the irq */
    if (Device->isrInitializeds[Core]) {
        if (Device->isrThread[Core]) {
            Device->killIsrThread = ljvTRUE;
            kthread_stop(Device->isrThread[Core]);
            Device->isrThread[Core] = ljvNULL;
        } else {
            if (Core == ljvCORE_VG) /* only VG request isr for every core */
                free_irq(Device->irqLines[Core], (void *)(Device->kernels[Core]));
        }

        Device->isrInitializeds[Core] = ljvFALSE;
    }

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

static int
threadRoutine(void *ctxt)
{
    ljkKERNEL kernel = (ljkKERNEL)ctxt;

    ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_DRIVER,
                   "Starting isr Thread with extension=%p", kernel);

    for (;;) {
        int down;

        down = down_interruptible(kernel->sema);
        if (down && down != -EINTR)
            return down;

        if (unlikely(kernel->killThread)) {
            /* The daemon exits. */
            while (!kthread_should_stop())
                ljkOS_Delay(kernel->os, 1);

            return 0;
        }

        ljkKERNEL_Notify(kernel, ljvNOTIFY_INTERRUPT);
    }
}

static ljeSTATUS
_StartThread(ljkDEVICE Device, ljeCORE Core)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkDEVICE device = Device;
    struct task_struct *task;

    if (device->kernels[Core] != ljvNULL) {
        /* Start the kernel thread. */
        task = kthread_run(threadRoutine, (void *)device->kernels[Core],
                           "ljmcore_deamon/%d-%d", Device->id, Core);

        if (IS_ERR(task)) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): Could not start the kernel thread.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_GENERIC_IO);
        }

        device->threadCtxts[Core] = task;
        device->threadInitializeds[Core] = ljvTRUE;
        device->kernels[Core]->threadInitialized = ljvTRUE;

        set_user_nice(task, -20);
    } else {
        device->threadInitializeds[Core] = ljvFALSE;
    }

OnError:
    return status;
}

static void
_StopThread(ljkDEVICE Device, ljeCORE Core)
{
    if (Device->threadInitializeds[Core]) {
        Device->kernels[Core]->killThread = ljvTRUE;
        up(Device->kernels[Core]->sema);

        kthread_stop(Device->threadCtxts[Core]);
        Device->threadCtxts[Core] = ljvNULL;
        Device->threadInitializeds[Core] = ljvFALSE;
    }
}

#if ljdENABLE_SW_PREEMPTION

static int
_ThreadPreempt(void *ctxt)
{
    ljkKERNEL kernel = (ljkKERNEL)ctxt;

    for (;;) {
        int down;

        down = down_interruptible(kernel->preemptSema);
        if (down && down != -EINTR)
            return down;

        if (unlikely(kernel->killPreemptThread)) {
            while (!kthread_should_stop())
                ljkOS_Delay(kernel->os, 1);

            return 0;
        }

        ljkKERNEL_PreemptionThread(kernel);
    }
}

static ljeSTATUS
_StartPreemptThread(ljkDEVICE Device, ljeCORE Core)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkDEVICE device = Device;
    struct task_struct *task;

    if (device->kernels[Core] != ljvNULL) {
        /* Start the kernel thread. */
        task = kthread_run(_ThreadPreempt, (void *)device->kernels[Core],
                           "ljmcore_preempt/%d-%d", device->id, Core);

        if (IS_ERR(task)) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): Could not start the kernel preempt thread.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_GENERIC_IO);
        }

        device->preemptThread[Core] = task;
        device->preemptThreadInits[Core] = ljvTRUE;
    } else {
        device->preemptThreadInits[Core] = ljvFALSE;
    }

OnError:
    return status;
}

static void
_StopPreemptThread(ljkDEVICE Device, ljeCORE Core)
{
    if (Device->preemptThreadInits[Core]) {
        Device->kernels[Core]->killPreemptThread = ljvTRUE;
        up(Device->kernels[Core]->preemptSema);

        kthread_stop((struct task_struct *)Device->preemptThread[Core]);
        Device->preemptThread[Core] = ljvNULL;
        Device->preemptThreadInits[Core] = ljvFALSE;
    }
}
#endif

#if USE_SCANOUT_PREALLOCATE
static ljeSTATUS ljkScanoutNode_Allocate(ljkGALDEVICE gal_device)
{
    ljkDEVICE    device;
    ljkKERNEL    kernel;
    ljeSTATUS    status = ljvSTATUS_OK;
    ljtSIZE_T    bytes;
    ljtUINT32    type;
    int          i, j;

    ljmkHEADER_ARG("gal_device=%p", gal_device);

    device = gal_device->devices[0];
    kernel = device->kernels[0];

    if (gal_device->scanoutNodeObjects[0][0] == ljvNULL) {
        /* allocate scanout node objects */
        for (i = 0; i < 4; i++) {
            /* typeï¿½ï¿½ï¿½Í¾ï¿½ï¿½ï¿½bankÎ»ï¿½ï¿½  */
            if (gal_device->exclusiveSize[0] > 0 || gal_device->externalSize[0] <= 0x40000000)
                type = ljvVIDMEM_TYPE_GENERIC;
            else
                type = ljvVIDMEM_TYPE_COLOR_BUFFER;
            for (j=0; j<2; j++) {
                bytes = MAX_SCANOUT_NODE_SIZE;
                if (gal_device->exclusiveSize[0] > 0) {
                    /* Allocate video memory node in exclusive */
                    ljmkONERROR(ljkVIDMEM_NODE_AllocateLinear(kernel, gal_device->exclusiveVidMem[0], ljvPOOL_LOCAL_EXCLUSIVE_EXTERNAL,
                                                              type++, ljvALLOC_FLAG_NONE, 64 << 20, ljvFALSE, &bytes,
                                                              &gal_device->scanoutNodeObjects[i][j]));
                }
                else {
                    /* Allocate video memory node in external */
                    ljmkONERROR(ljkVIDMEM_NODE_AllocateLinear(kernel, gal_device->externalVidMem[0], ljvPOOL_LOCAL_EXTERNAL,
                                                              type++, ljvALLOC_FLAG_NONE, 64 << 20, ljvFALSE, &bytes,
                                                              &gal_device->scanoutNodeObjects[i][j]));
                }
                /* Allocate handle for this VidMem */
                ljmkONERROR(ljkVIDMEM_HANDLE_Allocate(kernel, gal_device->scanoutNodeObjects[i][j],
                                                      &gal_device->scanoutNodeHandles[i][j]));
                ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE, "new scanout (%d:%d): handle=%d object=%p offset=%lx\n",
                    i, j, gal_device->scanoutNodeHandles[i][j], gal_device->scanoutNodeObjects[i][j],
                    (unsigned long)gal_device->scanoutNodeObjects[i][j]->node->VidMem.offset);

                ljmkONERROR(ljkVIDMEM_HANDLE_Reference(kernel, 1, gal_device->scanoutNodeHandles[i][j]));
                ljmkONERROR(ljkVIDMEM_NODE_Reference(kernel, gal_device->scanoutNodeObjects[i][j]));
            }
        }
    }
    return status;

OnError:
    ljmkFOOTER_ARG("status=%d", status);
    return status;
}
#endif

/*******************************************************************************
 *
 *  ljkGALDEVICE_Construct
 *
 *  Constructor.
 *
 *  INPUT:
 *
 *  OUTPUT:
 *
 *      ljkGALDEVICE *Device
 *          Pointer to a variable receiving the ljkGALDEVICE object pointer on
 *          success.
 */
ljeSTATUS
ljkGALDEVICE_Construct(ljsPLATFORM *Platform,
                       const ljsMODULE_PARAMETERS *Args,
                       ljkGALDEVICE *Device)
{
    ljkKERNEL kernel = ljvNULL;
    ljkGALDEVICE gal_device;
    ljkDEVICE device;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtUINT64 isrPolling = -1;
    ljtINT32 i = 0;
    ljtUINT32 devIndex, globalIndex = 0;
    ljtUINT32 global2DIndex = 0;
    ljtUINT32 index = 0;
    ljtUINT32 sysReserveBits;

    ljmkHEADER_ARG("Platform=%lx Args=%lx", (unsigned long)Platform, (unsigned long)Args);

    /* Allocate device structure. */
    gal_device = kmalloc(sizeof(struct _ljkGALDEVICE), GFP_KERNEL | __GFP_NOWARN);

    if (!gal_device)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    memset(gal_device, 0, sizeof(struct _ljkGALDEVICE));

    gal_device->platform = Platform;
    gal_device->platform->dev = ljvNULL;

    gal_device->args = *Args;

    if (gal_device->args.devCount > ljdDEVICE_COUNT)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    isrPolling = Args->isrPoll;

    /* Construct the ljkOS object. */
    ljmkONERROR(ljkOS_Construct(gal_device, &gal_device->os));

    /* Loop all the hardware devices. */
    for (devIndex = 0; devIndex < gal_device->args.devCount; devIndex++) {
        /* Construct the ljkDEVICE object for os independent core management. */
        ljmkONERROR(ljkDEVICE_Construct(gal_device->os, &gal_device->devices[devIndex]));

        device = gal_device->devices[devIndex];

        device->id = devIndex;
        device->vidMemIndex = Args->devMemIDs[devIndex];
        device->configCoreCount = Args->devCoreCounts[devIndex];
        device->config2DCoreCount = Args->dev2DCoreCounts[devIndex];
        device->showSRAMMapInfo = Args->showArgs;
        device->showMemInfo = Args->showArgs;
        device->platformIndex = Args->platformIDs[devIndex];
        device->dev = Args->devices[device->platformIndex];
        device->processPageTable = (gal_device->args.devCount > 1) ? ljvFALSE : ljvTRUE;

        /* Set the base address */
        device->baseAddress = Args->baseAddress;
        device->physBase = Args->baseAddress;
        device->physSize = Args->physSize;

        /* Loop all the cores in one deivce. */
        for (i = 0; i < device->configCoreCount; i++) {
            /* Get arguments from global core index. */
            device->irqLines[i] = Args->irqs[globalIndex];
            device->requestedRegisterMemBases[i] = Args->registerBases[globalIndex];
            device->requestedRegisterMemSizes[i] = Args->registerSizes[globalIndex];
            device->registerBasesMapped[i] = Args->registerBasesMapped[globalIndex];

            globalIndex++;
        }

        for (i = 0; i < device->config2DCoreCount; i++) {
            /* Get arguments from global core index. */
            device->irqLines[ljvCORE_2D + i] = Args->irq2Ds[global2DIndex];
            device->requestedRegisterMemBases[ljvCORE_2D + i] = Args->register2DBases[global2DIndex];
            device->requestedRegisterMemSizes[ljvCORE_2D + i] = Args->register2DSizes[global2DIndex];
            device->registerBasesMapped[ljvCORE_2D + i] = Args->register2DBasesMapped[global2DIndex];

            global2DIndex++;
        }

        if (Args->irqVG != -1) {
            /* Get arguments from global core index. */
            device->irqLines[ljvCORE_VG] = Args->irqVG;
            device->requestedRegisterMemBases[ljvCORE_VG] = Args->registerVGBase;
            device->requestedRegisterMemSizes[ljvCORE_VG] = Args->registerVGSize;
            device->registerBasesMapped[ljvCORE_VG] = Args->registerVGBaseMapped;
        }

        for (i = 0; i < ljvCORE_COUNT; i++) {
            unsigned long physical = (unsigned long)device->requestedRegisterMemBases[i];

            /* Set up register memory region. */
            if (physical) {
                if (device->registerBasesMapped[i]) {
                    device->registerBases[i] = device->registerBasesMapped[i];
                    device->registerSizes[i] = device->requestedRegisterMemSizes[i];
                    device->requestedRegisterMemBases[i] = 0;
                } else {
#if USE_LINUX_PCIE
                    ljmkPRINT("register should be mapped in platform layer");
#endif

#if 0/*abc*/
                    if (!request_mem_region(physical,
                                            device->requestedRegisterMemSizes[i],
                                            "ljmcore register region")) {
                        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                                       "%s(%d): Failed to claim %lu bytes @ 0x%llx\n",
                                       __func__, __LINE__,
                                       device->requestedRegisterMemSizes[i], physical);

                        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
                    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
                    device->registerBases[i] =
                        (ljtPOINTER)ioremap(physical, device->requestedRegisterMemSizes[i]);
#else
                    device->registerBases[i] =
                        (ljtPOINTER)ioremap_nocache(physical, device->requestedRegisterMemSizes[i]);
#endif
                    device->registerSizes[i] = device->requestedRegisterMemSizes[i];

                    if (!device->registerBases[i]) {
                        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                                       "%s(%d): Unable to map %ld bytes @ 0x%zx\n",
                                       __func__, __LINE__,
                                       physical, device->requestedRegisterMemSizes[i]);

                        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
                    }
#else
                    device->registerBases[i] = (ljtPOINTER)Args->registerBases[i];/*abc*/
                    device->registerSizes[i] = device->requestedRegisterMemSizes[i];
#endif
                }
            }
        }
    }

#if ljdENABLE_TRUST_APPLICATION
    if (gal_device->devices[0]->irqLines[ljvCORE_MAJOR] != -1 ||
        ljmBITTEST(isrPolling, ljvCORE_MAJOR)!= 0) {
        ljmkONERROR(ljtaOS_ConstructOS(gal_device->os, &gal_device->taos));
    }
#endif
    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        /* set -1 as default value, to avoid the misc device random minor is 0 */
        gal_device->devIDs[i] = -1;
    }

    gal_device->platform->dev = gal_device->devices[0];

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        /* Setup contiguous video memory pool. */
        ljmkONERROR(_SetupContiguousVidMem(gal_device, i, Args));
    }

    /* Set the external SRAM base address. */
    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        gal_device->extSRAMBases[i] = Args->extSRAMBases[i];
        gal_device->extSRAMSizes[i] = Args->extSRAMSizes[i];
    }

    /* Setup all the local memory for GAL device. */
    for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
        char name[20];

        gal_device->externalBase[i] = Args->externalBase[i];
        gal_device->externalSize[i] = Args->externalSize[i];
        gal_device->exclusiveBase[i] = Args->exclusiveBase[i];
        gal_device->exclusiveSize[i] = Args->exclusiveSize[i];

        if (gal_device->externalSize[i] > 0) {
            /* make two banks at least by hym */
            ljtSIZE_T bankSize = 256 << 20;
            if (gal_device->externalSize[i] <= bankSize) {
                bankSize = (gal_device->externalSize[i] + ((128 << 20) - 4)) / (128 << 20) * (128 << 20) / 2;
            }

            /* create the external memory heap */
            status = ljkVIDMEM_Construct(gal_device->os,
                                         gal_device->externalBase[i],
                                         gal_device->externalSize[i],
                                         64,
                                         bankSize,//0,
                                         &gal_device->externalVidMem[i]);

            if (ljmIS_ERROR(status)) {
                /* Error, disable external heap. */
                gal_device->externalSize[i] = 0;
            } else {
                sprintf(name, "ljExtMem%d", i);
                /* Map external memory. */
                ljmkONERROR(ljkOS_RequestReservedMemory(gal_device->os,
                                                        gal_device->externalBase[i],
                                                        gal_device->externalSize[i],
                                                        name,
                                                        ljvTRUE,
                                                        &gal_device->externalPhysical[i]));

                gal_device->externalVidMem[i]->physical = gal_device->externalPhysical[i];
            }
        }

        if (gal_device->exclusiveSize[i] > 0) {
            /* create the exclusive memory heap */
            status = ljkVIDMEM_Construct(gal_device->os,
                                         gal_device->exclusiveBase[i],
                                         gal_device->exclusiveSize[i],
                                         64,
                                         256<<20,//0,
                                         &gal_device->exclusiveVidMem[i]);

            if (ljmIS_ERROR(status)) {
                /* Error, disable exclusive heap. */
                gal_device->exclusiveSize[i] = 0;
            } else {
                ljkALLOCATOR allocator;

                sprintf(name, "ljExtMem%d", i);

                /* Map exclusive memory. */
                ljmkONERROR(ljkOS_RequestReservedMemory(gal_device->os,
                                                        gal_device->exclusiveBase[i],
                                                        gal_device->exclusiveSize[i],
                                                        name,
                                                        ljvTRUE,
                                                        &gal_device->exclusivePhysical[i]));

                allocator = ((PLINUX_MDL)gal_device->exclusivePhysical[i])->allocator;
                gal_device->exclusiveVidMem[i]->physical = gal_device->exclusivePhysical[i];
                gal_device->exclusiveVidMem[i]->capability |= allocator->capability;
            }
        }
    }

#if ljdEXTERNAL_SRAM_USAGE
    /* Setup external SRAM video memory pool. */
    ljmkONERROR(_SetupExternalSRAMVidMem(gal_device, Args));
#endif

    for (devIndex = 0; devIndex < gal_device->args.devCount; devIndex++) {
        device = gal_device->devices[devIndex];

        if (devIndex)
            device->database = gal_device->devices[0]->database;

        /* Query device memory configuration. */
        device->externalBase = gal_device->externalBase[device->vidMemIndex];
        device->externalSize = gal_device->externalSize[device->vidMemIndex];
        device->externalPhysical = gal_device->externalPhysical[device->vidMemIndex];
        device->externalVidMem = gal_device->externalVidMem[device->vidMemIndex];

        device->exclusiveBase = gal_device->exclusiveBase[device->vidMemIndex];
        device->exclusiveSize = gal_device->exclusiveSize[device->vidMemIndex];
        device->exclusivePhysical = gal_device->exclusivePhysical[device->vidMemIndex];
        device->exclusiveVidMem = gal_device->exclusiveVidMem[device->vidMemIndex];

        sysReserveBits = Args->devSysMemIDs[devIndex];
        index = 0;

        for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
            if (sysReserveBits & (1 << i)) {
                device->contiguousBases[index] = gal_device->contiguousBases[i];
                device->contiguousSizes[index] = gal_device->contiguousSizes[i];
                device->contiguousPhysicals[index] = gal_device->contiguousPhysicals[i];
                device->contiguousVidMems[index] = gal_device->contiguousVidMems[i];
                index++;
            }
        }

        device->lowContiguousBase = gal_device->lowContiguousBase;
        device->lowContiguousSize = gal_device->lowContiguousSize;

        index = 0;

        for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
            if (Args->devSRAMIDs[devIndex] & (1 << i)) {
                device->extSRAMBases[index] = gal_device->extSRAMBases[i];
                device->extSRAMSizes[index] = gal_device->extSRAMSizes[i];
                index++;
            }
        }

        /* Add core for all available major cores. */
        for (i = ljvCORE_MAJOR; i <= ljvCORE_3D_MAX; i++) {
            if (device->irqLines[i] != -1 || ljmBITTEST(isrPolling, i) != 0) {
                ljmkONERROR(ljkDEVICE_AddCore(device,
                                              (ljeCORE)i,
                                              Args->chipIDs[i],
                                              gal_device,
                                              &device->kernels[i]));

                if (!kernel)
                    kernel = device->kernels[i];

#if ljdENABLE_TRUST_APPLICATION
                if (device->kernels[i]->hardware->options.secureMode == ljvSECURE_IN_TA &&
                    devIndex == 0) {
                    ljmkONERROR(ljTA_Construct(gal_device->taos, (ljeCORE)i, &globalTA[i]));
                }
#endif

                ljmkONERROR(ljkHARDWARE_SetFastClear(device->kernels[i]->hardware,
                                                     Args->fastClear,
                                                     Args->compression));

                ljmkONERROR(ljkHARDWARE_EnablePowerManagement(device->kernels[i]->hardware,
                                                              Args->powerManagement));

#if ljdENABLE_FSCALE_VAL_ADJUST
                ljmkONERROR(ljkHARDWARE_SetMinFscaleValue(device->kernels[i]->hardware,
                                                          Args->gpu3DMinClock));

#endif
            } else {
                device->kernels[i] = ljvNULL;
            }
        }

        for (i = ljvCORE_2D; i <= ljvCORE_2D_MAX; i++) {
#if !ljdCAPTURE_ONLY_MODE
            if (device->irqLines[i] != -1 || ljmBITTEST(isrPolling, i)!= 0) {
                ljmkONERROR(ljkDEVICE_AddCore(device,
                                              (ljeCORE)i,
                                              Args->chipIDs[i],
                                              gal_device,
                                              &device->kernels[i]));

                if (!kernel)
                    kernel = device->kernels[i];

                ljmkONERROR(ljkHARDWARE_EnablePowerManagement(device->kernels[i]->hardware,
                                                              0 /* Args->powerManagement */));

#if ljdENABLE_FSCALE_VAL_ADJUST
                ljmkONERROR(ljkHARDWARE_SetMinFscaleValue(device->kernels[i]->hardware, 1));
#endif
            } else {
                device->kernels[i] = ljvNULL;
            }
#else
            device->kernels[i] = ljvNULL;
#endif
        }

#if !ljdCAPTURE_ONLY_MODE
        if (device->irqLines[ljvCORE_VG] != -1 || ljmBITTEST(isrPolling, ljvCORE_VG)!= 0) {
#if ljdENABLE_VG
             ljmkONERROR(ljkDEVICE_AddCore(device,
                                           ljvCORE_VG,
                                           ljvCHIP_ID_DEFAULT,
                                           gal_device,
                                           &device->kernels[ljvCORE_VG]));

            if (!kernel)
                kernel = device->kernels[i];

            ljmkONERROR(ljkVGHARDWARE_EnablePowerManagement(device->kernels[ljvCORE_VG]->vg->hardware,
                                                            Args->powerManagement));
#endif
        } else {
            device->kernels[ljvCORE_VG] = ljvNULL;
        }
#else
        device->kernels[ljvCORE_VG] = ljvNULL;
#endif
    }

#if !ljdEXTERNAL_SRAM_USAGE
    /* Setup external SRAM video memory pool. */
    ljmkONERROR(_SetupExternalSRAMVidMem(gal_device, Args));
#endif

    if (!kernel)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    if (gal_device->internalPhysical)
        gal_device->internalPhysName = ljmPTR_TO_NAME(gal_device->internalPhysical);

#if ljdENABLE_40BIT_VA
    if (gal_device->lowContiguousPhysical)
        gal_device->lowContiguousPhysName = ljmPTR_TO_NAME(gal_device->lowContiguousPhysical);
#endif

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        if (gal_device->contiguousPhysicals[i])
            gal_device->contiguousPhysNames[i] = ljmPTR_TO_NAME(gal_device->contiguousPhysicals[i]);
    }

    for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
        if (gal_device->externalPhysical[i])
            gal_device->externalPhysName[i] = ljmPTR_TO_NAME(gal_device->externalPhysical[i]);

        if (gal_device->exclusivePhysical[i])
            gal_device->exclusivePhysName[i] = ljmPTR_TO_NAME(gal_device->exclusivePhysical[i]);
    }

    for (devIndex = 0; devIndex < gal_device->args.devCount; devIndex++) {
        device = gal_device->devices[devIndex];

        /* Query device memory configuration. */
        device->externalPhysName = gal_device->externalPhysName[device->vidMemIndex];
        device->exclusivePhysName = gal_device->exclusivePhysName[device->vidMemIndex];

        sysReserveBits = Args->devSysMemIDs[devIndex];

        for (i = 0, index = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
            if (sysReserveBits & (1 << i)) {
                device->contiguousPhysNames[index] = gal_device->contiguousPhysNames[i];
                index++;
            }
        }
    }

    ljmkONERROR(_DebugfsInit(gal_device));

    /* check if 2D only, copy the 2d kernel pointer to major */
    if((!gal_device->devices[0]->kernels[ljvCORE_MAJOR]) && gal_device->devices[0]->kernels[ljvCORE_2D])
        gal_device->devices[0]->kernels[ljvCORE_MAJOR] = gal_device->devices[0]->kernels[ljvCORE_2D];

    /* Return pointer to the device. */
    galDevice = gal_device;
    *Device = gal_device;

#if 0
    /* create database for pid 1 */
    for (i = 0; i < ljdMAX_GPU_COUNT; ++i)
    {
        if (gal_device->devices[0]->kernels[i])
        {
            ljmkONERROR(ljkKERNEL_AttachProcessEx(gal_device->devices[0]->kernels[i], ljvTRUE, 1));
        }
    }
#endif

#if USE_SCANOUT_PREALLOCATE
    ljkScanoutNode_Allocate(gal_device); /* ï¿½ï¿½ï¿½ï¿½galdeviceï¿½ï¿½scanout nodeï¿½ï¿½handleï¿½Ä³ï¿½Ê¼ï¿½ï¿½   */
#endif

OnError:
    if (ljmIS_ERROR(status)) {
        /* Roll back. */
        ljmkVERIFY_OK(ljkGALDEVICE_Destroy(gal_device));
    }

    ljmkFOOTER_ARG("*Device=%p", *Device);
    return status;
}

/*******************************************************************************
 *
 *  ljkGALDEVICE_Destroy
 *
 *  Class destructor.
 *
 *  INPUT:
 *
 *      Nothing.
 *
 *  OUTPUT:
 *
 *      Nothing.
 *
 *  RETURNS:
 *
 *      Nothing.
 */
ljeSTATUS
ljkGALDEVICE_Destroy(ljkGALDEVICE gal_device)
{
    ljtINT i, j = 0;
    ljkKERNEL kernel = ljvNULL;
    ljtINT32 devIndex;
    ljkDEVICE device;

    ljmkHEADER_ARG("gal_device=%p", gal_device);

    if (gal_device) {
        if (!gal_device->devices[0])
            goto free_gal_device;

        kernel = gal_device->devices[0]->kernels[ljvCORE_MAJOR];
        if (!kernel)
            kernel = gal_device->devices[0]->kernels[ljvCORE_2D];
        if (!kernel)
            kernel = gal_device->devices[0]->kernels[ljvCORE_VG];

        if (!kernel) {
            ljmkFOOTER_NO();
            return ljvSTATUS_INVALID_ARGUMENT;
        }

        ljmkASSERT(kernel);

#if 0
        /* remove database for s_drm_probe_pid */
        for (i = 0; i < ljdMAX_GPU_COUNT; ++i)
        {
            if (gal_device->devices[0]->kernels[i])
            {
                ljkKERNEL_AttachProcessEx(gal_device->devices[0]->kernels[i], ljvFALSE, 1);
            }
        }
#endif

        /* Free all the local memories. */
        for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
            if (gal_device->externalPhysName[i]) {
                ljmRELEASE_NAME(gal_device->externalPhysName[i]);
                gal_device->externalPhysName[i] = 0;
            }

            if (gal_device->exclusivePhysName[i]) {
                ljmRELEASE_NAME(gal_device->exclusivePhysName[i]);
                gal_device->exclusivePhysName[i] = 0;
            }
        }

        for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
            if (gal_device->contiguousPhysNames[i]) {
                ljmRELEASE_NAME(gal_device->contiguousPhysNames[i]);
                gal_device->contiguousPhysNames[i] = 0;
            }
        }

#if ljdENABLE_40BIT_VA
        if (gal_device->lowContiguousPhysName) {
            ljmRELEASE_NAME(gal_device->lowContiguousPhysName);
            gal_device->lowContiguousPhysName = 0;
        }
#endif
        _DebugfsCleanup(gal_device);

        for (devIndex = gal_device->args.devCount - 1; devIndex >= 0; devIndex--) {
            device = gal_device->devices[devIndex];

            if (!device)
                continue;

            for (i = 0; i < ljvCORE_COUNT; i++) {
                if (device->kernels[i]) {
                    kernel = device->kernels[i];
#if ljdENABLE_TRUST_APPLICATION
                    if (kernel->core != ljvCORE_VG &&
                        kernel->hardware->options.secureMode == ljvSECURE_IN_TA &&
                        globalTA[i] && !devIndex) {
                        ljTA_Destroy(globalTA[i]);
                        globalTA[i] = ljvNULL;
                    }
#endif

                    if (gal_device->gotoShutdown)
                        kernel->dbCreated = ljvFALSE;

                    for (j = ljvSRAM_INTERNAL0; j < ljvSRAM_INTER_COUNT; j++) {
                        if (kernel->sRAMPhysical[j] != ljvNULL) {
                            /* Release reserved SRAM memory. */
                            ljkOS_ReleaseReservedMemory(device->os,
                                                        kernel->sRAMPhysical[j]);

                            kernel->sRAMPhysical[j] = ljvNULL;
                        }

                        if (kernel->sRAMVidMem[j] != ljvNULL) {
                            /* Destroy the SRAM contiguous heap. */
                            ljmkVERIFY_OK(ljkVIDMEM_Destroy(kernel->sRAMVidMem[j]));
                            kernel->sRAMVidMem[j] = ljvNULL;
                        }
                    }
                }
            }

            ljmkVERIFY_OK(ljkDEVICE_DestroyCores(device));

            for (i = 0; i < ljvCORE_COUNT; i++) {
                if (device->registerBases[i]) {
                    /* Unmap register memory. */
                    if (device->requestedRegisterMemBases[i] != 0) {
                        iounmap(device->registerBases[i]);

                        release_mem_region(device->requestedRegisterMemBases[i],
                                           device->requestedRegisterMemSizes[i]);
                    }

                    device->registerBases[i] = ljvNULL;
                    device->requestedRegisterMemBases[i] = 0;
                    device->requestedRegisterMemSizes[i] = 0;
                }
            }

            ljmkVERIFY_OK(ljkDEVICE_Destroy(device->os, device));

            device = ljvNULL;
        }

        for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
            if (gal_device->extSRAMPhysicals[i]) {
                ljkOS_ReleaseReservedMemory(gal_device->os,
                                            gal_device->extSRAMPhysicals[i]);

                gal_device->extSRAMPhysicals[i] = ljvNULL;
            }

            if (gal_device->extSRAMVidMems[i]) {
                ljmkVERIFY_OK(ljkVIDMEM_Destroy(gal_device->extSRAMVidMems[i]));
                gal_device->extSRAMVidMems[i] = ljvNULL;
            }
        }

        /* Free all the local memories. */
        for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
            if (gal_device->externalPhysical[i]) {
                ljkOS_ReleaseReservedMemory(gal_device->os,
                                            gal_device->externalPhysical[i]);

                gal_device->externalPhysical[i] = ljvNULL;
            }

            if (gal_device->externalVidMem[i]) {
                /* destroy the external heap */
                ljmkVERIFY_OK(ljkVIDMEM_Destroy(gal_device->externalVidMem[i]));
                gal_device->externalVidMem[i] = ljvNULL;
            }

            if (gal_device->exclusivePhysical[i]) {
                ljkOS_ReleaseReservedMemory(gal_device->os,
                                            gal_device->exclusivePhysical[i]);

                gal_device->exclusivePhysical[i] = ljvNULL;
            }

            if (gal_device->exclusiveLogical[i])
                gal_device->exclusiveLogical[i] = ljvNULL;

            if (gal_device->exclusiveVidMem[i]) {
                /* destroy the external heap */
                ljmkVERIFY_OK(ljkVIDMEM_Destroy(gal_device->exclusiveVidMem[i]));
                gal_device->exclusiveVidMem[i] = ljvNULL;
            }
        }

         /*
          * Destroy contiguous memory pool after ljkDEVICE destroyed. ljkDEVICE
          * may allocates GPU memory types from SYSTEM pool.
          */
        for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
            if (gal_device->contiguousPhysicals[i]) {
                if (!gal_device->requestedContiguousBases[i]) {
                    ljmkVERIFY_OK(_FreeMemory(gal_device,
                                              gal_device->contiguousLogicals[i],
                                              gal_device->contiguousPhysicals[i]));
                } else {
                    ljkOS_ReleaseReservedMemory(gal_device->os,
                                                gal_device->contiguousPhysicals[i]);

                    gal_device->contiguousPhysicals[i] = ljvNULL;
                    gal_device->requestedContiguousBases[i] = 0;
                    gal_device->requestedContiguousSizes[i] = 0;
                    gal_device->contiguousLogicals[i] = ljvNULL;
                    gal_device->contiguousPhysicals[i] = ljvNULL;
                }
            }

            if (gal_device->contiguousVidMems[i]) {
                /* Destroy the contiguous heap. */
                ljmkVERIFY_OK(ljkVIDMEM_Destroy(gal_device->contiguousVidMems[i]));
                gal_device->contiguousVidMems[i] = ljvNULL;
            }
        }

#if ljdENABLE_40BIT_VA
        if (gal_device->lowContiguousPhysical != ljvNULL) {
            ljkOS_ReleaseReservedMemory(gal_device->os,
                                        gal_device->lowContiguousPhysical);

            gal_device->lowContiguousPhysical = ljvNULL;
        }

        if (gal_device->lowContiguousVidMem != ljvNULL) {
            /* Destroy the contiguous heap. */
            ljmkVERIFY_OK(ljkVIDMEM_Destroy(gal_device->lowContiguousVidMem));
            gal_device->lowContiguousVidMem = ljvNULL;
        }
#endif

#if ljdENABLE_TRUST_APPLICATION
        if (gal_device->taos) {
            ljmkVERIFY_OK(ljtaOS_DestroyOS(gal_device->taos));
            gal_device->taos = ljvNULL;
        }
#endif

        /* Destroy the ljkOS object. */
        if (gal_device->os != ljvNULL) {
            ljmkVERIFY_OK(ljkOS_Destroy(gal_device->os));
            gal_device->os = ljvNULL;
        }

        /* Free the device. */
        kfree(gal_device);
    }

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

free_gal_device:
    kfree(gal_device);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

#if ljdENABLE_DEVFREQ
#include <../drivers/devfreq/governor.h>
static ljtUINT32 cur_freq = 64;
static int lj_df_target(struct device *dev, unsigned long *freq, u32 flags)
{
    ljtUINT32 i = 0;
    ljtUINT32 _freq = 1;
    ljkHARDWARE hardware;
    ljkDEVICE device = galDevice->devices[0];

    _freq = *freq;

    for (i = 0; i < ljvCORE_3D_MAX; i++) {
        if (device->kernels[i]) {
            hardware = device->kernels[i]->hardware;
            ljkHARDWARE_SetClock(hardware, _freq, _freq);
        }
    }

    cur_freq = _freq;

    return 0;
}

static ljeSTATUS lj_df_status(struct device *dev, struct devfreq_dev_status *stat)
{
    ljkHARDWARE Hardware = ljvNULL;
    ljtUINT32 load = 0;
    ljkDEVICE device = galDevice->devices[0];
    ljeSTATUS status = ljvSTATUS_OK;
    ljtUINT32 hi_total_cycle_count = 0, hi_total_idle_cycle_count = 0;

    ljmkHEADER();

    Hardware = device->kernels[0]->hardware;

    ljmkONERROR(ljkHARDWARE_QueryCycleCount(Hardware, &hi_total_cycle_count, &hi_total_idle_cycle_count));

    load = div_u64((ljtUINT64)(hi_total_cycle_count - hi_total_idle_cycle_count) * 100, hi_total_cycle_count);

    stat->current_frequency = (unsigned long)cur_freq;
    stat->busy_time = (unsigned long)load;
    stat->total_time = (unsigned long)100;

    ljmkONERROR(ljkHARDWARE_CleanCycleCount(Hardware));

OnError:
    ljmkFOOTER_NO();
    return status;
}

static int lj_df_get_cur_freq(struct device *dev, unsigned long *freq)
{
    *freq = cur_freq;
    return 0;
}

#define POLLING_MS        500
#define UP_THRESHOLD       90
#define DOWN_DIFFERENCTIAL 5

static struct devfreq *df;

static struct devfreq_simple_ondemand_data ljmcore_gov_data;

static struct devfreq_dev_profile lj_df_profile = {
    .polling_ms = POLLING_MS,
    .target = lj_df_target,
    .get_dev_status = lj_df_status,
    .get_cur_freq = lj_df_get_cur_freq,
};

static ljeSTATUS lj_df_governor_target(struct devfreq *df, unsigned long *freq)
{
    struct devfreq_dev_status *stat;
    unsigned long long a, b;

    ljtUINT32 up_threshold = UP_THRESHOLD;
    ljtUINT32 down_differential = DOWN_DIFFERENCTIAL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER();

    ljmkONERROR(devfreq_update_stats(df));

    stat = &df->last_status;

    /* Set MAX if it's busy enough */
    if (stat->busy_time * 100 >
        stat->total_time * up_threshold) {
        *freq = 64;
        goto OnError;
    }

    /* Set MAX if we do not know the initial frequency */
    if (stat->current_frequency == 0) {
        *freq = 64;
        goto OnError;
    }

    /* Keep the current frequency */
    if (stat->busy_time * 100 >
        stat->total_time * (up_threshold - down_differential)) {
        *freq = stat->current_frequency;
        goto OnError;
    }

    /* Set the desired frequency based on the load */
    a = stat->busy_time;
    a *= stat->current_frequency;
    b = div_u64(a, stat->total_time);
    b *= 100;
    b = div_u64(b, (up_threshold - down_differential / 2));

    if (b <= 1)
        *freq = 1;
    else if (b > 1 && b <= 2)
        *freq = 2;
    else if (b > 2 && b <= 4)
        *freq = 4;
    else if (b > 4 && b <= 8)
        *freq = 8;
    else if (b > 8 && b <= 16)
        *freq = 16;
    else if (b > 16 && b <= 32)
        *freq = 32;
    else if (b > 32 && b <= 48)
        *freq = 48;
    else
        *freq = 64;

OnError:
    ljmkFOOTER_NO();
    return status;
}

static int lj_df_governor_handler(struct devfreq *df, unsigned int event, void *data)
{
    switch (event) {
    case DEVFREQ_GOV_START:
        devfreq_monitor_start(df);
        break;

    case DEVFREQ_GOV_STOP:
        devfreq_monitor_stop(df);
        break;

    default:
        break;
    }

    return 0;
}

static struct devfreq_governor lj_df_governor = {
    .name = "ljmcore_gov",
    .get_target_freq = lj_df_governor_target,
    .event_handler = lj_df_governor_handler,
};

static ljeSTATUS
lj_df_init(ljkGALDEVICE galDevice)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtINT32 err = 0;
    ljtUINT32 i = 0;
    ljkHARDWARE Hardware;
    struct device *ljmcore_device = galDevice->devices[0]->dev;
    ljkDEVICE device = galDevice->devices[0];

    ljmkHEADER();

    dev_pm_opp_add(ljmcore_device, 1, 0);
    dev_pm_opp_add(ljmcore_device, 2, 0);
    dev_pm_opp_add(ljmcore_device, 4, 0);
    dev_pm_opp_add(ljmcore_device, 8, 0);
    dev_pm_opp_add(ljmcore_device, 16, 0);
    dev_pm_opp_add(ljmcore_device, 32, 0);
    dev_pm_opp_add(ljmcore_device, 48, 0);
    dev_pm_opp_add(ljmcore_device, 64, 0);

    err = devfreq_add_governor(&lj_df_governor);

    if (err) {
        ljmkPRINT("Failed to add governor: %d", err);
        return ljvSTATUS_NOT_SUPPORTED;
    }

    df = devm_devfreq_add_device(ljmcore_device, &lj_df_profile,
                                 "ljmcore_gov", &ljmcore_gov_data);
    if (IS_ERR(df)) {
        ljmkPRINT("Error: init devfreq %lx", (unsigned long)ljmcore_device);
        status = ljvSTATUS_NOT_SUPPORTED;
    }

    for (i = 0; i < ljvCORE_3D_MAX; i++) {
        if (device->kernels[i]) {
            Hardware = device->kernels[i]->hardware;

            if (Hardware->options.powerManagement) {
                ljmkONERROR(ljkHARDWARE_EnablePowerManagement(Hardware, ljvFALSE));

                ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_ON_AUTO));

                Hardware->options.powerManagement = ljvFALSE;
            }

            ljmkONERROR(ljkHARDWARE_SetGpuProfiler(Hardware, ljvTRUE));

            ljmkONERROR(ljkHARDWARE_InitProfiler(Hardware));

            Hardware->waitCount = 200 * 100;

            ljmkONERROR(ljkHARDWARE_CleanCycleCount(Hardware));
        }
    }

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

static ljeSTATUS
lj_df_exit(ljkGALDEVICE Device)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtINT32 err = 0;
    struct device *ljmcore_device = Device->devices[0]->dev;

    devm_devfreq_remove_device(ljmcore_device, df);

    err = devfreq_remove_governor(&lj_df_governor);

    if (err) {
        ljmkPRINT("%s %d: failed remove governor %d", __FUNCTION__, __LINE__,  err);
        status = ljvSTATUS_NOT_SUPPORTED;
    }

    return status;
}
#endif

/*******************************************************************************
 *
 *  ljkGALDEVICE_Start
 *
 *  Start the gal device, including the following actions: setup the isr routine
 *  and start the daemoni thread.
 *
 *  INPUT:
 *
 *      ljkGALDEVICE Device
 *          Pointer to an ljkGALDEVICE object.
 *
 *  OUTPUT:
 *
 *      Nothing.
 *
 *  RETURNS:
 *
 *      ljvSTATUS_OK
 *          Start successfully.
 */
ljeSTATUS
ljkGALDEVICE_Start(ljkGALDEVICE Device)
{
    ljtUINT i, devIndex;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkDEVICE device = ljvNULL;
    ljkKERNEL kernel = ljvNULL;

    ljmkHEADER_ARG("Device=%p", Device);

    for (devIndex = 0; devIndex < Device->args.devCount; devIndex++) {
        device = Device->devices[devIndex];

        /* Start the kernel threads. */
        for (i = 0; i < ljvCORE_COUNT; ++i) {
            if (i == ljvCORE_VG)
                continue;

            ljmkONERROR(_StartThread(device, i));

#if ljdENABLE_SW_PREEMPTION
            ljmkONERROR(_StartPreemptThread(device, i));
#endif
        }

        /* Power off needs to be after all the command start. */
        for (i = 0; i < ljvCORE_COUNT; i++) {
            kernel = device->kernels[i];

            if (kernel == ljvNULL)         /* 3d fpga/asic use this!!! */
            //if (kernel == ljvNULL || i == 0) /* only 2d fpga use this!!! */
                continue;

            if (!i && kernel->processPageTable &&
                (kernel->command->pool == ljvPOOL_VIRTUAL ||
                 !kernel->flatMapping))
                ljmkONERROR(ljkMMU_ConstructMmuCopy(kernel, &kernel->mmuCopy));

#if LEGACY_INT
            /* Setup the ISR routine. */
            ljmkONERROR(_SetupIsr(device, i));
#else
            if (i == ljvCORE_VG) {
                /* Setup the ISR routine. */
                ljmkONERROR(_SetupIsr(device, i));
            }
#endif

            if (i == ljvCORE_VG) {
#if ljdENABLE_VG
                /* Switch to SUSPEND power state. */
                ljmkONERROR(ljkVGHARDWARE_SetPowerState(kernel->vg->hardware,
                                                        ljvPOWER_OFF_BROADCAST));
#endif
            } else {
                /* Switch to SUSPEND power state. */
                ljmkONERROR(ljkHARDWARE_SetPowerState(kernel->hardware,
                                                      ljvPOWER_OFF_BROADCAST));
            }
        }

#if !LEGACY_INT
        /* request isr for 2D and 3D cores */
        ljmkONERROR(_SetupIsrEx(device));   /* by hym on 20220915 */
#endif
    }

#if ljdENABLE_DEVFREQ
    ljmkONERROR(lj_df_init(Device));
#endif

OnError:
    ljmkFOOTER();
    return status;
}

/*******************************************************************************
 *
 *  ljkGALDEVICE_Stop
 *
 *  Stop the gal device, including the following actions: stop the daemon
 *  thread, release the irq.
 *
 *  INPUT:
 *
 *      ljkGALDEVICE Device
 *          Pointer to an ljkGALDEVICE object.
 *
 *  OUTPUT:
 *
 *      Nothing.
 *
 *  RETURNS:
 *
 *      Nothing.
 */
ljeSTATUS
ljkGALDEVICE_Stop(ljkGALDEVICE Device)
{
    ljtUINT i, devIndex;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkDEVICE device = ljvNULL;
    ljkKERNEL kernel = ljvNULL;
    ///* ljtUINT32 pid    = _GetProcessID(); */

    ljmkHEADER_ARG("Device=%p", Device);

    ljmkVERIFY_ARGUMENT(Device != NULL);

    for (devIndex = 0; devIndex < Device->args.devCount; devIndex++) {
        device = Device->devices[devIndex];

        for (i = 0; i < ljvCORE_COUNT; i++) {
            kernel = device->kernels[i];
            if (!kernel)
                continue;

            //ljmkONERROR(ljkKERNEL_AttachProcessEx(kernel, ljvFALSE, kernel->os->kernelProcessID /* pid */));

            if (!Device->gotoShutdown) {
                if (i == ljvCORE_VG) {
#if ljdENABLE_VG
                    /* Switch to OFF power state. */
                    ljmkONERROR(ljkVGHARDWARE_SetPowerState(kernel->vg->hardware,
                                                            ljvPOWER_OFF));
#endif
                } else {
                    ljmkONERROR(ljkHARDWARE_EnablePowerManagement(kernel->hardware,
                                                                  ljvTRUE));

                    /* Switch to OFF power state. */
                    ljmkONERROR(ljkHARDWARE_SetPowerState(kernel->hardware,
                                                          ljvPOWER_OFF));
                }
            }

            if (!i && kernel->processPageTable &&
                (kernel->command->pool == ljvPOOL_VIRTUAL ||
                 !kernel->flatMapping))
                ljmkONERROR(ljkMMU_DestroyMmuCopy(kernel->mmuCopy));

            /* Stop the ISR routine. */
            ljmkONERROR(_ReleaseIsr(device, i));

            _StopThread(device, i);
#if ljdENABLE_SW_PREEMPTION
            _StopPreemptThread(device, i);
#endif
        }
    }

#if ljdENABLE_DEVFREQ
    ljmkONERROR(lj_df_exit(Device));
#endif

OnError:
    ljmkFOOTER();
    return status;
}

extern unsigned long ljm_get_vram_size_from_boardinfo(int card);

int gp_meminfo_get(int card, struct ljm_ioctl_info *pgpu_info)
{
    unsigned long vram_total = ljm_get_vram_size_from_boardinfo(card);
    unsigned long usable_total = galDevice-> externalSize[card];
    unsigned long usable_free = galDevice->externalVidMem[card]->freeBytes;
    long usable_used;

    if (galDevice->exclusiveVidMem[card]) {
        usable_total += galDevice->exclusiveSize[card];
        usable_free += galDevice->exclusiveVidMem[card]->freeBytes;
    }
    usable_used = usable_total - usable_free;

    pgpu_info->vram_total = vram_total / 1024 / 1024;
    pgpu_info->vram_used = usable_used / 1024 / 1024;
    pgpu_info->vram_free = pgpu_info->vram_total - pgpu_info->vram_used;

    return 0;
}

static ljeSTATUS ljkAllocate_ContinuousVirtual_Prior(ljkKERNEL Kernel, ljePOOL Pool,
                                            ljeVIDMEM_TYPE Type, ljtSIZE_T *Bytes,
                                            ljkVIDMEM_NODE *NodeObject)
{
    ljeSTATUS status = ljvSTATUS_OK;

    status = ljkVIDMEM_NODE_AllocateVirtual(Kernel, Pool, Type,
                                            ljvALLOC_FLAG_CONTIGUOUS,
                                            Bytes, NodeObject);

    if (status == ljvSTATUS_OK)
        return status;

    status = ljkVIDMEM_NODE_AllocateVirtual(Kernel, Pool, Type,
                                            ljvALLOC_FLAG_NON_CONTIGUOUS,
                                            Bytes, NodeObject);
    return status;
}

/*****************************************************************************
 *  函 数 名: ljkDma_Sync_Localoffset
 *  功能描述: 通过DMA同步方法备份显存到系统内存中
 *  输入参数: int card, ljtSIZE_T offset, ljtSIZE_T bytes
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         2. 日    期:   2023年1月7日
 *            作    者:   胡艳明、徐仕楠
 *            修改内容: 利用DMA搬移备份已知offset和bytes的显存数据
 *
 *****************************************************************************/
static ljeSTATUS ljkDma_Sync_Localoffset(int card, ljtSIZE_T offset, ljtSIZE_T bytes)
{
    ljkKERNEL           kernel   = _GetValidKernel(card, galDevice);
    ljsPLATFORM        *platform = kernel->os->device->platform;

    struct ljkVramNode *vramNode = ljvNULL;
    ljtPOINTER          pointer  = ljvNULL;
    ljeSTATUS           status   = ljvSTATUS_OK;
    ljtSIZE_T           tmp_bytes;
    ljsDMA_TRANS_INFO info;

    ljmkHEADER_ARG("card=%d", card);

    /* Allocate the ljkVramNode from the heap. */
    status = ljkOS_Allocate(kernel->os, ljmSIZEOF(struct ljkVramNode), &pointer);
    if (status != ljvSTATUS_OK) {
        ljmkPRINT("out of memory when suspend\n");
        goto OnError;
    }

    vramNode                = (struct ljkVramNode *)pointer;
    vramNode->next          = ljvNULL;
    vramNode->vramNode      = ljmINT2PTR(offset);
    vramNode->bytes         = bytes;
    vramNode->beenMirrored  = ljvFALSE;
    vramNode->useNodeOffset = ljvTRUE;

    tmp_bytes               = bytes;

    status = ljkAllocate_ContinuousVirtual_Prior(kernel, ljvPOOL_VIRTUAL, ljvVIDMEM_TYPE_GENERIC,
                                               &tmp_bytes, &vramNode->virtualNode);
    if (status != ljvSTATUS_OK) {
        ljmkPRINT("out of memory when allocate a virtual node for offset %lx(%lx)\n", offset, vramNode->bytes);
        ljkOS_Free(kernel->os, pointer);
        goto OnError;
    }

    if (tmp_bytes < vramNode->bytes) {
        ljmkPRINT("less memory is backup when suspend (%uz != %uz)\n", tmp_bytes, vramNode->bytes);
        vramNode->bytes = tmp_bytes;
    }

    info.src_node = vramNode->vramNode;
    info.dst_node = vramNode->virtualNode;
    info.offset = 0;
    info.bytes = vramNode->bytes;
    info.reason = ljvSYNC_MEMORY_DIRECTION_LOCALOFFSET_TO_SYSTEM;
    /* backup data of vram node to virtual node */
#if 0
    status = platform->ops->syncMemory(kernel, vramNode->virtualNode, vramNode->vramNode,
                                     ljvSYNC_MEMORY_DIRECTION_LOCALOFFSET_TO_SYSTEM, vramNode->bytes, 0);
#else
    status = platform->ops->dmaCopy(kernel, &info);
#endif
    if (status != ljvSTATUS_OK) {
        ljmkPRINT("synchronize vram to system memory with error\n");
        /* release virtual node */
        ljkVIDMEM_NODE_Dereference(kernel, vramNode->virtualNode);
        ljkOS_Free(kernel->os, pointer);
        goto OnError;
    }

    ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
        "%s(%d): vram offset %lx >> %p (%x) isMirrored=%d\n",
        __FUNCTION__, __LINE__,
        offset, vramNode->virtualNode,
        vramNode->bytes, vramNode->beenMirrored);

    if (galDevice->vramNodeList == ljvNULL) {
        galDevice->vramNodeList = galDevice->curVramNode = vramNode;
    }
    else {
        galDevice->curVramNode->next = vramNode;
        galDevice->curVramNode = vramNode;
    }
    ljmkFOOTER();
    return 0;

OnError:
    ljmkFOOTER_ARG("status=%d", status);
    return status;

}

#define BACKUP_VRAM_SIZE_DMA_THRESHOLD 0x400000

static ljeSTATUS ljkGpu_Combine_Block_Vram(int card, ljkVIDMEM vidmem, ljtSIZE_T offset_base)
{
    ljuVIDMEM_NODE_PTR node   = ljvNULL;
    ljtSIZE_T block_bytes, block_offset = 0;
    ljtINT i;
    ljeSTATUS status = ljvSTATUS_OK;

    if (vidmem == ljvNULL){
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    /* backup vram */
    for (i = 0; i < 8; i++) {
        if(vidmem->sentinel[i].VidMem.next == ljvNULL) /* 若sentinel数据为空，则不进行遍历 */
            continue;
        block_bytes = 0;         /* 确保每一个sentinel循环前block_bytes清零 */

        for (node = vidmem->sentinel[i].VidMem.next; node != &vidmem->sentinel[i]; node = node->VidMem.next) {
            /* free node */
            if (node->VidMem.nextFree != ljvNULL) {
                if (node->VidMem.bytes < 0x1000 && block_bytes > 0)  /* 过滤第一个node为free */
                    ; // block_bytes += node->VidMem.bytes;
                else if (block_bytes > 0) {
                    status = ljkDma_Sync_Localoffset(card, block_offset, block_bytes);
                    block_bytes = 0;
                }
                continue;/* 跳出空节点 */
            }

            if (block_bytes == 0)
                block_offset = node->VidMem.offset + offset_base;

            /* huym: 当存在占位字符时(padding bytes), block_bytes累加少算了占位字符 */
            // block_bytes     += node->VidMem.bytes;
            // ljmkASSERT(block_offset - offset_base == node->VidMem.offset + node->VidMem.bytes - block_bytes);
            block_bytes = node->VidMem.offset + offset_base - block_offset + node->VidMem.bytes;

            if (block_bytes > BACKUP_VRAM_SIZE_DMA_THRESHOLD) { /* 对bytes小于0x400000的块进行累加，以减少DMA传输次数                  */
                status = ljkDma_Sync_Localoffset(card, block_offset, block_bytes);
                block_bytes = 0; /* clean the tmp block */
            }
        }
        if (block_bytes > 0) { /* 备份尾端藏有小块node */
            status = ljkDma_Sync_Localoffset(card, block_offset, block_bytes);
            block_bytes = 0;
        }
    }

    return status;
}

#define WALK_DATABASE 1
/*****************************************************************************
 *  函 数 名: gpu_backup_exclusive_vram
 *  功能描述: 备份GPU不可见显存到系统内存中
 *  输入参数: int card
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年12月12日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int gpu_backup_exclusive_vram(int card)
{
    ljeSTATUS status        = ljvSTATUS_OK;
#if WALK_DATABASE
    ljkKERNEL       kernel   = _GetValidKernel(card, galDevice);
    ljsPLATFORM    *platform = kernel->os->device->platform;
    ljsDATABASE_PTR database;
    ljtINT          i,j;

    ljsDATABASE_RECORD_PTR record;
    struct ljkVramNode *vramNode = ljvNULL;
    ljsDMA_TRANS_INFO info;

    ljtBOOL dbAcquired      = ljvFALSE;
    ljtBOOL handleAcquired  = ljvFALSE;

    /* Acquire the database mutex. */
    ljmkVERIFY_OK(ljkOS_AcquireMutex(kernel->os, kernel->db->dbMutex, ljvINFINITE));
    dbAcquired = ljvTRUE;
    /* Walk all the databases to find video memory */
    for (i = 0; i < ljmCOUNTOF(kernel->db->db); ++i) {
        /* Walk the databases in the current slot */
        for (database = kernel->db->db[i]; database != ljvNULL; database = database->next) {
            /* Acquire the handle database muxtex */
            ljmkVERIFY_OK(ljkOS_AcquireMutex(kernel->os, database->handleDatabaseMutex, ljvINFINITE));
            handleAcquired = ljvTRUE;

            /* Walk the record lists */
            for (j=0; j < ljmCOUNTOF(database->list); ++j) {
                /* Walk the records in the current slot */
                for (record = database->list[j]; record != ljvNULL; record = record->next) {
                        /* Skip if it is not video memory */
                    if ((record->type & ljdDATABASE_TYPE_MASK) != ljvDB_VIDEO_MEMORY)
                        continue;

                    /* vram node */
                    {
                        ljtPOINTER       pointer      = ljvNULL;
                        ljkVIDMEM_HANDLE handleObject = ljvNULL;
                        ljtSIZE_T        bytes        = record->bytes;
//                        ljtSIZE_T        tmp_bytes    = record->bytes;

                        ljmkONERROR(ljkKERNEL_QueryIntegerId(database->handleDatabase, ljmPTR2INT32(record->data),
                                                                (ljtPOINTER *)&handleObject));

                        /* check node pool */
                        if (!(handleObject->node->pool == ljvPOOL_LOCAL_EXCLUSIVE))
                            continue;

                        /* Allocate the ljkVramNode from the heap. */
                        status = ljkOS_Allocate(kernel->os, ljmSIZEOF(struct ljkVramNode), &pointer);
                        if (status != ljvSTATUS_OK) {
                            ljmkPRINT("gpu_backup_exclusive_vram: out of memory when allocate a node (%d)\n", ljmPTR2INT32(record->data));

                            if (handleAcquired)
                                ljkOS_ReleaseMutex(kernel->os, database->handleDatabaseMutex);
                            if (dbAcquired)
                                ljkOS_ReleaseMutex(kernel->os, kernel->db->dbMutex);
                            goto OnError;
                        }

                        /* reference vram node */
                        //ljkVIDMEM_NODE_Reference(kernel, handleObject->node);

                        vramNode                = (struct ljkVramNode *)pointer;
                        vramNode->next          = ljvNULL;
                        vramNode->vramNode      = handleObject->node;
                        vramNode->bytes         = bytes;
                        vramNode->useNodeOffset = ljvFALSE;
                        if (handleObject->node->mirror.mirrorNode == ljvNULL)
                        {
#if 0
                            vramNode->beenMirrored = ljvFALSE;
                            status = ljkAllocate_ContinuousVirtual_Prior(kernel, ljvPOOL_VIRTUAL, ljvSURF_TYPE_UNKNOWN,
                                                                        &tmp_bytes, &vramNode->virtualNode);
                            if (status != ljvSTATUS_OK ){
                                ljmkPRINT("gpu_backup_exclusive_vram: Error allocate virtual node\n");
                                continue;
                            }
#endif
                            continue;
                        } else {
                            vramNode->beenMirrored = ljvTRUE;
                            vramNode->virtualNode = handleObject->node->mirror.mirrorNode;
                        }

                        ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                            "%s(%d): node %d (%p >> %p)(%x) isMirrored=%d\n",
                            __FUNCTION__, __LINE__,
                            ljmPTR2INT32(record->data), vramNode->vramNode, vramNode->virtualNode,
                            vramNode->bytes, vramNode->beenMirrored);

                        if (galDevice->vramNodeList == ljvNULL) {
                            galDevice->vramNodeList = galDevice->curVramNode = vramNode;
                        }
                        else {
                            galDevice->curVramNode->next = vramNode;
                            galDevice->curVramNode = vramNode;
                        }

                        info.src_node = vramNode->vramNode;
                        info.dst_node = vramNode->virtualNode;
                        info.offset = 0;
                        info.bytes = vramNode->bytes;
                        info.reason = ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM;
                        /* backup data of vram node to virtual node */
                        //ljmkONERROR(platform->ops->syncMemory(kernel, vramNode->virtualNode, vramNode->vramNode,
                        //            ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM, vramNode->bytes, 0));
                        ljmkONERROR(platform->ops->dmaCopy(kernel, &info));
                    }
                }
            }

            /* Release the handle database mutex. */
            ljmkVERIFY_OK(ljkOS_ReleaseMutex(kernel->os, database->handleDatabaseMutex));
            handleAcquired = ljvFALSE;
        }
    }

    /* Release the database mutex. */
    ljmkVERIFY_OK(ljkOS_ReleaseMutex(kernel->os, kernel->db->dbMutex));
    dbAcquired = ljvFALSE;
#else
    ljkVIDMEM exclusive_vidmem = galDevice->exclusiveVidMem[card];/* 获取总不可见显存 */

    if (exclusive_vidmem ) {
        /* backup exclusive vram */
        status = ljkGpu_Combine_Block_Vram(card, exclusive_vidmem, galDevice->exclusiveVidMem[card]->physicalBase - VRAM_GP200_BUS40_GPU_START);
        if (status != ljvSTATUS_OK) {
            goto OnError;
        }
    }
#endif
    return 0;

OnError:
    ljmkPRINT("backup exclusive vram with error\n");
    return (int)status;
}


/*****************************************************************************
 *  函 数 名: gpu_backup_external_vram
 *  功能描述: 备份GPU可见显存到系统内存中
 *  输入参数: int card
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年11月29日
 *            作    者: 胡艳明
 *            修改内容: 通过DMA搬移备份显存数据
 *
 *****************************************************************************/
int gpu_backup_external_vram(int card)
{
    ljkVIDMEM external_vidmem = galDevice->externalVidMem[card];/* 获取总可见显存 */
    ljeSTATUS status          = ljvSTATUS_OK;

    /* backup external vram */
    status = ljkGpu_Combine_Block_Vram(card, external_vidmem, 0);   /* offset base is 0 for external vram */

    if (status == ljvSTATUS_OK) {
        struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)galDevice->platform;

        /* backup fix vram for VPU */
        status = ljkDma_Sync_Localoffset(card,
            pcie_platform->pcie_info[card].bar[0].size - VRAM_FIX_PCI_DMA_SIZE - VRAM_FIX_INST_SIZE - VRAM_FIX_VPU_SIZE,
            VRAM_FIX_VPU_SIZE + VRAM_FIX_INST_SIZE);
    }

    return (int)status;
}

/*****************************************************************************
 *  函 数 名: gpu_restore_vram
 *  功能描述: 从系统主存恢复数据到GPU显存中
 *  输入参数: int card
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年11月29日
 *            作    者: 胡艳明
 *            修改内容: 通过DMA搬移恢复显存数据
 *
 *****************************************************************************/
int gpu_restore_vram(int card)
{
    ljkKERNEL   kernel = _GetValidKernel(card, galDevice);
    ljsPLATFORM *platform = kernel->os->device->platform;
    ljsDMA_TRANS_INFO info;

    struct ljkVramNode *vramNode, *vramNodeOld;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("card=%d", card);

    /* restore all vram */
    for (vramNode = galDevice->vramNodeList, galDevice->vramNodeList = ljvNULL;
        vramNode != ljvNULL;
        ) {
        if (! vramNode->useNodeOffset) {
            info.src_node = vramNode->virtualNode;
            info.dst_node = vramNode->vramNode;
            info.offset = 0;
            info.bytes = vramNode->bytes;
            info.reason = ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL;
            /* restore data of vram node from virtual node */
            //status = platform->ops->syncMemory(kernel, vramNode->vramNode, vramNode->virtualNode,
            //                          ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL, vramNode->bytes, 0);
            status = platform->ops->dmaCopy(kernel, &info);

            if (status != ljvSTATUS_OK) {
                ljmkPRINT("synchronize to exclusive vram with error (%p >> %p)\n",
                vramNode->virtualNode, vramNode->vramNode);
            }

            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                "%s(%d): (%p >> %p)(%x) isMirrored=%d\n",
                __FUNCTION__, __LINE__,
                vramNode->virtualNode, vramNode->vramNode,
                vramNode->bytes, vramNode->beenMirrored);
        } else {
            info.src_node = vramNode->virtualNode;
            info.dst_node = vramNode->vramNode;
            info.offset = 0;
            info.bytes = vramNode->bytes;
            info.reason = ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCALOFFSET;
            /* restore data of vram node from virtual node */
            //status = platform->ops->syncMemory(kernel, vramNode->vramNode, vramNode->virtualNode,
            //                          ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCALOFFSET, vramNode->bytes, 0);
            status = platform->ops->dmaCopy(kernel, &info);

            if (status != ljvSTATUS_OK) {
                ljmkPRINT("synchronize to vram with error (%p >> %lx)\n",
                vramNode->virtualNode, ljmPTR2INT32(vramNode->vramNode));
            }

            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                "%s(%d): (%p >> %lx)(%x) isMirrored=%d\n",
                __FUNCTION__, __LINE__,
                vramNode->virtualNode, ljmPTR2INT32(vramNode->vramNode),
                vramNode->bytes, vramNode->beenMirrored);
        }

        if (! vramNode->beenMirrored) {
            /* release virtual node */
            ljkVIDMEM_NODE_Dereference(kernel, vramNode->virtualNode);
        }

        vramNodeOld = vramNode;
        vramNode = vramNode->next;
        ljkOS_Free(kernel->os, vramNodeOld);
    }

    ljmkFOOTER();
    return 0;
}

int gpu_config_power_management(int core, bool enable)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkKERNEL kernel;
    int i;

    ljmkHEADER();

    //printk("gpu power management %s\n", enable ? "true" : "false");

    for (i=0; i<ljvCORE_COUNT; i++) {
        kernel = galDevice->devices[0]->kernels[i];
        if (kernel) {
            ljmkONERROR(ljkHARDWARE_EnablePowerManagement(kernel->hardware, enable));

            if (enable == ljvFALSE)
                ljmkONERROR(ljkHARDWARE_SetPowerState(kernel->hardware, ljvPOWER_ON));
        }
    }

    ljmkFOOTER_NO();
    return (int)ljvSTATUS_OK;

OnError:
    ljmkFOOTER();
    return (int)status;
}

void ljkVIDMEM_Node_DumpFree(ljkVIDMEM Memory);

void gpu_dump_external_free(void)
{
    ljkVIDMEM_Node_DumpFree(galDevice->externalVidMem[0]);
}

void gpu_dump_exclusive_free(void)
{
    ljkVIDMEM_Node_DumpFree(galDevice->exclusiveVidMem[0]);
}

ljeSTATUS
ljkVIDMEM_NODE_GetDmaAddress(ljkKERNEL Kernel, ljkVIDMEM_NODE NodeObject,
                             ljtSIZE_T Offset, ljtPHYS_ADDR_T *DmaAddress)
{
    ljeSTATUS status;
    ljuVIDMEM_NODE_PTR node = ljvNULL;
    ljkVIDMEM_BLOCK vidMemBlock = ljvNULL;
    ljkVIDMEM memory = ljvNULL;
    PLINUX_MDL mdl = ljvNULL;
    ljkALLOCATOR allocator = ljvNULL;
    ljmkHEADER_ARG("NodeObject=%p", NodeObject);

    node = NodeObject->node;

    if (!node)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    memory = node->VidMem.parent;
    vidMemBlock = node->VirtualChunk.parent;

    if (memory && memory->object.type == ljvOBJ_VIDMEM) {
        if (Offset >= node->VidMem.bytes) {
            /* Exceeds node size. */
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        *DmaAddress = node->VidMem.parent->physicalBase +
                           node->VidMem.offset + Offset;
    } else if (vidMemBlock && vidMemBlock->object.type == ljvOBJ_VIDMEM_BLOCK) {
        if (Offset >= node->VirtualChunk.bytes) {
            /* Exceeds node size. */
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
        mdl = (PLINUX_MDL)vidMemBlock->physical;
        allocator = mdl->allocator;

        ljmkONERROR(allocator->ops->GetDmaAddress(allocator, mdl, Offset, DmaAddress));
    } else {
        if (Offset >= node->VidMem.bytes) {
            /* Exceeds node size. */
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
        mdl = (PLINUX_MDL)node->VidMem.physical;
        allocator = mdl->allocator;

        ljmkONERROR(allocator->ops->GetDmaAddress(allocator, mdl, Offset, DmaAddress));
    }

    ljmkFOOTER_ARG("*PhysicalAddress=0x%llx", *DmaAddress);
    return ljvSTATUS_OK;

OnError:
    ljmkFOOTER();
    return status;
}
