/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_device_h_
#define __lj_hal_kernel_device_h_

#include "lj_hal_kernel_debugfs.h"

#if ljdENABLE_TTM
#include "lj_hal_kernel_ttm.h"
#endif

#if ljdENABLE_TRUST_APPLICATION
#include "lj_hal_ta.h"

extern ljTA             globalTA[ljvCORE_COUNT];
#endif

/*******************************************************************************
 ************************* ljkGALDEVICE Structure ******************************
 *******************************************************************************/

struct ljkVramNode {
    struct ljkVramNode *next;
    ljtPOINTER vramNode;        /* vram node or offset in video memory */
    ljkVIDMEM_NODE virtualNode; /* virtual mem node in system ram */
    ljtUINT64 bytes;            /* size of node */
    ljtBOOL beenMirrored;       /* vramNode has been mirrored */
    ljtBOOL useNodeOffset;      /*use node or offset of video memory*/
};

typedef struct _ljkGALDEVICE {
    /* Objects. */
    ljkOS               os;

    ljsPLATFORM         *platform;

    /* Attributes. */
    ljtPHYS_ADDR_T      internalBase;
    ljtSIZE_T           internalSize;
    ljtPHYS_ADDR        internalPhysical;
    ljtUINT32           internalPhysName;
    ljtPOINTER          internalLogical;
    ljkVIDMEM           internalVidMem;

    ljtPHYS_ADDR_T      externalBase[ljdDEVICE_COUNT];
    ljtSIZE_T           externalSize[ljdDEVICE_COUNT];
    ljtPHYS_ADDR        externalPhysical[ljdDEVICE_COUNT];
    ljtUINT32           externalPhysName[ljdDEVICE_COUNT];
    ljkVIDMEM           externalVidMem[ljdDEVICE_COUNT];

    /* Shared external SRAMs. */
    ljtPHYS_ADDR_T      extSRAMBases[ljvSRAM_EXT_COUNT];
    ljtSIZE_T           extSRAMSizes[ljvSRAM_EXT_COUNT];
    ljtPHYS_ADDR        extSRAMPhysicals[ljvSRAM_EXT_COUNT];
    ljkVIDMEM           extSRAMVidMems[ljvSRAM_EXT_COUNT];

    ljtPHYS_ADDR_T      contiguousBases[ljdSYSTEM_RESERVE_COUNT];
    ljtSIZE_T           contiguousSizes[ljdSYSTEM_RESERVE_COUNT];
    ljtPHYS_ADDR        contiguousPhysicals[ljdSYSTEM_RESERVE_COUNT];
    ljtUINT32           contiguousPhysNames[ljdSYSTEM_RESERVE_COUNT];
    ljtPOINTER          contiguousLogicals[ljdSYSTEM_RESERVE_COUNT];
    ljkVIDMEM           contiguousVidMems[ljdSYSTEM_RESERVE_COUNT];

    ljtPHYS_ADDR_T      requestedContiguousBases[ljdSYSTEM_RESERVE_COUNT];
    ljtSIZE_T           requestedContiguousSizes[ljdSYSTEM_RESERVE_COUNT];


    ljtPHYS_ADDR_T      lowContiguousBase;
    ljtSIZE_T           lowContiguousSize;
    ljtPHYS_ADDR        lowContiguousPhysical;
    ljtUINT32           lowContiguousPhysName;
    ljkVIDMEM           lowContiguousVidMem;

    ljtPHYS_ADDR_T      exclusiveBase[ljdDEVICE_COUNT];
    ljtSIZE_T           exclusiveSize[ljdDEVICE_COUNT];
    ljtPHYS_ADDR        exclusivePhysical[ljdDEVICE_COUNT];
    ljtUINT32           exclusivePhysName[ljdDEVICE_COUNT];
    ljtPOINTER          exclusiveLogical[ljdDEVICE_COUNT];
    ljkVIDMEM           exclusiveVidMem[ljdDEVICE_COUNT];

    ljsDEBUGFS_DIR      debugfsDir;

    ljkDEVICE           devices[ljdDEVICE_COUNT];
    ljtINT32            devIDs[ljdDEVICE_COUNT];
    ljtUINT             devCount;

    ljsMODULE_PARAMETERS args;

#if ljdENABLE_TRUST_APPLICATION
    /* ljtsOs object for trust application. */
    ljtaOS              taos;
#endif

    /* If we are going to shut down. */
    ljtBOOL             gotoShutdown;

#if ljdENABLE_DRM
    void                *drm;
    void *              ljm_dev;

    /* add by hym on 20230112 to handle first line scan out error */
    ljkVIDMEM_NODE      scanoutNodeObjects[4][2];
    uint32_t            scanoutNodeHandles[4][2];
#if ljdENABLE_TTM
    ljm_ttm_t            ttm;
#endif
#endif

    /* backup external vram for suspend */
    struct ljkVramNode *vramNodeList;
    struct ljkVramNode *curVramNode;
} *ljkGALDEVICE;

typedef struct _ljsHAL_PRIVATE_DATA {
    ljkGALDEVICE        device;
    /*
     * 'fput' schedules actual work in '__fput' in a different thread.
     * So the process opens the device may not be the same as the one that
     * closes it.
     */
    ljtUINT32           pidOpen;
    ljtBOOL             isLocked;
    ljtUINT32           devIndex;
} ljsHAL_PRIVATE_DATA, *ljsHAL_PRIVATE_DATA_PTR;

ljeSTATUS
ljkGALDEVICE_Start(ljkGALDEVICE Device);

ljeSTATUS
ljkGALDEVICE_Stop(ljkGALDEVICE Device);

ljeSTATUS
ljkGALDEVICE_Construct(ljsPLATFORM *Platform, const ljsMODULE_PARAMETERS *Args, ljkGALDEVICE *Device);

ljeSTATUS
ljkGALDEVICE_Destroy(ljkGALDEVICE Device);

static ljmkINLINE ljkKERNEL
_GetValidKernel(int card, ljkGALDEVICE Device)
{
    ljkDEVICE device;

    if (card < 0 || card >= ljdDEVICE_COUNT)
        card = 0;
    device = (Device->devices[card] == ljvNULL) ? Device->devices[0] : Device->devices[card];

    if (device->kernels[ljvCORE_MAJOR]) {
        return device->kernels[ljvCORE_MAJOR];
    } else if (device->kernels[ljvCORE_2D]) {
        return device->kernels[ljvCORE_2D];
    } else if (device->kernels[ljvCORE_VG]) {
        return device->kernels[ljvCORE_VG];
    } else {
        ljmkASSERT(ljvFALSE);
        return ljvNULL;
    }
}

#define MAX_SCANOUT_NODE_SIZE   (32<<20)

#endif /* __lj_hal_kernel_device_h_ */
