/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_dma_engine_h_
#define __lj_hal_kernel_dma_engine_h_
#include <linux/device.h>
#include <linux/err.h>
#include <linux/scatterlist.h>
#include "lj_hal.h"

typedef struct _ljsDMA_DEVICE ljsDMA_DEVICE;

typedef enum _ljeDMA_STATUS
{
    ljeDMA_STATUS_COMPLETE,
    ljeDMA_STATUS_IN_PROGRESS,
    ljeDMA_STATUS_PAUSED,
    ljeDMA_STATUS_ERROR,
}
ljeDMA_STATUS;

typedef enum _ljeDMA_TX_TYPE
{
    ljeDMA_TX_TYPE_MEMCPY,
    ljeDMA_TX_TYPE_XOR,
    ljeDMA_TX_TYPE_PQ,
    ljeDMA_TX_TYPE_XOR_VAL,
    ljeDMA_TX_TYPE_PQ_VAL,
    ljeDMA_TX_TYPE_MEMSET,
    ljeDMA_TX_TYPE_MEMSET_SG,
    ljeDMA_TX_TYPE_INTERRUPT,
    ljeDMA_TX_TYPE_PRIVATE,
    ljeDMA_TX_TYPE_ASYNC_TX,
    ljeDMA_TX_TYPE_SLAVE,
    ljeDMA_TX_TYPE_CYCLIC,
    ljeDMA_TX_TYPE_INTERLEAVE,

    ljeDMA_TX_TYPE_END,
}
ljeDMA_TX_TYPE;

typedef union _ljuDMA_CAP_MASK
{
    ljtUINT32 dmaCapMaskMemcpy:1;
    ljtUINT32 dmaCapMaskXOR:1;
    ljtUINT32 dmaCapMaskPQ:1;
    ljtUINT32 dmaCapMaskXORVal:1;
    ljtUINT32 dmaCapMaskPQVal:1;
    ljtUINT32 dmaCapMaskMemset:1;
    ljtUINT32 dmaCapMaskMemsetSG:1;
    ljtUINT32 dmaCapMaskInterrupt:1;
    ljtUINT32 dmaCapMaskPrivate:1;
    ljtUINT32 dmaCapMaskAsyncTX:1;
    ljtUINT32 dmaCapMaskSlave:1;
    ljtUINT32 dmaCapMaskCyclic:1;
    ljtUINT32 dmaCapMaskInterleave:1;
    ljtUINT32 dmaCapMaskReserve:19;

    ljtUINT32 dmaCapMasks;
}
ljuDMA_CAP_MASK;

typedef enum _ljeDMA_TX_RESULT
{
    ljeDMA_TX_RESULT_OK = 0,
    ljeDMA_TX_RESULT_READ_FAILED,
    ljeDMA_TX_RESULT_WRITE_FAILED,
    ljeDMA_TX_RESULT_ABORTED,
}
ljeDMA_TX_RESULT;

typedef struct _ljsDMA_RESULT
{
    ljeDMA_TX_RESULT Result;
    ljtUINT32 Residue;
}
ljsDMA_RESULT;

typedef enum _ljeDMA_ALIGNMENT
{
    ljeDMA_ALIGNMENT_NONE = 0,
    ljeDMA_ALIGNMENT_1_BYTE,
    ljeDMA_ALIGNMENT_2_BYTES,
    ljeDMA_ALIGNMENT_4_BYTES,
    ljeDMA_ALIGNMENT_8_BYTES,
    ljeDMA_ALIGNMENT_16_BYTES,
    ljeDMA_ALIGNMENT_32_BYTES,
    ljeDMA_ALIGNMENT_64_BYTES,
}
ljeDMA_ALIGNMENT;

typedef enum _ljeDMA_TX_DIRECTION
{
    ljeDMA_TX_DIRECTION_MEM_TO_MEM = 0,
    ljeDMA_TX_DIRECTION_MEM_TO_DEV,
    ljeDMA_TX_DIRECTION_DEV_TO_MEM,
    ljeDMA_TX_DIRECTION_DEV_TO_DEV,
    ljeDMA_TX_DIRECTION_TRANS_NONE,
}
ljeDMA_TX_DIRECTION;

typedef enum _ljeDMA_RESIDUE_GRANULARITY
{
    ljeDMA_RESIDUE_GRANULARITY_DESCRIPTOR = 0,
    ljeDMA_RESIDUE_GRANULARITY_SEGMENT,
    ljeDMA_RESIDUE_GRANULARITY_BURST,
}
ljeDMA_RESIDUE_GRANULARITY;

typedef enum _ljeDMA_SLAVE_BUS_WIDTH
{
    ljeDMA_SLAVE_BUS_WIDTH_NONE = 0,
    ljeDMA_SLAVE_BUS_WIDTH_1BYTE = 1,
    ljeDMA_SLAVE_BUS_WIDTH_2BYTES = 2,
    ljeDMA_SLAVE_BUS_WIDTH_3BYTES = 3,
    ljeDMA_SLAVE_BUS_WIDTH_4BYTES = 4,
    ljeDMA_SLAVE_BUS_WIDTH_8BYTES = 8,
}
ljeDMA_SLAVE_BUS_WIDTH;

typedef ljtINT32 ljdDMA_TX_IDENTIFIER;

#define ljdDMA_MINI_MARK        1

typedef struct _ljsDMA_CHANNEL
{
    ljsDMA_DEVICE *device;

    ljtINT id;
    ljdDMA_TX_IDENTIFIER identifier;
    ljdDMA_TX_IDENTIFIER completedIdentifier;
    struct list_head devNode;
    ljtUINT32 clientCount;
    ljtPOINTER Private;
}
ljsDMA_CHANNEL;

typedef struct _ljsDMA_SLAVE_CAPS
{
    ljtUINT32 srcAddrWidth;
    ljtUINT32 dstAddrWidth;
    ljtUINT32 directions;
    ljtUINT32 copyAlign;

    ljtBOOL canPause;
    ljtBOOL canResume;
    ljtBOOL canTerminate;
    ljtBOOL canDescriptorReuse;

    ljeDMA_RESIDUE_GRANULARITY residueGranularity;
}
ljsDMA_SLAVE_CAPS;

typedef struct _ljsDMA_SLAVE_CONFIG
{
    ljeDMA_TX_DIRECTION direction;
    phys_addr_t srcAddr;
    phys_addr_t dstAddr;
    ljeDMA_SLAVE_BUS_WIDTH srcAddrBusWidth;
    ljeDMA_SLAVE_BUS_WIDTH dstAddrBusWidth;
}
ljsDMA_SLAVE_CONFIG;


typedef ljtBOOL (*ljkDMAENGINE_FilterFunc)(ljsDMA_CHANNEL *Chan, ljtPOINTER FilterParam);
typedef void (*ljkDMAENGINE_TxCallback)(ljtPOINTER TxCallbackParam);
typedef void (*ljkDMAENGINE_TxCallbackResult)(ljtPOINTER TxCallbackParam, ljsDMA_RESULT *Result);

typedef struct _ljsDMAENGINE_DESC_CALLBACK
{
    ljkDMAENGINE_TxCallback ljkDMA_TxCallback;
    ljkDMAENGINE_TxCallbackResult ljkDMA_TxCallbackResult;
    void *txCallbackParam;
}
ljsDMAENGINE_DESC_CALLBACK;

typedef struct _ljsDMA_TX_STATE
{
    ljdDMA_TX_IDENTIFIER lastCompleted;
    ljdDMA_TX_IDENTIFIER lastIssued;
    ljtUINT32 residue;
}
ljsDMA_TX_STATE;

typedef struct _ljsDMA_TX_DESCRIPTOR
{
    ljdDMA_TX_IDENTIFIER identifier;
    ljsDMA_CHANNEL *chan;
    ljtUINT64 flags;
    ljdDMA_TX_IDENTIFIER (*ljkDMA_TxSubmit)(struct _ljsDMA_TX_DESCRIPTOR *TxDesc);

    ljkDMAENGINE_TxCallback ljkDMA_TxCallback;
    ljkDMAENGINE_TxCallbackResult ljkDMA_TxCallbackResult;
    ljtPOINTER txCallbackParam;
}
ljsDMA_TX_DESCRIPTOR;

typedef struct _ljsDMA_SLAVE_MAP
{
    const ljtSTRING devName;
    const ljtSTRING slave;
    void *param;
}
ljsDMA_SLAVE_MAP;

typedef struct _ljsDMA_FILTER
{
    ljkDMAENGINE_FilterFunc func;
    ljtUINT mapCount;
    const ljsDMA_SLAVE_MAP *maps;
}
ljsDMA_FILTER;

struct _ljsDMA_DEVICE
{
    struct device       *dev;
    ljtUINT             chanCount;
    ljtUINT             chanCountRequired;

    ljuDMA_CAP_MASK     devCapMasks;
    ljeDMA_ALIGNMENT    copyAlign;

    struct list_head    chanHead;
    struct list_head    glabolDevNode;

    ljtUINT             id;
    ljtCHAR             name[20];

    ljsDMA_SLAVE_CAPS   caps;
    ljsDMA_FILTER       filter;

    ljeSTATUS (*ljkDMADEV_AllocateChannelResource)(ljsDMA_CHANNEL *Chan);
    void (*ljkDMADEV_FreeChannelResource)(ljsDMA_CHANNEL *Chan);

    ljsDMA_TX_DESCRIPTOR* (*ljkDMADEV_PrepareDmaMemcpy)(ljsDMA_CHANNEL *Chan,
                                              dma_addr_t Src,
                                              dma_addr_t Dst,
                                              ljtSIZE_T Len,
                                              ljtUINT32 Flag);

    ljsDMA_TX_DESCRIPTOR* (*ljkDMADEV_PrepareSlaveSG)(ljsDMA_CHANNEL *Chan,
                                struct scatterlist *SGList , ljtSIZE_T SGLen,
                                ljeDMA_TX_DIRECTION Dir,
                                ljtUINT32 Flag, ljtPOINTER Context);

    ljeSTATUS (*ljkDMADEV_Config)(ljsDMA_CHANNEL *Chan, ljsDMA_SLAVE_CONFIG *Config);

    ljeSTATUS (*ljkDMADEV_Pause)(ljsDMA_CHANNEL *Chan);

    ljeSTATUS (*ljkDAMDEV_Resume)(ljsDMA_CHANNEL *Chan);

    ljeSTATUS (*ljkDMADEV_TerminateAll)(ljsDMA_CHANNEL *Chan);

    void (*ljkDMADEV_Synchronize)(ljsDMA_CHANNEL *Chan);

    ljeDMA_STATUS (*ljkDMADEV_TxStatus)(ljsDMA_CHANNEL *Chan,
                                        ljdDMA_TX_IDENTIFIER Identifier,
                                        ljsDMA_TX_STATE *TxState);

    void (*ljkDMADEV_IssuePending)(ljsDMA_CHANNEL *Chan);
};

ljeDMA_STATUS ljkDMA_SyncWait(
    ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljtSIZE_T Timeout
    );

ljeSTATUS ljkDMA_IssuePendingAll(void);

ljeSTATUS ljkDMA_GetSlaveCaps(ljsDMA_CHANNEL *Chan, ljsDMA_SLAVE_CAPS *Caps);

ljeSTATUS ljkDMA_RequestSpecialSlaveChannel(ljsDMA_CHANNEL *Chan);

ljeSTATUS ljkDMA_RequestAnySlaveChannel(ljsDMA_DEVICE *dmaDev, ljsDMA_CHANNEL **chan);

ljeSTATUS ljkDMA_RequestChannel(
    const ljuDMA_CAP_MASK *mask,
    ljkDMAENGINE_FilterFunc func,
    void *funcParam,
    ljsDMA_CHANNEL **chan
    );

ljeSTATUS ljkDMA_RequestSlaveChannel(const ljtSTRING devName, const ljtSTRING Slave, ljsDMA_CHANNEL **chan);

ljeSTATUS ljkDMA_RequestChannelByMask(const ljuDMA_CAP_MASK *mask, ljsDMA_CHANNEL **chan);

void ljkDMA_ReleaseChannel(ljsDMA_CHANNEL *Chan);

void ljkDMAENGINE_Get(void);

void ljkDMAENGINE_Put(void);

ljeSTATUS ljkDMA_DeviceRegister(ljsDMA_DEVICE *dmaDev);

void ljkDMA_DeviceUnregister(ljsDMA_DEVICE *dmaDev);

void ljkDMA_TxDescriptorInit(ljsDMA_TX_DESCRIPTOR *Desc, ljsDMA_CHANNEL *Chan);

ljeDMA_STATUS ljkDMA_WaitForTransaction(ljsDMA_TX_DESCRIPTOR *Desc);

void ljkDMA_IdentifierInit(ljsDMA_CHANNEL *Chan);

ljdDMA_TX_IDENTIFIER ljkDMA_IdentifierAssign(ljsDMA_TX_DESCRIPTOR *Desc);

void ljkDMA_IdentifierComplete(ljsDMA_TX_DESCRIPTOR *Desc);

ljeDMA_STATUS ljkDMA_TxIsComplete(ljdDMA_TX_IDENTIFIER Identifier,
    ljdDMA_TX_IDENTIFIER lastCompleted, ljdDMA_TX_IDENTIFIER lastIssued);

ljeDMA_STATUS ljkDMA_IdentifierStatus(ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier, ljsDMA_TX_STATE *txState);

ljeDMA_STATUS ljkDMA_GetStatus(ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljdDMA_TX_IDENTIFIER *Completed,
    ljdDMA_TX_IDENTIFIER *Issued
    );

void ljkDMA_SetTxState(ljsDMA_TX_STATE *txState,
    ljdDMA_TX_IDENTIFIER lastCompleted, ljdDMA_TX_IDENTIFIER lastIssued, ljtUINT32 Residue);


ljtBOOL ljkDMA_SubmitError(ljdDMA_TX_IDENTIFIER Identifier);


void ljkDMA_SetResidue(ljsDMA_TX_STATE *txState, ljtUINT32 Residue);

ljtCHAR* ljkDMA_GetDeviceName(ljsDMA_DEVICE *Dev);

ljtINT ljkDMA_GetDeviceId(ljsDMA_DEVICE *Dev);

ljtINT ljkDMA_GetChannelId(ljsDMA_CHANNEL *Chan);

void ljkDMA_ZeroCapMask(ljuDMA_CAP_MASK *Caps);

void ljkDMA_SetCapMasks(ljuDMA_CAP_MASK *Caps, ljeDMA_TX_TYPE Type);

void ljkDMA_ClearCapMasks(ljuDMA_CAP_MASK *Caps, ljeDMA_TX_TYPE Type);

ljtUINT32 ljkDMA_HasCapMask(ljuDMA_CAP_MASK *Caps, ljeDMA_TX_TYPE Type);

ljeSTATUS ljkDMAENGINE_SlaveConfig(
    ljsDMA_CHANNEL *Chan,
    ljsDMA_SLAVE_CONFIG *Config
    );

ljtBOOL ljkDMAENGINE_IsSlaveDirection(
    ljeDMA_TX_DIRECTION direction
    );

ljsDMA_TX_DESCRIPTOR *ljkDMAENGINE_PrepareSlaveSingle(
    ljsDMA_CHANNEL *Chan,
    dma_addr_t Addr,
    ljtSIZE_T Len,
    ljeDMA_TX_DIRECTION Dir,
    ljtUINT64 Flags
    );

ljsDMA_TX_DESCRIPTOR *ljkDMAENGINE_PrepareSlaveSG(
    ljsDMA_CHANNEL *Chan,
    struct scatterlist *SGList,
    ljtSIZE_T SGLen,
    ljeDMA_TX_DIRECTION Dir,
    ljtUINT64 Flags
    );

ljsDMA_TX_DESCRIPTOR *ljkDMAENGINE_PrepareMemcpy(
    ljsDMA_CHANNEL *Chan,
    dma_addr_t Src,
    dma_addr_t Dst,
    ljtSIZE_T Len,
    ljtUINT64 Flag
    );

ljeSTATUS ljkDMAENGINE_TerminateAll(ljsDMA_CHANNEL *Chan);

ljeSTATUS ljkDMAENGINE_Synchronize(ljsDMA_CHANNEL *Chan);

ljeSTATUS ljkDMAENGINE_TerminateSync(ljsDMA_CHANNEL *Chan);

ljeSTATUS ljkDMAENGINE_Pause(ljsDMA_CHANNEL *Chan);

ljeSTATUS ljkDMAENGINE_Resume(ljsDMA_CHANNEL *Chan);

ljeDMA_STATUS ljkDMAENGINE_TxStatus(
    ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljsDMA_TX_STATE *TxState
    );

void ljkDMAENGINE_IssuePending(
    ljsDMA_CHANNEL *Chan
    );

ljdDMA_TX_IDENTIFIER ljkDMAENGINE_Submit(ljsDMA_TX_DESCRIPTOR *Desc);

void ljkDMAINGINE_DescGetCallback(ljsDMA_TX_DESCRIPTOR *Desc,
   ljsDMAENGINE_DESC_CALLBACK *Cb
    );

void ljkDMAENGINE_DescCallbackInvoke(
    ljsDMAENGINE_DESC_CALLBACK *Cb,
    ljsDMA_RESULT *Result
    );

void ljkDMAENGINE_DescGetCallbackAndInvoke(
    ljsDMA_TX_DESCRIPTOR *Desc,
    ljsDMA_RESULT *Result
    );

ljtBOOL ljkDMAENGINE_DescCallbackValidate(
    ljsDMAENGINE_DESC_CALLBACK *Cb
    );

#endif
