/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/device.h>
#include <linux/slab.h>
#include <linux/miscdevice.h>
#include <linux/uaccess.h>
#include <linux/console.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
#include <drm/drm_drv.h>
#include <drm/drm_file.h>
#include <drm/drm_ioctl.h>
#include <drm/drm_print.h>
#include <drm/drm_vblank.h>
#include <drm/drm_framebuffer.h>
#include <drm/drm_fourcc.h>
#include <drm/drm_modes.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include <drm/drm_irq.h>
#endif
#else
#include <drm/drmP.h>
#endif

#include <drm/drm_fb_helper.h>
#include <linux/pm_runtime.h>

#include "lj_hal_kernel_linux.h"
#include "lj_hal_driver.h"
#include "lj_hal_kernel_parameter.h"

#include <linux/platform_device.h>

#if ljdENABLE_VPU
#include "hantrodec.h"
#endif
#if ljdENABLE_VPUENC
#include "vc8000_vcmd_driver.h"
#endif
#include "ljmicro_family.h"
#include "ljmicro_device.h"
/* Zone used for header/footer. */
#define _LJ_OBJ_ZONE ljvZONE_DRIVER

MODULE_DESCRIPTION("Ljmicro Graphics Driver");
MODULE_LICENSE("Dual MIT/GPL");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)
MODULE_IMPORT_NS(VFS_internal_I_am_really_a_filesystem_and_am_NOT_a_driver);
#endif

static struct class *gpu_class = ljvNULL;

static ljsPLATFORM *platform;

static ljkGALDEVICE galDevice;

static int gpu3DMinClock = 1;
static int contiguousRequested;
static ulong bankSize;

static struct miscdevice *gal_misc_device[ljdDEVICE_COUNT];

static uint32_t
ljm_get_dev_index(uint32_t minor)
{
    int dev_index = 0;

    for (dev_index = 0; dev_index < galDevice->args.devCount; dev_index++) {
        if (galDevice->devIDs[dev_index] == minor)
            return dev_index;
    }

    /* If not found the correct device index. */
    return ljdDEVICE_COUNT;
}

static void
_InitModuleParam(ljsMODULE_PARAMETERS *ModuleParam)
{
    ljtUINT i, j;
    ljsMODULE_PARAMETERS *p = ModuleParam;

    for (i = 0; i < ljdDEVICE_COUNT; i++)
        p->devices[i] = ljvNULL;

    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        p->irqs[i] = irqs[i];
        if (irqs[i] != -1 || ljmBITTEST(isrPoll, i) != 0) {
            p->registerBases[i] = registerBases[i];
            p->registerSizes[i] = registerSizes[i];
        }
#if USE_LINUX_PCIE
        p->bars[i] = bars[i];
        p->regOffsets[i] = regOffsets[i];
#endif
        p->registerBasesMapped[i] = ljvNULL;

        for (j = 0; j < ljvSRAM_INTER_COUNT; j++) {
            p->sRAMBases[i][j] = sRAMBases[i * ljvSRAM_INTER_COUNT + j];
            p->sRAMSizes[i][j] = sRAMSizes[i * ljvSRAM_INTER_COUNT + j];
        }
    }

    for (i = 0; i < ljvCORE_COUNT; i++)
        p->chipIDs[i] = chipIDs[i];

    for (i = 0; i < ljdGLOBAL_2D_COUNT; i++) {
        p->irq2Ds[i] = irq2Ds[i];

        if (irq2Ds[i] != -1) {
            p->register2DBases[i] = register2DBases[i];
            p->register2DSizes[i] = register2DSizes[i];
        }

#if USE_LINUX_PCIE
        p->bar2Ds[i] = bar2Ds[i];
        p->reg2DOffsets[i] = reg2DOffsets[i];
#endif
    }

    if (irqLine != -1) {
        p->irqs[ljvCORE_MAJOR] = irqLine;
        p->registerBases[ljvCORE_MAJOR] = registerMemBase;
        p->registerSizes[ljvCORE_MAJOR] = registerMemSize;
        /* Check legacy style. */
#if USE_LINUX_PCIE
        if (bar != -1) {
            if (p->bars[ljvCORE_MAJOR] == -1)
                p->bars[ljvCORE_MAJOR] = bar;
        }
#endif
    }

    if (irqLine2D != -1) {
        p->irq2Ds[0] = irqLine2D;
        p->register2DBases[0] = registerMemBase2D;
        p->register2DSizes[0] = registerMemSize2D;
#if USE_LINUX_PCIE
        if (bar2D != -1) {
            if (p->bar2Ds[0] == -1)
                p->bar2Ds[0] = bar2D;
        }
#endif
    }

    p->irqVG = irqLineVG;

    if (irqLineVG != -1) {
        p->registerVGBase = registerMemBaseVG;
        p->registerVGSize = registerMemSizeVG;
    }

#if USE_LINUX_PCIE
    p->barVG = barVG;
    p->regVGOffset = regVGOffset;
#endif

#if ljdDEC_ENABLE_AHB
    if (registerMemBaseDEC300 && registerMemSizeDEC300) {
        p->registerBases[ljvCORE_DEC] = registerMemBaseDEC300;
        p->registerSizes[ljvCORE_DEC] = registerMemSizeDEC300;
    }
#endif

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        p->contiguousBases[i] = contiguousBases[i];
        p->contiguousSizes[i] = contiguousSizes[i];
    }

    p->contiguousBase = contiguousBase;
    p->contiguousSize = contiguousSize;

    if (!p->contiguousSizes[0] && contiguousSize)
        p->contiguousSizes[0] = contiguousSize;

    p->contiguousRequested = contiguousRequested; /* not a module param. */

    p->devCount = 0;

    for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
        p->externalBase[i]  = externalBase[i];
        p->externalSize[i]  = externalSize[i];
        p->exclusiveBase[i] = exclusiveBase[i];
        p->exclusiveSize[i] = exclusiveSize[i];
    }

    for (i = 0; i < ljdPLATFORM_COUNT; i++)
        p->hwDevCounts[i] = hwDevCounts[i];

    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        p->devMemIDs[i] = devMemIDs[i];
        p->devSysMemIDs[i] = devSysMemIDs[i];
        p->devSRAMIDs[i] = devSRAMIDs[i];
        p->devCoreCounts[i] = devCoreCounts[i];
        p->dev2DCoreCounts[i] = dev2DCoreCounts[i];
    }

    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        p->extSRAMBases[i] = extSRAMBases[i];
        p->extSRAMSizes[i] = extSRAMSizes[i];
#if USE_LINUX_PCIE
        p->sRAMBars[i] = sRAMBars[i];
        p->sRAMOffsets[i] = sRAMOffsets[i];
#endif
    }

    for (i = 0; i < ljdMAX_MAJOR_CORE_COUNT; i++)
        p->userClusterMasks[i] = userClusterMasks[i];

    p->sRAMRequested = sRAMRequested;
    p->sRAMLoopMode  = sRAMLoopMode;

    p->baseAddress = baseAddress;
    p->physSize    = physSize;
    p->bankSize    = bankSize; /* not a module param. */

    p->recovery = recovery;
    p->powerManagement = powerManagement;

    p->softReset = softReset;

    p->enableMmu = mmu;
    p->fastClear = fastClear;

    p->compression = (compression == -1) ?
                     ljvCOMPRESSION_OPTION_DEFAULT : (ljeCOMPRESSION_OPTION)compression;
    p->gpu3DMinClock = gpu3DMinClock; /* not a module param. */
    p->enableNN      = enableNN;
    p->registerAPB   = registerAPB;
    p->smallBatch    = smallBatch;

    p->stuckDump  = stuckDump;

    p->deviceType = type;
    p->showArgs   = showArgs;

    p->mmuPageTablePool = mmuPageTablePool;
    p->mmuCmdPool       = mmuCmdPool;
    p->sharedPageTable  = sharedPageTable;
    p->processPageTable = processPageTable;
    p->flatMapping      = flatMapping;

    p->mmuDynamicMap = mmuDynamicMap;
    p->allMapInOne   = allMapInOne;

    p->vGPUType = vGPUType;
    p->vGPUId   = vGPUId;

    p->isrPoll = isrPoll;
#if !ljdENABLE_3D
    p->irqs[0] = -1;
    irqLine = -1;
    p->registerBases[0] = 0;
    registerMemBase     = 0;
    p->registerSizes[0] = 0;
    registerMemSize     = 0;
#endif

#if !ljdENABLE_2D
    p->irq2Ds[0] = -1;
    irqLine2D = -1;
    p->register2DBases[0] = 0;
    registerMemBase2D     = 0;
    p->register2DSizes[0] = 0;
    registerMemSize2D     = 0;
#endif

#if !ljdENABLE_VG
    p->irqVG = -1;
    irqLineVG = -1;
    p->registerVGBase = 0;
    registerMemBaseVG = 0;
    p->registerVGSize = 0;
    registerMemSizeVG = 0;
#endif

#if ljdENABLE_TTM
    p->flatMapping = 0;
    p->processPageTable = 0;
#endif
}

static void
_SyncModuleParam(ljsMODULE_PARAMETERS *ModuleParam)
{
    ljtUINT i, j;
    ljsMODULE_PARAMETERS *p = ModuleParam;

    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        irqs[i]          = p->irqs[i];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
        registerBases[i] = p->registerBases[i];
#else
        registerBases[i] = (ulong)p->registerBases[i];
#endif
        registerSizes[i] = (ulong)p->registerSizes[i];
#if USE_LINUX_PCIE
        bars[i]          = p->bars[i];
        regOffsets[i]    = p->regOffsets[i];
#endif
        for (j = 0; j < ljvSRAM_INTER_COUNT; j++) {
            sRAMBases[i * ljvSRAM_INTER_COUNT + j] = p->sRAMBases[i][j];
            sRAMSizes[i * ljvSRAM_INTER_COUNT + j] = p->sRAMSizes[i][j];
        }
    }

    for (i = 0; i < ljdGLOBAL_2D_COUNT; i++) {
        irq2Ds[i] = p->irq2Ds[i];
        register2DBases[i] = (ulong)p->register2DBases[i];
        register2DSizes[i] = (ulong)p->register2DSizes[i];
#if USE_LINUX_PCIE
        bar2Ds[i] = p->bar2Ds[i];
        reg2DOffsets[i] = p->reg2DOffsets[i];
#endif
    }

    /* Sync to legacy style. */
    irqLine2D = p->irq2Ds[0];
    irqLineVG = p->irqVG;
    registerMemBaseVG = (ulong)p->registerVGBase;
    registerMemSizeVG = (ulong)p->registerVGSize;

    for (i = 0; i < ljvCORE_COUNT; i++)
        chipIDs[i] = p->chipIDs[i];

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
        contiguousBases[i] = p->contiguousBases[i];
#else
        contiguousBases[i] = (ulong)p->contiguousBases[i];
#endif

        contiguousSizes[i] = (ulong)p->contiguousSizes[i];
    }

    if (p->contiguousSize && p->contiguousSize != ljdDEFAULT_CONTIGUOUS_SIZE)
        p->contiguousSizes[0] = contiguousSizes[0] = p->contiguousSize;

    if (p->contiguousBase)
        p->contiguousBases[0] = contiguousBases[0] = p->contiguousBase;

    contiguousRequested = p->contiguousRequested;   /* not a module param. */

    for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
        externalBase[i]  = p->externalBase[i];
        externalSize[i]  = p->externalSize[i];
        exclusiveBase[i] = p->exclusiveBase[i];
        exclusiveSize[i] = p->exclusiveSize[i];
    }

    for (i = 0; i < ljdPLATFORM_COUNT; i++)
        hwDevCounts[i] = p->hwDevCounts[i];

    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        platformIDs[i] = p->platformIDs[i];
        dev2DCoreCounts[i] = p->dev2DCoreCounts[i];
        devMemIDs[i] = p->devMemIDs[i];
        devSysMemIDs[i] = p->devSysMemIDs[i];
        devSRAMIDs[i] = p->devSRAMIDs[i];
        devCoreCounts[i] = p->devCoreCounts[i];

        if (p->devCoreCounts[i])
            p->devCount++;
    }

    if (!p->devCount) {
        for (i = 0; i < ljdDEVICE_COUNT; i++) {
            if (dev2DCoreCounts[i])
                p->devCount++;
        }
    }

    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        extSRAMBases[i] = p->extSRAMBases[i];
        extSRAMSizes[i] = p->extSRAMSizes[i];

#if USE_LINUX_PCIE
        sRAMBars[i] = p->sRAMBars[i];
        sRAMOffsets[i] = p->sRAMOffsets[i];
#endif
    }

    for (i = 0; i < ljdMAX_MAJOR_CORE_COUNT; i++)
        userClusterMasks[i] = p->userClusterMasks[i];

    sRAMRequested = p->sRAMRequested;
    sRAMLoopMode  = p->sRAMLoopMode;

    baseAddress = (ulong)p->baseAddress;
    physSize    = p->physSize;
    bankSize    = p->bankSize; /* not a module param. */

    recovery = p->recovery;
    powerManagement = p->powerManagement;

    mmu           = p->enableMmu;
    fastClear     = p->fastClear;
    compression   = p->compression;
    gpu3DMinClock = p->gpu3DMinClock; /* not a module param. */
    enableNN      = p->enableNN;
    registerAPB   = p->registerAPB;
    smallBatch    = p->smallBatch;

    stuckDump = p->stuckDump;
    type      = p->deviceType;
    showArgs  = p->showArgs;

    mmuPageTablePool = p->mmuPageTablePool;
    mmuCmdPool       = p->mmuCmdPool;
    sharedPageTable  = p->sharedPageTable;
    processPageTable = p->processPageTable;
    flatMapping      = p->flatMapping;
    mmuDynamicMap    = p->mmuDynamicMap;
    allMapInOne      = p->allMapInOne;
    isrPoll          = p->isrPoll;

    vGPUType         = p->vGPUType;
    vGPUId           = p->vGPUId;
}

void
ljkOS_DumpParam(void)
{
    ljtINT i;

    pr_warn("Ljmcore options:\n");

#if ljdDEC_ENABLE_AHB
    pr_warn("  registerMemBaseDEC300 = 0x%08lX\n", registerMemBaseDEC300);
    pr_warn("  registerMemSizeDEC300 = 0x%08lX\n", registerMemSizeDEC300);
#endif

    pr_warn("  bankSize          = 0x%08lX\n", bankSize);
    pr_warn("  fastClear         = %d\n",      fastClear);
    pr_warn("  compression       = %d\n",      compression);
    pr_warn("  powerManagement   = %d\n",      powerManagement);
    pr_warn("  baseAddress       = 0x%08lX\n", baseAddress);
    pr_warn("  physSize          = 0x%08lX\n", physSize);
    pr_warn("  recovery          = %d\n",      recovery);
    pr_warn("  stuckDump         = %d\n",      stuckDump);
    pr_warn("  GPU smallBatch    = %d\n",      smallBatch);
    pr_warn("  allMapInOne       = %d\n",      allMapInOne);
    pr_warn("  enableNN          = 0x%x\n",    enableNN);

    pr_warn("  userClusterMasks  = ");
    for (i = 0; i < ljdMAX_MAJOR_CORE_COUNT; i++)
        pr_warn("%x, ", userClusterMasks[i]);
    pr_warn("\n");

    pr_warn("  irqs              = ");
    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        if (irqs[i] != -1)
            pr_warn("%d, ", irqs[i]);
    }

    if (irqLine2D != -1)
        pr_warn("  irq2Ds            = ");

    for (i = 0; i < ljdGLOBAL_2D_COUNT; i++) {
        if (irq2Ds[i] != -1)
            pr_warn("%d, ", irq2Ds[i]);
    }
    pr_warn("\n");

    if (irqLineVG != -1) {
        pr_warn("  irqVG: ");
        pr_warn("%d, ", irqLineVG);
        pr_warn("\n");
    }

#if USE_LINUX_PCIE
    pr_warn("Bars configuration: \n");

    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        if (bars[i] != -1) {
            pr_warn("  bars[%d] = %d, regOffsets[%d] = %x, ",
                    i, bars[i], i, regOffsets[i]);
            pr_warn("\n");
        }
    }
#endif

    pr_warn("System reserve memory configuration: \n");

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        if (contiguousSizes[i]) {
            pr_warn("  contiguousSizes[%d] = 0x%lx\n", i, contiguousSizes[i]);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
            pr_warn("  contiguousBases[%d] = 0x%llX\n", i, contiguousBases[i]);
#else
            pr_warn("  contiguousBases[%d] = 0x%lX\n", i, contiguousBases[i]);
#endif
        }
    }

    pr_warn("Registers configuration: \n");

    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        if (registerSizes[i]) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
            pr_warn("  bases[%d] = 0x%llx, sizes[%d] = 0x%lx, ",
                    i, registerBases[i], i, registerSizes[i]);
#else
            pr_warn("  bases[%d] = 0x%lx, sizes[%d] = 0x%lx, ",
                    i, registerBases[i], i, registerSizes[i]);
#endif
            pr_warn("\n");
        }
    }

    for (i = 0; i < ljdCORE_2D_COUNT; i++) {
        if (register2DSizes[i]) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
            pr_warn("  2D bases[%d] = 0x%llx, sizes[%d] = 0x%lx, ",
                    i, register2DBases[i], i, register2DSizes[i]);
#else
            pr_warn("  2D bases[%d] = 0x%lx, sizes[%d] = 0x%lx, ",
                    i, register2DBases[i], i, register2DSizes[i]);
#endif
            pr_warn("\n");
        }
    }

    pr_warn("  External sRAMBases = ");
    for (i = 0; i < ljvSRAM_EXT_COUNT; i++)
        pr_warn("0x%llx, ", extSRAMBases[i]);
    pr_warn("\n");

    pr_warn("  mmu               = %d\n", mmu);
    pr_warn("  mmuPageTablePool  = %d\n", mmuPageTablePool);
    pr_warn("  mmuCmdPool        = %d\n", mmuCmdPool);
    pr_warn("  mmuDynamicMap     = %d\n", mmuDynamicMap);
    pr_warn("  isrPoll           = 0x%08X\n", isrPoll);

    pr_warn("Build options:\n");
    pr_warn("  ljdGPU_TIMEOUT    = %d\n", ljdGPU_TIMEOUT);
    pr_warn("  ljdGPU_2D_TIMEOUT = %d\n", ljdGPU_2D_TIMEOUT);
    pr_warn("  ljdINTERRUPT_STATISTIC = %d\n", ljdINTERRUPT_STATISTIC);
}

static int drv_open(struct inode *inode, struct file *filp)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsHAL_PRIVATE_DATA_PTR data = ljvNULL;
    ljtUINT i, dev_index;
    ljtUINT attached = 0;
    ljkDEVICE device = ljvNULL;

    ljmkHEADER_ARG("inode=%p filp=%p", inode, filp);

    dev_index = iminor(file_inode(filp));

    if (type == 1)
        dev_index = ljm_get_dev_index(dev_index);

    if (dev_index >= ljdDEVICE_COUNT)
        return -ENODEV;

    data = kmalloc(sizeof(ljsHAL_PRIVATE_DATA), GFP_KERNEL | __GFP_NOWARN);

    if (data == ljvNULL) {
        ljmkFOOTER_ARG("status=%d", ljvSTATUS_OUT_OF_MEMORY);
        return -ENOMEM;
    }

    data->isLocked = ljvFALSE;
    data->device   = galDevice;
    data->pidOpen  = _GetProcessID();

    device = galDevice->devices[dev_index];

    /* Attached the process. */
    for (i = 0; i < ljvCORE_COUNT; i++) {
        if (device->kernels[i]) {
            status = ljkKERNEL_AttachProcess(device->kernels[i], ljvTRUE);

            if (ljmIS_ERROR(status))
                goto OnError;

            attached = i;
        }
    }

    filp->private_data = data;

    /* Success. */
    ljmkFOOTER_NO();
    return 0;

OnError:
    for (i = 0; i < attached; i++) {
        if (device && device->kernels[i])
            ljmkVERIFY_OK(ljkKERNEL_AttachProcess(device->kernels[i], ljvFALSE));
    }

    kfree(data);

    ljmkFOOTER_ARG("status=%d", status);
    return -ENOTTY;
}

static int drv_release(struct inode *inode, struct file *filp)
{
    int ret = -ENOTTY;
    ljeSTATUS status = ljvSTATUS_OK;
    ljsHAL_PRIVATE_DATA_PTR data;
    ljkGALDEVICE gal_device = ljvNULL;
    ljkDEVICE device = ljvNULL;
    ljtUINT i, dev_index;

    ljmkHEADER_ARG("inode=%p filp=%p", inode, filp);

    dev_index = iminor(file_inode(filp));

    if (type == 1)
        dev_index = ljm_get_dev_index(dev_index);

    if (dev_index >= ljdDEVICE_COUNT)
        return -ENODEV;

    data = filp->private_data;

    if (data == ljvNULL) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): private_data is NULL\n",
                       __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    gal_device = data->device;

    if (gal_device == ljvNULL) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): device is NULL\n",
                       __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (data->isLocked) {
        device = gal_device->devices[data->devIndex];

        /* Release the mutex. */
        ljmkONERROR(ljkOS_ReleaseMutex(device->os, device->commitMutex));
        data->isLocked = ljvFALSE;
    }

    /* A process gets detached. */
    device = gal_device->devices[dev_index];

    /* A process gets detached. */
    for (i = 0; i < ljvCORE_COUNT; i++) {
        if (device && device->kernels[i]) {
            ljmkVERIFY_OK(ljkKERNEL_AttachProcessEx(device->kernels[i],
                                                    ljvFALSE, data->pidOpen));
        }
    }

    kfree(data);
    filp->private_data = NULL;

    /* Success. */
    ret = 0;

OnError:
    ljmkFOOTER();
    return ret;
}

static long drv_ioctl(struct file *filp, unsigned int ioctlCode, unsigned long arg)
{
    long ret = -ENOTTY;
    ljeSTATUS status = ljvSTATUS_OK;
    ljsHAL_INTERFACE iface = {0};
    ljtUINT32 copyLen;
    DRIVER_ARGS drvArgs;
    ljkGALDEVICE gal_device;
    ljkDEVICE device;
    ljsHAL_PRIVATE_DATA_PTR data;
    ljtUINT dev_index;
    ljtUINT32 copy_size;
#if LJMICRO_PROFILER
    static ljsHAL_PROFILER_INTERFACE iface_profiler = {0};
#endif

    ljmkHEADER_ARG("filp=%p ioctlCode=%u arg=%lu", filp, ioctlCode, arg);

    dev_index = iminor(file_inode(filp));

    if (type == 1)
        dev_index = ljm_get_dev_index(dev_index);

    if (dev_index >= ljdDEVICE_COUNT)
        return -ENODEV;

    data = filp->private_data;

    if (data == ljvNULL) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): private_data is NULL\n",
                       __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    gal_device = data->device;

    if (gal_device == ljvNULL) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): device is NULL\n",
                       __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    switch (ioctlCode) {
    case IOCTL_LJHAL_INTERFACE:
        /* Get the drvArgs. */
        copyLen = copy_from_user(&drvArgs, (void *)arg, sizeof(DRIVER_ARGS));

        if (copyLen != 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): error copying of the input arguments.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

#if !ljdIGNORE_DRIVER_VERSIONS_MISMATCH
        /* Now bring in the ljsHAL_INTERFACE structure. */
        if (drvArgs.InputBufferSize != sizeof(ljsHAL_INTERFACE) ||
            drvArgs.OutputBufferSize != sizeof(ljsHAL_INTERFACE)) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): input or/and output structures are invalid.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        copy_size = sizeof(ljsHAL_INTERFACE);
#else
        copy_size = ljmMIN(drvArgs.InputBufferSize, sizeof(ljsHAL_INTERFACE));
#endif /* !ljdIGNORE_DRIVER_VERSIONS_MISMATCH */

        copyLen = copy_from_user(&iface,
                                 ljmUINT64_TO_PTR(drvArgs.InputBuffer),
                                 copy_size);

        if (copyLen != 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): error copying of input HAL interface.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        if (iface.command == ljvHAL_DEVICE_MUTEX) {
            if (iface.u.DeviceMutex.isMutexLocked == ljvTRUE)
                data->isLocked = ljvTRUE;
            else
                data->isLocked = ljvFALSE;
        }

        /* Record the last device id. */
#if ljdENABLE_MULTI_DEVICE_MANAGEMENT
        if (!dev_index && dev_index != iface.devIndex)
            dev_index = iface.devIndex;
#endif

        data->devIndex = dev_index;

        if (iface.command == ljvHAL_CHIP_INFO) {
            ljtUINT i, count = 0;

            for (i = 0; i < gal_device->args.devCount; i++) {
                device = gal_device->devices[i];

                ljmkONERROR(ljkDEVICE_ChipInfo(device, &iface, &count));
            }
        } else
        {
#if ljdENABLE_MULTI_DEVICE_MANAGEMENT
            if (iface.devIndex >= ljdDEVICE_COUNT)
                ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
#endif

            device = gal_device->devices[dev_index];

            status = ljkDEVICE_Dispatch(device, &iface);

            /* Redo system call after pending signal is handled. */
            if (status == ljvSTATUS_INTERRUPTED) {
                ret = -ERESTARTSYS;
                //ljmkONERROR(status);  //solve the error : ljmkONERROR: status=-26(ljvSTATUS_INTERRUPTED) @ drv_ioctl(982), The error has no effect on the operation, but an error is reported.
                ljmkFOOTER();
                return ret;
            }
        }

        /* Copy data back to the user. */
        copyLen = copy_to_user(ljmUINT64_TO_PTR(drvArgs.OutputBuffer),
                               &iface,
                               copy_size);

        if (copyLen != 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): error copying of output HAL interface.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
        break;

    case IOCTL_LJHAL_PROFILER_INTERFACE:
#if LJMICRO_PROFILER
        /* Get the drvArgs. */
        copyLen = copy_from_user(&drvArgs, (void *)arg, sizeof(DRIVER_ARGS));

        if (copyLen != 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): error copying of the input arguments.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

#if !ljdIGNORE_DRIVER_VERSIONS_MISMATCH
        /* Now bring in the ljsHAL_INTERFACE structure. */
        if (drvArgs.InputBufferSize != sizeof(ljsHAL_PROFILER_INTERFACE) ||
            drvArgs.OutputBufferSize != sizeof(ljsHAL_PROFILER_INTERFACE)) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): input or/and output structures are invalid.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        copy_size = sizeof(ljsHAL_PROFILER_INTERFACE);
#else
        copy_size = ljmMIN(drvArgs.InputBufferSize, sizeof(ljsHAL_PROFILER_INTERFACE));
#endif /* !ljdIGNORE_DRIVER_VERSIONS_MISMATCH */

        copyLen = copy_from_user(&iface_profiler,
                                 ljmUINT64_TO_PTR(drvArgs.InputBuffer),
                                 copy_size);

        if (copyLen != 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): error copying of input HAL interface.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        device = gal_device->devices[dev_index];

        status = ljkDEVICE_Profiler_Dispatch(device, &iface_profiler);

        /* Redo system call after pending signal is handled. */
        if (status == ljvSTATUS_INTERRUPTED) {
            ret = -ERESTARTSYS;
            ljmkONERROR(status);
        }

        /* Copy data back to the user. */
        copyLen = copy_to_user(ljmUINT64_TO_PTR(drvArgs.OutputBuffer),
                               &iface_profiler,
                               copy_size);

        if (copyLen != 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): error copying of output HAL interface.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
#endif
        break;

    default:
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): unknown command %d\n",
                       __func__, __LINE__, ioctlCode);

        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    /* Success. */
    ret = 0;

OnError:
    ljmkFOOTER();
    return ret;
}

static const struct file_operations driver_fops = {
    .owner = THIS_MODULE,
    .open = drv_open,
    .release = drv_release,
    .unlocked_ioctl = drv_ioctl,

#if LINUX_VERSION_CODE <= KERNEL_VERSION(5, 8, 18)

#ifdef HAVE_COMPAT_IOCTL
    .compat_ioctl = drv_ioctl,
# endif

#else
    .compat_ioctl = drv_ioctl,
#endif
};

ljeSTATUS ljm_misc_device_node_create(uint32_t dev_index)
{
    ljeSTATUS status = ljvSTATUS_OK;
    struct miscdevice *dev = kzalloc(sizeof(struct miscdevice), GFP_KERNEL);
    ljtINT32 ret;

    if (!dev) {
        ret = -ENOMEM;

        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): misc_register fails.\n",
                        __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    dev->minor = MISC_DYNAMIC_MINOR;
    dev->fops = &driver_fops;
    if (dev_index > 0)
        dev->name = kasprintf(GFP_KERNEL, "ljmcore%d", dev_index);
    else
        dev->name = kasprintf(GFP_KERNEL, "ljmcore");

    if (!dev->name) {
        ret = -ENOMEM;

        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): misc_register fails.\n",
                        __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    ret = misc_register(dev);
    if (ret != 0) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): misc_register fails.\n",
                        __func__, __LINE__);

        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    gal_misc_device[dev_index] = dev;
    galDevice->devIDs[dev_index] = dev->minor;

    return ljvSTATUS_OK;

OnError:
    /* If one node loads failed, try to release all the node resource. */
    for (; dev_index >= 0; dev_index--) {
        if (gal_misc_device[dev_index]) {
            misc_deregister(gal_misc_device[dev_index]);
            kfree(gal_misc_device[dev_index]->name);
            gal_misc_device[dev_index]->name = ljvNULL;
            kfree(gal_misc_device[dev_index]);
            gal_misc_device[dev_index] = ljvNULL;
        }
    }

    return status;
}

void ljm_misc_device_node_destroy(uint32_t dev_index)
{
    ljmkASSERT(gal_misc_device[dev_index]);
    misc_deregister(gal_misc_device[dev_index]);
    kfree(gal_misc_device[dev_index]->name);
    gal_misc_device[dev_index]->name = ljvNULL;
    kfree(gal_misc_device[dev_index]);
    gal_misc_device[dev_index] = ljvNULL;
}

void ljm_char_device_node_destroy(uint32_t dev_index)
{
    ljmkASSERT(gpu_class != ljvNULL);
    device_destroy(gpu_class, MKDEV(major, dev_index));
}

ljeSTATUS ljm_char_device_node_create(uint32_t dev_index)
{
    ljeSTATUS status = ljvSTATUS_OK;
    int result = -EINVAL;

    ljmkHEADER();

    if (!gpu_class) {
        /* Register the character device. */
        result = register_chrdev(major, DEVICE_NAME, &driver_fops);

        if (result < 0) {
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): Could not allocate major number for mmap.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
        }

        if (major == 0)
            major = result;

        /* Create the device class. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
        gpu_class = class_create(THIS_MODULE, CLASS_NAME);
#else
        gpu_class = class_create(CLASS_NAME);
#endif

        if (IS_ERR(gpu_class)) {
            gpu_class = ljvNULL;
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): Failed to create the class.\n",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
        }
    }

    if (dev_index > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
        device_create(gpu_class, NULL, MKDEV(major, dev_index), NULL, "ljmcore%d", dev_index);
#else
        device_create(gpu_class, NULL, MKDEV(major, dev_index), "ljmcore%d", dev_index);
#endif
    } else { /* Compatible with old style. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
        device_create(gpu_class, NULL, MKDEV(major, dev_index), NULL, DEVICE_NAME);
#else
        device_create(gpu_class, NULL, MKDEV(major, dev_index), DEVICE_NAME);
#endif
    }

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    /* Roll back. */
    if (gpu_class)
        class_destroy(gpu_class);

    if (result < 0)
        unregister_chrdev(result, DEVICE_NAME);

    ljmkFOOTER();
    return status;
}

static int drv_init(void)
{
    ljeSTATUS status;
    ljkGALDEVICE device = ljvNULL;
    ljtUINT dev_index = 0;

    ljmkHEADER();

    pr_info("Ljmcore version %s\n", ljvVERSION_STRING);

    if (showArgs)
        ljkOS_DumpParam();

    /* Create the GAL device. */
    status = ljkGALDEVICE_Construct(platform, &platform->params, &device);

    if (ljmIS_ERROR(status)) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): Failed to create the GAL device: status=%d\n",
                       __func__, __LINE__, status);

        goto OnError;
    }

    /* Start the GAL device. */
    ljmkONERROR(ljkGALDEVICE_Start(device));

    /* Set global galDevice pointer. */
    galDevice = device;

    for (dev_index = 0; dev_index < device->args.devCount; dev_index++) {
        if (type == 1)
            ljmkONERROR(ljm_misc_device_node_create(dev_index));
        else
            ljmkONERROR(ljm_char_device_node_create(dev_index));
    }
    /* Success. */
    ljmkFOOTER();
    return ljvSTATUS_OK;

OnError:
    if (device) {
        ljmkVERIFY_OK(ljkGALDEVICE_Stop(device));
        ljmkVERIFY_OK(ljkGALDEVICE_Destroy(device));
    }

    ljmkFOOTER();
    return status;
}

static void drv_exit(void)
{
    ljtUINT dev_index;

    ljmkHEADER();

    if (type == 1) {
        for (dev_index = 0; dev_index < galDevice->args.devCount; dev_index++)
            ljm_misc_device_node_destroy(dev_index);
    } else {
        ljmkASSERT(gpu_class != ljvNULL);

        for (dev_index = 0; dev_index < galDevice->args.devCount; dev_index++)
            ljm_char_device_node_destroy(dev_index);

        class_destroy(gpu_class);
        gpu_class = ljvNULL;

        unregister_chrdev(major, DEVICE_NAME);
    }

    ljmkVERIFY_OK(ljkGALDEVICE_Stop(galDevice));

    /* huym: cancel destroy to support drm display with vram node,
     * otherwise NULL pointer would be meet in ljm_crtc_set_base() to unpin a vram node */
    /* ljmkVERIFY_OK(ljkGALDEVICE_Destroy(galDevice)); */

    ljmkFOOTER_NO();
}

#if ljdENABLE_DRM
int ljm_drm_probe(struct pci_dev *pdev, unsigned long flags);
int ljm_drm_remove(ljkGALDEVICE gal_dev);

int ljmicro_rw_init(void);
void ljmicro_rw_exit(void);

int ljmicro_get_args_power_management(void)
{
    return platform->params.powerManagement;
}
#endif

/*****************************************************************************
 *  ??? ljmicro_zero_exclusive_vram
 *  功能描述: 通过DMA同步方法清零不可见显? *  输入参数: struct _ljsPLATFORM_PCIE *pcie_platform
 *  输出参数: ? *  ??? ljeSTATUS
 *
 *  ޸ʷ:
 *         2.     :   2024530
 *                :
 *            ޸:
 *
 *****************************************************************************/
ljeSTATUS ljmicro_zero_exclusive_vram(struct _ljsPLATFORM_PCIE *pcie_platform)
{
    ljkKERNEL           kernel       = _GetValidKernel(0, galDevice);
    ljsPLATFORM        *platform     =  kernel->os->device->platform;

    struct ljkVramNode *vramNode     =  ljvNULL;
    ljtPOINTER          pointer      =  ljvNULL;
    ljeSTATUS           status       =  ljvSTATUS_OK;
    ljtSIZE_T           tmp_bytes =  0x1000000;  /* 10MB */
    ljsDMA_TRANS_INFO   info;

    int                 pdev_index       = 0;
    unsigned long       externalSize     = 0;
    unsigned long       exclusiveBase    = 0;
    unsigned long       exclusiveSize    = 0;
    unsigned long       exclusive_offset = 0;
    unsigned long       vram_size        = 0;
    ljtUINT8_PTR        logical          = ljvNULL;

    unsigned long flags   = pcie_platform->pcie_info[0].flags;

    /* Allocate the ljkVramNode from the heap. */
    status = ljkOS_Allocate(kernel->os, ljmSIZEOF(struct ljkVramNode), &pointer);
    if (status != ljvSTATUS_OK) {
        ljmkPRINT("out of memory when suspend\n");
        return 0;
    }

    /* 获取不可见显?*/
    externalSize = pcie_platform->pcie_info[pdev_index].bar[0].size;
    exclusiveBase = externalSize;
    //INVISBLE DIE1 16GB-32GB
    vram_size = min(pcie_platform->pcie_info[pdev_index].gpu_vram_size, 0xC0000000UL);   /* VRAM size is top to 3G */
    if (vram_size > pcie_platform->pcie_info[pdev_index].bar[0].size + ((flags & LJMICRO_IS_S2C) ?  VRAM_INVISIBLE_BAD_SIZE : 0)) {
        exclusiveSize = vram_size - pcie_platform->pcie_info[pdev_index].bar[0].size
            - ((flags & LJMICRO_IS_S2C) ?  VRAM_INVISIBLE_BAD_SIZE : 0);
    } else {
        exclusiveSize = 0;
    }

    vramNode                = (struct ljkVramNode *)pointer;
    vramNode->next          = ljvNULL;
    vramNode->bytes         = tmp_bytes;
    vramNode->beenMirrored  = ljvFALSE;
    vramNode->useNodeOffset = ljvTRUE;

    status = ljkVIDMEM_NODE_AllocateVirtual(kernel, ljvPOOL_VIRTUAL, (ljeVIDMEM_TYPE)ljvSURF_TYPE_UNKNOWN,
                                            ljvALLOC_FLAG_NONE, &tmp_bytes, &vramNode->virtualNode);
    if (status != ljvSTATUS_OK) {
        ljmkPRINT("out of memory when allocate a virtual node for bytes: %lx\n", vramNode->bytes);
        ljkOS_Free(kernel->os, pointer);
        return 0;
    }

    /* 初始化virtual_node大小 */
    /* Lock for kernel side CPU access. */
    ljkVIDMEM_NODE_LockCPU(kernel, vramNode->virtualNode, ljvFALSE, ljvFALSE, (ljtPOINTER *)&logical);
        /* Set the whole buffer to characteristic value */

    memset(logical, 0, tmp_bytes);

    info.src_node = vramNode->virtualNode;
    info.offset   = 0;
    info.bytes    = vramNode->bytes;
    info.reason   = ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCALOFFSET;

    /* backup data of virtual node to vram node */
    while(exclusive_offset < exclusiveSize)
    {
        vramNode->vramNode = ljmINT2PTR(exclusiveBase + exclusive_offset);
        info.dst_node      = vramNode->vramNode;

        status = platform->ops->dmaCopy(kernel, &info);
        exclusive_offset  += tmp_bytes;
    }

    if (status != ljvSTATUS_OK) {
        ljmkPRINT("synchronize vram to system memory with error\n");
    }

    /* release virtual node */
    ljkVIDMEM_NODE_UnlockCPU(kernel, vramNode->virtualNode, 0, ljvFALSE, ljvFALSE);
    ljkVIDMEM_NODE_Dereference(kernel, vramNode->virtualNode);
    ljkOS_Free(kernel->os, pointer);

    return status;

}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
static int ljm_dev_probe(struct platform_device *pdev)
#else
static int __devinit ljm_dev_probe(struct platform_device *pdev)
#endif
{
    int  ret = -ENODEV;
    bool getPowerFlag = ljvFALSE;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
    static u64 dma_mask = DMA_BIT_MASK(64);
#else
    static u64 dma_mask = DMA_64BIT_MASK;
#endif

#if ljdCAPTURE_ONLY_MODE
    ljtPHYS_ADDR_T contiguousBaseCap = 0;
    ljtSIZE_T contiguousSizeCap = 0;
    ljtPHYS_ADDR_T sRAMBaseCap[ljdGLOBAL_CORE_COUNT][ljvSRAM_INTER_COUNT];
    ljtUINT32 sRAMSizeCap[ljdGLOBAL_CORE_COUNT][ljvSRAM_INTER_COUNT];
    ljtPHYS_ADDR_T extSRAMBaseCap[ljvSRAM_EXT_COUNT];
    ljtUINT32 extSRAMSizeCap[ljvSRAM_EXT_COUNT];
    ljtUINT i = 0, j = 0;
#endif

    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)platform;
    unsigned long flags = pcie_platform->pcie_info[0].flags;
    ljtUINT k;

    ljmkHEADER_ARG("pdev=%lx platform=%lx device_number=%d",
        (unsigned long)pdev, (unsigned long)platform, pcie_platform->device_number);

    if (pcie_platform->device_number == 0) {
        printk(KERN_WARNING "LJMICRO GPU DEVICE is not found\n");
        return ret;
    }

    if (flags & LJMICRO_IS_BUS32)    /*LJMICRO_IS_BUS32*/
        dma_mask = DMA_BIT_MASK(31);

    platform->device = pdev;    /* as assigned here, no need to assign in ljkPLATFORM_Init */

    /* todo: 实现ljmcore_device指向平台设备的dev?
     * X86 + ubuntu平台: ljmcore_device指向平台设备的dev, 可以通过dma_set_mask和dma_set_coherent_mask函数设置dma_mask.
     * FT2000-4 + uos平台: 如果ljmcore_device指向平台设备的dev, dma_set_mask和dma_set_coherent_mask函数返回错误.
     * 以上平台采用pci设备的dev, dma_set_mask和dma_set_coherent_mask函数能够执行成功.
     * by hym on 20220518
     */
    //ljmcore_device = &pcie_platform->pcie_info[0].pdev->dev;  /* ljmcore_device is removed by hym on 20220907 */
    //ljmcore_device = &pdev->dev; /* change from pci dev to platform dev by hym on 20220518 */
    //printk("----%s(%d): platform->device=%lx ljmcore_device=%lx &pcie_platform->pcie_info[0].pdev->dev=%lx----\n",
    //    __FUNCTION__, __LINE__,
    //    (unsigned long)platform->device, (unsigned long)ljmcore_device, (unsigned long)&pcie_platform->pcie_info[0].pdev->dev);

    if (!mmu) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
        dma_mask = DMA_BIT_MASK(32);
#else
        dma_mask = DMA_32BIT_MASK;
#endif
    }

    pdev->dev.dma_mask = &dma_mask;
    pdev->dev.coherent_dma_mask = dma_mask;

    if (platform->ops->getPower) {
        if (ljmIS_ERROR(platform->ops->getPower(platform))) {
            ljmkFOOTER_NO();
            return ret;
        }
        getPowerFlag = ljvTRUE;
    }

    /* Gather module parameters. */
    _InitModuleParam(&platform->params);

    platform->params.devices[0] = &pdev->dev;

    platform->params.devices[0]->coherent_dma_mask = dma_mask;

    if (platform->params.devices[0]->dma_mask)
        *platform->params.devices[0]->dma_mask = dma_mask;
    else
        platform->params.devices[0]->dma_mask = &platform->params.devices[0]->coherent_dma_mask;

#if 0
    if (platform->ops->dmaInit) {
        if (ljmIS_ERROR(platform->ops->dmaInit(platform))) {
            ljmkFOOTER_NO();
            return ret;
        }
    }
#endif

#if ljdCAPTURE_ONLY_MODE
    contiguousBaseCap = platform->params.contiguousBases[0];
    contiguousSizeCap = platform->params.contiguousSizes[0];

    ljmkPRINT("Capture only mode is enabled in Hal Kernel.");

    if ((contiguousBaseCap + contiguousSizeCap) > 0x80000000)
        ljmkPRINT("Capture only mode: contiguousBase + contiguousSize > 2G, there is error in CModel and old MMU version RTL simulation.");

    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        for (j = 0; j < ljvSRAM_INTER_COUNT; j++) {
            sRAMBaseCap[i][j] = platform->params.sRAMBases[i][j];
            sRAMSizeCap[i][j] = platform->params.sRAMSizes[i][j];
        }
    }

    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        extSRAMBaseCap[i] = platform->params.extSRAMBases[i];
        extSRAMSizeCap[i] = platform->params.extSRAMSizes[i];
    }
#endif

    if (platform->ops->adjustParam)
        /* Override default module param. */
        platform->ops->adjustParam(platform, &platform->params);

#if ljdCAPTURE_ONLY_MODE
    platform->params.contiguousBases[0] = contiguousBaseCap;
    platform->params.contiguousSizes[0] = contiguousSizeCap;

    for (i = 0; i < ljdGLOBAL_CORE_COUNT; i++) {
        for (j = 0; j < ljvSRAM_INTER_COUNT; j++) {
            platform->params.sRAMBases[i][j] = sRAMBaseCap[i][j];
            platform->params.sRAMSizes[i][j] = sRAMSizeCap[i][j];
        }
    }

    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        platform->params.extSRAMBases[i] = extSRAMBaseCap[i];
        platform->params.extSRAMSizes[i] = extSRAMSizeCap[i];
    }
#endif
    /* Update module param because drv_init() uses them directly. */
    _SyncModuleParam(&platform->params);

    if (powerManagement == 0)
        ljmkPRINT("[ljmcore warning]: power saving is disabled.");

    if (debugLevel)
        ljkOS_SetDebugLevelZone(ljvLEVEL_VERBOSE, ljdZONE_ALL);

    ret = drv_init();

    if (!ret) {
        platform_set_drvdata(pdev, galDevice);

        for (k=0; k<pcie_platform->device_number; k++) {
            pci_set_drvdata(pcie_platform->pcie_info[k].pdev, galDevice);
            ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                           "%s(%d): %d pci_set_drvdata(pdev=%p dev=%p galDevice=%p)\n",
                           __func__, __LINE__,
                           k, pcie_platform->pcie_info[k].pdev, &pcie_platform->pcie_info[k].pdev->dev, galDevice
                           );
        }

        /* enable EDMA interrupt */
        if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP200) {
            //printk("----enable EDMA and AMBA DMA interrupt----\n");
            LJM_MISC_W32(0, MISC_INTR_EN, LJM_MISC_R32(0, MISC_INTR_EN) | m_INTR_EN_EDMA);
        }

        if (flags & LJMICRO_IS_FPGA) {
            /* 将不可见显存部分初始化为0 */
            ljmicro_zero_exclusive_vram(pcie_platform);
        }

#if ljdENABLE_DRM
        /* use dev of pci_device for drm, otherwise, using platform dev, modesetting could not detect the device. by hym on 20210427 */
        ret = ljm_drm_probe(pcie_platform->pcie_info[0].pdev, pcie_platform->pcie_info[0].flags);

        ljmicro_rw_init();

        g_card_cnt = pcie_platform->device_number;
#endif  /* ~ljdENABLE_DRM */

#if ljdENABLE_VPU
        /* VPU init */
        if (flags & LJMICRO_HAS_VPU) { /*LJMICRO_HAS_VPU*/
            hantrodec_init(pcie_platform->pcie_info[0].pdev);
        }
#endif  /* ~ljdENABLE_VPU */

#if ljdENABLE_VPUENC                        /* VPU init */
        if (flags & LJMICRO_HAS_VPU_ENC) { /*LJMICRO_HAS_VPU*/
#ifndef NEW_VIDMEM_MANAGE
            int ret=enc_vcmd_init();
            if (ret < 0) {
                printk("enc init failed ret =%d \n",ret);
            }
#else
            int ret= enc_vcmd_init(pcie_platform->pcie_info[0].pdev);
            if (ret < 0) {
                printk("enc init failed ret =%d \n",ret);
            }
#endif  /* ~NEW_VIDMEM_MANAGE */

        }
#endif  /* ~ljdENABLE_VPUENC */
    }

    if (ret < 0) {
        int i;
        if (platform->ops->putPower) {
            if (getPowerFlag == ljvTRUE)
                platform->ops->putPower(platform);
        }

        for (i = 0; i < ljdDEVICE_COUNT; i++) {
            if (platform->params.devices[i])
                platform->params.devices[i] = NULL;
        }

        ljmkFOOTER_ARG(KERN_INFO "Failed to register gpu driver: %d\n", ret);
    } else {
        ljmkFOOTER_ARG(KERN_INFO "Success ret=%d", ret);
    }

    return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
static int ljm_dev_remove(struct platform_device *pdev)
#else
static int __devexit ljm_dev_remove(struct platform_device *pdev)
#endif
{
    ljtUINT i;
    uint32_t int_mask = 0;
#if ljdENABLE_VPU || ljdENABLE_VPUENC
    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)platform;
    unsigned long flags = pcie_platform->pcie_info[0].flags;
#endif
    ljmkHEADER_ARG("pdev=%px", pdev);

    int_mask = LJM_MISC_R32(0, MISC_INTR_EN);

#if ljdENABLE_VPU
    if (flags & LJMICRO_HAS_VPU) { /*LJMICRO_HAS_VPU*/
#ifndef NEW_VIDMEM_MANAGE
        hantrodec_cleanup();
#else
        hantrodec_cleanup(pcie_platform->pcie_info[0].pdev);
#endif
    }
    /* mask VPU interrupt */
    int_mask &= ~(m_INTR_EN_VPU0 | m_INTR_EN_VPU1);
    LJM_MISC_W32(0, MISC_INTR_EN, int_mask);
#endif  /* ~ljdENABLE_VPU */

#if ljdENABLE_VPUENC
    if (flags & LJMICRO_HAS_VPU_ENC) { /* LJMICRO_HAS_VPU_ENC */
#ifndef NEW_VIDMEM_MANAGE
        enc_vcmd_cleanup();
#else
        enc_vcmd_cleanup(pcie_platform->pcie_info[0].pdev);
#endif
    }
#endif

#if ljdENABLE_DRM
    //printk("%s%d: ljmicro_rw_exit.....\n", __FUNCTION__, __LINE__);
    ljmicro_rw_exit();

#if 0
    //printk("%s%d: ljm_drm_remove galDevice=%p (%px).....\n", __FUNCTION__, __LINE__, galDevice, galDevice);
    ljm_drm_remove(galDevice);

    /* mask VGA interrupt */
    int_mask &= ~m_INTR_EN_VGA;
    LJM_MISC_W32(0, MISC_INTR_EN, int_mask);
#endif
#endif

    //printk("%s%d: drv_exit.....\n", __FUNCTION__, __LINE__); //msleep(5000);
    drv_exit();

    /* mask core interrupt */
    int_mask &= ~(m_INTR_EN_G2D | m_INTR_EN_G3D0 | m_INTR_EN_G3D1);
    LJM_MISC_W32(0, MISC_INTR_EN, int_mask);

    //printk("%s%d: platform->ops->putPower.....\n", __FUNCTION__, __LINE__);
    if (platform->ops->putPower) {
        platform->ops->putPower(platform);
        //printk("%s%d: ----platform->ops->putPower----\n", __FUNCTION__, __LINE__);
    }

    //printk("%s%d: platform->ops->dmaExit.....\n", __FUNCTION__, __LINE__);
    if (platform->ops->dmaExit)
        platform->ops->dmaExit(/*platform*/);

    //printk("----%s%d: mask edma interrupt.....\n", __FUNCTION__, __LINE__);
    /* mask edma interrupt */
    int_mask &= ~m_INTR_EN_EDMA;
    LJM_MISC_W32(0, MISC_INTR_EN, int_mask);

    /* disable all interrupts */
    LJM_MISC_W32(0, MISC_INTR_EN, 0);

    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (platform->params.devices[i])
            platform->params.devices[i] = NULL;
    }

#if 0
    {   /* unmap registers */
        int i;
        for (i=0; i<g_card_cnt; i++) {
            if (g_mmio_bases[i]) {
                iounmap(g_mmio_bases[i]);
                g_mmio_bases[i] = NULL;
            }
        }
    }
#endif

    ljmkFOOTER_NO();
    return 0;
}

static void ljm_dev_shutdown(struct platform_device *pdev)
{
    galDevice->gotoShutdown = ljvTRUE;

    ljm_dev_remove(pdev);
}

static int ljm_dev_suspend(struct platform_device *dev, pm_message_t state)
{
    ljeSTATUS status;
    ljkGALDEVICE gal_device;
    ljkDEVICE device;
    ljtUINT i, dev_index;

    ljmkHEADER_ARG("dev=%p state=%d", dev, state.event);

    //printk("----ljm_dev_suspend enter----\n");

    gal_device = platform_get_drvdata(dev);

    if (!gal_device) {
        ljmkFOOTER_ARG("r=%d", -1);
        return -1;
    }

    for (dev_index = 0; dev_index < gal_device->args.devCount; dev_index++) {
        device = gal_device->devices[dev_index];

        for (i = 0; i < ljvCORE_COUNT; i++) {
            if (device->kernels[i] != ljvNULL) {
                /* Store states. */
#if ljdENABLE_VG
                if (i == ljvCORE_VG)
                    status = ljkVGHARDWARE_QueryPowerManagementState(device->kernels[i]->vg->hardware,
                                                                     &device->statesStored[i]);
                else
                    status = ljkHARDWARE_QueryPowerState(device->kernels[i]->hardware,
                                                         &device->statesStored[i]);
#else
                status = ljkHARDWARE_QueryPowerState(device->kernels[i]->hardware,
                                                     &device->statesStored[i]);
#endif

                if (ljmIS_ERROR(status)) {
                    ljmkFOOTER_ARG("r=%d", -1);
                    return -1;
                }

#if ljdENABLE_VG
                if (i == ljvCORE_VG)
                    status = ljkVGHARDWARE_SetPowerState(device->kernels[i]->vg->hardware, ljvPOWER_OFF);
                else
                    status = ljkHARDWARE_SetPowerState(device->kernels[i]->hardware, ljvPOWER_OFF);
#else
                status = ljkHARDWARE_SetPowerState(device->kernels[i]->hardware, ljvPOWER_OFF);
#endif

                if (ljmIS_ERROR(status)) {
                    ljmkFOOTER_ARG("r=%d", -1);
                    return -1;
                }
            }
        }
    }

    //printk("----ljm_dev_suspend exist----\n");

    ljmkFOOTER_NO();
    return 0;
}

static int ljm_dev_resume(struct platform_device *dev)
{
    ljeSTATUS status;
    ljkGALDEVICE gal_device;
    ljkDEVICE device;
    ljtUINT i, dev_index;
    ljeCHIPPOWERSTATE statesStored;

    ljmkHEADER_ARG("dev=%p", dev);

    //printk("----ljm_dev_resume enter----\n");

    gal_device = platform_get_drvdata(dev);

    if (!gal_device) {
        ljmkFOOTER_ARG("r=%d", -1);
        return -1;
    }

    for (dev_index = 0; dev_index < gal_device->args.devCount; dev_index++) {
        device = gal_device->devices[dev_index];

        for (i = 0; i < ljvCORE_COUNT; i++) {
            if (device->kernels[i] != ljvNULL) {
#if ljdENABLE_VG
                if (i == ljvCORE_VG)
                    status = ljkVGHARDWARE_SetPowerState(device->kernels[i]->vg->hardware, ljvPOWER_ON);
                else
#endif
                    status = ljkHARDWARE_SetPowerState(device->kernels[i]->hardware, ljvPOWER_ON);

                if (ljmIS_ERROR(status)) {
                    ljmkFOOTER_ARG("r=%d", -1);
                    return -1;
                }

                /* Convert global state to crossponding internal state. */
                switch (device->statesStored[i]) {
                case ljvPOWER_ON:
                    statesStored = ljvPOWER_ON_AUTO;
                    break;
                case ljvPOWER_IDLE:
                    statesStored = ljvPOWER_IDLE_BROADCAST;
                    break;
                case ljvPOWER_SUSPEND:
                    statesStored = ljvPOWER_SUSPEND_BROADCAST;
                    break;
                case ljvPOWER_OFF:
                    statesStored = ljvPOWER_OFF_BROADCAST;
                    break;
                default:
                    statesStored = device->statesStored[i];
                    break;
                }

                /* Restore states. */
#if ljdENABLE_VG
                if (i == ljvCORE_VG)
                    status = ljkVGHARDWARE_SetPowerState(device->kernels[i]->vg->hardware, statesStored);
                else
#endif
                    status = ljkHARDWARE_SetPowerState(device->kernels[i]->hardware, statesStored);

                if (ljmIS_ERROR(status)) {
                    ljmkFOOTER_ARG("r=%d", -1);
                    return -1;
                }
            }
        }
    }

    //printk("----ljm_dev_resume exit----\n");

    ljmkFOOTER_NO();
    return 0;
}

#if defined(CONFIG_PM) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
#if 0
#ifdef CONFIG_PM_SLEEP
static int ljm_dev_system_suspend(struct device *dev)
{
    pm_message_t state = { 0 };
    int r;

    ljmkHEADER_ARG("dev=%p", dev);

    r = ljm_dev_suspend(to_platform_device(dev), state);

    ljmkFOOTER_ARG("r=%d", r);
    return r;
}

static int ljm_dev_system_resume(struct device *dev)
{
    int r;

    ljmkHEADER_ARG("dev=%p", dev);

    r = ljm_dev_resume(to_platform_device(dev));

    ljmkFOOTER_ARG("r=%d", r);
    return r;
}
# endif
#endif

int gpu_backup_external_vram(int card);
int gpu_backup_exclusive_vram(int card);
int gpu_restore_vram(int card);

int ljm_suspend_vga(struct drm_device *ddev);
int ljm_resume_vga(struct drm_device *ddev);

/**
 * gpu_suspend_kms - initiate device suspend
 *
 * @dev: dev pointer
 * @state: suspend state
 *
 * Puts the hw in the suspend state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver suspend.
 */
int gpu_suspend_kms(struct device *dev, struct drm_device *drm_dev, bool suspend, bool fbcon)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE gal_dev = platform_get_drvdata(platform_dev);
    struct ljm_device *ljm_dev = (struct ljm_device *)gal_dev->ljm_dev;
    struct pci_dev *pdev = ljm_dev->pdev;
    int ret;

    ljmkHEADER_ARG("dev=%p drm_dev=%p suspend=%d fbcon=%d", dev, drm_dev, suspend, fbcon);

    if (drm_dev == NULL || drm_dev->dev_private == NULL) {
        printk(KERN_ERR "\ndrm is NULL");
        ljmkFOOTER_ARG("r=%d", -ENODEV);
        return -ENODEV;
    }


    if (drm_dev->switch_power_state == DRM_SWITCH_POWER_OFF) {
        ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_DRIVER,
                       "%s(%d): switch_power_state is off\n",
                       __func__, __LINE__);
        ljmkFOOTER_NO();
        return 0;
    }

    ret = ljm_suspend_vga(drm_dev);
    if(ret) {
        printk(KERN_ERR "ljm_dev_suspend failed (%d)\n", ret);
        ljmkFOOTER_ARG("r=%d", ret);
        return ret;
    }

    /* synchronize vram to mem */
    gpu_backup_external_vram(ljm_dev->cur_card);
    gpu_backup_exclusive_vram(ljm_dev->cur_card);

    /* todo: suspend edma or wait edma to finish if needed */

    pci_save_state(pdev);

    if (suspend) {
        /* Shut down the device */
        pci_disable_device(pdev);
        pci_set_power_state(pdev, PCI_D3hot);
    }

    if (fbcon) {
        console_lock();
        if (ljm_dev->fbdev)
           drm_fb_helper_set_suspend(&ljm_dev->fbdev->helper, 1);
        console_unlock();
    }

    ljmkFOOTER_NO();
    return 0;
}

int gp100_pre_init(struct pci_dev *pdev, unsigned long flags);
int gp200_pre_init(int card, struct pci_dev *pdev, unsigned long flags, unsigned long *pddr_size);
int pcie_edma_init(int card);

/**
 * gpu_resume_kms - initiate device resume
 *
 * @drm_dev: drm dev pointer
 *
 * Bring the hw back to operating state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver resume.
 */
int gpu_resume_kms(struct device *dev, struct drm_device *drm_dev, bool resume, bool fbcon)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE gal_dev = platform_get_drvdata(platform_dev);
    struct ljm_device *ljm_dev = (struct ljm_device *)gal_dev->ljm_dev;
    struct pci_dev *pdev = ljm_dev->pdev;
    int ret = 0;

    ljmkHEADER_ARG("dev=%p drm_dev=%p resume=%d fbcon=%d", dev, drm_dev, resume, fbcon);

    if (drm_dev->switch_power_state == DRM_SWITCH_POWER_OFF) {
        ljmkFOOTER_NO();
        return 0;
    }

    if (fbcon) {
        console_lock();
    }

    if (resume) {
        pci_set_power_state(pdev, PCI_D0);
        pci_restore_state(pdev);
        if (pci_enable_device(pdev)) {
            if (fbcon)
                console_unlock();
            ljmkFOOTER_ARG("r=%d", -1);
            return -1;
        }
    }

    /* KMS pre-init */
    switch (ljm_dev->family) {
    case CHIP_GP200:
        ret = gp200_pre_init(ljm_dev->cur_card, pdev, ljm_dev->flags, NULL);
        break;
    default:
        break;
    }

    if (ret) {
        printk(KERN_ERR "\ninitialize vram failed: %d\n", ret);
    }
#if ljdENABLE_VPU
    /* todo: resume VPU */
    printk(KERN_INFO "vpu begin resume...");
    ret=hantrodec_resume();
    if (ret)
        printk(KERN_ERR "\nvpu_resume return %d\n", ret);
    printk(KERN_INFO "done\n");
#endif
    /* resume EDMA */
    pcie_edma_init(ljm_dev->cur_card);
    /* enable EDMA interrupt */
    LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN) | m_INTR_EN_EDMA);
    /* synchronize mem to vram */
    gpu_restore_vram(ljm_dev->cur_card);

    ljm_resume_vga(drm_dev);

    if (fbcon) {
        drm_helper_resume_force_mode(drm_dev);

        if(ljm_dev->fbdev) {
            drm_fb_helper_set_suspend(&ljm_dev->fbdev->helper, 0);
        }
        console_unlock();
    }

#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    /* need to be fixed */
#else
    drm_kms_helper_poll_enable(drm_dev);
#endif

    return 0;
}

#if !defined(CONFIG_ARM64)
static int gpu_pmops_suspend(struct device *dev)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE device = platform_get_drvdata(platform_dev);
    struct drm_device *drm_dev = (struct drm_device *)device->drm;
    int ret;
    pm_message_t state = { 0 };

    ljmkHEADER_ARG("dev=%p", dev);

    printk(KERN_INFO "gpu pmops suspend...");

    ret = ljm_dev_suspend(platform_dev, state);

    /* todo: add vpu suspend here */
    printk(KERN_INFO "vpu  suspend...");
    hantrodec_suspend();
    //hantrodec_cleanup(pcie_platform->pcie_info[0].pdev);
    ret = gpu_suspend_kms(dev, drm_dev, true, true);
    if (ret) {
        printk(KERN_ERR "\ngpu_suspend_kms return %d\n", ret);
    } else
        printk(KERN_INFO "done\n");

    ljmkFOOTER_ARG("r=%d", ret);
    return ret;
}
#endif

static int gpu_pmops_resume(struct device *dev)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE gal_dev = platform_get_drvdata(platform_dev);
    struct drm_device *drm_dev = (struct drm_device *)gal_dev->drm;
    int ret;

    ljmkHEADER_ARG("dev=%p", dev);

    printk(KERN_INFO "gpu pmops resume...");

    ret = gpu_resume_kms(dev, drm_dev, true, true);
    if (ret)
        printk(KERN_ERR "\ngpu_resume_kms return %d\n", ret);

    /* resume GPU cores */
    ret = ljm_dev_resume(platform_dev);
    if (ret)
        printk(KERN_ERR "\ngpu_dev_resume return %d\n", ret);

    ljmkFOOTER_ARG("r=%d", ret); msleep(5000);
    return ret;
}

static int gpu_pmops_freeze(struct device *dev)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE gal_dev;
    struct drm_device *drm_dev;
    int ret;
    pm_message_t state = { 0 };
    ljmkHEADER_ARG("dev=%p platform_dev=%p", dev, platform_dev);

    gal_dev = platform_get_drvdata(platform_dev);

    if (!gal_dev) {
        ljmkFOOTER_ARG("r=%d", -1);
        return -1;
    }

    drm_dev = (struct drm_device *)gal_dev->drm;

    printk(KERN_INFO "gpu pmops freeze...");

    /* gpu core suspend */
    ret = ljm_dev_suspend(platform_dev, state);

    /* todo: add vpu suspend */
#if ljdENABLE_VPU
    hantrodec_suspend();
#endif
    /* pci device suspend */
    ret = gpu_suspend_kms(dev, drm_dev, false, true);
    if (ret) {
        printk(KERN_ERR "\ngpu_suspend_kms return %d\n", ret);
    } else
        printk(KERN_INFO "done\n");

    ljmkFOOTER_ARG("r=%d", ret);
    return ret;
}

static int gpu_pmops_thaw(struct device *dev)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE device = platform_get_drvdata(platform_dev);
    struct drm_device *drm_dev = (struct drm_device *)device->drm;
    int ret;

    ljmkHEADER_ARG("dev=%p", dev);

    //printk(KERN_ERR "----gpu pmops thaw...");

    ret = gpu_resume_kms(dev, drm_dev, false, true);
    //printk(KERN_ERR "done----\n");

    ljmkFOOTER_ARG("r=%d", ret); msleep(5000);
    return ret;
}

#if 1
static int gpu_pmops_runtime_suspend(struct device *dev)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE device = platform_get_drvdata(platform_dev);
    struct drm_device *drm_dev = (struct drm_device *)device->drm;
    struct ljm_device *ljm_dev = (struct ljm_device *)device->ljm_dev;
    struct pci_dev *pdev = ljm_dev->pdev;

    //printk(KERN_ERR "----gpu pmops runtime suspend...");

    drm_dev->switch_power_state = DRM_SWITCH_POWER_CHANGING;

#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    /* need to be fixed */
#else
    drm_kms_helper_poll_disable(drm_dev);
#endif
    //vga_switcheroo_set_dynamic_switch(pdev, VGA_SWITCHEROO_OFF);

    gpu_suspend_kms(dev, drm_dev, false, false);

    pci_save_state(pdev);
    pci_disable_device(pdev);
    pci_ignore_hotplug(pdev);
    pci_set_power_state(pdev, PCI_D3cold);
    drm_dev->switch_power_state = DRM_SWITCH_POWER_DYNAMIC_OFF;

    //printk(KERN_ERR "done----\n");
    return 0;
}

static int gpu_pmops_runtime_resume(struct device *dev)
{
    struct platform_device *platform_dev = to_platform_device(dev);
    ljkGALDEVICE device = platform_get_drvdata(platform_dev);
    struct drm_device *drm_dev = (struct drm_device *)device->drm;
    struct ljm_device *ljm_dev = (struct ljm_device *)device->ljm_dev;
    struct pci_dev *pdev = ljm_dev->pdev;
    int ret;

    //printk(KERN_ERR "----gpu pmops runtime resume...");

    drm_dev->switch_power_state = DRM_SWITCH_POWER_CHANGING;

    pci_set_power_state(pdev, PCI_D0);
    pci_restore_state(pdev);
    ret = pci_enable_device(pdev);
    if (ret)
        return ret;
    pci_set_master(pdev);

    ret = gpu_resume_kms(dev, drm_dev, false, false);

#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    /* need to be fixed */
#else
    drm_kms_helper_poll_enable(drm_dev);
#endif
    //vga_switcheroo_set_dynamic_switch(pdev, VGA_SWITCHEROO_ON);
    drm_dev->switch_power_state = DRM_SWITCH_POWER_ON;

    //printk(KERN_ERR "done----\n");
    return 0;
}

static int gpu_pmops_runtime_idle(struct device *dev)
{
    pm_runtime_mark_last_busy(dev);
    pm_runtime_autosuspend(dev);

    return 1;
}
#endif

#if 0
static const struct dev_pm_ops ljm_dev_pm_ops = {
    SET_SYSTEM_SLEEP_PM_OPS(ljm_dev_system_suspend, ljm_dev_system_resume)
};
#else   /* !0 */
static const struct dev_pm_ops ljm_dev_pm_ops = {
#if defined CONFIG_ARM64
    .suspend = gpu_pmops_freeze,
#else
    .suspend = gpu_pmops_suspend,
#endif
    .resume = gpu_pmops_resume,
    .freeze = gpu_pmops_freeze,
    .thaw = gpu_pmops_thaw,
    .poweroff = gpu_pmops_freeze,
    .restore = gpu_pmops_resume,
    .runtime_suspend = gpu_pmops_runtime_suspend,
    .runtime_resume = gpu_pmops_runtime_resume,
    .runtime_idle = gpu_pmops_runtime_idle,
};
#endif  /* ~0 */
#endif  /* ~CONFIG_PM && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30) */

static struct platform_driver ljm_dev_driver = {
    .probe = ljm_dev_probe,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
    .remove = ljm_dev_remove,
#else
    .remove = __devexit_p(ljm_dev_remove),
#endif

#if !defined(CONFIG_PM)
    .suspend = ljm_dev_suspend,
    .resume = ljm_dev_resume,
#endif
    .shutdown = ljm_dev_shutdown,

    .driver = {
        .owner = THIS_MODULE,
        .name = DEVICE_NAME,
#if defined(CONFIG_PM) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
        .pm = &ljm_dev_pm_ops,
#endif
    }
};

static int __init ljm_dev_init(void)
{
    int ret = 0;

    ret = ljkPLATFORM_Init(&ljm_dev_driver, &platform);

    if (ret || !platform) {
        pr_err("ljmcore: Soc platform init failed.\n");
        return -ENODEV;
    }

    ret = platform_driver_register(&ljm_dev_driver);

    if (ret) {
        pr_err("ljmcore: gpu_init() failed to register driver!\n");
        ljkPLATFORM_Terminate(platform);
        platform = NULL;
        return -ENODEV;
    }

    platform->driver = &ljm_dev_driver;

    return 0;
}

static void __exit ljm_dev_exit(void)
{
    platform_driver_unregister(&ljm_dev_driver);

    ljkPLATFORM_Terminate(platform);
    platform = NULL;
}

module_init(ljm_dev_init);

module_exit(ljm_dev_exit);
