/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#if ljdENABLE_DRM

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
#include <drm/drm_drv.h>
#include <drm/drm_file.h>
#include <drm/drm_ioctl.h>
#include <drm/drm_print.h>
#include <drm/drm_vblank.h>
#include <drm/drm_crtc.h>
#include <drm/drm_framebuffer.h>
#include <drm/drm_fourcc.h>
#include <drm/drm_modes.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include <drm/drm_irq.h>
#else
#include <drm/drm_legacy.h>
#endif
#else
#include <drm/drmP.h>
#endif
#include <drm/drm_gem.h>
#include <linux/dma-buf.h>
#include "lj_hal_kernel_linux.h"
#include "lj_hal_drm.h"
#include "ljmicro_device.h"

#define _LJ_OBJ_ZONE ljvZONE_KERNEL

uint hDisplay = 0;
module_param(hDisplay, uint, 0644);
MODULE_PARM_DESC(hDisplay, "Horizontal display size");

uint vDisplay = 0;
module_param(vDisplay, uint, 0644);
MODULE_PARM_DESC(vDisplay, "Vertical display size");

uint hdmi0ToDvi = 0;	/* 0: hdmi port 0 connects hdmi monitor; 1: hdmi port 0 connects dvi monitor */
module_param(hdmi0ToDvi, uint, 0644);
MODULE_PARM_DESC(hdmi0ToDvi, "0 - HDMI 0 connects to HDMI monitor (Default), 1 - HDMI 0 connects to DVI monitor");

uint hdmi1ToDvi = 0;	/* 0: hdmi port 1 connects hdmi monitor; 1: hdmi port 1 connects dvi monitor */
module_param(hdmi1ToDvi, uint, 0644);
MODULE_PARM_DESC(hdmi1ToDvi, "0 - HDMI 1 connects to HDMI monitor (Default), 1 - HDMI 1 connects to DVI monitor");

#if NOEDID
uint ddcEnable = 0;
module_param(ddcEnable, uint, 0644);
MODULE_PARM_DESC(ddcEnable, "0 - CRTC disable DDC (Default), 1 - CRTC enable DDC");
#else
uint ddcEnable = 1;
module_param(ddcEnable, uint, 0644);
MODULE_PARM_DESC(ddcEnable, "0 - CRTC disable DDC, 1 - CRTC enable DDC (Default)");
#endif

int bweIntEnable = 0;
module_param(bweIntEnable, int, 0644);
MODULE_PARM_DESC(bweIntEnable, "Disable/Enable bandwith error interrupt");

ulong bweIntCount = 0;
module_param(bweIntCount, ulong, 0644);
MODULE_PARM_DESC(bweIntCount, "Bandwith error interrupt counter");

int ljm_driver_load_kms(struct drm_device *ddev, unsigned long flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
void ljm_driver_unload_kms(struct drm_device *ddev);
#else
int ljm_driver_unload_kms(struct drm_device *ddev);
#endif
void ljm_driver_irq_preinstall_kms(struct drm_device *ddev);
int ljm_driver_irq_postinstall_kms(struct drm_device *ddev);
void ljm_driver_irq_uninstall_kms(struct drm_device *ddev);
irqreturn_t ljm_driver_irq_handler_kms(int irq, void *arg);

#if USE_KMS_VBLANK
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
u32 ljm_get_vblank_counter_kms(struct drm_device *ddev, unsigned int pipe);
int ljm_enable_vblank_kms(struct drm_device *ddev, unsigned int pipe);
void ljm_disable_vblank_kms(struct drm_device *ddev, unsigned int pipe);
//bool ljm_get_vblank_timestamp_kms(struct drm_device *ddev, unsigned int pipe, int *max_error, ktime_t *vblank_time, bool in_vblank_irq)
#endif
#endif

#if !ljdENABLE_TTM

/*******************************************************************************
 ****************************** ljkKERNEL DRM Code *****************************
 *******************************************************************************/

#if 0
struct ljm_gem_object {
    struct drm_gem_object   base;
    uint32_t                node_handle;
    ljkVIDMEM_NODE          node_object;
    ljtBOOL                 cacheable;

    unsigned              pin_count;
    void                  *kptr;      /* kernel virtual address */
};
#else
#define ljm_gem_object ljm_gem_object
#endif

#define  BANK_CIRCLE   6;
static unsigned int bank_index = 0;
static unsigned int pick_bank[8] = {
    ljvVIDMEM_TYPE_COLOR_BUFFER,
    ljvVIDMEM_TYPE_DEPTH_BUFFER,
    ljvVIDMEM_TYPE_TEXTURE,
    ljvVIDMEM_TYPE_VERTEX_BUFFER,
    ljvVIDMEM_TYPE_INDEX_BUFFER,
    ljvVIDMEM_TYPE_TILE_STATUS,
    ljvVIDMEM_TYPE_COMMAND,
    ljvVIDMEM_TYPE_GENERIC
};

static const struct vm_operations_struct ljm_gem_vm_ops = {
//    .fault  = ljm_vm_fault,
    .open   = drm_gem_vm_open,
    .close  = drm_gem_vm_close,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
void ljm_gem_free_object(struct drm_gem_object *gem_obj);
struct dma_buf *ljm_gem_prime_export(struct drm_gem_object *gem_obj, int flags);

/* static *//*const struct drm_gem_object_funcs ljm_gem_object_funcs = {
    .free = ljm_gem_free_object,*/
/*static*/ const struct drm_gem_object_funcs ljm_gem_object_funcs = {
    .free   = ljm_gem_free_object,
    .export = ljm_gem_prime_export,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,6,0)
    .vm_ops = &ljm_gem_vm_ops,
#endif
};
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)
struct dma_buf *ljm_gem_prime_export(struct drm_gem_object *gem_obj, int flags)
{
    struct drm_device *drm = gem_obj->dev;
#else
struct dma_buf *ljm_gem_prime_export(struct drm_device *drm, struct drm_gem_object *gem_obj, int flags)
{
#endif
    struct ljm_gem_object *ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);
    struct dma_buf *dmabuf  = ljvNULL;
    ljkGALDEVICE gal_dev = (ljkGALDEVICE)drm->dev_private;

    if (gal_dev) {
        ljkDEVICE device = gal_dev->devices[0];
        ljkKERNEL kernel = device->kernels[0];

        ljmkVERIFY_OK(ljkVIDMEM_NODE_Export(kernel, ljm_obj->node_object, flags,
                                            (ljtPOINTER *)&dmabuf, ljvNULL));
    }

    return dmabuf;
}

struct drm_gem_object *ljm_gem_prime_import(struct drm_device *drm, struct dma_buf *dmabuf)
{
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj;

    ljsHAL_INTERFACE iface;
    ljkGALDEVICE gal_dev;
    ljkDEVICE device;
    ljkKERNEL kernel;
    ljtUINT32 processID;
    ljkVIDMEM_NODE nodeObject;
    ljeSTATUS status = ljvSTATUS_OK;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    ljkOS_ZeroMemory(&iface, sizeof(iface));
    iface.command = ljvHAL_WRAP_USER_MEMORY;
    iface.hardwareType = device->defaultHwType;
    iface.u.WrapUserMemory.desc.flag = ljvALLOC_FLAG_DMABUF;
    iface.u.WrapUserMemory.desc.handle = -1;
    iface.u.WrapUserMemory.desc.dmabuf = ljmPTR_TO_UINT64(dmabuf);
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

    kernel = device->kernels[0];
    ljmkONERROR(ljkOS_GetProcessID(&processID));
    ljmkONERROR(ljkVIDMEM_HANDLE_Lookup(kernel,
                                        processID,
                                        iface.u.WrapUserMemory.node,
                                        &nodeObject));

    /* ioctl output */
    gem_obj = kzalloc(sizeof(*ljm_obj), GFP_KERNEL);
    drm_gem_private_object_init(drm, gem_obj, dmabuf->size);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
    gem_obj->funcs = &ljm_gem_object_funcs;
#endif
    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);
    ljm_obj->node_handle = iface.u.WrapUserMemory.node;
    ljm_obj->node_object = nodeObject;

OnError:
    return gem_obj;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,11,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,7,0)
#else
int ljm_gem_prime_handle_to_fd(struct drm_device *dev,
                               struct drm_file *file_priv, uint32_t handle,
                               uint32_t flags,
                               int *prime_fd)
{
    struct drm_gem_object *obj;
    unsigned int refs = 0;
    int ret;
    obj = drm_gem_object_lookup(file_priv, handle);
    if (obj) {
        drm_gem_object_unreference_unlocked(obj);
        refs = obj->refcount.refcount.counter;
//      printk("---ljm_gem_prime_handle_to_fd--before---refcount=%u-----\n",obj->refcount.refcount.counter);
    }
    ret = drm_gem_prime_handle_to_fd(dev,file_priv,handle,flags,prime_fd);
    if (obj) {
        while(obj->refcount.refcount.counter > refs)
            drm_gem_object_unreference_unlocked(obj);
//          printk("---ljm_gem_prime_handle_to_fd--after---refcount=%u-----\n",obj->refcount.refcount.counter);
    }
    return ret;
}
#endif
#endif

void ljm_gem_free_object(struct drm_gem_object *gem_obj)
{
    struct ljm_gem_object *ljm_obj  = container_of(gem_obj, struct ljm_gem_object, base);
    struct drm_device *drm = gem_obj->dev;

    ljsHAL_INTERFACE iface;
    ljkGALDEVICE gal_dev = (ljkGALDEVICE)drm->dev_private;
    ljkDEVICE device;

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    if (ljm_obj->node_handle == (uint32_t)(-1)) {
        DRM_INFO("free vram: %zu\n", ljm_obj->base.size);
        ljkVIDMEM_NODE_Dereference(ljm_obj->node_object->kernel, ljm_obj->node_object);
    } else {
        device = gal_dev->devices[0];

        ljkOS_ZeroMemory(&iface, sizeof(iface));
        iface.command = ljvHAL_RELEASE_VIDEO_MEMORY;
        iface.hardwareType = device->defaultHwType;
        iface.u.ReleaseVideoMemory.node = ljm_obj->node_handle;
        iface.u.ReleaseVideoMemory.pid = ljm_obj->pid;
        ljmkVERIFY_OK(ljkDEVICE_Dispatch(device, &iface));
    }

    drm_gem_object_release(gem_obj);
    kfree(gem_obj);
}

static int ljm_ioctl_gem_create(struct drm_device *drm, void *data, struct drm_file *file)
{
    int ret = 0;
    struct drm_ljm_gem_create *args = (struct drm_ljm_gem_create *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljkGALDEVICE gal_dev;
    ljkKERNEL kernel;
    ljtUINT32 processID;
    ljkVIDMEM_NODE nodeObject;
    uint32_t nodeHandle;
    ljtUINT32 flags = ljvALLOC_FLAG_DMABUF_EXPORTABLE;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtUINT64 alignSize = PAGE_ALIGN(args->size);
    ljkDEVICE device;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    flags |= ljvALLOC_FLAG_FROM_USER;

    if (args->flags & DRM_VIV_GEM_CONTIGUOUS)
        flags |= ljvALLOC_FLAG_CONTIGUOUS;

    if (args->flags & DRM_VIV_GEM_CACHED) {
        DRM_DEBUG_KMS("ljm_ioctl_gem_create(%d): DRM_VIV_GEM_CACHED\n", __LINE__);
        flags |= ljvALLOC_FLAG_CACHEABLE;
    }

    if (args->flags & DRM_VIV_GEM_SECURE)
        flags |= ljvALLOC_FLAG_SECURITY;

    if (args->flags & DRM_VIV_GEM_EXCLUSIVE_NONEED_VIRTUAL)
        flags |= ljvALLOC_FLAG_EXCLUSIVE_NONEED_VIRTUAL;

    ljkOS_ZeroMemory(&iface, sizeof(iface));
    iface.command = ljvHAL_ALLOCATE_LINEAR_VIDEO_MEMORY;
    iface.hardwareType = device->defaultHwType;
    iface.u.AllocateLinearVideoMemory.bytes = alignSize;
    iface.u.AllocateLinearVideoMemory.alignment = args->alignment;  //256;
    iface.u.AllocateLinearVideoMemory.flag = flags;

    if(args->flags & DRM_VIV_GEM_CHANGE_BANK) {
        iface.u.AllocateLinearVideoMemory.type = pick_bank[bank_index++];
        bank_index %= BANK_CIRCLE;
    }
    else {
        iface.u.AllocateLinearVideoMemory.type = ljvVIDMEM_TYPE_GENERIC;
    }

    switch (args->domains) {
        case LJM_GEM_DOMAIN_DEFAULT:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_DEFAULT;//ljvPOOL_LOCAL_EXTERNAL;
            break;
        case LJM_GEM_DOMAIN_VRAM:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_LOCAL_EXTERNAL_EXCLUSIVE;
            break;
        case LJM_GEM_DOMAIN_INVISIBLE_VRAM:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_LOCAL_EXCLUSIVE_EXTERNAL;
            break;
        case LJM_GEM_DOMAIN_GTT:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_VIRTUAL;
            break;
        case LJM_GEM_DOMAIN_VRAM_NOEVICT:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_LOCAL_EXTERNAL;
            break;
        case LJM_GEM_DOMAIN_INVISIBLE_NOEVICT:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_LOCAL_EXCLUSIVE;
            break;
        case LJM_GEM_DOMAIN_INVISIBLE_FIRST:
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_LOCAL_EXCLUSIVE_EXTERNAL_VIRTUAL;
            break;
        default :
            iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_DEFAULT;// ljvPOOL_LOCAL_EXTERNAL;
    }

    kernel = device->kernels[0];

#if USE_SCANOUT_PREALLOCATE
    if ((args->flags & DRM_VIV_GEM_PAGEFLIP_SCANOUT) && (args->size <= MAX_SCANOUT_NODE_SIZE))
    {
        int crtc, scanout_id;
        int i, j;

        processID = 1;  /* processID=1 for pre-allocated scanout objects */
        //ljmkONERROR(ljkOS_GetProcessID(&processID));

        crtc = (args->flags & DRM_VIV_GEM_CRTC_INDEX_MASK) >> DRM_VIV_GEM_CRTC_INDEX_SHIFT;
        scanout_id = (args->flags & DRM_VIV_GEM_PAGEFLIP_SCANOUT_BACK) ? 1 : 0;

        if (gal_dev->scanoutNodeObjects[0][0] == ljvNULL) {
            /* allocate scanout node objects */
            //printk("----allocate all scanouts (%d:%d)----\n", crtc, scanout_id);

            iface.u.AllocateLinearVideoMemory.pid   = processID;
            iface.u.AllocateLinearVideoMemory.pool  = ljvPOOL_LOCAL_EXCLUSIVE_EXTERNAL;
            iface.u.AllocateLinearVideoMemory.bytes = MAX_SCANOUT_NODE_SIZE;
            for (i=0; i<4; i++) {
                iface.u.AllocateLinearVideoMemory.type = pick_bank[i];
                iface.u.AllocateLinearVideoMemory.alignment = (128<<20);

                for (j=0; j<2; j++) {
                    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));
                    ljmkONERROR(ljkVIDMEM_HANDLE_Lookup(kernel, processID,
                                                        iface.u.AllocateLinearVideoMemory.node,
                                                        &gal_dev->scanoutNodeObjects[i][j]));
                    gal_dev->scanoutNodeHandles[i][j] = iface.u.AllocateLinearVideoMemory.node;

                    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE, "new scanout (%d:%d): handle=%d object=%p offset=%lx\n",
                        i, j, gal_dev->scanoutNodeHandles[i][j], gal_dev->scanoutNodeObjects[i][j],
                        (unsigned long)gal_dev->scanoutNodeObjects[i][j]->node->VidMem.offset);

                    ljmkONERROR(ljkVIDMEM_HANDLE_Reference(kernel, processID, gal_dev->scanoutNodeHandles[i][j]));
                    ljmkONERROR(ljkVIDMEM_NODE_Reference(kernel, gal_dev->scanoutNodeObjects[i][j]));
                }
            }
        }

        if (gal_dev->scanoutNodeObjects[crtc][scanout_id] == ljvNULL)
            ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);

        nodeHandle = gal_dev->scanoutNodeHandles[crtc][scanout_id];
        nodeObject = gal_dev->scanoutNodeObjects[crtc][scanout_id];
    }
    else
#endif
    {
        ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));    /* iface.u.AllocateLinearVideoMemory.pid = 0 */

        ljmkONERROR(ljkOS_GetProcessID(&processID));
        ljmkONERROR(ljkVIDMEM_HANDLE_Lookup(kernel, processID,
                                            iface.u.AllocateLinearVideoMemory.node,
                                            &nodeObject));
        nodeHandle = iface.u.AllocateLinearVideoMemory.node;
    }

    /* ioctl output */
    gem_obj = kzalloc(sizeof(*ljm_obj), GFP_KERNEL);
    drm_gem_private_object_init(drm, gem_obj, (size_t)alignSize);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
    //gem_obj->funcs = &ljm_gem_object_funcs;
    gem_obj->funcs = &ljm_gem_object_funcs;
#endif
    ret = drm_gem_handle_create(file, gem_obj, &args->handle);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);
    ljm_obj->node_handle = nodeHandle;
    ljm_obj->node_object = nodeObject;
    ljm_obj->cacheable = flags & ljvALLOC_FLAG_CACHEABLE;
    ljm_obj->pid = processID;

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p %d",
                   ljm_obj->node_handle, ljm_obj->node_object, processID);

    /* drop reference from allocate - handle holds it now */
    drm_gem_object_unreference_unlocked(gem_obj);

OnError:
    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_lock(struct drm_device *drm, void *data, struct drm_file *file)
{
    struct drm_ljm_gem_lock *args = (struct drm_ljm_gem_lock *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;
    ljkKERNEL kernel;
    struct _ljsPLATFORM_PCIE *pcie_platform;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    ljkOS_ZeroMemory(&iface, sizeof(iface));
    iface.command = ljvHAL_LOCK_VIDEO_MEMORY;
    iface.hardwareType = device->defaultHwType;
    iface.u.LockVideoMemory.op = ljvLOCK_VIDEO_MEMORY_OP_LOCK |
                                 ljvLOCK_VIDEO_MEMORY_OP_MAP;
    iface.u.LockVideoMemory.node = ljm_obj->node_handle;
    iface.u.LockVideoMemory.cacheable = args->cacheable;
    iface.u.LockVideoMemory.pid = ljm_obj->pid;
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

    args->logical = iface.u.LockVideoMemory.memory;
    args->physical = iface.u.LockVideoMemory.address;

    /* get GPU bus address for VPU, dma, etc */
    kernel = ljm_obj->node_object->kernel;
    if (ljm_obj->node_object->pool == ljvPOOL_LOCAL_EXTERNAL) {
        /* visible memory */
        pcie_platform = (struct _ljsPLATFORM_PCIE *)gal_dev->platform;
        args->gpuBusAddr = pcie_platform->pcie_info[kernel->device->platformIndex].gpu_vram_start + ljm_obj->node_object->node->VidMem.offset;
    } else if (ljm_obj->node_object->pool == ljvPOOL_LOCAL_EXCLUSIVE) {
        /* invisible memory */
        args->gpuBusAddr = gal_dev->exclusiveBase[kernel->device->platformIndex] + ljm_obj->node_object->node->VidMem.offset;
    } else {
        /* system memory */
        args->gpuBusAddr = iface.u.LockVideoMemory.physicalAddress;
    }

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_unlock(struct drm_device *drm, void *data, struct drm_file *file)
{
    struct drm_ljm_gem_unlock *args = (struct drm_ljm_gem_unlock *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    memset(&iface, 0, sizeof(iface));
    iface.command = ljvHAL_UNLOCK_VIDEO_MEMORY;
    iface.hardwareType = device->defaultHwType;
    iface.u.UnlockVideoMemory.op = ljvLOCK_VIDEO_MEMORY_OP_UNLOCK |
                                       ljvLOCK_VIDEO_MEMORY_OP_UNMAP;
    iface.u.UnlockVideoMemory.node = (ljtUINT64)ljm_obj->node_handle;
    iface.u.UnlockVideoMemory.type = ljvVIDMEM_TYPE_GENERIC;
    iface.u.UnlockVideoMemory.pid = ljm_obj->pid;
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

    memset(&iface, 0, sizeof(iface));
    iface.command = ljvHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY;
    iface.hardwareType = device->defaultHwType;
    iface.u.BottomHalfUnlockVideoMemory.node = (ljtUINT64)ljm_obj->node_handle;
    iface.u.BottomHalfUnlockVideoMemory.type = ljvVIDMEM_TYPE_GENERIC;
    iface.u.BottomHalfUnlockVideoMemory.pid = ljm_obj->pid;
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_cache(struct drm_device *drm, void *data, struct drm_file *file)
{
    struct drm_ljm_gem_cache *args = (struct drm_ljm_gem_cache *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;
    ljeCACHEOPERATION cache_op = 0;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    switch (args->op) {
    case DRM_VIV_GEM_CLEAN_CACHE:
        cache_op = ljvCACHE_CLEAN;
        break;
    case DRM_VIV_GEM_INVALIDATE_CACHE:
        cache_op = ljvCACHE_INVALIDATE;
        break;
    case DRM_VIV_GEM_FLUSH_CACHE:
        cache_op = ljvCACHE_FLUSH;
        break;
    case DRM_VIV_GEM_MEMORY_BARRIER:
        cache_op = ljvCACHE_MEMORY_BARRIER;
        break;
    default:
        break;
    }

    ljkOS_ZeroMemory(&iface, sizeof(iface));
    iface.command           = ljvHAL_CACHE;
    iface.hardwareType      = device->defaultHwType;
    iface.u.Cache.node      = ljm_obj->node_handle;
    iface.u.Cache.operation = cache_op;
    iface.u.Cache.logical   = args->logical;
    iface.u.Cache.bytes     = args->bytes;
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_query(struct drm_device *drm, void *data, struct drm_file *file)
{
    struct drm_ljm_gem_query *args = (struct drm_ljm_gem_query *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;
    struct ljm_device *ljm_dev;

    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkKERNEL kernel;
    struct _ljsPLATFORM_PCIE *pcie_platform;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    ljm_dev = (struct ljm_device *)gal_dev->ljm_dev;
    pcie_platform = (struct _ljsPLATFORM_PCIE *)gal_dev->platform;

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    kernel = ljm_obj->node_object->kernel;

    switch (args->param) {
    case DRM_VIV_GEM_PARAM_POOL:
        args->value = (__u64)ljm_obj->node_object->pool;
        break;
    case DRM_VIV_GEM_PARAM_SIZE:
        args->value = (__u64)gem_obj->size;
        break;
    case DRM_VIV_GEM_PARAM_OFFSET:
        args->value = (__u64)(BO_IS_VIDMEM(ljm_obj) ? BO_GET_VIDMEM_OFFSET(ljm_obj) : 0);
        break;
    case DRM_VIV_GEM_PARAM_GPU_ADDR:
        args->value = (__u64)(BO_IS_VIDMEM(ljm_obj) ? (kernel->externalBaseAddress + BO_GET_VIDMEM_OFFSET(ljm_obj)) : 0);
        break;
    case DRM_VIV_GEM_PARAM_CPU_ADDR:
        args->value = BO_IS_VIDMEM(ljm_obj) ? (BO_GET_VIDMEM_START(ljm_obj) + BO_GET_VIDMEM_OFFSET(ljm_obj)) : 0x0;
        break;
    case DRM_VIV_GEM_PARAM_CPU_TYPE:
        args->value = 0;	/* CPU_TYPE_UNKNOWN; */
/*        if (is_ft1500a()) args->value = CPU_TYPE_FT1500A;
        else if (is_ft2000ahk()) args->value = CPU_TYPE_FT2000AHK;
        else if (is_ft2000plus()) args->value = CPU_TYPE_FT2000PLUS;
        else if (is_ft2000a4()) args->value = CPU_TYPE_FT2000A_4;
*/
        break;
    case DRM_VIV_GEM_PARAM_SWCURSOR:    /* set software cursor */
//            ljm_dev = (struct ljm_device *)gal_dev->ljm_dev;
//            ljm_dev->hw_cursor = false;
            break;
    case DRM_VIV_GEM_PARAM_HWCURSOR:    /* set hardware cursor */
//        ljm_dev = (struct ljm_device *)gal_dev->ljm_dev;
//        ljm_dev->hw_cursor = true;
        break;
    case DRM_VIV_GEM_PARAM_VERSION:     /* get csmcore version */
//        args->value = ((__u64)CSM_VERSION_MAJOR<<48) | ((__u64)CSM_VERSION_MINOR<<40) | ((__u64)CSM_VERSION_PATCH<<32) | CSM_VERSION_BUILD;
        args->value = 0x2000012345678ULL;
        break;
    case DRM_VIV_GEM_PARAM_REF:
//        printk("----ljm_gem_query--%p--refcount %d-----------\n", ljm_obj->node_object, kref_read(&gem_obj->refcount)-1);
        break;
    case DRM_VIV_GEM_IOMMU:
        args->value = 0;
        break;
    default:
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
    }

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_timestamp(struct drm_device *drm,
                                   void *data, struct drm_file *file)
{
    struct drm_ljm_gem_timestamp *args    = (struct drm_ljm_gem_timestamp *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    ljm_obj->node_object->timeStamp += args->inc;
    args->timestamp = ljm_obj->node_object->timeStamp;

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_set_tiling(struct drm_device *drm,
                                    void *data, struct drm_file *file)
{
    struct drm_ljm_gem_set_tiling *args    = (struct drm_ljm_gem_set_tiling *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    ljm_obj->node_object->tilingMode  = args->tiling_mode;
    ljm_obj->node_object->tsMode      = args->ts_mode;
    ljm_obj->node_object->tsCacheMode = args->ts_cache_mode;
    ljm_obj->node_object->clearValue  = args->clear_value;

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_get_tiling(struct drm_device *drm,
                                    void *data, struct drm_file *file)
{
    struct drm_ljm_gem_get_tiling *args = (struct drm_ljm_gem_get_tiling *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    args->tiling_mode   = ljm_obj->node_object->tilingMode;
    args->ts_mode       = ljm_obj->node_object->tsMode;
    args->ts_cache_mode = ljm_obj->node_object->tsCacheMode;
    args->clear_value   = ljm_obj->node_object->clearValue;

OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_attach_aux(struct drm_device *drm,
                                    void *data, struct drm_file *file)
{
    struct drm_ljm_gem_attach_aux *args = (struct drm_ljm_gem_attach_aux *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;
    struct drm_gem_object *gem_ts_obj = ljvNULL;

    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;
    ljkVIDMEM_NODE nodeObj = ljvNULL;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    nodeObj = ljm_obj->node_object;

    /* do not support re-attach */
    if (nodeObj->tsNode)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    if (args->ts_handle) {
        struct ljm_gem_object *ljm_ts_obj;
        ljkKERNEL kernel = device->kernels[0];
        ljsHAL_INTERFACE iface;
        ljtBOOL is128BTILE = ljkHARDWARE_IsFeatureAvailable(kernel->hardware,
                                                            ljvFEATURE_128BTILE);
        ljtBOOL is2BitPerTile = is128BTILE ?
                                ljvFALSE :
                                ljkHARDWARE_IsFeatureAvailable(kernel->hardware,
                                                               ljvFEATURE_TILE_STATUS_2BITS);
        ljtBOOL isCompressionDEC400 = ljkHARDWARE_IsFeatureAvailable(kernel->hardware,
                                                                     ljvFEATURE_COMPRESSION_DEC400);
        ljtPOINTER entry = ljvNULL;
        ljkVIDMEM_NODE ObjNode = ljvNULL;
        ljtUINT32 processID = 0;
        ljtUINT32 tileStatusFiller = (isCompressionDEC400 ||
                                           ((kernel->hardware->identity.chipModel == ljv500) &&
                                            (kernel->hardware->identity.chipRevision > 2))) ?
                                          0xFFFFFFFF :
                                          is2BitPerTile ? 0x55555555 : 0x11111111;

        gem_ts_obj = drm_gem_object_lookup(file, args->ts_handle);
        if (!gem_ts_obj)
            ljmkONERROR(ljvSTATUS_NOT_FOUND);

        ljm_ts_obj = container_of(gem_ts_obj, struct ljm_gem_object, base);

        ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                       "ljm_ts_obj: %d %p",
                       ljm_ts_obj->node_handle, ljm_ts_obj->node_object);

        ljmkONERROR(ljkVIDMEM_NODE_Reference(kernel, ljm_ts_obj->node_object));
        nodeObj->tsNode = ljm_ts_obj->node_object;

        ljmkONERROR(ljkOS_GetProcessID(&processID));
        if (ljm_ts_obj->pid > 0)
            processID = ljm_ts_obj->pid;

        /* Fill tile status node with tileStatusFiller value first time to avoid GPU hang. */
        /* Lock tile status node. */
        ljkOS_ZeroMemory(&iface, sizeof(iface));
        iface.command = ljvHAL_LOCK_VIDEO_MEMORY;
        iface.hardwareType = device->defaultHwType;
        iface.u.LockVideoMemory.op = ljvLOCK_VIDEO_MEMORY_OP_LOCK |
                                       ljvLOCK_VIDEO_MEMORY_OP_MAP;
        iface.u.LockVideoMemory.node = ljm_ts_obj->node_handle;
        iface.u.LockVideoMemory.cacheable = ljm_ts_obj->cacheable;
        iface.u.LockVideoMemory.pid = processID;
        ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

        ljmkONERROR(ljkVIDMEM_HANDLE_Lookup(kernel, processID, ljm_ts_obj->node_handle, &ObjNode));
        ljmkONERROR(ljkVIDMEM_NODE_LockCPU(kernel, ObjNode, ljvFALSE, ljvFALSE, &entry));

        /* Fill tile status node with tileStatusFiller. */
        memset(entry, tileStatusFiller, (__u64)gem_ts_obj->size);
        ljmkONERROR(ljkVIDMEM_NODE_UnlockCPU(kernel, ObjNode, 0, ljvFALSE, ljvFALSE));

        /* UnLock tile status node. */
        memset(&iface, 0, sizeof(iface));
        iface.command = ljvHAL_UNLOCK_VIDEO_MEMORY;
        iface.hardwareType = device->defaultHwType;
        iface.u.UnlockVideoMemory.op = ljvLOCK_VIDEO_MEMORY_OP_UNLOCK |
                                       ljvLOCK_VIDEO_MEMORY_OP_UNMAP;
        iface.u.UnlockVideoMemory.node = (ljtUINT64)ljm_ts_obj->node_handle;
        iface.u.UnlockVideoMemory.type = ljvSURF_TYPE_UNKNOWN;
        iface.u.UnlockVideoMemory.pid = processID;
        ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

        memset(&iface, 0, sizeof(iface));
        iface.command = ljvHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY;
        iface.hardwareType = device->defaultHwType;
        iface.u.BottomHalfUnlockVideoMemory.node = (ljtUINT64)ljm_ts_obj->node_handle;
        iface.u.BottomHalfUnlockVideoMemory.type = ljvSURF_TYPE_UNKNOWN;
        iface.u.BottomHalfUnlockVideoMemory.pid = processID;
        ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));
    }

OnError:
    if (gem_obj) {
        drm_gem_object_unreference_unlocked(gem_obj);

        if (gem_ts_obj)
            drm_gem_object_unreference_unlocked(gem_ts_obj);
    }
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

static int ljm_ioctl_gem_ref_node(struct drm_device *drm,
                                  void *data, struct drm_file *file)
{
    struct drm_ljm_gem_ref_node *args = (struct drm_ljm_gem_ref_node *)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljeSTATUS status  = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;
    ljkKERNEL kernel = ljvNULL;
    ljtUINT32 processID;
    ljkVIDMEM_NODE nodeObj;
    ljeDATABASE_TYPE type;
    ljtUINT32 nodeHandle = 0, tsNodeHandle = 0;
    ljtBOOL referred = ljvFALSE;
    ljtBOOL isContiguous = ljvFALSE;
    int ret = 0;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    kernel = device->kernels[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    nodeObj = ljm_obj->node_object;

    ljmkONERROR(ljkOS_GetProcessID(&processID));
    if (ljm_obj->pid > 0)
        processID = ljm_obj->pid;
    ljmkONERROR(ljkVIDMEM_HANDLE_Allocate(kernel, processID, nodeObj, &nodeHandle));

    type = ljvDB_VIDEO_MEMORY |
           (nodeObj->type << ljdDB_VIDEO_MEMORY_TYPE_SHIFT) |
           (nodeObj->pool << ljdDB_VIDEO_MEMORY_POOL_SHIFT);

    ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, nodeObj, &isContiguous));
    if (isContiguous)
        type |= (ljvDB_CONTIGUOUS << ljdDB_VIDEO_MEMORY_DBTYPE_SHIFT);

    ljmkONERROR(ljkKERNEL_AddProcessDB(kernel, processID, type,
                                       ljmINT2PTR(nodeHandle), ljvNULL, 0));
    ljmkONERROR(ljkVIDMEM_NODE_Reference(kernel, nodeObj));
    referred = ljvTRUE;
    if (nodeObj->tsNode) {
        type = ljvDB_VIDEO_MEMORY |
               (nodeObj->tsNode->type << ljdDB_VIDEO_MEMORY_TYPE_SHIFT) |
               (nodeObj->tsNode->pool << ljdDB_VIDEO_MEMORY_POOL_SHIFT);

        ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, nodeObj->tsNode, &isContiguous));
        if (isContiguous)
            type |= (ljvDB_CONTIGUOUS << ljdDB_VIDEO_MEMORY_DBTYPE_SHIFT);

        ljmkONERROR(ljkVIDMEM_HANDLE_Allocate(kernel, processID, nodeObj->tsNode, &tsNodeHandle));
        ljmkONERROR(ljkKERNEL_AddProcessDB(kernel, processID, type,
                                           ljmINT2PTR(tsNodeHandle), ljvNULL, 0));
        ljmkONERROR(ljkVIDMEM_NODE_Reference(kernel, nodeObj->tsNode));
    }
    args->node = nodeHandle;
    args->ts_node = tsNodeHandle;

OnError:
    if (ljmIS_ERROR(status) && kernel) {
        ljtUINT32 processID;

        ljmkVERIFY_OK(ljkOS_GetProcessID(&processID));
        if (ljm_obj->pid > 0)
            processID = ljm_obj->pid;

        if (tsNodeHandle)
            ljkVIDMEM_HANDLE_Dereference(kernel, processID, tsNodeHandle);

        if (nodeHandle)
            ljkVIDMEM_HANDLE_Dereference(kernel, processID, nodeHandle);

        if (referred)
            ljmkONERROR(ljkVIDMEM_NODE_Dereference(kernel, nodeObj));

        args->node = 0;
        args->ts_node = 0;

        ret = -ENOTTY;
    }

    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_NO();
    return ret;
}


static int ljm_ioctl_gem_deref_node(struct drm_device *drm, void *data,
                                 struct drm_file *file)
{
    struct drm_ljm_gem_deref_node *args = (struct drm_ljm_gem_deref_node*)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device = ljvNULL;
    ljkKERNEL kernel = ljvNULL;
    ljtUINT32 processID = 0;
    ljkVIDMEM_NODE nodeObj = ljvNULL;
    ljeDATABASE_TYPE type;
    ljtUINT32 nodeHandle = 0, tsNodeHandle = 0;
    int ret = 0;
    ljtBOOL isContiguous = ljvFALSE;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];
    kernel = device->kernels[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
        ljmkONERROR(ljvSTATUS_NOT_FOUND);

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    nodeObj = ljm_obj->node_object;

    ljmkONERROR(ljkOS_GetProcessID(&processID));
    if (ljm_obj->pid > 0)
        processID = ljm_obj->pid;

    if (nodeObj->tsNode) {
        ljmkONERROR(ljkVIDMEM_NODE_Dereference(kernel, nodeObj->tsNode));

        tsNodeHandle = args->ts_node;
        if (tsNodeHandle == 0) {
            printk(KERN_ERR "%s(%d) error: tsNodeHandle is 0\n", __FUNCTION__, __LINE__);
            goto OnNodeDeref;
        }

        type = ljvDB_VIDEO_MEMORY |
               (nodeObj->tsNode->type << ljdDB_VIDEO_MEMORY_TYPE_SHIFT) |
               (nodeObj->tsNode->pool << ljdDB_VIDEO_MEMORY_POOL_SHIFT);

        ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, nodeObj->tsNode, &isContiguous));
        if (isContiguous)
            type |= (ljvDB_CONTIGUOUS << ljdDB_VIDEO_MEMORY_DBTYPE_SHIFT);

        ljmkONERROR(
            ljkKERNEL_RemoveProcessDB(kernel, processID, type, ljmINT2PTR(tsNodeHandle)));

        ljmkONERROR(ljkVIDMEM_HANDLE_Dereference(kernel, processID, tsNodeHandle));
    }

OnNodeDeref:
    ljmkONERROR(ljkVIDMEM_NODE_Dereference(kernel, nodeObj));

    nodeHandle = args->node;
    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "args->node=%d ljm_obj->node_handle=%d",
                   args->node, ljm_obj->node_handle);
    if (nodeHandle == 0) {
        printk(KERN_ERR "%s(%d) error: nodeHandle is 0\n", __FUNCTION__, __LINE__);
        ljmkONERROR(ljvSTATUS_INVALID_DATA);
    }

    type = ljvDB_VIDEO_MEMORY |
           (nodeObj->type << ljdDB_VIDEO_MEMORY_TYPE_SHIFT) |
           (nodeObj->pool << ljdDB_VIDEO_MEMORY_POOL_SHIFT);

    ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, nodeObj, &isContiguous));

    if (isContiguous)
        type |= (ljvDB_CONTIGUOUS << ljdDB_VIDEO_MEMORY_DBTYPE_SHIFT);

    ljmkONERROR(ljkKERNEL_RemoveProcessDB(kernel, processID, type, ljmINT2PTR(nodeHandle)));

    ljmkONERROR(ljkVIDMEM_HANDLE_Dereference(kernel, processID, nodeHandle));

OnError:
    if (ljmIS_ERROR(status) && kernel)
        ret = -ENOTTY;

    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_NO();
    return ret;
}

static int ljm_ioctl_gem_lock_gpu(struct drm_device *drm, void *data,
                           struct drm_file *file)
{
    struct drm_ljm_gem_lock_gpu *args = (struct drm_ljm_gem_lock_gpu*)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;
    ljkKERNEL kernel = ljvNULL;
    struct _ljsPLATFORM_PCIE *pcie_platform;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    device = gal_dev->devices[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
    {
        ljmkONERROR(ljvSTATUS_NOT_FOUND);
    }
    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    ljkOS_ZeroMemory(&iface, sizeof(iface));
    iface.command = ljvHAL_LOCK_VIDEO_MEMORY;
    //iface.hardwareType = device->defaultHwType;
    iface.hardwareType = ljvHARDWARE_2D;// ljvHARDWARE_3D2D;// ljvHARDWARE_2D;
    iface.u.LockVideoMemory.op = ljvLOCK_VIDEO_MEMORY_OP_LOCK;
    iface.u.LockVideoMemory.node = ljm_obj->node_handle;
    iface.u.LockVideoMemory.cacheable = args->cacheable;
    iface.u.LockVideoMemory.pid = ljm_obj->pid;
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

    args->physical = iface.u.LockVideoMemory.address;

    /* get GPU bus address for VPU, dma, etc */
    kernel = ljm_obj->node_object->kernel;
    if (ljm_obj->node_object->pool == ljvPOOL_LOCAL_EXTERNAL) {
        /* visible memory */
        pcie_platform = (struct _ljsPLATFORM_PCIE *)gal_dev->platform;
        args->gpuBusAddr = pcie_platform->pcie_info[kernel->device->platformIndex].gpu_vram_start + ljm_obj->node_object->node->VidMem.offset;
    } else if (ljm_obj->node_object->pool == ljvPOOL_LOCAL_EXCLUSIVE) {
        /* invisible memory */
        args->gpuBusAddr = gal_dev->exclusiveBase[kernel->device->platformIndex] + ljm_obj->node_object->node->VidMem.offset;
    } else {
        /* system memory */
        args->gpuBusAddr = iface.u.LockVideoMemory.physicalAddress;
    }

    OnError:
    if (gem_obj)
         drm_gem_object_unreference_unlocked(gem_obj);
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}


static int ljm_ioctl_gem_unlock_gpu(struct drm_device *drm, void *data,
                               struct drm_file *file)
{
    struct drm_ljm_gem_unlock_gpu *args = (struct drm_ljm_gem_unlock_gpu*)data;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    ljkDEVICE device;

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
       ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    device = gal_dev->devices[0];

    gem_obj = drm_gem_object_lookup(file, args->handle);
    if (!gem_obj)
       ljmkONERROR(ljvSTATUS_NOT_FOUND);
    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "ljm_obj: %d %p",
                   ljm_obj->node_handle, ljm_obj->node_object);

    memset(&iface, 0, sizeof(iface));
    iface.command = ljvHAL_BOTTOM_HALF_UNLOCK_VIDEO_MEMORY;
    //iface.hardwareType = device->defaultHwType;
    iface.hardwareType = ljvHARDWARE_2D;//ljvHARDWARE_3D2D;//ljvHARDWARE_2D;
    iface.u.BottomHalfUnlockVideoMemory.node = (ljtUINT64)ljm_obj->node_handle;
    iface.u.BottomHalfUnlockVideoMemory.type = ljm_obj->node_object->type;//ljvVIDMEM_TYPE_GENERIC;
    iface.u.BottomHalfUnlockVideoMemory.pid = ljm_obj->pid;
    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

//    printk(KERN_INFO "------ljm_ioctl_gem_unlock_gpu:ljm_obj->node_object->type=%d \n", ljm_obj->node_object->type);
    OnError:
    if (gem_obj)
        drm_gem_object_unreference_unlocked(gem_obj);

    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

#define MISC_PVT_DATA_REG     0x308574
#define MISC_PVT_SAMP_CTL_REG 0x308570
#define CORE_TMP0_CMD    0x000f0000
//#define CORE_TMP1_CMD    0x000f0001

#define CORE_VOL0_CMD    0x000f0100
//#define CORE_VOL1_CMD    0x000f0101

extern ljm_board_info *ljm_get_board_info(int card);
extern uint32_t ljm_gp200_get_freq(int card, int type);

typedef enum GPUStringInfoFlag
{
    /* flag for get gpu name */
    GPU_NAME = 0x10600,
    /* flag for get gpu vendor id */
    GPU_VENDOR_ID = 0x10601,
    /* flag for get gpu driver version */
    GPU_DRIVER_VERSION = 0x10602,
    /* flag for get gpu internal name */
    GPU_INTERNAL_NAME = 0x10603,
    /* flag for get gpu internal version */
    GPU_VENDOR_INTERNAL_VERSION = 0x10604,
    /* flag for get gpu shader type */
    GPU_SHADER_TYPE = 0x10605,
    /* flag for get gpu video storage type */
    GPU_VIDEO_STORAGE_TYPE = 0x10606,
    /* flag for get gpu chipset publish date */
    GPU_CHIPSET_PUBLISH_DATE = 0x10607,
    /* flag for get gpu driver publish date */
    GPU_DRIVER_PUBLISH_DATE = 0x10608,
    /* flag for get gpu manufacturer name */
    GPU_MANUFACTURER_NAME = 0x10609,
    /* flag for get gpu chipset id */
    GPU_CHIPSET_ID = 0x10610,
    /* flag for get gpu screen interface type */
    GPU_SCREEN_INTERFACE_TYPE = 0x10611,
    /* flag for get gpu supported OpenGL version */
    GPU_SUPPORT_OPENGL_VERSION = 0x10612,
    /* flag for get gpu supported Vulkan version */
    GPU_SUPPORT_VULKAN_VERSION = 0x10613,
    /* flag for get gpu supported OpenCL version */
    GPU_SUPPORT_OPENCL_VERSION = 0x10614,
    /* flag for get gpu bus interface type */
    GPU_BUS_INTERFACE_TYPE = 0x10615,
    /* flag for getting crtc hw id */
    GPU_CRTC_HW_ID = 0x10616
} GPUStringInfoFlag;

typedef enum GPUIntInfoFlag
{
    /* flag for get gpu generation process (Unit: nm) */
    GPU_GENERATION_PROCESS = 0x10701,
    /* flag for get gpu power dissipation (Unit: W) */
    GPU_POWER_DISSIPATION = 0x10702,
    /* flag for get gpu video storage (Unit: M) */
    GPU_VIDEO_STORAGE = 0x10703,
    /* flag for get gpu raster operations unit count */
    GPU_ROP_COUNT = 0x10704,
    /* flag for get gpu texture mapping unit count */
    GPU_TMU_COUNT = 0x10705,
    /* flag for get gpu shader count */
    GPU_SHADER_COUNT = 0x10706,
    /* flag for get gpu memory bandwidth (Unit: MBps) */
    GPU_MEMORY_BANDWIDTH = 0x10707,
    /* flag for get gpu pixel fillrate (Unit: MPixel/s) */
    GPU_PIXEL_FILLRATE = 0x10708,
    /* flag for get gpu texture fillrate (Unit: MTexel/s) */
    GPU_TEXTURE_FILLRATE = 0x10709,
    /* flag for get gpu max resolution width (Unit: Pixel) */
    GPU_MAX_RESOLUTION_WIDTH = 0x10710,
    /* flag for get gpu max resolution height (Unit: Pixel) */
    GPU_MAX_RESOLUTION_HEIGHT = 0x10711,
    /* flag for get gpu max screen count */
    GPU_MAX_SCREEN_COUNT = 0x10712,
    /* flag for get gpu effective bandwidth (Unit: M) */
    GPU_EFFECTIVE_BANDWIDTH = 0x10713,
    /* flag for get gpu normal accelerate frequency (Unit: MHz) */
    GPU_NORM_ACCELERATE_FREQUENCY = 0x10714,
    /* flag for get gpu oc accelerate frequency (Unit: MHz) */
    GPU_OC_ACCELERATE_FREQUENCY = 0x10715,
    /* flag for get gpu default frequency (Unit: MHz) */
    GPU_DEFAULT_FREQUENCY = 0x10716,
    /* flag for get gpu max frequency (Unit: MHz) */
    GPU_MAX_FREQUENCY = 0x10717,
    /* flag for get gpu memory bus width (Unit: bit) */
    GPU_MEMORY_BUS_WIDTH = 0x10718,
    /* flag for get gpu current frequency (Unit: MHz).dynamic information */
    GPU_CURRENT_FREQUENCY = 0x10719,
    /* flag for get gpu current memory frequency (Unit: MHz).dynamic information */
    GPU_CURRENT_MEMORY_FREQUENCY = 0x10720,
    /* flag for get gpu usage video storage percent.dynamic information */
    GPU_USAGE_VIDEO_STROAGE_PERCENT = 0x10721,
    /* flag for get gpu used video storage size (Unit: M).dynamic information */
    GPU_USAGE_VIDEO_STROAGE_SIZE = 0x10722
} GPUIntInfoFlag;

typedef enum GPUFloatInfoFlag
{
    /* flag for get gpu temperature (Unit: Celsius).dynamic information */
    GPU_TEMPERATURE = 0x10801,
    /* flag for get gpu vddc (Unit: Volts).dynamic information */
    GPU_VDDC = 0x10802
} GPUFloatInfoFlag;


static int ljm_ioctl_gpu_info(struct drm_device *drm, void *data, struct drm_file *file)
{
    struct drm_ljm_gpu_info *args = (struct drm_ljm_gpu_info*)data;

    struct ljm_device *ljm_dev;
    ljkGALDEVICE gal_dev = ljvNULL;

    ljeSTATUS status               = ljvSTATUS_OK;
    int card                       = 0;
    ljm_board_info *board_info;
    ljtUINT32 base;
    ljtUINT32 usable_total, usable_free;

    base = MISC_PVT_DATA_REG;

    ljmkHEADER_ARG("drm=%p data=%p file=%p pid=%d(%s)",
        drm, data, file, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)drm->dev_private;
    if (!gal_dev)
       ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    ljm_dev = (struct ljm_device *)gal_dev->ljm_dev;

    /* 计算显存使用大小 */
    usable_total = gal_dev->externalSize[card];
    usable_free = gal_dev->externalVidMem[card]->freeBytes;

    if ( gal_dev->exclusiveVidMem[card] ) {
        usable_total += gal_dev->exclusiveSize[card];
        usable_free += gal_dev->exclusiveVidMem[card]->freeBytes;
    }

    switch(args->infoFlag) {
        case GPU_NAME:
            strcpy(args->pdata, "GP201");
            break;
        case GPU_VENDOR_ID:
            strcpy(args->pdata, "LinJoWing Cop");
            break;
        case GPU_DRIVER_VERSION:
            strcpy(args->pdata, LJM_VERSION_STRING);   /* 获取当前显卡驱动版本号 */
            break;
        case GPU_INTERNAL_NAME:
            strcpy(args->pdata, "GP201");
            break;
        case GPU_VENDOR_INTERNAL_VERSION:
            strcpy(args->pdata, "unknown");    /* 获取芯片内部版本 */
            break;
        case GPU_SHADER_TYPE:
            strcpy(args->pdata, "union shader unit");
            break;
        case GPU_VIDEO_STORAGE_TYPE:        /* 获取显存类型 */
            board_info = ljm_get_board_info(card);
            if (board_info->ddr_imem_start == 0X80000)
                strcpy(args->pdata, "DDR4");
            else if (board_info->ddr_imem_start == 0X90000)
                strcpy(args->pdata, "LPDDR4");
            else if (board_info->ddr_imem_start == 0X90000)
                strcpy(args->pdata, "DDR3");
            break;
        case GPU_CHIPSET_PUBLISH_DATE:
            strcpy(args->pdata, "2022-06");
            break;
        case GPU_DRIVER_PUBLISH_DATE:
            strcpy(args->pdata, LJM_VERSION_DATE);
            break;
        case GPU_MANUFACTURER_NAME:
            strcpy(args->pdata, "LinJoWing");
            break;
        case GPU_CHIPSET_ID:
            strcpy(args->pdata, "VEN0709&DEV0201");
            break;
        case GPU_SCREEN_INTERFACE_TYPE:
            strcpy(args->pdata, "HDMI/DP/DVI/VGA");
            break;
        case GPU_SUPPORT_OPENGL_VERSION:
            strcpy(args->pdata, "OpenGL(4.0)");
            break;
        case GPU_SUPPORT_VULKAN_VERSION:
            strcpy(args->pdata, "Vulkan(3.0)");
            break;
        case GPU_SUPPORT_OPENCL_VERSION:
            strcpy(args->pdata, "OpenCL(3.0)");
            break;
        case GPU_BUS_INTERFACE_TYPE:
            strcpy(args->pdata, "PCI-Express");
            break;
        case GPU_GENERATION_PROCESS:            /***********开始获取Int类型************/
            *((ljtUINT32*)args->pdata) = 16;
            break;
        case GPU_POWER_DISSIPATION:        /* 功耗10-35 */
            *((ljtUINT32*)args->pdata) = 25;
            break;
        case GPU_VIDEO_STORAGE:
            *((ljtUINT32*)args->pdata) = ljm_dev->smem_len / 1024 / 1024;
            break;
        case GPU_ROP_COUNT:
            *((ljtUINT32*)args->pdata) = 32;
            break;
        case GPU_TMU_COUNT:
            *((ljtUINT32*)args->pdata) = 128;
            break;
        case GPU_SHADER_COUNT:
            *((ljtUINT32*)args->pdata) = 512;
            break;
        case GPU_MEMORY_BANDWIDTH:
            *((ljtUINT32*)args->pdata) = ljm_gp200_get_freq(card, freq_ddr) * 4;
            break;
        case GPU_PIXEL_FILLRATE:
            *((ljtUINT32*)args->pdata) = 32000;
            break;
        case GPU_TEXTURE_FILLRATE:
            *((ljtUINT32*)args->pdata) = 64000;
            break;
        case GPU_MAX_RESOLUTION_WIDTH:
            *((ljtUINT32*)args->pdata) = 3840;     /* 屏幕分辨率 */
            break;
        case GPU_MAX_RESOLUTION_HEIGHT:
            *((ljtUINT32*)args->pdata) = 2160;
            break;
        case GPU_MAX_SCREEN_COUNT:
            *((ljtUINT32*)args->pdata) = 6;
            break;
        case GPU_EFFECTIVE_BANDWIDTH:
            *((ljtUINT32*)args->pdata) = 0x21400;
            break;
        case GPU_NORM_ACCELERATE_FREQUENCY:
            *((ljtUINT32*)args->pdata) = 920;
            break;
        case GPU_OC_ACCELERATE_FREQUENCY:
            *((ljtUINT32*)args->pdata) = 1024;
            break;
        case GPU_DEFAULT_FREQUENCY:
            *((ljtUINT32*)args->pdata) = 920;
            break;
        case GPU_MAX_FREQUENCY:
            *((ljtUINT32*)args->pdata) = 1024;
            break;
        case GPU_MEMORY_BUS_WIDTH:
            *((ljtUINT32*)args->pdata) = 256;
            break;
        case GPU_CURRENT_FREQUENCY:
            *((ljtUINT32*)args->pdata) = ljm_gp200_get_freq(card, freq_3d);
            break;
        case GPU_CURRENT_MEMORY_FREQUENCY:
            *((ljtUINT32*)args->pdata) = ljm_gp200_get_freq(card, freq_axi);
            break;
        case GPU_USAGE_VIDEO_STROAGE_PERCENT:
            *((ljtUINT32*)args->pdata) = ((usable_total - usable_free) / 1024 / 1024 * 100) / (usable_total / 1024 / 1024);
            break;
        case GPU_USAGE_VIDEO_STROAGE_SIZE:
            *((ljtUINT32*)args->pdata) = (usable_total - usable_free) / 1024 / 1024;
            break;
        case GPU_TEMPERATURE:
            REG_W32(0, base - MISC_PVT_DATA_REG + MISC_PVT_SAMP_CTL_REG, CORE_TMP0_CMD);
            *((ljtUINT32*)args->pdata) = REG_R32(0, base);
            break;
        case GPU_VDDC:
            REG_W32(0, base - MISC_PVT_DATA_REG + MISC_PVT_SAMP_CTL_REG, CORE_VOL0_CMD);
            *((ljtUINT32*)args->pdata) = REG_R32(0, base);
            break;
        case GPU_CRTC_HW_ID:
            //printk("viv ioctl gpu info：%d\n", *(ljtUINT32*)args->pdata);
            {
                struct drm_crtc *crtc = NULL;
                struct ljm_crtc *ljm_crtc = NULL;

                crtc = drm_crtc_find(drm, file, *(ljtUINT32*)args->pdata);
                ljm_crtc = to_ljm_crtc(crtc);
                *((ljtUINT32*)args->pdata) = ljm_crtc->id;
            }
            break;
        default:
            return -EINTR;
    }

    ljmkFOOTER_NO();
    return status;

OnError:

    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENOTTY : 0;

}

static const struct drm_ioctl_desc ljm_ioctls[] = {
    DRM_IOCTL_DEF_DRV(VIV_GEM_CREATE,        ljm_ioctl_gem_create,     DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_LOCK,          ljm_ioctl_gem_lock,       DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_UNLOCK,        ljm_ioctl_gem_unlock,     DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_CACHE,         ljm_ioctl_gem_cache,      DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_QUERY,         ljm_ioctl_gem_query,      DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_TIMESTAMP,     ljm_ioctl_gem_timestamp,  DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_SET_TILING,    ljm_ioctl_gem_set_tiling, DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_GET_TILING,    ljm_ioctl_gem_get_tiling, DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_ATTACH_AUX,    ljm_ioctl_gem_attach_aux, DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_REF_NODE,      ljm_ioctl_gem_ref_node,   DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_DEREF_NODE,    ljm_ioctl_gem_deref_node, DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_LOCK_GPU,      ljm_ioctl_gem_lock_gpu,   DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GEM_UNLOCK_GPU,    ljm_ioctl_gem_unlock_gpu, DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_GPU_INFO,          ljm_ioctl_gpu_info,       DRM_AUTH | DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_PLANE_REQUEST,     ljm_ioctl_plane_request,  DRM_AUTH|DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_PLANE_FREE,        ljm_ioctl_plane_free,     DRM_AUTH|DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_PROPPLANE_SET,     ljm_ioctl_plane_property_set, DRM_AUTH|DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_PROPPLANE_GET,     ljm_ioctl_plane_property_get, DRM_AUTH|DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_PROPZVIDEO_SET,    ljm_ioctl_zv_property_set, DRM_AUTH|DRM_RENDER_ALLOW),
    DRM_IOCTL_DEF_DRV(VIV_PROPZVIDEO_GET,    ljm_ioctl_zv_property_get, DRM_AUTH|DRM_RENDER_ALLOW),
};

int ljm_drm_open(struct drm_device *drm, struct drm_file *file)
{
    ljtINT i, dev_index;
    ljtUINT32 pid = _GetProcessID();
    ljkGALDEVICE gal_dev = (ljkGALDEVICE)drm->dev_private;
    ljkDEVICE device;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("drm=%p file=%p pid=%d name=%s", drm, file, pid, _GetProcessComm());

    for (dev_index = 0; dev_index < gal_dev->args.devCount; dev_index++) {
        device = gal_dev->devices[dev_index];

        for (i = 0; i < ljdMAX_GPU_COUNT; ++i) {
            if (device->kernels[i])
                ljmkONERROR(ljkKERNEL_AttachProcessEx(device->kernels[i], ljvTRUE, pid));
        }
    }

    file->driver_priv = ljmINT2PTR(pid);

OnError:
    ljmkFOOTER_NO();
    return ljmIS_ERROR(status) ? -ENODEV : 0;
}

void ljm_drm_postclose(struct drm_device *drm, struct drm_file *file)
{
    ljtINT i, dev_index;
    ljtUINT32 pid = ljmPTR2SIZE(file->driver_priv);
    ljkGALDEVICE gal_dev = (ljkGALDEVICE)drm->dev_private;
    ljkDEVICE device;

    ljmkHEADER_ARG("drm=%p file=%p pid=%d name=%s", drm, file, pid, _GetProcessComm());

    for (dev_index = 0; dev_index < gal_dev->args.devCount; dev_index++) {
        device = gal_dev->devices[dev_index];

        for (i = 0; i < ljdMAX_GPU_COUNT; ++i) {
            if (device->kernels[i])
                ljmkVERIFY_OK(ljkKERNEL_AttachProcessEx(device->kernels[i], ljvFALSE, pid));
        }
    }

    ljmkFOOTER_NO();
}

int ljm_dumb_create(
    struct drm_file *File,
    struct drm_device *Drm,
    struct drm_mode_create_dumb *Args
    )
{
    int ret = 0;
    struct drm_gem_object *gem_obj = ljvNULL;
    struct ljm_gem_object *ljm_obj = ljvNULL;

    ljsHAL_INTERFACE iface;
    ljkGALDEVICE gal_dev;
    ljkDEVICE device;
    ljkKERNEL kernel = ljvNULL;
    ljtUINT32 processID;
    ljkVIDMEM_NODE nodeObject;
    ljtUINT32 flags = ljvALLOC_FLAG_DMABUF_EXPORTABLE;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("File=0x%p Drm=0x%p Args=0x%p", File, Drm, Args);

    gal_dev = (ljkGALDEVICE)Drm->dev_private;
    if (!gal_dev)
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    device = gal_dev->devices[0];

    if (Args->flags & DRM_VIV_GEM_CONTIGUOUS)
    {
        flags |= ljvALLOC_FLAG_CONTIGUOUS;
    }
    if (Args->flags & DRM_VIV_GEM_CACHED)
    {
        flags |= ljvALLOC_FLAG_CACHEABLE;
    }
    if (Args->flags & DRM_VIV_GEM_SECURE)
    {
        flags |= ljvALLOC_FLAG_SECURITY;
    }

    //Args->pitch = Args->width * ((Args->bpp + 7) / 8);
    Args->pitch = DIV_ROUND_UP(Args->width * Args->bpp, 8);
    Args->size = PAGE_ALIGN(Args->pitch * Args->height);

    ljkOS_ZeroMemory(&iface, sizeof(iface));
    iface.command = ljvHAL_ALLOCATE_LINEAR_VIDEO_MEMORY;
    iface.hardwareType = device->defaultHwType;
    iface.u.AllocateLinearVideoMemory.bytes = Args->size;
    iface.u.AllocateLinearVideoMemory.alignment = PAGE_SIZE;	/*256*/	/* align page size to support remap_pfn_range */
    /*
     * In this version, gem in VRAM is only for framebuffer. Framebuffer should be in different physical bank from 3D VRAM to improve VRAM performance.
     */
    iface.u.AllocateLinearVideoMemory.type = ljvVIDMEM_TYPE_GENERIC;
    iface.u.AllocateLinearVideoMemory.flag = flags;
    iface.u.AllocateLinearVideoMemory.pool = ljvPOOL_LOCAL_EXTERNAL;

    ljmkONERROR(ljkDEVICE_Dispatch(device, &iface));

    kernel = device->kernels[0];
    ljmkONERROR(ljkOS_GetProcessID(&processID));
    ljmkONERROR(ljkVIDMEM_HANDLE_Lookup(kernel, processID, iface.u.AllocateLinearVideoMemory.node, &nodeObject));

    /* ioctl output */
    gem_obj = kzalloc(sizeof(struct ljm_gem_object), GFP_KERNEL);
    if (!gem_obj) {
        ljmkFOOTER_ARG("err: %d", -ENOMEM);
        return -ENOMEM;
    }

    drm_gem_private_object_init(Drm, gem_obj, Args->size);    /* no shmfs backing */

    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);
    ljm_obj->node_handle = iface.u.AllocateLinearVideoMemory.node;
    ljm_obj->node_object = nodeObject;
    ljm_obj->cacheable = flags & ljvALLOC_FLAG_CACHEABLE;
    ljm_obj->pid = processID;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    ljm_obj->base.funcs = &ljm_gem_object_funcs;
#endif
    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                       "ljm_obj: %d %p",
                       ljm_obj->node_handle, ljm_obj->node_object);

    ret = drm_gem_handle_create(File, gem_obj, &Args->handle);  /* call ljm_gem_object_open */
    if (ret) {
        printk(KERN_ERR "ljm_dumb_create: drm_gem_handle_create error (%d)\n", ret);
        drm_gem_object_release(gem_obj);
        goto OnError;
    }

    /* drop reference from allocate - handle holds it now */
    drm_gem_object_unreference_unlocked(gem_obj);

    ljmkFOOTER_ARG("pid=%d ljm_obj=%p", processID, ljm_obj);
    return ljmIS_ERROR(status) ? -ENOTTY : ret;

OnError:
    if (gem_obj) kfree(gem_obj);

    ljmkFOOTER_ARG("%d", ljmIS_ERROR(status) ? -ENOTTY : 0);
    return ljmIS_ERROR(status) ? -ENOTTY : 0;
}

int ljm_dumb_mmap_offset(
    struct drm_file *File,
    struct drm_device *Drm,
    uint32_t Handle,
    uint64_t *Offset
    )
{
    struct drm_gem_object *obj;
    int ret;

    ljmkHEADER_ARG("File=0x%p Drm=0x%p Handle=0x%d Offset=0x%p", File, Drm, Handle, Offset);

    obj = drm_gem_object_lookup(File, Handle);
    if (!obj)
    {
        printk(KERN_ERR "failed to lookup GEM object\n");
        ljmkFOOTER_ARG("%d", -ENOENT);
        return -ENOENT;
    }

    /* Don't allow imported objects to be mapped */
    if (obj->import_attach) {
        ret = -EINVAL;
        goto out;
    }

    ret = drm_gem_create_mmap_offset(obj);
    if (ret) {
        printk(KERN_ERR "ljm_dumb_mmap_offset: drm_gem_create_mmap_offset error (%d)\n", ret);
        goto out;
    }

    *Offset = drm_vma_node_offset_addr(&obj->vma_node);

out:
    drm_gem_object_unreference_unlocked(obj);

    ljmkFOOTER_NO();
    return ret;
}

static int ljm_gem_mmap_obj(struct ljm_gem_object *ljm_obj,
                struct vm_area_struct *vma)
{
    int ret;
    unsigned long pfn;

    ljmkHEADER_ARG("ljm_gem_mmap_obj: ljm_obj=0x%p vma=0x%p vm_flags=%lx vm_pgoff=%lx", ljm_obj, vma, vma->vm_flags, vma->vm_pgoff);

    /*
     * Set the ljdVM_FLAGS flag to make this mapping non-cached, and set the
     * vm_pgoff (used as a fake buffer offset by DRM) to 0 as we want to map
     * the whole buffer.
     */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    vma->vm_flags |= ljdVM_FLAGS;
#else
    vm_flags_set(vma, ljdVM_FLAGS);
#endif
    vma->vm_pgoff = 0;

    if (is_ft1500a() || is_ft2000ahk())
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#ifdef CONFIG_MIPS
    else
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#else
    else
        vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
#endif

//    printk("----physicalBase=0x%llx offset=0x%zx size=%zx vma->vm_start=%lx vma->vm_end=%lx----\n",
//        BO_GET_VIDMEM_START(ljm_obj), BO_GET_VIDMEM_OFFSET(ljm_obj), BO_GET_SIZE(ljm_obj),
//        vma->vm_start, vma->vm_end
//        );

    pfn = (BO_GET_VIDMEM_START(ljm_obj) + BO_GET_VIDMEM_OFFSET(ljm_obj)) >> PAGE_SHIFT;
    ret = remap_pfn_range(vma, vma->vm_start,
                          pfn,
                          vma->vm_end - vma->vm_start,
                          vma->vm_page_prot);
    if (ret < 0) {
        printk(KERN_ERR "remap_pfn_range error (0x%lx @ 0x%lx)\n", vma->vm_end - vma->vm_start, pfn << PAGE_SHIFT);

        drm_gem_vm_close(vma);
    }

    ljmkFOOTER_ARG("%d", ret);
    return ret;
}

/**
 * ljm_gem_mmap - memory-map a CSM GEM object
 * @filp: file object
 * @vma: VMA for the area to be mapped
 *
 * This function implements an augmented version of the GEM DRM file mmap
 * operation for CSM objects: In addition to the usual GEM VMA setup it
 * immediately faults in the entire object instead of using on-demaind
 * faulting. Drivers which employ the CSM helpers should use this function
 * as their ->mmap() handler in the DRM device file's file_operations
 * structure.
 *
 * Returns:
 * 0 on success or a negative error code on failure.
 */
static int ljm_gem_mmap(struct file *filp, struct vm_area_struct *vma)
{
    struct ljm_gem_object *ljm_obj;
    struct drm_gem_object *gem_obj;
    int ret;

    ljmkHEADER_ARG("filp=0x%p vma=0x%p", filp, vma);

    ret = drm_gem_mmap(filp, vma);
    if (ret) {
        printk(KERN_ERR "ljm_gem_mmap: drm_gem_mmap is error (%d)\n", ret);  // err here
        ljmkFOOTER_ARG("%d", ret);
        return ret;
    }

    gem_obj = vma->vm_private_data;
    ljm_obj = container_of(gem_obj, struct ljm_gem_object, base);

    ret = ljm_gem_mmap_obj(ljm_obj, vma);

    ljmkFOOTER_ARG("%d", ret);
    return ret;
}

static const struct file_operations ljm_drm_fops = {
    .owner              = THIS_MODULE,
    .open               = drm_open,
    .release            = drm_release,
    .unlocked_ioctl     = drm_ioctl,
#ifdef CONFIG_COMPAT
    .compat_ioctl       = drm_compat_ioctl,
#endif
    .poll               = drm_poll,
    .read               = drm_read,
    .llseek             = no_llseek,
    .mmap               = ljm_gem_mmap,
};
/*static void ljm_gem_object_close(struct drm_gem_object *obj,
			     struct drm_file *file_priv)
{
	struct ljm_gem_object *ljm_obj = container_of(obj, struct ljm_gem_object, base);

	printk("----ljm_gem_object_close--%p--refcount %d-----------\n", ljm_obj->node_object, kref_read(&obj->refcount));
}*/

static struct drm_driver ljm_drm_driver = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)
    .driver_features    = DRIVER_GEM | DRIVER_RENDER | DRIVER_MODESET,
#else
    .driver_features    = DRIVER_GEM | DRIVER_PRIME | DRIVER_RENDER | DRIVER_MODESET | DRIVER_HAVE_IRQ | DRIVER_IRQ_SHARED,
#endif
    .load               = ljm_driver_load_kms,
    .unload             = ljm_driver_unload_kms,
    .open               = ljm_drm_open,
    .postclose          = ljm_drm_postclose,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4,12,0)
    .set_busid          = drm_pci_set_busid,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6,2,0)
    .irq_preinstall     = ljm_driver_irq_preinstall_kms,
    .irq_postinstall    = ljm_driver_irq_postinstall_kms,
    .irq_uninstall      = ljm_driver_irq_uninstall_kms,
    .irq_handler        = ljm_driver_irq_handler_kms,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,11,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,7,0)
    .gem_free_object_unlocked = ljm_gem_free_object,
    .prime_handle_to_fd = drm_gem_prime_handle_to_fd,
#else
    .gem_free_object    = ljm_gem_free_object,
    .prime_handle_to_fd = ljm_gem_prime_handle_to_fd,
#endif
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(6,6,0)
    .gem_prime_mmap = drm_gem_prime_mmap,
#endif
    .prime_handle_to_fd = drm_gem_prime_handle_to_fd,
#endif
    .prime_fd_to_handle = drm_gem_prime_fd_to_handle,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,11,0)
    .gem_prime_export   = ljm_gem_prime_export,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6,2,0) || LINUX_VERSION_CODE >= KERNEL_VERSION(6,6,0)
    .gem_prime_import   = ljm_gem_prime_import,
#endif
 //   .gem_close_object =  ljm_gem_object_close,
    /* added for modesettings */
    .dumb_create        = ljm_dumb_create,
    .dumb_map_offset    = ljm_dumb_mmap_offset,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6,2,0)
    .dumb_destroy       = drm_gem_dumb_destroy,
    .gem_vm_ops         = &ljm_gem_vm_ops,
#endif

#if USE_KMS_VBLANK
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
    .get_vblank_counter = ljm_get_vblank_counter_kms,
    .enable_vblank = ljm_enable_vblank_kms,
    .disable_vblank = ljm_disable_vblank_kms,
    //.get_vblank_timestamp = ljm_get_vblank_timestamp_kms,
#endif
#endif

    .ioctls             = ljm_ioctls,
    .num_ioctls         = DRM_VIV_NUM_IOCTLS,
    .fops               = &ljm_drm_fops,
    .name               = "ljm",
    .desc               = "ljm DRM",
    .date               = "20170808",
    .major              = 1,
    .minor              = 0,
};

int
ljm_drm_probe(struct pci_dev *pdev, unsigned long flags)
{
    int ret = 0;
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE gal_dev = ljvNULL;
    struct drm_device *drm = ljvNULL;

    ljmkHEADER_ARG("pdev=%p flags=%lx pid=%d name=%s", pdev, flags, _GetProcessID(), _GetProcessComm());

    gal_dev = (ljkGALDEVICE)dev_get_drvdata(&pdev->dev);
    if (!gal_dev) {
        ret = -ENODEV;
        ljmkONERROR(ljvSTATUS_INVALID_OBJECT);
    }

    drm = drm_dev_alloc(&ljm_drm_driver, &pdev->dev);
    if (IS_ERR(drm)) {
        ret = PTR_ERR(drm);
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
    }
    drm->dev_private = (void *)gal_dev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    drm->pdev = pdev;
#endif

    ret = drm_dev_register(drm, flags);
    if (ret)
        ljmkONERROR(ljvSTATUS_GENERIC_IO);

    gal_dev->drm = (void *)drm;

OnError:
    if (ljmIS_ERROR(status)) {
        if (drm)
            drm_dev_unref(drm);
        pr_err("ljmcore: Failed to setup drm device.\n");
    }
    //printk("----leave ljm_drm_probe(ret=%d)----\n", ret);

    ljmkFOOTER_NO();
    return ret;
}

int
ljm_drm_remove(ljkGALDEVICE gal_dev)
{
#if ljdDEBUG
    ljtUINT32 pid = _GetProcessID();
#endif

    ljmkHEADER_ARG("gal_dev=%p (%px) drm=%p (%px) pid=%d name=%s", gal_dev, gal_dev, gal_dev->drm, gal_dev->drm, pid, _GetProcessComm());

    if (gal_dev) {
        struct drm_device *drm = (struct drm_device*)gal_dev->drm;

        drm_dev_unregister(drm);
        drm_dev_unref(drm);
    }

    ljmkFOOTER_NO();
    return 0;
}

#endif /* !ljdENABLE_TTM */
#endif /* ljdENABLE_DRM */
