/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_hardware_h_
#define __lj_hal_kernel_hardware_h_

#include "lj_hal_kernel_hardware_func.h"

#if ljdENABLE_VG
#include "lj_hal_kernel_hardware_vg.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define EVENT_ID_INVALIDATE_PIPE 29

#define ljmSEMAPHORESTALL(buffer)                                                                       \
    do {                                                                                                \
        /* Arm the PE-FE Semaphore. */                                                                  \
        *(buffer)++ = ljmSETFIELDVALUE(0, AQ_COMMAND_LOAD_STATE_COMMAND, OPCODE, LOAD_STATE) |         \
                           ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, COUNT, 1) |                  \
                           ljmSETFIELD(0, AQ_COMMAND_LOAD_STATE_COMMAND, ADDRESS, 0x0E02); \
                                                                                                        \
        *(buffer)++ = ljmSETFIELDVALUE(0, AQ_SEMAPHORE, SOURCE, FRONT_END) |                       \
                      ljmSETFIELDVALUE(0, AQ_SEMAPHORE, DESTINATION, PIXEL_ENGINE);                     \
                                                                                                        \
        /* STALL FE until PE is done flushing. */                                                       \
        *(buffer)++ = ljmSETFIELDVALUE(0, STALL_COMMAND, OPCODE, STALL);                                \
                                                                                                        \
        *(buffer)++ = ljmSETFIELDVALUE(0, STALL_STALL, SOURCE, FRONT_END) |                        \
                      ljmSETFIELDVALUE(0, STALL_STALL, DESTINATION, PIXEL_ENGINE);                      \
    } while (0)

typedef enum {
    ljvHARDWARE_FUNCTION_MMU,
    ljvHARDWARE_FUNCTION_FLUSH,
    ljvHARDWARE_FUNCTION_NUM,
} ljeHARDWARE_FUNCTION;

typedef struct _ljkASYNC_FE *ljkASYNC_FE;
typedef struct _ljkWLFE     *ljkWLFE;
typedef struct _ljkMCFE     *ljkMCFE;

typedef struct _ljsSTATETIMER {
    ljtUINT64                   start;
    ljtUINT64                   recent;

    /* Elapse of each power state. */
    ljtUINT64                   elapse[4];
} ljsSTATETIMER;

typedef struct _ljsHARDWARE_SIGNATURE {
    /* Chip model. */
    ljeCHIPMODEL                chipModel;

    /* Revision value.*/
    ljtUINT32                   chipRevision;

    /* Supported feature fields. */
    ljtUINT32                   chipFeatures;

    /* Supported minor feature fields. */
    ljtUINT32                   chipMinorFeatures;

    /* Supported minor feature 1 fields. */
    ljtUINT32                   chipMinorFeatures1;

    /* Supported minor feature 2 fields. */
    ljtUINT32                   chipMinorFeatures2;
} ljsHARDWARE_SIGNATURE;

typedef struct _ljsMMU_TABLE_ARRAY_ENTRY {
    ljtUINT32                   low;
    ljtUINT32                   high;
} ljsMMU_TABLE_ARRAY_ENTRY;

typedef struct _ljsHARDWARE_PAGETABLE_ARRAY {
    /* Number of entries in page table array. */
    ljtUINT                     num;

    /* Video memory node. */
    ljkVIDMEM_NODE              videoMem;

    /* Size in bytes of array. */
    ljtSIZE_T                   size;

    /* Physical address of array. */
    ljtPHYS_ADDR_T              address;

    /* Logical address of array. */
    ljtPOINTER                  logical;
} ljsHARDWARE_PAGETABLE_ARRAY;

/* ljkHARDWARE object. */
struct _ljkHARDWARE {
    /* Object. */
    ljsOBJECT                   object;

    /* Pointer to ljtKERNEL object. */
    ljkKERNEL                   kernel;

    /* Pointer to ljtOS object. */
    ljkOS                       os;

    /* Core */
    ljeCORE                     core;

    /* Type */
    ljeHARDWARE_TYPE            type;

    /* Chip characteristics. */
    ljsHAL_QUERY_CHIP_IDENTITY  identity;
    ljsHAL_QUERY_CHIP_OPTIONS   options;
    ljtUINT32                   powerBaseAddress;
    ljtBOOL                     extraEventStates;

    /* Big endian */
    ljtBOOL                     bigEndian;

    /* Base address. */
    ljtUINT32                   baseAddress;

    /* FE modules. */
    ljkWLFE                     wlFE;
    ljkASYNC_FE                 asyncFE;
    ljkMCFE                     mcFE;

    /* Chip status */
    ljtPOINTER                  powerMutex;
    ljeCHIPPOWERSTATE           chipPowerState;
    ljtBOOL                     clockState;
    ljtBOOL                     powerState;
    ljtPOINTER                  globalSemaphore;
    ljtBOOL                     isLastPowerGlobal;

    /* Wait Link FE only. */
    ljtADDRESS                  lastWaitLink;
    ljtADDRESS                  lastEnd;

    ljtUINT32                   mmuVersion;

#if ljdPOWEROFF_TIMEOUT
    ljeCHIPPOWERSTATE           nextPowerState;
    ljtPOINTER                  powerStateTimer;
#endif

#if ljdENABLE_FSCALE_VAL_ADJUST
    ljtUINT32                   powerOnFscaleVal;
    ljtUINT32                   powerOnShaderFscaleVal;
#endif
    ljtPOINTER                  pageTableDirty[ljvENGINE_GPU_ENGINE_COUNT];

#if ljdLINK_QUEUE_SIZE
    struct _ljkQUEUE            linkQueue;
#endif
    ljtBOOL                     stallFEPrefetch;

    ljtUINT32                   minFscaleValue;
    ljtUINT                     waitCount;

    ljtUINT64                   mcClk;
    ljtUINT64                   shClk;

    ljtPOINTER                  pendingEvent;

    ljsFUNCTION_EXECUTION_PTR   functions;

    ljsSTATETIMER               powerStateCounter;
    ljtUINT32                   executeCount;
    ljtADDRESS                  lastExecuteAddress;

    /* Head for hardware list in ljkMMU. */
    ljsLISTHEAD                 mmuHead;

    /* Internal SRAMs info. */
    ljkVIDMEM                   sRAMVidMem[ljvSRAM_INTER_COUNT];
    ljtPHYS_ADDR                sRAMPhysical[ljvSRAM_INTER_COUNT];

    ljtPOINTER                  featureDatabase;
    ljtBOOL                     hasL2Cache;

    /* MCFE channel bindings, temporary. */
    ljeMCFE_CHANNEL_TYPE        mcfeChannels[64];
    ljtUINT32                   mcfeChannelCount;

    ljsHARDWARE_SIGNATURE       signature;

    ljtUINT32                   maxOutstandingReads;

    ljsHARDWARE_PAGETABLE_ARRAY pagetableArray;

    ljtUINT64                   contextID;

    ljtBOOL                     hasQchannel;

    ljtUINT32                   powerOffTimeout;

    ljtUINT32                   devID;

    ljtBOOL                     largeVA;
};

ljeSTATUS
ljkHARDWARE_GetBaseAddress(ljkHARDWARE Hardware, ljtUINT32_PTR BaseAddress);

ljeSTATUS
ljkHARDWARE_NeedBaseAddress(ljkHARDWARE Hardware, ljtUINT32 State, ljtBOOL_PTR NeedBase);

ljeSTATUS
ljkHARDWARE_GetFrameInfo(ljkHARDWARE Hardware, ljsHAL_FRAME_INFO *FrameInfo);

ljeSTATUS
ljkHARDWARE_DumpGpuProfile(ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_HandleFault(ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_ExecuteFunctions(ljsFUNCTION_EXECUTION_PTR Execution);

ljeSTATUS
ljkHARDWARE_DummyDraw(ljkHARDWARE Hardware,
                      ljtPOINTER Logical,
                      ljtADDRESS Address,
                      ljeDUMMY_DRAW_TYPE DummyDrawType,
                      ljtUINT32 *Bytes);

ljeSTATUS
ljkHARDWARE_EnterQueryClock(ljkHARDWARE Hardware, ljtUINT64 *McStart, ljtUINT64 *ShStart);

ljeSTATUS
ljkHARDWARE_ExitQueryClock(ljkHARDWARE Hardware,
                           ljtUINT64 McStart,
                           ljtUINT64 ShStart,
                           ljtUINT64 *McClk,
                           ljtUINT64 *ShClk);

ljeSTATUS
ljkHARDWARE_QueryFrequency(ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_QueryFScale(ljkHARDWARE Hardware, ljtUINT8 *Scale);

ljeSTATUS
ljkHARDWARE_SetClock(ljkHARDWARE Hardware, ljtUINT32 MCScale, ljtUINT32 SHScale);

ljeSTATUS
ljkHARDWARE_PowerControlClusters(ljkHARDWARE Hardware,
                                 ljtUINT32 PowerControlValue,
                                 ljtBOOL PowerState);

ljeSTATUS
ljkHARDWARE_QueryCycleCount(ljkHARDWARE Hardware,
                            ljtUINT32 *hi_total_cycle_count,
                            ljtUINT32 *hi_total_idle_cycle_count);

ljeSTATUS
ljkHARDWARE_CleanCycleCount(ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_QueryCoreLoad(ljkHARDWARE Hardware,
                          ljtUINT32 Mdelay,
                          ljtUINT32 *Load);

ljeSTATUS
ljkHARDWARE_CancelJob(ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_FlushCache(ljkHARDWARE Hardware, ljkCOMMAND Command);

#if ljdENABLE_AHBXTTD
ljeSTATUS
ljkHARDWARE_ResetAHBXTTD(ljkHARDWARE Hardware, ljkKERNEL Kernel);
#endif

ljeSTATUS
ljkHARDWARE_FillMMUDescriptor(ljkHARDWARE Hardware,
                              ljtUINT32 Index,
                              ljtPHYS_ADDR_T MtlbPhysical);

ljeSTATUS
ljkHARDWARE_SwitchPageTableId(ljkHARDWARE Hardware, ljkMMU Mmu,
                              ljkCOMMAND Command);

ljeSTATUS
ljkHARDWARE_ResumeWLFE(ljkHARDWARE Hardware);

#define ljmkWRITE_MEMORY(logical, data)                                       \
    do {                                                                      \
        ljmkVERIFY_OK(ljkOS_WriteMemory(os, logical, data));                  \
        logical++;                                                            \
    } while (0)

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_kernel_hardware_h_ */

