/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_hardware_fe_h_
#define __lj_hal_kernel_hardware_fe_h_

#include "lj_hal.h"

/******************************************************************************/
/* Wait-Link FE commands. */

/* Construct Wait-Link FE. */
ljeSTATUS
ljkWLFE_Construct(ljkHARDWARE Hardware, ljkWLFE *FE);

void
ljkWLFE_Destroy(ljkHARDWARE Hardware, ljkWLFE FE);

/* Initialize Wait-Link FE, when hardware reset. */
ljeSTATUS
ljkWLFE_Initialize(ljkHARDWARE Hardware, ljkWLFE FE);

/* Add a WAIT/LINK pair in the command queue. */
ljeSTATUS
ljkWLFE_WaitLink(ljkHARDWARE Hardware,
                 ljtPOINTER Logical,
                 ljtADDRESS Address,
                 ljtUINT32 Offset,
                 ljtUINT32 *Bytes,
                 ljtUINT32 *WaitOffset,
                 ljtUINT32 *WaitBytes);

/* Add a LINK command in the command queue. */
ljeSTATUS
ljkWLFE_Link(ljkHARDWARE Hardware,
             ljtPOINTER Logical,
             ljtADDRESS FetchAddress,
             ljtUINT32 FetchSize,
             ljtUINT32 *Bytes,
             ljtUINT32 *Low,
             ljtUINT32 *High);

/* Add an END command in the command queue. */
ljeSTATUS
ljkWLFE_EndEx(ljkHARDWARE Hardware,
              ljtPOINTER Logical,
              ljtADDRESS Address,
              ljtBOOL User,
              ljtUINT32 *Bytes);

ljeSTATUS
ljkWLFE_End(ljkHARDWARE Hardware,
            ljtPOINTER Logical,
            ljtADDRESS Address,
            ljtUINT32 *Bytes);

/* Add a NOP command in the command queue. */
ljeSTATUS
ljkWLFE_Nop(ljkHARDWARE Hardware,
            ljtPOINTER Logical,
            ljtSIZE_T *Bytes);

/* Add an EVENT command in the command queue. */
ljeSTATUS
ljkWLFE_Event(ljkHARDWARE Hardware,
              ljtPOINTER Logical,
              ljtUINT8 Event,
              ljeKERNEL_WHERE FromWhere,
              ljtUINT32 *Bytes);

ljeSTATUS
ljkWLFE_ChipEnable(ljkHARDWARE Hardware,
                   ljtPOINTER Logical,
                   ljeCORE_3D_MASK ChipEnable,
                   ljtSIZE_T *Bytes);

/* Kickstart the command processor. */
ljeSTATUS
ljkWLFE_Execute(ljkHARDWARE Hardware, ljtADDRESS Address, ljtUINT32 Bytes);

/******************************************************************************/
/* ASync FE commands. */

ljeSTATUS
ljkASYNC_FE_Construct(ljkHARDWARE Hardware, ljkASYNC_FE *FE);

void
ljkASYNC_FE_Destroy(ljkHARDWARE Hardware, ljkASYNC_FE FE);

/* Initialize Async FE, when hardware reset. */
ljeSTATUS
ljkASYNC_FE_Initialize(ljkHARDWARE Hardware, ljkASYNC_FE FE);

/* Add a NOP command in the command queue. */
ljeSTATUS
ljkASYNC_FE_Nop(ljkHARDWARE Hardware, ljtPOINTER Logical, ljtSIZE_T *Bytes);

/* Add an EVENT command in the command queue. */
ljeSTATUS
ljkASYNC_FE_Event(ljkHARDWARE Hardware,
                  ljtPOINTER Logical,
                  ljtUINT8 Event,
                  ljeKERNEL_WHERE FromWhere,
                  ljtUINT32 *Bytes);

/* Kickstart the command processor. */
ljeSTATUS
ljkASYNC_FE_Execute(ljkHARDWARE Hardware, ljtADDRESS Address, ljtUINT32 Bytes);

ljeSTATUS
ljkASYNC_FE_ReserveSlot(ljkHARDWARE Hardware, ljtBOOL *Available);

void
ljkASYNC_FE_UpdateAvaiable(ljkHARDWARE Hardware);

/******************************************************************************/
/* MC FE commands. */

/* One MCFE includes max 64 engine, each engine contains 2 channels. */
ljeSTATUS
ljkMCFE_Construct(ljkHARDWARE Hardware, ljkMCFE *FE);

void
ljkMCFE_Destroy(ljkHARDWARE Hardware, ljkMCFE FE);

/* Initialize MC FE, when hardware reset. */
ljeSTATUS
ljkMCFE_Initialize(ljkHARDWARE Hardware, ljtBOOL MMUEnabled, ljkMCFE FE);

/* Add a NOP command in the command queue. */
ljeSTATUS
ljkMCFE_Nop(ljkHARDWARE Hardware, ljtPOINTER Logical, ljtSIZE_T *Bytes);

/* Add an EVENT command in the command queue. */
ljeSTATUS
ljkMCFE_Event(ljkHARDWARE Hardware,
              ljtPOINTER Logical,
              ljtUINT8 Event,
              ljeKERNEL_WHERE FromWhere,
              ljtUINT32 *Bytes);

/* Add a SendSemaphore command in the command queue. */
ljeSTATUS
ljkMCFE_SendSemaphore(ljkHARDWARE Hardware,
                      ljtPOINTER Logical,
                      ljtUINT32 SemaId,
                      ljtUINT32 *Bytes);

/* Add a WaitSemaphore command in the command queue. */
ljeSTATUS
ljkMCFE_WaitSemaphore(ljkHARDWARE Hardware,
                      ljtPOINTER Logical,
                      ljtUINT32 SemaId,
                      ljtUINT32 *Bytes);

/* Kickstart the command processor. */
ljeSTATUS
ljkMCFE_Execute(ljkHARDWARE Hardware,
                ljtBOOL Priority,
                ljtUINT32 ChannelId,
                ljtADDRESS Address,
                ljtUINT32 Bytes);

/* Query hardware module idle */
ljeSTATUS
ljkMCFE_HardwareIdle(ljkHARDWARE Hardware, ljtBOOL_PTR IsIdle);
#endif

