/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_hardware_func_h_
#define __lj_hal_kernel_hardware_func_h_
#include "lj_hal.h"
#ifdef __cplusplus
extern "C" {
#endif

#define ljdFLOP_RESET       1
#define ljdFLOP_RESET_PPU   1
#define ljdFLOP_RESET_NN    1
#define ljdFLOP_RESET_TP    1
#define ljdFLOP_RESET_DEBUG 0
/*
 * The following macros are for old FLOP
 * reset path.
 * Use ljdFLOP_RESET=1 for the new path.
 * The related code will be removed as long
 * as the new path is stable.
 */
#define ljdINITIALIZE_PPU   1
#define ljdINITIALIZE_PPU_C 0
#define ljdRESET_USC1       0
#define ljdRESET_USC_C      0
#define ljdRESET_USC2       1

typedef struct _ljsFUNCTION_EXECUTION *ljsFUNCTION_EXECUTION_PTR;

typedef enum {
    ljvFUNCTION_EXECUTION_MMU,
    ljvFUNCTION_EXECUTION_FLUSH,
#if ljdFLOP_RESET
    ljvFUNCTION_EXECUTION_FLOP_RESET,
#else
#if ljdRESET_USC1
    ljvFUNCTION_EXECUTION_RESET_USC,
# endif
#if ljdRESET_USC2
    ljvFUNCTION_EXECUTION_RESET_USC2,
# endif
#if ljdINITIALIZE_PPU
    ljvFUNCTION_EXECUTION_INITIALIZE_PPU,
# endif
#endif

    ljvFUNCTION_EXECUTION_NUM
} ljeFUNCTION_EXECUTION;

typedef struct {
    /* Data Vidmem object */
    ljkVIDMEM_NODE          bufVidMem;

    /* Total bytes of the data. */
    ljtSIZE_T               bufVidMemBytes;

    /* Entry of the data. */
    ljtADDRESS              address;

    /* CPU address of the data. */
    ljtPOINTER              logical;

    /* Actually bytes of the data. */
    ljtSIZE_T               bytes;
} ljsFUNCTION_EXECUTION_DATA, *ljsFUNCTION_EXECUTION_DATA_PTR;

typedef struct {
    ljeSTATUS (*validate)(ljsFUNCTION_EXECUTION_PTR Execution);
    ljeSTATUS (*init)(ljsFUNCTION_EXECUTION_PTR Execution);
    ljeSTATUS (*execute)(ljsFUNCTION_EXECUTION_PTR Execution);
    ljeSTATUS (*release)(ljsFUNCTION_EXECUTION_PTR Execution);
} ljsFUNCTION_API, *ljsFUNCTION_API_PTR;

typedef struct _ljsFUNCTION_COMMAND {
    /* Function Vidmem object */
    ljkVIDMEM_NODE          funcVidMem;

    /* Total bytes of the function. */
    ljtSIZE_T               funcVidMemBytes;

    /* Entry of the function. */
    ljtADDRESS              address;

    /* CPU address of the function. */
    ljtPOINTER              logical;

    /* Physical address of this command. */
    ljtPHYS_ADDR_T          physical;

    /* Actually bytes of the function. */
    ljtUINT32               bytes;

    /* Hardware address of END in this function. */
    ljtADDRESS              endAddress;

    /* Logical of END in this function. */
    ljtUINT8_PTR            endLogical;

    /* Physical of End in this function. */
    ljtPHYS_ADDR_T          endPhysical;

    /* mcfe channel set. */
    ljtUINT32               channelId;

    /* Function private data */
    ljtUINT32               dataCount;
    ljsFUNCTION_EXECUTION_DATA_PTR      data;
#if ljdFLOP_RESET_DEBUG
    ljtPOINTER              golden;
    ljtPOINTER              outlogical;
    ljtUINT32               outSize;
#endif
} ljsFUNCTION_COMMAND, *ljsFUNCTION_COMMAND_PTR;

typedef struct _ljsFUNCTION_EXECUTION {
    ljtPOINTER              hardware;

    /* Function name */
    ljtCHAR                 funcName[16];

    /* Function ID */
    ljeFUNCTION_EXECUTION   funcId;

    /* Function count */
    ljtUINT8                funcCmdCount;

    /* Function array */
    ljsFUNCTION_COMMAND_PTR funcCmd;

    /* API of functions */
    ljsFUNCTION_API         funcExecution;

    /* Need this function or not */
    ljtBOOL                 valid;

    /* Function has inited */
    ljtBOOL                 inited;
} ljsFUNCTION_EXECUTION;

ljeSTATUS
ljkFUNCTION_Construct(ljtPOINTER Hardware);

ljeSTATUS
ljkFUNCTION_Destory(ljtPOINTER Hardware);

ljeSTATUS
ljkFUNCTION_CheckCHIPID(ljsFUNCTION_EXECUTION_PTR Execution);

ljeSTATUS
ljkFUNCTION_Validate(ljsFUNCTION_EXECUTION_PTR Execution, ljtBOOL_PTR Valid);

ljeSTATUS
ljkFUNCTION_Init(ljsFUNCTION_EXECUTION_PTR Execution);

ljeSTATUS
ljkFUNCTION_Execute(ljsFUNCTION_EXECUTION_PTR Execution);

ljeSTATUS
ljkFUNCTION_Release(ljsFUNCTION_EXECUTION_PTR Execution);

void
ljkFUNCTION_Dump(ljsFUNCTION_EXECUTION_PTR Execution);
#ifdef __cplusplus
}
#endif
#endif

