/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_device.h"

#include <linux/iommu.h>
#include <linux/platform_device.h>

#define _LJ_OBJ_ZONE ljvZONE_OS

void
ljkIOMMU_Destory(IN ljkOS Os, IN ljkIOMMU Iommu)
{
    ljmkHEADER_ARG("Os=%p Iommu=%p", Os, Iommu);

    if (Iommu) {
        if (Iommu->paddingPageDmaHandle) {
            dma_unmap_page(Iommu->device, Iommu->paddingPageDmaHandle,
                           PAGE_SIZE, DMA_FROM_DEVICE);
        }

        ljmkOS_SAFE_FREE(Os, Iommu);
    }

    ljmkFOOTER_NO();
}

ljeSTATUS
ljkIOMMU_Construct(IN ljkOS Os, OUT ljkIOMMU *Iommu)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkIOMMU iommu = ljvNULL;
    struct device *dev;
    dma_addr_t dmaHandle;
    ljtUINT64 phys;
    struct iommu_domain *domain;
    ljtUINT32 gfp = GFP_KERNEL;
    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)Os->device->platform;

    ljmkHEADER_ARG("Os=%p", Os);

    //dev = &Os->device->platform->device->dev;
    dev = &pcie_platform->pcie_info[0].pdev->dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
    domain = iommu_get_domain_for_dev(dev);
#else
    domain = ljvNULL;
#endif

    if (domain) {
        struct page *page;

        page = alloc_page(gfp);
        if (!page)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        phys = page_to_phys(page);

        dmaHandle = dma_map_page(dev, page, 0, PAGE_SIZE, DMA_TO_DEVICE);

        if (dmaHandle)
            dma_unmap_page(dev, dmaHandle, PAGE_SIZE, DMA_FROM_DEVICE);

        __free_page(page);

        /* Iommu bypass */
        if (phys == dmaHandle) {
            *Iommu = ljvNULL;

            ljmkFOOTER();
            return status;
        }
    } else {
        /* Don't enable iommu */
        *Iommu = ljvNULL;

        ljmkFOOTER();
        return status;
    }

    ljmkONERROR(ljkOS_Allocate(Os, ljmSIZEOF(ljsIOMMU), (ljtPOINTER *)&iommu));

    ljkOS_ZeroMemory(iommu, ljmSIZEOF(ljsIOMMU));

    if (Os->paddingPage)
        iommu->paddingPageDmaHandle = dma_map_page(dev, Os->paddingPage, 0, PAGE_SIZE, DMA_TO_DEVICE);

    iommu->domain = domain;
    iommu->device = dev;

    *Iommu = iommu;

    ljmkPRINT("[ljmcore]: Enable IOMMU\n");
OnError:
    if (ljmIS_ERROR(status))
        ljkIOMMU_Destory(Os, iommu);

    ljmkFOOTER();
    return status;
}
