/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_linux_h_
#define __lj_hal_kernel_linux_h_

#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#include <linux/vmalloc.h>
#include <linux/dma-mapping.h>
#include <linux/kthread.h>

#if ljdENABLE_TTM
#include <linux/pfn_t.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
# include <linux/iommu.h>
# include <linux/iova.h>
#endif

#include <linux/idr.h>

#ifdef MODVERSIONS
# include <linux/modversions.h>
#endif
#include <linux/io.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 7, 0)
# include <linux/uaccess.h>
#else
# include <asm/uaccess.h>
#endif

#if ENABLE_GPU_CLOCK_BY_DRIVER && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
# include <linux/clk.h>
#endif

#define NTSTRSAFE_NO_CCH_FUNCTIONS
#include "lj_hal.h"
#include "lj_hal_driver.h"
#include "lj_hal_kernel.h"
#include "lj_hal_kernel_platform.h"
#include "lj_hal_kernel_device.h"
#include "lj_hal_kernel_os.h"
#include "lj_hal_kernel_debugfs.h"

#if ljdENABLE_TRUST_APPLICATION
#include "lj_hal_ta.h"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 31)
# define FIND_TASK_BY_PID(x) pid_task(find_vpid(x), PIDTYPE_PID)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
# define FIND_TASK_BY_PID(x) find_task_by_vpid(x)
#else
# define FIND_TASK_BY_PID(x) find_task_by_pid(x)
#endif

#ifndef DEVICE_NAME
# define DEVICE_NAME "ljmcore"
#endif

#ifndef CLASS_NAME
# define CLASS_NAME "graphics_class"
#endif

#define GetPageCount(size, offset) \
    ((((size) + ((offset) & ~PAGE_MASK)) + PAGE_SIZE - 1) >> PAGE_SHIFT)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
# define ljdVM_FLAGS (VM_IO | VM_DONTCOPY | VM_DONTEXPAND | VM_DONTDUMP)
#else
# define ljdVM_FLAGS (VM_IO | VM_DONTCOPY | VM_DONTEXPAND | VM_RESERVED)
#endif

/* Protection bit when mapping memory to user sapce */
#define ljmkPAGED_MEMROY_PROT(x) pgprot_writecombine(x)

#define ljdSUPPRESS_OOM_MESSAGE 1

#if ljdSUPPRESS_OOM_MESSAGE
# define ljdNOWARN __GFP_NOWARN
#else
# define ljdNOWARN        0
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 1, 0)
#ifdef ljdIRQ_SHARED
#  define ljdIRQF_FLAG    (IRQF_SHARED)
# else
#  define ljdIRQF_FLAG    0
# endif
#else
#ifdef ljdIRQ_SHARED
#  define ljdIRQF_FLAG    (IRQF_DISABLED | IRQF_SHARED | IRQF_TRIGGER_HIGH)
# else
#  define ljdIRQF_FLAG    (IRQF_DISABLED | IRQF_TRIGGER_HIGH)
# endif
#endif

/* ljdLINUX_SYNC_FILE and CONFIG_SYNC_FILE. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)
# define dma_fence                           fence
# define dma_fence_array                     fence_array
# define dma_fence_ops                       fence_ops

# define dma_fence_default_wait fence_default_wait

# define dma_fence_signal(f)                 fence_signal(f)
# define dma_fence_signal_locked(f)          fence_signal_locked(f)
# define dma_fence_get(f)                    fence_get(f)
# define dma_fence_put(f)                    fence_put(f)
# define dma_fence_is_array(f)               fence_is_array(f)
# define dma_fence_is_signaled(f)            fence_is_signaled(f)
# define to_dma_fence_array(f)               to_fence_array(f)
# define dma_fence_wait_timeout(f, n, t)     fence_wait_timeout((f), (n), (t))
# define dma_fence_init(f, o, l, t, s)       fence_init((f), (o), (l), (t), (s))
# define dma_fence_context_alloc(s)          fence_context_alloc(s)

#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
# define current_mm_mmap_sem         current->mm->mmap_lock
#else
# define current_mm_mmap_sem         current->mm->mmap_sem
#endif

#ifndef PHYS_PFN
# define PHYS_PFN(x)                 ((unsigned long)((x) >> PAGE_SHIFT))
#endif

#ifndef untagged_addr
# define untagged_addr(addr)         (addr)
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 20, 17) &&  \
    !defined(CONFIG_ARCH_NO_SG_CHAIN)) ||               \
    (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 6, 0) &&   \
     (defined(ARCH_HAS_SG_CHAIN) || defined(CONFIG_ARCH_HAS_SG_CHAIN)))
# define ljdUSE_LINUX_SG_TABLE_API   1
#else
# define ljdUSE_LINUX_SG_TABLE_API   0
#endif

#ifndef ljdWAR_PCIE_WC
#define ljdWAR_PCIE_WC 0
#endif

/******************************************************************************
 ********************************** Structures ********************************
 *****************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
/* Just to compile without error */
struct iommu_domain {
    void *priv;
}
#endif

typedef struct _ljsIOMMU {
    struct iommu_domain   *domain;
    struct device         *device;
    dma_addr_t            paddingPageDmaHandle;
} ljsIOMMU;

typedef struct _ljsIOMMU        *ljkIOMMU;

typedef struct _ljsINTEGER_DB   *ljsINTEGER_DB_PTR;

typedef struct _ljsINTEGER_DB {
    struct idr                  idr;
    /* protect data of _ljsINTEGER_DB */
    spinlock_t                  lock;
    ljtINT                      curr;
} ljsINTEGER_DB;

struct _ljkOS {
    /* Object. */
    ljsOBJECT                   object;

    /* Pointer to device */
    ljkGALDEVICE                device;

    /* Memory management */
    struct mutex                mdlMutex;
    struct list_head            mdlHead;

    /* Kernel process ID. */
    ljtUINT32                   kernelProcessID;

    /* Signal management. */

    /* Lock. */
    spinlock_t                  signalLock;

    /* signal id database. */
    ljsINTEGER_DB               signalDB;

    /* workqueue for os timer. */
    struct workqueue_struct     *workqueue;

    /* Allocate extra page to avoid cache overflow */
    struct page                 *paddingPage;

    /* Detect unfreed allocation. */
    atomic_t                    allocateCount;

    struct list_head            allocatorList;

    ljsDEBUGFS_DIR              allocatorDebugfsDir;

    /* Lock for register access check. */
    spinlock_t                  registerAccessLock;

    /* External power states. */
    ljtBOOL                     powerStates[ljdDEVICE_COUNT][ljdMAX_GPU_COUNT];

    /* External clock states. */
    ljtBOOL                     clockStates[ljdDEVICE_COUNT][ljdMAX_GPU_COUNT];

    /* IOMMU. */
    ljkIOMMU                    iommu;

    /* Dump in kernel. */
    struct file                 *dumpFilp;
    /* dump file mutex */
    struct mutex                dumpFilpMutex;

    int                         dumpTarget;
    char                        dumpFileName[256];
    ljsDEBUGFS_DIR              dumpDebugfsDir;

    atomic_t                    nodeID;
};

typedef struct _ljsSIGNAL *ljsSIGNAL_PTR;

typedef struct _ljsSIGNAL {
    /* Kernel sync primitive. */
    volatile unsigned int       done;
    /* protect signal */
    spinlock_t                  lock;

    wait_queue_head_t           wait;

    /* Manual reset flag. */
    ljtBOOL                     manualReset;

    /* The reference counter. */
    atomic_t                    ref;

    /* The owner of the signal. */
    ljtHANDLE                   process;

    /* ID. */
    ljtUINT32                   id;

    /* Signal return status. */
    ljeSIGNAL_STATUS            status;

#if ljdLINUX_SYNC_FILE
#ifndef CONFIG_SYNC_FILE
    /* Parent timeline. */
    struct sync_timeline        *timeline;
# else
    struct dma_fence            *fence;
# endif
#endif
} ljsSIGNAL;

typedef struct _ljsOSTIMER *ljsOSTIMER_PTR;

typedef struct _ljsOSTIMER {
    struct delayed_work     work;
    ljtTIMERFUNCTION        function;
    ljtPOINTER              data;
} ljsOSTIMER;

ljeSTATUS
ljkOS_ImportAllocators(ljkOS Os);

ljeSTATUS
ljkOS_FreeAllocators(ljkOS Os);

ljeSTATUS
_ConvertLogical2Physical(ljkOS Os, ljtPOINTER Logical, ljtUINT32 ProcessID,
                         PLINUX_MDL Mdl, ljtPHYS_ADDR_T *Physical);

ljeSTATUS
_QuerySignal(ljkOS Os, ljtSIGNAL Signal);

static inline ljtINT
_GetProcessID(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
    return task_tgid_nr(current);
#else
    return current->tgid;
#endif
}

char *_GetProcessComm(void);

static inline void
_MemoryBarrier(void)
{
#if defined(CONFIG_ARM) && (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 34))
    dsb();
#else
    /* memory barrier */
    mb();
#endif
}

static inline void
_Barrier(void)
{
    barrier();
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 25)
static inline int
is_vmalloc_addr(void *Addr)
{
    unsigned long addr = (unsigned long)Addr;

    return addr >= VMALLOC_START && addr < VMALLOC_END;
}
#endif

void
ljm_device_node_destroy(uint32_t dev_index);

ljeSTATUS
ljm_device_node_create(uint32_t dev_index);

void
ljkIOMMU_Destory(ljkOS Os, ljkIOMMU Iommu);

ljeSTATUS
ljkIOMMU_Construct(ljkOS Os, ljkIOMMU *Iommu);

#endif /* __lj_hal_kernel_linux_h_ */
