/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef _lj_hal_kenrel_mmu_h_
#define _lj_hal_kenrel_mmu_h_

#if defined(EMULATOR) && defined(_WIN32)
# include "ljDefines.h"
#if defined(ljdVIRTUAL_ADDRESS_WIDTH) && (ljdVIRTUAL_ADDRESS_WIDTH == 40)
#  define ljdENABLE_40BIT_VA        1
#  define ljdMMU_40VA_40PA          1
# endif
#endif

#ifndef ljdENABLE_40BIT_VA
# define ljdENABLE_40BIT_VA         0
#endif

#ifndef ljdMMU_40VA_40PA
# define ljdMMU_40VA_40PA           0
#endif

#define ljd4G_VA_FM_SIZE            0x40000000

#ifndef ljdDEBUG_MMU_SWITCH
#define ljdDEBUG_MMU_SWITCH         0
#endif

#ifndef ljdMMU_DESC_SIZE
#define ljdMMU_DESC_SIZE            (1 << 16)
#endif

/*******************************************************************************
 ***** New MMU Defination ******************************************************/

#if ljdENABLE_MMU_1KMODE
/* 1k mode */
# define ljdMMU_MTLB_SHIFT       24
# define ljdMMU_VA_BITS          32

#else /* ljdENABLE_MMU_1KMODE */
/* 4K mode */
# define ljdMMU_MTLB_SHIFT       22
# define ljdMMU_VA_BITS          32
#endif

#if ljdENABLE_40BIT_VA
# undef ljdMMU_VA_BITS
# undef ljdMMU_MTLB_SHIFT

#if ljdMMU_40VA_40PA
#  define ljdMMU_MTLB_SHIFT      30
# else
#  define ljdMMU_MTLB_SHIFT      24
# endif

# define ljdMMU_VA_BITS          40

#ifndef ljdDDR_SIZE_MAX
# define ljdDDR_SIZE_MAX         0x800000000
# endif
#endif

#define ljdMMU_STLB_4K_SHIFT        12
#define ljdMMU_STLB_64K_SHIFT       16
#define ljdMMU_STLB_1M_SHIFT        20
#define ljdMMU_STLB_16M_SHIFT       24

#define ljdMMU_MTLB_BITS            (ljdMMU_VA_BITS - ljdMMU_MTLB_SHIFT)
#define ljdMMU_PAGE_4K_BITS         ljdMMU_STLB_4K_SHIFT
#define ljdMMU_STLB_4K_BITS         (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_PAGE_4K_BITS)
#define ljdMMU_PAGE_64K_BITS        ljdMMU_STLB_64K_SHIFT
#define ljdMMU_STLB_64K_BITS        (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_PAGE_64K_BITS)
#define ljdMMU_PAGE_1M_BITS         ljdMMU_STLB_1M_SHIFT
#define ljdMMU_STLB_1M_BITS         (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_PAGE_1M_BITS)
#define ljdMMU_PAGE_16M_BITS        ljdMMU_STLB_16M_SHIFT

#if ljdENABLE_40BIT_VA && ljdMMU_40VA_40PA
# define ljdMMU_STLB_16M_BITS    (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_PAGE_16M_BITS)
#else
# define ljdMMU_STLB_16M_BITS    4
#endif

#ifndef ljdMMU_MTLB_ENTRY_NUM

#if defined(EMULATOR) && ljdENABLE_40BIT_VA
# define ljdMMU_MTLB_ENTRY_NUM   40
#elif ljdENABLE_40BIT_VA
# define ljdMMU_MTLB_ENTRY_NUM   ((ljdDDR_SIZE_MAX * 2) >> ljdMMU_MTLB_SHIFT)
#else
# define ljdMMU_MTLB_ENTRY_NUM   (1 << ljdMMU_MTLB_BITS)
#endif

#endif

#define ljdMMU_MTLB_SIZE            (ljdMMU_MTLB_ENTRY_NUM << 2)
#define ljdMMU_STLB_4K_ENTRY_NUM    (1 << ljdMMU_STLB_4K_BITS)
#define ljdMMU_STLB_4K_SIZE         (ljdMMU_STLB_4K_ENTRY_NUM << 2)
#define ljdMMU_PAGE_4K_SIZE         (1 << ljdMMU_STLB_4K_SHIFT)
#define ljdMMU_STLB_64K_ENTRY_NUM   (1 << ljdMMU_STLB_64K_BITS)
#define ljdMMU_STLB_64K_SIZE        (ljdMMU_STLB_64K_ENTRY_NUM << 2)
#define ljdMMU_PAGE_64K_SIZE        (1 << ljdMMU_STLB_64K_SHIFT)
#define ljdMMU_STLB_1M_ENTRY_NUM    (1 << ljdMMU_STLB_1M_BITS)
#define ljdMMU_STLB_1M_SIZE         (ljdMMU_STLB_1M_ENTRY_NUM << 2)
#define ljdMMU_PAGE_1M_SIZE         (1 << ljdMMU_STLB_1M_SHIFT)
#define ljdMMU_STLB_16M_ENTRY_NUM   (1 << ljdMMU_STLB_16M_BITS)
#define ljdMMU_STLB_16M_SIZE        (ljdMMU_STLB_16M_ENTRY_NUM << 2)
#define ljdMMU_PAGE_16M_SIZE        (1 << ljdMMU_STLB_16M_SHIFT)

#if ljdENABLE_40BIT_VA
# define ljdMMU_VA_MASK          ((1ULL << ljdMMU_VA_BITS) - 1)
# define ljdMMU_MTLB_MASK        (~((1ULL << ljdMMU_MTLB_SHIFT) - 1) & ljdMMU_VA_MASK)

# define ljdMMU_STLB_4K_MASK                                                                    \
     (((~0ULL << ljdMMU_STLB_4K_SHIFT) ^ ljdMMU_MTLB_MASK) & ljdMMU_VA_MASK)
# define ljdMMU_STLB_64K_MASK                                                                   \
     (((~((1ULL << ljdMMU_STLB_64K_SHIFT) - 1)) ^ ljdMMU_MTLB_MASK) & ljdMMU_VA_MASK)
# define ljdMMU_STLB_1M_MASK                                                                    \
     (((~((1ULL << ljdMMU_STLB_1M_SHIFT) - 1)) ^ ljdMMU_MTLB_MASK) & ljdMMU_VA_MASK)
# define ljdMMU_STLB_16M_MASK                                                                   \
     (((~((1ULL << ljdMMU_STLB_16M_SHIFT) - 1)) ^ ljdMMU_MTLB_MASK) & ljdMMU_VA_MASK)
#else
# define ljdMMU_MTLB_MASK        (~((1U << ljdMMU_MTLB_SHIFT) - 1))
# define ljdMMU_STLB_4K_MASK     ((~0U << ljdMMU_STLB_4K_SHIFT) ^ ljdMMU_MTLB_MASK)
# define ljdMMU_STLB_64K_MASK    ((~((1U << ljdMMU_STLB_64K_SHIFT) - 1)) ^ ljdMMU_MTLB_MASK)
# define ljdMMU_STLB_1M_MASK     ((~((1U << ljdMMU_STLB_1M_SHIFT) - 1)) ^ ljdMMU_MTLB_MASK)
# define ljdMMU_STLB_16M_MASK    0x0F000000
#endif

#define ljdMMU_PAGE_4K_MASK         (ljdMMU_PAGE_4K_SIZE - 1)
#define ljdMMU_PAGE_64K_MASK        (ljdMMU_PAGE_64K_SIZE - 1)
#define ljdMMU_PAGE_1M_MASK         (ljdMMU_PAGE_1M_SIZE - 1)
#define ljdMMU_PAGE_16M_MASK        (ljdMMU_PAGE_16M_SIZE - 1)

/* Page offset definitions. */
#define ljdMMU_OFFSET_4K_BITS       (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_STLB_4K_BITS)
#define ljdMMU_OFFSET_4K_MASK       ((1U << ljdMMU_OFFSET_4K_BITS) - 1)
#define ljdMMU_OFFSET_64K_BITS      (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_STLB_64K_BITS)
#define ljdMMU_OFFSET_64K_MASK      ((1U << ljdMMU_OFFSET_64K_BITS) - 1)
#define ljdMMU_OFFSET_1M_BITS       (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS - ljdMMU_STLB_1M_BITS)
#define ljdMMU_OFFSET_1M_MASK       ((1U << ljdMMU_OFFSET_1M_BITS) - 1)
#define ljdMMU_OFFSET_16M_BITS      (ljdMMU_VA_BITS - ljdMMU_MTLB_BITS)
#define ljdMMU_OFFSET_16M_MASK      ((1U << ljdMMU_OFFSET_16M_BITS) - 1)

#define ljdMMU_MTLB_ENTRY_HINTS_BITS 6
#define ljdMMU_MTLB_ENTRY_STLB_MASK  (~((1U << ljdMMU_MTLB_ENTRY_HINTS_BITS) - 1))

#define ljdMMU_STLB_EXCEPTION       0x00000002

#define ljd1M_PAGE_SIZE             (1 << 20)
#define ljd1M_PAGE_SHIFT            20
#define ljd4K_PAGE_SIZE             (1 << 12)

#define ljd4G_SIZE                  0x100000000
#define ljdVA_RESERVED_SIZE         (16 << 20)

/* VIP SRAM start virtual address. */
#define ljdRESERVE_START            (4 << 20)

#define ljdRESERVE_ALIGN            (4 << 10)

#define ljmENTRY_TYPE(x)            ((x) & 0xF0)

#define ljmENTRY_COUNT(x)           (((x) & 0xFFFFFFFF00) >> 8)

#define ljdMMU_TABLE_DUMP           0

#define ljdVERTEX_START             (128 << 10)

#define ljdFLAT_MAPPING_MODE        ljvPAGE_TYPE_16M

typedef enum _ljeMMU_TYPE {
    ljvMMU_USED   = (0 << 4),
    ljvMMU_SINGLE = (1 << 4),
    ljvMMU_FREE   = (2 << 4),
} ljeMMU_TYPE;

typedef struct _ljsMMU_STLB_CHUNK *ljsMMU_STLB_CHUNK_PTR;

typedef struct _ljsMMU_STLB_CHUNK {
    ljkVIDMEM_NODE        videoMem;
    ljtUINT32_PTR         logical;
    ljtSIZE_T             size;
    ljtPHYS_ADDR_T        physBase;
    ljtSIZE_T             pageCount;
    ljtUINT32             mtlbIndex;
    ljtUINT32             mtlbEntryNum;
    ljsMMU_STLB_CHUNK_PTR next;
} ljsMMU_STLB_CHUNK;

typedef struct _ljsFreeSpaceNode *ljsFreeSpaceNode_PTR;
typedef struct _ljsFreeSpaceNode {
    ljtUINT32 start;
    ljtUINT32 entries;
} ljsFreeSpaceNode;

#if ljdENDIAN_BIG

# define _WritePageEntry(pageEntry, entryValue) \
        (*(ljtUINT32_PTR)(pageEntry) = ljmBSWAP32((ljtUINT32)(entryValue)))

# define _ReadPageEntry(pageEntry) \
        ljmBSWAP32(*(ljtUINT32_PTR)(pageEntry))

#else

# define _WritePageEntry(pageEntry, entryValue) \
        (*(ljtUINT32_PTR)(pageEntry) = (ljtUINT32)(entryValue))

# define _ReadPageEntry(pageEntry) \
        (*(ljtUINT32_PTR)(pageEntry))

#endif

typedef enum _ljeMMU_INIT_MODE {
    ljvMMU_INIT_FROM_REG,
    ljvMMU_INIT_FROM_CMD,
} ljeMMU_INIT_MODE;

typedef struct _ljsADDRESS_AREA *ljsADDRESS_AREA_PTR;
typedef struct _ljsADDRESS_AREA {
    /* Page table / STLB table information. */
    ljtSIZE_T                   stlbSize;
    ljkVIDMEM_NODE              stlbVideoMem;
    ljtUINT32_PTR               stlbLogical;
    ljtUINT32                   stlbEntries;
    /* stlb physical address. */
    ljtPHYS_ADDR_T              stlbPhysical;

    /* Free entries. */
    ljtUINT32                   heapList;
    ljtBOOL                     freeNodes;

    ljeAREA_TYPE                areaType;

    ljtUINT32                   mappingStart;
    ljtUINT32                   mappingEnd;

    ljtUINT64_PTR               mapLogical;

    ljtUINT32                   usedIndex;
} ljsADDRESS_AREA;

/* ljkMMU object. */
struct _ljkMMU {
    /* The object. */
    ljsOBJECT                   object;

    /* Pointer to ljkOS object. */
    ljkOS                       os;

    /* Pointer to ljkHARDWARE object. */
    ljkHARDWARE                 hardware;

    /* The page table mutex. */
    ljtPOINTER                  pageTableMutex;

    /* Master TLB information. */
    ljtSIZE_T                   mtlbSize;
    ljkVIDMEM_NODE              mtlbVideoMem;
    ljtUINT32_PTR               mtlbLogical;
    ljtUINT32                   mtlbEntries;
    /* mtlb physical address. */
    ljtPHYS_ADDR_T              mtlbPhysical;

    /* memory pool used for page table */
    ljePOOL                     pool;

    ljtPOINTER                  staticSTLB;
    ljtBOOL                     enabled;

    ljtSIZE_T                   safePageSize;
    ljkVIDMEM_NODE              safePageVideoMem;
    ljtPOINTER                  safePageLogical;
    ljtADDRESS                  safeAddress;
    /* Safe page physical address. */
    ljtPHYS_ADDR_T              safePagePhysical;

    /* GPU physical address flat mapping area. */
    ljtUINT32                   gpuPhysicalRangeCount;
    ljsFLAT_MAPPING_RANGE       gpuPhysicalRanges[ljdMAX_FLAT_MAPPING_COUNT];

    /* GPU virtual address flat mapping area*/
    ljtUINT32                   gpuAddressRangeCount;
    ljsFLAT_MAPPING_RANGE       gpuAddressRanges[ljdMAX_FLAT_MAPPING_COUNT];

    /* List of hardware which uses this MMU. */
    ljsLISTHEAD                 hardwareList;

    struct _ljkQUEUE            recentFreedAddresses;

    ljsADDRESS_AREA             dynamicArea1M;
    ljsADDRESS_AREA             dynamicArea4K;
    ljsADDRESS_AREA             dynamicLowArea1M;
    ljsADDRESS_AREA             dynamicLowArea4K;
    ljsADDRESS_AREA             secureArea;

    ljtBOOL                     dynamicAreaSetuped;

    ljtBOOL                     sRAMMapped;

    ljtADDRESS                  contiguousBaseAddresses[ljdSYSTEM_RESERVE_COUNT];
    ljtADDRESS                  externalBaseAddress;
    ljtADDRESS                  internalBaseAddress;
    ljtADDRESS                  exclusiveBaseAddress;

    ljeMMU_INIT_MODE            initMode;
    ljtBOOL                     pageTableOver4G;

    /* Flat mapping for reserve memory. */
    ljtBOOL                     flatMapping;

    ljePAGE_TYPE                flatMappingMode;

    /* If the stlb is allocated when page size is 16M . */
    ljtBOOL                     stlbAllocated[ljdMMU_STLB_16M_ENTRY_NUM];

    /* The reserve size in page table. */
    ljtSIZE_T                   reserveRangeSize;

    ljsLISTHEAD                 nodeList;
    ljtPOINTER                  nodeListMutex;

    ljtUINT32                   descIndex;
};

ljeSTATUS
ljkMMU_GetPageEntry(ljkMMU Mmu, ljePAGE_TYPE PageType, ljtBOOL LowVA,
                    ljtADDRESS Address, ljtUINT32_PTR *PageTable);

ljeSTATUS
ljkMMU_Update(ljkMMU Mmu, ljtUINT64 Address, ljtUINT64 Physical);

ljeSTATUS
ljkMMU_SetupSRAM(ljkMMU Mmu, ljkHARDWARE Hardware, ljkDEVICE Device);

ljeSTATUS
ljkMMU_SetupDynamicSpace(ljkMMU Mmu);

void
ljkMMU_DumpRecentFreedAddress(ljkMMU Mmu);

ljeSTATUS
ljkMMU_DestroyProcessMMU(ljkMMU Mmu);

ljeSTATUS
ljkMMU_ConstructProcessMMU(ljkKERNEL Kernel, ljtUINT32 ProcessID, ljkMMU *Mmu);

ljeSTATUS
ljkMMU_CopyDynamicAreas(ljkKERNEL Kernel, ljkMMU dstMMU);

ljeSTATUS
ljkMMU_SwitchMtlb(ljkMMU dstMMU, ljkMMU srcMMU);

ljeSTATUS
ljkMMU_AttachNode(ljkMMU Mmu, ljuVIDMEM_NODE_PTR Node);

ljeSTATUS
ljkMMU_DetachNode(ljkMMU Mmu, ljuVIDMEM_NODE_PTR Node);

ljeSTATUS
ljkMMU_DestroyMmuCopy(ljkMMU Mmu);

ljeSTATUS
ljkMMU_ConstructMmuCopy(ljkKERNEL Kernel, ljkMMU *MmuCopy);

#endif /* _lj_hal_kernel_mmu_h */
