/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef _LJ_HAL_KERNEL_PARAMETER_H_
#define _LJ_HAL_KERNEL_PARAMETER_H_

/******************************************************************************
 *                       register memory related                              *
 ******************************************************************************/

static ulong registerMemBase = 0x80000000;
module_param(registerMemBase, ulong, 0644);
MODULE_PARM_DESC(registerMemBase, "Base of bus address of LJ core AHB register");

static ulong registerMemSize = 2 << 16;
module_param(registerMemSize, ulong, 0644);
MODULE_PARM_DESC(registerMemSize, "Size of bus address range of LJ core AHB register");

static ulong registerMemBase2D = 0x00000000;
module_param(registerMemBase2D, ulong, 0644);
MODULE_PARM_DESC(registerMemBase2D, "Base of bus address of G2D core, if registerMemBase2D is used for a G3D core");

static ulong registerMemSize2D = 2 << 16;
module_param(registerMemSize2D, ulong, 0644);
MODULE_PARM_DESC(registerMemSize2D, "Size of bus address range of G2D core, if registerMemSize is used for a G3D core");

static ulong registerMemBaseVG = 0x00000000;
module_param(registerMemBaseVG, ulong, 0644);
MODULE_PARM_DESC(registerMemBaseVG, "Base of bus address of VG core");

static ulong registerMemSizeVG = 2 << 10;
module_param(registerMemSizeVG, ulong, 0644);
MODULE_PARM_DESC(registerMemSizeVG, "Size of bus address range of VG core");

#if ljdDEC_ENABLE_AHB
static ulong registerMemBaseDEC300 = 0x00000000;
module_param(registerMemBaseDEC300, ulong, 0644);

static ulong registerMemSizeDEC300 = 2 << 10;
module_param(registerMemSizeDEC300, ulong, 0644);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
static ljtPHYS_ADDR_T registerBases[ljdGLOBAL_CORE_COUNT];
module_param_array(registerBases, ullong, NULL, 0644);
#else
static ulong registerBases[ljdGLOBAL_CORE_COUNT];
module_param_array(registerBases, ulong, NULL, 0644);
#endif
MODULE_PARM_DESC(registerBases, "Array of bases of bus address of register of multi-core");

static ulong registerSizes[ljdGLOBAL_CORE_COUNT] = {
    [0 ... ljdGLOBAL_CORE_COUNT - 1] = 2 << 16
};
module_param_array(registerSizes, ulong, NULL, 0644);
MODULE_PARM_DESC(registerSizes, "Array of sizes of bus address range of register of multi-core");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
static ljtPHYS_ADDR_T register2DBases[ljdCORE_2D_COUNT];
module_param_array(register2DBases, ullong, NULL, 0644);
#else
static ulong register2DBases[ljdCORE_2D_COUNT];
module_param_array(register2DBases, ulong, NULL, 0644);
#endif
MODULE_PARM_DESC(register2DBases, "Array of bases of bus address of register of multi-2D");

static ulong register2DSizes[ljdCORE_2D_COUNT] = {
    [0 ... ljdCORE_2D_COUNT - 1] = 2 << 16
};
module_param_array(register2DSizes, ulong, NULL, 0644);
MODULE_PARM_DESC(register2DSizes, "Array of sizes of bus address range of register of multi-2D");

static uint registerAPB = 0x300000;
module_param(registerAPB, uint, 0644);
MODULE_PARM_DESC(registerAPB, "The offset of APB register to the register base address.");

/******************************************************************************
 *                         memory pool related                                *
 ******************************************************************************/

#ifndef ljdDEFAULT_CONTIGUOUS_SIZE
# define ljdDEFAULT_CONTIGUOUS_SIZE (4 << 20)
#endif

static ulong contiguousSize = ljdDEFAULT_CONTIGUOUS_SIZE;
module_param(contiguousSize, ulong, 0644);
MODULE_PARM_DESC(contiguousSize, "Size of reserved system memory");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
static ljtPHYS_ADDR_T contiguousBase = 0;
module_param(contiguousBase, ullong, 0644);
#else
static ulong contiguousBase = 0;
module_param(contiguousBase, ulong, 0644);
#endif
MODULE_PARM_DESC(contiguousBase, "Base address of reserved system memory");

static ulong contiguousSizes[ljdSYSTEM_RESERVE_COUNT] = {
    [0 ... ljdSYSTEM_RESERVE_COUNT - 1] = 0
};
module_param_array(contiguousSizes, ulong, NULL, 0644);
MODULE_PARM_DESC(contiguousSizes, "Sizes of reserved system memory array");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
static ljtPHYS_ADDR_T contiguousBases[ljdSYSTEM_RESERVE_COUNT] = {
    [0 ... ljdSYSTEM_RESERVE_COUNT - 1] = 0
};
module_param_array(contiguousBases, ullong, NULL, 0644);
#else
static ulong contiguousBases[ljdSYSTEM_RESERVE_COUNT] = {
    [0 ... ljdSYSTEM_RESERVE_COUNT - 1] = 0
};
module_param_array(contiguousBases, ulong, NULL, 0644);
#endif
MODULE_PARM_DESC(contiguousBases, "Base addresses of reserved system memory array");

static ulong externalSize[ljdLOCAL_MEMORY_COUNT] = {
    [0 ... ljdLOCAL_MEMORY_COUNT - 1] = 0
};
module_param_array(externalSize, ulong, NULL, 0644);
MODULE_PARM_DESC(externalSize, "Size of external local memory, if it is 0, means there is no external pool");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
static ljtPHYS_ADDR_T externalBase[ljdLOCAL_MEMORY_COUNT] = {
    [0 ... ljdLOCAL_MEMORY_COUNT - 1] = 0
};
module_param_array(externalBase, ullong, NULL, 0644);
#else
static ulong externalBase[ljdLOCAL_MEMORY_COUNT] = {
    [0 ... ljdLOCAL_MEMORY_COUNT - 1] = 0
};
module_param_array(externalBase, ulong, NULL, 0644);
#endif
MODULE_PARM_DESC(externalBase, "Base address of external memory");

static ulong exclusiveSize[ljdLOCAL_MEMORY_COUNT] = {
    [0 ... ljdLOCAL_MEMORY_COUNT - 1] = 0
};
module_param_array(exclusiveSize, ulong, NULL, 0644);
MODULE_PARM_DESC(exclusiveSize, "Size of exclusive local memory, if it is 0, means there is no exclusive pool");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
static ljtPHYS_ADDR_T exclusiveBase[ljdLOCAL_MEMORY_COUNT] = {
    [0 ... ljdLOCAL_MEMORY_COUNT - 1] = 0
};
module_param_array(exclusiveBase, ullong, NULL, 0644);
#else
static ulong exclusiveBase[ljdLOCAL_MEMORY_COUNT] = {
    [0 ... ljdLOCAL_MEMORY_COUNT - 1] = 0
};
module_param_array(exclusiveBase, ulong, NULL, 0644);
#endif
MODULE_PARM_DESC(exclusiveBase, "Base address of exclusive memory(GPU access only)");

/******************************************************************************
 *                          interrupt related                                 *
 ******************************************************************************/

static int irqLine = -1;
module_param(irqLine, int, 0644);
MODULE_PARM_DESC(irqLine, "IRQ number of LJ core");

static int irqLine2D = -1;
module_param(irqLine2D, int, 0644);
MODULE_PARM_DESC(irqLine2D, "IRQ number of G2D core if irqLine is used for a G3D core");

static int irqLineVG = -1;
module_param(irqLineVG, int, 0644);
MODULE_PARM_DESC(irqLineVG, "IRQ number of VG core");

static int irqs[ljdGLOBAL_CORE_COUNT] = {[0 ... ljdGLOBAL_CORE_COUNT - 1] = -1};
module_param_array(irqs, int, NULL, 0644);
MODULE_PARM_DESC(irqs, "Array of IRQ numbers of multi-core");

static int irq2Ds[ljdCORE_2D_COUNT] = {[0 ... ljdCORE_2D_COUNT - 1] = -1};
module_param_array(irq2Ds, int, NULL, 0644);
MODULE_PARM_DESC(irq2Ds, "Array of IRQ numbers of multi-2D");

static uint isrPoll = 0;
module_param(isrPoll, uint, 0644);
MODULE_PARM_DESC(isrPoll, "Bits isr polling for per-core, default 0'1b means disable, 1'1b means auto enable isr polling mode");

/******************************************************************************
 *                        identification related                              *
 ******************************************************************************/

static uint platformIDs[ljdPLATFORM_COUNT] = {[0 ... ljdDEVICE_COUNT - 1] = 0};     /* 平台序号, hym on 20220913 */
module_param_array(platformIDs, uint, NULL, 0644);
MODULE_PARM_DESC(platformIDs, "Array of platform id of each hardware device");

static uint hwDevCounts[ljdPLATFORM_COUNT] = {[0 ... ljdPLATFORM_COUNT - 1] = 1};   /* 平台中设备的数量 */
module_param_array(hwDevCounts, uint, NULL, 0644);
MODULE_PARM_DESC(hwDevCounts, "Array of hw device count of each platform");

static uint devCoreCounts[ljdDEVICE_COUNT] = {ljdCORE_3D_COUNT, 0};                 /* 设备中3D核的数量 */
module_param_array(devCoreCounts, uint, NULL, 0644);
MODULE_PARM_DESC(devCoreCounts, "Array of core count of each hardware device");

static uint dev2DCoreCounts[ljdDEVICE_COUNT] = {ljdCORE_2D_COUNT, 0};               /* 设备中2D核的数量 */
module_param_array(dev2DCoreCounts, uint, NULL, 0644);
MODULE_PARM_DESC(dev2DCoreCounts, "Array of core 2D count of each hardware device");

static uint devMemIDs[ljdDEVICE_COUNT] = {[0 ... ljdDEVICE_COUNT - 1] = 0};
module_param_array(devMemIDs, uint, NULL, 0644);
MODULE_PARM_DESC(devMemIDs, "Array of local memory index of each hardware device");

static uint devSysMemIDs[ljdDEVICE_COUNT] = {[0 ... ljdDEVICE_COUNT - 1] = 1};
module_param_array(devSysMemIDs, uint, NULL, 0644);
MODULE_PARM_DESC(devSysMemIDs, "Array of system reserved memory index of each hardware device, each bit represents memory id");

static uint devSRAMIDs[ljdDEVICE_COUNT] = {[0 ... ljdDEVICE_COUNT - 1] = 1};
module_param_array(devSRAMIDs, uint, NULL, 0644);
MODULE_PARM_DESC(devSRAMIDs, "Array of SRAM index of each hardware device, each bit represents memory id");

static uint chipIDs[ljvCORE_COUNT] = {
    [0 ... ljvCORE_COUNT - 1] = ljvCHIP_ID_DEFAULT
};
module_param_array(chipIDs, uint, NULL, 0644);
MODULE_PARM_DESC(chipIDs, "Array of chipIDs of multi-chips");

/******************************************************************************
 *                           driver function                                  *
 ******************************************************************************/

static int fastClear = -1;
module_param(fastClear, int, 0644);
MODULE_PARM_DESC(fastClear, "Disable fast clear if set it to 0, enabled by default");

static int compression = -1;
module_param(compression, int, 0644);
MODULE_PARM_DESC(compression, "Disable compression if set it to 0, enabled by default");

static int powerManagement = 1;
module_param(powerManagement, int, 0644);
MODULE_PARM_DESC(powerManagement, "Disable auto power saving if set it to 0, enabled by default");

static uint recovery = 1;
module_param(recovery, uint, 0644);
MODULE_PARM_DESC(recovery, "Recover GPU from stuck (1: Enable, 0: Disable)");

/*
 * Level of stuck dump content, 0~5 and 11~15.
 * 0: Disable. 1: Dump nearby memory. 2: Dump user command.
 * 3: Commit stall besides level2. 4: Dump kernel command buffer besides level3.
 * 5: Dump all the cores with level4.
 *
 * Level 1~5 + 10 means force dump, whether the recovery is enabled or not.
 */
static uint stuckDump = 0;
module_param(stuckDump, uint, 0644);
MODULE_PARM_DESC(stuckDump, "Level of stuck dump content.");

/*
 * Level of debug zone.
 * 0: Disable.  1: debug level verbose and zone all.
 */
static uint debugLevel = 0;
module_param(debugLevel, uint, 0644);
MODULE_PARM_DESC(debugLevel, "Level of debug.");

static int showArgs = 0;
module_param(showArgs, int, 0644);
MODULE_PARM_DESC(showArgs, "Display parameters value when driver loaded");

static uint userClusterMasks[ljdMAX_MAJOR_CORE_COUNT] = {
    [0 ... ljdMAX_MAJOR_CORE_COUNT - 1] = 0xff
};
module_param_array(userClusterMasks, uint, NULL, 0644);
MODULE_PARM_DESC(userClusterMasks, "Array of user defined per-core cluster enable mask");

static uint enableNN = 0xFF;
module_param(enableNN, uint, 0644);
MODULE_PARM_DESC(enableNN, "How many NN cores will be enabled in one VIP, 0xFF means all enabled, 0 means all disabled, 1 means enable 1 NN core...");

/* GPU small batch feature. */
static int smallBatch = 1;
module_param(smallBatch, int, 0644);
MODULE_PARM_DESC(smallBatch, "Enable/disable GPU small batch feature, enable by default");

static int allMapInOne = 1;
module_param(allMapInOne, int, 0644);
MODULE_PARM_DESC(allMapInOne, "Mapping kernel video memory to user, 0 means mapping every time, otherwise only mapping one time");

static uint softReset = 1;
module_param(softReset, uint, 0644);
MODULE_PARM_DESC(softReset, "Disable soft reset when insert the driver if set it to 0, enabled by default.");

/******************************************************************************
 *                            SRAM related                                    *
 ******************************************************************************/

static uint sRAMLoopMode = 0;
module_param(sRAMLoopMode, uint, 0644);
MODULE_PARM_DESC(sRAMLoopMode, "Default 0 means SRAM pool must be specified when allocating SRAM memory, 1 means SRAM memory will be looped as default pool.");

/*
 * Per-core SRAM physical address base, the order of configuration is according
 * to access speed, ljvINVALID_PHYSICAL_ADDRESS means no bus address.
 */
static ljtPHYS_ADDR_T sRAMBases[ljvSRAM_INTER_COUNT * ljdGLOBAL_CORE_COUNT] = {
    [0 ... ljvSRAM_INTER_COUNT * ljdGLOBAL_CORE_COUNT - 1] = ljvINVALID_PHYSICAL_ADDRESS
};
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
module_param_array(sRAMBases, ullong, NULL, 0644);
MODULE_PARM_DESC(sRAMBases, "Array of base of bus address of SRAM,INTERNAL, EXTERNAL0, EXTERNAL1..., ljvINVALID_PHYSICAL_ADDRESS means no bus address");
#endif

/* Per-core SRAM size. */
static uint sRAMSizes[ljvSRAM_INTER_COUNT * ljdGLOBAL_CORE_COUNT] = {
    [0 ... ljvSRAM_INTER_COUNT * ljdGLOBAL_CORE_COUNT - 1] = 0
};
module_param_array(sRAMSizes, uint, NULL, 0644);
MODULE_PARM_DESC(sRAMSizes, "Array of size of per-core SRAMs, 0 means no SRAM");

/* Shared SRAM physical address bases. */
static ljtPHYS_ADDR_T extSRAMBases[ljvSRAM_EXT_COUNT] = {
    [0 ... ljvSRAM_EXT_COUNT - 1] = ljvINVALID_PHYSICAL_ADDRESS
};
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
module_param_array(extSRAMBases, ullong, NULL, 0644);
MODULE_PARM_DESC(extSRAMBases, "Shared SRAM physical address bases.");
#endif

/* Shared SRAM sizes. */
static uint extSRAMSizes[ljvSRAM_EXT_COUNT] = {[0 ... ljvSRAM_EXT_COUNT - 1] = 0};
module_param_array(extSRAMSizes, uint, NULL, 0644);
MODULE_PARM_DESC(extSRAMSizes, "Shared SRAM sizes.");

static uint sRAMRequested = 1;
module_param(sRAMRequested, uint, 0644);
MODULE_PARM_DESC(sRAMRequested, "Default 1 means AXI-SRAM is already reserved for GPU, 0 means GPU driver need request the memory region.");

/******************************************************************************
 *                             mmu related                                    *
 ******************************************************************************/

static int mmu = 1;
module_param(mmu, int, 0644);
MODULE_PARM_DESC(mmu, "Disable MMU if set it to 0, enabled by default");

static uint flatMapping = 1;
module_param(flatMapping, int, 0644);
MODULE_PARM_DESC(flatMapping, "Flat mapping for reserve memory, enabled by default");

static uint mmuPageTablePool = 1;
module_param(mmuPageTablePool, uint, 0644);
MODULE_PARM_DESC(mmuPageTablePool, "Default 1 means alloc mmu page table in virtual memory(external if PCIE), 0 means auto select memory pool.");

static uint mmuCmdPool = 1;
module_param(mmuCmdPool, uint, 0644);
MODULE_PARM_DESC(mmuCmdPool, "Default 1 means auto select memory pool to allocate mmu initial command, specific the pool type with ljvPOOL_XXXX");

static uint mmuDynamicMap = 1;
module_param(mmuDynamicMap, uint, 0644);
MODULE_PARM_DESC(mmuDynamicMap, "Default 1 means enable mmu dynamic mapping in virsual memory, 0 means disable dynnamic mapping.");

static uint sharedPageTable = 1;
module_param(sharedPageTable, uint, 0644);
MODULE_PARM_DESC(sharedPageTable, "Default 1 means all the cores in one HW device share one MMU page table, 0 means per core page table.");

/*
 * If hardware do not support switch mmu, driver will set back processPageTable to 0.
 * So if want to enable processPageTable with switch mmu by software, need to modify code in kernel construct.
 */
static uint processPageTable = 1;
module_param(processPageTable, uint, 0644);
MODULE_PARM_DESC(processPageTable, "Default 0 means the page table is shared by all the cores in one device or per-core. 1 means the page table is per-process.");

static ulong baseAddress = 0;
module_param(baseAddress, ulong, 0644);
MODULE_PARM_DESC(baseAddress, "The pre-flatmapping CPU view base address in MMU page table. It's shared for all core");

static ulong physSize = 0;
module_param(physSize, ulong, 0644);
MODULE_PARM_DESC(physSize, "The pre-flatmapping size in MMU page table. If set it to 0, driver will skip all the pre-flatmapping related work");

/******************************************************************************
 *                            PCIE related                                    *
 ******************************************************************************/

#if USE_LINUX_PCIE
static int bar = 1;
module_param(bar, int, 0644);
MODULE_PARM_DESC(bar, "PCIE Bar index of LJ core");

static int bar2D = 1;
module_param(bar2D, int, 0644);
MODULE_PARM_DESC(bar2D, "PCIE Bar index of LJ 2D core");

static int barVG = -1;
module_param(barVG, int, 0644);
MODULE_PARM_DESC(barVG, "PCIE Bar index of LJ VG core");

static int bars[ljdGLOBAL_CORE_COUNT] = {[0 ... ljdGLOBAL_CORE_COUNT - 1] = -1};
module_param_array(bars, int, NULL, 0644);
MODULE_PARM_DESC(bars, "Array of bar index of PCIE platform for multi-GPU");

static int bar2Ds[ljdCORE_2D_COUNT] = {[0 ... ljdCORE_2D_COUNT - 1] = -1};
module_param_array(bar2Ds, int, NULL, 0644);
MODULE_PARM_DESC(bar2Ds, "Array of 2D bar index of PCIE platform for multi-GPU");

static uint regOffsets[ljdGLOBAL_CORE_COUNT] = {
    [0 ... ljdGLOBAL_CORE_COUNT - 1] = 0
};
module_param_array(regOffsets, uint, NULL, 0644);
MODULE_PARM_DESC(regOffsets, "Array of register offsets in corresponding BAR space");

static uint reg2DOffsets[ljdGLOBAL_CORE_COUNT] = {
    [0 ... ljdGLOBAL_CORE_COUNT - 1] = 0
};
module_param_array(reg2DOffsets, uint, NULL, 0644);
MODULE_PARM_DESC(reg2DOffsets, "Array of register 2D offsets in corresponding BAR space");

static uint regVGOffset = 0;
module_param(regVGOffset, int, 0644);
MODULE_PARM_DESC(regVGOffset, "register VG offset.");

static int sRAMBars[ljvSRAM_EXT_COUNT] = {[0 ... ljvSRAM_EXT_COUNT - 1] = -1};
module_param_array(sRAMBars, int, NULL, 0644);
MODULE_PARM_DESC(sRAMBars, "Array of SRAM bar index of shared external SRAMs.");

static int sRAMOffsets[ljvSRAM_EXT_COUNT] = {[0 ... ljvSRAM_EXT_COUNT - 1] = -1};
module_param_array(sRAMOffsets, int, NULL, 0644);
MODULE_PARM_DESC(sRAMOffsets, "Array of SRAM offset inside bar of shared external SRAMs.");

#endif

/******************************************************************************
 *                               others                                       *
 ******************************************************************************/

static uint major = 199;
module_param(major, uint, 0644);
MODULE_PARM_DESC(major, "major device number for LJ device");

static uint type = 0;
module_param(type, uint, 0664);
MODULE_PARM_DESC(type, "0 - Char Driver (Default), 1 - Misc Driver");

static uint vGPUType;
module_param(vGPUType, uint, 0664);
MODULE_PARM_DESC(type, "0 - vGPU is disabled(Default), 1 - MDEV vGPU, 2 - SRIOV vGPU");

static uint vGPUId;
module_param(vGPUId, uint, 0664);
MODULE_PARM_DESC(type, "The vGPU id will be used set for debug register when the vGPU is enabled");

#endif /* _LJ_HAL_KERNEL_PARAMETER_H_ */
