/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef _lj_hal_kernel_platform_h_
#define _lj_hal_kernel_platform_h_
#include <linux/mm.h>
#include <linux/platform_device.h>
#if USE_LINUX_PCIE
# include <linux/pci.h>
#endif
#include "lj_hal_kernel_platform_edma.h"
//#include "lj_hal_kernel_platform_edma_dev.h"
#include "lj_hal_kernel.h"

#if USE_LINUX_PCIE
# define ljdMAX_PCIE_BAR 6
#endif

typedef struct _ljsMODULE_PARAMETERS {
    ljtUINT                 chipIDs[ljvCORE_COUNT];
    ljtINT                  irqs[ljdGLOBAL_CORE_COUNT];
    ljtPHYS_ADDR_T          registerBases[ljdGLOBAL_CORE_COUNT];
    ljtSIZE_T               registerSizes[ljdGLOBAL_CORE_COUNT];
    ljtPOINTER              registerBasesMapped[ljdGLOBAL_CORE_COUNT];

    ljtINT                  irq2Ds[ljdGLOBAL_2D_COUNT];
    ljtPHYS_ADDR_T          register2DBases[ljdGLOBAL_2D_COUNT];
    ljtSIZE_T               register2DSizes[ljdGLOBAL_2D_COUNT];
    ljtPOINTER              register2DBasesMapped[ljdGLOBAL_2D_COUNT];

    ljtINT                  irqVG;
    ljtPHYS_ADDR_T          registerVGBase;
    ljtSIZE_T               registerVGSize;
    ljtPOINTER              registerVGBaseMapped;

    /* Contiguous memory pool. */
    ljtPHYS_ADDR_T          contiguousBase;
    ljtSIZE_T               contiguousSize;
    ljtPHYS_ADDR_T          contiguousBases[ljdSYSTEM_RESERVE_COUNT];
    ljtSIZE_T               contiguousSizes[ljdSYSTEM_RESERVE_COUNT];
    ljtBOOL                 contiguousRequested;

    /* External memory pool. */
    ljtPHYS_ADDR_T          externalBase[ljdDEVICE_COUNT];
    ljtSIZE_T               externalSize[ljdDEVICE_COUNT];

    /* External memory pool. */
    ljtPHYS_ADDR_T          exclusiveBase[ljdDEVICE_COUNT];
    ljtSIZE_T               exclusiveSize[ljdDEVICE_COUNT];

    /* Per-core SRAM. */
    ljtPHYS_ADDR_T          sRAMBases[ljdGLOBAL_CORE_COUNT][ljvSRAM_INTER_COUNT];
    ljtUINT32               sRAMSizes[ljdGLOBAL_CORE_COUNT][ljvSRAM_INTER_COUNT];

    /* Shared SRAM. */
    ljtPHYS_ADDR_T          extSRAMBases[ljvSRAM_EXT_COUNT];
    ljtUINT32               extSRAMSizes[ljvSRAM_EXT_COUNT];
#if USE_LINUX_PCIE
    ljtINT                  bars[ljdGLOBAL_CORE_COUNT];
    ljtUINT32               regOffsets[ljdGLOBAL_CORE_COUNT];
    ljtINT                  bar2Ds[ljdGLOBAL_2D_COUNT];
    ljtUINT32               reg2DOffsets[ljdGLOBAL_2D_COUNT];
    ljtINT                  barVG;
    ljtUINT32               regVGOffset;
    ljtINT32                sRAMBars[ljvSRAM_EXT_COUNT];
    ljtINT32                sRAMOffsets[ljvSRAM_EXT_COUNT];
#endif

    ljtUINT                 platformIDs[ljdDEVICE_COUNT];
    ljtUINT                 hwDevCounts[ljdPLATFORM_COUNT];
    ljtUINT                 devCoreCounts[ljdDEVICE_COUNT];
    ljtUINT                 dev2DCoreCounts[ljdDEVICE_COUNT];
    ljtUINT                 devMemIDs[ljdDEVICE_COUNT];
    ljtUINT                 devSysMemIDs[ljdDEVICE_COUNT];
    ljtUINT                 devSRAMIDs[ljdDEVICE_COUNT];
    ljtUINT                 devCount;

    ljtBOOL                 sRAMRequested;
    ljtUINT32               sRAMLoopMode;

    ljtPHYS_ADDR_T          baseAddress;
    ljtSIZE_T               physSize;
    ljtSIZE_T               bankSize;

    ljtUINT                 recovery;
    ljtINT                  powerManagement;

    ljtINT                  enableMmu;
    ljtINT                  fastClear;
    ljeCOMPRESSION_OPTION   compression;
    ljtUINT                 gpu3DMinClock;
    ljtUINT                 userClusterMasks[ljdMAX_MAJOR_CORE_COUNT];
    ljtUINT                 smallBatch;

    /* Debug or other information. */
    ljtUINT                 stuckDump;
    ljtUINT                 softReset;
    /* device type, 0 for char device, 1 for misc device. */
    ljtUINT                 deviceType;
    ljtUINT                 showArgs;

    /* mmu page table pool, 0 mean auto, 1 means virtual. */
    ljtUINT                 mmuPageTablePool;

    /* mmu page table pool, 1 mean auto, others means specific pool. */
    ljtUINT                 mmuCmdPool;

    ljtUINT                 mmuDynamicMap;

    /* Shared MMU page table or per core page table. */
    ljtUINT                 sharedPageTable;

    /* If the page table is per process. */
    ljtUINT                 processPageTable;

    /* Flat mapping for reserve memory. */
    ljtUINT                 flatMapping;

    ljtUINT                 allMapInOne;

    ljtUINT                 isrPoll;

    /* APB register offset to the register base address. */
    ljtUINT64               registerAPB;

    /* Enabled NN clusters number */
    ljtUINT                 enableNN;

    /* Struct device array. */
    struct device           *devices[ljdDEVICE_COUNT];

    /* vGPU type and id. 0 means no support vGPU, 1 means mdev, 2 means SRIOV.*/
    ljtUINT32               vGPUType;
    ljtUINT32               vGPUId;
} ljsMODULE_PARAMETERS;

typedef struct _ljsDMA_TRANS_INFO {
    ljtPOINTER src_node;
    ljtPOINTER dst_node;
    ljtSIZE_T offset;
    ljtSIZE_T bytes;
    ljtUINT32 reason;
} ljsDMA_TRANS_INFO;

typedef struct _ljsPLATFORM ljsPLATFORM;

typedef struct _ljsPLATFORM_OPERATIONS {
    /**********************************************************************
     **
     **  adjustParam
     **
     **  Override content of arguments, if a argument is not changed here,
     **  it will keep as default value or value set by insmod command line.
     */
    ljeSTATUS
    (*adjustParam)(ljsPLATFORM *Platform, ljsMODULE_PARAMETERS *Args);

    /**********************************************************************
     **
     **  getPower
     **
     **  Prepare power and clock operation.
     */
    ljeSTATUS
    (*getPower)(ljsPLATFORM *Platform);

    /**********************************************************************
     **
     **  putPower
     **
     **  Finish power and clock operation.
     */
    ljeSTATUS
    (*putPower)(ljsPLATFORM *Platform);

    /**********************************************************************
     **
     **  setPower
     **
     **  Set power state of specified GPU.
     **
     **  INPUT:
     **
     **      ljeCORE GPU
     **          GPU neeed to config.
     **
     **      ljeBOOL Enable
     **          Enable or disable power.
     */
    ljeSTATUS
    (*setPower)(ljsPLATFORM *Platform, ljtUINT32 DevIndex, ljeCORE GPU, ljtBOOL Enable);

    /**********************************************************************
     **
     **  setClock
     **
     **  Set clock state of specified GPU.
     **
     **  INPUT:
     **
     **      ljeCORE GPU
     **          GPU neeed to config.
     **
     **      ljeBOOL Enable
     **          Enable or disable clock.
     */
    ljeSTATUS
    (*setClock)(ljsPLATFORM *Platform, ljtUINT32 DevIndex, ljeCORE GPU, ljtBOOL Enable);

    /**********************************************************************
     **
     **  reset
     **
     **  Reset GPU outside.
     **
     **  INPUT:
     **
     **      ljeCORE GPU
     **          GPU neeed to reset.
     */
    ljeSTATUS
    (*reset)(ljsPLATFORM *Platform, ljtUINT32 DevIndex, ljeCORE GPU);

    /**********************************************************************
     **
     **  getGPUPhysical
     **
     **  Convert CPU physical address to GPU physical address if they are
     **  different.
     */
    ljeSTATUS
    (*getGPUPhysical)(ljsPLATFORM *Platform, ljtPHYS_ADDR_T CPUPhysical, ljtPHYS_ADDR_T *GPUPhysical);

    /**********************************************************************
     **
     **  getCPUPhysical
     **
     **  Convert GPU physical address to CPU physical address if they are
     **  different.
     */
    ljeSTATUS
    (*getCPUPhysical)(ljsPLATFORM *Platform, ljtPHYS_ADDR_T GPUPhysical, ljtPHYS_ADDR_T *CPUPhysical);

    /**********************************************************************
     **
     **  adjustProt
     **
     **  Override Prot flag when mapping paged memory to userspace.
     */
    ljeSTATUS
    (*adjustProt)(struct vm_area_struct *vma);

    /**********************************************************************
     **
     **  shrinkMemory
     **
     **  Do something to collect memory, eg, act as oom killer.
     */
    ljeSTATUS
    (*shrinkMemory)(ljsPLATFORM *Platform);

    /**********************************************************************
     **
     ** getPolicyID
     **
     ** Get policyID for a specified surface type.
     */
    ljeSTATUS
    (*getPolicyID)(ljsPLATFORM *Platform, ljeVIDMEM_TYPE Type, ljtUINT32_PTR PolicyID, ljtUINT32_PTR AXIConfig);

#if 0
    /*******************************************************************************
    **
    ** syncMemory
    **
    */
    ljeSTATUS
    (*syncMemory)(
        IN ljtPOINTER Object,
        IN ljtPOINTER DstNode,
        IN ljtPOINTER SrcNode,
        IN ljtUINT32  Reason,
        IN ljtUINT64  SyncLen,
        IN ljtUINT64  Offset
    );
#endif
    /**********************************************************************
     **
     **  _ExternalCacheOperation
     **
     **  External device cache operation, if support. If the core has any
     **  additional caches they must be invalidated after this function
     **  returns.
     **  If the core does not have any additional caches the
     **  externalCacheOperation in the platform->ops should remain NULL.
     **  The function may be called by multiple thread, so need to add
     **  mutex in the callback when there is shared resource.
     **
     **  INPUT:
     **      ljeCACHEOPERATION Operation
     **          Cache Operation: ljvCACHE_FLUSH, ljvCACHE_CLEAN or
     **          ljvCACHE_INVALIDATE.
     **
     **  OUTPUT:
     **
     **      Nothing.
     */
    void
    (*externalCacheOperation)(ljsPLATFORM *Platform, ljeCACHEOPERATION Operation);

    /**********************************************************************
     **
     ** dmaCopy
     **
     ** copy videm memory by dma
     */
    ljeSTATUS
    (*dmaCopy)(ljtPOINTER Object, IN ljsDMA_TRANS_INFO *Info);

    /**********************************************************************
     **
     ** dmaInit
     **
     ** Init platform specified dma
     */
    ljeSTATUS
    (*dmaInit)(ljsPLATFORM *Platform);

    /**********************************************************************
     **
     ** dmaExit
     **
     ** deInit platform dma
     */
    ljeSTATUS
    (*dmaExit)(void/*ljsPLATFORM *Platform*/);

#if ljdENABLE_MP_SWITCH
    /**********************************************************************
     ** switchCoreCount
     **
     ** Switch the core count according to specific conditions.
     **
     */
    ljeSTATUS
    (*switchCoreCount)(ljsPLATFORM *Platform, ljtUINT32 *Count);
#endif
    /*******************************************************************************
    **
    ** dmaOperation
    **
    */
    ljeSTATUS
    (*dmaOperation)(
        IN ljtPOINTER Object,
        IN ljtPOINTER Node_Src,
        IN ljtPOINTER Node_Dst,
        IN ljtUINT32  Dir,
        IN ljtUINT64  TransLen
    );
} ljsPLATFORM_OPERATIONS;

struct _ljsPLATFORM {
    struct platform_device *device;
    struct platform_driver *driver;

    const char             *name;
    ljsPLATFORM_OPERATIONS *ops;

    ljkDEVICE               dev;

    /* PLATFORM specific flags */
    ljtUINT32               flagBits;

    /* Real-time core count. */
    ljtUINT32               coreCount;

    void                   *priv;
    /* Module special parameters */
    ljsMODULE_PARAMETERS    params;
};

typedef void __iomem    *iomem_t;

#define write_mreg32(_base, _reg, _val) writel((_val), (volatile iomem_t)(_base) + (_reg))
#define read_mreg32(_base, _reg)        readl((volatile iomem_t)(_base) + (_reg))
#define write_mreg8(_base, _reg, _val)  writeb((_val), (volatile iomem_t)(_base) + (_reg))
#define read_mreg8(_base, _reg)         readb((volatile iomem_t)(_base) + (_reg))

typedef struct _ljsBARINFO
{
    ljtPHYS_ADDR_T base;
    ljtSIZE_T size;
    ljtPOINTER logical;
    ljtBOOL available;
    ljtUINT64 reg_max_offset;
    ljtUINT32 reg_size;
}
ljsBARINFO, *ljkBARINFO;

struct _ljsPCIEInfo
{
    ljsBARINFO bar[ljdMAX_PCIE_BAR];
    struct pci_dev *pdev;
    unsigned long flags;
    ljtPHYS_ADDR_T gpu_vram_start;  /* gpu vram physical start address */
    ljtSIZE_T gpu_vram_size;        /* gpu vram size, including the external and exclusive */
};

struct _ljsPLATFORM_PCIE
{
    struct _ljsPLATFORM base;
    struct _ljsPCIEInfo pcie_info[ljdDEVICE_COUNT];
    edma_device dma[ljdDEVICE_COUNT];
    unsigned int device_number;
};

int ljkPLATFORM_Init(struct platform_driver *pdrv, ljsPLATFORM **platform);
int ljkPLATFORM_Terminate(ljsPLATFORM *platform);

#endif
