/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_platform_edma_h_
#define __lj_hal_kernel_platform_edma_h_
#include <linux/kthread.h>
#include <linux/interrupt.h>
#include "lj_hal.h"
#include "lj_hal_kernel_dma_engine.h"

/*old dma is in green fpga board with 1 channel; new dma is in z1 with channal 8+8*/
#define DMA_NEW             1
#define EDMA_DEBUG          0
#define FPGA_HAPS           0

#if !DMA_NEW
#include <linux/pci.h>
#endif
#if DMA_NEW
#if !FPGA_HAPS
#define DMA_BASE_ADDR                                       0x1000
#else
#define DMA_BASE_ADDR                                       0x0
#endif
#define DMA_WR_BASE                                         0x1200
#define DMA_RD_BASE                                         0x1300
#else
#define DMA_BASE_ADDR                                       0x700//0x1000
#define DMA_WR_BASE                                         0x1200
#define DMA_RD_BASE                                         0x1300
#endif

#if DMA_NEW//DMA_NEW
#define DMA_CTRL_DATA_ARB_PRIOR_REG                         (DMA_BASE_ADDR+0x00L)
#define DMA_CHANNEL_CTRL_REG                                (DMA_BASE_ADDR+0x08L)
#define DMA_WRITE_ENGINE_EN_REG                             (DMA_BASE_ADDR+0x0cL)
#define DMA_WRITE_DOORBELL_REG                              (DMA_BASE_ADDR+0x10L)
#define	DMA_WRITE_ENGINE_PF_REG								(DMA_BASE_ADDR+0x14L)
#define DMA_WRITE_CHANNEL_ARB_WEIGHT_LOW_REG                (DMA_BASE_ADDR+0x18L)
#define DMA_WRITE_CHANNEL_ARB_WEIGHT_HIGH_REG               (DMA_BASE_ADDR+0x1cL)
#define DMA_READ_ENGINE_EN_REG                              (DMA_BASE_ADDR+0x2cL)
#define DMA_READ_DOORBELL_REG                               (DMA_BASE_ADDR+0x30L)
#define DMA_READ_ENGINE_PF_REG                              (DMA_BASE_ADDR+0x34L)
#define DMA_READ_CHANNEL_ARB_WEIGHT_LOW_REG                 (DMA_BASE_ADDR+0x38L)
#define DMA_READ_CHANNEL_ARB_WEIGHT_HIGH_REG                (DMA_BASE_ADDR+0x3cL)
#define DMA_WRITE_INT_STATUS_REG                            (DMA_BASE_ADDR+0x4cL)
#define DMA_WRITE_INT_MASK_REG                              (DMA_BASE_ADDR+0x54L)
#define DMA_WRITE_INT_CLEAR_REG                             (DMA_BASE_ADDR+0x58L)
#define DMA_WRITE_ERR_STATUS_REG                            (DMA_BASE_ADDR+0x5cL)
#define DMA_WRITE_DONE_IMWR_LOW_REG                         (DMA_BASE_ADDR+0x60L)
#define DMA_WRITE_DONE_IMWR_HIGH_REG                        (DMA_BASE_ADDR+0x64L)
#define DMA_WRITE_ABORT_IMWR_LOW_REG                        (DMA_BASE_ADDR+0x68L)
#define DMA_WRITE_ABORT_IMWR_HIGH_REG                       (DMA_BASE_ADDR+0x6cL)
#define DMA_WRITE_CH01_IMWR_DATA_REG                        (DMA_BASE_ADDR+0x70L)
#define DMA_WRITE_CH23_IMWR_DATA_REG                        (DMA_BASE_ADDR+0x74L)
#define DMA_WRITE_CH45_IMWR_DATA_REG                        (DMA_BASE_ADDR+0x78L)
#define DMA_WRITE_CH67_IMWR_DATA_REG                        (DMA_BASE_ADDR+0x7cL)
#define DMA_WRITE_LINKED_LIST_ERR_EN_REG                    (DMA_BASE_ADDR+0x90L)
#define DMA_READ_INT_STATUS_REG                             (DMA_BASE_ADDR+0xa0L)
#define DMA_READ_INT_MASK_REG                               (DMA_BASE_ADDR+0xa8L)
#define DMA_READ_INT_CLEAR_REG                              (DMA_BASE_ADDR+0xacL)
#define DMA_READ_ERR_STATUS_LOW_REG                         (DMA_BASE_ADDR+0xb4L)
#define DMA_READ_ERR_STATUS_HIGH_REG                        (DMA_BASE_ADDR+0xb8L)
#define DMA_READ_LINKED_LIST_ERR_EN_REG                     (DMA_BASE_ADDR+0xc4L)
#define DMA_READ_DONE_IMWR_LOW_REG                          (DMA_BASE_ADDR+0xccL)
#define DMA_READ_DONE_IMWR_HIGH_REG                         (DMA_BASE_ADDR+0xd0L)
#define DMA_READ_ABORT_IMWR_LOW_REG                         (DMA_BASE_ADDR+0xd4L)
#define DMA_READ_ABORT_IMWR_HIGH_REG                        (DMA_BASE_ADDR+0xd8L)
#define DMA_READ_CH01_IMWR_DATA_REG                         (DMA_BASE_ADDR+0xdcL)
#define DMA_READ_CH23_IMWR_DATA_REG                         (DMA_BASE_ADDR+0xe0L)
#define DMA_READ_CH45_IMWR_DATA_REG                         (DMA_BASE_ADDR+0xe4L)
#define DMA_READ_CH67_IMWR_DATA_REG                         (DMA_BASE_ADDR+0xe8L)
#define DMA_WRITE_ENGINE_CHGROUP_REG                        (DMA_BASE_ADDR+0x100L)
#define DMA_READ_ENGINE_CHGROUP_REG                        	(DMA_BASE_ADDR+0x104L)
#define DMA_WRITE_ENGINE_HSHAKE_CNT_LOW_REG                 (DMA_BASE_ADDR+0x108L)
#define DMA_WRITE_ENGINE_HSHAKE_CNT_HIGH_REG                (DMA_BASE_ADDR+0x10cL)
#define DMA_READ_ENGINE_HSHAKE_CNT_LOW_REG                 	(DMA_BASE_ADDR+0x118L)
#define DMA_READ_ENGINE_HSHAKE_CNT_HIGH_REG                	(DMA_BASE_ADDR+0x11cL)
#define DMA_WRITE_CH0_PWR_EN_REG							(DMA_BASE_ADDR+0x128L)
#define DMA_WRITE_CH1_PWR_EN_REG							(DMA_BASE_ADDR+0x12cL)
#define DMA_WRITE_CH2_PWR_EN_REG							(DMA_BASE_ADDR+0x130L)
#define DMA_WRITE_CH3_PWR_EN_REG							(DMA_BASE_ADDR+0x134L)
#define DMA_WRITE_CH4_PWR_EN_REG							(DMA_BASE_ADDR+0x138L)
#define DMA_WRITE_CH5_PWR_EN_REG							(DMA_BASE_ADDR+0x13cL)
#define DMA_WRITE_CH6_PWR_EN_REG							(DMA_BASE_ADDR+0x140L)
#define DMA_WRITE_CH7_PWR_EN_REG							(DMA_BASE_ADDR+0x144L)
#define DMA_READ_CH0_PWR_EN_REG								(DMA_BASE_ADDR+0x168L)
#define DMA_READ_CH1_PWR_EN_REG								(DMA_BASE_ADDR+0x16cL)
#define DMA_READ_CH2_PWR_EN_REG								(DMA_BASE_ADDR+0x170L)
#define DMA_READ_CH3_PWR_EN_REG								(DMA_BASE_ADDR+0x174L)
#define DMA_READ_CH4_PWR_EN_REG								(DMA_BASE_ADDR+0x178L)
#define DMA_READ_CH5_PWR_EN_REG								(DMA_BASE_ADDR+0x17cL)
#define DMA_READ_CH6_PWR_EN_REG								(DMA_BASE_ADDR+0x180L)
#define DMA_READ_CH7_PWR_EN_REG								(DMA_BASE_ADDR+0x184L)
//
#define DMA_CH_CONTROL1_WRCH_REG(i)                         (DMA_BASE_ADDR+i*0x200+0x200+0x00L)
#define DMA_CH_CONTROL2_WRCH_REG(i)                         (DMA_BASE_ADDR+i*0x200+0x200+0x04L)
#define DMA_TRANSFER_SIZE_WRCH_REG(i)                       (DMA_BASE_ADDR+i*0x200+0x200+0x08L)
#define DMA_SAR_LOW_WRCH_REG(i)                             (DMA_BASE_ADDR+i*0x200+0x200+0x0cL)
#define DMA_SAR_HIGH_WRCH_REG(i)                            (DMA_BASE_ADDR+i*0x200+0x200+0x10L)
#define DMA_DAR_LOW_WRCH_REG(i)                             (DMA_BASE_ADDR+i*0x200+0x200+0x14L)
#define DMA_DAR_HIGH_WRCH_REG(i)                            (DMA_BASE_ADDR+i*0x200+0x200+0x18L)
#define DMA_LLP_LOW_WRCH_REG(i)                             (DMA_BASE_ADDR+i*0x200+0x200+0x1cL)
#define DMA_LLP_HIGH_WRCH_REG(i)                            (DMA_BASE_ADDR+i*0x200+0x200+0x20L)
#define DMA_CH_CONTROL1_RDCH_REG(i)                         (DMA_BASE_ADDR+i*0x200+0x300+0x00L)
#define DMA_CH_CONTROL2_RDCH_REG(i)                         (DMA_BASE_ADDR+i*0x200+0x300+0x04L)
#define DMA_TRANSFER_SIZE_RDCH_REG(i)                       (DMA_BASE_ADDR+i*0x200+0x300+0x08L)
#define DMA_SAR_LOW_RDCH_REG(i)                             (DMA_BASE_ADDR+i*0x200+0x300+0x0c)
#define DMA_SAR_HIGH_RDCH_REG(i)                            (DMA_BASE_ADDR+i*0x200+0x300+0x10L)
#define DMA_DAR_LOW_RDCH_REG(i)                             (DMA_BASE_ADDR+i*0x200+0x300+0x14L)
#define DMA_DAR_HIGH_RDCH_REG(i)                            (DMA_BASE_ADDR+i*0x200+0x300+0x18L)
#define DMA_LLP_LOW_RDCH_REG(i)                             (DMA_BASE_ADDR+i*0x200+0x300+0x1cL)
#define DMA_LLP_HIGH_RDCH_REG(i)                            (DMA_BASE_ADDR+i*0x200+0x300+0x20L)

#else/*!DMA_NEW*/

#define DMA_WRITE_ENGINE_EN_REG                             (DMA_BASE_ADDR+0x27C)
#define DMA_WRITE_DOORBELL_REG                              (DMA_BASE_ADDR+0x280)
#define DMA_READ_ENGINE_EN_REG                              (DMA_BASE_ADDR+0x29c)
#define DMA_READ_DOORBELL_REG                               (DMA_BASE_ADDR+0x2a0)
#define DMA_WRITE_INT_STATUS_REG                            (DMA_BASE_ADDR+0x2bc)
#define DMA_WRITE_INT_MASK_REG                              (DMA_BASE_ADDR+0x2c4)
#define DMA_WRITE_INT_CLEAR_REG                             (DMA_BASE_ADDR+0x2c8)
#define DMA_WRITE_ERR_STATUS_REG                            (DMA_BASE_ADDR+0x2cc)
#define DMA_WRITE_DONE_IMWR_LOW_REG                         (DMA_BASE_ADDR+0x2d0)
#define DMA_WRITE_DONE_IMWR_HIGH_REG                        (DMA_BASE_ADDR+0x2d4)
#define DMA_WRITE_ABORT_IMWR_LOW_REG                        (DMA_BASE_ADDR+0x2d8)
#define DMA_WRITE_ABORT_IMWR_HIGH_REG                       (DMA_BASE_ADDR+0x2dc)
#define DMA_WRITE_CH0_IMWR_DATA_REG                        	(DMA_BASE_ADDR+0x2e0)
#define DMA_WRITE_LINKED_LIST_ERR_EN_REG                    (DMA_BASE_ADDR+0x300)
#define DMA_READ_INT_STATUS_REG                             (DMA_BASE_ADDR+0x310)
#define DMA_READ_INT_MASK_REG                               (DMA_BASE_ADDR+0x318)
#define DMA_READ_INT_CLEAR_REG                              (DMA_BASE_ADDR+0x31c)
#define DMA_READ_ERR_STATUS_LOW_REG                         (DMA_BASE_ADDR+0x324)
#define DMA_READ_ERR_STATUS_HIGH_REG                        (DMA_BASE_ADDR+0x328)
#define DMA_READ_LINKED_LIST_ERR_EN_REG                     (DMA_BASE_ADDR+0x334)
#define DMA_READ_DONE_IMWR_LOW_REG                          (DMA_BASE_ADDR+0x33c)
#define DMA_READ_DONE_IMWR_HIGH_REG                         (DMA_BASE_ADDR+0x340)
#define DMA_READ_ABORT_IMWR_LOW_REG                         (DMA_BASE_ADDR+0x344)
#define DMA_READ_ABORT_IMWR_HIGH_REG                        (DMA_BASE_ADDR+0x348)


#define DMA_CHANNEL_CTX_INDEX_REG                         	(DMA_BASE_ADDR+0x36c)
#define DMA_CHANNEL_CTRL_REG                        		(DMA_BASE_ADDR+0x370)
#define DMA_TRANSFER_SIZE_REG                      			(DMA_BASE_ADDR+0x378)
#define DMA_SAR_LOW_REG                             		(DMA_BASE_ADDR+0x37c)
#define DMA_SAR_HIGH_REG                            		(DMA_BASE_ADDR+0x380)
#define DMA_DAR_LOW_REG                             		(DMA_BASE_ADDR+0x384)
#define DMA_DAR_HIGH_REG                            		(DMA_BASE_ADDR+0x388)
#define DMA_LLP_LOW_REG                             		(DMA_BASE_ADDR+0x38c)
#define DMA_LLP_HIGH_REG                            		(DMA_BASE_ADDR+0x390)
#endif
/*==========================================================*/
/*   PCIe embed DMA bit define, struct and enum typedef     */
/*==========================================================*/

//DMA_WRITE_INT_STATUS_REG,DMA_READ_INT_STATUS_REG
#define DMA_DONE_INT(x)                 (0x1L << x)
#define DMA_ABORT_INT(x)                (0x1L << (16 + x))
//DMA_WRITE_LINKED_LIST_ERR_EN_REG,DMA_READ_LINKED_LIST_ERR_EN_REG
#define DMA_LLRAIE(x)                   (0x1L << x)
#define DMA_LLLAIE(x)                   (0x1L << (16 + x))
//DMA_WRITE_ENGINE_EN_REG
#define DMA_WRITE_ENGINE_ENABLE         (0x1L)
//DMA_READ_ENGINE_EN_REG
#define DMA_READ_ENGINE_ENABLE          (0x1L)
//DMA_WRITE_DOORBELL_REG,DMA_READ_DOORBELL_REG
#define DMA_TRANSFER_STOP               (0x1L << 31)
//DMA_VIEWPORT_SEL_REG
#define DMA_CHANNEL_DIR_WRITE           (0x0L << 31)
#define DMA_CHANNEL_DIR_READ            (0x1L << 31)
//*******************DMA_CH_CONTROL1_REG************************//
//Link and intrrupt bit define
#define DMA_CB                          (0x1L << 0)
#define DMA_TCB                         (0x1L << 1)
#define DMA_LLP                         (0x1L << 2)
#define DMA_LIE                         (0x1L << 3)
#define DMA_RIE                         (0x1L << 4)
#define DMA_CCS                         (0x1L << 8)
#define DMA_LLE                         (0x1L << 9)

//dma TLP relative define
#define DMA_FUNC_NUM(x)                 (x << 12)
#if DMA_NEW
#define DMA_NS_DST                      (0x1L << 23)
#define DMA_NS_SRC                      (0x1L << 24)
#else
#define DMA_NS	                        (0x1L << 24)
#endif
#define DMA_RO                          (0x1L << 25)
#define DMA_TD                          (0x1L << 26)
#define DMA_TC(x)                       (x << 27)
#define DMA_AT(x)                       (x << 30)
//channel status CS
#define DMA_CS_RUNNING                  (0x1L << 5)
#define DMA_CS_HALTED                   (0x2L << 5)
#define DMA_CS_STOPPED                  (0x3L << 5)
//*******************DMA_CH_CONTROL2_REG************************//
//virtual funcyion relative define
#define DMA_VF_EN                       (0x1L << 0)
#define DMA_VF_NUM(x)                   (x << 1)
//dma TLP relative define
#define DMA_ST(x)                       (x << 17)
#define DMA_PH(x)                       (x << 29)
#define DMA_TH                          (0x1L << 31)

#define DMA_ENGINE_EN                   0x1
#define MASK_ALL_INTE                   0x00FF00FF
#define CLEAR_ALL_INTE                  0x00FF00FF
/*
 * edma link table address in slice0 ddr. There are 4 rc2ep channels and
 * 4 ep2rc channels,total 8 channels, every channel have EDMA_LT_SIZE space
 * to save link table.
 * sequence: rc2ep0,rc2ep1,rc2ep2,rc2ep3,ep2rc0,ep2rc1,ep2rc2,ep2rc3
 */
#define EDMA_LT_SIZE                    0x40000
#define LINKTABLE_OFFSET(base, c)       ((base) + (c)*EDMA_LT_SIZE)

#define QWORD_HI(value)                 ((value >> 32) & 0xFFFFFFFF)
#define QWORD_LO(value)                 (value & 0xFFFFFFFF)

typedef void __iomem *                  iomem_t;

typedef struct _edma_desc_ll
{
    uint32_t control;
    uint32_t size;
    uint32_t sar_low;
    uint32_t sar_high;
    uint32_t dar_low;
    uint32_t dar_high;
}
edma_desc_ll;

typedef struct _edma_desc_llp
{
    uint32_t control;
    uint32_t reserved;
    uint32_t llp_low;
    uint32_t llp_high;
}
edma_desc_llp;

/*
* edma_status
* @EDMA_IDLE: idle
* @EDMA_OK: init or config or satrt transfer OK
* @EDMA_DONE: transfer success
* @EDMA_ERROR: config or transfer error
* @EDMA_ABORT: transfer abort
* @EDMA_BUSY: it is running at current time
* @EDMA_TIMEOUT: transfer timeout
*/
typedef enum
{
    EDMA_IDLE                   =0x00,
    EDMA_OK,
    EDMA_DONE,
    EDMA_ERROR,
    EDMA_ABORT,
    EDMA_BUSY,
    EDMA_TIMEOUT,
}
edma_status;
/*
* edma_check_mode
* @EDMA_CHECK_MODE_POLLING: polling
* @EDMA_CHECK_MODE_INTERRUPT: interrupt
*/
typedef enum
{
    EDMA_CHECK_MODE_POLLING     =0x00,
    EDMA_CHECK_MODE_INTERRUPT   =0x01,
}
edma_check_mode;
/*
* edma_cpu_mode
* @EDMA_CPU_MODE_LOCAL: local interrupt
* @EDMA_CPU_MODE_REMOTE: remote interrupt
*/
typedef enum
{
    EDMA_CPU_MODE_LOCAL         = 0x00,
    EDMA_CPU_MODE_REMOTE        = 0x01,
}
edma_cpu_mode;
/*
* edma_transfer_mode
* @EDMA_TRANSFER_MODE_NORMAL: norml read and write mode
* @EDMA_TRANSFER_MODE_LINK: link list mode
*/
typedef enum
{
    EDMA_TRANSFER_MODE_NORMAL   = 0x00,
    EDMA_TRANSFER_MODE_LINK     = 0x01,
}
edma_transfer_mode;

typedef struct
{
#if DMA_NEW
	iomem_t reg_base;
#else
	struct pci_dev *reg_base;
#endif
    uint8_t channel;
    uint32_t interrupt_mask;
    uint8_t arbi_weight;
    uint16_t imwr_data;
    uint32_t done_imwr_addr_low;
    uint32_t done_imwr_addr_high;
    uint32_t abort_imwr_addr_low;
    uint32_t abort_imwr_addr_high;
    uint32_t link_err_en;
    uint32_t control1;
    uint32_t control2;
    uint32_t transfer_size;
    /* for no-link mode */
    uint32_t sar_low;
    uint32_t sar_high;
    uint32_t dar_low;
    uint32_t dar_high;

    /* for link mode */
    uint32_t llp_low;
    uint32_t llp_high;

    uint32_t timeout;
}
edma_hw_config;

typedef struct
{
#if DMA_NEW
	iomem_t reg_base;
#else
	struct pci_dev *reg_base;
#endif
    uint8_t channel;
    edma_cpu_mode cpu_mode;
    edma_check_mode check_mode;
    edma_transfer_mode transfer_mode;
    uint64_t src_addr;
    uint64_t dst_addr;
    size_t transfer_size;
    uint64_t lle;
    uint64_t msi_addr;
    uint32_t msi_data;
    uint32_t timeout;
}
edma_transfer_info;

typedef struct _edma_descriptor
{
    struct _ljsDMA_TX_DESCRIPTOR txd;
    edma_cpu_mode cpu_mode;
    edma_check_mode check_mode;
    edma_transfer_mode transfer_mode;
    struct list_head desc_node;

    bool is_done;
    uint32_t timeout;
    size_t total_len;
    uint32_t residue;

    uint64_t src_addr;
    uint64_t dst_addr;
    uint64_t lle;
}
edma_descriptor;

typedef struct _edma_chan
{
    ljsDMA_CHANNEL chan;
#if DMA_NEW
    iomem_t reg_base;
#else
	struct pci_dev *reg_base;
#endif
    spinlock_t lock;
    struct list_head active_list;
    struct list_head pending_list;
    /* record current desc which is in process */
    struct list_head *current_active;
    ljsDMA_SLAVE_CONFIG sconfig;
    uint32_t descs_allocated;
    ljeDMA_STATUS status;

    uint32_t bus_width;
    size_t total_len;
    uint32_t residue;
    ljeDMA_TX_DIRECTION direction;

    edma_descriptor *desc;
}
edma_channel;

typedef struct _edma_device
{
    ljsDMA_DEVICE slave_rdma;
    ljsDMA_DEVICE slave_wdma;

#if DMA_NEW
    iomem_t reg_base;
    iomem_t reg_base_2;
#else
    struct pci_dev *reg_base;
#endif
    struct tasklet_struct tasklet;
    spinlock_t irq_lock;
    edma_channel *rchan;
    uint32_t rchan_count;
    edma_channel *wchan;
    uint32_t wchan_count;
    struct pci_dev *pdev;
    int irq;
    uint64_t msi_addr;
    uint32_t msi_data;
    uint32_t rchan_in_use;
    uint32_t wchan_in_use;
    bool interrupt_enable;
    /* edma link table */
    iomem_t vlink_table_base;
    uint64_t plink_table_base;

    struct task_struct *isr_thread;
    bool kill_thread;

    bool hdma_flag;
    bool msi_flag;
    bool single_chan;
} edma_device;
#if DMA_NEW
static inline void edma_write(iomem_t base, unsigned int addr, unsigned int val)
{
#if EDMA_DEBUG
    //dump_stack();
    printk("edma_write: addr=0x%x; val=0x%x\n", addr, val);
#endif
    writel((val), ( volatile iomem_t)(((addr) + (u64)(base))));
}

static inline unsigned int edma_read(iomem_t base, unsigned int addr)
{
    unsigned int data_read = 0;
    data_read = readl(( volatile iomem_t)((addr)+(u64)(base)));
#if EDMA_DEBUG
    printk("edma_read after: addr=0x%x; data_read=0x%x\n", addr, data_read);
#endif
    return data_read;
}
#else
static inline void edma_write(struct pci_dev *pdev, unsigned int addr, unsigned int val)
{
    pci_write_config_dword(pdev, addr, val);
#if EDMA_DEBUG
    printk("edma_write: addr=0x%x; val=0x%x\n", addr, val);
#endif
}

static inline unsigned int edma_read(struct pci_dev *pdev, unsigned int addr)
{
	unsigned int data_read = 0;
	pci_read_config_dword(pdev, addr, &data_read);
#if EDMA_DEBUG
	printk("edma_read: addr=0x%x; data_read=0x%x\n", addr, data_read);
#endif
    return data_read;
}
#endif

int edma_device_probe(edma_device *edma_dev);
void edma_device_remove(edma_device *edma_dev);
edma_descriptor *edma_get_first_active(edma_channel *edma_chan);
edma_descriptor *edma_descriptor_alloc(edma_channel *edma_chan);
void edma_descriptor_free(edma_channel *edma_chan, edma_descriptor *edma_desc);
void edma_descriptor_complete(
    edma_channel *edma_chan,
    edma_descriptor *desc,
    bool invoke_callback
    );
void edma_free_descriptor_list(
    edma_channel *edma_chan,
    struct list_head *list
    );
edma_channel *to_edma_chan(ljsDMA_CHANNEL *chan);
edma_device *to_edma_dev(ljsDMA_DEVICE *ddma);
edma_descriptor *edma_find_desc(
    edma_channel *edma_chan,
    ljdDMA_TX_IDENTIFIER identifier
    );
uint32_t edma_get_residue(
    edma_channel *edma_chan,
    ljdDMA_TX_IDENTIFIER identifier
    );

#define edma_desc_entry(d)    list_entry(d, edma_descriptor, desc_node)

#endif
