/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/module.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/uaccess.h>
#include <linux/delay.h>
#include <asm-generic/siginfo.h>
#include <asm/signal.h>
#include "lj_hal_kernel_dma_engine.h"
#include "lj_hal_kernel_platform_edma_dev.h"
#include "lj_hal_kernel_platform_gp.h"
#include "lj_hal_kernel_platform.h"
#include "lj_hal_kernel_debug.h"
#include "lj_hal_kernel.h"
#include "ljmicro_family.h"

#define _LJ_OBJ_ZONE    ljvZONE_DEVICE

static DEFINE_MUTEX(cdmaMutex);
static ljsEDMA *galEDMA = ljvNULL;

ljeSTATUS ljkEDMA_DeviceInit(    edma_device *edma)
{
    int i;
    ljeSTATUS status = ljvSTATUS_OK;

    if (!edma)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    if (!galEDMA)
    {
        galEDMA = kzalloc(sizeof(*galEDMA), GFP_KERNEL);
        if (!galEDMA)
        {
            return ljvSTATUS_OUT_OF_MEMORY;
        }

        galEDMA->edma_dev = edma;
        galEDMA->rdChan = kmalloc(sizeof(ljsEDMA_CHAN) * edma->rchan_count, GFP_KERNEL);
        if (!galEDMA->rdChan)
        {
            status = ljvSTATUS_OUT_OF_MEMORY;
            goto err_rd_chan;
        }

        galEDMA->wrChan = kmalloc(sizeof(ljsEDMA_CHAN)* edma->wchan_count, GFP_KERNEL);
        if (!galEDMA->wrChan)
        {
            status = ljvSTATUS_OUT_OF_MEMORY;
            goto err_wr_chan;
        }

        for (i = 0; i< edma->rchan_count; i++)
        {
            galEDMA->rdChan[i].dmaChan = ljvNULL;
            galEDMA->rdChan[i].state = ljvDMA_UNUSE;
            galEDMA->rdChan[i].sgt = ljvNULL;
        }

        for (i = 0; i< edma->wchan_count; i++)
        {
            galEDMA->wrChan[i].dmaChan = ljvNULL;
            galEDMA->wrChan[i].state = ljvDMA_UNUSE;
            galEDMA->wrChan[i].sgt = ljvNULL;
        }

//		galEDMA->dmaBusy = ljvFALSE;    fq

        mutex_init(&galEDMA->dmaMutex);
    }

    return status;

err_wr_chan:
    kfree(galEDMA->rdChan);
err_rd_chan:
    kfree(galEDMA);

    return status;
}


void ljkEDMA_DeviceExit(void)
{
    if (!galEDMA)
    {
        return;
    }

    kfree(galEDMA->rdChan);
    kfree(galEDMA->wrChan);
    kfree(galEDMA);

    return;
}

static void ljkEDMA_TransactionCompleteCB(void *veri_chan)
{
    ljsEDMA_CHAN *vchan = (ljsEDMA_CHAN *)veri_chan;
    /*
     * Step8, indicate the dma transaction has completed
     */

    complete(&vchan->comp);
}

static ljeSTATUS ljkEDMA_ChanProbe(ljsEDMA_CHAN *vchan, ljvEDMA_DIRECTION dir)
{
    ljsDMA_CHANNEL *chan;
    ljuDMA_CAP_MASK mask;
    ljeDMA_TX_DIRECTION dmaDir = ljeDMA_TX_DIRECTION_MEM_TO_DEV;
    ljeSTATUS status = ljvSTATUS_FALSE;

    if (!vchan)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    if (dir != ljvEDMA_DEV_TO_MEM && dir != ljvEDMA_MEM_TO_DEV)
    {
        printk("invalid dma direction config\n");
        return ljvSTATUS_INVALID_ARGUMENT;
    }
    /*
     * Step 1, zero out the capability mask then iniilialize
     * it for a slave channel that is private.
     */
    ljkDMA_ZeroCapMask(&mask);
    ljkDMA_SetCapMasks(&mask, ljeDMA_TX_TYPE_PRIVATE);
    ljkDMA_SetCapMasks(&mask, ljeDMA_TX_TYPE_SLAVE);
    /*
     * Step 2, request the transmit and recieive channels for edma
     * channels from dma engine.
     */
    if (dir == ljvEDMA_MEM_TO_DEV)
    {
        dmaDir = ljeDMA_TX_DIRECTION_MEM_TO_DEV;
        status = ljkDMA_RequestAnySlaveChannel(&galEDMA->edma_dev->slave_rdma, &chan);
    }
    else if (dir == ljvEDMA_DEV_TO_MEM)
    {
        dmaDir = ljeDMA_TX_DIRECTION_DEV_TO_MEM;
        status = ljkDMA_RequestAnySlaveChannel(&galEDMA->edma_dev->slave_wdma, &chan);
    }

    if (ljmIS_ERROR(status))
    {
        if (status != ljvSTATUS_PROBE_LATER)
        {
            printk("can not request channel for dir %d\n", dir);
        }
        return status;
    }

    vchan->dmaChan = chan;
    vchan->dir = dmaDir;
    vchan->state = ljvDMA_USED;

    return ljvSTATUS_OK;
}

static ljeSTATUS ljkEDMA_ConfigAndSubmit(ljsEDMA_CHAN *vchan)
{
    ljsDMA_TX_DESCRIPTOR *desc;
    ljsDMA_SLAVE_CONFIG config;
    ljdDMA_TX_IDENTIFIER identifier;
    ljeSTATUS status;

    if (!vchan || !vchan->dmaChan)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    /*
     * Step3, set config to channel.
     * In memcpy, the config will be ingored, in slave sg mode, src is slave address
     * in write(dev2mem), and dst is slave address in read (mem2dev).
     */
    config.dstAddr = vchan->dst;
    config.srcAddr = vchan->src;
    config.direction = vchan->dir;
    config.srcAddrBusWidth = 1 << ljeDMA_SLAVE_BUS_WIDTH_4BYTES;
    config.dstAddrBusWidth = 1 << ljeDMA_SLAVE_BUS_WIDTH_4BYTES;

    status = ljkDMAENGINE_SlaveConfig(vchan->dmaChan, &config);
    if (ljmIS_ERROR(status))
    {
        return status;
    }
    /*
     * Step4, create a (or a chain) buffer(channel) descriptor for dma buffer
     */

    if (vchan->sgt)
    {
        desc = ljkDMAENGINE_PrepareSlaveSG(vchan->dmaChan,
                                    vchan->sgt->sgl,
                                    vchan->sgt->nents,
                                    vchan->dir,
                                    0);
    }
    else
    {
        desc = ljkDMAENGINE_PrepareMemcpy(vchan->dmaChan,
                                    vchan->src,
                                    vchan->dst,
                                    vchan->len,
                                    0);
    }

    /* Make sure the operation was completed successfully */
    if (!desc)
    {
        printk("ljkDMAENGINE_PrepareSlaveSG error\n");
        status = ljvSTATUS_INVALID_REQUEST;
    }
    else
    {
        desc->ljkDMA_TxCallback = ljkEDMA_TransactionCompleteCB;
        desc->txCallbackParam = vchan;

        /*
         * Step5, submit the transaction to dma engine so that it can be queued to
         * precess later and get a identifier to trace its status.
         */

        identifier = ljkDMAENGINE_Submit(desc);
        if (ljkDMA_SubmitError(identifier))
        {
            return ljvSTATUS_GENERIC_IO;
        }
        vchan->identifier = identifier;
    }

    return status;
}

static ljeSTATUS ljkEDMA_StartTransfer(ljsEDMA_CHAN *vchan)
{
    ljtUINT32 timeout = msecs_to_jiffies(10000);
    ljeDMA_STATUS status = ljeDMA_STATUS_ERROR;

    if (!vchan || !vchan->dmaChan)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }
    /*
     * step6, initizlize the completion and start dma transaction which previously
     * queued in the dma engine
     */
    init_completion(&vchan->comp);
    ljkDMAENGINE_IssuePending(vchan->dmaChan);

    if (vchan->sync_wait)
    {
        /*
         * Step7, wait for transaction to complete, timeout or
         * get an error.
         */
        timeout = wait_for_completion_timeout(&vchan->comp, timeout);
        if (timeout == 0)
        {
            printk("\n--------ljkEDMA_StartTransfer---timeout---src:0x%llx;dst:0x%llx;len:0x%lx;chan_id:%d----\n",vchan->src, vchan->dst, vchan->len, vchan->dmaChan->id);

//Reset edma test
#if 1
            //status = ljkDMAENGINE_Resume(vchan->dmaChan);
            status = (ljeDMA_STATUS)ljkDMAENGINE_TerminateAll(vchan->dmaChan);
            //printk("\n--------ljkEDMA_StartTransfer---status = %d---\n",status);

            //if (status != 0)
            //    return ljvSTATUS_GENERIC_IO;
#endif
           // timeout = msecs_to_jiffies(10000);
           // timeout = wait_for_completion_timeout(&vchan->comp, timeout);
           // printk("\n--------ljkEDMA_StartTransfer---timeout---%d----\n",timeout);

            //if (timeout == 0)
                return ljvSTATUS_TIMEOUT;
        }

        status = ljkDMA_SyncWait(vchan->dmaChan, vchan->identifier, 0);

        if (timeout == 0)
        {
            printk("DMA transfer 0x%x is timeout\n", vchan->identifier);
        }
        else if (status != ljeDMA_STATUS_COMPLETE)
        {
            printk("DMA transfer 0x%x is %s\n",
                vchan->identifier, status == ljeDMA_STATUS_ERROR ? "error" : "in progress");
        }
    }

    return status == ljeDMA_STATUS_COMPLETE ? ljvSTATUS_OK : ljvSTATUS_GENERIC_IO;
}

static void ljkEDMA_ChanRelease(ljsEDMA_CHAN *vchan)
{
    if (!vchan)
    {
        return;
    }

    if (vchan->dmaChan)
    {
        ljkDMA_ReleaseChannel(vchan->dmaChan);
    }

    mutex_lock(&galEDMA->dmaMutex);
    vchan->dmaChan = ljvNULL;
    vchan->state = ljvDMA_UNUSE;
    vchan->sgt = ljvNULL;
//	galEDMA->dmaBusy = ljvFALSE;  fq
    mutex_unlock(&galEDMA->dmaMutex);
}

static ljsEDMA_CHAN *ljkEDMA_GetUnuseChan(ljvEDMA_DIRECTION dir)
{
    ljsEDMA_CHAN *vchan = ljvNULL;
    ljtUINT index;

    mutex_lock(&galEDMA->dmaMutex);

#if 0   //fq
    if (galEDMA->dmaBusy)
    {
		mutex_unlock(&galEDMA->dmaMutex);
		return ljvNULL;
    }
#endif
    /* request write channel */
    if (dir == ljvEDMA_DEV_TO_MEM)
    {
        for (index = 0; index < galEDMA->edma_dev->wchan_count; index++)
        {
            if  (galEDMA->wrChan[index].state == ljvDMA_UNUSE)
            {
                galEDMA->wrChan[index].state = ljvDMA_USED;
                vchan = &galEDMA->wrChan[index];

//				galEDMA->dmaBusy = ljvTRUE;   fq
                break;
            }
        }
    }
    else if (dir == ljvEDMA_MEM_TO_DEV)
    {
        for (index = 0; index < galEDMA->edma_dev->rchan_count; index++)
        {
            if  (galEDMA->rdChan[index].state == ljvDMA_UNUSE)
            {
                galEDMA->rdChan[index].state = ljvDMA_USED;
                vchan = &galEDMA->rdChan[index];

//				galEDMA->dmaBusy = ljvTRUE;  fq
                break;
            }
        }
    }
    else
    {
        printk("unsupport dma direction config\n");
    }
    mutex_unlock(&galEDMA->dmaMutex);

    return vchan;
}

static struct sg_table *ljkEDMA_DMABufMap(ljkOS Os, ljuVIDMEM_NODE_PTR node, enum dma_data_direction dir, ljtUINT64  syncLen, ljtUINT64  Offset)
{
    struct sg_table *sgt = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;
    int mapped_nents;
    do
    {
        ljkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
        ljtPHYS_ADDR physical = ljvNULL;
        ljtSIZE_T offset = 0;
        ljtSIZE_T bytes = 0;

        if (vidMemBlock && vidMemBlock->object.type == ljvOBJ_VIDMEM_BLOCK) /* virtual block memory */
        {
            physical = vidMemBlock->physical;
            offset = node->VirtualChunk.offset;
            bytes = node->VirtualChunk.bytes;
        }
        else if (node->VidMem.parent == ljvNULL) /* virtual mem in system ram */
        {
            physical = node->VidMem.physical;
            offset = 0;
            bytes = node->VidMem.bytes;
        }
        else    /* video ram */
        {
            physical = node->VidMem.parent->physical;
            offset = node->VidMem.offset;
            bytes = node->VidMem.bytes;
        }

        /*printk("------%s--------%d-------1------bytes = %ld--\n", __func__, __LINE__, bytes); */
        if (syncLen)
        {
            bytes = syncLen;
        }
        if (Offset)
        {
            offset = Offset;
        }
        //printk("------%s(%d)------------bytes = 0x%lx- offset = 0x%lx-\n", __func__, __LINE__, bytes, offset);

        ljmkERR_BREAK(
            ljkOS_MemoryGetSGT(Os, physical, offset, bytes, (ljtPOINTER*)&sgt));

        mapped_nents = dma_map_sg(&galEDMA->edma_dev->pdev->dev, sgt->sgl, sgt->nents, dir);
        if (mapped_nents == 0)
        {
            sg_free_table(sgt);
            kfree(sgt);
            sgt = ljvNULL;
            ljmkERR_BREAK(ljvSTATUS_GENERIC_IO);
        }
        sgt->nents = mapped_nents;
    }
    while (ljvFALSE);

    return sgt;
}

static void ljkEDMA_DMABufUnmap(struct sg_table *sgt, enum dma_data_direction dir)
{
    dma_unmap_sg(&galEDMA->edma_dev->pdev->dev, sgt->sgl, sgt->orig_nents, dir);

    sg_free_table(sgt);
    kfree(sgt);
}

#if 0
static ljeSTATUS ljkDMA_GetGPUPhyicalAddress(
    IN ljkKERNEL Kernel,
    IN ljkVIDMEM_NODE NodeObject,
    OUT ljtPHYS_ADDR_T * PhysicalAddress)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkONERROR(ljkVIDMEM_NODE_GetPhysical(Kernel, NodeObject, 0, PhysicalAddress));
    ljmkONERROR(ljkOS_CPUPhysicalToGPUPhysical(Kernel->os,
        *PhysicalAddress,
        PhysicalAddress
        ));

OnError:
    return status;
}
#endif

#ifdef DEBUG
#define CHECK_VRAM_PHYSICAL if(physical < VRAM_GP200_BUS40_GPU_START) { \
        status = ljvSTATUS_INVALID_ADDRESS; \
        printk("%s(%d): invalid vram addr 0x%llx\n", __FUNCTION__, __LINE__, physical); \
        goto OnError; \
    }
#define CHECK_OFFSET if(Offset + SyncLen > ljmMIN(size_src, size_dst)) { \
    status = ljvSTATUS_INVALID_ARGUMENT; \
    printk("%s(%d): invalid Offset 0x%lx\n", __FUNCTION__, __LINE__, Offset); \
    goto OnError; \
}
#else
#define CHECK_VRAM_PHYSICAL
#define CHECK_OFFSET
#endif

extern struct _ljsPLATFORM_PCIE gp_platform;

#if 0
ljeSTATUS ljkEDMA_SyncMemory(
        IN ljtPOINTER Object,
        IN ljtPOINTER DstNode,
        IN ljtPOINTER SrcNode,
        IN ljtUINT32  Reason,
        IN ljtUINT64  SyncLen,
        IN ljtUINT64  Offset
    )
{
    ljkKERNEL kernel = (ljkKERNEL)Object;
    ljsEDMA_CHAN *vchan = ljvNULL;
    ljkVIDMEM_NODE dst_node_obj = ljvNULL;
    ljkVIDMEM_NODE src_node_obj = ljvNULL;
    ljtSIZE_T      size_src, size_dst/*, i = 0*/;
    struct sg_table *sgt = ljvNULL;
    ljtPHYS_ADDR_T physical = ljvINVALID_PHYSICAL_ADDRESS;
    ljeSTATUS status = ljvSTATUS_OK;
    ljvEDMA_DIRECTION dir;
    ljtBOOL contiguous;
    ljtPHYS_ADDR_T transmitPhysical = ljvINVALID_PHYSICAL_ADDRESS;

    /*abc: for debug*/
//    printk("abc: ljkEDMA_SyncMemory need to be debug!!!\n");
//    return -1;

    if (Reason == ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM ||
        Reason == ljvSYNC_MEMORY_DIRECTION_LOCALOFFSET_TO_SYSTEM)
        dir = ljvEDMA_DEV_TO_MEM;
    else if (Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL ||
        Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCALOFFSET)
        dir = ljvEDMA_MEM_TO_DEV;
    else {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    /* get channel slot */
    do {
        vchan = ljkEDMA_GetUnuseChan(dir);
    } while(vchan == ljvNULL);

    /* allocate channel resource */
    ljmkONERROR(ljkEDMA_ChanProbe(vchan, dir));

    if (Reason == ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM || Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL)
    {
        dst_node_obj   = (ljkVIDMEM_NODE)DstNode;
        src_node_obj   = (ljkVIDMEM_NODE)SrcNode;

        if ((!dst_node_obj->node) || (!src_node_obj->node))
        {
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        ljkVIDMEM_NODE_GetSize(kernel, src_node_obj, &size_src);
        ljkVIDMEM_NODE_GetSize(kernel, dst_node_obj, &size_dst);
        printk("\n----ljkEDMA_SyncMemory-------size_src: 0x%lx--size_dst:0x%lx--SyncLen:0x%llx; dir=%d; Offset=0x%llx\n", size_src, size_dst, SyncLen, dir, Offset);
        if (dir == ljvEDMA_MEM_TO_DEV)
        {
            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, src_node_obj, &contiguous));
            ljmkONERROR(ljkVIDMEM_NODE_GetGPUPhysical(kernel, dst_node_obj, 0, &physical));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory-------contiguous--ljvEDMA_MEM_TO_DEV--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetCPUPhysical(kernel, src_node_obj, 0, &transmitPhysical));
                vchan->src = transmitPhysical + Offset;
                vchan->dst = physical + Offset;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
                CHECK_OFFSET;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory------non-contiguous--ljvEDMA_MEM_TO_DEV--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, src_node_obj->node, DMA_TO_DEVICE, SyncLen, Offset);/*当SyncLen为0的时候，此函数会去取主存的长度来创建sgt表。此处应该是取主存和显存之间较小的长度来创建sgt表*/
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = ljvINVALID_PHYSICAL_ADDRESS;
                vchan->dst = physical + Offset;
                CHECK_VRAM_PHYSICAL;
            }
        }

        else
        {
            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, dst_node_obj, &contiguous));
            ljmkONERROR(ljkVIDMEM_NODE_GetGPUPhysical(kernel, src_node_obj, 0, &physical));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory------contiguous--ljvEDMA_DEV_TO_MEM--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetCPUPhysical(kernel, dst_node_obj, 0, &transmitPhysical));
                vchan->src = physical + Offset;
                vchan->dst = transmitPhysical + Offset;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
                CHECK_OFFSET;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory-----non--contiguous--ljvEDMA_DEV_TO_MEM--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, dst_node_obj->node, DMA_FROM_DEVICE, SyncLen, Offset);
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = physical + Offset;
                vchan->dst = ljvINVALID_PHYSICAL_ADDRESS;
                CHECK_VRAM_PHYSICAL;
            }
        }
    }
    else
    {
        if (dir == ljvEDMA_MEM_TO_DEV)
        {
            src_node_obj = (ljkVIDMEM_NODE)SrcNode;
            ljkVIDMEM_NODE_GetSize(kernel, src_node_obj, &size_src);

            physical = gp_platform.pcie_info[kernel->device->platformIndex].gpu_vram_start + ljmPTR2INT32(DstNode);
            size_dst = SyncLen;

            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                            "%s(%d): node %p >> %lx\n",
                            __FUNCTION__, __LINE__,
                            src_node_obj, (unsigned long)physical);

            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, src_node_obj, &contiguous));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory-------contiguous--ljvEDMA_MEM_TO_DEV--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetCPUPhysical(kernel, src_node_obj, 0, &transmitPhysical));
                vchan->src = transmitPhysical;
                vchan->dst = physical;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory------non-contiguous--ljvEDMA_MEM_TO_DEV--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, src_node_obj->node, DMA_TO_DEVICE, SyncLen, Offset);
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = ljvINVALID_PHYSICAL_ADDRESS;
                vchan->dst = physical;
                CHECK_VRAM_PHYSICAL;
            }
        }
        else
        {
            physical = gp_platform.pcie_info[kernel->device->platformIndex].gpu_vram_start + ljmPTR2INT32(SrcNode);
            size_src = SyncLen;

            dst_node_obj = (ljkVIDMEM_NODE)DstNode;
            ljkVIDMEM_NODE_GetSize(kernel, dst_node_obj, &size_dst);

            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                            "%s(%d): %lx >> node %p\n",
                            __FUNCTION__, __LINE__,
                            (unsigned long)physical, dst_node_obj);

            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, dst_node_obj, &contiguous));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory------contiguous--ljvEDMA_DEV_TO_MEM--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetCPUPhysical(kernel, dst_node_obj, 0, &transmitPhysical));
                vchan->src = physical;
                vchan->dst = transmitPhysical;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory-----non--contiguous--ljvEDMA_DEV_TO_MEM--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, dst_node_obj->node, DMA_FROM_DEVICE, SyncLen, Offset);
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = physical;
                vchan->dst = ljvINVALID_PHYSICAL_ADDRESS;
                CHECK_VRAM_PHYSICAL;
            }
        }
    }

    if(SyncLen != 0)
        vchan->len = SyncLen;
    else
        vchan->len = ljmMIN(size_src, size_dst);

#if 1
    /*config and submit transcation to dma engine */
    ljmkONERROR(ljkEDMA_ConfigAndSubmit(vchan));

    /* start dma */
    vchan->sync_wait = ljvTRUE;
    ljkEDMA_StartTransfer(vchan);

#else
    for ( i = 0; i < 3; i++)
    {
        /*config and submit transcation to dma engine */
        ljmkONERROR(ljkEDMA_ConfigAndSubmit(vchan));

        /* start dma */
        vchan->sync_wait = ljvTRUE;
        status = ljkEDMA_StartTransfer(vchan);

        if(status != ljvSTATUS_TIMEOUT)
        {
            break;
        }
        else
        {
            printk("\n--------retry---status = %d---\n",status);
        }
    }
    if (i == 3)
    {
        status =  ljvSTATUS_TIMEOUT;
        vchan->sync_wait = ljvFALSE;
    }
#endif

OnError:
    if (sgt)
    {
        /* release resource */
        if (dir == ljvEDMA_MEM_TO_DEV)
        {
            ljkEDMA_DMABufUnmap(sgt, DMA_TO_DEVICE);
        }
        else if (dir == ljvEDMA_DEV_TO_MEM)
        {
            ljkEDMA_DMABufUnmap(sgt, DMA_FROM_DEVICE);
        }
    }
    ljkEDMA_ChanRelease(vchan);

    return status;
}
#else
ljeSTATUS ljkVIDMEM_NODE_GetDmaAddress(ljkKERNEL Kernel, ljkVIDMEM_NODE NodeObject,
                                               ljtSIZE_T Offset, ljtPHYS_ADDR_T *DmaAddress);
ljeSTATUS ljkEDMA_dmaCopy(
        IN ljtPOINTER Object,
        IN ljsDMA_TRANS_INFO *Info
    )
{
    ljkKERNEL kernel = (ljkKERNEL)Object;
    ljsEDMA_CHAN *vchan = ljvNULL;
    ljkVIDMEM_NODE dst_node_obj = (ljkVIDMEM_NODE)Info->dst_node;
    ljkVIDMEM_NODE src_node_obj = (ljkVIDMEM_NODE)Info->src_node;
    ljtSIZE_T Offset = Info->offset;
    ljtSIZE_T SyncLen = Info->bytes;
    ljtUINT32 Reason = Info->reason;

    ljtSIZE_T      size_src, size_dst, i = 0;
    struct sg_table *sgt = ljvNULL;
    ljtPHYS_ADDR_T physical = ljvINVALID_PHYSICAL_ADDRESS;
    ljeSTATUS status = ljvSTATUS_OK;
    ljvEDMA_DIRECTION dir;
    ljtBOOL contiguous;
    ljtPHYS_ADDR_T transmitPhysical = ljvINVALID_PHYSICAL_ADDRESS;

    /*abc: for debug*/
//    printk("abc: ljkEDMA_SyncMemory need to be debug!!!\n");
//    return -1;

    if (Reason == ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM ||
        Reason == ljvSYNC_MEMORY_DIRECTION_LOCALOFFSET_TO_SYSTEM)
        dir = ljvEDMA_DEV_TO_MEM;
    else if (Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL ||
        Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCALOFFSET)
        dir = ljvEDMA_MEM_TO_DEV;
    else {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    /* get channel slot */
    do {
        vchan = ljkEDMA_GetUnuseChan(dir);
    } while(vchan == ljvNULL);

    /* allocate channel resource */
    ljmkONERROR(ljkEDMA_ChanProbe(vchan, dir));

    if (Reason == ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM || Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL)
    {
        //dst_node_obj   = (ljkVIDMEM_NODE)DstNode;
        //src_node_obj   = (ljkVIDMEM_NODE)SrcNode;

        if ((!dst_node_obj->node) || (!src_node_obj->node))
        {
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        ljkVIDMEM_NODE_GetSize(kernel, src_node_obj, &size_src);
        ljkVIDMEM_NODE_GetSize(kernel, dst_node_obj, &size_dst);
        //printk("\n----ljkEDMA_SyncMemory-------size_src: 0x%lx--size_dst:0x%lx--SyncLen:0x%lx; dir=%d; Offset=0x%lx\n", size_src, size_dst, SyncLen, dir, Offset);
        if (dir == ljvEDMA_MEM_TO_DEV)
        {
            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, src_node_obj, &contiguous));
            ljmkONERROR(ljkVIDMEM_NODE_GetGPUPhysical(kernel, dst_node_obj, 0, &physical));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory-------contiguous--ljvEDMA_MEM_TO_DEV--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetDmaAddress(kernel, src_node_obj, 0, &transmitPhysical));
                vchan->src = transmitPhysical + Offset;
                vchan->dst = physical + Offset;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
                CHECK_OFFSET;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory------non-contiguous--ljvEDMA_MEM_TO_DEV--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, src_node_obj->node, DMA_TO_DEVICE, SyncLen, Offset);/*当SyncLen为0的时候，此函数会去取主存的长度来创建sgt表。此处应该是取主存和显存之间较小的长度来创建sgt表*/
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = ljvINVALID_PHYSICAL_ADDRESS;
                vchan->dst = physical + Offset;
                CHECK_VRAM_PHYSICAL;
            }
        }

        else
        {
            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, dst_node_obj, &contiguous));
            ljmkONERROR(ljkVIDMEM_NODE_GetGPUPhysical(kernel, src_node_obj, 0, &physical));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory------contiguous--ljvEDMA_DEV_TO_MEM--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetDmaAddress(kernel, dst_node_obj, 0, &transmitPhysical));
                vchan->src = physical + Offset;
                vchan->dst = transmitPhysical + Offset;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
                CHECK_OFFSET;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory-----non--contiguous--ljvEDMA_DEV_TO_MEM--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, dst_node_obj->node, DMA_FROM_DEVICE, SyncLen, Offset);
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = physical + Offset;
                vchan->dst = ljvINVALID_PHYSICAL_ADDRESS;
                CHECK_VRAM_PHYSICAL;
            }
        }
    }
    else
    {
        if (dir == ljvEDMA_MEM_TO_DEV)
        {
            //src_node_obj = (ljkVIDMEM_NODE)SrcNode;
            ljkVIDMEM_NODE_GetSize(kernel, src_node_obj, &size_src);

            physical = gp_platform.pcie_info[kernel->device->platformIndex].gpu_vram_start + ljmPTR2INT32(dst_node_obj);
            size_dst = SyncLen;

            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                            "%s(%d): node %p >> %lx\n",
                            __FUNCTION__, __LINE__,
                            src_node_obj, (unsigned long)physical);

            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, src_node_obj, &contiguous));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory-------contiguous--ljvEDMA_MEM_TO_DEV--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetDmaAddress(kernel, src_node_obj, 0, &transmitPhysical));
                vchan->src = transmitPhysical;
                vchan->dst = physical;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory------non-contiguous--ljvEDMA_MEM_TO_DEV--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, src_node_obj->node, DMA_TO_DEVICE, SyncLen, Offset);
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = ljvINVALID_PHYSICAL_ADDRESS;
                vchan->dst = physical;
                CHECK_VRAM_PHYSICAL;
            }
        }
        else
        {
            physical = gp_platform.pcie_info[kernel->device->platformIndex].gpu_vram_start + ljmPTR2INT32(src_node_obj);
            size_src = SyncLen;

            //dst_node_obj = (ljkVIDMEM_NODE)DstNode;
            ljkVIDMEM_NODE_GetSize(kernel, dst_node_obj, &size_dst);

            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE,
                            "%s(%d): %lx >> node %p\n",
                            __FUNCTION__, __LINE__,
                            (unsigned long)physical, dst_node_obj);

            ljmkONERROR(ljkVIDMEM_NODE_IsContiguous(kernel, dst_node_obj, &contiguous));
            if (contiguous)
            {
//                printk("\n----ljkEDMA_SyncMemory------contiguous--ljvEDMA_DEV_TO_MEM--\n");
                ljmkONERROR(ljkVIDMEM_NODE_GetDmaAddress(kernel, dst_node_obj, 0, &transmitPhysical));
                vchan->src = physical;
                vchan->dst = transmitPhysical;
                vchan->sgt = ljvNULL;
                CHECK_VRAM_PHYSICAL;
            }
            else
            {
//                printk("\n----ljkEDMA_SyncMemory-----non--contiguous--ljvEDMA_DEV_TO_MEM--\n");
                sgt = ljkEDMA_DMABufMap(kernel->os, dst_node_obj->node, DMA_FROM_DEVICE, SyncLen, Offset);
                if (!sgt)
                {
                    ljmkONERROR(ljvSTATUS_GENERIC_IO);
                }
                vchan->sgt = sgt;
                vchan->src = physical;
                vchan->dst = ljvINVALID_PHYSICAL_ADDRESS;
                CHECK_VRAM_PHYSICAL;
            }
        }
    }

    if(SyncLen != 0)
        vchan->len = SyncLen;
    else
        vchan->len = ljmMIN(size_src, size_dst);

#if 0
    /*config and submit transcation to dma engine */
    ljmkONERROR(ljkEDMA_ConfigAndSubmit(vchan));

    /* start dma */
    vchan->sync_wait = ljvTRUE;
    ljkEDMA_StartTransfer(vchan);

#else
    for ( i = 0; i < 3; i++)
    {
        /*config and submit transcation to dma engine */
        ljmkONERROR(ljkEDMA_ConfigAndSubmit(vchan));

        /* start dma */
        vchan->sync_wait = ljvTRUE;
        status = ljkEDMA_StartTransfer(vchan);

        if(status != ljvSTATUS_TIMEOUT)
        {
            break;
        }
        else
        {
            printk("\n--------retry---status = %d---\n",status);
        }
    }
    if (i == 3)
    {
        status =  ljvSTATUS_TIMEOUT;
        vchan->sync_wait = ljvFALSE;
    }
#endif

OnError:
    if (sgt)
    {
        /* release resource */
        if (dir == ljvEDMA_MEM_TO_DEV)
        {
            ljkEDMA_DMABufUnmap(sgt, DMA_TO_DEVICE);
        }
        else if (dir == ljvEDMA_DEV_TO_MEM)
        {
            ljkEDMA_DMABufUnmap(sgt, DMA_FROM_DEVICE);
        }
    }
    ljkEDMA_ChanRelease(vchan);

    return status;
}

#endif

static ljeSTATUS ljkCDMADEV_dma_trans_block(
ljkKERNEL kernel, ljtUINT64 sar, ljtUINT64 dar, ljtUINT64 trans_len, ljtUINT32 channel)
{
    ljtUINT32 inst_rom [12];
    ljtUINT32 sar0 = sar;
    ljtUINT32 dar0 = dar;
    ljtUINT32 i, addr_offset, n0, ccr0, ccr1, remain_burst;
    ljtUINT32 ret=0, num=0;
    ljtUINT32 *inst_base_v;
    ljtUINT32 tmpdata;
    ljtUINT32 core_offset = 0;
    ljtUINT64 inst_base_offset = gp_platform.pcie_info[0].bar[PCI_BAR_0].size
        - ((gp_platform.pcie_info[0].flags & LJMICRO_IS_BUS32) ? VRAM_FIX_GART_PAGE_SIZE : 0)
        - VRAM_FIX_PCI_DMA_SIZE
        - VRAM_FIX_INST_SIZE;

    if(((dar>>32)!=0x80)||((sar>>32)!=0x80))
    {
        printk("dar or sar cannot above 40bit. dar:0x%llx; sar:0x%llx\n", dar, sar);
        return -1;
    }

    if((trans_len <= AMBA_TR_LEN_LIMIT_3)&&(trans_len > 256))
    {
        if((trans_len%16) != 0)
        {
            printk("trans_len <= 64k(>256byte) must be an integer multiple of 16Byte. trans_len:0x%llx\n", trans_len);
            return -1;
        }
    }else if((trans_len > AMBA_TR_LEN_LIMIT_3)&&(trans_len < AMBA_TR_LEN_LIMIT_1))
    {
        if((trans_len%256) != 0)
        {
            printk("trans_len > 64k must be an integer multiple of 256Byte. trans_len:0x%llx\n", trans_len);
            return -1;
        }
    }else if(trans_len >= AMBA_TR_LEN_LIMIT_1)
    {
        printk("trans_len cannot above 16M. trans_len:0x%llx\n", trans_len);
        return -1;
    }

    if(channel > 7)
    {
        printk("channel cannot above 7. channel:0x%x\n", channel);
        return -1;
    }

    core_offset = (kernel->core==ljvCORE_MAJOR) ? REG_3D_OFFSET : REG_2D_OFFSET;
    //printk("\n-------ljkCDMADEV_dma_trans_block-------core_offset---%x---\n",core_offset);
    //writel(0x8080, (void __iomem *)demo_dev.base2_logical + MISC_REG_BASE + MISC_DMA_AADDR_HIGH_OFFSET);
    ljkOS_WriteRegisterEx(kernel->os,kernel,
                            MISC_REG_BASE + MISC_DMA_AADDR_HIGH_OFFSET - core_offset,
                            0x8080);

    /* enable interrupt register */
    //writel(0x1, (void __iomem *)demo_dev.base2_logical + 0x330020);
    ljkOS_WriteRegisterEx(kernel->os,kernel,
                            INTERRUPT_ENABLE_REGISTER - core_offset,
                            0x1);

    if(trans_len < 256)
    {
        n0 = trans_len;
        /*assign ccr0 = {1'b0,3'b000,
            3'b000,3'b010,4'b0000,3'b000,1'b1,
            3'b000,3'b010,4'b0000,3'b000,1'b1};*/
        ccr0 = 0b100000000100001000000001;/* 0x00804201 */

        inst_rom[0] = (ccr0&0xffff)<<16 | 0x01bc;
        inst_rom[1] = 0x00bc<<16 | ccr0>>16;
        inst_rom[2] = sar0;
        inst_rom[3] = (dar0&0xffff)<<16 | 0x02bc;
        inst_rom[4] = (n0-1)<<24 | 0x20<<16 | dar0>>16;
        inst_rom[5] = 0x02380804;
        inst_rom[6] = 0x00003413;
        inst_rom[7] = 0x0;
        inst_rom[8] = 0x0;
        inst_rom[9] = 0x0;
        inst_rom[10] = 0x0;
        inst_rom[11] = 0x0;
    }
    else if(trans_len <= AMBA_TR_LEN_LIMIT_3)
    {
        n0 = trans_len / ((128/8)*16);

        /*assign ccr0 = {1'b0,3'b000,
            3'b000,3'b010,4'b1111,3'b100,1'b1,
            3'b000,3'b010,4'b1111,3'b100,1'b1};*/
        ccr0 = 0b101111100100001011111001;/* 0x00be42f9 */

        remain_burst = (trans_len % ((128/8)*16))/16;
        /*assign ccr1 = {1'b0,3'b000,
            3'b000,3'b010,remain_burst-4'h1,3'b100,1'b1,
            3'b000,3'b010,remain_burst-4'h1,3'b100,1'b1};*/
        ccr1 = 0b100000100100001000001001 | ((remain_burst-1)<<4) | ((remain_burst-1)<<18);

        inst_rom[0] = (ccr0&0xffff)<<16 | 0x01bc;
        inst_rom[1] = 0x00bc<<16 | ccr0>>16;
        inst_rom[2] = sar0;
        inst_rom[3] = (dar0&0xffff)<<16 | 0x02bc;
        inst_rom[4] = (n0-1)<<24 | 0x20<<16 | dar0>>16;
        inst_rom[5] = 0x02380804;
        if(remain_burst != 0){
            inst_rom[6] = (ccr1&0xffff)<<16 | 0x01bc;
            inst_rom[7] = 0x0804<<16 | ccr1>>16;
            inst_rom[8] = 0x00003413;
            inst_rom[9] = 0x0;
            inst_rom[10] = 0x0;
            inst_rom[11] = 0x0;
        }else{
            inst_rom[6] = 0x00003413;
            inst_rom[7] = 0x0;
            inst_rom[8] = 0x0;
            inst_rom[9] = 0x0;
            inst_rom[10] = 0x0;
            inst_rom[11] = 0x0;
        }
    }
    else{/* trans_len > 64k */
        //printk("\n-------ljkCDMADEV_dma_trans_block-------trans_len---%llx---\n",trans_len);
        n0 = trans_len / (256*(128/8)*16);

        /*assign ccr0 = {1'b0,3'b000,
            3'b000,3'b010,4'b1111,3'b100,1'b1,
            3'b000,3'b010,4'b1111,3'b100,1'b1};*/
        ccr0 = 0b101111100100001011111001;/* 0x00be42f9 */

        remain_burst = (trans_len % (256*(128/8)*16))/256;

        inst_rom[0] = (ccr0&0xffff)<<16 | 0x01bc;
        inst_rom[1] = 0x00bc<<16 | ccr0>>16;
        inst_rom[2] = sar0;
        inst_rom[3] = (dar0&0xffff)<<16 | 0x02bc;
        inst_rom[4] = 0x1818<<16 | dar0>>16;
        inst_rom[5] = 0xff22<<16 | (n0-1)<<8 | 0x20;
        inst_rom[6] = 0x023c0804;

        if(remain_burst != 0){
            inst_rom[7] = (remain_burst-1)<<24 | 0x20<<16 | 0x0638;
            inst_rom[8] = 0x02380804;
            inst_rom[9] = 0x00003413;
            inst_rom[10] = 0x0;
            inst_rom[11] = 0x0;
        }else{
            inst_rom[7] = 0x34130638;
            inst_rom[8] = 0xffff0000;
            inst_rom[9] = 0x0;
            inst_rom[10] = 0x0;
            inst_rom[11] = 0x0;
        }
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    inst_base_v = ioremap(gp_platform.pcie_info[0].bar[PCI_BAR_0].base
        + inst_base_offset, VRAM_FIX_INST_SIZE);
#else
    inst_base_v = ioremap_nocache(gp_platform.pcie_info[0].bar[PCI_BAR_0].base
        + inst_base_offset, VRAM_FIX_INST_SIZE);
#endif

    //printk("\n-------ljkCDMADEV_dma_trans_block-----bar0 base:%llx------\n",gp_platform.pcie_info[0].bar[PCI_BAR_0].base);
    //printk("\n-------ljkCDMADEV_dma_trans_block---------VRAM_FIX_INST_SIZE----%x-------------\n",VRAM_FIX_INST_SIZE);
    //printk("\n-------ljkCDMADEV_dma_trans_block---------inst_base_offset----%llx-------------\n",inst_base_offset);
    //printk("\n-------ljkCDMADEV_dma_trans_block---------inst_base_v----%p-------------\n",inst_base_v);
    for(i=0;i<12;i= i+1)
    {
        addr_offset = i*4;
        writel(inst_rom[i], (void __iomem *)inst_base_v + addr_offset);
    }
    iounmap(inst_base_v);

    ///////////////start_dma///////////////////////////////////
    //boot_from_pc
    //writel(0x0100, (void __iomem *)demo_dev.base2_logical + MISC_REG_BASE + MISC_DMA_CONFIG_OFFSET);
    ljkOS_WriteRegisterEx(kernel->os,kernel,
                            MISC_REG_BASE + MISC_DMA_CONFIG_OFFSET - core_offset,
                            0x0100);
    //printk("\n-------ljkCDMADEV_dma_trans_block---------aa---------\n");

    //read debug register
    // if 0, send dmago instruction
    do{
        if(num++ > 99)
        {
            printk("amba dma not idle; timeout!!!\n ");
            return -1;  //break;
        }
        udelay(50);

        ljkOS_ReadRegisterEx(kernel->os,kernel,
            DEBUG_STATUS_REGISTER - core_offset,
            &tmpdata);
    }while(tmpdata);
    //printk("\n-------ljkCDMADEV_dma_trans_block---------bb---------\n");
    //debug0 register 01a2_0100
    //writel(channel<<24 | 0xa2<<16 | channel<<8, (void __iomem *)demo_dev.base2_logical + 0x330d08);
    ljkOS_WriteRegisterEx(kernel->os,kernel,
                            DEBUG_INSTRUCTION_0_REGISTER - core_offset,
                            channel<<24 | 0xa2<<16 | channel<<8);

    //debug1 register
    //writel(inst_base, (void __iomem *)demo_dev.base2_logical + 0x330d0c);
    ljkOS_WriteRegisterEx(kernel->os,kernel,
                            DEBUG_INSTRUCTION_1_REGISTER - core_offset,
                            inst_base_offset);

	//wirte 0 to dbgcmd register,start dma
	//writel(0x0, (void __iomem *)demo_dev.base2_logical + 0x330d04);
	ljkOS_WriteRegisterEx(kernel->os,kernel,
                            DEBUG_COMMAND_REGISTER - core_offset,
                            0x0);

    //printk("amba dma start\n");
    //printk("\n-------ljkCDMADEV_dma_trans_block---------cc---------\n");
    //read int status register: poll
    do{
        if(num++ > 9999)
        {
            ljkOS_ReadRegisterEx(kernel->os,kernel,
                INTERRUPT_STATUS_REGISTER - core_offset,
                &tmpdata);
            //printk("amba dma int cannot come; timeout!!!: int status: 0x%x\n ", readl((void __iomem *)demo_dev.base2_logical + 0x330028));
            printk("amba dma int cannot come; timeout!!!: int status: 0x%x\n ", tmpdata);
            return -1;  //break;
        }
        //printk("amba dma int is coming: int status: 0x%x; misc int status: 0x%x\n ", readl((void __iomem *)demo_dev.base2_logical + 0x330028), readl((void __iomem *)demo_dev.base2_logical + 0x308440));
        mdelay(10);

        ljkOS_ReadRegisterEx(kernel->os,kernel,
            INTERRUPT_STATUS_REGISTER - core_offset,
            &tmpdata);
    }while(!tmpdata);
    //printk("\n-------ljkCDMADEV_dma_trans_block---------ddd---------\n");
    //clear interrupt
    //writel(0x1, (void __iomem *)demo_dev.base2_logical + 0x33002c);
    ljkOS_WriteRegisterEx(kernel->os,kernel,
                            INTERRUPT_CLEAR_REGISTER - core_offset,
                            0x1);
    //printk("amba dma end\n");

    return ret;
}

ljeSTATUS ljkDMA_Operaton(
        IN ljtPOINTER Object,
        IN ljtPOINTER Node_Src,
        IN ljtPOINTER Node_Dst,
        IN ljtUINT32  Dir,
        IN ljtUINT64  TransLen
    )
{
    ljkKERNEL kernel = (ljkKERNEL)Object;
    ljkVIDMEM_NODE nodeObject_Src = (ljkVIDMEM_NODE)Node_Src;
    ljkVIDMEM_NODE nodeObject_Dst = (ljkVIDMEM_NODE)Node_Dst;
    ljtPHYS_ADDR_T physical_mem = ljvINVALID_PHYSICAL_ADDRESS;
    ljtPHYS_ADDR_T physical_ddr = ljvINVALID_PHYSICAL_ADDRESS;
    ljtSIZE_T bytes;
    ljtUINT32  Reason;
    ljeSTATUS  status = ljvSTATUS_OK;
    ljsDMA_TRANS_INFO info;

    if ((!nodeObject_Src->node) || (!nodeObject_Dst->node))
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if ((Dir != ljvDMA_DIR_MEM_TO_DDR) && (Dir != ljvDMA_DIR_DDR_TO_MEM) && (Dir != ljvDMA_DIR_DDR_TO_DDR))
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (Dir == ljvDMA_DIR_DDR_TO_DDR)
    {
//        printk("\n------ljkDMA_Operaton----nodeObject_Dst----%p---------\n",nodeObject_Dst);
//        printk("\n------ljkDMA_Operaton----nodeObject_Dst----%p---------\n",nodeObject_Src);
        ljmkONERROR(ljkVIDMEM_NODE_GetGPUPhysical(kernel, nodeObject_Dst, 0, &physical_ddr));
        ljmkONERROR(ljkVIDMEM_NODE_GetGPUPhysical(kernel, nodeObject_Src, 0, &physical_mem));
        bytes = ljmMIN(nodeObject_Dst->node->VidMem.bytes, nodeObject_Src->node->VidMem.bytes);

        ljmkMUTEX_LOCK(cdmaMutex);

        while(bytes > AMBA_TR_LEN_LIMIT_1)
        {
            ljmkONERROR(ljkCDMADEV_dma_trans_block(kernel, physical_mem, physical_ddr, AMBA_TR_LEN_LIMIT_2, 2));
            physical_mem += AMBA_TR_LEN_LIMIT_2;
            physical_ddr += AMBA_TR_LEN_LIMIT_2;
            bytes -= AMBA_TR_LEN_LIMIT_2;
        }

        if((bytes > AMBA_TR_LEN_LIMIT_3)&&(bytes < AMBA_TR_LEN_LIMIT_1)&&(bytes&0xff))
        {
            ljmkONERROR(ljkCDMADEV_dma_trans_block(kernel, physical_mem, physical_ddr, bytes - (bytes&0xff), 1));
            physical_mem += bytes - (bytes&0xff);
            physical_ddr += bytes - (bytes&0xff);
            bytes = (bytes&0xff);
        }

        ljmkONERROR(ljkCDMADEV_dma_trans_block(kernel, physical_mem, physical_ddr, bytes, 0));

        ljmkMUTEX_UNLOCK(cdmaMutex);
    }
    else
    {
        if (Dir == ljvDMA_DIR_DDR_TO_MEM)
        {
            Reason = ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM;
        }
        else if (Dir == ljvDMA_DIR_MEM_TO_DDR)
        {
            Reason = ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL;
        }
        else
        {
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
        info.src_node = Node_Src;
        info.dst_node = Node_Dst;
        info.offset = 0;
        info.bytes = TransLen;
        info.reason = Reason;
        //ljmkONERROR(ljkEDMA_SyncMemory(Object, Node_Dst, Node_Src, Reason, TransLen, 0));
        ljmkONERROR(ljkEDMA_dmaCopy(Object, &info));
    }

OnError:
    return status;
}

ljeSTATUS ljkEDMA_Exit(void)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsEDMA_CHAN *vchan = ljvNULL;
    ljtUINT index;

    //printk(KERN_ERR "----ljkEDMA_Exit: enter----\n");

    mutex_lock(&galEDMA->dmaMutex);

    for (index = 0; index < galEDMA->edma_dev->wchan_count; index++)
    {
        if  (galEDMA->wrChan[index].state == ljvDMA_USED)
        {
            vchan = &galEDMA->wrChan[index];

            /* terminate EDMA channel */
            ljmkONERROR(ljkDMAENGINE_TerminateAll(vchan->dmaChan));
        }
    }

    for (index = 0; index < galEDMA->edma_dev->rchan_count; index++)
    {
        if  (galEDMA->rdChan[index].state == ljvDMA_USED)
        {
            vchan = &galEDMA->rdChan[index];

            /* terminate EDMA channel */
            ljmkONERROR(ljkDMAENGINE_TerminateAll(vchan->dmaChan));
        }
    }
    mutex_unlock(&galEDMA->dmaMutex);
    //printk(KERN_ERR "----ljkEDMA_Exit: exit----\n");

    return ljvSTATUS_OK;

OnError:
    mutex_unlock(&galEDMA->dmaMutex);
    printk(KERN_ERR "EDMA exit with %d\n", status);
    return status;
}
