/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_platform.h"
#include "lj_hal_kernel_platform_gp.h"
#include "ljmicro_family.h"
#include "ljmicro_proc.h"

#define _LJ_OBJ_ZONE    ljvZONE_DEVICE

ljeSTATUS
_AdjustParam(
    IN ljsPLATFORM *Platform,
    OUT ljsMODULE_PARAMETERS *Args
    );

ljeSTATUS
_GetGPUPhysical(
    IN ljsPLATFORM * Platform,
    IN ljtPHYS_ADDR_T CPUPhysical,
    OUT ljtPHYS_ADDR_T *GPUPhysical
    );

ljeSTATUS
_GetCPUPhysical(
    IN ljsPLATFORM * Platform,
    IN ljtPHYS_ADDR_T GPUPhysical,
    OUT ljtPHYS_ADDR_T * CPUPhysical
    );


static struct _ljsPLATFORM_OPERATIONS gp_ops =
{
    .adjustParam = _AdjustParam,
    .getGPUPhysical = _GetGPUPhysical,
    .getCPUPhysical = _GetCPUPhysical,
//    .syncMemory = ljkEDMA_SyncMemory,
    .dmaCopy            = ljkEDMA_dmaCopy,
    .dmaOperation = ljkDMA_Operaton,
    .dmaExit = ljkEDMA_Exit,
};

#if USE_LINUX_PCIE

struct _ljsPLATFORM_PCIE gp_platform =
{
    .base =
    {
        .name = __FILE__,
        .ops  = &gp_ops,
    },
};

void
_QueryBarInfo(
    struct pci_dev *Pdev,
    ljtPHYS_ADDR_T *BarAddr,
    ljtSIZE_T *BarSize,
    ljtUINT BarNum
    )
{
    ljtUINT addr;
    ljtUINT size;

    /* Read the bar address */
    if (pci_read_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, &addr) < 0)
    {
        return;
    }

    /* Read the bar size */
    if (pci_write_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, 0xffffffff) < 0)
    {
        return;
    }

    if (pci_read_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, &size) < 0)
    {
        return;
    }

    size &= 0xfffffff0;
    size  = ~size;
    size += 1;

    /* Write back the bar address */
    if (pci_write_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, addr) < 0)
    {
        return;
    }

    ljmkPRINT("Bar%d addr=0x%x size=0x%x", BarNum, addr, size);
    //printk("----Bar%d addr=0x%x size=0x%x----\n", BarNum, addr, size);

    *BarAddr = addr;
    *BarSize = size;
}

#else

static struct _ljsPLATFORM gp_platform =
{
    .name = __FILE__,
    .ops  = &gp_ops,
};
#endif

int gp200_pre_init(int card, struct pci_dev *pdev, unsigned long flags, unsigned long *pddr_size);

ljm_board_info *ljm_get_board_info(int card);

ljeSTATUS
_AdjustParam(
    IN ljsPLATFORM *Platform,
    OUT ljsMODULE_PARAMETERS *Args
    )
{
#if USE_LINUX_PCIE
    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)Platform;
    struct pci_dev *pdev = pcie_platform->pcie_info[0].pdev;
    unsigned long flags = pcie_platform->pcie_info[0].flags;
    ljtSIZE_T vram_size;
    unsigned int core = 0;
    unsigned int core_count = 0;
    unsigned int core_2d_count = 0;
    unsigned int core_2d = 0;
    unsigned int pdev_index, dev_index = 0;
    unsigned int index = 0;
    unsigned int hw_dev_index = 0;

    //printk("----enter _AdjustParam----\n");
    if (!pdev)
    {
        /* in z1 we need to use wdz_drm, so we need to add args in insmod ljmcore. */
        pcie_platform->pcie_info[0].bar[0].base = Args->externalBase[0] + VRAM_RESERVED_FB_SIZE;
        pcie_platform->pcie_info[0].bar[0].size = Args->externalSize[0]
            - VRAM_RESERVED_FB_SIZE
            - VRAM_FIX_PCI_DMA_SIZE
            - VRAM_FIX_INST_SIZE
            - VRAM_FIX_VPU_SIZE;
        //Platform->flagBits = ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS;
        //printk("\n----_AdjustParam:pdev is NULL, base:0x%llx, size:0x%lx----\n",
        //        pcie_platform->pcie_info[0].bar[0].base, pcie_platform->pcie_info[0].bar[0].size);
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    for (pdev_index = 0; pdev_index < pcie_platform->device_number; pdev_index++) {
        struct pci_dev *pcie_dev = pcie_platform->pcie_info[pdev_index].pdev;
        unsigned long pcie_flags = pcie_platform->pcie_info[pdev_index].flags;

        Args->devices[dev_index] = &pcie_dev->dev;
        Args->platformIDs[dev_index] = pdev_index;

        if ((pcie_flags & LJMICRO_FAMILY_MASK) == CHIP_GP300)
            Args->hwDevCounts[dev_index] = 2;   /* maybe two device per platform for GP301 */
        else
            Args->hwDevCounts[dev_index] = 1;   /* one device per platform for GP201 */

        hw_dev_index += Args->hwDevCounts[dev_index];
        //printk("----hw_dev_index=%d index=%d----\n", hw_dev_index, index);

        for (; index < hw_dev_index; index++) {
            Args->devMemIDs[index] = pdev_index;    /* by hym on 20220914 */

            /*
             * 3D cores
             */
#if ljdENABLE_3D
            if (pcie_flags & LJMICRO_HAS_3D1) {
                ljm_board_info *boardinfo = ljm_get_board_info(index);
                if (boardinfo->vendor_id == LJM_VENDOR_ID && boardinfo->core_3d_cnt == 0)
                    Args->devCoreCounts[index] = 1; /* one 3D core */
                else
                    Args->devCoreCounts[index] = 2; /* two 3D cores */
            }
            else if (pcie_flags & LJMICRO_HAS_3D)
                Args->devCoreCounts[index] = 1; /* one 3D core */
            else {
                Args->devCoreCounts[index] = 0; /* no 3D core */
                Args->irqs[core] = -1;
            }
#else
            Args->devCoreCounts[index] = 0; /* no 3D core */
            Args->irqs[core] = -1;
#endif

            core_count += Args->devCoreCounts[index];
            //printk("----core_count=%d core=%d----\n", core_count, core);

            Args->regOffsets[core] = REG_3D_OFFSET; /* reg offset of first 3D core is 0x40000 */
            for (; core < core_count; core++) {
                Args->registerBasesMapped[core] =
                    pcie_platform->pcie_info[pdev_index].bar[PCI_BAR_2].logical
                    + Args->regOffsets[core];

                Args->registerBases[core] =
                    pcie_platform->pcie_info[pdev_index].bar[PCI_BAR_2].base
                    + Args->regOffsets[core];

                if (ljmBITTEST(Args->isrPoll, core) != 0)
                    Args->irqs[core] = -1;
                else
#if !MSI_MULTI_VECTOR_TEST
                    Args->irqs[core] = pcie_dev->irq;
#else
                    Args->irqs[core] = pci_irq_vector(pdev, 3);
#endif

                /* check core valid */
                if (read_mreg32(Args->registerBasesMapped[core], REG_CHIP_ID_ADR) == 0) {
                    Args->irqs[core] = -1;
                    printk("3D %d is invalid\n", core);
                }

                //printk("----registerBasesMapped[%d]=%lx registerBases[%d]=%lx irqs[%d]=%d----\n",
                //    core, (unsigned long)Args->registerBasesMapped[core],
                //    core, (unsigned long)Args->registerBases[core],
                //    core, Args->irqs[core]);

                Args->regOffsets[core + 1] = Args->regOffsets[core] + REG_3D_SIZE;
            }

            /*
             * 2D cores
             */
#if ljdENABLE_2D
            if (pcie_flags & LJMICRO_HAS_2D)
                Args->dev2DCoreCounts[index] = 1;   /* 2D core */
            else {
                Args->dev2DCoreCounts[index] = 0;   /* no 2D core */
                Args->irq2Ds[core_2d] = -1;
            }
#else
            Args->dev2DCoreCounts[index] = 0;   /* no 2D core */
            Args->irq2Ds[core_2d] = -1;
#endif

            core_2d_count += Args->dev2DCoreCounts[index];
            //printk("----core_2d_count=%d core_2d=%d----\n", core_2d_count, core_2d);

            Args->reg2DOffsets[core_2d] = REG_2D_OFFSET; /* reg offset of first 2D core is 0xc0000 */
            for (; core_2d < core_2d_count; core_2d++) {
                Args->register2DBasesMapped[core_2d] =
                    pcie_platform->pcie_info[pdev_index].bar[PCI_BAR_2].logical
                    + Args->reg2DOffsets[core_2d];

                Args->register2DBases[core_2d] =
                    pcie_platform->pcie_info[pdev_index].bar[PCI_BAR_2].base
                    + Args->reg2DOffsets[core_2d];

#if !MSI_MULTI_VECTOR_TEST
                Args->irq2Ds[core_2d] = pcie_dev->irq;
#else
                Args->irq2Ds[core_2d] = pci_irq_vector(pdev, 1);
#endif

                /* check core valid */
                if (read_mreg32(Args->register2DBasesMapped[core_2d], REG_CHIP_ID_ADR) == 0) {
                    Args->irq2Ds[core_2d] = -1;
                    printk("2D %d is invalid\n", core_2d);
                }

                //printk("----Args->register2DBasesMapped[%d]=%lx Args->irq2Ds[%d]=%d----\n",
                //    core_2d, (unsigned long)Args->register2DBasesMapped[core_2d], core_2d, Args->irq2Ds[core_2d]);

                Args->reg2DOffsets[core_2d] += REG_2D_SIZE;
            }
        }

        /* visible VRAM */
        Args->externalBase[pdev_index] = pcie_platform->pcie_info[pdev_index].bar[0].base
            + ((flags & LJMICRO_RSV_FB) ? VRAM_RESERVED_FB_SIZE : 0);
        Args->externalSize[pdev_index] = pcie_platform->pcie_info[pdev_index].bar[0].size
            - ((flags & LJMICRO_RSV_FB) ? VRAM_RESERVED_FB_SIZE : 0)    /* reserved framebuffer */
            - ((flags & LJMICRO_IS_BUS32) ? VRAM_FIX_GART_PAGE_SIZE : 0)
            - VRAM_FIX_PCI_DMA_SIZE
            - VRAM_FIX_INST_SIZE
            - VRAM_FIX_VPU_SIZE;
        ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                       "externalBase[%d]=%llx externalSize[%d]=%zx\n",
                       pdev_index, Args->externalBase[pdev_index],
                       pdev_index, Args->externalSize[pdev_index]
                       );

        /* invisible VRAM */
#if 1
        Args->exclusiveBase[pdev_index] = VRAM_INVISIBLE_FLAG + pcie_platform->pcie_info[pdev_index].gpu_vram_start
            + pcie_platform->pcie_info[pdev_index].bar[0].size
            + ((flags & LJMICRO_IS_S2C) ?  VRAM_INVISIBLE_BAD_SIZE : 0);
#else   /*die1 local vram*/
        //INVISBLE DIE1 16GB-32GB
        Args->exclusiveBase[pdev_index] = pcie_platform->pcie_info[pdev_index].gpu_vram_start + 0x400000000;
#endif
        vram_size = min(pcie_platform->pcie_info[pdev_index].gpu_vram_size, 0xC0000000UL);   /* VRAM size is top to 3G */
        if (vram_size > pcie_platform->pcie_info[pdev_index].bar[0].size + ((flags & LJMICRO_IS_S2C) ?  VRAM_INVISIBLE_BAD_SIZE : 0)) {
            Args->exclusiveSize[pdev_index] = vram_size - pcie_platform->pcie_info[pdev_index].bar[0].size
                - ((flags & LJMICRO_IS_S2C) ?  VRAM_INVISIBLE_BAD_SIZE : 0);
        } else {
            Args->exclusiveSize[pdev_index] = 0;
        }
        ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                       "exclusiveBase[%d]=%llx exclusiveSize[%d]=%zx\n",
                       pdev_index, Args->exclusiveBase[0],
                       pdev_index, Args->exclusiveSize[0]
                       );

        dev_index++;
    }

    Args->contiguousBase = 0;
    Args->contiguousSize = 0;

    /* add by hym on 20220914 */
    Args->contiguousSizes[0] = 0;
    Args->devSysMemIDs[0] = 0;
    Args->devSRAMIDs[0] = 0;

    Args->mmuPageTablePool = 0; /* mmuPageTable: 0 means auto (in vram), 1 means virtual */
#if (defined CONFIG_MIPS || defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
    Args->powerManagement = 0;  /* 3D core would hang at the first frame of 3D test (glmark2), if enabled for 3A6000 */
#else
    Args->powerManagement = 0;  /* 压力测试下，部分显卡会概率hang，如GP201A-MXM-11显卡 */
#endif

#endif

    //printk("----leave _AdjustParam----\n");
    return ljvSTATUS_OK;
}

#ifndef LJM_MEMIO_BANKS
#define LJM_MEMIO_BANKS    8

struct ljm_memio_bank {
    unsigned long start;        /* cpu physical address start */
    unsigned long end;          /* cpu physical address end */
    unsigned long gpu_start;    /* gpu physical address start */
};

struct ljm_memio_info {
    int nr_banks;               /* count of banks */
    struct ljm_memio_bank bank[LJM_MEMIO_BANKS];
};
#endif

#define GTT_BANK_BLOCK_SIZE 0x2000000       /* 32MB block size */
#define GTT_BANK_BLOCK_MASK 0xfffffe000000  /* 16MB block mask */

struct ljm_memio_info g_gtt_memio_info = {};

/* get gtt memio info */
void gp100_get_memio_info(struct ljm_memio_info *gtt_memio_info)
{
    struct resource *iomem = iomem_resource.child;
    unsigned long left_memio = (unsigned long)VRAM_FIX_GART_PAGE_SIZE / 4 * GPU_PAGE_SIZE;
    int i;
    unsigned long total_size;
    unsigned long cur_size;

    memset((void*)gtt_memio_info, 0, sizeof(gtt_memio_info));

    //printk(KERN_NOTICE "--physical memory table (0x%lx) ----\n", left_memio);
    while (iomem) {
        //printk(KERN_NOTICE "----start=0x%llx end=0x%llx flags=0x%lx name=%s\n",
        //        iomem->start, iomem->end, iomem->flags, iomem->name);
        if ((iomem->flags & IORESOURCE_MEM) && (iomem->name && strcmp("System RAM", iomem->name) == 0)) {
            //printk("----index %d: start=0x%llx end=0x%llx flags=0x%lx name=%s\n",
            //    gtt_memio_info->nr_banks, iomem->start, iomem->end, iomem->flags, iomem->name);
            //if (gtt_memio_info->nr_banks > 0) {
            //    printk("----rank %d: start=0x%lx end=0x%lx left_memio=0x%lx\n",
            //        gtt_memio_info->nr_banks-1, gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].start,
            //        gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].end,
            //        left_memio);
            //}

            iomem->start &= GTT_BANK_BLOCK_MASK;    /* align to GTT_BANK_BLOCK_SIZE */
            if (iomem->start >= (unsigned long)0x100000000) break;

            /* check and merge into the former bank */
            if (gtt_memio_info->nr_banks > 0
                && iomem->start <= gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].end + 1) {
                //printk("----merge into rank %d\n", gtt_memio_info->nr_banks - 1);
                /* align to GTT_BANK_BLOCK_SIZE */

                left_memio += gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].end - gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].start + 1;
                cur_size = (iomem->end - gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].start + GTT_BANK_BLOCK_SIZE) & GTT_BANK_BLOCK_MASK;
                gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].end = gtt_memio_info->bank[gtt_memio_info->nr_banks - 1].start + cur_size - 1;
                left_memio -= cur_size;
            }
            else {
                gtt_memio_info->bank[gtt_memio_info->nr_banks].start = iomem->start;
                if (gtt_memio_info->bank[gtt_memio_info->nr_banks].start + left_memio >= (unsigned long)0x100000000) {    /* over 4GB, just add largest bank */
                    gtt_memio_info->bank[gtt_memio_info->nr_banks].end = (unsigned long)0x100000000 - 1;
                    gtt_memio_info->nr_banks++;
                    break;
                }
                else {
                    cur_size = (iomem->end - gtt_memio_info->bank[gtt_memio_info->nr_banks].start + GTT_BANK_BLOCK_SIZE) & GTT_BANK_BLOCK_MASK;
                    gtt_memio_info->bank[gtt_memio_info->nr_banks].end = gtt_memio_info->bank[gtt_memio_info->nr_banks].start + cur_size - 1;

                    if (cur_size >= left_memio) {
                        gtt_memio_info->bank[gtt_memio_info->nr_banks].end = gtt_memio_info->bank[gtt_memio_info->nr_banks].start + left_memio - 1;
                        gtt_memio_info->nr_banks++;
                        break;
                    }
                    else {
                        left_memio -= cur_size;
                        gtt_memio_info->nr_banks++;
                    }
                }
            }
        }

        if (left_memio == 0) break;

        // skip children
        //if (iomem->child)
        //    iomem = iomem->child;
        //while (!iomem->sibling && iomem->parent)
        //    iomem = iomem->parent;
        iomem = iomem->sibling;
    }

    for (i=0, total_size = 0; i<gtt_memio_info->nr_banks; i++) {
//      printk(KERN_NOTICE "memio: rank %d start=0x%lx end=0x%lx\n",
//      i, gtt_memio_info->bank[i].start, gtt_memio_info->bank[i].end);
        total_size += gtt_memio_info->bank[i].end - gtt_memio_info->bank[i].start + 1;
    }
    //printk(KERN_NOTICE "memio (low 4GB): total_size=0x%lx--------------------\n", total_size);
}

#if !LINEAR_GART
inline int ljm_gtt_get_gpu_phys(ljtPHYS_ADDR_T CPUPhysical, ljtPHYS_ADDR_T *GPUPhysical)
{
    int i;
    for (i=0; i<g_gtt_memio_info.nr_banks; i++) {
        if (CPUPhysical >= g_gtt_memio_info.bank[i].start && CPUPhysical <= g_gtt_memio_info.bank[i].end) {
            *GPUPhysical = CPUPhysical - g_gtt_memio_info.bank[i].start + g_gtt_memio_info.bank[i].gpu_start;
            return 0;
        }
    }
    return -1;
}

inline int ljm_gtt_get_cpu_phys(ljtPHYS_ADDR_T GPUPhysical, ljtPHYS_ADDR_T *CPUPhysical)
{
    int i;
    unsigned long size;

    for (i=0; i<g_gtt_memio_info.nr_banks; i++) {
        size = g_gtt_memio_info.bank[i].end - g_gtt_memio_info.bank[i].start + 1;
        if (GPUPhysical >= g_gtt_memio_info.bank[i].gpu_start && GPUPhysical < g_gtt_memio_info.bank[i].gpu_start + size) {
            *CPUPhysical = GPUPhysical - g_gtt_memio_info.bank[i].gpu_start + g_gtt_memio_info.bank[i].start;
            return 0;
        }
    }
    return -1;
}

bool ljm_gtt_is_valid(unsigned long Physical)
{
    int i;

    for (i=0; i<g_gtt_memio_info.nr_banks; i++) {
        if (Physical >= g_gtt_memio_info.bank[i].start && Physical <= g_gtt_memio_info.bank[i].end)
            return true;
    }

    return false;
}
#endif

int gp200_get_gpu_phys(u64 cpu_physical, u64 *gpu_physical);
int gp200_get_cpu_phys(u64 gpu_physical, u64 *cpu_physical);

ljeSTATUS
_GetGPUPhysical(
    IN ljsPLATFORM * Platform,
    IN ljtPHYS_ADDR_T CPUPhysical,
    OUT ljtPHYS_ADDR_T *GPUPhysical
    )
{
#if USE_LINUX_PCIE
    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)Platform;
    unsigned int pdev_index;
    ljtPHYS_ADDR_T base;
    ljtSIZE_T size;

    /* check visible VRAM region */
    for (pdev_index = 0; pdev_index < pcie_platform->device_number; pdev_index++) {
        base = pcie_platform->pcie_info[pdev_index].bar[0].base;
        size = pcie_platform->pcie_info[pdev_index].bar[0].size;
        if (CPUPhysical >= base && CPUPhysical < (base + size)) {
            *GPUPhysical = CPUPhysical - base + pcie_platform->pcie_info[pdev_index].gpu_vram_start;
            goto on_end;
        }
    }

    /* check invisible VRAM region */
    if (CPUPhysical & VRAM_INVISIBLE_FLAG) {
        *GPUPhysical = CPUPhysical & ~VRAM_INVISIBLE_FLAG;
        goto on_end;
    }

    /* check 0 address */
    if (CPUPhysical == 0) {
        printk(KERN_WARNING "unexpected cpu physical address (0x0)\n");
        *GPUPhysical = pcie_platform->pcie_info[0].gpu_vram_start;
        goto on_end;
    }

    /* As CPU physical address is in CPU SDRAM, it should be transformed to GART address (based on HOST_MEM_BASE).
     * GART transforms GPU address based on HOST_MEM_BASE to below of
     */
#if LINEAR_GART
    if(CPUPhysical >= GP200_BUS32_DMA_ADDR_TOP)/*get gpu physical in ljoOS_CPUPhysicalToGPUPhysical*/
    {
        printk(KERN_ERR "error: In LINEAR_GART mode cpu physical(0x%llx) cannot above DMA_ADDR_TOP\n", CPUPhysical);
    }
#else
    for (pdev_index = 0; pdev_index < pcie_platform->device_number; pdev_index++) {
        unsigned long flags = pcie_platform->pcie_info[pdev_index].flags;

        if ((flags & LJMICRO_IS_BUS32)
            && ljm_gtt_get_gpu_phys(CPUPhysical, GPUPhysical) == 0)
        {
            /* found and translated in gart table */
            goto on_end;
        }
    }
#endif
#endif

    gp200_get_gpu_phys(CPUPhysical, GPUPhysical);
    //*GPUPhysical = CPUPhysical;

on_end:
    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "CPUPhysical=%llx -> *GPUPhysical=%llx\n",
                   CPUPhysical, *GPUPhysical);

    return ljvSTATUS_OK;
}

ljeSTATUS
_GetCPUPhysical(
    IN ljsPLATFORM * Platform,
    IN ljtPHYS_ADDR_T GPUPhysical,
    OUT ljtPHYS_ADDR_T *CPUPhysical
    )
{
#if USE_LINUX_PCIE
    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)Platform;
    unsigned int pdev_index = 0;
    ljtPHYS_ADDR_T base;
    ljtSIZE_T size;
    unsigned long flags;

    if (unlikely(pcie_platform->device_number > 1)) {
        printk(KERN_ERR "As device_number (%d) is over 1, _GetCPUPhysical is untrustwrothy\n",
            pcie_platform->device_number);
    }

    base = pcie_platform->pcie_info[pdev_index].bar[0].base;
    size = pcie_platform->pcie_info[pdev_index].bar[0].size;
    flags = pcie_platform->pcie_info[pdev_index].flags;
    if (GPUPhysical >= pcie_platform->pcie_info[pdev_index].gpu_vram_start
        && GPUPhysical < (pcie_platform->pcie_info[pdev_index].gpu_vram_start + size)) {
        /* in visible vram */
        *CPUPhysical = GPUPhysical - pcie_platform->pcie_info[pdev_index].gpu_vram_start + base;
    } else if ((flags & LJMICRO_IS_BUS32)
            && GPUPhysical >= GP200_BUS32_HOST_MEM_START
            && GPUPhysical < GP200_BUS32_HOST_MEM_START + GP200_BUS32_HOST_MEM_SIZE) {
#if !LINEAR_GART
        /* GP200 FPGA with 32-bit bus: in cpu dma memory, use gart table to translate. readonly */
        if (ljm_gtt_get_cpu_phys(GPUPhysical, CPUPhysical) != 0)
            printk(KERN_ERR "bug1: wrong GPUPhysical(0x%llx)!!!\n", GPUPhysical);
#else
        /* GP100: in cpu dma memory, use line gart table to translate. readonly */
        *CPUPhysical = GPUPhysical - GP100_HOST_MEM_START;
#endif
    } else
#endif
    {
        gp200_get_cpu_phys(GPUPhysical, CPUPhysical);
        //*CPUPhysical = GPUPhysical;
    }

    ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_VIDMEM,
                   "*CPUPhysical=0x%llx <- GPUPhysical=0x%llx",
                   *CPUPhysical, GPUPhysical);

    return ljvSTATUS_OK;
}


#if USE_LINUX_PCIE
static const struct pci_device_id ljm_pci_ids[] = {
    GP_PCI_IDS
};

MODULE_DEVICE_TABLE(pci, ljm_pci_ids);

static int get_bar_resource(struct _ljsPCIEInfo *pcie, uint32_t id, bool virtual)
{
    if (!pcie || !pcie->pdev || id > 5)
    {
        return -EINVAL;
    }

    pcie->bar[id].base = pci_resource_start(pcie->pdev, id);
    pcie->bar[id].size = pci_resource_len(pcie->pdev, id);
    if (virtual)
    {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
        pcie->bar[id].logical = ioremap(pcie->bar[id].base, pcie->bar[id].size);
#else
        pcie->bar[id].logical = ioremap_nocache(pcie->bar[id].base, pcie->bar[id].size);
#endif
        //pcie->bar[id].logical = pci_iomap(pcie->pdev, pcie->bar[id].base, pcie->bar[id].size);
        if (!pcie->bar[id].logical)
        {
            printk(KERN_ERR "map bar %d failed.\n", id);
            return -ENODEV;
        }
        //printk(KERN_INFO "map bar %d to 0x%lx\n", id, (unsigned long)pcie->bar[id].logical);
    }

    return 0;
}

int ljm_ddr_test(void *vram_base, resource_size_t vram_start, size_t vram_size);
int ljm_ddr_set(void *vram_base, resource_size_t vram_start, size_t vram_size, unsigned int value);

static void put_bar_resource(struct _ljsPCIEInfo *pcie, uint32_t id)
{
    if (!pcie || !pcie->pdev || id > 5)
    {
        return;
    }

    if (pcie->bar[id].logical)
    {
        iounmap(pcie->bar[id].logical);
        //pci_iounmap(pcie->pdev, pcie->bar[id].logical);
    }

    pcie->bar[id].base = 0;
    pcie->bar[id].size = 0;
    pcie->bar[id].logical = 0;
}

int pcie_edma_init(int card)
{
    int ret;
    ljeSTATUS status;
    unsigned int cur_device = card;
    unsigned long flags = gp_platform.pcie_info[cur_device].flags;
    u64 edma_link_table_offset = gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].size
        - ((flags & LJMICRO_IS_BUS32) ? VRAM_FIX_GART_PAGE_SIZE : 0)
        - VRAM_FIX_PCI_DMA_SIZE;

    //printk("----enter pcie_edma_init----\n");

    gp_platform.dma[cur_device].pdev = gp_platform.pcie_info[cur_device].pdev;
#if DMA_NEW
    gp_platform.dma[cur_device].reg_base = gp_platform.pcie_info[cur_device].bar[PCI_BAR_5].logical;
    gp_platform.dma[cur_device].reg_base_2 = gp_platform.pcie_info[cur_device].bar[PCI_BAR_2].logical;
    //printk("reg_base:0x%llx; reg_base_2:0x%llx; \n", (u64)gp_platform.dma[cur_device].reg_base, (u64)gp_platform.dma[cur_device].reg_base_2);
#else
    gp_platform.dma[cur_device].reg_base = gp_platform.dma[cur_device].pdev;
#endif
#if !MSI_MULTI_VECTOR_TEST
    gp_platform.dma[cur_device].irq = gp_platform.pcie_info[cur_device].pdev->irq;
#else
    gp_platform.dma[cur_device].irq = pci_irq_vector(gp_platform.dma[cur_device].pdev, 2);
#endif
    gp_platform.dma[cur_device].plink_table_base = gp_platform.pcie_info[cur_device].gpu_vram_start + edma_link_table_offset;
#if (defined CONFIG_MIPS || defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
    gp_platform.dma[cur_device].vlink_table_base = ioremap_nocache(
#else
    gp_platform.dma[cur_device].vlink_table_base = ioremap_wc(
#endif
    //gp_platform.dma.vlink_table_base = pci_iomap(gp_platform.pcie_info[cur_device].pdev,
    gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].base + edma_link_table_offset, VRAM_FIX_PCI_DMA_SIZE);
    //printk("nc:pbase: 0x%llx; vlink_table_base: 0x%llx\n",gp_platform.dma[cur_device].plink_table_base, (u64)gp_platform.dma[cur_device].vlink_table_base);
    if (!gp_platform.dma[cur_device].vlink_table_base)
    {
        printk(KERN_ERR "edma: map vlink_table failed\n");
        return -1;
    }

    gp_platform.dma[cur_device].hdma_flag = flags & LJMICRO_HAS_HDMA;
    //gp_platform.dma[cur_device].single_chan = flags & LJMICRO_IS_FPGA;  /**< fpga must set this define, fpga channel is 2+2 */
    gp_platform.dma[cur_device].single_chan = (flags & LJMICRO_IS_FPGA)||(gp_platform.dma[cur_device].hdma_flag);  /**< GP201 set single channel for more stable on FT D2000 */
    ret = edma_device_probe(&gp_platform.dma[cur_device]);
    if (ret)
    {
        printk(KERN_ERR "edma: edma_device_probe failed\n");
        return ret;
    }

    status = ljkEDMA_DeviceInit(&gp_platform.dma[cur_device]);
    if (status != ljvSTATUS_OK)
    {
        printk(KERN_ERR "edma: ljkEDMA_DeviceInit failed\n");
        edma_device_remove(&gp_platform.dma[cur_device]);
        return -1;
    }

#if 0   /* move to the end of 3D/2D initialization by hym on 20230428 */
    /* enable EDMA interrupt */
    if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP200) {
        //printk("----enable EDMA and AMBA DMA interrupt----\n");
        LJM_MISC_W32(cur_device, MISC_INTR_EN, LJM_MISC_R32(cur_device, MISC_INTR_EN) | m_INTR_EN_EDMA);
    }
#endif

    //printk("----leave pcie_edma_init----\n");
    return 0;
}

static void pcie_edma_remove(int card)
{
    unsigned int cur_device = card;
    unsigned long flags = gp_platform.pcie_info[cur_device].flags;

    /* disable EDMA interrupt */
    if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP200) {
        //printk("----disable EDMA and AMBA DMA interrupt----\n");
        LJM_MISC_W32(card, MISC_INTR_EN, LJM_MISC_R32(card, MISC_INTR_EN) & ~m_INTR_EN_EDMA);
    }

    if (gp_platform.dma[cur_device].vlink_table_base)
    {
        iounmap(gp_platform.dma[cur_device].vlink_table_base);
        //pci_iounmap(pcie_platform->pcie_info[cur_device].pdev, gp_platform.dma.vlink_table_base);
    }
    ljkEDMA_DeviceExit();

    edma_device_remove(&gp_platform.dma[cur_device]);
}

static int gpu_sub_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
    static u64 dma_mask;
    unsigned long flags = ent->driver_data;
    int r = 0;
    unsigned int cur_device = gp_platform.device_number++;
    unsigned long ddr_size = 0xC0000000;    /* default 3GB */
#if MSI_MULTI_VECTOR_TEST
    int nr_irqs;
#endif

    ljmkHEADER_ARG("cur_device=%d pdev=%p flags=%lx pid=%d(%s)",
        cur_device, pdev, flags, _GetProcessID(), _GetProcessComm());

    //printk("----enter gpu_sub_probe(cur_device=%d pdev=%p flags=%lx pid=%d(%s))----\n",
    //    cur_device, pdev, flags, _GetProcessID(), _GetProcessComm());

    ljmkPRINT("PCIE DRIVER PROBED");

    r = pci_enable_device(pdev);
    if (r) {
        printk(KERN_ERR "ljmcore: pci_enable_device() failed.\n");
        ljmkFOOTER_ARG("r=%d", r);
        return r;
    }

    if (flags & LJMICRO_IS_BUS32) { /*LJMICRO_IS_BUS32*/
        dma_mask = DMA_BIT_MASK(32);
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
        if ((r = pci_set_dma_mask(pdev, dma_mask)) && (r = pci_set_consistent_dma_mask(pdev, dma_mask))) {
#else
        if ((r = dma_set_mask(&pdev->dev, dma_mask)) && (r = dma_set_coherent_mask(&pdev->dev, dma_mask))) {
#endif
            printk(KERN_ERR "ljmcore: Failed to set DMA mask.\n");
            goto error_out;
        }
    }
    else {
        /* try 64 bit DMA mask */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
        dma_mask = DMA_BIT_MASK(64);
#else
        dma_mask = DMA_64BIT_MASK;
#endif
	    //pci_set_consistent_dma_mask(pdev, dma_mask);
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
        if (pci_set_dma_mask(pdev, dma_mask) && pci_set_consistent_dma_mask(pdev, dma_mask)) {
#else
        if (dma_set_mask(&pdev->dev, dma_mask) && dma_set_coherent_mask(&pdev->dev, dma_mask)) {
#endif
            printk(KERN_WARNING "ljmcore: Failed to set 64 bit DMA mask.\n");

            /* try 40 bit DMA mask */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
            dma_mask = DMA_BIT_MASK(40);
#else
            dma_mask = DMA_40BIT_MASK;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
            if (pci_set_dma_mask(pdev, dma_mask) && pci_set_consistent_dma_mask(pdev, dma_mask)) {
#else
            if (dma_set_mask(&pdev->dev, dma_mask) && dma_set_coherent_mask(&pdev->dev, dma_mask)) {
#endif
                printk(KERN_WARNING "ljmcore: Failed to set 40 bit DMA mask.\n");

                /* try 32 bit DMA mask */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
                dma_mask = DMA_BIT_MASK(32);
#else
                dma_mask = DMA_32BIT_MASK;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
                if ((r = pci_set_dma_mask(pdev, dma_mask)) && (r = pci_set_consistent_dma_mask(pdev, dma_mask))) {
#else
                if ((r = dma_set_mask(&pdev->dev, dma_mask)) && (r = dma_set_coherent_mask(&pdev->dev, dma_mask))) {
#endif
                    printk(KERN_ERR "ljmcore: Failed to set DMA mask.\n");
                    goto error_out;
                }
            }
        }
    }

    pci_set_master(pdev);

    //abc: for bar0 will failed to reserve 256M
    //if (pci_request_regions(pdev, "ljmcore")) {
    if (pci_request_selected_regions(pdev, PCI_BAR_2, "ljmcore")) {
        printk(KERN_ERR "ljmcore: Failed to get ownership of BAR region.\n");
    }

    gp_platform.pcie_info[cur_device].pdev = pdev;
    gp_platform.pcie_info[cur_device].flags = flags;
    gp_platform.pcie_info[cur_device].bar[0].available = ljvTRUE;
    gp_platform.pcie_info[cur_device].bar[2].available = ljvTRUE;
    gp_platform.pcie_info[cur_device].bar[5].available = ljvTRUE;

    /* bar 0 for VRAM */
    get_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_0, false);
    //gp_platform.pcie_info[cur_device].bar[0].size = 64UL << 20; /* only for 64MB */

    /* bar 2 for registers */
    r = get_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_2, true);
    if (r)
    {
        goto error_iomap_bar2;
    }

    if (g_mmio_bases[cur_device] == NULL) {
        g_mmio_bases[cur_device] = gp_platform.pcie_info[cur_device].bar[PCI_BAR_2].logical;
    }

    if (0 == cur_device)
        g_mmio_base = gp_platform.pcie_info[0].bar[PCI_BAR_2].logical;

    /* now RREG8/16/32 and WREG8/16/32 are available */

    if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP200) {
        gp200_pre_init(cur_device, pdev, flags, &ddr_size);
    }

    printk(KERN_INFO "ljmcore: irq=%d\n", pdev->irq);
#ifdef CONFIG_MIPS/* LS3A4000 use msi will cpu hang (1+1+1) */
    flags = flags & (~LJMICRO_IS_MSI);
#endif
    if (flags & LJMICRO_IS_MSI)
    {
#if !MSI_MULTI_VECTOR_TEST
        if (pci_enable_msi(pdev))
        {
            printk(KERN_ERR "ljmcore: Failed to enable MSI.\n");
            gp_platform.dma[cur_device].msi_flag = 0;
        }
        else
        {
            gp_platform.dma[cur_device].msi_flag = 1;
        }
        //printk(KERN_INFO "ljmcore: after enable MSI, irq=%d\n", pdev->irq);
#else
        /* IRQs allocation */
        nr_irqs = pci_alloc_irq_vectors(pdev, 1, 32, PCI_IRQ_MSI);
        if (nr_irqs < 1) {
            pci_err(pdev, "fail to alloc IRQ vector (number of IRQs=%u)\n",
            nr_irqs);
            return -EPERM;
        }
        printk("----nr_irqs:%d----\n", nr_irqs);
        printk("----irq:%d----\n", pdev->irq);
        printk("----irq0 default(vga...):%d----\n", pci_irq_vector(pdev, 0));
        printk("----irq1 2d:%d----\n", pci_irq_vector(pdev, 1));
        printk("----irq2 edma(16ch):%d----\n", pci_irq_vector(pdev, 2));
        printk("----irq3 3d:%d----\n", pci_irq_vector(pdev, 3));
        LJM_MISC_W32(cur_device, MISC_INTR_VGA_MSI_MSG_DATA_OFFSET, 0);
        LJM_MISC_W32(cur_device, MISC_INTR_3D_MSI_MSG_DATA_OFFSET, 3);
        LJM_MISC_W32(cur_device, MISC_INTR_2D_MSI_MSG_DATA_OFFSET, 1);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR0_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR1_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR2_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR3_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR4_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR5_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR6_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_WR7_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD0_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD1_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD2_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD3_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD4_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD5_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD6_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_INTR_EDMA_RD7_MSI_MSG_DATA_OFFSET, 2);
        LJM_MISC_W32(cur_device, MISC_MSI_MSG_THROTTLE_US_OFFSET, 20);
#endif
    }
    else
    {
        printk(KERN_ERR "ljmcore: NO MSI int\n");
        gp_platform.dma[cur_device].msi_flag = 0;
    }

#if !LINEAR_GART
    /* make sure to get g_gtt_memio_info */
    if (g_gtt_memio_info.nr_banks == 0)
        gp100_get_memio_info(&g_gtt_memio_info);
#endif

    /* GPU local VRAM */
    gp_platform.pcie_info[cur_device].gpu_vram_start =
        (flags & LJMICRO_IS_BUS32) ? VRAM_GP200_BUS32_GPU_START : VRAM_GP200_BUS40_GPU_START;
    if (flags & LJMICRO_IS_FPGA) {
        gp_platform.pcie_info[cur_device].gpu_vram_size = (flags & LJMICRO_VRAM_2G) ?  0x80000000 : 0x40000000;
        //gp_platform.pcie_info[cur_device].gpu_vram_size = (flags & LJMICRO_VRAM_2G) ?  (64UL << 20) : 0x40000000;   /* only for 64MB */
    } else {
        gp_platform.pcie_info[cur_device].gpu_vram_size = ddr_size;
    }

    switch (flags & LJMICRO_FAMILY_MASK) {
    case CHIP_GP200:
        printk(KERN_INFO "LJ GP200 series: Detected Video RAM %zuM, BAR=%zuM\n",
            gp_platform.pcie_info[cur_device].gpu_vram_size >> 20,
            gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].size >> 20);
        break;
    case CHIP_GP300:
        printk(KERN_INFO "LJ GP300 series: Detected VRAM RAM %zuM, BAR=%zuM\n",
            gp_platform.pcie_info[cur_device].gpu_vram_size >> 20,
            gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].size >> 20);
        break;
    default:
        printk(KERN_INFO "Unkown gpu\n");
    }

    /* bar 5 for dma registers */
    r = get_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_5, true);
    if (r)
    {
        goto error_iomap_bar5;
    }

    /* enable 2D and 3D interrupt */
    if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP200) {
        //printk("----enable 2D and 3D interrupt----\n");
        LJM_MISC_W32(cur_device, MISC_INTR_EN, LJM_MISC_R32(cur_device, MISC_INTR_EN) | m_INTR_EN_G3D0 | m_INTR_EN_G3D1 | m_INTR_EN_G2D);
    }

    if (flags & LJMICRO_IS_FPGA) {
        /** for fpga board with crc we need to init all ddr */
        r = ljm_ddr_set(NULL,
            gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].base,
            gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].size - (((flags & LJMICRO_IS_BUS32)) ? VRAM_FIX_GART_PAGE_SIZE : 0),
            0);
        if (r)
            printk(KERN_ERR "VRAM mem set error\n");
    }

#if 0
    r = ljm_ddr_test(NULL,
        gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].base,
        gp_platform.pcie_info[cur_device].bar[PCI_BAR_0].size - (((flags & LJMICRO_IS_BUS32)) ? VRAM_FIX_GART_PAGE_SIZE : 0));
    if (r) {
        printk(KERN_ERR "VRAM test error\n");
        //goto error_iomap_bar2;
    }
#endif

    /* FIXME: every PCIE device should has an edma, by hym */
    if (cur_device == 0) {
        r = pcie_edma_init(cur_device);
        if (r)
        {
            goto error_edma_init;
        }
    }

    gpuinfo_proc_init(cur_device);

    ljmkFOOTER_NO();
    return 0;

error_edma_init:
    put_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_2);
    put_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_0);
error_iomap_bar2:
    pci_disable_msi(pdev);
    pci_release_selected_regions(pdev, PCI_BAR_2);
    pci_clear_master(pdev);
error_iomap_bar5:
    pci_disable_msi(pdev);
    pci_release_selected_regions(pdev, PCI_BAR_5);
    pci_clear_master(pdev);

error_out:
    pci_disable_device(pdev);

    ljmkFOOTER_ARG("r=%d", r);
    return r;

}

static void gpu_sub_remove(struct pci_dev *pdev)
{
    unsigned int cur_device = gp_platform.device_number - 1;
    unsigned long flags = gp_platform.pcie_info[cur_device].flags;

    ljmkHEADER_ARG("pdev=%p pid=%d(%s)", pdev, _GetProcessID(), _GetProcessComm());

    ljmicro_proc_exit(cur_device);

    /* disable 2D and 3D interrupt */
    if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP200) {
        //printk("----disable 2D and 3D interrupt----\n");
        LJM_MISC_W32(cur_device, MISC_INTR_EN, LJM_MISC_R32(cur_device, MISC_INTR_EN) & ~m_INTR_EN_G3D0 & ~m_INTR_EN_G3D1 & ~m_INTR_EN_G2D);
    }

    put_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_2);
    put_bar_resource(&gp_platform.pcie_info[cur_device], PCI_BAR_0);
    if (0 == cur_device)
        pcie_edma_remove(cur_device);
    pci_set_drvdata(pdev, NULL);
    /* Freeing IRQs */
#if MSI_MULTI_VECTOR_TEST
    pci_free_irq_vectors(pdev);
#else
    pci_disable_msi(pdev);
#endif
    pci_release_selected_regions(pdev, PCI_BAR_2);  // pci_release_regions(pdev);
    pci_clear_master(pdev);
    pci_disable_device(pdev);

    gp_platform.device_number--;    /* decrease device_number at last */

    ljmkFOOTER_NO();
}

static struct pci_driver gpu_pci_subdriver = {
    .name = DEVICE_NAME,
    .id_table = ljm_pci_ids,
    .probe = gpu_sub_probe,
    .remove = gpu_sub_remove
};

#endif

static struct platform_device *gp_platform_device;

int ljkPLATFORM_Init(struct platform_driver *pdrv, struct _ljsPLATFORM **platform)
{
    int ret = 0;

    //printk("----enter ljkPLATFORM_Init----\n");
    gp_platform_device = platform_device_alloc(pdrv->driver.name, -1);

    if (!gp_platform_device) {
        printk(KERN_ERR "ljmcore: platform_device_alloc failed.\n");
        return -ENOMEM;
    }
    //printk("----platform_device_alloc: gp_platform_device=%lx----\n",(unsigned long)gp_platform_device);

    /* Add device */
    ret = platform_device_add(gp_platform_device);
    if (ret) {
        printk(KERN_ERR "ljmcore: platform_device_add failed.\n");
        goto put_dev;
    }

#if 0   /* will be assigned in ljm_dev_probe */
    /* assign platform device to gp_platform. added by hym on 20210329 */
    printk("%s(%d): gp_platform.base.device=%lx gp_platform_device=%lx----",
        __FUNCTION__, __LINE__,
        (unsigned long)gp_platform.base.device, (unsigned long)gp_platform_device);
    gp_platform.base.device = gp_platform_device;
#endif

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "%s(%d): &gp_platform=%lx gp_platform_device=%lx &gp_platform_device->dev=%lx\n",
                   __FUNCTION__, __LINE__,
                   (unsigned long)&gp_platform, (unsigned long)gp_platform_device, (unsigned long)&gp_platform_device->dev
                   );

#if USE_LINUX_PCIE
    ret = pci_register_driver(&gpu_pci_subdriver);
    if (ret)
    {
        //platform_device_unregister(gp_platform_device);
        //return ret;
        printk(KERN_ERR "ljmcore: pci_register_driver failed.\n");
        goto put_dev;
    }
#endif

    //printk("----leave ljkPLATFORM_Init (*platform=%lx)----\n",(unsigned long)&gp_platform);

    *platform = (ljsPLATFORM *)&gp_platform;
    return 0;

put_dev:
    platform_device_put(gp_platform_device);

    return ret;
}

int ljkPLATFORM_Terminate(struct _ljsPLATFORM *platform)
{
#if USE_LINUX_PCIE
    unsigned int              dev_index;
    struct _ljsPLATFORM_PCIE *pcie_platform = (struct _ljsPLATFORM_PCIE *)platform;

    for (dev_index = 0; dev_index < pcie_platform->device_number; dev_index++) {
        unsigned int i;

        for (i = 0; i < ljdMAX_PCIE_BAR; i++) {
            if (pcie_platform->pcie_info[dev_index].bar[i].logical != 0) {
                iounmap(pcie_platform->pcie_info[dev_index].bar[i].logical);
                //pci_iounmap(pcie_platform->pcie_info[dev_index].pdev,
                //            pcie_platform->pcie_info[dev_index].bar[i].logical);
            }
        }
    }

    pci_unregister_driver(&gpu_pci_subdriver);
#endif
    if (gp_platform_device) {
        platform_device_unregister(gp_platform_device);
        gp_platform_device = NULL;
    }

    return 0;
}
