/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_sync_h_
#define __lj_hal_kernel_sync_h_

#include <linux/types.h>

#ifndef CONFIG_SYNC_FILE

/* sync.h is in drivers/staging/android/ for now. */
#include <sync.h>

#include "lj_hal.h"
#include "lj_hal_base.h"

struct ljm_sync_timeline {
    /* Parent object. */
    struct sync_timeline obj;

    /* Timestamp when sync_pt is created. */
    ljtUINT64            stamp;

    /* Pointer to os struct. */
    ljkOS                os;
};

struct ljm_sync_pt {
    /* Parent object. */
    struct sync_pt       pt;

    /* Reference signal. */
    ljtSIGNAL            signal;

    /* Timestamp when sync_pt is created. */
    ljtUINT64            stamp;
};

/* Create ljm_sync_timeline object. */
struct ljm_sync_timeline *ljm_sync_timeline_create(const char *name, ljkOS Os);

/* Create ljm_sync_pt object. */
struct sync_pt *ljm_sync_pt_create(struct ljm_sync_timeline *obj, ljtSIGNAL signal);

#else

#include <linux/sync_file.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)
#include <linux/fence.h>
#include <linux/fence-array.h>
# else
#include <linux/dma-fence.h>
#include <linux/dma-fence-array.h>
# endif

#include "lj_hal.h"
#include "lj_hal_base.h"
#include "lj_hal_kernel_linux.h"

struct ljm_sync_timeline {
    char        name[64];

    /* Parent object. */
    u64         context;

    /* Timestamp when sync_pt is created. */
    atomic64_t  seqno;

    /* Pointer to os struct. */
    ljkOS       os;
};

struct ljm_fence {
    /* must be the first. */
    struct dma_fence          base;
    /* spinlock_t */
    spinlock_t                lock;

    struct ljm_sync_timeline *parent;

    /* link with signal. */
    ljtSIGNAL                 signal;
};

struct ljm_sync_timeline *ljm_sync_timeline_create(const char *name, ljkOS Os);

void ljm_sync_timeline_destroy(struct ljm_sync_timeline *timeline);

struct dma_fence *ljm_fence_create(struct ljm_sync_timeline *timeline, ljsSIGNAL *signal);

#endif /* CONFIG_SYNC_FILE */

#endif /* __lj_hal_kernel_sync_h_ */
