/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef _LJ_HAL_SECURITY_INTERFACE_H_
#define _LJ_HAL_SECURITY_INTERFACE_H_
struct kernel_start_command {
    kernel_packet_command_t command; /*! The command (always needs to be the first entry in a structure). */
    ljtUINT8                gpu;     /*! Which GPU. */
    ljtADDRESS              address;
    ljtUINT32               bytes;
};

/*!
 *  @brief ljkCOMMAND Object requests TrustZone to submit command buffer.
 *  @discussion
 *  Code in trustzone will check content of command buffer after copying command buffer to TrustZone.
 */
struct kernel_submit {
    kernel_packet_command_t command;                /*! The command (always needs to be the first entry in a structure). */
    ljtUINT8                gpu;                    /*! Which GPU. */
    ljtUINT8                kernel_command;         /*! Whether it is a kernel command. */
    ljtUINT32               command_buffer_handle;  /*! Handle to command buffer. */
    ljtUINT32               offset;                 /* Offset in command buffer. */
    ljtUINT32               *command_buffer;        /*! Content of command buffer need to be submit. */
    ljtUINT32               command_buffer_length;  /*! Length of command buffer. */
};

/*!
 *  @brief ljkVIDMEM Object requests TrustZone to allocate security memory.
 *  @discussion
 *  Allocate a buffer from security GPU memory.
 */
struct kernel_allocate_security_memory {
    kernel_packet_command_t command;        /*! The command (always needs to be the first entry in a structure). */
    ljtUINT32               bytes;          /*! Requested bytes. */
    ljtUINT32               memory_handle;  /*! Handle of allocated memory. */
};

/*!
 *  @brief ljkVIDMEM Object requests TrustZone to allocate security memory.
 *  @discussion
 *  Free a video memory buffer from security GPU memory.
 */
struct kernel_free_security_memory {
    kernel_packet_command_t command;        /*! The command (always needs to be the first entry in a structure). */
    ljtUINT32               memory_handle;  /*! Handle of allocated memory. */
};

struct kernel_execute {
    kernel_packet_command_t command;                /*! The command (always needs to be the first entry in a structure). */
    ljtUINT8                gpu;                    /*! Which GPU. */
    ljtUINT8                kernel_command;         /*! Whether it is a kernel command. */
    ljtUINT32               *command_buffer;        /*! Content of command buffer need to be submit. */
    ljtUINT32               command_buffer_length;  /*! Length of command buffer. */
};

typedef struct kernel_map_scatter_gather {
    ljtUINT32                         bytes;
    ljtUINT32                         physical;
    struct kernel_map_scatter_gather  *next;
} kernel_map_scatter_gather_t;

struct kernel_map_memory {
    kernel_packet_command_t      command;
    kernel_map_scatter_gather_t  *scatter;
    ljtUINT32                    *physicals;
    ljtPHYS_ADDR_T               physical;  /*! Contiguous physical address range. */
    ljtUINT32                    pageCount;
    ljtADDRESS                   gpuAddress;
};

struct kernel_unmap_memory {
    ljtADDRESS gpuAddress;
    ljtUINT32  pageCount;
};

struct kernel_read_mmu_exception {
    ljtUINT32 mmuStatus;
    ljtUINT32 mmuException;
};

struct kernel_handle_mmu_exception {
    ljtUINT32      mmuStatus;
    ljtPHYS_ADDR_T physical;
    ljtADDRESS     gpuAddress;
};

typedef struct _ljsTA_INTERFACE {
    kernel_packet_command_t command;
    union {
        struct kernel_submit                   Submit;
        struct kernel_start_command            StartCommand;
        struct kernel_allocate_security_memory AllocateSecurityMemory;
        struct kernel_execute                  Execute;
        struct kernel_map_memory               MapMemory;
        struct kernel_unmap_memory             UnmapMemory;
        struct kernel_read_mmu_exception       ReadMMUException;
        struct kernel_handle_mmu_exception     HandleMMUException;
    } u;
    ljeSTATUS result;
} ljsTA_INTERFACE;

#endif


