/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_statistics_h_
#define __lj_hal_statistics_h_

#define VIV_STAT_ENABLE_STATISTICS              0

/*  Toal number of frames for which the frame time is accounted. We have storage
 *  to keep frame times for last this many frames.
 */
#define VIV_STAT_FRAME_BUFFER_SIZE              30

/*
 *   Total number of frames sampled for a mode. This means
 *
 *   # of frames for HZ Current  : VIV_STAT_EARLY_Z_SAMPLE_FRAMES
 *   # of frames for HZ Switched : VIV_STAT_EARLY_Z_SAMPLE_FRAMES
 * +
 * --------------------------------------------------------
 *                               : (2 * VIV_STAT_EARLY_Z_SAMPLE_FRAMES) frames needed
 *
 *   IMPORTANT: This total must be smaller than VIV_STAT_FRAME_BUFFER_SIZE
 */
#define VIV_STAT_EARLY_Z_SAMPLE_FRAMES          7
#define VIV_STAT_EARLY_Z_LATENCY_FRAMES         2

/*
 * Multiplication factor for previous Hz off mode.
 * Make it more than 1.0 to advertise HZ on.
 */
#define VIV_STAT_EARLY_Z_FACTOR                 (1.05f)

/* HAL statistics information. */
typedef struct _ljsSTATISTICS_EARLYZ {
    ljtUINT                     switchBackCount;
    ljtUINT                     nextCheckPoint;
    ljtBOOL                     disabled;
} ljsSTATISTICS_EARLYZ;

/* HAL statistics information. */
typedef struct _ljsSTATISTICS {
    ljtUINT64                   frameTime[VIV_STAT_FRAME_BUFFER_SIZE];
    ljtUINT64                   previousFrameTime;
    ljtUINT                     frame;
    ljsSTATISTICS_EARLYZ        earlyZ;
} ljsSTATISTICS;

/* Add a frame based data into current statistics. */
void
ljfSTATISTICS_AddData(IN ljeSTATISTICS Key, IN ljtUINT Value);

/* Marks the frame end and triggers statistical calculations and decisions.*/
void
ljfSTATISTICS_MarkFrameEnd(void);

/* Sets whether the dynamic HZ is disabled or not .*/
void
ljfSTATISTICS_DisableDynamicEarlyZ(IN ljtBOOL Disabled);

#endif /*__lj_hal_statistics_h_ */


