/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef _LJ_HAL_TA_H_
#define _LJ_HAL_TA_H_
#include "lj_hal_types.h"
#include "lj_hal_security_interface.h"

#ifdef __cplusplus
extern "C" {
#endif


typedef struct _ljtaOS          * ljtaOS;
typedef struct _ljTA            * ljTA;

typedef struct _ljTA_HARDWARE   * ljTA_HARDWARE;
typedef struct _ljTA_MMU        * ljTA_MMU;

/*
    Trust Application is a object needed to be created as a context in trust zone.
    One client for a core.
*/
typedef struct _ljTA {
    /* ljtaOS object */
    ljtaOS          os;

    ljeCORE         core;

    ljTA_MMU        mmu;

    ljTA_HARDWARE   hardware;

    ljtBOOL         destoryMmu;
} ljsTA;

typedef struct _ljTA_MMU
{
    ljtaOS          os;

    ljtSIZE_T       mtlbBytes;
    ljtPOINTER      mtlbLogical;
    ljtPHYS_ADDR    mtlbPhysical;

    ljtPOINTER      stlbs;

    ljtPOINTER      safePageLogical;
    ljtPHYS_ADDR    safePagePhysical;

    ljtPOINTER      nonSecureSafePageLogical;
    ljtPHYS_ADDR    nonSecureSafePagePhysical;

    ljtPOINTER      mutex;
}
ljsTA_MMU;

ljeSTATUS HALDECL
TAEmulator(
    ljeCORE Core,
    void * Interface
    );

int
ljTA_Construct(
    IN ljtaOS Os,
    IN ljeCORE Core,
    OUT ljTA *TA
);

int
ljTA_Destroy(
    IN ljTA TA
);

int
ljTA_Dispatch(
    IN ljTA TA,
    IN OUT ljsTA_INTERFACE * Interface
);

/*************************************
* Porting layer
*/

ljeSTATUS
ljtaOS_ConstructOS(
    IN ljkOS Os,
    OUT ljtaOS *TAos
    );

ljeSTATUS
ljtaOS_DestroyOS(
    IN ljtaOS Os
    );

ljeSTATUS
ljtaOS_Allocate(
    IN ljtUINT32 Bytes,
    OUT ljtPOINTER *Pointer
    );

ljeSTATUS
ljtaOS_Free(
    IN ljtPOINTER Pointer
    );

ljeSTATUS
ljtaOS_AllocateSecurityMemory(
    IN ljtaOS Os,
    IN ljtSIZE_T *Bytes,
    OUT ljtPOINTER *Logical,
    OUT ljtPOINTER *Physical
    );

ljeSTATUS
ljtaOS_FreeSecurityMemory(
    IN ljtaOS Os,
    IN ljtSIZE_T  Bytes,
    IN ljtPOINTER Logical,
    OUT ljtPOINTER Physical
    );

ljeSTATUS
ljtaOS_AllocateNonSecurityMemory(
    IN ljtaOS Os,
    IN ljtSIZE_T *Bytes,
    OUT ljtPOINTER *Logical,
    OUT ljtPOINTER *Physical
    );

ljeSTATUS
ljtaOS_FreeNonSecurityMemory(
    IN ljtaOS Os,
    IN ljtSIZE_T  Bytes,
    IN ljtPOINTER Logical,
    OUT ljtPOINTER Physical
    );



ljeSTATUS
ljtaOS_GetPhysicalAddress(
    IN ljtaOS Os,
    IN ljtPOINTER Logical,
    OUT ljtPHYS_ADDR_T * Physical
    );

ljeSTATUS ljtaOS_WriteRegister(
    IN ljtaOS Os, IN ljeCORE Core,
    IN ljtUINT32 Address,
    IN ljtUINT32 Data
    );

ljeSTATUS ljtaOS_ReadRegister(
    IN ljtaOS Os, IN ljeCORE Core,
    IN ljtUINT32 Address,
    IN ljtUINT32 *Data
    );

ljeSTATUS
ljtaOS_MemCopy(
    IN ljtUINT8_PTR Dest,
    IN ljtUINT8_PTR Src,
    IN ljtUINT32 Bytes
    );

ljeSTATUS
ljtaOS_ZeroMemory(
    IN ljtUINT8_PTR Dest,
    IN ljtUINT32 Bytes
    );

void
ljtaOS_CacheFlush(
    IN ljtUINT8_PTR Dest,
    IN ljtUINT32 Bytes
    );

void
ljtaOS_CacheClean(
    IN ljtUINT8_PTR Dest,
    IN ljtUINT32 Bytes
    );

void
ljtaOS_CacheInvalidate(
    IN ljtUINT8_PTR Dest,
    IN ljtUINT32 Bytes
    );

ljeSTATUS
ljtaOS_IsPhysicalSecure(
    IN ljtaOS Os,
    IN ljtUINT32 Physical,
    OUT ljtBOOL *Secure
    );

ljeSTATUS
ljtaOS_Delay(
    IN ljtaOS Os,
    IN ljtUINT32 Delay
    );

ljeSTATUS
ljtaOS_SetGPUPower(
    IN ljtaOS Os,
    IN ljtUINT32 Core,
    IN ljtBOOL Clock,
    IN ljtBOOL Power
    );

/*
** ljtaHARDWARE
*/
ljeSTATUS
ljtaHARDWARE_Construct(
    IN ljTA TA,
    OUT ljTA_HARDWARE * Hardware
    );

ljeSTATUS
ljtaHARDWARE_Destroy(
    IN ljTA_HARDWARE Hardware
    );

ljeSTATUS
ljtaHARDWARE_Execute(
    IN ljTA TA,
    IN ljtUINT32 Address,
    IN ljtUINT32 Bytes
    );

ljeSTATUS
ljtaHARDWARE_End(
    IN ljTA_HARDWARE Hardware,
    IN ljtPOINTER Logical,
    IN OUT ljtUINT32 * Bytes
    );

ljeSTATUS
ljtaHARDWARE_SetMMU(
    IN ljTA_HARDWARE Hardware,
    IN ljtPOINTER Logical
    );

ljeSTATUS
ljtaHARDWARE_IsFeatureAvailable(
    IN ljTA_HARDWARE Hardware,
    IN ljeFEATURE Feature
    );

ljeSTATUS
ljtaHARDWARE_PrepareFunctions(
    IN ljTA_HARDWARE Hardware
    );

ljeSTATUS
ljtaHARDWARE_DumpMMUException(
    IN ljTA_HARDWARE Hardware
    );

ljeSTATUS
ljtaHARDWARE_HandleMMUException(
    IN ljTA_HARDWARE Hardware,
    IN ljtUINT32 MMUStatus,
    IN ljtPHYS_ADDR_T Physical,
    IN ljtUINT32 GPUAddress
    );

ljeSTATUS
ljtaHARDWARE_ReadMMUException(
    IN ljTA_HARDWARE Hardware,
    OUT ljtUINT32_PTR MMUStatus,
    OUT ljtUINT32_PTR MMUException
    );

ljeSTATUS
ljtaMMU_Construct(
    IN ljTA TA,
    OUT ljTA_MMU *Mmu
    );

ljeSTATUS
ljtaMMU_Destory(
    IN ljTA_MMU Mmu
    );

ljeSTATUS
ljtaMMU_SetPage(
    IN ljTA_MMU Mmu,
    IN ljtUINT32 PageAddress,
    IN ljtUINT32 *PageEntry
    );

ljeSTATUS
ljtaMMU_GetPageEntry(
    IN ljTA_MMU Mmu,
    IN ljtUINT32 Address,
    OUT ljtUINT32_PTR MtlbEntry,
    OUT ljtUINT32_PTR *PageTable,
    OUT ljtBOOL * Secure
    );

void
ljtaMMU_DumpPagetableEntry(
    IN ljTA_MMU Mmu,
    IN ljtUINT32 Address
    );

ljeSTATUS
ljtaMMU_FreePages(
    IN ljTA_MMU Mmu,
    IN ljtUINT32 Address,
    IN ljtUINT32 PageCount
    );

#ifdef __cplusplus
}
#endif
#endif


