/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_types_shared_h_
#define __lj_hal_types_shared_h_

#if !defined(GC_KMD)
#if defined(__KERNEL__)
#  include "linux/version.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
typedef unsigned long uintptr_t;
#  endif
#  include "linux/types.h"
# elif defined(UNDER_CE)
#  include <crtdefs.h>
typedef signed char        int8_t;
typedef short              int16_t;
typedef int                int32_t;
typedef long long          int64_t;
typedef unsigned char      uint8_t;
typedef unsigned short     uint16_t;
typedef unsigned int       uint32_t;
typedef unsigned long long uint64_t;
# elif defined(_MSC_VER) && (_MSC_VER <= 1500)
#  include <crtdefs.h>
#  include "vadefs.h"
# elif defined(__QNXNTO__)
#  define _QNX_SOURCE
#  include <stdint.h>
#  include <stddef.h>
# else
#  include <stdlib.h>
#  include <stddef.h>
#  include <stdint.h>
# endif
#endif

#ifdef _WIN32
# pragma warning(disable : 4127) /* Conditional expression is constant (do { } while(0)). */
# pragma warning(disable : 4100) /* Unreferenced formal parameter. */
# pragma warning(disable : 4204) /* Non-constant aggregate initializer (C99). */
# pragma warning(disable : 4131) /* Uses old-style declarator. */
# pragma warning(disable : 4206) /* Translation unit is empty. */
# pragma warning(disable : 4214) /* Nonstandard extension used :
                                  * bit field types other than int.
                                  */
#endif

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************\
 *  Platform macros.
 */

#if defined(__GNUC__)
# define ljdHAS_ELLIPSIS     1       /* GCC always has it. */
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)
# define ljdHAS_ELLIPSIS     1       /* C99 has it. */
#elif defined(_MSC_VER) && (_MSC_VER >= 1500)
# define ljdHAS_ELLIPSIS     1       /* MSVC 2007+ has it. */
#elif defined(UNDER_CE)
#if UNDER_CE >= 600
#  define ljdHAS_ELLIPSIS 1
# else
#  define ljdHAS_ELLIPSIS 0
# endif
#else
# error "ljdHAS_ELLIPSIS: Platform could not be determined"
#endif

/******************************************************************************
 *********************************** Keyword **********************************
 ******************************************************************************/

#if defined(ANDROID) && defined(__BIONIC_FORTIFY)
#if defined(__clang__)
#if (__clang_major__ >= 10)
#   define ljmINLINE           __inline__ __attribute__ ((always_inline))
#  else
#   define ljmINLINE           __inline__ __attribute__ ((always_inline)) __attribute__ ((gnu_inline))
#  endif
# else
#  define ljmINLINE            __inline__ __attribute__ ((always_inline)) __attribute__ ((gnu_inline)) __attribute__ ((artificial))
# endif
#elif ((defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)) || defined(__APPLE__))
# define ljmINLINE            inline      /* C99 keyword. */
#elif defined(__GNUC__)
# define ljmINLINE            __inline__  /* GNU keyword. */
#elif defined(_MSC_VER) || defined(UNDER_CE)
# define ljmINLINE            __inline    /* Internal keyword. */
#else
# error "ljmINLINE: Platform could not be determined"
#endif

/* kernel layer keyword. */
#define ljmkINLINE            inline

/* kernel fall-through keyword. */
#if defined __has_attribute
#if __has_attribute(__fallthrough__)
#  define ljmkFALLTHRU                   __attribute__((__fallthrough__))
# else
#  define ljmkFALLTHRU                   do {} while (0)  /* fallthrough */
# endif
#else
# define ljmkFALLTHRU                    do {} while (0)  /* fallthrough */
#endif

/* Possible debug flags. */
#define ljdDEBUG_NONE           0
#define ljdDEBUG_ALL            (1 << 0)
#define ljdDEBUG_FATAL          (1 << 1)
#define ljdDEBUG_TRACE          (1 << 2)
#define ljdDEBUG_BREAK          (1 << 3)
#define ljdDEBUG_ASSERT         (1 << 4)
#define ljdDEBUG_CODE           (1 << 5)
#define ljdDEBUG_STACK          (1 << 6)

#define ljmIS_DEBUG(flag)       (ljdDEBUG & ((flag) | ljdDEBUG_ALL))

#ifndef ljdDEBUG
#if (defined(DBG) && DBG) || defined(DEBUG) || defined(_DEBUG)
#  define ljdDEBUG        ljdDEBUG_ALL
# else
#  define ljdDEBUG        ljdDEBUG_NONE
# endif
#endif

#ifdef _USRDLL
#ifdef _MSC_VER
#ifdef HAL_EXPORTS
#   define HALAPI      __declspec(dllexport)
#  else
#   define HALAPI      __declspec(dllimport)
#  endif
#  define HALDECL      __cdecl
# else
#ifdef HAL_EXPORTS
#   define HALAPI
#  else
#   define HALAPI      extern
#  endif
# endif
#else
# define HALAPI
# define HALDECL
#endif

/******************************************************************************
 ********************************* Common Types *******************************
 ******************************************************************************/

#define ljvFALSE                0
#define ljvTRUE                 1

#define ljvINFINITE             ((ljtUINT32)~0U)

#define ljvINVALID_HANDLE       ((ljtHANDLE)ljmINT2PTR(~0U))

typedef int                     ljtBOOL;
typedef ljtBOOL                 *ljtBOOL_PTR;

typedef int                     ljtINT;
typedef signed char             ljtINT8;
typedef signed short            ljtINT16;
typedef signed int              ljtINT32;
typedef signed long long        ljtINT64;

typedef ljtINT                  *ljtINT_PTR;
typedef ljtINT8                 *ljtINT8_PTR;
typedef ljtINT16                *ljtINT16_PTR;
typedef ljtINT32                *ljtINT32_PTR;
typedef ljtINT64                *ljtINT64_PTR;

typedef unsigned int            ljtUINT;
typedef unsigned char           ljtUINT8;
typedef unsigned short          ljtUINT16;
typedef unsigned int            ljtUINT32;
typedef unsigned long long      ljtUINT64;
typedef uintptr_t               ljtUINTPTR_T;
typedef ptrdiff_t               ljtPTRDIFF_T;

typedef ljtUINT                 *ljtUINT_PTR;
typedef ljtUINT8                *ljtUINT8_PTR;
typedef ljtUINT16               *ljtUINT16_PTR;
typedef ljtUINT32               *ljtUINT32_PTR;
typedef ljtUINT64               *ljtUINT64_PTR;

typedef size_t                  ljtSIZE_T;
typedef ljtSIZE_T               *ljtSIZE_T_PTR;
typedef ljtUINT32               ljtTRACE;

#ifdef __cplusplus
# define ljvNULL                0
#else
# define ljvNULL                ((void *)0)
#endif

#define ljvMAXINT8              0x7f
#define ljvMININT8              0x80
#define ljvMAXINT16             0x7fff
#define ljvMININT16             0x8000
#define ljvMAXINT32             0x7fffffff
#define ljvMININT32             0x80000000
#define ljvMAXINT64             0x7fffffffffffffff
#define ljvMININT64             0x8000000000000000
#define ljvMAXUINT8             0xff
#define ljvMINUINT8             0x0
#define ljvMAXUINT16            0xffff
#define ljvMINUINT16            0x0
#define ljvMAXUINT32            0xffffffff
#define ljvMINUINT32            0x0
#define ljvMAXUINT64            0xffffffffffffffff
#define ljvMINUINT64            0x0
#define ljvMAXUINTPTR_T         (~(ljtUINTPTR_T)0)
#define ljvMAXSIZE_T            ((ljtSIZE_T)(-1))

typedef float                   ljtFLOAT;
typedef double                  ljtDOUBLE;
typedef signed int              ljtFIXED_POINT;
typedef float                   *ljtFLOAT_PTR;
typedef double                  *ljtDOUBLE_PTR;

typedef void                    *ljtPHYS_ADDR;
typedef void                    *ljtHANDLE;
typedef void                    *ljtFILE;
typedef void                    *ljtSIGNAL;
typedef void                    *ljtWINDOW;
typedef void                    *ljtIMAGE;
typedef void                    *ljtSHBUF;

typedef void                    *ljtSEMAPHORE;

typedef void                    *ljtPOINTER;
typedef const void              *ljtCONST_POINTER;

typedef char                    ljtCHAR;
typedef signed char             ljtSIGNED_CHAR;
typedef unsigned char           ljtUNSIGNED_CHAR;
typedef char                    *ljtSTRING;
typedef const char              *ljtCONST_STRING;

typedef ljtUINT64               ljtPHYS_ADDR_T;
typedef ljtUINT64               ljtADDRESS;

typedef struct _ljsCOUNT_STRING {
    ljtSIZE_T                   Length;
    ljtCONST_STRING             String;
} ljsCOUNT_STRING;

typedef union _ljuFLOAT_UINT32 {
    ljtFLOAT    f;
    ljtUINT32   u;
} ljuFLOAT_UINT32;

/* Fixed point constants. */
#define ljvZERO_X               ((ljtFIXED_POINT)0x00000000)
#define ljvHALF_X               ((ljtFIXED_POINT)0x00008000)
#define ljvONE_X                ((ljtFIXED_POINT)0x00010000)
#define ljvNEGONE_X             ((ljtFIXED_POINT)0xFFFF0000)
#define ljvTWO_X                ((ljtFIXED_POINT)0x00020000)

/* No special needs. */
#define ljvALLOC_FLAG_NONE                  0x00000000

/* Physical contiguous. */
#define ljvALLOC_FLAG_CONTIGUOUS            0x00000001
/* Physical non contiguous. */
#define ljvALLOC_FLAG_NON_CONTIGUOUS        0x00000002

/* Should not swap out. */
#define ljvALLOC_FLAG_NON_PAGED             0x00000004

/* CPU access explicitly needed. */
#define ljvALLOC_FLAG_CPU_ACCESS            0x00000008
/* Can be remapped as cacheable. */
#define ljvALLOC_FLAG_CACHEABLE             0x00000010

/* Need 32bit address. */
#define ljvALLOC_FLAG_4GB_ADDR              0x00000020

/* Secure buffer. */
#define ljvALLOC_FLAG_SECURITY              0x00000040
/* Can be exported as dmabuf-fd */
#define ljvALLOC_FLAG_DMABUF_EXPORTABLE     0x00000080
/* Do not try slow pools (ljvPOOL_VIRTUAL) */
#define ljvALLOC_FLAG_FAST_POOLS            0x00000100

/* Only accessed by GPU */
#define ljvALLOC_FLAG_NON_CPU_ACCESS        0x00000200
/* Do not be moved */
#define ljvALLOC_FLAG_NO_EVICT              0x00000400

/* Allocate from user space. */
#define ljvALLOC_FLAG_FROM_USER             0x00000800

/* Import DMABUF. */
#define ljvALLOC_FLAG_DMABUF                0x00001000
/* Import USERMEMORY. */
#define ljvALLOC_FLAG_USERMEMORY            0x00002000
/* Import an External Buffer, this flag is no longer used. */
#define ljvALLOC_FLAG_EXTERNAL_MEMORY       0x00004000
/* Import linux reserved memory. */
#define ljvALLOC_FLAG_LINUX_RESERVED_MEM    0x00008000

/* 1M pages unit allocation. */
#define ljvALLOC_FLAG_1M_PAGES              0x00010000

/* Non 1M pages unit allocation. */
#define ljvALLOC_FLAG_4K_PAGES              0x00020000

/* Lower 4G VA range. */
#define ljvALLOC_FLAG_32BIT_VA              0x00040000
#define ljvALLOC_FLAG_PRIOR_32BIT_VA        0x00080000

/* Dynamically allocate local memory pool memory. */
#define ljvALLOC_FLAG_DYNAMIC_ALLOC_LOCAL   0x00100000

/* Real allocation happens when GPU page fault. */
#define ljvALLOC_FLAG_ALLOC_ON_FAULT        0x01000000
/* Alloc with memory limit. */
#define ljvALLOC_FLAG_MEMLIMIT              0x02000000
/* Alloc memory with mirror */
#define ljvALLOC_FLAG_WITH_MIRROR           0x04000000

#define ljvALLOC_FLAG_EXCLUSIVE_NONEED_VIRTUAL              0x10000000

#define ljvALLOC_FLAG_DMA_ONLY              0x20000000 /* for mmu cmd */

#define ljmFIXEDCLAMP_NEG1_TO_1(_x)                                 \
    (((_x) < ljvNEGONE_X) ? ljvNEGONE_X :                           \
                            (((_x) > ljvONE_X) ? ljvONE_X : (_x)))

#define ljmFLOATCLAMP_NEG1_TO_1(_f)                                 \
    (((_f) < -1.0f) ? -1.0f : (((_f) > 1.0f) ? 1.0f : (_f)))

#define ljmFIXEDCLAMP_0_TO_1(_x)                                    \
    (((_x) < 0) ? 0 : (((_x) > ljvONE_X) ? ljvONE_X : (_x)))

#define ljmFLOATCLAMP_0_TO_1(_f)                                    \
    (((_f) < 0.0f) ? 0.0f : (((_f) > 1.0f) ? 1.0f : (_f)))

/******************************************************************************
 ****************************** Multicast Values ******************************
 ******************************************************************************/

/* Value unions. */
typedef union _ljuVALUE {
    ljtUINT                     uintValue;
    ljtFIXED_POINT              fixedValue;
    ljtFLOAT                    floatValue;
    ljtINT                      intValue;
} ljuVALUE;

/* Stringizing macro. */
#define ljmSTRING(Value)        #Value

/******************************************************************************
 ****************************** Fixed Point Math ******************************
 ******************************************************************************/

#define ljmXMultiply(x1, x2)            ljoMATH_MultiplyFixed(x1, x2)
#define ljmXDivide(x1, x2)              ljoMATH_DivideFixed(x1, x2)
#define ljmXMultiplyDivide(x1, x2, x3)  ljoMATH_MultiplyDivideFixed(x1, x2, x3)

/* 2D Engine profile. */
typedef struct _ljs2D_PROFILE {
    /* Cycle count.
     * 32bit counter incremented every 2D clock cycle.
     * Wraps back to 0 when the counter overflows.
     */
    ljtUINT32       cycleCount;

    /* Pixels rendered by the 2D engine.
     * Resets to 0 every time it is read.
     */
    ljtUINT32       pixelsRendered;
} ljs2D_PROFILE;

#define ljmPRINTABLE(c) ((((c) >= ' ') && ((c) <= '}')) ? ((c) != '%' ? (c) : ' ') : ' ')

#define ljmCC_PRINT(cc) \
    ljmPRINTABLE((char) ((cc)        & 0xFF)), \
    ljmPRINTABLE((char) (((cc) >>  8) & 0xFF)), \
    ljmPRINTABLE((char) (((cc) >> 16) & 0xFF)), \
    ljmPRINTABLE((char) (((cc) >> 24) & 0xFF))

/******************************************************************************
 ***************************** Function Parameters ****************************
 ******************************************************************************/

#define IN
#define OUT
#define INOUT
#define OPTIONAL

/******************************************************************************
 ******************************** Status Macros *******************************
 ******************************************************************************/

#define ljmIS_ERROR(status)         ((status) < 0)
#define ljmNO_ERROR(status)         ((status) >= 0)
#define ljmIS_SUCCESS(status)       ((status) == ljvSTATUS_OK)
#define ljmERROR2PTR(err)           ((ljtPOINTER)(err))
#define ljmPTR2ERROR(ptr)           ((ljtINT64)(ptr))
/******************************************************************************
 ******************************** Field Macros ********************************
 ******************************************************************************/

#define __ljmSTART(reg_field)                        \
    (0 ? reg_field)

#define __ljmEND(reg_field)                          \
    (1 ? reg_field)

#define __ljmGETSIZE(reg_field)                      \
    (__ljmEND(reg_field) - __ljmSTART(reg_field) + 1)

#define __ljmALIGN(data, reg_field)                  \
    (((ljtUINT32)(data)) << __ljmSTART(reg_field))

#define __ljmMASK(reg_field)                         \
    ((ljtUINT32) ((__ljmGETSIZE(reg_field) == 32)    \
        ?  ~0U \
        : (~(~0U << __ljmGETSIZE(reg_field)))))


/*******************************************************************************
 **
 **  ljmFIELDMASK
 **
 **      Get aligned field mask.
 **
 **  ARGUMENTS:
 **
 **      reg     Name of register.
 **      field   Name of field within register.
 */
#define ljmFIELDMASK(reg, field)                     \
    (__ljmALIGN(__ljmMASK(reg##_##field), reg##_##field))

/*******************************************************************************
 **
 **  ljmGETFIELD
 **
 **      Extract the value of a field from specified data.
 **
 **  ARGUMENTS:
 **
 **      data    Data value.
 **      reg     Name of register.
 **      field   Name of field within register.
 */
#define ljmGETFIELD(data, reg, field)                       \
(\
    ((((ljtUINT32)(data)) >> __ljmSTART(reg##_##field))     \
        & __ljmMASK(reg##_##field))                         \
)

/*******************************************************************************
 **
 **  ljmSETFIELD
 **
 **      Set the value of a field within specified data.
 **
 **  ARGUMENTS:
 **
 **      data    Data value.
 **      reg     Name of register.
 **      field   Name of field within register.
 **      value   Value for field.
 */
#define ljmSETFIELD(data, reg, field, value)                    \
(\
    (((ljtUINT32)(data))                                       \
        & ~__ljmALIGN(__ljmMASK(reg##_##field), reg##_##field)) \
        |  __ljmALIGN((ljtUINT32)(value)                       \
            & __ljmMASK(reg##_##field), reg##_##field)          \
)

/*******************************************************************************
 **
 **  ljmSETFIELDVALUE
 **
 **      Set the value of a field within specified data with a
 **      predefined value.
 **
 **  ARGUMENTS:
 **
 **      data    Data value.
 **      reg     Name of register.
 **      field   Name of field within register.
 **      value   Name of the value within the field.
 */
#define ljmSETFIELDVALUE(data, reg, field, value)               \
(\
    (((ljtUINT32)(data))                                       \
        & ~__ljmALIGN(__ljmMASK(reg##_##field), reg##_##field)) \
        |  __ljmALIGN(reg##_##field##_##value                   \
            & __ljmMASK(reg##_##field), reg##_##field)          \
)

/*******************************************************************************
 **
 **  ljmGETMASKEDFIELDMASK
 **
 **      Determine field mask of a masked field.
 **
 **  ARGUMENTS:
 **
 **      reg     Name of register.
 **      field   Name of field within register.
 */
#define ljmGETMASKEDFIELDMASK(reg, field) \
(\
    ljmSETFIELD(0, reg, field, ~0U) | \
    ljmSETFIELD(0, reg, MASK_ ## field, ~0U)   \
)

/*******************************************************************************
 **
 **  ljmSETMASKEDFIELD
 **
 **      Set the value of a masked field with specified data.
 **
 **  ARGUMENTS:
 **
 **      reg     Name of register.
 **      field   Name of field within register.
 **      value   Value for field.
 */
#define ljmSETMASKEDFIELD(reg, field, value) \
(\
    ljmSETFIELD     (~0U, reg, field, value) & \
    ljmSETFIELDVALUE(~0U, reg, MASK_ ## field, ENABLED) \
)

/*******************************************************************************
 **
 **  ljmSETMASKEDFIELDVALUE
 **
 **      Set the value of a masked field with specified data.
 **
 **  ARGUMENTS:
 **
 **      reg     Name of register.
 **      field   Name of field within register.
 **      value   Value for field.
 */
#define ljmSETMASKEDFIELDVALUE(reg, field, value) \
(\
    ljmSETFIELDVALUE(~0U, reg, field, value) & \
    ljmSETFIELDVALUE(~0U, reg, MASK_ ## field, ENABLED) \
)

/*******************************************************************************
 **
 **  ljmVERIFYFIELDVALUE
 **
 **      Verify if the value of a field within specified data equals a
 **      predefined value.
 **
 **  ARGUMENTS:
 **
 **      data    Data value.
 **      reg     Name of register.
 **      field   Name of field within register.
 **      value   Name of the value within the field.
 */
#define ljmVERIFYFIELDVALUE(data, reg, field, value)     \
(\
    (((ljtUINT32)(data)) >> __ljmSTART(reg##_##field) &  \
                            __ljmMASK(reg##_##field))    \
        ==                                               \
    (reg##_##field##_##value & __ljmMASK(reg##_##field)) \
)

/*******************************************************************************
 **  Bit field macros.
 */

#define __ljmSTARTBIT(Field)                   \
    (1 ? Field)

#define __ljmBITSIZE(Field)                                                                        \
    (0 ? Field)

#define __ljmBITMASK(Field)                     \
(\
    (1 << __ljmBITSIZE(Field)) - 1              \
)

#define ljmGETBITS(Value, Type, Field)          \
(\
    (((Type)(Value)) >> __ljmSTARTBIT(Field))   \
    &                                           \
    __ljmBITMASK(Field)                         \
)

#define ljmSETBITS(Value, Type, Field, NewValue)                            \
(\
    (((Type)(Value)) & ~(__ljmBITMASK(Field) << __ljmSTARTBIT(Field))) |    \
    ((((Type)(NewValue)) & __ljmBITMASK(Field)) << __ljmSTARTBIT(Field))    \
)

/*******************************************************************************
 **
 **  ljmISINREGRANGE
 **
 **      Verify whether the specified address is in the register range.
 **
 **  ARGUMENTS:
 **
 **      Address  the Address to be verified.
 **      Name     Name of a register.
 */

#define ljmISINREGRANGE(Address, Name)                          \
(\
    ((Address & (~0U << Name##_LSB)) == (Name##_Address >> 2))  \
)

/******************************************************************************
 ******************************* Ceiling Macro ********************************
 ******************************************************************************/
#define ljmCEIL(x)                                                      \
(\
    ((x) - (ljtUINT32)(x)) == 0 ? (ljtUINT32)(x) : (ljtUINT32)(x) + 1   \
)

/******************************************************************************
 ******************************* Min/Max Macros *******************************
 ******************************************************************************/

#define ljmMIN(x, y)            (((x) <= (y)) ? (x) : (y))
#define ljmMAX(x, y)            (((x) >= (y)) ? (x) : (y))
#define ljmCLAMP(x, min, max)   (((x) < (min)) ? (min) : ((x) > (max)) ? (max) : (x))
#define ljmABS(x)               (((x) < 0) ? -(x) : (x))
#define ljmNEG(x)               (((x) < 0) ? (x) : -(x))

/******************************************************************************
 ******************************* Bit Macro ************************************
 ******************************************************************************/
#define ljmBITSET(x, bit)       ((x) | (1 << (bit)))
#define ljmBITCLEAR(x, bit)     ((x) & ~(1 << (bit)))
#define ljmBITTEST(x, bit)      ((x) & (1 << (bit)))

/*******************************************************************************
 **
 **  ljmPTR2SIZE
 **
 **      Convert a pointer to an integer value.
 **
 **  ARGUMENTS:
 **
 **      p       Pointer value.
 */
#define ljmPTR2SIZE(p)      ((ljtUINTPTR_T)(p))

#define ljmPTR2INT32(p)     ((ljtUINT32)(ljtUINTPTR_T)(p))

/*******************************************************************************
 **
 **  ljmINT2PTR
 **
 **      Convert an integer value into a pointer.
 **
 **  ARGUMENTS:
 **
 **      v       Integer value.
 */

#define ljmINT2PTR(i)       ((ljtPOINTER)(ljtUINTPTR_T)(i))

/*******************************************************************************
 **
 **  ljmOFFSETOF
 **
 **      Compute the byte offset of a field inside a structure.
 **
 **  ARGUMENTS:
 **
 **      s       Structure name.
 **      field   Field name.
 */
#define ljmOFFSETOF(s, field)       (ljmPTR2INT32(&(((struct s *)0)->field)))

#define __ljmOFFSETOF(type, field)  (ljmPTR2INT32(&(((type *)0)->field)))

/*******************************************************************************
 **
 **  ljmCONTAINEROF
 **
 **      Get containing structure of a member.
 **
 **  ARGUMENTS:
 **
 **      Pointer the Pointer of member.
 **      Type    Structure name.
 **      Name    Field name.
 */
#define ljmCONTAINEROF(Pointer, Type, Member)                         \
(\
    (Type *)((ljtUINTPTR_T)(Pointer) - __ljmOFFSETOF(Type, Member))   \
)

/*******************************************************************************
 **
 ** ljmBSWAP16/32/64
 **
 **      Return a value with all bytes in the 16/32/64 bit argument swapped.
 */
#if !defined(__KERNEL__) && defined(__GNUC__) &&                                \
    (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__ >= 40300) && \
    !defined(__VXWORKS__)
# define ljmBSWAP16(x)       __builtin_bswap16(x)
# define ljmBSWAP32(x)       __builtin_bswap32(x)
# define ljmBSWAP64(x)       __builtin_bswap64(x)
#else
# define ljmBSWAP16(x) ((ljtUINT16)(\
    (((ljtUINT16)(x) & (ljtUINT16)0x00FF) << 8)  |        \
    (((ljtUINT16)(x) & (ljtUINT16)0xFF00) >> 8)))

# define ljmBSWAP32(x) ((ljtUINT32)(\
    (((ljtUINT32)(x) & (ljtUINT32)0x000000FFU) << 24) |   \
    (((ljtUINT32)(x) & (ljtUINT32)0x0000FF00U) << 8)  |   \
    (((ljtUINT32)(x) & (ljtUINT32)0x00FF0000U) >> 8)  |   \
    (((ljtUINT32)(x) & (ljtUINT32)0xFF000000U) >> 24)))

# define ljmBSWAP64(x) ((ljtUINT64)(\
    (((ljtUINT64)(x) & (ljtUINT64)0x00000000000000FFULL) << 56) |   \
    (((ljtUINT64)(x) & (ljtUINT64)0x000000000000FF00ULL) << 40) |   \
    (((ljtUINT64)(x) & (ljtUINT64)0x0000000000FF0000ULL) << 24) |   \
    (((ljtUINT64)(x) & (ljtUINT64)0x00000000FF000000ULL) << 8)  |   \
    (((ljtUINT64)(x) & (ljtUINT64)0x000000FF00000000ULL) >> 8)  |   \
    (((ljtUINT64)(x) & (ljtUINT64)0x0000FF0000000000ULL) >> 24) |   \
    (((ljtUINT64)(x) & (ljtUINT64)0x00FF000000000000ULL) >> 40) |   \
    (((ljtUINT64)(x) & (ljtUINT64)0xFF00000000000000ULL) >> 56)))
#endif

/*******************************************************************************
 **
 ** ljmBSWAP16IN32
 **
 **      Return a value with every 16 bit swapped of a 32 bit data type.
 */
#define ljmBSWAP16IN32(x) ((ljtUINT32)(\
    (((ljtUINT32)(x) & (ljtUINT32)0x000000FFU) << 8)  | \
    (((ljtUINT32)(x) & (ljtUINT32)0x0000FF00U) >> 8)  | \
    (((ljtUINT32)(x) & (ljtUINT32)0x00FF0000U) << 8)  | \
    (((ljtUINT32)(x) & (ljtUINT32)0xFF000000U) >> 8)))

/*******************************************************************************
 **
 ** ljmBSWAP16IN32EX
 **
 **      Return a value with whole 16 bit swapped of a 32 bit data type.
 */
#define ljmBSWAP16IN32EX(x) ((ljtUINT32)(\
    (((ljtUINT32)(x) & (ljtUINT32)0x0000FFFFU) << 16)  | \
    (((ljtUINT32)(x) & (ljtUINT32)0xFFFF0000U) >> 16)))

/*******************************************************************************
 **
 ** ljmBSWAP32IN64
 **
 **      Return a value with whole 32 bit swapped of a 64 bit data type.
 */
#define ljmBSWAP32IN64(x) ((ljtUINT64)(\
    (((ljtUINT64)(x) & (ljtUINT64)0x00000000FFFFFFFFULL) << 32) | \
    (((ljtUINT64)(x) & (ljtUINT64)0xFFFFFFFF00000000ULL) >> 32)))

/******************************************************************************
 **** Database ****************************************************************/

typedef struct _ljsDATABASE_COUNTERS {
    /* Number of currently allocated bytes. */
    ljtUINT64                   bytes;

    /* Maximum number of bytes allocated (memory footprint). */
    ljtUINT64                   maxBytes;

    /* Total number of bytes allocated. */
    ljtUINT64                   totalBytes;

    /* The numbers of times video memory was allocated. */
    ljtUINT32                   allocCount;

    /* The numbers of times video memory was freed. */
    ljtUINT32                   freeCount;
} ljsDATABASE_COUNTERS;

typedef struct _ljuDATABASE_INFO {
    /* Counters. */
    ljsDATABASE_COUNTERS        counters;

    /* Time value. */
    ljtUINT64                   time;
} ljuDATABASE_INFO;

/******************************************************************************
 **** Frame database **********************************************************/

/* ljsHAL_FRAME_INFO */
typedef struct _ljsHAL_FRAME_INFO {
    /* Current timer tick. */
    OUT ljtUINT64               ticks;

    /* Bandwidth counters. */
    OUT ljtUINT                 readBytes8[8];
    OUT ljtUINT                 writeBytes8[8];

    /* Counters. */
    OUT ljtUINT                 cycles[8];
    OUT ljtUINT                 idleCycles[8];
    OUT ljtUINT                 mcCycles[8];
    OUT ljtUINT                 readRequests[8];
    OUT ljtUINT                 writeRequests[8];

    /* 3D counters. */
    OUT ljtUINT                 vertexCount;
    OUT ljtUINT                 primitiveCount;
    OUT ljtUINT                 rejectedPrimitives;
    OUT ljtUINT                 culledPrimitives;
    OUT ljtUINT                 clippedPrimitives;
    OUT ljtUINT                 outPrimitives;
    OUT ljtUINT                 inPrimitives;
    OUT ljtUINT                 culledQuadCount;
    OUT ljtUINT                 totalQuadCount;
    OUT ljtUINT                 quadCount;
    OUT ljtUINT                 totalPixelCount;

    /* PE counters. */
    OUT ljtUINT                 colorKilled[8];
    OUT ljtUINT                 colorDrawn[8];
    OUT ljtUINT                 depthKilled[8];
    OUT ljtUINT                 depthDrawn[8];

    /* Shader counters. */
    OUT ljtUINT                 shaderCycles;
    OUT ljtUINT                 vsInstructionCount;
    OUT ljtUINT                 vsTextureCount;
    OUT ljtUINT                 psInstructionCount;
    OUT ljtUINT                 psTextureCount;

    /* Texture counters. */
    OUT ljtUINT                 bilinearRequests;
    OUT ljtUINT                 trilinearRequests;
    OUT ljtUINT                 txBytes8;
    OUT ljtUINT                 txHitCount;
    OUT ljtUINT                 txMissCount;
} ljsHAL_FRAME_INFO;

typedef struct _ljkLINKDATA *ljkLINKDATA;
struct _ljkLINKDATA {
    ljtADDRESS                  start;
    ljtADDRESS                  end;
    ljtUINT32                   pid;
    ljtUINT32                   linkLow;
    ljtUINT32                   linkHigh;
};

typedef struct _ljkADDRESSDATA *ljkADDRESSDATA;
struct _ljkADDRESSDATA {
    ljtADDRESS                  start;
    ljtADDRESS                  end;
};

typedef union _ljuQUEUEDATA {
    struct _ljkLINKDATA         linkData;

    struct _ljkADDRESSDATA      addressData;
} ljuQUEUEDATA;

typedef struct _ljkQUEUE *ljkQUEUE;
struct _ljkQUEUE {
    ljuQUEUEDATA                *datas;
    ljtUINT32                   rear;
    ljtUINT32                   front;
    ljtUINT32                   count;
    ljtUINT32                   size;
};

typedef struct _ljsLISTHEAD *ljsLISTHEAD_PTR;
typedef struct _ljsLISTHEAD {
    ljsLISTHEAD_PTR     prev;
    ljsLISTHEAD_PTR     next;
} ljsLISTHEAD;

/*
 * 'Patch' here means a mechanism to let kernel side modify user space reserved
 * command buffer location, or something the like, during the command buffer
 * commit.
 *
 * Reasons of using 'patch':
 * 1. Some resources/states are managed globally only in kernel side, such as
 *    MCFE semaphore, etc.
 * 2. For the sake of security or optimization, like video memory address.
 *
 * Patches are arranged in arrays, each array has the same type. The 'patchArray'
 * in 'ljsHAL_PATCH_LIST' pointers the concrete patch item array.
 *
 * NOTICE:
 * Be aware of the order and values! Tables in lj_hal_user_buffer.c and
 * lj_hal_kernel_command.c depend on this.
 */
/* The patch array. */
typedef struct _ljsHAL_PATCH_LIST {
    /* Patch type. */
    ljtUINT32           type;

    /* Patch item count. */
    ljtUINT32           count;

    /*
     * Pointer to the patch items.
     *
     * ljsHAL_PATCH_VIDMEM_ADDRESS * patchArray;
     * ljsHAL_PATCH_MCFE_SEMAPHORE * patchArray;
     * ljsHAL_PATCH_VIDMEM_TIMESTAMP * patchArray;
     * ...
     */
    ljtUINT64           patchArray;

    /* struct _ljsHAL_PATCH_LIST * next; */
    ljtUINT64           next;
} ljsHAL_PATCH_LIST;

/*
 * Patch a GPU address in the place (ljvHAL_PATCH_VIDMEM_ADDRESS).
 * Size of a GPU address is always 32 bits.
 */
typedef struct _ljsHAL_PATCH_VIDMEM_ADDRESS {
    /* Patch location in the command buffer. */
    ljtUINT32           location;

    /* Handle of the video memory node. */
    ljtUINT32           node;

    /* Address offset in the video memory node. */
    ljtUINT32           offset;
} ljsHAL_PATCH_VIDMEM_ADDRESS;

/*
 * Patch a MCFE semaphore command in the place (ljvHAL_PATCH_MCFE_SEMAPHORE).
 * Size of the semaphore command is fixed at _64_ bits!
 */
typedef struct _ljsHAL_PATCH_MCFE_SEMAPHORE {
    /* Patch location in the command buffer. */
    ljtUINT32           location;

    /* semaphore direction: 1 = Send, 0 = Wait. */
    ljtUINT32           sendSema;

    /* Handle of the semaphore. */
    ljtUINT32           semaHandle;
} ljsHAL_PATCH_MCFE_SEMAPHORE;

/*
 * Patch timestamp of given video memory node (ljvHAL_PATCH_VIDMEM_TIMESTAMP).
 * Pure software-wise, not command relevant.
 */
typedef struct _ljsHAL_PATCH_VIDMEM_TIMESTAMP {
    /* Handle of a video memory node. */
    ljtUINT32           handle;

    ljtUINT32           flag;
} ljsHAL_PATCH_VIDMEM_TIMESTAMP;

/* Put together patch list handling variables. */
typedef struct _ljsPATCH_LIST_VARIABLE {
    /* ljvHAL_PATCH_VIDMEM_TIMESTAMP. */
    ljtUINT64           maxAsyncTimestamp;

    /* ljvHAL_PATCH_MCFE_SEMAPHORE. */
    ljtBOOL             semaUsed;
} ljsPATCH_LIST_VARIABLE;

/*
 *  ljvFEATURE_DATABASE_DATE_MASK
 *
 *  Mask used to control which bits of chip date will be used to
 *  query feature database, ignore release date for fpga and emulator.
 */
#if (ljdFPGA_BUILD || defined(EMULATOR))
# define ljvFEATURE_DATABASE_DATE_MASK   (0U)
#else
# define ljvFEATURE_DATABASE_DATE_MASK   (~0U)
#endif

#if defined(__GNUC__)
#if defined(__BYTE_ORDER__) && defined(__ORDER_BIG_ENDIAN__) && \
     (__BYTE_ORDER__ == __ORDER_BIG_ENDIAN__)
#  define ljdENDIAN_BIG   1
# else
#  define ljdENDIAN_BIG   0
# endif
#else
# define ljdENDIAN_BIG    0
#endif

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_types_shared_h_ */


