/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_family.h"
#include "ljmicro_device.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

/* GP201 */
int gp201_init(struct ljm_device *ljm_dev);
int gp201_resume(struct drm_device *ddev);
void gp200_fini(struct ljm_device *ljm_dev);
int gp200_irq_set(struct ljm_device *ljm_dev);
int gp200_irq_vblank_set(struct ljm_device *ljm_dev);
int gp201_irq_process(struct ljm_device *ljm_dev);
void gp200_irq_postinstall(struct ljm_device *ljm_dev);
void gp200_hpd_init(struct ljm_device *ljm_dev);
void gp200_hpd_fini(struct ljm_device *ljm_dev);
bool gp200_hpd_sense(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector);
void gp200_hpd_set_polarity(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector);
void gp200_encoder_dpms(struct drm_encoder *encoder, int mode);
void gp200_encoder_mode_set(struct drm_encoder *encoder, struct drm_display_mode *mode, struct drm_display_mode *adjustedMode);
bool gp201_encoder_mode_fixup(struct drm_encoder *encoder, const struct drm_display_mode *mode, struct drm_display_mode *adjustedMode);

void gp200_crtc_cursor_move(struct drm_crtc *crtc, int x, int y);
void gp200_crtc_cursor_hide(struct ljm_device *ljm_dev, struct drm_crtc *crtc);
void gp200_crtc_cursor_show(struct ljm_device *ljm_dev, struct drm_crtc *crtc);
bool gp200_crtc_mode_fixup(struct ljm_device *ljm_dev, const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode);
int gp200_crtc_mode_set(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, struct drm_display_mode *adjusted_mode);
void gp200_crtc_mode_enable(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, bool enable);
int gp200_mode_set_base(struct drm_crtc *crtc, int bpp, int pitches, int base);
int gp200_crtc_set_base(struct ljm_device *ljm_dev, int crtc_id, u64 crtc_base);
int gp200_crtc_prog_reset(struct ljm_device *ljm_dev,  int crtc_used, long min_hscan_time);
void gp200_crtc_commit_for_dp(struct ljm_device *ljm_dev);
int gp200_connector_detect(struct drm_connector *connector);
int gp200_connector_mode_valid(struct drm_display_mode *mode, struct ljm_connector *ljm_connector);
struct edid *gp200_connector_get_edid(IN struct drm_connector *connector, OUT int *size);
int gp200_suspend(struct drm_device *ddev);
u32 gp200_get_vblank_counter(struct ljm_device *ljm_dev, int crtc);
void gp200_page_flip(struct ljm_device *ljm_dev, int crtc_id, u64 crtc_base, bool async);
bool gp200_page_flip_pending(struct ljm_device *ljm_dev, int crtc);
int gp200_get_crtc_scanout_pos(struct ljm_device *ljm_dev, unsigned int pipe, int *vpos, int *hpos);
int gp201_crtc_load_lut(struct ljm_crtc *ljm_crtc, u16 *red, u16 *green, u16 *blue, uint32_t gamma_size);
// int gp200_power_set_state(struct ljm_device *ljm_dev, int power_state);
int gp201_power_set_state(struct ljm_device *ljm_dev, int power_state);

/* GP202 */
int gp202_init(struct ljm_device *ljm_dev);
int gp202_resume(struct drm_device *ddev);
int gp202_irq_process(struct ljm_device *ljm_dev);
bool gp202_encoder_mode_fixup(struct drm_encoder *encoder, const struct drm_display_mode *mode, struct drm_display_mode *adjustedMode);
int gp202_crtc_load_lut(struct ljm_crtc *ljm_crtc, u16 *red, u16 *green, u16 *blue, uint32_t gamma_size);
int gp202_power_set_state(struct ljm_device *ljm_dev, int power_state);

static struct ljm_asic gp201_asic = {
    .init = &gp201_init,
    .fini = &gp200_fini,
    .suspend = &gp200_suspend,
    .resume = &gp201_resume,
    .power_set_state = &gp201_power_set_state,
    .irq = {
        .set = &gp200_irq_set,
        .vblank_set = &gp200_irq_vblank_set,
        .process = &gp201_irq_process,
        .postinstall = &gp200_irq_postinstall,
    },
    .hpd = {
        .init = &gp200_hpd_init,
        .fini =&gp200_hpd_fini,
        .sense = &gp200_hpd_sense,
        .set_polarity = &gp200_hpd_set_polarity,
    },
    .display = {
        .encoder_dpms = &gp200_encoder_dpms,
        .encoder_mode_set = &gp200_encoder_mode_set,
        .encoder_mode_fixup = &gp201_encoder_mode_fixup,
        .crtc_cursor_move = &gp200_crtc_cursor_move,
        .crtc_cursor_hide = &gp200_crtc_cursor_hide,
        .crtc_cursor_show = &gp200_crtc_cursor_show,
        .crtc_mode_fixup = &gp200_crtc_mode_fixup,
        .crtc_mode_set = &gp200_crtc_mode_set,
        .crtc_mode_enable = &gp200_crtc_mode_enable,
        .crtc_prog_reset = &gp200_crtc_prog_reset,
        .crtc_mode_set_base = &gp200_mode_set_base,
        .crtc_set_base = &gp200_crtc_set_base,
        .crtc_commit_dp = &gp200_crtc_commit_for_dp,
        .connector_detect = &gp200_connector_detect,
        .connector_mode_valid = &gp200_connector_mode_valid,
        .connector_get_edid = &gp200_connector_get_edid,
        .get_vblank_counter = &gp200_get_vblank_counter,
        .get_crtc_scanout_pos = &gp200_get_crtc_scanout_pos,
        .crtc_load_lut = &gp201_crtc_load_lut
    },
    .pflip = {
        .page_flip = &gp200_page_flip,
        .page_flip_pending = &gp200_page_flip_pending,
    },
};

static struct ljm_asic gp202_asic = {
    .init = &gp202_init,
    .fini = &gp200_fini,
    .suspend = &gp200_suspend,
    .resume = &gp202_resume,
    .power_set_state = &gp202_power_set_state,
    .irq = {
        .set = &gp200_irq_set,
        .vblank_set = &gp200_irq_vblank_set,
        .process = &gp202_irq_process,
        .postinstall = &gp200_irq_postinstall,
    },
    .hpd = {
        .init = &gp200_hpd_init,
        .fini =&gp200_hpd_fini,
        .sense = &gp200_hpd_sense,
        .set_polarity = &gp200_hpd_set_polarity,
    },
    .display = {
        .encoder_dpms = &gp200_encoder_dpms,
        .encoder_mode_set = &gp200_encoder_mode_set,
        .encoder_mode_fixup = &gp202_encoder_mode_fixup,
        .crtc_cursor_move = &gp200_crtc_cursor_move,
        .crtc_cursor_hide = &gp200_crtc_cursor_hide,
        .crtc_cursor_show = &gp200_crtc_cursor_show,
        .crtc_mode_fixup = &gp200_crtc_mode_fixup,
        .crtc_mode_set = &gp200_crtc_mode_set,
        .crtc_mode_enable = &gp200_crtc_mode_enable,
        .crtc_prog_reset = &gp200_crtc_prog_reset,
        .crtc_mode_set_base = &gp200_mode_set_base,
        .crtc_set_base = &gp200_crtc_set_base,
        .crtc_commit_dp = &gp200_crtc_commit_for_dp,
        .connector_detect = &gp200_connector_detect,
        .connector_mode_valid = &gp200_connector_mode_valid,
        .connector_get_edid = &gp200_connector_get_edid,
        .get_vblank_counter = &gp200_get_vblank_counter,
        .get_crtc_scanout_pos = &gp200_get_crtc_scanout_pos,
        .crtc_load_lut = &gp202_crtc_load_lut
    },
    .pflip = {
        .page_flip = &gp200_page_flip,
        .page_flip_pending = &gp200_page_flip_pending,
    },
};

int ljm_dev_info_init(struct ljm_device *ljm_dev);

int ljm_asic_init(struct ljm_device *ljm_dev)
{
    HEADER_ARG("ljm_dev=%p", ljm_dev);

    if(ljm_dev_info_init(ljm_dev))
    {
        FOOTER();
        return -ENOENT;
    }

    switch (ljm_dev->family) {
    case CHIP_GP200:
        /** GP200 series base functions */
        if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
            ljm_dev->asic = &gp202_asic;
        else
            ljm_dev->asic = &gp201_asic;
        break;
    default:
        /* FIXME: not supported yet */
        DRM_ERROR("device family 0x%x is not supported\n", ljm_dev->family);
        return -EINVAL;
    }

    FOOTER();
    return 0;
}

