/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/**************************************************************************
* 版权所有 (C)2022, 武汉凌久微电子有限公司。
*
* 文件名称： ljmicro_audio_mdrv.c
* 文件标识：
* 内容摘要： 本文件定义dai_link，并注册声卡
* 其它说明：
* 当前版本： v1.0
* 作    者： zhoupw, 武汉凌久微电子有限公司
* 完成日期： 2022年4月27日
***************************************************************************/
#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)

#include <sound/soc.h>
#include <linux/version.h>
#include "ljmicro_device.h"

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
SND_SOC_DAILINK_DEFS(ljmicro_audio,
    DAILINK_COMP_ARRAY(COMP_EMPTY()),
    DAILINK_COMP_ARRAY(COMP_CODEC("", "ljm-i2s-hifi")),
    DAILINK_COMP_ARRAY(COMP_EMPTY()));
#endif

struct snd_soc_dai_link ljmicro_link[] = {
    {
        .name = "LJM-AUDIO",
        .stream_name = "ljm-pcm",
        .dai_fmt = SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS,
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
        SND_SOC_DAILINK_REG(ljmicro_audio),
#endif
    },
};

static struct snd_soc_card ljm_audio_card = {
    .name = "ljm-audio-card",
    .owner = THIS_MODULE,
    .dai_link = ljmicro_link,
    .num_links = ARRAY_SIZE(ljmicro_link),
};

int ljm_audio_card_register(struct ljm_device *ljm_dev)
{
    int i;
    int ret;
    struct snd_soc_card *card;

    card = &ljm_audio_card;
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
    for (i = 0; i < card->num_links; i++) {
        card->dai_link[i].cpus->dai_name = dev_name(ljm_dev->dev);
        card->dai_link[i].num_cpus = 1;
        card->dai_link[i].platforms->name = dev_name(ljm_dev->dev);
        card->dai_link[i].num_platforms = 1;
        card->dai_link[i].codecs->name = dev_name(ljm_dev->dev);
        card->dai_link[i].num_codecs = 1;
    }
#else
    for (i = 0; i < card->num_links; i++) {
        card->dai_link[i].cpu_dai_name = dev_name(ljm_dev->dev);
        card->dai_link[i].platform_name = dev_name(ljm_dev->dev);
        card->dai_link[i].codec_dai_name = "ljm-i2s-hifi";
        card->dai_link[i].codec_name = dev_name(ljm_dev->dev);
        card->dai_link[i].num_codecs = 1;
    }
#endif
    card->dev = ljm_dev->dev;
    ret = devm_snd_soc_register_card(card->dev, card);
    if (ret) {
        dev_err(card->dev,
                "devm_snd_soc_register_card(%s) failed: %d\n",
                card->name, ret);
        return ret;
    }
    return 0;
}

#endif