/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_base.h
 *  版 本 号: V1.0
 *  作    者: 胡艳明，刘天玥
 *  生成日期: 2022年4月15日
 *
 *  功能描述: 架构无关的硬件相关头文件
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 创建文件
 *
 ******************************************************************************/

#ifndef __LJMICRO_BASE_H__
#define __LJMICRO_BASE_H__

#ifdef UEFI_GOP

#include <IndustryStandard/Pci.h>
#include <Library/BaseMemoryLib.h>
#include <Library/DebugLib.h>
#include <Library/DevicePathLib.h>
#include <Library/MemoryAllocationLib.h>
#include <Library/PcdLib.h>
#include <Library/TimerLib.h>
#include <Library/UefiBootServicesTableLib.h>
#include <Library/UefiDriverEntryPoint.h>
#include <Library/UefiLib.h>
#include <Protocol/DevicePath.h>
#include <Protocol/DriverSupportedEfiVersion.h>
#include <Protocol/EdidActive.h>
#include <Protocol/EdidDiscovered.h>
#include <Protocol/EdidOverride.h>
#include <Protocol/GraphicsOutput.h>
#include <Protocol/PciIo.h>
#include <Protocol/UgaDraw.h>
#include <Uefi.h>

#include "ljmicro_family.h"

#define EPERM   1  /* Operation not permitted */
#define ENOENT  2  /* No such file or directory */
#define ESRCH   3  /* No such process */
#define EINTR   4  /* Interrupted system call */
#define EIO     5  /* I/O error */
#define ENXIO   6  /* No such device or address */
#define E2BIG   7  /* Argument list too long */
#define ENOEXEC 8  /* Exec format error */
#define EBADF   9  /* Bad file number */
#define ECHILD  10 /* No child processes */
#define EAGAIN  11 /* Try again */
#define ENOMEM  12 /* Out of memory */
#define EACCES  13 /* Permission denied */
#define EFAULT  14 /* Bad address */
#define ENOTBLK 15 /* Block device required */
#define EBUSY   16 /* Device or resource busy */
#define EEXIST  17 /* File exists */
#define EXDEV   18 /* Cross-device link */
#define ENODEV  19 /* No such device */
#define ENOTDIR 20 /* Not a directory */
#define EISDIR  21 /* Is a directory */
#define EINVAL  22 /* Invalid argument */
#define ENFILE  23 /* File table overflow */
#define EMFILE  24 /* Too many open files */
#define ENOTTY  25 /* Not a typewriter */
#define ETXTBSY 26 /* Text file busy */
#define EFBIG   27 /* File too large */
#define ENOSPC  28 /* No space left on device */
#define ESPIPE  29 /* Illegal seek */
#define EROFS   30 /* Read-only file system */
#define EMLINK  31 /* Too many links */
#define EPIPE   32 /* Broken pipe */
#define EDOM    33 /* Math argument out of domain of func */
#define ERANGE  34 /* Math result not representable */

#define bool BOOLEAN
#define true TRUE
#define false FALSE
//#define resource_size_t   UINT32
//#define size_t            UINT32
#define uint32_t UINT32

#define u8  UINT8
#define u16 UINT16
#define u32 UINT32
#define u64 UINT64

#define BIT(nr) (1UL << (nr))

#define msleep(msecs)          gBS->Stall(msecs * 1000)
#define usleep_range(min, max) gBS->Stall(min > 1000 ? (min / 1000) : 1)
#define udelay(usecs)          gBS->Stall(usecs > 1000 ? (usecs / 1000) : 1)

UINT8 LJM_READ_MMIO8(void *Private, UINTN Offset);
VOID  LJM_WRITE_MMIO8(void *Private, UINTN Offset, UINT8 Data);

UINT16 LJM_READ_MMIO16(void *Private, UINTN Offset);
VOID   LJM_WRITE_MMIO16(void *Private, UINTN Offset, UINT16 Data);

UINT32 LJM_READ_MMIO32(void *Private, UINTN Offset);
VOID   LJM_WRITE_MMIO32(void *Private, UINTN Offset, UINT32 Data);

extern int   gCardCnt;
extern void *gPrivate;
extern void *gPrivates[LJM_MAX_CARDS_NUM];

#if 0
inline void *gBSAllocatePool(unsigned long size)
{
    void *     Buf;
    EFI_STATUS Status;

    Status = gBS->AllocatePool(EfiBootServicesData, size, (VOID **)&Buf);
    if (EFI_ERROR(Status))
    {
        return NULL;
    }

    return Buf;
}
#endif

#define kzalloc(size, flag)    AllocateZeroPool(size)
#define kfree(pointer)         FreePool(pointer)
#define memcpy(dst, src, size) CopyMem(dst, src, size)

#else /* !UEFI_GOP */

#include <asm/io.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
#include <drm/drm_drv.h>
#include <drm/drm_file.h>
#include <drm/drm_ioctl.h>
#include <drm/drm_print.h>
#include <drm/drm_vblank.h>
#include <drm/drm_framebuffer.h>
#include <drm/drm_fourcc.h>
#include <drm/drm_modes.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include <drm/drm_irq.h>
#endif
#else
#include <drm/drmP.h>
#endif

#include <drm/drm_crtc_helper.h>
#include <drm/drm_edid.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include <drm/drm_dp_helper.h>
#else
#include <drm/display/drm_dp_helper.h>
#endif
#include <linux/kern_levels.h>
#include <linux/module.h>

#include "ljmicro_compat.h"
#include "ljmicro_drv.h"

#ifndef LJMICRO_DRV_ONLY
#include "lj_hal_base.h"
#include "lj_hal_kernel_linux.h"
#include "lj_hal_types.h"
#endif

#endif /* ~UEFI_GOP */

#include "ljmicro_family.h"

#ifndef NULL
#ifdef __cplusplus
#define NULL 0
#else
#define NULL ((void *)0)
#endif
#endif

#define IN
#define OUT
#define INOUT
#define OPTIONAL

typedef enum _CONNECTOR_FLAG
{
    CONNECTOR_TYPE_NONE     = 0x00,
    CONNECTOR_TYPE_HDMI     = 0x01,
    CONNECTOR_TYPE_DVI      = 0x02,
    CONNECTOR_TYPE_VGA      = 0x04,
    CONNECTOR_CMD_EDID      = 0x10, /* HDMI read EDID command */
    CONNECTOR_CMD_DELAY_HPD = 0x20, /* HDMI delay to enable hpd interrupt */
} CONNECTOR_FLAG;

#define CONNECTOR_isTypeHDMI(flag)    ((flag & CONNECTOR_TYPE_HDMI) != 0)
#define CONNECTOR_isTypeDVI(flag)     ((flag & CONNECTOR_TYPE_DVI) != 0)
#define CONNECTOR_isTypeNone(flag)    (!CONNECTOR_isTypeHDMI(flag) && !CONNECTOR_isTypeDVI(flag))
#define CONNECTOR_isCmdEDID(flag)     ((flag & CONNECTOR_CMD_EDID) != 0)
#define CONNECTOR_isCmdDelayHPD(flag) ((flag & CONNECTOR_CMD_DELAY_HPD) != 0)

#define CONNECTOR_SetIsTypeHDMI(flag, v) \
    (flag = (flag & ~CONNECTOR_TYPE_HDMI) | ((v) == false ? 0 : CONNECTOR_TYPE_HDMI))

#define CONNECTOR_SetIsTypeDVI(flag, v) (flag = (flag & ~CONNECTOR_TYPE_DVI) | ((v) == false ? 0 : CONNECTOR_TYPE_DVI))

#define CONNECTOR_SetIsCmdEDID(flag, v) (flag = (flag & ~CONNECTOR_CMD_EDID) | ((v) == false ? 0 : CONNECTOR_CMD_EDID))

/* reg for hw init */
struct ljm_init_reg
{
    unsigned char addr;
    unsigned char data;
};

#ifdef PRINT
#undef PRINT
#endif

#ifdef UEFI_GOP
#define PRINT(fmt, ...)       DEBUG((EFI_D_INFO | EFI_D_ERROR, fmt, ##__VA_ARGS__))
#define ERR_PRINT(fmt, ...)   DEBUG((EFI_D_ERROR, fmt, ##__VA_ARGS__))
#define INFO_PRINT(fmt, ...)  DEBUG((EFI_D_INFO, fmt, ##__VA_ARGS__))
#define DEBUG_PRINT(fmt, ...) DEBUG((EFI_D_INFO | EFI_D_ERROR, fmt, ##__VA_ARGS__))
#define WARN_PRINT(fmt, ...)  DEBUG((EFI_D_WARN, fmt, ##__VA_ARGS__))
#else /* !UEFI_GOP */
#define PRINT(fmt, ...)       printk(KERN_INFO fmt, ##__VA_ARGS__)
#define ERR_PRINT(fmt, ...)   DRM_ERROR(fmt, ##__VA_ARGS__)
#define INFO_PRINT(fmt, ...)  DRM_INFO(fmt, ##__VA_ARGS__)
#define DEBUG_PRINT(fmt, ...) DRM_DEBUG_KMS(fmt, ##__VA_ARGS__)
#define WARN_PRINT(fmt, ...)  DRM_WARN(fmt, ##__VA_ARGS__)
#endif /* ~UEFI_GOP */

void dump_bytes(const unsigned char *block, int size, const char *headinfo);

#if defined(_DEBUG) || defined(UEFI_GOP)
#if defined(ljmkTRACE_ZONE)
#define HEADER          ljmkHEADER
#define HEADER_ARG      ljmkHEADER_ARG

#define FOOTER     ljmkFOOTER_NO
#define FOOTER_ARG ljmkFOOTER_ARG

#define TRACE_NO(Text) ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, "%s(%d): " Text, __FUNCTION__, __LINE__)
#define TRACE(Text, ...) \
    ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, "%s(%d): " Text, __FUNCTION__, __LINE__, ##__VA_ARGS__)
#define VERIFY_ARGUMENT(arg) \
    do \
    { \
        if (!(arg)) \
        { \
            ljmkTRACE_ZONE(ljdHEADER_LEVEL, _LJ_OBJ_ZONE, "%s(%d): VERIFY_ARGUMENT failed: %s", __FUNCTION__, \
                           __LINE__, #arg); \
            return -EINVAL; \
        } \
    } \
    while (0)
#else /* !ljmkTRACE_ZONE */
extern char *indent_h(void);
extern char *indent_f(void);
extern char *valid_p(void);

#define TRACE(Text, ...) \
    do \
    { \
        PRINT("%s  %s(%d): " Text "\n", valid_p(), __FUNCTION__, __LINE__, ##__VA_ARGS__); /*msleep(100);*/ \
    } \
    while (0)
#define TRACE_NO(Text) TRACE(Text)

#define HEADER() \
    do \
    { \
        PRINT("%s++%s(%d)\n", indent_h(), __FUNCTION__, __LINE__); /*msleep(100);*/ \
    } \
    while (0)
#define HEADER_ARG(Text, ...) \
    do \
    { \
        PRINT("%s++%s(%d): " Text "\n", indent_h(), __FUNCTION__, __LINE__, ##__VA_ARGS__); /*msleep(100);*/ \
    } \
    while (0)

#define FOOTER() \
    do \
    { \
        PRINT("%s--%s(%d)\n", indent_f(), __FUNCTION__, __LINE__); /*msleep(100);*/ \
    } \
    while (0)
#define FOOTER_ARG(Text, ...) \
    do \
    { \
        PRINT("%s--%s(%d): " Text "\n", indent_f(), __FUNCTION__, __LINE__, ##__VA_ARGS__); /*msleep(100);*/ \
    } \
    while (0)
#define VERIFY_ARGUMENT(arg) \
    do \
    { \
        if (!(arg)) \
        { \
            PRINT("%s(%d): VERIFY_ARGUMENT failed: %s\n", __FUNCTION__, __LINE__, #arg); /*msleep(100);*/ \
            return -EINVAL; \
        } \
    } \
    while (0)
#endif  /* ~ljmkTRACE_ZONE */
#else /* !(_DEBUG || UEFI_GOP) */
#define TRACE(Text, ...)
#define TRACE_NO(Text)

#define HEADER()
#define HEADER_ARG(Text, ...)

#define FOOTER()
#define FOOTER_ARG(Text, ...)
#define VERIFY_ARGUMENT(arg)
#endif /* ~ (_DEBUG || UEFI_GOP) */

#define DEBUG_DUMP_REG  0 /* dump registers switch in debug mode */
#define DEBUG_DUMP_EDID 0

#if DEBUG_DUMP_REG
#define DUMP_WREG8(Adr, Val)  PRINT("%s(%d): R[0x%x]=0x%02x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(u8)(Val))
#define DUMP_WREG16(Adr, Val) PRINT("%s(%d): R[0x%x]=0x%04x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(u16)(Val))
#define DUMP_WREG32(Adr, Val) PRINT("%s(%d): R[0x%x]=0x%08x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(Val))
#define DUMP_RREG8(Adr)       PRINT("%s(%d): R[0x%x]=?\n", __FUNCTION__, __LINE__, (u32)(Adr))
#define DUMP_RREG16(Adr)      PRINT("%s(%d): R[0x%x]=?\n", __FUNCTION__, __LINE__, (u32)(Adr))
#define DUMP_RREG32(Adr)      PRINT("%s(%d): R[0x%x]=?\n", __FUNCTION__, __LINE__, (u32)(Adr))
#define DUMP_WCR(Adr, Val)    PRINT("%s(%d): CR[0x%x]=0x%02x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(u8)(Val))
#define DUMP_WSR(Adr, Val)    PRINT("%s(%d): SR[0x%x]=0x%02x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(u8)(Val))
#define DUMP_WGR(Adr, Val)    PRINT("%s(%d): GR[0x%x]=0x%02x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(u8)(Val))
#define DUMP_WAR(Adr, Val)    PRINT("%s(%d): AR[0x%x]=0x%02x\n", __FUNCTION__, __LINE__, (u32)(Adr), (u32)(u8)(Val))

enum gp_read_type
{
    GP_READ_TYPE_AR,
    GP_READ_TYPE_SR,
    GP_READ_TYPE_GR,
    GP_READ_TYPE_CR,
};

void dump_vga_regs(int Card, struct ljm_init_reg *Regs, int Size, enum gp_read_type Type);

#define DUMP_VGA_REGS(Card, Regs, Size, Type) dump_vga_regs(Card, Regs, Size, Type)
#else /* !DEBUG_DUMP_REG */
#define DUMP_WREG8(Adr, Val)
#define DUMP_WREG16(Adr, Val)
#define DUMP_WREG32(Adr, Val)
#define DUMP_RREG8(Adr)
#define DUMP_RREG16(Adr)
#define DUMP_RREG32(Adr)
#define DUMP_WCR(Adr, Val)
#define DUMP_WSR(Adr, Val)
#define DUMP_WGR(Adr, Val)
#define DUMP_WAR(Adr, Val)

#define DUMP_VGA_REGS(Card, Regs, Size, Type)
#endif /* ~DEBUG_DUMP_REG */

/*
 * register operation
 */
#ifdef UEFI_GOP
/* Read / write 8-bit register */
#define RREG8(reg) LJM_READ_MMIO8(gPrivate, reg)
#define WREG8(reg, v) \
    do \
    { \
        DUMP_WREG8(reg, v); \
        LJM_WRITE_MMIO8(gPrivate, reg, (UINT8)v); \
    } \
    while (0)

/* Read / Write 32-bit register */
#define RREG32(reg) LJM_READ_MMIO32(gPrivate, reg)
#define WREG32(reg, v) \
    do \
    { \
        DUMP_WREG32(reg, v); \
        LJM_WRITE_MMIO32(gPrivate, reg, (UINT32)v); \
    } \
    while (0)

/* Read / write 8-bit register */
#define RREG8_BASE(private, reg) LJM_READ_MMIO8(private, reg)
#define WREG8_BASE(private, reg, v) \
    do \
    { \
        DUMP_WREG8(reg, v); \
        LJM_WRITE_MMIO8(private, reg, (UINT8)v); \
    } \
    while (0)

/* Read / Write 32-bit register */
#define RREG32_BASE(private, reg) LJM_READ_MMIO32(private, reg)
#define WREG32_BASE(private, reg, v) \
    do \
    { \
        DUMP_WREG32(reg, v); \
        LJM_WRITE_MMIO32(private, reg, (UINT32)v); \
    } \
    while (0)

/* Read / write 8-bit register */
#define REG_R8(card, reg) LJM_READ_MMIO8(gPrivates[card], reg)
#define REG_W8(card, reg, v) \
    do \
    { \
        DUMP_WREG8(reg, v); \
        LJM_WRITE_MMIO8(gPrivates[card], reg, (UINT8)v); \
    } \
    while (0)

/* Read / write 8-bit register */
#define REG_R16(card, reg) LJM_READ_MMIO16(gPrivates[card], reg)
#define REG_W16(card, reg, v) \
    do \
    { \
        DUMP_WREG8(reg, v); \
        LJM_WRITE_MMIO16(gPrivates[card], reg, (UINT16)v); \
    } \
    while (0)


/* Read / Write 32-bit register */
#define REG_R32(card, reg) LJM_READ_MMIO32(gPrivates[card], reg)
#define REG_W32(card, reg, v) \
    do \
    { \
        DUMP_WREG32(reg, v); \
        LJM_WRITE_MMIO32(gPrivates[card], reg, (UINT32)v); \
    } \
    while (0)

#else /* !UEFI_GOP */
/* Read / write 8-bit register */
#define RREG8(reg) readb(g_mmio_base + reg)
#define WREG8(reg, v) \
    do \
    { \
        DUMP_WREG8(reg, v); \
        writeb((u8)(v), (g_mmio_base + reg)); \
    } \
    while (0)

/* Read / write 16-bit register */
#define RREG16(reg) readw(g_mmio_base + reg)
#define WREG16(reg, v) \
    do \
    { \
        DUMP_WREG16(reg, v); \
        writew((u16)(v), g_mmio_base + reg); \
    } \
    while (0)

/* Read / Write 32-bit register */
#define RREG32(reg) readl(g_mmio_base + reg)
#define WREG32(reg, v) \
    do \
    { \
        DUMP_WREG32(reg, v); \
        writel((u32)(v), g_mmio_base + reg); \
    } \
    while (0)

/* Read / write 8-bit register */
#define RREG8_BASE(base, offset) readb((void __iomem *)(base) + (offset))
#define WREG8_BASE(base, offset, v) \
    do \
    { \
        DUMP_WREG8(offset, v); \
        writeb((u8)(v), (void __iomem *)(base) + (offset)); \
    } \
    while (0)
/* Read / write 16-bit register */
#define RREG16_BASE(base, offset) readw((void __iomem *)(base) + (offset))
#define WREG16_BASE(base, offset, v) \
    do \
    { \
        DUMP_WREG16(offset, v); \
        writew((u16)(v), (void __iomem *)(base) + (offset)); \
    } \
    while (0)
/* Read / Write 32-bit register */
#define RREG32_BASE(base, offset) readl((void __iomem *)(base) + (offset))
#define WREG32_BASE(base, offset, v) \
    do \
    { \
        DUMP_WREG32(offset, v); \
        writel((u32)(v), (void __iomem *)(base) + (offset)); \
    } \
    while (0)

/* Read / write 8-bit register */
#define REG_R8(card, reg) readb(g_mmio_bases[card] + reg)
#define REG_W8(card, reg, v) \
    do \
    { \
        DUMP_WREG8(reg, v); \
        writeb((u8)(v), (g_mmio_bases[card] + reg)); \
    } \
    while (0)

/* Read / write 16-bit register */
#define REG_R16(card, reg) readw(g_mmio_bases[card] + reg)
#define REG_W16(card, reg, v) \
    do \
    { \
        DUMP_WREG16(reg, v); \
        writew((u16)(v), g_mmio_bases[card] + reg); \
    } \
    while (0)

/* Read / Write 32-bit register */
#define REG_R32(card, reg) readl(g_mmio_bases[card] + reg)
#define REG_W32(card, reg, v) \
    do \
    { \
        DUMP_WREG32(reg, v); \
        writel((u32)(v), g_mmio_bases[card] + reg); \
    } \
    while (0)
#endif /* ~UEFI_GOP */

/**********************************************
 * General VGA registers and they operate funcs
 ***********************************************/
#define MMIO_OFFSET 0x0 /* Offset of mmio (for display controller) */
#define MMIO_SIZE   0x1000000 /* Size of mmio for GP device */

#define REG_VGA_OFFSET (0x10000)

#define CRTC_XIN_GP101 27000 /* CRTC clock in: 27000 kHz */

/* EDID chip */
#define DDC_EDID_SIZE    0x80 /* EDID block size */
#define DDC_EDID_ADDR    0x50 /* EDID chip (EEPROM) address is 0x50 */
#define DDC_WR_BIT       0x0
#define DDC_RD_BIT       0x1
#define DDC_EDID_WR_ADDR ((DDC_EDID_ADDR << 1) | DDC_WR_BIT) /* I2C chip write address */
#define DDC_EDID_RD_ADDR ((DDC_EDID_ADDR << 1) | DDC_RD_BIT) /* I2C chip read address */
#define DDC_SEG_PTR_ADDR 0x30 /* E-DDC segment pointer address */
#define DDC_SCDC_ADDR    0x54 /* SCDC slave address */

#define DDC_TIMEOUT_CNT 0xffff /* Read ST max times to wait TIP down */
#define DDC_ERROR       (-1) /* DDC error */
#define DDC_TIMEOUT     (-2) /* wait TIP timeout */
#define DDC_NACK        (-3) /* DDC NACK */

#define VGAREG_ACTL_ADDRESS    (0x3c0 + REG_VGA_OFFSET)
#define VGAREG_ACTL_WRITE_DATA (0x3c0 + REG_VGA_OFFSET)
#define VGAREG_ACTL_READ_DATA  (0x3c1 + REG_VGA_OFFSET)

#define VGAREG_INPUT_STATUS      (0x3c2 + REG_VGA_OFFSET)
#define VGAREG_WRITE_MISC_OUTPUT (0x3c2 + REG_VGA_OFFSET)
#define VGAREG_VIDEO_ENABLE      (0x3c3 + REG_VGA_OFFSET)
#define VGAREG_SEQU_ADDRESS      (0x3c4 + REG_VGA_OFFSET)
#define VGAREG_SEQU_DATA         (0x3c5 + REG_VGA_OFFSET)

#define VGAREG_PEL_MASK          (0x3c6 + REG_VGA_OFFSET)
#define VGAREG_DAC_STATE         (0x3c7 + REG_VGA_OFFSET)
#define VGAREG_DAC_READ_ADDRESS  (0x3c7 + REG_VGA_OFFSET)
#define VGAREG_DAC_WRITE_ADDRESS (0x3c8 + REG_VGA_OFFSET)
#define VGAREG_DAC_DATA          (0x3c9 + REG_VGA_OFFSET)

#define VGAREG_READ_FEATURE_CTL (0x3ca + REG_VGA_OFFSET)
#define VGAREG_READ_MISC_OUTPUT (0x3cc + REG_VGA_OFFSET)

#define VGAREG_GRDC_ADDRESS (0x3ce + REG_VGA_OFFSET)
#define VGAREG_GRDC_DATA    (0x3cf + REG_VGA_OFFSET)

#define VGAREG_MDA_CRTC_ADDRESS (0x3b4 + REG_VGA_OFFSET)
#define VGAREG_MDA_CRTC_DATA    (0x3b5 + REG_VGA_OFFSET)
#define VGAREG_VGA_CRTC_ADDRESS (0x3d4 + REG_VGA_OFFSET)
#define VGAREG_VGA_CRTC_DATA    (0x3d5 + REG_VGA_OFFSET)

#define VGAREG_MDA_WRITE_FEATURE_CTL (0x3ba + REG_VGA_OFFSET)
#define VGAREG_VGA_WRITE_FEATURE_CTL (0x3da + REG_VGA_OFFSET)
#define VGAREG_MDA_ACTL_RESET        (0x3ba + REG_VGA_OFFSET)
#define VGAREG_VGA_ACTL_RESET        (0x3da + REG_VGA_OFFSET)

#define VGAREG_INPUT_STATUS_1      (0x3da + REG_VGA_OFFSET) /* Input status register 1, read only */
#define INPUT_STATUS_1_ACTIVE_MASK 0x1 /* 0: composite blank, 1: display active */
#define INPUT_STATUS_1_VBLANK_MASK 0x8 /* 0: not vertical blank, 1: vertical blank */

#define VGAREG_MDA_MODECTL (0x3b8 + REG_VGA_OFFSET)
#define VGAREG_CGA_MODECTL (0x3d8 + REG_VGA_OFFSET)
#define VGAREG_CGA_PALETTE (0x3d9 + REG_VGA_OFFSET)

/* CRTC base mode */
#define CRTC_BASE_LOW    (0xC) /* base address [15:8], 32-bit address */
#define CRTC_BASE_LOWEST (0xD) /* base address [7:0], 32-bit address */
#define CRTC_PITCH_LOW   (0x13) /* pitch [7:0], 64 bit address */
#define CRTC_SYNC_CTRL   (0x17) /* hsync and vsync enable [7]: 0: disable, crtc0 don't read vram; 1: enable */

#ifdef UEFI_GOP
/* CRTC registers */
#define CR_WR(card, ind, val) \
    do \
    { \
        DUMP_WCR((ind), (val)); \
        LJM_WRITE_MMIO16(gPrivates[card], VGAREG_VGA_CRTC_ADDRESS,  ((UINT16)(val)<<8) | (UINT16)(ind)); \
    } \
    while (0)
#define CR_RD(card, ind) \
    (LJM_WRITE_MMIO8(gPrivates[card], VGAREG_VGA_CRTC_ADDRESS, (UINT8)(ind)), \
     LJM_READ_MMIO8(gPrivates[card], VGAREG_VGA_CRTC_DATA))

/* graphics registers */
#define GR_WR(card, ind, val) \
    do \
    { \
        DUMP_WGR(ind, val); \
        LJM_WRITE_MMIO8(gPrivates[card], VGAREG_GRDC_ADDRESS, (UINT8)(ind)); \
        LJM_WRITE_MMIO8(gPrivates[card], VGAREG_GRDC_DATA, (UINT8)(val)); \
    } \
    while (0)
#define GR_RD(card, ind) \
    (LJM_WRITE_MMIO8(gPrivates[card], VGAREG_GRDC_ADDRESS, (UINT8)(ind)), \
     LJM_READ_MMIO8(gPrivates[card], VGAREG_GRDC_DATA))

/* sequencer registers */
#define SR_WR(card, ind, val) \
    do \
    { \
        DUMP_WSR(ind, val); \
        LJM_WRITE_MMIO8(gPrivates[card], VGAREG_SEQU_ADDRESS, (UINT8)(ind)); \
        LJM_WRITE_MMIO8(gPrivates[card], VGAREG_SEQU_DATA, (UINT8)(val)); \
    } \
    while (0)
#define SR_RD(card, ind) \
    (LJM_WRITE_MMIO8(gPrivates[card], VGAREG_SEQU_ADDRESS, (UINT8)(ind)), \
     LJM_READ_MMIO8(gPrivates[card], VGAREG_SEQU_DATA))

/* attribute registers */
/* 索引端口: 0x3C0(可读写)，数据端口: 0x3C0(只写)/0x3C1(只读)
 * 0x3DA端口的读指向索引端口，对0x3C0端口的写改变0x3C0端口的指向。
 */
#define AR_WR(card, ind, val) \
    do \
    { \
        DUMP_WAR(ind, val); \
        LJM_READ_MMIO8(gPrivates[card], VGAREG_VGA_ACTL_RESET); \
        LJM_WRITE_MMIO8(gPrivates[card], VGAREG_ACTL_ADDRESS, (UINT8)(ind)); \
        LJM_WRITE_MMIO8(gPrivates[card], VGAREG_ACTL_WRITE_DATA, (UINT8)(val)); \
    } \
    while (0)
#define AR_RD(card, ind) \
    (LJM_READ_MMIO8(gPrivates[card], VGAREG_VGA_ACTL_RESET), \
     LJM_WRITE_MMIO8(gPrivates[card], VGAREG_ACTL_ADDRESS, (UINT8)(ind)), \
     LJM_READ_MMIO8(gPrivates[card], VGAREG_ACTL_READ_DATA))

#else /* !UEFI_GOP */
/* CRTC registers */
void ljm_crtc_reg_write(int card, uint8_t ind, uint8_t val);
uint8_t ljm_crtc_reg_read(int card, uint8_t ind);
#define CR_WR(card, ind, val) \
    do \
    { \
        DUMP_WCR((ind), (val)); \
        ljm_crtc_reg_write((card), (ind), (val)); \
    } \
    while (0)
#define CR_RD(card, ind) ljm_crtc_reg_read((card), (ind))

/* graphics registers */
#define GR_WR(card, ind, val) \
    do \
    { \
        DUMP_WGR(ind, val); \
        writeb((u8)(ind), g_mmio_bases[card] + VGAREG_GRDC_ADDRESS); \
        writeb((u8)(val), g_mmio_bases[card] + VGAREG_GRDC_DATA); \
    } \
    while (0)
#define GR_RD(card, ind) \
    (writeb((u8)(ind), g_mmio_bases[card] + VGAREG_GRDC_ADDRESS), readb(g_mmio_bases[card] + VGAREG_GRDC_DATA))

/* sequencer registers */
#define SR_WR(card, ind, val) \
    do \
    { \
        DUMP_WSR(ind, val); \
        writeb((u8)(ind), g_mmio_bases[card] + VGAREG_SEQU_ADDRESS); \
        writeb((u8)(val), g_mmio_bases[card] + VGAREG_SEQU_DATA); \
    } \
    while (0)
#define SR_RD(card, ind) \
    (writeb((u8)(ind), g_mmio_bases[card] + VGAREG_SEQU_ADDRESS), readb(g_mmio_bases[card] + VGAREG_SEQU_DATA))

/* attribute registers */
/* 索引端口: 0x3C0(可读写)，数据端口: 0x3C0(只写)/0x3C1(只读)
 * 0x3DA端口的读指向索引端口，对0x3C0端口的写改变0x3C0端口的指向。
 */
#define AR_WR(card, ind, val) \
    do \
    { \
        DUMP_WAR(ind, val); \
        readb(g_mmio_bases[card] + VGAREG_VGA_ACTL_RESET); \
        writeb((u8)(ind), g_mmio_bases[card] + VGAREG_ACTL_ADDRESS); \
        writeb((u8)(val), g_mmio_bases[card] + VGAREG_ACTL_WRITE_DATA); \
    } \
    while (0)
#define AR_RD(card, ind) \
    (readb(g_mmio_bases[card] + VGAREG_VGA_ACTL_RESET), writeb((u8)(ind), g_mmio_bases[card] + VGAREG_ACTL_ADDRESS), \
     readb(g_mmio_bases[card] + VGAREG_ACTL_READ_DATA))

#ifdef LJMICRO_DRV_ONLY
#define GetLjmDevFrmDrm(ddev)          ((struct ljm_device *)(ddev->dev_private))
#define SetLjmDevFrmDrm(ddev, ljm_dev) ((ddev)->dev_private = (void *)(ljm_dev))
//#define GetPdevFrmDrm(ddev)             (ddev->pdev)
#else
#define GetGalFrmDrm(ddev)             ((ljkGALDEVICE)((ddev)->dev_private))
#define GetGalFrmDev(dev)              (ljkGALDEVICE)(dev_get_drvdata(dev))
#define GetLjmDevFrmDrm(ddev)          ((struct ljm_device *)(((ljkGALDEVICE)ddev->dev_private)->ljm_dev))
#define SetLjmDevFrmDrm(ddev, ljm_dev) (((ljkGALDEVICE)ddev->dev_private)->ljm_dev = (struct ljm_device *)ljm_dev)
//#define GetPdevFrmDrm(ddev)             (((struct _gcsPLATFORM_PCIE
//*)(((ljkGALDEVICE)ddev->dev_private)->platform))->pcie.pdev)
#endif
#endif /* ~UEFI_GOP */

#define LjmIsFpga(flags) (flags & LJMICRO_IS_FPGA)

/* HDMI interrupt type */
#define LJM_IRQ_TYPE_NONE  0x0
#define LJM_IRQ_TYPE_HPD   0x1 /* interrupt for hot plug detect */
#define LJM_IRQ_TYPE_VSYNC 0x2 /* interrupt for active vsync edge */
#define LJM_IRQ_TYPE_EDID  0x4 /* interrupt for EDID ready */

#ifdef UEFI_GOP
/* Video mode flags */
/* bit compatible with the xorg definitions. */
#define DRM_MODE_FLAG_PHSYNC    (1 << 0)
#define DRM_MODE_FLAG_NHSYNC    (1 << 1)
#define DRM_MODE_FLAG_PVSYNC    (1 << 2)
#define DRM_MODE_FLAG_NVSYNC    (1 << 3)
#define DRM_MODE_FLAG_INTERLACE (1 << 4)
#define DRM_MODE_FLAG_DBLSCAN   (1 << 5)
#define DRM_MODE_FLAG_CSYNC     (1 << 6)
#define DRM_MODE_FLAG_PCSYNC    (1 << 7)
#define DRM_MODE_FLAG_NCSYNC    (1 << 8)
#define DRM_MODE_FLAG_HSKEW     (1 << 9) /* hskew provided */
#define DRM_MODE_FLAG_BCAST     (1 << 10)
#define DRM_MODE_FLAG_PIXMUX    (1 << 11)
#define DRM_MODE_FLAG_DBLCLK    (1 << 12)
#define DRM_MODE_FLAG_CLKDIV2   (1 << 13)

/* DPMS flags */
/* bit compatible with the xorg definitions. */
#define DRM_MODE_DPMS_ON      0
#define DRM_MODE_DPMS_STANDBY 1
#define DRM_MODE_DPMS_SUSPEND 2
#define DRM_MODE_DPMS_OFF     3

/**
 * struct drm_display_mode - DRM kernel-internal display mode structure
 * @hdisplay: horizontal display size
 * @hsync_start: horizontal sync start
 * @hsync_end: horizontal sync end
 * @htotal: horizontal total size
 * @vdisplay: vertical display size
 * @vsync_start: vertical sync start
 * @vsync_end: vertical sync end
 * @vtotal: vertical total size
 *
 * The horizontal and vertical timings are defined per the following diagram.
 *
 * ::
 *
 *
 *               Active                 Front           Sync           Back
 *              Region                 Porch                          Porch
 *     <-----------------------><----------------><-------------><-------------->
 *       //////////////////////|
 *      ////////////////////// |
 *     //////////////////////  |..................               ................
 *                                                _______________
 *     <----- [hv]display ----->
 *     <------------- [hv]sync_start ------------>
 *     <--------------------- [hv]sync_end --------------------->
 *     <-------------------------------- [hv]total ----------------------------->*
 */
struct drm_display_mode
{
    /**
     * @clock:
     *
     * Pixel clock in kHz.
     */
    int clock; /* in kHz */
    int hdisplay;
    int hsync_start;
    int hsync_end;
    int htotal;
    int vdisplay;
    int vsync_start;
    int vsync_end;
    int vtotal;
    int vscan;
    /**
     * @flags:
     *
     * Sync and timing flags:
     *
     *  - DRM_MODE_FLAG_PHSYNC: horizontal sync is active high.
     *  - DRM_MODE_FLAG_NHSYNC: horizontal sync is active low.
     *  - DRM_MODE_FLAG_PVSYNC: vertical sync is active high.
     *  - DRM_MODE_FLAG_NVSYNC: vertical sync is active low.
     *  - DRM_MODE_FLAG_INTERLACE: mode is interlaced.
     *  - DRM_MODE_FLAG_DBLSCAN: mode uses doublescan.
     *  - DRM_MODE_FLAG_CSYNC: mode uses composite sync.
     *  - DRM_MODE_FLAG_PCSYNC: composite sync is active high.
     *  - DRM_MODE_FLAG_NCSYNC: composite sync is active low.
     *  - DRM_MODE_FLAG_HSKEW: hskew provided (not used?).
     *  - DRM_MODE_FLAG_BCAST: not used?
     *  - DRM_MODE_FLAG_PIXMUX: not used?
     *  - DRM_MODE_FLAG_DBLCLK: double-clocked mode.
     *  - DRM_MODE_FLAG_CLKDIV2: half-clocked mode.
     *
     * Additionally there's flags to specify how 3D modes are packed:
     *
     *  - DRM_MODE_FLAG_3D_NONE: normal, non-3D mode.
     *  - DRM_MODE_FLAG_3D_FRAME_PACKING: 2 full frames for left and right.
     *  - DRM_MODE_FLAG_3D_FIELD_ALTERNATIVE: interleaved like fields.
     *  - DRM_MODE_FLAG_3D_LINE_ALTERNATIVE: interleaved lines.
     *  - DRM_MODE_FLAG_3D_SIDE_BY_SIDE_FULL: side-by-side full frames.
     *  - DRM_MODE_FLAG_3D_L_DEPTH: ?
     *  - DRM_MODE_FLAG_3D_L_DEPTH_GFX_GFX_DEPTH: ?
     *  - DRM_MODE_FLAG_3D_TOP_AND_BOTTOM: frame split into top and bottom
     *    parts.
     *  - DRM_MODE_FLAG_3D_SIDE_BY_SIDE_HALF: frame split into left and
     *    right parts.
     */
    unsigned int flags;
    /**
     * @vrefresh:
     *
     * Vertical refresh rate, for debug output in human readable form. Not
     * used in a functional way.
     *
     * This value is in Hz.
     */
     /*内核5.10.0以上版本去掉vrefresh,需要计算得出*/
    /*int vrefresh;*/
};

struct fb_videomode
{
    const char *name;    /* optional */
    u32         refresh; /* optional */
    u32         xres;
    u32         yres;
    u32         pixclock;
    u32         left_margin;
    u32         right_margin;
    u32         upper_margin;
    u32         lower_margin;
    u32         hsync_len;
    u32         vsync_len;
    u32         sync;
    u32         vmode;
    u32         flag;
};
#endif /* ~UEFI_GOP */

#define IMAGE_HEAD_IDENTIFICATION 0x0709

/*
 * board information definition
 */
typedef struct _ljm_board_info
{
    uint16_t vendor_id;         /**< 0x0: 0x0709 */
    uint16_t device_id;         /**< 0x2: 0x0201 */
    uint16_t version;           /**< 0x4: xx.yy */
    uint16_t size;              /**< 0x6: size of board info - 1 */
    uint32_t ddr_imem_start;    /**< 0x8: imem start address for ddr */
    uint32_t ddr_cfg_start;     /**< 0xC: config start address for ddr */
    uint32_t vga0_en:1;         /**< 0x10: [0] VGA0 enable */
    uint32_t vga1_en:1;         /**< 0x10: [1] VGA1 enable */
    uint32_t vga2_en:1;         /**< 0x10: [2] VGA2 enable */
    uint32_t vga3_en:1;         /**< 0x10: [3] VGA3 enable */
    uint32_t hdmi0_en:1;        /**< 0x10: [4] HDMI0 enable */
    uint32_t hdmi1_en:1;        /**< 0x10: [5] HDMI1 enable */
    uint32_t hdmi2_en:1;        /**< 0x10: [6] HDMI2 enable */
    uint32_t hdmi3_en:1;        /**< 0x10: [7] HDMI3 enable */
    uint32_t dvo0_en:1;         /**< 0x11: [0] DVO0 enable */
    uint32_t dvo1_en:1;         /**< 0x11: [1] DVO1 enable */
    uint32_t dvo2_en:1;         /**< 0x11: [2] DVO2 enable */
    uint32_t dvo3_en:1;         /**< 0x11: [3] DVO3 enable */
    uint32_t dp0_en:1;          /**< 0x11: [4] DP0 enable */
    uint32_t dp1_en:1;          /**< 0x11: [5] DP1 enable */
    uint32_t dp2_en:1;          /**< 0x11: [6] DP2 enable */
    uint32_t dp3_en:1;          /**< 0x11: [7] DP3 enable */
    uint32_t hdmi0_vga_en:1;    /**< 0x12: [0] HDMI0 VGA enable for DVI-I */
    uint32_t hdmi0_rsv:1;       /**< 0x12: [1] HDMI0 reserved */
    uint32_t hdmi0_vga_sel:2;   /**< 0x12: [3:2] which VGA channel is selected by HDMI0 if hdmi0_vga_en==1 */
    uint32_t hdmi1_vga_en:1;    /**< 0x12: [4] HDMI1 VGA enable for DVI-I */
    uint32_t hdmi1_rsv:1;       /**< 0x12: [5] HDMI1 reserved */
    uint32_t hdmi1_vga_sel:2;   /**< 0x12: [7:6] which VGA channel is selected by HDMI1 if hdmi1_vga_en==1 */
    uint32_t hdmi2_vga_en:1;    /**< 0x13: [0] HDMI2 VGA enable for DVI-I */
    uint32_t hdmi2_rsv:1;       /**< 0x13: [1] HDMI2 reserved */
    uint32_t hdmi2_vga_sel:2;   /**< 0x13: [3:2] which VGA channel is selected by HDMI2 if hdmi2_vga_en==1 */
    uint32_t hdmi3_vga_en:1;    /**< 0x13: [4] HDMI3 VGA enable for DVI-I */
    uint32_t hdmi3_rsv:1;       /**< 0x13: [5] HDMI3 reserved */
    uint32_t hdmi3_vga_sel:2;   /**< 0x13: [7:6] which VGA channel is selected by HDMI3 if hdmi3_vga_en==1 */
    uint32_t dvo0_vga_en:1;     /**< 0x14: [0] HDMI0 VGA enable for DVI-I */
    uint32_t dvo0_rsv:1;        /**< 0x14: [1] HDMI0 reserved */
    uint32_t dvo0_vga_sel:2;    /**< 0x14: [3:2] which VGA channel is selected by HDMI0 if hdmi0_vga_en==1 */
    uint32_t dvo1_vga_en:1;     /**< 0x14: [4] HDMI1 VGA enable for DVI-I */
    uint32_t dvo1_rsv:1;        /**< 0x14: [5] HDMI1 reserved */
    uint32_t dvo1_vga_sel:2;    /**< 0x14: [7:6] which VGA channel is selected by HDMI1 if hdmi1_vga_en==1 */
    uint32_t dvo2_vga_en:1;     /**< 0x15: [0] HDMI2 VGA enable for DVI-I */
    uint32_t dvo2_rsv:1;        /**< 0x15: [1] HDMI2 reserved */
    uint32_t dvo2_vga_sel:2;    /**< 0x15: [3:2] which VGA channel is selected by HDMI2 if hdmi2_vga_en==1 */
    uint32_t dvo3_vga_en:1;     /**< 0x15: [4] HDMI3 VGA enable for DVI-I */
    uint32_t dvo3_rsv:1;        /**< 0x15: [5] HDMI3 reserved */
    uint32_t dvo3_vga_sel:2;    /**< 0x15: [7:6] which VGA channel is selected by HDMI3 if hdmi3_vga_en==1 */
    uint32_t dp0_edp_en:1;      /**< 0x16: [0] DP0 enable edp */
    uint32_t dp0_aux_gpio_en:1; /**< 0x16: [1] DP0 aux using GPIO enable */
    uint32_t dp0_aux_gpio_sel:2;/**< 0x16: if dp0_aux_gpio_en==1, [3:2] DP0 aux GPIO group selection, two sequent GPIOs in a group.
                                 *         if dp0_aux_gpio_en==0, [2] DP0 aux virtual DPCD and EDID enable and [3] reserved */
    uint32_t dp1_edp_en:1;      /**< 0x16: [4] DP1 enable edp */
    uint32_t dp1_aux_gpio_en:1; /**< 0x16: [5] DP1 aux using GPIO enable */
    uint32_t dp1_aux_gpio_sel:2;/**< 0x16: if dp1_aux_gpio_en==1, [7:6] DP1 aux GPIO group selection, two sequent GPIOs in a group.
                                 *         if dp1_aux_gpio_en==0, [6] DP1 aux virtual DPCD and EDID enable and [7] reserved */
    uint32_t dp2_edp_en:1;      /**< 0x17: [0] DP2 enable edp */
    uint32_t dp2_aux_gpio_en:1; /**< 0x17: [1] DP2 aux using GPIO enable */
    uint32_t dp2_aux_gpio_sel:2;/**< 0x17: if dp2_aux_gpio_en==1, [3:2] DP2 aux GPIO group selection, two sequent GPIOs in a group.
                                 *         if dp2_aux_gpio_en==0, [2] DP2 aux virtual DPCD and EDID enable and [3] reserved */
    uint32_t dp3_edp_en:1;      /**< 0x17: [4] DP3 enable edp */
    uint32_t dp3_aux_gpio_en:1; /**< 0x17: [5] DP3 aux using GPIO enable */
    uint32_t dp3_aux_gpio_sel:2;/**< 0x17: if dp3_aux_gpio_en==1, [7:6] DP3 aux GPIO group selection, two sequent GPIOs in a group.
                                 *         if dp3_aux_gpio_en==0, [6] DP3 aux virtual DPCD and EDID enable and [7] reserved */
    uint32_t ddr_ch_mask:8;     /**< 0x18: mask channel for ddr, 0x1 for channel 0, 0x3 for channel both 0 and 1, 0xf for channel 0, 1, 2, 3 */
    uint32_t ddr_total_size:4;  /**< 0x19: [3:0]: ddr size N, total ddr size is 256MB * pow(2, N) */
#define DDR_TYPE_DDR4   0
#define DDR_TYPE_DDR4_1 1
#define DDR_TYPE_LPDDR4 2
    uint32_t ddr_type:4;        /**< 0x19: [7:4]: ddr type, 0 or 1 => DDR4, 2 => LPDDR4/LPDDR4X */
    uint32_t max_hdisplay:16;   /**< 0x1A: max horizontal display for vbios */
    uint32_t date;              /**< 0x1C: binary date data for the board info, such as 0x20220808 */
    uint8_t  board_name[16];    /**< 0x20: board name, such as "PE-10", must be less than 16 bytes */
    uint8_t  ddr_chip_name[20]; /**< 0x30: ddr chip name, such as "MT40A256M16LY-062E" */
    uint32_t core_3d_cnt:4;     /**< 0x44: [3:0]: 3D core num - 1 */
    uint32_t :4;                /**< 0x44: [7:4]: reserved */
    uint32_t ddr_freq_id:8;     /**< 0x45: 0: 800Mbps; 1: 1066Mbps; 2: 1600Mbps; 3: 2400Mbps; 4: 2666Mbps (default); 5: 2933Mbps; 6: 3200Mbps */
    uint32_t ddr_chip_factory:4;/**< 0x46: [3:0]: factory code, 0: changxin, 1: sanxing */
    uint32_t ddr_adv_ch_mask:4;  /**< 0x46: [7:4]: advanced ddr channel mask, 0x1: channel 0, 0x3: channel 0 and 1, 0xc: channel 2 and 3, 0xf: channel 0, 1, 2, 3 */
    uint32_t :8;                /**< 0x47: reserved */
    uint32_t dp0_pwr_gpio_sel:2;/**< 0x48: [1:0] DP0 power GPIO group selection, two sequent GPIOs in a group */
    uint32_t dp0_pwr_gpio_en:1; /**< 0x48: [2] DP0 power GPIO group enable */
    uint32_t :1;                /**< 0x48: [3] reserved */
    uint32_t dp0_pwm_sel:2;     /**< 0x48: [5:4] DP0 PWM selection */
    uint32_t dp0_pwm_en:1;      /**< 0x48: [6] DP0 PWM enable */
    uint32_t :1;                /**< 0x48: [7] reserved */
    uint32_t dp1_pwr_gpio_sel:2;/**< 0x49: [1:0] DP1 power GPIO group selection, two sequent GPIOs in a group */
    uint32_t dp1_pwr_gpio_en:1; /**< 0x49: [2] DP1 power GPIO group enable */
    uint32_t :1;                /**< 0x49: [3] reserved */
    uint32_t dp1_pwm_sel:2;     /**< 0x49: [5:4] DP1 PWM selection */
    uint32_t dp1_pwm_en:1;      /**< 0x49: [6] DP1 PWM enable */
    uint32_t :1;                /**< 0x49: [7] reserved */
    uint32_t dp2_pwr_gpio_sel:2;/**< 0x4A: [1:0] DP2 power GPIO group selection, two sequent GPIOs in a group */
    uint32_t dp2_pwr_gpio_en:1; /**< 0x4A: [2] DP2 power GPIO group enable */
    uint32_t :1;                /**< 0x4A: [3] reserved */
    uint32_t dp2_pwm_sel:2;     /**< 0x4A: [5:4] DP2 PWM selection */
    uint32_t dp2_pwm_en:1;      /**< 0x4A: [6] DP2 PWM enable */
    uint32_t :1;                /**< 0x4A: [7] reserved */
    uint32_t dp3_pwr_gpio_sel:2;/**< 0x4B: [1:0] DP3 power GPIO group selection, two sequent GPIOs in a group */
    uint32_t dp3_pwr_gpio_en:1; /**< 0x4B: [2] DP3 power GPIO group enable */
    uint32_t :1;                /**< 0x4B: [3] reserved */
    uint32_t dp3_pwm_sel:2;     /**< 0x4B: [5:4] DP3 PWM selection */
    uint32_t dp3_pwm_en:1;      /**< 0x4B: [6] DP3 PWM enable */
    uint32_t :1;                /**< 0x4B: [7] reserved */
    uint16_t extra_reg_count;   /**< 0x4C: count of extra registers */
    uint16_t crc;               /**< 0x4E: 16-bit checksum for above items */
} __attribute__((packed)) ljm_board_info;

ljm_board_info *ljm_ddrv2_get_board_info(int card, void *mmio_base);
int ljm_ddr_head_get(void *mmio_base, int card, unsigned long flags);
unsigned long ljm_ddrv2_get_ddr_size(int card, void *mmio_base);

static inline const char *enableddisabled(bool v)
{
    return v ? "enabled" : "disabled";
}

/*
 * break interface
 */
#if defined(_DEBUG) && USE_DEBUG_BREAK

#define DEBUG_BREAK_ALL     0xffffffff
#define DEBUG_BREAK_NEXT    0xfffffffe
#define DEBUG_BREAK_NONE    0x0

unsigned int debug_read_break_id(void);
void debug_write_break_id(unsigned int break_id);
void debug_print_vars(int vars_cnt, ...);
void debug_bug_on(int break_id);
void debug_warn_on(int break_id);

/*****************************************************************************
 *  函 数 名: DEBUG_BREAK宏函数
 *  功能描述: 设置断点，并打印变量的指针
 *  输入参数: BREAK_ID  断点序号，当BREAK_ID==breakId时，运行到该宏函数时将暂停
 *            VARS...   可变参数，通常为需要调试的结构体或变量的指针
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2023年2月10日
 *            作    者: hym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
#define DEBUG_BREAK(BREAK_ID, VARS...) do {                                                     \
    unsigned int __tmp_old = debug_read_break_id();                                             \
    unsigned int __tmp;                                                                         \
    if (__tmp_old == DEBUG_BREAK_ALL || __tmp_old == BREAK_ID) {    /* break */                 \
        int __vars_cnt, __i;                                                                    \
        char *__str;                                                                            \
        for (__i=0, __vars_cnt=1, __str = #VARS; __i<strlen(__str); __i++) {                    \
            if (__str[__i] == ',')                                                              \
                __vars_cnt++;                                                                   \
        }                                                                                       \
        printk(KERN_INFO "%s:%s:%d(%u): " #VARS, __FILE__, __FUNCTION__, __LINE__, BREAK_ID);   \
        debug_print_vars(__vars_cnt, ##VARS);                                                   \
        do {                                                                                    \
            msleep(100);                                                                        \
            __tmp = debug_read_break_id();                                                      \
        } while (__tmp == DEBUG_BREAK_ALL || __tmp == BREAK_ID);                                \
        if (__tmp == DEBUG_BREAK_NEXT)     /* next: restore break id */                         \
            debug_write_break_id(__tmp_old);                                                    \
    }                                                                                           \
} while (0)

#define DEBUG_BUG_ON(BREAK_ID) debug_bug_on(BREAK_ID)
#define DEBUG_WARN_ON(BREAK_ID) debug_warn_on(BREAK_ID)
#else
#define DEBUG_BREAK(BREAK_ID, VARS...)
#define DEBUG_BUG_ON(BREAK_ID)
#define DEBUG_WARN_ON(BREAK_ID)
#endif

#endif /* ~__LJMICRO_BASE_H__ */
