/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#if USE_CLOCK_REGISTER

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/clk.h>
#include <linux/clkdev.h>
#include <linux/clk-provider.h>


#include "ljmicro_device.h"
#include "ljmicro_base.h"
#include "ljmicro_clk.h"
#include "ljmicro_vgav2.h"
#include "ljmicro_family.h"


#define KHZ 1000UL
#define MHZ (KHZ * KHZ)
#define HOST_FREQ (27 * MHZ)

/*Total number of clocks*/
#define NR_CLKS  16

extern void __iomem *g_mmio_base;
extern void __iomem *g_mmio_bases[LJM_MAX_CARDS_NUM];
static void __iomem *ljm_clk_vbase;


#define clk_readl(offset) readl((void __iomem *)(ljm_clk_vbase) + (offset))
#define clk_writel(offset, v) \
do \
{ \
    DUMP_WREG32(offset, v); \
    writel((u32)(v), (void __iomem *)(ljm_clk_vbase) + (offset)); \
    } \
while (0)

struct ljm_clk_pll
{
    struct clk_hw hw;
    u32 reg_ofs;
    u8 reset_bit;
    u8 bypass_bit;
    u32 en_ofs;
    signed char en_bit1;
    signed char en_bit2;
};
#define to_ljm_clk_pll(_hw) container_of(_hw, struct ljm_clk_pll, hw)

struct ljm_clk_pll_data
{
    unsigned int id;
    const char *name;
    const char *parent_name;
    u32 reg_ofs;
    u8 reset_bit;
    u8 bypass_bit;
    u32 en_ofs;
    signed char en_bit1;
    signed char en_bit2;
    unsigned long flags;
};

#define PLL(_id, _name, _parent_name, _reg, _reset_bit, _bypass_bit, _en_ofs, _en_bits1, _en_bits2, _flags) \
{ \
    .id = _id, \
    .name = _name, \
    .parent_name = _parent_name, \
    .flags = _flags, \
    .reg_ofs = _reg, \
    .reset_bit = _reset_bit, \
    .bypass_bit = _bypass_bit, \
    .en_ofs = _en_ofs, \
    .en_bit1 = _en_bits1, \
    .en_bit2 = _en_bits2, \
}

struct ljm_clk_divider
{
    struct clk_hw hw;
    u32 reg_ofs;
    u8 shift;
    u8 width;
};
#define to_ljm_clk_divider(_hw) container_of(_hw, struct ljm_clk_divider, hw)


struct ljm_clk_divider_data
{
    unsigned int id;
    const char *name;
    const char *parent_name;
    u32 reg_ofs;
    u8 shift;
    u8 width;
    unsigned long flags;
};

#define DIV(_id, _name, _parent_name, _reg_ofs, _shift, _width, _flags) \
{ \
    .id = _id, \
    .name = _name, \
    .parent_name = _parent_name, \
    .reg_ofs = _reg_ofs, \
    .shift = _shift, \
    .width = _width, \
    .flags = _flags, \
}

struct ljm_clk_mux
{
    struct clk_hw hw;
    u32 reg_ofs;
    u8 mux_bit;
};
#define to_ljm_clk_mux(_hw) container_of(_hw, struct ljm_clk_mux, hw)

struct ljm_clk_mux_data
{
    unsigned int id;
    const char *name;
    const char *const *parent_names;
    u8 num_parents;
    u32 reg_ofs;
    u8 mux_bits;
    unsigned long flags;
};
#define MUX(_id, _name, _parent_names, _reg_ofs, _mux_bits, _flags) \
{ \
    .id = _id, \
    .name = _name, \
    .parent_names = _parent_names, \
    .num_parents = ARRAY_SIZE(_parent_names), \
    .reg_ofs = _reg_ofs, \
    .mux_bits = _mux_bits, \
    .flags = _flags, \
}


struct ljm_clk_gate
{
    struct clk_hw hw;
    u32 reg_ofs;
    u8 gate_bits;
};
#define to_ljm_clk_gate(_hw) container_of(_hw, struct ljm_clk_gate, hw)


struct ljm_clk_gate_data
{
    unsigned int id;
    const char *name;
    const char *parent_name;
    u32 reg_ofs;
    u8 gate_bits;
    unsigned long flags;
};

#define GATE(_id, _name, _parent_name, _reg, _shift, _flags) \
{ \
    .id = _id, \
    .name = _name, \
    .parent_name = _parent_name, \
    .reg_ofs = _reg, \
    .gate_bits = _shift, \
    .flags = _flags, \
}

enum GP201_CLK_LIST{
    DEVPLL, DEV2PLL, GPUPLL, GPU2PLL, DDRPLL0, DDRPLL1, DDRPLL2, DDRPLL3, PXLPLL0, PXLPLL1, PXLPLL2, PXLPLL3,
    DEVPLL_0, DEVPLL_1, DEVPLL_2, DEVPLL_3, CLK_I2S0, CLK_I2S1,
    CLK_BUS_AXI, CLK_BUS_NOC, CLK_BUS_AHB, CLK_BUS_APB, CLK_PCIE_AXI, CLK_GPU3D_CORE, CLK_GPU3D_SH, CLK_GPU2D_CORE,
    CLK_VPU_CORE, CLK_DDRC0_CORE, CLK_DDRC1_CORE, CLK_DDRC2_CORE, CLK_DDRC3_CORE, HOST_LINE,
    CLK_NUMS,
};

#define PNAME(X) static const char *const X[]

PNAME(clk_bus_axi_p) = { "DEVPLL_0", "HOST", };
PNAME(clk_bus_noc_p) = { "DEVPLL_0", "HOST", };
PNAME(clk_bus_ahb_p) = { "DEVPLL_1", "HOST", };
PNAME(clk_bus_apb_p) = { "DEVPLL_2", "HOST", };
PNAME(clk_pcie_axi_p) = { "DEV2PLL", "HOST", };
PNAME(clk_gpu3d_core_p) = { "GPUPLL", "GPU2PLL", };
PNAME(clk_gpu3d_sh_p) = { "GPUPLL", "GPU2PLL", };
PNAME(clk_gpu2d_core_p) = { "DEVPLL_3", "GPUPLL" };


static const struct ljm_clk_pll_data gp201_clk_plls[] = {
    PLL(DEVPLL, "DEVPLL", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DEVPLL, 31, 15, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 0,-1, 0),
    PLL(DEV2PLL, "DEV2PLL", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DEV2PLL, 30, 14, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 1, -1, 0),
    PLL(GPUPLL, "GPUPLL", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_GPUPLL, 29, 13, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 2, 3, 0),
    PLL(GPU2PLL, "GPU2PLL", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_GPU2PLL, 28, 12, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, -1, -1, 0),
    PLL(DDRPLL0, "DDRPLL0", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DDRPLL, 27, 11, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, -1, -1, 0),
    PLL(DDRPLL1, "DDRPLL1", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DDRPLL, 26, 10, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, -1, -1, 0),
    PLL(DDRPLL2, "DDRPLL2", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DDRPLL, 25, 9, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, -1, -1, 0),
    PLL(DDRPLL3, "DDRPLL3", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DDRPLL, 24, 8, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, -1, -1, 0),
    PLL(PXLPLL0, "PXLPLL0", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_PXLPLL0, 23, 7, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 28, -1, 0),
    PLL(PXLPLL1, "PXLPLL1", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_PXLPLL1, 22, 6, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 29, -1, 0),
    PLL(PXLPLL2, "PXLPLL2", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_PXLPLL2, 21, 5, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 30, -1, 0),
    PLL(PXLPLL3, "PXLPLL3", "HOST", GP201_CRU_CONTROL_REGISTER + GP201_CRU_DDRPLL, 20, 4, GP201_CRU_CONTROL_REGISTER + GP201_CRU_PLLEN, 31, -1, 0),
};

static const struct ljm_clk_divider_data gp201_clk_divider[] = {
    DIV(DEVPLL_0, "DEVPLL_0", "DEVPLL", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKDIV, 8, 4, 0),
    DIV(DEVPLL_1, "DEVPLL_1", "DEVPLL", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKDIV, 4, 4, 0),
    DIV(DEVPLL_2, "DEVPLL_2", "DEVPLL", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKDIV, 0, 4, 0),
    DIV(DEVPLL_3, "DEVPLL_3", "DEVPLL", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKDIV, 0, 0, 0),
    DIV(CLK_I2S0, "CLK_I2S0", "CLK_BUS_AXI", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CNTCLK, 0, 16, 0),
    DIV(CLK_I2S1, "CLK_I2S1", "CLK_BUS_AXI", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CNTCLK, 16, 16, 0),
};

static const struct ljm_clk_mux_data gp201_clk_mux[] = {
    MUX(CLK_BUS_AXI, "CLK_BUS_AXI", clk_bus_axi_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 0, 0),
    MUX(CLK_BUS_NOC, "CLK_BUS_NOC", clk_bus_noc_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 0, 0),
    MUX(CLK_BUS_AHB, "CLK_BUS_AHB", clk_bus_ahb_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 0, 0),
    MUX(CLK_BUS_APB, "CLK_BUS_APB", clk_bus_apb_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 0, 0),
    MUX(CLK_PCIE_AXI, "CLK_PCIE_AXI", clk_pcie_axi_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 0, 0),
    MUX(CLK_GPU3D_CORE, "CLK_GPU3D_CORE", clk_gpu3d_core_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 8, 0),
    MUX(CLK_GPU3D_SH, "CLK_GPU3D_SH", clk_gpu3d_sh_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 9, 0),
    MUX(CLK_GPU2D_CORE, "CLK_GPU3D_CORE", clk_gpu2d_core_p, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKMUX, 10, 0),
};

static const struct ljm_clk_gate_data gp201_clk_gate[] = {
    GATE(CLK_VPU_CORE, "CLK_VPU_CORE", "DEV2PLL", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKEN0, 2, 0),
    GATE(CLK_DDRC0_CORE, "CLK_DDRC0_CORE", "DDRPLL0", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKEN1, 11, 0),
    GATE(CLK_DDRC1_CORE, "CLK_DDRC1_CORE", "DDRPLL1", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKEN1, 10, 0),
    GATE(CLK_DDRC2_CORE, "CLK_DDRC2_CORE", "DDRPLL2", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKEN1, 9, 0),
    GATE(CLK_DDRC3_CORE, "CLK_DDRC3_CORE", "DDRPLL3", GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKEN1, 8, 0),
};


/*Recalculate the rate of this clock, by querying hardware*/
static unsigned long ljm_clk_pll_recalc_rate(struct clk_hw *hw, unsigned long  parent_rate)
{
    struct ljm_clk_pll *clk = to_ljm_clk_pll(hw);
    unsigned long fin = parent_rate;
    unsigned long red_val;
    unsigned long divq;

    red_val = clk_readl(GP201_CRU_CONTROL_REGISTER);
    if (red_val & BIT(clk->bypass_bit))
    {
        /*by pass*/
        return fin;
    }
    else if (red_val & BIT(clk->reset_bit))
    {
        /*power off */
        fin = 0;
        return fin;
    }
    else
    {
        red_val = clk_readl(clk->reg_ofs);
        divq = (red_val & 0x70) >> 4;
        divq = 1 << divq;
        fin = (parent_rate  * 2 * (((red_val & 0x1ff0000) >> 16) + 1) + (((red_val & 0x1f00) >> 8) + 1) * divq / 2)
                     / ((((red_val & 0x1f00) >> 8) + 1) * divq );
        return fin;
    }
    return -EPERM;
}


/*设置当前的pll时钟频率*/
static int ljm_clk_pll_set_rate(struct clk_hw *hw, unsigned long rate, unsigned long parent_rate)
{
    struct ljm_clk_pll *clk = to_ljm_clk_pll(hw);
    uint32_t *pll_para = NULL;

    BUG_ON(parent_rate < MHZ || parent_rate % MHZ);
    ljm_vgav2_pllcalc((unsigned int)parent_rate, 1, rate, pll_para);
    clk_writel(*pll_para, clk->reg_ofs);
    return 0;
}

/*Given a target rate as input, returns the closest rate actually supported by the clock*/
static long   ljm_clk_pll_round_rate(struct clk_hw *hw, unsigned long rate,
                unsigned long *parent_rate)
{
    uint32_t *pll_para = NULL;
    uint32_t red_val;
    unsigned long fin,divq;

    ljm_vgav2_pllcalc(*parent_rate, 1, rate, pll_para);
    red_val = *pll_para;
    divq = (red_val & 0x70) >> 4;
    divq = 1 << divq;
    fin = (*parent_rate / MHZ  * 2 * (((red_val & 0x1ff0000) >> 16) + 1) + (((red_val & 0x1f00) >> 8) + 1) * divq / 2)
                     / ((((red_val & 0x1f00) >> 8) + 1) * divq ) * MHZ;

    return (long)fin;
}

/*Given a target rate as input, returns the closest rate
 *		actually supported by the clock, and optionally the parent clock
 *		that should be used to provide the clock rate*/
static int     ljm_clk_pll_determine_rate(struct clk_hw *hw,
                  struct clk_rate_request *req)
{
    struct ljm_clk_pll *clk = to_ljm_clk_pll(hw);
    struct clk_hw *parent;
    unsigned long parent_rate;

    parent = clk_hw_get_parent(hw);
    if(!parent)
    return -EINVAL;

    parent_rate = clk_hw_get_rate(parent);
    req->best_parent_hw = parent;
    req->best_parent_rate = parent_rate;

    switch (clk->reg_ofs)
    {
        case GP201_CRU_CONTROL_REGISTER + GP201_CRU_DEVPLL:
            req->rate = 800 * MHZ;
            req->min_rate = 100 * MHZ;
            req->max_rate = 1200 * MHZ;
            break;
        case GP201_CRU_CONTROL_REGISTER + GP201_CRU_DEV2PLL:
            req->rate = 500 * MHZ;
            req->min_rate = 500 *MHZ;
            req->max_rate = 500 *MHZ;
            break;
        case GP201_CRU_CONTROL_REGISTER + GP201_CRU_GPUPLL:
            req->rate = 900 * MHZ;
            req->min_rate = 100 * MHZ;
            req->max_rate = 1200 * MHZ;
            break;
        case GP201_CRU_CONTROL_REGISTER + GP201_CRU_GPU2PLL:
            req->rate = 900 * MHZ;
            req->min_rate = 100 * MHZ;
            req->max_rate = 1200 * MHZ;
            break;
        case GP201_CRU_CONTROL_REGISTER + GP201_CRU_DDRPLL:
            req->rate = 1000 * MHZ;
            req->min_rate = 800 * MHZ;
            req->max_rate = 1200 * MHZ;
            break;
    }
    return 0;
}

static void ljm_clk_pll_init(struct clk_hw *hw)
{
    struct ljm_clk_pll *clk = to_ljm_clk_pll(hw);
    unsigned long en_val;
    en_val = clk_readl(clk->en_ofs);
    clk_writel(clk->en_ofs, en_val & (~(1<<clk->reset_bit)) &(~(1<<clk->bypass_bit)));
}

static const struct clk_ops ljm_pll_ops = {
    .recalc_rate = ljm_clk_pll_recalc_rate,
    .set_rate = ljm_clk_pll_set_rate,
    .round_rate = ljm_clk_pll_round_rate,
    .determine_rate = ljm_clk_pll_determine_rate,
    .init = ljm_clk_pll_init,
};


static unsigned long ljm_clk_divider_recalc_rate (struct clk_hw *hw, unsigned long parent_rate)
{
    struct ljm_clk_divider *clk = to_ljm_clk_divider(hw);
    unsigned long val;
    unsigned long rate;

    val = clk_readl(clk->reg_ofs);
    val = (val & (((1 << clk->width) -1) << clk->shift)) >> clk->shift;
    if(clk->width == 0 || val == 0)
    {
        rate = parent_rate;
    }
    else
    {
        rate = parent_rate / (val + 1);
    }


    return rate;
}

static int ljm_clk_divider_set_rate(struct clk_hw * hw, unsigned long rate, unsigned long parent_rate)
{

    struct ljm_clk_divider *clk = to_ljm_clk_divider(hw);
    uint32_t val;
    val = 10;
        //divider_get_val(rate, parent_rate, clk->table, clk->width, CLK_DIVIDER_ROUND_CLOSEST);
    val = val << clk->shift;
    clk_writel(val, GP201_CRU_CONTROL_REGISTER + GP201_CRU_CKDIV);
    return 0;
}

static long ljm_clk_divider_round_rate(struct clk_hw *hw, unsigned long rate,
                unsigned long *parent_rate)
{
    //struct ljm_clk_divider *clk = to_ljm_clk_divider(hw);

    return 100;
        //divider_round_rate(hw, rate, parent_rate, clk->table, clk->width, CLK_DIVIDER_ROUND_CLOSEST);
}


const struct clk_ops ljm_divider_ops = {
	.recalc_rate = ljm_clk_divider_recalc_rate,
	.round_rate = ljm_clk_divider_round_rate,
	.set_rate = ljm_clk_divider_set_rate,
};


static int  ljm_clk_mux_set_parent(struct clk_hw *hw, u8 index)
{
    struct ljm_clk_mux *clk = to_ljm_clk_mux(hw);
    uint32_t val;

    val = clk_readl(clk->reg_ofs);
    if (index)
        val = val | (1 << clk->mux_bit);
    else
        val = val & (~(1 << clk->mux_bit));
    clk_writel(val, clk->reg_ofs);
    return 0;
}
static u8 ljm_clk_mux_get_parent(struct clk_hw *hw)
{
    struct ljm_clk_mux *clk = to_ljm_clk_mux(hw);
    u8 index;
    uint32_t val;

    val = clk_readl(clk->reg_ofs);
    index = (val & (1 << clk->mux_bit)) << clk->mux_bit;

    return index;
}

const struct clk_ops ljm_mux_ops = {
    .set_parent = ljm_clk_mux_set_parent,
    .get_parent = ljm_clk_mux_get_parent,
    .determine_rate = __clk_mux_determine_rate,
};

static int ljm_clk_gate_enable(struct clk_hw *hw)
{
    struct ljm_clk_gate *clk = to_ljm_clk_gate(hw);
    unsigned long gate_val;

    gate_val = clk_readl(clk->reg_ofs);
    clk_writel(clk->reg_ofs, gate_val & (1 << clk->gate_bits));
    return 0;
}

static void ljm_clk_gate_disable(struct clk_hw *hw)
{
    struct ljm_clk_gate *clk = to_ljm_clk_gate(hw);
    unsigned long gate_val;

    gate_val = clk_readl(clk->reg_ofs);
    clk_writel(clk->reg_ofs, gate_val & (~(1 << clk->gate_bits)));
}

static int ljm_clk_gate_is_enabled(struct clk_hw *hw)
{
    struct ljm_clk_gate *clk = to_ljm_clk_gate(hw);
    unsigned long gate_val;

    gate_val = clk_readl(clk->reg_ofs);
    gate_val &= BIT(clk->gate_bits);
    return gate_val ? 1 : 0;
}

const struct clk_ops ljm_gate_ops = {
    .prepare = ljm_clk_gate_enable,
    .unprepare = ljm_clk_gate_disable,
    .is_prepared = ljm_clk_gate_is_enabled,
};

static struct ljm_clk_pll gp201_pll[ARRAY_SIZE(gp201_clk_plls)];
static struct ljm_clk_divider gp201_divider[ARRAY_SIZE(gp201_clk_divider)];
static struct ljm_clk_mux gp201_mux[ARRAY_SIZE(gp201_clk_mux)];
static struct ljm_clk_gate gp201_gate[ARRAY_SIZE(gp201_clk_gate)];
static struct clk * gp201_register_clk[CLK_NUMS];
static struct clk_onecell_data ljm_clk_data;


struct clk * ljm_devm_register_pll(struct device *dev, const struct ljm_clk_pll_data *data)
{
    struct clk_init_data init;
    static int i = 0;

    init.name = data->name;
    init.ops = &ljm_pll_ops;
    init.flags = data->flags;
    init.parent_names = (data->parent_name ? &data->parent_name : NULL);
    init.num_parents = (data->parent_name ? 1 : 0);

    gp201_pll[i].hw.init = &init;
    gp201_pll[i].bypass_bit = data->bypass_bit;
    gp201_pll[i].en_bit1 = data->en_bit1;
    gp201_pll[i].en_bit2 = data->en_bit2;
    gp201_pll[i].en_ofs = data->en_ofs;
    gp201_pll[i].reg_ofs = data->reg_ofs;
    gp201_pll[i].reset_bit = data->reset_bit;

    return devm_clk_register(dev, &gp201_pll[i++].hw);
}

static int ljm_clk_register_plls(struct device *dev)
{
    struct clk *clk;
    int i;

    for (i = 0; i < ARRAY_SIZE(gp201_clk_plls); i++)
    {
        const struct ljm_clk_pll_data *data = &gp201_clk_plls[i];
        clk = ljm_devm_register_pll(dev, data);
#if 0
        if (IS_ERR(clk))
        {
            ERR_PRINT("failed to register clock %s\n", data->name);
            return -i;
        }
#endif
        gp201_register_clk[data->id] = clk;
    }
    return i;
}

struct clk * ljm_devm_register_divider(struct device *dev, const struct ljm_clk_divider_data *data)
{
    struct clk_init_data init;
    static int i = 0;

    init.name = data->name;
    init.ops = &ljm_divider_ops;
    init.flags = data->flags;
    init.parent_names = (data->parent_name ? &data->parent_name : NULL);
    init.num_parents = (data->parent_name ? 1 : 0);

    gp201_divider[i].hw.init = &init;
    gp201_divider[i].reg_ofs = data->reg_ofs;
    gp201_divider[i].width = data->width;
    gp201_divider[i].shift = data->shift;

    return devm_clk_register(dev, &gp201_divider[i++].hw);
}

static int ljm_clk_register_divider(struct device *dev)
{
    struct clk *clk;
    int i;

    for (i = 0; i < ARRAY_SIZE(gp201_clk_divider); i++)
    {
        const struct ljm_clk_divider_data *data = &gp201_clk_divider[i];
        clk = ljm_devm_register_divider(dev, data);
#if 0
        if (IS_ERR(clk))
        {
            ERR_PRINT("failed to register clock %s\n", data->name);
            return -i;
        }
#endif
        gp201_register_clk[data->id] = clk;
    }
    return i;
}

struct clk * ljm_devm_register_mux(struct device *dev, const struct ljm_clk_mux_data *data)
{
    struct clk_init_data init;
    static int i = 0;

    init.name = data->name;
    init.ops = &ljm_mux_ops;
    init.flags = data->flags;
    init.parent_names = data->parent_names;
    init.num_parents = data->num_parents;

    gp201_mux[i].hw.init = &init;
    gp201_mux[i].reg_ofs = data->reg_ofs;
    gp201_mux[i].mux_bit = data->mux_bits;

    return devm_clk_register(dev, &gp201_mux[i++].hw);
}

static int ljm_clk_register_mux(struct device *dev)
{
    struct clk *clk;
    int i;

    for (i = 0; i < ARRAY_SIZE(gp201_clk_mux); i++)
    {
        const struct ljm_clk_mux_data *data = &gp201_clk_mux[i];
        clk = ljm_devm_register_mux(dev, data);
#if 0
        if (IS_ERR(clk))
        {
            ERR_PRINT("failed to register clock %s\n", data->name);
            return -i;
        }
#endif
        gp201_register_clk[data->id] = clk;
    }
    return i;
}

struct clk * ljm_devm_register_gate(struct device *dev, const struct ljm_clk_gate_data *data)
{
    struct clk_init_data init;
    static int i = 0;

    init.name = data->name;
    init.ops = &ljm_gate_ops;
    init.flags = data->flags;
    init.parent_names = (data->parent_name ? &data->parent_name : NULL);
    init.num_parents = (data->parent_name ? 1 : 0);

    gp201_gate[i].hw.init = &init;
    gp201_gate[i].reg_ofs = data->reg_ofs;
    gp201_gate[i].gate_bits = data->gate_bits;

    return devm_clk_register(dev, &gp201_gate[i++].hw);
}

static int ljm_clk_register_gate(struct device *dev)
{
    struct clk *clk;
    int i;

    for (i = 0; i < ARRAY_SIZE(gp201_clk_gate); i++)
    {
        const struct ljm_clk_gate_data *data = &gp201_clk_gate[i];
        clk = ljm_devm_register_gate(dev, data);
#if 0
        if (IS_ERR(clk))
        {
            ERR_PRINT("failed to register clock %s\n", data->name);
            return -i;
        }
#endif
        gp201_register_clk[data->id] = clk;
    }
    return i;
}

#define CLK_IS_ROOT     BIT(4)
int  ljm_clk_init(struct ljm_device *ljm_dev)
{
    int pll_num, divider_num, mux_num, gate_num;

    if (0 == ljm_dev->cur_card)
    {
        ljm_clk_vbase = g_mmio_base;
    }
    else
    {
        ljm_clk_vbase = g_mmio_bases[ljm_dev->cur_card];
    }

    gp201_register_clk[HOST_LINE] = clk_register_fixed_rate(ljm_dev->dev, "HOST", NULL, CLK_IS_ROOT, 27 * MHZ);
    BUG_ON(!gp201_register_clk[HOST_LINE]);

    pll_num = ljm_clk_register_plls(ljm_dev->dev);
    if (pll_num < 0)
    {
        ERR_PRINT("fail to register pll %d", pll_num);
        return -1;
    }

    divider_num = ljm_clk_register_divider(ljm_dev->dev);
    if (divider_num < 0)
    {
        ERR_PRINT("fail to register divider %d", divider_num);
        return -1;
    }

    mux_num = ljm_clk_register_mux(ljm_dev->dev);
    if (mux_num < 0)
    {
        ERR_PRINT("fail to register mux %d", mux_num);
        return -1;
    }

    gate_num = ljm_clk_register_gate(ljm_dev->dev);
    if (gate_num < 0)
    {
        ERR_PRINT("fail to register gate %d", gate_num);
        return -1;
    }

    if (gp201_register_clk != NULL)
    {
        ljm_clk_data.clks = gp201_register_clk;
        ljm_clk_data.clk_num = CLK_NUMS;
    }
    else
    {
        return -1;
    }

    return of_clk_add_provider(ljm_dev->dev->of_node, of_clk_src_onecell_get, &ljm_clk_data);
}


int ljm_clk_remove(struct ljm_device *ljm_dev)
{
    of_clk_del_provider(ljm_dev->dev->of_node);
    return 0;
}

#endif  /* ~USE_CLOCK_REGISTER */