/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_COMPAT_H__
#define __LJMICRO_COMPAT_H__

#include <linux/version.h>

#ifndef FPGA_HAPS
#define FPGA_HAPS   0
#endif

/*** compact for different kernels ********************************************/
#if LINUX_VERSION_CODE == KERNEL_VERSION(4,4,32) && defined CONFIG_MIPS
    #define DRM_GEM_OBJECT_LOOKUP(Drm, Filp, Handle)    drm_gem_object_lookup(Drm, Filp, Handle)

    //static inline bool drm_vma_node_has_offset(struct drm_vma_offset_node *node)
    #define DRM_MM_NODE_ALLOCATED(VMA_NODE)    drm_vma_node_has_offset(&VMA_NODE)

    #define DRM_VMA_OFFSET_EXACT_LOOKUP_LOCKED  drm_vma_offset_exact_lookup

    #define CMD_ARGS struct drm_mode_fb_cmd2

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,0) || (LINUX_VERSION_CODE == KERNEL_VERSION(3,10,0) && defined CONFIG_MIPS)
    /* 飞腾+银河麒麟(USE_FT_CPU): KERNEL_VERSION(4,4,6), 4.4.0-21.x86_64
     * ubuntu16: KERNEL_VERSION(4,4,6), 4.4.0-21.x86_64
     * Fedora 25: KERNEL_VERSION(4,8,6), 4.8.6-300.x86_64
     * 银河麒麟社区版: KERNEL_VERSION(4,11,12), 4.11.0-14.x86_64
     * 中标麒麟: KERNEL_VERSION(3,10,0)
     */
    //struct drm_gem_object *drm_gem_object_lookup(struct drm_device *dev, struct drm_file *filp, u32 handle)
    #define DRM_GEM_OBJECT_LOOKUP(Drm, Filp, Handle)    drm_gem_object_lookup(Filp, Handle)

    //static inline bool drm_mm_node_allocated(struct drm_mm_node *node)
    #define DRM_MM_NODE_ALLOCATED(VMA_NODE)    drm_mm_node_allocated(&VMA_NODE.vm_node)

    #define DRM_VMA_OFFSET_EXACT_LOOKUP_LOCKED  drm_vma_offset_exact_lookup_locked

    #define CMD_ARGS const struct drm_mode_fb_cmd2
#else
    /* Fedora 23: KERNEL_VERSION(4,2,3), 4.2.3-300.x86_64
     * 中标麒麟: KERNEL_VERSION(4,1,0), 4.1.0-1.x86
     * Centos 6.9: KERNEL_VERSION(2,6,32), 2.6.32-696.el6.x86_64
     */
    //struct drm_gem_object *drm_gem_object_lookup(struct drm_file *filp, u32 handle)
    #define DRM_GEM_OBJECT_LOOKUP(Drm, Filp, Handle)    drm_gem_object_lookup(Drm, Filp, Handle)

    //static inline bool drm_vma_node_has_offset(struct drm_vma_offset_node *node)
    #define DRM_MM_NODE_ALLOCATED(VMA_NODE)    drm_vma_node_has_offset(&VMA_NODE)

    #define DRM_VMA_OFFSET_EXACT_LOOKUP_LOCKED  drm_vma_offset_exact_lookup

    #define CMD_ARGS struct drm_mode_fb_cmd2
#endif

/* KERNEL_VERSION(4,0xb,0xc): 4.0.2-desktop-sp2_Community-20171127, hym on 20190805 */
/* KERNEL_VERSION(4,4,167): linux-4.15.0-45, modify by xwh at 20190528 */
/* KERNEL_VERSION(4,8,6): linux-4.8.6-300, modify by hzh at 20190911 */
/* KERNEL_VERSION(4,0x13,0x53): linux-4.19.0-arm64-desktop #1072, modified for adapting Deepin by zpw at 20191214 */
//#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0x13, 0x53)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 18)
    #define DRM_ENCODER_FIND(Drm, X, Id) drm_encoder_find(Drm, NULL, Id)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,0xb,0xc) || LINUX_VERSION_CODE <= KERNEL_VERSION(4,8,6)
    #define DRM_ENCODER_FIND(Drm, X, Id) drm_encoder_find(Drm, Id)
#else
    #define DRM_ENCODER_FIND(Drm, X, Id) drm_encoder_find(Drm, NULL, Id)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    #define FAULT_ARGS struct vm_fault *Vmf
#else
    #define FAULT_ARGS struct vm_area_struct *Vma, struct vm_fault *Vmf
#endif

/******************************************************************************/

/*
 * compatible
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
#define drm_gem_object_unreference_unlocked drm_gem_object_put
#define drm_gem_object_reference drm_gem_object_get
#elif LINUX_VERSION_CODE >= KERNEL_VERSION (5, 4, 0)
#define drm_gem_object_unreference_unlocked drm_gem_object_put_unlocked
#define drm_gem_object_reference drm_gem_object_get
#endif

#ifdef CONFIG_ARM64
/* fixup ft1500a and ft2000ahk */
#ifndef is_ft1500a
#include <asm/cputype.h>

#ifndef ARM_CPU_IMP_PHYTIUM  /* not defined by cputype.h of deepin, added by zpw */
#define ARM_CPU_IMP_PHYTIUM 0x70
#endif

#ifndef PHYTIUM_CPU_PART_1500A
#define PHYTIUM_CPU_PART_1500A      0X660
#endif
#define PHYTIUM_CPU_PART_2000AHK    0X661
#define PHYTIUM_CPU_PART_2000PLUS   0X662
#define PHYTIUM_CPU_PART_2000A_4    0X663

#ifndef MIDR_FT_1500A
#define MIDR_FT_1500A   MIDR_CPU_MODEL(ARM_CPU_IMP_PHYTIUM, PHYTIUM_CPU_PART_1500A)
#endif
#define MIDR_FT_2000AHK MIDR_CPU_MODEL(ARM_CPU_IMP_PHYTIUM, PHYTIUM_CPU_PART_2000AHK)
#define MIDR_FT_2000PLUS    MIDR_CPU_MODEL(ARM_CPU_IMP_PHYTIUM, PHYTIUM_CPU_PART_2000PLUS)
#define MIDR_FT_2000A_4 MIDR_CPU_MODEL(ARM_CPU_IMP_PHYTIUM, PHYTIUM_CPU_PART_2000A_4)

/********************* for all Phytium CPUs ******************/
static inline bool is_part(u32 cpuid)
{
    return ((read_cpuid_id() & MIDR_CPU_MODEL_MASK) == cpuid);
}

static inline bool is_implementer(u32 implementer)
{
    return ((read_cpuid_implementor() == implementer));
}

#define is_ft1500a()    is_part(MIDR_FT_1500A)
#define is_ft2000ahk()  is_part(MIDR_FT_2000AHK)
#define is_ft2000plus() is_part(MIDR_FT_2000PLUS)
#define is_ft2000a4()   is_part(MIDR_FT_2000A_4)
#define is_ft_all() is_implementer(ARM_CPU_IMP_PHYTIUM)

#else   /* is_ft1500a */

#ifndef is_ft2000a4
#ifndef MIDR_FT_2000A_4
#define MIDR_FT_2000A_4 MIDR_CPU_MODEL(ARM_CPU_IMP_PHYTIUM, PHYTIUM_CPU_PART_2000A_4)
#endif
#ifndef __PHYPIUM_PLATFORM__
static inline bool is_part(u32 cpuid)
{
    return ((read_cpuid_id() & MIDR_CPU_MODEL_MASK) == cpuid);
}
#endif
#define PHYTIUM_CPU_PART_2000A_4    0X663

#define is_ft2000a4()   is_part(MIDR_FT_2000A_4)
#endif

#endif  /* is_ft1500a */

#else   /* CONFIG_ARM64 */
#define is_ft1500a()    0
#define is_ft2000ahk()  0
#define is_ft2000plus() 0
#define is_ft2000a4()   0
#define is_ft_all()     0
#endif  /* CONFIG_ARM64 */

#endif
