/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/**************************************************************************
* 版权所有 (C)2022, 武汉凌久微电子有限公司。
*
* 文件名称： ljmicro_connector_code.c
* 文件标识：
* 内容摘要： 本文件定义显示接口音频层，调用DP、HDMI的函数实现，完成所有
*            接口的音频驱动，并注册hdmi_codec
* 其它说明：
* 当前版本： v1.0
* 作     者： zhoupw, 武汉凌久微电子有限公司
* 完成日期： 2022年4月27日
***************************************************************************/
#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)

#include <linux/string.h>
#include <sound/core.h>
#include <sound/jack.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/tlv.h>
#include <sound/pcm_drm_eld.h>
#include <sound/hdmi-codec.h>
#include <sound/pcm_iec958.h>
#include <drm/drm_crtc.h>

#include "ljmicro_device.h"
#include "ljmicro_hdmiv2.h"
#include "ljmicro_connector_codec.h"

static int ljm_connector_codec_hw_params(struct device *dev, void *data,
        struct hdmi_codec_daifmt *fmt, struct hdmi_codec_params *hparams)
{
#if ENABLE_HDMI_AUDIO || ENABLE_DP_AUDIO
    int i;
    int ret;
    struct ljm_connector_codec_data *audio_data = data;
#endif

#if ENABLE_HDMI_AUDIO
    for (i = 0; i < audio_data->num_hdmis; i++) {
        audio_data->hdmi_audio_data->hdmi = audio_data->hdmi_devs[i];
        ret = audio_data->ljm_hdmi_codec_ops->hw_params(dev, (void *)audio_data->hdmi_audio_data, fmt, hparams);
        if (ret) {
            dev_err(dev, "Set hdmi i2s hw params failed.\n");
            return ret;
        }
    }
#endif

#if ENABLE_DP_AUDIO
    for (i = 0; i < audio_data->num_dps; i++) {
        audio_data->dp_audio_data->dp = audio_data->dp_devs[i];
        ret = audio_data->ljm_dp_codec_ops->hw_params(dev, (void *)audio_data->dp_audio_data, fmt, hparams);
        if (ret) {
            dev_err(dev, "Set dp i2s hw params failed.\n");
            return ret;
        }
    }
#endif

    return 0;
}

static int ljm_connector_codec_audio_startup(struct device *dev, void *data)
{
#if ENABLE_HDMI_AUDIO || ENABLE_DP_AUDIO
    int i;
    struct ljm_connector_codec_data *audio_data = data;
#endif

#if ENABLE_HDMI_AUDIO
    for (i = 0; i < audio_data->num_hdmis; i++) {
        audio_data->hdmi_audio_data->hdmi = audio_data->hdmi_devs[i];
        audio_data->ljm_hdmi_codec_ops->audio_startup(dev, (void *)audio_data->hdmi_audio_data);
    }
#endif

#if ENABLE_DP_AUDIO
    for (i = 0; i < audio_data->num_dps; i++) {
        audio_data->dp_audio_data->dp = audio_data->dp_devs[i];
        audio_data->ljm_dp_codec_ops->audio_startup(dev, (void *)audio_data->dp_audio_data);
    }
#endif

    return 0;
}

static void ljm_connector_codec_audio_shutdown(struct device *dev, void *data)
{
#if ENABLE_HDMI_AUDIO || ENABLE_DP_AUDIO
    int i;
    struct ljm_connector_codec_data *audio_data = data;
#endif

#if ENABLE_HDMI_AUDIO
    for (i = 0; i < audio_data->num_hdmis; i++) {
        audio_data->hdmi_audio_data->hdmi = audio_data->hdmi_devs[i];
        audio_data->ljm_hdmi_codec_ops->audio_shutdown(dev, (void *)audio_data->hdmi_audio_data);
    }
#endif

#if ENABLE_DP_AUDIO
    for (i = 0; i < audio_data->num_dps; i++) {
        audio_data->dp_audio_data->dp = audio_data->dp_devs[i];
        audio_data->ljm_dp_codec_ops->audio_shutdown(dev, (void *)audio_data->dp_audio_data);
    }
#endif
}

static int ljm_connector_codec_get_eld(struct device *dev, void *data, uint8_t *buf, size_t len)
{
#if ENABLE_HDMI_AUDIO || ENABLE_DP_AUDIO
    int i;
    struct ljm_connector_codec_data *audio_data = data;
#endif

#if ENABLE_HDMI_AUDIO
    for (i = 0; i < audio_data->num_hdmis; i++) {
        audio_data->hdmi_audio_data->hdmi = audio_data->hdmi_devs[i];
        audio_data->hdmi_audio_data->eld = audio_data->hdmi_devs[i]->connector.eld;
        if (audio_data->ljm_hdmi_codec_ops->get_eld) {
            audio_data->ljm_hdmi_codec_ops->get_eld(dev, (void *)audio_data->hdmi_audio_data, buf, len);
        }
    }
#endif

#if ENABLE_DP_AUDIO
    for (i = 0; i < audio_data->num_dps; i++) {
        audio_data->dp_audio_data->dp = audio_data->dp_devs[i];
        audio_data->dp_audio_data->eld = devm_kzalloc(audio_data->dp_audio_data->dp->dev,
            min_t(size_t, MAX_ELD_BYTES, len), GFP_KERNEL);
        audio_data->dp_audio_data->eld = audio_data->dp_devs[i]->connector->eld;
        if (audio_data->ljm_dp_codec_ops->get_eld)
            audio_data->ljm_dp_codec_ops->get_eld(dev, (void *)audio_data->dp_audio_data, buf, len);
    }
#endif

    return 0;
}

static int ljm_connector_codec_get_dai_id(struct snd_soc_component *component, struct device_node *endpoint)
{
    struct of_endpoint of_ep;
    int ret;

    ret = of_graph_parse_endpoint(endpoint, &of_ep);
    if (ret < 0)
        return ret;

    if (of_ep.port == 2)
        return 0;

    return -EINVAL;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
static int ljm_connector_codec_hook_plugged_cb(struct device *dev, void *data,
        hdmi_codec_plugged_cb fn, struct device *codec_dev)
{
#if ENABLE_HDMI_AUDIO || ENABLE_DP_AUDIO
    int i;
    struct ljm_connector_codec_data *audio_data = data;
#endif

#if ENABLE_HDMI_AUDIO
    for (i = 0; i < audio_data->num_hdmis; i++) {
        audio_data->hdmi_audio_data->hdmi = audio_data->hdmi_devs[i];
        audio_data->ljm_hdmi_codec_ops->hook_plugged_cb(dev, (void *)audio_data->hdmi_audio_data, fn, codec_dev);
    }
#endif

#if ENABLE_DP_AUDIO
    for (i = 0; i < audio_data->num_dps; i++) {
        audio_data->dp_audio_data->dp = audio_data->dp_devs[i];
        audio_data->ljm_dp_codec_ops->hook_plugged_cb(dev, (void *)audio_data->dp_audio_data, fn, codec_dev);
    }
#endif

    return 0;
}
#endif

static struct hdmi_codec_ops ljm_connector_codec_ops = {
    .hw_params = ljm_connector_codec_hw_params,
    .audio_startup = ljm_connector_codec_audio_startup,
    .audio_shutdown = ljm_connector_codec_audio_shutdown,
    .get_eld = ljm_connector_codec_get_eld,
    .get_dai_id = ljm_connector_codec_get_dai_id,
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
    .hook_plugged_cb = ljm_connector_codec_hook_plugged_cb,
#endif
};

/* General codec layer */
#define CONNECTOR_CODEC_CHMAP_IDX_UNKNOWN -1

struct connector_codec_channel_map_table {
    unsigned char map; /* ALSA API channel map position */
    unsigned long spk_mask; /* speaker position bit mask */
};

/*
 * CEA speaker placement for HDMI 1.4:
 *
 *  FL  FLC   FC   FRC   FR   FRW
 *
 *                                  LFE
 *
 *  RL  RLC   RC   RRC   RR
 *
 *  Speaker placement has to be extended to support HDMI 2.0
 */
enum connector_codec_cea_spk_placement {
    FL  = BIT(0), /* Front Left           */
    FC  = BIT(1), /* Front Center         */
    FR  = BIT(2), /* Front Right          */
    FLC = BIT(3), /* Front Left Center    */
    FRC = BIT(4), /* Front Right Center   */
    RL  = BIT(5), /* Rear Left            */
    RC  = BIT(6), /* Rear Center          */
    RR  = BIT(7), /* Rear Right           */
    RLC = BIT(8), /* Rear Left Center     */
    RRC = BIT(9), /* Rear Right Center    */
    LFE = BIT(10), /* Low Frequency Effect */
};

/*
 * cea Speaker allocation structure
 */
struct connector_codec_cea_spk_alloc {
    const int ca_id;
    unsigned int n_ch;
    unsigned long mask;
};

/* Channel maps  stereo HDMI */
static const struct snd_pcm_chmap_elem connector_codec_stereo_chmaps[] = {
    {
        .channels = 2,
        .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR }
    },
    { }
};

/* Channel maps for multi-channel playbacks, up to 8 n_ch */
static const struct snd_pcm_chmap_elem connector_codec_8ch_chmaps[] = {
    { .channels = 2, /* CA_ID 0x00 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR } },
    { .channels = 4, /* CA_ID 0x01 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA } },
    { .channels = 4, /* CA_ID 0x02 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC } },
    { .channels = 4, /* CA_ID 0x03 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC } },
    { .channels = 6, /* CA_ID 0x04 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 6, /* CA_ID 0x05 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 6, /* CA_ID 0x06 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 6, /* CA_ID 0x07 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 6, /* CA_ID 0x08 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR } },
    { .channels = 6, /* CA_ID 0x09 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR } },
    { .channels = 6, /* CA_ID 0x0A */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR } },
    { .channels = 6, /* CA_ID 0x0B */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR } },
    { .channels = 8, /* CA_ID 0x0C */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 8, /* CA_ID 0x0D */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 8, /* CA_ID 0x0E */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 8, /* CA_ID 0x0F */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RC, SNDRV_CHMAP_NA } },
    { .channels = 8, /* CA_ID 0x10 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RLC, SNDRV_CHMAP_RRC } },
    { .channels = 8, /* CA_ID 0x11 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RLC, SNDRV_CHMAP_RRC } },
    { .channels = 8, /* CA_ID 0x12 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RLC, SNDRV_CHMAP_RRC } },
    { .channels = 8, /* CA_ID 0x13 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_RL, SNDRV_CHMAP_RR,
           SNDRV_CHMAP_RLC, SNDRV_CHMAP_RRC } },
    { .channels = 8, /* CA_ID 0x14 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x15 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x16 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x17 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x18 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x19 */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x1A */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x1B */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x1C */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x1D */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_NA, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x1E */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { .channels = 8, /* CA_ID 0x1F */
      .map = { SNDRV_CHMAP_FL, SNDRV_CHMAP_FR, SNDRV_CHMAP_LFE,
           SNDRV_CHMAP_FC, SNDRV_CHMAP_NA, SNDRV_CHMAP_NA,
           SNDRV_CHMAP_FLC, SNDRV_CHMAP_FRC } },
    { }
};

/*
 * hdmi_codec_channel_alloc: speaker configuration available for CEA
 *
 * This is an ordered list that must match with hdmi_codec_8ch_chmaps struct
 * The preceding ones have better chances to be selected by
 * hdmi_codec_get_ch_alloc_table_idx().
 */
static const struct connector_codec_cea_spk_alloc connector_codec_channel_alloc[] = {
    { .ca_id = 0x00, .n_ch = 2,
      .mask = FL | FR},
    /* 2.1 */
    { .ca_id = 0x01, .n_ch = 4,
      .mask = FL | FR | LFE},
    /* Dolby Surround */
    { .ca_id = 0x02, .n_ch = 4,
      .mask = FL | FR | FC },
    /* surround51 */
    { .ca_id = 0x0b, .n_ch = 6,
      .mask = FL | FR | LFE | FC | RL | RR},
    /* surround40 */
    { .ca_id = 0x08, .n_ch = 6,
      .mask = FL | FR | RL | RR },
    /* surround41 */
    { .ca_id = 0x09, .n_ch = 6,
      .mask = FL | FR | LFE | RL | RR },
    /* surround50 */
    { .ca_id = 0x0a, .n_ch = 6,
      .mask = FL | FR | FC | RL | RR },
    /* 6.1 */
    { .ca_id = 0x0f, .n_ch = 8,
      .mask = FL | FR | LFE | FC | RL | RR | RC },
    /* surround71 */
    { .ca_id = 0x13, .n_ch = 8,
      .mask = FL | FR | LFE | FC | RL | RR | RLC | RRC },
    /* others */
    { .ca_id = 0x03, .n_ch = 8,
      .mask = FL | FR | LFE | FC },
    { .ca_id = 0x04, .n_ch = 8,
      .mask = FL | FR | RC},
    { .ca_id = 0x05, .n_ch = 8,
      .mask = FL | FR | LFE | RC },
    { .ca_id = 0x06, .n_ch = 8,
      .mask = FL | FR | FC | RC },
    { .ca_id = 0x07, .n_ch = 8,
      .mask = FL | FR | LFE | FC | RC },
    { .ca_id = 0x0c, .n_ch = 8,
      .mask = FL | FR | RC | RL | RR },
    { .ca_id = 0x0d, .n_ch = 8,
      .mask = FL | FR | LFE | RL | RR | RC },
    { .ca_id = 0x0e, .n_ch = 8,
      .mask = FL | FR | FC | RL | RR | RC },
    { .ca_id = 0x10, .n_ch = 8,
      .mask = FL | FR | RL | RR | RLC | RRC },
    { .ca_id = 0x11, .n_ch = 8,
      .mask = FL | FR | LFE | RL | RR | RLC | RRC },
    { .ca_id = 0x12, .n_ch = 8,
      .mask = FL | FR | FC | RL | RR | RLC | RRC },
    { .ca_id = 0x14, .n_ch = 8,
      .mask = FL | FR | FLC | FRC },
    { .ca_id = 0x15, .n_ch = 8,
      .mask = FL | FR | LFE | FLC | FRC },
    { .ca_id = 0x16, .n_ch = 8,
      .mask = FL | FR | FC | FLC | FRC },
    { .ca_id = 0x17, .n_ch = 8,
      .mask = FL | FR | LFE | FC | FLC | FRC },
    { .ca_id = 0x18, .n_ch = 8,
      .mask = FL | FR | RC | FLC | FRC },
    { .ca_id = 0x19, .n_ch = 8,
      .mask = FL | FR | LFE | RC | FLC | FRC },
    { .ca_id = 0x1a, .n_ch = 8,
      .mask = FL | FR | RC | FC | FLC | FRC },
    { .ca_id = 0x1b, .n_ch = 8,
      .mask = FL | FR | LFE | RC | FC | FLC | FRC },
    { .ca_id = 0x1c, .n_ch = 8,
      .mask = FL | FR | RL | RR | FLC | FRC },
    { .ca_id = 0x1d, .n_ch = 8,
      .mask = FL | FR | LFE | RL | RR | FLC | FRC },
    { .ca_id = 0x1e, .n_ch = 8,
      .mask = FL | FR | FC | RL | RR | FLC | FRC },
    { .ca_id = 0x1f, .n_ch = 8,
      .mask = FL | FR | LFE | FC | RL | RR | FLC | FRC },
};

struct connector_codec_priv {
    struct hdmi_codec_pdata hcd;
    uint8_t eld[MAX_ELD_BYTES];
    struct snd_pcm_chmap *chmap_info;
    unsigned int chmap_idx;
    unsigned long busy;
    struct snd_soc_jack *jack;
    unsigned int jack_status;
};

struct connector_codec_priv *g_hcp;

static const struct snd_soc_dapm_widget connector_widgets[] = {
    SND_SOC_DAPM_OUTPUT("TX"),
};

enum {
    DAI_ID_I2S = 0,
};

static int connector_eld_ctl_info(struct snd_kcontrol *kcontrol,
        struct snd_ctl_elem_info *uinfo)
{
    uinfo->type = SNDRV_CTL_ELEM_TYPE_BYTES;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    uinfo->count = sizeof_field(struct connector_codec_priv, eld);
#else
    uinfo->count = FIELD_SIZEOF(struct connector_codec_priv, eld);
#endif
    return 0;
}

static int connector_eld_ctl_get(struct snd_kcontrol *kcontrol,
        struct snd_ctl_elem_value *ucontrol)
{
    struct connector_codec_priv *hcp = g_hcp;

    memcpy(ucontrol->value.bytes.data, hcp->eld, sizeof(hcp->eld));

    return 0;
}

static unsigned long connector_codec_spk_mask_from_alloc(int spk_alloc)
{
    int i;
    static const unsigned long connector_codec_eld_spk_alloc_bits[] = {
        [0] = FL | FR, [1] = LFE, [2] = FC, [3] = RL | RR,
        [4] = RC, [5] = FLC | FRC, [6] = RLC | RRC,
    };
    unsigned long spk_mask = 0;

    for (i = 0; i < ARRAY_SIZE(connector_codec_eld_spk_alloc_bits); i++) {
        if (spk_alloc & (1 << i))
            spk_mask |= connector_codec_eld_spk_alloc_bits[i];
    }

    return spk_mask;
}

static void connector_codec_eld_chmap(struct connector_codec_priv *hcp)
{
    u8 spk_alloc;
    unsigned long spk_mask;

    spk_alloc = drm_eld_get_spk_alloc(hcp->eld);
    spk_mask = connector_codec_spk_mask_from_alloc(spk_alloc);

    /* Detect if only stereo supported, else return 8 channels mappings */
    if ((spk_mask & ~(FL | FR)) && hcp->chmap_info->max_channels > 2)
        hcp->chmap_info->chmap = connector_codec_8ch_chmaps;
    else
        hcp->chmap_info->chmap = connector_codec_stereo_chmaps;
}

static int connector_codec_get_ch_alloc_table_idx(struct connector_codec_priv *hcp,
        unsigned char channels)
{
    int i;
    u8 spk_alloc;
    unsigned long spk_mask;
    const struct connector_codec_cea_spk_alloc *cap = connector_codec_channel_alloc;

    spk_alloc = drm_eld_get_spk_alloc(hcp->eld);
    spk_mask = connector_codec_spk_mask_from_alloc(spk_alloc);

    for (i = 0; i < ARRAY_SIZE(connector_codec_channel_alloc); i++, cap++) {
        /* If spk_alloc == 0, HDMI is unplugged return stereo config*/
        if (!spk_alloc && cap->ca_id == 0)
            return i;
        if (cap->n_ch != channels)
            continue;
        if (!(cap->mask == (spk_mask & cap->mask)))
            continue;
        return i;
    }

    return -EINVAL;
}
static int connector_codec_chmap_ctl_get(struct snd_kcontrol *kcontrol,
        struct snd_ctl_elem_value *ucontrol)
{
    unsigned const char *map;
    unsigned int i;
    struct snd_pcm_chmap *info = snd_kcontrol_chip(kcontrol);
    struct connector_codec_priv *hcp = g_hcp;

    map = info->chmap[hcp->chmap_idx].map;

    for (i = 0; i < info->max_channels; i++) {
        if (hcp->chmap_idx == CONNECTOR_CODEC_CHMAP_IDX_UNKNOWN)
            ucontrol->value.integer.value[i] = 0;
        else
            ucontrol->value.integer.value[i] = map[i];
    }

    return 0;
}

static int connector_codec_startup(struct snd_pcm_substream *substream,
        struct snd_soc_dai *dai)
{
    struct connector_codec_priv *hcp = g_hcp;
    int ret = 0;

    ret = test_and_set_bit(0, &hcp->busy);
    if (ret) {
        dev_err(dai->dev, "Only one simultaneous stream supported!\n");
        return -EINVAL;
    }

    if (hcp->hcd.ops->audio_startup) {
        ret = hcp->hcd.ops->audio_startup(dai->dev, hcp->hcd.data);
        if (ret)
            goto err;
    }

    if (hcp->hcd.ops->get_eld) {
        ret = hcp->hcd.ops->get_eld(dai->dev, hcp->hcd.data, hcp->eld, sizeof(hcp->eld));

#if IS_ENABLED(CONFIG_SND_PCM_ELD)
        if (!ret) {
            ret = snd_pcm_hw_constraint_eld(substream->runtime, hcp->eld);
            if (ret)
                goto err;
        }
#endif
        /* Select chmap supported */
        connector_codec_eld_chmap(hcp);
    }
    return 0;

err:
    /* Release the exclusive lock on error */
    clear_bit(0, &hcp->busy);
    return ret;
}

static void connector_codec_shutdown(struct snd_pcm_substream *substream,
        struct snd_soc_dai *dai)
{
    struct connector_codec_priv *hcp = g_hcp;

    hcp->chmap_idx = CONNECTOR_CODEC_CHMAP_IDX_UNKNOWN;
    hcp->hcd.ops->audio_shutdown(dai->dev, hcp->hcd.data);

    clear_bit(0, &hcp->busy);
}

static int connector_codec_hw_params(struct snd_pcm_substream *substream,
        struct snd_pcm_hw_params *params,
        struct snd_soc_dai *dai)
{
    struct connector_codec_priv *hcp = g_hcp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    struct hdmi_codec_daifmt *cf = snd_soc_dai_dma_data_get_playback(dai);
#else
    struct hdmi_codec_daifmt *cf = dai->playback_dma_data;
#endif
    struct hdmi_codec_params hp = {
        .iec = {
            .status = { 0 },
            .subcode = { 0 },
            .pad = 0,
            .dig_subframe = { 0 },
        }
    };
    int ret, idx;

    ret = snd_pcm_create_iec958_consumer_hw_params(params, hp.iec.status,
              sizeof(hp.iec.status));
    if (ret < 0) {
        dev_err(dai->dev, "Creating IEC958 channel status failed %d\n", ret);
        return ret;
    }

    hdmi_audio_infoframe_init(&hp.cea);
    hp.cea.channels = params_channels(params);
    hp.cea.coding_type = HDMI_AUDIO_CODING_TYPE_STREAM;
    hp.cea.sample_size = HDMI_AUDIO_SAMPLE_SIZE_STREAM;
    hp.cea.sample_frequency = HDMI_AUDIO_SAMPLE_FREQUENCY_STREAM;

    /* Select a channel allocation that matches with ELD and pcm channels */
    idx = connector_codec_get_ch_alloc_table_idx(hcp, hp.cea.channels);
    if (idx < 0) {
        dev_err(dai->dev, "Not able to map channels to speakers (%d)\n", idx);
        hcp->chmap_idx = CONNECTOR_CODEC_CHMAP_IDX_UNKNOWN;
        return idx;
    }
    hp.cea.channel_allocation = connector_codec_channel_alloc[idx].ca_id;
    hcp->chmap_idx = connector_codec_channel_alloc[idx].ca_id;

    hp.sample_width = params_width(params);
    hp.sample_rate = params_rate(params);
    hp.channels = params_channels(params);

    return hcp->hcd.ops->hw_params(dai->dev, hcp->hcd.data, cf, &hp);
}

static int connector_codec_i2s_set_fmt(struct snd_soc_dai *dai,
        unsigned int fmt)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    struct hdmi_codec_daifmt *cf = snd_soc_dai_dma_data_get_playback(dai);
#else
    struct hdmi_codec_daifmt *cf = dai->playback_dma_data;
#endif

    /* Reset daifmt */
    memset(cf, 0, sizeof(*cf));

    switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
        case SND_SOC_DAIFMT_CBM_CFM:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
            cf->bit_clk_provider = 1;
            cf->frame_clk_provider = 1;
#else
            cf->bit_clk_master = 1;
            cf->frame_clk_master = 1;
#endif
            break;
        case SND_SOC_DAIFMT_CBS_CFM:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
            cf->frame_clk_provider = 1;
#else
            cf->frame_clk_master = 1;
#endif
            break;
        case SND_SOC_DAIFMT_CBM_CFS:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
            cf->bit_clk_provider = 1;
#else
            cf->bit_clk_master = 1;
#endif
            break;
        case SND_SOC_DAIFMT_CBS_CFS:
            break;
        default:
            return -EINVAL;
    }

    switch (fmt & SND_SOC_DAIFMT_INV_MASK) {
        case SND_SOC_DAIFMT_NB_NF:
            break;
        case SND_SOC_DAIFMT_NB_IF:
            cf->frame_clk_inv = 1;
            break;
        case SND_SOC_DAIFMT_IB_NF:
            cf->bit_clk_inv = 1;
            break;
        case SND_SOC_DAIFMT_IB_IF:
            cf->frame_clk_inv = 1;
            cf->bit_clk_inv = 1;
            break;
    }

    switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
        case SND_SOC_DAIFMT_I2S:
            cf->fmt = HDMI_I2S;
            break;
        case SND_SOC_DAIFMT_DSP_A:
            cf->fmt = HDMI_DSP_A;
            break;
        case SND_SOC_DAIFMT_DSP_B:
            cf->fmt = HDMI_DSP_B;
            break;
        case SND_SOC_DAIFMT_RIGHT_J:
            cf->fmt = HDMI_RIGHT_J;
            break;
        case SND_SOC_DAIFMT_LEFT_J:
            cf->fmt = HDMI_LEFT_J;
            break;
        case SND_SOC_DAIFMT_AC97:
            cf->fmt = HDMI_AC97;
            break;
        default:
            dev_err(dai->dev, "Invalid DAI interface format\n");
            return -EINVAL;
    }

    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int connector_codec_mute_stream(struct snd_soc_dai *dai, int mute, int direction)
{
    struct connector_codec_priv *hcp = g_hcp;

    if (hcp->hcd.ops->mute_stream && (direction == SNDRV_PCM_STREAM_PLAYBACK || !hcp->hcd.ops->no_capture_mute))
        return hcp->hcd.ops->mute_stream(dai->dev->parent, hcp->hcd.data, mute, direction);

    return -ENOTSUPP;
}
#else
static int connector_codec_digital_mute(struct snd_soc_dai *dai, int mute)
{
    struct connector_codec_priv *hcp = g_hcp;

    if (hcp->hcd.ops->digital_mute)
        return hcp->hcd.ops->digital_mute(dai->dev, hcp->hcd.data, mute);

    return 0;
}
#endif

static int connector_dai_probe(struct snd_soc_dai *dai)
{
    struct snd_soc_dapm_context *dapm;
    struct hdmi_codec_daifmt *daifmt;
    struct snd_soc_dapm_route route = {
        .sink = "TX",
        .source = dai->driver->playback.stream_name,
    };
    int ret;

    dapm = snd_soc_component_get_dapm(dai->component);
    ret = snd_soc_dapm_add_routes(dapm, &route, 1);
    if (ret)
        return ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    daifmt = devm_kzalloc(dai->dev, sizeof(*daifmt), GFP_KERNEL);
#else
    daifmt = kzalloc(sizeof(*daifmt), GFP_KERNEL);
#endif
    if (!daifmt)
        return -ENOMEM;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    snd_soc_dai_dma_data_set_playback(dai, daifmt);
#else
    dai->playback_dma_data = daifmt;
#endif
    return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
static int connector_codec_dai_remove(struct snd_soc_dai *dai)
{
    kfree(dai->playback_dma_data);
    return 0;
}
#endif

static int connector_codec_pcm_new(struct snd_soc_pcm_runtime *rtd,
        struct snd_soc_dai *dai)
{
    struct snd_soc_dai_driver *drv = dai->driver;
    struct connector_codec_priv *hcp = g_hcp;
    struct snd_kcontrol *kctl;
    struct snd_kcontrol_new connector_eld_ctl = {
        .access = SNDRV_CTL_ELEM_ACCESS_READ | SNDRV_CTL_ELEM_ACCESS_VOLATILE,
        .iface = SNDRV_CTL_ELEM_IFACE_PCM,
        .name = "ELD",
        .info = connector_eld_ctl_info,
        .get = connector_eld_ctl_get,
        .device = rtd->pcm->device,
    };
    int ret;

    ret =  snd_pcm_add_chmap_ctls(rtd->pcm, SNDRV_PCM_STREAM_PLAYBACK,
               NULL, drv->playback.channels_max, 0,
               &hcp->chmap_info);
    if (ret < 0)
        return ret;

    /* override handlers */
    hcp->chmap_info->private_data = hcp;
    hcp->chmap_info->kctl->get = connector_codec_chmap_ctl_get;

    /* default chmap supported is stereo */
    hcp->chmap_info->chmap = connector_codec_stereo_chmaps;
    hcp->chmap_idx = CONNECTOR_CODEC_CHMAP_IDX_UNKNOWN;

    /* add ELD ctl with the device number corresponding to the PCM stream */
    kctl = snd_ctl_new1(&connector_eld_ctl, dai->component);
    if (!kctl)
        return -ENOMEM;

    return snd_ctl_add(rtd->card->snd_card, kctl);
}

static const struct snd_soc_dai_ops connector_codec_i2s_dai_ops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    .probe = connector_dai_probe,
#endif
    .startup = connector_codec_startup,
    .shutdown = connector_codec_shutdown,
    .hw_params = connector_codec_hw_params,
    .set_fmt = connector_codec_i2s_set_fmt,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    .mute_stream = connector_codec_mute_stream,
#else
    .digital_mute = connector_codec_digital_mute,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    .pcm_new = connector_codec_pcm_new,
#endif
};

#define CONNECTOR_RATES (SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 |\
     SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_88200 |\
     SNDRV_PCM_RATE_96000 | SNDRV_PCM_RATE_176400 |\
     SNDRV_PCM_RATE_192000)

#define I2S_FORMATS (SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S16_BE |\
     SNDRV_PCM_FMTBIT_S20_3LE | SNDRV_PCM_FMTBIT_S20_3BE |\
     SNDRV_PCM_FMTBIT_S24_3LE | SNDRV_PCM_FMTBIT_S24_3BE |\
     SNDRV_PCM_FMTBIT_S24_LE | SNDRV_PCM_FMTBIT_S24_BE |\
     SNDRV_PCM_FMTBIT_S32_LE | SNDRV_PCM_FMTBIT_S32_BE)

static const struct snd_soc_dai_driver connector_i2s_daidrv = {
    .name = "ljm-i2s-hifi",
    .id = DAI_ID_I2S,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    .probe = connector_dai_probe,
    .remove = connector_codec_dai_remove,
#endif
    .playback = {
        .stream_name = "I2S Playback",
        .channels_min = 2,
        .channels_max = 8,
        .rates = CONNECTOR_RATES,
        .formats = I2S_FORMATS,
        .sig_bits = 24,
    },
    .ops = &connector_codec_i2s_dai_ops,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    .pcm_new = connector_codec_pcm_new,
#endif
};

static int connector_of_xlate_dai_id(struct snd_soc_component *component,
        struct device_node *endpoint)
{
    struct connector_codec_priv *hcp = g_hcp;
    int ret = -ENOTSUPP; /* see snd_soc_get_dai_id() */

    if (hcp->hcd.ops->get_dai_id)
        ret = hcp->hcd.ops->get_dai_id(component, endpoint);

    return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
static void hdmi_codec_jack_report(struct connector_codec_priv *hcp,
        unsigned int jack_status)
{
    if (jack_status != hcp->jack_status) {
        if (hcp->jack)
            snd_soc_jack_report(hcp->jack, jack_status, SND_JACK_LINEOUT);
        hcp->jack_status = jack_status;
    }
}

static void plugged_cb(struct device *dev, bool plugged)
{
    struct connector_codec_priv *hcp = dev_get_drvdata(dev);

    if (plugged) {
        if (hcp->hcd.ops->get_eld) {
        hcp->hcd.ops->get_eld(dev->parent, hcp->hcd.data,
                              hcp->eld, sizeof(hcp->eld));
        }
        hdmi_codec_jack_report(hcp, SND_JACK_LINEOUT);
    }
    else {
        hdmi_codec_jack_report(hcp, 0);
        memset(hcp->eld, 0, sizeof(hcp->eld));
    }
}

static int hdmi_probe(struct snd_soc_component *component)
{
    struct connector_codec_priv *hcp = g_hcp;
    int ret = 0;

    if (hcp->hcd.ops->hook_plugged_cb) {
        ret = hcp->hcd.ops->hook_plugged_cb(component->dev->parent,
                            hcp->hcd.data,
                            plugged_cb,
                            component->dev);
    }

    return ret;
}

static void hdmi_remove(struct snd_soc_component *component)
{
    struct connector_codec_priv *hcp = g_hcp;

    if (hcp->hcd.ops->hook_plugged_cb)
        hcp->hcd.ops->hook_plugged_cb(component->dev->parent,
                      hcp->hcd.data, NULL, NULL);
}
#endif

static const struct snd_soc_component_driver component_driver = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
     .probe = hdmi_probe,
     .remove = hdmi_remove,
#endif
    .dapm_widgets = connector_widgets,
    .num_dapm_widgets = ARRAY_SIZE(connector_widgets),
    .of_xlate_dai_id = connector_of_xlate_dai_id,
    .idle_bias_on = 1,
    .use_pmdown_time = 1,
    .endianness = 1,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    .non_legacy_dai_naming = 1,
#endif
};

int ljm_codec_register(struct ljm_device *ljm_dev)
{
    int dai_count = 1;
    int ret;
    struct hdmi_codec_pdata *hcd;
    struct snd_soc_dai_driver *daidrv;
    struct device *dev = ljm_dev->dev;
    struct connector_codec_priv *hcp;
    struct ljm_connector_codec_data *audio_data;

    audio_data = devm_kzalloc(dev, sizeof(struct ljm_connector_codec_data), GFP_KERNEL);
    audio_data->hdmi_audio_data = devm_kzalloc(dev, sizeof(struct ljm_hdmi_codec_data), GFP_KERNEL);
    audio_data->dp_audio_data = devm_kzalloc(dev, sizeof(struct ljm_dp_codec_data), GFP_KERNEL);
    audio_data->hdmi_devs = devm_kzalloc(dev, sizeof(struct hdmi_dev *) * LJM_MAX_HDMIS, GFP_KERNEL);
    audio_data->dp_devs = devm_kzalloc(dev, sizeof(struct dp_dev *) * LJM_MAX_DPS, GFP_KERNEL);

    hcd = devm_kzalloc(dev, sizeof(*hcd), GFP_KERNEL);
    hcd->ops = &ljm_connector_codec_ops;
    hcd->i2s = 1;
    hcd->max_i2s_channels = 8;
    hcd->data = audio_data;

#if ENABLE_HDMI_AUDIO
    ljm_hdmi_codec_init(audio_data, ljm_dev);
#endif

#if (ENABLE_DP_AUDIO && USE_DP_CONNECTOR)
    ljm_dp_codec_init(audio_data, ljm_dev);
#endif

    daidrv = devm_kcalloc(dev, dai_count, sizeof(*daidrv), GFP_KERNEL);

    hcp = devm_kzalloc(dev, sizeof(*hcp), GFP_KERNEL);
    hcp->hcd = *hcd;

    *daidrv = connector_i2s_daidrv;
    daidrv->playback.channels_max = hcd->max_i2s_channels;

    g_hcp = hcp;

    ret = devm_snd_soc_register_component(dev, &component_driver, daidrv, dai_count);
    if (ret) {
        dev_err(dev, "%s: snd_soc_register_component() failed (%d)\n", __func__, ret);
        return ret;
    }
    return 0;
}

#endif
