/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <drm/drm_gem_framebuffer_helper.h>

#include "ljmicro_device.h"
#include "ljmicro_mode.h"
#include "ljmicro_object.h"
#include "ljmicro_compat.h"
#include "ljmicro_edid.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
#include <drm/drm_modeset_helper_vtables.h>
#include <drm/drm_modeset_helper.h>
#endif

#ifdef DELAY_4K_DISPLAY
#include "ljmicro_vgav2.h"
#endif

#define _LJ_OBJ_ZONE    ljvZONE_KMS

#if USE_KMS_VBLANK
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
u32 ljm_get_vblank_counter_kms(struct drm_crtc *crtc);
int ljm_enable_vblank_kms(struct drm_crtc *crtc);
void ljm_disable_vblank_kms(struct drm_crtc *crtc);
//bool ljm_get_vblank_timestamp_kms(struct drm_crtc *crtc, int *max_error, ktime_t *vblank_time, bool in_vblank_irq);
#endif
#endif

void ljmmode_from_dispmode(struct ljm_crtc_mode *ljm_mode, struct drm_display_mode *disp_mode)
{
    HEADER_ARG("ljm_mode=0x%p disp_mode=0x%p",ljm_mode, disp_mode);

    ljm_mode->clock = disp_mode->clock;
    ljm_mode->hdisplay = disp_mode->hdisplay;
    ljm_mode->hsync_start = disp_mode->hsync_start;
    ljm_mode->hsync_end = disp_mode->hsync_end;
    ljm_mode->htotal = disp_mode->htotal;
    ljm_mode->vdisplay = disp_mode->vdisplay;
    ljm_mode->vsync_start = disp_mode->vsync_start;
    ljm_mode->vsync_end = disp_mode->vsync_end;
    ljm_mode->vtotal = disp_mode->vtotal;
    ljm_mode->type = disp_mode->type;
    ljm_mode->flags = disp_mode->flags;

    FOOTER();
}

void ljmmode_to_dispmode(struct ljm_crtc_mode *ljm_mode, struct drm_display_mode *disp_mode)
{
    HEADER_ARG("ljm_mode=0x%p disp_mode=0x%p",ljm_mode, disp_mode);

    disp_mode->clock = ljm_mode->clock;
    disp_mode->hdisplay = ljm_mode->hdisplay;
    disp_mode->hsync_start = ljm_mode->hsync_start;
    disp_mode->hsync_end = ljm_mode->hsync_end;
    disp_mode->htotal = ljm_mode->htotal;
    disp_mode->vdisplay = ljm_mode->vdisplay;
    disp_mode->vsync_start = ljm_mode->vsync_start;
    disp_mode->vsync_end = ljm_mode->vsync_end;
    disp_mode->vtotal = ljm_mode->vtotal;
    disp_mode->type = ljm_mode->type;
    disp_mode->flags = ljm_mode->flags;

    FOOTER();
}

static struct ljm_crtc_mode vesa_mode[] = {
    {148500, 1920, 2008, 2052, 2200, 1080, 1084, 1089, 1125, 0x48, 0x5},    /* 1920*1080@60 */
    {119000, 1680, 1728, 1760, 1840, 1050, 1053, 1059, 1080, 0x40, 0x9},    /* 1680 *1050@60 */
    {88750 , 1440, 1488, 1520, 1600, 900 , 903 , 909 , 926 , 0x40, 0x9},    /* 1440*900@60 */
    {65000 , 1024, 1048, 1184, 1344, 768 , 771 , 777 , 806 , 0x40, 0xa},    /* 1024*768@60 */
};

static unsigned int vesa_mode_count = sizeof(vesa_mode) / sizeof(vesa_mode[0]);

/* Screen A CRTC HV, CR78 config */
/* name[15], w, h, pixelclk, crtc_hv[16], crtc_cr78 */
#define RES640x480    "640x480",   640,  480,  25175,  {0x03,0x20,0x02,0x80,0x08,0x08,0x60,0x28,0x02,0x0D,0x01,0xE0,0x08,0x08,0x02,0x19}, 0x00
#define RES800x600    "800x600",   800,  600,  40000,  {0x04,0x20,0x03,0x20,0x00,0x00,0x80,0x58,0x02,0x74,0x02,0x58,0x00,0x00,0x04,0x17}, 0x00
#define RES1024x768   "1024x768",  1024, 768,  65000,  {0x05,0x40,0x04,0x00,0x00,0x00,0x88,0xA0,0x03,0x26,0x03,0x00,0x00,0x00,0x06,0x1D}, 0x00
#define RES1280x1024  "1280x1024", 1280, 1024, 108000, {0x06,0x98,0x05,0x00,0x00,0x00,0x70,0xF8,0x04,0x2A,0x04,0x00,0x00,0x00,0x03,0x26}, 0x00
#define RES1400x900   "1400x900",  1400, 900,  136750, {0x07,0x90,0x05,0xA0,0x00,0x00,0x98,0xF8,0x03,0xAE,0x03,0x84,0x00,0x00,0x06,0x21}, 0x00
#define RES1600x1200  "1600x1200", 1600, 1200, 162000, {0x08,0x70,0x06,0x40,0x00,0x00,0xC0,0x30,0x04,0xE2,0x04,0xB0,0x00,0x00,0x03,0x2E}, 0x01
#define RES1680x1050  "1680x1050", 1680, 1050, 146250, {0x08,0xC0,0x06,0x90,0x00,0x00,0xB0,0x18,0x04,0x41,0x04,0x1A,0x00,0x00,0x06,0x1E}, 0x01
#define RES1920x1080  "1920x1080", 1920, 1080, 148500, {0x08,0x98,0x07,0x80,0x00,0x00,0x2c,0x94,0x04,0x65,0x04,0x38,0x00,0x00,0x05,0x24}, 0x00
#define RES1920x1200  "1920x1200", 1920, 1200, 154000, {0x08,0x20,0x07,0x80,0x00,0x00,0x20,0x50,0x04,0xD3,0x04,0xB0,0x00,0x00,0x06,0x1A}, 0x00
#define RES3840x2160  "3840x2160", 3840, 2160, 297000, {0x11,0x30,0x0f,0x00,0x00,0x00,0x58,0x28,0x08,0xca,0x08,0x70,0x00,0x00,0x0a,0x48}, 0x01

/* Screen A CRTC config for GP101 */
const crtc_mod_reg gp_mode[] = {
    {"\0", 0, 0, 0, {0}, 0, {0}}
};

/* Screen B CRTC common config */
#define B_RES640x480    "640x480",   640,  480,  25175,  {0x02800320, 0x00000808, 0x00280060, 0x01E0020D, 0x19020808, 0x01400004, 0x0
#define B_RES800x600    "800x600",   800,  600,  40000,  {0x03200420, 0x00000000, 0x00580080, 0x02580274, 0x17040000, 0x01900004, 0x0
#define B_RES1024x768   "1024x768",  1024, 768,  65000,  {0x04000540, 0x00000000, 0x00A00088, 0x03000326, 0x1D060000, 0x02000004, 0x0
#define B_RES1280x1024  "1280x1024", 1280, 1024, 108000, {0x05000698, 0x00000000, 0x00F80070, 0x0400042A, 0x26030000, 0x02800004, 0x0
#define B_RES1440x900   "1440x900",  1440,  900, 136750, {0x05A00790, 0x00000000, 0x00F80098, 0x038403AE, 0x21060000, 0x02d00004, 0x0
#define B_RES1600x1200  "1600x1200", 1600, 1200, 162000, {0x06400870, 0x00000000, 0x013000C0, 0x04B004E2, 0x2E030000, 0x03200004, 0x0
#define B_RES1680x1050  "1680x1050", 1680, 1050, 146250, {0x069008C0, 0x00000000, 0x011800B0, 0x041A0441, 0x1E060000, 0x03480004, 0x0
#define B_RES1920x1080  "1920x1080", 1920, 1080, 148500, {0x07800898, 0x00000000, 0x0094002c, 0x04380465, 0x24050000, 0x03c00004, 0x0
#define B_RES1920x1200  "1920x1200", 1920, 1200, 154000, {0x07800820, 0x00000000, 0x00500020, 0x04B004D3, 0x1A060000, 0x03c00004, 0x0
#define B_RES3840x2160  "3840x2160", 3840, 2160, 297000, {0x0F001130, 0x00000000, 0x01280058, 0x087008CA, 0x480A0000, 0x07800004, 0x0

/* Screen B CRTC config for GP101 */
const crtc_b_mode_reg gp_b_mode[] = {
    {"\0", 0, 0, 0, {0}}
};

/* Screen B CRTC config for GP101 FPGA 2D */
const crtc_b_mode_reg gp_b_mode_fpga[] = {
    {B_RES640x480,   0x12120607, 0x00050506}},  //mode 101h ,110h,111h,112h   640*480* 25.175M  60hz
    {B_RES800x600,   0x06060708, 0x00010101}},  //mode 102h ,103h,113h,114h,115h   800*600* 40M  60hz
    {B_RES1024x768,  0x06070505, 0x00020101}},  //mode 104h ,105h,116h,117h,118h   1024*768* 75M  70hz
    {B_RES1280x1024, 0x0d0e0203, 0x00070203}},  //mode 106h ,107h,119h,11ah,11bh   1280*1024* 108M 60hz
    {B_RES1440x900,  0x2e2f0202, 0x00060809}},  //mode 13ch,13dh,13eh,13fH   1440*900* 136.75M   75hz
    {B_RES1600x1200, 0x28290203, 0x00030505}},  //mode 11ch,11dh,11eh,11fH   1600*1200* 162M 60HZ
    {B_RES1680x1050, 0x3a3b0203, 0x00030808}},  //mode 12ch,12dh,12eh,12fH   1680*1050* 146.25M  60HZ
    {B_RES1920x1080, 0x18190102, 0x00070506}},  //mode 15ch,15dh,15eh,15fH   1920*1080* 148.5M 60HZ
    {B_RES1920x1200, 0x26270203, 0x00030505}},  //mode 14ch,14dh,14eh,14fH   1920*1200* 154M 60HZ
    {"\0", 0, 0, 0, {0}}
};

bool ljm_match_vesa_mode(struct ljm_crtc_mode *ljm_mode)
{
    int i;
    HEADER_ARG("ljm_mode=%p", ljm_mode);

    for (i = 0; i < vesa_mode_count; i++) {
        if (ljm_mode->clock == vesa_mode[i].clock
            && ljm_mode->vdisplay == vesa_mode[i].vdisplay
            && ljm_mode->hdisplay == vesa_mode[i].hdisplay) {
            FOOTER_ARG("ret=%d", true);
            return true;
        }
    }

    FOOTER_ARG("ret=%d", false);
    return false;
}

int ljm_get_vesa_mode(struct ljm_crtc_mode *ljm_mode)
{
    int index_pclk = -1;
    int index = -1;
    int i;

    HEADER_ARG("ljm_mode=%p", ljm_mode);

    for (i = 0; i < vesa_mode_count; i++) {
        if (index_pclk < 0) {
            if (ljm_mode->clock == vesa_mode[i].clock
            && ljm_mode->vdisplay == vesa_mode[i].vdisplay
            && ljm_mode->hdisplay == vesa_mode[i].hdisplay) {
                index_pclk = i;
            }
        }

        if (index < 0) {
            if (ljm_mode->vdisplay == vesa_mode[i].vdisplay
                && ljm_mode->hdisplay == vesa_mode[i].hdisplay) {
                index = i;
            }
        }

        if (index_pclk >= 0 && index >= 0) break;
    }

    if (index_pclk >= 0){
        FOOTER_ARG("ret=%d", index_pclk);
        return index_pclk;
    }
    else{
        FOOTER_ARG("ret=%d", index);
        return index;
    }
    FOOTER();
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
static int ljm_crtc_gamma_set(struct drm_crtc *crtc, u16 *red, u16 *green,
                 u16 *blue, uint32_t size,
                 struct drm_modeset_acquire_ctx *ctx)
#else
static int ljm_crtc_gamma_set(struct drm_crtc *crtc, u16 *red, u16 *green,
                 u16 *blue, uint32_t size)
#endif
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);

    return ljm_crtc_load_lut(ljm_dev, ljm_crtc, red, green, blue, size);
}

/* vblank相关功能，有待支持 */
/*static int ljm_crtc_page_flip_target(struct drm_crtc *crtc,
                    struct drm_framebuffer *fb,
                    struct drm_pending_vblank_event *event,
                    uint32_t page_flip_flags,
                    uint32_t target,
                    struct drm_modeset_acquire_ctx *ctx)
{
    FOOTER();
    return 0;
}*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 18)
static int ljm_crtc_set_config(struct drm_mode_set *set, struct drm_modeset_acquire_ctx *ctx)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,0xb,0xc) || LINUX_VERSION_CODE <= KERNEL_VERSION(4,8,6)
static int ljm_crtc_set_config(struct drm_mode_set *set)
#else
static int ljm_crtc_set_config(struct drm_mode_set *set, struct drm_modeset_acquire_ctx *ctx)
#endif
{
    int r = 0;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(set->crtc->dev);

    HEADER_ARG("set=0x%p", set);
    DRM_DEBUG_KMS("enter ljm_crtc_helper_set_config: set=0x%p\n", set);

    //DEBUG_BREAK(1, set, set->mode, set->crtc, set->connectors[0]);
    if (!ljm_dev->is_shutdown) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 18)
        r = drm_crtc_helper_set_config(set, ctx);
#else
        r = drm_crtc_helper_set_config(set);
#endif
    }

    DRM_DEBUG_KMS("leave ljm_crtc_helper_set_config\n");
    FOOTER_ARG("r=%d", r);
    return r;
}

void ljm_crtc_destroy(struct drm_crtc *crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    HEADER_ARG("crtc = 0x%p", crtc);

#ifdef DELAY_4K_DISPLAY
    del_timer(&ljm_crtc->enable_timer);
#endif
    //drm_crtc_cleanup(crtc);
    destroy_workqueue(ljm_crtc->flip_queue); //释放vblank工作对列
    kfree(ljm_crtc);

    FOOTER();
}

#if USE_KMS_VBLANK
/**
 * ljm_unpin_work_func - unpin old buffer object
 *
 * @__work - kernel work item
 *
 * Unpin the old frame buffer object outside of the interrupt handler
 */
static void ljm_unpin_work_func(struct work_struct *__work)
{
    struct ljm_flip_work *work =
        container_of(__work, struct ljm_flip_work, unpin_work);
    int r;

    //HEADER_ARG("__work=0x%p", __work);

    DRM_DEBUG_VBL("unpin and unreference obj(0x%p)\n", work->old_bo);

    r = ljm_bo_unpin(work->old_bo);
    if (unlikely(r != 0)) {
        DRM_ERROR("failed to unpin buffer after flip\n");
    }

    drm_gem_object_unreference_unlocked(&work->old_bo->base);
    kfree(work);

    //FOOTER();
}

/**
 * ljm_save_pflip_base - pageflip callback.
 *
 * @ljm_dev: ljm_device pointer
 * @crtc_id: crtc to cleanup pageflip on
 * @pflip_base: new page flip address for the crtc (GPU MC address)
 *
 * Save the crtc base address in ljm_crtc.
 */
#if 0
static void ljm_save_pflip_base(struct ljm_device *ljm_dev, int crtc_id, u64 pflip_base, bool async)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[crtc_id]);

    HEADER_ARG("ljm_dev=%p crtc_id=%d pflip_base=%llx async=%d", ljm_dev, crtc_id, pflip_base, async);

    /* crtc先保存pflip_base，在vsync上升沿中断（作为vblank中断）时写入base寄存器，crtc在vysnc下降沿时base地址生效 */
    ljm_crtc->pflip_base = pflip_base;

    FOOTER();
}
#endif

void ljm_crtc_handle_vblank(struct ljm_device *ljm_dev, int crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[crtc]);
    unsigned long flags;
    //u32 update_pending;

    //HEADER_ARG("ljm_dev=%p crtc_id=%d", ljm_dev, crtc_id); //驱动运行时该函数被频繁调用，打印过多。

    /* can happen during initialization */
    if (ljm_crtc == NULL) {
        //FOOTER();
        return;
    }

    spin_lock_irqsave(&ljm_dev->ddev->event_lock, flags);
    if (ljm_crtc->flip_status != LJM_FLIP_SUBMITTED) {
        DRM_DEBUG_DRIVER("ljm_crtc->flip_status = %d != LJM_FLIP_SUBMITTED(%d)\n",
                         ljm_crtc->flip_status, LJM_FLIP_SUBMITTED);
        spin_unlock_irqrestore(&ljm_dev->ddev->event_lock, flags);
        //FOOTER();
        return;
    }
    /* pflip completion happens at the end of vsync, while vblank interrupts happens at begin of vsync */
    //update_pending = ljm_page_flip_pending(ljm_dev, ljm_crtc->id);
    spin_unlock_irqrestore(&ljm_dev->ddev->event_lock, flags);

    //if (!update_pending)
        ljm_crtc_handle_flip(ljm_dev, crtc);

    //FOOTER();
}

/**
 * ljm_crtc_handle_flip - page flip completed
 *
 * @ljm_dev: radeon device pointer
 * @crtc_id: crtc number this event is for
 *
 * Called when we are sure that a page flip for this crtc is completed.
 */
void ljm_crtc_handle_flip(struct ljm_device *ljm_dev, int crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[crtc]);
    struct ljm_flip_work *work;
    unsigned long flags;

    //HEADER_ARG("ljm_dev=0x%p crtc_id=%d", ljm_dev, crtc_id);

    /* this can happen at init */
    if (ljm_crtc == NULL) {
        //FOOTER();
        return;
    }

    spin_lock_irqsave(&ljm_dev->ddev->event_lock, flags);
    work = ljm_crtc->flip_work;
    if (ljm_crtc->flip_status != LJM_FLIP_SUBMITTED) {
        DRM_DEBUG_DRIVER("ljm_crtc->flip_status = %d != LJM_FLIP_SUBMITTED(%d)\n",
            ljm_crtc->flip_status, LJM_FLIP_SUBMITTED);
        spin_unlock_irqrestore(&ljm_dev->ddev->event_lock, flags);
        //FOOTER();
        return;
    }

    /* set crtc base address at posedge of a vsync, and it will work at negedge of the vsync */
    //ljm_crtc_set_base(ljm_dev, crtc_id, ljm_crtc->pflip_base);

    /* Pageflip completed. Clean up. */
    ljm_crtc->flip_status = LJM_FLIP_NONE;
    ljm_crtc->flip_work = NULL;
    DRM_DEBUG_VBL("crtc %d: handle flip completed\n", crtc);

    /* wakeup userspace */
    if (work->event)
        drm_crtc_send_vblank_event(&ljm_crtc->base, work->event);
    spin_unlock_irqrestore(&ljm_dev->ddev->event_lock, flags);

    drm_crtc_vblank_put(&ljm_crtc->base);
    ljm_irq_kms_pflip_irq_put(ljm_dev, work->crtc);
    queue_work(ljm_crtc->flip_queue, &work->unpin_work);

    //FOOTER();
}

/**
 * ljm_flip_work_func - page flip framebuffer
 *
 * @work - kernel work item
 *
 * Wait for the buffer object to become idle and do the actual page flip
 */
static void ljm_flip_work_func(struct work_struct *__work)
{
    struct ljm_flip_work *work =
        container_of(__work, struct ljm_flip_work, flip_work);
    struct ljm_device *ljm_dev = work->ljm_dev;
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[work->crtc]);

    struct drm_crtc *crtc = &ljm_crtc->base;
    unsigned long flags;

    //HEADER_ARG("__work=0x%p", __work);

    down_read(&ljm_dev->exclusive_lock);
#if 0
    if (work->fence) {
        struct radeon_fence *fence;

        fence = to_radeon_fence(work->fence);
        if (fence && fence->ljm_dev == ljm_dev) {
            r = radeon_fence_wait(fence, false);
            if (r == -EDEADLK) {
                up_read(&ljm_dev->exclusive_lock);
                do {
                    r = radeon_gpu_reset(ljm_dev);
                } while (r == -EAGAIN);
                down_read(&ljm_dev->exclusive_lock);
            }
        } else
            r = fence_wait(work->fence, false);

        if (r)
            DRM_ERROR("failed to wait on page flip fence (%d)!\n", r);

        /* We continue with the page flip even if we failed to wait on
         * the fence, otherwise the DRM core and userspace will be
         * confused about which BO the CRTC is scanning out
         */

        fence_put(work->fence);
        work->fence = NULL;
    }
#endif

    /* We borrow the event spin lock for protecting flip_status */
    spin_lock_irqsave(&crtc->dev->event_lock, flags);

    /* set the proper interrupt */
    ljm_irq_kms_pflip_irq_get(ljm_dev, ljm_crtc->base.index);

    /* do the flip (mmio) */
    ljm_page_flip(ljm_dev, ljm_crtc->id, work->base, work->async);
    /* delay to excute in vblank irq process */
    //ljm_save_pflip_base(ljm_dev, ljm_crtc->id, work->base, work->async);

    ljm_crtc->flip_status = LJM_FLIP_SUBMITTED;
    spin_unlock_irqrestore(&crtc->dev->event_lock, flags);
    up_read(&ljm_dev->exclusive_lock);

    //FOOTER();
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,15,0)
static int ljm_crtc_page_flip(struct drm_crtc *crtc,
    struct drm_framebuffer *fb,
    struct drm_pending_vblank_event *event,
    uint32_t page_flip_flags)
#else
static int ljm_crtc_page_flip(struct drm_crtc *crtc,
    struct drm_framebuffer *fb,
    struct drm_pending_vblank_event *event,
    uint32_t page_flip_flags,
    struct drm_modeset_acquire_ctx *ctx)
#endif
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct drm_gem_object *obj;
    struct ljm_flip_work *work;
    struct ljm_gem_object *new_bo;
    uint64_t base;
    unsigned long flags;
    int r;
    int bpp;

    //HEADER_ARG("crtc%d=0x%p fb=0x%p event=%p page_flip_flags=%x", crtc->index, crtc, fb, event, page_flip_flags);

    work = kzalloc(sizeof *work, GFP_KERNEL);
    if (work == NULL) {
        DRM_ERROR("ljm_crtc_page_flip: Out of memory\n");
        //FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    INIT_WORK(&work->flip_work, ljm_flip_work_func);
    INIT_WORK(&work->unpin_work, ljm_unpin_work_func);

    work->ljm_dev = ljm_dev;
    work->crtc = ljm_crtc->base.index;
    work->event = event;
    work->async = (page_flip_flags & DRM_MODE_PAGE_FLIP_ASYNC) != 0;

    /* schedule unpin of the old buffer */
    obj = drm_gem_fb_get_obj(crtc->primary->fb, 0);

    /* take a reference to the old object */
    drm_gem_object_reference(obj);
    work->old_bo = to_ljm_gem(obj);

    new_bo = to_ljm_gem(drm_gem_fb_get_obj(fb, 0));

    /* pin the new buffer */
    //DRM_DEBUG_DRIVER("flip-ioctl() cur_bo = %p, new_bo = %p\n",
                     //work->old_bo, new_bo);

    r = ljm_bo_pin(new_bo, &base);
    if (unlikely(r != 0)) {
        r = -EINVAL;
        DRM_ERROR("failed to pin new bo buffer before flip\n");
        goto cleanup;
    }
    //DRM_DEBUG_VBL("new fb offset: 0x%llx\n", base);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    bpp = (fb->format->cpp[0] + fb->format->cpp[1] + fb->format->cpp[2]) * 8;
#else
    bpp = fb->bits_per_pixel;
#endif
    base += crtc->y * fb->pitches[0] + crtc->x * bpp / 8;
    work->base = base;

    r = drm_crtc_vblank_get(crtc);  /* get a reference count on vblank events */
    if (r) {
        DRM_ERROR("failed to get vblank before flip\n");
        goto pflip_cleanup;
    }

    /* We borrow the event spin lock for protecting flip_work */
    spin_lock_irqsave(&crtc->dev->event_lock, flags);
    if (ljm_crtc->flip_status != LJM_FLIP_NONE) {
        DRM_DEBUG_DRIVER("flip queue: crtc already busy\n");
        spin_unlock_irqrestore(&crtc->dev->event_lock, flags);
        r = -EBUSY;
        DRM_DEBUG_VBL("page flip is busy\n");
        goto vblank_cleanup;
    }
    ljm_crtc->flip_status = LJM_FLIP_PENDING;
    ljm_crtc->flip_work = work;

    /* update crtc fb */
    crtc->primary->fb = fb;
    spin_unlock_irqrestore(&crtc->dev->event_lock, flags);

    queue_work(ljm_crtc->flip_queue, &work->flip_work);

    //FOOTER();
    return 0;

vblank_cleanup:
    drm_crtc_vblank_put(crtc);

pflip_cleanup:
    ljm_bo_unpin(new_bo);
cleanup:
    drm_gem_object_unreference_unlocked(obj);
//  fence_put(work->fence);
    kfree(work);

    //FOOTER_ARG("r=%d", r);
    return r;
}
#endif

static const struct drm_crtc_funcs ljm_crtc_funcs = {
    .cursor_set2 = ljm_crtc_cursor_set2,
    .cursor_move = ljm_crtc_cursor_move,
    .gamma_set = ljm_crtc_gamma_set,
    .set_config = ljm_crtc_set_config,
    .destroy = ljm_crtc_destroy,
#if USE_KMS_VBLANK
    .page_flip = ljm_crtc_page_flip,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    .get_vblank_counter = ljm_get_vblank_counter_kms,
    .enable_vblank = ljm_enable_vblank_kms,
    .disable_vblank = ljm_disable_vblank_kms,
//  .get_vblank_timestamp = ljm_get_vblank_timestamp_kms,
#endif
#endif
};

#ifdef DELAY_4K_DISPLAY
static bool ljm_crtc_mode_check_4k(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc)
{
    u32 data = 0;

    if (ljm_crtc->id == 0) {
        data = CR_RD(ljm_dev->cur_card, CRTC_TIMING_HDISPLAY_HIGH);
        data = (data << 8) | CR_RD(ljm_dev->cur_card, CRTC_TIMING_HDISPLAY_LOW);
    }
    else {
        data = REG_R16(ljm_dev->cur_card, CRTC1_HDISPALY + CRTCx_REG_SIZE * (ljm_crtc->id-1));
    }

    if (data >= 3840) {
        return true;
    }
    else {
        return false;
    }
}

void ljm_crtc_mode_delay_enable(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, bool enable)
{
    del_timer(&ljm_crtc->enable_timer);
    ljm_crtc->enable_timer.expires = jiffies + msecs_to_jiffies(2500);
    add_timer(&ljm_crtc->enable_timer);
}
#endif

static void ljm_crtc_dpms(struct drm_crtc *crtc, int mode)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);

    HEADER_ARG("crtc%d=0x%p, mode=%d", crtc->index, crtc, mode);

    switch (mode) {
        case DRM_MODE_DPMS_ON:
            ljm_crtc->enabled = true;
#ifdef DELAY_4K_DISPLAY
            if (ljm_crtc_mode_check_4k(ljm_dev, ljm_crtc)) {
                ljm_crtc_mode_delay_enable(ljm_dev, ljm_crtc, true);
            }
            else {
                ljm_crtc_mode_enable(ljm_dev, ljm_crtc, true);
            }
#else
            ljm_crtc_mode_enable(ljm_dev, ljm_crtc, true);
#endif

#if USE_KMS_VBLANK
            if (ljm_dev->num_crtcs > ljm_crtc->id)
                drm_crtc_vblank_on(crtc);
#endif
            break;
        case DRM_MODE_DPMS_STANDBY:
        case DRM_MODE_DPMS_SUSPEND:
        case DRM_MODE_DPMS_OFF:
#if USE_KMS_VBLANK
            if (ljm_dev->num_crtcs > ljm_crtc->id)
                drm_crtc_vblank_off(crtc);
#endif
            ljm_crtc_mode_enable(ljm_dev, ljm_crtc, false);
            ljm_crtc->enabled = false;
            break;
    }

    FOOTER();
}

int ljm_crtc_mode_set_base(struct drm_crtc *crtc,
                int x, int y, struct drm_framebuffer *old_fb
                )
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct drm_framebuffer *new_fb = NULL;
    struct ljm_gem_object *old_bo = NULL;
    struct ljm_gem_object *new_bo = NULL;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
    u64 vram_offset = 0;
    u32 base = 0;
    int bpp = 0;
    unsigned int pitches = 0;
    int r = 0;

    HEADER_ARG("id=%d, crtc=0x%p x=%d y=%d old_fb=0x%p", ljm_crtc->id, crtc,  x, y, old_fb);

    DRM_DEBUG_KMS("\n");
    if (!crtc->primary->fb) {
        DRM_DEBUG_KMS("CRTC(%d): No FB bound\n", ljm_crtc->id);

        FOOTER();
        return 0;
    }

#if defined(DEBUG)
    if (old_fb && (old_fb != crtc->primary->fb)) {
        DRM_INFO("fb %d is different from old %d. A new framebuffer is created.\n",
            crtc->primary->fb->base.id, old_fb->base.id);
    }
    else if (old_fb) {
        DRM_INFO("fb is same as crts's fb\n");
    }
    else {
        DRM_INFO("fb is NULL.\n");
    }
#endif

    new_fb = crtc->primary->fb;
    new_bo = to_ljm_gem(drm_gem_fb_get_obj(new_fb, 0));

    r = ljm_bo_pin(new_bo, &vram_offset);
    if (unlikely(r != 0)) {
        FOOTER_ARG("ERR: r=%d", r);
        return r;
    }

#if 0   /* no need to clear new framebuffer */
    if (ljm_dev->fb_bo_offset != gpu_offset) {  /* make sure to clear the new framebuffer */
        ljm_dev->fb_bo_offset = gpu_offset;
        ljm_dev->fb_bo_size = get_size_from_bo(new_bo);
        ljm_fb_clear(ljm_dev);
    }
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    bpp = (new_fb->format->cpp[0] + new_fb->format->cpp[1] + new_fb->format->cpp[2]) * 8;
#else
    bpp = new_fb->bits_per_pixel;   //fedora19->21
#endif

    pitches = new_fb->pitches[0];
    base = y * pitches + x * bpp/8 + vram_offset;

    TRACE("CRTC %d base=0x%08X pitches=%d bpp=%d",
        ljm_crtc->id, base, pitches, bpp
        );

    r = ljm_mode_set_base(ljm_dev, crtc, bpp, pitches, base);

    if (old_fb && old_fb != new_fb) {
        old_bo = to_ljm_gem(drm_gem_fb_get_obj(old_fb, 0));
        ljm_bo_unpin(old_bo);
    }

    if (unlikely(r != 0)) {
        FOOTER_ARG("r=%d", r);
    }
    else {
        FOOTER();
    }
    return r;
}

static bool ljm_crtc_mode_fixup(struct drm_crtc *crtc,
                   const struct drm_display_mode *mode,
                   struct drm_display_mode *adjusted_mode)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
    bool r;

    HEADER_ARG("crtc%d=0x%p mode=0x%p adjustedmode=0x%p", crtc->index, crtc, mode, adjusted_mode);
//  if (!ljm_crtc_scaling_mode_fixup(crtc, mode, adjusted_mode)){
//      FOOTER_ARG("%s", "false");
//      return false;
//  }

    r = ljm_mode_fixup(ljm_dev, mode, adjusted_mode);

    FOOTER();
    return r;
}

static int ljm_crtc_mode_set(struct drm_crtc *crtc,
                struct drm_display_mode *mode,
                struct drm_display_mode *adjusted_mode,
                int x, int y, struct drm_framebuffer *old_fb)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
#if !FPGA_HAPS
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
#endif
    int r;

    HEADER_ARG("id=%d crtc=0x%p mode=0x%p adjustedmode=0x%p X=%d Y=%d OldFb=0x%p ",
                 ljm_crtc->id, crtc, mode, adjusted_mode, x, y, old_fb);

    DRM_DEBUG_KMS("clock=%d hdisplay=%d hsync_start=%d hsync_end=%d htotal=%d hskew=%d\n",
        adjusted_mode->clock, adjusted_mode->hdisplay, adjusted_mode->hsync_start, adjusted_mode->hsync_end,
        adjusted_mode->htotal, adjusted_mode->hskew
        );
    DRM_DEBUG_KMS("vdisplay=%d vsync_start=%d vsync_end=%d vtotal=%d vscan=%d flags=%d\n",
        adjusted_mode->vdisplay, adjusted_mode->vsync_start, adjusted_mode->vsync_end, adjusted_mode->vtotal,
        adjusted_mode->vscan, adjusted_mode->flags
        );

    DRM_DEBUG_KMS("vrefresh=%d\n", ljm_mode_vrefresh(adjusted_mode));

    /* set timing */
#if !FPGA_HAPS
    ljm_mode_set(ljm_dev, ljm_crtc, adjusted_mode);
#endif

    /* set vram base address */
    r = ljm_crtc_mode_set_base(crtc, x, y, old_fb);
    if (r) {
        DRM_ERROR("ljm_crtc_mode_set() returns %d\n", r);
        FOOTER_ARG("r=%d", r);
        return r;
    }

    ljm_cursor_reset(crtc);

    FOOTER();
    return 0;
}

static void ljm_crtc_prepare(struct drm_crtc *crtc)
{
    HEADER_ARG("crtc%d=0x%p crtc->enabled=%d", crtc->index, crtc, crtc->enabled);

    ljm_crtc_dpms(crtc, DRM_MODE_DPMS_OFF);

    FOOTER();
}

static int ljm_crtc_reset(struct ljm_device *ljm_dev)
{
    struct drm_crtc *crtc = NULL;
    struct ljm_crtc *ljm_crtc = NULL;
    int i, crtc_used = 0;
    long hscan_time, min_hscan_time = 0;

    for(i = 0; i < ljm_dev->num_crtcs; i++)
    {
        crtc = ljm_dev->crtcs[i];
        if(!crtc)
            continue;

        ljm_crtc = to_ljm_crtc(crtc);

        if(ljm_crtc->enabled)
        {
            crtc_used++;
            hscan_time = ((long)crtc->mode.htotal * 1000000) / (long)crtc->mode.clock;

            if((min_hscan_time == 0) || (min_hscan_time > hscan_time))
                min_hscan_time = hscan_time;
        }
    }

    ljm_crtc_prog_reset(ljm_dev, crtc_used, min_hscan_time);
    return 0;
}

static void ljm_crtc_commit(struct drm_crtc *crtc)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);

    HEADER_ARG("crtc%d=%p crtc->enabled=%d", crtc->index, crtc, crtc->enabled);

    ljm_crtc_dpms(crtc, DRM_MODE_DPMS_ON);
    if(!ljm_dev->once_edp)
    {
        /**
         * @brief ljm_crtc_reset函数用于延时复位crtc，以减轻gp201显示带宽不足的影响。
         * 实际应用中发现该功能与once_edp存在冲突，在某些环境下易导致edp不显示。考虑到
         * edp和高带宽需求通常不同时存在，只在非once_Edp模式下使用执行。
         * 
         */
        ljm_crtc_reset(ljm_dev);
        ljm_crtc_commit_dp(ljm_dev);
    }

    FOOTER();
}

static void ljm_crtc_disable(struct drm_crtc *crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);

    HEADER_ARG("crtc%d=0x%p", crtc->index, crtc);

    /* only disable crtc */
    ljm_crtc_mode_enable(ljm_dev, ljm_crtc, false);
    ljm_crtc->enabled = false;

    FOOTER();
}

/* drm_crtc legacy helper functions */
static const struct drm_crtc_helper_funcs ljm_crtc_helper_funcs = {
    .dpms = ljm_crtc_dpms,
    .mode_fixup = ljm_crtc_mode_fixup,
    .mode_set = ljm_crtc_mode_set,
    .mode_set_base = ljm_crtc_mode_set_base,
//  .mode_set_base_atomic = ljm_crtc_set_base_atomic,
    .prepare = ljm_crtc_prepare,
    .commit  = ljm_crtc_commit,
    .disable = ljm_crtc_disable
};

#ifdef DELAY_4K_DISPLAY
static void crtc_enable_timer_callback(struct timer_list *callback_timer)
{
    struct ljm_crtc *ljm_crtc = container_of(callback_timer, struct ljm_crtc, enable_timer);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm((&ljm_crtc->base)->dev);

    ljm_crtc_mode_enable(ljm_dev, ljm_crtc, true);
}
#endif

void ljm_crtc_init(struct drm_device *ddev, int index)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct ljm_crtc *ljm_crtc;
    struct drm_crtc *crtc;
    uint32_t gamma_size = 256; //gamma_size默认值

    HEADER_ARG("ddev=0x%p index=%d ljm_dev=%p", ddev, index, ljm_dev);

    ljm_crtc = kzalloc(sizeof(struct ljm_crtc)/* + (LJMFB_CONN_LIMIT * sizeof(struct drm_connector *))*/, GFP_KERNEL);
    if (ljm_crtc == NULL) {
        FOOTER_ARG("%d", -ENOMEM);
        return;
    }

    drm_crtc_init(ddev, &ljm_crtc->base, &ljm_crtc_funcs);

    ljm_crtc->id = index;
    ljm_crtc->flip_queue = alloc_workqueue("ljm-crtc", WQ_HIGHPRI, 0);

    switch (ljm_dev->family) {
        case CHIP_GP100:
        case CHIP_GP200:
            ljm_crtc->max_cursor_width = CURSOR_MAX_WIDTH;
            ljm_crtc->max_cursor_height = CURSOR_MAX_HEIGHT;
            break;
        default:
            break;
    }

    ddev->mode_config.cursor_width = CURSOR_WIDTH;
    ddev->mode_config.cursor_height = CURSOR_HEIGHT;

    crtc = &ljm_crtc->base;
    ljm_dev->crtcs[index] = crtc;

    drm_crtc_helper_add(crtc, &ljm_crtc_helper_funcs);

    sema_init(&ljm_crtc->vblank_sema, 0);

    ljm_crtc->lut.red = kmalloc((sizeof(uint16_t)*gamma_size), GFP_KERNEL);
    ljm_crtc->lut.green = kmalloc((sizeof(uint16_t)*gamma_size), GFP_KERNEL);
    ljm_crtc->lut.blue = kmalloc((sizeof(uint16_t)*gamma_size), GFP_KERNEL);

    drm_mode_crtc_set_gamma_size(&ljm_crtc->base, gamma_size);

#ifdef DELAY_4K_DISPLAY
    timer_setup(&ljm_crtc->enable_timer, crtc_enable_timer_callback, 0);
#endif

    DRM_DEBUG_KMS("new crtc(%lx): index=%d base.id=%d\n", (unsigned long)crtc, crtc->index, crtc->base.id);

    FOOTER_ARG("new crtc%d=%p(%lx) %d", crtc->index, crtc, (unsigned long)crtc, crtc->base.id);
}

