/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_device.h"
#include "ljmicro_mode.h"
#include "ljmicro_object.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

static void ljm_lock_cursor(struct drm_crtc *crtc, bool lock)
{

}

static void ljm_hide_cursor(struct drm_crtc *crtc)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
//    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

    HEADER_ARG("crtc=%p", crtc);

    ljm_dev->curor_en[crtc->index] = false;

    ljm_crtc_cursor_hide(ljm_dev, crtc);

    FOOTER();
}

static void ljm_show_cursor(struct drm_crtc *crtc)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
//    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

    HEADER_ARG("crtc=0x%p cursor %d show", crtc, crtc->index);

    ljm_dev->curor_en[crtc->index] = true;

    ljm_crtc_cursor_show(ljm_dev, crtc);

    FOOTER();
}

/*static int ljm_cursor_move_locked(struct drm_crtc *crtc, int x, int y)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
    int x_orig = 0, y_orig = 0;
    int w = ljm_crtc->cursor_width;

    ljm_crtc->cursor_x = x;
    ljm_crtc->cursor_y = y;


    return 0;
}*/

int ljm_crtc_cursor_move(struct drm_crtc *crtc,
    int x, int y)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(crtc->dev);
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

    ljm_cursor_move(ljm_dev, crtc, x, y);

    ljm_crtc->cursor_x = x;
    ljm_crtc->cursor_y = y;

    return 0;
}


int ljm_crtc_cursor_set2(struct drm_crtc *crtc,
                struct drm_file *file_priv,
                uint32_t handle,
                uint32_t width,
                uint32_t height,
                int32_t hot_x,
                int32_t hot_y)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
    struct drm_gem_object *obj;
    struct ljm_gem_object *ljm_obj;
    u64 cursor_offset;

    HEADER_ARG("crtc=0x%p filp=0x%p handle=%x width=%d height=%d hot_x=%d hot_y=%d",
                crtc, file_priv, handle, width, height, hot_x, hot_y);

    if (!handle) {
        /* turn off cursor */
        ljm_hide_cursor(crtc);
        if (ljm_crtc->cursor_obj) {
            drm_gem_object_unreference_unlocked(ljm_crtc->cursor_obj);
        }
        ljm_crtc->cursor_obj = NULL;
        FOOTER();
        return 0;
    }

    if ((width != 64 && width != 32) || width != height) {
        DRM_ERROR("bad cursor width or height %d x %d\n", width, height);
        FOOTER_ARG("r=%d", -EINVAL);
        return -EINVAL;
    }

    obj = DRM_GEM_OBJECT_LOOKUP(crtc->dev, file_priv, handle);
    if (!obj) {
        DRM_ERROR("Cannot find cursor object %x for crtc %d\n", handle, crtc->index);
        FOOTER_ARG("r=%d", -ENOENT);
        return -ENOENT;
    }

    ljm_obj = to_ljm_gem(obj);
    if(ljm_bo_pin(ljm_obj, &cursor_offset))
    {
        DRM_ERROR("Failed to pin new cursor bo\n");
        drm_gem_object_unreference_unlocked(obj);
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    ljm_crtc->cursor_addr = (uint32_t)cursor_offset;
    ljm_crtc->cursor_size = width;

    if (hot_x != ljm_crtc->cursor_hot_x || hot_y != ljm_crtc->cursor_hot_y) {
        int x, y;
        DRM_DEBUG_KMS("cursor hot move (%d,%d) to (%d,%d)\n",
            ljm_crtc->cursor_hot_x, ljm_crtc->cursor_hot_y, hot_x, hot_y
            );

        x = ljm_crtc->cursor_x + ljm_crtc->cursor_hot_x - hot_x;
        y = ljm_crtc->cursor_y + ljm_crtc->cursor_hot_y - hot_y;

        ljm_crtc_cursor_move(crtc, x, y);

        ljm_crtc->cursor_hot_x = hot_x;
        ljm_crtc->cursor_hot_y = hot_y;
    }

    ljm_show_cursor(crtc);

    if (ljm_crtc->cursor_obj) {
        ljm_obj = to_ljm_gem(ljm_crtc->cursor_obj);
        ljm_bo_unpin(ljm_obj);
        drm_gem_object_unreference_unlocked(ljm_crtc->cursor_obj);
    }

    ljm_crtc->cursor_obj = obj;

    FOOTER();

    return 0;
}

void ljm_cursor_reset(struct drm_crtc *crtc)
{
    struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

    HEADER_ARG("crtc=%p", crtc);

    DRM_DEBUG_KMS("crtc %d: cursor reset\n", ljm_crtc->id);

    if (ljm_crtc->cursor_obj) {
        ljm_lock_cursor(crtc, true);

        ljm_crtc_cursor_move(crtc, ljm_crtc->cursor_x, ljm_crtc->cursor_y);

        ljm_show_cursor(crtc);

        ljm_lock_cursor(crtc, false);
    }
    FOOTER();
}

