/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_base.h"
#include "ljmicro_device.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

/* DAC 0 registers in CRTC region */
#define DAC0_HPD_ST_CR_ADR      0X8A    /* [0]: 0-connected, 1-disconnected;   与VGA寄存器文档描述相反 ——sxk
                                           [1]: 0-not hpd interrupted, 1-hpd interrupted (write anything to clear) */
#define DAC0_HPD_INT_EN_CR_ADR  0X8B    /* [0]: 0-disable hpd interrupt, 1-enable hpd interrupt */

#define DAC0_CH0_GAIN_CR_ADR    0X8C    /* [5:0]: channel 0 current gain */
#define DAC0_CH1_GAIN_CR_ADR    0X8D    /* [5:0]: channel 1 current gain */
#define DAC0_CH2_GAIN_CR_ADR    0X8E    /* [5:0]: channel 2 current gain */

#define DAC0_PWR_CTRL_CR_ADR    0X8F    /* [2:0]: channel [2:0] power off enable, 0-power on, 1-power off;
                                           [3]: dac power off enable, 0-power on, 1-power off */

#define DAC0_CTRL_LOW_CR_ADR    0XAC    /* low byte of the control register */
#define DAC0_CTRL_HIGH_CR_ADR   0XAD    /* high byte of the control register */


/* there is only a dac0 in GP201 chip */
int ljm_dac_fs_current_set(int card, int index, int ch, u8 val);
int ljm_dac_ch_gain_set(struct vga_dev *vga, u8 red, u8 green, u8 blue);

/* power down DAC */
int ljm_dac_pwr_down(struct vga_dev *vga, int card, int index)
{
    u8 red, green, blue;
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0) {
        /* current gain to min */

        red = CR_RD(card, DAC0_CH0_GAIN_CR_ADR);
        green = CR_RD(card, DAC0_CH1_GAIN_CR_ADR);
        blue = CR_RD(card, DAC0_CH2_GAIN_CR_ADR);
        ljm_dac_ch_gain_set(vga, red, green, blue);

        ljm_dac_fs_current_set(card, index, 0, 0);
        ljm_dac_fs_current_set(card, index, 1, 0);
        ljm_dac_fs_current_set(card, index, 2, 0);
        /* power down channels */
        CR_WR(card, DAC0_PWR_CTRL_CR_ADR, 0xf);
    }

    return 0;
}

/* power up DAC (default) */
int ljm_dac_pwr_up(struct vga_dev *vga, int card, int index)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0) {
        /* current gain to max */
        ljm_dac_fs_current_set(card, index, 0, vga->dac_ch_red);
        ljm_dac_fs_current_set(card, index, 1, vga->dac_ch_green);
        ljm_dac_fs_current_set(card, index, 2, vga->dac_ch_blue);
        /* power on all channel */
        CR_WR(card, DAC0_PWR_CTRL_CR_ADR, 0x0);
    }

    return 0;
}

void ljm_dac_dpms(struct vga_dev *vga, int card, int index, int mode)
{
    HEADER_ARG("card=%d index=%d mode=%d", card, index, mode);

    /* TODO: add dpms */
    switch (mode) {
    case DRM_MODE_DPMS_ON:
        /* power up dac */
        ljm_dac_pwr_up(vga, card, index);
        break;
    case DRM_MODE_DPMS_STANDBY:
    case DRM_MODE_DPMS_SUSPEND:
    case DRM_MODE_DPMS_OFF:
        /* power down dac */
        ljm_dac_pwr_down(vga, card, index);
        break;
    }

    FOOTER();
}

/* dac full scale current control
 * GC[5:0]    FS Curent(mA)
 * 0x0        17
 * 0x10       21.3
 * 0x20       25.5
 * 0x30       30
 * 0x3f       33.7
 * Ifs = 17mA + GC[5:0]/64*17mA = 17mA + GC[5:0]*0.27mA
 */
int ljm_dac_fs_current_set(int card, int index, int ch, u8 val)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0) {
        if (ch < 0 || ch > 2) {
            printk(KERN_ERR "dac channel %d is out of range\n", ch);
            return -EINVAL;
        }

        CR_WR(card, DAC0_CH0_GAIN_CR_ADR + ch, val);
    }

    return 0;
}

/* DAC control set */
int ljm_dac_ctrl_set(int card, int index, u16 val)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0) {
        CR_WR(card, DAC0_CTRL_LOW_CR_ADR, val & 0xff);
        CR_WR(card, DAC0_CTRL_HIGH_CR_ADR, (val & 0xff00) >> 8);
    }

    return 0;
}

/* DAC control get */
int ljm_dac_ctrl_get(int card, int index, u16 *val)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0)
        *val = (u16)CR_RD(card, DAC0_CTRL_LOW_CR_ADR) +
               ((u16)CR_RD(card, DAC0_CTRL_HIGH_CR_ADR) << 8);
    else
        return -EINVAL;

    return 0;
}

/* Is DAC connected with a monitor */
bool ljm_dac_is_connected(int card, int index)
{
    bool connected = false;

    HEADER_ARG("card=%d index=%d", card, index);

    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0)
        connected = !!(CR_RD(card, DAC0_HPD_ST_CR_ADR) & 0x1);

    FOOTER_ARG("connected=%d", connected);
    return connected;
}

/* Is DAC met with a hpd interrupt */
bool ljm_dac_is_hpd_irq(int card, int index)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0)
        return !!(CR_RD(card, DAC0_HPD_ST_CR_ADR) & 0x2);
    else
        return false;
}

/* DAC hpd interrupt clear */
int ljm_dac_hpd_irq_clear(int card, int index)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0)
        CR_WR(card, DAC0_HPD_ST_CR_ADR, 0x2);

    return 0;
}

/*
 * ljm_dac_hpd_irq_enable - dac hpd enable
 *
 * @hdmi_dev: pointer to hdmi device
 * @enable: 0 to disable, otherwise enable
 */
int ljm_dac_hpd_irq_enable(int card, int index, int enable)
{
    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0) {
        if (enable)
            CR_WR(card, DAC0_HPD_INT_EN_CR_ADR, 0x1);
        else
            CR_WR(card, DAC0_HPD_INT_EN_CR_ADR, 0x0);
    }

    return 0;
}

/**
 * ljm_dac_irq_process - dac irq process
 *
 * @hdmi_dev: pointer to hdmi device
 *
 * Return: irq type
 */
int ljm_dac_irq_process(int card, int index)
{
    int r = LJM_IRQ_TYPE_NONE;

    VERIFY_ARGUMENT(card >= 0 && card < g_card_cnt);
    VERIFY_ARGUMENT(index >= 0 && index < 4);

    if (index == 0 && (CR_RD(card, DAC0_HPD_ST_CR_ADR) & 0x2)) {
        r |= LJM_IRQ_TYPE_HPD;
        ljm_dac_hpd_irq_clear(card, index);
    }

    return r;
}

/*****************************************************************************
 *  函 数 名: ljm_dac_mode_valid
 *  功能描述: DAC模块支持最大的输入时钟为300MHz，超过该频率的分辨率不支持
 *  输入参数: const struct drm_display_mode *mode
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月21日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dac_mode_valid(const struct drm_display_mode *mode)
{
    bool r;

//    HEADER_ARG("mode=0x%p", mode);

    if (mode->clock > 300000) {
        r = false;
    }
    else {
        r = true;
    }

    if (r) {
        TRACE("%dx%d@%dkHz is valid", mode->hdisplay, mode->vdisplay, mode->clock);
    } else {
        TRACE("%dx%d@%dkHz is invalid", mode->hdisplay, mode->vdisplay, mode->clock);
    }

//    FOOTER_ARG("%d", r);
    return r;
}

/*****************************************************************************
 *  函 数 名：ljm_dac_ch_gain_set
 *  功能描述：预设DAC颜色通道的增益值
 *  输入参数：struct ljm_port *ljm_port
 *            u8 red
 *            u8 green
 *            u8 blue
 *  输出参数：无
 *  返 回 值：int
 *
 *  修改历史：
 *         1. 日    期：2023年3月17日
 *            作    者：苏徐昆
 *            修改内容：新生成函数
 *
 *****************************************************************************/
int ljm_dac_ch_gain_set(struct vga_dev *vga, u8 red, u8 green, u8 blue)
{
//    HEADER_ARG("ljm_port=0x%p, red=%d, green=%d, blue=%d", ljm_port, red, green, blue);

    if (red > 0x3f)
        red = 0x3f;

    if (green > 0x3f)
        green = 0x3f;

    if (blue > 0x3f)
        blue = 0x3f;

    vga->dac_ch_red = red;
    vga->dac_ch_green = green;
    vga->dac_ch_blue = blue;

//    FOOTER_ARG();
    return 0;
}

