/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_DDRPHY_H__
#define __LJMICRO_DDRPHY_H__

#include "ljmicro_ddrcom.h"
/*
    ********************
     APBONLY reg group
    ********************
*/

#define DDR_PHY_MICROCONTMUXSEL 0x00 /* pmu config mux select*/
/*
   REG : MicroContMuxSel offset : 000
   bit[0]  MicroContMuxSel      : 1 = MicroController/PHY Init Engine has control of csr bus.
                                  0 = MicroController/PHY Init Engine csr requests are ignored.
*/
#define m_MICROCONTMUXSEL      (1U << 0)



#define DDR_PHY_UCTSHADOWREGS (0x4 * 4)
/*
    bit[0] UctWriteProtShadow  When set to 0, the PMU has a message for the user
*/
#define m_UCTWRITEPROTSHADOW (1U << 0)



#define DDR_PHY_DCTWRITEPROT (0x31 * 4)
/*
    bit[0] DctWriteProt By setting this register to 0, the user acknowledges the receipt of the message.
*/
#define m_DCTWRITEPROT (1U << 0)



#define DDR_PHY_UCTWRITEONLYSHADOW (0x32 * 4)
/*
    [15:0] UctWriteOnlyShadow Used to pass the message ID for major messages. Also used to pass the lower
                              16 bits for streaming messages.
*/


#define DDR_PHY_MICRORESET (0x99 * 4) /* pmu config mux select*/
/*
    REG :  MicroReset offset:0x99
    bit[3] ResetToMicro : Set this bit to apply synchronous reset to the microcontroller
                              Clear this bit to release reset to the microcontroller.
    bit[0] StallToMicro : Set this bit to stall the microcontroller by hardware.
                              Clear this bit to allow the microcontroller to continue executing from its current
                              program counter location.
*/
#define m_RESETTOMICRO (1U << 3)
#define m_STALLTOMICRO (1U << 0)


#define DDR_PHY_SEQUENCEROVERRIDE (0xe7 * 4)  //Reserved for PHY training firmware use



/*
    **********************
     master phy reg group
    **********************
*/
//These registers are used to control various delays during PLL, DLL initialization.
// delay unit different between DDR_PHY_SEQ0BDLY0~DDR_PHY_SEQ0BDLY3
#define DDR_PHY_SEQ0BDLY0 (0xb * 4) // PHY Initialization Engine (PIE) Delay Register 0
#define DDR_PHY_SEQ0BDLY1 (0xc * 4) //PHY Initialization Engine (PIE) Delay Register 1
#define DDR_PHY_SEQ0BDLY2 (0xd * 4) //PHY Initialization Engine (PIE) Delay Register 2
#define DDR_PHY_SEQ0BDLY3 (0xe * 4) //PHY Initialization Engine (PIE) Delay Register 3

uint32_t ddr_phy_pie_delay[7][4] = {
    {0x32, 0x64, 0x3e8, 0x2c},//800
    {0x32, 0x64, 0x3e8, 0x2c},//1066
    {0x32, 0x64, 0x3e8, 0x2c},//1600
    {0x4b, 0x96, 0x5dc, 0x2c},//2400
    {0x53, 0xa6, 0x682, 0x2c},//2666
    {0x5b, 0xb7, 0x728, 0x2c},//2933
    {0x64, 0xc8, 0x7d0, 0x2c},//3200
};


#define DDR_PHY_CALZAP (0x89 * 4)
// bit[0] calzap :Setting this csr resets the calibrator to its idle state.
#define m_CALZAP (0x1 << 0)

#define DDR_PHY_CALRATE (0x88 * 4)
/*
    REG CalRate 0x88
    bit[3:0] CalInterval : This CSR specifies the interval between successive calibrations, in mS.
    bit[4]   CalRun      : 1: A calibration sequence will be triggered by the 0->1 transition of this
                              bit, as determined by CSR CalOnce
*/


#define  DDR_PHY_CALBUSY (0x97 * 4)
//bit[0] CalBusy : Read 1 if the calibrator is actively calibrating.
#define m_CALBUSY (0x1 << 0)


/*
    **********************
     drtub (pub) reg group
    **********************
*/
#define DDR_PHY_UCCLKHCLKENABLES (0x80 * 4)
//bit[0] UcclkEn : When training has completed , the enable should be set to 0 to reduce power.
//bit[1] HclkEn  : When training has completed , the enable should be set to 0 to reduce power.

/*
    **********************
     init eng reg group
    **********************
*/
#define DDR_PHY_SEQ0BDISABLEFLAG0 (0xc *4)
#define DDR_PHY_SEQ0BDISABLEFLAG1 (0xd *4)
#define DDR_PHY_SEQ0BDISABLEFLAG2 (0xe *4)
#define DDR_PHY_SEQ0BDISABLEFLAG3 (0xf *4)
#define DDR_PHY_SEQ0BDISABLEFLAG4 (0x10 *4)
#define DDR_PHY_SEQ0BDISABLEFLAG5 (0x11 *4)
#define DDR_PHY_SEQ0BDISABLEFLAG6 (0x12 *4)
#define DDR_PHY_SEQ0BDISABLEFLAG7 (0x13 *4)

uint32_t ddr_phymaster_tim_cfg[7][7] = {
    {0x0000000b, 0x00000001, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x0000000a},
    {0x0000000b, 0x00000001, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x0000000a},
    {0x0000000b, 0x00000001, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x0000000a},
    {0x0000000a, 0x00000002, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x00000006},
    {0x00000019, 0x00000002, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x00000007},
    {0x00000019, 0x00000002, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x00000007},
    {0x00000019, 0x00000002, 0x00000008, 0x00000002, 0x00000212, 0x00000061, 0x00000007},
};

uint32_t ddr_phymaster_comm_cfg[7][7] = {
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x00000190, 0x00000009, 0x00000248}, //800M
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x00000190, 0x00000009, 0x00000248}, //1066M
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x00000190, 0x00000009, 0x00000248}, //1600M
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x00000258, 0x00000009, 0x00000208}, //2400M
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x0000029b, 0x00000009, 0x00000208}, //2666M
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x000002dd, 0x00000009, 0x00000208}, //2933M
    {0x00000005, 0x00000002, 0x00000011, 0x00000002, 0x00000320, 0x00000009, 0x00000208}, //3200M
};

ddr_reg_val_v2 ddr_initeng_cfg[] = {
    {0x00000000, 0x00000010},
    {0x00000004, 0x00000400},
    {0x00000008, 0x0000010e},
    {0x0000000c, 0x00000000},
    {0x00000010, 0x00000000},
    {0x00000014, 0x00000008},
    {0x000000a4, 0x0000000b},
    {0x000000a8, 0x00000480},
    {0x000000ac, 0x00000109},
    {0x000000b0, 0x00000008},
    {0x000000b4, 0x00000448},
    {0x000000b8, 0x00000139},
    {0x000000bc, 0x00000008},
    {0x000000c0, 0x00000478},
    {0x000000c4, 0x00000109},
    {0x000000c8, 0x00000002},
    {0x000000cc, 0x00000010},
    {0x000000d0, 0x00000139},
    {0x000000d4, 0x0000000b},
    {0x000000d8, 0x000007c0},
    {0x000000dc, 0x00000139},
    {0x000000e0, 0x00000044},
    {0x000000e4, 0x00000633},
    {0x000000e8, 0x00000159},
    {0x000000ec, 0x0000014f},
    {0x000000f0, 0x00000630},
    {0x000000f4, 0x00000159},
    {0x000000f8, 0x00000047},
    {0x000000fc, 0x00000633},
    {0x00000100, 0x00000149},
    {0x00000104, 0x0000004f},
    {0x00000108, 0x00000633},
    {0x0000010c, 0x00000179},
    {0x00000110, 0x00000008},
    {0x00000114, 0x000000e0},
    {0x00000118, 0x00000109},
    {0x0000011c, 0x00000000},
    {0x00000120, 0x000007c8},
    {0x00000124, 0x00000109},
    {0x00000128, 0x00000000},
    {0x0000012c, 0x00000001},
    {0x00000130, 0x00000008},
    {0x00000134, 0x00000000},
    {0x00000138, 0x0000045a},
    {0x0000013c, 0x00000009},
    {0x00000140, 0x00000000},
    {0x00000144, 0x00000448},
    {0x00000148, 0x00000109},
    {0x0000014c, 0x00000040},
    {0x00000150, 0x00000633},
    {0x00000154, 0x00000179},
    {0x00000158, 0x00000001},
    {0x0000015c, 0x00000618},
    {0x00000160, 0x00000109},
    {0x00000164, 0x000040c0},
    {0x00000168, 0x00000633},
    {0x0000016c, 0x00000149},
    {0x00000170, 0x00000008},
    {0x00000174, 0x00000004},
    {0x00000178, 0x00000048},
    {0x0000017c, 0x00004040},
    {0x00000180, 0x00000633},
    {0x00000184, 0x00000149},
    {0x00000188, 0x00000000},
    {0x0000018c, 0x00000004},
    {0x00000190, 0x00000048},
    {0x00000194, 0x00000040},
    {0x00000198, 0x00000633},
    {0x0000019c, 0x00000149},
    {0x000001a0, 0x00000010},
    {0x000001a4, 0x00000004},
    {0x000001a8, 0x00000018},
    {0x000001ac, 0x00000000},
    {0x000001b0, 0x00000004},
    {0x000001b4, 0x00000078},
    {0x000001b8, 0x00000549},
    {0x000001bc, 0x00000633},
    {0x000001c0, 0x00000159},
    {0x000001c4, 0x00000d49},
    {0x000001c8, 0x00000633},
    {0x000001cc, 0x00000159},
    {0x000001d0, 0x0000094a},
    {0x000001d4, 0x00000633},
    {0x000001d8, 0x00000159},
    {0x000001dc, 0x00000441},
    {0x000001e0, 0x00000633},
    {0x000001e4, 0x00000149},
    {0x000001e8, 0x00000042},
    {0x000001ec, 0x00000633},
    {0x000001f0, 0x00000149},
    {0x000001f4, 0x00000001},
    {0x000001f8, 0x00000633},
    {0x000001fc, 0x00000149},
    {0x00000200, 0x00000000},
    {0x00000204, 0x000000e0},
    {0x00000208, 0x00000109},
    {0x0000020c, 0x0000000a},
    {0x00000210, 0x00000010},
    {0x00000214, 0x00000109},
    {0x00000218, 0x00000009},
    {0x0000021c, 0x000003c0},
    {0x00000220, 0x00000149},
    {0x00000224, 0x00000009},
    {0x00000228, 0x000003c0},
    {0x0000022c, 0x00000159},
    {0x00000230, 0x00000018},
    {0x00000234, 0x00000010},
    {0x00000238, 0x00000109},
    {0x0000023c, 0x00000000},
    {0x00000240, 0x000003c0},
    {0x00000244, 0x00000109},
    {0x00000248, 0x00000018},
    {0x0000024c, 0x00000004},
    {0x00000250, 0x00000048},
    {0x00000254, 0x00000018},
    {0x00000258, 0x00000004},
    {0x0000025c, 0x00000058},
    {0x00000260, 0x0000000b},
    {0x00000264, 0x00000010},
    {0x00000268, 0x00000109},
    {0x0000026c, 0x00000001},
    {0x00000270, 0x00000010},
    {0x00000274, 0x00000109},
    {0x00000278, 0x00000005},
    {0x0000027c, 0x000007c0},
    {0x00000280, 0x00000109},
    {0x00000284, 0x00000000},
    {0x00000288, 0x00008140},
    {0x0000028c, 0x0000010c},
    {0x00000290, 0x00000010},
    {0x00000294, 0x00008138},
    {0x00000298, 0x00000104},
    {0x0000029c, 0x00000008},
    {0x000002a0, 0x00000448},
    {0x000002a4, 0x00000109},
    {0x000002a8, 0x0000000f},
    {0x000002ac, 0x000007c0},
    {0x000002b0, 0x00000109},
    {0x000002b4, 0x00000047},
    {0x000002b8, 0x00000630},
    {0x000002bc, 0x00000109},
    {0x000002c0, 0x00000008},
    {0x000002c4, 0x00000618},
    {0x000002c8, 0x00000109},
    {0x000002cc, 0x00000008},
    {0x000002d0, 0x000000e0},
    {0x000002d4, 0x00000109},
    {0x000002d8, 0x00000000},
    {0x000002dc, 0x000007c8},
    {0x000002e0, 0x00000109},
    {0x000002e4, 0x00000008},
    {0x000002e8, 0x00008140},
    {0x000002ec, 0x0000010c},
    {0x000002f0, 0x00000000},
    {0x000002f4, 0x00000478},
    {0x000002f8, 0x00000109},
    {0x000002fc, 0x00000000},
    {0x00000300, 0x00000001},
    {0x00000304, 0x00000008},
    {0x00000308, 0x00000008},
    {0x0000030c, 0x00000004},
    {0x00000310, 0x00000000},
    {0x00000018, 0x00000008},
    {0x0000001c, 0x000007c8},
    {0x00000020, 0x00000109},
    {0x00000024, 0x00000000},
    {0x00000028, 0x00000400},
    {0x0000002c, 0x00000106},
};

#endif

