/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
*
*  版权所有(c)，2020-2099，中船第709研究所
*
*******************************************************************************
*
*  文件名称: ljmicro_ddrv2.c
*  版 本 号: V1.0
*  作    者: huym，liuty
*  生成日期: 2022年4月15日
*
*  功能描述: DDR3 V1控制器驱动
*
*  修改历史:
*         1. 日    期: 2022年3月15日
*            作    者: huym，liuty
*            修改内容: 创建文件
*
******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/

#include "gp200.h"
#include "ljmicro_ddrv2.h"
#include "ljmicro_ddrphy.h"
#include "ljmicro_ddrctl.h"
#include "ljmicro_ddrcom.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

/**
 * DDR
 */
#define DDR_STAT 0x4
#define m_STAT   (0x7 << 0)

#define IMEM_SIZE 0x8000 /* 32768 */

#define DDR_ONLY_DUMP_REG 0
#define DDR_DUMP_REG      0

#define DDR_CRU_STATUS_STA_DDR0PLL_LOCK_SHIFT   (27)
#define DDR_CRU_STATUS_STA_DDR0PLL_LOCK_MASK    (1 << DDR_CRU_STATUS_STA_DDR0PLL_LOCK_SHIFT)
#define DDR_CRU_STATUS_STA_DDR0PLL_LOCK_LOCKED  \
    (((1) << (DDR_CRU_STATUS_STA_DDR0PLL_LOCK_SHIFT)) & (DDR_CRU_STATUS_STA_DDR0PLL_LOCK_MASK))


#define DDR_CTR_R32_BASE(BASE, INDEX, ADDR) \
    readl((void __iomem *)(BASE) + ljm_gpu_hardware.reg_ddr_ctr_offset[INDEX] + (ADDR))

#if DDR_ONLY_DUMP_REG
#define REG_W32_BASE(BASE, ADDR, VAL)        PRINT("%08x %08x W FFFFFFFF 32\n", (ADDR), (VAL))
#define REG_CHECK_REG(BASE, ADDR, VAL, MASK) PRINT("%08x %08x R %08X 32\n", (ADDR), (VAL), (MASK))

#define DDR_CTR_W32_BASE(BASE, INDEX, ADDR, VAL) \
    PRINT("%08x %08x W FFFFFFFF         32\n", (ljm_gpu_hardware.reg_ddr_ctr_offset[INDEX] + (ADDR)), (u32)(VAL))

#define DDR_PHY_W32_BASE(BASE, INDEX, ADDR, VAL) \
    PRINT("%08x %08x W FFFFFFFF         32\n", (ljm_gpu_hardware.reg_ddr_phy_offset[INDEX] + (ADDR)), (u32)(VAL))

#define DDR_CTR_CHECK_REG(BASE, INDEX, ADDR, VAL, MASK) \
    PRINT("%08x %08x R %08X 32\n", ljm_gpu_hardware.reg_ddr_ctr_offset[INDEX] + (ADDR), (VAL), (MASK))

#define DDR_PHY_CHECK_REG(BASE, INDEX, ADDR, VAL, MASK) \
    PRINT("%08x %08x R %08X 32\n", ljm_gpu_hardware.reg_ddr_phy_offset[INDEX] + (ADDR), (VAL), (MASK))
#else /* !DDR_ONLY_DUMP_REG */
#define REG_W32_BASE(BASE, ADDR, VAL) writel((u32)(VAL), (void __iomem *)(BASE) + (ADDR))
#define REG_CHECK_REG(BASE, ADDR, VAL, MASK) \
    do { \
        int _j; \
        for (_j = 0; _j < 0xFF; _j++) { \
            if ((readl((void __iomem *)(BASE) + (ADDR)) & (MASK)) == ((VAL) & (MASK))) { \
                    break; \
                } \
            msleep(1); \
        } \
        if (_j == 0xFF) { \
            ERR_PRINT("DDR(%d): STAT error\n", __LINE__); \
            ret = -1; \
            /*goto OnError;*/ \
        } \
    } \
    while (0)

#define DDR_CTR_W32_BASE(BASE, INDEX, ADDR, VAL) \
    writel((u32)(VAL), (void __iomem *)(BASE) + ljm_gpu_hardware.reg_ddr_ctr_offset[INDEX] + (ADDR))
#define DDR_CTR_CHECK_REG(BASE, INDEX, ADDR, VAL, MASK) \
    do { \
        int _j; \
        for (_j = 0; _j < 0xFF; _j++) { \
            if ((DDR_CTR_R32_BASE(BASE, INDEX, ADDR) & MASK) == VAL) { \
                break; \
            } \
            msleep(1); \
        } \
        if (_j == 0xFF) { \
            ERR_PRINT("DDR(%d): STAT error\n", __LINE__); \
            ret = -1; \
            /*goto OnError;*/ \
        } \
    } \
    while (0)

#define DDR_PHY_R32_BASE(BASE, INDEX, ADDR) \
        readl((void __iomem *)(BASE) + ljm_gpu_hardware.reg_ddr_phy_offset[INDEX] + (ADDR))
#define DDR_PHY_W32_BASE(BASE, INDEX, ADDR, VAL) \
    writel((u32)(VAL), (void __iomem *)(BASE) + ljm_gpu_hardware.reg_ddr_phy_offset[INDEX] + (ADDR))
#define DDR_PHY_CHECK_REG(BASE, INDEX, ADDR, VAL, MASK) \
    do { \
        int _j; \
        for (_j = 0; _j < 0xFF; _j++) { \
            if ((DDR_PHY_R32_BASE(BASE, INDEX, ADDR) & MASK) == VAL) { \
                break; \
            } \
            msleep(1); \
        } \
        if (_j == 0xFF) { \
            ERR_PRINT("DDR(%d): STAT error\n", __LINE__); \
            ret = -1; \
            /*goto OnError;*/ \
        } \
    } \
    while (0)
#endif /* ~DDR_ONLY_DUMP_REG */

struct ddrv2_ctrl_reg {
    u16 addr;
    u32 val;
};

struct ddrv2_phy_reg {
    u32 addr;
    u16 val;
};

struct ddrv2_ctrl_phy_reg {
    uint32_t addr;
    uint32_t val;
};

int ddrChannelMask = 0x1;
module_param(ddrChannelMask, int, 0644);
MODULE_PARM_DESC(ddrChannelMask, "video ddr channel mask, 0x1: ch 0; 0x1: ch 1; 0x3: ch 1&2; 0xc: ch 3&4; 0xf: ch 1&2&3&4");

int ddrFreqId = 0x4;
module_param(ddrFreqId, int, 0644);
MODULE_PARM_DESC(ddrFreqId, "video ddr frequency id, 0: 800Mbps; 1: 1066Mbps; 2: 1600Mbps; 3: 2400Mbps; 4: 2666Mbps; 5: 2933Mbps; 6: 3200Mbps");

enum ljm_ddr_freq_id {
    LJM_DDR_FREQ_800M  = 0x0,
    LJM_DDR_FREQ_1066M = 0x1,
    LJM_DDR_FREQ_1600M = 0x2,
    LJM_DDR_FREQ_2400M = 0x3,
    LJM_DDR_FREQ_2666M = 0x4,
    LJM_DDR_FREQ_2933M = 0x5,
    LJM_DDR_FREQ_3200M = 0x6,
};

/*
  ljm_ddr_reg_blk_id 是block type的值，是register group 的简称
  该值用于地址寻址，不可任意更改
  参考phy_pub_databook中Register Address Overview章节
*/
enum ljm_ddr_reg_blk_id {
    LJM_DDR_ANIB = 0x0,
    LJM_DDR_DBYTE = 0x1,
    LJM_DDR_MASTER = 0x2,
    LJM_DDR_ACSM = 0x4,
    LJM_DDR_UCTL_MEMORY = 0x5,
    LJM_DDR_PPGC = 0x7,
    LJM_DDR_INITENG = 0x9,
    LJM_DDR_PUB = 0xC,
    LJM_DDR_APBONLY = 0xD,
};

/*
  ljm_ddr_train_stage_id是从消息邮箱读取firmware返回的消息块
  此处值参考UctWriteOnlyShadowValue配置，符合协议,此处值不可更改
*/
enum ljm_ddr_train_stage_id {
    LJM_DDR_INIT_END = 0x00,
    LJM_DDR_FINE_WR_LVL = 0x01,
    LJM_DDR_RD_EN_TRAIN = 0x02,
    LJM_DDR_RD_DELAY_CENTER = 0x03,
    LJM_DDR_WR_DELAY_CENTER = 0x04,
    LJM_DDR_2D_RD = 0x05,
    LJM_DDR_2D_WR = 0x06,
    LJM_DDR_TRAIN_RUN_SUCC = 0x07,
    LJM_DDR_MAX_RD_LATENCY = 0x09,
    LJM_DDR_RD_DQ_DESKEW = 0x0A,
    LJM_DDR_ALL_DB_TRAIN = 0x0C,
    LJM_DDR_CA_TRAIN  = 0x0D,
    LJM_DDR_MPR_RD_DELAY = 0xFD,
    LJM_DDR_COARSE_WR_LVL = 0xFE,
    LJM_DDR_TRAIN_RUN_FAIL = 0xFF
};

u32 ljm_ddr_id_freq_clk[7][3] = {
    {0, 800,  0x00b10241},
    {1, 1066, 0x004E0043},
    {2, 1600, 0x00b10231},
    {3, 2400, 0x01090231},
    {4, 2666, 0x01270231},
    {5, 2933, 0x01450231},
    {6, 3200, 0x00af0221},
};

struct ljm_ddr_phy_initeng {
    uint32_t addr;
    uint32_t val;
};


int ljm_ddrv2_init_v2(int card, void *mmio_base, unsigned long *pddr_size, bool is_flash, unsigned long flags);
int ljmSpiReadBuffer(void *mmio_base, char *buf, int addr, int size);


/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_ctr
 *  功能描述: ddrv2控制器寄存器配置
 *  输入参数: void *mmio_base   MMIO基地址
 *            int channel_mask  DDR通道mask位
 *            u32 addr          寄存器地址
 *            u32 val           寄存器值
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月13日
 *            作    者: huym, liuty
 *            修改内容: 新生成函数
 *         2. 日    期: 2023年10月31日
 *            作    者: huym
 *            修改内容: 增加ljm_ddr_funcs函数的调用
 *
 *****************************************************************************/
int ljm_ddrv2_set_ctr(void *mmio_base, int channel_mask, u32 addr, u32 val)
{
    int ret = 0;
    int i;

    for (i = 0; i < ljm_gpu_hardware.max_ddr_channels; i++) {
        if (channel_mask & (0x1 << i)) {
            if (addr & 0x80000000) {
                DDR_CTR_CHECK_REG(mmio_base, i, addr - 0x80000000, val, 0xffffffff);
#if DDR_DUMP_REG
                PRINT("%08x %08x R %08X 32\n", addr - 0x80000000 + ljm_gpu_hardware.reg_ddr_ctr_offset[i], val, 0xffffffff);
#endif
                if (ret < 0) {
                    ERR_PRINT("%d: ddr channel %d check [%08x] == 0x%08x timeout\n", __LINE__,
                              i, addr - 0x80000000 + ljm_gpu_hardware.reg_ddr_ctr_offset[i], val);
                }
            }
            else {
                DDR_CTR_W32_BASE(mmio_base, i, addr, val);
#if DDR_DUMP_REG
                PRINT("%08x %08x W FFFFFFFF 32\n", addr + ljm_gpu_hardware.reg_ddr_ctr_offset[i], val);
#endif
            }

            if (ret) {
                /* check status timeout */
                return ret;
            }
        }
    }

    return ret;
}

#if 0
/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_monitor
 *  功能描述: monitor寄存器配置
 *  输入参数: void *mmio_base   MMIO基地址
 *            u32 addr          寄存器地址
 *            u32 val           寄存器值
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2024年4月8日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *****************************************************************************/
void ljm_ddrv2_set_monitor(void *mmio_base, u32 addr, u32 val)
{
    REG_W32_BASE(mmio_base, (0x309000 + addr), val);
}
#endif

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_phy
 *  功能描述: ddrv2 PHY寄存器配置
 *  输入参数: void *mmio_base   MMIO基地址
 *            int channel_mask  DDR通道mask位
 *            u32 addr          寄存器地址
 *            u32 val           寄存器值
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月13日
 *            作    者: huym, liuty
 *            修改内容: 新生成函数
 *         2. 日    期: 2023年10月31日
 *            作    者: huym
 *            修改内容: 增加ljm_ddr_funcs函数的调用
 *
 *****************************************************************************/
int ljm_ddrv2_set_phy(void *mmio_base, int channel_mask, u32 addr, u32 val)
{
    int ret = 0;
    int i;

    for (i = 0; i < ljm_gpu_hardware.max_ddr_channels; i++) {
        if (channel_mask & (0x1<<i)) {
            if (addr & 0x80000000) {
                DDR_PHY_CHECK_REG(mmio_base, i, addr - 0x80000000, val, 0xffffffff);
#if DDR_DUMP_REG
                PRINT("%08x %08x R %08X 32\n", addr - 0x80000000 + ljm_gpu_hardware.reg_ddr_phy_offset[i], val, 0xffffffff);
#endif
                if (ret < 0) {
                    ERR_PRINT("%d: check [%08x] == 0x%08x timeout\n", __LINE__, ljm_gpu_hardware.reg_ddr_phy_offset[i] + addr - 0x80000000, val);
                }
            }
            else {
                DDR_PHY_W32_BASE(mmio_base, i, addr, val);
#if DDR_DUMP_REG
                PRINT("%08x %08x W FFFFFFFF 32\n", addr + ljm_gpu_hardware.reg_ddr_phy_offset[i], val);
#endif
            }
        }
    }

    return ret;
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_misc
 *  功能描述: ddrv2控制器寄存器配置
 *  输入参数: void *mmio_base   MMIO基地址
 *            uint8_t channel_mask  DDR通道mask位
 *            u32 val           寄存器值
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月13日
 *            作    者: huym, liuty
 *            修改内容: 新生成函数
 *         2. 日    期: 2023年10月31日
 *            作    者: huym
 *            修改内容: 增加ljm_ddr_funcs函数的调用
 *****************************************************************************/
int ljm_ddrv2_set_misc(void *mmio_base, uint8_t channel_mask, u32 val)
{
    int i;

    for (i = 0; i < ljm_gpu_hardware.max_ddr_channels; i++) {
        if (channel_mask & (0x1<<i)) {
            LJM_MISC_W32_BASE(mmio_base, MISC_DDR_PHY_ADDR0 + 4 * i, val);
#if DDR_DUMP_REG
            PRINT("%08x %08x W FFFFFFFF 32\n", ljm_gpu_hardware.reg_misc_offset + MISC_DDR_PHY_ADDR0 + 4 * i, val);
#endif
        }
    }
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_other
 *  功能描述: ddrv2控制器寄存器配置
 *  输入参数: void *mmio_base   MMIO基地址
 *            int channel_mask  DDR通道mask位
 *            u32 addr          寄存器地址
 *            u16 val           寄存器值
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月13日
 *            作    者:  liuty
 *            修改内容: 新生成函数
 *         2. 日    期: 2023年10月31日
 *            作    者: huym
 *            修改内容: 增加ljm_ddr_funcs函数的调用
 *
 *****************************************************************************/
int ljm_ddrv2_set_other(void *mmio_base, uint8_t channel_mask, u32 addr, u32 val)
{
    int ret = 0;

    if (addr & 0x80000000) {
        u32 mask = 0xffffffff;

        if ((addr - 0x80000000) == ljm_gpu_hardware.reg_cru_offset + CRU_STATUS) {
            mask = 0x08000000;  /* only check ddr0pll locked */
        }

        REG_CHECK_REG(mmio_base,
            addr - 0x80000000,
            val,
            mask);
#if DDR_DUMP_REG
        PRINT("%08x %08x R %08X 32\n", addr, val, 0xffffffff);
#endif
        if (ret < 0) {
            ERR_PRINT("%d: check [%08x] == 0x%08x timeout\n", __LINE__, addr - 0x80000000, val);
        }
    }
    else {
        REG_W32_BASE(mmio_base, addr, val);
#if DDR_DUMP_REG
        PRINT("%08x %08x W FFFFFFFF 32\n", addr, val);
#endif
        udelay(70); /* wait 70us for DDRPLL locked */

        if (addr == ljm_gpu_hardware.reg_cru_offset + CRU_DDRPLL) {
            u32 pll_lock_status_mask = 0x0;
            u32 pll_lock_status = 0;
            int i;

            for (i = 0; i < ljm_gpu_hardware.max_ddr_channels; i++) {
                if (channel_mask & (0x1 << i)) {
                    pll_lock_status_mask |=  (DDR_CRU_STATUS_STA_DDR0PLL_LOCK_MASK >> i);
                    pll_lock_status |= (DDR_CRU_STATUS_STA_DDR0PLL_LOCK_LOCKED >> i);
                }
            }

            REG_CHECK_REG(mmio_base,
                          ljm_gpu_hardware.reg_cru_offset + CRU_STATUS,
                          pll_lock_status,
                          pll_lock_status_mask);
        }
    }
    return ret;
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_ctr_misc_phy
 *  功能描述: ddrv2控制器寄存器配置
 *  输入参数: void *mmio_base   MMIO基地址
 *            int channel_mask  DDR通道mask位
 *            u32 addr          寄存器地址
 *            u16 val           寄存器值
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月13日
 *            作    者: huym, liuty
 *            修改内容: 新生成函数
 *         2. 日    期: 2023年10月31日
 *            作    者: huym
 *            修改内容: 增加ljm_ddr_funcs函数的调用
 *
 *****************************************************************************/
int ljm_ddrv2_set_ctr_misc_phy(void *mmio_base, int channel_mask, u32 addr, u32 val)
{
    int ret = 0;

    if ((addr & 0x7fffffff) >= ljm_gpu_hardware.reg_ddr_ctr_offset[0]
        && (addr & 0x7fffffff) < ljm_gpu_hardware.reg_ddr_ctr_offset[0] + REG_DDR_CTRL_SIZE) {
        ret = ljm_ddrv2_set_ctr(mmio_base, channel_mask, addr - ljm_gpu_hardware.reg_ddr_ctr_offset[0], val);
        if (ret) {
            /* print err infomation */
            return ret;
        }
        return 0;
    }

    if ((addr & 0x7fffffff) >= ljm_gpu_hardware.reg_ddr_phy_offset[0]
        && (addr & 0x7fffffff) < ljm_gpu_hardware.reg_ddr_phy_offset[0] + REG_DDR_PHY_SIZE) {
        ret = ljm_ddrv2_set_phy(mmio_base, channel_mask, addr - ljm_gpu_hardware.reg_ddr_phy_offset[0], val);
        if (ret) {
            /* print err infomation */
            return ret;
        }
        return 0;
    }

    if (addr == ljm_gpu_hardware.reg_misc_offset + MISC_DDR_PHY_ADDR0) {
        ret = ljm_ddrv2_set_misc(mmio_base, channel_mask, val);
        if (ret) {
            /* print err infomation */
            return ret;
        }
        return 0;
    }

    ljm_ddrv2_set_other(mmio_base, channel_mask, addr, val);

    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_is_init
 *  功能描述: 判断ddr4控制器是否已经初始化
 *  输入参数: void *mmio_base 寄存器空间基地址
 *            int channel     通道序号
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年3月18日
 *            作    者: huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_ddrv2_is_init(void *mmio_base, int channel)
{
    HEADER_ARG("mmio_base=%p channel=%d", mmio_base, channel);

    if ((DDR_CTR_R32_BASE(mmio_base, channel, DDR_STAT) & m_STAT) == 0x1) {
        FOOTER_ARG("%s", "true");
        return true; /* VRAM controller has been initialized in BIOS */
    } else {
        FOOTER_ARG("%s", "false");
        return false;
    }
}

/*
 * ddr configuration definition
 */
typedef struct _ljm_ddr_cfg
{
    uint16_t vendor_id;         /* 0x0: 0x0709 */
    uint16_t device_id;         /* 0x2: 0x0201 */
    uint16_t block0_offset;     /* 0x4: offset of block 0 from this header */
    uint16_t block0_size;       /* 0x6: size of block 0 */
    uint16_t block0_crc;        /* 0x8: crc of block 0 */
    uint16_t block0_ext_offset; /* 0xA: offset of extended image for block 0 if ljm_board_info.version >= 0x140;
                                        ddr bandwidth, such as 0x3200 for 3200Mbps, if ljm_board_info.version < 0x140 */
    uint16_t block1_offset;     /* 0xC: offset of block 1 from this header */
    uint16_t block1_size;       /* 0xE: size of block 1 */
    uint16_t block1_crc;        /* 0x10: crc of block 1 */
    uint16_t block1_ext_offset; /* 0x12: offset of extended image for block 1 if ljm_board_info.version >= 0x140, otherwise reserved */
    uint16_t block2_offset;     /* 0x14: offset of block 2 from this header */
    uint16_t block2_size;       /* 0x16: size of block 2 */
    uint16_t block2_crc;        /* 0x18: crc of block 2 */
    uint16_t block2_ext_offset; /* 0x1A: offset of extended image for block 2 if ljm_board_info.version >= 0x140, otherwise reserved */
    uint16_t version;           /* 0x1C: 0x0200 => version 2.0 */
    uint16_t  crc;              /* 0x1E: 16-bit checksum for above items */
} __attribute__((packed)) ljm_ddr_cfg;

typedef struct _ljm_ddr_cfg_ext_image_header
{
    uint16_t vendor_id;     /* 0x0: 0x0709 */
    uint16_t image_type;    /* 0x2: ddr_image=0x5244 */
    uint16_t channel_mask;  /* 0x4: channel mask of ddr, 0 indicates ignore channel,
                             *      bit0 indicates channel0, bit1 indicates channel1, and so on */
    uint16_t data_type;     /* 0x6: edid_data=0,dpcd_data=1 */
    uint16_t data_offest;   /* 0x8: offest of data from the image start address */
    uint16_t data_size;     /* 0xa: size of image data,exclude image header */
    uint16_t crc;           /* 0xc: 16-bit checksum for image data */
    uint16_t image_offest;  /* 0xe: offest address of the next image, 0:no next image */
    uint8_t reserved[48];   /* 0x10~0x3f: reserved=oxff */
} __attribute__((packed)) ljm_ddr_cfg_ext_image_header;

//block 0 iccm     block 1 dccm
typedef struct _ljm_ddr_cfg_v2
{
    uint16_t vendor_id;     /* 0x0: 0x0709 */
    uint16_t device_id;     /* 0x2: 0x0201 */
    uint16_t block0_offset; /* 0x4: offset of block 0 from this header */
    uint16_t block0_size;   /* 0x6: size of block 0 */
    uint16_t block0_crc;    /* 0x8: crc of block 0 */
    uint16_t reserved0;     /* 0xA */
    uint16_t block1_offset; /* 0xC: offset of block 1 from this header */
    uint16_t block1_size;   /* 0xE: size of block 1 */
    uint16_t block1_crc;    /* 0x10: crc of block 1 */
    uint16_t reserved1;     /* 0x12 */
    uint16_t version;       /* 0x1C: xx.yy */
    uint16_t crc;           /* 0x1E: 16-bit checksum for above items */
} __attribute__((packed)) ljm_ddr_cfg_v2;


#define SPI_BOADR_INFO_OFFSET   0x70000

/*****************************************************************************
 *  函 数 名: calc_crc16
 *  功能描述: 计算16位CRC校验码
 *  输入参数: uint16_t *pv16
 *            int count
 *  输出参数: 无
 *  返 回 值: static uint16_t
 *
 *  修改历史:
 *         1. 日    期: 2022年8月10日
 *            作    者: huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static uint16_t calc_crc16(uint16_t *pv16, int count)
{
    int i;
    uint16_t sum = 0;

    for (i=0; i<count; i++)
        sum += *pv16++;

    return (0 - sum);
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_cfg
 *  功能描述: 配置DDR控制器和PHY寄存器
 *  输入参数: void *mmio_base
 *            uint8_t channel_mask
 *            struct ddrv2_ctrl_phy_reg *cfg
 *            int count
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年8月10日
 *            作    者: liuty
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrv2_cfg(void *mmio_base, uint8_t channel_mask, struct ddrv2_ctrl_phy_reg *cfg, int count)
{
    int i;

    HEADER_ARG("mmio_base=%p channel_mask=%x cfg=%p count=%d", mmio_base, channel_mask, cfg, count);

    for (i=0; i < count; i++) {
        if (cfg->addr == 0xffffffff || cfg->addr == 0x0) {
            ERR_PRINT("error: i=%d address=%08x", i, cfg->addr);
            break;
        }
        ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, cfg->addr, cfg->val);
        cfg++;
    }

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_imem
 *  功能描述: 配置DDR IMEM
 *  输入参数: void *mmio_base
 *            uint8_t channel_mask
 *            uint16_t *pval
 *            int count
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年8月10日
 *            作    者: liuty
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrv2_imem(void *mmio_base, uint8_t channel_mask, uint16_t *pval, int count)
{
    int i;

    HEADER_ARG("mmio_base=%p channel_mask=%x pval=%p count=%d", mmio_base, channel_mask, pval, count);

    for (i = 0; i < count; ++i) {
        ljm_ddrv2_set_phy(mmio_base, channel_mask, 4 * i, (u32)(*pval));
        pval++;
    }

    FOOTER();
}

ljm_board_info g_board_info[LJM_MAX_CARDS_NUM] = {};

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_get_board_info
 *  功能描述: 获取板载信息
 *  输入参数: void *mmio_base
 *  输出参数: 无
 *  返 回 值: ljm_board_info *
 *
 *  修改历史:
 *         1. 日    期: 2022年8月13日
 *            作    者: liuty, huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
ljm_board_info *ljm_ddrv2_get_board_info(int card, void *mmio_base)
{
    int ret = 0;
    uint16_t crc;

    if (g_board_info[card].vendor_id == LJM_VENDOR_ID) {
        return &g_board_info[card];
    } else {
        /* read board info */
        ret = ljmSpiReadBuffer(mmio_base, (char*)&g_board_info[card], SPI_BOADR_INFO_OFFSET, sizeof(g_board_info[card]));
        if (ret) {
            ERR_PRINT("read board info from flash failed.\n");
            g_board_info[card].vendor_id = 0;
            g_board_info[card].extra_reg_count = 0;
            return NULL;
        }

        /** check vendor id */
        if (g_board_info[card].vendor_id != LJM_VENDOR_ID) {
            WARN_PRINT("board info id error.");
            dump_bytes((const unsigned char *)&g_board_info[card], sizeof(g_board_info[card]), "board info");
            g_board_info[card].vendor_id = 0;
            g_board_info[card].extra_reg_count = 0;
            return NULL;
        }

        /** check crc */
        crc = calc_crc16((uint16_t *)&g_board_info[card], (sizeof(g_board_info[card]))/2 - 1);
        if (crc != g_board_info[card].crc) {
            ERR_PRINT("board info crc error.");
            dump_bytes((const unsigned char *)&g_board_info[card], sizeof(g_board_info[card]), "board info");
            g_board_info[card].vendor_id = 0;
            g_board_info[card].extra_reg_count = 0;
            return NULL;
        }

        /** dump the correct boardinfo */
        dump_bytes((const unsigned char *)&g_board_info[card], sizeof(g_board_info[card]), "board info");

        ddrChannelMask = g_board_info[card].ddr_ch_mask;
        ddrChannelMask |= g_board_info[card].ddr_adv_ch_mask;

        if (g_board_info[card].version >= 0x200)    /* ddr_freq_id is supported from version 2.00 */
            ddrFreqId = g_board_info[card].ddr_freq_id;

        return &g_board_info[card];
    }
}

/*****************************************************************************
 *  函 数 名: ljm_get_board_info
 *  功能描述: 返回板载信息
 *  输入参数: int card
 *  输出参数: 无
 *  返 回 值: ljm_board_info *
 *
 *  修改历史:
 *         1. 日    期: 2023年4月20日
 *            作    者: Xusn
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

ljm_board_info *ljm_get_board_info(int card)
{
    return &g_board_info[card];
}

/*****************************************************************************
 *  函 数 名: ljm_get_vram_size_from_boardinfo
 *  功能描述: 从板载信息获取显存大小
 *  输入参数: int card  板卡编号
 *  输出参数: 无
 *  返 回 值: unsigned long 显存大小(单位字节)
 *
 *  修改历史:
 *         1. 日    期: 2022年8月13日
 *            作    者: liuty, huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
unsigned long ljm_get_vram_size_from_boardinfo(int card)
{
    unsigned long size;

    if (g_board_info[card].vendor_id == LJM_VENDOR_ID) {
        uint8_t ddr_ch_mask, ddr_adv_ch_mask;
        int ddr_ch_cnt, ddr_adv_ch_cnt;
        int i;

        ddr_ch_mask = g_board_info[card].ddr_ch_mask;
        for (i=0, ddr_ch_cnt=0; i<8; i++) {
            if (ddr_ch_mask & (1U << i))
                ddr_ch_cnt++;
        }

        ddr_adv_ch_mask = ddr_ch_mask | g_board_info[card].ddr_adv_ch_mask;
        for (i=0, ddr_adv_ch_cnt=0; i<8; i++) {
            if (ddr_adv_ch_mask & (1U << i))
                ddr_adv_ch_cnt++;
        }
        size =  (0x10000000UL * (1UL << ((g_board_info[card].ddr_total_size) & 0xf)));
        if (ddr_ch_cnt > 0 && ddr_adv_ch_cnt / ddr_ch_cnt > 1)
            size *= ddr_adv_ch_cnt / ddr_ch_cnt;

        if (size < 0x80000000UL) {
            size = 0x80000000UL;
            WARN_PRINT("VRAM size is less than 2GB from board info, force to 2GB");
        }
        return size;
    } else {
        if (ddrChannelMask == 0x1 || ddrChannelMask == 0x2)
            return 0x80000000UL;   /* 2GB */
        else if (ddrChannelMask == 0x3 || ddrChannelMask == 0xc)
            return 0x100000000UL;  /* 4GB */
        else if (ddrChannelMask == 0xf)
            return 0x200000000UL;   /* 8GB */
        else
            return 0x80000000UL;   /* 2GB */
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_get_ddr_size
 *  功能描述: 获取板载DDR容量
 *  输入参数: void *mmio_base
 *  输出参数: 无
 *  返 回 值: size_t
 *
 *  修改历史:
 *         1. 日    期: 2022年8月13日
 *            作    者: liuty, huym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
unsigned long ljm_ddrv2_get_ddr_size(int card, void *mmio_base)
{
    ljm_ddrv2_get_board_info(card, mmio_base);

    return ljm_get_vram_size_from_boardinfo(card);
}

static bool ljm_ddr_blk_ext_img_header_valid(ljm_ddr_cfg_ext_image_header *header)
{
    if (!(header->vendor_id == IMAGE_HEAD_IDENTIFICATION &&
                header->image_type == IMAGE_TYPE_DDR)) {
        ERR_PRINT("ddr block ext img type error! image_type = 0x%x\n", header->image_type);
        return LJM_DDR_ERR_IMG_INVALID;
    }

    return LJM_DDR_SUCCESS;
}

static bool ljm_ddr_blk_ext_img_valid(ljm_ddr_cfg_ext_image_header *header, void *data)
{
    if (header->crc != calc_crc16((uint16_t *)data, (header->data_size + 1) / 2)) {
        ERR_PRINT("ddr block ext img crc error!\n");
        return LJM_DDR_ERR_CRC;
    }

    return LJM_DDR_SUCCESS;
}

int ljm_ddrv2_set_ext_img_cfg(void *mmio_base, uint32_t ddr_cfg_start_addr, uint32_t ext_img_offset)
{
    uint32_t i;
    ljm_ddr_cfg_ext_image_header ext_img_header;
    uint32_t size = 0;
    void *pdata = NULL;
    struct ddrv2_ctrl_phy_reg *ext_regs = NULL;
    uint32_t count = 0;
    enum LjmDdrErrCode ret = LJM_DDR_SUCCESS;

    HEADER_ARG("mmio_base=%p, ddr_cfg_start_addr = 0x%x, ext_img_offset=0x%x\n",
               mmio_base, ddr_cfg_start_addr, ext_img_offset);

    if (0 == ext_img_offset) {
        DEBUG_PRINT("no ext img!\n");
        FOOTER();
        return  LJM_DDR_SUCCESS;
    }

    if (ext_img_offset > 0x10000) {
        ERR_PRINT("error ext img offset, ext_img_offset = %d\n", ext_img_offset);
        ret = LJM_DDR_ERR_EXT_IMG_SIZE;
        goto OnError;
    }

    memset((void *)&ext_img_header, 0, sizeof(ext_img_header));

    /* read ext img header*/
    ret = ljmSpiReadBuffer(mmio_base, (char *)&ext_img_header,
                            ddr_cfg_start_addr + ext_img_offset, sizeof(ext_img_header));

    DEBUG_PRINT("read ext_img_header ok!\n");

    ret = ljm_ddr_blk_ext_img_header_valid(&ext_img_header);
    if (ret != LJM_DDR_SUCCESS) {
        goto OnError;
    }

    DEBUG_PRINT("ljm_ddr_blk_ext_img_header_valid ok!\n");

    size = ext_img_header.data_size;
    pdata = kzalloc(size, GFP_KERNEL);
    if (!((pdata != NULL) && (size > 0 && size < 0x8000)))
    {
        ERR_PRINT("out of memory: %d\n", size);
        ret = LJM_DDR_ERR_ALLOC;
        goto OnError;
    }

    /* read ext img data*/
    ret = ljmSpiReadBuffer(mmio_base,
                            (char *)pdata,
                            ddr_cfg_start_addr + ext_img_offset + ext_img_header.data_offest,
                            size);

    ret = ljm_ddr_blk_ext_img_valid(&ext_img_header, pdata);
    if (ret != LJM_DDR_SUCCESS) {
        ERR_PRINT("ddr block ext img invalid!!!\n");
        dump_bytes((const unsigned char *)pdata, size, "ddr block ext image");
        goto OnError;
    }

    ext_regs = (struct ddrv2_ctrl_phy_reg *)pdata;

    count = size / (sizeof(uint32_t) * 2);

    for (i = 0; i < count; i++)
    {
        if (ext_regs->addr == 0xffffffff || ext_regs->addr == 0x0)
        {
            ERR_PRINT("error: i = %d address = %08x\n", i, ext_regs->addr);
            break;
        }
        ljm_ddrv2_set_other(mmio_base, ddrChannelMask, ext_regs->addr, ext_regs->val);
        ext_regs++;
    }

    if (pdata) {
        kfree(pdata);
    }

    FOOTER_ARG("%s", "ddr block0 ext img cfg ok!");
    return LJM_DDR_SUCCESS;

OnError:
    if (pdata) {
        kfree(pdata);
    }
    ERR_PRINT("ddr init fail from ddr cfg block ext image, offset = 0x%x\n",
                ext_img_offset);

    FOOTER_ARG("ret = %d\n", ret);
    return ret;
}


/*****************************************************************************
 *  函 数 名: ljm_ddrv2_init
 *  功能描述: initialize ddrv2 controller and phy
 *  输入参数: int card  显卡序号
 *            void *mmio_base   寄存器空间基地址
 *            unsigned long *pddr_size  返回显存的容量
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年3月18日
 *            作    者: liuty
 *            修改内容: 新生成函数
 *         2. 日    期: 2022年8月10日
 *            作    者: liuty, huym
 *            修改内容: 根据SPI板级信息进行DDR配置
 *         3. 日    期: 2023年10月31日
 *            作    者: huym
 *            修改内容: 适配GP202
 *
 *****************************************************************************/
int ljm_ddrv2_init(int card, void *mmio_base, unsigned long *pddr_size)
{
    int ret = 0;
    struct ddrv2_ctrl_phy_reg *extra_regs = NULL;
    ljm_ddr_cfg ddr_cfg;
    uint16_t crc = 0;
    void *pdata = NULL;
    int size = 0;
    int i;

    INFO_PRINT("use old interface to init ddr\n");

    /* check and read extra configuration for registers */
    if (g_board_info[card].extra_reg_count > 0) {
        extra_regs = kzalloc(g_board_info[card].extra_reg_count * sizeof(struct ddrv2_ctrl_phy_reg), GFP_KERNEL);
        ret = ljmSpiReadBuffer(mmio_base,
            (char*)extra_regs,
            SPI_BOADR_INFO_OFFSET + sizeof(g_board_info[card]),
            g_board_info[card].extra_reg_count * sizeof(struct ddrv2_ctrl_phy_reg)
            );
        if (ret) {  /* read extra registers error */
            ERR_PRINT("ljm_ddr_init: read extra registers info error\n");
            g_board_info[card].extra_reg_count = 0;

            kfree(extra_regs);
            extra_regs = NULL;
        } else {
            INFO_PRINT("find extra regs: %d", g_board_info[card].extra_reg_count);
            dump_bytes((const unsigned char *)extra_regs, g_board_info[card].extra_reg_count * sizeof(struct ddrv2_ctrl_phy_reg), "extra registers");
        }
    }

    /* read ddr cfg */
    ret = ljmSpiReadBuffer(mmio_base, (char*)&ddr_cfg, g_board_info[card].ddr_cfg_start, sizeof(ddr_cfg));
    if (ret) {
        ERR_PRINT("read ddr cfg from SPI flash failed.");
        return -1;
    }
    crc = calc_crc16((uint16_t *)&ddr_cfg, sizeof(ddr_cfg)/2 - 1);
    if (crc != ddr_cfg.crc) {
        ERR_PRINT("ddr cfg crc error.");
        dump_bytes((const unsigned char *)&ddr_cfg, sizeof(ddr_cfg), "ddr configuration");
        return -2;
    }

    /* 1: configure ddr cfg block 0 */
    size = ddr_cfg.block0_size;
    pdata = kzalloc(size, GFP_KERNEL);
    if (!pdata) {
        ERR_PRINT("out of memory: %d", size);
        return ret;
    }
    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_cfg_start + ddr_cfg.block0_offset, ddr_cfg.block0_size);
    if (ret) {
        ERR_PRINT("read ddr cfg block 0 from SPI flash failed.");
        kfree(pdata);
        return -3;
    }
    crc = calc_crc16((uint16_t *)pdata, ddr_cfg.block0_size/2);
    if (crc != ddr_cfg.block0_crc) {
        ERR_PRINT("ddr cfg block 0 crc error.");
        dump_bytes((const unsigned char *)pdata, ddr_cfg.block0_size, "ddr configuration block 0");
        kfree(pdata);
        return -4;
    }
    ljm_ddrv2_cfg(mmio_base, ddrChannelMask, (struct ddrv2_ctrl_phy_reg *)pdata, ddr_cfg.block0_size/8);

    if (g_board_info[card].version >= 0x140) {
         /* 2: configure ddr cfg block 0 ext image */
        ret = ljm_ddrv2_set_ext_img_cfg(mmio_base, g_board_info[card].ddr_cfg_start, ddr_cfg.block0_ext_offset);
        if (ret != LJM_DDR_SUCCESS) {
            ERR_PRINT("block0 ext img cfg error, errcode = %d\n", ret);
        }
    }

    /* 3: load imem0 */
    if (size < IMEM_SIZE) {
        kfree(pdata);
        size = IMEM_SIZE;
        pdata = kzalloc(size, GFP_KERNEL);
        if (!pdata) {
            ERR_PRINT("out of memory: %d", size);
            return -5;
        }
    }
    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_imem_start, IMEM_SIZE);
    if (ret) {
        ERR_PRINT("read ddr imem0 from SPI flash failed.");
        kfree(pdata);
        return -6;
    }
    ljm_ddrv2_set_misc(mmio_base, ddrChannelMask, 0x5);
    ljm_ddrv2_imem(mmio_base, ddrChannelMask, (uint16_t *)pdata, IMEM_SIZE/2);

    /* 4: configure ddr cfg block 1 */
    if (size < ddr_cfg.block1_size) {
        kfree(pdata);
        size = ddr_cfg.block1_size;
        pdata = kzalloc(size, GFP_KERNEL);
        if (!pdata) {
            ERR_PRINT("out of memory: %d", size);
            return -7;
        }
    }
    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_cfg_start + ddr_cfg.block1_offset, ddr_cfg.block1_size);
    if (ret) {
        ERR_PRINT("read ddr cfg block 1 from SPI flash failed.");
        kfree(pdata);
        return -8;
    }
    crc = calc_crc16((uint16_t *)pdata, ddr_cfg.block1_size/2);
    if (crc != ddr_cfg.block1_crc) {
        ERR_PRINT("ddr cfg block 1 crc error.");
        dump_bytes((const unsigned char *)pdata, ddr_cfg.block1_size, "ddr configuration block 1");
        kfree(pdata);
        return -9;
    }
    for (i = 0; i < ljm_gpu_hardware.max_ddr_channels; i++) {
        u8 cur_mask = ddrChannelMask & (1u << i);
        if (cur_mask)
            ljm_ddrv2_cfg(mmio_base, cur_mask, (struct ddrv2_ctrl_phy_reg *)pdata, ddr_cfg.block1_size/8);
    }

    if (g_board_info[card].version >= 0x140) {
        ret = ljm_ddrv2_set_ext_img_cfg(mmio_base, g_board_info[card].ddr_cfg_start,
                                        ddr_cfg.block1_ext_offset);
        if (ret != LJM_DDR_SUCCESS) {
            ERR_PRINT("block1 ext img cfg error, errcode = %d\n", ret);
        }
    }
    DEBUG_PRINT("1D training is finished\n");

    /* 2D training */
    if (ddr_cfg.block2_size > 0x0) {
        /* 5: load imem1 */
        ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_imem_start + IMEM_SIZE, IMEM_SIZE);
        if (ret) {
            ERR_PRINT("read ddr imem0 from SPI flash failed.");
            kfree(pdata);
            return -10;
        }
        ljm_ddrv2_set_misc(mmio_base, ddrChannelMask, 0x5);
        ljm_ddrv2_imem(mmio_base, ddrChannelMask, (uint16_t *)pdata, IMEM_SIZE/2);

        /* 6: configure ddr cfg block 2 */
        if (size < ddr_cfg.block2_size) {
            kfree(pdata);
            size = ddr_cfg.block2_size;
            pdata = kzalloc(size, GFP_KERNEL);
            if (!pdata) {
                ERR_PRINT("out of memory: %d", size);
                return -11;
            }
        }
        ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_cfg_start + ddr_cfg.block2_offset, ddr_cfg.block2_size);
        if (ret) {
            ERR_PRINT("read ddr cfg block 2 from SPI flash failed.");
            kfree(pdata);
            return -12;
        }
        crc = calc_crc16((uint16_t *)pdata, ddr_cfg.block2_size/2);
        if (crc != ddr_cfg.block2_crc) {
            ERR_PRINT("ddr cfg block 2 crc error.");
            dump_bytes((const unsigned char *)pdata, ddr_cfg.block2_size, "ddr configuration block 2");
            kfree(pdata);
            return -13;
        }
        ljm_ddrv2_cfg(mmio_base, ddrChannelMask, (struct ddrv2_ctrl_phy_reg *)pdata, ddr_cfg.block2_size/8);

        if (g_board_info[card].version >= 0x140) {
            ret = ljm_ddrv2_set_ext_img_cfg(mmio_base, g_board_info[card].ddr_cfg_start,
                                            ddr_cfg.block2_ext_offset);
            if (ret != LJM_DDR_SUCCESS) {
                ERR_PRINT("block2 ext img cfg error, errcode = %d\n", ret);
            }
        }
        DEBUG_PRINT("2D training is finished\n");
    }

    if (pdata) {
        kfree(pdata);
    }

    /* 7: configure extra registers */
    if (g_board_info[card].extra_reg_count > 0) {
        ljm_ddrv2_cfg(mmio_base,
            ddrChannelMask,
            (struct ddrv2_ctrl_phy_reg *)extra_regs,
            g_board_info[card].extra_reg_count
            );

        if (extra_regs) {
            kfree(extra_regs);
            extra_regs = NULL;
        }
    }

    if (pddr_size) {
        *pddr_size = ljm_get_vram_size_from_boardinfo(card);
        DEBUG_PRINT("found ddr size: %zuB from board info (%x)", (*pddr_size), g_board_info[card].ddr_total_size);
    }

    return 0;
}

#if 0
/*****************************************************************************
 *  函 数 名: ljm_ddrv2_set_urgent_qos
 *  功能描述: 设置DDR通道的URGENT和QOS
 *  输入参数: void *mmio_base   寄存器空间基地址
 *            int channel_mask  显存通道使能位
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2023年10月17日
 *            作    者: huym
 *            修改内容: 创建函数，提升CRTC访问显存优先级到最高。
 *                      测试发现0x003c06a4寄存器写入失败，待完善。
  *****************************************************************************/
void ljm_ddrv2_set_urgent_qos(void *mmio_base, int channel_mask)
{
    /** DDR port 2: Port disable, HIF and DQ diable */
    ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, 0x003c0304, 0x00000003);    /**< HifDis, DqDis */
    ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, 0x003c06a0, 0x00000000);    /**< PortDis */

    /** set CRTC and ZoomVideo urgent value */
    WREG32_BASE(mmio_base, 0x309038, 0x0fffffff);   /**< VGA_H0-11 and VGA_N0-3 urgent enable */
    /** DDR port 2: enable DDR Read Port urgent */
    ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, 0x003c0614,
        RREG32_BASE(mmio_base, 0x003c0614) | (0x1U<<13));

    /** set CRTC and ZoomVideo Qos value */
    WREG32_BASE(mmio_base, 0x309020, 0x55555555);   /**< VGA_H0-7 qos: 5 */
    WREG32_BASE(mmio_base, 0x309024, 0x55555555);   /**< VGA_H8-11 and VGA_N0-3(ZV) qos: 5 */

    /** DDR port 2: region 0 to LPR, level1 to 4; region 1 to HPR */
    ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, 0x003c06a4, 0x00200004);

    /** HIF and DQ enable, Port enable */
    ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, 0x003c0304, 0x00000000);    /**< HifEn, DqEn */
    ljm_ddrv2_set_ctr_misc_phy(mmio_base, channel_mask, 0x003c06a0, 0x00000001);    /**< PortEn */
}
#endif

int ljm_ddrc_check_value(void *mmio_base, int channel_id, unsigned int addr, unsigned int value)
{
    int i=0;

    for (i=0; i<0xff; i++) {
        if (DDR_CTR_R32_BASE(mmio_base, channel_id, addr) == value) {
            return 0;
        } else {
            msleep(1);
        }
    }

    return -1;
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_ddrc_init_stat_wait(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_ddrc_init_stat_wait(void *mmio_base, int channel_mask)
{
    int ret = 0;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DBG1,  m_DIS_DQ);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL,  m_SELFREF_EN);
        ret = ljm_ddrc_check_value(mmio_base, channel_id, DDR_CTL_STAT, 0x0);
        if (ret < 0) {
            ERR_PRINT("%d: ljm_ddr_ddrc_init_stat_wait timeout! value=0x%x\n", __LINE__, DDR_CTR_R32_BASE(mmio_base, channel_id, DDR_CTL_STAT));
        }
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_for_ddr4_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_for_ddr4_precfg(void *mmio_base, int channel_mask)
{
    int channel_id;
    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_MSTR,          0x81040010);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_MRCTRL0,       0x40007030);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_MRCTRL1,       0x0001a789);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_MRCTRL2,       0xb9171b20);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DERATEEN,      0x00001606);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DERATEINT,     0x103ed8bb);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_MSTR2,         0x00000001);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DERATECTL,     0x00000001);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL,        0x000000a2);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PWRTMG,        0x00221306);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_HWLPCTL,       0x00ae0000);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_HWFFCCTL,      0x00000030);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_HWFFCEX_RANK1, 0x6a030004);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_scrub_check(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_scrub_check(void *mmio_base, int channel_mask)
{
    int ret = 0;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;

        DDR_CTR_W32_BASE(mmio_base, channel_id,  DDR_CTL_DBG1,               0x0);
        DDR_CTR_W32_BASE(mmio_base, channel_id,  DDR_CTL_PWRCTL,             0xa0);
        DDR_CTR_W32_BASE(mmio_base, channel_id,  DDR_CTL_DERATEEN,           0x1606); //derate
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_DERATEEN + 0x2000), 0x1606); //derate freq1
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_DERATEEN + 0x3000), 0x1606); //derate freq2
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_DERATEEN + 0x4000), 0x1606); //derate freq3
#if 0
        ERR_PRINT("%d: channel:%d  dbg1=0x%x, pwrctl=0x%x, derateen=0x%x 0x%x 0x%x 0x%x\n", __LINE__, channel_id, \
                DDR_CTR_R32_BASE(mmio_base, channel_id,  DDR_CTL_DBG1), DDR_CTR_R32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL), \
                DDR_CTR_R32_BASE(mmio_base, channel_id,  DDR_CTL_DERATEEN), DDR_CTR_R32_BASE(mmio_base, channel_id,  DDR_CTL_DERATEEN+0x2000), \
                DDR_CTR_R32_BASE(mmio_base, channel_id,  DDR_CTL_DERATEEN+0x3000),DDR_CTR_R32_BASE(mmio_base, channel_id,  DDR_CTL_DERATEEN+0x4000));
        //DDR_CTR_CHECK_REG(mmio_base, channel_id, DDR_CTL_SBRCTL, 0x00021e12, 0xffffffff);//ctl into init state
#endif
        ret = ljm_ddrc_check_value(mmio_base, channel_id, DDR_CTL_SBRCTL, 0x00021e12);
        if (ret < 0) {
            ERR_PRINT("%d: ljm_ddr_scrub_check timeout!\n", __LINE__);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_ctl_precfg_end(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_ctl_precfg_end(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        //0x0 (SwDone0): Program this register to 0 to enable quasi-dynamic programming
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_SWCTL,   0);
        //bit[0]:0x1 (DfiInitCompleteEn): The dfi_init_complete signal can be used to trigger SDRAM initialisation
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC, m_DISDYNADRTRI | m_DFI_INIT_COMPLETE_EN);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC, m_DISDYNADRTRI);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_rfsh_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_rfsh_precfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RFSHCTL0,  ddr_ctl_rfsh[freq_id][0]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RFSHCTL1,  ddr_ctl_rfsh[freq_id][1]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RFSHCTL3,  ddr_ctl_rfsh[freq_id][2]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RFSHTMG,   ddr_ctl_rfsh[freq_id][3]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RFSHTMG1,  ddr_ctl_rfsh[freq_id][4]);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_ecc_crc_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_ecc_crc_precfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ECCCFG0, 0x01027f10);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ECCCFG1, 0x000007b0);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ECCCCTL, 0x00000700);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ECCPOISONADDR0, 0x00000681);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ECCPOISONADDR1, 0x07023881);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_CRCPARCTL0, 0x00000000);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_CRCPARCTL1, 0x00001000);
    }
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_sdram_intreg_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_sdram_intreg_precfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;

        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT0,  ddr_ctl_intreg[freq_id][0]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT1,  ddr_ctl_intreg[freq_id][1]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT2,  ddr_ctl_intreg[freq_id][2]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT3,  ddr_ctl_intreg[freq_id][3]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT4,  ddr_ctl_intreg[freq_id][4]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT5,  ddr_ctl_intreg[freq_id][5]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT6,  ddr_ctl_intreg[freq_id][6]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_INT7,  ddr_ctl_intreg[freq_id][7]);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_dimm_rank_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_dimm_rank_precfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DIMMCTL, 0x20);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RANKCTL, 0xec7);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_dramtmg_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_dramtmg_precfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;

        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG0,  ddr_ctl_dramtmg[freq_id][0]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG1,  ddr_ctl_dramtmg[freq_id][1]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG2,  ddr_ctl_dramtmg[freq_id][2]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG3,  ddr_ctl_dramtmg[freq_id][3]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG4,  ddr_ctl_dramtmg[freq_id][4]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG5,  ddr_ctl_dramtmg[freq_id][5]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG6,  ddr_ctl_dramtmg[freq_id][6]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG7,  ddr_ctl_dramtmg[freq_id][7]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG8,  ddr_ctl_dramtmg[freq_id][8]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG9,  ddr_ctl_dramtmg[freq_id][9]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG10, ddr_ctl_dramtmg[freq_id][10]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG11, ddr_ctl_dramtmg[freq_id][11]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG12, ddr_ctl_dramtmg[freq_id][12]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG13, ddr_ctl_dramtmg[freq_id][13]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG14, ddr_ctl_dramtmg[freq_id][14]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG15, ddr_ctl_dramtmg[freq_id][15]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DRAMTMG17, ddr_ctl_dramtmg[freq_id][16]);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_zqctl_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_zqctl_precfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;

        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ZQCTL0, 0x01000040);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ZQCTL1, ddr_ctl_zqctl1[freq_id]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ZQCTL2, 0x00000000);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_dfi_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_dfi_precfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFITMG0, ddr_ctl_dfitmg[freq_id][0]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFITMG1, ddr_ctl_dfitmg[freq_id][1]);


        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTLDFILPCFG0, 0x07f04011);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTLDFILPCFG1, 0x000000b0);

        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIUPD0,  0xe0400018);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIUPD1,  0x0048005a);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIUPD2,  0x80000000);

        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC , ddr_ctl_dfimisc[freq_id][0]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC1, ddr_ctl_dfimisc[freq_id][1]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC2, ddr_ctl_dfimisc[freq_id][2]);

        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DBICTL,     0x1);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIPHYMSTR, 0x0);
    }

}


/*****************************************************************************
 *  函 数 名: ljm_ddrc_addrmap_precfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_addrmap_precfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP0,  ddr_ctl_addrmap[freq_id][0]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP1,  ddr_ctl_addrmap[freq_id][1]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP2,  ddr_ctl_addrmap[freq_id][2]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP3,  ddr_ctl_addrmap[freq_id][3]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP4,  ddr_ctl_addrmap[freq_id][4]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP5,  ddr_ctl_addrmap[freq_id][5]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP6,  ddr_ctl_addrmap[freq_id][6]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP7,  ddr_ctl_addrmap[freq_id][7]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP8,  ddr_ctl_addrmap[freq_id][8]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP9,  ddr_ctl_addrmap[freq_id][9]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP10, ddr_ctl_addrmap[freq_id][10]);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ADDRMAP11, ddr_ctl_addrmap[freq_id][11]);
    }
}



/*****************************************************************************
 *  函 数 名: ljm_ddrc_port_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_port_cfg(void *mmio_base, int channel_mask, int freq_id)
{
    int i;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PCCFG, ddr_ctl_pccfg[freq_id][0]);
        for (i=0; i<=3; i++) {
            // 7为下面配置的寄存器个数
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PCFGR_N + i*DDR_CTL_PORT_STEP),     ddr_ctl_pccfg[freq_id][1+i*7]);
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PCFGW_N + i*DDR_CTL_PORT_STEP),     ddr_ctl_pccfg[freq_id][2+i*7]);
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTRL_PCTRL_N + i*DDR_CTL_PORT_STEP),    ddr_ctl_pccfg[freq_id][3+i*7]);
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PCFGQOS0_N + i*DDR_CTL_PORT_STEP),  ddr_ctl_pccfg[freq_id][4+i*7]);
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PCFGQOS1_N + i*DDR_CTL_PORT_STEP),  ddr_ctl_pccfg[freq_id][5+i*7]);
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PCFGWQOS0_N + i*DDR_CTL_PORT_STEP), ddr_ctl_pccfg[freq_id][6+i*7]);
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PCFGWQOS1_N + i*DDR_CTL_PORT_STEP), ddr_ctl_pccfg[freq_id][7+i*7]);
        }
    }
}


/****************************************************************************
 *  函 数 名: ljm_ddrc_sbr_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_sbr_cfg(void *mmio_base, int channel_mask)
{
    int i;
    uint32_t val, addr;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        for (i=0; i<sizeof(ddr_ctl_sbr)/sizeof(ddr_ctl_sbr[0]); i++) {
            addr = ddr_ctl_sbr[i].addr;
            val = ddr_ctl_sbr[i].val;
            DDR_CTR_W32_BASE(mmio_base, channel_id, addr, val);
        }
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_phy_reg_blk_set
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_reg_blk_set(void *mmio_base, uint8_t channel_id, u32 blk_id)
{
    LJM_MISC_W32_BASE(mmio_base, MISC_DDR_PHY_ADDR0 + 4 * channel_id, blk_id);
}

/* 存放ddr dccm 位置大小数据的头信息 */
ljm_ddr_cfg_v2 g_ddr_cfg_info;

inline int ljm_ddr_get_dccm_start(int card, unsigned long flags)
{
    unsigned int addr = 0x58000;

    switch (flags & LJMICRO_CHIP_MASK) {
    case CHIP_GP201:
        if (g_board_info[card].vendor_id == LJM_VENDOR_ID) {
            if (g_board_info[card].ddr_imem_start & 0x10000)
                addr = g_board_info[card].ddr_imem_start + 0x10000; /* LPDDR4 dccm: 0x90000 + 0x18000 = 0xa0000 */
            else
                addr = g_board_info[card].ddr_imem_start + 0x20000; /* DDR4 dccm: 0x80000 + 0x20000 = 0xa0000 */
        } else
            addr = 0xa0000;   /* DDR4 dccm default at 0xa0000 */
        break;
    case CHIP_GP202:
        addr = 0x58000;
        break;
    default :
        addr = 0x58000;
        break;
    }

    return addr;
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_head_get
 *  功能描述: ddr iccm dccm 头获取
 *  输入参数:
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_ddr_head_get(void *mmio_base, int card, unsigned long flags)
{
    int ret = 0;
    uint16_t crc;

    ret = ljmSpiReadBuffer(mmio_base, (char*)&g_ddr_cfg_info, ljm_ddr_get_dccm_start(card, flags), sizeof(g_ddr_cfg_info));
    if (ret) {
        ERR_PRINT("read ddr cfg from SPI flash failed.");
        ret = -1;
    }
    crc = calc_crc16((uint16_t *)&g_ddr_cfg_info, sizeof(g_ddr_cfg_info)/2 - 1);
    if (crc != g_ddr_cfg_info.crc) {
        ERR_PRINT("ddr cfg crc error.");
        ret = -1;
        dump_bytes((const unsigned char *)&g_ddr_cfg_info, sizeof(g_ddr_cfg_info), "ddr cfg info");
    }
   return ret;
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1diccm_flash
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1diccm_flash(void *mmio_base, int card, int channel_mask)
{
    int ret = 0;
    void *pdata = NULL;
    int i = 0;
    uint16_t *pdata_tmp;
    int channel_id;

    pdata = kzalloc(IMEM_SIZE, GFP_KERNEL);
    if (!pdata) {
        ERR_PRINT("ddr: out of memory: %d", IMEM_SIZE);
    }

    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_imem_start, IMEM_SIZE);
    if (ret) {
        ERR_PRINT("ddr: read ddr imem0 from SPI flash failed.");
    }
    pdata_tmp = (uint16_t *) pdata;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<=IMEM_SIZE/2; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, (i*4), (u32)(*(pdata_tmp+i)));
        }
    }

    if (pdata) {
        kfree(pdata);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1diccm_tmp
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1diccm_tmp(void *mmio_base, int card, int channel_mask)
{
    uint16_t i;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<sizeof(ljm_ddr_1d_iccm)/sizeof(ljm_ddr_1d_iccm[0]); i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, (i*4), (u32)ljm_ddr_1d_iccm[i]);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1diccm
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1diccm(void *mmio_base, int card, int channel_mask, bool is_flash)
{
    if (is_flash == true)
        ljm_ddr_load_1diccm_flash(mmio_base, card, channel_mask);
    else
        ljm_ddr_load_1diccm_tmp(mmio_base, card, channel_mask);
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1ddccm
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1ddccm_right(void *mmio_base, int channel_id, int freq_id)
{
    int i;

    switch (freq_id) {
        case LJM_DDR_FREQ_800M:
            for(i=0; i<sizeof(ddr_800_1ddccm)/sizeof(ddr_800_1ddccm[0]); i++) {
               DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_800_1ddccm[i].addr+0x10000, ddr_800_1ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_1066M:
            for(i=0; i<sizeof(ddr_1066_1ddccm)/sizeof(ddr_1066_1ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_1066_1ddccm[i].addr+0x10000, ddr_1066_1ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_1600M:
            for(i=0; i<sizeof(ddr_1600_1ddccm)/sizeof(ddr_1600_1ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_1600_1ddccm[i].addr+0x10000, ddr_1600_1ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_2400M:
            for(i=0; i<sizeof(ddr_2400_1ddccm)/sizeof(ddr_2400_1ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_2400_1ddccm[i].addr+0x10000, ddr_2400_1ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_2933M:
            for(i=0; i<sizeof(ddr_2933_1ddccm)/sizeof(ddr_2933_1ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_2933_1ddccm[i].addr+0x10000, ddr_2933_1ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_3200M:
            for(i=0; i<sizeof(ddr_3200_1ddccm)/sizeof(ddr_3200_1ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_3200_1ddccm[i].addr+0x10000, ddr_3200_1ddccm[i].val);
            }
            break;
        default:
            break;
        }
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1ddccm_flash
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1ddccm_flash(void *mmio_base, int card, int channel_mask, int freq_id, unsigned long flags)
{
    int ret = 0;
    int size = 0;
    void *pdata = NULL;
    int i = 0;
    uint16_t crc;
    uint16_t *pdata_tmp = NULL;
    int channel_id;

    //(void)ljm_ddr_head_get(mmio_base, card, flags);
    size = g_ddr_cfg_info.block0_size;

    pdata = kzalloc(size, GFP_KERNEL);
    if (!pdata) {
        ERR_PRINT("out of memory: %d !", size);
    }
    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, ljm_ddr_get_dccm_start(card, flags) + g_ddr_cfg_info.block0_offset, g_ddr_cfg_info.block0_size);
    if (ret) {
        ERR_PRINT("read ddr dccm block 0 from SPI flash failed.\n");
    }

    crc = calc_crc16((uint16_t *)pdata, g_ddr_cfg_info.block0_size/2);
    if (crc != g_ddr_cfg_info.block0_crc) {
        ERR_PRINT("dccm 0 crc check err!\n");
        dump_bytes((const unsigned char *)pdata, g_ddr_cfg_info.block0_size, "ddr dccm block 0");
    }

    pdata_tmp = (uint16_t *) pdata;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;

        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<=g_ddr_cfg_info.block0_size/2; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, (i*4)+0x10000, (u32)(*(pdata_tmp+i)));
        }

        if (freq_id != LJM_DDR_FREQ_2666M) {
            ljm_ddr_load_1ddccm_right(mmio_base, channel_id, freq_id);
        }
    }

    if (pdata) {
        kfree(pdata);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1ddccm_tmp
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1ddccm_tmp(void *mmio_base, int card, int channel_mask, int freq_id)
{
    uint16_t i;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<870; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, 0x10000+(i*4), (u32)ljm_ddr_1d_dccm[i]);
        }

        if (freq_id != LJM_DDR_FREQ_2666M) {
            ljm_ddr_load_1ddccm_right(mmio_base, channel_id, freq_id);
        }
    }

}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_1ddccm_flash
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_1ddccm(void *mmio_base, int card, int channel_mask, int freq_id, bool is_flash,
                            unsigned long flags)
{
    if (is_flash == true)
        ljm_ddr_load_1ddccm_flash(mmio_base, card, channel_mask, freq_id, flags);
    else
        ljm_ddr_load_1ddccm_tmp(mmio_base, card, channel_mask, freq_id);
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2diccm_flash
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2diccm_flash(void *mmio_base, int card, int channel_mask)
{
     int ret = 0;
    void *pdata = NULL;
    int i = 0;
    uint16_t *pdata_tmp;
    int channel_id;

    pdata = kzalloc(IMEM_SIZE, GFP_KERNEL);
    if (!pdata) {
        ERR_PRINT("ddr: out of memory: %d", IMEM_SIZE);
    }

    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, g_board_info[card].ddr_imem_start+IMEM_SIZE, IMEM_SIZE);
    if (ret) {
        ERR_PRINT("ddr: read ddr imem0 from SPI flash failed.");
    }
    pdata_tmp = (uint16_t *) pdata;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<=IMEM_SIZE/2; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, (i*4), (u32)(*(pdata_tmp+i)));
        }
    }

    if (pdata) {
        kfree(pdata);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2diccm_tmp
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2diccm_tmp(void *mmio_base, int card, int channel_mask)
{
    int i;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<IMEM_SIZE/2; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, (i*4), (u32)ljm_ddr_2d_iccm[i]);
        }
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2diccm
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2diccm(void *mmio_base, int card, int channel_mask, bool is_flash)
{
    if (is_flash == true)
        ljm_ddr_load_2diccm_flash(mmio_base, card, channel_mask);
    else
        ljm_ddr_load_2diccm_tmp(mmio_base, card, channel_mask);
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2ddccm
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2ddccm_right(void *mmio_base, int channel_id, int freq_id)
{

    int i;

    switch (freq_id) {
        case LJM_DDR_FREQ_2400M:
            for(i=0; i<sizeof(ddr_2400_2ddccm)/sizeof(ddr_2400_2ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_2400_2ddccm[i].addr+0x10000, ddr_2400_2ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_2933M:
            for(i=0; i<sizeof(ddr_2933_2ddccm)/sizeof(ddr_2933_2ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_2933_2ddccm[i].addr+0x10000, ddr_2933_2ddccm[i].val);
            }
            break;
        case LJM_DDR_FREQ_3200M:
            for(i=0; i<sizeof(ddr_3200_2ddccm)/sizeof(ddr_3200_2ddccm[0]); i++) {
                DDR_PHY_W32_BASE(mmio_base, channel_id, ddr_3200_2ddccm[i].addr+0x10000, ddr_3200_2ddccm[i].val);
            }
            break;
        default:
            break;
    }

}


/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2ddccm_flash
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2ddccm_flash(void *mmio_base, int card, int channel_mask, int freq_id, unsigned long flags)
{
    int ret = 0;
    int size = 0;
    void *pdata = NULL;
    int i = 0;
    uint16_t crc;
    uint16_t *pdata_tmp = NULL;
    int channel_id;

    size = g_ddr_cfg_info.block0_size;

    pdata = kzalloc(size, GFP_KERNEL);
    if (!pdata) {
        ERR_PRINT("out of memory: %d !", size);
    }
    ret = ljmSpiReadBuffer(mmio_base, (char*)pdata, ljm_ddr_get_dccm_start(card, flags) + g_ddr_cfg_info.block1_offset, g_ddr_cfg_info.block1_size);
    if (ret) {
        ERR_PRINT("read ddr 2d dccm block 1 from SPI flash failed.\n");
    }

    crc = calc_crc16((uint16_t *)pdata, g_ddr_cfg_info.block1_size/2);
    if (crc != g_ddr_cfg_info.block1_crc) {
        ERR_PRINT("dccm 1 crc check err!\n");
        dump_bytes((const unsigned char *)pdata, g_ddr_cfg_info.block1_size, "ddr dccm block 1");
    }

    pdata_tmp = (uint16_t *) pdata;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;

        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<=g_ddr_cfg_info.block1_size/2; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, (i*4)+0x10000, (u32)(*(pdata_tmp+i)));
        }

        if (freq_id != LJM_DDR_FREQ_2666M) {
            ljm_ddr_load_2ddccm_right(mmio_base, channel_id, freq_id);
        }
    }

    if (pdata) {
        kfree(pdata);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2ddccm_tmp
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2ddccm_tmp(void *mmio_base, int card, int channel_mask, int freq_id)
{
    int i;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_UCTL_MEMORY);
        for (i=0; i<738; i++) {
            DDR_PHY_W32_BASE(mmio_base, channel_id, 0x10000+(i*4), (u32)ljm_ddr_2d_dccm[i]);
        }
        if (freq_id != LJM_DDR_FREQ_2666M) {
            ljm_ddr_load_2ddccm_right(mmio_base, channel_id, freq_id);
        }
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_load_2ddccm
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_load_2ddccm(void *mmio_base, int card, int channel_mask, int freq_id, bool is_flash,
                            unsigned long flags)
{
    if (is_flash == true)
        ljm_ddr_load_2ddccm_flash(mmio_base, card, channel_mask, freq_id, flags);
    else
        ljm_ddr_load_2ddccm_tmp(mmio_base, card, channel_mask, freq_id);
}


/*****************************************************************************
 *  函 数 名: ljm_ddrc_freq0_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_freq0_cfg(void *mmio_base, int channel_mask, int freq_id)
{
    int i;
    uint32_t val, addr;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_ODTCFG, ddr_ctl_odtcfg[freq_id]);//ODTCFG
        for (i=0; i<sizeof(ddr_ctl_freq0)/sizeof(ddr_ctl_freq0[0]); i++) {
            addr = ddr_ctl_freq0[i].addr;
            val = ddr_ctl_freq0[i].val;
            DDR_CTR_W32_BASE(mmio_base, channel_id, addr, val);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddrc_freq1_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_freq1_cfg(void *mmio_base, int channel_mask)
{
    int i;
    uint32_t val, addr;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        for (i=0; i<sizeof(ddr_ctl_freq1)/sizeof(ddr_ctl_freq1[0]); i++) {
            addr = ddr_ctl_freq1[i].addr;
            val = ddr_ctl_freq1[i].val;
            DDR_CTR_W32_BASE(mmio_base, channel_id, addr, val);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddrc_freq2_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_freq2_cfg(void *mmio_base, int channel_mask)
{
    int i;
    uint32_t val, addr;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        for (i=0; i<sizeof(ddr_ctl_freq2)/sizeof(ddr_ctl_freq2[0]); i++) {
            addr = ddr_ctl_freq2[i].addr;
            val = ddr_ctl_freq2[i].val;
            DDR_CTR_W32_BASE(mmio_base, channel_id, addr, val);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddrc_freq3_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_freq3_cfg(void *mmio_base, int channel_mask)
{
    int i;
    uint32_t val, addr;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        for (i=0; i<sizeof(ddr_ctl_freq3)/sizeof(ddr_ctl_freq3[0]); i++) {
            addr = ddr_ctl_freq3[i].addr;
            val = ddr_ctl_freq3[i].val;
            DDR_CTR_W32_BASE(mmio_base, channel_id, addr, val);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_phy_reg_blk_set
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
/*
void ljm_ddr_pll_set(void *mmio_base, int freq_id)
{
    u32 timeout = 1000;

    LJM_CRU_W32_BASE(mmio_base, CRU_DDRPLL, ljm_ddr_pll_val[freq_id]); //pll set
    LJM_CRU_W32_BASE(mmio_base, CRU_PLLEN, m_GPUPLL_CKEN1 | m_GPUPLL_CKEN0 | m_DEV2PLL_CKEN1 | m_DEVPLL_CKEN3);//pll enable
    while ((LJM_CRU_R32(ljm_dev->cur_card, CRU_STATUS) != 0xfff00000)) {
            if (--timeout == 0) {
                ERR_PRINT("wait ddr pll lock timeout\n");
                break;
            }
        }
}
*/



/*****************************************************************************
 *  函 数 名: ljm_ddr_train_stage
 *  功能描述: 通过读邮箱来判断训练阶段是否完成
 *  输入参数: u32 train_stage 训练阶段
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_train_stage(void *mmio_base, int channel_id, u32 train_stage_id)
{
    int ret = 0;

    DDR_PHY_CHECK_REG(mmio_base, channel_id, DDR_PHY_UCTSHADOWREGS, 0x2, 0xffffffff); //waiting pmu send message .bit[0] UctWriteProtShadow == 0
    if (ret < 0) {
        WARN_PRINT("%d: ddr pmu send message timeout, stage 0x%x !\n", __LINE__, train_stage_id);
    }
    DDR_PHY_CHECK_REG(mmio_base, channel_id, DDR_PHY_UCTWRITEONLYSHADOW, train_stage_id, 0xffffffff); //获取主消息号 UctWriteOnlyShadow
    if (ret < 0) {
        WARN_PRINT("+++++%d: major message:0x%x ++++\n", channel_id, DDR_PHY_R32_BASE(mmio_base, channel_id, DDR_PHY_UCTWRITEONLYSHADOW));
    }
    if (ret < 0) {
        WARN_PRINT("%d: ddr get major message timeout, stage 0x%x !\n", __LINE__, train_stage_id);
    }
    DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_DCTWRITEPROT, 0x0);//告知用户已获取 DctWriteProt=0
    DDR_PHY_CHECK_REG(mmio_base, channel_id, DDR_PHY_UCTSHADOWREGS, 0x3, 0xffffffff);//轮询DctWriteProt== 1，用户获取标记位清除
    if (ret < 0) {
        WARN_PRINT("%d: ddr clear message flag timeout, stage 0x%x !\n", __LINE__, train_stage_id);
    }
    DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_DCTWRITEPROT, m_DCTWRITEPROT); //整个读取邮箱数据流程结束
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_wait_1D_train
 *  功能描述: 通过读邮箱来判断训练阶段是否完成
 *  输入参数: u32 train_stage 训练阶段
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_wait_1D_train(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_INIT_END);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_RD_EN_TRAIN);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_FINE_WR_LVL);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_RD_DQ_DESKEW);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_MPR_RD_DELAY);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_COARSE_WR_LVL);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_WR_DELAY_CENTER);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_RD_DELAY_CENTER);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_MAX_RD_LATENCY);
     	ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_TRAIN_RUN_SUCC);
    }
 }

/*****************************************************************************
 *  函 数 名: ljm_ddr_wait_1D_train
 *  功能描述: 通过读邮箱来判断训练阶段是否完成
 *  输入参数: u32 train_stage 训练阶段
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_wait_2D_train(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_INIT_END);
	    ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_2D_WR);
	    ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_2D_RD);
	    ljm_ddr_train_stage(mmio_base, channel_id, LJM_DDR_TRAIN_RUN_SUCC);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_clk_set(int freq_id)
 *  功能描述: ddr时钟配置
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_clk_set(void *mmio_base, int freq_id)
{
    int ret = 0;

    LJM_CRU_W32_BASE(mmio_base, 0x14, ljm_ddr_id_freq_clk[freq_id][2]); //ddr clk set
    LJM_CRU_W32_BASE(mmio_base, 0x0, 0xf); //enable pll
    /* check pll lock state */
    REG_CHECK_REG(mmio_base, ljm_gpu_hardware.reg_cru_offset + 0x50, 0x0c000000, 0x0c000000);
    if (ret < 0) {
        ERR_PRINT("%d: ddr pll lock state error !\n", __LINE__);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_reset_ddrc_normal(void)
 *  功能描述: ddrc解复位
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_reset_ddrc_normal(void *mmio_base)
{
    LJM_CRU_W32_BASE(mmio_base, 0x60, 0x0); // ddr block reset to normal
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_reset(void)
 *  功能描述: ddrc解复位
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_reset(void *mmio_base)
{
    LJM_MISC_W32_BASE(mmio_base, 0x138, 0x0); //ddr0/1/2/3 power off
    LJM_MISC_W32_BASE(mmio_base, 0x138, 0xf); //ddr0/1/2/3 power on
    LJM_CRU_W32_BASE(mmio_base,  0x60,  0x00ff0000); // ddr block set
}


/*****************************************************************************
 *  函 数 名: ljm_ddrc_pre_cfg(void)
 *  功能描述: ddrc pre config
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddrc_pre_cfg(void *mmio_base, int channel_mask, int freq_id)
{

    ljm_ddrv2_set_ctr(mmio_base, channel_mask, 0x304, 0x1);
    /* 一系列配置 */
    ljm_ddrv2_set_ctr(mmio_base, channel_mask, 0x30, 0xa2);
    LJM_CRU_W32_BASE(mmio_base, 0x60, 0x00000000); // ddr block reset
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_power_stat_check(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_power_stat_check(void *mmio_base, int channel_mask)
{
    int ret;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ret = ljm_ddrc_check_value(mmio_base, channel_id, DDR_CTL_PWRCTL, 0xa2);
        if (ret < 0) {
            ERR_PRINT("%d: ljm_ddr_power_stat_check timeout! value =0x%x\n", __LINE__, DDR_CTR_R32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL));
        }
        DDR_CTR_W32_BASE( mmio_base, channel_id, DDR_CTL_PWRCTL, 0xa2);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_qos_cfg(void)
 *  功能描述: 调度优先级配置
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_qos_cfg(void *mmio_base, int freq_id)
{
#if 0   /* move to gp201.c for special chip */
    ljm_ddrv2_set_monitor(mmio_base, 0x38, 0x0fffffff);
    ljm_ddrv2_set_monitor(mmio_base, 0x20, 0x55555555);
    ljm_ddrv2_set_monitor(mmio_base, 0x24, 0x55555555);
#endif
}


/*****************************************************************************
 *  函 数 名: ljm_phy_master_timing_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_phy_master_timing_cfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x314, ddr_phymaster_tim_cfg[freq_id][0]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xb8,  ddr_phymaster_tim_cfg[freq_id][1]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x90,  ddr_phymaster_tim_cfg[freq_id][2]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xe8,  ddr_phymaster_tim_cfg[freq_id][3]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1f4, ddr_phymaster_tim_cfg[freq_id][4]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1f0, ddr_phymaster_tim_cfg[freq_id][5]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x158, ddr_phymaster_tim_cfg[freq_id][6]);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_phy_TxSlewRate_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_phy_txslewrate_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_DBYTE);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x017c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x057c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x417c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x457c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x817c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x857c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xc17c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xc57c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1017c, 0x2f1);//pstate 1
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1057c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1417c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1457c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1817c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1857c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1c17c, 0x2f1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1c57c, 0x2f1);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_host_odt_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_host_odt_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_DBYTE);

        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x0134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x0534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x4134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x4534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x8134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x8534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xc134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xc534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x10134, 0x38);//pstate 1
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x10534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x14134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x14534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x18134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x18534, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1c134, 0x38);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1c534, 0x38);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_dq_dependance_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_dq_dependance_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0x0124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0x0524, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0x4124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0x4524, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0x8124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0x8524, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0xc124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,    0xc524, 0xe3f);
        //pstate 1
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x10124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x10524, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x14124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x14524, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x18124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x18524, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1c124, 0xe3f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x1c524, 0xe3f);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_phy_atxslewrate_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_phy_atxslewrate_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_ANIB);

        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x0154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x4154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x8154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xc154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x10154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x14154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x18154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x1c154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x20154, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x24154, 0x7f);
    }
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_phy_atximpedance_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_phy_atximpedance_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_ANIB);

        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x010c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x410c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x810c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0xc10c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x1010c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x1410c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x1810c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x1c10c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x2010c, 0x7f);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x2410c, 0x7f);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_phy_master_common_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_phy_master_common_cfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);

        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x60,  ddr_phymaster_comm_cfg[freq_id][0]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x1d4, ddr_phymaster_comm_cfg[freq_id][1]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x140, ddr_phymaster_comm_cfg[freq_id][2]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x26c, ddr_phymaster_comm_cfg[freq_id][3]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x20,  ddr_phymaster_comm_cfg[freq_id][4]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x220, ddr_phymaster_comm_cfg[freq_id][5]);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x2c8, ddr_phymaster_comm_cfg[freq_id][6]);
    }
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_dq_dqs_timing_precfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_dq_dqs_timing_precfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_DBYTE);

        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x10c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,   0x50c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x410c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x450c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x810c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0x850c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0xc10c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id,  0xc50c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1010c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1050c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1410c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1450c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1810c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1850c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1c10c, 0x5b1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1c50c, 0x5b1);
    }
}


/*****************************************************************************
 *  函 数 名: ljm__phy_dfi_frq_trans_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_phy_dfi_frq_trans_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);

        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3e8, 0x1);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x064, 0x4);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3c0, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3c4, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3c8, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3cc, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3d0, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3d4, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3d8, 0x5555);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x3dc, 0xf000);
    }
}



/*****************************************************************************
 *  函 数 名: ljm_ddr_phy_DqDqsRcvCntrl_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_phy_DqDqsRcvCntrl_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_DBYTE);

        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x0128,  0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x4128,  0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x8128,  0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0xc128,  0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x10128, 0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x14128, 0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x18128, 0x500);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x1c128, 0x500);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_dmi_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_dmi_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);

        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x94,  0x0);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0xb4,  0x0);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0xb0,  0x0);
        //dram reset
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x180, 0x2);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_dmi_cfg(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_micro_release_csr(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
            ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_APBONLY);

            DDR_PHY_W32_BASE(mmio_base, channel_id, 0x0, 0x0);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_micro_start(void)
 *  功能描述: micro control start
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_micro_start(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_APBONLY);

        //phy engine control scr
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, m_MICROCONTMUXSEL);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, m_MICROCONTMUXSEL);

        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICRORESET, m_RESETTOMICRO | m_STALLTOMICRO);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICRORESET, m_STALLTOMICRO);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICRORESET, 0x0);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_micro_request_ignore(void)
 *  功能描述: phy engine control stall and release control of csr
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_micro_request_ignore(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_APBONLY);

        //phy engine control scr
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, m_MICROCONTMUXSEL);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, 0);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_stall_microcontroller(void)
 *  功能描述: phy engine control stall and release control of csr
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_stall_micro(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        //phy engine control scr
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICRORESET, m_STALLTOMICRO);

        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, 0x0);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, m_MICROCONTMUXSEL);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, 0x0);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_pie_vector(void)
 *  功能描述: 调度优先级配置
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_pie_vector(void *mmio_base, int channel_mask)
{
    int i;
    uint32_t addr, val;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_INITENG);
        for (i=0; i<sizeof(ddr_initeng_cfg)/sizeof(ddr_initeng_cfg[0]); i++) {
                addr = ddr_initeng_cfg[i].addr;
                val = ddr_initeng_cfg[i].val;
                DDR_PHY_W32_BASE(mmio_base, channel_id, addr, val);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_pie_flag(void)
 *  功能描述: 调度优先级配置
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_pie_flag(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_APBONLY);
        //tell firmware to do initeng?
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQUENCEROVERRIDE, 0x400);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_start_vector_cfg(void)
 *  功能描述: 调度优先级配置
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_start_vector_cfg(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_INITENG);

        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x5c, 0x0);
        DDR_PHY_W32_BASE(mmio_base, channel_id, 0x98, 0x2a);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_start_vector_cfg(void)
 *  功能描述: pll dll delay cfg
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_pie_delay_cfg(void *mmio_base, int channel_mask, int freq_id)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);

        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDLY0, ddr_phy_pie_delay[freq_id][0]);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDLY1, ddr_phy_pie_delay[freq_id][1]);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDLY2, ddr_phy_pie_delay[freq_id][2]);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDLY3, ddr_phy_pie_delay[freq_id][3]);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_seq_flag_set(void)
 *  功能描述: 与频点无关
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_seq_flag_set(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_INITENG);

        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG0, 0x0);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG1, 0x173);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG2, 0x60);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG3, 0x6110);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG4, 0x2152);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG5, 0xdfbd);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG6, 0xffff);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_SEQ0BDISABLEFLAG7, 0x6152);
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_calibration_set(void)
 *  功能描述:
 *  输入参数: freq_id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
 void ljm_ddr_calibration_set(void *mmio_base, int channel_mask, int freq_id)
 {
     int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        //1066频点及以下无需配置
        if (freq_id <= 1) {
            ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);
            // cal into idle
            DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_CALZAP, 0x1);
            // cal mode cfg
            DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_CALRATE, 0x19);
        }
    }
 }

 /*****************************************************************************
  *  函 数 名: ljm_ddr_ucclk_hclk_disable(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
  void ljm_ddr_ucclk_hclk_disable(void *mmio_base, int channel_mask)
  {
      int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_PUB);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_UCCLKHCLKENABLES, 0x0); // no need train, dis clk to reduce power
    }
  }


/*****************************************************************************
  *  函 数 名: ljm_ddr_apb_control_internal_csr(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
void ljm_ddr_apb_control_internal_csr(void *mmio_base, int channel_mask)
{
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_APBONLY);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, m_MICROCONTMUXSEL);
        DDR_PHY_W32_BASE(mmio_base, channel_id, DDR_PHY_MICROCONTMUXSEL, 0);
    }
}


/*****************************************************************************
  *  函 数 名: ljm_ddr_wait_calibration_done(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
void ljm_ddr_wait_calibration_done(void *mmio_base, int channel_mask)
{
    int ret =0;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        ljm_ddr_reg_blk_set(mmio_base, channel_id, LJM_DDR_MASTER);
        REG_CHECK_REG(mmio_base, ljm_gpu_hardware.reg_ddr_phy_offset[channel_id] + DDR_PHY_CALBUSY, 0, m_CALBUSY);
        if (ret < 0) {
        ERR_PRINT("%d: ljm_ddr_wait_calibration_done  time out \n", __LINE__);
        }
    }
}


/*****************************************************************************
  *  函 数 名: ljm_ddr_wait_dfi_init_done(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
void ljm_ddr_wait_dfi_init_done(void *mmio_base, int channel_mask)
{
    int ret = 0;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        //Triggers the PHY init start request
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC, m_DISDYNADRTRI | m_DFI_INITSTART);
        //wait DFI initialization completed
        //DDR_CTR_CHECK_REG(mmio_base, channel_id, DDR_CTL_DFISTAT, m_DFI_INIT_COMPELETE, 0xffffffff);
        ret = ljm_ddrc_check_value(mmio_base, channel_id, DDR_CTL_DFISTAT, m_DFI_INIT_COMPELETE);
        if (ret < 0) {
            ERR_PRINT("%d: DFI initialization timeout!\n", __LINE__);
        }
        //Triggers clear
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC, m_DISDYNADRTRI);
    }
}


/*****************************************************************************
  *  函 数 名: ljm_ddr_read_qos_cfg(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
void ljm_ddr_read_qos_cfg(void *mmio_base, int channel_mask, unsigned long flags)
{
    int channel_id;
    unsigned int ddr_ctl_port_used = GP202_DDR_CTL_PORT_USED;

    switch (flags & LJMICRO_CHIP_MASK) {
        case CHIP_GP201 :
            ddr_ctl_port_used = GP201_DDR_CTL_PORT_USED;
            break;
        case CHIP_GP202 :
            ddr_ctl_port_used = GP202_DDR_CTL_PORT_USED;
            break;
        default :
            ddr_ctl_port_used = GP202_DDR_CTL_PORT_USED;
            break;
    }

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DBG1, m_DIS_HIF | m_DIS_DQ);// some reg(Dynamic reg group) set ennd ctl empty
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PORT_STEP * ddr_ctl_port_used + DDR_CTRL_PCTRL_N), 0);
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PORT_STEP * ddr_ctl_port_used + DDR_CTL_PCFGQOS0_N), 0x200004);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DBG1, 0x0);
    }
}


/*****************************************************************************
  *  函 数 名: ljm_ddr_wait_ctl_normal(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
void ljm_ddr_wait_ctl_normal(void *mmio_base, int channel_mask)
{
    int ret = 0;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DFIMISC, m_DISDYNADRTRI | m_DFI_INIT_COMPLETE_EN);// enable dfi_init_compelete
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL, m_LPDDR4_SR_ALLOWED | m_DIS_CAM_DRAIN_SELFREF);
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_SWCTL, m_SW_DONE);//dynamic regs programed

        //DDR_CTR_CHECK_REG(mmio_base, channel_id, DDR_CTL_SWSTAT, 0x1, 0xffffffff);//wait sw_done
        ret = ljm_ddrc_check_value(mmio_base, channel_id, DDR_CTL_SWSTAT, 0x1);
        if (ret < 0) {
            ERR_PRINT("%d:dynamic regs program timeout!\n", __LINE__);
        }
        //DDR_CTR_CHECK_REG(mmio_base, channel_id, DDR_CTL_STAT, m_DDR_CTL_STAT_NORMAL, 0xffffffff);//wait sw_done
        ret = ljm_ddrc_check_value(mmio_base, channel_id, DDR_CTL_STAT, m_DDR_CTL_STAT_NORMAL);
        if (ret < 0) {
            ERR_PRINT("%d: ddr controler to normal state timeout!\n", __LINE__);
        }
    }
}


/*****************************************************************************
  *  函 数 名: ljm_ddr_ctl_runtime_cfg(void)
  *  功能描述:
  *  输入参数: freq_id
  *  输出参数: 无
  *  返 回 值: 无
  *
  *  修改历史:
  *         1. 日    期: 2024年4月7日
  *            作    者: xuht
  *            修改内容: 新生成函数
  *
  *****************************************************************************/
void ljm_ddr_ctl_runtime_cfg(void *mmio_base, int channel_mask, int freq_id)
{
    int i;
    int channel_id;

    for (channel_id = 0; channel_id < ljm_gpu_hardware.max_ddr_channels; channel_id++) {
        if (!(channel_mask & (0x1 << channel_id))) continue;
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_DERATEEN, 0x1606); //derate
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_DERATEEN + 0x2000), 0x1606); //derate freq1
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_DERATEEN + 0x3000), 0x1606); //derate freq2
        DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_DERATEEN + 0x4000), 0x1606); //derate freq3
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_SBRCTL, 0x21e12); // scrub ecc disable
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL, m_LPDDR4_SR_ALLOWED | m_DIS_CAM_DRAIN_SELFREF);
        // ctl deep powerdone enable
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_PWRCTL, m_LPDDR4_SR_ALLOWED | m_DIS_CAM_DRAIN_SELFREF | m_POWERDOWN_EN);
        // refresh mode cfg
        DDR_CTR_W32_BASE(mmio_base, channel_id, DDR_CTL_RFSHCTL3, ddr_ctl_refresh_mode[freq_id]);
        //axi port enable
        for (i=0; i<4; i++) {
            DDR_CTR_W32_BASE(mmio_base, channel_id, (DDR_CTL_PORT_STEP * i + DDR_CTRL_PCTRL_N), m_PORT_EN);
        }
    }
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_phy_to_normal_state
 *  功能描述: ddr  寄存器域设置
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_phy_to_normal_state(void *mmio_base, int channel_mask, int freq_id, unsigned long flags)
{
    ljm_ddr_pie_flag(mmio_base, channel_mask);
    ljm_ddr_start_vector_cfg(mmio_base, channel_mask);
    ljm_ddr_pie_delay_cfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_seq_flag_set(mmio_base, channel_mask);
    ljm_ddr_calibration_set(mmio_base, channel_mask, freq_id);
    ljm_ddr_ucclk_hclk_disable(mmio_base, channel_mask);
    ljm_ddr_apb_control_internal_csr(mmio_base, channel_mask);
    ljm_ddr_wait_calibration_done(mmio_base, channel_mask);
    ljm_ddr_wait_dfi_init_done(mmio_base, channel_mask);
    ljm_ddr_read_qos_cfg(mmio_base, channel_mask, flags);
    ljm_ddr_wait_ctl_normal(mmio_base, channel_mask);
    ljm_ddr_ctl_runtime_cfg(mmio_base, channel_mask, freq_id);
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_initphyconfig
 *  功能描述: ddr     pre phy config
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_initphyconfig(void *mmio_base, int channel_mask, int freq_id)
{
    ljm_phy_txslewrate_cfg(mmio_base, channel_mask);
    ljm_phy_atxslewrate_cfg(mmio_base, channel_mask);
    ljm_phy_master_timing_cfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_host_odt_cfg(mmio_base, channel_mask);
    ljm_ddr_dq_dependance_cfg(mmio_base, channel_mask);
    ljm_ddr_phy_atximpedance_cfg(mmio_base, channel_mask);
    ljm_ddr_phy_master_common_cfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_dq_dqs_timing_precfg(mmio_base, channel_mask);
    ljm_phy_dfi_frq_trans_cfg(mmio_base, channel_mask);
    ljm_ddr_phy_DqDqsRcvCntrl_cfg(mmio_base, channel_mask);
    ljm_ddr_dmi_cfg(mmio_base, channel_mask);
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_initctlconfig
 *  功能描述: ddr     pre ctl config
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_initctlconfig(void *mmio_base, int channel_mask, int freq_id)
{
    ljm_ddr_ddrc_init_stat_wait(mmio_base, channel_mask);
    ljm_ddr_for_ddr4_precfg(mmio_base, channel_mask);
    ljm_ddr_rfsh_precfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_ecc_crc_precfg(mmio_base, channel_mask);
    ljm_ddr_sdram_intreg_precfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_dimm_rank_precfg(mmio_base, channel_mask);
    ljm_ddr_dramtmg_precfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_zqctl_precfg(mmio_base, channel_mask, freq_id);
    ljm_ddr_dfi_precfg(mmio_base, channel_mask, freq_id);
    ljm_ddrc_addrmap_precfg(mmio_base, channel_mask, freq_id);
    ljm_ddrc_freq0_cfg(mmio_base, channel_mask, freq_id);
    ljm_ddrc_port_cfg(mmio_base, channel_mask, freq_id);
    ljm_ddrc_sbr_cfg(mmio_base, channel_mask);
    ljm_ddrc_freq1_cfg(mmio_base, channel_mask);
    ljm_ddrc_freq2_cfg(mmio_base, channel_mask);
    ljm_ddrc_freq3_cfg(mmio_base, channel_mask);
    ljm_ddr_power_stat_check(mmio_base, channel_mask);
    ljm_ddr_reset_ddrc_normal(mmio_base);
    ljm_ddr_qos_cfg(mmio_base, freq_id);
    ljm_ddr_scrub_check(mmio_base, channel_mask);
    ljm_ddr_ctl_precfg_end(mmio_base, channel_mask);
}


/*****************************************************************************
 *  函 数 名: ljm_ddr_init_freq
 *  功能描述: ddr  单频点初始化
 *  输入参数: u32 blk_id    寄存器域id
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_ddr_init_freq(void *mmio_base, int card, int channel_mask, int freq_id, bool is_flash,
                            unsigned long flags)
{
    ljm_ddr_clk_set(mmio_base, freq_id);
    ljm_ddr_reset(mmio_base);

    /* init ctl config */
    ljm_ddr_initctlconfig(mmio_base, channel_mask, freq_id);
    /* init phy config */
    ljm_ddr_initphyconfig(mmio_base, channel_mask, freq_id);
    ljm_ddr_micro_release_csr(mmio_base, channel_mask);

    /* load IMEM Memory */
    ljm_ddr_load_1diccm(mmio_base, card, channel_mask, is_flash);
    ljm_ddr_micro_request_ignore(mmio_base, channel_mask);

    /* load dccm memory */
    ljm_ddr_load_1ddccm(mmio_base, card, channel_mask, freq_id, is_flash, flags);
    /* execute firmware */
    ljm_ddr_micro_start(mmio_base, channel_mask);
    /* read message block result */
    ljm_ddr_wait_1D_train(mmio_base, channel_mask);
    ljm_ddr_stall_micro(mmio_base, channel_mask);

    if (freq_id >= LJM_DDR_FREQ_2400M) {
        ljm_ddr_load_2diccm(mmio_base, card, channel_mask, is_flash);
        ljm_ddr_micro_request_ignore(mmio_base, channel_mask);
        ljm_ddr_load_2ddccm(mmio_base, card, channel_mask, freq_id, is_flash, flags);
        ljm_ddr_micro_start(mmio_base, channel_mask);
        /* read message block result */
        ljm_ddr_wait_2D_train(mmio_base, channel_mask);
        ljm_ddr_stall_micro(mmio_base, channel_mask);
    }

    ljm_ddr_pie_vector(mmio_base, channel_mask);
    /* init phy to mission mode via dfi request*/
    ljm_ddr_phy_to_normal_state(mmio_base, channel_mask, freq_id, flags);

    INFO_PRINT("ddr freq %d init finshed!\n", freq_id);
}

/*****************************************************************************
 *  函 数 名: ljm_ddrv2_init_v2
 *  功能描述: ddr   初始化
 *  输入参数:
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

int ljm_ddrv2_init_v2(int card, void *mmio_base, unsigned long *pddr_size, bool is_flash, unsigned long flags)
{
    int ret = 0;
    struct ddrv2_ctrl_phy_reg *extra_regs = NULL;

    /* check and read extra configuration for registers */
    if (g_board_info[card].extra_reg_count > 0) {
        extra_regs = kzalloc(g_board_info[card].extra_reg_count * sizeof(struct ddrv2_ctrl_phy_reg), GFP_KERNEL);
        ret = ljmSpiReadBuffer(mmio_base,
            (char*)extra_regs,
            SPI_BOADR_INFO_OFFSET + sizeof(g_board_info[card]),
            g_board_info[card].extra_reg_count * sizeof(struct ddrv2_ctrl_phy_reg)
            );
        if (ret) {  /* read extra registers error */
            ERR_PRINT("ljm_ddr_init: read extra registers info error\n");
            g_board_info[card].extra_reg_count = 0;

            kfree(extra_regs);
            extra_regs = NULL;
        } else
            dump_bytes((const unsigned char *)extra_regs, g_board_info[card].extra_reg_count * sizeof(struct ddrv2_ctrl_phy_reg), "extra registers");
    }

    ljm_ddr_init_freq(mmio_base, card, ddrChannelMask, ddrFreqId, is_flash, flags);

     /* 6: configure extra registers */
    if (g_board_info[card].extra_reg_count > 0) {
        ljm_ddrv2_cfg(mmio_base,
            ddrChannelMask,
            (struct ddrv2_ctrl_phy_reg *)extra_regs,
            g_board_info[card].extra_reg_count
            );

        if (extra_regs) {
            kfree(extra_regs);
            extra_regs = NULL;
        }
    }

    if (pddr_size) {
        *pddr_size = ljm_get_vram_size_from_boardinfo(card);
        DEBUG_PRINT("found ddr size: %zuB from board info (%x)", (*pddr_size), g_board_info[card].ddr_total_size);
    }

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_ddr_init
 *  功能描述: ddr   初始化
 *  输入参数:
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2024年4月7日
 *            作    者: xuht
 *            修改内容: 新生成函数
 *
 *****************************************************************************/

void ljm_ddr_init(int card, void *mmio_base, unsigned long *pddr_size, unsigned long flags)
{

    ljm_board_info *pboard_info;
    int ret = 0;

    pboard_info = ljm_ddrv2_get_board_info(card, mmio_base);

    /* boardinfo 为空时iccm dccm数据用代码段硬编码 */
    if (pboard_info == NULL) {
        INFO_PRINT("board info err, use default iccm dccm data\n");
        ljm_ddrv2_init_v2(card, mmio_base, pddr_size, false, flags);
    }
    /* version大于200时，先校验dccm数据头，校验错误时数据用代码段硬编码 */
    else if ((pboard_info->version >= 0x200)) {
        if(ljm_ddr_head_get(mmio_base, card, flags) != (-1)) {
            INFO_PRINT("use new interface & iccm dccm data from flash\n");
            ljm_ddrv2_init_v2(card, mmio_base, pddr_size, true, flags);
        }
        else {
            INFO_PRINT("dccm err & use default iccm dccm data\n");
            ljm_ddrv2_init_v2(card, mmio_base, pddr_size, false, flags);
        }
    }
    /* version小于200时，老流程初始化出错时数据用代码段硬编码 */
    else {
        ret = ljm_ddrv2_init(card, mmio_base, pddr_size);
        if (ret != 0) {
            INFO_PRINT("old interface err (%d) & use default iccm dccm data\n", ret);
            ljm_ddrv2_init_v2(card, mmio_base, pddr_size, false, flags);
        }
    }
}

