/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_debug.c
 *  版 本 号: V1.0
 *  作    者: 胡艳明
 *  生成日期: 2022年4月15日
 *
 *  功能描述: 信息打印输出函数
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 创建文件
 *
 ******************************************************************************/

#ifndef UEFI_GOP
#include <asm/io.h>
#endif
#include "ljmicro_base.h"

/******************************************************************************
 *                                    函数声明                                    *
 ******************************************************************************/

char *valid_p(void);
// void print_init(void);
char *indent_h(void);
char *indent_f(void);
int   dbg_check(void);

char p_buf[] = "                              ";
#define s_buf sizeof(p_buf)
char *p_b = (p_buf + s_buf + 1);

/******************************************************************************
 *                                    函数实现                                    *
 ******************************************************************************/
char *indent_h(void)
{
    --p_b;
    --p_b;
    return valid_p();
}

char *indent_f(void)
{
    char *ret = valid_p();
    p_b++;
    p_b++;
    return ret;
}

char *valid_p(void)
{
    if (p_b < p_buf)
        return p_buf;
    if (p_b > (p_buf + s_buf - 1))
        return (p_buf + s_buf - 1);
    return p_b;
}

int dbg_check(void)
{
    if (p_b != (p_buf + s_buf - 1))
        return 1;
    return 0;
}

#ifdef UEFI_GOP
void dump_bytes(const unsigned char *block, int size, const char *headinfo)
{
    int           i;
    unsigned char data;
    char          line[] = "       0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f \n";
    char          star[] = "******************************************************\n";

    /* dump data */
    if (headinfo)
    {
        PRINT("Dump %s with %d bytes\n", headinfo, size);
    }
    PRINT(line);
    PRINT(star);
    for (i = 0; i < size; ++i)
    {
        if (i % 16 == 0)
        {
            if (i != 0)
            {
                PRINT("\n");
            }

            PRINT("%04x: ", i); /* print address */
        }

        data = block[i];
        PRINT("%02x ", data);
    }
}
#else /* !UEFI_GOP */
void dump_bytes(const unsigned char *block, int size, const char *headinfo)
{
    int           i, offset;
    unsigned char data;
    char          line[] = "       0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f \n";
    char          star[] = "******************************************************\n";

    /* dump data */
    if (headinfo)
    {
        DEBUG_PRINT("Dump %s with %d bytes\n", headinfo, size);
    }
    DEBUG_PRINT("%s", line);
    DEBUG_PRINT("%s", star);
    for (i = 0, offset = 0; i < size; ++i)
    {
        if (i % 16 == 0)
        {
            if (i != 0)
                DEBUG_PRINT("%s\n", line);
            offset = 0;
            offset += sprintf(line + offset, "%04x: ", i); /* print address */
        }

        data = block[i];
        offset += sprintf(line + offset, "%02x ", data);
    }

    if (offset)
    {
        line[offset] = '\0';
        DEBUG_PRINT("%s\n", line);
    }
}
#endif /* ~UEFI_GOP */

#if DEBUG_DUMP_REG
/*****************************************************************************
 *  函 数 名: dump_vga_regs
 *  功能描述: 打印VGA控制器计算器
 *  输入参数: int Card                   插卡编号
 *            struct ljm_init_reg *Regs  寄存器数组
 *            int Size                   寄存器数量
 *            enum gp_read_type Type     寄存器类型
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void dump_vga_regs(int Card, struct ljm_init_reg *Regs, int Size, enum gp_read_type Type)
{
    int          i, j;
    unsigned int addr, data;
    char         line[] = "     0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f \n";
    int          flag   = 0; /* address disorder flag */
    char         dash[] = "====================================================\n";

    PRINT("%s", line);
    PRINT("%s", dash);
    for (i = 0, j = 0; i < Size; ++i, ++j)
    {
        if (j % 16 == 0)
        {
            if (j != 0)
            {
                PRINT("\n");
            }
            PRINT("%02x: ", j); /* print address */
        }
        addr = Regs[i].addr;

        /* print registers which are not read */
        while (j < addr)
        {
            PRINT("?? ");
            j++;
            if (j % 16 == 0)
            {
                PRINT("\n");
                PRINT("%02x: ", j); /* print address */
            }
        }

        if (j > addr)
        {
            PRINT("\n");
            flag = 1;
        }

        /* read register */
        switch (Type)
        {
            case GP_READ_TYPE_AR:
                data = AR_RD(Card, addr);
                break;
            case GP_READ_TYPE_SR:
                data = SR_RD(Card, addr);
                break;
            case GP_READ_TYPE_GR:
                data = GR_RD(Card, addr);
                break;
            case GP_READ_TYPE_CR:
                data = CR_RD(Card, addr);
                break;
            default:
                data = AR_RD(Card, addr);
                break;
        }

        if (flag)
        {
            PRINT("[%02x]=%02x\n", addr, data);
            flag = 0;
        }
        else
        {
            PRINT("%02x ", data);
        }
    }

    PRINT("%s", dash);
}
#endif /* ~DEBUG_DUMP_REG */

/*
 * function for break
 */
#if defined(_DEBUG) && USE_DEBUG_BREAK

static uint breakId = 0;
module_param(breakId, uint, 0644);
MODULE_PARM_DESC(breakId, "break index from 1 to 0xfffffffd, and 0: none break, 0xffffffff: break all, 0xfffffffe: break next");

unsigned int debug_read_break_id(void)
{
    return breakId;
}

void debug_write_break_id(unsigned int break_id)
{
    breakId = break_id;
}

#define MAX_BUF_LEN 512
/*****************************************************************************
 *  函 数 名: debug_print_vars
 *  功能描述: 打印可变参数内的指针值
 *  输入参数: int vars_cnt  可变参数数量
 *            ...           指针类型可变参数
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2023年2月10日
 *            作    者: hym
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void debug_print_vars(int vars_cnt, ...)
{
    char buf[MAX_BUF_LEN];
    va_list args;
    void *ptr;
    int i;

    buf[0] = '\0';

    va_start(args, vars_cnt);
    for (i=0; i<vars_cnt; i++) {
        ptr = va_arg(args, void *);
        snprintf(buf + strlen(buf), MAX_BUF_LEN - strlen(buf), " 0x%lx", (unsigned long)ptr);
    }
    va_end(args);

    printk(KERN_INFO "%s\n", buf);
    return;
}

void debug_bug_on(int break_id)
{
    BUG_ON(breakId == break_id);
}

void debug_warn_on(int break_id)
{
    WARN_ON(breakId == break_id);
}
#endif
