/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/pm_runtime.h>
#include <linux/console.h>
#include "ljmicro_base.h"
#include "ljmicro_device.h"
#include "ljmicro_family.h"
#include "ljmicro_mode.h"
#include "ljmicro_i2s.h"
#include "ljmicro_pm.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

int gp200_pre_init(int card, struct pci_dev *pdev, unsigned long flags, unsigned long *pddr_size);

void __iomem *g_mmio_base = NULL;
int g_card_cnt = 0;
void __iomem *g_mmio_bases[LJM_MAX_CARDS_NUM] = {NULL};

static struct ljm_device *g_ljm_dev;
static struct ljm_device *g_ljm_devs[LJM_MAX_CARDS_NUM];

bool is_fpga = false;

/** hardware special data */
struct ljm_gpu_hardware_t ljm_gpu_hardware = {};

struct ljm_device * get_ljm_dev(void)
{
    return g_ljm_dev;
}

/*****************************************************************************
 *  函 数 名: get_ljm_devs
 *  功能描述: 根据卡序号返回其drm驱动私有数据结构体指针
 *  输入参数: int card  卡序号
 *  输出参数: 无
 *  返 回 值: 如果卡存在，则返回drm私有结构体指针，否则返回NULL
 *
 *  修改历史:
 *         1. 日    期: 2022年4月24日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
struct ljm_device * get_ljm_devs(int card)
{
    if (card < 0 || card >= LJM_MAX_CARDS_NUM) {
        return NULL;
    }
    return g_ljm_devs[card];
}

void ljm_disp_attr_init(struct ljm_device *ljm_dev)
{
    switch (ljm_dev->family) {
        case CHIP_GP100:
            ljm_dev->disp_attr.max_hdisplay = 3840;
            ljm_dev->disp_attr.max_vdisplay = 2160;
            ljm_dev->disp_attr.min_hdisplay = 320;
            ljm_dev->disp_attr.min_vdisplay = 240;
            ljm_dev->disp_attr.max_pclk = 340000;   /* HDMI 1.4: UHD-30@297M 3840x2160@30 */
            ljm_dev->disp_attr.default_hdisplay = 1920;
            ljm_dev->disp_attr.default_vdisplay = 1080;
            ljm_dev->disp_attr.max_dvi_hdisplay = 1920;
            ljm_dev->disp_attr.max_dvi_vdisplay = 1200;
            ljm_dev->disp_attr.max_dvi_pclk = 165000;
            break;
        case CHIP_GP200:
        case CHIP_GP300:
            ljm_dev->disp_attr.max_hdisplay = 4096;
            ljm_dev->disp_attr.max_vdisplay = 2160;
            ljm_dev->disp_attr.min_hdisplay = 320;
            ljm_dev->disp_attr.min_vdisplay = 240;
            ljm_dev->disp_attr.max_pclk = 600000;   /* HDMI 2.0: 4K-60@594M 4090x2160@60 */
            ljm_dev->disp_attr.default_hdisplay = 1920;
            ljm_dev->disp_attr.default_vdisplay = 1080;
            ljm_dev->disp_attr.max_dvi_hdisplay = 1920;
            ljm_dev->disp_attr.max_dvi_vdisplay = 1200;
            ljm_dev->disp_attr.max_dvi_pclk = 165000;
            break;
        default:
            break;
    }
}

int ljm_ddr_set(void *vram_base, resource_size_t vram_start, size_t vram_size, unsigned int value)
{
    unsigned int *base;
    int i;
    int r = 0;

    HEADER_ARG("vram_base=0x%px vram_start=0x%lx vram_size=%zx value=%u",
        vram_base, (unsigned long)vram_start, vram_size, value);

    if (vram_base)
        base = (unsigned int *)vram_base;
    else {
        base = (unsigned int *)ioremap_wc(vram_start, vram_size);
    }

    if (base == NULL) {
        printk(KERN_ERR "%s: ioremap(%lx, %zx) failed\n", __FUNCTION__, (unsigned long)vram_start, vram_size);
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    for(i=0; i < vram_size/sizeof(unsigned int); i+=0x1) {/* for fpga board with crc we need to init all ddr */
        base[i] = value;
    }

    if (vram_base == NULL && base)
        iounmap(base);

    FOOTER_ARG("r=%d", r);
    return r;
}

int ljm_ddr_test(void *vram_base, resource_size_t vram_start, size_t vram_size)
{
    unsigned int *base;
    const unsigned int data=0x5a5a55aa;
    int i;
    int r = 0;

    HEADER_ARG("vram_base=0x%lx vram_start=0x%lx vram_size=%zx",
        (unsigned long)vram_base, (unsigned long)vram_start, vram_size);

    if (vram_base)
        base = (unsigned int *)vram_base;
    else
        base = (unsigned int *)ioremap_wc(vram_start, vram_size);

    if (base == NULL) {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
        printk(KERN_ERR "ljm_ddr_test: ioremap(%lx, %zx) failed\n", (unsigned long)vram_start, vram_size);
#else
        printk(KERN_ERR "ljm_ddr_test: ioremap_nocache(%lx, %zx) failed\n", (unsigned long)vram_start, vram_size);
#endif
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    for(i=0; i < vram_size/sizeof(unsigned int); i+=0x100000)
        base[i] = data;

    for(i=0; i < vram_size/sizeof(unsigned int); i+=0x100000) {
        if (base[i] != data) {
            ERR_PRINT("%s(%d): VRAM (%x) should be %x, but %x\n",
                      __FUNCTION__, __LINE__,
                      i, data, base[i]
                      );
            r = -1;
            break;
        }
    }

    if (vram_base == NULL && base)
        iounmap(base);

    FOOTER_ARG("r=%d", r);
    return r;
}

/* initialize GART controller */
int ljm_pre_init(int card, struct pci_dev *pdev, unsigned long flags)
{
    int r = 0;
    enum ljm_family family = flags & LJMICRO_FAMILY_MASK;
    unsigned long ddr_size = 0x100000000;

    HEADER_ARG("card=%d pdev=%p flags=%lx", card, pdev, flags);

    switch (family) {
        case CHIP_GP200:
            r = gp200_pre_init(card, pdev, flags, &ddr_size);
#ifndef _DEBUG
            if (ddr_size <= 2048UL * 1024 * 1024)
                ddr_size = 4096UL * 1024 * 1024;
#endif
            printk(KERN_INFO "LJ GP201: detect VRAM RAM=%zuMB, BAR=256MB\n", ddr_size >> 20);
            break;
        case CHIP_GP300:
            printk(KERN_INFO "LJ GP301: detect VRAM RAM=%zuMB, BAR=256MB\n", ddr_size >> 20);
            break;
        default:
            break;
    }

    if (flags & LJMICRO_IS_FPGA) {  /** for fpga board with crc we need to init all ddr */
        r = ljm_ddr_set(NULL, pci_resource_start(pdev, 0), pci_resource_len(pdev, 0), 0);
        if (r)
            printk(KERN_ERR "VRAM mem set error\n");
    }

    FOOTER();
    return r;
}

#ifdef LJMICRO_DRV_ONLY
int vpu_suspend(void)
{
    HEADER();

    /* TODO */

    FOOTER();
    return 0;
}

int vpu_hard_int(void)
{
    HEADER();

    /* TODO */

    FOOTER();
    return 0;
}


int gpu_system_suspend(struct device *dev)
{
    HEADER_ARG("dev=%p", dev);

    /* TODO */

    FOOTER();
    return 0;
}

int gpu_system_resume(struct device *dev)
{
    HEADER_ARG("dev=%p", dev);

    /* TODO */

    FOOTER();
    return 0;
}

int ljmicro_get_args_power_management(void)
{
    return 0;
}
#else
int vpu_suspend(void)
{
    HEADER();

    /* TODO with GPU core */

    FOOTER();
    return 0;
}

int vpu_hard_int(void)
{
    HEADER();

    /* TODO with GPU core */

    FOOTER();
    return 0;
}


int gpu_system_suspend(struct device *dev)
{
    HEADER_ARG("dev=%p", dev);

    /* TODO with GPU core */

    FOOTER();
    return 0;
}

int gpu_system_resume(struct device *dev)
{
    HEADER_ARG("dev=%p", dev);

    /* TODO with GPU core */

    FOOTER();
    return 0;
}

int ljmicro_get_args_power_management(void);
#endif

/**
 * ljm_device_init - initialize the driver
 *
 * @ljm_dev: ljm_device pointer
 * @pdev: drm dev pointer
 * @pdev: pci dev pointer
 * @flags: driver flags
 *
 * Initializes the driver info and hw (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver startup.
 */
int ljm_device_init(struct ljm_device *ljm_dev,
    struct drm_device *ddev,
    struct pci_dev *pdev,
    uint32_t flags
    )
{
    int r = 0;

    HEADER_ARG("ljm_dev=%p (%px) ddev=%p (%px) ddev->dev=%p (%px) pdev=%p (%px)(%x:%x) flags=%x",
        ljm_dev, ljm_dev, ddev, ddev, ddev->dev, ddev->dev, pdev, pdev, pdev->vendor, pdev->device, flags);

    ljm_dev->dev = ddev->dev;
    ljm_dev->ddev = ddev;
    ljm_dev->pdev = pdev;
    ljm_dev->flags = flags;
    ljm_dev->family = flags & LJMICRO_FAMILY_MASK;
    ljm_dev->cur_card = g_card_cnt++;

    /* set display attributes like max_pclk */
    ljm_disp_attr_init(ljm_dev);

    spin_lock_init(&ljm_dev->CR_RD_lock);
    spin_lock_init(&ljm_dev->cru_lock);
    mutex_init(&ljm_dev->hw_mutex);
    mutex_init(&ljm_dev->i2c_mutex);

    /* Set asic functions */
    r = ljm_asic_init(ljm_dev);
    if (r) goto OnERROR;

    /* get ljm device register memory */
    ljm_dev->mmio_start = pci_resource_start(pdev, 2) + MMIO_OFFSET;
    ljm_dev->mmio_len = MMIO_SIZE;
    if (ljm_dev->mmio_len > pci_resource_len(pdev, 2))
        ljm_dev->mmio_len = pci_resource_len(pdev, 2);

    if (g_mmio_bases[ljm_dev->cur_card] == NULL) {
        /* ioremap ljm device register memory */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
        ljm_dev->mmio_base = ioremap(ljm_dev->mmio_start, ljm_dev->mmio_len);
#else
        ljm_dev->mmio_base = ioremap_nocache(ljm_dev->mmio_start, ljm_dev->mmio_len);
#endif
        if (!ljm_dev->mmio_base) {
            DRM_ERROR("%s: Cannot map register area.\n", __FUNCTION__);
            r = -ENOMEM;
            goto OnERROR;
        }
        g_mmio_bases[ljm_dev->cur_card] = ljm_dev->mmio_base;
    }
    else {
        ljm_dev->mmio_base = g_mmio_bases[ljm_dev->cur_card];
    }

    if (0 == ljm_dev->cur_card) {
        g_mmio_base = g_mmio_bases[ljm_dev->cur_card];
        g_ljm_dev = ljm_dev;
    }

    g_ljm_devs[ljm_dev->cur_card] = ljm_dev;

    /* now RREG8/16/32 and WREG8/16/32 are available */

    /* get ljm video memory */
    ljm_dev->smem_start = pci_resource_start(pdev, 0);
    ljm_dev->smem_len = pci_resource_len(pdev, 0);

    /* gem init */
    r = ljm_gem_init(ljm_dev);
    if (r) goto OnERROR;

#ifdef LJMICRO_DRV_ONLY
    /* ioremap ljm device register memory */
    ljm_dev->smem_base = ioremap_wc(ljm_dev->smem_start, ljm_dev->smem_len);
    if (!ljm_dev->smem_base) {
        DRM_ERROR("%s: Cannot map bar 0 area.\n", __FUNCTION__);
    }
#else
    /* ioremap_wc() would meet with following error:
     * kernel: x86/PAT: insmod:1771 conflicting memory types c0000000-d0000000 write-combining<->uncached-minus
     * kernel: x86/PAT: reserve_memtype failed [mem 0xc0000000-0xcfffffff], track uncached-minus, req uncached-minus
     * kernel: ioremap reserve_memtype failed -16

     * ioremap/ioremap_nocache() would meet with following error:
     * kernel: x86/PAT: insmod:1152 conflicting memory types c0000000-d0000000 write-combining<->uncached-minus
     * kernel: x86/PAT: reserve_memtype failed [mem 0xc0000000-0xcfffffff], track uncached-minus, req uncached-minus
     * kernel: ioremap reserve_memtype failed -16
     */
#endif
    TRACE("mmio_start=%llx mmio_len=%zx mmio_base=%lx smem_start=%llx smem_len=%zx smem_base=%lx",
          ljm_dev->mmio_start, ljm_dev->mmio_len, (unsigned long)ljm_dev->mmio_base,
          ljm_dev->smem_start, ljm_dev->smem_len, (unsigned long)ljm_dev->smem_base
          );

    /* Is GP a GP101 FPGA ? */
    if (LjmIsFpga(flags)) {
        DRM_WARN("GPU is an FPGA version\n");

        /* TODO: FPGA spectial functions */
        is_fpga = true;
    }

    ljm_dev->dp_irq_is_extra = false;

    /* get original power state from powerManage argument of cores */
    if (ljmicro_get_args_power_management())
        ljm_dev->power_state = LJM_POWER_STATE_SAVE;
    else
        ljm_dev->power_state = LJM_POWER_STATE_PERFORMANCE;

    r = ljm_hard_init(ljm_dev);

#ifdef LJMICRO_DRV_ONLY
    r = ljm_ddr_test(ljm_dev->smem_base, ljm_dev->smem_start, ljm_dev->smem_len);
    if (r) {
        DRM_ERROR("VRAM test error\n");
        goto OnERROR;
    } else {
        DRM_INFO("VRAM test success\n");
    }
#endif

    /* now hardware functions could be called correctly */

    FOOTER();
    return r;

OnERROR:
    if (ljm_dev) {
        if (ljm_dev->mmio_base) {
            iounmap(ljm_dev->mmio_base);
            ljm_dev->mmio_base = NULL;
            g_mmio_bases[ljm_dev->cur_card] = NULL;
            g_ljm_devs[ljm_dev->cur_card] = NULL;
            if (0 == ljm_dev->cur_card) {
                g_mmio_base = NULL;
                g_ljm_dev = NULL;
            }
        }
        kfree(ljm_dev);
        ljm_dev = NULL;
        SetLjmDevFrmDrm(ddev, ljm_dev);
    }
    FOOTER_ARG("r=%d", r);
    return r;
}

/**
 * ljm_device_fini - tear down the driver
 *
 * @ljm_dev: ljm_device pointer
 *
 * Tear down the driver info (all asics).
 * Called at driver shutdown.
 */
void ljm_device_fini(struct ljm_device *ljm_dev)
{
    int i;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    if (ljm_dev) {
        /* free i2_devs */
        for (i=0; i<ljm_dev->i2c_dev_cnt; i++) {
            if (ljm_dev->i2c_devs[i] != NULL) {
                kfree(ljm_dev->i2c_devs[i]);
                ljm_dev->i2c_devs[i] = NULL;
            }
        }

        for (i=0; i<LJM_MAX_PORTS; i++) {
            if (ljm_dev->ports[i] != NULL) {
                kfree(ljm_dev->ports[i]);
                ljm_dev->ports[i] = NULL;
            }
        }

#if 0
        /* iounmap fix space */
        if (!ljm_dev->smem_base && ljm_dev->fix_base) {
            iounmap(ljm_dev->fix_base);
            ljm_dev->fix_base = NULL;
        }
#endif

#if DYNAMIC_GART
        /* iounmap gart space */
        if (!ljm_dev->smem_base && ljm_dev->gart_base) {
            iounmap(ljm_dev->gart_base);
            ljm_dev->gart_base = NULL;
        }
#endif

        /* iounmap register memory space */
        if(ljm_dev->smem_base) {
            iounmap(ljm_dev->smem_base);
            ljm_dev->smem_base = NULL;
//          release_mem_region(ljm_dev->smem_start, ljm_dev->smem_len);
        }

#if 0   /* 关机时还需要调drv_exit()函数, 暂不iounmap */
        /* iounmap register memory space */
        if(ljm_dev->mmio_base) {
            iounmap(ljm_dev->mmio_base);
            ljm_dev->mmio_base = NULL;
//          release_mem_region(ljm_dev->mmio_start, ljm_dev->mmio_len);
        }
#endif
        ljm_gem_fini(ljm_dev);
        ljm_pwm_fan_exit(ljm_dev);
        ljm_pm_exit(ljm_dev);

        /* free drm_info*/
        kfree(ljm_dev->drm_info);
        ljm_dev->drm_info = NULL;
    }

    FOOTER();
}

/*
 * Suspend & resume.
 */

int ljm_suspend_vga(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct drm_connector *connector;
    int r;

    HEADER_ARG("drm=%p", ddev);

#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    /* need to be fixed */
#else
    drm_kms_helper_poll_disable(ddev);
#endif

    /*
     * suspend DC
     */
    ljm_dev->suspend = true;

    r = ljm_hard_suspend(ljm_dev, ddev);
    if (r) {
        printk(KERN_ERR "\ndisplay suspend error\n");
        FOOTER();
        return r;
    }

    drm_modeset_lock_all(ddev);
    /* turn off display hw */
    list_for_each_entry(connector, &ddev->mode_config.connector_list, head) {
        drm_helper_connector_dpms(connector, DRM_MODE_DPMS_OFF);
    }
    drm_modeset_unlock_all(ddev);

    /* unpin the front buffers and cursors */
    /* FIXME: todo */

#ifdef LJMICRO_DRV_ONLY
    /* backup vram
     * 1GB可见显存保存成功，但是2GB可见显存保存会出错
     */
    TRACE("save visible vram...");
    ljm_dev->save_external_base = vmalloc(ljm_dev->smem_len);
    if (ljm_dev->save_external_base == NULL) {
        printk(KERN_ERR "\nout of memory for external vram backup\n");
        //FOOTER_ARG("r=%d", -ENOMEM);
        //return -ENOMEM;
        r = -ENOMEM;
    } else {
        void *vram_base;

        if (ljm_dev->smem_base)
            vram_base = (char *)ljm_dev->smem_base;
        else
            vram_base = (char *)ioremap_wc(ljm_dev->smem_start, ljm_dev->smem_len);
        //printk(KERN_ERR "----2----display off----\n");
        if (vram_base) {
            /* FIXME: 如果bar0为2GB空间，执行2GB的memcpy会出错：INFO: rcu_sched self-detected stall on CPU */
            memcpy(ljm_dev->save_external_base, vram_base, ljm_dev->smem_len);

            if (!ljm_dev->smem_base)
                iounmap(vram_base);
            //printk(KERN_ERR "----3----display off----\n");
        }
        else {
            printk(KERN_ERR "\nioremap_wc external vram failed\n");
//            FOOTER_ARG("r=%d", -EIO);
//            return -EIO;
            r = -EIO;
        }
    }
    TRACE("save visible vram done(r=%d)", r);
#endif

    FOOTER();
    return r;
}

/**
 * ljm_suspend_kms - initiate device suspend
 *
 * @pdev: drm dev pointer
 * @state: suspend state
 *
 * Puts the hw in the suspend state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver suspend.
 */
int ljm_suspend_kms(struct drm_device *ddev, bool suspend, bool fbcon)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct pci_dev *pdev = ljm_dev->pdev;
//    struct drm_crtc *crtc;
    struct drm_connector *connector;
    int r;

    HEADER_ARG("drm=%p suspend=%d fbcon=%d",
        ddev, suspend, fbcon);

    if (ddev == NULL || ddev->dev_private == NULL) {
        FOOTER_ARG("r=%d", -ENODEV);
        return -ENODEV;
    }

    if (ddev->switch_power_state == DRM_SWITCH_POWER_OFF) {
        FOOTER();
        return 0;
    }

#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    /* need to be fixed */
#else
    drm_kms_helper_poll_disable(ddev);
#endif

    drm_modeset_lock_all(ddev);
    /* turn off display hw */
    list_for_each_entry(connector, &ddev->mode_config.connector_list, head) {
        drm_helper_connector_dpms(connector, DRM_MODE_DPMS_OFF);
    }
    drm_modeset_unlock_all(ddev);

    /*
     * suspend DC
     */
    r = ljm_suspend_vga(ddev);
    ljm_hpd_fini(ljm_dev);
#if 0
    /* unpin the front buffers and cursors */
    list_for_each_entry(crtc, &ddev->mode_config.crtc_list, head) {
        struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);
        struct drm_framebuffer *fb = crtc->primary->fb;
        struct ljm_gem_object *ljm_obj;

        if (ljm_crtc->cursor_obj) {
            struct ljm_gem_object *ljm_obj = to_ljm_gem(ljm_crtc->cursor_obj);
            ljm_put_vram(ddev, ljm_obj);    /* free curor bo */
        }

        if (fb == NULL || fb->obj[0] == NULL) {
            continue;
        }
        ljm_obj = to_ljm_gem(fb->obj[0]);
        /* don't unpin kernel fb objects */
        if (!ljm_fbdev_bo_is_fb(ljm_dev, ljm_obj)) {
            ljm_put_vram(ddev, ljm_obj);    /* free bo except framebuffer's */
        }
    }
#endif
    pci_save_state(pdev);

    if (suspend) {
        /* Shut down the device */
        pci_disable_device(pdev);
        pci_set_power_state(pdev, PCI_D3hot);
    }

    if (fbcon) {
        console_lock();
        if(ljm_dev->fbdev)
            drm_fb_helper_set_suspend(&ljm_dev->fbdev->helper, 1);
        console_unlock();
    }

    FOOTER();
    return 0;
}

void ljm_crtc_pre_resume(struct ljm_device *ljm_dev)
{
    ljm_dev->suspend = false;

#if 0
    /* delay a second to enable display */
    del_timer_sync(&ljm_dev->resume_timer);
    ljm_dev->resume_timer.expires = jiffies + msecs_to_jiffies(1000);
    add_timer(&ljm_dev->resume_timer);
#endif
}

int ljm_resume_vga(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
#ifdef LJMICRO_DRV_ONLY
    void *vram_base;
#endif
    int r = 0;

    HEADER_ARG("ddev=%p", ddev);

#ifdef LJMICRO_DRV_ONLY
    /* restore whole visible vram */
    TRACE("restore visible vram...");
    if (ljm_dev->smem_base)
        vram_base = ljm_dev->smem_base;
    else
        vram_base = ioremap_wc(ljm_dev->smem_start, ljm_dev->smem_len);

    if (vram_base) {
        if (likely(ljm_dev->save_external_base))
            memcpy(vram_base, ljm_dev->save_external_base, ljm_dev->smem_len);
        else {
            printk(KERN_ERR "visible vram is not saved\n");
            r = -EIO;
        }
    } else {
        printk(KERN_ERR "ioremap_wc external vram failed\n");

        if (!ljm_dev->smem_base)
            iounmap(vram_base);

        //FOOTER_ARG("r=%d", -EIO); msleep(1000);
        //return -EIO;
        r = -EIO;
    }
    if (!ljm_dev->smem_base)
        iounmap(vram_base);

    if (ljm_dev->save_external_base)
        vfree(ljm_dev->save_external_base);

    TRACE("restore visible vram done(%d)", r);
#endif

    /*
     * resume DC
     */
    /* hardware resume: ddr init, gart init (if there is), vga init, etc. */
    ljm_hard_resume(ljm_dev, ddev);

    /* resume hpd */
    ljm_hpd_init(ljm_dev);

    ljm_dev->suspend = false;    /* end suspend state */

    FOOTER();
    return r;
}


/**
 * ljm_resume_kms - initiate device resume
 *
 * @pdev: drm dev pointer
 *
 * Bring the hw back to operating state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver resume.
 */
int ljm_resume_kms(struct drm_device *ddev, bool resume, bool fbcon)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct pci_dev *pdev = ljm_dev->pdev;
    struct drm_connector *connector;
    struct drm_crtc *crtc;
    int r;

    HEADER_ARG("drm=%p resume=%d fbcon=%d",
        ddev, resume, fbcon);

    if (ddev->switch_power_state == DRM_SWITCH_POWER_OFF) {
        FOOTER();
        return 0;
    }

    //printk(KERN_ERR "----1----\n");msleep(1000);

//    if (fbcon) {
//        console_lock();   // hang when resume
//    }

    //printk(KERN_ERR "----2----\n");msleep(1000);

    if (resume) {
        //printk(KERN_ERR "----2.1----\n");msleep(1000);
        pci_set_power_state(pdev, PCI_D0);
        //printk(KERN_ERR "----2.2----\n");msleep(1000);
        pci_restore_state(pdev);
        //printk(KERN_ERR "----2.3----\n");msleep(1000);
        if (pci_enable_device(pdev)) {
            if (fbcon)
                console_unlock();

            FOOTER();
            return -1;
        }
        //printk(KERN_ERR "----2.4----\n");msleep(1000);
    }

    //printk(KERN_ERR "----3----\n");msleep(1000);

    /* pin cursors */
    list_for_each_entry(crtc, &ddev->mode_config.crtc_list, head) {
        struct ljm_crtc *ljm_crtc = to_ljm_crtc(crtc);

        if (ljm_crtc->cursor_obj) {
            struct ljm_gem_object *ljm_obj = to_ljm_gem(ljm_crtc->cursor_obj);
            r = ljm_get_vram(ddev, ljm_obj);
            if (r != 0)
                DRM_ERROR("Failed to pin cursor BO (%d)\n", r);
        }
    }

    //printk(KERN_ERR "----4----\n");msleep(1000);

    ljm_resume_vga(ddev);

    //printk(KERN_ERR "----5----\n");msleep(1000);

    /* reset hpd state */
    ljm_hpd_init(ljm_dev);

    //printk(KERN_ERR "----6----\n");msleep(1000);

    /* blat the mode back in */
    if (fbcon) {
        drm_helper_resume_force_mode(ddev);
        /* turn on display hw */
        drm_modeset_lock_all(ddev);
        list_for_each_entry(connector, &ddev->mode_config.connector_list, head) {
            drm_helper_connector_dpms(connector, DRM_MODE_DPMS_ON);
        }
        drm_modeset_unlock_all(ddev);
    }

    //printk(KERN_ERR "----7----\n");msleep(1000);

#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    /* need to be fixed */
#else
    drm_kms_helper_poll_enable(ddev);
#endif

//    if (fbcon) {
//        if(ljm_dev->fbdev) {
//            drm_fb_helper_set_suspend(&ljm_dev->fbdev->helper, 0);
//        }
//        console_unlock();
//    }

    /* delay to enable crtc to avoid blurred screen */
    //ljm_crtc_pre_resume(ljm_dev);

    FOOTER();
    return 0;
}

/****************************** for ljmicro_util **********************************/
unsigned long get_mmio_base(int card)
{
    //struct ljm_device *ljm_dev = get_ljm_devs(card);

    //return (unsigned long)ljm_dev->mmio_base;
    return (unsigned long)g_mmio_bases[card];
}

struct pci_dev *get_drm_pdev(int card)
{
    struct ljm_device *ljm_dev = get_ljm_devs(card);
    return ljm_dev->pdev;
}

unsigned long get_vram_start(int card)
{
    struct ljm_device *ljm_dev = get_ljm_devs(card);

    return (unsigned long)ljm_dev->smem_start;
}

unsigned long get_vram_base(int card)
{
    struct ljm_device *ljm_dev = get_ljm_devs(card);

    return (unsigned long)ljm_dev->smem_base;
}


unsigned long get_vram_len(int card)
{
    struct ljm_device *ljm_dev = get_ljm_devs(card);

    return (unsigned long)ljm_dev->smem_len;
}

/*****************************************************************************
*  函 数 名: get_ljm_connector
*  功能描述: 根据显示接口index号获取connector结构体指针
*  输入参数: int card   显卡序号
*            int index  接口序号
*  输出参数: 无
*  返 回 值: 如果接口存在，则返回connector结构体指针；否则返回NULL
*
*  修改历史:
*         1. 日    期: 2022年4月24日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
struct drm_connector *get_ljm_connector(int card, int index)
{
    struct ljm_device *ljm_dev = get_ljm_devs(card);

    if (index < 0 || index >= ljm_dev->num_ports) {
        printk(KERN_ERR "connector%d is out of range\n", index);
        return NULL;
    }
    else
        return &ljm_dev->ports[index]->vdev_connector->base;
}

#ifdef LJMICRO_DRV_ONLY
int gp_meminfo_get(int card, struct ljm_ioctl_info *pgpu_info)
{

    struct ljm_device *ljm_dev = get_ljm_devs(card);

    HEADER();

    pgpu_info->vram_total = ljm_dev->smem_len / 1024 / 1024;
    pgpu_info->vram_free = ljm_dev->vram_free.vram_size / 1024 / 1024;
    pgpu_info->vram_used = pgpu_info->vram_total - pgpu_info->vram_free;

    FOOTER();
    return 0;
}
#endif

void ljm_crtc_reg_write(int card, uint8_t ind, uint8_t val)
{
    uint16_t x;

    x = ((uint16_t)val << 8) | (uint16_t)ind;
    writew(x, g_mmio_bases[card] + VGAREG_VGA_CRTC_ADDRESS);

    return;
}

uint8_t ljm_crtc_reg_read(int card, uint8_t ind)
{
    uint8_t val;
    unsigned long flags = 0;
    struct ljm_device *ljm_dev = g_ljm_devs[card];

    if (in_irq())
        spin_lock(&ljm_dev->CR_RD_lock);
    else
        spin_lock_irqsave(&ljm_dev->CR_RD_lock, flags);

    writeb((u8)(ind), g_mmio_bases[card] + VGAREG_VGA_CRTC_ADDRESS);
    val = readb(g_mmio_bases[card] + VGAREG_VGA_CRTC_DATA);

    if (in_irq())
        spin_unlock(&ljm_dev->CR_RD_lock);
    else
        spin_unlock_irqrestore(&ljm_dev->CR_RD_lock, flags);

    return val;
}

uint8_t ljm_crtc_reg_nolock_read(int card, uint8_t ind)
{
    uint8_t val;

    writeb((u8)(ind), g_mmio_bases[card] + VGAREG_VGA_CRTC_ADDRESS);
    val = readb(g_mmio_bases[card] + VGAREG_VGA_CRTC_DATA);

    return val;
}

int ljm_set_power(struct ljm_device *ljm_dev, int power_state)
{
    return ljm_power_set_state(ljm_dev, power_state);
}
