/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_DEVICE_H__
#define __LJMICRO_DEVICE_H__

#include <linux/version.h>
#include <drm/drm_fb_helper.h>
#ifdef DELAY_4K_ENABLE
#include <linux/timer.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
#include <drm/drm_drv.h>
#include <drm/drm_file.h>
#include <drm/drm_ioctl.h>
#include <drm/drm_print.h>
#include <drm/drm_vblank.h>
#include <drm/drm_framebuffer.h>
#include <drm/drm_fourcc.h>
#include <drm/drm_modes.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include <drm/drm_irq.h>
#endif
#else
#include <drm/drmP.h>
#endif

#include <drm/drm_gem.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,4,0)
#include <drm/drm_probe_helper.h>
#endif
#include <drm/drm_encoder.h>
#include "ljmicro_family.h"
#include "ljmicro_base.h"
#include "ljmicro_compat.h"
#include "ljmicro_ioctl.h"
#include "ljmicro_hdmi.h"
#include "ljmicro_i2c.h"
#if USE_DP_CONNECTOR
#include "ljmicro_dp.h"
#endif
#include "ljmicro_i2s.h"
#include "ljmicro_plane.h"
#include "ljmicro_zoomvideo.h"
#include "ljmicro_mcu.h"
#include "ljmicro_fan.h"

/* FIXME: if vblank is enabled, should set MODE_SET_BASE_HDMI_PWR_OFF 0 ? */
#define MODE_SET_BASE_HDMI_PWR_OFF  1

/* DYNAMIC_GART=1 means dynamic gart system */
#ifndef DYNAMIC_GART
#define DYNAMIC_GART        0
#endif

/* LINEAR_GART=1 means linear gart system for x86 and old 32bit bus fpga  */
#ifndef LINEAR_GART
#define LINEAR_GART         0
#endif

#ifndef NOEDID
#define NOEDID              0
#endif

#ifndef USE_KMS_VBLANK
#define USE_KMS_VBLANK      1
#endif

#ifndef USE_DP_CONNECTOR
#define USE_DP_CONNECTOR    1
#endif

#ifndef USE_DAC_CONNECTOR
#define USE_DAC_CONNECTOR   1
#endif

#ifndef USE_HDMI_CONNECTOR
#define USE_HDMI_CONNECTOR  1
#endif

#ifndef USE_IN_SDT
#define USE_IN_SDT          0
#endif

#define CURSOR_WIDTH        64 //32
#define CURSOR_HEIGHT       64 //32

#define CURSOR_MAX_WIDTH    64
#define CURSOR_MAX_HEIGHT   64

#define LJM_MAX_VBLANK_CNT  0xffffffff

enum Proc_freq
{
     freq_axi = 0,
     freq_ahb,
     freq_apb,
     freq_pcie,
     freq_vpu,
     freq_3d,
     freq_2d,
     freq_ddr,
};

#ifdef LJMICRO_DRV_ONLY
struct ljm_gem_object {
    struct drm_gem_object   base;

    /* Protected by ljm_dev->gem.mutex */
    struct list_head    list;
    struct ljm_device   *ljm_dev;

    unsigned            pin_count;
    size_t              size;       /* 0 means memory not pinned */
    unsigned long       offset;     /* offset from video memory base */
    resource_size_t     gpu_addr;   /* physical address for gpu */
    resource_size_t     cpu_addr;   /* physica address for cpu */
    void                *kptr;      /* kernel virtual address */
};

#define get_cpu_addr_from_bo(bo)    (bo->cpu_addr)
#define get_size_from_bo(bo)        (bo->size)
#define get_offset_from_bo(bo)      (bo->offset)
#else
struct ljm_gem_object {
    struct drm_gem_object base;

    uint32_t              node_handle;
    ljkVIDMEM_NODE        node_object;
    ljtBOOL               cacheable;

    uint32_t              pid;
};

#define BO_IS_VIDMEM(BO)                (BO->node_object->node->VidMem.parent && BO->node_object->node->VidMem.parent->object.type == ljvOBJ_VIDMEM)
#define BO_IS_VIDMEM_BLOCK(BO)          (BO->node_object->node->VirtualChunk.parent && BO->node_object->node->VirtualChunk.parent->object.type == ljvOBJ_VIDMEM_BLOCK)
#define BO_GET_SIZE(BO)                 (BO_IS_VIDMEM_BLOCK(BO) ? BO->node_object->node->VirtualChunk.bytes : BO->node_object->node->VidMem.bytes)
//#define BO_GET_VIDMEM_OFFSET(BO)        BO->node_object->node->VidMem.offset
//#define BO_GET_VIDMEM_START(BO)         BO->node_object->node->VidMem.parent->physicalBase
#define BO_GET_VIDMEM_START(BO)         (BO->node_object->node->VidMem.parent->physicalBase)
#define BO_GET_VIDMEM_OFFSET(BO)        ((BO->node_object->pool == ljvPOOL_LOCAL_EXCLUSIVE) ? \
                                            (BO->node_object->node->VidMem.offset + BO_GET_VIDMEM_START(BO) - VRAM_GP200_BUS40_GPU_START) : \
                                            BO->node_object->node->VidMem.offset)
#define BO_GET_VIRTUAL_PAGECOUNT(BO)    BO->node_object->node->Virtual.pageCount

#define get_cpu_addr_from_bo(bo)        (BO_GET_VIDMEM_START(bo) + BO_GET_VIDMEM_OFFSET(bo))
#define get_size_from_bo(bo)            BO_GET_SIZE(bo)
#define get_offset_from_bo(bo)          BO_GET_VIDMEM_OFFSET(bo)
#endif

/*
 * GEM objects.
 */
struct ljm_gem {
    struct mutex        mutex;
    struct list_head    objects;
};

#ifdef LJMICRO_DRV_ONLY
/* free memory manager */
struct ljm_mm_obj {
    struct list_head list;      /* free memory block list */
    unsigned long offset;       /* physical address offset from total vram_start */
    size_t size;                /* size of this free memory block */
};

struct ljm_mm_list {
    struct list_head    list;           /* free vram block list */
    unsigned long       vram_start;     /* gpu vram start address, 0x40000000 for GP100 */
    size_t              vram_size;      /* vram size */
    struct mutex        mutex;
};
#endif

#define FB_INFO_NAME            "LJMDRMFB"  /* frame buffer info name */

#if 0
/* framebuffer object for GP device */
struct ljm_framebuffer {
    struct drm_framebuffer base;
    struct drm_gem_object* obj;
};
#define to_ljm_framebuffer(x) container_of(x, struct ljm_framebuffer, base)
#endif

struct ljm_fbdev {
    struct drm_fb_helper helper;
    struct drm_framebuffer fb;
//  struct list_head fbdev_list;
//  struct drm_display_mode *our_mode;
//  spinlock_t dirty_lock;
};

struct ljm_irq {
    bool                installed;
    spinlock_t          lock;
    bool                crtc_vblank_int[LJM_MAX_CRTCS];
    atomic_t            pflip[LJM_MAX_CRTCS];
    bool                crtc_vblank_int_en[LJM_MAX_CRTCS];  /* GP101采用vblank物理中断生成帧计数，crtc_vblank_int_en表示vblank功能是否开启 */
    wait_queue_head_t   vblank_queue;
};

struct ljm_disp_attr {
    int max_hdisplay;
    int max_vdisplay;
    int min_hdisplay;
    int min_vdisplay;
    int default_hdisplay;
    int default_vdisplay;
    int max_pclk;
    int max_dvi_hdisplay;
    int max_dvi_vdisplay;
    int max_dvi_pclk;
};

struct ljm_asic;

enum ljm_port_type
{
    LJM_PORT_UNKNOWN = -1,
    LJM_PORT_CLOSE = 0,
    LJM_PORT_VGA,
    LJM_PORT_HDMI,
    LJM_PORT_DVO,
    LJM_PORT_DP,
};

struct vga_dev
{
    struct i2c_dev *i2c_dev;
    bool hw_connected;

    /* dac channel set*/
    u8 dac_ch_red;
    u8 dac_ch_green;
    u8 dac_ch_blue;
};

struct dvo_dev
{
    struct i2c_dev *i2c_dev;
    bool is_dvi;

    struct drm_property *broadcast_rgb_property;
};

enum ljm_encoder_hard_type {
    LJM_ENCODER_NONE = 0,
    LJM_ENCODER_DAC,
    LJM_ENCODER_HDMIV1,
    LJM_ENCODER_HDMIV2,
    LJM_ENCODER_DVO,
    LJM_ENCODER_DP
};

struct ljm_encoder {
    struct drm_encoder base;
    uint32_t id;    /* interface id. For GP200 series, DAC: 0~3, HDMI: 4~7, DVO: 8~11, DP: 12~15
                     * For GP100 series, HDMI: 0~1, DAC: 2~3 */

    enum ljm_encoder_hard_type hard_type;       /* hardware type */

    bool once_edp_modeset;
    int once_edp_crtc;    /* EDP接口记录VBIOS阶段绑定的CRTC ID */
    struct drm_display_mode *once_edp_mode;
};

enum ljm_hpd_id {
    LJM_HPD_0 = 0,
    LJM_HPD_1,
    LJM_HPD_2,
    LJM_HPD_3,
    LJM_HPD_4,
    LJM_HPD_5,
    LJM_HPD_6,
    LJM_HPD_7,
    LJM_HPD_8,
    LJM_HPD_9,
    LJM_HPD_10,
    LJM_HPD_11,
    LJM_HPD_12,
    LJM_HPD_13,
    LJM_HPD_14,
    LJM_HPD_15,
    LJM_HPD_NONE = 0xff,
};

enum ljm_hpd_state {
    LJM_HPD_STATE_OUT = 0,
    LJM_HPD_STATE_IN,
    LJM_HPD_STATE_UNKOWN = 0xff,
};

struct ljm_hpd {
    enum ljm_hpd_id hpd;
    enum ljm_hpd_state plugged_state;
};

enum ljm_ddc_type {
    LJM_DDC_TYPE_NONE = 0,
    LJM_DDC_TYPE_DAC,
    LJM_DDC_TYPE_HDMIV1,
    LJM_DDC_TYPE_HDMIV2,
    LJM_DDC_TYPE_DP,
    LJM_DDC_TYPE_OCI2C,
    LJM_DDC_TYPE_DWI2C
};

enum ljm_hpd_type {
    LJM_HPD_TYPE_NONE = 0,
    LJM_HPD_TYPE_DAC,
    LJM_HPD_TYPE_HDMIV1,
    LJM_HPD_TYPE_HDMIV2,
    LJM_HPD_TYPE_CRTC,
    LJM_HPD_TYPE_DP,
    LJM_HPD_TYPE_EIRQ
};

struct ljm_crtc_state {
    struct drm_crtc_state base;
};

/* "Broadcast RGB" property */
#define LJM_BROADCAST_RGB_AUTO 0
#define LJM_BROADCAST_RGB_FULL 1
#define LJM_BROADCAST_RGB_LIMITED 2

enum pipe {
    INVALID_PIPE = -1,

    PIPE_A = 0,
    PIPE_B,
    PIPE_C,
    _PIPE_EDP,
};

struct ljm_connector;
struct ljm_port;

struct ljm_panel {
    /* backlight */
    struct {
        bool present;
        u32 level;
        u32 min;
        u32 max;
        bool enabled;
        /* Connector and platform specific backlight functions */
        int (*setup)(struct ljm_port *ljm_port/*, enum pipe pipe*/);
        void (*disable)(struct ljm_port *ljm_port, const struct drm_connector_state *conn_state);
        void (*enable)(const struct ljm_crtc_state *crtc_state, const struct drm_connector_state *conn_state, struct ljm_port *ljm_port);
    } backlight;
};

struct ljm_connector {
    struct drm_connector base;
    uint32_t id;    /* interface id. For GP200 series, DAC: 0~3, HDMI: 4~7, DVO: 8~11, DP: 12~15
                     * For GP100 series, HDMI: 0~1, DAC: 2~3 */

    /*
     * The fixed encoder this connector is connected to.
     */
    struct ljm_encoder *encoder;

    struct edid * edid;
    bool edid_dirty;
    struct semaphore edid_sema; /* hdmi edid synchronization */

    struct ljm_hpd hpd; /* 热插拔状态 */

    enum ljm_ddc_type ddc_type;     /* ddc类型，硬件初始化时设置 */

    enum ljm_hpd_type hpd_type;     /* 热插拔类型，硬件初始化时设置 */

    /* DP */
    struct ljm_panel panel; /* Panel info for eDP and LVDS */
};

struct ljm_crtc_mode {
    /**
    * @clock:
    *
    * Pixel clock in kHz.
    */
    int clock;      /* in kHz */
    int hdisplay;
    int hsync_start;
    int hsync_end;
    int htotal;
    int vdisplay;
    int vsync_start;
    int vsync_end;
    int vtotal;
    /**
    * @type:
    *
    * A bitmask of flags, mostly about the source of a mode. Possible flags
    * are:
    *
    *  - DRM_MODE_TYPE_BUILTIN: Meant for hard-coded modes, effectively
    *    unused.
    *  - DRM_MODE_TYPE_PREFERRED: Preferred mode, usually the native
    *    resolution of an LCD panel. There should only be one preferred
    *    mode per connector at any given time.
    *  - DRM_MODE_TYPE_DRIVER: Mode created by the driver, which is all of
    *    them really. Drivers must set this bit for all modes they create
    *    and expose to userspace.
    *
    * Plus a big list of flags which shouldn't be used at all, but are
    * still around since these flags are also used in the userspace ABI:
    *
    *  - DRM_MODE_TYPE_DEFAULT: Again a leftover, use
    *    DRM_MODE_TYPE_PREFERRED instead.
    *  - DRM_MODE_TYPE_CLOCK_C and DRM_MODE_TYPE_CRTC_C: Define leftovers
    *    which are stuck around for hysterical raisins only. No one has an
    *    idea what they were meant for. Don't use.
    *  - DRM_MODE_TYPE_USERDEF: Mode defined by userspace, again a vestige
    *    from older kms designs where userspace had to first add a custom
    *    mode to the kernel's mode list before it could use it. Don't use.
    */
    unsigned int type;
    /**
    * @flags:
    *
    * Sync and timing flags:
    *
    *  - DRM_MODE_FLAG_PHSYNC: horizontal sync is active high.
    *  - DRM_MODE_FLAG_NHSYNC: horizontal sync is active low.
    *  - DRM_MODE_FLAG_PVSYNC: vertical sync is active high.
    *  - DRM_MODE_FLAG_NVSYNC: vertical sync is active low.
    *  - DRM_MODE_FLAG_INTERLACE: mode is interlaced.
    *  - DRM_MODE_FLAG_DBLSCAN: mode uses doublescan.
    *  - DRM_MODE_FLAG_CSYNC: mode uses composite sync.
    *  - DRM_MODE_FLAG_PCSYNC: composite sync is active high.
    *  - DRM_MODE_FLAG_NCSYNC: composite sync is active low.
    *  - DRM_MODE_FLAG_HSKEW: hskew provided (not used?).
    *  - DRM_MODE_FLAG_BCAST: not used?
    *  - DRM_MODE_FLAG_PIXMUX: not used?
    *  - DRM_MODE_FLAG_DBLCLK: double-clocked mode.
    *  - DRM_MODE_FLAG_CLKDIV2: half-clocked mode.
    *
    * Additionally there's flags to specify how 3D modes are packed:
    *
    *  - DRM_MODE_FLAG_3D_NONE: normal, non-3D mode.
    *  - DRM_MODE_FLAG_3D_FRAME_PACKING: 2 full frames for left and right.
    *  - DRM_MODE_FLAG_3D_FIELD_ALTERNATIVE: interleaved like fields.
    *  - DRM_MODE_FLAG_3D_LINE_ALTERNATIVE: interleaved lines.
    *  - DRM_MODE_FLAG_3D_SIDE_BY_SIDE_FULL: side-by-side full frames.
    *  - DRM_MODE_FLAG_3D_L_DEPTH: ?
    *  - DRM_MODE_FLAG_3D_L_DEPTH_GFX_GFX_DEPTH: ?
    *  - DRM_MODE_FLAG_3D_TOP_AND_BOTTOM: frame split into top and bottom
    *    parts.
    *  - DRM_MODE_FLAG_3D_SIDE_BY_SIDE_HALF: frame split into left and
    *    right parts.
    */
    unsigned int flags;
};

enum ljm_flip_status {
    LJM_FLIP_NONE,
    LJM_FLIP_PENDING,
    LJM_FLIP_SUBMITTED
};

struct ljm_flip_work {
    struct work_struct  flip_work;
    struct work_struct  unpin_work;
    struct ljm_device   *ljm_dev;
    int                 crtc;
    uint64_t            base;
    struct drm_pending_vblank_event *event;
    struct ljm_gem_object *old_bo;
//  struct fence        *fence;
    bool                async;
};

/**
 * struct ljm_lut - 保存CRTC颜色查找表数据(Look-Up-Table)，用于色温亮度调节
 * 
 * @gamma_changed: lut是否发生改变；
 * @red: 红色lut数组指针；
 * @green: 绿色lut数组指针；
 * @blue: 蓝色lut数组指针；
 * @gamma_size: 单个色彩的lut长度，通常为256。
*/
struct ljm_lut 
{
    bool lut_changed;
    uint16_t *red;
    uint16_t *green;
    uint16_t *blue;
    uint32_t gamma_size;
};

struct ljm_crtc {
    struct drm_crtc base;
    int id;
    bool enabled;

//  struct drm_encoder *encoder;
//  struct drm_connector *connector;

    /* cursor */
    int cursor_x;
    int cursor_y;
    int cursor_hot_x;
    int cursor_hot_y;
    int cursor_width;
    int cursor_height;
    int max_cursor_width;
    int max_cursor_height;
    bool cursor_hw_show;

    int cursor_size;
    uint32_t cursor_addr;
    struct drm_gem_object *cursor_obj;

    uint32_t vram_offset;           /* vram offset for framebuffer of the crtc */

    struct semaphore vblank_sema;   /* semaphore for vblank */

//  struct timer_list hpd_timer;    /* timer to delay to enable hpd interrupt */
//  struct delayed_work hpd_work;   /* delayed_work after hpd_timer callback */
//  struct timer_list enable_timer; /* timer to delay to enable hpd interrupt */
#ifdef DELAY_4K_DISPLAY
    struct timer_list enable_timer; /* timer to delay to enable crtc display */
#endif

#if MODE_SET_BASE_HDMI_PWR_OFF
    unsigned long stamp_dpms_off;   /* record stamp at dpms off */
#endif

    bool save_crtc_on;  /* crtc status is on or off while suspend */

//  struct ljm_crtc_mode csmmode;
//  bool changed_to_vesa_mode;

    /* page flipping */
    struct workqueue_struct *flip_queue;
    struct ljm_flip_work *flip_work;
    enum ljm_flip_status flip_status;
    u64 pflip_base; /* new flipping base address for crtc */

    struct drm_display_mode *once_edp_mode;
    struct ljm_lut lut;

};

struct ljm_port
{
    struct ljm_encoder base;

    int id;
    char name[20];

    struct ljm_port_info *info;

    enum ljm_port_type type;
    union
    {
        struct vga_dev vga;
        struct dvo_dev dvo;
        struct hdmi_dev hdmi;
#if USE_DP_CONNECTOR
        struct dp_dev dp;
#endif
    }vdev;

    struct ljm_connector *vdev_connector;

    bool hpd_en;
};

#include "ljmicro_info.h"
/* GP device private data */
struct ljm_device {
    struct device *dev;
    struct drm_device *ddev;
    struct pci_dev *pdev;

    enum ljm_family family;
    unsigned int flags;
    struct ljm_drm_info *drm_info;

    int cur_card;     /* current card index */

    struct ljm_asic *asic;

    /* diaplay attribute */
    struct ljm_disp_attr disp_attr;

    bool mode_config_initialized;

    struct rw_semaphore exclusive_lock; /* used for drm ioctrl sync */

#ifdef LJMICRO_DRV_ONLY
    struct ljm_gem      gem;
    struct ljm_mm_list  vram_free;  /* free (not pinned) VRAM list */
#endif

    resource_size_t     smem_start; /* physical VRAM start address of GP. PCIE bar 0 */
    size_t              smem_len;   /* VGA (framebuffer) RAM size for GP. */
    void __iomem *      smem_base;  /* kernel virtual address of GP. PCIE bar 0 */

#if DYNAMIC_GART
    void __iomem *      gart_base;  /* mapped address for GART table region */
#else
//  void __iomem *      gart_base;  /* mapped address for GART table region */
#endif
//    void __iomem *      fix_base;   /* mapped address for FIX region */

    resource_size_t     mmio_start; /* start of physical address for GP register memory. PCIE bar 2 */
    size_t              mmio_len;   /* memory size for GP register memory */
    void __iomem *      mmio_base;  /* mapped address for GP register memory */

    struct ljm_fbdev *      fbdev;

    unsigned int            num_ports;
    struct ljm_port *       ports[LJM_MAX_PORTS];

    unsigned int            num_crtcs;
    struct drm_crtc *       crtcs[LJM_MAX_CRTCS];

    unsigned int            num_planes;
    struct ljm_plane *      planes[LJM_MAX_LAYERS];

    unsigned int            num_zvideos;
    struct ljm_zoomvideo *  zvideos[LJM_MAX_ZVIDEOS];

    /* port type: NONE (not connected), HDMI, DVI, both HDMI and DVI, and command type: NONE, HDMI read EDID */
    unsigned int            connector_flag[LJM_MAX_CONNECTORS];

    bool                    curor_en[LJM_MAX_CONNECTORS];   /* cursor */

    struct ljm_irq      irq;
    int msi_enabled; /* msi enabled */

    struct delayed_work hotplug_work;

//    struct timer_list   resume_timer;
    bool                suspend;    /* true: in resume state */
    void                *save_external_base;
    uint32_t            misc_intr_maskpcie;

#if 0   /* no need to clear new framebuffer */
    unsigned long       fb_bo_offset;   /* framebuffer bo offset */
    unsigned long       fb_bo_size;     /* framebuffer bo size */
#endif

    struct i2c_dev *i2c_devs[LJM_MAX_I2CS];
    int i2c_dev_cnt;

    bool dp_irq_is_extra ;
    struct drm_property *broadcast_rgb_property;

    //struct backlight_device *backlight_dev[LJM_MAX_BLACKLIGHTS];
    //int backlight_dev_cnt;

    struct ljm_i2s_dev *i2s_devs[LJM_MAX_I2SS];
    int axi_bus_clk; /* 计算i2s发送时钟 */
    struct hrtimer hrt; /* 高精度定时查询i2s的FIFO是否为空 */

    /*
     * 在一次寄存器读写中，对索引寄存器和数据寄存器的操作应是不可分割的，谨防中断导致读写值错误。
     * 设置自旋锁以避免上述情况出现，在寄存器读取时使用此锁。
     */
    spinlock_t CR_RD_lock;

    /* MCU设备指针 */
    struct ljm_mcu *mcu;

    /* 风扇设备指针 */
    struct ljm_fan *fan;

    bool     once_edp; //只对edp做一次配置，仅GP201使用此模式。
    uint32_t once_edp_vga_mux;
    bool     once_edp_vga_inited; //edp已在vbios阶段初始化过。

    /*
     * power state of LJM_POWER_STATE_PERFORMANCE, LJM_POWER_STATE_SAVE, LJM_POWER_STATE_STANDBY
     */
    uint32_t power_state;

    bool killMonitorThread;
    struct task_struct *monitorTask;
    struct mutex        hw_mutex;   /* access muxtex for hardware devices, such as PVT */
    struct mutex        i2c_mutex;  /* access muxtex for i2c devices */
    spinlock_t          cru_lock;

    bool is_shutdown;
};

#define to_ljm_crtc(x) container_of(x, struct ljm_crtc, base)
#define to_ljm_port(x) container_of(x, struct ljm_port, base)
#define to_ljm_encoder(x) container_of(x, struct ljm_encoder, base)
#define to_ljm_connector(x) container_of(x, struct ljm_connector, base)
#define to_ljm_fbdev(x) container_of(x, struct ljm_fbdev, helper)
#define to_ljm_gem(x) container_of(x, struct ljm_gem_object, base)

#define LJMFB_CONN_LIMIT            4
struct ljm_asic {
    int (*init)(struct ljm_device *ljm_dev);
    void (*fini)(struct ljm_device *ljm_dev);
    int (*resume)(struct drm_device *ddev);
    int (*suspend)(struct drm_device *ddev);
    int (*power_set_state)(struct ljm_device *ljm_dev, int power_state);
    void (*vga_set_state)(struct ljm_device *ljm_dev, bool state);
    int (*asic_reset)(struct ljm_device *ljm_dev, bool hard);
    /* Flush the HDP cache via MMIO */
    void (*mmio_hdp_flush)(struct ljm_device *ljm_dev);
    /* check if 3D engine is idle */
    bool (*gui_idle)(struct ljm_device *ljm_dev);
    /* wait for mc_idle */
    int (*mc_wait_for_idle)(struct ljm_device *ljm_dev);
    /* get the reference clock */
    u32 (*get_xclk)(struct ljm_device *ljm_dev);
    /* get the gpu clock counter */
    uint64_t (*get_gpu_clock_counter)(struct ljm_device *ljm_dev);
    /* get register for info ioctl */
    int (*get_allowed_info_register)(struct ljm_device *ljm_dev, u32 reg, u32 *val);
    /* gart */
    struct {
        void (*tlb_flush)(struct ljm_device *ljm_dev);
        uint64_t (*get_page_entry)(uint64_t addr, uint32_t flags);
        void (*set_page)(struct ljm_device *ljm_dev, unsigned i, uint64_t entry);
    } gart;
    struct {
        unsigned long (*get_gpu_start)(unsigned long flags);
        unsigned long (*get_normal_size)(unsigned long flags);
    } vram;
    struct {
        int (*init)(struct ljm_device *ljm_dev);
        void (*fini)(struct ljm_device *ljm_dev);
//      void (*copy_pages)(struct ljm_device *ljm_dev,
//                 struct ljm_ib *ib,
//                 uint64_t pe, uint64_t src,
//                 unsigned count);
//      void (*write_pages)(struct ljm_device *ljm_dev,
//                  struct ljm_ib *ib,
//                  uint64_t pe,
//                  uint64_t addr, unsigned count,
//                  uint32_t incr, uint32_t flags);
//      void (*set_pages)(struct ljm_device *ljm_dev,
//                struct ljm_ib *ib,
//                uint64_t pe,
//                uint64_t addr, unsigned count,
//                uint32_t incr, uint32_t flags);
//      void (*pad_ib)(struct ljm_ib *ib);
    } vm;
    /* ring specific callbacks */
//  const struct ljm_asic_ring *ring[RADEON_NUM_RINGS];
    /* irqs */
    struct {
        int (*set)(struct ljm_device *ljm_dev);
        int (*vblank_set)(struct ljm_device *ljm_dev);
        int (*process)(struct ljm_device *ljm_dev);
        void (*postinstall)(struct ljm_device *ljm_dev);
    } irq;
    /* displays */
    struct {
        /* display watermarks */
        void (*bandwidth_update)(struct ljm_device *ljm_dev);
        /* get frame count */
        u32 (*get_vblank_counter)(struct ljm_device *ljm_dev, int crtc);
        /* wait for vblank */
        void (*wait_for_vblank)(struct ljm_device *ljm_dev, int crtc);
        /* set backlight level */
    //  void (*set_backlight_level)(struct ljm_encoder *ljm_encoder, u8 level);
        /* get backlight level */
//      u8 (*get_backlight_level)(struct ljm_encoder *ljm_encoder);
        /* audio callbacks */
        void (*hdmi_enable)(struct drm_encoder *encoder, bool enable);
        void (*hdmi_setmode)(struct drm_encoder *encoder, struct drm_display_mode *mode);
        void (*encoder_dpms)(struct drm_encoder *encoder, int mode);
        void (*encoder_mode_set)(struct drm_encoder *encoder, struct drm_display_mode *mode, struct drm_display_mode *adjustedMode);
        bool (*encoder_mode_fixup)(struct drm_encoder *encoder, const struct drm_display_mode *mode, struct drm_display_mode *adjustedMode);
        void (*crtc_cursor_move)(struct drm_crtc *crtc, int x, int y);
        void (*crtc_cursor_hide)(struct ljm_device *ljm_dev, struct drm_crtc *crtc);
        void (*crtc_cursor_show)(struct ljm_device *ljm_dev, struct drm_crtc *crtc);
        bool (*crtc_mode_fixup)(struct ljm_device *ljm_dev, const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode);
        int (*crtc_mode_set)(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, struct drm_display_mode *adjusted_mode);
        int (*crtc_mode_set_base)(struct drm_crtc *crtc, int bpp, int pitches, int base);
        int (*crtc_set_base)(struct ljm_device *ljm_dev, int crtc_id, u64 crtc_base);
        int (*crtc_prog_reset)(struct ljm_device *ljm_dev, int crtc_used, long min_hscan_time);
        void (*crtc_mode_enable)(struct ljm_device *ljm_dev, struct ljm_crtc *ljm_crtc, bool enable);
        void (*crtc_commit_dp)(struct ljm_device *ljm_dev);
        int (*connector_detect)(struct drm_connector *connector);
        int (*connector_mode_valid)(struct drm_display_mode *mode, struct ljm_connector *ljm_connector);
        struct edid *(*connector_get_edid)(struct drm_connector *connector, int *size);
        int (*get_crtc_scanout_pos)(struct ljm_device *ljm_dev, unsigned int pipe, int *vpos, int *hpos);
        int (*crtc_load_lut)(struct ljm_crtc *ljm_crtc, u16 *red, u16 *green, u16 *blue, uint32_t gamma_size);
    } display;
#if 0
    /* copy functions for bo handling */
    struct {
        struct ljm_fence *(*blit)(struct ljm_device *ljm_dev,
                         uint64_t src_offset,
                         uint64_t dst_offset,
                         unsigned num_gpu_pages,
                         struct reservation_object *resv);
        u32 blit_ring_index;
        struct ljm_fence *(*dma)(struct ljm_device *ljm_dev,
                        uint64_t src_offset,
                        uint64_t dst_offset,
                        unsigned num_gpu_pages,
                        struct reservation_object *resv);
        u32 dma_ring_index;
        /* method used for bo copy */
        struct ljm_fence *(*copy)(struct ljm_device *ljm_dev,
                         uint64_t src_offset,
                         uint64_t dst_offset,
                         unsigned num_gpu_pages,
                         struct reservation_object *resv);
        /* ring used for bo copies */
        u32 copy_ring_index;
    } copy;
#endif
    /* surfaces */
    struct {
        int (*set_reg)(struct ljm_device *ljm_dev, int reg,
                       uint32_t tiling_flags, uint32_t pitch,
                       uint32_t offset, uint32_t obj_size);
        void (*clear_reg)(struct ljm_device *ljm_dev, int reg);
    } surface;
    /* hotplug detect */
    struct {
        void (*init)(struct ljm_device *ljm_dev);
        void (*fini)(struct ljm_device *ljm_dev);
        bool (*sense)(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector);
        void (*set_polarity)(struct ljm_device *ljm_dev, struct ljm_connector *ljm_connector);
        void (*set)(struct ljm_device *ljm_dev, struct drm_connector *connector, bool value, int msec);
    } hpd;
    /* static power management */
    struct {
        void (*misc)(struct ljm_device *ljm_dev);
        void (*prepare)(struct ljm_device *ljm_dev);
        void (*finish)(struct ljm_device *ljm_dev);
        void (*init_profile)(struct ljm_device *ljm_dev);
        void (*get_dynpm_state)(struct ljm_device *ljm_dev);
        uint32_t (*get_engine_clock)(struct ljm_device *ljm_dev);
        void (*set_engine_clock)(struct ljm_device *ljm_dev, uint32_t eng_clock);
        uint32_t (*get_memory_clock)(struct ljm_device *ljm_dev);
        void (*set_memory_clock)(struct ljm_device *ljm_dev, uint32_t mem_clock);
        int (*get_pcie_lanes)(struct ljm_device *ljm_dev);
        void (*set_pcie_lanes)(struct ljm_device *ljm_dev, int lanes);
        void (*set_clock_gating)(struct ljm_device *ljm_dev, int enable);
        int (*set_uvd_clocks)(struct ljm_device *ljm_dev, u32 vclk, u32 dclk);
        int (*set_vce_clocks)(struct ljm_device *ljm_dev, u32 evclk, u32 ecclk);
        int (*get_temperature)(struct ljm_device *ljm_dev);
    } pm;
    /* dynamic power management */
    struct {
        int (*init)(struct ljm_device *ljm_dev);
        void (*setup_asic)(struct ljm_device *ljm_dev);
        int (*enable)(struct ljm_device *ljm_dev);
        int (*late_enable)(struct ljm_device *ljm_dev);
        void (*disable)(struct ljm_device *ljm_dev);
        int (*pre_set_power_state)(struct ljm_device *ljm_dev);
        int (*set_power_state)(struct ljm_device *ljm_dev);
        void (*post_set_power_state)(struct ljm_device *ljm_dev);
        void (*display_configuration_changed)(struct ljm_device *ljm_dev);
        void (*fini)(struct ljm_device *ljm_dev);
        u32 (*get_sclk)(struct ljm_device *ljm_dev, bool low);
        u32 (*get_mclk)(struct ljm_device *ljm_dev, bool low);
//      void (*print_power_state)(struct ljm_device *ljm_dev, struct ljm_ps *ps);
        void (*debugfs_print_current_performance_level)(struct ljm_device *ljm_dev, struct seq_file *m);
//      int (*force_performance_level)(struct ljm_device *ljm_dev, enum ljm_dpm_forced_level level);
        bool (*vblank_too_short)(struct ljm_device *ljm_dev);
        void (*powergate_uvd)(struct ljm_device *ljm_dev, bool gate);
        void (*enable_bapm)(struct ljm_device *ljm_dev, bool enable);
        void (*fan_ctrl_set_mode)(struct ljm_device *ljm_dev, u32 mode);
        u32 (*fan_ctrl_get_mode)(struct ljm_device *ljm_dev);
        int (*set_fan_speed_percent)(struct ljm_device *ljm_dev, u32 speed);
        int (*get_fan_speed_percent)(struct ljm_device *ljm_dev, u32 *speed);
        u32 (*get_current_sclk)(struct ljm_device *ljm_dev);
        u32 (*get_current_mclk)(struct ljm_device *ljm_dev);
    } dpm;
    /* pageflipping */
    struct {
        void (*page_flip)(struct ljm_device *ljm_dev, int crtc, u64 crtc_base, bool async);
        bool (*page_flip_pending)(struct ljm_device *ljm_dev, int crtc);
    } pflip;
};

/* ASICs macro */
#define ljm_hard_init(ljm_dev) (ljm_dev)->asic->init((ljm_dev))
#define ljm_hard_fini(ljm_dev) (ljm_dev)->asic->fini((ljm_dev))

//#define ljm_vram_get_gpu_start(ljm_dev)     (ljm_dev)->asic->vram.get_gpu_start() /* vram physical start address for GPU */
//#define ljm_vram_get_normal_size(ljm_dev)   (ljm_dev)->asic->vram.get_normal_size()   /* visiable vram size for display */
#define ljm_irq_set(ljm_dev)        (ljm_dev)->asic->irq.set((ljm_dev))
#define ljm_irq_vblank_set(ljm_dev) (ljm_dev)->asic->irq.vblank_set((ljm_dev))
#define ljm_irq_process(ljm_dev)    (ljm_dev)->asic->irq.process((ljm_dev))
#define ljm_irq_postinstall(ljm_dev)    (ljm_dev)->asic->irq.postinstall((ljm_dev))
#define ljm_hpd_init(ljm_dev)       (ljm_dev)->asic->hpd.init((ljm_dev))
#define ljm_hpd_fini(ljm_dev)       (ljm_dev)->asic->hpd.fini((ljm_dev))
#define ljm_hpd_sense(ljm_dev, ljm_connector)           (ljm_dev)->asic->hpd.sense((ljm_dev), (ljm_connector))
#define ljm_hpd_set_polarity(ljm_dev, ljm_connector)    (ljm_dev)->asic->hpd.set_polarity((ljm_dev), (ljm_connector))

#define encoder_dpms(ljm_dev, encoder, mode)            (ljm_dev)->asic->display.encoder_dpms((encoder), (mode))
#define encoder_mode_set(ljm_dev, encoder, mode, adjustedMode)   (ljm_dev)->asic->display.encoder_mode_set((encoder), (mode), (adjustedMode))
#define encoder_mode_fixup(ljm_dev, encoder, mode, adjustedMode) (ljm_dev)->asic->display.encoder_mode_fixup((encoder), (mode), (adjustedMode))

#define ljm_cursor_move(ljm_dev, crtc, x, y)            (ljm_dev)->asic->display.crtc_cursor_move((crtc), (x), (y))
#define ljm_crtc_cursor_hide(ljm_dev, crtc)             (ljm_dev)->asic->display.crtc_cursor_hide((ljm_dev),(crtc))
#define ljm_crtc_cursor_show(ljm_dev, crtc)             (ljm_dev)->asic->display.crtc_cursor_show((ljm_dev), (crtc))
#define ljm_mode_fixup(ljm_dev, mode, adjusted_mode)    (ljm_dev)->asic->display.crtc_mode_fixup((ljm_dev), (mode), (adjusted_mode))
#define ljm_mode_set(ljm_dev, ljm_crtc, adjusted_mode)  (ljm_dev)->asic->display.crtc_mode_set((ljm_dev), (ljm_crtc), (adjusted_mode))
#define ljm_crtc_mode_enable(ljm_dev, ljm_crtc, en)     (ljm_dev)->asic->display.crtc_mode_enable((ljm_dev), (ljm_crtc), (en))
#define ljm_mode_set_base(ljm_dev, crtc, bpp, pitches, base)    (ljm_dev)->asic->display.crtc_mode_set_base((crtc), (bpp), (pitches), (base))
#define ljm_crtc_set_base(ljm_dev, crtc_id, crtc_base)  (ljm_dev)->asic->display.crtc_set_base((ljm_dev), (crtc_id), (crtc_base))
#define ljm_crtc_prog_reset(ljm_dev, crtc_used, min_hscan_time)           (ljm_dev)->asic->display.crtc_prog_reset((ljm_dev), (crtc_used), (min_hscan_time))
#define ljm_crtc_commit_dp(crtc)                        (ljm_dev)->asic->display.crtc_commit_dp(crtc)
#define ljm_connector_detect(ljm_dev, connector)        (ljm_dev)->asic->display.connector_detect(connector)
#define ljm_connector_mode_set_valid(ljm_dev, mode, ljm_connector)      (ljm_dev)->asic->display.connector_mode_valid((mode), (ljm_connector))
#define ljm_connector_get_edid(ljm_dev, connector, size)    (ljm_dev)->asic->display.connector_get_edid(connector, size)
#define ljm_get_vblank_counter(ljm_dev, crtc)           (ljm_dev)->asic->display.get_vblank_counter((ljm_dev), (crtc))
#define ljm_page_flip(ljm_dev, crtc_id, base, async)    (ljm_dev)->asic->pflip.page_flip(ljm_dev, crtc_id, base, async)
#define ljm_page_flip_pending(ljm_dev, crtc_id)         (ljm_dev)->asic->pflip.page_flip_pending((ljm_dev), (crtc_id))
#define ljm_get_crtc_scanout_pos(ljm_dev, pipe, vpos, hpos) (ljm_dev)->asic->display.get_crtc_scanout_pos(ljm_dev, pipe, vpos, hpos)
#define ljm_crtc_load_lut(ljm_dev, ljm_crtc, red, green, blue, size) (ljm_dev)->asic->display.crtc_load_lut(ljm_crtc, red, green, blue, size)

#define ljm_hard_suspend(ljm_dev, ddev)    (ljm_dev)->asic->suspend(ddev)
#define ljm_hard_resume(ljm_dev, ddev)     (ljm_dev)->asic->resume(ddev)

#define ljm_power_set_state(ljm_dev, power_state)   (ljm_dev)->asic->power_set_state((ljm_dev), (power_state))

/* ASIC functions */
int ljm_asic_init(struct ljm_device *ljm_dev);

/* modeset functions */
int ljm_modeset_init(struct ljm_device *ljm_dev);
void ljm_modeset_fini(struct ljm_device *ljm_dev);

int ljm_gem_init(struct ljm_device *ljm_dev);
void ljm_gem_fini(struct ljm_device *ljm_dev);

int ljm_gem_object_create(struct ljm_device *ljm_dev, unsigned long size, struct drm_gem_object **obj);
void ljm_gem_object_free(struct drm_gem_object *gobj);

/* connector functions */
void ljm_connector_hotplug(struct drm_connector *connector);

/* function before drm is loaded */
int ljm_pre_init(int card, struct pci_dev *pdev, unsigned long flags);

struct ljm_device * get_ljm_devs(int card);

int ljm_i2s_register(struct ljm_device *ljm_dev); /* 注册iis */

int ljm_set_power(struct ljm_device *ljm_dev, int power_state);

/** 获取GOP版本信息 */
int ljm_gop_info_get(struct pci_dev *pdev);

#endif
