/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_device.h"
#include "ljmicro_mode.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

static int ljm_modeset_create_props(struct ljm_device *ljm_dev)
{
    HEADER_ARG("ljm_dev=%p", ljm_dev);

    /* create properties for display */

    FOOTER();
    return 0;
}

/* Callback used by helpers to inform the driver of output configuration changes.
 * FIXME: call drm_fb_helper_hotplug_changed ?
 */
static void ljm_output_poll_changed(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    HEADER_ARG("ddev=0x%p", ddev);

    if (ljm_dev->fbdev)
        drm_fb_helper_hotplug_event(&ljm_dev->fbdev->helper);

    FOOTER();
}

static const struct drm_mode_config_funcs ljm_mode_funcs = {
    .fb_create = ljm_user_framebuffer_create,
    .output_poll_changed = ljm_output_poll_changed,
};

int ljm_irq_kms_init(struct ljm_device *ljm_dev);

int ljm_modeset_init(struct ljm_device *ljm_dev)
{
    int i;
    int r;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    drm_mode_config_init(ljm_dev->ddev);
    ljm_dev->mode_config_initialized = true;

    ljm_dev->ddev->mode_config.funcs = &ljm_mode_funcs;

    switch (ljm_dev->family) {
        case CHIP_GP100:
            ljm_dev->ddev->mode_config.max_width = 4096 * 2;
            ljm_dev->ddev->mode_config.max_height = 4096 * 2;
            break;
        case CHIP_GP200:
        case CHIP_GP300:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202) {
                ljm_dev->ddev->mode_config.max_width = 4096 * 2;
                ljm_dev->ddev->mode_config.max_height = 4096 * 2;
            } else {
                ljm_dev->ddev->mode_config.max_width = 4096 * 4;
                ljm_dev->ddev->mode_config.max_height = 4096 * 4;
            }
            break;
        default:
            break;
    }

    ljm_dev->ddev->mode_config.preferred_depth = 24;
    ljm_dev->ddev->mode_config.prefer_shadow = 0;   /* FIXME: ljm use 1 ? */

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    ljm_dev->ddev->mode_config.fb_base = ljm_dev->smem_start;
#endif
    TRACE("fb_base=%llx", ljm_dev->smem_start);

    r = ljm_modeset_create_props(ljm_dev);
    if (r) {
        FOOTER_ARG("r=%d", r);
        return r;
    }

    /*allocate crtcs*/
    for (i = 0; i < ljm_dev->num_crtcs; i++) {
        ljm_crtc_init(ljm_dev->ddev, i);
    }

    /*init all encoders&connectors*/
    r = ljm_setup_enc_conn(ljm_dev->ddev);
    if (r) {
        FOOTER_ARG("r=%d", r);
        return r;
    }
    ljm_plane_init(ljm_dev);
    ljm_zoomvideo_init(ljm_dev);
    /* enable hpd */
    ljm_hpd_init(ljm_dev);

    /* setup afmt */
//  ljm_afmt_init(ljm_dev);

    ljm_dev->fbdev = ljm_fbdev_init(ljm_dev->ddev);
    drm_kms_helper_poll_init(ljm_dev->ddev);

    /* do pm late init */
//  ret = ljm_pm_late_init(ljm_dev);
    FOOTER();
    return 0;
}

void ljm_modeset_fini(struct ljm_device *ljm_dev)
{
    int i;

    HEADER_ARG("LjmDev=0x%p", ljm_dev);

    if (ljm_dev->mode_config_initialized) {
        drm_kms_helper_poll_fini(ljm_dev->ddev);
        ljm_hpd_fini(ljm_dev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)
        drm_helper_force_disable_all(ljm_dev->ddev);
#else
        drm_crtc_force_disable_all(ljm_dev->ddev);
#endif

        /* hardware finish, such as disable vblank interrupt */
        ljm_hard_fini(ljm_dev);

        // if(ljm_dev->fbdev)
        //    ljm_fbdev_fini(ljm_dev);

#if 0
        drm_mode_config_cleanup(ljm_dev->ddev);
        ljm_dev->mode_config_initialized = false;

        for (i=0; i<ljm_dev->num_crtcs; i++) {
            ljm_dev->crtcs[i] = NULL;
        }
        ljm_dev->num_crtcs = 0;

        for (i = 0; i < ljm_dev->num_ports; i++) {
            ljm_dev->ports[i] = NULL;
        }
        ljm_dev->num_ports = 0;
#else
        for (i = 0; i < ljm_dev->num_ports; i++) {
            ljm_connector_destroy(&ljm_dev->ports[i]->vdev_connector->base);
            ljm_encoder_destroy(&ljm_dev->ports[i]->base.base);
            ljm_dev->ports[i] = NULL;
        }
        ljm_dev->num_ports = 0;

        for (i=0; i<ljm_dev->num_crtcs; i++) {
            ljm_crtc_destroy(ljm_dev->crtcs[i]);
            ljm_dev->crtcs[i] = NULL;
        }
        ljm_dev->num_crtcs = 0;

        ljm_dev->mode_config_initialized = false;
#endif
    }

    FOOTER();
}

/**
 * ljm_irq_kms_pflip_irq_get - enable pageflip interrupt
 *
 * @ljm_dev: GP device pointer
 * @crtc: crtc whose interrupt you want to enable
 *
 * Enables the pageflip interrupt for a specific crtc (all asics).
 * For pageflips we use the vblank interrupt source.
 */
void ljm_irq_kms_pflip_irq_get(struct ljm_device *ljm_dev, int crtc)
{
    unsigned long irqflags;

    DRM_DEBUG_VBL("ljm_dev=%p crtc=%d\n", ljm_dev, crtc);

    if (crtc < 0 || crtc >= ljm_dev->num_crtcs)
        return;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    if (!ljm_dev->ddev->irq_enabled)
        return;
#else
    if (!ljm_dev->irq.installed)
        return;
#endif

    if (atomic_inc_return(&ljm_dev->irq.pflip[crtc]) == 1) {
        spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
        ljm_irq_vblank_set(ljm_dev);
        spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);
    }
}

/**
 * ljm_irq_kms_pflip_irq_put - disable pageflip interrupt
 *
 * @ljm_dev: ljm device pointer
 * @crtc: crtc whose interrupt you want to disable
 *
 * Disables the pageflip interrupt for a specific crtc (all asics).
 * For pageflips we use the vblank interrupt source.
 */
void ljm_irq_kms_pflip_irq_put(struct ljm_device *ljm_dev, int crtc_id)
{
    unsigned long irqflags;

    DRM_DEBUG_VBL("ljm_dev=%p crtc_id=%d\n", ljm_dev, crtc_id);

    if (crtc_id < 0 || crtc_id >= ljm_dev->num_crtcs)
        return;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    if (!ljm_dev->ddev->irq_enabled)
        return;
#else
    if (!ljm_dev->irq.installed)
        return;
#endif

    if (atomic_dec_and_test(&ljm_dev->irq.pflip[crtc_id])) {
        spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
        ljm_irq_vblank_set(ljm_dev);
        spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);
    }
}

/*
 * Retrieve current video scanout position of crtc on a given gpu, and
 * an optional accurate timestamp of when query happened.
 *
 * \param dev Device to query.
 * \param crtc Crtc to query.
 * \param flags Flags from caller (DRM_CALLED_FROM_VBLIRQ or 0).
 *              For driver internal use only also supports these flags:
 *
 *              USE_REAL_VBLANKSTART to use the real start of vblank instead
 *              of a fudged earlier start of vblank.
 *
 *              GET_DISTANCE_TO_VBLANKSTART to return distance to the
 *              fudged earlier start of vblank in *vpos and the distance
 *              to true start of vblank in *hpos.
 *
 * \param *vpos Location where vertical scanout position should be stored.
 * \param *hpos Location where horizontal scanout position should go.
 * \param *stime Target location for timestamp taken immediately before
 *               scanout position query. Can be NULL to skip timestamp.
 * \param *etime Target location for timestamp taken immediately after
 *               scanout position query. Can be NULL to skip timestamp.
 *
 * Returns vpos as a positive number while in active scanout area.
 * Returns vpos as a negative number inside vblank, counting the number
 * of scanlines to go until end of vblank, e.g., -1 means "one scanline
 * until start of active scanout / end of vblank."
 *
 * \return Flags, or'ed together as follows:
 *
 * DRM_SCANOUTPOS_VALID = Query successful.
 * DRM_SCANOUTPOS_INVBL = Inside vblank.
 * DRM_SCANOUTPOS_ACCURATE = Returned position is accurate. A lack of
 * this flag means that returned position may be offset by a constant but
 * unknown small number of scanlines wrt. real scanout position.
 *
 */
int ljm_get_crtc_scanoutpos(struct drm_device *ddev, unsigned int pipe,
                   int *vpos, int *hpos,
                   ktime_t *stime, ktime_t *etime,
                   const struct drm_display_mode *mode)
{
    int ret;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    /* preempt_disable_rt() should go right here in PREEMPT_RT patchset. */

    /* Get optional system timestamp before query. */
    if (stime)
        *stime = ktime_get();

    ret = ljm_get_crtc_scanout_pos(ljm_dev, pipe, vpos, hpos);

    /* Get optional system timestamp after query. */
    if (etime)
        *etime = ktime_get();

    if (*vpos >= mode->vdisplay) {
        *vpos -= mode->vtotal;
    }

    if (*hpos >= mode->hdisplay)
        *hpos -= mode->htotal;

    return ret;
}


