/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#if USE_DP_CONNECTOR

#include "ljmicro_base.h"
#include "ljmicro_device.h"
#include "ljmicro_dp.h"
#include "ljmicro_ddc.h"
#include "ljmicro_i2c.h"
#include "ljmicro_gpio.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include "drm/drm_dp_helper.h"
#else
#include "drm/display/drm_dp_helper.h"
#endif
#include "linux/backlight.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

#define TIMEOUT_CNT_DPTX    10000
#define LJM_DPCD_TIMEOUT_MS 5000

#ifdef UEFI_GOP
/* FIXME: add register opertion for UEFI */

/* 32-bit register for DP controller */
#define DPTX_REG_WR(dp_dev, addr, val)
#define DPTX_REG_RD(dp_dev, addr)

/* 16-bit register for DP PHY */
#define DPPHY_REG_WR(dp_dev, addr, val)
#define DPPHY_REG_RD(dp_dev, addr)

#define DPPHY_CDB_WR(dp_dev, addr, val)
#define DPPHY_CDB_RD(dp_dev, addr)
#else
/* 32-bit register opertion for DP transmitter */
#define DPTX_REG_WR(dp_dev, addr, val)  WREG32_BASE(dp_dev->regbase, (dp_dev->offset + addr + 0x10000), val)
#define DPTX_REG_RD(dp_dev, addr)       RREG32_BASE(dp_dev->regbase, (dp_dev->offset + addr + 0x10000))

/* 32-bit PHY signal configuration */
#define DPPHY_REG_WR(dp_dev, addr, val)  WREG32_BASE(dp_dev->regbase, dp_dev->offset + addr, val)
#define DPPHY_REG_RD(dp_dev, addr)       RREG32_BASE(dp_dev->regbase, dp_dev->offset + addr)

/* 16-bit register opertion for DP PHY */
#define DPPHY_CDB_WR(dp_dev, addr, val)\
    do {\
        WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_ADDR), addr);\
        WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_WDATA), val);\
        WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_SEL_WRITE), 0x3);\
        WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_RSTN), 0x1);\
        WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_ENABLE), 0x1);\
        while ((RREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_EN_READY)) & 0x1) == 0x0) ;\
        WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_SEL_WRITE), 0x0);\
    } while (0)

inline u16 DPPHY_CDB_RD(struct dp_dev *dp_dev, int addr)
{
    u16 val;

    WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_ADDR), addr);
    WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_SEL_WRITE), 0x2);   /* {i_cdb_sel, i_cdb_write} = 2'b10 */
    WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_RSTN), 0x1);        /* i_cdb_reset_n = 1 */
    WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_ENABLE), 0x1);      /* i_cdb_enable = 1 */
    while ((RREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_EN_READY)) & 0x1) == 0x0) {
        udelay(1);
    }
    val = (u16)(RREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_RDATA)) & 0xffff);
    WREG32_BASE(dp_dev->regbase, (dp_dev->offset + PHY_CDB_SEL_WRITE), 0x0);

    return val;
}
#endif

static inline void DPTX_REG_MSK(struct dp_dev *dp_dev, int addr, u32 msk, u32 val)
{
    u32 tmp;

    tmp = DPTX_REG_RD(dp_dev, addr) & (0xffffffff - (msk));
    DPTX_REG_WR(dp_dev, addr, (tmp | ((val) & (msk))));
}

/******************************DP Vitual configuration********************************/
/* BOE NV156FHM-N52 V8.0 */
static uint8_t BOEn52_vedid[] = {
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x09, 0xe5, 0x47, 0x08, 0x00, 0x00, 0x00, 0x00,
0x2d, 0x1c, 0x01, 0x04, 0xb5, 0x22, 0x13, 0x78, 0x06, 0xb3, 0x35, 0xa5, 0x54, 0x4f, 0x9c, 0x27,
0x0f, 0x50, 0x54, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x42, 0x37, 0x80, 0xcc, 0x70, 0x38, 0x1e, 0x40, 0x30, 0x20,
0x36, 0x00, 0x58, 0xc2, 0x10, 0x00, 0x00, 0x1a, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x42,
0x4f, 0x45, 0x20, 0x43, 0x51, 0x0a, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfe,
0x00, 0x4e, 0x56, 0x31, 0x35, 0x36, 0x46, 0x48, 0x4d, 0x2d, 0x4e, 0x35, 0x32, 0x0a, 0x01, 0x70,
0x02, 0x03, 0x0f, 0x00, 0xe3, 0x05, 0x80, 0x00, 0xe6, 0x06, 0x05, 0x01, 0x60, 0x60, 0x28, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
};
static struct virtual_dpcd BOEn52_vdpcd[] = {
    /* Receiver Capability */
    { .addr = 0x0,  .val = 0x14,},
    { .addr = 0x1,  .val = 0x0a,},
    { .addr = 0x2,  .val = 0xc2,},
    { .addr = 0x3,  .val = 0x41,},
    { .addr = 0x4,  .val = 0x0,},
    { .addr = 0x5,  .val = 0x0,},
    { .addr = 0x6,  .val = 0x1,},
    { .addr = 0x7,  .val = 0xc0,},
    { .addr = 0x8,  .val = 0x2,},
    { .addr = 0x9,  .val = 0x0,},
    { .addr = 0xa, .val = 0xa4,},
    { .addr = 0xb, .val = 0x1f,},
    { .addr = 0xc, .val = 0x30,},
    { .addr = 0xd, .val = 0x2a,},
    { .addr = 0xe, .val = 0x76,},
    { .addr = 0xf, .val = 0x2f,},
    { .addr = 0x10, .val = 0x2f,},
    { .addr = 0x11, .val = 0x2f,},
    { .addr = 0x12, .val = 0x2f,},
    { .addr = 0x13, .val = 0x2f,},
    { .addr = 0x14, .val = 0x2f,},
    { .addr = 0x15, .val = 0x2f,},
    { .addr = 0x16, .val = 0x2f,},
    { .addr = 0x17, .val = 0x2f,},

    //{ .addr = 0x17, .val = ,},
    //{ .addr = 0x17, .val = ,},

    /* Link/Sink Device Status */
    //{ .addr = 0x200, .val = },
    //{ .addr = 0x200, .val = },

    /* eDP-specific */
    { .addr = 0x700, .val = 0x4,},
    { .addr = 0x701, .val = 0x99,},
    { .addr = 0x702, .val = 0xa7,},
    //{ .addr = 0x722, .val = ,},
    //{ .addr = 0x723, .val = ,},
};

#define VDPCD_LIST_LEN(x)   (sizeof((x)) / sizeof(struct virtual_dpcd))
#define VEDID_LEN(x)   sizeof((x))

static int ljm_dp_vedid_add(struct dp_dev *dp_dev, struct board_info_image_header *h)
{
    if(!h)
    {
        memcpy(dp_dev->vedid, BOEn52_vedid, sizeof(BOEn52_vedid));
        return 0;
    }

    if(h->data_size == DDC_EDID_SIZE || h->data_size == DDC_EDID_SIZE*2)
        memcpy(dp_dev->vedid, h->data, h->data_size);

    return 0;
}

static int ljm_dp_vdpcd_add(struct dp_dev *dp_dev, struct board_info_image_header *h)
{
    int i=0, j, vdpcd_index, BOEn52_vdpcd_size;
    uint32_t addr, len;
    uint8_t *raw, *val;
    struct virtual_dpcd *vdpcd;

    if(!h)
    {
        BOEn52_vdpcd_size = sizeof(BOEn52_vdpcd)/sizeof(struct virtual_dpcd);
        memcpy(dp_dev->vdpcd, BOEn52_vdpcd, sizeof(BOEn52_vdpcd));
        dp_dev->vdpcd[BOEn52_vdpcd_size].addr = VDPCD_ADDR_END; //添加vdpcd结束符
        return 0;
    }

    raw = h->data;
    vdpcd = &dp_dev->vdpcd[0];

    //查找vdpcd数组的结尾。
    for(vdpcd_index=0; vdpcd_index<VDPCD_SIZE_MAX; vdpcd_index++)
    {
        if(vdpcd[vdpcd_index].addr == VDPCD_ADDR_END)
            break;
    }
    if(vdpcd_index == VDPCD_SIZE_MAX)
        return 0;

    while(i < h->data_size)
    {
        addr = ((raw[i+2]&0xf) << 16) | (raw[i+1] << 8) | raw[i];
        len = ((raw[i+2] & 0xf0) >> 4) + 1;
        val = &raw[i+3];
        for(j=0; j<len; j++)
        {
            vdpcd[vdpcd_index].addr = addr + j;
            vdpcd[vdpcd_index].val = val[j];
            //将vdpcd_index向后移，若数组溢出则不再继续添加vdpcd
            if(++vdpcd_index == VDPCD_SIZE_MAX)
                return 0;
        }
        //计算raw的偏移：三个地址字节+若干数据字节
        i = i + 3 + len;
    }

    //给数组添加结束标志
    if(vdpcd_index < VDPCD_SIZE_MAX)
        vdpcd[vdpcd_index].addr = VDPCD_ADDR_END;

    return 0;
}

int ljm_dp_vedid_vdpcd_add(struct dp_dev *dp_dev, struct ljm_port_info *info)
{
    int i;
    bool has_vedid = false, has_vdpcd = false;

    //初始化vdpcd数组
    dp_dev->vdpcd[0].addr = VDPCD_ADDR_END;

    for(i=0; i<PORT_INFO_IMAGE_MAX; i++)
    {
        if(info->image[i])
        {
            if(info->image[i]->data_type == IMAGE_DATA_TYPE_EDID)
            {
                ljm_dp_vedid_add(dp_dev, info->image[i]);
                has_vedid = true;
            }
            else if (info->image[i]->data_type == IMAGE_DATA_TYPE_DPCD)
            {
                ljm_dp_vdpcd_add(dp_dev, info->image[i]);
                has_vdpcd = true;
            }
        }
    }

    if(!has_vedid)
        ljm_dp_vedid_add(dp_dev, NULL);
    if(!has_vdpcd)
        ljm_dp_vdpcd_add(dp_dev, NULL);

    return 0;
}

struct edid * vedid_get(struct dp_dev *dp_dev)
{
    struct edid *vedid;
    vedid = kmalloc(256, GFP_KERNEL);
    if(!vedid)
    {
        ERR_PRINT("Failed to get Vitual Edid.\n");
        return vedid;
    }
    memcpy(vedid, dp_dev->vedid, 256);
    return vedid;
}

static int vdpcd_read_byte(struct dp_dev *dp_dev, uint32_t addr, uint8_t *val)
{
    int i;
    for(i = 0; i < VDPCD_SIZE_MAX; i++)
    {
        /**
         * addr == VDPCD_ADDR_END时，表示vdpcd数组已结束。
         * 由于dp_dev->vdpcd数组较长，以此方法提高读取vdpcd效率。
        */
        if(addr == VDPCD_ADDR_END)
            break;

        if(addr == dp_dev->vdpcd[i].addr)
        {
            *val = dp_dev->vdpcd[i].val;
            return 0;
        }
    }

    /* If the read failed, return 0. */
    *val = 0;
    return 0;
}

static int vdpcd_read(struct dp_dev *dp_dev, uint32_t addr, uint8_t *buff, int size)
{
    int i, byte = 0;
    DEBUG_PRINT("[dp %d] Visual dpcd read addr=0x%x size=0x%x\n", dp_dev->index, addr, size);
    for(i = 0; i< size; i++)
    {
        if(vdpcd_read_byte(dp_dev, addr + i, buff + i) < 0)
            return -EFAULT;
        byte++;
    }
    return byte;
}

static ssize_t _ljm_dp_aux_transfer_mcu(struct dp_dev *dp, struct drm_dp_aux_msg *msg)
{
    struct mcu_arg_aux args;
    ssize_t ret;
    struct ljm_device *ljm_dev = get_ljm_devs(0);

    switch (msg->request & ~DP_AUX_I2C_MOT){
        case DP_AUX_NATIVE_WRITE:
        case DP_AUX_NATIVE_READ:
            args.gpio_direction = dp->gpio_direction;
            args.gpio_data = dp->gpio_data;
            args.cmd = msg->request;
            args.addr = msg->address;
            args.buff = msg->buffer;
            args.len = msg->size;
            ret =  ljm_mcu_handle(ljm_dev->mcu, MCU_CMD_DPCD, &args);
            msg->reply = args.reply;
            return ret;
        case DP_AUX_I2C_WRITE:
        case DP_AUX_I2C_WRITE_STATUS_UPDATE:
        case DP_AUX_I2C_READ:
            DEBUG_PRINT("Can't handle requests of i2c.\n");
            return -EINVAL;
        default:
            WARN_PRINT("dp %d: invalid aux request %x\n", dp->index, msg->request);
            return -EINVAL;
    }
}

static ssize_t _ljm_dp_aux_transfer_deaf(struct dp_dev *dp, struct drm_dp_aux_msg *msg)
{
    int ret = 0;
    bool is_write = false;

    if (!ljm_dptx_aux_ready(dp)) {
        DEBUG_PRINT("dp %d: aux is busy\n", dp->index);
        return -EBUSY;
    }

    switch (msg->request & ~DP_AUX_I2C_MOT){
        case DP_AUX_NATIVE_WRITE:
        case DP_AUX_I2C_WRITE:
        case DP_AUX_I2C_WRITE_STATUS_UPDATE:
            if(dp->once_edp)
            {
                msg->reply = (u8)(TR_DPTX_AUX_REPLY_ACK & 0xf);
                return 0;
            }
            is_write = true;
            ljm_dptx_aux_write_fifo(dp, msg->buffer, msg->size);
            ljm_dptx_aux_send_request(dp, ((uint32_t)msg->request) << 8, msg->address, msg->size, (msg->size == 0), false);
            ljm_dptx_aux_wait_reply(dp);
            ljm_dptx_reg_read(dp, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
            /* FIXME: we would think dpcd had been writen successfully whatever was received when writing */
            msg->reply = TR_DPTX_AUX_REPLY_ACK;
            return msg->size;
        case DP_AUX_NATIVE_READ:
        case DP_AUX_I2C_READ:
                ret = vdpcd_read(dp, msg->address, msg->buffer, msg->size);
                return ret;
            break;
        default:
            WARN_PRINT("dp %d: invalid aux request %x\n", dp->index, msg->request);
            return -EINVAL;
    }
    return 0;
}

static ssize_t _ljm_dp_aux_transfer_native(struct dp_dev *dp, struct drm_dp_aux_msg *msg)
{
    int ret = 0;
    bool is_write = false;
    uint32_t reply_code;

    DEBUG_WARN_ON(1);    /* print call trace if breakId == 1 */
    DEBUG_PRINT("dp_aux %d: address=%x request=%02x replay=%02x size=%zu",
        dp->index, msg->address, msg->request, msg->reply, msg->size);

    if (WARN_ON(msg->size > 16)) {
        return -E2BIG;
    }

    if (!ljm_dptx_aux_ready(dp)) {
        DEBUG_PRINT("dp %d: aux is busy\n", dp->index);
        return -EBUSY;
    }

    switch (msg->request & ~DP_AUX_I2C_MOT){
        case DP_AUX_NATIVE_WRITE:
        case DP_AUX_I2C_WRITE:
        case DP_AUX_I2C_WRITE_STATUS_UPDATE:
            is_write = true;
            ljm_dptx_aux_write_fifo(dp, msg->buffer, msg->size);
            break;
        case DP_AUX_NATIVE_READ:
        case DP_AUX_I2C_READ:
            break;
        default:
            WARN_PRINT("dp %d: invalid aux request %x\n", dp->index, msg->request);
            return -EINVAL;
    }

    ljm_dptx_aux_send_request(dp, ((uint32_t)msg->request) << 8, msg->address, msg->size, (msg->size == 0), false);
    ret = ljm_dptx_aux_wait_reply(dp);
    ljm_dptx_reg_read(dp, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state

    if(ret < 0)
    {
        DEBUG_PRINT("dp %d: aux reply error(%d) when %s %s [0x%x]: %*ph\n",
            dp->index, ret,
            is_write ? "write" : "read",
            (msg->request == DP_AUX_NATIVE_READ || msg->request == DP_AUX_NATIVE_WRITE) ? "DPCD" : "I2C",
            msg->address, (int)msg->size, msg->buffer);
        goto error;
    }


    reply_code = ljm_dptx_aux_reply_code(dp);
    msg->reply = (u8)(reply_code & 0xf);

    if (reply_code == TR_DPTX_AUX_REPLY_ACK) {
        if (!is_write && msg->size > 0)
            ret = ljm_dptx_aux_read_fifo(dp, msg->buffer, msg->size);
        else
            ret = msg->size;
    }
    else {
        DEBUG_PRINT("dp %d: aux reply error code(0x%x)\n", dp->index, reply_code);
        ret = -EREMOTEIO;
        goto error;
    }

    return ret;
error:
    if (ret < 0 && !is_write) {
        int reply_size = ljm_dptx_aux_read_fifo(dp, msg->buffer, msg->size);
        if (reply_size > 0) {
            DEBUG_PRINT("dp %d aux error (%d), but read %s [0x%x@%d]: %*ph\n",
                dp->index, ret,
                (msg->request == DP_AUX_NATIVE_READ) ? "DPCD" : "I2C",
                msg->address, (int)msg->size, reply_size, msg->buffer);
        }
    }
    return ret;
}

static ssize_t ljm_dp_aux_transfer_mcu(struct drm_dp_aux *aux, struct drm_dp_aux_msg *msg)
{
    struct dp_dev *dp = container_of(aux, struct dp_dev, aux);

    if(!dp->is_edp)
        return _ljm_dp_aux_transfer_mcu(dp, msg);

    switch (msg->request & ~DP_AUX_I2C_MOT){
        case DP_AUX_NATIVE_WRITE:
            return _ljm_dp_aux_transfer_mcu(dp, msg);
        case DP_AUX_NATIVE_READ:
            return _ljm_dp_aux_transfer_deaf(dp, msg);
        case DP_AUX_I2C_WRITE:
        case DP_AUX_I2C_WRITE_STATUS_UPDATE:
        case DP_AUX_I2C_READ:
            DEBUG_PRINT("Can't handle requests of i2c.\n");
            return -EINVAL;
        default:
            WARN_PRINT("dp %d: invalid aux request %x\n", dp->index, msg->request);
            return -EINVAL;
    }
}

static ssize_t ljm_dp_aux_transfer_nomcu(struct drm_dp_aux *aux, struct drm_dp_aux_msg *msg)
{
    struct dp_dev *dp = container_of(aux, struct dp_dev, aux);

    if(dp->disable_aux)
        return _ljm_dp_aux_transfer_deaf(dp, msg);
    else
        return _ljm_dp_aux_transfer_native(dp, msg);

}

/**
 * ljm_edp_delayed_display_work - 延时打开edp屏幕
 * @work: work指针，程序依此获取dp_dev；
 *
 * 该函数主要用于延时打开edp屏幕的显示。
 *
 * 在mode_set过程中，edp屏会出现几秒钟的花屏。为避免花屏显示在屏幕上，在mode_set开始时，会通过DP_EDP_DISPLAY_CONTROL_REGISTER寄存器强制输
 * 出黑色图像，此时edp源的图像数据不会被显示。mode_set结束时，启动此delay_work，在一定时间延时后，待信号稳定再关闭BLACK_VIDEO_ENABLE模式。
 *
*/
static void ljm_edp_delayed_display_work(struct work_struct *work)
{
    struct dp_dev *dp_dev = container_of(work, struct dp_dev, delayed_display_wk.work);
    drm_dp_dpcd_writeb(&dp_dev->aux, DP_EDP_DISPLAY_CONTROL_REGISTER, 0x0);   /* 关闭BLACK_VIDEO_ENABLE模式，恢复正常显示 */
    return;
}

void ljm_dp_backlight_gpio_sel(struct dp_dev *dp, int bl_pwr_sel, int pwm_sel)
{
    switch(bl_pwr_sel)
    {
        case 0: dp->gpio_backlight = 0; dp->gpio_power = 1; break;
        case 1: dp->gpio_backlight = 2; dp->gpio_power = 3; break;
        case 2: dp->gpio_backlight = 4; dp->gpio_power = 5; break;
        case 3: dp->gpio_backlight = 6; dp->gpio_power = 7; break;
        default:
            dp->gpio_backlight = NO_BL_PWM_SEL;
            dp->gpio_power = NO_BL_PWM_SEL;
            break;
    }

    dp->pwm_ch = pwm_sel;
    return;
}

/**
 * ljm_dp_mcu_init - 分配gpio给dp
 * @dp_dev: dp/edp设备指针；
 *
*/
void ljm_dp_mcu_init(struct dp_dev *dp_dev, int extra_pipe_index)
{
    switch(extra_pipe_index)
    {
        case 0: dp_dev->gpio_data = 0; dp_dev->gpio_direction = 1; dp_dev->mcu_aux = true; break;
        case 1: dp_dev->gpio_data = 2; dp_dev->gpio_direction = 3; dp_dev->mcu_aux = true; break;
        case 2: dp_dev->gpio_data = 4; dp_dev->gpio_direction = 5; dp_dev->mcu_aux = true; break;
        case 3: dp_dev->gpio_data = 6; dp_dev->gpio_direction = 7; dp_dev->mcu_aux = true; break;
        default: dp_dev->mcu_aux = false;
    }
    return;
}

/**
 * ljm_dp_init - init dp controller
 * @dp_dev: pointer to dp device
 * @base: virtual base adddress
 * @offset: dp register offset from mmio_base
 * @index: dp index, such as 0, 1, 2, 3
 *
 * This function initializes dp controller and phy.
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_dp_init(struct dp_dev *dp_dev, void *base, int offset, struct ljm_port_info *info)
{
    int ret;
    int index = info->sub_id;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);

    HEADER_ARG("dp_dev=%lx base=%lx offset=%x index=%d",
        (unsigned long)dp_dev, (unsigned long)base, offset, index);

    dp_dev->is_edp = info->is_edp;
    dp_dev->once_edp = dp_dev->is_edp ? ljm_dev->once_edp : false;
    dp_dev->disable_aux = info->disable_aux;
    ljm_dptx_init(dp_dev, base, offset, info->sub_id);

    //mutex_init(&dp_dev->lock);

    dp_dev->connected = false;
    dp_dev->max_lanes = 4;  /* max lanes is 4 */
    ljm_dp_mcu_init(dp_dev, info->extra_pipe_index);
    ljm_dp_vedid_vdpcd_add(dp_dev, info);

    dp_dev->aux.name = kasprintf(GFP_KERNEL, "DPDDC-%c", index);
    if(dp_dev->mcu_aux)
        dp_dev->aux.transfer = ljm_dp_aux_transfer_mcu;
    else
        dp_dev->aux.transfer = ljm_dp_aux_transfer_nomcu;
    dp_dev->aux.dev = dp_dev->ddev->dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    dp_dev->aux.drm_dev = dp_dev->ddev;
#endif
    ret = drm_dp_aux_register(&dp_dev->aux);
    if (ret) {
        FOOTER_ARG("r=%d", ret);
        return ret;
    }

    if(dp_dev->is_edp)
    {
        INIT_DELAYED_WORK(&dp_dev->delayed_display_wk, ljm_edp_delayed_display_work);
        ljm_dp_backlight_gpio_sel(dp_dev, info->bl_pwr_sel, info->pwm_sel);
    }

    FOOTER();
    return 0;
}

/**
 * ljm_dp_init - finish dp controller
 * @dp_dev: pointer to dp device
 *
 * This function finishes dp controller and phy.
 */
void ljm_dp_fini(struct dp_dev *dp_dev)
{
    HEADER_ARG("dp_dev=%lx", (unsigned long)dp_dev);

    drm_dp_aux_unregister(&dp_dev->aux);
    kfree(dp_dev->aux.name);

    //mutex_destroy(&dp_dev->lock);

    FOOTER();
}

#if 0
/**
 * ljm_dp_aux_chk_reply - DP AUX check reply
 * @dp_dev: pointer to DP device
 *
 * This function checks AUX reply
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
static int ljm_dp_aux_chk_reply(struct dp_dev *dp_dev)
{
    unsigned int tmp;

    /* FIXME: clear INTERRUPT_MASK of REPLY_TIMEOUT_MASK and REPLY_RECEIVED_MASK? */
    /* check REPLY_RECEIVED or REPLY_TIMEOUT */
    tmp = DPTX_REG_RD(dp_dev, DPTX_INTERRUPT_STATE_RO);
    while ((tmp & (m_INT_REPLY_TIMEOUT | m_INT_REPLY_RECEIVED)) == 0) {
        udelay(1);
        tmp = DPTX_REG_RD(dp_dev, DPTX_INTERRUPT_STATE_RO);
    }

    if (tmp & m_INT_REPLY_TIMEOUT)
        return -ETIMEDOUT;

    /* read AUX_REPLY_CODE */
    tmp = DPTX_REG_RD(dp_dev, DPTX_AUX_REPLAY_CODE);
    if (tmp & AUX_NACK)
        return -DP_AUX_NACK;
    else if (tmp & AUX_DEFER)
        return -DP_AUX_DEFER;
    else if (tmp & AUX_I2C_NACK)
        return -DP_AUX_I2C_NACK;
    else if (tmp & AUX_I2C_DEFER)
        return -DP_AUX_I2C_DEFER;

    return 0;
}

/**
 * ljm_dp_aux_write_req - DP AUX write request
 * @dp_dev: pointer to DP device
 * @addr: AUX address
 * @buf: buffer to write
 * @size: size of buffer
 *
 * This function write a buffer to AUX device start with an address
 *
 * Returns: the number of bytes transferred on success, or a negative error
 * code on failure. -ETIMEDOUT is returned if mailbox message was not send
 * success; -EINVAL is returned if get the wrong data size after message
 * is sent
 */
int ljm_dp_aux_write_req(struct dp_dev *dp_dev, unsigned int addr, unsigned char *buf, int size)
{
    int left_size = size;
    int cur_size;
    int i;
    int r;

    while (left_size > 0) {
        /* write AUX_ADDRESS */
        DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, addr);

        /* write AUX_WRITE_FIFO */
        if (left_size > 16)
            cur_size = 16;
        else
            cur_size = left_size;
        left_size -= cur_size;
        for (i=0; i<cur_size; i++) {
            DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, *buf++);
        }

        /* write AUX_COMMAND */
        DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_WRITE) | (cur_size - 1));

        /* check reply */
        r = ljm_dp_aux_chk_reply(dp_dev);
        if (r) {
            DRM_ERROR("dp %d: aux write reply error %d\n", dp_dev->index, r);
            return r;
        }
    }

    return 0;
}

/**
 * ljm_dp_aux_read_req - DP AUX write request
 * @dp_dev: pointer to DP device
 * @addr: AUX address
 * @buf: buffer for read
 * @size: size of buffer
 *
 * This function read a buffer from AUX device start with an address
 *
 * Returns: Returns the number of bytes transferred on success, or a negative
 * error code on failure. -ETIMEDOUT is returned if mailbox message was
 * not send successfully;
 */
int ljm_dp_aux_read_req(struct dp_dev *dp_dev, unsigned int addr, unsigned char *buf, int size)
{
    int left_size = size;
    int cur_size;
    int i;
    int r;

    while (left_size > 0) {
        /* write AUX_ADDRESS */
        DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, addr);

        /* write AUX_WRITE_FIFO */
        if (left_size > 16)
            cur_size = 16;
        else
            cur_size = left_size;
        left_size -= cur_size;

        /* write AUX_COMMAND */
        DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_READ) | (cur_size - 1));

        /* check reply */
        r = ljm_dp_aux_chk_reply(dp_dev);
        if (r) {
            DRM_ERROR("dp %d: aux read reply error %d\n", dp_dev->index, r);
            return r;
        }

        for (i=0; i<cur_size; i++)
            *buf++ = (unsigned char)DPTX_REG_RD(dp_dev, DPTX_AUX_REPLAY_DATA);
    }

    return 0;
}

/**
 * ljm_dp_i2c_read_req - DP I2C read request over AUX
 * @hw_dev: pointer to DP device
 * @target: 7-bit i2c target address
 * @addr: i2c sub address
 * @buf: buffer for read
 * @size: size of buffer
 *
 * This function read a buffer from AUX device start with an address
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_dp_i2c_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct dp_dev *dp_dev = (struct dp_dev *)hw_dev;
    int left_size = size;
    int cur_size;
    int i;
    int r;

    if (size <= 0)
        return 0;

    /* write AUX_ADDRESS */
    DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, (target << 1) | DDC_WR_BIT);  /* FIXME: 7-bit target address or 8-bit target address with RD/WR ? */
    /* write sub address */
    DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, addr);

    /* write AUX_COMMAND */
    DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_WRITE_MOT));
    /* check reply */
    r = ljm_dp_aux_chk_reply(dp_dev);
    if (r) {
        DRM_ERROR("dp %d: aux i2c read reply error %d\n", dp_dev->index, r);
        return r;
    }

    while (left_size > 0) {
        /* write AUX_WRITE_FIFO */
        if (left_size > 16)
            cur_size = 16;
        else
            cur_size = left_size;
        left_size -= cur_size;

        DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, (target << 1) | DDC_RD_BIT);  /* FIXME: 7-bit target address or 8-bit target address with RD/WR ? */

        if (left_size == 0) {
            if (cur_size > 1) {
                /* write AUX_COMMAND */
                DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_READ_MOT) | (cur_size - 2));

                /* check reply */
                r = ljm_dp_aux_chk_reply(dp_dev);
                if (r) {
                    DRM_ERROR("dp %d: aux i2c read reply error %d\n", dp_dev->index, r);
                    return r;
                }

                /* read the buffer */
                for (i=0; i < (cur_size - 1); i++)
                    *buf++ = (unsigned char)DPTX_REG_RD(dp_dev, DPTX_AUX_REPLAY_DATA);
            }

            /* read the last byte */
            /* write AUX_COMMAND */
            DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_READ));

            /* check reply */
            r = ljm_dp_aux_chk_reply(dp_dev);
            if (r) {
                DRM_ERROR("dp %d: aux i2c read reply error %d\n", dp_dev->index, r);
                return r;
            }

            *buf++ = (unsigned char)DPTX_REG_RD(dp_dev, DPTX_AUX_REPLAY_DATA);
            return 0;
        } else {
            /* write AUX_COMMAND */
            DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_READ_MOT) | (cur_size - 1));

            /* check reply */
            r = ljm_dp_aux_chk_reply(dp_dev);
            if (r) {
                DRM_ERROR("dp %d: aux i2c read reply error %d\n", dp_dev->index, r);
                return r;
            }

            /* read the buffer */
            for (i=0; i < cur_size; i++)
                *buf++ = (unsigned char)DPTX_REG_RD(dp_dev, DPTX_AUX_REPLAY_DATA);
        }
    }

    return 0;
}

/**
 * ljm_dp_i2c_write_req - DP I2C write request over AUX
 * @hw_dev: pointer to DP device
 * @target: 7-bit i2c target address
 * @addr: i2c sub address
 * @buf: buffer for write
 * @size: size of buffer
 *
 * This function read a buffer from AUX device start with an address
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_dp_i2c_write_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct dp_dev *dp_dev = (struct dp_dev *)hw_dev;
    int left_size = size;
    int cur_size;
    int i;
    int r;

    if (size <= 0)
        return 0;

    /* write AUX_ADDRESS */
    DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, (target << 1) | DDC_WR_BIT);  /* FIXME: 7-bit target address or 8-bit target address with RD/WR ? */
    /* write sub address */
    DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, addr);

    /* write AUX_COMMAND */
    DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_WRITE_MOT));
    /* check reply */
    r = ljm_dp_aux_chk_reply(dp_dev);
    if (r) {
        DRM_ERROR("dp %d: aux i2c write reply error %d\n", dp_dev->index, r);
        return r;
    }

    while (left_size > 0) {
        /* write AUX_WRITE_FIFO */
        if (left_size > 16)
            cur_size = 16;
        else
            cur_size = left_size;
        left_size -= cur_size;

        DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, (target << 1) | DDC_WR_BIT);  /* FIXME: 7-bit target address or 8-bit target address with RD/WR ? */
        for (i=0; i < (cur_size - 1); i++) {
            DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, *buf++);
        }

        if (left_size == 0) {
            if (cur_size > 1) {
                /* write AUX_COMMAND */
                DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_WRITE_MOT) | (cur_size - 2));

                /* check reply */
                r = ljm_dp_aux_chk_reply(dp_dev);
                if (r) {
                    DRM_ERROR("dp %d: aux i2c write reply error %d\n", dp_dev->index, r);
                    return r;
                }
            }

            /* write the last byte */
            DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, *buf++);
            /* write AUX_COMMAND */
            DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_WRITE));

            /* check reply */
            r = ljm_dp_aux_chk_reply(dp_dev);
            if (r) {
                DRM_ERROR("dp %d: aux i2c read reply error %d\n", dp_dev->index, r);
                return r;
            }

            return 0;
        } else {
            /* write the current last byte */
            DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, *buf++);
            /* write AUX_COMMAND */
            DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_WRITE_MOT) | (cur_size - 1));

            /* check reply */
            r = ljm_dp_aux_chk_reply(dp_dev);
            if (r) {
                DRM_ERROR("dp %d: aux i2c read reply error %d\n", dp_dev->index, r);
                return r;
            }
        }
    }

    return 0;
}

/**
 * ljm_dp_i2c_raw_write_req - DP I2C raw write request over AUX
 * @hw_dev: pointer to DP device
 * @target: 7-bit i2c target address
 * @data: data for write
 * @size: size of buffer
 *
 * This function read a buffer from AUX device start with an address
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_dp_i2c_raw_write_req(void *hw_dev, unsigned char target, unsigned char data)
{
    struct dp_dev *dp_dev = (struct dp_dev *)hw_dev;
    int r;

    /* write AUX_ADDRESS */
    DPTX_REG_WR(dp_dev, DPTX_AUX_ADDRESS, (target << 1) | DDC_WR_BIT);  /* FIXME: 7-bit target address or 8-bit target address with RD/WR ? */
    /* write data */
    DPTX_REG_WR(dp_dev, DPTX_AUX_WRTIE_FIFO, data);

    /* write AUX_COMMAND */
    DPTX_REG_WR(dp_dev, DPTX_AUX_COMMAND, v_COMMAND(AUX_I2C_WRITE));
    /* check reply */
    r = ljm_dp_aux_chk_reply(dp_dev);
    if (r) {
        DRM_ERROR("dp %d: aux i2c write reply error %d\n", dp_dev->index, r);
        return r;
    }

    return 0;
}
#endif

int ljm_dp_train_link(struct dp_dev *dp);
int ljm_dptx_train_link_no_aux(struct dp_dev *dp_dev, uint32_t bw_code, uint32_t lane_count, bool edp_port);
int ljm_dptx_mode_set(struct dp_dev *dp_dev, const struct drm_display_mode *mode);

void ljm_dp_dpms(struct dp_dev *dp_dev, int mode)
{
    HEADER_ARG("dp_dev=%lx mode=%d", (unsigned long)dp_dev, mode);

    DEBUG_WARN_ON(2);    /* print call trace if breakId == 2 */

    /* TODO: add dpms */
    switch (mode) {
    case DRM_MODE_DPMS_ON:
        /* power up dp */
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SOURCE_ENABLE, 0x1);
        /* added by hym on 20220301 */



            /* link training */

            /* If training result is changed, update the video config */
        break;
    case DRM_MODE_DPMS_STANDBY:
    case DRM_MODE_DPMS_SUSPEND:
    case DRM_MODE_DPMS_OFF:
        /* power down dp */
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SOURCE_ENABLE, 0x0);
        break;
    }

    FOOTER();
}

/**
 * ljm_dp_hpd_sense - get sink hpd state
 * @ dp_dev: pointer to dp device
 *
 * Return: true if connected or false if disconnected.
 */
bool ljm_dp_hpd_sense(struct dp_dev *dp_dev)
{
    return ljm_dptx_hpd_sense(dp_dev);
}


/*****************************************************************************
 *  函 数 名: ljm_dp_hpd_enable
 *  功能描述: 设置DP接口热插拔使能
 *  输入参数: struct dp_dev *dp_dev  DP接口结构体指针
 *            int enable             使能参数
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年4月26日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dp_hpd_enable(struct dp_dev *dp_dev, int enable)
{
    unsigned char data;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);

    HEADER_ARG("dp_dev=%lx enable=%d", (unsigned long)dp_dev, enable);

    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
    {
        ljm_dptx_reg_write(dp_dev, TR_DPTX_INTERRUPT_MASK, ~(TR_DPTX_INTERRUPT_HPD_IRQ | TR_DPTX_INTERRUPT_HPD_EVENT));
        FOOTER();
        return;
    }

    if (ljm_dev->dp_irq_is_extra)
    {
        if (enable)
        {
            ljm_i2c_dw_read_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_CONFIGURATION_SUBADR, &data, 1);
            data |= (1u << dp_dev->index);
            ljm_i2c_dw_write_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_CONFIGURATION_SUBADR, &data, 1);
        }
        else
        {
            ljm_i2c_dw_read_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_CONFIGURATION_SUBADR, &data, 1);
            data &= ~(1u << dp_dev->index);
            ljm_i2c_dw_write_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_CONFIGURATION_SUBADR, &data, 1);
        }
    }
    FOOTER();
}

int ljm_dp_dc_irq_process(struct dp_dev *dp_dev)
{
    uint32_t status;

    status = ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE);
    /**
     * 实际应用中发现，中断发生时只有bit0会被置1，bit1始终为0。
     * 在mode_set中(尚不明确是在fast_train还是full_train过程中)可能导致bit1为1。
     */
    if(status & TR_DPTX_INTERRUPT_HPD_EVENT)
        return LJM_IRQ_TYPE_HPD;

    return LJM_IRQ_TYPE_NONE;
}

int ljm_dp_phy_irq_process(struct dp_dev *dp_dev)
{
    return ljm_dptx_phy_irq_clear(dp_dev);
}

/**
 * ljm_dp_get_edid - get EDID data
 * @ dp_dev: pointer to dp device
 * @ psize: pointer to edid size
 *
 * Return: Pointer to valid EDID or NULL if we couldn't find any.
 */
struct edid * ljm_dp_get_edid(struct ljm_connector *ljm_connector, int *psize)
{
    struct edid *cur_edid = NULL;
    struct ljm_port *ljm_port = to_ljm_port(ljm_connector->encoder);
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    struct dp_dev *dp_dev = &ljm_port->vdev.dp;
    uint8_t *block, *new;
    int size;
    struct mcu_arg_edid args;

    HEADER_ARG("dp_dev=%p psize=%p", dp_dev, psize);

    if((ljm_dev->mcu) && (dp_dev->mcu_aux))
    {
        if(!dp_dev->is_edp)
        {
            block =  kmalloc(DDC_EDID_SIZE, GFP_KERNEL);
            args.gpio_direction = dp_dev->gpio_direction;
            args.gpio_data = dp_dev->gpio_data;
            args.subaddr = 0;
            args.len = DDC_EDID_SIZE;
            args.buff = block;
            if(ljm_mcu_handle(ljm_dev->mcu, MCU_CMD_EDID, &args) < 0 )
                goto OnError;

            if (block[DDC_EDID_SIZE-2] != 0)    /* there are extensions, continue ddc reading. */
            {
                size = (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE;
                new = krealloc(block, size, GFP_KERNEL);
                if (!new)
                    goto OnError;

                memset(&args, 0, sizeof(struct mcu_arg_edid));
                args.gpio_direction = dp_dev->gpio_direction;
                args.gpio_data = dp_dev->gpio_data;
                args.subaddr = 0x80;
                args.len = DDC_EDID_SIZE;
                args.buff = new+DDC_EDID_SIZE;
                if(ljm_mcu_handle(ljm_dev->mcu, MCU_CMD_EDID, &args) < 0)
                    goto OnError;

                block = new;
            }
            cur_edid = (struct edid *)block;
        }
        else
            cur_edid = vedid_get(dp_dev);
    }
    else
    {
        if(!dp_dev->disable_aux)
            cur_edid = drm_get_edid(&ljm_connector->base, &dp_dev->aux.ddc);
        else
            cur_edid = vedid_get(dp_dev);
    }

    if (!cur_edid)
        goto OnError;
/*dp由于aux不稳定的原因，读取edid容易出错。这段代码用于将edid修正为仅block0，并纠正checksum*/
#if 0
    {
        int i;
        uint8_t sum=0;
        if(cur_edid->extensions)
        {
            cur_edid->checksum += cur_edid->extensions;
            cur_edid->extensions -= cur_edid->extensions;
        }
        /*重新计算checksum*/
        block = (uint8_t *)cur_edid;
        for(i=0; i<127; i++)
        {
            sum += block[i];

        }
        cur_edid->checksum = 0 - sum;
    }
#endif

#ifdef DEBUG
    {
        int i, offset;
        unsigned char data;
        char line[] = "       0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f \n";
        char star[] = "******************************************************\n";
        block = (uint8_t *)cur_edid;
        /* dump EDID */
        printk("Dump EDID with %d bytes\n", (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE);
        printk(star);
        for (i = 0, offset = 0; i < (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE; ++i) {
            if (i % 16 == 0) {
                printk(line);
                offset = 0;
                offset += sprintf(line + offset, "%04x: ", i);  /* print address */
            }

            data = block[i];
            offset += sprintf(line + offset, "%02x ", data);
        }

        if (offset) {
            line[offset] = '\0';
            printk(line);
        }
    }
#endif /*DEBUG*/
    if (psize) {
        *psize = DDC_EDID_SIZE;
        *psize += DDC_EDID_SIZE * cur_edid->extensions;
    }

    FOOTER_ARG("edid=%p", cur_edid);
    return cur_edid;

OnError:
    WARN_PRINT("DP %d: read edid failed\n", dp_dev->index);
    if (psize) *psize = 0;
    FOOTER_ARG("edid=%p", cur_edid);
    return NULL;
}

static bool ljm_dp_bandwidth_ok(struct dp_dev *dp,
                   const struct drm_display_mode *mode, u32 bpp,
                   unsigned int lanes, unsigned int rate)
{
    u32 max_bw, req_bw;

    req_bw = mode->clock * bpp / 8;
    max_bw = lanes * rate;
    if (req_bw > max_bw)
        return false;

    return true;
}

static void ljm_dp_link_caps_reset(struct ljm_dp_link_caps *caps)
{
    caps->enhanced_framing = false;
    caps->tps3_supported = false;
    caps->tps4_supported = false;
    caps->fast_training = false;
    caps->channel_coding = false;
}

static void ljm_dp_link_reset(struct ljm_dp_link *link)
{
    link->vsc_sdp_extension_for_colorimetry_supported = 0;
    link->sink_count = 0;
    link->revision = 0;

    ljm_dp_link_caps_reset(&link->caps);
    memset(link->dpcd, 0, sizeof(link->dpcd));

    link->rate = 0;
    link->lanes = 0;
}

static int ljm_dp_read_extended_dpcd_caps(struct drm_dp_aux *aux,
                      u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
    u8 dpcd_ext[6];
    int ret;

    /*
     * Prior to DP1.3 the bit represented by
     * DP_EXTENDED_RECEIVER_CAP_FIELD_PRESENT was reserved.
     * If it is set DP_DPCD_REV at 0000h could be at a value less than
     * the true capability of the panel. The only way to check is to
     * then compare 0000h and 2200h.
     */
    if (!(dpcd[DP_TRAINING_AUX_RD_INTERVAL] & (1 << 7))) // kernel-drm: #define DP_EXTENDED_RECEIVER_CAP_FIELD_PRESENT (1 << 7)
        return 0;

    ret = drm_dp_dpcd_read(aux, DP_DP13_DPCD_REV, &dpcd_ext,
                   sizeof(dpcd_ext));
    if (ret < 0)
        return ret;
    if (ret != sizeof(dpcd_ext))
        return -EIO;

    if (dpcd[DP_DPCD_REV] > dpcd_ext[DP_DPCD_REV]) {
        DRM_DEBUG_KMS("%s: Extended DPCD rev less than base DPCD rev (%d > %d)\n",
                  aux->name, dpcd[DP_DPCD_REV],
                  dpcd_ext[DP_DPCD_REV]);
        return 0;
    }

    if (!memcmp(dpcd, dpcd_ext, sizeof(dpcd_ext)))
        return 0;

    DRM_DEBUG_KMS("%s: Base DPCD: %*ph\n",
              aux->name, DP_RECEIVER_CAP_SIZE, dpcd);

    memcpy(dpcd, dpcd_ext, sizeof(dpcd_ext));

    return 0;
}

int ljm_dp_read_dpcd_caps(struct drm_dp_aux *aux,
              u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
    int ret;

    ret = drm_dp_dpcd_read(aux, DP_DPCD_REV, dpcd, DP_RECEIVER_CAP_SIZE);
    if (ret < 0)
        return ret;
    if (ret != DP_RECEIVER_CAP_SIZE || dpcd[DP_DPCD_REV] == 0)
        return -EIO;

    ret = ljm_dp_read_extended_dpcd_caps(aux, dpcd);
    if (ret < 0)
        return ret;

    return ret;
}

static inline bool
ljm_dp_fast_training_cap(const u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
    return dpcd[DP_DPCD_REV] >= 0x11 &&
        (dpcd[DP_MAX_DOWNSPREAD] & DP_NO_AUX_HANDSHAKE_LINK_TRAINING);
}

static inline bool
ljm_dp_channel_coding_supported(const u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
    return dpcd[DP_MAIN_LINK_CHANNEL_CODING] & DP_SET_ANSI_8B10B;
}

/**
 * ljm_dp_link_probe - get sink capablility data
 * @ dp_dev: pointer to dp device
 *
 * Return: 0 success, otherwise error code.
 */
int ljm_dp_link_probe(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    u8 dpcd;
    int ret;

    HEADER_ARG("dp_dev=%p", dp);

    ljm_dp_link_reset(link);

    ret = ljm_dp_read_dpcd_caps(&dp->aux, link->dpcd);
    if (ret < 0)
        return ret;

    drm_dp_read_desc(&dp->aux, &link->desc, drm_dp_is_branch(link->dpcd));

    /* Note: ljm_drv can't suppport multi-sinks. Just jump it!  */
    //if (dw_dp_has_sink_count(link->dpcd, &link->desc)) {
        //ret = drm_dp_read_sink_count(&dp->aux);
        //if (ret < 0)
            //return ret;

        //link->sink_count = ret;

        /* Dongle connected, but no display */
        //if (!link->sink_count)
            //return -ENODEV;
    //}

    ret = drm_dp_dpcd_readb(&dp->aux, DP_DPRX_FEATURE_ENUMERATION_LIST, &dpcd);
    if (ret < 0)
        return ret;

    link->vsc_sdp_extension_for_colorimetry_supported = !!(dpcd & DP_VSC_SDP_EXT_FOR_COLORIMETRY_SUPPORTED);

    link->revision = link->dpcd[DP_DPCD_REV];
    link->rate = drm_dp_max_link_rate(link->dpcd);
    link->lanes = drm_dp_max_lane_count(link->dpcd);

    link->caps.enhanced_framing = drm_dp_enhanced_frame_cap(link->dpcd);
    link->caps.tps3_supported = drm_dp_tps3_supported(link->dpcd);
    link->caps.tps4_supported = drm_dp_tps4_supported(link->dpcd);
    link->caps.fast_training = ljm_dp_fast_training_cap(link->dpcd);
    link->caps.channel_coding = ljm_dp_channel_coding_supported(link->dpcd);
    link->caps.ssc = !!(link->dpcd[DP_MAX_DOWNSPREAD] & DP_MAX_DOWNSPREAD_0_5);

    FOOTER_ARG("r=%d", ret);
    return ret;
}

int ljm_dp_mode_set(struct dp_dev *dp_dev, struct drm_display_mode *mode)
{
    HEADER_ARG("dp_dev=%lx mode=%p", (unsigned long)dp_dev, mode);

    drm_mode_copy(&dp_dev->mode, mode);
    ljm_dptx_mode_set(dp_dev, mode);

    FOOTER();
    return 0;
}

static void ljm_dp_link_train_init(struct ljm_dp_link_train *train)
{
    struct ljm_dp_link_train_set *request = &train->request;
    struct ljm_dp_link_train_set *adjust = &train->adjust;
    unsigned int i;

    for (i = 0; i < 4; i++) {
        request->voltage_swing[i] = 0;
        adjust->voltage_swing[i] = 0;

        request->pre_emphasis[i] = 0;
        adjust->pre_emphasis[i] = 0;

        request->voltage_max_reached[i] = false;
        adjust->voltage_max_reached[i] = false;

        request->pre_max_reached[i] = false;
        adjust->pre_max_reached[i] = false;
    }

    train->clock_recovered = false;
    train->channel_equalized = false;
}

static bool ljm_dp_link_train_valid(const struct ljm_dp_link_train *train)
{
    return train->clock_recovered && train->channel_equalized;
}

static int ljm_dp_link_train_set_pattern(struct dp_dev *dp, u32 pattern)
{
    u8 buf = 0;
    int ret;
    if (pattern && pattern != DP_TRAINING_PATTERN_4) {
        buf |= DP_LINK_SCRAMBLING_DISABLE;
        ljm_dptx_reg_write(dp, TR_DPTX_DISABLE_SCRAMBLING, 0x01);
    } else {
        ljm_dptx_reg_write(dp, TR_DPTX_DISABLE_SCRAMBLING, 0x0);
    }

    switch (pattern) {
    case DP_TRAINING_PATTERN_DISABLE:
        ljm_dptx_reg_write(dp, TR_DPTX_TRAINING_PATTERN_SET, 0x00); // Training off
        break;
    case DP_TRAINING_PATTERN_1:
        ljm_dptx_reg_write(dp, TR_DPTX_TRAINING_PATTERN_SET, 0x01); // Training pattern 1, clock recovery
        break;
    case DP_TRAINING_PATTERN_2:
        ljm_dptx_reg_write(dp, TR_DPTX_TRAINING_PATTERN_SET, 0x02); // Training pattern 2, inter-lane alignment and symbol recovery
        break;
    case DP_TRAINING_PATTERN_3:
        ljm_dptx_reg_write(dp, TR_DPTX_TRAINING_PATTERN_SET, 0x03); // Training pattern 3, inter-lane alignment and symbol recovery
        break;
    case DP_TRAINING_PATTERN_4:
        ljm_dptx_reg_write(dp, TR_DPTX_TRAINING_PATTERN_SET, 0x04); // Training pattern 4, inter-lane alignment and symbol recovery
        break;
    default:
        return -EINVAL;
    }

    ret = drm_dp_dpcd_writeb(&dp->aux, DP_TRAINING_PATTERN_SET,
                 buf | pattern);
    if (ret < 0)
        return ret;

    return 0;
}

static u8 ljm_dp_voltage_max(u8 preemph)
{
    switch (preemph & DP_TRAIN_PRE_EMPHASIS_MASK) {
    case DP_TRAIN_PRE_EMPH_LEVEL_0:
        return DP_TRAIN_VOLTAGE_SWING_LEVEL_3;
    case DP_TRAIN_PRE_EMPH_LEVEL_1:
        return DP_TRAIN_VOLTAGE_SWING_LEVEL_2;
    case DP_TRAIN_PRE_EMPH_LEVEL_2:
        return DP_TRAIN_VOLTAGE_SWING_LEVEL_1;
    case DP_TRAIN_PRE_EMPH_LEVEL_3:
    default:
        return DP_TRAIN_VOLTAGE_SWING_LEVEL_0;
    }
}

static int ljm_dp_link_train_update_vs_emph(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    struct ljm_dp_link_train_set *train_set = &link->train.request;
    unsigned int lanes = link->lanes;
    unsigned int *vs, *pe;
    u8 buf[4];
    int i, ret;
    unsigned int voltage[4]={0}, pre[4]={0};

    vs = train_set->voltage_swing;
    pe = train_set->pre_emphasis;

	for (i = 0; i < lanes; i++) {
		voltage[i] = vs[i];
		pre[i] = pe[i];
	}

    // Set VS and PE based on the DPCD values
    ljm_dptx_set_voltage_swing(dp, voltage[0], voltage[1], voltage[2], voltage[3]);
    ljm_dptx_set_preemphasis(dp, pre[0], pre[1], pre[2], pre[3]);

    for (i = 0; i < lanes; i++) {
        buf[i] = (vs[i] << DP_TRAIN_VOLTAGE_SWING_SHIFT) |
             (pe[i] << DP_TRAIN_PRE_EMPHASIS_SHIFT);
        if (train_set->voltage_max_reached[i])
            buf[i] |= DP_TRAIN_MAX_SWING_REACHED;
        if (train_set->pre_max_reached[i])
            buf[i] |= DP_TRAIN_MAX_PRE_EMPHASIS_REACHED;
    }

    ret = drm_dp_dpcd_write(&dp->aux, DP_TRAINING_LANE0_SET, buf, lanes);

    if (ret < 0)
        return ret;

    return 0;
}

static void ljm_dp_link_get_adjustments(struct ljm_dp_link *link,
                       u8 status[DP_LINK_STATUS_SIZE])
{
    struct ljm_dp_link_train_set *adjust = &link->train.adjust;
    u8 v = 0;
    u8 p = 0;
    unsigned int i;

    for (i = 0; i < link->lanes; i++) {
        v = drm_dp_get_adjust_request_voltage(status, i);
        p = drm_dp_get_adjust_request_pre_emphasis(status, i);
        if (p >=  DP_TRAIN_PRE_EMPH_LEVEL_3) {
            adjust->pre_emphasis[i] = DP_TRAIN_PRE_EMPH_LEVEL_3 >>
                          DP_TRAIN_PRE_EMPHASIS_SHIFT;
            adjust->pre_max_reached[i] = true;
        } else {
            adjust->pre_emphasis[i] = p >> DP_TRAIN_PRE_EMPHASIS_SHIFT;
            adjust->pre_max_reached[i] = false;
        }
        v = min(v, ljm_dp_voltage_max(p));
        if (v >= DP_TRAIN_VOLTAGE_SWING_LEVEL_3) {
            adjust->voltage_swing[i] = DP_TRAIN_VOLTAGE_SWING_LEVEL_3 >>
                           DP_TRAIN_VOLTAGE_SWING_SHIFT;
            adjust->voltage_max_reached[i] = true;
        } else {
            adjust->voltage_swing[i] = v >> DP_TRAIN_VOLTAGE_SWING_SHIFT;
            adjust->voltage_max_reached[i] = false;
        }
    }
}

static void ljm_dp_link_train_adjust(struct ljm_dp_link_train *train)
{
    struct ljm_dp_link_train_set *request = &train->request;
    struct ljm_dp_link_train_set *adjust = &train->adjust;
    unsigned int i;

    for (i = 0; i < 4; i++) {
        if (request->voltage_swing[i] != adjust->voltage_swing[i])
            request->voltage_swing[i] = adjust->voltage_swing[i];
        if (request->voltage_max_reached[i] != adjust->voltage_max_reached[i])
            request->voltage_max_reached[i] = adjust->voltage_max_reached[i];
    }

    for (i = 0; i < 4; i++) {
        if (request->pre_emphasis[i] != adjust->pre_emphasis[i])
            request->pre_emphasis[i] = adjust->pre_emphasis[i];
        if (request->pre_max_reached[i] != adjust->pre_max_reached[i])
            request->pre_max_reached[i] = adjust->pre_max_reached[i];
    }
}

static int ljm_dp_link_clock_recovery(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    u8 status[DP_LINK_STATUS_SIZE];
    unsigned int tries = 0;
    int ret;

    ret = ljm_dp_link_train_set_pattern(dp, DP_TRAINING_PATTERN_1);
    if (ret)
        return ret;

    for (;;) {
        ret = ljm_dp_link_train_update_vs_emph(dp);
        if (ret)
            return ret;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
        drm_dp_link_train_clock_recovery_delay(link->dpcd);
#else
        drm_dp_link_train_clock_recovery_delay(&dp->aux, link->dpcd);
#endif
        ret = drm_dp_dpcd_read_link_status(&dp->aux, status);
        if (ret < 0) {
            ERR_PRINT("failed to read link status: %d\n", ret);
            return ret;
        }

        if (drm_dp_clock_recovery_ok(status, link->lanes)) {
            link->train.clock_recovered = true;
            break;
        }

        ljm_dp_link_get_adjustments(link, status);

        if (link->train.request.voltage_swing[0] ==
            link->train.adjust.voltage_swing[0])
            tries++;
        else
            tries = 0;

        if (tries == 5)
            break;

        ljm_dp_link_train_adjust(&link->train);
    }

    return 0;
}

static int ljm_dp_link_channel_equalization(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    u8 status[DP_LINK_STATUS_SIZE], pattern;
    unsigned int tries;
    int ret;

    if (link->caps.tps4_supported)
        pattern = DP_TRAINING_PATTERN_4;
    else if (link->caps.tps3_supported)
        pattern = DP_TRAINING_PATTERN_3;
    else
        pattern = DP_TRAINING_PATTERN_2;
    ret = ljm_dp_link_train_set_pattern(dp, pattern);
    if (ret)
        return ret;

    for (tries = 1; tries < 5; tries++) {
        ret = ljm_dp_link_train_update_vs_emph(dp);
        if (ret)
            return ret;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
        drm_dp_link_train_channel_eq_delay(link->dpcd);
#else
        drm_dp_link_train_channel_eq_delay(&dp->aux, link->dpcd);
#endif
        ret = drm_dp_dpcd_read_link_status(&dp->aux, status);
        if (ret < 0)
            return ret;

        if (!drm_dp_clock_recovery_ok(status, link->lanes)) {
            dev_err(dp->dev, "clock recovery lost while equalizing channel\n");
            link->train.clock_recovered = false;
            break;
        }

        if (drm_dp_channel_eq_ok(status, link->lanes)) {
            link->train.channel_equalized = true;
            break;
        }

        ljm_dp_link_get_adjustments(link, status);
        ljm_dp_link_train_adjust(&link->train);
    }
    return 0;
}

static int ljm_dp_phy_configure(struct dp_dev *dp, unsigned int rate,
                   unsigned int lanes, bool ssc)
{
    unsigned int bw_code = drm_dp_link_rate_to_bw_code(rate);
    ljm_dptx_set_link_rate(dp, bw_code);
    ljm_dptx_set_lane_count(dp, lanes);

    return 0;
}

static int ljm_dp_link_configure(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    u8 buf[2];
    int ret;

    ret = ljm_dp_phy_configure(dp, link->rate, link->lanes, link->caps.ssc);
    if (ret)
        return ret;

    buf[0] = drm_dp_link_rate_to_bw_code(link->rate); // = link_rate / 27000;
    buf[1] = link->lanes;

    if (link->caps.enhanced_framing) {
        buf[1] |= DP_LANE_COUNT_ENHANCED_FRAME_EN;
        ljm_dptx_reg_write(dp, TR_DPTX_ENHANCED_FRAMING_ENABLE, 0x01);
    } else {
        ljm_dptx_reg_write(dp, TR_DPTX_ENHANCED_FRAMING_ENABLE, 0x0);
    }

    ret = drm_dp_dpcd_write(&dp->aux, DP_LINK_BW_SET, buf, sizeof(buf));
    if (ret < 0)
        return ret;

    buf[0] = link->caps.ssc ? DP_SPREAD_AMP_0_5 : 0;
    buf[1] = link->caps.channel_coding ? DP_SET_ANSI_8B10B : 0;

    ret = drm_dp_dpcd_write(&dp->aux, DP_DOWNSPREAD_CTRL, buf,
                sizeof(buf));
    if (ret < 0)
        return ret;

    return 0;
}

static int ljm_dp_link_downgrade(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;

    switch (link->rate) {
    case 162000:
        return -EINVAL;
    case 270000:
        link->rate = 162000;
        break;
    case 540000:
        link->rate = 270000;
        break;
    case 810000:
        link->rate = 540000;
        break;
    }

    if (!ljm_dp_bandwidth_ok(dp, &dp->mode, 24, link->lanes,
                link->rate))
        return -E2BIG;

    return 0;
}

static int ljm_dp_link_train_full(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    int ret;

retry:
    ljm_dp_link_train_init(&link->train);
    DEBUG_PRINT("full-training link: %u lane%s at %u MHz\n",
         link->lanes, (link->lanes > 1) ? "s" : "", link->rate / 100);

    ret = ljm_dp_link_configure(dp);
    if (ret < 0) {
        ERR_PRINT("failed to configure DP link: %d\n", ret);
        return ret;
    }

    ret = ljm_dp_link_clock_recovery(dp);
    if (ret < 0) {
        ERR_PRINT("clock recovery failed: %d\n", ret);
        goto out;
    }

    if (!link->train.clock_recovered) {
        DEBUG_PRINT("clock recovery failed, downgrading link\n");

        ret = ljm_dp_link_downgrade(dp);
        if (ret < 0)
            goto out;
        else
            goto retry;
    }

    WARN_PRINT("clock recovery succeeded\n");

    ret = ljm_dp_link_channel_equalization(dp);
    if (ret < 0) {
        ERR_PRINT("channel equalization failed: %d\n", ret);
        goto out;
    }

    if (!link->train.channel_equalized) {
        DEBUG_PRINT("channel equalization failed, downgrading link\n");

        ret = ljm_dp_link_downgrade(dp);
        if (ret < 0)
            goto out;
        else
            goto retry;
    }

    WARN_PRINT("channel equalization succeeded\n");

out:
    ljm_dp_link_train_set_pattern(dp, DP_TRAINING_PATTERN_DISABLE);
    return ret;
}

static int ljm_dp_link_train_fast(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    u8 status[DP_LINK_STATUS_SIZE], pattern;
    int ret;

    ljm_dp_link_train_init(&link->train);

    DEBUG_PRINT("fast-training link: %u lane%s at %u MHz\n",
         link->lanes, (link->lanes > 1) ? "s" : "", link->rate / 100);

    ret = ljm_dp_link_configure(dp);
    if (ret < 0) {
        ERR_PRINT("failed to configure DP link: %d\n", ret);
        return ret;
    }

    ret = ljm_dp_link_train_set_pattern(dp, DP_TRAINING_PATTERN_1);
    if (ret)
        goto out;

    usleep_range(500, 1000);

    if (link->caps.tps4_supported)
        pattern = DP_TRAINING_PATTERN_4;
    else if (link->caps.tps3_supported)
        pattern = DP_TRAINING_PATTERN_3;
    else
        pattern = DP_TRAINING_PATTERN_2;
    ret = ljm_dp_link_train_set_pattern(dp, pattern);
    if (ret)
        goto out;

    usleep_range(500, 1000);

    ret = drm_dp_dpcd_read_link_status(&dp->aux, status);
    if (ret < 0) {
        ERR_PRINT("failed to read link status: %d\n", ret);
        goto out;
    }

    if (!drm_dp_clock_recovery_ok(status, link->lanes)) {
        ERR_PRINT("clock recovery failed\n");
        ret = -EIO;
        goto out;
    }

    if (!drm_dp_channel_eq_ok(status, link->lanes)) {
        ERR_PRINT("channel equalization failed\n");
        ret = -EIO;
        goto out;
    }

out:
    ljm_dp_link_train_set_pattern(dp, DP_TRAINING_PATTERN_DISABLE);
    return ret;
}

int ljm_dp_train_link(struct dp_dev *dp)
{
    struct ljm_dp_link *link = &dp->link;
    int ret;

    HEADER_ARG("dp_dev=%p", dp);

    if (link->caps.fast_training) {
        if (ljm_dp_link_train_valid(&link->train)) {
            ret = ljm_dp_link_train_fast(dp);
            if (ret < 0)
                ERR_PRINT("fast link training failed: %d\n", ret);
            else
                return 0;
        }
    }

    ret = ljm_dp_link_train_full(dp);
    if (ret < 0) {
        ERR_PRINT("full link training failed: %d\n", ret);
        return ret;
    }
    WARN_PRINT("dp-%d training success!\n", dp->index);
    FOOTER();
    return 0;
}


int ljm_dp_hpd_set_polarity(struct dp_dev *dp_dev)
{
    unsigned char data;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);


    HEADER_ARG("dp_dev=%p", dp_dev);

    if (ljm_dev->family == CHIP_GP200)/*如果芯片为201情况下*/
    {
        ljm_i2c_dw_read_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_POLARITY_SUBADR, &data, 1);
        ljm_i2c_dw_write_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_POLARITY_SUBADR, &data, 1);
    }

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dp_mode_valid
 *  功能描述: 判断DP接口是否支持该分辨率
 *  输入参数: struct dp_dev *dp_dev
 *            const struct drm_display_mode *mode
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月21日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dp_mode_valid(struct dp_dev *dp_dev, const struct drm_display_mode *mode)
{
    bool r;

//    HEADER_ARG("dp_dev=%p mode=0x%p", dp_dev, mode);

    if (mode->clock > 340000) {  /* 600000 for 4K@60Hz (max), 340000 for 4K@30Hz (max) */
        r = false;
    }
    else {
        r = true;
    }

    if (r) {
        DEBUG_PRINT("DP%d: %dx%d@%dkHz is valid", dp_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    } else {
        DEBUG_PRINT("DP%d: %dx%d@%dkHz is invalid", dp_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    }

//    FOOTER_ARG("%d", r);
    return r;
}

bool ljm_dp_mode_valid_ex(struct dp_dev *dp_dev, const struct drm_display_mode *mode)
{
    bool r;

//    HEADER_ARG("dp_dev=%p mode=0x%p", dp_dev, mode);

    if (mode->clock > 600000) {  /* 600000 for 4K@60Hz (max), 340000 for 4K@30Hz (max) */
        r = false;
    }
    else {
        r = true;
    }

    if (r) {
        DEBUG_PRINT("DP%d: %dx%d@%dkHz is valid", dp_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    } else {
        DEBUG_PRINT("DP%d: %dx%d@%dkHz is invalid", dp_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    }

//    FOOTER_ARG("%d", r);
    return r;
}

/*****************************************************************************
 *  函 数 名: ljm_dp_read_dpcd
 *  功能描述: 通过aux对象读取DPCD信息
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年9月3日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dp_read_dpcd(struct dp_dev *dp_dev)
{
    bool r = false;

    HEADER_ARG("dp_dev=%p", dp_dev);

    if (drm_dp_dpcd_read(&dp_dev->aux, 0x000, dp_dev->dpcd, sizeof(dp_dev->dpcd)) < 0) {
        FOOTER_ARG("r=%s", "false");
        return r; /* aux transfer failed */
    }

    DRM_DEBUG_KMS("DPCD: %*ph\n", (int) sizeof(dp_dev->dpcd), dp_dev->dpcd);

    r = dp_dev->dpcd[DP_DPCD_REV] != 0;

    FOOTER_ARG("r=%s", r ? "true" : "false");
    return r;
}

static void ljm_dp_pps_init(struct dp_dev *dp_dev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    u8 bl_en, pwr_en, pwm;

    HEADER_ARG("dp_dev=%lx", (unsigned long)dp_dev);

    if(dp_dev->gpio_backlight < 0 || dp_dev->gpio_power < 0)
        return;

    bl_en = dp_dev->gpio_backlight;
    pwr_en = dp_dev->gpio_power;
    DEBUG_PRINT("DP%d: backlight GPIO: bl_en=%d pwr_en=%d\n", dp_dev->index, bl_en, pwr_en);

    if (ljm_dev->once_edp && ljm_dev->once_edp_vga_inited) {
        DEBUG_PRINT("DP%d: skip GPIO config\n", dp_dev->index);
        FOOTER();
        return;
    }

    /* pps include vdd bl_en pwm and panel power */
    ljm_gpio_set_mask(ljm_dev->cur_card, bl_en, 1);
    ljm_gpio_set_mask(ljm_dev->cur_card, pwr_en, 1);
    ljm_gpio_set_oen(ljm_dev->cur_card, bl_en, 0);
    ljm_gpio_set_oen(ljm_dev->cur_card, pwr_en, 0);
    /* power on LED and LCD for current board */
    ljm_gpio_set_output(ljm_dev->cur_card, bl_en, 1);
    ljm_gpio_set_output(ljm_dev->cur_card, pwr_en, 1);
    msleep(1);

    if(dp_dev->pwm_ch < 0)
        return;

    pwm = dp_dev->pwm_ch;
    DEBUG_PRINT("DP%d: backlight PWM: pwm=%d\n", dp_dev->index, pwm);

    /* 将4路pwm的一级分频器全部设置为0x01*/
    CR_WR(ljm_dev->cur_card, PWM_LED1_2_FREQ_FISET, 0x01);
    CR_WR(ljm_dev->cur_card, PWM_LED3_4_FREQ_FISET, 0x01);
    /* 将4路pwm的二级分频器全部设置为0xa00 */
    CR_WR(ljm_dev->cur_card, PWM_LED1_FREQ_SECOND_LOWBYTE, 0x00);
    CR_WR(ljm_dev->cur_card, PWM_LED1_FREQ_SECOND_HIGHBYTE, 0xa);
    CR_WR(ljm_dev->cur_card, PWM_LED2_FREQ_SECOND_LOWBYTE, 0x00);
    CR_WR(ljm_dev->cur_card, PWM_LED2_FREQ_SECOND_HIGHBYTE, 0xa);
    CR_WR(ljm_dev->cur_card, PWM_LED3_FREQ_SECOND_LOWBYTE, 0x00);
    CR_WR(ljm_dev->cur_card, PWM_LED3_FREQ_SECOND_HIGHBYTE, 0xa);
    CR_WR(ljm_dev->cur_card, PWM_LED4_FREQ_SECOND_LOWBYTE, 0x00);
    CR_WR(ljm_dev->cur_card, PWM_LED4_FREQ_SECOND_HIGHBYTE, 0xa);

    /* 将指定pwm通道的占空比寄存器设置为0x500，即占空比为50% */
    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_HIGHBYTE+(0x2*pwm), 0x5);
    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_LOWBYTE+(0x2*pwm), 0x00);

    /* enable pwm */
    CR_WR(ljm_dev->cur_card, PWM_LED1_2_3_4_ENABLE, 0x01);
    msleep(80);

    /* enable LED_BL_EN for current board */
    FOOTER();
}

/*****************************************************************************
 *  函 数 名：ljm_edp_self_test_set
 *  功能描述：设置edp的自检模式
 *  输入参数：struct dp_dev *dp_dev
 *            u8 state
 *  输出参数：无
 *  返 回 值：int
 *
 *  修改历史：
 *         1. 日    期：2023年2月22日
 *            作    者：苏徐昆
 *            修改内容：新生成函数
 *
 *****************************************************************************/
int ljm_edp_self_test_set(struct dp_dev *dp_dev, u8 state)
{
    u8 reg_val;

    if (drm_dp_dpcd_readb(&dp_dev->aux, DP_EDP_GENERAL_CAP_1, &reg_val) < 0)
    {
        DRM_DEBUG_KMS("Failed to read DPCD register 0x%x\n",
                      DP_EDP_GENERAL_CAP_1);
        return -EINVAL;
    }

    if (!(reg_val & DP_EDP_PANEL_SELF_TEST_AUX_ENABLE_CAP))
    {
         DRM_DEBUG_KMS("Panel self test incapable to enable, Failed to set state%d\n", state);
         return -EPERM;
    }

    if (drm_dp_dpcd_readb(&dp_dev->aux, DP_EDP_CONFIGURATION_SET, &reg_val) < 0)
    {
        DRM_DEBUG_KMS("Failed to read DPCD register 0x%x\n",
                      DP_EDP_CONFIGURATION_SET);
        return -EINVAL;
    }

    if ((reg_val & DP_PANEL_SELF_TEST_ENABLE) >> 7 == state)
    {
        DRM_DEBUG_KMS("eDP state is same as state%d", state);
        return 0;
    }

    reg_val = state << 7;
    if (drm_dp_dpcd_writeb(&dp_dev->aux, DP_EDP_CONFIGURATION_SET, reg_val) < 0)
    {
        DRM_DEBUG_KMS("Failed to write edp self test mode\n");
        return -EINVAL;
    }

    return 0;
}
#if 0
/* Enable backlight in the panel power control. */
static void ljm_edp_backlight_on(struct dp_dev *dp_dev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    u8 gpio_out, gpio_mask, gpio_oen;

    HEADER_ARG("dp_dev=%lx", (unsigned long)dp_dev);

    gpio_out = REG_R8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_DATA_O);
    gpio_mask = REG_R8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_MASK);
    gpio_oen = REG_R8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_PAD_OEN);

    if((gpio_mask & 0x07) != 0x7 || (gpio_oen & 0x07) != 0x0)
    {
        REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_MASK, gpio_mask | 0x07);
        REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_PAD_OEN, gpio_oen & 0xf8);
    }

    /* gpio2 to bl power */
    REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_DATA_O, (gpio_out & 0xf8) + 0x5);
    msleep(1);
    /* 设置初始PWM数值，pwm 为0xa000，占空比为50% */
    CR_WR(ljm_dev->cur_card, PWM_LED1_2_FREQ_FISET, 0x01);
    CR_WR(ljm_dev->cur_card, PWM_LED1_FREQ_SECOND_LOWBYTE, 0x00);
    CR_WR(ljm_dev->cur_card, PWM_LED1_FREQ_SECOND_HIGHBYTE, 0xa);

    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_HIGHBYTE, 0x2);
    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_LOWBYTE, 0x00);

    /*enable pwm*/
    CR_WR(ljm_dev->cur_card, PWM_LED1_2_3_4_ENABLE, 0x01);

    msleep(80);
    REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_DATA_O, (gpio_out & 0xf8) + 0x4);

    FOOTER();
}

/* Disable backlight in the panel power control. */
static void ljm_edp_backlight_off(struct dp_dev *dp_dev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    u8 gpio_out, gpio_mask, gpio_oen;

    HEADER_ARG("dp_dev=%lx", (unsigned long)dp_dev);

    gpio_out = REG_R8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_DATA_O);
    gpio_mask = REG_R8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_MASK);
    gpio_oen = REG_R8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_PAD_OEN);

    if((gpio_mask & 0x07) != 0x7 || (gpio_oen & 0x07) != 0x0)
    {
        REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_MASK, gpio_mask | 0x07);
        REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_PAD_OEN, gpio_oen & 0xf8);
    }

    /* gpio2 to bl power gpio0 to bl en */
    REG_W8(ljm_dev->cur_card, LJM_IO_MUX_REG + LJM_GPIO_DATA_O, (gpio_out & 0xf8) + 0x1);
    /* 在CPLD控制时，设置使能pwm情况下，占空比为100%，屏幕为全黑 */
    CR_WR(ljm_dev->cur_card, PWM_LED1_2_FREQ_FISET, 0x01);
    CR_WR(ljm_dev->cur_card, PWM_LED1_FREQ_SECOND_LOWBYTE, 0x00);
    CR_WR(ljm_dev->cur_card, PWM_LED1_FREQ_SECOND_HIGHBYTE, 0xa);

    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_HIGHBYTE, 0xa);
    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_LOWBYTE, 0x00);

    /*enable pwm*/
    CR_WR(ljm_dev->cur_card, PWM_LED1_2_3_4_ENABLE, 0x01);

    FOOTER();
}

/*
 * Hook for controlling the panel power control backlight through the bl_power
 * sysfs attribute. Take care to handle multiple calls.
 */
 void ljm_edp_backlight_power(struct ljm_port *ljm_port, bool enable)
{
    struct dp_dev *dp_dev = &ljm_port->vdev.dp;

    DRM_DEBUG_KMS("panel power control backlight %s\n",
                  enable ? "enable" : "disable");

    if (enable)
        ljm_edp_backlight_on(dp_dev);
    else
        ljm_edp_backlight_off(dp_dev);
}
#endif
bool ljm_edp_init_connector(struct ljm_port *ljm_port)
{
    struct ljm_connector *ljm_connector = ljm_port->vdev_connector;
    struct dp_dev *dp_dev = &ljm_port->vdev.dp;

    HEADER_ARG("ljm_port=%px", ljm_port);

    if (!(ljm_connector->ddc_type == LJM_DDC_TYPE_DP
        && dp_dev
        && dp_dev->is_edp)) {
        FOOTER();
        return true;
    }

    /*power control*/
    ljm_dp_pps_init(dp_dev);

    FOOTER_ARG("r=%d", true);
    return true;
}

int ljm_dptx_mode_set(struct dp_dev *dp_dev, const struct drm_display_mode *mode)
{
    uint32_t nvid = 0;
    uint32_t tmp32;
    int no_video_time = 0;

    HEADER_ARG("dp_dev=0x%lx mode=0x%lx: %dx%d@%d",
        (unsigned long)dp_dev, (unsigned long)mode, mode->hdisplay, mode->vdisplay, mode->clock);

    DEBUG_WARN_ON(3);    /* print call trace if breakId == 3 */

    if(ljm_dp_link_probe(dp_dev) < 0 )
    {
        WARN_PRINT("Failed to dp link probe. Force to 1.62Gbps with 4 lanes\n");

        dp_dev->link.rate = drm_dp_bw_code_to_link_rate(0x06); //1.62GHz
        dp_dev->link.lanes = 4;
    }
    DEBUG_PRINT("dp %d: rate=%dMHz lanes=%d", dp_dev->index, dp_dev->link.rate/100, dp_dev->link.lanes);

    /* 打开BLACK_VIDEO_ENABLE模式，输出纯黑图像，避免mode_set过程中花屏 */
    if(dp_dev->is_edp)
        drm_dp_dpcd_writeb(&dp_dev->aux, DP_EDP_DISPLAY_CONTROL_REGISTER, DP_EDP_BLACK_VIDEO_ENABLE);

    /* link training */
    if (dp_dev->once_edp)
        ljm_dptx_train_link_no_aux(dp_dev, drm_dp_link_rate_to_bw_code(dp_dev->link.rate), dp_dev->link.lanes, true);
    else
        ljm_dp_train_link(dp_dev);

    /* mode set */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HTOTAL, mode->htotal);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VTOTAL, mode->vtotal);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HSWIDTH, mode->hsync_end - mode->hsync_start);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VSWIDTH, mode->vsync_end - mode->vsync_start);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HRES, mode->hdisplay);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VRES, mode->vdisplay);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HSTART, mode->htotal - mode->hsync_start);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VSTART, mode->vtotal - mode->vsync_start);
    DEBUG_PRINT("dp %d mode: %d %d %d %d %d %d %d %d %d\n", dp_dev->index,
        mode->hdisplay, mode->hsync_start, mode->hsync_end, mode->htotal,
        mode->vdisplay, mode->vsync_start, mode->vsync_end, mode->vtotal,
        mode->clock);

    tmp32 = 0;
    if (mode->flags & DRM_MODE_FLAG_NHSYNC) tmp32 |= 0x1;
    if (mode->flags & DRM_MODE_FLAG_NVSYNC) tmp32 |= 0x2;
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_POLARITY, tmp32);

    tmp32 = 0xc;
    if (mode->flags & DRM_MODE_FLAG_PHSYNC) tmp32 |= 0x1;
    if (mode->flags & DRM_MODE_FLAG_PVSYNC) tmp32 |= 0x2;
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_SYNC_POLARITY, tmp32);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_MISC0, 0x20);   /* 8bits depth, rgb, synchronous clock (Fix me -sxk) */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_MISC1, 0x0);    /* no stereo video */

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MVID, mode->clock); /* m-vid, user data pixel clock value in KHz */

    ljm_dptx_set_tu_size(dp_dev, mode->clock, 24);  /* TR_DPTX_SRC0_TU_CONFIG (0x1b0) */

    nvid = ljm_dptx_reg_read(dp_dev, TR_DPTX_LINK_BW_SET) * 27000;
    if(unlikely(!nvid))
        nvid = 0x06 * 27000; //nvid为空，则设为默认值0x06。

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_NVID, nvid);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_PIXEL_COUNT, 0x1);

    ljm_dptx_set_user_data_count(dp_dev, mode->hdisplay, 3, dp_dev->link.lanes);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_USER_DATA_CONTROL, 0x2004);  /* src data control 0_1 0x2006 */
    // sparse mode enable
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_CONTROL, 0x0);

    if (dp_dev->is_edp)
    {
        //ljm_edp_self_test_set(dp_dev, 0x1);/* Set edp self test mode, 1 = open, 0 = close */
        ljm_dptx_reg_write(dp_dev, 0x0c8, 0x2004);  /* user data control */

        // enable VSC packets
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC0_INFOFRAME_ENABLE, 0x0); // infoframes off
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC0_3D_VSC_DATA, 0x0);      // clear the 3D/VSC byte
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC_PSR_CONFIG, 0x0);        // psr off
        ljm_dptx_reg_write(dp_dev, 0x368, 0x0000);  /* sec_psr_state */
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_SECONDARY_DATA_WINDOW, 0x80); // secondary data window

        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC0_AUDIO_ENABLE, 0x0);     // audio off
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SECONDARY_STREAM_ENABLE, 0x0);   // secondary channel disabled (Fix me -sxk)
        ljm_dptx_reg_write(dp_dev, TR_DPTX_EDP_CRC_ENABLE, 0x0);        // crc enabled
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC_PSR_3D_ENABLE, 0x0);     // psr/vsc disabled

        ljm_dptx_tmr_wait(dp_dev, 1);   // wait 1us

    }

    DEBUG_PRINT("DP%d: enable the link\n", dp_dev->index);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SOURCE_ENABLE, 0x1); // input source enable
    no_video_time = 8192 * 6 * 2 / (nvid / 1000); // 8192 * 6 * 1.1 / (nvid / 1000)
    ljm_dptx_tmr_wait(dp_dev, no_video_time);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_VIDEO_STREAM_ENABLE, 0x1);   // video stream enable
    ljm_dptx_tmr_wait(dp_dev, 5);    // wait for 5us
    //ljm_dptx_reg_write(dp_dev, TR_DPTX_FORCE_SCRAMBLER_RESET, 0x1);

    if(dp_dev->is_edp)
        schedule_delayed_work(&dp_dev->delayed_display_wk, msecs_to_jiffies(2000));

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_train_link_no_aux
 *  功能描述: 不依赖AUX接口的DP接口连接训练
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t speed
 *            uint32_t lane_count
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_train_link_no_aux(struct dp_dev *dp_dev, uint32_t bw_code, uint32_t lane_count, bool edp_port)
{
    uint8_t rd_data[16], write_byte;    // 16 bytes to hold read data

    DEBUG_WARN_ON(4);    /* print call trace if breakId == 4 */

    HEADER_ARG("dp_dev=0x%lx bw_code=0x%x lanes=%d", (unsigned long)dp_dev, bw_code, lane_count);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_DISABLE_SCRAMBLING, 0x01);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_ENHANCED_FRAMING_ENABLE, 0x01);
    ljm_dptx_set_link_rate(dp_dev, bw_code);
    ljm_dptx_set_lane_count(dp_dev, lane_count);

    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_LINK_BANDWIDTH_SET, bw_code);
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_LANE_COUNT_SET, 0x80 | lane_count);  // dpcd: enable enhanced_frame_en and set lane count
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_DOWNSPREAD_CONTROL, 0x10); //downspread_ctrl  main link signal is down spread by equal to or less tan 0.5%
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_MAIN_LINK_CODING_SET, 1);    //main_link_channel_coding_set  set ansi 80/10b

    /* make sure sink is powered on */
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_SET_POWER, 1);     // Powering up the link

    /* link training pattern sequence 1 */
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_TRAINING_PATTERN_SET, 0x21); // scrambling disable & TPS1
    ljm_dptx_reg_write(dp_dev, TR_DPTX_TRAINING_PATTERN_SET, 0x01); // Clock recovery - set for training pattern 1
    ljm_dptx_tmr_wait(dp_dev, 100);                                 // wait 100us

    /* link training pattern sequence 3 */
    write_byte = 0x22;
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_TRAINING_PATTERN_SET, 0x22);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_TRAINING_PATTERN_SET, 0x02); // set for training pattern 3
    ljm_dptx_tmr_wait(dp_dev, 400);                                 // wait 400us

    drm_dp_dpcd_read(&dp_dev->aux, TR_DPCD_SINK_COUNT, rd_data, 16);
    DEBUG_PRINT("DPCD[0x%lx]: %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n", TR_DPCD_SINK_COUNT, rd_data[0], rd_data[1],
          rd_data[2], rd_data[3], rd_data[4], rd_data[5], rd_data[6], rd_data[7], rd_data[8], rd_data[9], rd_data[10],
          rd_data[11], rd_data[12], rd_data[13], rd_data[14], rd_data[15]);
    if ((rd_data[2] & 0x2) != 0x2)
    {
        DEBUG_PRINT("Link training failed, dq failed, status = 0x%02x", rd_data[2]);
    }
    if ((rd_data[2] & 0x4) != 0x4)
    {
        DEBUG_PRINT("Link training failed, symbol failed to lock, status = 0x%02x", rd_data[2]);
    }
    if ((rd_data[4] & 0x1) != 0x1)
    {
        DEBUG_PRINT("Link training failed, interlane_align notddone, status = 0x%02x", rd_data[4]);
    }

    /* reset */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_TRAINING_PATTERN_SET, 0x0);
    drm_dp_dpcd_writeb(&dp_dev->aux, TR_DPCD_TRAINING_PATTERN_SET, 0);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_DISABLE_SCRAMBLING, 0x0);

    FOOTER();
    return 0;
}

int ljm_resume_dp(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    uint32_t i;
    struct ljm_port *port;
    struct dp_dev *dp;

    for(i = 0, port = ljm_dev->ports[0];
        i < ljm_dev->num_ports; port = ljm_dev->ports[++i])
    {
        if(port->type != LJM_PORT_DP)
            continue;

        dp = &port->vdev.dp;
#ifdef UEFI_GOP
        ljm_dptx_init(dp, dp->Private, dp->offset, dp->index);
#else
        ljm_dptx_init(dp, dp->regbase, dp->offset, dp->index);
#endif
        if (dp->is_edp)
            ljm_dp_pps_init(dp);
    }

    return 0;
}

static const struct drm_prop_enum_list broadcast_rgb_names[] = {
    { LJM_BROADCAST_RGB_AUTO, "Automatic" },
    { LJM_BROADCAST_RGB_FULL, "Full" },
    { LJM_BROADCAST_RGB_LIMITED, "Limited 16:235" },
};

void ljm_attach_broadcast_rgb_property(struct drm_connector *connector)
{
    struct drm_device *dev = connector->dev;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);
    struct drm_property *prop;

    HEADER_ARG("connector=%p", connector);

    prop = ljm_dev->broadcast_rgb_property;
    if (prop == NULL) {
        prop = drm_property_create_enum(dev, DRM_MODE_PROP_ENUM,
                                        "Broadcast RGB",
                                        broadcast_rgb_names,
                                        ARRAY_SIZE(broadcast_rgb_names));
        if (prop == NULL) {
            FOOTER();
            return;
        }

        ljm_dev->broadcast_rgb_property = prop;
    }

    drm_object_attach_property(&connector->base, prop, 0);

    FOOTER();
}

void ljm_dp_add_properties(struct dp_dev *dp_dev, struct drm_connector *connector)
{
    //struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    //struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);

    HEADER_ARG("dp_dev=%p connector=%p", dp_dev, connector);

    ljm_attach_broadcast_rgb_property(connector);

    FOOTER();
}

void ljm_dptx_soft_reset(struct dp_dev *dp)
{
    ljm_dptx_reg_write(dp, TR_DPTX_SOFT_RESET, 0x1);
    return;
}

static int ljm_bl_get_lightness(struct backlight_device *bd)
{
    struct dp_dev *dp_dev = bd->dev.driver_data;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    uint8_t pwm, low, high;

    if (dp_dev->pwm_ch < 0)
    {
        return -ENODEV;
    }
    pwm = dp_dev->pwm_ch;
    high = CR_RD(ljm_dev->cur_card, PWM_LED1_DUTY_HIGHBYTE + (0x2 * pwm));
    low = CR_RD(ljm_dev->cur_card, PWM_LED1_DUTY_LOWBYTE + (0x2 * pwm));

    return (high << 8 | low) / 10 - 1;/* val = (brightness + 1) * 10 , brightness:0~255 */
}

static int ljm_bl_update_status(struct backlight_device *bd)
{
    struct dp_dev *dp_dev = bd->dev.driver_data;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    uint8_t pwm;
    uint16_t val;
    int brightness = bd->props.brightness;

    if (dp_dev->pwm_ch < 0)
    {
        return -ENODEV;
    }
    pwm = dp_dev->pwm_ch;

    val = (brightness + 1) * 10; /* val = (brightness + 1) * 10 , brightness:0~255 */
    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_HIGHBYTE + (0x2 * pwm), ((val & 0xff00) >> 8));
    CR_WR(ljm_dev->cur_card, PWM_LED1_DUTY_LOWBYTE + (0x2 * pwm), (val & 0x00ff));
    return 0;
}

static const struct backlight_ops edp_bl_ops = {
    .get_brightness = ljm_bl_get_lightness,
    .update_status = ljm_bl_update_status,
};

void ljm_backlight_dev_init(struct dp_dev *dp_dev, struct drm_connector *drm_connector)
{
    char bl_name[16];
    struct backlight_device *bd;
    struct backlight_properties props;

    memset(&props, 0, sizeof(props));
    props.max_brightness = LJM_EDP_MAX_BL_LEVEL;
    props.type = BACKLIGHT_RAW;/* kernel:enum backlight_type - the type of backlight control; */
    snprintf(bl_name, sizeof(bl_name), "ljmicro_bl%d", dp_dev->index);

    bd = backlight_device_register(bl_name, drm_connector->kdev, dp_dev, &edp_bl_ops, &props);

    return;
}

#endif /* #if USE_DP_CONNECTOR */