/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_DP_H__
#define __LJMICRO_DP_H__

#include "ljmicro_dp_hw.h"

#define DPTX_LINK_BW_SET                0x0
enum {
    LINK_BW_1_62 = 0x06,
    LINK_BW_2_16 = 0x08,
    LINK_BW_2_43 = 0x09,
    LINK_BW_2_7  = 0x0a,
    LINK_BW_3_24 = 0x0c,
    LINK_BW_4_32 = 0x10,
    LINK_BW_5_4  = 0x14,
    LINK_BW_8_1  = 0x1e,
};

#ifndef DP_LINK_BW_8_1
#define DP_LINK_BW_8_1                  LINK_BW_8_1
#endif

#define LJM_DP_MAX_LINK_RATE            DP_LINK_BW_8_1

/* io_mux register group */
#define LJM_IO_MUX_REG        0x30a000
#define LJM_GPIO_DATA_O       0x00
#define LJM_GPIO_MASK         0x08
#define LJM_GPIO_PAD_OEN           0x10

/* pwm backlight crtc register*/
#define PWM_LED1_FREQ_SECOND_LOWBYTE  0xC0
#define PWM_LED1_FREQ_SECOND_HIGHBYTE 0xC1
#define PWM_LED1_DUTY_LOWBYTE 0xC2
#define PWM_LED1_DUTY_HIGHBYTE 0xC3
#define PWM_LED2_FREQ_SECOND_LOWBYTE  0xC4
#define PWM_LED2_FREQ_SECOND_HIGHBYTE 0xC5
#define PWM_LED2_DUTY_LOWBYTE 0xC6
#define PWM_LED2_DUTY_HIGHBYTE 0xC7
#define PWM_LED3_FREQ_SECOND_LOWBYTE  0xC8
#define PWM_LED3_FREQ_SECOND_HIGHBYTE 0xC9
#define PWM_LED3_DUTY_LOWBYTE 0xCA
#define PWM_LED3_DUTY_HIGHBYTE 0xCB
#define PWM_LED4_FREQ_SECOND_LOWBYTE  0xCC
#define PWM_LED4_FREQ_SECOND_HIGHBYTE 0xCD
#define PWM_LED4_DUTY_LOWBYTE 0xCE
#define PWM_LED4_DUTY_HIGHBYTE 0xCF
#define PWM_LED1_2_FREQ_FISET 0xD0
#define PWM_LED3_4_FREQ_FISET 0xD1
#define PWM_LED1_2_3_4_ENABLE 0xD2
#define PWM_LED_OUT_SINGLE 0xD3 /*only read*/


/* DPCD */
#define DPTX_LANE_COUNT_SET             0x004
#define DPTX_ENHANCED_FRAME_EN          0x008
#define DPTX_TRAINING_PATTERN_SET       0x00c
#define DPTX_SCRAMBLING_DISABLE         0x014

#define DPTX_TRANSMITTER_OUTPUT_ENABLE  0x080
#define DPTX_VIDEO_STREAM_ENABLE        0x084
#define DPTX_SOURCE_ENABLE              0x094
#define DPTX_FORCE_SCRAMBLER_RESET      0x0c0
#define DPTX_USER_DATA_CONTROL          0x0c8

#define DPTX_AUX_COMMAND                0x100
#define m_BYTE_COUNT            (0xf << 0)
#define v_BYTE_COUNT(n)         (((n) & 0xf) << 0)
/* AUX Channel Command
 * 0x0: I2C over AUX Write
 * 0x1: I2C over AUX Read
 * 0x2: I2C over AUX Write Status
 *
 * 0x4: I2C over AUX Write, Middle of Transaction bit set
 * 0x5: I2C over AUX Read, Middle of Transaction bit set
 *
 * 0x8: AUX Write
 * 0x9: AUX Read
 */
enum {
    AUX_I2C_WRITE = 0,
    AUX_I2C_READ,
    AUX_I2C_WRITE_STATUS,
    AUX_I2C_WRITE_MOT = 4,
    AUX_I2C_READ_MOT,
    AUX_WRITE = 8,
    AUX_READ,
};
#define m_COMMAND               (0xf << 8)
#define v_COMMAND(n)            (((n) & 0xf) << 8)
#define m_ADDRESS_ONLY          BIT(12)
#define v_ADDRESS_ONLY(n)       (((n) & 0x01) << 12)
#define m_AUX_PHY_WAKE          BIT(13)
#define v_AUX_PHY_WAKE(n)       (((n) & 0x01) << 13)

#define DPTX_AUX_WRTIE_FIFO             0x104
#define DPTX_AUX_ADDRESS                0x108
#define DPTX_AUX_CLOCK_DIVIDER          0x10c
#define DPTX_AUX_REPLAY_TIMEOUT_INTERVAL 0x110
#define DPTX_SINK_HPD_STATE             0x128
#define m_SINK_HPD_STATE        BIT(0)
#define DPTX_INTERRUPT_STATE_RO         0x130
#define m_INT_HDCP_TIMER_EVENT  BIT(5)
#define m_INT_GP_TIMER_EVENT    BIT(4)
#define m_INT_REPLY_TIMEOUT     BIT(3)
#define m_INT_REPLY_RECEIVED    BIT(2)
#define m_INT_HPD_IRQ           BIT(1)
#define m_INT_HPD_EVENT         BIT(0)
#define DPTX_AUX_REPLAY_DATA            0x134
#define DPTX_AUX_REPLAY_CODE            0x138
/* AUX channel replay code received from the sink device
 * 0x0: Native AUX ACK or I2C over AUX ACK
 * 0x1: Native AUX NACK
 * 0x2: Native AUX DEFER
 *
 * 0x4: I2C over AUX NACK
 * 0x8: I2C over AUX DEFER
 */
enum {
    AUX_ACK = 0,
    AUX_NACK,
    AUX_DEFER,
    AUX_I2C_NACK = 4,
    AUX_I2C_DEFER = 8,
};
#define m_AUX_NACK              BIT(0)
#define m_AUX_DEFFER            BIT(1)
#define m_AUX_I2C_NACK          BIT(2)
#define m_AUX_I2C_DEFER         BIT(3)

#define DPTX_AUX_REPLAY_COUNT_RW        0x13c   /* write a '1' to this register clears the count */
#define DPTX_INTERRUPT_STATE_RW         0x140   /* a read from this register clears all values */
#define DPTX_INTERRUPT_MASK             0x144
#define DPTX_REPLAY_DATA_COUNT          0x148   /* 4:0 - Total number of data bytes received during the replay phase of the AUX transaction */
#define DPTX_AUX_STATUS                 0x14c
#define m_AUX_REPLY_RECEIVED        BIT(0)
#define m_AUX_REPLY_IN_PROGRESS     BIT(1)
#define m_AUX_REQUEST_IN_PROGRESS   BIT(2)
#define m_AUX_REPLY_ERROR           BIT(3)

/* main stream attributes */
#define DPTX_MAIN_STREAM_HTOTAL         0x180
#define DPTX_MAIN_STREAM_VTOTAL         0x184
#define DPTX_MAIN_STREAM_POLARITY       0x188
#define m_VSYNC_POLARITY_N          BIT(1)
#define v_VSYNC_POLARITY_N(n)       (((n) & 0x01) << 1)
#define m_HSYNC_POLARITY_N          BIT(0)
#define v_HSYNC_POLARITY_N(n)       (((n) & 0x01) << 0)
#define DPTX_MAIN_STREAM_HSWIDTH        0x18c
#define DPTX_MAIN_STREAM_VSWIDTH        0x190
#define DPTX_MAIN_STREAM_HRES           0x194
#define DPTX_MAIN_STREAM_VRES           0x198
#define DPTX_MAIN_STREAM_HSTART         0x19c
#define DPTX_MAIN_STREAM_VSTART         0x1a0
#define DPTX_MAIN_STREAM_MISC0          0x1a4
enum {
    BIT_DEPTH_6 = 0,
    BIT_DEPTH_8,
    BIT_DEPTH_10,
    BIT_DEPTH_12,
    BIT_DEPTH_16,
};
#define m_BIT_DEPTH                 (0x7 << 5)
#define v_BIT_DEPTH(n)              (((n) & 0x7) << 5)
enum {
    YCBCR_COLORIMETRY_ITU_601 = 0,
    YCBCR_COLORIMETRY_ITU_709,
};
#define m_YCBCR_COLORIMETRY         BIT(4)
#define v_YCBCR_COLORIMETRY(n)      (((n) & 0x01) << 4)
enum {
    DYNAMIC_RANGE_VESA = 0,
    DYNAMIC_RANGE_CEA,
};
#define m_DYNAMIC_RANGE             BIT(3)
#define v_DYNAMIC_RANGE(n)          (((n) & 0x01) << 3)
enum {
    COMPONENT_FORMAT_RGB = 0,
    COMPONENT_FORMAT_YCBCR422,
    COMPONENT_FORMAT_YCBCR444,
    COMPONENT_FORMAT_RESERVED,
};
#define m_COMPONENT_FORMAT          (0x3 << 1)
#define v_COMPONENT_FORMAT(n)       (((n) & 0x3) << 1)
enum {
    ASYNCHRONOUS_CLOCK = 0,
    SYNCHRONOUS_CLOCK,
};
#define m_SYNCHRONOUS_CLOCK         BIT(0)
#define v_SYNCHRONOUS_CLOCK(n)      (((n) & 0x01) << 0)

#define DPTX_MAIN_STREAM_MISC1          0x1a8
enum {
    STEREO_VIDEO_ATTRIBUTE_NO = 0,
    STEREO_VIDEO_ATTRIBUTE_TOP_OR_RIGHTEYE,
    STEREO_VIDEO_ATTRIBUTE_RESERVED,
    STEREO_VIDEO_ATTRIBUTE_TOP_OR_LEFTEYE,
};
#define m_STEREO_VIDEO_ATTRIBUTE    (0x3 << 1)
#define v_STEREO_VIDEO_ATTRIBUTE(n) (((n) & 0x3) << 1)
enum {
    INTERLACED_TOTAL_ODD = 0,
    INTERLACED_TOTAL_EVEN,
};
#define m_INTERLACED_TOTAL_EVEN     BIT(0)
#define v_INTERLACED_TOTAL_EVEN(n)  (((n) & 0x01) << 0)

#define DPTX_MVID                       0x1ac
#define DPTX_TRANSFER_UNIT_CONFIG_SRC_0 0x1b0
#define m_FRAC_SYMBOLS_PER_TU       (0xf << 24)
#define v_FRAC_SYMBOLS_PER_TU(n)    (((n) & 0xf) << 24)
#define m_SYMBOLS_PER_TU            (0xff << 16)
#define v_SYMBOLS_PER_TU(n)         (((n) & 0xff) << 16)
#define m_TRANSFER_UNIT_SIZE        (0x7f << 0)
#define v_TRANSFER_UNIT_SIZE(n)     (((n) & 0x7f) << 0)
#define DPTX_NVID                       0x1b4
#define DPTX_USER_PIXEL_COUNT           0x1b8
#define DPTX_USER_DATA_COUNT            0x1bc
#define DPTX_MAIN_STREAM_INTERLACED     0x1c0
#define DPTX_USER_SYNC_POLARITY         0x1c4
#define DPTX_USER_CONTROL               0x1c8

/*
 * APB space for DPTX_PHY
 */
/* PHY internal configuration data bus (CDB) access */
#define PHY_CDB_ADDR                0x0     /* address */
#define PHY_CDB_WDATA               0x4     /* write data */
#define PHY_CDB_SEL_WRITE           0x8     /* {i_cdb_sel, i_cdb_write} */
#define PHY_CDB_EN_READY            0xc     /* {i_cdb_enable_to_phy, i_cdb_ready} */
#define PHY_CDB_RDATA               0x10    /* read data */
#define PHY_CDB_RSTN                0x14    /* reset_n */
#define PHY_CDB_ENABLE              0x18    /* enable */

/* PHY configuration */
#define PHY_CFG_RSTN                0x40    /* i_phy_config_rstn <= i_apb_wdata[0] */
#define PHY_CFG_PMA_CMN_READY       0x44    /* i_apb_rdata[0] <= i_pma_cmn_ready */
#define PHY_CFG_LANE_ENABLE         0x48    /* i_tx_phy_config_lane_enable <= i_apb_wdata[3:0] */
#define PHY_CFG_LANE_RSTN           0x4c    /* i_tx_phy_config_lane_rstn <= i_apb_wdata[3:0] */
#define PHY_CFG_DATA_ENABLE         0x50    /* i_tx_phy_config_data_enable <= i_apb_wdata[0] */
#define PHY_CFG_AUX                 0x54    /* i_config_aux_en, i_aux_bg_en, i_aux_decap_en_delayed, i_aux_decap_en} <= i_apb_wdata[3:0] */
#define PHY_CFG_LANE_ELEC_IDLE      0x58    /* i_phy_lane_elec_idle <= i_apb_wdata[3:0] */

/* PHY XCVR power state request/ack */
#define PHY_XCVR_PWR_STATE          0x60    /* i_xcvr_pwr_state_req <= i_apb_wdata[5:0] */
#define PHY_XCVR_PWR_STATE_ACK      0x64    /* i_apb_rdata[5:0] <= i_xcvr_pwr_state_ack */
#define PHY_XCVR_PLLCLK_EN          0x68    /* i_xcvr_pllclk_en <= i_apb_wdata[0] */
#define PHY_XCVR_PLLCLK_EN_ACK      0x6c    /* i_apb_rdata[0] <= i_xcvr_pllclk_en_ack */

/* PMA powerdown control */
#define PHY_IDDQ_EN_PMA_PWRDOWN     0x70    /* i_iddq_en_pma_powerdown <= i_apb_wdata[0] */

/* link rate control configuration */
#define PHY_LINK_RATE               0x78    /* i_tx_phy_link_rate <= i_apb_wdata[7:0] */

/* AUX/HPD block non-zero static configuration RW access */
#define PHY_AUXHPD                  0x8c    /* i_auxhpd_atbsel_onehot <= i_apb_wdata[7:0]
                                             * i_auxhpd_bg_start_adj, i_auxhpd_term_seg_en} <= i_apb_wdata[14:8]
                                             * i_auxhpd_rx_deglitch_filter, i_auxhpd_rx_hyst_lvl} <= i_apb_wdata[20:16]
                                             * i_auxhpd_tx_slew, i_auxhpd_tx_cur_ctrl} <= i_apb_wdata[30:24]
                                             */

/* emphasis and vmargin for each lane */
#define PHY_PMA_EMPHASIS_0          0xa0    /* i_pma_emphasis[0][17:0] <= i_apb_wdata[17:0] */
#define PHY_PMA_EMPHASIS_1          0xa4    /* i_pma_emphasis[1][17:0] <= i_apb_wdata[17:0] */
#define PHY_PMA_EMPHASIS_2          0xa8    /* i_pma_emphasis[2][17:0] <= i_apb_wdata[17:0] */
#define PHY_PMA_EMPHASIS_3          0xac    /* i_pma_emphasis[3][17:0] <= i_apb_wdata[17:0] */
#define PHY_PMA_VMARGIN_0           0xb0    /* i_pma_vmargin[0] <= i_apb_wdata[2:0] */
#define PHY_PMA_VMARGIN_1           0xb4    /* i_pma_vmargin[1] <= i_apb_wdata[2:0] */
#define PHY_PMA_VMARGIN_2           0xb8    /* i_pma_vmargin[2] <= i_apb_wdata[2:0] */
#define PHY_PMA_VMARGIN_3           0xbc    /* i_pma_vmargin[3] <= i_apb_wdata[2:0] */
#define PHY_PMA_LOW_PWR_SWING       0xc0    /* i_apb_rdata[3:0] <= i_pma_low_power_swing */

/* test rd/wr register & core id */
#define PHY_CONFIG_RDWR_TEST_REG    0xf8    /* i_config_rdwr_test_register[31:0] <= i_apb_wdata[31:0] */
#define PHY_TRILINEAR_PHY_ID_REV    0xfc    /* i_apb_rdata[31:0] <= {kTRILINEAR_PHY_ID, kTRILINEAR_PHY_REV} */

/* PHY misc I/O read/write */
#define PHY_PMA_XCVR_SCANIN_0_32    0x300   /* i_pma_xcvr_scanin[0][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_0_64    0x304   /* i_pma_xcvr_scanin[0][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_1_32    0x308   /* i_pma_xcvr_scanin[1][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_1_64    0x30c   /* i_pma_xcvr_scanin[1][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_2_32    0x310   /* i_pma_xcvr_scanin[2][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_2_64    0x314   /* i_pma_xcvr_scanin[2][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_3_32    0x318   /* i_pma_xcvr_scanin[3][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANIN_3_64    0x31c   /* i_pma_xcvr_scanin[3][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_0_32   0x320   /* i_pma_xcvr_scanin[0][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_0_64   0x324   /* i_pma_xcvr_scanin[0][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_1_32   0x328   /* i_pma_xcvr_scanin[1][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_1_64   0x32c   /* i_pma_xcvr_scanin[1][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_2_32   0x330   /* i_pma_xcvr_scanin[2][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_2_64   0x334   /* i_pma_xcvr_scanin[2][63:32] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_3_32   0x338   /* i_pma_xcvr_scanin[3][31:0] <= i_apb_wdata */
#define PHY_PMA_XCVR_SCANOUT_3_64   0x33c   /* i_pma_xcvr_scanin[3][63:32] <= i_apb_wdata */
#define PHY_DTB_DATA                0x340   /* i_apb_rdata[31:0] <= i_phy_dtb_data */

#define DP_GP201_HPD_TARGET_ADR                    0x38
#define  DP_GP201_HPD_TARGET_INPUTPORT_SUBADR      0x00
#define  DP_GP201_HPD_TARGET_OUPUTPORT_SUBADR      0x01
#define  DP_GP201_HPD_TARGET_POLARITY_SUBADR       0x02      /*默认值为0x00*/
#define  DP_GP201_HPD_TARGET_CONFIGURATION_SUBADR  0x03

enum {
    DP_OK = 0,
    DP_AUX_TIMEOUT,
    DP_AUX_NACK,
    DP_AUX_DEFER,
    DP_AUX_I2C_TIMEOUT,
    DP_AUX_I2C_NACK,
    DP_AUX_I2C_DEFER,
};

enum voltage_swing_level {
    VOLTAGE_LEVEL_0,
    VOLTAGE_LEVEL_1,
    VOLTAGE_LEVEL_2,
    VOLTAGE_LEVEL_3,
};

enum pre_emphasis_level {
    PRE_EMPHASIS_LEVEL_0,
    PRE_EMPHASIS_LEVEL_1,
    PRE_EMPHASIS_LEVEL_2,
    PRE_EMPHASIS_LEVEL_3,
};

#if 0
struct dp_dev {
#ifdef UEFI_GOP
    GMI_PRIVATE_DATA *Private;  /* UEFI private data pointer */
#else
    void __iomem        *regbase;
    struct drm_encoder  *encoder;
#endif

    unsigned int offset;        /* base offset */
    int index;

    struct drm_display_mode mode;


    struct mutex lock;
    bool connected;
    bool suspended;

    struct drm_dp_aux aux;

    u8 max_lanes;

    struct drm_dp_link link;

    u8 train_set[4];
    u8 dpcd[DP_RECEIVER_CAP_SIZE];
};
#endif

#if 0
int ljm_dp_aux_write_req(struct dp_dev *dp_dev, unsigned int addr, unsigned char *buf, int size);
int ljm_dp_aux_read_req(struct dp_dev *dp_dev, unsigned int addr, unsigned char *buf, int size);
#endif

/* backlight */
#define LJM_EDP_MAX_BL_LEVEL 0xff

/* initialize */
struct ljm_port_info;
int ljm_dp_init(struct dp_dev *dp_dev, void *base, int offset, struct ljm_port_info *info);
void ljm_dp_fini(struct dp_dev *dp_dev);

/* mode */
int ljm_dp_mode_set(struct dp_dev *dp_dev, struct drm_display_mode *mode);
void ljm_dp_dpms(struct dp_dev *dp_dev, int mode);
bool ljm_dp_hpd_sense(struct dp_dev *dp_dev);
void ljm_dp_hpd_enable(struct dp_dev *dp_dev, int enable);
struct ljm_connector;
struct edid * ljm_dp_get_edid(struct ljm_connector *ljm_connector, int *psize);
int ljm_dp_get_dpcd(struct dp_dev *dp_dev);
bool ljm_dp_mode_valid(struct dp_dev *dp_dev, const struct drm_display_mode *mode);
bool ljm_dp_mode_valid_ex(struct dp_dev *dp_dev, const struct drm_display_mode *mode);
void ljm_dp_hpd_process(struct dp_dev *dp_dev);
int ljm_dp_hpd_set_polarity( struct dp_dev * dp_dev);
int ljm_edp_self_test_set(struct dp_dev *dp_dev, u8 state);
int ljm_resume_dp(struct drm_device *ddev);
void ljm_dp_add_properties(struct dp_dev *dp_dev, struct drm_connector *connector);
int ljm_dptx_mode_set(struct dp_dev *dp_dev, const struct drm_display_mode *mode);
void ljm_dptx_soft_reset(struct dp_dev *dp);
int ljm_dp_dc_irq_process(struct dp_dev *dp_dev);
int ljm_dp_phy_irq_process(struct dp_dev *dp_dev);

#endif /* __LJMICRO_DP_H__ */
