/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/************************************************************************
* 版权所有 (C)2022, 武汉凌久微电子有限公司。
*
* 文件名称： ljmicro_dp_codec.c
* 文件标识：
* 内容摘要： 本文件定义dp接口音频驱动代码
* 其它说明：
* 当前版本： v1.0
* 作     者： yuqj, 武汉凌久微电子有限公司
* 完成日期： 2022年5月17日
************************************************************************/
#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO) && USE_DP_CONNECTOR

#include <linux/dma-mapping.h>
#include <drm/drm_crtc.h>
#include <linux/err.h>

#include "ljmicro_dp.h"
#include "ljmicro_device.h"
#include "ljmicro_connector_codec.h"

uint32_t ljm_gp200_get_freq(int card, int type);

static inline void dp_write(struct ljm_dp_codec_data *audio, u32 val, int offset)
{
    struct dp_dev *dp = audio->dp;
    writel(val,dp->regbase + dp->offset + 0x10000 + offset);
}

static inline u32 dp_read(struct ljm_dp_codec_data *audio, int offset)
{
    struct dp_dev *dp = audio->dp;
    return readl(dp->regbase + dp->offset + 0x10000 + offset);
}

static void ljm_dp_audio_enable(struct ljm_dp_codec_data *audio)
{
    audio->dp->audio_enable = true;
    if(!dp_read(audio, SECONDARY_STREAM_ENABLE))
        dp_write(audio, m_SECONDARY_STREAM_ENABLE, SECONDARY_STREAM_ENABLE);
    if(!(dp_read(audio, TR_DPTX_SOURCE_ENABLE) & BIT(0)))
        dp_write(audio, dp_read(audio, TR_DPTX_SOURCE_ENABLE) | BIT(0), TR_DPTX_SOURCE_ENABLE);
    dp_write(audio, m_DP_AUDIO_ENABLE, DP_AUDIO_ENABLE);
}

static void ljm_dp_audio_disable(struct ljm_dp_codec_data *audio)
{
    audio->dp->audio_enable = false;
    dp_write(audio, 0, DP_AUDIO_ENABLE);
    dp_write(audio, 0, SECONDARY_STREAM_ENABLE);
}

static int ljm_dp_audio_startup(struct device *dev, void *data)
{
    return 0;
}

static void ljm_dp_audio_shutdown(struct device *dev, void *data)
{
    struct ljm_dp_codec_data *audio = data;
    struct dp_dev *dp = audio->dp;
    unsigned long flags;

    spin_lock_irqsave(&dp->audio_lock, flags);
    ljm_dp_audio_disable(audio);
    spin_unlock_irqrestore(&dp->audio_lock, flags);
}

static void ljm_dp_set_channel_status(struct ljm_dp_codec_data *audio, unsigned char *st)
{
    unsigned int val32[5];
    int i;

    val32[0]  = (st[0] & 0x38) >> 3; /* DP_CS_SOURCE_FORMAT */
    val32[0] |= (st[0] & 0x2) << 1;
    val32[0] |= (st[2] & 0xf) << 4;

    val32[1]  = st[1]; /* DP_CS_CATEGORY_CODE */
    /* fd = b1111_1101 "101 1111L" other
       b5 = b1011_0101 "101 01101" reserved
       85 = b1000_0101 "101 0000L" synthesizer
    */

    val32[2]  = (st[4] & 0xf0) >> 4; /*DP_CS_LENGTH_ORIG_FREQ*/
    val32[2] |= (st[4] & 0xf)  << 4;

    val32[3]  = (st[3] & 0x30) >> 4; /*DP_CS_FREQ_CLOCK_ACCURACY*/
    val32[3] |= (st[3] & 0xf)  << 4;

    val32[4]  = (st[0] & 0x4) >> 2; /*DP_CS_COPYRIGHT*/
    val32[4] |= (st[5] & 0x3) << 1;

    for(i=0; i < 5; i++) {
        dp_write(audio, val32[i], DP_CS_SOURCE_FORMAT + 4 * i);
    }
}

static unsigned int dp_compute_mn(unsigned int freq, unsigned char bw_code, unsigned int *maud, unsigned int *naud)
{
    unsigned int n = 0;
    unsigned int m = 0;
    unsigned int mult = 1;

    while (freq > 48000) {
        mult *= 2;
        freq /= 2;
    }

    switch (freq) {
        case 32000:
            if (bw_code == 0x06) /*1620MHz*/
                n = 10125;
            else if (bw_code == 0x0a) /*2700MHz*/
                n = 16875;
            else if (bw_code == 0x14) /*5400MHz*/
                n = 33750;
            else if (bw_code == 0x1e) /*8100MHz*/
                n = 50625;
            m = 1024 * mult;
            break;

        case 44100:
            if (bw_code == 0x06) /*1620MHz*/
                n = 5625;
            else if (bw_code == 0x0a) /*2700MHz*/
                n = 9375;
            else if (bw_code == 0x14) /*5400MHz*/
                n = 18750;
            else if (bw_code == 0x1e) /*8100MHz*/
                n = 28125;
            m = 784 * mult;
            break;

        case 48000:
            if (bw_code == 0x06) /*1620MHz*/
                n = 3375;
            else if (bw_code == 0x0a) /*2700MHz*/
                n = 5625;
            else if (bw_code == 0x14) /*5400MHz*/
               n = 11250;
            else if (bw_code == 0x1e) /*8100MHz*/
               n = 16875;
            m = 512 * mult;
            break;

        default:
            break;
    }

    *maud = m;
    *naud = n;

    return 0;
}

static int ljm_dp_audio_config_sample_rate(struct ljm_dp_codec_data *audio, int sample_rate)
{
    struct dp_dev *dp = audio->dp;
    int maud = 0;
    int naud = 0;
    u32 link_rate;

    link_rate = dp_read(audio, TR_DPTX_LINK_BW_SET);
    mutex_lock(&dp->audio_mutex);
    dp_compute_mn(sample_rate, link_rate, &maud, &naud);/*asdf: get_pixclk; */

    spin_lock_irq(&dp->audio_lock);
    dp_write(audio, maud, DP_MAUD);
    dp_write(audio, naud, DP_NAUD);
    dp_write(audio, m_DP_AUDIO_CLOCK_MODE, DP_AUDIO_CLOCK_MODE);/*asdf: sync mode*/
    spin_unlock_irq(&dp->audio_lock);

    mutex_unlock(&dp->audio_mutex);

    return 0;
}

static int  ljm_dp_audio_config_channels(struct ljm_dp_codec_data *audio, int channels)
{
    struct dp_dev *dp = audio->dp;
    int val = 0;

    mutex_lock(&dp->audio_mutex);
    switch (channels) {
        case 7 ... 8:
            val = v_DP_CHANNEL_COUNT_7_1;
            break;
        case 5 ... 6:
            val = v_DP_CHANNEL_COUNT_5_1;
            break;
        case 3 ... 4:
            val = v_DP_CHANNEL_COUNT_2_1;
            break;
        case 2:
            val = v_DP_CHANNEL_COUNT_2;
            break;
        case 0:
            val = v_DP_CHANNEL_COUNT_MUTE;
            break;
    }
    dp_write(audio, val, DP_CHANNEL_COUNT);
    dp_write(audio, 0x76543210, DP_AUDIO_CHANNEL_MAP);
    mutex_unlock(&dp->audio_mutex);

    return 0;
}

static int  ljm_dp_audio_config(struct ljm_dp_codec_data *audio, struct params_info *info)
{
    struct dp_dev *dp = audio->dp;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp->ddev);
    int ret = 0;
    u32 val = 0, apb_freq;

    mutex_lock(&dp->audio_mutex);
    switch (info->format) {
        case DP_I2S:
            val = v_DP_INPUT_I2S;
            break;
        case DP_DS_FIFO:
            val = v_DP_INPUT_DS_FIFO;
            break;
        case DP_SPDIF:
            val = v_DP_INPUT_SPDIF;
            info->channels = 2;
            break;
        case DP_DMA:
            val = v_DP_INPUT_DMA;
            break;
        default:
            ret = -EINVAL;
            return ret;
    }
    dp_write(audio, val, DP_INPUT_SELECT); /*select input type*/
    mutex_unlock(&dp->audio_mutex);

    apb_freq = ljm_gp200_get_freq(ljm_dev->cur_card, freq_apb);
    if (info->format == DP_DS_FIFO) {
        /* APB Frequency(hz) / Audio Sample Rate(hz) */
        val = (apb_freq * 1000000)/ info->sample_rate; /*TODO: need to cal*/
        dp_write(audio, val, DP_DIRECT_CLKDIV);
    }

    ljm_dp_audio_config_channels(audio, info->channels);

    val = 0xf8;
    dp_write(audio, val, TR_DPTX_SRC0_SECONDARY_DATA_WINDOW);

    ljm_dp_set_channel_status(audio, info->status);
    ljm_dp_audio_config_sample_rate(audio, info->sample_rate); /*maud, naud*/

    val = 0x5;
    dp_write(audio, val, DP_TIMESTAMP_INTERVAL);
    dp_write(audio, 0x2, TR_DPTX_SRC0_USER_CONTROL); /*sdp no vertical sync*/
    dp_write(audio, 0, DP_AUDIO_FIFO_DEPTH); /*disable dma fifo*/

    return 0;
}

static int ljm_dp_hw_params(struct device *dev, void *data,
        struct hdmi_codec_daifmt *fmt, struct hdmi_codec_params *hparms)
{
    struct ljm_dp_codec_data *audio = data;
    int ret;
    struct params_info info = {
        .sample_rate = hparms->sample_rate,
        .channels = hparms->channels,
        .sample_width = hparms->sample_width,
    };

    memcpy(info.status,hparms->iec.status,sizeof(info.status));

    /* I2S only */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    if (fmt->bit_clk_provider | fmt->frame_clk_provider) {
#else
    if (fmt->bit_clk_master | fmt->frame_clk_master) {
#endif
        dev_err(dev, "LJM DP CODEC: unsupported clock settings\n");
        ret = -EINVAL;
        goto out;
    }

    switch (fmt->fmt) {
        case HDMI_I2S:
            info.format = DP_I2S;
            break;
        case HDMI_SPDIF:
            info.format = DP_SPDIF;
            info.channels = 2;
            /* uncompressed*/
            break;
        default:
            dev_err(dev, "LJM DP CODEC: invalid format %d\n", fmt->fmt);
            ret = -EINVAL;
            goto out;
    }

    ret = ljm_dp_audio_config(audio, &info);
    if (!ret)
        audio->info = info;
    ljm_dp_audio_enable(audio);

out:
    return ret;
}

static int ljm_dp_get_eld(struct device *dev, void *data, uint8_t *buf, size_t len)
{
    return 0;
}

static int ljm_dp_get_dai_id(struct snd_soc_component *component, struct device_node *endpoint)
{
    struct of_endpoint of_ep;
    int ret;

    ret = of_graph_parse_endpoint(endpoint, &of_ep);
    if (ret < 0)
        return ret;

    if (of_ep.port == 2)
        return 0;

    return -EINVAL;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
static int ljm_dp_hook_plugged_cb(struct device *dev, void *data, hdmi_codec_plugged_cb fn, struct device *codec_dev)
{
    struct ljm_dp_codec_data *audio = data;
    struct dp_dev *dp = audio->dp;
    bool plugged;

    mutex_lock(&dp->mutex);
    dp->plugged_cb = fn;
    dp->codec_dev = codec_dev;
    plugged = dp->last_connector_result == connector_status_connected;

    if (fn && codec_dev)
        fn(codec_dev, plugged);
    mutex_unlock(&dp->mutex);

    return 0;
}
#endif

static struct hdmi_codec_ops ljm_dp_codec_ops = {
    .hw_params = ljm_dp_hw_params,
    .audio_startup = ljm_dp_audio_startup,
    .audio_shutdown = ljm_dp_audio_shutdown,
    .get_eld = ljm_dp_get_eld,
    .get_dai_id = ljm_dp_get_dai_id,
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
    .hook_plugged_cb = ljm_dp_hook_plugged_cb,
#endif
};

static const struct regmap_config dp_regmap_32bit_config = {
    .reg_bits = 32,
    .val_bits = 32,
    .reg_stride = 1,
    .max_register = TR_DPTX_SRC3_USER_CONTROL,
};

int ljm_dp_codec_init(struct ljm_connector_codec_data *audio_data, struct ljm_device *ljm_dev)
{
    int i;
    int j = 0;

    audio_data->ljm_dp_codec_ops = &ljm_dp_codec_ops;

    for (i = 0; i < LJM_MAX_PORTS; i++) {
        if (!ljm_dev->ports[i]) {
            continue;
        }
        /* 初始化与DP音频相关部分 */
        if (ljm_dev->ports[i]->type == LJM_PORT_DP) {
            ljm_dev->ports[i]->vdev.dp.dev = ljm_dev->dev;
            audio_data->dp_devs[j] = &ljm_dev->ports[i]->vdev.dp;
            j++;
        }
        else {
            continue;
        }
    }
    audio_data->num_dps = j;
    audio_data->dp_audio_data->write = dp_write;
    audio_data->dp_audio_data->read = dp_read;

    return 0;
}

#endif
