/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，武汉凌久微电子有限公司
 *
 ******************************************************************************
 *
 *  文件名称：ljmicro_dp_hw.c
 *  版 本 号：V1.0
 *  作    者：胡艳明
 *  生成日期：2022年7月6日
 *
 *  功能描述：DP显示接口硬件驱动
 *
 *  修改历史：
 *         1. 日    期：2022年7月6日
 *            作    者：胡艳明
 *            修改内容：创建文件
 *
 *****************************************************************************/

#if USE_DP_CONNECTOR

#include "ljmicro_base.h"
#include "ljmicro_proc.h"
#include "ljmicro_dp_hw.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS


/******************************************************************************
 *                                    函数实现                                    *
 ******************************************************************************/

#if DEBUG_DUMP_REG_DP
inline uint32_t dp_reg_r32(struct dp_dev *dp_dev, uint32_t addr)
{
    uint32_t r;
    r = RREG32_BASE(dp_dev->regbase, dp_dev->offset + addr);
    PRINT("DP %d: R[%x]=%08x", , dp_dev->offset + addr, r);
    return r;
}
#endif

/*
 * DP PHY interface
 */

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_reg_read
 *  功能描述: 读DP接口PHY寄存器
 *  输入参数: struct dp_dev *dp_dev  DP结构体指针
 *            uint32_t reg_offset    寄存器偏移地址
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_phy_reg_read(struct dp_dev *dp_dev, uint32_t reg_offset)
{
    CHECK_DP_DEV_VALID(dp_dev);

    return DP_REG_R32(dp_dev, reg_offset);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_reg_write
 *  功能描述: 写DP接口PHY寄存器
 *  输入参数: struct dp_dev *dp_dev  DP结构体指针
 *            uint32_t reg_offset    寄存器偏移地址
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_phy_reg_write(struct dp_dev *dp_dev, uint32_t reg_offset, uint32_t data)
{
    CHECK_DP_DEV_VALID_VOID(dp_dev);

    DP_REG_W32(dp_dev, reg_offset, data);
}

/*
 * DP Controller interface
 */

/*****************************************************************************
 *  函 数 名: ljm_dptx_reg_read
 *  功能描述: 读DP接口控制器寄存器
 *  输入参数: struct dp_dev *dp_dev  DP结构体指针
 *            uint32_t reg_offset    寄存器偏移地址
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_reg_read(struct dp_dev *dp_dev, uint32_t reg_offset)
{
    CHECK_DP_DEV_VALID(dp_dev);

    return DP_REG_R32(dp_dev, reg_offset + 0x10000);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_reg_write
 *  功能描述: 写DP接口控制器寄存器
 *  输入参数: struct dp_dev *dp_dev  DP结构体指针
 *            uint32_t reg_offset    寄存器偏移地址
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_reg_write(struct dp_dev *dp_dev, uint32_t reg_offset, uint32_t data)
{
    CHECK_DP_DEV_VALID_VOID(dp_dev);

    DP_REG_W32(dp_dev, reg_offset + 0x10000, data);
}

/*
 * DP timer interface
 */

/*****************************************************************************
 *  函 数 名: dptx_tmr_init
 *  功能描述: DP接口Timer初始化
 *  输入参数: struct dp_dev *dp_dev  DP结构体指针
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_tmr_init(struct dp_dev *dp_dev)
{
    ljm_dptx_reg_write(dp_dev, TR_DPTX_HOST_TIMER, 0ul);
}

/*****************************************************************************
 *  函 数 名: dptx_tmr_set_value
 *  功能描述: 设置timer值
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t new_value
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_tmr_set_value(struct dp_dev *dp_dev, uint32_t new_value)
{
    uint32_t timer_reg = ljm_dptx_reg_read(dp_dev, TR_DPTX_HOST_TIMER);

    timer_reg = ((timer_reg & TR_DPTX_TIMER_FLAGS_ALL) | (TR_DPTX_TIMER_MAX_COUNT & new_value));
    ljm_dptx_reg_write(dp_dev, TR_DPTX_HOST_TIMER, timer_reg);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_tmr_enable
 *  功能描述: 使能DP接口timer
 *  输入参数: struct dp_dev *dp_dev
 *            bool enable
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_tmr_enable(struct dp_dev *dp_dev, bool enable)
{
    uint32_t timer_reg = ljm_dptx_reg_read(dp_dev, TR_DPTX_HOST_TIMER);

    if (enable == true)
    {
        timer_reg = timer_reg | TR_DPTX_TIMER_FLAG_ENABLE;
    }
    else
    {
        timer_reg = timer_reg & ~TR_DPTX_TIMER_FLAG_ENABLE;
    }
    ljm_dptx_reg_write(dp_dev, TR_DPTX_HOST_TIMER, timer_reg);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_tmr_get_value
 *  功能描述: DP接口获取timer计数值
 *  输入参数: struct dp_dev *dp_dev
 *            bool enable
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_tmr_get_value(struct dp_dev *dp_dev)
{
    return (ljm_dptx_reg_read(dp_dev, TR_DPTX_HOST_TIMER) & TR_DPTX_TIMER_MAX_COUNT);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_tmr_wait
 *  功能描述: DP接口通过timer进行延时控制
 *  输入参数: struct dp_dev *dp_dev
 *            bool enable
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_tmr_wait(struct dp_dev *dp_dev, uint32_t us)
{
    ljm_dptx_tmr_init(dp_dev);
    ljm_dptx_tmr_set_value(dp_dev, us);
    ljm_dptx_tmr_enable(dp_dev, true);
    while (ljm_dptx_tmr_get_value(dp_dev) != 0)
        ;
}

/*
 * DP cdb bus interface
 */

/*****************************************************************************
 *  函 数 名: ljm_cdb_ready
 *  功能描述: 等待cdb总线空闲
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t delay_time
 *            uint32_t loop_ct
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_cdb_ready(struct dp_dev *dp_dev, uint32_t delay_time, uint32_t loop_ct)
{
    uint32_t         reg_val, idx;
    dptx_phy_error_t status = dptx_phy_error_bus_not_idle;

    for (idx = 0; idx < loop_ct; idx++)
    {
        reg_val = ljm_dptx_phy_reg_read(dp_dev, TR_DPTX_CDB_READY);
        reg_val &= 0x01ul;
        if (reg_val == 0x01ul)
        {
            status = dptx_phy_error_none;
            break;
        }
        else
        {
            ljm_dptx_tmr_wait(dp_dev, delay_time);
        }
    }
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_cdb_write
 *  功能描述: DP接口CDB总线写寄存器
 *  输入参数: struct dp_dev *dp_dev
 *            uint16_t addr
 *            uint16_t data
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_cdb_write(struct dp_dev *dp_dev, uint16_t addr, uint16_t val)
{
    uint32_t         data   = (uint32_t)val;
    dptx_phy_error_t status = ljm_cdb_ready(dp_dev, 5, 100); //	wait 5us, 100 loops (up to 500us)

    if (status == dptx_phy_error_none)
    {   //	wait 5us, 100 loops
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_ADDR, addr);
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_WDATA, data);
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_SEL_WR, 3ul); //	select & write flags active
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_ENABLE, 1ul); //	execute configured CDB transaction
    }
    else
    {
        ERR_PRINT("dp %d: cdb bus is busy when write [0x%x] with 0x%x\n", dp_dev->index, addr, data);
    }

    return (int)status;
}

/*****************************************************************************
 *  函 数 名: ljm_cdb_read
 *  功能描述: DP接口CDB总线读数据
 *  输入参数: struct dp_dev *dp_dev
 *            uint16_t addr
 *            uint16_t pval          pointer of read data from cdb bus
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_cdb_read(struct dp_dev *dp_dev, uint16_t addr, uint16_t *pval)
{
    uint32_t         data   = 0ul;
    dptx_phy_error_t status = ljm_cdb_ready(dp_dev, 5, 100); //	wait 5us, 100 loops (up to 500us)

    if (status == dptx_phy_error_none)
    {   //	wait 5us, 100 loops
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_ADDR, addr);
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_SEL_WR, 0x02ul); // select flag active
        ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_ENABLE, 0x01ul); // execute configured CDB transaction
        status = ljm_cdb_ready(dp_dev, 5, 100);                 // wait for the transaction to complete
        if (status == dptx_phy_error_none)
        {   //	wait 5us, 100 loops
            data = ljm_dptx_phy_reg_read(dp_dev, TR_DPTX_CDB_READY);    // chech the status
            if (data == 1)
            {
                data = ljm_dptx_phy_reg_read(dp_dev, TR_DPTX_CDB_RDATA);// get the data
                *pval = data & 0xffff;
                return 0;
            }
            else
            {
                *pval = 0;
                return data;
            }
        }
    }
    else
    {
        ERR_PRINT("dp %d: cdb bus is busy when read [0x%x]\n", dp_dev->index, addr);
    }

    return (int)status;
}

/*
 * DP AUX interface
 */

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_ready
 *  功能描述: cdb接口是否空闲
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dptx_aux_ready(struct dp_dev *dp_dev)
{
    return (ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_STATUS) & TR_DPTX_AUX_STATUS_REQUEST_IN_PROGRESS) ? false : true;
}

/*****************************************************************************
 *  函 数 名: dptx_aux_hpd_status
 *  功能描述: DP接口热插拔状态
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dptx_aux_hpd_status(struct dp_dev *dp_dev)
{
    return (ljm_dptx_reg_read(dp_dev, TR_DPTX_HPD_INPUT_STATE)) ? true : false;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_send_request
 *  功能描述: 发送AUX总线操作请求
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t cmd_code
 *            uint32_t address
 *            uint32_t byte_ct
 *            bool addr_only
 *            bool phy_wake
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_aux_send_request(struct dp_dev *dp_dev, uint32_t cmd_code, uint32_t address, uint32_t byte_ct,
                               bool addr_only, bool phy_wake)
{
    uint32_t cmd = 0ul;

    HEADER_ARG("dp_dev=%p cmd_code=%x address=%x byte_ct=%d addr_only=%d phy_wake=%d",
        dp_dev, cmd_code, address, byte_ct, addr_only, phy_wake);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_ADDRESS, address);   // Set address

    cmd = (cmd_code | ((byte_ct == 0 ? 0U : (byte_ct - 1U)) & 0xFul));

    if (addr_only == true)
    {   // Set the address only bit in the HW if param is true
        cmd |= TR_DPTX_AUX_CMD_ADDRESS_ONLY;
    }
    if (phy_wake == true)
    {   // Set PHY_WAKE bit in the HW if param is true
        cmd |= TR_DPTX_AUX_CMD_PHY_WAKE;
    }
    ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_COMMAND, cmd); // Submit the whole command to the HW

    DEBUG_PRINT("dp %d: aux send request 0x%x @ 0x%x\n", dp_dev->index, cmd, address);

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_reply_timeout
 *  功能描述: AUX应答是否超时
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dptx_aux_reply_timeout(struct dp_dev *dp_dev)
{
    return (ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_STATE) & TR_DPTX_INTERRUPT_STATE_REPLY_TIMEOUT) ? true : false;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_reply_error
 *  功能描述: AUX接口应答错误
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dptx_aux_reply_error(struct dp_dev *dp_dev)
{
    return (ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_STATUS) & TR_DPTX_AUX_STATUS_REPLY_ERROR) ? true : false;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_reply_received
 *  功能描述: AUX接口接收到数据
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_dptx_aux_reply_received(struct dp_dev *dp_dev)
{
    return (ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_STATE) & TR_DPTX_INTERRUPT_STATE_REPLY_RECEIVED);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_wait_reply
 *  功能描述: AUX接口等待应答
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_aux_wait_reply(struct dp_dev *dp_dev)
{
    uint32_t failsafe = 100;   /* FIXME: 100? */
    uint32_t idx;

    for (idx = 0; idx < failsafe; idx++)
    {
        if (ljm_dptx_aux_reply_timeout(dp_dev) == true)
        {
            return -ETIMEDOUT;
        }
        if (ljm_dptx_aux_reply_error(dp_dev) == true)
        {
            return -EREMOTEIO;
        }
        if (ljm_dptx_aux_reply_received(dp_dev) == true)
        {
            return 0;
        }
        usleep_range(10, 11);
    }
#if 0
    if (ljm_dptx_aux_reply_timeout(dp_dev) == true)
    {
        return TR_DPTX_AUX_STATUS_TIMEOUT;
    }
#endif
    /* wait 1000us without replay */
    DEBUG_PRINT("DP%d: aux bus failed", dp_dev->index);
    return -ENXIO;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_reply_code
 *  功能描述: 读取AUX总线返回码
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_aux_reply_code(struct dp_dev *dp_dev)
{
    return ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_REPLY_CODE);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_read_fifo
 *  功能描述: 从RFIFO中读取AUX总线返回数据
 *  输入参数: struct dp_dev *dp_dev
 *            uint8_t *read_buffer
 *            uint32_t byte_ct
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_aux_read_fifo(struct dp_dev *dp_dev, uint8_t *read_buffer, uint32_t byte_ct)
{
    uint32_t i;
    uint32_t reply_data_count;
#ifdef _DEBUG
    uint8_t tmp[16] = {'d', 'e', 'a', 'd', 'b', 'e', 'e', 'f', 'd', 'e', 'a', 'd', 'b', 'e', 'e', 'f'};
#else
    uint8_t tmp[16];
#endif

    /* read all data from internal RFIFO */
    reply_data_count = ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_REPLY_DATA_COUNT) & 0x1f;
    if (reply_data_count > 16) {
        ERR_PRINT("dp %d: aux reply data count %d is over range\n", dp_dev->index, reply_data_count);
        reply_data_count = 16;
    }
    for (i = 0; i < reply_data_count; i++) {
        tmp[i] = ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_REPLY_DATA);
    }

    /* dump aux read fifo */
    if (reply_data_count > 0) {
        DEBUG_PRINT("dp %d: read fifo: %*ph\n",
            dp_dev->index, reply_data_count, tmp);
    }

    if (reply_data_count < byte_ct) {
        DEBUG_PRINT("dp %d: aux reply %d bytes, less than requested %d\n",
            dp_dev->index, reply_data_count, byte_ct);
        byte_ct = reply_data_count;
    }
    for (i = 0; i < byte_ct; i++) {
        read_buffer[i] = tmp[i];
    }
    return (int)byte_ct;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_write_fifo
 *  功能描述: 写AUX总线WFIFO
 *  输入参数: struct dp_dev *dp_dev
 *            uint8_t *read_buffer
 *            uint32_t byte_ct
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_aux_write_fifo(struct dp_dev *dp_dev, uint8_t *write_buffer, uint32_t byte_ct)
{
    uint32_t i;
    for (i = 0; i < byte_ct; i++)
    {
        ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_WRITE_FIFO, write_buffer[i]);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_dpcd_read_blk
 *  功能描述: AUX总线写一块数据
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t addr
 *            uint8_t *rd_data
 *            uint32_t ct
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_aux_dpcd_read_blk(struct dp_dev *dp_dev, uint32_t addr, uint8_t *rd_data, uint32_t ct)
{
    uint32_t reply_status;
    uint32_t reply_code;
    uint32_t txn_status = 0;
    int timeout= 100;

    memset(rd_data, 0, ct);
    if (ljm_dptx_aux_ready(dp_dev) == true)
    {
        do {
            txn_status = 0x10000;
            ljm_dptx_aux_send_request(dp_dev, TR_DPTX_AUX_CMD_READ, addr, ct, false, false);
            //udelay(1);
            reply_status = ljm_dptx_aux_wait_reply(dp_dev);
            ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
            reply_code = ljm_dptx_aux_reply_code(dp_dev);
            txn_status |= (reply_status << 8) | reply_code;
            if ((reply_status == TR_DPTX_AUX_STATUS_RECEIVED) && (reply_code == TR_DPTX_AUX_REPLY_ACK))
            {
                ljm_dptx_aux_read_fifo(dp_dev, rd_data, ct);
            }

            if (reply_code == TR_DPTX_AUX_REPLY_DEFER)
            {
                ljm_dptx_tmr_wait(dp_dev, 10);
            }
        } while (reply_code == TR_DPTX_AUX_REPLY_DEFER && --timeout > 0);
    }

    return txn_status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_i2c_raw_write
 *  功能描述: DP接口通过AUX通道直接写一个字节数据到I2C设备
 *  输入参数: struct dp_dev *dp_dev
 *            uint8_t target
 *            uint8_t data
 *  输出参数: 无
 *  返 回 值: static uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月11日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static uint32_t ljm_dptx_aux_i2c_raw_write(struct dp_dev *dp_dev, uint8_t target, uint8_t data)
{
    uint32_t reply_status;
    uint32_t reply_code;
    uint32_t txn_status = 0;
    int timeout= 100;

    if (ljm_dptx_aux_ready(dp_dev) == true)
    {
        do {
            txn_status = 0x10000;
            /* write data to wfifo */
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_WRITE_FIFO, data);
            /* request i2c write */
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_WRITE_FIFO, data);
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_ADDRESS, target);                    // Set address
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_COMMAND, TR_DPTX_AUX_CMD_I2C_WRITE); // Submit the whole command to the HW
            //udelay(1);
            reply_status = ljm_dptx_aux_wait_reply(dp_dev);
            ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
            reply_code = ljm_dptx_aux_reply_code(dp_dev);
            txn_status |= (reply_status << 8) | reply_code;

            if (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER)
            {
                ljm_dptx_tmr_wait(dp_dev, 10);
            }
        } while (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER && --timeout > 0);
    }

    return txn_status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_i2c_read_blk
 *  功能描述: DP接口通过AUX从I2C设备指定地址读一块数据
 *  输入参数: struct dp_dev *dp_dev
 *            uint8_t target
 *            uint8_t addr
 *            uint8_t *rd_data
 *            uint32_t ct
 *  输出参数: 无
 *  返 回 值: static uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月11日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static uint32_t ljm_dptx_aux_i2c_read_blk(struct dp_dev *dp_dev, uint8_t target, uint8_t addr, uint8_t *rd_data, uint32_t ct)
{
    uint32_t reply_status;
    uint32_t reply_code;
    uint32_t txn_status = 0;
    int timeout= 100;
    int left_size, cur_size;

    if (ljm_dptx_aux_ready(dp_dev) == true)
    {
        do {
            txn_status = 0x10000;
            /* request i2c write */
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_WRITE_FIFO, addr);
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_ADDRESS, target);    // Set address
            ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_COMMAND, TR_DPTX_AUX_CMD_I2C_WRITE_MOT); // Submit the whole command to the HW
            //udelay(1);
            reply_status = ljm_dptx_aux_wait_reply(dp_dev);
            ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
            reply_code = ljm_dptx_aux_reply_code(dp_dev);
            txn_status |= (reply_status << 8) | reply_code;

            if (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER)
            {
                ljm_dptx_tmr_wait(dp_dev, 10);
            }
        } while (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER && --timeout > 0);
    }

    if (((txn_status & 0xffff) || ((txn_status & 0x10000) == 0)))
    {
        TRACE("dp %d: aux i2c write 0x%02x with 0x%02x error", dp_dev->index, target, addr);
        return txn_status;
    }

    ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_ADDRESS, target);    // Set address

    memset(rd_data, 0, ct);
    left_size = ct;
    while (left_size > 0) {
        if (left_size > 16) cur_size = 16;
        else cur_size = left_size;
        left_size -= cur_size;

        if (left_size == 0) /* last block */
        {
            if (cur_size > 1) {
                /* i2c read, middle of transaction bit set */
                do {
                    ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_COMMAND, TR_DPTX_AUX_CMD_I2C_READ_MOT | (cur_size - 2));
                    //udelay(1);
                    reply_status = ljm_dptx_aux_wait_reply(dp_dev);
                    ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
                    reply_code = ljm_dptx_aux_reply_code(dp_dev);
                    txn_status |= (reply_status << 8) | reply_code;

                    if ((reply_status == TR_DPTX_AUX_STATUS_RECEIVED) && (reply_code == TR_DPTX_AUX_REPLY_I2C_ACK))
                    {
                        ljm_dptx_aux_read_fifo(dp_dev, rd_data, cur_size - 1);
                        rd_data += cur_size - 1;
                    }

                    if (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER)
                    {
                        ljm_dptx_tmr_wait(dp_dev, 10);
                    }
                } while (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER && --timeout > 0);

                if (txn_status & 0xffff)
                {
                    TRACE("dp %d: aux i2c read reply error 0x%x", dp_dev->index, txn_status);
                    return txn_status;
                }
            }

            /* i2c read the last byte */
            do {
                ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_COMMAND, TR_DPTX_AUX_CMD_I2C_READ);
                //udelay(1);
                reply_status = ljm_dptx_aux_wait_reply(dp_dev);
                ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
                reply_code = ljm_dptx_aux_reply_code(dp_dev);
                txn_status |= (reply_status << 8) | reply_code;

                if ((reply_status == TR_DPTX_AUX_STATUS_RECEIVED) && (reply_code == TR_DPTX_AUX_REPLY_I2C_ACK))
                {
                    ljm_dptx_aux_read_fifo(dp_dev, rd_data, 1);
                    rd_data += 1;
                }

                if (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER)
                {
                    ljm_dptx_tmr_wait(dp_dev, 10);
                }
            } while (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER && --timeout > 0);
        }
        else
        {
            /* i2c read, middle of transaction bit set */
            do {
                ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_COMMAND, TR_DPTX_AUX_CMD_I2C_READ_MOT | (cur_size - 1));
                //udelay(1);
                reply_status = ljm_dptx_aux_wait_reply(dp_dev);
                ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
                reply_code = ljm_dptx_aux_reply_code(dp_dev);
                txn_status |= (reply_status << 8) | reply_code;

                if ((reply_status == TR_DPTX_AUX_STATUS_RECEIVED) && (reply_code == TR_DPTX_AUX_REPLY_I2C_ACK))
                {
                    ljm_dptx_aux_read_fifo(dp_dev, rd_data, cur_size);
                    rd_data += cur_size;
                }

                if (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER)
                {
                    ljm_dptx_tmr_wait(dp_dev, 10);
                }
            } while (reply_code == TR_DPTX_AUX_REPLY_I2C_DEFER && --timeout > 0);

            if (txn_status & 0xffff)
            {
                ERR_PRINT("dp %d: aux i2c read reply error 0x%x\n", dp_dev->index, txn_status);
                return txn_status;
            }
        }
    }

    return txn_status;
}

/*****************************************************************************
 *  函 数 名: ljm_dp_i2c_read_req
 *  功能描述: DP接口i2c读一块数据
 *  输入参数: void *hw_dev
 *            unsigned char target
 *            unsigned char addr
 *            unsigned char *buf
 *            int size
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月11日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_i2c_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct dp_dev *dp_dev = (struct dp_dev *)hw_dev;
    uint32_t txn_status = 0;

    HEADER_ARG("hw_dev=%p target=%x addr=%x buf=%p size=%x", hw_dev, target, addr, buf, size);

    txn_status = ljm_dptx_aux_i2c_read_blk(dp_dev, target, addr, buf, size);
    if (((txn_status & 0xffff) || ((txn_status & 0x10000) == 0)))
    {
        ERR_PRINT("dp %d: aux i2c read buffer error\n", dp_dev->index);

        FOOTER();
        return -1;
    }

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dp_i2c_raw_write_req
 *  功能描述: DP接口I2C直接写一个字节数据
 *  输入参数: void *hw_dev
 *            unsigned char target
 *            unsigned char data
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月11日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_i2c_raw_write_req(void *hw_dev, unsigned char target, unsigned char data)
{
    struct dp_dev *dp_dev = (struct dp_dev *)hw_dev;
    uint32_t txn_status = 0;

    HEADER_ARG("hw_dev=%p target=%d data=%d", hw_dev, target, data);

    txn_status = ljm_dptx_aux_i2c_raw_write(dp_dev, target, data);
    if (((txn_status & 0xffff) || ((txn_status & 0x10000) == 0)))
    {
        WARN_PRINT("dp %d: aux i2c write data error\n", dp_dev->index);

        FOOTER();
        return -1;
    }

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_dpcd_write_blk
 *  功能描述: AUX总线写一组数据
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t addr
 *            uint8_t *rd_data
 *            uint32_t ct
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_aux_dpcd_write_blk(struct dp_dev *dp_dev, uint32_t addr, uint8_t *wr_data, uint32_t ct)
{
    uint32_t reply_status;
    uint32_t reply_code;
    uint32_t txn_status = 0;
    int timeout= 100;

    if (ljm_dptx_aux_ready(dp_dev) == true)
    {
        do {
            txn_status = 0x10000;
            ljm_dptx_aux_write_fifo(dp_dev, wr_data, ct);
            ljm_dptx_aux_send_request(dp_dev, TR_DPTX_AUX_CMD_WRITE, addr, ct, false, false);
            //udelay(1);
            reply_status = ljm_dptx_aux_wait_reply(dp_dev);
            ljm_dptx_reg_read(dp_dev, TR_DPTX_INTERRUPT_CAUSE); // clear [0x140] state
            reply_code = ljm_dptx_aux_reply_code(dp_dev);
            txn_status |= (reply_status << 8) | reply_code;
        } while (reply_code == TR_DPTX_AUX_REPLY_DEFER && --timeout > 0);
    }

    return txn_status;
}



/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_read
 *  功能描述: DP AUX总线读取一个字节
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t addr
 *            uint8_t *pval
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_aux_read(struct dp_dev *dp_dev, uint32_t addr, uint8_t *pval)
{
    uint32_t txn_status;
    txn_status = ljm_dptx_aux_dpcd_read_blk(dp_dev, addr, pval, 1);

    if (((txn_status & 0xffff) || ((txn_status & 0x10000) == 0)))
    {
        WARN_PRINT("dp %d: aux read byte error\n", dp_dev->index);
        return -1;
    }

    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_read_blk
 *  功能描述: DP AUX总线读一块数据（最大16字节）
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t addr
 *            int size
 *            uint8_t *buf
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_aux_read_blk(struct dp_dev *dp_dev, uint32_t addr, uint8_t *buf, int size)
{
    uint32_t txn_status;
    int left_size = size;
    int cur_size;

    while (left_size > 0)
    {
        if (left_size > 16)
            cur_size = 16;
        else
            cur_size = left_size;
        left_size -= cur_size;

        txn_status = ljm_dptx_aux_dpcd_read_blk(dp_dev, addr, buf, cur_size);
        if (((txn_status & 0xffff) || ((txn_status & 0x10000) == 0)))
        {
            ERR_PRINT("dp %d: aux read buffer error\n", dp_dev->index);
            return -1;
        }

        buf += cur_size;
        addr += cur_size;
    }

    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_aux_write
 *  功能描述: DP AUX总线写一个字节
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t addr
 *            uint8_t val
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_aux_write(struct dp_dev *dp_dev, uint32_t addr, uint8_t val)
{
    uint32_t txn_status;
    txn_status = ljm_dptx_aux_dpcd_write_blk(dp_dev, addr, &val, 1);

    if (((txn_status & 0xffff) || ((txn_status & 0x10000) == 0)))
    {
        DEBUG_PRINT("dp %d: aux write byte error(0x%x)\n", dp_dev->index, txn_status);
        return -1;
    }

    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_common_ready_ack
 *  功能描述: 等待DP PHY应答
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t delay_time
 *            uint32_t loop_ct
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_common_ready_ack(struct dp_dev *dp_dev, uint32_t delay_time, uint32_t loop_ct)
{
    uint32_t         reg_val, idx;
    dptx_phy_error_t status = dptx_phy_error_bus_not_idle;

    for (idx = 0; idx < loop_ct; idx++)
    {
        reg_val = ljm_dptx_phy_reg_read(dp_dev, TR_DPTX_PHY_CMN_READY);
        reg_val &= 0x01ul;
        if (reg_val == 0x01ul)
        {
            status = dptx_phy_error_none;
            break;
        }
        else
        {
            ljm_dptx_tmr_wait(dp_dev, delay_time);
        }
    }
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_lane_pll_ack
 *  功能描述: 等待PLL锁定应答
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t delay_time
 *            uint32_t loop_ct
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_lane_pll_ack(struct dp_dev *dp_dev, uint32_t delay_time, uint32_t loop_ct)
{
    uint32_t         reg_val, idx;
    dptx_phy_error_t status = dptx_phy_error_bus_not_idle;

    for (idx = 0; idx < loop_ct; idx++)
    {
        reg_val = ljm_dptx_phy_reg_read(dp_dev, TR_DPTX_PHY_LANE_PLLCLK_EN_ACK);
        reg_val &= 0x01ul;
        if (reg_val == 0x01ul)
        {
            status = dptx_phy_error_none;
            break;
        }
        else
        {
            ljm_dptx_tmr_wait(dp_dev, delay_time);
        }
    }
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_pwr_state_ack
 *  功能描述: 等待电源状态应答
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t pwr_state
 *            uint32_t delay_time
 *            uint32_t loop_ct
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_pwr_state_ack(struct dp_dev *dp_dev, uint32_t pwr_state, uint32_t delay_time, uint32_t loop_ct)
{
    uint32_t         reg_val, idx;
    dptx_phy_error_t status = dptx_phy_error_pwr_ack;

    for (idx = 0; idx < loop_ct; idx++)
    {
        reg_val = ljm_dptx_phy_reg_read(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_ACK);
        reg_val &= 0x1ful;
        if (reg_val == pwr_state)
        {
            status = dptx_phy_error_none;
            break;
        }
        else
        {
            ljm_dptx_tmr_wait(dp_dev, delay_time);
        }
    }
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_power
 *  功能描述: 设置DP PHY电源状态
 *  输入参数: struct dp_dev *dp_dev
 *            dptx_power_state_t pwr_state
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
dptx_phy_error_t ljm_dptx_phy_power(struct dp_dev *dp_dev, dptx_power_state_t pwr_state)
{
    dptx_phy_error_t status = dptx_phy_error_none;

//      Notes:
//      - State A1 is not a used power state for DisplayPort.
//      - States A2 & A3 are configured identically for DisplayPort.
//      - State A3 requires special handshaking for entry/exit.
    switch (pwr_state)
    {
        case dptx_power_a0:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_A0_TXRX_ACTIVE);
            status = ljm_pwr_state_ack(dp_dev, DPTX_PHY_PWR_STATE_A0_TXRX_ACTIVE, 1, 50);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_IDLE);
            break;

        case dptx_power_a1:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_A1_POWERDOWN1);
            status = ljm_pwr_state_ack(dp_dev, DPTX_PHY_PWR_STATE_A1_POWERDOWN1, 1, 50);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_IDLE);
            break;

        case dptx_power_a2:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_A2_POWERDOWN2);
            status = ljm_pwr_state_ack(dp_dev, DPTX_PHY_PWR_STATE_A2_POWERDOWN2, 1, 50);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_IDLE);
            break;

        case dptx_power_a3:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_A3_POWERDOWN3);
            status = ljm_pwr_state_ack(dp_dev, DPTX_PHY_PWR_STATE_A3_POWERDOWN3, 1, 50);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, DPTX_PHY_PWR_STATE_IDLE);
            break;

        default:
            status = dptx_phy_error_pwr_type;
            break;
    }
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_init
 *  功能描述: 初始化DP PHY
 *  输入参数: struct dp_dev *dp_dev
 *            dptx_link_clk_t bw_code
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *         2. 日    期: 2024年5月9日
 *            作    者: 翟一夫
 *            修改内容: 弃用dptx_link_clk_t枚举变量，换用内核风格的bw_code
 *
 *****************************************************************************/
int ljm_dptx_phy_init(struct dp_dev *dp_dev, unsigned char bw_code)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    dptx_phy_error_t status = dptx_phy_error_none;
    uint32_t idx, dp_counts;

    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_CDB_RSTN, 1ul); // enable PHY's local configuration bus
    ljm_cdb_write(dp_dev, kPHY_CMN_CDIAG_REFCLK_OVRD, 4ul); // this must be the first address written after release of reset

    // reset AUXHPD block
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_AUXHPD_ENABLE_BITS, 0ul);   // all enable bits cleared
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_TX_ELEC_IDLE_BITS, 0x0ful); // set TX lanes to electrically idle state

    // reset PHY
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_DATA_ENABLE, 0ul);   // clear data enable bit
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_RSTN_BITS, 0ul);// set (low) lane reset_n signals
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_EN_BITS, 0ul);  // clear lane enable signals
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_CONFIG_RSTN, 0ul);   // set (low) the PHY wrapper's "global" configuration-based reset_n signal

    // enable AUXHPD block
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_AUXHPD_ENABLE_BITS, 1ul); // decap enable
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_AUXHPD_ENABLE_BITS, 3ul); // decap enable (delayed version)
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_AUXHPD_ENABLE_BITS, 7ul); // bg enable

    // reset PLLs
    ljm_cdb_write(dp_dev, kPHY_PMA_PLL_RAW_CTRL, 0x0000); // force both PLL0 and PLL1 into reset state
    ljm_cdb_write(dp_dev, kPHY_PLL_CFG, 0x0000);          // use PLL0 only- PLL1 disabled
    ljm_cdb_write(dp_dev, kPHY_PMA_ISO_PLL_CTRL0, 0x000f); // these "isolation" control registers must be set this way for DP lanes to function
    ljm_cdb_write(dp_dev, kPHY_PMA_ISO_PLL_CTRL1, 0x2222);

    // RST VAL for all link rate
    /*
    ljm_cdb_write(dp_dev, kPMA_CMN_SSM_BIAS_TMR,         0x0019);
    ljm_cdb_write(dp_dev, kPMA_CMN_PLLSM0_PLLPRE_TMR,    0x0032);
    ljm_cdb_write(dp_dev, kPMA_CMN_PLLSM0_PLLLOCK_TMR,   0x00d1);
    ljm_cdb_write(dp_dev, kPMA_CMN_PLLSM1_PLLPRE_TMR,    0x0032);
    ljm_cdb_write(dp_dev, kPMA_CMN_PLLSM1_PLLLOCK_TMR,   0x00d1);
    ljm_cdb_write(dp_dev, kPMA_CMN_BGCAL_INIT_TMR,       0x007d);
    ljm_cdb_write(dp_dev, kPMA_CMN_BGCAL_ITER_TMR,       0x007d);
    ljm_cdb_write(dp_dev, kPMA_CMN_IBCAL_INIT_TMR,       0x0019);
    ljm_cdb_write(dp_dev, kPMA_CMN_IBCAL_ITER_TMR,       0x0019);
    ljm_cdb_write(dp_dev, kPMA_CMN_TXPUCAL_INIT_TMR,     0x001e);
    ljm_cdb_write(dp_dev, kPMA_CMN_TXPUCAL_ITER_TMR,     0x0006);
    ljm_cdb_write(dp_dev, kPMA_CMN_TXPDCAL_INIT_TMR,     0x001e);
    ljm_cdb_write(dp_dev, kPMA_CMN_TXPDCAL_ITER_TMR,     0x0006);
    ljm_cdb_write(dp_dev, kPMA_CMN_RXCAL_INIT_TMR,       0x02ee);
    ljm_cdb_write(dp_dev, kPMA_CMN_RXCAL_ITER_TMR,       0x0006);
    ljm_cdb_write(dp_dev, kPMA_CMN_SD_CAL_INIT_TMR,      0x0006); // FIXME: 0x0002 in simulation task, diff with ds
    ljm_cdb_write(dp_dev, kPMA_CMN_SD_CAL_ITER_TMR,      0x0002);
    ljm_cdb_write(dp_dev, kPMA_CMN_SD_CAL_REFTIM_START,  0x000e);
    ljm_cdb_write(dp_dev, kPMA_CMN_SD_CAL_PLLCNT_START,  0x012b);
    */

    if((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
        dp_counts = 1;
    else
        dp_counts = 4;

    // configure lanes
    for (idx = 0; idx < dp_counts; idx++)
    {
        ljm_cdb_write(dp_dev, kXCVR_DIAG_PLLDRC_CTL + (idx * kLANE1_OFFSET), 0x0001);
        ljm_cdb_write(dp_dev, kXCVR_DIAG_BIDI_CTRL + (idx * kLANE1_OFFSET), 0x000f);

        if (bw_code == 0x06)
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0002);   // for 1.62G
        else if (bw_code == 0x09)
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0001);   // for 2.43G
        else if (bw_code == 0x0a)
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0001);   // for 2.70G
        else if (bw_code == 0x0c)
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0002);   // for 3.24G
        else if (bw_code == 0x14)
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0000);   // for 5.40G
        else if (bw_code == 0x1e)
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0000);   // for 8.10G
        else
            ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + (idx * kLANE1_OFFSET), 0x0002);   // for 1.62G

        ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_SEL + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kTX_PSC_A0 + (idx * kLANE1_OFFSET), 0x00fb);
        ljm_cdb_write(dp_dev, kTX_PSC_A2 + (idx * kLANE1_OFFSET), 0x04aa);
        ljm_cdb_write(dp_dev, kTX_PSC_A3 + (idx * kLANE1_OFFSET), 0x04aa);
        ljm_cdb_write(dp_dev, kRX_PSC_A0 + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kRX_PSC_A2 + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kRX_PSC_A3 + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kRX_PSC_CAL + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kRX_REE_GCSM1_CTRL + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kRX_REE_GCSM2_CTRL + (idx * kLANE1_OFFSET), 0x0000);
        ljm_cdb_write(dp_dev, kRX_REE_PERGCSM_CTRL + (idx * kLANE1_OFFSET), 0x0000);
    }

    ljm_cdb_write(dp_dev, kPLL0_DSM_DIAG_M0, 0x0004);
    ljm_cdb_write(dp_dev, kPDIAG_PLL0_CP_PADJ_M0, 0x0509);
    ljm_cdb_write(dp_dev, kPDIAG_PLL0_CP_IADJ_M0, 0x0f00);
    ljm_cdb_write(dp_dev, kPDIAG_PLL0_FILT_PADJ_M0, 0x0f08);
    ljm_cdb_write(dp_dev, kPLL0_FRACDIVH_M0, 0x0002);
    ljm_cdb_write(dp_dev, kPDIAG_PLL0_CTRL_M0, 0x0002);
    ljm_cdb_write(dp_dev, kPLL0_VCOCAL_INIT_TMR, 0x00fa);
    ljm_cdb_write(dp_dev, kPLL0_VCOCAL_ITER_TMR, 0x0004);
    ljm_cdb_write(dp_dev, kPLL0_VCOCAL_REFTIM_START, 0x0317);
    ljm_cdb_write(dp_dev, kPLL0_VCOCAL_TCTRL, 0x0003);
    ljm_cdb_write(dp_dev, kPLL0_LOCK_REFCNT_START, 0x00c7);
    ljm_cdb_write(dp_dev, kPLL0_LOCK_PLLCNT_START, 0x00c7);

    if (bw_code == 0x06)
    {
        // Configure PLL0, 25M ref clock, link rate 1.62G, 9.72GHz VCO
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, 0x0317);
        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, 0x0184);
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, 0xcccd);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, 0x0104);
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, 0x0f01); // CMN_PDIAG_PLL0_CLK_SEL_M0
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, 0x001e);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, 0x000c);
    }
    else if (bw_code == 0x09)
    {
        // Configure PLL0, 25M ref clock, link rate 2.43G, 9.72GHz VCO
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, 0x0317);
        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, 0x0184);
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, 0xcccd);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, 0x0104);
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, 0x0701);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, 0x001e);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, 0x000c);
    }
    else if (bw_code == 0x0a)
    {
        // Configure PLL0, 25M ref clock, link rate 2.7G, 10.8GHz VCO
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, 0x0317);
        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, 0x01b0);
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, 0x0000);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, 0x0120);
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, 0x0701);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, 0x008e);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, 0x0042);
    }
    else if (bw_code == 0x0c)
    {
        // Configure PLL0, 25M ref clock, link rate 3.24G, 9.72GHz VCO
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, 0x0317);
        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, 0x0184);
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, 0xcccd);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, 0x0104);
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, 0x0b00);

        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, 0x001e);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, 0x000c);
    }
    else if (bw_code == 0x14)
    {
        // Configure PLL0, 25M ref clock, link rate 5.4G, 10.8GHz VCO
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, 0x0318);
        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, 0x01b0);
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, 0x0000);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, 0x0120); //	PLL_HIGH_THR
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, 0x0301); //CMN_PDIAG_PLL0_CLK_SEL_M0
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, 0x008e);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, 0x0042);
    }
    else if (bw_code == 0x1e)
    {
        // Configure PLL0, 25M ref clock, link rate 8.1G, 8.1GHz VCO
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, 0x0318);
        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, 0x0144);
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, 0x0000);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, 0x00d8);
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, 0x0200);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, 0x006a);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, 0x0030);
    }
    else
    {
        ERR_PRINT("error: bw_code=0x%x is not supported yet! force to 1.62G\n", bw_code);
    }

    // release forced reset of PLL0 (leave PLL1 off)
    ljm_cdb_write(dp_dev, kPHY_PMA_PLL_RAW_CTRL, PMA_PLL_RAW_CTRL_PLL0_FORCE_RSTN_BIT);

    // enable all four lanes
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PLLCLK_ENABLE, 0x0001ul);
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_RSTN_BITS, 0x000ful);
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_EN_BITS, 0x000ful);
    //ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_EN_BITS, 0x0003ul);

    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_TX_ELEC_IDLE_BITS, 0x0000ul);
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_CONFIG_RSTN, 0x0001ul); // release the PHY global configuration-sourced reset (active low)
    status = ljm_common_ready_ack(dp_dev, 1, 500);
    if (status != dptx_phy_error_none)
    {
        DEBUG_PRINT("%d: common_ready_ack status %d is error\n", __LINE__, status);
    }
    status = ljm_lane_pll_ack(dp_dev, 1, 500);
    if (status != dptx_phy_error_none)
    {
        DEBUG_PRINT("%d: lane_pll_ack status %d is error\n", __LINE__, status);
    }

    status = ljm_dptx_phy_power(dp_dev, dptx_power_a2); //	bring the power state back up to active
    if (status != dptx_phy_error_none)
    {
        DEBUG_PRINT("%d: dptx_phy_power_a2 status %d is error\n", __LINE__, status);
    }
    status = ljm_dptx_phy_power(dp_dev, dptx_power_a0);
    if (status != dptx_phy_error_none)
    {
        DEBUG_PRINT("%d: dptx_phy_power_a0 status %d is error\n", __LINE__, status);
    }
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_DATA_ENABLE, 1ul);       // set data enable bit
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_AUXHPD_ENABLE_BITS, 0x0ful); // aux lines enabled for TX and RX

    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_clk_divider
 *  功能描述: 设置APB时钟频率，以获取1MB的AUX总线时钟
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t apb_clk_mhz
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_set_clk_divider(struct dp_dev *dp_dev, uint32_t apb_clk_mhz)
{
    ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_CLOCK_DIVIDER, apb_clk_mhz);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_get_clk_divider
 *  功能描述: 获取APB时钟值
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_get_clk_divider(struct dp_dev *dp_dev)
{
    return ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_CLOCK_DIVIDER);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_reset
 *  功能描述: DP PHY复位控制
 *  输入参数: struct dp_dev *dp_dev
 *            dptx_phy_reset_t reset_type
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
dptx_phy_error_t ljm_dptx_phy_reset(struct dp_dev *dp_dev, dptx_phy_reset_t reset_type)
{
    dptx_phy_error_t status = dptx_phy_error_none;

    switch (reset_type)
    {
        case dptx_phy_reset_all:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_DATA_ENABLE, 0ul);   // clear data enable bit
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_RSTN_BITS, 0ul);// set (low) lane reset_n signals
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_EN_BITS, 0ul);  // clear lane enable signals
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_CONFIG_RSTN, 0ul);   // set (low) the PHY wrapper's "global" configuration-based reset_n signal
            status = dptx_phy_error_none;
            break;

        case dptx_phy_reset_min:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PWR_STATE_REQ, 0ul);    // requested power state reset to "no request"
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PLLCLK_ENABLE, 0ul);    // disable link lanes' xcvr_pllclk_en and activate lanes' reset_n
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_RSTN_BITS, 0ul);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_EN_BITS, 0ul);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_CONFIG_RSTN, 0ul);   // set the PHY global configuration-sourced reset (active low)
            status = dptx_phy_error_none;
            break;

        default:
            status = dptx_phy_error_reset_type;
            break;
    }

    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_pll_raw_ctrl_ack
 *  功能描述: 等待PLL设置完成
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t delay_time
 *            uint32_t loop_ct
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_pll_raw_ctrl_ack(struct dp_dev *dp_dev, uint32_t delay_time, uint32_t loop_ct)
{
    uint16_t         reg_val;
    dptx_phy_error_t status = dptx_phy_error_bus_not_idle;
    uint32_t idx;

    for (idx = 0; idx < loop_ct; idx++)
    {
        ljm_cdb_read(dp_dev, kPHY_PMA_CMN_CTRL2, &reg_val);
        reg_val &= PMA_CMN_CTRL2_PLL0_DISABLED_BIT;
        if (reg_val)
        {
            status = dptx_phy_error_none;
            break;
        }
        else
        {
            ljm_dptx_tmr_wait(dp_dev, delay_time);
        }
    }
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_pll_disable
 *  功能描述: 禁用DP PHY PLL
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_pll_disable(struct dp_dev *dp_dev)
{
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_DATA_ENABLE, 0ul);   // clear data enable bit
    ljm_dptx_phy_power(dp_dev, dptx_power_a3);                      // set to low power

    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PLLCLK_ENABLE, 0ul);    //	disable lane pll clock
    ljm_lane_pll_ack(dp_dev, 1, 100);                                       //	wait for PLL lane ACK

    ljm_cdb_write(dp_dev, kPHY_PMA_PLL_RAW_CTRL, 0ul);  //	disable PLL0 (PLL1 is already off)
    ljm_pll_raw_ctrl_ack(dp_dev, 1, 100);               //	wait for operation to complete

    return dptx_phy_error_none;
}

/*****************************************************************************
 *  函 数 名: ljm_pll_enable
 *  功能描述: 使能DP PHY PLL
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_pll_enable(struct dp_dev *dp_dev)
{
    ljm_cdb_write(dp_dev, kPHY_PMA_PLL_RAW_CTRL, 1ul); //	enable PLL0
    ljm_pll_raw_ctrl_ack(dp_dev, 1, 100);              //	wait for operation to complete

    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_PLLCLK_ENABLE, 1ul); //	enable lane pll clock
    ljm_lane_pll_ack(dp_dev, 1, 100);                                    //	wait for PLL lane ACK

    ljm_dptx_phy_power(dp_dev, dptx_power_a2); //	bring the power state back up to active
    ljm_dptx_phy_power(dp_dev, dptx_power_a0);

    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_DATA_ENABLE, 1ul); //	set data enable bit

    return dptx_phy_error_none;
}

typedef struct
{
    uint32_t hsclk_div;
    uint32_t clk_sel;
    uint32_t pll_intdiv;
    uint32_t pll_fracdivl;
    uint32_t pll_high_thr;
    uint32_t pll_vcocal_pllcnt_start;
    uint32_t pll_dsm_fbh_ovrd;
    uint32_t pll_dsm_fbl_ovrd;
} phy_link_cfg_t;

static const phy_link_cfg_t link_cfg[] = {
    {0x0002, 0x0f01, 0x0184, 0xcccd, 0x0104, 0x0317, 0x001e, 0x000c}, // 1.62gbps VCO 9.72G
    {0x0001, 0x0701, 0x0184, 0xcccd, 0x0104, 0x0317, 0x001e, 0x000c}, // 2.43gpbs VCO 9.72G
    {0x0001, 0x0701, 0x01b0, 0x0000, 0x0120, 0x0318, 0x008e, 0x0042}, // 2.70gbps VCO 10.8G
    {0x0002, 0x0b00, 0x0184, 0xcccd, 0x0104, 0x0317, 0x001e, 0x000c}, // 3.24gbps VCO 9.72G
    {0x0000, 0x0301, 0x01b0, 0x0000, 0x0120, 0x0318, 0x008e, 0x0042}, // 5.40gbps VCO 10.8G
    {0x0000, 0x0200, 0x0144, 0x0000, 0x00d8, 0x0318, 0x006a, 0x0030}, // 8.10gbps VCO 8.1G
};

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_update_link_rate
 *  功能描述: 更新DP接口连接速度
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t bw_code
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *         2. 日    期: 2024年5月9日
 *            作    者: 翟一夫
 *            修改内容: 更改变量名
 *****************************************************************************/
dptx_phy_error_t ljm_dptx_phy_update_link_rate(struct dp_dev *dp_dev, uint32_t bw_code)
{
    const phy_link_cfg_t *cfg    = NULL;
    dptx_phy_error_t      status = dptx_phy_error_none;

    status = ljm_pll_disable(dp_dev);
    if (status != dptx_phy_error_none)
    {
        return status;
    }

    switch (bw_code)
    {
        case 0x6: // dptx_link_clk_1p62gbps:
            cfg = &link_cfg[0];
            break;

        case 0x9: // dptx_link_clk_2p43gbps:
            cfg = &link_cfg[1];
            break;

        case 0xa: // dptx_link_clk_2p70gbps:
            cfg = &link_cfg[2];
            break;

        case 0xc: // dptx_link_clk_3p24gbps:
            cfg = &link_cfg[3];
            break;

        case 0x14: // dptx_link_clk_5p40gbps:
            cfg = &link_cfg[4];
            break;

        case 0x1e: // dptx_link_clk_8p10gbps:
            cfg = &link_cfg[5];
            break;

        default:
            status = dptx_phy_error_link_rate;
            break;
    }

    if (cfg != NULL)
    {
        ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + kLANE0_OFFSET, cfg->hsclk_div); // reprogram link-speed
        ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + kLANE1_OFFSET, cfg->hsclk_div);
        ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + kLANE2_OFFSET, cfg->hsclk_div);
        ljm_cdb_write(dp_dev, kXCVR_DIAG_HSCLK_DIV + kLANE3_OFFSET, cfg->hsclk_div);

        ljm_cdb_write(dp_dev, kPLL0_INTDIV_M0, cfg->pll_intdiv); // reprogram PLL0 (PLL1 disabled)
        ljm_cdb_write(dp_dev, kPLL0_FRACDIVL_M0, cfg->pll_fracdivl);
        ljm_cdb_write(dp_dev, kPLL0_HIGH_THR_M0, cfg->pll_high_thr);
        ljm_cdb_write(dp_dev, kPDIAG_PLL0_CLK_SEL_M0, cfg->clk_sel);
        ljm_cdb_write(dp_dev, kPLL0_VCOCAL_PLLCNT_START, cfg->pll_vcocal_pllcnt_start);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBH_OVRD_M0, cfg->pll_dsm_fbh_ovrd);
        ljm_cdb_write(dp_dev, kPLL0_DSM_FBL_OVRD_M0, cfg->pll_dsm_fbl_ovrd);
    }

    status = ljm_pll_enable(dp_dev);
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_link_rate
 *  功能描述: 设置DP接口连接速度
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t bw_code
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *         2. 日    期: 2024年5月9日
 *            作    者: 翟一夫
 *            修改内容: 更改变量名
 *****************************************************************************/
uint32_t ljm_dptx_set_link_rate(struct dp_dev *dp_dev, uint32_t bw_code)
{
    uint32_t status = 0;

    ljm_dptx_reg_write(dp_dev, TR_DPTX_TRANSMITTER_ENABLE, 0); // Disable the core before continuing
    ljm_dptx_reg_write(dp_dev, TR_DPTX_LINK_BW_SET, bw_code);    // Set the link rate in the core (for completeness)
    status = ljm_dptx_phy_update_link_rate(dp_dev, bw_code);     // Set the link rate in the PHY
    if (status)
    {
        ERR_PRINT("ljm_dptx_phy_update_link_rate returns %d\n", status);
    }
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SOFT_RESET, 0x01);      // Reset the link
    ljm_dptx_tmr_wait(dp_dev, 2);                              // Small delay
    ljm_dptx_reg_write(dp_dev, TR_DPTX_TRANSMITTER_ENABLE, 1); // Re-enable the core for use

    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_set_lane_count
 *  功能描述: DP接口PHY设置连接通道数
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t lanecount
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
dptx_phy_error_t ljm_dptx_phy_set_lane_count(struct dp_dev *dp_dev, uint32_t lanecount)
{
    dptx_phy_error_t status  = dptx_phy_error_none;
    dptx_lane_en_t   lane_en = 0x0;

    status = ljm_pll_disable(dp_dev);
    if (status != dptx_phy_error_none)
    {
        return status;
    }

    if (lanecount == 1)
    {
        lane_en = dptx_lane_en_1;
    }
    else if (lanecount == 2)
    {
        lane_en = dptx_lane_en_2;
    }
    else if (lanecount == 4)
    {
        lane_en = dptx_lane_en_4;
    }
    else
    {
        WARN_PRINT("dptx_phy_set_lane_count: lane count error %d, force to lane count 1\n", lanecount);
        lane_en = dptx_lane_en_1;
    }

    switch (lane_en)
    {
        case dptx_lane_en_1:
        case dptx_lane_en_2:
        case dptx_lane_en_4:
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_RSTN_BITS, lane_en);
            ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_LANE_EN_BITS, lane_en);
            break;

        default:
            status = dptx_phy_error_lane_count;
            break;
    }

    status = ljm_pll_enable(dp_dev);
    return status;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_lane_count
 *  功能描述: DP接口设置连接通道数
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t lanecount
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_set_lane_count(struct dp_dev *dp_dev, uint32_t new_lc)
{
    uint32_t curr_lc;

    curr_lc = ljm_dptx_reg_read(dp_dev, TR_DPTX_LANE_COUNT_SET);

    if (curr_lc != new_lc)
    {   // reconfigure only on a change
        ljm_dptx_reg_write(dp_dev, TR_DPTX_LANE_COUNT_SET, new_lc);
        ljm_dptx_phy_power(dp_dev, dptx_power_a2);
        ljm_dptx_phy_set_lane_count(dp_dev, new_lc);
        //ljm_dptx_phy_reset(dp_dev, dptx_phy_reset_all); // reset the entire PHY state on a lane change
        ljm_dptx_phy_power(dp_dev, dptx_power_a0);
    }
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_set_vswing
 *  功能描述: 设置DP PHY信号电压幅值
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t lane0
 *            uint32_t lane1
 *            uint32_t lane2
 *            uint32_t lane3
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_dptx_phy_set_vswing(struct dp_dev *dp_dev, uint32_t lane0, uint32_t lane1, uint32_t lane2, uint32_t lane3)
{
    /* 在硬件设计上 lane 值设置与标准 DP lane 值设置相反了，此处对 lane 值按位取反以作调整 */
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN0, ~lane0); // vswing lane 0
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN1, ~lane1); // vswing lane 1
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN2, ~lane2); // vswing lane 2
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN3, ~lane3); // vswing lane 3

    return dptx_phy_error_none;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_voltage_swing
 *  功能描述: 设置DP PHY信号伏值摆幅
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t lane0
 *            uint32_t lane1
 *            uint32_t lane2
 *            uint32_t lane3
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月7日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_set_voltage_swing(struct dp_dev *dp_dev, uint32_t lane0, uint32_t lane1, uint32_t lane2, uint32_t lane3)
{
    lane0 &= 0x03;
    lane1 &= 0x03;
    lane2 &= 0x03;
    lane3 &= 0x03;

    ljm_dptx_phy_set_vswing(dp_dev, lane0, lane1, lane2, lane3);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_set_preemphasis
 *  功能描述: 设置DP PHY每个通道的预加重
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t lane0
 *            uint32_t lane1
 *            uint32_t lane2
 *            uint32_t lane3
 *  输出参数: 无
 *  返 回 值: static dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static dptx_phy_error_t ljm_dptx_phy_set_preemphasis(struct dp_dev *dp_dev, uint32_t lane0, uint32_t lane1, uint32_t lane2, uint32_t lane3)
{
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS0, ~lane0); // pre-emphasis lane 0
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS1, ~lane1); // pre-emphasis lane 1
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS2, ~lane2); // pre-emphasis lane 2
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS3, ~lane3); // pre-emphasis lane 3

    return dptx_phy_error_none;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_preemphasis
 *  功能描述: 设置DP接口的预加重
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t lane0
 *            uint32_t lane1
 *            uint32_t lane2
 *            uint32_t lane3
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_dptx_set_preemphasis(struct dp_dev *dp_dev, uint32_t lane0, uint32_t lane1, uint32_t lane2, uint32_t lane3)
{
    lane0 &= 0x03;
    lane1 &= 0x03;
    lane2 &= 0x03;
    lane3 &= 0x03;

    ljm_dptx_phy_set_preemphasis(dp_dev, lane0, lane1, lane2, lane3);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_set_vswing_all
 *  功能描述: 设置DP PHY所有通道的伏值摆幅
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t vs
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
dptx_phy_error_t ljm_dptx_phy_set_vswing_all(struct dp_dev *dp_dev, uint32_t vs)
{
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN0, vs); // vswing lane 0
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN1, vs); // vswing lane 1
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN2, vs); // vswing lane 2
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_VMARGIN3, vs); // vswing lane 3

    return dptx_phy_error_none;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_phy_set_preemphasis_all
 *  功能描述: 设置DP PHY所有通道的预加重
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t pe
 *  输出参数: 无
 *  返 回 值: dptx_phy_error_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
dptx_phy_error_t ljm_dptx_phy_set_preemphasis_all(struct dp_dev *dp_dev, uint32_t pe)
{
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS0, pe); // pre-emphasis lane 0
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS1, pe); // pre-emphasis lane 1
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS2, pe); // pre-emphasis lane 2
    ljm_dptx_phy_reg_write(dp_dev, TR_DPTX_PHY_EMPHASIS3, pe); // pre-emphasis lane 3

    return dptx_phy_error_none;
}

/*****************************************************************************
 *  函 数 名: ljm_dp_init
 *  功能描述: 初始化DP接口
 *  输入参数: struct dp_dev *dp_dev
 *            void *base
 *            int offset
 *            int index
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_dptx_init(struct dp_dev *dp_dev, void *base, int offset, int index)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dp_dev->ddev);
    uint32_t apb_clk;
    HEADER_ARG("dp_dev=%lx base=%lx offset=%x index=%d", (unsigned long)dp_dev, (unsigned long)base, offset, index);

#ifdef UEFI_GOP
    dp_dev->Private = base;
#else
    dp_dev->regbase = base;
#endif

    dp_dev->offset = offset;
    dp_dev->index = index;

    if(ljm_dev->once_edp)
    {
        apb_clk = ljm_dptx_reg_read(dp_dev, TR_DPTX_AUX_CLOCK_DIVIDER);
        if (apb_clk != 0 && dp_dev->is_edp) {
            DEBUG_PRINT("EDP%d: skip dptx init\n", index);
            FOOTER_ARG("dp %d: offset=%x", dp_dev->index, dp_dev->offset);
            return 0;
        }
    }

    /*
     * set aux clock to 1Mhz from APB frequency (100Mhz)
     */
    apb_clk = ljm_gp200_get_freq(ljm_dev->cur_card, freq_apb);
    //apb_clk = (apb_clk > 110) ? (apb_clk * 95 /100) : apb_clk;  /* FIXME: by sxk */
    ljm_dptx_set_clk_divider(dp_dev, apb_clk);

    /*
     * Initialize the DisplayPort TX core
     */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SOURCE_ENABLE, 0x00);     // disable MST source 0 (SST mode from init)
    ljm_dptx_reg_write(dp_dev, TR_DPTX_TRANSMITTER_ENABLE, 0x00);// Disable the transmitter

    /* tmr init */
    ljm_dptx_tmr_init(dp_dev); // Initialize the timer block in the DisplayPort TX core

    /* phy init */
    ljm_dptx_phy_init(dp_dev, 0x0a); // Init the PHY

    /* FIXME: for test */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_LANE_COUNT_SET, 2);

    //ljm_dptx_reg_write(dp_dev, TR_DPTX_TRANSMITTER_ENABLE, 0x01);   // Enable the transmitter
    ljm_dptx_reg_write(dp_dev, TR_DPTX_INTERRUPT_MASK, 0x0f);       // Unmask all interrupts (Fix me -sxk)
    ljm_dptx_reg_write(dp_dev, TR_DPTX_AUX_REPLY_TIMEOUT_INTERVAL, 440);    // Set the AUX reply timeout to spec + 10% (400us by spec + 10% = 440us)
    //ljm_dptx_reg_write(dp_dev, TR_DPTX_SOURCE_ENABLE, 0x01);        // Enable MST source 0 (SST mode from init)

    FOOTER_ARG("dp %d: offset=%x", dp_dev->index, dp_dev->offset);
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_dp_hpd_sense
 *  功能描述: 获取DP接口设备热插拔状态
 *  输入参数: struct dp_dev *dp_dev
 *  输出参数: 无
 *  返 回 值: bool
 *
 *  修改历史:
 *         1. 日    期: 2022年7月11日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 ******************2***********************************************************/
bool ljm_dptx_hpd_sense(struct dp_dev *dp_dev)
{
    return !!(ljm_dptx_reg_read(dp_dev, TR_DPTX_HPD_INPUT_STATE) & 0x1);
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_tu_size
 *  功能描述: 计算和设置DP接口传输单位
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t pixel_clock
 *            uint32_t bits_per_pixel
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_set_tu_size(struct dp_dev *dp_dev, uint32_t pixel_clock, uint32_t bits_per_pixel)
{
    uint32_t bw          = 0;
    uint32_t tu_size     = 0;
    uint32_t data_per_tu = 0;

    bw = ljm_dptx_reg_read(dp_dev, TR_DPTX_LANE_COUNT_SET) * ljm_dptx_reg_read(dp_dev, TR_DPTX_LINK_BW_SET) * 27;

    if (bw != 0)
    {
        data_per_tu = (pixel_clock * 8 * bits_per_pixel) / bw;
        tu_size     = ((data_per_tu % 1000) * 16) / 1000;
        tu_size     = (tu_size << 8) + (data_per_tu / 1000);
        tu_size     = (tu_size << 16) | 0x0040; // fixed 64 TU size
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_TU_CONFIG, tu_size);
    }

    TRACE("tu_size=0x%08x", tu_size);

    return tu_size;
}

/*****************************************************************************
 *  函 数 名: ljm_dptx_set_user_data_count
 *  功能描述: 计算和设置DP接口User Data Count
 *    SYMBOL_COUNT = ((HRES * bits per pixel) + 7) / 8
 *    UDC = (SYMBOL_COUNT + lane_count – 1) / lane_count
 *  输入参数: struct dp_dev *dp_dev
 *            uint32_t hres
 *            uint32_t bytes_per_pixel
 *            uint32_t lane_count
 *  输出参数: 无
 *  返 回 值: uint32_t
 *
 *  修改历史:
 *         1. 日    期: 2022年7月16日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
uint32_t ljm_dptx_set_user_data_count(struct dp_dev *dp_dev, uint32_t hres, uint32_t bytes_per_pixel, uint32_t lane_count)
{
    uint32_t udc;

    udc = ((hres * bytes_per_pixel * 8) + 7) >> 3; // Compute user data count (data per lane)
    udc = (udc + lane_count - 1) / lane_count;
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_DATA_COUNT, udc);
    TRACE("udc=%08x", udc);

    return udc;
}

/**
 * ljm_dptx_phy_irq_clear - 清除phy中断
 *
 * @dp_dev: dp接口设备指针。
 *
 * 对各条lane对应的cdb寄存器做读操作，以清除中断。
*/
int ljm_dptx_phy_irq_clear(struct dp_dev *dp_dev)
{
    uint16_t v0, v1, v2, v3;

    ljm_cdb_read(dp_dev, kPHY_INTERRUPT_STATUS_LANE0, &v0);
    ljm_cdb_read(dp_dev, kPHY_INTERRUPT_STATUS_LANE1, &v1);
    ljm_cdb_read(dp_dev, kPHY_INTERRUPT_STATUS_LANE2, &v2);
    ljm_cdb_read(dp_dev, kPHY_INTERRUPT_STATUS_LANE3, &v3);

    return 0;
}

# if 0
int ljm_dptx_mode_set(struct dp_dev *dp_dev, const struct drm_display_mode *mode)
{
    int r = 0;
    uint8_t rd_data[16];    // 16 bytes to hold read data
    uint32_t tp2_time = 400; // AUX training time for TP2
    uint32_t nvid;
    uint32_t tmp32;
    int no_video_time = 0;

    HEADER_ARG("dp_dev=0x%lx mode=0x%lx: %dx%d@%d",
        (unsigned long)dp_dev, (unsigned long)mode, mode->hdisplay, mode->vdisplay, mode->clock);

    r = drm_dp_dpcd_read(&dp_dev->aux, TR_DPCD_REVISION, rd_data, 16);
    if (r < 0) {
        INFO_PRINT("aux read linkrate from monitor error, force to 1.62Gbps with 4 lanes\n");

        dp_dev->link_rate = ljm_link_clk_to_speed(dptx_link_clk_1p62gbps);
        dp_dev->lane_count = 4;
    }
    else {
        DEBUG_PRINT("DP %d DPCD[0x%x]: %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n", dp_dev->index, TR_DPCD_REVISION, rd_data[0], rd_data[1],
           rd_data[2], rd_data[3], rd_data[4], rd_data[5], rd_data[6], rd_data[7], rd_data[8], rd_data[9], rd_data[10],
           rd_data[11], rd_data[12], rd_data[13], rd_data[14], rd_data[15]);

        dp_dev->link_rate = rd_data[1];
        dp_dev->lane_count = rd_data[2] & 0x1f;
    }
    DEBUG_PRINT("dp %d: speed=%x lanes=%d", dp_dev->index, dp_dev->link_rate, dp_dev->lane_count);

    if (rd_data[14] > 0)
    {
        tp2_time = (tp2_time * 10) * rd_data[14];
    }

    /* link training */
    if (dp_dev->is_edp)
        ljm_dptx_train_link_no_aux(dp_dev, dp_dev->link_rate, dp_dev->lane_count, tp2_time, true);
    else
        ljm_dptx_train_link(dp_dev, dp_dev->link_rate, dp_dev->lane_count, tp2_time, false);

    /* mode set */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HTOTAL, mode->htotal);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VTOTAL, mode->vtotal);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HSWIDTH, mode->hsync_end - mode->hsync_start);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VSWIDTH, mode->vsync_end - mode->vsync_start);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HRES, mode->hdisplay);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VRES, mode->vdisplay);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_HSTART, mode->htotal - mode->hsync_start);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_VSTART, mode->vtotal - mode->vsync_start);

    tmp32 = 0;
    if (mode->flags & DRM_MODE_FLAG_NHSYNC) tmp32 |= 0x1;
    if (mode->flags & DRM_MODE_FLAG_NVSYNC) tmp32 |= 0x2;
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_POLARITY, tmp32);

    tmp32 = 0xc;
    if (mode->flags & DRM_MODE_FLAG_PHSYNC) tmp32 |= 0x1;
    if (mode->flags & DRM_MODE_FLAG_PVSYNC) tmp32 |= 0x2;
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_SYNC_POLARITY, tmp32);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_MISC0, 0x21);   /* 8bits depth, rgb, synchronous clock */
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MAIN_STREAM_MISC1, 0x0);    /* no stereo video */

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_MVID, mode->clock); /* m-vid, user data pixel clock value in KHz */

    ljm_dptx_set_tu_size(dp_dev, mode->clock, 24);  /* TR_DPTX_SRC0_TU_CONFIG (0x1b0) */

    nvid = ljm_dptx_reg_read(dp_dev, TR_DPTX_LINK_BW_SET) * 27000;
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_NVID, nvid);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_PIXEL_COUNT, 0x1);

    ljm_dptx_set_user_data_count(dp_dev, mode->hdisplay, 3, dp_dev->lane_count);

    ljm_dptx_reg_write(dp_dev, TR_DPTX_USER_DATA_CONTROL, 0x2004);  /* src data control 0_1 0x2006 */

    if (dp_dev->index == 3)
    {
        //ljm_edp_self_test_set(dp_dev, 0x1);/* Set edp self test mode, 1 = open, 0 = close */
        ljm_dptx_reg_write(dp_dev, 0x0c8, 0x4000);  /* user data control */
        // sparse mode enable
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_CONTROL, 0x1); // sparse mode enable

        // enable VSC packets
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC0_INFOFRAME_ENABLE, 0x0); // infoframes off
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC0_3D_VSC_DATA, 0x0);      // clear the 3D/VSC byte
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC_PSR_CONFIG, 0x0);        // psr off
        ljm_dptx_reg_write(dp_dev, 0x368, 0x0000);  /* sec_psr_state */
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_SECONDARY_DATA_WINDOW, 0x80); // secondary data window

        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC0_AUDIO_ENABLE, 0x0);     // audio off
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SECONDARY_STREAM_ENABLE, 0x1);   // secondary channel enabled
        ljm_dptx_reg_write(dp_dev, TR_DPTX_EDP_CRC_ENABLE, 0x0);        // crc enabled
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC_PSR_3D_ENABLE, 0x0);     // psr/vsc disabled

        tmp32 = ljm_dptx_reg_read(dp_dev, TR_DPTX_SEC_PSR_STATE_INTERNAL);
        if (tmp32 != 0x1)
            INFO_PRINT("PSR function not DISABLED (0x%x)\n", tmp32);

        ljm_dptx_reg_write(dp_dev, TR_DPTX_SEC_PSR_3D_ENABLE, 0x1);     // psr/vsc enabled
        ljm_dptx_tmr_wait(dp_dev, 1);   // wait 1us

        tmp32 = ljm_dptx_reg_read(dp_dev, TR_DPTX_SEC_PSR_STATE_INTERNAL);
        if (tmp32 != 0x2)
            INFO_PRINT("PSR transition error: DISABLED -> INACTIVE (0x%x)\n", tmp32);
    }
    else
    {
        // sparse mode disable
        ljm_dptx_reg_write(dp_dev, TR_DPTX_SRC0_USER_CONTROL, 0x0);
    }

    DEBUG_PRINT("DP%d: enable the link\n", dp_dev->index);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_SOURCE_ENABLE, 0x1); // input source enable
    no_video_time = 8192 * 6 * 2 / (nvid / 1000); // 8192 * 6 * 1.1 / (nvid / 1000)
    ljm_dptx_tmr_wait(dp_dev, no_video_time);
    ljm_dptx_reg_write(dp_dev, TR_DPTX_VIDEO_STREAM_ENABLE, 0x1);   // video stream enable
    ljm_dptx_tmr_wait(dp_dev, 5);    // wait for 5us
    //ljm_dptx_reg_write(dp_dev, TR_DPTX_FORCE_SCRAMBLER_RESET, 0x1);

    FOOTER();
    return 0;
}
#endif

#endif /* USE_DP_CONNECTOR */
