/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，武汉凌久微电子有限公司
 *
 ******************************************************************************
 *
 *  文件名称：ljmicro_dp_hw.h
 *  版 本 号：V1.0
 *  作    者：胡艳明
 *  生成日期：2022年7月5日
 *
 *  功能描述：DP接口硬件相关信息的头文件
 *
 *  修改历史：
 *         1. 日    期：2022年7月5日
 *            作    者：胡艳明
 *            修改内容：创建文件
 *
 *****************************************************************************/

#ifndef __LJMICRO_DP_HW__
#define __LJMICRO_DP_HW__

/******************************************************************************
 *                                     头文件                                 *
 ******************************************************************************/

#include "ljmicro_dpcd.h"
#include <linux/version.h>
#if !defined(UEFI_GOP) && !defined(UTIL_TOOL)
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include <drm/drm_dp_helper.h>
#else
#include <drm/display/drm_dp_helper.h>
#endif
#endif

#include <sound/hdmi-codec.h>
#include <linux/regmap.h>
#include <linux/version.h>

/******************************************************************************
 *                                     宏定义                                 *
 ******************************************************************************/

/*
 * DP 控制器寄存器
 */

//------------------------------------------------------------------------------
//  link configuration fields
//------------------------------------------------------------------------------
#define TR_DPTX_LINK_BW_SET                                 0x000ul
#define TR_DPTX_LANE_COUNT_SET                              0x004ul
#define TR_DPTX_ENHANCED_FRAMING_ENABLE                     0x008ul
#define TR_DPTX_TRAINING_PATTERN_SET                        0x00cul
#define TR_DPTX_LINK_QUAL_PATTERN_SET                       0x010ul
#define TR_DPTX_DISABLE_SCRAMBLING                          0x014ul
#define TR_DPTX_DOWNSPREAD_CONTROL                          0x018ul
#define TR_DPTX_EDP_CAPABILITY_CONFIG                       0x01cul
#define TR_DPTX_HBR2_SCRAMBLER_RESET                        0x020ul
#define TR_DPTX_DISPLAYPORT_VERSION                         0x024ul
#define TR_DPTX_ALPM_POWER_SET                              0x028ul
#define TR_DPTX_LANE_REMAP                                  0x02cul
#define TR_DPTX_CUSTOM_80BIT_PATTERN_31_0                   0x030ul
#define TR_DPTX_CUSTOM_80BIT_PATTERN_63_32                  0x034ul
#define TR_DPTX_CUSTOM_80BIT_PATTERN_79_64                  0x038ul

//------------------------------------------------------------------------------
//  core enables
//------------------------------------------------------------------------------
#define TR_DPTX_TRANSMITTER_ENABLE                          0x080ul
#define TR_DPTX_VIDEO_STREAM_ENABLE                         0x084ul
#define TR_DPTX_SECONDARY_STREAM_ENABLE                     0x088ul
#define TR_DPTX_SRC0_SECONDARY_DATA_WINDOW                  0x08cul
#define TR_DPTX_SOFT_RESET                                  0x090ul
#define TR_DPTX_SOURCE_ENABLE                               0x094ul
#define TR_DPTX_FEC_ENABLE                                  0x098ul

//------------------------------------------------------------------------------
//  misc control registers
//------------------------------------------------------------------------------
#define TR_DPTX_FORCE_SCRAMBLER_RESET                       0x0c0ul
#define TR_DPTX_USER_DATA_CONTROL                           0x0c8ul
#define TR_DPTX_CORE_FEATURES                               0x0f8ul
#define TR_DPTX_CORE_REVISION                               0x0fcul

//------------------------------------------------------------------------------
//  AUX channel
//------------------------------------------------------------------------------
#define TR_DPTX_AUX_COMMAND                                 0x100ul
#define TR_DPTX_AUX_WRITE_FIFO                              0x104ul
#define TR_DPTX_AUX_ADDRESS                                 0x108ul
#define TR_DPTX_AUX_CLOCK_DIVIDER                           0x10cul
#define TR_DPTX_AUX_REPLY_TIMEOUT_INTERVAL                  0x110ul
#define TR_DPTX_HPD_INPUT_STATE                             0x128ul
#define TR_DPTX_INTERRUPT_STATE                             0x130ul
#define TR_DPTX_AUX_REPLY_DATA                              0x134ul
#define TR_DPTX_AUX_REPLY_CODE                              0x138ul
#define TR_DPTX_AUX_REPLY_COUNT                             0x13cul
#define TR_DPTX_INTERRUPT_CAUSE                             0x140ul
#define TR_DPTX_INTERRUPT_MASK                              0x144ul
#define TR_DPTX_INTERRUPT_HPD_IRQ                      (0X1 << 1)
#define TR_DPTX_INTERRUPT_HPD_EVENT                    (0X1 << 0)
#define TR_DPTX_AUX_REPLY_DATA_COUNT                        0x148ul
#define TR_DPTX_AUX_STATUS                                  0x14cul
#define TR_DPTX_AUX_REPLY_CLOCK_WIDTH                       0x150ul
#define TR_DPTX_AUX_PHY_WAKE_ACK_DETECTED                   0x154ul
#define TR_DPTX_HOST_TIMER                                  0x158ul

//------------------------------------------------------------------------------
//	Main stream control, virtual source 0
//------------------------------------------------------------------------------
#define TR_DPTX_SRC0_MAIN_STREAM_HTOTAL                     0x180ul
#define TR_DPTX_SRC0_MAIN_STREAM_VTOTAL                     0x184ul
#define TR_DPTX_SRC0_MAIN_STREAM_POLARITY                   0x188ul
#define TR_DPTX_SRC0_MAIN_STREAM_HSWIDTH                    0x18cul
#define TR_DPTX_SRC0_MAIN_STREAM_VSWIDTH                    0x190ul
#define TR_DPTX_SRC0_MAIN_STREAM_HRES                       0x194ul
#define TR_DPTX_SRC0_MAIN_STREAM_VRES                       0x198ul
#define TR_DPTX_SRC0_MAIN_STREAM_HSTART                     0x19cul
#define TR_DPTX_SRC0_MAIN_STREAM_VSTART                     0x1a0ul
#define TR_DPTX_SRC0_MAIN_STREAM_MISC0                      0x1a4ul
#define TR_DPTX_SRC0_MAIN_STREAM_MISC1                      0x1a8ul
#define TR_DPTX_SRC0_MVID                                   0x1acul
#define TR_DPTX_SRC0_TU_CONFIG                              0x1b0ul
#define TR_DPTX_SRC0_NVID                                   0x1b4ul
#define TR_DPTX_SRC0_USER_PIXEL_COUNT                       0x1b8ul
#define TR_DPTX_SRC0_USER_DATA_COUNT                        0x1bcul
#define TR_DPTX_SRC0_MAIN_STREAM_INTERLACED                 0x1c0ul
#define TR_DPTX_SRC0_USER_SYNC_POLARITY                     0x1c4ul
#define TR_DPTX_SRC0_USER_CONTROL                           0x1c8ul

//------------------------------------------------------------------------------
//	eDP CRC, virtual source 0
//------------------------------------------------------------------------------
#define TR_DPTX_EDP_CRC_ENABLE                              0x1d0ul
#define TR_DPTX_EDP_CRC_RED                                 0x1d4ul
#define TR_DPTX_EDP_CRC_GREEN                               0x1d8ul
#define TR_DPTX_EDP_CRC_BLUE                                0x1dcul

//------------------------------------------------------------------------------
//  PHY status
//------------------------------------------------------------------------------
#define TR_DPTX_PHY_STATUS                                  0x280ul

//------------------------------------------------------------------------------
/* Stream Registers */
//------------------------------------------------------------------------------
#define SECONDARY_STREAM_ENABLE         0x088
#define m_SECONDARY_STREAM_ENABLE       BIT(0)
#define v_SECONDARY_STREAM_ENABLE       BIT(0)

//------------------------------------------------------------------------------
/* Audio Registers */
//------------------------------------------------------------------------------
#define DP_AUDIO_ENABLE                 0x300
#define m_DP_AUDIO_ENABLE               BIT(0)
#define DP_INPUT_SELECT                 0x304
#define m_DP_INPUT_SELECT               0x3
#define v_DP_INPUT_I2S                  0x0
#define v_DP_INPUT_DS_FIFO              0x1
#define v_DP_INPUT_SPDIF                0x2
#define v_DP_INPUT_DMA                  0x3
//#define m_DP_INPUT_SELECT               0x3
//#define v_DP_INPUT_SELECT(n)            (((n) & 0x3) << 0)
#define DP_CHANNEL_COUNT                0x308
#define v_DP_CHANNEL_COUNT_MUTE         0x0
#define v_DP_CHANNEL_COUNT_2            0x2
#define v_DP_CHANNEL_COUNT_2_1          0x3
#define v_DP_CHANNEL_COUNT_5_1          0x6
#define v_DP_CHANNEL_COUNT_7_1          0x8
#define DP_DIRECT_CLKDIV                0x30C
#define m_DP_DIRECT_CLKDIV              0xffff
/* 15:0 v = apb/sample */
#define DP_SDP_PACKET_ENABLE            0x310
#define m_DP_SDP_PACKET_ENABLE          0x7f
/* 6:0 enable*/
#define DP_SDP_PACKET_RATE              0x314
/* 6:0 enable continuous transmission*/
#define DP_MAUD                         0x318
#define DP_NAUD                         0x31C
/* 23:0 maud naud, in sync mode, clock freq, refer to [2.2.5.2.2] */
#define DP_AUDIO_CLOCK_MODE             0x320
#define m_DP_AUDIO_CLOCK_MODE           BIT(0)
#define v_DP_AUDIO_CLOCK_MODE(n)        (((n) & 0x1) << 0)
/* 0, 0 for sync mode, 1 for async mode  */
#define DP_3D_VSC_DATA                  0x324
/* 3:0 code, 7:4 param      , refer to [2.2.5.2.2]*/
#define DP_AUDIO_FIFO                   0x328
/* 3:0 sample FIFO entry */
#define DP_AUDIO_FIFO_DEPTH             0x32C
/* 9:0 max number of samples stored in FIFO  */
#define DP_AUDIO_FIFO_READY             0x330
#define m_DP_AUDIO_FIFO_READY           BIT(0)
/* 0 ready, read only   */
#define DP_SDP_BUFFER_SELECT            0x334
/* 2:0, value 0-6, select buffer to write */
#define DP_SDP_BUFFER_DATA              0x338
/* 7:0 byte value write to SDP SRAM   */
#define DP_TIMESTAMP_INTERVAL           0x33C
/* 14:0 mseconds, wait time between transmissions of each audio timestamp packet */
#define DP_CS_SOURCE_FORMAT             0x340
/* 7:4 SOURCE_NUMBER id */
/* 3 LINEAR_PCM,  0 for linear, 1 for encoded */
/* 2:0 PCM_AUDIO_FORMAT,  channel status */
/* fields in the channel status information sent with each audio frame.*/
#define DP_CS_CATEGORY_CODE             0x344
/* 7:0 category code, type  */
#define DP_CS_LENGTH_ORIG_FREQ          0x348
/* 7:4 SAMPLE_WORD_LENGTH, 32-35 BIT in CS field in IEC 60958-3 SPEC, */
/* 3:0 ORIGIN_SAMPLEING_FREQ, 39-36 BIT in CS field */
#define DP_CS_FREQ_CLOCK_ACCURACY       0x34C
/* 7:4 SAMPLING_FREQ 24-27 BIT in CS, 3:0 CLOCK_ACCURACY 28-29 BIT in CS  */
#define DP_CS_COPYRIGHT                 0x350
/* 2:1 CGMS-A info, 0 COPYRIGHT CP-bit  */
#define DP_GTC_COUNT_CONFIG             0x354
/* GTC for global time code */
/* 23:16 GTC_COUNT_INT int portion of GTC count, */
/* 15:0 GTC_COUNT_FRAC fractional portion, v = f_count_decimal * 65536 */
#define DP_GTC_COMMAND_EDGE             0x358
/* 31:0 internal GTC counter value, read only */
#define DP_AUDIO_CHANNEL_MAP            0x35C
/* 8 channels remap, value 0-7 */
/* 31:28         27:24  23:20  19:16  15:12  11:8  7:4  3:0 */
/* CHANNEL_7_MAP   6      5      4      3      2    1    0  */
#define DP_SEC_AUDIO_FIFO_OVERFLOW      0x370
#define DP_SEC_PACKET_COUNT             0x374
#define DP_SEC_GTC_AUX_FRAME_SYNC       0x378
#define DP_SEC_CS_USER_DATA             0x37C
#define DP_SEC_DATA_PACKET_ID           0x380
#define DP_SEC_ADAPTICE_SYNC_SDP_ENABLE 0x384
#define DP_SEC_DB_LANE_SELECT           0x3E0
#define DP_SEC_DB_WRITE_INDEX           0x3E4
#define DP_SEC_DB_DATA_COUNT            0x3E8
#define DP_SEC_DB_DATA                  0x3EC
#define DP_SEC_DB_READY                 0x3F0
#define DP_SEC_DB_BUSY                  0x3F4
#define DP_SEC_DB_ENABLE                0x3F8
#define DP_HDCP_ENABLE                  0x400
#define DP_MST_ENABLE                   0x500
#define TR_DPTX_SRC3_USER_CONTROL       0x6C8

//------------------------------------------------------------------------------
//	Secondary channel, source 0
//------------------------------------------------------------------------------
#define TR_DPTX_SEC0_AUDIO_ENABLE                           0x300ul
#define TR_DPTX_SEC0_INPUT_SELECT                           0x304ul
#define TR_DPTX_SEC0_CHANNEL_COUNT                          0x308ul
#define TR_DPTX_SEC0_DIRECT_CLKDIV                          0x30cul
#define TR_DPTX_SEC0_INFOFRAME_ENABLE                       0x310ul
#define TR_DPTX_SEC0_INFOFRAME_RATE                         0x314ul
#define TR_DPTX_SEC0_MAUD                                   0x318ul
#define TR_DPTX_SEC0_NAUD                                   0x31cul
#define TR_DPTX_SEC0_AUDIO_CLOCK_MODE                       0x320ul
#define TR_DPTX_SEC0_3D_VSC_DATA                            0x324ul
#define TR_DPTX_SEC0_AUDIO_FIFO                             0x328ul
#define TR_DPTX_SEC0_AUDIO_FIFO_DEPTH                       0x32cul
#define TR_DPTX_SEC0_AUDIO_FIFO_READY                       0x330ul
#define TR_DPTX_SEC0_INFOFRAME_SELECT                       0x334ul
#define TR_DPTX_SEC0_INFOFRAME_DATA                         0x338ul
#define TR_DPTX_SEC0_TIMESTAMP_INTERVAL                     0x33cul
#define TR_DPTX_SEC0_CS_SOURCE_FORMAT                       0x340ul
#define TR_DPTX_SEC0_CS_CATEGORY_CODE                       0x344ul
#define TR_DPTX_SEC0_CS_LENGTH_ORIG_FREQ                    0x348ul
#define TR_DPTX_SEC0_CS_FREQ_CLOCK_ACCURACY                 0x34cul
#define TR_DPTX_SEC0_CS_COPYRIGHT                           0x350ul
#define TR_DPTX_SEC0_GTC_COUNT_CONFIG                       0x354ul
#define TR_DPTX_SEC0_GTC_COMMAND_EDGE                       0x358ul
#define TR_DPTX_SEC0_AUDIO_CHANNEL_MAP                      0x35cul
#define TR_DPTX_SEC_PSR_3D_ENABLE                           0x360ul
#define TR_DPTX_SEC_PSR_CONFIG                              0x364ul
#define TR_DPTX_SEC_PSR_STATE                               0x368ul
#define TR_DPTX_SEC_PSR_STATE_INTERNAL                      0x36cul

/*
 * AUX寄存器相关
 */

//------------------------------------------------------------------------------
//  TR_DPTX_AUX_COMMAND (offset 0x100)
//
//  Initiates an AUX channel command of the specified length when written.
//  This register is written last as part of the AUX channel request set up process.
//  When written, the internal state machines will begin transmitting the request to the sink device.
//  The AUX_ADDRESS and, when applicable, the AUX_WRITE_FIFO must be set up before writing this register.
//
//  13 - AUX_PHY_WAKE:
//      Set to a 1 to send the AUX_PHY_WAKE signal for embedded DisplayPort operations.
//      This signal is sent as a standalone operation and is not sent as a part of a larger AUX transaction.
//
//  12 - ADDRESS_ONLY:
//      Set to a 1 to initiate an address only request.
//
//  11:8 - COMMAND:
//      AUX Channel Command. The field is equivalent to the commands defined in
//      section 2.4.1 of the DisplayPort specification.
//      0x8 = AUX Write
//      0x9 = AUX Read
//      0x0 = I2C over AUX Write
//      0x4 = I2C over AUX Write, Middle of Transaction bit set
//      0x1 = I2C over AUX Read
//      0x5 = I2C over AUX Read, Middle of Transaction bit set
//      0x2 = I2C over AUX Write Status
//
//  3:0 - BYTE_COUNT:
//      Specifies the number of bytes to transfer with the current command.
//      The range of the register is 0 to 15 indicating between 1 and 16 bytes of data.
//------------------------------------------------------------------------------
#define TR_DPTX_AUX_CMD_PHY_WAKE                (0x01ul << 13)
#define TR_DPTX_AUX_CMD_ADDRESS_ONLY            (0x01ul << 12)

#define TR_DPTX_AUX_CMD_MASK                    0x0f00ul
#define TR_DPTX_AUX_CMD_READ                    0x0900ul
#define TR_DPTX_AUX_CMD_WRITE                   0x0800ul

#define TR_DPTX_AUX_CMD_I2C_WRITE               0x0000ul
#define TR_DPTX_AUX_CMD_I2C_READ                0x0100ul
#define TR_DPTX_AUX_CMD_I2C_WRITE_STATUS        0x0200ul
#define TR_DPTX_AUX_CMD_I2C_READ_MOT            0x0500ul
#define TR_DPTX_AUX_CMD_I2C_WRITE_MOT           0x0400ul
#define TR_DPTX_AUX_CMD_I2C_WRITE_STATUS_MOT    0x0600ul

#define TR_DPTX_AUX_BYTE_CT_MASK                0x000ful

//------------------------------------------------------------------------------
//  TR_DPTX_AUX_REPLY_CODE (offset 0x138)
//  Contains the reply code received from the most recent AUX Channel request.
//  This code is the two, two-bit fields received by the reply controller from the sink device.
//  The value of the code maps to the DisplayPort specification section 2.4.1.2 which
//  is repeated below for clarity.
//
//  1:0  AUX channel native reply codes received from the sink device.
//  0x0 = Native AUX ACK
//  0x1 = Native AUX NACK
//  0x2 = Native AUX Defer
//
//  3:2  AUX channel I2C reply codes received from the sink device.
//  0x0 = I2C over AUX ACK
//  0x4 = I2C over AUX NACK
//  0x8 = I2C over AUX Defer
//------------------------------------------------------------------------------
#define TR_DPTX_AUX_REPLY_NATIVE_MASK           0x03ul
#define TR_DPTX_AUX_REPLY_ACK                   0x00ul
#define TR_DPTX_AUX_REPLY_NACK                  0x01ul
#define TR_DPTX_AUX_REPLY_DEFER                 0x02ul

#define TR_DPTX_AUX_REPLY_I2C_MASK              0x0cul
#define TR_DPTX_AUX_REPLY_I2C_ACK               0x00ul
#define TR_DPTX_AUX_REPLY_I2C_NACK              0x04ul
#define TR_DPTX_AUX_REPLY_I2C_DEFER             0x08ul

//------------------------------------------------------------------------------
//  TR_DPTX_AUX_STATUS (offset 0x14c)
//
//  This register contains the status of the internal AUX channel controllers.
//  The progress of request and reply transactions are monitored and reply transactions
//  are checked for errors. These bits are always valid:
//
//  3   REPLY_ERROR: When set to a 1, the AUX reply logic has detected an error in
//      the reply to the most recent AUX transaction. Errors are detected when the
//      precharge and sync phases of the reply last more than 38 cycles instead of the maximum 32.
//      This condition typically indicates noise on the AUX channel data signals.
//
//  2   REQUEST_IN_PROGRESS: The AUX transaction request controller sets this bit to a 1
//      while actively transmitting a request on the AUX channel. The bit is set to 0 when
//      the AUX transaction request controller is idle.
//
//  1   REPLY_IN_PROGRESS: The AUX reply detection logic sets this bit to a 1 while
//      receiving a reply on the AUX channel. The bit is 0 otherwise.
//
//  0   REPLY_RECEIVED: This bit is set to 0 when the AUX request controller begins
//      sending bits on the AUX serial bus. The AUX reply controller sets this bit
//      to 1 when a complete and valid reply transaction has been received.
//      This bit is cleared when a request transaction has been initiated by the request controller.
//------------------------------------------------------------------------------
#define TR_DPTX_AUX_STATUS_REPLY_ERROR              (1ul<<3)
#define TR_DPTX_AUX_STATUS_REQUEST_IN_PROGRESS      (1ul<<2)
#define TR_DPTX_AUX_STATUS_REPLY_IN_PROGRESS        (1ul<<1)
#define TR_DPTX_AUX_STATUS_REPLY_RECEIVED           (1ul<<0)

//------------------------------------------------------------------------------
// Aux channel reply constants
//------------------------------------------------------------------------------
#define TR_DPTX_AUX_STATUS_RECEIVED 0x00ul
#define TR_DPTX_AUX_STATUS_TIMEOUT  0x01ul
#define TR_DPTX_AUX_STATUS_ERROR    0x02ul
#define TR_DPTX_AUX_STATUS_FAILSAFE 0x03ul

//------------------------------------------------------------------------------
//	TR_DPTX_INTERRUPT_STATE (offset 0x0130)
//
//	Holds status bits which report the internal state of the interrupt management
//	logic. These signals are used to generate interrupts and can be used as polled
//	status for systems that do not implement interrupts.
//
//	A read from this register clears all values.
//------------------------------------------------------------------------------
#define TR_DPTX_INTERRUPT_STATE_HDCP_TIMER_EVENT \
    (1ul << 5) //	indicates that an event has been generated by the HDCP timer block.
#define TR_DPTX_INTERRUPT_STATE_GP_TIMER_EVENT \
    (1ul << 4) //	indicates that an event has been generated by the general purpose timer block.
#define TR_DPTX_INTERRUPT_STATE_REPLY_TIMEOUT  (1ul << 3) //	indicates that a reply timeout has occurred
#define TR_DPTX_INTERRUPT_STATE_REPLY_RECEIVED (1ul << 2) //	indicates that a reply has been received
#define TR_DPTX_INTERRUPT_STATE_HPD_IRQ        (1ul << 1) //	indicates that an IRQ interrupt has been received.
#define TR_DPTX_INTERRUPT_STATE_HPD_EVENT \
    (1ul << 0) //	indicates that an HPD connect or disconnect event has been detected.

//------------------------------------------------------------------------------
//	TR_DPTX_INTERRUPT_CAUSE (offset 0x0140)
//
//	The transmitter core interrupt status register contains the cause of an
//	interrupt asserted by the core. The specific events that can cause an
//	interrupt and the associated status bits are shown below.
//
//	A read from this register clears all values.
//------------------------------------------------------------------------------
#define TR_DPTX_INTERRUPT_CAUSE_HDCP_TIMER_IRQ (1ul << 5) //	the HDCP internal timer has generated an interrupt.
#define TR_DPTX_INTERRUPT_CAUSE_GP_TIMER_IRQ \
    (1ul << 4) //	the internal general-purpose timer has generated an interrupt.
#define TR_DPTX_INTERRUPT_CAUSE_REPLY_TIMEOUT \
    (1ul << 3) //	a reply timeout has occurred when the sink has not sent a response 400us after the transmitter has
               // sent a request.
#define TR_DPTX_INTERRUPT_CAUSE_REPLY_RECIEVED \
    (1ul << 2) //	an AUX reply transaction has been detected. This value may be used to allow a system to process
               // other events while waiting for a response from the sink device.
#define TR_DPTX_INTERRUPT_CAUSE_HPD_IRQ \
    (1ul << 1) //	an IRQ framed with the proper timing on the HPD signal has been detected.
#define TR_DPTX_INTERRUPT_CAUSE_HPD_EVENT \
    (1ul << 0) //	the core has detected the presence of the HPD signal. This interrupt asserts immediately after the
               // detection of HPD and after the loss of HPD for 2 msec.

//------------------------------------------------------------------------------
//	TR_DPTX_INTERRUPT_MASK (offset 0x0144)
//
//	Each interrupt source for the transmitter core may be individually masked.
//	When the corresponding bit in this register is set to 1, the interrupt is
//	masked and will not generate an interrupt for the event.
//
//	At power up,  this register has a value of 0x3F.
//	That is, all interrupt sources are masked.
//------------------------------------------------------------------------------
#define TR_DPTX_INTERRUPT_MASK_HDCP_TIMER_IRQ_MASK \
    (1ul << 5) //	this control bit allows the internal general purpose timer to cause an interrupt to be asserted.
#define TR_DPTX_INTERRUPT_MASK_GP_TIMER_IRQ_MASK \
    (1ul << 4) //	this control bit allows the internal general purpose timer to cause an interrupt to be asserted.
#define TR_DPTX_INTERRUPT_MASK_REPLY_TIMEOUT_MASK \
    (1ul << 3) //	this bit to allow reply timeout events to cause an interrupt to be asserted.
#define TR_DPTX_INTERRUPT_MASK_REPLY_RECEIVED_MASK \
    (1ul << 2) //	this bit to allow reply received events to cause an interrupt.
#define TR_DPTX_INTERRUPT_MASK_HPD_IRQ_EVENT  (1ul << 1) //	this bit to allow HPD_IRQ events to cause an interrupt
#define TR_DPTX_INTERRUPT_MASK_HPD_EVENT_MASK (1ul << 0) //	this bit to allow HPD present events to cause an interrupt.

//------------------------------------------------------------------------------
// Timer block
//------------------------------------------------------------------------------
#define TR_DPTX_TIMER_MAX_COUNT 0x1FFFFFul

//------------------------------------------------------------------------------
// Timer Register
// bit 31     = Timer enable
// bit 30     = Auto-reload
// bit 29     = Interrupt enable
// bit 28 : 0 = Timer value
//------------------------------------------------------------------------------
#define TR_DPTX_TIMER_FLAG_ENABLE        0x80000000ul
#define TR_DPTX_TIMER_FLAG_AUTORELOAD_EN 0x40000000ul
#define TR_DPTX_TIMER_FLAG_INTERRUPT_EN  0x20000000ul
#define TR_DPTX_TIMER_FLAGS_ALL \
    (TR_DPTX_TIMER_FLAG_ENABLE | TR_DPTX_TIMER_FLAG_AUTORELOAD_EN | TR_DPTX_TIMER_FLAG_INTERRUPT_EN)

/*
 * DP PHY相关寄存器
 */

//------------------------------------------------------------------------------
// configuration for 10.8GHz VCO           25MHz   (refclk)
// ----------------------------------------------------------------------
//   PMA:CMN_PLL0_INTDIV_M0               16'h01b0
//   PMA:CMN_PLL0_FRACDIVL_M0             16'h0000
//   PMA:CMN_PLL0_HIGH_THR_M0             16'h0120
//
// configuration for 9.72GHz VCO           25MHz   (refclk)
// ----------------------------------------------------------------------
//   PMA:CMN_PLL0_INTDIV_M0               16'h0184
//   PMA:CMN_PLL0_FRACDIVL_M0             16'hcccd
//   PMA:CMN_PLL0_HIGH_THR_M0             16'h0104
//
// configuration for 8.1GHz VCO            25MHz   (refclk)
// ----------------------------------------------------------------------
//   PMA:CMN_PLL0_INTDIV_M0               16'h0144
//   PMA:CMN_PLL0_FRACDIVL_M0             16'h0000
//   PMA:CMN_PLL0_HIGH_THR_M0             16'h00d8
//
// common configuration between VCOs       25MHz   (refclk)
// ----------------------------------------------------------------------
//   PMA:CMN_SSM_BIAS_TMR                 16'h0019
//   PMA:CMN_PLLSM0_PLLPRE_TMR            16'h0032
//   PMA:CMN_PLLSM0_PLLLOCK_TMR           16'h00d1 (16'h00f0          16'h0104)
//   PMA:CMN_PLLSM1_PLLPRE_TMR            16'h0032
//   PMA:CMN_PLLSM1_PLLLOCK_TMR           16'h00d1 (16'h00f0          16'h0104)
//   PMA:CMN_BGCAL_INIT_TMR               16'h007d
//   PMA:CMN_BGCAL_ITER_TMR               16'h007d
//   PMA:CMN_IBCAL_INIT_TMR               16'h0019
//   PMA:CMN_IBCAL_ITER_TMR               16'h0007 (16'h0011          16'h0012)
//   PMA:CMN_TXPUCAL_INIT_TMR             16'h001e
//   PMA:CMN_TXPUCAL_ITER_TMR             16'h0006
//   PMA:CMN_TXPDCAL_INIT_TMR             16'h001e
//   PMA:CMN_TXPDCAL_ITER_TMR             16'h0006
//   PMA:CMN_RXCAL_INIT_TMR               16'h02ee
//   PMA:CMN_RXCAL_ITER_TMR               16'h0006
//   PMA:CMN_SD_CAL_INIT_TMR              16'h0006
//   PMA:CMN_SD_CAL_ITER_TMR              16'h0002
//   PMA:CMN_SD_CAL_REFTIM_START          16'h000e
//   PMA:CMN_SD_CAL_PLLCNT_START          16'h012b
//   PMA:CMN_PDIAG_PLL0_CP_PADJ_M0                 16'h0509
//   PMA:CMN_PDIAG_PLL0_CP_IADJ_M0                 16'h0f00
//   PMA:CMN_PDIAG_PLL0_FILT_PADJ_M0               16'h0f08
//   PMA:CMN_PDIAG_PLL0_CTRL_M0                    16'h0002
//   PMA:CMN_PLL0_FRACDIVH_M0                      16'h0002
//   PMA:TX_PSC_A0                                 16'h00fb
//   PMA:TX_PSC_A2                                 16'h04aa
//   PMA:TX_PSC_A3                                 16'h04aa
//   PMA:RX_PSC_A0                                 16'h0000
//   PMA:RX_PSC_A2                                 16'h0000
//   PMA:RX_PSC_A3                                 16'h0000
//   PMA:RX_PSC_CAL                                16'h0000
//   PMA:XCVR_DIAG_BIDI_CTRL                       16'h000f
//   PMA:RX_REE_GCSM1_CTRL                         16'h0000
//   PMA:RX_REE_GCSM2_CTRL                         16'h0000
//   PMA:RX_REE_PERGCSM_CTRL                       16'h0000
//   PMA:CMN_PLL0_DSM_DIAG_M0                      16'h0004
//   PMA:CMN_PLL0_VCOCAL_TCTRL                     16'h0003
//   PMA:CMN_PLL0_VCOCAL_PLLCNT_START     16'h0317
//   PMA:CMN_PLL0_LOCK_REFCNT_START       16'h00c7
//   PMA:CMN_PLL0_LOCK_PLLCNT_START       16'h00c7
//   PMA:CMN_PLL0_VCOCAL_INIT_TMR         16'h00fa
//   PMA:CMN_PLL0_VCOCAL_ITER_TMR         16'h0004
//   PMA:CMN_PLL0_VCOCAL_REFTIM_START     16'h0317
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Unique configuration by link rate (assumes a 25MHz reference clock):
//
// link rate  VCO Freq    CLK_SEL_M0    HSCLK_DIV
//------------------------------------------------------------------------------
//   1.62      9.72GHz    16'h0f01      16'h0002
//   2.43      9.72GHz    16'h0701      16'h0001
//   2.7      10.8 GHz    16'h0701      16'h0001
//   3.24      9.72GHz    16'h0b00      16'h0002
//   5.4      10.8 GHz    16'h0301      16'h0000
//   8.1       8.1 GHz    16'h0200      16'h0000
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//   register map
//------------------------------------------------------------------------------
#define TR_DPTX_CDB_ADDR                        0x0000
#define TR_DPTX_CDB_WDATA                       0x0004
#define TR_DPTX_CDB_SEL_WR                      0x0008  // cdb_select @ bit 1, cdb_write @ bit 0
#define TR_DPTX_CDB_READY                       0x000c  // read-only address... will readback as 1 when transaction is complete
#define TR_DPTX_CDB_RDATA                       0x0010  // read-only address
#define TR_DPTX_CDB_RSTN                        0x0014  // cdb_rstn at bit 0
#define TR_DPTX_CDB_ENABLE                      0x0018  // write a 1 to enable configured transaction (resets CDB_READY flag)
#define TR_DPTX_PHY_CONFIG_RSTN                 0x0040
#define TR_DPTX_PHY_CMN_READY                   0x0044  // read-only address
#define TR_DPTX_PHY_LANE_EN_BITS                0x0048
#define TR_DPTX_PHY_LANE_RSTN_BITS              0x004c
#define TR_DPTX_PHY_DATA_ENABLE                 0x0050
#define TR_DPTX_AUXHPD_ENABLE_BITS              0x0054
#define TR_DPTX_TX_ELEC_IDLE_BITS               0x0058
#define TR_DPTX_PHY_LANE_PWR_STATE_REQ          0x0060
#define TR_DPTX_PHY_LANE_PWR_STATE_ACK          0x0064  // read-only address
#define TR_DPTX_PHY_LANE_PLLCLK_ENABLE          0x0068
#define TR_DPTX_PHY_LANE_PLLCLK_EN_ACK          0x006c  // read-only address
#define TR_DPTX_PHY_POWERDOWN                   0x0070
#define TR_DPTX_PHY_TX_LINK_RATE                0x0078
#define TR_DPTX_AUXHPD_CONFIG                   0x008c  // non-zero static configuration inputs to AUX/HPD block
#define TR_DPTX_PHY_EMPHASIS0                   0x00a0
#define TR_DPTX_PHY_EMPHASIS1                   0x00a4
#define TR_DPTX_PHY_EMPHASIS2                   0x00a8
#define TR_DPTX_PHY_EMPHASIS3                   0x00ac
#define TR_DPTX_PHY_VMARGIN0                    0x00b0
#define TR_DPTX_PHY_VMARGIN1                    0x00b4
#define TR_DPTX_PHY_VMARGIN2                    0x00b8
#define TR_DPTX_PHY_VMARGIN3                    0x00bc
#define TR_DPTX_PHY_LOW_POWER_SWING             0x00c0
#define TR_DPTX_PHY_TEST_RDWR_ADDRESS           0x00f8  // test register for connection sanity check
#define TR_DPTX_PHY_TRILINEAR_ID_REV            0x00fc  // read-only address

#define TR_DPTX_PHY_XCVR0_SCANIN0               0x0300
#define TR_DPTX_PHY_XCVR0_SCANIN1               0x0304
#define TR_DPTX_PHY_XCVR1_SCANIN0               0x0308
#define TR_DPTX_PHY_XCVR1_SCANIN1               0x030c
#define TR_DPTX_PHY_XCVR2_SCANIN0               0x0310
#define TR_DPTX_PHY_XCVR2_SCANIN1               0x0314
#define TR_DPTX_PHY_XCVR3_SCANIN0               0x0318
#define TR_DPTX_PHY_XCVR3_SCANIN1               0x031c
#define TR_DPTX_PHY_XCVR0_SCANOUT0              0x0320  // read-only address
#define TR_DPTX_PHY_XCVR0_SCANOUT1              0x0324  // read-only address
#define TR_DPTX_PHY_XCVR1_SCANOUT0              0x0328  // read-only address
#define TR_DPTX_PHY_XCVR1_SCANOUT1              0x032c  // read-only address
#define TR_DPTX_PHY_XCVR2_SCANOUT0              0x0330  // read-only address
#define TR_DPTX_PHY_XCVR2_SCANOUT1              0x0334  // read-only address
#define TR_DPTX_PHY_XCVR3_SCANOUT0              0x0338  // read-only address
#define TR_DPTX_PHY_XCVR3_SCANOUT1              0x033c  // read-only address
#define TR_DPTX_PHY_DTB_DATA_BUS                0x0340  // read-only address

//------------------------------------------------------------------------------
//  Cadence PHY config & diagnostics bus
//------------------------------------------------------------------------------
#define kPHY_PLL_CFG                            0xc00e
#define kPHY_PMA_CMN_CTRL1                      0xe000
#define kPHY_PMA_CMN_CTRL2                      0xe001
#define kPHY_PMA_PLL_RAW_CTRL                   0xe003
#define kPHY_PMA_ISO_PLL_CTRL0                  0xe005
#define kPHY_PMA_ISO_PLL_CTRL1                  0xe006
#define kPHY_PMA_ISO_TX_CTRL                    0xe00f
#define kXCVR_DIAG_BIDI_CTRL                    0x40ea
#define kXCVR_DIAG_HSCLK_DIV                    0x40e7
#define kXCVR_DIAG_HSCLK_SEL                    0x40e6
#define kXCVR_DIAG_PLLDRC_CTL                   0x40e5
#define kTX_PSC_A0                              0x4100
#define kTX_PSC_A1                              0x4101
#define kTX_PSC_A2                              0x4102
#define kTX_PSC_A3                              0x4103
#define kRX_PSC_A0                              0x8000
#define kRX_PSC_A1                              0x8001
#define kRX_PSC_A2                              0x8002
#define kRX_PSC_A3                              0x8003
#define kRX_PSC_CAL                             0x8006
#define kRX_REE_GCSM1_CTRL                      0x8108
#define kRX_REE_GCSM2_CTRL                      0x8110
#define kRX_REE_PERGCSM_CTRL                    0x8118

//------------------------------------------------------------------------------
//  Cadence PHY Configuration & Debug Bus (cdb)
//------------------------------------------------------------------------------
#define kPMA_CMN_SSM_BIAS_TMR                   0x0022
#define kPMA_CMN_PLLSM0_PLLPRE_TMR              0x002a
#define kPMA_CMN_PLLSM0_PLLLOCK_TMR             0x002c
#define kPMA_CMN_PLLSM1_PLLPRE_TMR              0x0032
#define kPMA_CMN_PLLSM1_PLLLOCK_TMR             0x0034
#define kPHY_CMN_CDIAG_REFCLK_OVRD              0x004c
#define kPMA_CMN_BGCAL_INIT_TMR                 0x0064
#define kPMA_CMN_BGCAL_ITER_TMR                 0x0065
#define kPMA_CMN_IBCAL_INIT_TMR                 0x0074
#define kPMA_CMN_IBCAL_ITER_TMR                 0x0075
#define kPMA_CMN_TXPUCAL_INIT_TMR               0x0104
#define kPMA_CMN_TXPUCAL_ITER_TMR               0x0105
#define kPMA_CMN_TXPDCAL_INIT_TMR               0x010a
#define kPMA_CMN_TXPDCAL_ITER_TMR               0x010b
#define kPMA_CMN_RXCAL_INIT_TMR                 0x0114
#define kPMA_CMN_RXCAL_ITER_TMR                 0x0115
#define kPMA_CMN_SD_CAL_INIT_TMR                0x0124
#define kPMA_CMN_SD_CAL_ITER_TMR                0x0125
#define kPMA_CMN_SD_CAL_REFTIM_START            0x0126
#define kPMA_CMN_SD_CAL_PLLCNT_START            0x0128

#define kPLL0_VCOCAL_TCTRL                      0x0082
#define kPLL0_VCOCAL_INIT_TMR                   0x0084
#define kPLL0_VCOCAL_ITER_TMR                   0x0085
#define kPLL0_VCOCAL_REFTIM_START               0x0086
#define kPLL0_VCOCAL_PLLCNT_START               0x0088
#define kPLL0_INTDIV_M0                         0x0090
#define kPLL0_FRACDIVL_M0                       0x0091
#define kPLL0_FRACDIVH_M0                       0x0092
#define kPLL0_HIGH_THR_M0                       0x0093
#define kPLL0_DSM_DIAG_M0                       0x0094
#define kPLL0_DSM_FBH_OVRD_M0                   0x0095
#define kPLL0_DSM_FBL_OVRD_M0                   0x0096
#define kPLL0_LOCK_REFCNT_START                 0x009c
#define kPLL0_LOCK_PLLCNT_START                 0x009e
#define kPDIAG_PLL0_CTRL_M0                     0x01a0
#define kPDIAG_PLL0_CLK_SEL_M0                  0x01a1
#define kPDIAG_PLL0_CP_PADJ_M0                  0x01a4
#define kPDIAG_PLL0_CP_IADJ_M0                  0x01a5
#define kPDIAG_PLL0_FILT_PADJ_M0                0x01a6

#define kPLL1_VCOCAL_TCTRL                      0x00c2
#define kPLL1_VCOCAL_INIT_TMR                   0x00c4
#define kPLL1_VCOCAL_ITER_TMR                   0x00c5
#define kPLL1_VCOCAL_REFTIM_START               0x00c6
#define kPLL1_VCOCAL_PLLCNT_START               0x00c8
#define kPLL1_INTDIV_M0                         0x00d0
#define kPLL1_FRACDIVL_M0                       0x00d1
#define kPLL1_FRACDIVH_M0                       0x00d2
#define kPLL1_HIGH_THR_M0                       0x00d3
#define kPLL1_DSM_DIAG_M0                       0x00d4
#define kPLL1_DSM_FBH_OVRD_M0                   0x00d5
#define kPLL1_DSM_FBL_OVRD_M0                   0x00d6
#define kPLL1_LOCK_REFCNT_START                 0x00dc
#define kPLL1_LOCK_PLLCNT_START                 0x00de
#define kPDIAG_PLL1_CTRL_M0                     0x01c0
#define kPDIAG_PLL1_CLK_SEL_M0                  0x01c1
#define kPDIAG_PLL1_CP_PADJ_M0                  0x01c4
#define kPDIAG_PLL1_CP_IADJ_M0                  0x01c5
#define kPDIAG_PLL1_FILT_PADJ_M0                0x01c6

//------------------------------------------------------------------------------
//  PHY interrupt status register(cdb)
//------------------------------------------------------------------------------
#define kPHY_INTERRUPT_STATUS_LANE0             0xd012 //A read from this register clears a interrupt.
#define kPHY_INTERRUPT_STATUS_LANE1             0xd112 //A read from this register clears a interrupt.
#define kPHY_INTERRUPT_STATUS_LANE2             0xd212 //A read from this register clears a interrupt.
#define kPHY_INTERRUPT_STATUS_LANE3             0xd312 //A read from this register clears a interrupt.

//------------------------------------------------------------------------------
//  PMA_CMN_CTRL1 bit defines
//------------------------------------------------------------------------------
#define PMA_CMN_CTRL1_CMN_READY_BIT             (1ul<<0)

//------------------------------------------------------------------------------
//  PMA_CMN_CTRL2 bit defines
//------------------------------------------------------------------------------
#define PMA_CMN_CTRL2_PLL0_READY_BIT            (1ul<<0)
#define PMA_CMN_CTRL2_PLL1_READY_BIT            (1ul<<1)
#define PMA_CMN_CTRL2_PLL0_DISABLED_BIT         (1ul<<2)
#define PMA_CMN_CTRL2_PLL1_DISABLED_BIT         (1ul<<3)
#define PMA_CMN_CTRL2_PLL0_CLK_EN_ACK_BIT       (1ul<<4)
#define PMA_CMN_CTRL2_PLL1_CLK_EN_ACK_BIT       (1ul<<5)
#define PMA_CMN_CTRL2_PLL0_LOCKED_BIT           (1ul<<6)
#define PMA_CMN_CTRL2_PLL1_LOCKED_BIT           (1ul<<7)

//------------------------------------------------------------------------------
//  PMA_PLL_RAW_CTRL bit defines
//------------------------------------------------------------------------------
#define PMA_PLL_RAW_CTRL_PLL0_FORCE_RSTN_BIT     (1ul<<0)
#define PMA_PLL_RAW_CTRL_PLL1_FORCE_RSTN_BIT     (1ul<<1)

//------------------------------------------------------------------------------
//  Cadence PHY IP internal configuration addresses
//------------------------------------------------------------------------------
#define kLANE0_OFFSET                           0x0000
#define kLANE1_OFFSET                           0x0200
#define kLANE2_OFFSET                           0x0400
#define kLANE3_OFFSET                           0x0600

//------------------------------------------------------------------------------
//  power state defines
//------------------------------------------------------------------------------
#define DPTX_PHY_PWR_STATE_IDLE                 0x0000
#define DPTX_PHY_PWR_STATE_A0_TXRX_ACTIVE       0x0001
#define DPTX_PHY_PWR_STATE_A1_POWERDOWN1        0x0002  // note: state A1 is not a used power state for DisplayPort
#define DPTX_PHY_PWR_STATE_A2_POWERDOWN2        0x0004  // note: states A2 & A3 are configured identically for DisplayPort
#define DPTX_PHY_PWR_STATE_A3_POWERDOWN3        0x0008  // note: state A3 requires special handshaking for entry/exit

//------------------------------------------------------------------------------
//  driver typedefs
//------------------------------------------------------------------------------
typedef enum {
    dptx_phy_error_none,
    dptx_phy_error_core_not_present,
    dptx_phy_error_invalid_ref_clk,
    dptx_phy_error_invalid_link_clk,
    dptx_phy_error_bus_not_idle,
    dptx_phy_error_reset_type,
    dptx_phy_error_reset_ns,
    dptx_phy_error_link_rate,
    dptx_phy_error_lane_count,
    dptx_phy_error_pwr_type,
    dptx_phy_error_pwr_ack,
    dptx_phy_error_pwr_ns,
} dptx_phy_error_t;

typedef enum {
    dptx_ref_clk_24mhz,
    dptx_ref_clk_26mhz,
} dptx_ref_clk_t;

typedef enum {
    dptx_phy_reset_all,
    dptx_phy_reset_min,
} dptx_phy_reset_t;

typedef enum {
    dptx_lane_en_1 = 0x01,
    dptx_lane_en_2 = 0x03,
    dptx_lane_en_4 = 0x0f,
} dptx_lane_en_t;

typedef enum {
    dptx_power_a0,
    dptx_power_a1,
    dptx_power_a2,
    dptx_power_a3,
} dptx_power_state_t;

/*
 * dp接口结构体
 */
#ifndef DP_RECEIVER_CAP_SIZE
#define DP_RECEIVER_CAP_SIZE    0xf
#endif

#if !defined(UEFI_GOP) && !defined(UTIL_TOOL)
struct ljm_psr {
    struct mutex lock;
    bool sink_support;
    struct intel_dp *enabled;
    bool active;
    struct work_struct work;
    unsigned busy_frontbuffer_bits;
    bool sink_psr2_support;
    bool link_standby;
    bool colorimetry_support;
    bool alpm;
    bool psr2_enabled;
    u8 sink_sync_latency;
    bool debug;
    ktime_t last_entry_attempt;
    ktime_t last_exit;
};

struct ljm_dp_compliance_data {
    unsigned long edid;
    uint8_t video_pattern;
    uint16_t hdisplay, vdisplay;
    uint8_t bpc;
};

struct ljm_dp_compliance {
    unsigned long test_type;
    struct ljm_dp_compliance_data test_data;
    bool test_active;
    int test_link_rate;
    u8 test_lane_count;
};
#endif

struct ljm_dp_link_caps {
    bool enhanced_framing;
    bool tps3_supported;
    bool tps4_supported;
    bool fast_training;
    bool channel_coding;
    bool ssc;
};

struct ljm_dp_link_train_set {
    unsigned int voltage_swing[4];
    unsigned int pre_emphasis[4];
    bool voltage_max_reached[4];
    bool pre_max_reached[4];
};

struct ljm_dp_link_train {
    struct ljm_dp_link_train_set request;
    struct ljm_dp_link_train_set adjust;
    bool clock_recovered;
    bool channel_equalized;
};

struct ljm_dp_link {
    u8 dpcd[DP_RECEIVER_CAP_SIZE];
    unsigned char revision;
    unsigned int rate; //单位：十千赫兹（10kHz）
    unsigned int lanes;
    struct ljm_dp_link_caps caps;
    struct ljm_dp_link_train train;
    struct drm_dp_desc desc;
    u8 sink_count;
    u8 vsc_sdp_extension_for_colorimetry_supported;
};

/**
 * struct virtual_dpcd - 虚拟dpcd。记录特定显示屏已知的dpcd信息，在
 * aux通道失效时使用。
 * @addr: dpcd寄存器地址
 * @val: dpcd值
 *
 * NOTE: 此功能作为不使用aux通道时的替代。仅记录dpcd中固定不可变的寄存器值，
 * 只能保证aux读操作无异常，aux写操作无法支持。
 */
#define VDPCD_SIZE_MAX 128
#define VDPCD_ADDR_END  0xffffffff
struct virtual_dpcd
{
    //vdpcd数组以addr == 0xffffffff作为结束，这样程序可以无需遍历整个数组即可知道数组已遍历完。
    uint32_t addr;
    uint8_t val;
};

struct dp_dev
{
#ifdef UEFI_GOP
    GMI_PRIVATE_DATA *Private;  /* UEFI private data pointer */
#else
    void __iomem       *regbase;
    struct drm_device  *ddev;
    struct drm_connector *connector;
#endif

    unsigned int offset;        /* base offset */
    int index;

    bool is_edp;
    bool once_edp;

    struct ljm_dp_link link;

    bool mcu_aux; /*是否可支持使用mcu-aux*/
    uint8_t     gpio_direction;
    uint8_t     gpio_data;

    /*edp 背光：负值表示没有使用gpio控制背光和电源*/
    int gpio_backlight;
    int gpio_power;
    int pwm_ch;

#if !defined(UEFI_GOP) && !defined(UTIL_TOOL)
    struct delayed_work delayed_display_wk; /* Wait until the panel is stable, then display it. */

//    struct mutex lock;
    bool connected;
    bool suspended;

    u8 max_lanes;

    bool disable_aux;
    struct drm_dp_aux aux;
    struct drm_display_mode mode;

    uint8_t vedid[256];
    struct virtual_dpcd vdpcd[VDPCD_SIZE_MAX];

//---------- intel ---------------------------
    u8 dpcd[DP_RECEIVER_CAP_SIZE];

#endif

    /*for dp audio*/
    struct device *dev;
    struct mutex mutex; /* for state below and previous_mode */

    u8 mc_clkdis; /* clock disable register */
    spinlock_t audio_lock;
    struct mutex audio_mutex;
    bool audio_enable;

    unsigned int reg_shift;
    struct regmap *regm;

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0) && defined(ljdENABLE_AUDIO)
    hdmi_codec_plugged_cb plugged_cb;
#endif
    struct device *codec_dev;
    enum drm_connector_status last_connector_result;
};

#ifdef UEFI_GOP
#define DP_REG_W32(dp_dev, addr, val) WREG32_BASE((dp_dev)->Private, (dp_dev)->offset + (addr), (val))
#define DP_REG_R32(dp_dev, addr)      RREG32_BASE((dp_dev)->Private, (dp_dev)->offset + (addr))

#define CHECK_DP_DEV_VALID(dp_dev)  \
    if (NULL == (dp_dev) || (dp_dev)->Private == NULL || (dp_dev)->offset == NULL) \
    {   \
        ERR_PRINT("%s(%d): invalid dp_dev\n", __FUNCTION__, __LINE__);  \
        return -EINVAL; \
    }

#define CHECK_DP_DEV_VALID_VOID(dp_dev)  \
    if (NULL == (dp_dev) || (dp_dev)->Private == NULL || (dp_dev)->offset == NULL) \
    {   \
        ERR_PRINT("%s(%d): invalid dp_dev\n", __FUNCTION__, __LINE__);  \
        return; \
    }
#else /* !UEFI_GOP */

#define DEBUG_DUMP_REG_DP   0

#if DEBUG_DUMP_REG_DP
#define DP_REG_W32(dp_dev, addr, val) do {  \
        PRINT("DP %d: W[%x]=%08x", (dp_dev)->index, (dp_dev)->offset + (addr), (val));  \
        WREG32_BASE((dp_dev)->regbase, (dp_dev)->offset + (addr), (val));   \
    } while(0)
extern uint32_t dp_reg_r32(struct dp_dev *dp_dev, uint32_t addr);
#define DP_REG_R32(dp_dev, addr)      dp_reg_r32(dp_dev, addr)
#else
#define DP_REG_W32(dp_dev, addr, val) WREG32_BASE((dp_dev)->regbase, (dp_dev)->offset + (addr), (val))
#define DP_REG_R32(dp_dev, addr)      RREG32_BASE((dp_dev)->regbase, (dp_dev)->offset + (addr))
#endif

#define CHECK_DP_DEV_VALID(dp_dev)  \
    if (NULL == (dp_dev) || ((dp_dev)->regbase + (dp_dev)->offset) == NULL) \
    {   \
        ERR_PRINT("%s(%d): invalid dp_dev\n", __FUNCTION__, __LINE__);  \
        return -EINVAL; \
    }

#define CHECK_DP_DEV_VALID_VOID(dp_dev)  \
    if (NULL == (dp_dev) || ((dp_dev)->regbase + (dp_dev)->offset) == NULL) \
    {   \
        ERR_PRINT("%s(%d): invalid dp_dev\n", __FUNCTION__, __LINE__);  \
        return; \
    }
#endif

bool ljm_dptx_aux_ready(struct dp_dev *dp_dev);
int ljm_dptx_aux_read_fifo(struct dp_dev *dp_dev, uint8_t *read_buffer, uint32_t byte_ct);
void ljm_dptx_aux_write_fifo(struct dp_dev *dp_dev, uint8_t *write_buffer, uint32_t byte_ct);
void ljm_dptx_aux_send_request(struct dp_dev *dp_dev, uint32_t cmd_code, uint32_t address, uint32_t byte_ct,
                               bool addr_only, bool phy_wake);
uint32_t ljm_dptx_aux_wait_reply(struct dp_dev *dp_dev);
uint32_t ljm_dptx_aux_reply_code(struct dp_dev *dp_dev);

/* DP PHY interface */
uint32_t ljm_dptx_phy_reg_read(struct dp_dev *dp_dev, uint32_t reg_offset);
void ljm_dptx_phy_reg_write(struct dp_dev *dp_dev, uint32_t reg_offset, uint32_t data);

/* DP controller interface */
void ljm_dptx_reg_write(struct dp_dev *dp_dev, uint32_t reg_offset, uint32_t data);
uint32_t ljm_dptx_reg_read(struct dp_dev *dp_dev, uint32_t reg_offset);

/* DP cdb interface */
int ljm_cdb_write(struct dp_dev *dp_dev, uint16_t addr, uint16_t val);
int ljm_cdb_read(struct dp_dev *dp_dev, uint16_t addr, uint16_t *pval);

/* DP aux interface */
int ljm_dptx_aux_read(struct dp_dev *dp_dev, uint32_t addr, uint8_t *pval);
int ljm_dptx_aux_read_blk(struct dp_dev *dp_dev, uint32_t addr, uint8_t *buf, int size);
int ljm_dptx_aux_write(struct dp_dev *dp_dev, uint32_t addr, uint8_t val);

int ljm_dptx_train_link(struct dp_dev *dp_dev, uint32_t speed, uint32_t lane_count, uint32_t tp2_time, bool edp_port);
int ljm_dptx_train_link_no_aux(struct dp_dev *dp_dev, uint32_t bw_code, uint32_t lane_count, bool edp_port);

/* DP function interface */
int ljm_dptx_init(struct dp_dev *dp_dev, void *base, int offset, int index);
int ljm_dptx_mode_set(struct dp_dev *dp_dev, const struct drm_display_mode *mode);

/* DP state of hot-plug detect */
bool ljm_dptx_hpd_sense(struct dp_dev *dp_dev);

/* DP i2c interface for DDC */
int ljm_dptx_i2c_raw_write_req(void *hw_dev, unsigned char target, unsigned char data);
int ljm_dptx_i2c_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size);

dptx_phy_error_t ljm_dptx_phy_update_link_rate(struct dp_dev *dp_dev, uint32_t bw_code);
dptx_phy_error_t ljm_dptx_phy_power(struct dp_dev *dp_dev, dptx_power_state_t pwr_state);
dptx_phy_error_t ljm_dptx_phy_set_lane_count(struct dp_dev *dp_dev, uint32_t lanecount);

void ljm_dptx_tmr_wait(struct dp_dev *dp_dev, uint32_t us);
uint32_t ljm_dptx_set_tu_size(struct dp_dev *dp_dev, uint32_t pixel_clock, uint32_t bits_per_pixel);
uint32_t ljm_dptx_set_user_data_count(struct dp_dev *dp_dev, uint32_t hres, uint32_t bytes_per_pixel, uint32_t lane_count);
uint32_t ljm_dptx_set_link_rate(struct dp_dev *dp_dev, uint32_t bw_code);
void ljm_dptx_set_lane_count(struct dp_dev *dp_dev, uint32_t new_lc);
void ljm_dptx_set_voltage_swing(struct dp_dev *dp_dev, uint32_t lane0, uint32_t lane1, uint32_t lane2, uint32_t lane3);
void ljm_dptx_set_preemphasis(struct dp_dev *dp_dev, uint32_t lane0, uint32_t lane1, uint32_t lane2, uint32_t lane3);

int ljm_dptx_phy_irq_clear(struct dp_dev *dp_dev);

#endif /* ~__LJMICRO_DP_HW__ */
