/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/stat.h>
#include <linux/syscalls.h>

#include "ljmicro_device.h"
#include "ljmicro_vgav2.h"
#include "ljmicro_fan.h"
#include "gp200.h"

#define m_SAMP_CTL_MODE_MASK_201     0x1ffffe
#define v_SAMP_CTL_MODE_TEMP0_201    0xf0000
#define m_SAMP_CTL_MODE_MASK_202     0xffffffff
#define v_SAMP_CTL_MODE_TEMP_202     0x01e201e2

struct ljm_temp_to_fan{
    uint32_t temp_code;//读取的寄存器值
    int temperature;   //实际温度，单位为毫度
};

static struct ljm_temp_to_fan ljm_fan_temp_table[] =
{
    {0XAB, -10000},
    {0XBD, -5000},
    {0XCF, 0},
    {0XE1, 5000},
    {0XF4, 10000},
    {0X107, 15000},
    {0X11B, 20000},
    {0X12F, 25000},
    {0X143, 30000},
    {0X158, 35000},
    {0X16D, 40000},
    {0X183, 45000},
    {0X199, 50000},
    {0X1AF, 55000},
    {0X1C7, 60000},
    {0X1DF, 65000},
    {0X1F7, 70000},
    {0X210, 75000},
    {0X22A, 80000},
    {0X244, 85000},
    {0X25F, 90000},
    {0X27B, 95000},
    {0X297, 100000},
    {0X2B5, 105000},
    {0X2D3, 110000},
    {0X2F1, 115000},
    {0X311, 120000},
    {0X331, 125000},
    {0X352, 130000},
    {0X395, 140000},
    {0X3DA, 150000}
};
static const int ljm_fan_temp_table_count = sizeof(ljm_fan_temp_table) / sizeof(ljm_fan_temp_table[0]);

static const struct ljm_cooling_states default_cooling_states[] =
{
    {
        .min_state = 1,
        .max_state = (4 * LJM_THERMAL_MAX_STATE) / 10,
    },
    {
        .min_state = (4 * LJM_THERMAL_MAX_STATE) / 10,
        .max_state = LJM_THERMAL_MAX_STATE,
    },
    {
        .min_state = LJM_THERMAL_MAX_STATE,
        .max_state = LJM_THERMAL_MAX_STATE,
    },
};

/* thermal_trip 设置，在此定义 trip 范围 */
static const struct thermal_trip default_thermal_trips[] = {
    {   /* In range - 10-40% PWM */
        .type       = THERMAL_TRIP_ACTIVE,
        .temperature        = LJM_THERMAL_TEMP_NORM,
        .hysteresis         = LJM_THERMAL_HYST_TEMP,
    },
    {   /* In range - 40-100% PWM */
        .type       = THERMAL_TRIP_ACTIVE,
        .temperature        = LJM_THERMAL_TEMP_HIGH,
        .hysteresis         = LJM_THERMAL_HYST_TEMP,
    },
    {   /* Warning */
        .type       = THERMAL_TRIP_HOT,
        .temperature        = LJM_THERMAL_TEMP_HOT,
    },
};
#define LJM_THERMAL_NUM_TRIPS   ARRAY_SIZE(default_thermal_trips)
/* Make sure all trips are writable */
#define LJM_THERMAL_TRIP_MASK   (BIT(LJM_THERMAL_NUM_TRIPS) - 1)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
static struct thermal_zone_params ljm_thermal_params =
{
    .no_hwmon = false,
};
#endif

static inline uint32_t abs_val(uint32_t val1, uint32_t val2)
{
    return (val1 > val2) ? (val1 - val2) : (val2 - val1);
}

uint32_t find_closest(uint32_t target)
{
    int i;
    uint32_t min_diff = 0x12;
    uint32_t closet_value = ljm_fan_temp_table[0].temp_code;

    for (i = 0; i < ljm_fan_temp_table_count; i++)
    {
        uint32_t diff = abs_val(target, ljm_fan_temp_table[i].temp_code);
        if (diff < min_diff)
        {
            min_diff = diff;
            closet_value = ljm_fan_temp_table[i].temp_code;
        }
    }
    return closet_value;
}

/**
* @brief 设置风扇转速，通过将 state 写入对应寄存器实现风扇转速的设置
* @param struct ljm_fan *fan ：fan 设备结构体
         int state           ：需要设置风扇状态，范围为 1-10
* @return 空
*/
void ljm_set_pwm_fan(struct ljm_fan *fan, int state)
{
    uint8_t fan_high;
    uint8_t fan_low;

    uint32_t pwm;
    pwm = state * 100;

    fan_high = (pwm & 0xff00) >> 8;
    fan_low = (pwm & 0xff);

    /* 以下寄存器操作需要确保 VGAREG_WRITE_MISC_OUTPUT 寄存器 BIT0 位值为 1*/
    /* PWN FAN 二级分频寄存器 */
    CR_WR(fan->card, CRTC_PWM_FAN_FRE_DEVISION_SEC_LOW, 0xE8);
    CR_WR(fan->card, CRTC_PWM_FAN_FRE_DEVISION_SEC_HIGH, 0x3);
    /* PWM FAN 一级分频系数 */
    CR_WR(fan->card, CRTC_PWM_FAN_FRE_DEVISION_ONE, 0x01);

    /* PWM FAN占空比寄存器 */
    CR_WR(fan->card, CRTC_PWM_FAN_DUTYCYCLE_LOW, fan_low);
    CR_WR(fan->card, CRTC_PWM_FAN_DUTYCYCLE_HIGH, fan_high);
    /* PWM FAN 使能 */
    CR_WR(fan->card, CRTC_PWM_FAN_EN, 0X01);

    //fan->current_state = state;

    return;
}

static inline uint32_t ljm_state_to_duty(int state)
{
    return DIV_ROUND_CLOSEST(state * LJM_THERMAL_MAX_DUTY, LJM_THERMAL_MAX_STATE);
}

static inline int ljm_duty_to_state(uint32_t duty)
{
    return DIV_ROUND_CLOSEST(duty * LJM_THERMAL_MAX_STATE, LJM_THERMAL_MAX_DUTY);
}

static int fan_get_max_state(struct thermal_cooling_device *cdev, unsigned long *state)
{
    /* state 范围为 1 - 10 */
    *state = LJM_THERMAL_MAX_STATE;
    return 0;
}

static int fan_get_cur_state(struct thermal_cooling_device *cdev, unsigned long *state)
{
    struct ljm_fan *fan = (struct ljm_fan *)cdev->devdata;
    *state = fan->current_state;
    return 0;
}

static int fan_set_cur_state(struct thermal_cooling_device *cdev, unsigned long state)
{
    struct ljm_fan *fan = (struct ljm_fan *)cdev->devdata;

    if (fan->manual)
        return 0;

    state = (state > 10) ? 10 : state;

    if (strcmp(cdev->type, "ljm_fan") != 0)
        return -EINVAL;

    if (state == 0) /* 第一次调用时，state = 0 */
    {
        state = fan->current_state;
    }

    /* PWM FAN 使能 */
    ljm_set_pwm_fan(fan, state);
    fan->current_state = state;

    return 0;
}

static struct thermal_cooling_device_ops fan_cooling_ops = {
    .get_max_state = fan_get_max_state,
    .get_cur_state = fan_get_cur_state,
    .set_cur_state = fan_set_cur_state,
};

/* 读取 GPU 温度 */
static int read_gpu_temp(struct ljm_fan *fan)
{
    uint32_t temp_code;
    int i;
    int temperature = 0;

    int card = fan->card;
    struct pci_dev *pdev= fan->pdev;

    if (pdev->device == CHIP_GP202)
    {
        temp_code = LJM_PVT_R32(card, PVT_SAMP_CTL);
        if ((temp_code & m_SAMP_CTL_MODE_MASK_202) != v_SAMP_CTL_MODE_TEMP_202) {
            /* switch to TEMP0 mode */
            LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_TEMP_202);
            msleep(1100);   /* sleep 1.1s to get valid data */
        }
        temp_code = LJM_PVT_R32(card, PVT_DATA) & m_PVT_DATA_MASK;
    }
    else //GP201
    {
        temp_code = LJM_PVT_R32(card, PVT_SAMP_CTL);
        if ((temp_code & m_SAMP_CTL_MODE_MASK_201) != v_SAMP_CTL_MODE_TEMP0_201) {
            LJM_PVT_W32(card, PVT_SAMP_CTL, v_SAMP_CTL_MODE_TEMP0_201);
            msleep(1100);   /* sleep 1.1s to get valid data */
        }

        temp_code = LJM_PVT_R32(card, PVT_DATA) & m_PVT_DATA_MASK;
    }

    temp_code = find_closest(temp_code);

    for (i = 0; i < ljm_fan_temp_table_count; i++)
    {
        if (temp_code == ljm_fan_temp_table[i].temp_code)
            temperature = ljm_fan_temp_table[i].temperature;
    }

    return temperature;
}

static int ljm_gpu_get_temp(struct thermal_zone_device *tz, int *temp)
{
    struct ljm_fan *fan = (struct ljm_fan *)tz->devdata;
    *temp = read_gpu_temp(fan);
    fan->current_temp = *temp;

    return 0;
}

#if 0
/**
* @brief 调试接口，设置GPU模拟温度
* @param struct thermal_zone_device *tz ：热区设备，指GPU温度
         int temp                       ：需要设置的模拟温度
* @return 成功返回0，失败返回其他值
*/
static int ljm_set_emul_temp(struct thermal_zone_device *tz, int temp)
{
    struct ljm_fan *fan = (struct ljm_fan *)tz->devdata;
    if (!fan)
        return -ENODEV;
    fan->emul_temp = temp;
    tz->emul_temperature = temp;
    pr_info("setting emulated temperature to %d mC\n", temp);
    return 0;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
static int ljm_get_trip_type(struct thermal_zone_device *tzdev, int trip, enum thermal_trip_type *type)
{
    struct ljm_fan *fan = (struct ljm_fan *)tzdev->devdata;
    if (trip < 0 || trip >= LJM_THERMAL_NUM_TRIPS)
        return -EINVAL;

    *type = fan->trips[trip].type;
    return 0;
}

static int ljm_get_trip_temp(struct thermal_zone_device *tzdev, int trip, int *temp)
{
    struct ljm_fan *fan = (struct ljm_fan *)tzdev->devdata;
    if (trip < 0 || trip >= LJM_THERMAL_NUM_TRIPS)
        return -EINVAL;

    *temp = fan->trips[trip].temperature;
    return 0;
}

static int ljm_set_trip_temp(struct thermal_zone_device *tzdev, int trip, int temp)
{
    struct ljm_fan *fan = (struct ljm_fan *)tzdev->devdata;
    if (trip < 0 || trip >= LJM_THERMAL_NUM_TRIPS)
        return -EINVAL;

    fan->trips[trip].temperature = temp;
    return 0;
}

static int ljm_get_trip_hyst(struct thermal_zone_device *tzdev, int trip, int *hyst)
{
    struct ljm_fan *fan = (struct ljm_fan *)tzdev->devdata;
    *hyst = fan->trips[trip].hysteresis;
    return 0;
}
#endif
/**
* @brief fan 与 thermal_zone 绑定
* @param struct thermal_zone_device *tzdev   ：热区设备，指GPU温度
         struct thermal_cooling_device *cdev ：冷却设备，指 fan
* @return 成功返回0，失败返回其他值
*/
static int ljm_thermal_bind(struct thermal_zone_device *tzdev, struct thermal_cooling_device *cdev)
{
    int i, err;
    struct ljm_fan *fan = (struct ljm_fan *)tzdev->devdata;

    if (strcmp(cdev->type, "ljm_fan") == 0)
    {
        for (i = 0; i < LJM_THERMAL_NUM_TRIPS; i++)
        {
            const struct ljm_cooling_states *coolState = &fan->cooling_states[i];

            err = thermal_zone_bind_cooling_device(tzdev, i, cdev,
                                coolState->max_state,
                                coolState->min_state,
                                THERMAL_WEIGHT_DEFAULT);
            if (err < 0)
            {
                printk("Failed to bind cooling device to trip %d\n", i);
                return err;
            }
        }
    }
	return 0;
}

static int ljm_thermal_unbind(struct thermal_zone_device *tzdev, struct thermal_cooling_device *cdev)
{
    int i, err;

    if (strcmp(cdev->type, "ljm_fan") == 0)
    {
        for (i = 0; i < LJM_THERMAL_NUM_TRIPS; i++)
        {
            err = thermal_zone_unbind_cooling_device(tzdev, i, cdev);
            if (err < 0)
            {
                printk("Failed to unbind cooling device\n");
                return err;
            }
        }
    }
    return 0;
}

static struct thermal_zone_device_ops ljm_thermal_ops = {
    .bind = ljm_thermal_bind,
    .unbind = ljm_thermal_unbind,
    .get_temp = ljm_gpu_get_temp,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    .get_trip_type = ljm_get_trip_type,
    .get_trip_temp = ljm_get_trip_temp,
    .set_trip_temp = ljm_set_trip_temp,
    .get_trip_hyst = ljm_get_trip_hyst,
    //.set_emul_temp = ljm_set_emul_temp,
#endif
};

/* tool 工具通过此处调用对风扇转速进行手动调整 */
long ljmicro_ioctl_thermal(unsigned int cmd, unsigned long data)
{
    uint32_t pwm;
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    struct ljm_fan *fan = ljm_dev->fan;

    switch(cmd)
    {
        case LJM_IOCTL_SET_FAN_MANUAL:
            if (copy_from_user(&pwm, (int *)data, sizeof(pwm)))
            {
                return -EFAULT;
            }
            fan->manual = true;
            ljm_set_pwm_fan(fan, pwm);
            break;
        case LJM_IOCTL_CLEAR_MANUAL:
            fan->manual = false;
            ljm_set_pwm_fan(fan, fan->current_state);
            break;
        default:
            return -ENOTTY;
    }
    return 0;
}

/**
* @brief thermal初始化，在 /sys/class/thermal/ 目录下生成 cooling_device 和 thermal_zone 节点
* @param struct ljm_device *ljm_dev  ： ljm_dev 设备结构体
* @return 成功返回0，失败返回其他值
*/
int ljm_gpu_thermal_init(struct ljm_device *ljm_dev)
{
    int ret;
    struct ljm_fan *fan;

    fan = devm_kzalloc(ljm_dev->dev, sizeof(struct ljm_fan), GFP_KERNEL);
    if (!fan)
        return -ENOMEM;

    /* fan 结构体赋初值 */
    fan->card = ljm_dev->cur_card;
    fan->pdev = ljm_dev->pdev;
    fan->current_state = 2;
    fan->current_temp = 10000; /* 初始温度设为10 C*/
    fan->max_state = LJM_THERMAL_MAX_STATE;
    fan->polling_delay = LJM_THERMAL_POLL_DELAY;
    fan->manual = false;
    memcpy(fan->trips, default_thermal_trips, sizeof(fan->trips));
    memcpy(fan->cooling_states, default_cooling_states, sizeof(fan->cooling_states));

    fan->cdev = thermal_cooling_device_register("ljm_fan", fan, &fan_cooling_ops);
    if (IS_ERR(fan->cdev))
    {
        ret = PTR_ERR(fan->cdev);
        pr_err("Failed to register cooling device: %d\n", ret);
        return ret;
    }
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    fan->tzdev = thermal_zone_device_register("ljm_thermal_zone",
                                        LJM_THERMAL_NUM_TRIPS,
                                        LJM_THERMAL_TRIP_MASK,
                                        fan,
                                        &ljm_thermal_ops,
                                        NULL, 0,
                                        fan->polling_delay);
#else
    fan->tzdev = thermal_zone_device_register_with_trips("ljm_thermal_zone",
                                                    fan->trips,
                                                    LJM_THERMAL_NUM_TRIPS,
                                                    LJM_THERMAL_TRIP_MASK,
                                                    fan,
                                                    &ljm_thermal_ops,
                                                    &ljm_thermal_params, 0,
                                                    fan->polling_delay);
#endif

    if (IS_ERR(fan->tzdev))
    {
        ret = PTR_ERR(fan->tzdev);
        pr_err("Failed to register thermal zone: %d\n", ret);
        thermal_cooling_device_unregister(fan->cdev);
        return ret;
    }

    ljm_dev->fan = fan;

    printk("Thermal Fan Load!\n");
    return 0;
}

void ljm_gpu_thermal_exit(struct ljm_device *ljm_dev)
{
    if (ljm_dev->fan->cdev)
        thermal_cooling_device_unregister(ljm_dev->fan->cdev);

    if (ljm_dev->fan->tzdev)
        thermal_zone_device_unregister(ljm_dev->fan->tzdev);

    kfree(ljm_dev->fan);

    printk("Thermal Fan UnLoad!\n");
}

/* 提供给外部调用，通过调用此函数即可完成热管理策略的注册实现 */
int ljm_pwm_fan_init(struct ljm_device *ljm_dev)
{
    int ret;
    ret = ljm_gpu_thermal_init(ljm_dev);
    if (ret)
        return ret;

    return 0;
}

void ljm_pwm_fan_exit(struct ljm_device *ljm_dev)
{
    ljm_gpu_thermal_exit(ljm_dev);
}

/**
* @brief 在 pre_init 函数中先调用，仅调用一次，根据百分比percent 设置风扇初转速
* @param int card     :显卡序号
*        int percent  :风扇转速(范围为 0-100)
* @return 无
*/
void ljm_fan_pre_init(int card, int percent)
{
    uint16_t data;

    /* 调用该函数在 ljm_vgav2_init 函数之前，需要确保 VGAREG_WRITE_MISC_OUTPUT 寄存器 BIT0 位值为 1 */

    /* 写 VGAREG_WRITE_MISC_OUTPUT 寄存器为 0x1 时，地址取 0X3Dx*/
    REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, 0x1);
    /* PWN FAN 二级分频寄存器 */
    CR_WR(card, CRTC_PWM_FAN_FRE_DEVISION_SEC_LOW, 0xE8);
    CR_WR(card, CRTC_PWM_FAN_FRE_DEVISION_SEC_HIGH, 0x3);
    /* PWM FAN 一级分频系数 */
    CR_WR(card, CRTC_PWM_FAN_FRE_DEVISION_ONE, 0x01);

    percent = (percent == 0) ? 10 : percent;
    data = 10 * percent; /* 10 * 100 = 0x3E8 */
    /* PWM FAN占空比寄存器 */
    CR_WR(card, CRTC_PWM_FAN_DUTYCYCLE_LOW, (uint8_t)(data & 0xFF));
    CR_WR(card, CRTC_PWM_FAN_DUTYCYCLE_HIGH, (uint8_t)((data & 0xFF00) >> 8));
    /* PWM FAN 使能 */
    CR_WR(card, CRTC_PWM_FAN_EN, 0X01);

    REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, 0x0);
    return;
}
