/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_device.h"
#include "ljmicro_object.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

int ljm_gem_object_create(struct ljm_device *ljm_dev, unsigned long size, struct drm_gem_object **obj)
{
    struct ljm_gem_object *bo;
    int r;

    HEADER_ARG("ljm_dev=%p size=%lu", ljm_dev, size);

#ifdef LJMICRO_DRV_ONLY
    /* Maximum bo size is the unpinned gtt size since we use the gtt to
     * handle vram to system pool migrations.
     */
    if (size > ljm_dev->vram_free.vram_size) {
        DRM_DEBUG("Allocation size %ldMb bigger than %ldMb limit\n",
              size >> 20, ljm_dev->vram_free.vram_size >> 20);
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }
#endif

    r = ljm_bo_create(ljm_dev, size, &bo);
    if (r) {
        DRM_ERROR("Failed to allocate GEM object (%ld, %d)\n", size, r);
        FOOTER_ARG("r=%d", r);
        return r;
    }
    *obj = &bo->base;

#ifdef LJMICRO_DRV_ONLY
    /* add to gem objects */
    mutex_lock(&ljm_dev->gem.mutex);
    list_add_tail(&bo->list, &ljm_dev->gem.objects);
    mutex_unlock(&ljm_dev->gem.mutex);
#endif

    FOOTER_ARG("new *obj=%p(%lx)", *obj, (unsigned long)*obj);
    return 0;
}

unsigned long ljm_vram_get_gpu_start(unsigned long flags)
{
    if ((flags & LJMICRO_FAMILY_MASK) == CHIP_GP100) TRACE("GP100: %llx", VRAM_GP100_GPU_START);
    else if (flags & LJMICRO_IS_BUS32) TRACE("GPU is bus 32: %llx", VRAM_GP200_BUS32_GPU_START);
    else TRACE("GPU is bus 40: %llx", VRAM_GP200_BUS40_GPU_START);

    return (flags & LJMICRO_FAMILY_MASK) == CHIP_GP100 ? VRAM_GP100_GPU_START :
        ((flags & LJMICRO_IS_BUS32) ? VRAM_GP200_BUS32_GPU_START : VRAM_GP200_BUS40_GPU_START);
}

/* initialize GP mmap */
int
ljm_gem_init(struct ljm_device *ljm_dev)
{
#ifdef LJMICRO_DRV_ONLY
    struct ljm_mm_obj *mo= NULL;
#endif

    HEADER_ARG("ljm_dev=0x%p", ljm_dev);

    init_rwsem(&ljm_dev->exclusive_lock);

#ifdef LJMICRO_DRV_ONLY
    /* Initialize gem and exclusive_lock. */
    INIT_LIST_HEAD(&ljm_dev->gem.objects);
    mutex_init(&ljm_dev->gem.mutex);

    /* initialize vram free */
    INIT_LIST_HEAD(&ljm_dev->vram_free.list);
    mutex_init(&ljm_dev->vram_free.mutex);
    ljm_dev->vram_free.vram_start = ljm_vram_get_gpu_start(ljm_dev->flags);
    ljm_dev->vram_free.vram_size = ljm_dev->smem_len;
    TRACE("vram_start=%lx size=%lx", ljm_dev->vram_free.vram_start, ljm_dev->vram_free.vram_size);

    /* Create and init the first free memory buffer object, which will contain all the free VRAM. */
    mo = kzalloc(sizeof(struct ljm_mm_obj), GFP_KERNEL);
    if (!mo) {
        DRM_ERROR("Failed to init memory manager: out of memory.\n");
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    mo->offset = 0;
    mo->size = ljm_dev->vram_free.vram_size;
    INIT_LIST_HEAD(&mo->list);

    mutex_lock(&ljm_dev->vram_free.mutex);
    list_add_tail(&mo->list, &ljm_dev->vram_free.list);
    mutex_unlock(&ljm_dev->vram_free.mutex);
#endif

    FOOTER();
    return 0;
}

/* Finish GP memory manager. */
void
ljm_gem_fini(
    IN struct ljm_device *ljm_dev
    )
{
    HEADER_ARG("ljm_dev=0x%p", ljm_dev);

#ifdef LJMICRO_DRV_ONLY
    ljm_bo_force_delete(ljm_dev);
#endif

    FOOTER();
}

