/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_GPIO_H__
#define __LJMICRO_GPIO_H__

#include "ljmicro_family.h"

/* io_mux register group */
#define REG_GPIO_DATA_O           0x00
#define REG_GPIO_DATA_I           0x04
#define REG_GPIO_MASK             0x08
#define REG_GP201_GPIO_PAD_OEN    0x10
#define REG_GP202_GPIO_PAD_OEN    0x0C

#define LJM_IOMUX_R8(CARD, ADDR)  readb(g_mmio_bases[CARD] + ljm_gpu_hardware.reg_iomux_offset + (ADDR))
#define LJM_IOMUX_W8(CARD, ADDR, VAL)  writeb((u8)(VAL), g_mmio_bases[CARD] + ljm_gpu_hardware.reg_iomux_offset + (ADDR))

/**
 * @brief 设置 GPIO MASK 位值
 * @param int card      : 使用的 DEVICE ID
 * @param uint8_t gpiox : 需要设置的引脚号
 * @param int mode      : 为 0 时设置对应位为（无效）关闭，为 1 时设置对应位为（有效）开启
 * @return 空
*/
void ljm_gpio_set_mask(int card, uint8_t gpiox, int mode);

/**
 * @brief 设置 GPIO OEN 位值
 * @param int card      : 使用的 DEVICE ID
 * @param uint8_t gpiox : 需要设置的引脚号
 * @param int mode      : 为 0 时设置对应位为（无效）关闭，为 1 时设置对应位为（有效）开启
 * @return 空
*/
void ljm_gpio_set_oen(int card, uint8_t gpiox, int mode);

/**
 * @brief 设置 GPIO OUTPUT 位值
 * @param int card      : 使用的 DEVICE ID
 * @param uint8_t gpiox : 需要设置的引脚号
 * @param int mode      : 为 0 时设置对应位为（无效）关闭，为 1 时设置对应位为（有效）开启
 * @return 空
*/
void ljm_gpio_set_output(int card, uint8_t gpiox, int mode);

/**
 * @brief 获取 GPIO MASK 位值
 * @param int card    : 使用的 DEVICE ID
 * @return uint8_t    : 读取到的GPIO MASK寄存器状态值
 *           GP202 :  返回值为 00 时，寄存器实际值为 ff (GP202 Mask 寄存器为 ff 时禁止数据写入)
 *                    返回值为 ff 时，寄存器实际值为 00 (GP202 Mask 寄存器为 00 时允许数据写入)
 *           GP201 :  返回值为 00 时，寄存器实际值为 00 (GP201 Mask 寄存器为 00 时禁止数据写入)
 *                    返回值为 ff 时，寄存器实际值为 ff (GP201 Mask 寄存器为 ff 时允许数据写入)
 */
uint8_t ljm_gpio_get_mask(int card);

uint8_t ljm_gpio_get_oen(int card);
uint8_t ljm_gpio_get_input(int card);
uint8_t ljm_gpio_get_output(int card);

#endif /* __LJMICRO_GPIO_H__ */
