/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_hdmi.h
 *  版 本 号: V1.0
 *  作    者: 胡艳明
 *  生成日期: 2022年4月15日
 *
 *  功能描述: HDMI控制器硬件驱动对外API声明头文件
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 创建文件
 *
 ******************************************************************************/

#ifndef __LJMICRO_HDMI_H__
#define __LJMICRO_HDMI_H__

/*******************************************************************************
 *                                 标准库头文件                                *
 *******************************************************************************/

#include <sound/hdmi-codec.h>
#include <linux/regmap.h>
#include <linux/fb.h>
/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/

#include "ljmicro_base.h"

/******************************************************************************
 *                                     宏定义                                     *
 ******************************************************************************/

#define HDMI_VIDEO_NORMAL      0
#define HDMI_VIDEO_DMT         BIT(9)
#define HDMI_VIDEO_YUV420      BIT(10)
#define HDMI_VIDEO_DISCRETE_VR BIT(11)

#define HDMI_VIC_MASK  (0xFF)
#define HDMI_TYPE_MASK (0xFF << 8)
#define HDMI_MAX_ID    4

#define OUT_P888          0 // 24bit screen,connect to lcdc D0~D23
#define OUT_P666          1 // 18bit screen,connect to lcdc D0~D17
#define OUT_P565          2
#define OUT_S888x         4
#define OUT_CCIR656       6
#define OUT_S888          8
#define OUT_S888DUMY      12
#define OUT_YUV_420       14
#define OUT_P101010       15
#define OUT_YUV_420_10BIT 16
#define OUT_YUV_422       12
#define OUT_YUV_422_10BIT 17
#define OUT_P16BPP4       24
#define OUT_D888_P666     0x21 // 18bit screen,connect to lcdc D2~D7, D10~D15, D18~D23
#define OUT_D888_P565     0x22

/******************************************************************************
 *                                   数据类型定义                                   *
 ******************************************************************************/

/* HDMI video information code according CEA-861-F */
enum hdmi_video_information_code
{
    HDMI_640X480P_60HZ = 1,
    HDMI_720X480P_60HZ_4_3,
    HDMI_720X480P_60HZ_16_9,
    HDMI_1280X720P_60HZ,
    HDMI_1920X1080I_60HZ, /*5*/
    HDMI_720X480I_60HZ_4_3,
    HDMI_720X480I_60HZ_16_9,
    HDMI_720X240P_60HZ_4_3,
    HDMI_720X240P_60HZ_16_9,
    HDMI_2880X480I_60HZ_4_3, /*10*/
    HDMI_2880X480I_60HZ_16_9,
    HDMI_2880X240P_60HZ_4_3,
    HDMI_2880X240P_60HZ_16_9,
    HDMI_1440X480P_60HZ_4_3,
    HDMI_1440X480P_60HZ_16_9, /*15*/
    HDMI_1920X1080P_60HZ,
    HDMI_720X576P_50HZ_4_3,
    HDMI_720X576P_50HZ_16_9,
    HDMI_1280X720P_50HZ,
    HDMI_1920X1080I_50HZ, /*20*/
    HDMI_720X576I_50HZ_4_3,
    HDMI_720X576I_50HZ_16_9,
    HDMI_720X288P_50HZ_4_3,
    HDMI_720X288P_50HZ_16_9,
    HDMI_2880X576I_50HZ_4_3, /*25*/
    HDMI_2880X576I_50HZ_16_9,
    HDMI_2880X288P_50HZ_4_3,
    HDMI_2880X288P_50HZ_16_9,
    HDMI_1440X576P_50HZ_4_3,
    HDMI_1440X576P_50HZ_16_9, /*30*/
    HDMI_1920X1080P_50HZ,
    HDMI_1920X1080P_24HZ,
    HDMI_1920X1080P_25HZ,
    HDMI_1920X1080P_30HZ,
    HDMI_2880X480P_60HZ_4_3, /*35*/
    HDMI_2880X480P_60HZ_16_9,
    HDMI_2880X576P_50HZ_4_3,
    HDMI_2880X576P_50HZ_16_9,
    HDMI_1920X1080I_50HZ_1250, /* V Line 1250 total*/
    HDMI_1920X1080I_100HZ,     /*40*/
    HDMI_1280X720P_100HZ,
    HDMI_720X576P_100HZ_4_3,
    HDMI_720X576P_100HZ_16_9,
    HDMI_720X576I_100HZ_4_3,
    HDMI_720X576I_100HZ_16_9, /*45*/
    HDMI_1920X1080I_120HZ,
    HDMI_1280X720P_120HZ,
    HDMI_720X480P_120HZ_4_3,
    HDMI_720X480P_120HZ_16_9,
    HDMI_720X480I_120HZ_4_3, /*50*/
    HDMI_720X480I_120HZ_16_9,
    HDMI_720X576P_200HZ_4_3,
    HDMI_720X576P_200HZ_16_9,
    HDMI_720X576I_200HZ_4_3,
    HDMI_720X576I_200HZ_16_9, /*55*/
    HDMI_720X480P_240HZ_4_3,
    HDMI_720X480P_240HZ_16_9,
    HDMI_720X480I_240HZ_4_3,
    HDMI_720X480I_240HZ_16_9,
    HDMI_1280X720P_24HZ, /*60*/
    HDMI_1280X720P_25HZ,
    HDMI_1280X720P_30HZ,
    HDMI_1920X1080P_120HZ,
    HDMI_1920X1080P_100HZ,
    HDMI_1280X720P_24HZ_21_9, /*65*/
    HDMI_1280X720P_25HZ_21_9,
    HDMI_1280X720P_30HZ_21_9,
    HDMI_1280X720P_50HZ_21_9,
    HDMI_1280X720P_60HZ_21_9,
    HDMI_1280X720P_100HZ_21_9, /*70*/
    HDMI_1280X720P_120HZ_21_9,
    HDMI_1920X1080P_24HZ_21_9,
    HDMI_1920X1080P_25HZ_21_9,
    HDMI_1920X1080P_30HZ_21_9,
    HDMI_1920X1080P_50HZ_21_9, /*75*/
    HDMI_1920X1080P_60HZ_21_9,
    HDMI_1920X1080P_100HZ_21_9,
    HDMI_1920X1080P_120HZ_21_9,
    HDMI_1680X720P_24HZ,
    HDMI_1680X720P_25HZ, /*80*/
    HDMI_1680X720P_30HZ,
    HDMI_1680X720P_50HZ,
    HDMI_1680X720P_60HZ,
    HDMI_1680X720P_100HZ,
    HDMI_1680X720P_120HZ, /*85*/
    HDMI_2560X1080P_24HZ,
    HDMI_2560X1080P_25HZ,
    HDMI_2560X1080P_30HZ,
    HDMI_2560X1080P_50HZ,
    HDMI_2560X1080P_60HZ, /*90*/
    HDMI_2560X1080P_100HZ,
    HDMI_2560X1080P_120HZ,
    HDMI_3840X2160P_24HZ,
    HDMI_3840X2160P_25HZ,
    HDMI_3840X2160P_30HZ, /*95*/
    HDMI_3840X2160P_50HZ,
    HDMI_3840X2160P_60HZ,
    HDMI_4096X2160P_24HZ,
    HDMI_4096X2160P_25HZ,
    HDMI_4096X2160P_30HZ, /*100*/
    HDMI_4096X2160P_50HZ,
    HDMI_4096X2160P_60HZ,
    HDMI_3840X2160P_24HZ_21_9,
    HDMI_3840X2160P_25HZ_21_9,
    HDMI_3840X2160P_30HZ_21_9, /*105*/
    HDMI_3840X2160P_50HZ_21_9,
    HDMI_3840X2160P_60HZ_21_9,
};

/* HDMI Extended Resolution */
enum
{
    HDMI_VIC_4KX2K_30HZ = 1,
    HDMI_VIC_4KX2K_25HZ,
    HDMI_VIC_4KX2K_24HZ,
    HDMI_VIC_4KX2K_24HZ_SMPTE
};

/* HDMI Video Format */
enum
{
    HDMI_VIDEO_FORMAT_NORMAL = 0,
    HDMI_VIDEO_FORMAT_4KX2K,
    HDMI_VIDEO_FORMAT_3D,
};

/* HDMI 3D type */
enum
{
    HDMI_3D_NONE              = -1,
    HDMI_3D_FRAME_PACKING     = 0,
    HDMI_3D_TOP_BOOTOM        = 6,
    HDMI_3D_SIDE_BY_SIDE_HALF = 8,
};

/* HDMI Video Data Color Mode */
enum hdmi_video_color_mode
{
    HDMI_COLOR_AUTO = 0,
    HDMI_COLOR_RGB_0_255,
    HDMI_COLOR_RGB_16_235,
    HDMI_COLOR_YCBCR444,
    HDMI_COLOR_YCBCR422,
    HDMI_COLOR_YCBCR420
};

/* HDMI Video Data Color Depth */
enum hdmi_deep_color
{
    HDMI_DEPP_COLOR_AUTO   = 0,
    HDMI_DEEP_COLOR_Y444   = 0x1,
    HDMI_DEEP_COLOR_30BITS = 0x2,
    HDMI_DEEP_COLOR_36BITS = 0x4,
    HDMI_DEEP_COLOR_48BITS = 0x8,
};

#if 0
enum hdmi_colorimetry {
    HDMI_COLORIMETRY_NO_DATA = 0,
    HDMI_COLORIMETRY_SMTPE_170M,
    HDMI_COLORIMETRY_ITU709,
    HDMI_COLORIMETRY_EXTEND_XVYCC_601,
    HDMI_COLORIMETRY_EXTEND_XVYCC_709,
    HDMI_COLORIMETRY_EXTEND_SYCC_601,
    HDMI_COLORIMETRY_EXTEND_ADOBE_YCC601,
    HDMI_COLORIMETRY_EXTEND_ADOBE_RGB,
    HDMI_COLORIMETRY_EXTEND_BT_2020_YCC_C, /*constant luminance*/
    HDMI_COLORIMETRY_EXTEND_BT_2020_YCC,
    HDMI_COLORIMETRY_EXTEND_BT_2020_RGB,
};
#endif

/* HDMI Audio Type */
enum hdmi_audio_type
{
    HDMI_AUDIO_NLPCM = 0,
    HDMI_AUDIO_LPCM  = 1,
    HDMI_AUDIO_AC3,
    HDMI_AUDIO_MPEG1,
    HDMI_AUDIO_MP3,
    HDMI_AUDIO_MPEG2,
    HDMI_AUDIO_AAC_LC, /*AAC */
    HDMI_AUDIO_DTS,
    HDMI_AUDIO_ATARC,
    HDMI_AUDIO_DSD, /* One bit Audio */
    HDMI_AUDIO_E_AC3,
    HDMI_AUDIO_DTS_HD,
    HDMI_AUDIO_MLP,
    HDMI_AUDIO_DST,
    HDMI_AUDIO_WMA_PRO
};

/* HDMI Audio Sample Rate */
enum hdmi_audio_samplerate
{
    HDMI_AUDIO_FS_32000  = 0x1,
    HDMI_AUDIO_FS_44100  = 0x2,
    HDMI_AUDIO_FS_48000  = 0x4,
    HDMI_AUDIO_FS_88200  = 0x8,
    HDMI_AUDIO_FS_96000  = 0x10,
    HDMI_AUDIO_FS_176400 = 0x20,
    HDMI_AUDIO_FS_192000 = 0x40
};

/* HDMI Audio Word Length */
enum hdmi_audio_word_length
{
    HDMI_AUDIO_WORD_LENGTH_16bit = 0x1,
    HDMI_AUDIO_WORD_LENGTH_20bit = 0x2,
    HDMI_AUDIO_WORD_LENGTH_24bit = 0x4
};

/* HDMI Hotplug Status */
enum hdmi_hotpulg_status
{
    HDMI_HPD_REMOVED = 0, /* HDMI is disconnected */
    HDMI_HPD_INSERT,      /* HDMI is connected, but HDP is low
                           * or TMDS link is not pull up to 3.3V.
                           */
    HDMI_HPD_ACTIVATED    /* HDMI is connected, all singnal
                           * is normal
                           */
};

enum hdmi_mute_status
{
    HDMI_AV_UNMUTE  = 0,
    HDMI_VIDEO_MUTE = 0x1,
    HDMI_AUDIO_MUTE = 0x2,
};

/* HDMI Error Code */
enum hdmi_error_code
{
    HDMI_ERROR_SUCCESS = 0,
    HDMI_ERROR_FALSE,
    HDMI_ERROR_I2C,
    HDMI_ERROR_EDID,
};

enum hdmi_hdr_eotf
{
    EOTF_TRADITIONAL_GMMA_SDR = 1,
    EOTF_TRADITIONAL_GMMA_HDR = 2,
    EOTF_ST_2084              = 4,
};

struct hdmi_hdr_metadata
{
    u32 prim_x0;
    u32 prim_y0;
    u32 prim_x1;
    u32 prim_y1;
    u32 prim_x2;
    u32 prim_y2;
    u32 white_px;
    u32 white_py;
    u32 max_dml;
    u32 min_dml;
    u32 max_cll;  /*max content light level*/
    u32 max_fall; /*max frame-average light level*/
};

struct hdmi_hdr
{
    u8 eotf;
    u8 metadata; /*Staic Metadata Descriptor*/
    u8 maxluminance;
    u8 max_average_luminance;
    u8 minluminance;
};

#define HDMI_MAX_EDID_BLOCK 8

struct edid_prop_value
{
    int vid;
    int pid;
    int sn;
    int xres;
    int yres;
    int vic;
    int width;
    int height;
    int x_w;
    int x_h;
    int hwrotation;
    int einit;
    int vsync;
    int panel;
    int scan;
};

struct edid_prop_data
{
    struct edid_prop_value value;

    int valid;
    int last_vid;
    int last_pid;
    int last_sn;
    int last_xres;
    int last_yres;
};

enum vendor_code
{
    EDID_VENDOR_DEFAULT = 0,
    TCD0709,
};

/* HDMI EDID Information */
struct hdmi_edid
{
    unsigned char sink_hdmi;   /* HDMI display device flag */
    unsigned char ycbcr444;    /* Display device support YCbCr444 */
    unsigned char ycbcr422;    /* Display device support YCbCr422 */
    unsigned char ycbcr420;    /* Display device support YCbCr420 */
    unsigned char deepcolor;    /* bit3:DC_48bit; bit2:DC_36bit;
                                 * bit1:DC_30bit; bit0:DC_Y444;
                                 */
    unsigned char deepcolor_420;
    unsigned int  cecaddress;   /* CEC physical address */
    unsigned int maxtmdsclock; /* Max supported tmds clock */
    unsigned char fields_present;   /* bit7: latency
                                    * bit6: i_lantency
                                    * bit5: hdmi_video
                                    */
    unsigned char video_latency;
    unsigned char audio_latency;
    unsigned char interlaced_video_latency;
    unsigned char interlaced_audio_latency;
    /* for hdmi 2.0 */
    unsigned char hf_vsdb_version;
    unsigned char scdc_present;
    unsigned char rr_capable;
    unsigned char lte_340mcsc_scramble;
    unsigned char independent_view;
    unsigned char dual_view;
    unsigned char osd_disparity_3d;

    //  struct edid_prop_value value;

    unsigned int colorimetry;
    //  struct fb_monspecs  *specs; /*Device spec*/
    //  struct list_head modelist;  /*Device supported display mode list*/
    unsigned char baseaudio_support;
    //  struct hdmi_audio *audio;   /*Device supported audio info*/
    //  unsigned int  audio_num;    /*Device supported audio type number*/
    //
    //  unsigned int status;        /*EDID read status, success or failed*/
    //    u8 *raw[HDMI_MAX_EDID_BLOCK];   /*Raw EDID Data*/
    union {
        u8  data[5];
        struct hdmi_hdr hdrinfo;
    } hdr;

    enum vendor_code vendor;
};

/* HDMI Video Timing */
struct hdmi_video_timing
{
    struct fb_videomode mode; /* Video timing*/
    unsigned int        vic;  /* Video information code*/
    unsigned int        vic_2nd;
    unsigned int        pixelrepeat; /* Video pixel repeat rate*/
    unsigned int        interface;   /* Video input interface*/
};

/* HDMI Video Parameters */
struct hdmi_video
{
    unsigned int        vic; /* Video information code*/
    struct fb_videomode mode;

    unsigned int color_input;        /* Input video color mode*/
    unsigned int color_output;       /* Output video color mode*/
    unsigned int color_output_depth; /* Output video Color Depth*/
    unsigned int colorimetry;        /* Output Colorimetry */
    unsigned int sink_hdmi;          /* Output signal is DVI or HDMI*/
    unsigned int format_3d;          /* Output 3D mode*/
    unsigned int eotf;               /* EOTF */
};

/* HDMI Audio Parameters */
struct hdmi_audio
{
    u32 type;        /*Audio type*/
    u32 channel;     /*Audio channel number*/
    u32 rate;        /*Audio sampling rate*/
    u32 word_length; /*Audio data word length*/
};

struct hdmi_dev
{
#ifdef UEFI_GOP
    void *Private; /* UEFI private data pointer */
#else
    void __iomem *      regbase;
    struct drm_encoder *encoder;
#endif

    unsigned int  offset; /* base offset */
    int           index;
    bool    is_dvi;
    unsigned long pixelclk;
    unsigned int  tmdsclk;
    unsigned int  pixelrepeat;
    unsigned char colordepth;

    bool tmdsclk_ratio_change;

    struct hdmi_edid edid; /* EDID information*/

    int mode_3d;             /* HDMI output video 3d mode*/
    int eotf;                /* HDMI HDR EOTF */
                             //    struct hdmi_hdr_metadata hdr;   /* HDMI HDR MedeData */
    struct hdmi_audio audio; /* HDMI output audio information.*/
    struct hdmi_video video; /* HDMI output video information.*/

    const char *  vendor_name;
    const char *  product_name;
    unsigned char deviceinfo;

    struct device *dev;
    struct drm_connector connector;
    struct mutex mutex; /* for state below and previous_mode */

    u8 mc_clkdis; /* clock disable register */
    spinlock_t audio_lock;
    struct mutex audio_mutex;
    unsigned int sample_rate;
    unsigned int audio_cts;
    unsigned int audio_n;
    bool audio_enable;

    unsigned int reg_shift;
    struct regmap *regm;
    void (*enable_audio)(struct hdmi_dev *hdmi);
    void (*disable_audio)(struct hdmi_dev *hdmi);

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0) && defined(ljdENABLE_AUDIO)
    hdmi_codec_plugged_cb plugged_cb;
#endif
    struct device *codec_dev;
    enum drm_connector_status last_connector_result;

    /*scdc work*/
    bool need_scdc;
    int scdc_work_confirm_count;
    int dpms_state;
    bool scdc_work_run;
    struct mutex i2cm_mutex;
    struct delayed_work scdc_work;
};

/* HDMI EDID Block Size */
#define HDMI_EDID_BLOCK_SIZE 128

/* SCDC Registers */
#define SCDC_SINK_VER        0x01 /* sink version     */
#define SCDC_SOURCE_VER      0x02 /* source version   */
#define SCDC_UPDATE_0        0x10 /* Update_0     */
#define SCDC_UPDATE_1        0x11 /* Update_1     */
#define SCDC_UPDATE_RESERVED 0x12 /* 0x12-0x1f - Reserved */
#define SCDC_TMDS_CONFIG     0x20 /* TMDS_Config   */
#define SCDC_SCRAMBLER_STAT  0x21 /* Scrambler_Status   */
#define SCDC_CONFIG_0        0x30 /* Config_0           */
#define SCDC_CONFIG_RESERVED 0x31 /* 0x31-0x3f - Reserved */
#define SCDC_STATUS_FLAG_0   0x40 /* Status_Flag_0        */
#define SCDC_STATUS_FLAG_1   0x41 /* Status_Flag_1        */
#define SCDC_STATUS_RESERVED 0x42 /* 0x42-0x4f - Reserved */
#define SCDC_ERR_DET_0_L     0x50 /* Err_Det_0_L          */
#define SCDC_ERR_DET_0_H     0x51 /* Err_Det_0_H          */
#define SCDC_ERR_DET_1_L     0x52 /* Err_Det_1_L          */
#define SCDC_ERR_DET_1_H     0x53 /* Err_Det_1_H          */
#define SCDC_ERR_DET_2_L     0x54 /* Err_Det_2_L          */
#define SCDC_ERR_DET_2_H     0x55 /* Err_Det_2_H          */
#define SCDC_ERR_DET_CHKSUM  0x56 /* Err_Det_Checksum     */
#define SCDC_TEST_CFG_0      0xc0 /* Test_config_0        */
#define SCDC_TEST_RESERVED   0xc1 /* 0xc1-0xcf        */
#define SCDC_MAN_OUI_3RD \
    0xd0 /* Manufacturer IEEE OUI, \
          * Third Octet \
          */
#define SCDC_MAN_OUI_2ND \
    0xd1 /* Manufacturer IEEE OUI, \
          * Second Octet \
          */
#define SCDC_MAN_OUI_1ST \
    0xd2 /* Manufacturer IEEE OUI, \
          * First Octet \
          */
#define SCDC_DEVICE_ID    0xd3 /* 0xd3-0xdd - Device ID            */
#define SCDC_MAN_SPECIFIC 0xde /* 0xde-0xff - ManufacturerSpecific */

#ifdef UEFI_GOP
#define HDMIDBG(x, format, ...)

#define HDMI_REG_W8(hdmi_dev, addr, val) WREG8_BASE((hdmi_dev)->Private, (hdmi_dev)->offset + (addr), (val))
#define HDMI_REG_R8(hdmi_dev, addr)      RREG8_BASE((hdmi_dev)->Private, (hdmi_dev)->offset + (addr))
#else /* !UEFI_GOP */
extern int hdmi_dbg_level;
#define HDMIDBG(x, format, ...) \
    do \
    { \
        if (unlikely(hdmi_dbg_level >= x)) \
            pr_info(format, ##__VA_ARGS__); \
    } \
    while (0)

#define HDMI_REG_W8(hdmi_dev, addr, val) WREG8_BASE((hdmi_dev)->regbase, (hdmi_dev)->offset + (addr), (val))
#define HDMI_REG_R8(hdmi_dev, addr)      RREG8_BASE((hdmi_dev)->regbase, (hdmi_dev)->offset + (addr))
#endif

/******************************************************************************
 *                                    函数声明                                    *
 ******************************************************************************/

/*
 * GP201
 */
struct ljm_port_info;
void ljm_hdmiv2_irq_init(struct hdmi_dev *hdmi_dev);
int ljm_hdmiv2_init(struct hdmi_dev *hdmi_dev, void *base, int offset, struct ljm_port_info *info);

/* i2c */
int ljm_hdmiv2_write_phy(struct hdmi_dev *hdmi_dev, u8 addr, u16 data);
int ljm_hdmiv2_read_phy(struct hdmi_dev *hdmi_dev, u8 addr, u16 *pdata);
int ljm_hdmiv2_i2cm_write(struct hdmi_dev *hdmi_dev, u8 slave, u8 addr, u8 data);
int ljm_hdmiv2_i2cm_read(struct hdmi_dev *hdmi_dev, u8 slave, u8 addr, u8 *pdata);
int ljm_hdmiv2_i2cm_read_ext(struct hdmi_dev *hdmi_dev, u8 seg_ptr, u8 slave, u8 addr, u8 *pdata);

/* mode */
int          ljm_hdmiv2_mode_set(struct hdmi_dev *hdmi_dev, struct hdmi_video *video);
void         ljm_hdmiv2_dpms(struct hdmi_dev *hdmi_dev, int mode);
bool         ljm_hdmiv2_hpd_sense(struct hdmi_dev *hdmi_dev);
struct edid *ljm_hdmiv2_get_edid(struct hdmi_dev *hdmi_dev, int *psize);
void         ljm_hdmiv2_hpd_enable(struct hdmi_dev *hdmi_dev, int enable);

/* scdc */
int ljm_hdmiv2_scdc_read_update(struct hdmi_dev *hdmi_dev, u16 *pdata);

/* interrupt process */
int ljm_hdmiv2_irq_process(struct hdmi_dev *hdmi_dev);

/* set hpd polarity */
int ljm_hdmiv2_hpd_set_polarity(struct hdmi_dev *hdmi_dev);

/* check mode */
bool ljm_hdmiv2_mode_valid(struct hdmi_dev *hdmi_dev, const struct drm_display_mode *mode);
bool ljm_hdmiv2_mode_valid_ex(struct hdmi_dev *hdmi_dev, const struct drm_display_mode *mode);

/*parse edid*/
int ljm_hdmiv2_parse_edid(struct hdmi_dev *hdmi_dev, struct edid *edid);

/* Get monitor vendor and product id*/
enum vendor_code ljm_hdmiv2_edid_vendor(struct hdmi_dev *hdmi_dev, struct edid *edid);

/* scdc set */
void ljm_hdmiv2_scrambling_set(struct hdmi_dev *hdmi_dev);

/*
 * GP101
 */
int ljm_hdmiv1_init(struct hdmi_dev *hdmi_dev, void *base, int offset, int index);

int          ljm_hdmiv1_mode_set(struct hdmi_dev *hdmi_dev, const struct drm_display_mode *mode, bool is_hdmi);
void         ljm_hdmiv1_enable(struct hdmi_dev *hdmi_dev, bool enable);
void         ljm_hdmiv1_dpms(struct hdmi_dev *hdmi_dev, int mode);
bool         ljm_hdmiv1_hpd_sense(struct hdmi_dev *hdmi_dev);
struct edid *ljm_hdmiv1_get_edid(struct hdmi_dev *hdmi_dev, int *psize, void *connector);

/* interrupt process */
int ljm_hdmiv1_irq_process(struct hdmi_dev *hdmi_dev);

void ljm_hdmiv1_hpd_clear(struct hdmi_dev *hdmi_dev);
void ljm_hdmiv1_hpd_enable(struct hdmi_dev *hdmi_dev, int enable);
bool ljm_hdmiv1_is_enable(struct hdmi_dev *hdmi_dev);
void ljm_hdmiv1_int_vsync_set(struct hdmi_dev *hdmi_dev, int enable);
void ljm_hdmiv1_int_edid_set(struct hdmi_dev *hdmi_dev, int enable);

/* lcdc interface */
int                             hdmi_videomode_to_vic(struct fb_videomode *vmode);
const struct hdmi_video_timing *hdmi_vic2timing(int vic);
const struct fb_videomode *     hdmi_vic_to_videomode(int vic);
int                             hdmi_displaymode_to_vic(struct drm_display_mode *dmode);
void hdmi_displaymode_to_videomode(struct drm_display_mode *dmode, struct fb_videomode *vmode);

#endif /* __LJMICRO_HDMI_H__ */
