/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
*
*  版权所有(c)，2020-2099，中船第709研究所
*
*******************************************************************************
*
*  文件名称: ljmicro_hdmiv2.c
*  版 本 号: V1.0
*  作    者: 胡艳明
*  生成日期: 2022年4月21日
*
*  功能描述: HDMI V2控制器硬件设备驱动
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 创建文件
*
******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/
#include "ljmicro_hdmiv2.h"
#include "ljmicro_info.h"
/******************************************************************************
 *                                     宏定义                                     *
 ******************************************************************************/
#define _LJ_OBJ_ZONE ljvZONE_KMS
int hdmi_dbg_level = 2;
#define PHY_MPLL_TABLE_SIZE (sizeof(PHY_MPLL_TABLE) / sizeof(struct phy_mpll_config_tab))
module_param(hdmi_dbg_level, int, S_IRUGO | S_IWUSR);

/******************* only for fpga board, skip for asic **********************/
#ifdef HDMI_REG_W8
#undef HDMI_REG_W8
#endif

extern bool is_fpga;

#define HDMI_REG_W8(hdmi_dev, adr, val) \
    do { \
        if (is_fpga) \
            DUMP_WREG8(hdmi_dev->offset + (adr), (val)); \
        else \
            WREG8_BASE(hdmi_dev->regbase, hdmi_dev->offset + (adr), (val)); \
    } while (0)
/*****************************************************************************/

static inline int HDMI_REG_MSK(struct hdmi_dev *hdmi_dev, u16 offset, u8 msk, u8 val)
{
    u8 temp;

    temp = HDMI_REG_R8(hdmi_dev, offset) & (0xFF - (msk));
    HDMI_REG_W8(hdmi_dev, offset, (temp | ((val) & (msk))));
    return 0;
}

struct phy_mpll_config_tab {
    u32 pix_clock;
    u32 tmdsclock;
    u8 pix_repet;
    u8 color_depth;
    u8 prep_div;
    u8 cko_div;
    u8 opmode;
    u8 ref_cntrl;
    u8 n_cntrl;
    u8 gmp_cntrl;
    u8 prop_cntrl;
    u8 int_cntrl;
    u8 vco_cntrl;
    u8 multiplier;
};

static const struct phy_mpll_config_tab PHY_MPLL_TABLE[] = {
/*  tmdsclk = ?
 *  opmode: 0:HDMI1.4   1:HDMI2.0
 *
 *  |pixclock|tmdsclock|pixrepet|colordepth|
 *      prep_div|cko_div|opmode|ref_cntrl|n_cntrl|
 *      gmp_cntrl|prop_cntrl|int_cntrl|vco_ctrl|multiplier|
 */
    {13500000, 27000000, 1, 8,
        0, 0, 0, 0, 3,
        0, 10, 0, 3, 80},
    {13500000, 33750000, 1, 10,
        1, 0, 0, 0, 3,
        0, 10, 0, 3, 100},
    {13500000, 40500000, 1, 12,
        2, 0, 0, 0, 3,
        0, 11, 0, 1, 120},
    {13500000, 54000000, 1, 16,
        3, 0, 0, 0, 2,
        1, 10, 0, 3, 80},
    {13500000, 54000000, 3, 8,
        0, 0, 0, 0, 2,
        1, 10, 0, 3, 80},
    {13500000, 67500000, 3, 10,
        1, 0, 0, 0, 2,
        1, 10, 0, 3, 100},
    {13500000, 81000000, 3, 12,
        2, 0, 0, 0, 2,
        1, 11, 0, 1, 120},
    {13500000, 108000000, 3, 16,
        3, 0, 0, 0, 1,
        2, 10, 0, 3, 80},
    {13500000, 108000000, 7, 8,
        0, 0, 0, 0, 1,
        2, 10, 0, 3, 80},
    {13500000, 135000000, 7, 10,
        1, 0, 0, 0, 1,
        2, 10, 0, 3, 100},
    {13500000, 162000000, 7, 12,
        2, 0, 0, 0, 1,
        2, 11, 0, 1, 120},
    {13500000, 216000000, 7, 16,
        3, 0, 0, 0, 0,
        3, 10, 0, 3, 80},
    {18000000, 54000000, 2, 8,
        0, 0, 0, 0, 2,
        1, 10, 0, 3, 60},
    {18000000, 67500000, 2, 10,
        1, 0, 0, 0, 2,
        1, 10, 0, 3, 75},
    {18000000, 81000000, 2, 12,
        2, 0, 0, 0, 2,
        1, 11, 0, 1, 90},
    {18000000, 108000000, 2, 16,
        3, 0, 0, 0, 1,
        2, 10, 0, 3, 60},
    {18000000, 108000000, 5, 8,
        0, 0, 0, 0, 1,
        2, 10, 0, 3, 60},
    {18000000, 135000000, 5, 10,
        1, 0, 0, 0, 1,
        2, 10, 0, 3, 75},
    {18000000, 162000000, 5, 12,
        2, 0, 0, 0, 1,
        2, 11, 0, 1, 90},
    {18000000, 216000000, 5, 16,
        3, 0, 0, 0, 0,
        3, 10, 0, 3, 60},
    {21600000, 108000000, 4, 8,
        0, 0, 0, 0, 1,
        2, 10, 1, 3, 50},
    {21600000, 162000000, 4, 12,
        2, 0, 0, 0, 1,
        2, 11, 0, 1, 75},
    {21600000, 216000000, 4, 16,
        3, 0, 0, 0, 0,
        3, 10, 1, 3, 50},
    {21600000, 216000000, 9, 8,
        0, 0, 0, 0, 0,
        3, 10, 1, 3, 50},
    {21600000, 324000000, 9, 12,
        2, 0, 0, 0, 0,
        3, 11, 0, 1, 75},
    {21600000, 432000000, 9, 16,
        3, 3, 1, 0, 0,
        3, 11, 0, 1, 100},
    {24000000, 216000000, 8, 8,
        0, 0, 0, 0, 0,
        3, 10, 2, 3, 45},
    {24000000, 432000000, 8, 16,
        3, 3, 1, 0, 0,
        3, 11, 0, 1, 90},
    {25175000, 25175000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 3, 3, 40},
    {25175000, 31468750, 0, 10,
        1, 0, 0, 0, 3,
        0, 10, 1, 3, 50},
    {25175000, 37762500, 0, 12,
        2, 0, 0, 0, 3,
        0, 11, 2, 1, 60},
    {25175000, 50350000, 0, 16,
        3, 0, 0, 0, 2,
        1, 10, 3, 3, 40},
    {27000000, 27000000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 3, 3, 40},
    {27000000, 33750000, 0, 10,
        1, 0, 0, 0, 3,
        0, 10, 1, 3, 50},
    {27000000, 40500000, 0, 12,
        2, 0, 0, 0, 3,
        0, 11, 2, 1, 60},
    {27000000, 54000000, 0, 16,
        3, 0, 0, 0, 2,
        1, 10, 3, 3, 40},
    {27000000, 54000000, 1, 8,
        0, 0, 0, 0, 2,
        1, 10, 3, 3, 40},
    {27000000, 67500000, 1, 10,
        1, 0, 0, 0, 2,
        1, 10, 1, 3, 50},
    {27000000, 81000000, 1, 12,
        2, 0, 0, 0, 2,
        1, 11, 2, 1, 60},
    {27000000, 108000000, 1, 16,
        3, 0, 0, 0, 1,
        2, 10, 3, 3, 40},
    {27000000, 108000000, 3, 8,
        0, 0, 0, 0, 1,
        2, 10, 3, 3, 40},
    {27000000, 135000000, 3, 10,
        1, 0, 0, 0, 1,
        2, 10, 1, 3, 50},
    {27000000, 162000000, 3, 12,
        2, 0, 0, 0, 1,
        2, 11, 2, 1, 60},
    {27000000, 216000000, 3, 16,
        3, 0, 0, 0, 0,
        3, 10, 3, 3, 40},
    {27000000, 216000000, 7, 8,
        0, 0, 0, 0, 0,
        3, 10, 3, 3, 40},
    {27000000, 270000000, 7, 10,
        1, 0, 0, 0, 0,
        3, 10, 1, 3, 50},
    {27000000, 324000000, 7, 12,
        2, 0, 0, 0, 0,
        3, 11, 2, 1, 60},
    {27000000, 432000000, 7, 16,
        3, 3, 1, 0, 0,
        3, 11, 0, 1, 80},
    {31500000, 31500000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 3, 3, 40},
    {33750000, 33750000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 3, 3, 40},
    {35500000, 35500000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 3, 3, 40},
    {36000000, 36000000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 5, 1, 40},
    {36000000, 108000000, 2, 8,
        0, 0, 0, 0, 1,
        2, 10, 5, 3, 30},
    {36000000, 135000000, 2, 10,
        1, 0, 0, 1, 1,
        2, 10, 0, 3, 75},
    {36000000, 162000000, 2, 12,
        2, 0, 0, 0, 1,
        2, 11, 3, 1, 45},
    {36000000, 216000000, 2, 16,
        3, 0, 0, 0, 0,
        3, 10, 5, 3, 30},
    {36000000, 216000000, 5, 8,
        0, 0, 0, 0, 0,
        3, 10, 5, 3, 30},
    {36000000, 270000000, 5, 10,
        1, 0, 0, 1, 0,
        3, 10, 0, 3, 75},
    {36000000, 324000000, 5, 12,
        2, 0, 0, 0, 0,
        3, 11, 3, 1, 45},
    {36000000, 432000000, 5, 16,
        3, 3, 1, 0, 0,
        3, 11, 2, 1, 60},
    {40000000, 40000000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 5, 1, 40},
    {40500000, 81000000, 1, 8,
        0, 0, 0, 0, 2,
        1, 10, 5, 1, 40},
    {40500000, 101250000, 1, 10,
        1, 0, 0, 0, 1,
        2, 8, 3, 3, 25},
    {40500000, 121500000, 1, 12,
        2, 0, 0, 0, 1,
        2, 10, 5, 3, 30},
    {40500000, 162000000, 1, 16,
        3, 0, 0, 0, 1,
        2, 10, 5, 1, 40},
    {40500000, 162000000, 3, 8,
        0, 0, 0, 0, 1,
        2, 10, 5, 1, 40},
    {40500000, 202500000, 3, 10,
        1, 0, 0, 0, 0,
        3, 8, 3, 3, 25},
    {40500000, 243000000, 3, 12,
        2, 0, 0, 0, 0,
        3, 10, 5, 3, 30},
    {40500000, 324000000, 3, 16,
        3, 0, 0, 0, 0,
        3, 10, 5, 1, 40},
    {43200000, 216000000, 4, 8,
        0, 0, 0, 0, 0,
        3, 8, 3, 3, 25},
    {43200000, 324000000, 4, 12,
        2, 0, 0, 1, 0,
        3, 11, 0, 1, 75},
    {43200000, 432000000, 4, 16,
        3, 3, 1, 0, 0,
        3, 11, 2, 1, 50},
    {44900000, 44900000, 0, 8,
        0, 0, 0, 0, 3,
        0, 10, 5, 1, 40},
    {49500000, 49500000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {50000000, 50000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {50350000, 50350000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {50350000, 62937500, 0, 10,
        1, 0, 0, 0, 2,
        1, 8, 3, 3, 25},
    {50350000, 75525000, 0, 12,
        2, 0, 0, 0, 2,
        1, 8, 2, 1, 30},
    {50350000, 100700000, 0, 16,
        3, 0, 0, 0, 1,
        2, 6, 3, 3, 20},
    {54000000, 54000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {54000000, 67500000, 0, 10,
        1, 0, 0, 0, 2,
        1, 8, 3, 3, 25},
    {54000000, 81000000, 0, 12,
        2, 0, 0, 0, 2,
        1, 8, 2, 1, 30},
    {54000000, 108000000, 0, 16,
        3, 0, 0, 0, 1,
        2, 6, 3, 3, 20},
    {54000000, 108000000, 1, 8,
        0, 0, 0, 0, 1,
        2, 6, 3, 3, 20},
    {54000000, 135000000, 1, 10,
        1, 0, 0, 0, 1,
        2, 8, 3, 3, 25},
    {54000000, 162000000, 1, 12,
        2, 0, 0, 0, 1,
        2, 8, 2, 1, 30},
    {54000000, 216000000, 1, 16,
        3, 0, 0, 0, 0,
        3, 6, 3, 3, 20},
    {54000000, 162000000, 2, 8,
        0, 0, 0, 0, 1,
        2, 8, 2, 1, 30},
    {54000000, 202500000, 2, 10,
        1, 0, 0, 3, 0,
        3, 10, 0, 3, 75},
    {54000000, 243000000, 2, 12,
        2, 0, 0, 1, 0,
        3, 10, 2, 3, 45},
    {54000000, 324000000, 2, 16,
        3, 0, 0, 0, 0,
        3, 8, 2, 1, 30},
    {54000000, 216000000, 3, 8,
        0, 0, 0, 0, 0,
        3, 6, 3, 3, 20},
    {54000000, 270000000, 3, 10,
        1, 0, 0, 3, 0,
        3, 10, 0, 3, 100},
    {54000000, 324000000, 3, 12,
        2, 0, 0, 0, 0,
        3, 8, 2, 1, 30},
    {54000000, 432000000, 3, 16,
        3, 3, 1, 0, 0,
        3, 10, 5, 1, 40},
    {56250000, 56250000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {59400000, 59400000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {59400000, 74250000, 0, 10,
        1, 0, 0, 0, 2,
        1, 6, 2, 1, 25},
    {59400000, 89100000, 0, 12,
        2, 0, 0, 0, 2,
        1, 8, 2, 1, 30},
    {59400000, 118800000, 0, 16,
        3, 0, 0, 0, 1,
        2, 6, 3, 3, 20},
    {65000000, 65000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {67500000, 135000000, 1, 8,
        0, 0, 0, 0, 1,
        2, 6, 3, 3, 20},
    {67500000, 168750000, 1, 10,
        1, 0, 0, 0, 1,
        2, 6, 2, 1, 25},
    {67500000, 202500000, 1, 12,
        2, 0, 0, 0, 0,
        3, 5, 1, 3, 15},
    {67500000, 270000000, 1, 16,
        3, 0, 0, 0, 0,
        3, 6, 3, 3, 20},
    {67500000, 270000000, 3, 8,
        0, 0, 0, 0, 0,
        3, 6, 3, 3, 20},
    {67500000, 337500000, 3, 10,
        1, 0, 0, 0, 0,
        3, 6, 2, 1, 25},
    {67500000, 405000000, 3, 12,
        2, 3, 1, 0, 0,
        3, 8, 2, 1, 30},
    {67500000, 540000000, 3, 16,
        3, 3, 1, 0, 0,
        3, 9, 0, 0, 40},
    {68250000, 68250000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {71000000, 71000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 6, 3, 3, 20},
    {72000000, 72000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {72000000, 90000000, 0, 10,
        1, 0, 0, 0, 2,
        1, 6, 2, 1, 25},
    {72000000, 108000000, 0, 12,
        2, 0, 0, 0, 1,
        2, 5, 1, 3, 15},
    {72000000, 144000000, 0, 16,
        3, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {72000000, 216000000, 2, 8,
        0, 0, 0, 0, 0,
        3, 5, 1, 3, 15},
    {72000000, 270000000, 2, 10,
        1, 0, 0, 3, 0,
        3, 10, 0, 3, 75},
    {72000000, 324000000, 2, 12,
        2, 0, 0, 1, 0,
        3, 11, 3, 1, 45},
    {72000000, 432000000, 2, 16,
        3, 3, 1, 0, 0,
        3, 8, 2, 1, 30},
    {73250000, 73250000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {74250000, 74250000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {74250000, 92812500, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {74250000, 111375000, 0, 12,
        2, 0, 0, 0, 1,
        2, 5, 1, 3, 15},
    {74250000, 148500000, 0, 16,
        3, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {75000000, 75000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {75525000, 75525000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {75525000, 94406250, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {75525000, 113287500, 0, 12,
        2, 0, 0, 0, 1,
        2, 5, 1, 3, 15},
    {75525000, 151050000, 0, 16,
        3, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {78750000, 78750000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {79500000, 79500000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {81000000, 81000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {81000000, 101250000, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {81000000, 121500000, 0, 12,
        2, 0, 0, 0, 1,
        2, 5, 1, 3, 15},
    {81000000, 162000000, 0, 16,
        3, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {81000000, 162000000, 1, 8,
        0, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {81000000, 202500000, 1, 10,
        1, 0, 0, 1, 0,
        3, 8, 3, 3, 25},
    {81000000, 243000000, 1, 12,
        2, 0, 0, 0, 0,
        3, 5, 1, 3, 15},
    {81000000, 324000000, 1, 16,
        3, 0, 0, 0, 0,
        3, 5, 2, 1, 20},
    {81000000, 324000000, 3, 8,
        0, 0, 0, 0, 0,
        3, 5, 2, 1, 20},
    {81000000, 405000000, 3, 10,
        1, 3, 1, 0, 0,
        3, 6, 2, 1, 25},
    {81000000, 486000000, 3, 12,
        2, 3, 1, 0, 0,
        3, 7, 0, 0, 30},
    {82500000, 82500000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {82500000, 103125000, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {82500000, 123750000, 0, 12,
        2, 0, 0, 0, 1,
        2, 5, 1, 3, 15},
    {82500000, 165000000, 0, 16,
        3, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {83500000, 83500000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {85500000, 85500000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {88750000, 88750000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {90000000, 90000000, 0, 8,
        0, 0, 0, 0, 2,
        1, 5, 2, 1, 20},
    {90000000, 112500000, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {90000000, 135000000, 0, 12,
        2, 0, 0, 0, 1,
        2, 5, 1, 3, 15},
    {90000000, 180000000, 0, 16,
        3, 0, 0, 0, 1,
        2, 5, 2, 1, 20},
    {90000000, 270000000, 2, 8,
        0, 0, 0, 0, 0,
        3, 5, 1, 3, 15},
    {90000000, 337500000, 2, 10,
        1, 0, 0, 3, 0,
        3, 11, 0, 1, 75},
    {90000000, 405000000, 2, 12,
        2, 3, 1, 1, 0,
        3, 10, 5, 1, 45},
    {90000000, 540000000, 2, 16,
        3, 3, 1, 0, 0,
        3, 7, 0, 0, 30},
    {94500000, 94500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {94500000, 189000000, 1, 8,
        0, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {94500000, 236250000, 1, 10,
        1, 0, 0, 1, 0,
        3, 8, 3, 3, 25},
    {94500000, 283500000, 1, 12,
        2, 0, 0, 0, 0,
        3, 5, 1, 3, 15},
    {94500000, 378000000, 1, 16,
        3, 3, 1, 0, 0,
        3, 5, 2, 1, 20},
    {94500000, 378000000, 3, 8,
        0, 3, 1, 0, 0,
        3, 5, 2, 1, 20},
    {94500000, 472500000, 3, 10,
        1, 3, 1, 0, 0,
        3, 7, 0, 0, 25},
    {94500000, 567000000, 3, 12,
        2, 3, 1, 0, 0,
        3, 7, 0, 0, 30},
    {99000000, 99000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {99000000, 123750000, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {99000000, 148500000, 0, 12,
        2, 0, 0, 0, 1,
        2, 4, 0, 1, 15},
    {99000000, 198000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {100700000, 100700000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {100700000, 125875000, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {100700000, 151050000, 0, 12,
        2, 0, 0, 0, 1,
        2, 4, 0, 1, 15},
    {100700000, 201400000, 0, 16,
        3, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {101000000, 101000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {102250000, 102250000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {106500000, 106500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {108000000, 108000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {108000000, 135000000, 0, 10,
        1, 0, 0, 1, 1,
        2, 8, 3, 3, 25},
    {108000000, 162000000, 0, 12,
        2, 0, 0, 0, 1,
        2, 4, 0, 1, 15},
    {108000000, 216000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {108000000, 216000000, 1, 8,
        0, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {108000000, 270000000, 1, 10,
        1, 0, 0, 3, 0,
        3, 10, 1, 3, 50},
    {108000000, 324000000, 1, 12,
        2, 0, 0, 0, 0,
        3, 4, 0, 1, 15},
    {108000000, 432000000, 1, 16,
        3, 3, 1, 0, 0,
        3, 5, 2, 1, 20},
    {108000000, 432000000, 2, 8,
        0, 3, 1, 0, 0,
        3, 5, 2, 1, 20},
    {108000000, 540000000, 2, 10,
        1, 3, 1, 0, 0,
        3, 7, 0, 0, 25},
    {108000000, 432000000, 3, 8,
        0, 3, 1, 0, 0,
        3, 5, 2, 1, 20},
    {108000000, 540000000, 3, 10,
        1, 3, 1, 0, 0,
        3, 7, 0, 0, 25},
    {115500000, 115500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {117500000, 117500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {118800000, 118800000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {118800000, 148500000, 0, 10,
        1, 0, 0, 1, 1,
        2, 6, 2, 1, 25},
    {118800000, 178200000, 0, 12,
        2, 0, 0, 0, 1,
        2, 4, 0, 1, 15},
    {118800000, 237600000, 0, 16,
        3, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {119000000, 119000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {121500000, 243000000, 1, 8,
        0, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {121500000, 303750000, 1, 10,
        1, 0, 0, 1, 0,
        3, 6, 2, 1, 25},
    {121500000, 364500000, 1, 12,
        2, 3, 1, 0, 0,
        3, 4, 0, 1, 15},
    {121500000, 486000000, 1, 16,
        3, 3, 1, 0, 0,
        3, 5, 0, 0, 20},
    {121500000, 486000000, 3, 8,
        0, 3, 1, 0, 0,
        3, 5, 0, 0, 20},
    {121750000, 121750000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {122500000, 122500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {125875000, 125875000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {125875000, 157343750, 0, 10,
        1, 0, 0, 1, 1,
        2, 6, 2, 1, 25},
    {125875000, 188812500, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {125875000, 251750000, 0, 16,
        3, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {126000000, 378000000, 2, 8,
        0, 3, 1, 0, 0,
        3, 4, 0, 1, 15},
    {126000000, 472500000, 2, 10,
        1, 3, 1, 3, 0,
        3, 11, 0, 0, 75},
    {126000000, 567000000, 2, 12,
        2, 3, 1, 1, 0,
        3, 9, 0, 0, 45},
    {135000000, 135000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {135000000, 168750000, 0, 10,
        1, 0, 0, 1, 1,
        2, 6, 2, 1, 25},
    {135000000, 202500000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {135000000, 270000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {135000000, 270000000, 1, 8,
        0, 0, 0, 0, 0,
        3, 3, 0, 3, 10},
    {135000000, 337500000, 1, 10,
        1, 0, 0, 1, 0,
        3, 6, 2, 1, 25},
    {135000000, 405000000, 1, 12,
        2, 3, 1, 0, 0,
        3, 4, 0, 1, 15},
    {135000000, 540000000, 1, 16,
        3, 3, 1, 0, 0,
        3, 5, 0, 0, 20},
    {135000000, 540000000, 3, 8,
        0, 3, 1, 0, 0,
        3, 5, 0, 0, 20},
    {136750000, 136750000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {140250000, 140250000, 0, 8,
        0, 0, 0, 0, 1,
        2, 3, 0, 3, 10},
    {144000000, 144000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {144000000, 180000000, 0, 10,
        1, 0, 0, 1, 1,
        2, 6, 2, 1, 25},
    {144000000, 216000000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {144000000, 288000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {144000000, 432000000, 2, 8,
        0, 3, 1, 0, 0,
        3, 4, 0, 1, 15},
    {144000000, 540000000, 2, 10,
        1, 3, 1, 3, 0,
        3, 11, 0, 0, 75},
    {146250000, 146250000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {148250000, 148250000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {148500000, 148500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {148500000, 185625000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {148500000, 222750000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {148500000, 297000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {148500000, 297000000, 1, 8,
        0, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {148500000, 371250000, 1, 10,
        1, 3, 1, 1, 0,
        3, 6, 2, 1, 25},
    {148500000, 445500000, 1, 12,
        2, 3, 1, 0, 0,
        3, 4, 0, 1, 15},
    {148500000, 594000000, 1, 16,
        3, 3, 1, 0, 0,
        3, 5, 0, 0, 20},
    {148500000, 594000000, 3, 8,
        0, 3, 1, 0, 0,
        3, 5, 0, 0, 20},
    {151050000, 151050000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {151050000, 188812500, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {151050000, 226575000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {151050000, 302100000, 0, 16,
        3, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {154000000, 154000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {156000000, 156000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {157000000, 157000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {157500000, 157500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {162000000, 162000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {162000000, 202500000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {162000000, 243000000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {162000000, 324000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {162000000, 324000000, 1, 8,
        0, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {162000000, 405000000, 1, 10,
        1, 3, 1, 1, 0,
        3, 6, 2, 1, 25},
    {162000000, 486000000, 1, 12,
        2, 3, 1, 0, 0,
        3, 4, 0, 0, 15},
    {162000000, 486000000, 2, 8,
        0, 3, 1, 0, 0,
        3, 4, 0, 0, 15},
    {165000000, 165000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {165000000, 206250000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {165000000, 247500000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {165000000, 330000000, 0, 16,
        3, 0, 0, 0, 0,
        3, 2, 0, 1, 10},
    {175500000, 175500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {175500000, 351000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {175500000, 438750000, 1, 10,
        1, 3, 1, 1, 0,
        3, 6, 2, 1, 25},
    {175500000, 526500000, 1, 12,
        2, 3, 1, 0, 0,
        3, 4, 0, 0, 15},
    {176225000, 176225000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {176225000, 220281250, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {176225000, 264337500, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {176225000, 352450000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {178200000, 178200000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {178200000, 222750000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {178200000, 267300000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {178200000, 356400000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {179500000, 179500000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {180000000, 180000000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {180000000, 225000000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {180000000, 270000000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {180000000, 360000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {180000000, 540000000, 2, 8,
        0, 3, 1, 0, 0,
        3, 4, 0, 0, 15},
    {182750000, 182750000, 0, 8,
        0, 0, 0, 0, 1,
        2, 2, 0, 1, 10},
    {185625000, 185625000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {185625000, 232031250, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {185625000, 278437500, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {185625000, 371250000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {187000000, 187000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {187250000, 187250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {189000000, 189000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {189000000, 236250000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {189000000, 283500000, 0, 12,
        2, 0, 0, 1, 0,
        3, 5, 1, 3, 15},
    {189000000, 378000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {189000000, 378000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {189000000, 472500000, 1, 10,
        1, 3, 1, 1, 0,
        3, 7, 0, 0, 25},
    {189000000, 567000000, 1, 12,
        2, 3, 1, 0, 0,
        3, 4, 0, 0, 15},
    {193250000, 193250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {198000000, 198000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {198000000, 247500000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {198000000, 297000000, 0, 12,
        2, 0, 0, 1, 0,
        3, 4, 0, 1, 15},
    {198000000, 396000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {198000000, 594000000, 2, 8,
        0, 3, 1, 0, 0,
        3, 4, 0, 0, 15},
    {201400000, 201400000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {201400000, 251750000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {201400000, 302100000, 0, 12,
        2, 0, 0, 1, 0,
        3, 4, 0, 1, 15},
    {201400000, 402800000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {202500000, 202500000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {202500000, 405000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {202500000, 506250000, 1, 10,
        1, 3, 1, 1, 0,
        3, 7, 0, 0, 25},
    {204750000, 204750000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {208000000, 208000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {214750000, 214750000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {216000000, 216000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {216000000, 270000000, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {216000000, 324000000, 0, 12,
        2, 0, 0, 1, 0,
        3, 4, 0, 1, 15},
    {216000000, 432000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {216000000, 432000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {216000000, 540000000, 1, 10,
        1, 3, 1, 1, 0,
        3, 7, 0, 0, 25},
    {218250000, 218250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {222750000, 222750000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {222750000, 278437500, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {222750000, 334125000, 0, 12,
        2, 0, 0, 1, 0,
        3, 4, 0, 1, 15},
    {222750000, 445500000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {226575000, 226575000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {226575000, 283218750, 0, 10,
        1, 0, 0, 3, 0,
        3, 8, 3, 3, 25},
    {226575000, 339862500, 0, 12,
        2, 0, 0, 1, 0,
        3, 4, 0, 1, 15},
    {226575000, 453150000, 0, 16,
        3, 3, 1, 0, 0,
        3, 2, 0, 1, 10},
    {229500000, 229500000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {234000000, 234000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {237600000, 237600000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {237600000, 297000000, 0, 10,
        1, 0, 0, 3, 0,
        3, 6, 2, 1, 25},
    {237600000, 356400000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {237600000, 475200000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {243000000, 243000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {243000000, 303750000, 0, 10,
        1, 0, 0, 3, 0,
        3, 6, 2, 1, 25},
    {243000000, 364500000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {243000000, 486000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {243000000, 486000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {245250000, 245250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {245500000, 245500000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {247500000, 247500000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {247500000, 309375000, 0, 10,
        1, 0, 0, 3, 0,
        3, 6, 2, 1, 25},
    {247500000, 371250000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {247500000, 495000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {251750000, 251750000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {251750000, 314687500, 0, 10,
        1, 0, 0, 3, 0,
        3, 6, 2, 1, 25},
    {251750000, 377625000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {251750000, 503500000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {261000000, 261000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {268250000, 268250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {268500000, 268500000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {270000000, 270000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {270000000, 337500000, 0, 10,
        1, 0, 0, 3, 0,
        3, 6, 2, 1, 25},
    {270000000, 405000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {270000000, 540000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {270000000, 540000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {276925000, 276925000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {276925000, 346156250, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {276925000, 415387500, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {276925000, 553850000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {281250000, 281250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 0, 3, 5},
    {288000000, 288000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {288000000, 360000000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {288000000, 432000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {288000000, 576000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {297000000, 297000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {297000000, 371250000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {297000000, 445500000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {297000000, 594000000, 0, 16,
        3, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {297000000, 594000000, 1, 8,
        0, 3, 1, 0, 0,
        3, 3, 0, 0, 10},
    {302100000, 302100000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {302100000, 377625000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {302100000, 453150000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 1, 15},
    {317000000, 317000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {324000000, 324000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {324000000, 405000000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {324000000, 486000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {327275000, 327275000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {327275000, 409093750, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {327275000, 490912500, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {330000000, 330000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {330000000, 412500000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {330000000, 495000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {333250000, 333250000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {340000000, 340000000, 0, 8,
        0, 0, 0, 0, 0,
        3, 1, 1, 1, 5},
    {348500000, 348500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {351000000, 351000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {351000000, 438750000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {351000000, 526500000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {352450000, 352450000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {352450000, 440562500, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {352450000, 528675000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {356400000, 356400000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {356400000, 445500000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {356400000, 534600000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {356500000, 356500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {360000000, 360000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {360000000, 450000000, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {360000000, 540000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {371250000, 371250000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {371250000, 464062500, 0, 10,
        1, 3, 1, 3, 0,
        3, 6, 2, 1, 25},
    {371250000, 556875000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {377625000, 377625000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {377625000, 472031250, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {377625000, 566437500, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {378000000, 378000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {378000000, 472500000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {378000000, 567000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {380500000, 380500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {396000000, 396000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {396000000, 495000000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {396000000, 594000000, 0, 12,
        2, 3, 1, 1, 0,
        3, 4, 0, 0, 15},
    {402800000, 402800000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {402800000, 503500000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {405000000, 405000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {405000000, 506250000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {412500000, 412500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {412500000, 515625000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {415800000, 415800000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {415800000, 519750000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {432000000, 432000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {432000000, 540000000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {443250000, 443250000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {445500000, 445500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {445500000, 556875000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {450000000, 450000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {450000000, 562500000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {453150000, 453150000, 0, 8,
        0, 3, 1, 0, 0,
        3, 1, 1, 1, 5},
    {453150000, 566437500, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {475200000, 475200000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {475200000, 594000000, 0, 10,
        1, 3, 1, 3, 0,
        3, 7, 0, 0, 25},
    {486000000, 486000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {495000000, 495000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {503500000, 503500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {505250000, 505250000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {519750000, 519750000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {534600000, 534600000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {540000000, 540000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {552750000, 552750000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {553850000, 553850000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {556875000, 556875000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {577500000, 577500000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
    {594000000, 594000000, 0, 8,
        0, 3, 1, 0, 0,
        3, 2, 0, 0, 5},
};
/******************************************************************************
 *                                    函数实现                                    *
 ******************************************************************************/




















/*****************************************************************************
*  函 数 名: get_phy_mpll_tab
*  功能描述: 获取phy_mpll配置
*  输入参数: unsigned int pixclock: 像素始终(Hz)
*            unsigned int tmdsclk: TMDS信号时钟(Hz)
*            char pixrepet
*            char colordepth: 色深(8/10/12/16)
*  输出参数: 无
*  返 回 值: static const struct phy_mpll_config_tab *
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
static const struct phy_mpll_config_tab *get_phy_mpll_tab(
        unsigned int pixclock, unsigned int tmdsclk,
        char pixrepet, char colordepth)
{
    int i = 0, index1 = 0, index2 = 0;

    if (pixclock == 0)
    {
        DEBUG_PRINT("[%s]ERROR:pixclock is invalid!\n", __FUNCTION__);
        return NULL;
    }

    if (tmdsclk == 0)
    {
        ERR_PRINT("[%s]ERROR:tmdsclk is invalid!\n", __FUNCTION__);
        return NULL;
    }

    if(pixclock < PHY_MPLL_TABLE[0].pix_clock)
    {
        for(i=0; i<PHY_MPLL_TABLE_SIZE; i++)
        {
            if((PHY_MPLL_TABLE[i].pix_repet == pixrepet) && (PHY_MPLL_TABLE[i].color_depth == colordepth))
            {
                DEBUG_PRINT("[%s]Warning: Not match exactly! Use approaching config PHY_MPLL_TABLE[%d]\n", __FUNCTION__, i);
                return &PHY_MPLL_TABLE[i];
            }
        }
        goto error;
    }
    else if (pixclock > PHY_MPLL_TABLE[PHY_MPLL_TABLE_SIZE-1].pix_clock)
    {
        for(i=PHY_MPLL_TABLE_SIZE-1; i>=0; i--)
        {
            if((PHY_MPLL_TABLE[i].pix_repet == pixrepet) && (PHY_MPLL_TABLE[i].color_depth == colordepth))
            {
                DEBUG_PRINT("[%s]Warning: Not match exactly! Use approaching config PHY_MPLL_TABLE[%d]\n", __FUNCTION__, i);
                return &PHY_MPLL_TABLE[i];
            }
        }
        goto error;
    }

   for(i=0; i<PHY_MPLL_TABLE_SIZE; i++)
   {
        if(pixclock <= PHY_MPLL_TABLE[i].pix_clock)
        {
            if((pixclock == PHY_MPLL_TABLE[i].pix_clock) &&
                (pixrepet == PHY_MPLL_TABLE[i].pix_repet) &&
                (colordepth == PHY_MPLL_TABLE[i].color_depth))
            {
                DEBUG_PRINT("[%s]Match successfully! Use config PHY_MPLL_TABLE[%d].\n", __FUNCTION__, i);
                return &PHY_MPLL_TABLE[i];
            }

            index1 = i - 1;
            index2 = i;

            while(index1 >= 0)
            {
                if((PHY_MPLL_TABLE[index1].pix_repet == pixrepet) && (PHY_MPLL_TABLE[index1].color_depth == colordepth))
                {
                    break;
                }
                index1--;
            }

            while(index2 < PHY_MPLL_TABLE_SIZE)
            {
                if((PHY_MPLL_TABLE[index2].pix_repet == pixrepet) && (PHY_MPLL_TABLE[index2].color_depth == colordepth))
                {
                    break;
                }
                index2
++;
            }

            if((index1 < 0) && (index2 < PHY_MPLL_TABLE_SIZE))
            {
                DEBUG_PRINT("[%s]Warning: Not match exactly! Use approaching config PHY_MPLL_TABLE[%d].\n", __FUNCTION__, index2);
                return &PHY_MPLL_TABLE[index2];
            }
            else if((index1 >= 0) && (index2 >= PHY_MPLL_TABLE_SIZE) )
            {
                DEBUG_PRINT("[%s]Warning: Not match exactly! Use approaching config PHY_MPLL_TABLE[%d].\n", __FUNCTION__, index1);
                return &PHY_MPLL_TABLE[index1];
            }
            else if((index1 >= 0 ) && (index2 < PHY_MPLL_TABLE_SIZE))
            {
                if( (pixclock - PHY_MPLL_TABLE[index1].pix_clock) < (PHY_MPLL_TABLE[index2].pix_clock - pixclock) )
                {
                    DEBUG_PRINT("[%s]Warning: Not match exactly! Use approaching config PHY_MPLL_TABLE[%d].\n", __FUNCTION__, index1);
                    return &PHY_MPLL_TABLE[index1];
                }
                else
                {
                    DEBUG_PRINT("[%s]Warning: Not match exactly! Use approaching config PHY_MPLL_TABLE[%d].\n", __FUNCTION__, index2);
                    return &PHY_MPLL_TABLE[index2];
                }
            }
            goto error;
        }
   }

error:
   ERR_PRINT("[%s]Error: Not match! Can't find a approximation.\n", __FUNCTION__);
   return NULL;
}

/* ddc i2c master reset */
/*static*/ void ljm_hdmiv2_i2cm_reset(struct hdmi_dev *hdmi_dev)
{
    HDMI_REG_MSK(hdmi_dev, I2CM_SOFTRSTZ, m_I2CM_SOFTRST, v_I2CM_SOFTRST(0));
    usleep_range(90, 100);
}
#if 0
/*set read/write offset,set read/write mode*/
static void ljm_hdmiv2_i2cm_write_request(struct hdmi_dev *hdmi_dev,
                           u8 offset, u8 data)
{
    HDMI_REG_W8(hdmi_dev, I2CM_ADDRESS, offset);
    HDMI_REG_W8(hdmi_dev, I2CM_DATAO, data);
    HDMI_REG_MSK(hdmi_dev, I2CM_OPERATION, m_I2CM_WR, v_I2CM_WR(1));
}

static void ljm_hdmiv2_i2cm_read_request(struct hdmi_dev *hdmi_dev,
                          u8 offset)
{
    HDMI_REG_W8(hdmi_dev, I2CM_ADDRESS, offset);
    HDMI_REG_MSK(hdmi_dev, I2CM_OPERATION, m_I2CM_RD, v_I2CM_RD(1));
}
#endif
#define I2C_DIV_FACTOR 1000000
static u16 i2c_count(u16 sfrclock, u16 sclmintime)
{
    unsigned long tmp_scl_period = 0;

    if (((sfrclock * sclmintime) % I2C_DIV_FACTOR) != 0)
        tmp_scl_period = (unsigned long)((sfrclock * sclmintime) +
                (I2C_DIV_FACTOR - ((sfrclock * sclmintime) %
                I2C_DIV_FACTOR))) / I2C_DIV_FACTOR;
    else
        tmp_scl_period = (unsigned long)(sfrclock * sclmintime) /
                I2C_DIV_FACTOR;

    return (u16)(tmp_scl_period);
}

#define EDID_I2C_MIN_SS_SCL_HIGH_TIME   9625
#define EDID_I2C_MIN_SS_SCL_LOW_TIME    10000

static void ljm_hdmiv2_i2cm_clk_init(struct hdmi_dev *hdmi_dev)
{
    int value;

    /* Set DDC I2C CLK which divided from DDC_CLK. */
    value = i2c_count(27000, EDID_I2C_MIN_SS_SCL_HIGH_TIME);
    HDMI_REG_W8(hdmi_dev, I2CM_SS_SCL_HCNT_0_ADDR,
            value & 0xff);
    HDMI_REG_W8(hdmi_dev, I2CM_SS_SCL_HCNT_1_ADDR,
            (value >> 8) & 0xff);
    value = i2c_count(27000, EDID_I2C_MIN_SS_SCL_LOW_TIME);
    HDMI_REG_W8(hdmi_dev, I2CM_SS_SCL_LCNT_0_ADDR,
            value & 0xff);
    HDMI_REG_W8(hdmi_dev, I2CM_SS_SCL_LCNT_1_ADDR,
            (value >> 8) & 0xff);
    HDMI_REG_MSK(hdmi_dev, I2CM_DIV, m_I2CM_FAST_STD_MODE,
             v_I2CM_FAST_STD_MODE(STANDARD_MODE));
}

static void ljm_hdmiv2_scdc_work_func(struct work_struct *work);

/**
 * ljm_hdmiv2_irq_init - hdmi中断配置初始化
 * @hdmi_dev: hdmi接口设备指针
 *
 * 该函数用于在hdmi硬件初始化流程中控制中断开关。
 * 函数内分别关闭了各中断的使能，但同时也打开了总的中断开关。
 * 这样做的目的，是让开发者在需要时只使能所需中断，而不用关注总的中断开关。
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
void ljm_hdmiv2_irq_init(struct hdmi_dev *hdmi_dev)
{
    /* mute all possible intterupt */
    HDMI_REG_W8(hdmi_dev, IH_MUTE_FC_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_FC_STAT1, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_FC_STAT2, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_AS_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CM_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_CEC_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_VP_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CMPHY_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE_AHBDMAAUD_STAT0, 0xff);
    HDMI_REG_W8(hdmi_dev, IH_MUTE, 0x0);

    return;
}

/**
 * ljm_hdmiv2_init - init hdmi controller
 * @hdmi_dev: pointer to hdmi device
 * @base: virtual base adddress
 * @offset: register offset
 *
 * This function initializes hdmi controller.
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_init(struct hdmi_dev *hdmi_dev, void *base, int offset, struct ljm_port_info *info)
{
    int index = info->sub_id;

    HEADER_ARG("hdmi_dev=%lx base=%lx offset=%x index=%d", (unsigned long)hdmi_dev, (unsigned long)base, offset, index);

#ifdef UEFI_GOP
    hdmi_dev->Private = base;
#else
    hdmi_dev->regbase = base;
#endif

    hdmi_dev->offset = offset;
    hdmi_dev->index = index;
    hdmi_dev->is_dvi = info->is_dvii;
    hdmi_dev->vendor_name = "ljm";
    hdmi_dev->product_name = "gp201";
    hdmi_dev->deviceinfo = 0x01;

    ljm_hdmiv2_i2cm_clk_init(hdmi_dev);
    ljm_hdmiv2_irq_init(hdmi_dev);

    mutex_init(&hdmi_dev->i2cm_mutex);
    INIT_DELAYED_WORK(&hdmi_dev->scdc_work, ljm_hdmiv2_scdc_work_func);

    FOOTER();
    return 0;
}

/**
 * ljm_hdmiv2_i2cm_write - dwc phy i2c write
 * @hdmi_dev: pointer to hdmi device
 * @slave: slave address
 * @addr: regitser address
 * @data: data to write
 *
 * This function writes dwc phy register
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_i2cm_write(struct hdmi_dev *hdmi_dev, u8 slave, u8 addr, u8 data)
{
    int timeout, retry = 0, ret = 0;
    u8 status;

    mutex_lock(&hdmi_dev->i2cm_mutex);
Retry:
    timeout = GP201_I2CM_TIMEOUT_CNT;
    if(retry > 0)
    {
        if(retry <= 3)
        {
            HDMI_REG_MSK(hdmi_dev, I2CM_OPERATION, m_I2CM_BUS_CLEAR, v_I2CM_BUS_CLEAR(1));
            WARN_PRINT("hdmi %d i2cm_write 0x%x 0x%x [ERROR: %d] retry:%d", hdmi_dev->index, slave, addr, ret, retry);
            udelay(5);
        }
        else
        {
            mutex_unlock(&hdmi_dev->i2cm_mutex);
            return ret;
        }
    }

    HDMI_REG_W8(hdmi_dev, IH_I2CM_STAT0, m_I2CM_STAT0_DONE | m_I2CM_STAT0_ERROR);
    HDMI_REG_W8(hdmi_dev, I2CM_SLAVE, slave);
    HDMI_REG_W8(hdmi_dev, I2CM_ADDRESS, addr);
    HDMI_REG_W8(hdmi_dev, I2CM_DATAO, data);    /* write data */

    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CM_STAT0, 0x3); /* mutes ih_i2cmphy_stat0[1:0] */
    HDMI_REG_W8(hdmi_dev, I2CM_OPERATION, m_I2CM_OP_WR);    /* i2c write */

    while (--timeout) {
        status = HDMI_REG_R8(hdmi_dev, IH_I2CM_STAT0);
        if (status & m_I2CM_STAT0_DONE)
            break;
        if (status & m_I2CM_STAT0_ERROR)
        if (status & m_I2CM_STAT0_ERROR)
        {
            retry++;
            ret = DDC_ERROR;
            goto Retry;   /* slave is not exist */
        }
        udelay(5);
    }

    if (timeout == 0)
    {
        ret = DDC_TIMEOUT; /* i2c read timeout */
        retry++;
        goto Retry;
    }

    TRACE("hdmi %d i2cm_write 0x%x [0x%x]=0x%x", hdmi_dev->index, slave, addr, data);
    mutex_unlock(&hdmi_dev->i2cm_mutex);
    return 0;
}

/**
 * ljm_hdmiv2_i2cm_read - dwc phy i2c read
 * @hdmi_dev: pointer to hdmi device
 * @slave: slave address
 * @addr: regitser address
 * @pdata: pointer to read data
 *
 * This function reads dwc phy register
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_i2cm_read(struct hdmi_dev *hdmi_dev, u8 slave, u8 addr, u8 *pdata)
{
    int timeout, retry = 0, ret = 0;
    u8 status;

    mutex_lock(&hdmi_dev->i2cm_mutex);
Retry:
    timeout = GP201_I2CM_TIMEOUT_CNT;
    if(retry > 0)
    {
        if(retry <= 3)
        {
            HDMI_REG_MSK(hdmi_dev, I2CM_OPERATION, m_I2CM_BUS_CLEAR, v_I2CM_BUS_CLEAR(1));
            DEBUG_PRINT("hdmi %d i2cm_read 0x%x 0x%x [ERROR: %d] retry:%d", hdmi_dev->index, slave, addr, ret, retry);
            udelay(100);
        }
        else
        {
            mutex_unlock(&hdmi_dev->i2cm_mutex);
            return ret;
        }
    }

    HDMI_REG_W8(hdmi_dev, IH_I2CM_STAT0, m_I2CM_STAT0_DONE | m_I2CM_STAT0_ERROR);
    HDMI_REG_W8(hdmi_dev, I2CM_SLAVE, slave);
    HDMI_REG_W8(hdmi_dev, I2CM_ADDRESS, addr);

    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CM_STAT0, 0x3); /* mutes ih_i2cmphy_stat0[1:0] */
    HDMI_REG_W8(hdmi_dev, I2CM_OPERATION, m_I2CM_OP_RD);    /* i2c read */

    while (--timeout) {
        status = HDMI_REG_R8(hdmi_dev, IH_I2CM_STAT0);
        if (status & m_I2CM_STAT0_DONE)
            break;
        if (status & m_I2CM_STAT0_ERROR)
        {
            retry++;
            ret = DDC_ERROR;
            goto Retry;   /* slave is not exist */
        }
        udelay(5);
    }

    if (timeout == 0)
    {
        ret = DDC_TIMEOUT; /* i2c read timeout */
        retry++;
        goto Retry;
    }

    *pdata = HDMI_REG_R8(hdmi_dev, I2CM_DATAI);

    //TRACE("hdmi %d i2cm_read 0x%x [0x%x]=0x%x", hdmi_dev->index, slave, addr, *pdata);
    mutex_unlock(&hdmi_dev->i2cm_mutex);
    return 0;
}

/**
 * ljm_hdmiv2_i2cm_read_ext - dwc phy i2c extended read
 * @hdmi_dev: pointer to hdmi device
 * @seg_ptr: E-DDC segment pointer
 * @slave: slave address
 * @addr: regitser address
 * @pdata: pointer to read data
 *
 * This function reads dwc phy register
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_i2cm_read_ext(struct hdmi_dev *hdmi_dev, u8 seg_ptr, u8 slave, u8 addr, u8 *pdata)
{
    int timeout, retry = 0, ret = 0;
    u8 status;

    mutex_lock(&hdmi_dev->i2cm_mutex);
Retry:
    timeout = GP201_I2CM_TIMEOUT_CNT;
    if(retry > 0)
    {
        if(retry <= 3)
        {
            HDMI_REG_MSK(hdmi_dev, I2CM_OPERATION, m_I2CM_BUS_CLEAR, v_I2CM_BUS_CLEAR(1));
            WARN_PRINT("hdmi %d i2cm_read_ext 0x%x 0x%x [ERROR: %d] retry:%d", hdmi_dev->index, slave, addr, ret, retry);
            udelay(5);
        }
        else
        {
            mutex_unlock(&hdmi_dev->i2cm_mutex);
            return ret;
        }
    }

    HDMI_REG_W8(hdmi_dev, IH_I2CM_STAT0, m_I2CM_STAT0_DONE | m_I2CM_STAT0_ERROR);
    HDMI_REG_W8(hdmi_dev, I2CM_SEGADDR, DDC_SEG_PTR_ADDR);
    HDMI_REG_W8(hdmi_dev, I2CM_SEGPTR, seg_ptr);
    HDMI_REG_W8(hdmi_dev, I2CM_SLAVE, slave);
    HDMI_REG_W8(hdmi_dev, I2CM_ADDRESS, addr);

    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CM_STAT0, 0x3); /* mutes ih_i2cmphy_stat0[1:0] */
    HDMI_REG_W8(hdmi_dev, I2CM_OPERATION, m_I2CM_OP_RD_EXT);    /* i2c extended read */

    while (--timeout) {
        status = HDMI_REG_R8(hdmi_dev, IH_I2CM_STAT0);
        if (status & m_I2CM_STAT0_DONE)
            break;
        if (status & m_I2CM_STAT0_ERROR)
        {
            retry++;
            ret = DDC_ERROR;
            goto Retry;   /* slave is not exist */
        }
        udelay(5);
    }

    if (timeout == 0)
    {
        ret = DDC_TIMEOUT; /* i2c read timeout */
        retry++;
        goto Retry;
    }

    *pdata = HDMI_REG_R8(hdmi_dev, I2CM_DATAI);

    TRACE("hdmi %d i2cm_read_ext 0x%x [0x%x]=0x%x", hdmi_dev->index, slave, addr, *pdata);

    mutex_unlock(&hdmi_dev->i2cm_mutex);
    return 0;
}

#if 0
static void ljm_hdmiv2_i2cm_mask_int(struct hdmi_dev *hdmi_dev, int mask)
{
    if (!mask) {
        HDMI_REG_MSK(hdmi_dev, I2CM_INT,
                 m_I2CM_DONE_MASK, v_I2CM_DONE_MASK(0));
        HDMI_REG_MSK(hdmi_dev, I2CM_CTLINT,
                 m_I2CM_NACK_MASK | m_I2CM_ARB_MASK,
                 v_I2CM_NACK_MASK(0) | v_I2CM_ARB_MASK(0));
    } else {
        HDMI_REG_MSK(hdmi_dev, I2CM_INT,
                 m_I2CM_DONE_MASK, v_I2CM_DONE_MASK(1));
        HDMI_REG_MSK(hdmi_dev, I2CM_CTLINT,
                 m_I2CM_NACK_MASK | m_I2CM_ARB_MASK,
                 v_I2CM_NACK_MASK(1) | v_I2CM_ARB_MASK(1));
    }
}
#endif
static int ljm_hdmiv2_scdc_get_sink_version(struct hdmi_dev *hdmi_dev)
{
    u8 data;
    ljm_hdmiv2_i2cm_read(hdmi_dev, DDC_SCDC_ADDR, SCDC_SINK_VER, &data);

    return data;
}

static void ljm_hdmiv2_scdc_set_source_version(struct hdmi_dev *hdmi_dev,
                            u8 version)
{
    ljm_hdmiv2_i2cm_write(hdmi_dev, DDC_SCDC_ADDR, SCDC_SOURCE_VER, version);
}

static void ljm_hdmiv2_scdc_read_request(struct hdmi_dev *hdmi_dev,
                          int enable)
{
    HDMI_REG_MSK(hdmi_dev, I2CM_SCDC_READ_UPDATE,
             m_I2CM_READ_REQ_EN, v_I2CM_READ_REQ_EN(enable));
    ljm_hdmiv2_i2cm_write(hdmi_dev, DDC_SCDC_ADDR, SCDC_CONFIG_0, enable);
}
#if 0
static void ljm_hdmiv2_scdc_update_read(struct hdmi_dev *hdmi_dev)
{
    HDMI_REG_MSK(hdmi_dev, I2CM_SCDC_READ_UPDATE,
             m_I2CM_READ_UPDATE, v_I2CM_READ_UPDATE(1));
}

static int ljm_hdmiv2_scdc_get_scambling_status(struct hdmi_dev *hdmi_dev)
{
    u8 val;

    ljm_hdmiv2_i2cm_read(hdmi_dev, DDC_SCDC_ADDR, SCDC_SCRAMBLER_STAT, &val);
    return val;
}

static void ljm_hdmiv2_scdc_enable_polling(struct hdmi_dev *hdmi_dev,
                        int enable)
{
    ljm_hdmiv2_scdc_read_request(hdmi_dev, enable);
    HDMI_REG_MSK(hdmi_dev, I2CM_SCDC_READ_UPDATE,
             m_I2CM_UPRD_VSYNC_EN, v_I2CM_UPRD_VSYNC_EN(enable));
}

static int ljm_hdmiv2_scdc_get_status_reg0(struct hdmi_dev *hdmi_dev)
{
    ljm_hdmiv2_scdc_read_request(hdmi_dev, 1);
    ljm_hdmiv2_scdc_update_read(hdmi_dev);
    return HDMI_REG_R8(hdmi_dev, I2CM_SCDC_UPDATE0);
}

static int ljm_hdmiv2_scdc_get_status_reg1(struct hdmi_dev *hdmi_dev)
{
    ljm_hdmiv2_scdc_read_request(hdmi_dev, 1);
    ljm_hdmiv2_scdc_update_read(hdmi_dev);
    return HDMI_REG_R8(hdmi_dev, I2CM_SCDC_UPDATE1);
}

static void ljm_hdmiv2_scdc_init(struct hdmi_dev *hdmi_dev)
{
    ljm_hdmiv2_i2cm_reset(hdmi_dev);
    ljm_hdmiv2_i2cm_mask_int(hdmi_dev, 1);
    ljm_hdmiv2_i2cm_clk_init(hdmi_dev);
    /* set scdc i2c addr */
    HDMI_REG_W8(hdmi_dev, I2CM_SLAVE, DDC_SCDC_ADDR);
    ljm_hdmiv2_i2cm_mask_int(hdmi_dev, 0);/*enable interrupt*/
}

static void ljm_hdmiv2_scdc_set_tmds_rate(struct hdmi_dev *hdmi_dev)
{
    u8 stat;

    //mutex_lock(&hdmi_dev->ddc_lock);
    ljm_hdmiv2_scdc_init(hdmi_dev);
    ljm_hdmiv2_i2cm_read(hdmi_dev, DDC_SCDC_ADDR, SCDC_TMDS_CONFIG, &stat);
    if (hdmi_dev->tmdsclk > 340000000)
        stat |= 2;
    else
        stat &= 0x1;
    ljm_hdmiv2_i2cm_write(hdmi_dev, DDC_SCDC_ADDR, SCDC_TMDS_CONFIG, stat);
    //mutex_unlock(&hdmi_dev->ddc_lock);
}
#endif

/**
 * ljm_hdmiv2_scdc_work_func - 监控scdc配置是否生效。
 *
 * 在配置scdc后调用此函数，间隔5s确认一次scdc是否生效，共确认两次。
 * 若未生效，则重新使能，重复上述步骤。
*/
static void ljm_hdmiv2_scdc_work_func(struct work_struct *work)
{
    int ret;
    u8 status, conf;
    struct hdmi_dev *hdmi = container_of(work, struct hdmi_dev, scdc_work.work);

    if(!hdmi->scdc_work_run)
        return;

    DEBUG_PRINT("[hdmi%d scdc]check work.\n", hdmi->index);

    if (hdmi->edid.scdc_present == 1) {
        if (hdmi->tmdsclk > 340000000 ||
            hdmi->edid.lte_340mcsc_scramble)
        {
            ret = ljm_hdmiv2_i2cm_read(hdmi, DDC_SCDC_ADDR, SCDC_SCRAMBLER_STAT, &status);
            ret |= ljm_hdmiv2_i2cm_read(hdmi, DDC_SCDC_ADDR, SCDC_TMDS_CONFIG, &conf);
            if(ret<0)
            {
                ERR_PRINT("[hdmi%d scdc] i2cm error.\n", hdmi->index);
                return;
            }

            DEBUG_PRINT("[hdmi%d scdc ] conf=0x%x status=0x%x.\n", hdmi->index, conf, status);
            if(((status & 0x1) == 0) || (conf != 0x3))
            {
                hdmi->scdc_work_confirm_count = 0;
                ljm_hdmiv2_i2cm_write(hdmi, DDC_SCDC_ADDR, SCDC_TMDS_CONFIG, 0x3);
                HDMI_REG_MSK(hdmi, FC_SCRAMBLER_CTRL, m_FC_SCRAMBLE_EN, v_FC_SCRAMBLE_EN(1));
                schedule_delayed_work(&hdmi->scdc_work, msecs_to_jiffies(5000));
            }
            else if(hdmi->scdc_work_confirm_count < 1) /* 状态正确后，间隔5s再次确认，防止状态反复。 */
            {
                DEBUG_PRINT("[hdmi%d scdc ] check work confirm 1.\n", hdmi->index);
                hdmi->scdc_work_confirm_count++;
                schedule_delayed_work(&hdmi->scdc_work, msecs_to_jiffies(5000));
            }
        }
    }
    return;
}

void ljm_hdmiv2_scdc_work_start(struct hdmi_dev *hdmi_dev)
{
    hdmi_dev->scdc_work_confirm_count = 0;
    hdmi_dev->scdc_work_run = true;
    schedule_delayed_work(&hdmi_dev->scdc_work, msecs_to_jiffies(5000));
    return;
}

void ljm_hdmiv2_scdc_work_stop(struct hdmi_dev *hdmi_dev)
{
    hdmi_dev->scdc_work_run = false;
    cancel_delayed_work_sync(&hdmi_dev->scdc_work);
    return;
}

static int ljm_hdmiv2_scrambling_enable(struct hdmi_dev *hdmi_dev,
                         int enable)
{
    u8 hdmi_reg_tmp;

    DEBUG_PRINT("[hdmi%d] scdc enable=%d\n", hdmi_dev->index, enable);
    ljm_hdmiv2_scdc_work_stop(hdmi_dev);
    if (enable == 1) {
        /* Write on Rx the bit Scrambling_Enable, register 0x20 */
        ljm_hdmiv2_i2cm_write(hdmi_dev, DDC_SCDC_ADDR, SCDC_TMDS_CONFIG, 0x3);
        ljm_hdmiv2_scdc_work_start(hdmi_dev);
        /* TMDS software reset request */
        HDMI_REG_MSK(hdmi_dev, MC_SWRSTZREQ,
                 m_TMDS_SWRST, v_TMDS_SWRST(0));
        /* Enable/Disable Scrambling */
        HDMI_REG_MSK(hdmi_dev, FC_SCRAMBLER_CTRL,
                 m_FC_SCRAMBLE_EN, v_FC_SCRAMBLE_EN(1));

        /* 解决单接口热插拔声音播放丢失问题 */
        hdmi_reg_tmp = HDMI_REG_R8(hdmi_dev, FC_CTRLDUR);
        HDMI_REG_W8(hdmi_dev, FC_CTRLDUR, hdmi_reg_tmp);

        hdmi_reg_tmp = HDMI_REG_R8(hdmi_dev, FC_EXCTRLDUR);
        HDMI_REG_W8(hdmi_dev, FC_EXCTRLDUR, hdmi_reg_tmp);
    } else {
        /* Enable/Disable Scrambling */
        HDMI_REG_MSK(hdmi_dev, FC_SCRAMBLER_CTRL,
                 m_FC_SCRAMBLE_EN, v_FC_SCRAMBLE_EN(0));
        /* TMDS software reset request */
        HDMI_REG_MSK(hdmi_dev, MC_SWRSTZREQ,
                 m_TMDS_SWRST, v_TMDS_SWRST(0));
        /* Write on Rx the bit Scrambling_Enable, register 0x20 */
        ljm_hdmiv2_i2cm_write(hdmi_dev, DDC_SCDC_ADDR, SCDC_TMDS_CONFIG, 0x0);
    }
    return 0;
}

void ljm_hdmiv2_scrambling_set(struct hdmi_dev *hdmi_dev)
{
    int sink_version;

    hdmi_dev->need_scdc = false;
    if (hdmi_dev->edid.scdc_present == 1) {
        if (hdmi_dev->tmdsclk > 340000000 ||
            hdmi_dev->edid.lte_340mcsc_scramble) {
            /* used for HDMI 2.0 TX */
            sink_version = ljm_hdmiv2_scdc_get_sink_version(hdmi_dev);
            DEBUG_PRINT("sink scdc version is %d\n", sink_version);
            sink_version = hdmi_dev->edid.hf_vsdb_version;
            ljm_hdmiv2_scdc_set_source_version(hdmi_dev, sink_version);
            if (hdmi_dev->edid.rr_capable == 1)
                ljm_hdmiv2_scdc_read_request(hdmi_dev, 1);
            ljm_hdmiv2_scrambling_enable(hdmi_dev, 1);
        } else {
            ljm_hdmiv2_scrambling_enable(hdmi_dev, 0);
        }
    } else {
        HDMI_REG_MSK(hdmi_dev, FC_SCRAMBLER_CTRL,
                     m_FC_SCRAMBLE_EN, v_FC_SCRAMBLE_EN(0));
    }

    return;
}

#define PHY_TIMEOUT     10000
/*****************************************************************************
*  函 数 名: ljm_hdmiv2_config_phy
*  功能描述: 配置hdmi v2硬件控制
*  输入参数: struct hdmi_dev *hdmi_dev: hdmi设备指针
*  输出参数: 无
*  返 回 值: static int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
static int ljm_hdmiv2_config_phy(struct hdmi_dev *hdmi_dev)
{
    int stat = 0, i = 0;
    const struct phy_mpll_config_tab *phy_mpll = NULL;

//  clk_set_rate(hdmi_dev->pclk_phy, pixelclk);

    HDMI_REG_MSK(hdmi_dev, PHY_I2CM_DIV, m_PHY_I2CM_FAST_STD, v_PHY_I2CM_FAST_STD(0));
    HDMI_REG_MSK(hdmi_dev, PHY_MASK, m_PHY_LOCK, v_PHY_LOCK(1));
    /* power off PHY */
    //HDMI_REG_W8(hdmi_dev, PHY_CONF0,
       //     m_SEL_DATAEN_POL | m_PDDQ_SIG | m_SVSRET_SIG);
	HDMI_REG_MSK(hdmi_dev, PHY_CONF0,
		     m_PDDQ_SIG | m_TXPWRON_SIG | m_SVSRET_SIG,
		     v_PDDQ_SIG(1) | v_TXPWRON_SIG(0) | v_SVSRET_SIG(1));

    /* reset PHY */
    HDMI_REG_W8(hdmi_dev, MC_PHYRSTZ, v_PHY_RSTZ(1));
    usleep_range(1000, 2000);
    HDMI_REG_W8(hdmi_dev, MC_PHYRSTZ, v_PHY_RSTZ(0));

    /* Set slave address as PHY GEN2 address */
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_SLAVE, PHY_GEN2_ADDR);

    /* config the required PHY I2C register */
    phy_mpll = get_phy_mpll_tab(hdmi_dev->pixelclk,
                                hdmi_dev->tmdsclk,
                                hdmi_dev->pixelrepeat - 1,
                                hdmi_dev->colordepth);
    if (phy_mpll) {
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_OPMODE_PLLCFG,
                      v_PREP_DIV(phy_mpll->prep_div) |
                      v_MPLL_CKO_DIV(phy_mpll->cko_div) |
                      v_OPMODE(phy_mpll->opmode) |
                      v_REF_CNTRL(phy_mpll->ref_cntrl) |
                      v_MPLL_N_CNTRL(phy_mpll->n_cntrl));
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_PLLCURRGMPCTRL,
                      v_MPLL_GMP_CNTRL(phy_mpll->gmp_cntrl) |
                      v_PROP_CNTRL(phy_mpll->prop_cntrl) |
                      v_INT_CNTRL(phy_mpll->int_cntrl));
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_PLLDIVCTRL,
                      v_VCO_CNTRL(phy_mpll->vco_cntrl) |
                      v_MPLL_MULTIPLIER(phy_mpll->multiplier));
    }

    /* if (hdmi_dev->phy_table) {
        for (i = 0; i < hdmi_dev->phy_table_size; i++)
            if (hdmi_dev->tmdsclk <= hdmi_dev->phy_table[i].maxfreq)
                break;
    }
    if (i != hdmi_dev->phy_table_size && hdmi_dev->phy_table) {
        stat = v_OVERRIDE(1) | v_TX_SYMON(1) | v_CLK_SYMON(1) |
               v_PREEMPHASIS(hdmi_dev->phy_table[i].pre_emphasis) |
               v_SLOPEBOOST(hdmi_dev->phy_table[i].slopeboost);
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_CLKSYMCTRL, stat);

        stat = v_SUP_CLKLVL(hdmi_dev->phy_table[i].clk_level) |
               v_SUP_TXLVL(hdmi_dev->phy_table[i].data0_level);
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_VLEVCTRL, stat);
    } else { */
        if (hdmi_dev->tmdsclk > 340000000)
        {
            //zhaiyf:340MHz以上的配置会再调整，当前暂用340MHz以下的配置
            ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_CLKSYMCTRL,
                        v_OVERRIDE(1) | v_SLOPEBOOST(0) |
                        v_TX_TRAON(1) | v_TX_TRBON(1) |
                        v_TX_SYMON(12) | v_CLK_SYMON(12));
        }
        else
        {
            ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_CLKSYMCTRL,
                        v_OVERRIDE(1) | v_SLOPEBOOST(0) |
                        v_TX_TRAON(1) | v_TX_TRBON(1) |
                        v_TX_SYMON(12) | v_CLK_SYMON(12));
        }

        if (hdmi_dev->tmdsclk > 340000000)
            ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_VLEVCTRL,
                          v_SUP_TXLVL(9));
        else if (hdmi_dev->tmdsclk > 165000000)
            ljm_hdmiv2_write_phy(hdmi_dev,
                          PHYTX_VLEVCTRL,
                          v_SUP_TXLVL(14));
        else
            ljm_hdmiv2_write_phy(hdmi_dev,
                          PHYTX_VLEVCTRL,
                          v_SUP_TXLVL(9));
    /* } */

    if (hdmi_dev->tmdsclk > 340000000)
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_TERM_RESIS,
                             v_TX_TERM(R50_OHMS));
    else
        ljm_hdmiv2_write_phy(hdmi_dev, PHYTX_TERM_RESIS,
                             v_TX_TERM(R100_OHMS));
    ljm_hdmiv2_write_phy(hdmi_dev, 0x05, 0x8000);
    if (hdmi_dev->tmdsclk_ratio_change)
        msleep(100);
    /* power on PHY */
    //HDMI_REG_W8(hdmi_dev, PHY_CONF0,
      //      m_SEL_DATAEN_POL | m_ENHPD_RXSENSE_SIG |
        //    m_TXPWRON_SIG | m_SVSRET_SIG);
    	HDMI_REG_W8(hdmi_dev, PHY_CONF0, 0x2e);

    /* check if the PHY PLL is locked */

    i = 0;
    while (i++ < PHY_TIMEOUT) {
        if ((i % 10) == 0) {
            stat = HDMI_REG_R8(hdmi_dev, PHY_STAT0);
            if (stat & m_PHY_LOCK)
                break;
            usleep_range(1000, 2000);
        }
    }
    if ((stat & m_PHY_LOCK) == 0) {
        stat = HDMI_REG_R8(hdmi_dev, MC_LOCKONCLOCK);
        DRM_ERROR("PHY PLL not locked: PCLK_ON=%ld,TMDSCLK_ON=%ld\n",
                  (stat & m_PCLK_ON) >> 6, (stat & m_TMDSCLK_ON) >> 5);
        return -1;
    }
    HDMI_REG_MSK(hdmi_dev, PHY_MASK, m_PHY_LOCK, v_PHY_LOCK(0));

    return 0;
}

/**
 * ljm_hdmiv2_write_phy - dwc phy i2c write
 * @hdmi_dev: pointer to hdmi device
 * @addr: regitser address
 * @data: data to write
 *
 * This function writes dwc phy register
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_write_phy(struct hdmi_dev *hdmi_dev, u8 addr, u16 data)
{
    int timeout = GP201_I2CM_TIMEOUT_CNT;
    u8 status;

    HDMI_REG_W8(hdmi_dev, IH_I2CMPHY_STAT0, m_I2CM_STAT0_DONE | m_I2CM_STAT0_ERROR);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_SLAVE, PHY_GEN2_ADDR);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_ADDRESS, addr);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_DATAO_1, (data & 0xff00) >> 8);  /* write data high */
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_DATAO_0, data & 0xff);   /* write data low */

    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CMPHY_STAT0, 0x3);  /* mutes ih_i2cmphy_stat0[1:0] */
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_OPERATION, m_I2CM_OP_WR);    /* i2c write */

    while (--timeout) {
        status = HDMI_REG_R8(hdmi_dev, IH_I2CMPHY_STAT0);
        if (status & m_I2CM_STAT0_DONE)
            break;
        if (status & m_I2CM_STAT0_ERROR)
            return DDC_ERROR;   /* slave is not exist */
        usleep_range(900, 1000);
    }

    if (timeout == 0)
        return DDC_TIMEOUT; /* i2c write timeout */

    TRACE("hdmi %d i2cmphy_write [0x%x]=0x%x", hdmi_dev->index, addr, data);
    return 0;
}

/**
 * ljm_hdmiv2_read_phy - dwc phy i2c read
 * @hdmi_dev: pointer to hdmi device
 * @addr: regitser address
 * @pdata: pointer to read data
 *
 * This function reads dwc phy register
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_read_phy(struct hdmi_dev *hdmi_dev, u8 addr, u16 *pdata)
{
    int timeout = GP201_I2CM_TIMEOUT_CNT;
    u8 status;

    HDMI_REG_W8(hdmi_dev, IH_I2CMPHY_STAT0, m_I2CM_STAT0_DONE | m_I2CM_STAT0_ERROR);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_SLAVE, PHY_GEN2_ADDR);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_ADDRESS, addr);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_DATAO_1, 0x00);
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_DATAO_0, 0x00);

    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CMPHY_STAT0, 0x3);  /* mutes ih_i2cmphy_stat0[1:0] */
    HDMI_REG_W8(hdmi_dev, PHY_I2CM_OPERATION, m_I2CM_OP_RD);    /* i2c read */

    while (--timeout) {
        status = HDMI_REG_R8(hdmi_dev, IH_I2CMPHY_STAT0);
        if (status & m_I2CM_STAT0_DONE)
            break;
        if (status & m_I2CM_STAT0_ERROR)
            return DDC_ERROR;   /* slave is not exist */
        usleep_range(900, 1000);
    }

    if (timeout == 0)
        return DDC_TIMEOUT; /* i2c write timeout */

    *pdata = ((u16)(HDMI_REG_R8(hdmi_dev, PHY_I2CM_DATAI_1)) << 8) + HDMI_REG_R8(hdmi_dev, PHY_I2CM_DATAI_0);

    TRACE("hdmi %d i2cmphy_read [0x%x]=0x%x", hdmi_dev->index, addr, *pdata);

    return 0;
}
/*****************************************************************************
*  函 数 名: ljm_hdmiv2_video_framecomposer
*  功能描述: 视频画面配置（色彩、分辨率等）
*  输入参数: struct hdmi_dev *hdmi_dev: hdmi设备指针
*            struct hdmi_video *vpara: 输入配置参数
*  输出参数: 无
*  返 回 值: static int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
static int ljm_hdmiv2_video_framecomposer(struct hdmi_dev *hdmi_dev, struct hdmi_video *vpara)
{
    int value, vsync_pol, hsync_pol, de_pol, interlace;
    struct hdmi_video_timing *timing = NULL;
    struct fb_videomode *mode = NULL;
    u32 tmdsclk;

    vsync_pol = (vpara->mode.sync & FB_SYNC_VERT_HIGH_ACT) ? 1: 0;
    hsync_pol = (vpara->mode.sync & FB_SYNC_HOR_HIGH_ACT) ? 1: 0;
    de_pol = 1;
    interlace = (vpara->mode.vmode & FB_VMODE_INTERLACED) ? 1 : 0;

    HDMI_REG_MSK(hdmi_dev, A_VIDPOLCFG,
                 m_DATAEN_POL | m_VSYNC_POL | m_HSYNC_POL,
                 v_DATAEN_POL(de_pol) |
                 v_VSYNC_POL(vsync_pol) |
                 v_HSYNC_POL(hsync_pol));

    if (vpara->vic > 0) {
        timing = (struct hdmi_video_timing *)hdmi_vic2timing(vpara->vic);
        if (!timing) {
            DRM_ERROR("[%s] not found vic %d\n", __func__, vpara->vic);
            return -ENOENT;
        }
        mode = &timing->mode;
        hdmi_dev->pixelrepeat = timing->pixelrepeat;
    } else {
        mode = &vpara->mode;
        hdmi_dev->pixelrepeat = 1;
    }

    mode = &vpara->mode;
    TRACE("mode=%lx %s: %d %d %d %d %d %d %d %d %d %x %x %x",
        (unsigned long)mode, mode->name,
        mode->xres, mode->yres, mode->pixclock, mode->left_margin, mode->right_margin,
        mode->upper_margin, mode->lower_margin, mode->hsync_len, mode->vsync_len, mode->sync, mode->vmode, mode->flag);

    if (vpara->color_input == HDMI_COLOR_YCBCR420)
        tmdsclk = mode->pixclock / 2;
    else if (vpara->format_3d == HDMI_3D_FRAME_PACKING)
        tmdsclk = 2 * mode->pixclock;
    else
        tmdsclk = mode->pixclock;

    if (vpara->color_output != HDMI_COLOR_YCBCR422) {
        switch (vpara->color_output_depth) {
        case 10:
            tmdsclk += tmdsclk / 4;
            break;
        case 12:
            tmdsclk += tmdsclk / 2;
            break;
        case 16:
            tmdsclk += tmdsclk;
            break;
        case 8:
            default:
            break;
        }
    } else if (vpara->color_output_depth > 12) {
        /* YCbCr422 mode only support up to 12bit */
        vpara->color_output_depth = 12;
    }

    /* FIXME: HDMI edid解析 */
    if ((tmdsclk > 594000000) ||
        (tmdsclk > 340000000  &&
        tmdsclk > hdmi_dev->edid.maxtmdsclock)) {
        if (vpara->format_3d == HDMI_3D_FRAME_PACKING) {
            ERR_PRINT("3d frame packing mode out of max tmdsclk\n");
            return -1;
        } else if (vpara->color_output == HDMI_COLOR_YCBCR444 &&
            hdmi_dev->edid.ycbcr422) {
            WARN_PRINT("out of max tmdsclk, down to YCbCr422");
            vpara->color_output = HDMI_COLOR_YCBCR422;
            tmdsclk = mode->pixclock;
        } else {
            WARN_PRINT("out of max tmds clock, limit to 8bit\n");
            vpara->color_output_depth = 8;
            if (vpara->color_input == HDMI_COLOR_YCBCR420)
                tmdsclk = mode->pixclock / 2;
            else
                tmdsclk = mode->pixclock;
        }
    }

    if ((tmdsclk > 340000000) ||
        (tmdsclk < 340000000 && hdmi_dev->tmdsclk > 340000000))
        hdmi_dev->tmdsclk_ratio_change = true;
    else
        hdmi_dev->tmdsclk_ratio_change = false;

    hdmi_dev->tmdsclk = tmdsclk;
    if (vpara->format_3d == HDMI_3D_FRAME_PACKING)
        hdmi_dev->pixelclk = 2 * mode->pixclock;
    else
        hdmi_dev->pixelclk = mode->pixclock;

    /* hdmi_dev->colordepth is used for find pll config.
     * For YCbCr422, tmdsclk is same on all color depth.
     */
    if (vpara->color_output == HDMI_COLOR_YCBCR422)
        hdmi_dev->colordepth = 8;
    else
        hdmi_dev->colordepth = vpara->color_output_depth;
    DEBUG_PRINT("pixel clk is %lu tmds clk is %u\n",
            hdmi_dev->pixelclk, hdmi_dev->tmdsclk);
    /* Start/stop HDCP keepout window generation */
    HDMI_REG_MSK(hdmi_dev, FC_INVIDCONF,
                 m_FC_HDCP_KEEPOUT, v_FC_HDCP_KEEPOUT(1));

   /* scdc support */
    ljm_hdmiv2_scrambling_set(hdmi_dev);

    HDMI_REG_MSK(hdmi_dev, FC_INVIDCONF,
                 m_FC_VSYNC_POL | m_FC_HSYNC_POL | m_FC_DE_POL |
                 m_FC_HDMI_DVI | m_FC_INTERLACE_MODE,
                 v_FC_VSYNC_POL(vsync_pol) | v_FC_HSYNC_POL(hsync_pol) |
                 v_FC_DE_POL(de_pol) | v_FC_HDMI_DVI(vpara->sink_hdmi) |
                 v_FC_INTERLACE_MODE(interlace));
    if (interlace && vpara->format_3d != HDMI_3D_FRAME_PACKING)
        HDMI_REG_MSK(hdmi_dev, FC_INVIDCONF,
                     m_FC_VBLANK, v_FC_VBLANK(1));
    else
        HDMI_REG_MSK(hdmi_dev, FC_INVIDCONF,
                     m_FC_VBLANK, v_FC_VBLANK(0));

    value = mode->xres;
    if (vpara->color_input == HDMI_COLOR_YCBCR420)
        value = value / 2;
    HDMI_REG_W8(hdmi_dev, FC_INHACTIV1, v_FC_HACTIVE1(value >> 8));
    HDMI_REG_W8(hdmi_dev, FC_INHACTIV0, (value & 0xff));

    if (vpara->format_3d == HDMI_3D_FRAME_PACKING) {
        if (interlace == 0)
            value = 2 * mode->yres +
                mode->upper_margin +
                mode->lower_margin +
                mode->vsync_len;
        else
            value = 2 * mode->yres +
                    3 * (mode->upper_margin +
                         mode->lower_margin +
                         mode->vsync_len) + 2;
    } else {
        value = mode->yres;
    }
    HDMI_REG_W8(hdmi_dev, FC_INVACTIV1, v_FC_VACTIVE1(value >> 8));
    HDMI_REG_W8(hdmi_dev, FC_INVACTIV0, (value & 0xff));

    value = mode->hsync_len + mode->left_margin + mode->right_margin;
    if (vpara->color_input == HDMI_COLOR_YCBCR420)
        value = value / 2;
    HDMI_REG_W8(hdmi_dev, FC_INHBLANK1, v_FC_HBLANK1(value >> 8));
    HDMI_REG_W8(hdmi_dev, FC_INHBLANK0, (value & 0xff));

    value = mode->vsync_len + mode->upper_margin + mode->lower_margin;
    HDMI_REG_W8(hdmi_dev, FC_INVBLANK0, (value & 0xff));

    value = mode->right_margin;
    if (vpara->color_input == HDMI_COLOR_YCBCR420)
        value = value / 2;
    HDMI_REG_W8(hdmi_dev, FC_HSYNCINDELAY1, v_FC_HSYNCINDEAY1(value >> 8));
    HDMI_REG_W8(hdmi_dev, FC_HSYNCINDELAY0, (value & 0xff));

    value = mode->lower_margin;
    HDMI_REG_W8(hdmi_dev, FC_VSYNCINDELAY0, (value & 0xff));

    value = mode->hsync_len;
    if (vpara->color_input == HDMI_COLOR_YCBCR420)
        value = value / 2;
    HDMI_REG_W8(hdmi_dev, FC_HSYNCINWIDTH1, v_FC_HSYNCWIDTH1(value >> 8));
    HDMI_REG_W8(hdmi_dev, FC_HSYNCINWIDTH0, (value & 0xff));

    value = mode->vsync_len;
    HDMI_REG_W8(hdmi_dev, FC_VSYNCINWIDTH, (value & 0xff));

    /* Set the control period minimum duration (min. of 12 pixel
    * clock cycles, refer to HDMI 1.4b specification)
    */
    HDMI_REG_W8(hdmi_dev, FC_CTRLDUR, 12);
    HDMI_REG_W8(hdmi_dev, FC_EXCTRLDUR, 32);

    /* spacing < 256^2 * config / tmdsClock, spacing <= 50ms
     * worst case: tmdsClock == 25MHz => config <= 19
     */
    HDMI_REG_W8(hdmi_dev, FC_EXCTRLSPAC,
                (hdmi_dev->tmdsclk / 1000) * 50 / (256 * 512));

    HDMI_REG_W8(hdmi_dev, FC_INFREQ0, FC_INFREQ0_VAL(hdmi_dev->tmdsclk / 1000)); //kHz
    HDMI_REG_W8(hdmi_dev, FC_INFREQ1, FC_INFREQ1_VAL(hdmi_dev->tmdsclk / 1000));
    HDMI_REG_W8(hdmi_dev, FC_INFREQ2, FC_INFREQ2_VAL(hdmi_dev->tmdsclk / 1000));

    HDMI_REG_W8(hdmi_dev, FC_PRCONF,
                v_FC_PR_FACTOR(hdmi_dev->pixelrepeat) |
                v_FC_PR_FACTOR_OUT(hdmi_dev->pixelrepeat - 1));

    return 0;
}

/*****************************************************************************
*  函 数 名: ljm_hdmiv2_video_packetizer
*  功能描述: 视频数据包配置
*  输入参数: struct hdmi_dev *hdmi_dev: hdmi设备指针
*            struct hdmi_video *vpara: 输入设置参数
*  输出参数: 无
*  返 回 值: static int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
static int ljm_hdmiv2_video_packetizer(struct hdmi_dev *hdmi_dev, struct hdmi_video *vpara)
{
    unsigned char color_depth = COLOR_DEPTH_24BIT_DEFAULT;
    unsigned char output_select = 0;
    unsigned char remap_size = 0;

    if (vpara->color_output == HDMI_COLOR_YCBCR422) {
        switch (vpara->color_output_depth) {
        case 8:
            remap_size = YCC422_16BIT;
            break;
        case 10:
            remap_size = YCC422_20BIT;
            break;
        case 12:
            remap_size = YCC422_24BIT;
            break;
        default:
            remap_size = YCC422_16BIT;
            break;
        }

        output_select = OUT_FROM_YCC422_REMAP;
        /*Config remap size for the different color Depth*/
        HDMI_REG_MSK(hdmi_dev, VP_REMAP,
                     m_YCC422_SIZE, v_YCC422_SIZE(remap_size));
    } else {
        switch (vpara->color_output_depth) {
        case 10:
            color_depth = COLOR_DEPTH_30BIT;
            output_select = OUT_FROM_PIXEL_PACKING;
            break;
        case 12:
            color_depth = COLOR_DEPTH_36BIT;
            output_select = OUT_FROM_PIXEL_PACKING;
            break;
        case 16:
            color_depth = COLOR_DEPTH_48BIT;
            output_select = OUT_FROM_PIXEL_PACKING;
            break;
        case 8:
        default:
            color_depth = COLOR_DEPTH_24BIT_DEFAULT;
            output_select = OUT_FROM_8BIT_BYPASS;
            break;
        }
    }
    /*Config Color Depth*/
    HDMI_REG_MSK(hdmi_dev, VP_PR_CD,
                 m_COLOR_DEPTH, v_COLOR_DEPTH(color_depth));
    /*Config pixel repettion*/
    HDMI_REG_MSK(hdmi_dev, VP_PR_CD, m_DESIRED_PR_FACTOR,
                 v_DESIRED_PR_FACTOR(hdmi_dev->pixelrepeat - 1));
    if (hdmi_dev->pixelrepeat > 1)
        HDMI_REG_MSK(hdmi_dev, VP_CONF,
                     m_PIXEL_REPET_EN | m_BYPASS_SEL,
                     v_PIXEL_REPET_EN(1) | v_BYPASS_SEL(0));
    else
        HDMI_REG_MSK(hdmi_dev, VP_CONF,
                     m_PIXEL_REPET_EN | m_BYPASS_SEL,
                     v_PIXEL_REPET_EN(0) | v_BYPASS_SEL(1));

    /*config output select*/
    if (output_select == OUT_FROM_PIXEL_PACKING) { /* pixel packing */
        HDMI_REG_MSK(hdmi_dev, VP_CONF,
                     m_BYPASS_EN | m_PIXEL_PACK_EN |
                     m_YCC422_EN | m_OUTPUT_SEL,
                     v_BYPASS_EN(0) | v_PIXEL_PACK_EN(1) |
                     v_YCC422_EN(0) | v_OUTPUT_SEL(output_select));
    } else if (output_select == OUT_FROM_YCC422_REMAP) { /* YCC422 */
        HDMI_REG_MSK(hdmi_dev, VP_CONF,
                     m_BYPASS_EN | m_PIXEL_PACK_EN |
                     m_YCC422_EN | m_OUTPUT_SEL,
                     v_BYPASS_EN(0) | v_PIXEL_PACK_EN(0) |
                     v_YCC422_EN(1) | v_OUTPUT_SEL(output_select));
    } else if (output_select == OUT_FROM_8BIT_BYPASS ||
               output_select == 3) { /* bypass */
        HDMI_REG_MSK(hdmi_dev, VP_CONF,
                     m_BYPASS_EN | m_PIXEL_PACK_EN |
                     m_YCC422_EN | m_OUTPUT_SEL,
                     v_BYPASS_EN(1) | v_PIXEL_PACK_EN(0) |
                     v_YCC422_EN(0) | v_OUTPUT_SEL(output_select));
    }

#if defined(HDMI_VIDEO_STUFFING)
    /* YCC422 and pixel packing stuffing */
    HDMI_REG_MSK(hdmi_dev, VP_STUFF, m_PR_STUFFING, v_PR_STUFFING(1));
    HDMI_REG_MSK(hdmi_dev, VP_STUFF,
                 m_YCC422_STUFFING | m_PP_STUFFING,
                 v_YCC422_STUFFING(1) | v_PP_STUFFING(1));
#endif
    return 0;
}
/*****************************************************************************
*  函 数 名: ljm_hdmiv2_video_sampler
*  功能描述: 视频采样设置
*  输入参数: struct hdmi_dev *hdmi_dev: hdmi设备指针
*            struct hdmi_video *vpara: 输入设置参数
*  输出参数: 无
*  返 回 值: static int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
static int ljm_hdmiv2_video_sampler(struct hdmi_dev *hdmi_dev, struct hdmi_video *vpara)
{
    int map_code = 0;

    if (vpara->color_input == HDMI_COLOR_YCBCR422) {
        /* YCC422 mapping is discontinued - only map 1 is supported */
        switch (vpara->color_output_depth) {
        case 8:
            map_code = VIDEO_YCBCR422_8BIT;
            break;
        case 10:
            map_code = VIDEO_YCBCR422_10BIT;
            break;
        case 12:
            map_code = VIDEO_YCBCR422_12BIT;
            break;
        default:
            map_code = VIDEO_YCBCR422_8BIT;
            break;
        }
    } else if (vpara->color_input == HDMI_COLOR_YCBCR420 || vpara->color_input == HDMI_COLOR_YCBCR444) {
        switch (vpara->color_output_depth) {
        case 10:
            map_code = VIDEO_YCBCR444_10BIT;
            break;
        case 12:
            map_code = VIDEO_YCBCR444_12BIT;
            break;
        case 16:
            map_code = VIDEO_YCBCR444_16BIT;
            break;
        case 8:
        default:
            map_code = VIDEO_YCBCR444_8BIT;
            break;
        }
    } else {
        switch (vpara->color_output_depth) {
        case 10:
            map_code = VIDEO_RGB444_10BIT;
            break;
        case 12:
            map_code = VIDEO_RGB444_12BIT;
            break;
        case 16:
            map_code = VIDEO_RGB444_16BIT;
            break;
        case 8:
        default:
            map_code = VIDEO_RGB444_8BIT;
            break;
        }
    }

    /* Set Data enable signal from external
     * and set video sample input mapping
     */
    HDMI_REG_MSK(hdmi_dev, TX_INVID0,
                 m_INTERNAL_DE_GEN | m_VIDEO_MAPPING,
                 v_INTERNAL_DE_GEN(0) | v_VIDEO_MAPPING(map_code));

#if defined(HDMI_VIDEO_STUFFING)
    HDMI_REG_W8(hdmi_dev, TX_GYDATA0, 0x00);
    HDMI_REG_W8(hdmi_dev, TX_GYDATA1, 0x00);
    HDMI_REG_MSK(hdmi_dev, TX_INSTUFFING,
                 m_GYDATA_STUFF, v_GYDATA_STUFF(1));
    HDMI_REG_W8(hdmi_dev, TX_RCRDATA0, 0x00);
    HDMI_REG_W8(hdmi_dev, TX_RCRDATA1, 0x00);
    HDMI_REG_MSK(hdmi_dev, TX_INSTUFFING,
                 m_RCRDATA_STUFF, v_RCRDATA_STUFF(1));
    HDMI_REG_W8(hdmi_dev, TX_BCBDATA0, 0x00);
    HDMI_REG_W8(hdmi_dev, TX_BCBDATA1, 0x00);
    HDMI_REG_MSK(hdmi_dev, TX_INSTUFFING,
                 m_BCBDATA_STUFF, v_BCBDATA_STUFF(1));
#endif
    return 0;
}

static const char coeff_csc[][24] = {
        /*   G        R        B        Bias
         *   A1    |    A2     |    A3     |    A4    |
         *   B1    |    B2     |    B3     |    B4    |
         *   C1    |    C2     |    C3     |    C4    |
         */
    {   /* CSC_BYPASS */
        0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x00, 0x00,
        0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x00, 0x00,
        0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x00, 0x00,
    },
    {   /* CSC_RGB_0_255_TO_RGB_16_235_8BIT */
        0x36, 0xf7, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40,        /*G*/
        0x00, 0x00, 0x36, 0xf7, 0x00, 0x00, 0x00, 0x40,        /*R*/
        0x00, 0x00, 0x00, 0x00, 0x36, 0xf7, 0x00, 0x40,        /*B*/
    },
    {   /* CSC_RGB_0_255_TO_RGB_16_235_10BIT */
        0x36, 0xf7, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00,        /*G*/
        0x00, 0x00, 0x36, 0xf7, 0x00, 0x00, 0x01, 0x00,        /*R*/
        0x00, 0x00, 0x00, 0x00, 0x36, 0xf7, 0x01, 0x00,        /*B*/
    },
    {   /* CSC_RGB_0_255_TO_ITU601_16_235_8BIT */
        0x20, 0x40, 0x10, 0x80, 0x06, 0x40, 0x00, 0x40,        /*Y*/
        0xe8, 0x80, 0x1c, 0x00, 0xfb, 0x80, 0x02, 0x00,        /*Cr*/
        0xed, 0x80, 0xf6, 0x80, 0x1c, 0x00, 0x02, 0x00,        /*Cb*/
    },
    {   /* CSC_RGB_0_255_TO_ITU601_16_235_10BIT */
        0x20, 0x40, 0x10, 0x80, 0x06, 0x40, 0x01, 0x00,        /*Y*/
        0xe8, 0x80, 0x1c, 0x00, 0xfb, 0x80, 0x08, 0x00,        /*Cr*/
        0xed, 0x80, 0xf6, 0x80, 0x1c, 0x00, 0x08, 0x00,        /*Cb*/
    },
    {   /* CSC_RGB_0_255_TO_ITU709_16_235_8BIT */
        0x27, 0x40, 0x0b, 0xc0, 0x04, 0x00, 0x00, 0x40,        /*Y*/
        0xe6, 0x80, 0x1c, 0x00, 0xfd, 0x80, 0x02, 0x00,        /*Cr*/
        0xea, 0x40, 0xf9, 0x80, 0x1c, 0x00, 0x02, 0x00,        /*Cb*/
    },
    {   /* CSC_RGB_0_255_TO_ITU709_16_235_10BIT */
        0x27, 0x40, 0x0b, 0xc0, 0x04, 0x00, 0x01, 0x00,        /*Y*/
        0xe6, 0x80, 0x1c, 0x00, 0xfd, 0x80, 0x08, 0x00,        /*Cr*/
        0xea, 0x40, 0xf9, 0x80, 0x1c, 0x00, 0x08, 0x00,        /*Cb*/
    },
    /* Y        Cr        Cb        Bias */
    {   /* CSC_ITU601_16_235_TO_RGB_0_255_8BIT */
        0x20, 0x00, 0x69, 0x26, 0x74, 0xfd, 0x01, 0x0e,        /*G*/
        0x20, 0x00, 0x2c, 0xdd, 0x00, 0x00, 0x7e, 0x9a,        /*R*/
        0x20, 0x00, 0x00, 0x00, 0x38, 0xb4, 0x7e, 0x3b,        /*B*/
    },
    {   /* CSC_ITU709_16_235_TO_RGB_0_255_8BIT */
        0x20, 0x00, 0x71, 0x06, 0x7a, 0x02, 0x00, 0xa7,        /*G*/
        0x20, 0x00, 0x32, 0x64, 0x00, 0x00, 0x7e, 0x6d,        /*R*/
        0x20, 0x00, 0x00, 0x00, 0x3b, 0x61, 0x7e, 0x25,        /*B*/
    },
};
/*****************************************************************************
*  函 数 名: ljm_hdmiv2_video_csc
*  功能描述: 色彩空间处理
*  输入参数: struct hdmi_dev *hdmi_dev: hdmi设备指针
*            struct hdmi_video *vpara: 输入要配置的参数
*  输出参数: 无
*  返 回 值: static int
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
static int ljm_hdmiv2_video_csc(struct hdmi_dev *hdmi_dev, struct hdmi_video *vpara)
{
    int i, mode, interpolation, decimation, csc_scale = 0;
    const char *coeff = NULL;
    unsigned char color_depth = 0;

    if (vpara->color_input == vpara->color_output) {
        HDMI_REG_MSK(hdmi_dev, MC_FLOWCTRL,
                     m_FEED_THROUGH_OFF, v_FEED_THROUGH_OFF(0));
        return 0;
    }

    if (vpara->color_input == HDMI_COLOR_YCBCR422 &&
        vpara->color_output != HDMI_COLOR_YCBCR422 &&
        vpara->color_output != HDMI_COLOR_YCBCR420) {
        interpolation = 1;
        HDMI_REG_MSK(hdmi_dev, CSC_CFG,
                     m_CSC_INTPMODE, v_CSC_INTPMODE(interpolation));
    }

    if ((vpara->color_input == HDMI_COLOR_RGB_0_255 ||
         vpara->color_input == HDMI_COLOR_YCBCR444) &&
         vpara->color_output == HDMI_COLOR_YCBCR422) {
        decimation = 1;
        HDMI_REG_MSK(hdmi_dev, CSC_CFG,
                     m_CSC_DECIMODE, v_CSC_DECIMODE(decimation));
    }

    mode = CSC_BYPASS;
    csc_scale = 0;

    switch (vpara->vic) {
    case HDMI_720X480I_60HZ_4_3:
    case HDMI_720X576I_50HZ_4_3:
    case HDMI_720X480P_60HZ_4_3:
    case HDMI_720X576P_50HZ_4_3:
    case HDMI_720X480I_60HZ_16_9:
    case HDMI_720X576I_50HZ_16_9:
    case HDMI_720X480P_60HZ_16_9:
    case HDMI_720X576P_50HZ_16_9:
        if (vpara->color_input == HDMI_COLOR_RGB_0_255 &&
            vpara->color_output >= HDMI_COLOR_YCBCR444) {
            mode = CSC_RGB_0_255_TO_ITU601_16_235_8BIT;
            csc_scale = 0;
        } else if (vpara->color_input >= HDMI_COLOR_YCBCR444 &&
                   vpara->color_output == HDMI_COLOR_RGB_0_255) {
            mode = CSC_ITU601_16_235_TO_RGB_0_255_8BIT;
            csc_scale = 1;
        }
        break;
    default:
        if (vpara->color_input == HDMI_COLOR_RGB_0_255 &&
            vpara->color_output >= HDMI_COLOR_YCBCR444) {
            mode = CSC_RGB_0_255_TO_ITU709_16_235_8BIT;
            csc_scale = 0;
        } else if (vpara->color_input >= HDMI_COLOR_YCBCR444 &&
                   vpara->color_output == HDMI_COLOR_RGB_0_255) {
            mode = CSC_ITU709_16_235_TO_RGB_0_255_8BIT;
            csc_scale = 1;
        }
        break;
    }

    if ((vpara->color_input == HDMI_COLOR_RGB_0_255) &&
        (vpara->color_output == HDMI_COLOR_RGB_16_235)) {
        mode = CSC_RGB_0_255_TO_RGB_16_235_8BIT;
        csc_scale = 0;
    }
    if (mode != CSC_BYPASS) {
        switch (vpara->color_output_depth) {
        case 10:
            color_depth = COLOR_DEPTH_30BIT;
            mode += 1;
            break;
        case 12:
            color_depth = COLOR_DEPTH_36BIT;
            mode += 2;
            break;
        case 16:
            color_depth = COLOR_DEPTH_48BIT;
            mode += 3;
            break;
        case 8:
        default:
            color_depth = COLOR_DEPTH_24BIT;
            break;
        }
    }
    coeff = coeff_csc[mode];
    for (i = 0; i < 24; i++)
        HDMI_REG_W8(hdmi_dev, CSC_COEF_A1_MSB + i, coeff[i]);

    HDMI_REG_MSK(hdmi_dev, CSC_SCALE,
                 m_CSC_SCALE, v_CSC_SCALE(csc_scale));
    /*config CSC_COLOR_DEPTH*/
    HDMI_REG_MSK(hdmi_dev, CSC_SCALE,
                 m_CSC_COLOR_DEPTH, v_CSC_COLOR_DEPTH(color_depth));

    /* enable CSC */
    if (mode == CSC_BYPASS)
        HDMI_REG_MSK(hdmi_dev, MC_FLOWCTRL,
                     m_FEED_THROUGH_OFF, v_FEED_THROUGH_OFF(0));
    else
        HDMI_REG_MSK(hdmi_dev, MC_FLOWCTRL,
                     m_FEED_THROUGH_OFF, v_FEED_THROUGH_OFF(1));
    return 0;
}

static void hdmi_dev_config_avi(struct hdmi_dev *hdmi_dev, struct hdmi_video *vpara)
{
    unsigned char colorimetry, ext_colorimetry = 0, aspect_ratio, y1y0;
    unsigned char rgb_quan_range = AVI_QUANTIZATION_RANGE_DEFAULT;

    HDMI_REG_MSK(hdmi_dev, FC_DATAUTO3, m_AVI_AUTO, v_AVI_AUTO(0));
    HDMI_REG_MSK(hdmi_dev, IH_FC_STAT1,
                 m_AVI_INFOFRAME, v_AVI_INFOFRAME(1));
    /* Set AVI infoFrame Data byte1 */
    if (vpara->color_output == HDMI_COLOR_YCBCR444)
        y1y0 = AVI_COLOR_MODE_YCBCR444;
    else if (vpara->color_output == HDMI_COLOR_YCBCR422)
        y1y0 = AVI_COLOR_MODE_YCBCR422;
    else if (vpara->color_output == HDMI_COLOR_YCBCR420)
        y1y0 = AVI_COLOR_MODE_YCBCR420;
    else
        y1y0 = AVI_COLOR_MODE_RGB;

    HDMI_REG_MSK(hdmi_dev, FC_AVICONF0,
                 m_FC_ACTIV_FORMAT | m_FC_RGC_YCC,
                 v_FC_RGC_YCC(y1y0) | v_FC_ACTIV_FORMAT(1));

#if 0
    /* Set AVI infoFrame Data byte2 */
    switch (vpara->vic) {
    case HDMI_720X480I_60HZ_4_3:
    case HDMI_720X576I_50HZ_4_3:
    case HDMI_720X480P_60HZ_4_3:
    case HDMI_720X576P_50HZ_4_3:
        aspect_ratio = AVI_CODED_FRAME_ASPECT_4_3;
        if (vpara->colorimetry == HDMI_COLORIMETRY_NO_DATA)
            colorimetry = AVI_COLORIMETRY_SMPTE_170M;
        break;
    case HDMI_720X480I_60HZ_16_9:
    case HDMI_720X576I_50HZ_16_9:
    case HDMI_720X480P_60HZ_16_9:
    case HDMI_720X576P_50HZ_16_9:
        aspect_ratio = AVI_CODED_FRAME_ASPECT_16_9;
        if (vpara->colorimetry == HDMI_COLORIMETRY_NO_DATA)
            colorimetry = AVI_COLORIMETRY_SMPTE_170M;
        break;
    default:
        aspect_ratio = AVI_CODED_FRAME_ASPECT_16_9;
        if (vpara->colorimetry == HDMI_COLORIMETRY_NO_DATA)
            colorimetry = AVI_COLORIMETRY_ITU709;
    }
#else
    aspect_ratio = AVI_CODED_FRAME_ASPECT_16_9;
    colorimetry = AVI_COLORIMETRY_ITU709;
#endif

#if 0
    if (vpara->colorimetry > HDMI_COLORIMETRY_ITU709) {
        colorimetry = AVI_COLORIMETRY_EXTENDED;
        ext_colorimetry = vpara->colorimetry -
            HDMI_COLORIMETRY_EXTEND_XVYCC_601;
    } else if (vpara->color_output == HDMI_COLOR_RGB_16_235 ||
         vpara->color_output == HDMI_COLOR_RGB_0_255) {
        colorimetry = AVI_COLORIMETRY_NO_DATA;
    } else if (vpara->colorimetry != HDMI_COLORIMETRY_NO_DATA) {
        colorimetry = vpara->colorimetry;
    }
#else
    if (vpara->color_output == HDMI_COLOR_RGB_16_235 ||
         vpara->color_output == HDMI_COLOR_RGB_0_255) {
        colorimetry = AVI_COLORIMETRY_NO_DATA;
    } else if (vpara->colorimetry != 0) {
        colorimetry = vpara->colorimetry;
    }
#endif

    HDMI_REG_W8(hdmi_dev, FC_AVICONF1,
                v_FC_COLORIMETRY(colorimetry) |
                v_FC_PIC_ASPEC_RATIO(aspect_ratio) |
                v_FC_ACT_ASPEC_RATIO(ACTIVE_ASPECT_RATE_DEFAULT));

    /* Set AVI infoFrame Data byte3 */
    HDMI_REG_MSK(hdmi_dev, FC_AVICONF2,
                 m_FC_EXT_COLORIMETRY | m_FC_QUAN_RANGE,
                 v_FC_EXT_COLORIMETRY(ext_colorimetry) |
                 v_FC_QUAN_RANGE(rgb_quan_range));

    /* Set AVI infoFrame Data byte4 */
    if ((vpara->vic > 92 && vpara->vic < 96) ||
        (vpara->vic == 98) ||
        (vpara->vic & HDMI_VIDEO_DMT) ||
        (vpara->vic & HDMI_VIDEO_DISCRETE_VR))
        HDMI_REG_W8(hdmi_dev, FC_AVIVID, 0);
    else
        HDMI_REG_W8(hdmi_dev, FC_AVIVID, vpara->vic & 0xff);
    /* Set AVI infoFrame Data byte5 */
    HDMI_REG_MSK(hdmi_dev, FC_AVICONF3, m_FC_YQ | m_FC_CN,
                 v_FC_YQ(YQ_LIMITED_RANGE) | v_FC_CN(CN_GRAPHICS));
    HDMI_REG_MSK(hdmi_dev, FC_DATAUTO3, m_AVI_AUTO, v_AVI_AUTO(1));
}

static int hdmi_dev_config_vsi(struct hdmi_dev *hdmi_dev, unsigned char vic_3d, unsigned char format)
{
    int i = 0, id = 0x000c03;
    unsigned char data[3] = {0};

    TRACE("[%s] vic %d format %d.\n", __FUNCTION__, vic_3d, format);

    HDMI_REG_MSK(hdmi_dev, FC_DATAUTO0, m_VSD_AUTO, v_VSD_AUTO(0));
    HDMI_REG_W8(hdmi_dev, FC_VSDIEEEID0, id & 0xff);
    HDMI_REG_W8(hdmi_dev, FC_VSDIEEEID1, (id >> 8) & 0xff);
    HDMI_REG_W8(hdmi_dev, FC_VSDIEEEID2, (id >> 16) & 0xff);

    data[0] = format << 5;    /* PB4 --HDMI_Video_Format */
    switch (format) {
    case HDMI_VIDEO_FORMAT_4KX2K:
        data[1] = vic_3d;    /* PB5--HDMI_VIC */
        data[2] = 0;
        break;
    case HDMI_VIDEO_FORMAT_3D:
        data[1] = vic_3d << 4;    /* PB5--3D_Structure field */
        data[2] = 0;        /* PB6--3D_Ext_Data field */
        break;
    default:
        data[1] = 0;
        data[2] = 0;
        break;
    }

    for (i = 0; i < 3; i++)
        HDMI_REG_W8(hdmi_dev, FC_VSDPAYLOAD0 + i, data[i]);
    HDMI_REG_W8(hdmi_dev, FC_VSDSIZE, 0x6);

    HDMI_REG_W8(hdmi_dev, FC_DATAUTO1, 0);
    HDMI_REG_W8(hdmi_dev, FC_DATAUTO2, 0x11);
    HDMI_REG_MSK(hdmi_dev, FC_DATAUTO0, m_VSD_AUTO, v_VSD_AUTO(1));
    return 0;
}

#define HDR_LSB(n) ((n) & 0xff)
#define HDR_MSB(n) (((n) & 0xff00) >> 8)

/*static*/ void hdmi_dev_config_hdr(struct hdmi_dev *hdmi_dev, int eotf, struct hdmi_hdr_metadata *hdr)
{
    /* hdr is supportted after disignid = 0x21 */
    if (!hdmi_dev || HDMI_REG_R8(hdmi_dev, DESIGN_ID) < 0x21)
        return;

    HDMI_REG_W8(hdmi_dev, FC_DRM_HB, 1);/*verion = 0x1*/
    HDMI_REG_W8(hdmi_dev, (FC_DRM_HB + 1), 26);/*length of following data*/
    HDMI_REG_W8(hdmi_dev, FC_DRM_PB, eotf / 2);
    HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 1, 0);

    if (hdr) {
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 2, HDR_LSB(hdr->prim_x0));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 3, HDR_MSB(hdr->prim_x0));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 4, HDR_LSB(hdr->prim_y0));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 5, HDR_MSB(hdr->prim_y0));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 6, HDR_LSB(hdr->prim_x1));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 7, HDR_MSB(hdr->prim_x1));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 8, HDR_LSB(hdr->prim_y1));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 9, HDR_MSB(hdr->prim_y1));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 10, HDR_LSB(hdr->prim_x2));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 11, HDR_MSB(hdr->prim_x2));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 12, HDR_LSB(hdr->prim_y2));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 13, HDR_MSB(hdr->prim_y2));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 14, HDR_LSB(hdr->white_px));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 15, HDR_MSB(hdr->white_px));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 16, HDR_LSB(hdr->white_py));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 17, HDR_MSB(hdr->white_py));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 18, HDR_LSB(hdr->max_dml));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 19, HDR_MSB(hdr->max_dml));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 20, HDR_LSB(hdr->min_dml));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 21, HDR_MSB(hdr->min_dml));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 22, HDR_LSB(hdr->max_cll));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 23, HDR_MSB(hdr->max_cll));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 24, HDR_LSB(hdr->max_fall));
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 25, HDR_MSB(hdr->max_fall));
    } else {
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 1, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 2, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 3, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 4, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 5, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 6, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 7, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 8, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 9, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 10, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 11, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 12, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 13, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 14, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 15, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 16, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 17, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 18, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 19, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 20, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 21, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 22, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 23, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 24, 0);
        HDMI_REG_W8(hdmi_dev, FC_DRM_PB + 25, 0);
    }
    if (eotf) {
        HDMI_REG_MSK(hdmi_dev, FC_PACK_TXE, m_DRM_TXEN, v_DRM_TXEN(1));
        HDMI_REG_MSK(hdmi_dev, FC_MASK2, m_DRM_MASK, v_DRM_MASK(0));
        HDMI_REG_MSK(hdmi_dev, FC_DRM_UP, m_DRM_PUPD, v_DRM_PUPD(1));
    } else {
        HDMI_REG_MSK(hdmi_dev, FC_PACK_TXE, m_DRM_TXEN, v_DRM_TXEN(0));
        HDMI_REG_MSK(hdmi_dev, FC_MASK2, m_DRM_MASK, v_DRM_MASK(1));
        HDMI_REG_MSK(hdmi_dev, FC_DRM_UP, m_DRM_PUPD, v_DRM_PUPD(1));
    }
}

static int hdmi_dev_config_spd(struct hdmi_dev *hdmi_dev, const char *vendor,
                   const char *product, char deviceinfo)
{
    int i, len;

    if (!hdmi_dev || !vendor || !product)
        return -1;

    HDMI_REG_MSK(hdmi_dev, FC_DATAUTO0, m_SPD_AUTO, v_SPD_AUTO(0));
    len = strlen(vendor);
    for (i = 0; i < 8; i++) {
        if (i < len)
            HDMI_REG_W8(hdmi_dev, FC_SPDVENDORNAME0 + i,
                    vendor[i]);
        else
            HDMI_REG_W8(hdmi_dev, FC_SPDVENDORNAME0 + i,
                    0);
    }
    len = strlen(product);
    for (i = 0; i < 16; i++) {
        if (i < len)
            HDMI_REG_W8(hdmi_dev, FC_SPDPRODUCTNAME0 + i,
                    product[i]);
        else
            HDMI_REG_W8(hdmi_dev, FC_SPDPRODUCTNAME0 + i,
                    0);
    }
    HDMI_REG_W8(hdmi_dev, FC_SPDDEVICEINF, deviceinfo);
    HDMI_REG_MSK(hdmi_dev, FC_DATAUTO0, m_SPD_AUTO, v_SPD_AUTO(1));
    return 0;
}


/**
 * ljm_hdmiv2_mode_set - init hdmi controller for GP200
 * @hdmi_dev: pointer to hdmi device
 * @vpara: hdmi video parameters
 *
 * This function sets dwc controller to a specified mode.
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_mode_set(struct hdmi_dev *hdmi_dev, struct hdmi_video *vpara)
{

    HEADER_ARG("hdmi %d: pixclock=%d xres=%d right_margin=%d hsync_len=%d left_margin=%d\n"
        "yres=%d low_margin=%d vsync_len=%d upper_margin=%d sync=%d hdmi=%s",
        hdmi_dev->index, vpara->mode.pixclock, vpara->mode.xres, vpara->mode.right_margin, vpara->mode.hsync_len, vpara->mode.left_margin,
        vpara->mode.yres, vpara->mode.lower_margin, vpara->mode.vsync_len, vpara->mode.upper_margin, vpara->mode.sync, vpara->sink_hdmi ? "Y" : "N");

    TRACE("%s vic %d 3dformat %d color mode %d color depth %d\n",
        __FUNCTION__, vpara->vic, vpara->format_3d, vpara->color_output, vpara->color_output_depth);

    /* clear interrupt */
    //HDMI_REG_W8(hdmi_dev, IH_MUTE, 0x0);
    TRACE("hdmi %d: IH_PHY_STAT0=%x", hdmi_dev->index, HDMI_REG_R8(hdmi_dev, IH_PHY_STAT0));

    /* before configure video, we power off phy */
    HDMI_REG_MSK(hdmi_dev, PHY_CONF0,
                 m_PDDQ_SIG | m_TXPWRON_SIG,
                 v_PDDQ_SIG(1) | v_TXPWRON_SIG(0));

    //HDMI_REG_W8(hdmi_dev, MC_CLKDIS, m_HDCPCLK_DISABLE);

    if (ljm_hdmiv2_video_framecomposer(hdmi_dev, vpara) < 0) {
        FOOTER_ARG("r=%d", -1);
        return -1;
    }

    if (ljm_hdmiv2_video_packetizer(hdmi_dev, vpara) < 0) {
        FOOTER_ARG("r=%d", -1);
        return -1;
    }

    /* Color space convert */
    if (ljm_hdmiv2_video_csc(hdmi_dev, vpara) < 0) {
        FOOTER_ARG("r=%d", -1);
        return -1;
    }

    if (ljm_hdmiv2_video_sampler(hdmi_dev, vpara) < 0) {
        FOOTER_ARG("r=%d", -1);
        return -1;
    }

    if (vpara->sink_hdmi == OUTPUT_HDMI) {
        hdmi_dev_config_avi(hdmi_dev, vpara);
        hdmi_dev_config_spd(hdmi_dev, hdmi_dev->vendor_name, hdmi_dev->product_name, hdmi_dev->deviceinfo);
        if (vpara->format_3d != HDMI_3D_NONE) {
            hdmi_dev_config_vsi(hdmi_dev, vpara->format_3d, HDMI_VIDEO_FORMAT_3D);
        } else if ((vpara->vic > 92 && vpara->vic < 96) || (vpara->vic == 98)) {
            vpara->vic = (vpara->vic == 98) ? 4 : (96 - vpara->vic);
            hdmi_dev_config_vsi(hdmi_dev, vpara->vic, HDMI_VIDEO_FORMAT_4KX2K);
        } else {
            hdmi_dev_config_vsi(hdmi_dev, vpara->vic, HDMI_VIDEO_FORMAT_NORMAL);
        }
        //hdmi_dev_config_hdr(hdmi_dev, vpara->eotf, &hdmi_dev->hdr);
        TRACE("[%s] success output HDMI.\n", __FUNCTION__);
    } else {
        TRACE("[%s] success output DVI.\n", __FUNCTION__);
    }

    ljm_hdmiv2_config_phy(hdmi_dev);

    FOOTER();

    return 0;
}
/*****************************************************************************
*  函 数 名: ljm_hdmiv2_dpms
*  功能描述: dpms电源管理
*  输入参数: struct hdmi_dev *hdmi_dev  :hdmi设备对象
*            int mode: dpms状态
*  输出参数: 无
*  返 回 值: void
*
*  修改历史:
*         1. 日    期: 2022年4月21日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
void ljm_hdmiv2_dpms(struct hdmi_dev *hdmi_dev, int mode)
{
    u16 tmp;

    HEADER_ARG("hdmi_dev=%lx mode=%d", (unsigned long)hdmi_dev, mode);

    ljm_hdmiv2_read_phy(hdmi_dev, PHY_PWRCTRL_OFFSET, &tmp);
    tmp |= m_PHY_PWRCTRL_OVERRIDE;
    switch (mode) {
    case DRM_MODE_DPMS_ON:
        hdmi_dev->dpms_state = DRM_MODE_DPMS_ON;
        /* power up hdmi phy */
        tmp |= m_PHY_PWRCTRL_BIAS_ON | m_PHY_PWRCTRL_TX_PWRON
            | m_PHY_PWRCTRL_TX_PWRON0 | m_PHY_PWRCTRL_TX_PWRON1 | m_PHY_PWRCTRL_TX_PWRON2 | m_PHY_PWRCTRL_CK_PWRON;
        break;
    case DRM_MODE_DPMS_STANDBY:
    case DRM_MODE_DPMS_SUSPEND:
    case DRM_MODE_DPMS_OFF:
        hdmi_dev->dpms_state = DRM_MODE_DPMS_OFF;
        /* power down hdmi phy */
        tmp &= ~m_PHY_PWRCTRL_BIAS_ON & ~m_PHY_PWRCTRL_TX_PWRON
            & ~m_PHY_PWRCTRL_TX_PWRON0 & ~m_PHY_PWRCTRL_TX_PWRON1 & ~m_PHY_PWRCTRL_TX_PWRON2 & ~m_PHY_PWRCTRL_CK_PWRON;
        break;
    }
    ljm_hdmiv2_write_phy(hdmi_dev, PHY_PWRCTRL_OFFSET, tmp);

    FOOTER();
}

bool ljm_hdmiv2_ddc_detect(struct hdmi_dev *hdmi_dev)
{
    int i;
    unsigned char const header[8] = {0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00};
    unsigned char val;

    for(i=0; i<8; i++)
    {
        if(ljm_hdmiv2_i2cm_read(hdmi_dev, DDC_EDID_ADDR, i, &val))
            goto error;
        if(val != header[i])
            goto error;
    }

    DEBUG_PRINT("[%s]DDC is connected\n", __FUNCTION__);
    return true;

error:
    DEBUG_PRINT("[%s]DDC is disconnected\n", __FUNCTION__);
    return false;
}

/**
 * ljm_hdmiv2_hpd_sense - detect whether hdmi is connected
 * @hdmi_dev: pointer to hdmi device
 *
 * This function detect hdmi connector status
 *
 * Returns:
 * true on connected or false.
 */
bool ljm_hdmiv2_hpd_sense(struct hdmi_dev *hdmi_dev)
{
    u8 status;
    status = HDMI_REG_R8(hdmi_dev, PHY_STAT0);
    if (status & m_PHY_STAT0_HPD)/* check hpd */
        return true;
    //else if(status & (m_PHY_STAT0_RX_SENSE_3 | m_PHY_STAT0_RX_SENSE_2 | m_PHY_STAT0_RX_SENSE_1_MASK | m_PHY_STAT0_RX_SENSE_0_MASK))
        //return true;
    /* 在 HDMI 拔出时，存在依然能读取到 edid 并返回连接状态为 true 的错误情况 。因此将此判断条件去掉 */
    // else if (ljm_hdmiv2_ddc_detect(hdmi_dev))
    //     return true;
    else
        return false;
}

/**
 * ljm_hdmiv2_get_edid - get EDID data
 * @hdmi_dev: pointer to hdmi device
 * @psize: pointer to edid size
 *
 * This function read monitor's edid by ddc
 *
 * Return: Pointer to valid EDID or NULL if we couldn't find any.
 */
struct edid * ljm_hdmiv2_get_edid(struct hdmi_dev *hdmi_dev, int *psize)
{
    int i;
    u8 *block, *new;
    int size;

    HEADER_ARG("hdmi_dev=%lx", (unsigned long)hdmi_dev);

/*#ifdef UEFI_GOP
    FIXME: todo for UEFI GOP
#else
*/
    if ((block = kzalloc(DDC_EDID_SIZE, GFP_KERNEL)) == NULL) {
/*#endif*/
        if (psize) *psize = 0;
        ERR_PRINT("out of memory\n");
        FOOTER_ARG("%lx", (unsigned long)block);
        return NULL;
    }
    size = DDC_EDID_SIZE;

    for (i=0; i<size; i++) {    /* block 0 */
        if (ljm_hdmiv2_i2cm_read(hdmi_dev, DDC_EDID_ADDR, (u8)i, &block[i]))
            goto OnError;
    }

    if (!drm_edid_block_valid(block, 0, true, NULL)){
        ERR_PRINT("hdmi %d: edid block 0 is invalid\n", hdmi_dev->index);
        goto OnError;
    }

    if (block[DDC_EDID_SIZE-2] == 0)    /* there is no extensions, ddc have done */
        goto OnSuccess;

    /* read extension of EDID */
    size = (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE;
/*#ifdef UEFI_GOP
    FIXME: todo for UEFI GOP
#else
*/
    new = krealloc(block, size, GFP_KERNEL);
/*#endif*/
    if (!new) {
        goto OnError;
    }
    block = new;

    for (; i<size; i++) {
        if (i < 256) {  /* block 0 */
            if (ljm_hdmiv2_i2cm_read(hdmi_dev, DDC_EDID_ADDR, (u8)i, &block[i]))
                goto OnError;
        }
        else {  /* block i/256 */
            if (ljm_hdmiv2_i2cm_read_ext(hdmi_dev, (u8)(i/256), DDC_EDID_ADDR, (u8)(i & 0xff), &block[i]))
                goto OnError;
        }
    }

    /* check EDID */
    if(!drm_edid_is_valid((struct edid *)block)) {
        ERR_PRINT("hdmi %d: edid is invalid\n", hdmi_dev->index);
        goto OnError;
    }

OnSuccess:
#ifdef DEBUG
    {
        int i, offset;
        unsigned char data;
        char line[] = "       0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f \n";
        char star[] = "******************************************************\n";

        /* dump EDID */
        printk("Dump EDID with %d bytes\n", (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE);
        printk(star);
        for (i = 0, offset = 0; i < (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE; ++i) {
            if (i % 16 == 0) {
                printk(line);
                offset = 0;
                offset += sprintf(line + offset, "%04x: ", i);  /* print address */
            }

            data = block[i];
            offset += sprintf(line + offset, "%02x ", data);
        }

        if (offset) {
            line[offset] = '\0';
            printk(line);
        }
    }
#endif

    if (psize) *psize = size;
    FOOTER_ARG("%lx", (unsigned long)block);
    return (struct edid *)block;

OnError:
    ERR_PRINT("hdmi %d: read edid failed\n", hdmi_dev->index);
    kfree(block);
    if (psize) *psize = 0;

    FOOTER();
    return NULL;
}

/*****************************************************************************
 *  函 数 名：ljm_hdmiv2_edid_checksum
 *  功能描述：edid校验和检查
 *  输入参数：unsigned char *buf
 *  输出参数：无
 *  返 回 值：static int 校验正确：0 校验错误：-1
 *
 *  修改历史：
 *         1. 日    期：2022年7月7日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
static int ljm_hdmiv2_edid_checksum(unsigned char *buf)
{
    int i;
    int checksum = 0;

    for (i = 0; i < DDC_EDID_SIZE; i++)
        checksum += buf[i];

    checksum &= 0xff;

    if (checksum == 0)
        return 0;
    else
        return -1;
}

/*****************************************************************************
 *  函 数 名：ljm_hdmiv2_edmi_parse_vsdb
 *  功能描述：解析Vendor-Specific Data  Block
 *  输入参数：unsigned char *buf
 *            struct hdmi_edid *pedid
 *            int cur_offset
 *            int IEEEOUI
 *  输出参数：无
 *  返 回 值：static int
 *
 *  修改历史：
 *         1. 日    期：2022年7月7日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
static int ljm_hdmiv2_edmi_parse_vsdb(unsigned char *buf, struct hdmi_edid *pedid,
                int cur_offset, int IEEEOUI)
{
    int count, buf_offset;

    HEADER_ARG("IEEEOUI: 0x%x", IEEEOUI);

    count = buf[cur_offset] & 0x1F;
    switch (IEEEOUI) {
    case 0x0c03:
        pedid->sink_hdmi = 1;
        pedid->cecaddress = buf[cur_offset + 5];
        pedid->cecaddress |= buf[cur_offset + 4] << 8;
        DEBUG_PRINT("[CEA] CEC Physical address is 0x%08x.\n",
             pedid->cecaddress);
        if (count > 6)
            pedid->deepcolor = (buf[cur_offset + 6] >> 3) & 0x0F;
        if (count > 7) {
            pedid->maxtmdsclock = buf[cur_offset + 7] * 5000000;
            DEBUG_PRINT("[CEA] maxtmdsclock is %d.\n",
                 pedid->maxtmdsclock);
        }
        if (count > 8) {
            pedid->fields_present = buf[cur_offset + 8];
            DEBUG_PRINT("[CEA] fields_present is 0x%02x.\n",
                 pedid->fields_present);
        }
        buf_offset = cur_offset + 9;
        if (pedid->fields_present & 0x80) {
            pedid->video_latency = buf[buf_offset++];
            pedid->audio_latency = buf[buf_offset++];
        }
        if (pedid->fields_present & 0x40) {
            pedid->interlaced_video_latency = buf[buf_offset++];
            pedid->interlaced_audio_latency = buf[buf_offset++];
        }
        if (pedid->fields_present & 0x20) {
            //hdmi_edid_parse_3dinfo(buf + buf_offset, &pedid->modelist);
        }
        break;
    case 0xc45dd8:
        pedid->sink_hdmi = 1;
        pedid->hf_vsdb_version = buf[cur_offset + 4];
        switch (pedid->hf_vsdb_version) {
        case 1:/*compliant with HDMI Specification 2.0*/
            pedid->maxtmdsclock =
                buf[cur_offset + 5] * 5000000;
            DEBUG_PRINT("[CEA] maxtmdsclock is %d.\n",
                 pedid->maxtmdsclock);
            pedid->scdc_present = buf[cur_offset + 6] >> 7;
            pedid->rr_capable = (buf[cur_offset + 6] & 0x40) >> 6;
            pedid->lte_340mcsc_scramble = (buf[cur_offset + 6] & 0x08) >> 3;
            pedid->independent_view = (buf[cur_offset + 6] & 0x04) >> 2;
            pedid->dual_view = (buf[cur_offset + 6] & 0x02) >> 1;
            pedid->osd_disparity_3d = buf[cur_offset + 6] & 0x01;
            pedid->deepcolor_420 = (buf[cur_offset + 7] & 0x7) << 1;
            break;
        default:
            DEBUG_PRINT("hf_vsdb_version = %d\n", pedid->hf_vsdb_version);
            break;
        }
        break;
    default:
        DEBUG_PRINT("IEEOUT = 0x%x\n", IEEEOUI);
        break;
    }

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名：ljm_hdmiv2_edid_parse_extensions_cea
 *  功能描述：解析CEA-861标准的extension block
 *  输入参数：unsigned char *buf
 *            struct hdmi_edid *pedid
 *  输出参数：无
 *  返 回 值：static int
 *
 *  修改历史：
 *         1. 日    期：2022年7月7日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
static int ljm_hdmiv2_edid_parse_extensions_cea(unsigned char *buf,
                      struct hdmi_edid *pedid)
{
    unsigned int ddc_offset, cur_offset = 4;
    unsigned int tag, IEEEOUI = 0, count, i;

    HEADER_ARG("buf=%lx pedid=%lx", (unsigned long)buf, (unsigned long)pedid);
    /* Check ces extension version */
    if (buf[1] != 3) {
        WARN_PRINT("[CEA] error version.\n");
        goto error;
    }

    ddc_offset = buf[2];
    pedid->baseaudio_support = (buf[3] >> 6) & 0x01;
    pedid->ycbcr444 = (buf[3] >> 5) & 0x01;
    pedid->ycbcr422 = (buf[3] >> 4) & 0x01;
    //native_dtd_num = buf[3] & 0x0F;

    /* Parse data block */
    while (cur_offset < ddc_offset) {
        tag = buf[cur_offset] >> 5;
        count = buf[cur_offset] & 0x1F;
        switch (tag) {
        case 0x02:    /* Video Data Block */
            DEBUG_PRINT("[CEA] Video Data Block.\n");
            //hdmi_edid_get_cea_svd(buf + cur_offset, pedid);
            break;
        case 0x01:    /* Audio Data Block */
            DEBUG_PRINT("[CEA] Audio Data Block.\n");
            //hdmi_edid_parse_cea_sad(buf + cur_offset, pedid);
            break;
        case 0x04:    /* Speaker Allocation Data Block */
            DEBUG_PRINT("[CEA] Speaker Allocatio Data Block.\n");
            break;
        case 0x03:    /* Vendor Specific Data Block */
            DEBUG_PRINT("[CEA] Vendor Specific Data Block.\n");

            IEEEOUI = buf[cur_offset + 3];
            IEEEOUI <<= 8;
            IEEEOUI += buf[cur_offset + 2];
            IEEEOUI <<= 8;
            IEEEOUI += buf[cur_offset + 1];
            DEBUG_PRINT("[CEA] IEEEOUI is 0x%08x.\n", IEEEOUI);

            ljm_hdmiv2_edmi_parse_vsdb(buf, pedid,
                         cur_offset, IEEEOUI);
            break;
        case 0x05:    /* VESA DTC Data Block */
            DEBUG_PRINT("[CEA] VESA DTC Data Block.\n");
            break;
        case 0x07:    /* Use Extended Tag */
            DEBUG_PRINT("[CEA] Use Extended Tag Data Block %02x.\n",
                 buf[cur_offset + 1]);
            switch (buf[cur_offset + 1]) {
            case 0x00:
                DEBUG_PRINT("[CEA] Video Capability Data Block\n");
                DEBUG_PRINT("value is %02x\n", buf[cur_offset + 2]);
                break;
            case 0x05:
                DEBUG_PRINT("[CEA] Colorimetry Data Block\n");
                DEBUG_PRINT("value is %02x\n", buf[cur_offset + 2]);
                pedid->colorimetry = buf[cur_offset + 2];
                break;
            case 0x06:
                DEBUG_PRINT("[CEA] HDR Static Metedata data Block\n");
                for (i = 0; i < count - 1; i++)
                    pedid->hdr.data[i] = buf[cur_offset + 2 + i];
                break;
            case 0x0e:
                DEBUG_PRINT("[CEA] YCBCR 4:2:0 Video Data Block\n");
                for (i = 0; i < count - 1; i++) {
                    DEBUG_PRINT("mode is %d\n", buf[cur_offset + 2 + i]);
                    pedid->ycbcr420 = 1;
                    //IEEEOUI = buf[cur_offset + 2 + i] | HDMI_VIDEO_YUV420;
                    //hdmi_add_vic(IEEEOUI, &pedid->modelist);
                }
                break;
            case 0x0f:
                DEBUG_PRINT("[CEA] YCBCR 4:2:0 Capability Map Data\n");
                //hdmi_edid_parse_yuv420cmdb(&buf[cur_offset + 2], count, &pedid->modelist);
                pedid->ycbcr420 = 1;
                break;
            }
            break;
        default:
            WARN_PRINT("[CEA] unkowned data block tag.\n");
            break;
        }
        cur_offset += (buf[cur_offset] & 0x1F) + 1;
    }

    FOOTER();
    return 0;

error:
    FOOTER();
    return -1;
}

/*****************************************************************************
 *  函 数 名：ljm_hdmiv2_parse_edid_ext
 *  功能描述：解析EDID ext block内容
 *  输入参数：struct hdmi_dev *hdmi_dev
 *            struct edid *edid
 *  输出参数：无
 *  返 回 值：int
 *
 *  修改历史：
 *         1. 日    期：2022年7月7日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
int ljm_hdmiv2_parse_edid_ext(struct hdmi_dev *hdmi_dev, struct edid *edid)
{
    int i;
    unsigned char *edid_ext = NULL;

    HEADER_ARG("hdmi_dev=%lx ", (unsigned long)hdmi_dev);

    for(i = 1; (i < edid->extensions + 1) && (i < HDMI_MAX_EDID_BLOCK); i++)
    {
        edid_ext = (u8 *)edid + i * DDC_EDID_SIZE;

        if(ljm_hdmiv2_edid_checksum(edid_ext))
        {
            ERR_PRINT("hdmi %d: EDID block %d check failed\n", hdmi_dev->index, i);
            continue;
        }

        switch(edid_ext[0])
        {
            case 0xF0:
                DEBUG_PRINT("[EDID-EXTEND] Iextensions block map.\n");
                break;
            case 0x02:
                DEBUG_PRINT("[EDID-EXTEND] CEA 861 Series Extension.\n");
                ljm_hdmiv2_edid_parse_extensions_cea(edid_ext, &hdmi_dev->edid);
                break;
            case 0x10:
                DEBUG_PRINT("[EDID-EXTEND] Video Timing Block Extension.\n");
                break;
            case 0x40:
                DEBUG_PRINT("[EDID-EXTEND] Display Information Extension.\n");
                break;
            case 0x50:
                DEBUG_PRINT("[EDID-EXTEND] Localized String Extension.\n");
                break;
            case 0x60:
                DEBUG_PRINT("[EDID-EXTEND] Digital Packet Video Link Extension.\n");
                break;
            default:
                WARN_PRINT("[EDID-EXTEND] Unkowned Extension.\n");
                break;
        }
    }

    FOOTER();
    return 0;
}

enum vendor_code ljm_hdmiv2_edid_vendor(struct hdmi_dev *hdmi_dev, struct edid *edid)
{
    if((edid->mfg_id[0] == 0x50) && (edid->mfg_id[1] == 0x64) &&
        (edid->prod_code[0] == 0x09) && (edid->prod_code[1] == 0x07))
        return TCD0709;

    return EDID_VENDOR_DEFAULT;
}

/**
 * ljm_hdmiv2_parse_edid - 解析edid, 解析结果保存在hdmi_dev.edid中作为私有数据。
 * @hdmi_dev: hdmi设备指针；
 * @edid:待解析的edid；
 *
 * 返回值：
 * 成功返回0，失败返回负值。
*/
int ljm_hdmiv2_parse_edid(struct hdmi_dev *hdmi_dev, struct edid *edid)
{
    hdmi_dev->edid.sink_hdmi = drm_detect_hdmi_monitor(edid);
    hdmi_dev->edid.vendor = ljm_hdmiv2_edid_vendor(hdmi_dev, edid);
    ljm_hdmiv2_parse_edid_ext(hdmi_dev, edid);
    return 0;
}

/**
 * ljm_hdmiv2_scdc_read_update - i2c master interface SCDC read update
 * @hdmi_dev: pointer to hdmi device
 * @pdata: pointer to read update
 *
 * This function reads SCDC read update
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_scdc_read_update(struct hdmi_dev *hdmi_dev, u16 *pdata)
{
    int timeout = GP201_I2CM_TIMEOUT_CNT;
    u8 status;

    HDMI_REG_W8(hdmi_dev, IH_I2CM_STAT0, m_I2CM_STAT0_DONE | m_I2CM_STAT0_ERROR);
    HDMI_REG_W8(hdmi_dev, I2CM_SLAVE, DDC_SCDC_ADDR);

    HDMI_REG_W8(hdmi_dev, IH_MUTE_I2CM_STAT0, 0x3); /* mutes ih_i2cmphy_stat0[1:0] */
    HDMI_REG_W8(hdmi_dev, I2CM_SCDC_READ_UPDATE, 0x1);  /* SCDC Update Read */

    while (--timeout) {
        status = HDMI_REG_R8(hdmi_dev, IH_I2CM_STAT0);
        if (status & m_I2CM_STAT0_DONE)
            break;
        if (status & m_I2CM_STAT0_ERROR)
            return DDC_ERROR;   /* slave is not exist */
        udelay(5);
    }

    if (timeout == 0)
        return DDC_TIMEOUT; /* i2c read timeout */

    *pdata = HDMI_REG_R8(hdmi_dev, I2CM_SCDC_UPDATE0) + ((u16)(HDMI_REG_R8(hdmi_dev, I2CM_SCDC_UPDATE1)) << 8);

    TRACE("hdmi %d i2cm_scdc_read_update 0x%x", hdmi_dev->index, *pdata);

    return 0;
}

/*
 * ljm_hdmiv2_hpd_clear - hdmi hpd clear
 *
 * @hdmi_dev: pointer to hdmi device
 *
 * This function clears hpd interrupt
 */
inline void ljm_hdmiv2_hpd_clear(struct hdmi_dev *hdmi_dev)
{
    HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, HDMI_REG_R8(hdmi_dev, IH_MUTE_PHY_STAT0) | m_IH_MUTE_PHY_STAT0_HPD);
    HDMI_REG_W8(hdmi_dev, IH_PHY_STAT0, m_HPD);
    //HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, HDMI_REG_R8(hdmi_dev, IH_MUTE_PHY_STAT0) & ~m_IH_MUTE_PHY_STAT0_HPD);
}

/*
 * ljm_hdmiv2_hpd_enable - hdmi hpd enable
 *
 * @hdmi_dev: pointer to hdmi device

 * @enable: 0 to disable, otherwise enable
 */
void ljm_hdmiv2_hpd_enable(struct hdmi_dev *hdmi_dev, int enable)
{
    if (enable) /* enable hpd interrupt */
        HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, HDMI_REG_R8(hdmi_dev, IH_MUTE_PHY_STAT0) & ~m_IH_MUTE_PHY_STAT0_HPD);
    else    /* mute hpd interrupt */
        HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, HDMI_REG_R8(hdmi_dev, IH_MUTE_PHY_STAT0) | m_IH_MUTE_PHY_STAT0_HPD);
}

/**
 * ljm_hdmiv2_irq_process - hdmi irq process
 *
 * @hdmi_dev: pointer to hdmi device
 *
 * Return: irq type
 */
int ljm_hdmiv2_irq_process(struct hdmi_dev *hdmi_dev)
{
    int r = LJM_IRQ_TYPE_NONE;

    /* check HPD interrupt */
    if (HDMI_REG_R8(hdmi_dev, IH_PHY_STAT0) & m_HPD) {
        r |= LJM_IRQ_TYPE_HPD;
        //ljm_hdmiv2_hpd_clear(hdmi_dev);  /* clear HPD interrupt */
        HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, HDMI_REG_R8(hdmi_dev, IH_MUTE_PHY_STAT0) | m_IH_MUTE_PHY_STAT0_HPD);

    }

    return r;
}

/**
 * ljm_hdmiv2_hpd_set_polarity - hdmi hpd set polarity
 *
 * @hdmi_dev: pointer to hdmi device
 *
 * Returns:
 * Zero on success or a negative error code on failure.
 */
int ljm_hdmiv2_hpd_set_polarity(struct hdmi_dev *hdmi_dev)
{
    int phy_int, phy_int0, phy_pol, phy_status;

    //HEADER_ARG("hdmi_dev=%p", hdmi_dev);

    phy_pol = HDMI_REG_R8(hdmi_dev, PHY_POL0);
    phy_status = HDMI_REG_R8(hdmi_dev, PHY_STAT0);
    phy_int0 = HDMI_REG_R8(hdmi_dev, PHY_INT0);
    phy_int = HDMI_REG_R8(hdmi_dev, IH_PHY_STAT0);

    phy_pol = (phy_int0 & (~phy_status)) | ((~phy_int0) & phy_pol);
    HDMI_REG_W8(hdmi_dev, PHY_POL0, phy_pol);
    HDMI_REG_W8(hdmi_dev, IH_PHY_STAT0, phy_int);

    HDMI_REG_W8(hdmi_dev, IH_MUTE_PHY_STAT0, HDMI_REG_R8(hdmi_dev, IH_MUTE_PHY_STAT0) & ~m_IH_MUTE_PHY_STAT0_HPD);

    //FOOTER();

    return 0;
}

bool ljm_hdmiv2_mode_valid(struct hdmi_dev *hdmi_dev, const struct drm_display_mode *mode)
{
    bool r = false;

    if(!hdmi_dev->is_dvi) {
        if(mode->clock > 340000) {   /* 600000 for 4K@60Hz (max), 340000 for 4K@30Hz (max) */
            r = false;
            goto end;
        }

        if(mode->hdisplay >= 3840 || mode->vdisplay >= 2160) {
            if((mode->clock * 1000 == mode->htotal * mode->vtotal * 30) ||
            (mode->clock * 1000 == mode->htotal * mode->vtotal * 60)) //4K分辨率只有30Hz和60Hz的mode可用
                r = true;
            else
                r = false;
            goto end;
        }

        r = true;
    }
    else {
        if(mode->clock > 165000)
            r = false;
        else
            r = true;
    }

end:
    if (r) {
        DEBUG_PRINT("HDMI%d: %dx%d@%dkHz is valid", hdmi_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    } else {
        DEBUG_PRINT("HDMI%d: %dx%d@%dkHz is invalid", hdmi_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    }
    return r;
}

bool ljm_hdmiv2_mode_valid_ex(struct hdmi_dev *hdmi_dev, const struct drm_display_mode *mode)
{
    bool r = false;

    if (hdmi_dev->is_dvi) {
        if (mode->clock > 165000)
            r = false;
        else
            r = true;
    } else {
        if (mode->clock > 600000)
            r = false;
        else
            r = true;
    }

    if (r) {
        DEBUG_PRINT("HDMI%d: %dx%d@%dkHz is valid", hdmi_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    } else {
        DEBUG_PRINT("HDMI%d: %dx%d@%dkHz is invalid", hdmi_dev->index, mode->hdisplay, mode->vdisplay, mode->clock);
    }
    return r;
}

