/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_HDMIV2_H__
#define __LJMICRO_HDMIV2_H__

#include "ljmicro_hdmi.h"

enum {
    OUTPUT_DVI = 0,
    OUTPUT_HDMI,
};

/* Color Space Conversion Mode */
enum {
    CSC_BYPASS,
    CSC_RGB_0_255_TO_RGB_16_235_8BIT,   /* RGB 0-255 input to RGB
                                         * 16-235 output that is 8bit
                                         * clolor depth
                                        */
    CSC_RGB_0_255_TO_RGB_16_235_10BIT,  /* RGB 0-255 input to RGB
                                         * 16-235 output that is 8bit
                                         * clolor depth
                                         */
    CSC_RGB_0_255_TO_ITU601_16_235_8BIT,    /* RGB 0-255 input to YCbCr
                                             * 16-235 output according
                                             * BT601 that is 8bit clolor
                                             * depth
                                             */
    CSC_RGB_0_255_TO_ITU601_16_235_10BIT,   /* RGB 0-255 input to YCbCr
                                             * 16-235 output according
                                             * BT601 that is 10bit clolor
                                             * depth
                                             */
    CSC_RGB_0_255_TO_ITU709_16_235_8BIT,    /* RGB 0-255 input to YCbCr
                                             * 16-235 output accroding
                                             * BT709 that is 8bit clolor
                                             * depth
                                             */
    CSC_RGB_0_255_TO_ITU709_16_235_10BIT,   /* RGB 0-255 input to YCbCr
                                             * 16-235 output accroding
                                             * BT709 that is 10bit clolor
                                             * depth
                                             */
    CSC_ITU601_16_235_TO_RGB_0_255_8BIT,    /* YCbCr 16-235 input to RGB
                                             * 0-255 output according
                                             * BT601 that is 8bit clolor
                                             * depth
                                             */
    CSC_ITU709_16_235_TO_RGB_0_255_8BIT,    /* YCbCr 16-235 input to RGB
                                             * 0-255 output according
                                             * BT709 that is 8bit clolor
                                             * depth
                                             */
    CSC_ITU601_16_235_TO_RGB_16_235_8BIT,   /* YCbCr 16-235 input to RGB
                                             * 16-235 output according
                                             * BT601 that is 8bit clolor
                                             * depth
                                             */
    CSC_ITU709_16_235_TO_RGB_16_235_8BIT    /* YCbCr 16-235 input to RGB
                                             * 16-235 output according
                                             * BT709 that is 8bit clolor
                                             * depth
                                             */
};

/*
 * GP201 HDMI Controller and PHY
 */

#define GP201_I2CM_TIMEOUT_CNT  0xff

#define PHY_GEN2_ADDR           0x69

/*********************** GP201 HDMI Controller (8bit) *************************/
/* Register and Field Descriptions */
/* Identification Registers */
#define IDENTIFICATION_BASE     0x0000

#define DESIGN_ID               0x0000
#define REVISION_ID             0x0001
#define PRODUCT_ID0             0x0002
#define PRODUCT_ID1             0x0003

#define CONFIG0_ID              0x0004
#define m_PREPEN                BIT(7)
#define m_AUDSPDIF              BIT(5)
#define m_AUDI2S                BIT(4)
#define m_HDMI14                BIT(3)
#define m_CSC                   BIT(2)
#define m_CEC                   BIT(1)
#define m_HDCP                  BIT(0)

#define CONFIG1_ID              0x0005
#define m_HDCP22                BIT(6)
#define m_HDMI20                BIT(5)
#define m_CONFAPB               BIT(1)

#define CONFIG2_ID              0x0006
enum PHYTYPE {
    HDMI_TX_PHY = 0x00,
    HDMI_MHL_WITH_HEAC_PHY = 0xb2,
    HDMI_MHL_PHY = 0xc2,
    HDMI_3D_TX_WITH_HEAC_PHY = 0xe2,
    HDMI_3D_TX_PHY = 0xf2,
    HDMI2_TX_PHY = 0xf3
};

#define CONFIG3_ID              0x0007
#define m_AHB_AUD_DMA           BIT(1)
#define m_GP_AUD                BIT(0)

/* Interrupt Registers */
#define INTERRUPT_BASE          0x0100

#define IH_FC_STAT0             0x0100
#define m_AUD_INFOFRAME         BIT(7)
#define m_AUD_CONTENT_PROTECT   BIT(6)
#define m_AUD_HBR               BIT(5)
#define m_AUD_SAMPLE            BIT(2)
#define m_AUD_CLK_REGEN         BIT(1)
#define m_NULL_PACKET           BIT(0)

#define IH_FC_STAT1             0x0101
#define m_GMD                   BIT(7)
#define m_ISCR1                 BIT(6)
#define m_ISCR2                 BIT(5)
#define m_VSD                   BIT(4)
#define m_SPD                   BIT(3)
#define m_AVI_INFOFRAME         BIT(1)
#define m_GCP                   BIT(0)

#define v_AVI_INFOFRAME(n)      (((n) & 0x01) << 1)

#define IH_FC_STAT2             0x0102
#define m_LOWPRIO_OVERFLOW      BIT(1)
#define m_HIGHPRIO_OVERFLOW     BIT(0)

#define IH_AS_SATA0             0x0103
#define m_FIFO_UNDERRUN         BIT(4)
#define m_FIFO_OVERRUN          BIT(3)
#define m_AUD_FIFO_UDFLOW_THR   BIT(2)
#define m_AUD_FIFO_UDFLOW       BIT(1)
#define m_AUD_FIFO_OVERFLOW     BIT(0)

#define IH_PHY_STAT0            0x0104  /* PHY Interface Interrupt Status Register */
#define m_RX_SENSE3             BIT(5)
#define m_RX_SENSE2             BIT(4)
#define m_RX_SENSE1             BIT(3)
#define m_RX_SENSE0             BIT(2)
#define m_TX_PHY_LOCK           BIT(1)
#define m_HPD                   BIT(0)

#define IH_I2CM_STAT0           0x0105  /* E-DDC I2C Master Interrupt Status Register
                                           [0]: 1-I2Cmphyerror
                                           [1]: 1-I2Cmphydone */
#define m_SCDC_READREQ          BIT(2)
#define m_I2CM_DONE             BIT(1)
#define m_I2CM_ERROR            BIT(0)

#define IH_VP_STAT0             0x0107  /* FIFO Full and Empty Interrupts */
#define IH_I2CMPHY_STAT0        0x0108  /* PHY GEN2 I2C Master Interrupt Status Register
                                           [0]: 1-I2Cmphyerror
                                           [1]: 1-I2Cmphydone */
#define m_I2CM_STAT0_ERROR      (1u<<0)
#define m_I2CM_STAT0_DONE       (1u<<1)

#define IH_DECODE               0x0170  /* Interruption Handler Decode Assist Register */
#define IH_MUTE_FC_STAT0        0x0180  /* Frame Composer Interrupt Mute Control Register 0 */
#define IH_MUTE_FC_STAT1        0x0181  /* Frame Composer Interrupt Mute Control Register 1 */
#define IH_MUTE_FC_STAT2        0x0182  /* Frame Composer Interrupt Mute Control Register 2 */
#define IH_MUTE_AS_STAT0        0x0183  /* Audio Sampler Interrupt Mute Control Register */
#define IH_MUTE_PHY_STAT0       0x0184  /* PHY Interface Interrupt Mute Control Register */
#define m_IH_MUTE_PHY_STAT0_HPD         (1u<<0)
#define m_IH_MUTE_PHY_STAT0_TX_PHY_LOCK (1u<<1)
#define m_IH_MUTE_PHY_STAT0_RX_SENSE_0  (1u<<2)
#define m_IH_MUTE_PHY_STAT0_RX_SENSE_1  (1u<<3)
#define m_IH_MUTE_PHY_STAT0_RX_SENSE_2  (1u<<4)
#define m_IH_MUTE_PHY_STAT0_RX_SENSE_3  (1u<<5)
#define IH_MUTE_I2CM_STAT0      0x185   /* E-DDC I2C Master Interrupt Mute Control Register */
#define IH_MUTE_CEC_STAT0       0x186   /* CEC Interrupt Mute Control Register */
#define IH_MUTE_VP_STAT0        0x0187  /* Video Packetizer Interrupt Mute Control Register */
#define IH_MUTE_I2CMPHY_STAT0   0x188   /* PHY GEN2 I2C Master Interrupt Mute Control Register */
#define IH_MUTE_AHBDMAAUD_STAT0 0x189   /* AHB Audio DMA Interrupt Mute Control Register */
#define IH_MUTE                 0x01FF  /* [0]: 1-mutes the main interrupt line (default). The sticky bit interrupts continue with their state.
                                           [1]: 1-mutes the main interrupt output port */
#define m_IH_MUTE_ALL           (1u<<0)
#define m_IH_MUTE_WAKEUP        (1u<<1)

/* Video Sampler Registers */
#define VIDEO_SAMPLER_BASE      0x0200

#define TX_INVID0               0x0200  /* Video Input Mapping and Internal Data Enable Configuration Register */
#define m_INTERNAL_DE_GEN       BIT(7)
#define v_INTERNAL_DE_GEN(n)    (((n) & 0x01) << 7)
enum VIDEO_MODE {
    VIDEO_RGB444_8BIT = 0x01,
    VIDEO_RGB444_10BIT = 0x03,
    VIDEO_RGB444_12BIT = 0x05,
    VIDEO_RGB444_16BIT = 0x07,
    VIDEO_YCBCR444_8BIT = 0x09,	/* or YCbCr420 */
    VIDEO_YCBCR444_10BIT = 0x0b,	/* or YCbCr420 */
    VIDEO_YCBCR444_12BIT = 0x0d,	/* or YCbCr420 */
    VIDEO_YCBCR444_16BIT = 0x0f,	/* or YCbCr420 */
    VIDEO_YCBCR422_12BIT = 0x12,
    VIDEO_YCBCR422_10BIT = 0x14,
    VIDEO_YCBCR422_8BIT = 0x16
};

#define m_VIDEO_MAPPING         (0x1f << 0)
#define v_VIDEO_MAPPING(n)      ((n) & 0x1f)

#define TX_INSTUFFING           0x0201
#define m_BCBDATA_STUFF         BIT(2)
#define v_BCBDATA_STUFF(n)      (((n) & 0x01) << 2)
#define m_RCRDATA_STUFF         BIT(1)
#define v_RCRDATA_STUFF(n)      (((n) & 0x01) << 1)
#define m_GYDATA_STUFF          BIT(0)
#define v_GYDATA_STUFF(n)       (((n) & 0x01) << 0)

#define TX_GYDATA0              0x0202
#define TX_GYDATA1              0x0203
#define TX_RCRDATA0             0x0204
#define TX_RCRDATA1             0x0205
#define TX_BCBDATA0             0x0206
#define TX_BCBDATA1             0x0207

/* Video Packetizer Registers */
#define VIDEO_PACKETIZER_BASE   0x0800

#define VP_STATUS               0x0800
#define m_PACKING_PHASE         (0x0f << 0)

#define VP_PR_CD                0x0801  /* Video Packetizer Pixel Repetition and Color Depth Register */
enum COLOR_DEPTH {
    COLOR_DEPTH_24BIT_DEFAULT = 0,
    COLOR_DEPTH_24BIT = 0x04,
    COLOR_DEPTH_30BIT,
    COLOR_DEPTH_36BIT,
    COLOR_DEPTH_48BIT
};

#define m_COLOR_DEPTH           (0x0f << 4)
#define v_COLOR_DEPTH(n)        (((n) & 0x0f) << 4)
enum PIXEL_REPET {
    NO_PIXEL_REPET = 0,
    PIXEL_SENT_2TIMES,
    PIXEL_SENT_3TIMES,
    PIXEL_SENT_4TIMES,
    PIXEL_SENT_5TIMES,
    PIXEL_SENT_6TIMES,
    PIXEL_SENT_7TIMES,
    PIXEL_SENT_8TIMES,
    PIXEL_SENT_9TIMES,
    PIXEL_SENT_10TIMES
};

#define m_DESIRED_PR_FACTOR     (0x0f << 0)
#define v_DESIRED_PR_FACTOR(n)  (((n) & 0x0f) << 0)

#define VP_STUFF                0x0802
#define m_IDEFAULT_PHASE        BIT(5)
#define v_IDEFAULT_PHASE(n)     (((n) & 0x01) << 5)
#define m_IFIX_PP_TO_LAST       BIT(4)
#define m_ICX_GOTO_P0_ST        BIT(3)
enum {
    DIRECT_MODE = 0,
    STUFFING_MODE
};

#define m_YCC422_STUFFING       BIT(2)
#define v_YCC422_STUFFING(n)    (((n) & 0x01) << 2)
#define m_PP_STUFFING           BIT(1)
#define v_PP_STUFFING(n)        (((n) & 0x01) << 1)
#define m_PR_STUFFING           BIT(0)
#define v_PR_STUFFING(n)        (((n) & 0x01) << 0)

#define VP_REMAP                0x0803
enum YCC422_SIZE {
    YCC422_16BIT = 0,
    YCC422_20BIT,
    YCC422_24BIT
};

#define m_YCC422_SIZE           (0x03 << 0)
#define v_YCC422_SIZE(n)        (((n) & 0x03) << 0)

#define VP_CONF                 0x0804
#define m_BYPASS_EN             BIT(6)
#define v_BYPASS_EN(n)          (((n) & 0x01) << 6)
#define m_PIXEL_PACK_EN         BIT(5)
#define v_PIXEL_PACK_EN(n)      (((n) & 0x01) << 5)
#define m_PIXEL_REPET_EN        BIT(4)
#define v_PIXEL_REPET_EN(n)     (((n) & 0x01) << 4)
#define m_YCC422_EN             BIT(3)
#define v_YCC422_EN(n)          (((n) & 0x01) << 3)
#define m_BYPASS_SEL            BIT(2)
#define v_BYPASS_SEL(n)         (((n) & 0x01) << 2)
enum {
    OUT_FROM_PIXEL_PACKING = 0,
    OUT_FROM_YCC422_REMAP,
    OUT_FROM_8BIT_BYPASS
};

#define m_OUTPUT_SEL            (0x03 << 0)
#define v_OUTPUT_SEL(n)         (((n) & 0x03) << 0)

#define VP_MASK                 0x0807
#define m_OINTFULL_REPET        BIT(7)
#define m_OINTEMPTY_REPET       BIT(6)
#define m_OINTFULL_PACK         BIT(5)
#define m_OINTEMPTY_PACK        BIT(4)
#define m_OINTFULL_REMAP        BIT(3)
#define m_OINTEMPTY_REMAP       BIT(2)
#define m_OINTFULL_BYPASS       BIT(1)
#define m_OINTEMPTY_BYPASS      BIT(0)

/* Frame Composer Registers */
#define FRAME_COMPOSER_BASE     0x1000

#define FC_INVIDCONF            0x1000  /* Frame Composer Input Video Configuration and HDCP Keepout Register */
#define m_FC_HDCP_KEEPOUT       BIT(7)
#define v_FC_HDCP_KEEPOUT(n)    (((n) & 0x01) << 7)
#define m_FC_VSYNC_POL          BIT(6)
#define v_FC_VSYNC_POL(n)       (((n) & 0x01) << 6)
#define m_FC_HSYNC_POL          BIT(5)
#define v_FC_HSYNC_POL(n)       (((n) & 0x01) << 5)
#define m_FC_DE_POL             BIT(4)
#define v_FC_DE_POL(n)          (((n) & 0x01) << 4)
#define m_FC_HDMI_DVI           BIT(3)
#define v_FC_HDMI_DVI(n)        (((n) & 0x01) << 3)
#define m_FC_VBLANK             BIT(1)
#define v_FC_VBLANK(n)          (((n) & 0x01) << 1)
#define m_FC_INTERLACE_MODE     BIT(0)
#define v_FC_INTERLACE_MODE(n)  (((n) & 0x01) << 0)

#define FC_INHACTIV0            0x1001  /* Frame Composer Input Video HActive Register 0 */

#define FC_INHACTIV1            0x1002  /* Frame Composer Input Video HActive Register 1 */
#define v_FC_HACTIVE1(n)        ((n) & 0x3f)
#define m_FC_H_ACTIVE_13        BIT(5)
#define v_FC_H_ACTIVE_13(n)     (((n) & 0x01) << 5)
#define m_FC_H_ACTIVE_12        BIT(4)
#define v_FC_H_ACTIVE_12(n)     (((n) & 0x01) << 4)
#define m_FC_H_ACTIVE           (0x0f << 0)
#define v_FC_H_ACTIVE(n)        (((n) & 0x0f) << 0)

#define FC_INHBLANK0            0x1003  /* Frame Composer Input Video HBlank Register 0 */

#define FC_INHBLANK1            0x1004  /* Frame Composer Input Video HBlank Register 1 */
#define v_FC_HBLANK1(n)         ((n) & 0x1f)
#define m_FC_H_BLANK_12_11      (0x07 << 2)
#define v_FC_H_BLANK_12_11(n)   (((n) & 0x07) << 2)
#define m_FC_H_BLANK            (0x03 << 0)
#define v_FC_H_BLANK(n)         (((n) & 0x03) << 0)

#define FC_INVACTIV0            0x1005  /* Frame Composer Input Video VActive Register 0 */

#define FC_INVACTIV1            0x1006  /* Frame Composer Input Video VActive Register 1 */
#define v_FC_VACTIVE1(n)        ((n) & 0x1f)
#define m_FC_V_ACTIVE_12_11     (0x03 << 3)
#define v_FC_V_ACTIVE_12_11(n)  (((n) & 0x03) << 3)
#define m_FC_V_ACTIVE           (0x07 << 0)
#define v_FC_V_ACTIVE(n)        (((n) & 0x07) << 0)

#define FC_INVBLANK0            0x1007  /* Frame Composer Input Video VBlank Register 0 */
#define FC_HSYNCINDELAY0        0x1008  /* Frame Composer Input Video HSync Front Porch Register 0 */

#define FC_HSYNCINDELAY1        0x1009  /* Frame Composer Input Video HSync Front Porch Register 1 */
#define v_FC_HSYNCINDEAY1(n)    ((n) & 0x1f)
#define m_FC_H_SYNCFP_12_11     (0x03 << 3)
#define v_FC_H_SYNCFP_12_11(n)  (((n) & 0x03) << 3)
#define m_FC_H_SYNCFP           (0x07 << 0)
#define v_FC_H_SYNCFP(n)        (((n) & 0x07) << 0)

#define FC_HSYNCINWIDTH0        0x100A  /* Frame Composer Input Video HSync Width Register 0 */

#define FC_HSYNCINWIDTH1        0x100B  /* Frame Composer Input Video HSync Width Register 1 */
#define v_FC_HSYNCWIDTH1(n)     ((n) & 0x03)
#define m_FC_HSYNC_9            BIT(1)
#define v_FC_HSYNC_9(n)         (((n) & 0x01) << 1)
#define m_FC_HSYNC              BIT(0)
#define v_FC_HSYNC(n)           (((n) & 0x01) << 0)

#define FC_VSYNCINDELAY0        0x100C  /* Frame Composer Input Video VSync Front Porch Register 0 */
#define FC_VSYNCINWIDTH         0x100D  /* Frame Composer Input Video VSync Width Register */
#define	FC_INFREQ0              0x100e
#define FC_INFREQ0_VAL(n)       ((n) & 0xff)
#define	FC_INFREQ1              0x100f
#define FC_INFREQ1_VAL(n)       (((n) & 0xff00) >> 8)
#define	FC_INFREQ2              0x1010
#define FC_INFREQ2_VAL(n)       (((n) & 0xff0000) >> 16)
#define FC_CTRLDUR              0x1011  /* Frame Composer Control Period Duration Register */
#define FC_EXCTRLDUR            0x1012  /* Frame Composer Extended Control Period Duration Register */
#define FC_EXCTRLSPAC           0x1013  /* Frame Composer Extended Control Period Maximum Spacing Register */
#define	FC_CH0PREAM             0x1014
#define	FC_CH1PREAM             0x1015
#define	FC_CH2PREAM             0x1016

#define FC_AVICONF3             0x1017  /* Frame Composer AVI Packet Configuration Register 3 */
enum YCC_QUAN_RANGE {
    YQ_LIMITED_RANGE = 0,
    YQ_FULL_RANGE,
    RESERVED,
};

#define m_FC_YQ                 (0x03 << 2)
#define v_FC_YQ(n)              (((n) & 0x03) << 2)
enum IT_CONTENT_TYPE {
    CN_GRAPHICS = 0,
    CN_PHOTO,
    CN_CINEMA,
    CN_GAME,
};

#define m_FC_CN                 (0x03 << 0)
#define v_FC_CN(n)              (((n) & 0x03) << 0)

#define FC_GCP                  0x1018  /* Frame Composer GCP Packet Configuration Register */
#define m_FC_DEFAULT_PHASE	BIT(2)
#define v_FC_DEFAULT_PHASE(n)	(((n) & 0x01) << 2)
#define m_FC_SET_AVMUTE		BIT(1)
#define v_FC_SET_AVMUTE(n)	(((n) & 0x01) << 1)
#define m_FC_CLR_AVMUTE		BIT(0)
#define v_FC_CLR_AVMUTE(n)	(((n) & 0x01) << 0)

enum {
	AVI_COLOR_MODE_RGB = 0,
	AVI_COLOR_MODE_YCBCR422,
	AVI_COLOR_MODE_YCBCR444,
	AVI_COLOR_MODE_YCBCR420
};

enum {
	AVI_COLORIMETRY_NO_DATA = 0,
	AVI_COLORIMETRY_SMPTE_170M,
	AVI_COLORIMETRY_ITU709,
	AVI_COLORIMETRY_EXTENDED
};

enum {
	AVI_CODED_FRAME_ASPECT_NO_DATA,
	AVI_CODED_FRAME_ASPECT_4_3,
	AVI_CODED_FRAME_ASPECT_16_9
};

enum {
	ACTIVE_ASPECT_RATE_DEFAULT = 0x08,
	ACTIVE_ASPECT_RATE_4_3,
	ACTIVE_ASPECT_RATE_16_9,
	ACTIVE_ASPECT_RATE_14_9
};

enum {
	AVI_QUANTIZATION_RANGE_DEFAULT = 0,
	AVI_QUANTIZATION_RANGE_LIMITED,
	AVI_QUANTIZATION_RANGE_FULL
};

#define FC_AVICONF0             0x1019  /* Frame Composer AVI Packet Configuration Register 0 */
#define m_FC_RGC_YCC_2		BIT(7)	/* use for HDMI2.0 TX */
#define v_FC_RGC_YCC_2(n)	(((n) & 0x01) << 7)
#define m_FC_ACTIV_FORMAT	BIT(6)
#define v_FC_ACTIV_FORMAT(n)	(((n) & 0x01) << 6)
#define m_FC_SCAN_INFO		(0x03 << 4)
#define v_FC_SCAN_INFO(n)	(((n) & 0x03) << 4)
#define m_FC_BAR_FORMAT		(0x03 << 2)
#define v_FC_BAR_FORMAT(n)	(((n) & 0x03) << 2)
#define m_FC_RGC_YCC		(0x03 << 0)
#define v_FC_RGC_YCC(n)		(((n) & 0x03) << 0)

#define FC_AVICONF1             0x101A  /* Frame Composer AVI Packet Configuration Register 1 */
#define m_FC_COLORIMETRY	(0x03 << 6)
#define v_FC_COLORIMETRY(n)	(((n) & 0x03) << 6)
#define m_FC_PIC_ASPEC_RATIO	(0x03 << 4)
#define v_FC_PIC_ASPEC_RATIO(n)	(((n) & 0x03) << 4)
#define m_FC_ACT_ASPEC_RATIO	(0x0f << 0)
#define v_FC_ACT_ASPEC_RATIO(n)	(((n) & 0x0f) << 0)

#define FC_AVICONF2             0x101B  /* Frame Composer AVI Packet Configuration Register 2 */
#define m_FC_IT_CONTENT		BIT(7)
#define v_FC_IT_CONTENT(n)	(((n) & 0x01) << 7)
#define m_FC_EXT_COLORIMETRY	(0x07 << 4)
#define v_FC_EXT_COLORIMETRY(n)	(((n) & 0x07) << 4)
#define m_FC_QUAN_RANGE		(0x03 << 2)
#define v_FC_QUAN_RANGE(n)	(((n) & 0x03) << 2)
#define m_FC_NUN_PIC_SCALE	(0x03 << 0)
#define v_FC_NUN_PIC_SCALE(n)	(((n) & 0x03) << 0)

#define FC_AVIVID               0x101C  /* Frame Composer AVI Packet VIC Register */
#define m_FC_AVIVID_H		BIT(7)	/* use for HDMI2.0 TX */
#define v_FC_AVIVID_H(n)	(((n) & 0x01) << 7)
#define m_FC_AVIVID		(0x7f << 0)
#define v_FC_AVIVID(n)		(((n) & 0x7f) << 0)

#define FC_AVIETB0              0x101D  /* Frame Composer AVI Packet End of Top Bar Register Array [0] */
#define FC_AVIETB1              0x101E  /* Frame Composer AVI Packet End of Top Bar Register Array [1] */
#define FC_AVISBB0              0x101F  /* Frame Composer AVI Packet Start of Bottom Bar Register Array [0] */
#define FC_AVISBB1              0x1020  /* Frame Composer AVI Packet Start of Bottom Bar Register Array [1] */
#define FC_AVIELB0              0x1021  /* Frame Composer AVI Packet End of Left Bar Register Array [0] */
#define FC_AVIELB1              0x1022  /* Frame Composer AVI Packet End of Left Bar Register Array [1] */
#define FC_AVISRB0              0x1023  /* Frame Composer AVI Packet Start of Right Bar Register Array [0] */
#define FC_AVISRB1              0x1024  /* Frame Composer AVI Packet Start of Right Bar Register Array [1] */

#define FC_AUDICONF0            0x1025  /* Frame Composer AUD Packet Configuration Register 0 */
#define m_FC_CHN_CNT            (0x07 << 4)
#define v_FC_CHN_CNT(n)         (((n) & 0x07) << 4)
#define m_FC_CODING_TYPE        (0x0f << 0)
#define v_FC_CODING_TYPE(n)     (((n) & 0x0f) << 0)

#define FC_AUDICONF1            0x1026  /* Frame Composer AUD Packet Configuration Register 1 */#define m_FC_SAMPLE_SIZE	(0x03 << 4)
#define v_FC_SAMPLE_SIZE(n)     (((n) & 0x03) << 4)
#define m_FC_SAMPLE_FREQ        (0x07 << 0)
#define v_FC_SAMPLE_FREQ(n)     (((n) & 0x07) << 0)

#define FC_AUDICONF2            0x1027  /* Frame Composer AUD Packet Configuration Register 2 */
#define FC_AUDICONF3            0x1028  /* Frame Composer AUD Packet Configuration Register 3 */
#define m_FC_LFE_PBL            (0x03 << 5)	/*only use for HDMI1.4 TX*/
#define v_FC_LFE_PBL(n)         (((n) & 0x03) << 5)
#define m_FC_DM_INH             BIT(4)
#define v_FC_DM_INH(n)          (((n) & 0x01) << 4)
#define m_FC_LSV                (0x0f << 0)
#define v_FC_LSV(n)             (((n) & 0x0f) << 0)

#define FC_INVBLANK1            0x102E  /* Frame Composer Input Video VBlank Register 1 */
#define FC_VSYNCINDELAY1        0x102F  /* Frame Composer Input Video VSync Front Porch Register 1 */
#define	FC_VSDIEEEID0           0x1029
#define	FC_VSDSIZE              0x102a
#define FC_VSDIEEEID1           0x1030  /* Frame Composer VSI Packet Data IEEE Register 1 */
#define FC_VSDIEEEID2           0x1031  /* Frame Composer VSI Packet Data IEEE Register 2 */
#define FC_VSDPAYLOAD0          0x1032  /* Frame Composer VSI Packet Data Payload Register Array */
#define FC_VSDPAYLOAD_SIZE      24      /* FC_VSDPAYLOAD_A array size */

#define	FC_SPDVENDORNAME0       0x104a	/* 0~7 */
#define	FC_SPDPRODUCTNAME0      0x1052	/* 0~15 */
#define	FC_SPDDEVICEINF         0x1062
#define	FC_AUDSCONF             0x1063
#define	HDMI_FC_AUDSCHNLS7      0x106e
#define	HDMI_FC_AUDSCHNLS8      0x106f

#define	FC_DATAUTO0             0x10b3
#define m_SPD_AUTO		BIT(4)
#define v_SPD_AUTO(n)		(((n) & 0x01) << 4)
#define m_VSD_AUTO		BIT(3)
#define v_VSD_AUTO(n)		(((n) & 0x01) << 3)
#define m_ISCR2_AUTO		BIT(2)
#define v_ISCR2_AUTO(n)		(((n) & 0x01) << 2)
#define m_ISCR1_AUTO		BIT(1)
#define v_ISCR1_AUTO(n)		(((n) & 0x01) << 1)
#define m_ACP_AUTO		BIT(0)
#define v_ACP_AUTO(n)		(((n) & 0x01) << 0)

#define	FC_DATAUTO1             0x10b4
#define	FC_DATAUTO2             0x10b5

#define	FC_DATMAN               0x10b6
#define m_SPD_MAN		BIT(4)
#define v_SPD_MAN(n)		(((n) & 0x01) << 4)
#define m_VSD_MAN		BIT(3)
#define v_VSD_MAN(n)		(((n) & 0x01) << 3)
#define m_ISCR2_MAN		BIT(2)
#define v_ISCR2_MAN(n)		(((n) & 0x01) << 2)
#define m_ISCR1_MAN		BIT(1)
#define v_ISCR1_MAN(n)		(((n) & 0x01) << 1)
#define m_ACP_MAN		BIT(0)
#define v_ACP_MAN(n)		(((n) & 0x01) << 0)

#define	FC_DATAUTO3             0x10b7
#define m_AVI_AUTO              BIT(3)
#define v_AVI_AUTO(n)           (((n) & 0x01) << 3)
#define m_GCP_AUTO              BIT(2)
#define v_GCP_AUTO(n)           (((n) & 0x01) << 2)
#define m_AAI_AUTO              BIT(1)
#define v_AAI_AUTO(n)           (((n) & 0x01) << 1)
#define m_ACR_AUTO              BIT(0)
#define v_ACR_AUTO(n)           (((n) & 0x01) << 0)

#define	FC_MASK0                0x10d2
#define	FC_MASK1                0x10d6
#define	FC_MASK2                0x10da

#define	FC_PRCONF               0x10e0
#define m_FC_PR_FACTOR          (0x0f << 4)
#define v_FC_PR_FACTOR(n)       (((n) & 0x0f) << 4)
#define m_FC_PR_FACTOR_OUT      (0x0f)
#define v_FC_PR_FACTOR_OUT(n)   ((n) & 0x0f)

#define	FC_SCRAMBLER_CTRL       0x10e1
#define m_FC_SCRAMBLE_UCP       BIT(4)
#define v_FC_SCRAMBLE_UCP(n)    (((n) & 0x01) << 4)
#define m_FC_SCRAMBLE_EN        BIT(0)
#define v_FC_SCRAMBLE_EN(n)     (((n) & 0x01) << 0)

#define FC_PACK_TXE             0x10E3  /* Frame Composer Packet Transmission Control Reigster 0 */
#define m_DRM_TXEN              BIT(7)
#define v_DRM_TXEN(n)           (((n) & 0x01) << 7)
#define FC_DRM_UP               0x1167
#define m_DRM_PUPD              BIT(0)
#define v_DRM_PUPD(n)           (((n) & 0x01) << 0)
#define FC_DRM_HB               0x1168
#define FC_DRM_PB               0x116a
#define m_DRM_MASK              BIT(4)
#define v_DRM_MASK(n)           (((n) & 0x01) << 4)

#define FC_DBGFORCE             0x1200
#define m_FC_FORCEAUDIO         BIT(4)
#define v_FC_FORCEAUDIO(n)      (((n) & 0x01) << 4)
#define m_FC_FORCEVIDEO	        BIT(0)
#define v_FC_FORCEVIDEO(n)      (((n) & 0x01) << 0)

#define	FC_DBGAUD0CH0			0x1201	/* aud0~aud2 ch0 */
#define	FC_DBGAUD0CH1			0x1204	/* aud0~aud2 ch1 */
#define	FC_DBGAUD0CH2			0x1207	/* aud0~aud2 ch2 */
#define	FC_DBGAUD0CH3			0x120a	/* aud0~aud2 ch3 */
#define	FC_DBGAUD0CH4			0x120d	/* aud0~aud2 ch4 */
#define	FC_DBGAUD0CH5			0x1210	/* aud0~aud2 ch5 */
#define	FC_DBGAUD0CH6			0x1213	/* aud0~aud2 ch6 */
#define	FC_DBGAUD0CH7			0x1216	/* aud0~aud2 ch7 */
#define	FC_DBGTMDS0			0x1219
#define	FC_DBGTMDS1			0x121a
#define	FC_DBGTMDS2			0x121b

#define FC_DBGFORCE             0x1200
#define m_FC_FORCEAUDIO         BIT(4)
#define v_FC_FORCEAUDIO(n)      (((n) & 0x01) << 4)
#define m_FC_FORCEVIDEO         BIT(0)
#define v_FC_FORCEVIDEO(n)      (((n) & 0x01) << 0)

#define	FC_DBGTMDS0             0x1219
#define	FC_DBGTMDS1             0x121a
#define	FC_DBGTMDS2             0x121b

#define PHY_CONF0               0x3000  /* PHY Configuration Register */
#define m_POWER_DOWN_EN         BIT(7)/* no use */
#define v_POWER_DOWN_EN(n)      (((n) & 0x01) << 7)
#define m_TMDS_EN               BIT(6)/* no use */
#define v_TMDS_EN(n)            (((n) & 0x01) << 6)
#define m_SVSRET_SIG            BIT(5)/* depend on PHY_MHL_COMB0=1 */
#define v_SVSRET_SIG(n)         (((n) & 0x01) << 5)
#define m_PDDQ_SIG              BIT(4)
/*1: power down phy; 0: power on phy */
#define v_PDDQ_SIG(n)           (((n) & 0x01) << 4)
#define m_TXPWRON_SIG           BIT(3)
/*1: power on transmitter; 0: power down transmitter */
#define v_TXPWRON_SIG(n)        (((n) & 0x01) << 3)
#define m_ENHPD_RXSENSE_SIG     BIT(2)
/*1: enable detect hdp & rx sense */
#define v_ENHPD_RXSENSE_SIG(n)  (((n) & 0x01) << 2)
#define m_SEL_DATAEN_POL        BIT(1)
#define v_SEL_DATAEN_POL(n)     (((n) & 0x01) << 1)
#define m_SEL_INTERFACE         BIT(0)
#define v_SEL_INTERFACE(n)      (((n) & 0x01) << 0)

#define PHY_STAT0               0x3004  /* PHY RXSENSE, PLL Lock, and HPD Status Register */
#define m_PHY_STAT0_RX_SENSE_3      (1u<<7)
#define m_PHY_STAT0_RX_SENSE_2      (1u<<6)
#define m_PHY_STAT0_RX_SENSE_1_MASK (1u<<5)
#define m_PHY_STAT0_RX_SENSE_0_MASK (1u<<4)
#define m_PHY_STAT0_HPD             (1u<<1)
#define m_PHY_STAT0_TX_PHY_LOCK     (1u<<0)
#define PHY_INT0                0x3005
#define PHY_MASK                0x3006
#define PHY_POL0                0x3007  /* PHY RXSENSE, PLL Lock, and HPD Polarity Register */
#define m_PHY_RX_SENSE3         (1u<<7)
#define v_PHY_TX_SENSE3(n)      (((n) & 0x01) << 7)
#define m_PHY_RX_SENSE2         (1u<<6)
#define v_PHY_TX_SENSE2(n)      (((n) & 0x01) << 6)
#define m_PHY_RX_SENSE1         (1u<<5)
#define v_PHY_TX_SENSE1(n)      (((n) & 0x01) << 5)
#define m_PHY_RX_SENSE0         (1u<<4)
#define v_PHY_TX_SENSE0(n)      (((n) & 0x01) << 4)
#define m_PHY_HPD               (1u<<1)
#define v_PHY_HPD(n)            (((n) & 0x01) << 1)
#define m_PHY_LOCK              (1u<<0)
#define v_PHY_LOCK(n)           (((n) & 0x01) << 0)

/* Audio Sampler Registers */
#define HDMI_AUD_CONF0           0x3100
enum {
/* MC_SWRSTZ field values */
	HDMI_MC_SWRSTZ_I2SSWRST_REQ = 0x08,
	HDMI_MC_SWRSTZ_TMDSSWRST_REQ = 0x02,

/* HDMI_AUD_INPUTCLKFS field values */
	HDMI_AUD_INPUTCLKFS_128FS = 0,
	HDMI_AUD_INPUTCLKFS_256FS = 1,
	HDMI_AUD_INPUTCLKFS_512FS = 2,
	HDMI_AUD_INPUTCLKFS_64FS = 4,

/* CONFIG3_ID field values */
	HDMI_CONFIG3_AHBAUDDMA = 0x02,
	HDMI_CONFIG3_GPAUD = 0x01,

/* FC_AUDICONF0 field values */
	HDMI_FC_AUDICONF0_CC_OFFSET = 4,
	HDMI_FC_AUDICONF0_CC_MASK = 0x70,
	HDMI_FC_AUDICONF0_CT_OFFSET = 0,
	HDMI_FC_AUDICONF0_CT_MASK = 0xF,

/* FC_AUDICONF1 field values */
	HDMI_FC_AUDICONF1_SS_OFFSET = 3,
	HDMI_FC_AUDICONF1_SS_MASK = 0x18,
	HDMI_FC_AUDICONF1_SF_OFFSET = 0,
	HDMI_FC_AUDICONF1_SF_MASK = 0x7,

/* FC_AUDICONF3 field values */
	HDMI_FC_AUDICONF3_LFEPBL_OFFSET = 5,
	HDMI_FC_AUDICONF3_LFEPBL_MASK = 0x60,
	HDMI_FC_AUDICONF3_DM_INH_OFFSET = 4,
	HDMI_FC_AUDICONF3_DM_INH_MASK = 0x10,
	HDMI_FC_AUDICONF3_LSV_OFFSET = 0,
	HDMI_FC_AUDICONF3_LSV_MASK = 0xF,

/* FC_AUDSCONF field values */
	HDMI_FC_AUDSCONF_AUD_PACKET_SAMPFIT_MASK = 0xF0,
	HDMI_FC_AUDSCONF_AUD_PACKET_SAMPFIT_OFFSET = 4,
	HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_MASK = 0x1,
	HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_OFFSET = 0,
	HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_LAYOUT1 = 0x1,
	HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_LAYOUT0 = 0x0,

/* AUD_CONF0 field values */
	HDMI_AUD_CONF0_SW_RESET = 0x80,
	HDMI_AUD_CONF0_I2S_SELECT = 0x20,
	HDMI_AUD_CONF0_I2S_EN3 = 0x08,
	HDMI_AUD_CONF0_I2S_EN2 = 0x04,
	HDMI_AUD_CONF0_I2S_EN1 = 0x02,
	HDMI_AUD_CONF0_I2S_EN0 = 0x01,

/* AUD_CONF1 field values */
	HDMI_AUD_CONF1_MODE_I2S = 0x00,
	HDMI_AUD_CONF1_MODE_RIGHT_J = 0x20,
	HDMI_AUD_CONF1_MODE_LEFT_J = 0x40,
	HDMI_AUD_CONF1_MODE_BURST_1 = 0x60,
	HDMI_AUD_CONF1_MODE_BURST_2 = 0x80,
	HDMI_AUD_CONF1_WIDTH_16 = 0x10,
	HDMI_AUD_CONF1_WIDTH_24 = 0x18,

/* AUD_CTS3 field values */
	HDMI_AUD_CTS3_N_SHIFT_OFFSET = 5,
	HDMI_AUD_CTS3_N_SHIFT_MASK = 0xe0,
	HDMI_AUD_CTS3_N_SHIFT_1 = 0,
	HDMI_AUD_CTS3_N_SHIFT_16 = 0x20,
	HDMI_AUD_CTS3_N_SHIFT_32 = 0x40,
	HDMI_AUD_CTS3_N_SHIFT_64 = 0x60,
	HDMI_AUD_CTS3_N_SHIFT_128 = 0x80,
	HDMI_AUD_CTS3_N_SHIFT_256 = 0xa0,
	/* note that the CTS3 MANUAL bit has been removed
	   from our part. Can't set it, will read as 0. */
	HDMI_AUD_CTS3_CTS_MANUAL = 0x10,
	HDMI_AUD_CTS3_AUDCTS19_16_MASK = 0x0f,

/* MC_CLKDIS field values */
	HDMI_MC_CLKDIS_HDCPCLK_DISABLE = 0x40,
	HDMI_MC_CLKDIS_CECCLK_DISABLE = 0x20,
	HDMI_MC_CLKDIS_CSCCLK_DISABLE = 0x10,
	HDMI_MC_CLKDIS_AUDCLK_DISABLE = 0x8,
	HDMI_MC_CLKDIS_PREPCLK_DISABLE = 0x4,
	HDMI_MC_CLKDIS_TMDSCLK_DISABLE = 0x2,
	HDMI_MC_CLKDIS_PIXELCLK_DISABLE = 0x1,
};

#define HDMI_AUD_CONF1                          0x3101
#define HDMI_AUD_INT                            0x3102
#define HDMI_AUD_CONF2                          0x3103
#define HDMI_AUD_N1                             0x3200
#define HDMI_AUD_N2                             0x3201
#define HDMI_AUD_N3                             0x3202
#define HDMI_AUD_CTS1                           0x3203
#define HDMI_AUD_CTS2                           0x3204
#define HDMI_AUD_CTS3                           0x3205
#define HDMI_AUD_INPUTCLKFS                     0x3206
#define HDMI_AUD_SPDIFINT                       0x3302
#define HDMI_AUD_CONF0_HBR                      0x3400
#define HDMI_AUD_HBR_STATUS                     0x3401
#define HDMI_AUD_HBR_INT                        0x3402
#define HDMI_AUD_HBR_POL                        0x3403
#define HDMI_AUD_HBR_MASK                       0x3404

#define PHY_I2CM_SLAVE          0x3020  /* PHY I2C Slave Address Configuration Register */
#define PHY_I2CM_ADDRESS        0x3021  /* PHY I2C Address Configuration Register */
#define PHY_I2CM_DATAO_1        0x3022  /* PHY I2C Data Write Register 1 */
#define PHY_I2CM_DATAO_0        0x3023  /* PHY I2C Data Write Register 0 */
#define PHY_I2CM_DATAI_1        0x3024  /* PHY I2C Data Read Register 1 */
#define PHY_I2CM_DATAI_0        0x3025  /* PHY I2C Data Read Register 0 */
#define PHY_I2CM_OPERATION      0x3026  /* PHY I2C RD/RD_EXT/WR Operation Register */
#define PHY_I2CM_INT            0x3027  /* PHY I2C Done Interrupt Register */
#define PHY_I2CM_CTLINT         0x3028  /* PHY I2C error Interrupt Register */

#define PHY_I2CM_DIV            0x3029  /* PHY I2C Speed control Register */
#define m_PHY_I2CM_FAST_STD     (1u<<3)
#define v_PHY_I2CM_FAST_STD(n)  (((n) & 0x01) << 3)

#define PHY_I2CM_SOFTRSTZ       0x302A  /* PHY I2C SW reset control Register */
#define m_PHY_I2CM_SOFTRST      BIT(0)
#define v_PHY_I2CM_SOFTRST(n)   (((n) & 0x01) << 0)

#define PHY_I2CM_SS_SCL_HCNT_1_ADDR 0x302b
#define PHY_I2CM_SS_SCL_HCNT_0_ADDR 0x302c
#define PHY_I2CM_SS_SCL_LCNT_1_ADDR 0x302d
#define PHY_I2CM_SS_SCL_LCNT_0_ADDR 0x302e
#define PHY_I2CM_FS_SCL_HCNT_1_ADDR 0x302f
#define PHY_I2CM_FS_SCL_HCNT_0_ADDR 0x3030
#define PHY_I2CM_FS_SCL_LCNT_1_ADDR 0x3031
#define PHY_I2CM_FS_SCL_LCNT_0_ADDR 0x3032
#define PHY_I2CM_SDA_HOLD           0x3033

#define JTAG_PHY_CONFIG         0x3034  /* PHY I2C/JTAG/I/O Configuration Control Register */

#define AUD_CONF0               0x3100  /* Audio I2S Software FIFO Reset, select, and enable control register */

#define MC_CLKDIS               0x4001  /* Main Controller Synchronous Clock Domain Disable Register */
#define m_HDCPCLK_DISABLE       BIT(6)
#define v_HDCPCLK_DISABLE(n)    (((n) & 0x01) << 6)
#define m_CECCLK_DISABLE        BIT(5)
#define v_CECCLK_DISABLE(n)     (((n) & 0x01) << 5)
#define m_CSCCLK_DISABLE        BIT(4)
#define v_CSCCLK_DISABLE(n)     (((n) & 0x01) << 4)
#define m_AUDCLK_DISABLE        BIT(3)
#define v_AUDCLK_DISABLE(n)     (((n) & 0x01) << 3)
#define m_PREPCLK_DISABLE       BIT(2)
#define v_PREPCLK_DISABLE(n)    (((n) & 0x01) << 2)
#define m_TMDSCLK_DISABLE       BIT(1)
#define v_TMDSCLK_DISABLE(n)    (((n) & 0x01) << 1)
#define m_PIXELCLK_DISABLE      BIT(0)
#define v_PIXELCLK_DISABLE(n)   (((n) & 0x01) << 0)

#define MC_SWRSTZREQ			0x4002
#define m_IGPA_SWRST		BIT(7)
#define v_IGPA_SWRST(n)		(((n) & 0x01) << 7)
#define m_CEC_SWRST		BIT(6)
#define v_CEC_SWRST(n)		(((n) & 0x01) << 6)
#define m_ISPDIF_SWRST          BIT(4)
#define v_ISPDIF_SWRST(n)       (((n) & 0x01) << 4)
#define m_II2S_SWRST            BIT(3)
#define v_II2S_SWRST(n)         (((n) & 0x01) << 3)
#define m_PREP_SWRST            BIT(2)
#define v_PREP_SWRST(n)         (((n) & 0x01) << 2)
#define m_TMDS_SWRST		BIT(1)
#define v_TMDS_SWRST(n)		(((n) & 0x01) << 1)
#define m_PIXEL_SWRST           BIT(0)
#define v_PIXEL_SWRST(n)        (((n) & 0x01) << 0)

#define MC_OPCTRL			0x4003
#define m_HDCP_BLOCK_BYP	BIT(0)
#define v_HDCP_BLOCK_BYP(n)	(((n) & 0x01) << 0)

#define MC_FLOWCTRL             0x4004
#define m_FEED_THROUGH_OFF      BIT(0)
#define v_FEED_THROUGH_OFF(n)   (((n) & 0x01) << 0)


#define MC_PHYRSTZ              0x4005  /* Main Controller PHY Reset Register */
#define m_PHY_RSTZ              BIT(0)
#define v_PHY_RSTZ(n)           (((n) & 0x01) << 0)

#define MC_LOCKONCLOCK          0x4006
#define m_IGPACLK_ON            BIT(7)
#define v_IGPACLK_ON(n)         (((n) & 0x01) << 7)
#define m_PCLK_ON               BIT(6)
#define v_PCLK_ON(n)            (((n) & 0x01) << 6)
#define m_TMDSCLK_ON            BIT(5)
#define v_TMDSCLK_ON(n)         (((n) & 0x01) << 5)
#define m_PREPCLK_ON            BIT(4)
#define v_PREPCLK_ON(n)         (((n) & 0x01) << 4)
#define m_I2SCLK_ON             BIT(3)
#define v_I2SCLK_ON(n)          (((n) & 0x01) << 3)
#define m_SPDIFCLK_ON           BIT(2)
#define v_SPDIFCLK_ON(n)        (((n) & 0x01) << 2)
#define m_CECCLK_ON             BIT(0)
#define v_CECCLK_ON(n)          (((n) & 0x01) << 0)

/* Color Space Converter Registers */
#define COLOR_SPACE_CONVERTER_BASE  0x4100

#define	CSC_CFG                 0x4100
#define m_CSC_INTPMODE          (0x03 << 4)
#define v_CSC_INTPMODE(n)       (((n) & 0x03) << 4)
#define m_CSC_DECIMODE          (0x03 << 0)
#define v_CSC_DECIMODE(n)       (((n) & 0x03) << 0)

#define	CSC_SCALE               0x4101
#define m_CSC_COLOR_DEPTH       (0x0f << 4)
#define v_CSC_COLOR_DEPTH(n)    (((n) & 0x0f) << 4)
#define m_CSC_SCALE             (0x03 << 0)
#define v_CSC_SCALE(n)          (((n) & 0x03) << 0)

#define	CSC_COEF_A1_MSB         0x4102
#define	CSC_COEF_A1_LSB         0x4103
#define	CSC_COEF_A2_MSB         0x4104
#define	CSC_COEF_A2_LSB         0x4105
#define	CSC_COEF_A3_MSB         0x4106
#define	CSC_COEF_A3_LSB         0x4107
#define	CSC_COEF_A4_MSB         0x4108
#define	CSC_COEF_A4_LSB         0x4109
#define	CSC_COEF_B1_MSB         0x410a
#define	CSC_COEF_B1_LSB         0x410b
#define	CSC_COEF_B2_MSB         0x410c
#define	CSC_COEF_B2_LSB         0x410d
#define	CSC_COEF_B3_MSB         0x410e
#define	CSC_COEF_B3_LSB         0x410f
#define	CSC_COEF_B4_MSB         0x4110
#define	CSC_COEF_B4_LSB         0x4111
#define	CSC_COEF_C1_MSB         0x4112
#define	CSC_COEF_C1_LSB         0x4113
#define	CSC_COEF_C2_MSB         0x4114
#define	CSC_COEF_C2_LSB         0x4115
#define	CSC_COEF_C3_MSB         0x4116
#define	CSC_COEF_C3_LSB         0x4117
#define	CSC_COEF_C4_MSB         0x4118
#define	CSC_COEF_C4_LSB         0x4119
#define	CSC_SPARE_1             0x411a
#define	CSC_SPARE_2             0x411b


#define	A_VIDPOLCFG             0x5009
#define m_UNENCRYT_CONF         (0x03 << 5)
#define v_UNENCRYT_CONF(n)      (((n) & 0x03) << 5)
#define m_DATAEN_POL            BIT(4)
#define v_DATAEN_POL(n)         (((n) & 0x01) << 4)
#define m_VSYNC_POL             BIT(3)
#define v_VSYNC_POL(n)          (((n) & 0x01) << 3)
#define m_HSYNC_POL             BIT(1)
#define v_HSYNC_POL(n)          (((n) & 0x01) << 1)

#define I2CM_SLAVE              0x7e00  /* I2C DDC Slave address Configuration Register */
#define I2CM_ADDRESS            0x7e01  /* I2C DDC Address Configuration Register */
#define I2CM_DATAO              0x7e02  /* I2C DDC Data Write Register */
#define I2CM_DATAI              0x7e03  /* I2C DDC Data Read Register */
#define I2CM_OPERATION          0x7e04  /* I2C DDC RD/RD_EXT/WR Register */
#define m_I2CM_BUS_CLEAR		BIT(5)
#define v_I2CM_BUS_CLEAR(n)		(((n) & 0x01) << 5)
#define m_I2CM_WR		BIT(4)
#define v_I2CM_WR(n)		(((n) & 0x01) << 4)
#define m_I2CM_RD8_EXT		BIT(3)
#define v_I2CM_RD8_EXT(n)	(((n) & 0x01) << 3)
#define m_I2CM_RD8		BIT(2)
#define v_I2CM_RD8(n)		(((n) & 0x01) << 2)
#define m_I2CM_RD_EXT		BIT(1)
#define v_I2CM_RD_EXT(n)	(((n) & 0x01) << 1)
#define m_I2CM_RD		BIT(0)
#define v_I2CM_RD(n)		(((n) & 0x01) << 0)

#define	I2CM_INT			0x7e05
#define m_I2CM_RD_REQ_MASK	BIT(6)
#define v_I2CM_RD_REQ_MASK(n)	(((n) & 0x01) << 6)
#define m_I2CM_DONE_MASK	BIT(2)
#define v_I2CM_DONE_MASK(n)	(((n) & 0x01) << 2)

#define	I2CM_CTLINT			0x7e06
#define m_I2CM_NACK_MASK	BIT(6)
#define v_I2CM_NACK_MASK(n)	(((n) & 0x01) << 6)
#define m_I2CM_ARB_MASK		BIT(2)
#define v_I2CM_ARB_MASK(n)	(((n) & 0x01) << 2)

#define m_I2CM_OP_RD            (1u<<0) /* read operation request */
#define m_I2CM_OP_RD_EXT        (1u<<1) /* extended read operation request */
#define m_I2CM_OP_WR            (1u<<4) /* write operation request */

#define I2CM_DIV                0x7e07
enum {
    STANDARD_MODE = 0,
    FAST_MODE
};
#define m_I2CM_FAST_STD_MODE    (1u<<3)
#define v_I2CM_FAST_STD_MODE(n) (((n) & 0x01) << 3)
#define I2CM_SEGADDR            0x7e08  /* I2C DDC Segment Address Configuration Register */

#define I2CM_SOFTRSTZ           0x7e09
#define m_I2CM_SOFTRST          (1u<<0)
#define v_I2CM_SOFTRST(n)       (((n) & 0x01) << 0)

#define I2CM_SEGPTR             0x7e0a  /* I2C DDC Segment Pointer Register */
#define I2CM_SS_SCL_HCNT_1_ADDR 0x7e0b
#define I2CM_SS_SCL_HCNT_0_ADDR 0x7e0c
#define I2CM_SS_SCL_LCNT_1_ADDR 0x7e0d
#define I2CM_SS_SCL_LCNT_0_ADDR 0x7e0e
#define I2CM_FS_SCL_HCNT_1_ADDR 0x7e0f
#define I2CM_FS_SCL_HCNT_0_ADDR 0x7e10
#define I2CM_FS_SCL_LCNT_1_ADDR 0x7e11
#define I2CM_FS_SCL_LCNT_0_ADDR 0x7e12
#define I2CM_SDA_HOLD           0x7e13

#define I2CM_SCDC_READ_UPDATE   0x7e14  /* SCDC Control Register */
#define m_I2CM_UPRD_VSYNC_EN	BIT(5)
#define v_I2CM_UPRD_VSYNC_EN(n) (((n) & 0x01) << 5)
#define m_I2CM_READ_REQ_EN	BIT(4)
#define v_I2CM_READ_REQ_EN(n)	(((n) & 0x01) << 4)
#define m_I2CM_READ_UPDATE	BIT(0)
#define v_I2CM_READ_UPDATE(n)	(((n) & 0x01) << 0)

#define I2CM_SCDC_UPDATE0       0x7e30  /* I2C SCDC Read Update Register 0 */
#define I2CM_SCDC_UPDATE1       0x7e31  /* I2C SCDC Read Update Register 1 */

/*********************** GP201 HDMI PHY (16bit) *******************************/
#define PHY_PWRCTRL_OFFSET          0x0 /* Power Control Register */
#define m_PHY_PWRCTRL_OVERRIDE  (1u<<15)    /* Override bit */
#define m_PHY_PWRCTRL_BIAS_ON   (1u<< 5)    /* Bias for Transmission Power On */
#define m_PHY_PWRCTRL_TX_PWRON  (1u<< 4)    /* Transmitter Power On */
#define m_PHY_PWRCTRL_TX_PWRON0 (1u<< 3)    /* Transmitter Power On Zero */
#define m_PHY_PWRCTRL_TX_PWRON1 (1u<< 2)    /* Transmitter Power On One */
#define m_PHY_PWRCTRL_TX_PWRON2 (1u<< 1)    /* Transmitter Power On Two */
#define m_PHY_PWRCTRL_CK_PWRON  (1u<< 0)    /* Clock Power On */
#define v_PHY_PWRCTRL_OVERRIDE(n)   (((n) & 1u)<<15)    /* Override bit */
#define v_PHY_PWRCTRL_BIAS_ON(n)    (((n) & 1u)<< 5)    /* Bias for Transmission Power On */
#define v_PHY_PWRCTRL_TX_PWRON(n)   (((n) & 1u)<< 4)    /* Transmitter Power On */
#define v_PHY_PWRCTRL_TX_PWRON0(n)  (((n) & 1u)<< 3)    /* Transmitter Power On Zero */
#define v_PHY_PWRCTRL_TX_PWRON1(n)  (((n) & 1u)<< 2)    /* Transmitter Power On One */
#define v_PHY_PWRCTRL_TX_PWRON2(n)  (((n) & 1u)<< 1)    /* Transmitter Power On Two */
#define v_PHY_PWRCTRL_CK_PWRON(n)   (((n) & 1u)<< 0)    /* Clock Power On */

#define PHY_SERDIVCTRL_OFFSET       0x1 /* Serializer Divider Control Register */
#define m_PHY_SERDIVCTRL_OVERRIDE       (1u<<15)    /* Override bit */
#define m_PHY_SERDIVCTRL_TX_CK_EN0      (1u<<5) /* Enable of TMDSCLKIN clock gating */
#define m_PHY_SERDIVCTRL_TX_CK_EN1      (1u<<4) /* Enable of TMDSCLKIN clock gating */
#define m_PHY_SERDIVCTRL_TX_CK_EN2      (1u<<3) /* Enable of TMDSCLKIN clock gating */
#define m_PHY_SERDIVCTRL_TX_SER_DIV_EN0 (1u<<2) /* Transmitter Serializer Divider Enable Zero */
#define m_PHY_SERDIVCTRL_TX_SER_DIV_EN1 (1u<<1) /* Transmitter Serializer Divider Enable One */
#define m_PHY_SERDIVCTRL_TX_SER_DIV_EN2 (1u<<0) /* Transmitter Serializer Divider Enable Two */

#define PHYTX_OPMODE_PLLCFG         0x6 /* Mode of Operation and PLL Dividers Control Register */
/* Digital Pixel repetition divider.
 * Controls the ratio at which internal TMDS clock is divided to generate PREPCLK
 * 00: Divides by 1 (8 bits)
 * 01: Divides by 1.25 (10 bits)
 * 10: Divides by 1.5 (12 bits)
 * 11: Divides by 2 (16 bits) */
enum {
    PREP_DIV_BY_1 = 0,  /* 8 bits */
    PREP_DIV_BY_125,    /* 10 bits */
    PREP_DIV_BY_15,     /* 12 bits */
    PREP_DIV_BY_2,      /* 16 bits */
};
#define m_PREP_DIV      (0x03 << 12)
#define v_PREP_DIV(n)   (((n) & 0x03) << 12)

/* Clock mpll_cko divider
 * 11: Divides by 4
 * 10: Divides by 3
 * 01: Not used
 * 00: Divides by 1 */
enum {
    MPLL_CKO_DIV_BY_1 = 0,
    MPLL_CKO_DIV_NOT_USED,
    MPLL_CKO_DIV_BY_3,
    MPLL_CKO_DIV_BY_4,
};
#define m_MPLL_CKO_DIV      (0x03 << 9)
#define v_MPLL_CKO_DIV(n)   (((n) & 0x03) << 9)

/* Mode of Operation selection
 * 00: HDMI 1.4
 * 01: HDMI 2.0 (Data rate greater than 3.4 Gbps)
 * 10: Not used
 * 11: Not used */
enum OPMODE {
    OP_HDMI_14 = 0,
    OP_HDMI_20,
};
#define m_OPMODE        (0x03 << 6)
#define v_OPMODE(n)     (((n) & 0x03) << 6)

/* Programmable Input Divider Control
 * 11: Divides by 4
 * 10: Divides by 3
 * 01: Divides by 2
 * 00: Divides by 1 */
enum {
    REF_CNTRL_BY_1 = 1,
    REF_CNTRL_BY_2,
    REF_CNTRL_BY_3,
    REF_CNTRL_BY_4,
};
#define m_REF_CNTRL     (0x03 << 3)
#define v_REF_CNTRL(n)  (((n) & 0x03) << 3)

/* Programmable Output Divider Control
 * 11: Divides by 8
 * 10: Divides by 4
 * 01: Divides by 2
 * 00: Divides by 1 */
enum {
    MPLL_N_CNTRL_BY_1 = 1,
    MPLL_N_CNTRL_BY_2,
    MPLL_N_CNTRL_BY_4,
    MPLL_N_CNTRL_BY_8,
};
#define m_MPLL_N_CNTRL      (0x03 << 0)
#define v_MPLL_N_CNTRL(n)   (((n) & 0x03) << 0)

#define PHYTX_CLKSYMCTRL            0x09
#define v_OVERRIDE(n)               (0x01 << 15)
#define m_SLOPEBOOST                (0x0f << 10)
#define v_SLOPEBOOST(n)             (((n) & 0x0f) << 10)
#define m_TX_TRAON                  (0x01 << 9)
#define v_TX_TRAON(n)               (((n) & 0x01) << 9)
#define m_TX_TRBON                  (0x01 << 8)
#define v_TX_TRBON(n)               (((n) & 0x01) << 8)
#define m_TX_SYMON                  (0x0f << 4)
#define v_TX_SYMON(n)               (((n) & 0x0f) << 4)
#define m_CLK_SYMON                 (0x0f << 0)
#define v_CLK_SYMON(n)              (((n) & 0x0f) << 0)

#define PHYTX_VLEVCTRL              0x0e
#define m_MPLL_MEANS_LV17_14        (0xf << 10)
#define v_MPLL_MEANS_LV17_14(n)     (((n) & 0xf) << 10)
#define m_SUP_TXLVL                 (0x1f << 5)
#define v_SUP_TXLVL(n)              (((n) & 0x1f) << 5)

#define PHYTX_PLLCURRGMPCTRL        0x10    /* PLL Current and Gmp Control Register */
/* MPLL gmp control. This two bits value controls the effective
loop-filter resistance (=1/gmp) to increase or decrease MPLL
bandwidth and to compensate changes in divider module
(mpll_n_cntrl) */
#define m_MPLL_GMP_CNTRL        (0x03 << 12)
#define v_MPLL_GMP_CNTRL(n)     (((n) & 0x03) << 12)

/* PLL proportional control. This 5-bit value controls the PLL
charge pump proportional current. The specific values are
defined in MPLL Generic Configuration Settings. */
#define m_PROP_CNTRL            (0x3f << 6)
#define v_PROP_CNTRL(n)         (((n) & 0x3f) << 6)

/* PLL charge pump integral control. PLL charge pump integral control. This 5-bit value controls
the PLL charge pump integral current. The specific values
are defined in MPLL Generic Configuration Settings. */
#define m_INT_CNTRL             (0x3f << 0)
#define v_INT_CNTRL(n)          (((n) & 0x3f) << 0)

#define PHYTX_PLLDIVCTRL        0x11    /* PLL Dividers Control Register */
/* VCO operating range, needs to be set appropriately desired operating frequency */
#define m_VCO_CNTRL             (0xf<<9)
#define v_VCO_CNTRL(n)          (((n) & 0xf)<<9)

/* Feedback divider control */
#define m_MPLL_MULTIPLIER       (0xff<<0)
#define v_MPLL_MULTIPLIER(n)    (((n) & 0xff)<<0)

enum TERM_RESIS {
    R50_OHMS = 0,
    R5714_OHMS,
    R6667_OHMS,
    R80_OHMS,
    R100_OHMS,
    R13333_OHMS,
    R200_OHMS,
    ROPEN_CIRCUIT,
};

#define PHYTX_TERM_RESIS            0x19
#define m_TX_TERM                   (0x07 << 0)
#define v_TX_TERM(n)                (((n) & 0x07) << 0)

#endif
