/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_i2c.h
 *  版 本 号: V1.0
 *  作    者: 胡艳明
 *  生成日期: 2022年4月15日
 *
 *  功能描述: I2C驱动头文件
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 创建文件
 *
 ******************************************************************************/

#ifndef __LJMICRO_I2C_H__
#define __LJMICRO_I2C_H__

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/

#include "ljmicro_base.h"

/******************************************************************************
 *                                   数据类型定义                                   *
 ******************************************************************************/

struct i2c_dev
{
#ifdef UEFI_GOP
    void *Private; /* UEFI private data pointer */
#else
    void __iomem *regbase;
#endif

    unsigned int offset; /* base offset */
    int          index;

    unsigned int   tx_len;
    unsigned char *tx_buf;

    unsigned int   rx_len;
    unsigned char *rx_buf;

    unsigned int max_speed_mode;
    unsigned int rx_fifo_size;
    unsigned int tx_fifo_size;
};

/******************************************************************************
 *                                     宏定义                                     *
 ******************************************************************************/

#ifdef UEFI_GOP
#define I2C_REG_W8(i2c_dev, addr, val)  WREG8_BASE(i2c_dev->Private, i2c_dev->offset + (addr), (val))
#define I2C_REG_R8(i2c_dev, addr)       RREG8_BASE(i2c_dev->Private, i2c_dev->offset + (addr))
#define I2C_REG_W32(i2c_dev, addr, val) WREG32_BASE(i2c_dev->Private, i2c_dev->offset + (addr), (val))
#define I2C_REG_R32(i2c_dev, addr)      RREG32_BASE(i2c_dev->Private, i2c_dev->offset + (addr))
#else
#define I2C_REG_W8(i2c_dev, addr, val)  WREG8_BASE(i2c_dev->regbase, i2c_dev->offset + (addr), (val))
#define I2C_REG_R8(i2c_dev, addr)       RREG8_BASE(i2c_dev->regbase, i2c_dev->offset + (addr))
#define I2C_REG_W32(i2c_dev, addr, val) WREG32_BASE(i2c_dev->regbase, i2c_dev->offset + (addr), (val))
#define I2C_REG_R32(i2c_dev, addr)      RREG32_BASE(i2c_dev->regbase, i2c_dev->offset + (addr))
#endif

/******************************************************************************
 *                                    函数声明                                    *
 ******************************************************************************/

int  ljm_i2c_oc_init(struct i2c_dev *i2c_dev, void *base, int offset, int index);
bool ljm_i2c_oc_detect(struct i2c_dev *i2c_dev);
int  ljm_i2c_oc_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size);
int  ljm_i2c_oc_write_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size);
int  ljm_i2c_oc_raw_write_req(void *hw_dev, unsigned char target, unsigned char data);

// struct edid *ljm_i2c_oc_get_edid(struct i2c_dev *i2c_dev, int *psize);

int  ljm_i2c_dw_init(struct i2c_dev *i2c_dev, void *base, int offset, int index);
bool ljm_i2c_dw_detect(struct i2c_dev *i2c_dev, unsigned char target);
int  ljm_i2c_dw_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size);
int  ljm_i2c_dw_write_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size);
int  ljm_i2c_dw_raw_write_req(void *hw_dev, unsigned char target, unsigned char data);

#endif /* __LJMICRO_I2C_H__ */
