/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/**************************************************************************
* 版权所有 (C)2022, 武汉凌久微电子有限公司。
*
* 文件名称： ljmicro_i2s.c
* 文件标识：
* 内容摘要： 本文件定义IIS驱动，注册CPU_DAI
* 其它说明：
* 当前版本： v1.0
* 作    者： zhoupw, 武汉凌久微电子有限公司
* 完成日期： 2022年4月27日
***************************************************************************/
#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)

#include <linux/irq.h>
#include <linux/pm_runtime.h>
#include <sound/designware_i2s.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>

#include "ljmicro_device.h"
#include "gp201.h"
#include "ljmicro_i2s.h"
#include "lj_hal_kernel_linux.h"

#define MISC_I2S_ENABLE    0x180
#define CLK_I2S_ENABLE     0x40

static void __iomem *clk_base;
static void __iomem *misc_base;

struct ljm_device *get_ljm_dev(void);

static int calculate_i2s_clk(struct ljm_i2s_dev *dev)
{
    int val = 0;
    char flag_even = 0;
    struct ljm_device *ljm_dev = get_ljm_dev();
    const int million = 1000000;

    val = ljm_dev->axi_bus_clk * million / (2 * dev->config.sample_rate * 64);

    flag_even = val % 2;

    if (!flag_even) {
        return val;
    }
    else {
        return val + 1;
    }
}

static void set_i2s_clk(struct ljm_i2s_dev *dev)
{
    int val = 0;
    val = calculate_i2s_clk(dev);
    val |= val << 16;

    writel(val, clk_base + CLK_I2S_ENABLE);
    writel(0xF, misc_base + MISC_I2S_ENABLE); //使能4路输出
}

static inline void i2s_disable_channels(struct ljm_i2s_dev *dev, u32 stream)
{
    u32 i = 0;

    if (stream == SNDRV_PCM_STREAM_PLAYBACK) {
        for (i = 0; i < 4; i++)
            i2s_write_reg(dev->i2s_base, TER(i), 0);
    }
    else {
        for (i = 0; i < 4; i++)
            i2s_write_reg(dev->i2s_base, RER(i), 0);
    }
}

static inline void i2s_clear_irqs(struct ljm_i2s_dev *dev, u32 stream)
{
    u32 i = 0;

    if (stream == SNDRV_PCM_STREAM_PLAYBACK) {
        for (i = 0; i < 4; i++)
            i2s_read_reg(dev->i2s_base, TOR(i));
    }
    else {
        for (i = 0; i < 4; i++)
            i2s_read_reg(dev->i2s_base, ROR(i));
    }
}

static inline void i2s_disable_irqs(struct ljm_i2s_dev *dev, u32 stream, int chan_nr)
{
    u32 i, irq;

    if (stream == SNDRV_PCM_STREAM_PLAYBACK) {
        for (i = 0; i < (chan_nr / 2); i++) {
            irq = i2s_read_reg(dev->i2s_base, IMR(i));
            i2s_write_reg(dev->i2s_base, IMR(i), irq | 0x30);
        }
    }
    else {
        for (i = 0; i < (chan_nr / 2); i++) {
            irq = i2s_read_reg(dev->i2s_base, IMR(i));
            i2s_write_reg(dev->i2s_base, IMR(i), irq | 0x03);
        }
    }
}

static inline void i2s_enable_irqs(struct ljm_i2s_dev *dev, u32 stream, int chan_nr)
{
    u32 i, irq;

    if (stream == SNDRV_PCM_STREAM_PLAYBACK) {
        for (i = 0; i < (chan_nr / 2); i++) {
            irq = i2s_read_reg(dev->i2s_base, IMR(i));
            i2s_write_reg(dev->i2s_base, IMR(i), irq & ~0x30);
        }
    }
    else {
        for (i = 0; i < (chan_nr / 2); i++) {
            irq = i2s_read_reg(dev->i2s_base, IMR(i));
            i2s_write_reg(dev->i2s_base, IMR(i), irq & ~0x03);
        }
    }
}

static void i2s_start(struct ljm_i2s_dev *dev, struct snd_pcm_substream *substream)
{
//    struct i2s_clk_config_data *config = &dev->config;

    i2s_write_reg(dev->i2s_base, IER, 1);
//    i2s_enable_irqs(dev, substream->stream, config->chan_nr);

    if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
        i2s_write_reg(dev->i2s_base, ITER, 1);
        i2s_write_reg(dev->i2s_base, DMACR, 0x0);
        i2s_write_reg(dev->i2s_base, DMACR, 0x20000);
    }
    else
        i2s_write_reg(dev->i2s_base, IRER, 1);

    i2s_write_reg(dev->i2s_base, CER, 1);
}

static void i2s_stop(struct ljm_i2s_dev *dev, struct snd_pcm_substream *substream)
{
    i2s_clear_irqs(dev, substream->stream);
    if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
        i2s_write_reg(dev->i2s_base, ITER, 0);
    else
        i2s_write_reg(dev->i2s_base, IRER, 0);

    i2s_disable_irqs(dev, substream->stream, 8);

    if (!dev->active) {
        i2s_write_reg(dev->i2s_base, CER, 0);
        i2s_write_reg(dev->i2s_base, IER, 0);
    }
}

static void ljm_i2s_config(struct ljm_i2s_dev *dev, int stream)
{
    u32 ch_reg;
    struct i2s_clk_config_data *config = &dev->config;

    i2s_disable_channels(dev, stream);

    for (ch_reg = 0; ch_reg < (config->chan_nr / 2); ch_reg++) {
        if (stream == SNDRV_PCM_STREAM_PLAYBACK) {
            i2s_write_reg(dev->i2s_base, TCR(ch_reg), dev->xfer_resolution);
            i2s_write_reg(dev->i2s_base, TFCR(ch_reg), dev->fifo_th - 1);
            i2s_write_reg(dev->i2s_base, TER(ch_reg), 1);
        }
        else {
            i2s_write_reg(dev->i2s_base, RCR(ch_reg), dev->xfer_resolution);
            i2s_write_reg(dev->i2s_base, RFCR(ch_reg), dev->fifo_th - 1);
            i2s_write_reg(dev->i2s_base, RER(ch_reg), 1);
        }
    }
}

static int ljm_i2s_startup(struct snd_pcm_substream *substream, struct snd_soc_dai *cpu_dai)
{
    int i;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;
    union ljm_i2s_snd_dma_data *dma_data = NULL;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        if (!(dev->capability & DWC_I2S_RECORD) && (substream->stream == SNDRV_PCM_STREAM_CAPTURE))
            return -EINVAL;

        if (!(dev->capability & DWC_I2S_PLAY) && (substream->stream == SNDRV_PCM_STREAM_PLAYBACK))
            return -EINVAL;

        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
            dma_data = &dev->play_dma_data;
        }
        else if (substream->stream == SNDRV_PCM_STREAM_CAPTURE) {
            dma_data = &dev->capture_dma_data;
        }

        snd_soc_dai_set_dma_data(cpu_dai, substream, (void *)dma_data);
    }

    return 0;
}

static void ljm_i2s_shutdown(struct snd_pcm_substream *substream, struct snd_soc_dai *dai)
{
    snd_soc_dai_set_dma_data(dai, substream, NULL);
}

static int ljm_i2s_hw_params(struct snd_pcm_substream *substream,
        struct snd_pcm_hw_params *params, struct snd_soc_dai *dai)
{
    int i;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;
    struct i2s_clk_config_data *config = NULL;
    int ret;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        config = &dev->config;
        dev->poll_time_ns = 1000000000 / params_rate(params) * dev->fifo_th / 4;

        switch (params_format(params)) {
            case SNDRV_PCM_FORMAT_S16_LE:
                config->data_width = 16;
                dev->ccr = 0x00;
                dev->xfer_resolution = 0x02;
                break;
            case SNDRV_PCM_FORMAT_S24_LE:
                config->data_width = 24;
                dev->ccr = 0x08;
                dev->xfer_resolution = 0x04;
                break;
            case SNDRV_PCM_FORMAT_S32_LE:
                config->data_width = 32;
                dev->ccr = 0x10;
                dev->xfer_resolution = 0x05;
                break;
            default:
                dev_err(dev->dev, "LJM IIS: unsupported PCM fmt");
                return -EINVAL;
        }

        config->chan_nr = params_channels(params);

        switch (config->chan_nr) {
            case EIGHT_CHANNEL_SUPPORT:
            case SIX_CHANNEL_SUPPORT:
            case FOUR_CHANNEL_SUPPORT:
            case TWO_CHANNEL_SUPPORT:
                break;
            default:
                dev_err(dev->dev, "LJM IIS: channel not supported\n");
                return -EINVAL;
        }

        ljm_i2s_config(dev, substream->stream);

        i2s_write_reg(dev->i2s_base, CCR, dev->ccr);

        config->sample_rate = params_rate(params);

        if (dev->capability & DW_I2S_MASTER) {
            set_i2s_clk(dev);
            if (dev->i2s_clk_cfg) {
                ret = dev->i2s_clk_cfg(config);
                if (ret < 0) {
                    dev_err(dev->dev, "LJM IIS: runtime audio clk config fail\n");
                    return ret;
                }
            }
            else {
#ifdef CONFIG_MIPS
#else
                u32 bitclk = config->sample_rate * config->data_width * 2;
                ret = clk_set_rate(dev->clk, bitclk);
                if (ret) {
                    dev_err(dev->dev, "LJM IIS: Can't set I2S clock rate: %d\n", ret);
                    return ret;
                }
#endif
            }
        }
    }

    return 0;
}

static int ljm_i2s_prepare(struct snd_pcm_substream *substream, struct snd_soc_dai *dai)
{
    int i;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
            i2s_write_reg(dev->i2s_base, TXFFR, 1);
        else
            i2s_write_reg(dev->i2s_base, RXFFR, 1);
    }

    return 0;
}

static int ljm_i2s_trigger(struct snd_pcm_substream *substream, int cmd, struct snd_soc_dai *dai)
{
    int i;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;
    int ret = 0;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        switch (cmd) {
            case SNDRV_PCM_TRIGGER_START:
            case SNDRV_PCM_TRIGGER_RESUME:
            case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
                dev->active++;
                i2s_start(dev, substream);
                break;
            case SNDRV_PCM_TRIGGER_STOP:
            case SNDRV_PCM_TRIGGER_SUSPEND:
            case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
                dev->active--;
                i2s_stop(dev, substream);
                break;
            default:
                ret = -EINVAL;
                break;
        }
    }

    return ret;
}

static int ljm_i2s_set_fmt(struct snd_soc_dai *cpu_dai, unsigned int fmt)
{
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;
    int i;
    int ret = 0;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        if (!dev)
            return -1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
        switch (fmt & SND_SOC_DAIFMT_CLOCK_PROVIDER_MASK) {
            case SND_SOC_DAIFMT_BC_FC:
                if (dev->capability & DW_I2S_SLAVE)
                    ret = 0;
                else
                    ret = -EINVAL;
                break;
            case SND_SOC_DAIFMT_BP_FP:
                if (dev->capability & DW_I2S_MASTER)
                    ret = 0;
                else
                    ret = -EINVAL;
                break;
            case SND_SOC_DAIFMT_BC_FP:
            case SND_SOC_DAIFMT_BP_FC:
                ret = -EINVAL;
                break;
            default:
                dev_dbg(dev->dev, "dwc : Invalid clock provider format\n");
                ret = -EINVAL;
                break;
        }

        switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
            case SND_SOC_DAIFMT_I2S:
            case SND_SOC_DAIFMT_LEFT_J:
            case SND_SOC_DAIFMT_RIGHT_J:
                break;
            case SND_SOC_DAIFMT_DSP_A:
                dev->frame_offset = 1;
                break;
            case SND_SOC_DAIFMT_DSP_B:
                dev->frame_offset = 0;
                break;
            default:
                dev_err(dev->dev, "DAI format unsupported");
                return -EINVAL;
        }
#else
        switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
            case SND_SOC_DAIFMT_CBM_CFM:
                if (dev->capability & DW_I2S_SLAVE)
                    ret = 0;
                else
                    ret = -EINVAL;
                break;
            case SND_SOC_DAIFMT_CBS_CFS:
                if (dev->capability & DW_I2S_MASTER)
                    ret = 0;
                else
                    ret = -EINVAL;
                break;
            case SND_SOC_DAIFMT_CBM_CFS:
            case SND_SOC_DAIFMT_CBS_CFM:
                ret = -EINVAL;
                break;
            default:
                dev_err(dev->dev, "LJM IIS: Invalid master/slave format\n");
                ret = -EINVAL;
                break;
        }
#endif
    }

    return ret;
}

static const struct snd_soc_dai_ops ljm_i2s_dai_ops = {
    .startup	= ljm_i2s_startup,
    .shutdown	= ljm_i2s_shutdown,
    .hw_params	= ljm_i2s_hw_params,
    .prepare	= ljm_i2s_prepare,
    .trigger	= ljm_i2s_trigger,
    .set_fmt	= ljm_i2s_set_fmt,
};

#ifdef CONFIG_PM
#if 0
static int ljm_i2s_runtime_suspend(struct device *dev)
{
    int i;
    struct ljm_device *ljm_dev = dev_get_drvdata(dev);
    struct ljm_i2s_dev *i2s_dev = NULL;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        i2s_dev = ljm_dev->i2s_devs[i];
        if (i2s_dev->capability & DW_I2S_MASTER)
            clk_disable(i2s_dev->clk);
    }

    return 0;
}

static int ljm_i2s_runtime_resume(struct device *dev)
{
    int i;
    struct ljm_device *ljm_dev = dev_get_drvdata(dev);
    struct ljm_i2s_dev *i2s_dev = NULL;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        i2s_dev = ljm_dev->i2s_devs[i];
        if (i2s_dev->capability & DW_I2S_MASTER)
            clk_enable(i2s_dev->clk);
    }

    return 0;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_i2s_suspend(struct snd_soc_component *component)
#else
static int ljm_i2s_suspend(struct snd_soc_dai *dai)
#endif
{
    int i;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        if (dev->capability & DW_I2S_MASTER)
            clk_disable(dev->clk);
    }

    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_i2s_resume(struct snd_soc_component *component)
#else
static int ljm_i2s_resume(struct snd_soc_dai *dai)
#endif
{
    int i;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_i2s_dev *dev = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    struct snd_soc_dai *dai;
    int stream;
#endif

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        if (dev->capability & DW_I2S_MASTER)
            clk_enable(dev->clk);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
        for_each_component_dais(component, dai) {
            for_each_pcm_streams(stream)
                if (snd_soc_dai_stream_active(dai, stream))
                    ljm_i2s_config(dev, stream);
        }
#else
        if (dai->playback_active)
            ljm_i2s_config(dev, SNDRV_PCM_STREAM_PLAYBACK);
        if (dai->capture_active)
            ljm_i2s_config(dev, SNDRV_PCM_STREAM_CAPTURE);
#endif
    }

    return 0;
}
#else
#define ljm_i2s_suspend NULL
#define ljm_i2s_resume  NULL
#endif

static const struct snd_soc_component_driver ljm_i2s_component = {
    .name = "ljm-i2s",
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    .suspend = ljm_i2s_suspend,
    .resume = ljm_i2s_resume,
#endif
};

/* Maximum bit resolution of a channel - not uniformly spaced */
static const u32 fifo_width[COMP_MAX_WORDSIZE] = {
    12, 16, 20, 24, 32, 0, 0, 0
};

/* Width of (DMA) bus */
static const u32 bus_widths[COMP_MAX_DATA_WIDTH] = {
    DMA_SLAVE_BUSWIDTH_1_BYTE,
    DMA_SLAVE_BUSWIDTH_2_BYTES,
    DMA_SLAVE_BUSWIDTH_4_BYTES,
    DMA_SLAVE_BUSWIDTH_UNDEFINED
};

/* PCM format to support channel resolution */
static const u32 formats[COMP_MAX_WORDSIZE] = {
    SNDRV_PCM_FMTBIT_S16_LE,
    SNDRV_PCM_FMTBIT_S16_LE,
    SNDRV_PCM_FMTBIT_S24_LE,
    SNDRV_PCM_FMTBIT_S24_LE,
    SNDRV_PCM_FMTBIT_S32_LE,
    0,
    0,
    0
};

static int ljm_configure_dai(struct ljm_i2s_dev *dev,
        struct snd_soc_dai_driver *ljm_i2s_dai,
        unsigned int rates)
{
    /*
     * Read component parameter registers to extract
     * the I2S block's configuration.
     */
    u32 comp1 = i2s_read_reg(dev->i2s_base, dev->i2s_reg_comp1);
    u32 comp2 = i2s_read_reg(dev->i2s_base, dev->i2s_reg_comp2);
    u32 fifo_depth = 1 << (1 + COMP1_FIFO_DEPTH_GLOBAL(comp1));
    u32 idx;

    if (dev->capability & DWC_I2S_RECORD && dev->quirks & DW_I2S_QUIRK_COMP_PARAM1)
        comp1 = comp1 & ~BIT(5);

    if (dev->capability & DWC_I2S_PLAY && dev->quirks & DW_I2S_QUIRK_COMP_PARAM1)
        comp1 = comp1 & ~BIT(6);

    if (COMP1_TX_ENABLED(comp1)) {
        idx = COMP1_TX_WORDSIZE_0(comp1);
        if (WARN_ON(idx >= ARRAY_SIZE(formats)))
            return -EINVAL;
        if (dev->quirks & DW_I2S_QUIRK_16BIT_IDX_OVERRIDE)
            idx = 1;
        ljm_i2s_dai->playback.channels_min = MIN_CHANNEL_NUM;
        ljm_i2s_dai->playback.channels_max = 1 << (COMP1_TX_CHANNELS(comp1) + 1);
        ljm_i2s_dai->playback.formats = formats[idx];
        ljm_i2s_dai->playback.rates = rates;
    }

    if (COMP1_RX_ENABLED(comp1)) {
        idx = COMP2_RX_WORDSIZE_0(comp2);
        if (WARN_ON(idx >= ARRAY_SIZE(formats)))
            return -EINVAL;
        if (dev->quirks & DW_I2S_QUIRK_16BIT_IDX_OVERRIDE)
            idx = 1;
        ljm_i2s_dai->capture.channels_min = MIN_CHANNEL_NUM;
        ljm_i2s_dai->capture.channels_max = 1 << (COMP1_RX_CHANNELS(comp1) + 1);
        ljm_i2s_dai->capture.formats = formats[idx];
        ljm_i2s_dai->capture.rates = rates;
    }

    if (COMP1_MODE_EN(comp1)) {
        dev->capability |= DW_I2S_MASTER;
    }
    else {
        dev->capability |= DW_I2S_SLAVE;
    }

    dev->fifo_th = fifo_depth / 2;
    return 0;
}

static int ljm_configure_dai_by_dt(struct ljm_i2s_dev *dev,
        struct snd_soc_dai_driver *ljm_i2s_dai,
        resource_size_t res_start)
{
    u32 comp1 = i2s_read_reg(dev->i2s_base, I2S_COMP_PARAM_1);
    u32 comp2 = i2s_read_reg(dev->i2s_base, I2S_COMP_PARAM_2);
    u32 fifo_depth = 1 << (1 + COMP1_FIFO_DEPTH_GLOBAL(comp1));
    u32 idx = COMP1_APB_DATA_WIDTH(comp1);
    u32 idx2;
    int ret;

    if (WARN_ON(idx >= ARRAY_SIZE(bus_widths)))
        return -EINVAL;

    ret = ljm_configure_dai(dev, ljm_i2s_dai, SNDRV_PCM_RATE_8000_192000);
    if (ret < 0)
        return ret;

    if (COMP1_TX_ENABLED(comp1)) {
        idx2 = COMP1_TX_WORDSIZE_0(comp1);

        dev->capability |= DWC_I2S_PLAY;
        dev->play_dma_data.dt.addr = res_start + I2S_TXDMA;
        dev->play_dma_data.dt.addr_width = bus_widths[idx];
        dev->play_dma_data.dt.fifo_size = fifo_depth * (fifo_width[idx2]) >> 8;
        dev->play_dma_data.dt.maxburst = 16;
    }
    if (COMP1_RX_ENABLED(comp1)) {
        idx2 = COMP2_RX_WORDSIZE_0(comp2);

        dev->capability |= DWC_I2S_RECORD;
        dev->capture_dma_data.dt.addr = res_start + I2S_RXDMA;
        dev->capture_dma_data.dt.addr_width = bus_widths[idx];
        dev->capture_dma_data.dt.fifo_size = fifo_depth * (fifo_width[idx2] >> 8);
        dev->capture_dma_data.dt.maxburst = 16;
    }

    return 0;
}

int ljm_i2s_register(struct ljm_device *ljm_dev)
{
    int i;
    int ret;
    struct snd_soc_dai_driver *ljm_i2s_dai;

    ljm_i2s_dai = devm_kzalloc(ljm_dev->dev, sizeof(*ljm_i2s_dai), GFP_KERNEL);
    if (!ljm_i2s_dai)
        return -ENOMEM;

    ljm_i2s_dai->ops = &ljm_i2s_dai_ops;
    ljm_i2s_dai->name = "ljm-i2s";
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
    ljm_i2s_dai->suspend = ljm_i2s_suspend;
    ljm_i2s_dai->resume = ljm_i2s_resume;
#endif

    clk_base = ljm_dev->mmio_base + ljm_gpu_hardware.reg_cru_offset;
    misc_base = ljm_dev->mmio_base + ljm_gpu_hardware.reg_misc_offset;


    for (i = 0; i < LJM_MAX_I2SS; i++) {

        ljm_dev->i2s_devs[i]->dev = ljm_dev->dev;

        ljm_dev->i2s_devs[i]->i2s_reg_comp1 = I2S_COMP_PARAM_1;
        ljm_dev->i2s_devs[i]->i2s_reg_comp2 = I2S_COMP_PARAM_2;

        ret = ljm_configure_dai_by_dt(ljm_dev->i2s_devs[i], ljm_i2s_dai,
            (resource_size_t)ljm_dev->i2s_devs[i]->i2s_base);
        if (ret < 0) {
            return ret;
        }
        ljm_dev->i2s_devs[i]->use_pio = true;
    }

    ret = devm_snd_soc_register_component(ljm_dev->dev, &ljm_i2s_component, ljm_i2s_dai, 1);
    if (ret != 0) {
        dev_err(ljm_dev->dev, "LJM IIS: not able to register dai.\n");
        return ret;
    }

    return 0;
}

#endif
