/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，武汉凌久微电子有限公司
 *
 ******************************************************************************
 *
 *  文件名称：ljmicro_info.c
 *  版 本 号：V1.0
 *  作    者：zhaiyf
 *  生成日期：2023年1月9日
 *
 *  功能描述：提供不同型号drm_info模板，根据board_info开关
              视频接口。针对drm_connector和drm_encoder属性的
              改大多可在此处进行。文件还包括了针对
              struct ljm_drm_info的一些操作函数
 *
 *  修改历史：
 *         1. 日    期：2023年1月5日
 *            作    者：zhaiyf
 *            修改内容：创建文件
 *         2. 日    期：2023年1月9日
 *            作    者：zhaiyf
 *            修改内容：添加注释
 *
 *****************************************************************************/

/******************************************************************************
 *                                   标准库头文件                                   *
 ******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/
#include "ljmicro_info.h"

/******************************************************************************
 *                                     宏定义                                     *
 ******************************************************************************/


/******************************************************************************
 *                                   数据类型定义                                   *
 ******************************************************************************/


/******************************************************************************
 *                                    常量定义                                    *
 ******************************************************************************/


/******************************************************************************
 *                                   GPU Info                                   *
 ******************************************************************************/
/*----------------------GP101-----------------------*/
struct ljm_drm_info ljm_gp101_info = {
    .gpu_family = CHIP_GP100,
    .ports_num = 4,

    .ports = {
        /* PORT0(VGA0): encoder-2 connector-2 */
        {
            LJM_PORT_INFO(0, LJM_PORT_VGA, 0, LJM_PORT_HPD_DISABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(2, LJM_ENCODER_NONE, 0x1, 0x0),
                LJM_CONNECTOR_INFO(2, DRM_MODE_CONNECTOR_VGA, DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT,
                    LJM_DDC_TYPE_OCI2C, LJM_HPD_TYPE_NONE))
        },
        /* PORT1(VGA1): encoder-3 connector-3 */
        {
            LJM_PORT_INFO(1, LJM_PORT_VGA, 1, LJM_PORT_HPD_DISABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(3, LJM_ENCODER_NONE, 0x2, 0x0),
                LJM_CONNECTOR_INFO(3, DRM_MODE_CONNECTOR_VGA, DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT,
                    LJM_DDC_TYPE_OCI2C, LJM_HPD_TYPE_NONE))
        },
        /* PORT(VGA2): CLOSE*/
        {0},
        /* PORT(VGA3): CLOSE*/
        {0},

        /* PORT2(HDMI0): encoder-0 connector-0 */
        {
            LJM_PORT_INFO(2, LJM_PORT_HDMI, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(0, LJM_ENCODER_HDMIV1, 0x1, 0x0),
                LJM_CONNECTOR_INFO(0, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV1, LJM_HPD_TYPE_HDMIV1))
        },
        /* PORT3(HDMI1): encoder-1 connector-1 */
        {
            LJM_PORT_INFO(3, LJM_PORT_HDMI, 1, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(1, LJM_ENCODER_HDMIV1, 0x2, 0x0),
                LJM_CONNECTOR_INFO(1, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV1, LJM_HPD_TYPE_HDMIV1))
        },

    },
};

/*----------------------GP201-----------------------*/
struct ljm_drm_info ljm_gp201_info = {
    .gpu_family = CHIP_GP200,
    .ports_num = 11,

    .ports = {
        /* PORT0(VGA0): encoder-0 connector-0*/
        {
            LJM_PORT_INFO(0, LJM_PORT_VGA, 0, LJM_PORT_HPD_DISABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(0, LJM_ENCODER_DAC, 0xe, 0xffff),
                LJM_CONNECTOR_INFO(0, DRM_MODE_CONNECTOR_VGA, DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT,
                    LJM_DDC_TYPE_OCI2C, LJM_HPD_TYPE_DAC))
        },
        /* PORT(VGA1): CLOSE*/
        {0},
        /* PORT(VGA2): CLOSE*/
        {0},
        /* PORT(VGA3): CLOSE*/
        {0},

        /* PORT1(HDMI0): encoder-4 connector-4*/
        {
            LJM_PORT_INFO(1, LJM_PORT_HDMI, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(4, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(4, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },
        /* PORT_2(HDMI1): encoder-5 connector-5*/
        {
            LJM_PORT_INFO(2, LJM_PORT_HDMI, 1, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(5, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(5, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },
        /* PORT_3(HDMI2): encoder-6 connector-6*/
        {
            LJM_PORT_INFO(3, LJM_PORT_HDMI, 2, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(6, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(6, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },
        /* PORT_4(HDMI3->DVO): encoder-7 connector-7*/
        {
            LJM_PORT_INFO(4, LJM_PORT_HDMI, 3, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(7, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(7, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },

        /* PORT_5(DVO0): encoder-8 connector-8*/
        {
            LJM_PORT_INFO(5, LJM_PORT_DVO, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(8, LJM_ENCODER_DVO, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(8, DRM_MODE_CONNECTOR_DVID, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_OCI2C, LJM_HPD_TYPE_CRTC))
        },
        /* PORT6(DVO1): encoder-9 connector-9*/
        {
            LJM_PORT_INFO(6, LJM_PORT_DVO, 1, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(9, LJM_ENCODER_DVO, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(9, DRM_MODE_CONNECTOR_DVID, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_OCI2C, LJM_HPD_TYPE_CRTC))
        },
        /* PORT(DVO2): CLOSE*/
        {0},
        /* PORT(DVO3): CLOSE*/
        {0},

        /* PORT7(DP0): encoder-12 connector-12*/
        {
            LJM_PORT_INFO(7, LJM_PORT_DP, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(12, LJM_ENCODER_DP, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(12, DRM_MODE_CONNECTOR_DisplayPort, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DP, LJM_HPD_TYPE_DP))
        },
        /* PORT8(DP1): encoder-13 connector-13*/
        {
            LJM_PORT_INFO(8, LJM_PORT_DP, 1, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(13, LJM_ENCODER_DP, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(13, DRM_MODE_CONNECTOR_DisplayPort, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DP, LJM_HPD_TYPE_DP))

        },
        /* PORT9(DP2): encoder-14 connector-14*/
        {
            LJM_PORT_INFO(9, LJM_PORT_DP, 2, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(14, LJM_ENCODER_DP, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(14, DRM_MODE_CONNECTOR_DisplayPort, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DP, LJM_HPD_TYPE_DP))
        },
        /* PORT10(edp->DP3): encoder-15 connector-15*/
        {
            LJM_PORT_INFO(10, LJM_PORT_DP, 3, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(15, LJM_ENCODER_DP, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(15, DRM_MODE_CONNECTOR_DisplayPort, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DP, LJM_HPD_TYPE_DP))
        }
    }
};

/*----------------------gp202-----------------------*/
struct ljm_drm_info ljm_gp202_info = {
    .gpu_family = CHIP_GP200,
    .ports_num = 6,

    .ports = {
        /* PORT0(VGA0): encoder-0 connector-0*/
        {
            LJM_PORT_INFO(0, LJM_PORT_VGA, 0, LJM_PORT_HPD_DISABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(0, LJM_ENCODER_DAC, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(0, DRM_MODE_CONNECTOR_VGA, DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT,
                    LJM_DDC_TYPE_DWI2C, LJM_HPD_TYPE_DAC))
        },
        /* PORT(VGA1): CLOSE*/
        {0},
        /* PORT(VGA2): CLOSE*/
        {0},
        /* PORT(VGA3): CLOSE*/
        {0},

        /* PORT1(HDMI0): encoder-4 connector-4*/
        {
            LJM_PORT_INFO(1, LJM_PORT_HDMI, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(4, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(4, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },
        /* PORT_2(HDMI1): encoder-5 connector-5*/
        {
            LJM_PORT_INFO(2, LJM_PORT_HDMI, 1, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(5, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(5, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },
        /* PORT_3(HDMI2): CLOSE*/
        {0},
        /* PORT_4(HDMI3): CLOSE*/
        {0},

        /* PORT_5(DVO0): encoder-8 connector-8*/
        {
            LJM_PORT_INFO(3, LJM_PORT_DVO, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(8, LJM_ENCODER_DVO, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(8, DRM_MODE_CONNECTOR_DVID, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DWI2C, LJM_HPD_TYPE_CRTC))
        },
        /* PORT6(DVO1): encoder-9 connector-9*/
        {
            LJM_PORT_INFO(4, LJM_PORT_DVO, 1, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(9, LJM_ENCODER_DVO, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(9, DRM_MODE_CONNECTOR_DVID, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DWI2C, LJM_HPD_TYPE_CRTC))
        },
        /* PORT(DVO2): CLOSE*/
        {0},
        /* PORT(DVO3): CLOSE*/
        {0},

        /* PORT7(DP0): encoder-12 connector-12*/
        {
            LJM_PORT_INFO(5, LJM_PORT_DP, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(12, LJM_ENCODER_DP, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(12, DRM_MODE_CONNECTOR_DisplayPort, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DP, LJM_HPD_TYPE_DP))
        },
        /* PORT8(DP1): CLOSE*/
        {0},
        /* PORT9(DP2): CLOSE*/
        {0},
        /* PORT10(edp->DP3): CLOSE*/
        {0}
    }
};

struct ljm_drm_info ljm_fpga_info = {
    .gpu_family = CHIP_GP200,
    .ports_num = 4,

    .ports = {
        /* PORT(VGA0): CLOSE*/
        {0},
        /* PORT(VGA1): CLOSE*/
        {0},
        /* PORT(VGA2): CLOSE*/
        {0},
        /* PORT(VGA3): CLOSE*/
        {0},
        /* PORT(HDMI0): CLOSE*/
        {
            LJM_PORT_INFO(1, LJM_PORT_HDMI, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(4, LJM_ENCODER_HDMIV2, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(4, DRM_MODE_CONNECTOR_HDMIA, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_HDMIV2, LJM_HPD_TYPE_HDMIV2))
        },
        /* PORT(HDMI1): CLOSE*/
        {0},
        /* PORT(HDMI2): CLOSE*/
        {0},
        /* PORT(HDMI3): CLOSE*/
        {0},
        /* PORT_5(DVO0): encoder-8 connector-8*/
        {
            LJM_PORT_INFO(5, LJM_PORT_DVO, 0, LJM_PORT_HPD_DISABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(8, LJM_ENCODER_DVO, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(8, DRM_MODE_CONNECTOR_DVID, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DWI2C, LJM_HPD_TYPE_CRTC))
        },
        /* PORT6(DVO1): encoder-9 connector-9*/
        {
            LJM_PORT_INFO(6, LJM_PORT_DVO, 1, LJM_PORT_HPD_DISABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(9, LJM_ENCODER_DVO, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(9, DRM_MODE_CONNECTOR_DVID, DRM_CONNECTOR_POLL_HPD,
                    LJM_DDC_TYPE_DWI2C, LJM_HPD_TYPE_CRTC))
        },
        /* PORT(DVO2): encoder-9 connector-9*/
        {0},
        /* PORT_8(DP0): CLOSE*/
        {
            LJM_PORT_INFO(8, LJM_PORT_DP, 0, LJM_PORT_HPD_ENABLE, LJM_PORT_NOT_DVI_I, LJM_PORT_NOT_EDP, 0,
                LJM_ENCODER_INFO(12, LJM_ENCODER_DP, 0xf, 0xffff),
                LJM_CONNECTOR_INFO(12, DRM_MODE_CONNECTOR_DisplayPort, DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT,
                    LJM_DDC_TYPE_DP, LJM_HPD_TYPE_DP))
        },
        /* PORT(DP1): CLOSE*/
        {0},
        /* PORT(DP2): CLOSE*/
        {0},
        /* PORT(DP3): CLOSE*/
        {0},
    }
};

/******************************************************************************
 *                                    函数声明                                    *
 ******************************************************************************/
extern ljm_board_info g_board_info[LJM_MAX_CARDS_NUM];

/******************************************************************************
 *                                    函数实现                                    *
 ******************************************************************************/


/*****************************************************************************
 *  函 数 名：ljm_board_info_port_enable
 *  功能描述：根据board_info开关视频接口
 *  输入参数：ljm_board_info *board_info
 *            struct ljm_drm_info *info
 *  输出参数：无
 *  返 回 值：static int
 *
 *  修改历史：
 *         1. 日    期：2023年1月9日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
static int ljm_board_info_port_enable(ljm_board_info *board_info, struct ljm_drm_info *info)
{
#if 0
    int i;
    u16 port_en, hdmi_dvii, dvo_dvii, dp_edp;
#endif

    if(board_info->vendor_id != LJM_VENDOR_ID)
    {
        WARN_PRINT("[%s]Invalid board_info. Defaults will be chosen.\n", __FUNCTION__);
        return 0;
    }

#if 0
    /* cannot take address of bit-field
    port_en = *((u16*)(&board_info->vga0_en));          // ((u16)board_info->dvo_dp_en << 8) | board_info->vga_hdmi_en;
    hdmi_dvii = *((u16*)(&board_info->hdmi0_vga_en));   // ((u16)board_info->dvi_i_cfg1 << 8) | board_info->dvi_i_cfg0;
    dvo_dvii = *((u16*)(&board_info->dvo0_vga_en));     // ((u16)board_info->dvi_i_cfg3 << 8) | board_info->dvi_i_cfg2;
    dp_edp = *((u16*)(&board_info->dp0_edp_en));        // ((u16)board_info->dp_cfg1 << 8) | board_info->dp_cfg0;
    */

    for(i=0; i<LJM_MAX_PORTS; i++)
    {
        if(!(port_en & BIT(i)))
        {
            if(info->ports[i].type > 0)
            {
                info->ports_num--;
                info->ports[i].type = LJM_PORT_CLOSE;
            }
            continue;
        }

        if(info->ports[i].type <= 0)
        {
            WARN_PRINT("[%s]Board_info try to enable a nonexistent port[%d]!\n", __FUNCTION__, i);
            continue;
        }

        switch(info->ports[i].type)
        {
            case LJM_PORT_HDMI:
                if(hdmi_dvii & BIT((i % 4) * 4))
                {
                    info->ports[i].is_dvii = true;
                    info->ports[i].extra_pipe_index = (hdmi_dvii & (0x3 << (i % 4 * 4 + 2))) >> (i % 4 * 4 + 2);
                }
                break;
            case LJM_PORT_DVO:
                if(dvo_dvii & BIT((i % 4) * 4))
                {
                    info->ports[i].is_dvii = true;
                    info->ports[i].extra_pipe_index = (dvo_dvii & (0x3 << (i % 4 * 4 + 2))) >> (i % 4 * 4 + 2);
                }
                break;
            case LJM_PORT_DP:
                if(dp_edp & BIT((i % 4) * 4))
                {
                    info->ports[i].is_edp = true;
                    info->ports[i].extra_pipe_index = (dp_edp & (0x3 << (i % 4 * 4 + 2))) >> (i % 4 * 4 + 2);
                }
                if(dp_edp & BIT((i % 4) * 4 + 1))
                {
                    info->ports[i].disable_aux = true;
                }
                else
                {
                    info->ports[i].disable_aux = false;
                }
                break;
            default:
                break;
        }
    }
#else
    /* VGA connector */
    if (!board_info->vga0_en && info->ports[0].type > 0) {
        info->ports_num--;
        info->ports[0].type = LJM_PORT_CLOSE;
    }

    if (!board_info->vga1_en && info->ports[1].type > 0) {
        info->ports_num--;
        info->ports[1].type = LJM_PORT_CLOSE;
    }

    if (!board_info->vga2_en && info->ports[2].type > 0) {
        info->ports_num--;
        info->ports[2].type = LJM_PORT_CLOSE;
    }

    if (!board_info->vga3_en && info->ports[3].type > 0) {
        info->ports_num--;
        info->ports[3].type = LJM_PORT_CLOSE;
    }

    /* HDMI connector */
    if (!board_info->hdmi0_en && info->ports[4].type > 0) {
        info->ports_num--;
        info->ports[4].type = LJM_PORT_CLOSE;
    } else if (board_info->hdmi0_vga_en) {
        info->ports[4].is_dvii = true;
        info->ports[4].extra_pipe_index = board_info->hdmi0_vga_sel;
    }

    if (!board_info->hdmi1_en && info->ports[5].type > 0) {
        info->ports_num--;
        info->ports[5].type = LJM_PORT_CLOSE;
    } else if (board_info->hdmi1_vga_en) {
        info->ports[5].is_dvii = true;
        info->ports[5].extra_pipe_index = board_info->hdmi1_vga_sel;
    }

    if (!board_info->hdmi2_en && info->ports[6].type > 0) {
        info->ports_num--;
        info->ports[6].type = LJM_PORT_CLOSE;
    } else if (board_info->hdmi2_vga_en) {
        info->ports[6].is_dvii = true;
        info->ports[6].extra_pipe_index = board_info->hdmi2_vga_sel;
    }

    if (!board_info->hdmi3_en && info->ports[7].type > 0) {
        info->ports_num--;
        info->ports[7].type = LJM_PORT_CLOSE;
    } else if (board_info->hdmi3_vga_en) {
        info->ports[7].is_dvii = true;
        info->ports[7].extra_pipe_index = board_info->hdmi3_vga_sel;
    }

    /* DVO connector */
    if (!board_info->dvo0_en && info->ports[8].type > 0) {
        info->ports_num--;
        info->ports[8].type = LJM_PORT_CLOSE;
    } else if (board_info->dvo0_vga_en) {
        info->ports[8].is_dvii = true;
        info->ports[8].extra_pipe_index = board_info->dvo0_vga_sel;
    }

    if (!board_info->dvo1_en && info->ports[9].type > 0) {
        info->ports_num--;
        info->ports[9].type = LJM_PORT_CLOSE;
    } else if (board_info->dvo1_vga_en) {
        info->ports[9].is_dvii = true;
        info->ports[9].extra_pipe_index = board_info->dvo1_vga_sel;
    }

    if (!board_info->dvo2_en && info->ports[10].type > 0) {
        info->ports_num--;
        info->ports[10].type = LJM_PORT_CLOSE;
    } else if (board_info->dvo2_vga_en) {
        info->ports[10].is_dvii = true;
        info->ports[10].extra_pipe_index = board_info->dvo2_vga_sel;
    }

    if (!board_info->dvo3_en && info->ports[11].type > 0) {
        info->ports_num--;
        info->ports[11].type = LJM_PORT_CLOSE;
    } else if (board_info->dvo3_vga_en) {
        info->ports[11].is_dvii = true;
        info->ports[11].extra_pipe_index = board_info->dvo3_vga_sel;
    }

    /* DP/EDP connector */
    if (!board_info->dp0_en && info->ports[12].type > 0) {
        info->ports_num--;
        info->ports[12].type = LJM_PORT_CLOSE;
    } else {
        info->ports[12].extra_pipe_index =  board_info->dp0_aux_gpio_en ? board_info->dp0_aux_gpio_sel: NO_EXTRA_PIPE_INDEX;
        if (board_info->dp0_edp_en) {
            info->ports[12].is_edp = true;
            info->ports[12].connector_info.connector_type = DRM_MODE_CONNECTOR_eDP;
            info->ports[12].bl_pwr_sel = board_info->dp0_pwr_gpio_en ? board_info->dp0_pwr_gpio_sel : NO_BL_PWM_SEL;
            info->ports[12].pwm_sel = board_info->dp0_pwm_en ? board_info->dp0_pwm_sel : NO_BL_PWM_SEL;
        }
        if (!board_info->dp0_aux_gpio_en && board_info->dp0_aux_gpio_sel == 1) {
            info->ports[12].disable_aux = true;
        }
    }

    if (!board_info->dp1_en && info->ports[13].type > 0) {
        info->ports_num--;
        info->ports[13].type = LJM_PORT_CLOSE;
    } else {
        info->ports[13].extra_pipe_index = board_info->dp1_aux_gpio_en ? board_info->dp1_aux_gpio_sel: NO_EXTRA_PIPE_INDEX;
        if (board_info->dp1_edp_en) {
            info->ports[13].is_edp = true;
            info->ports[13].connector_info.connector_type = DRM_MODE_CONNECTOR_eDP;
            info->ports[13].bl_pwr_sel = board_info->dp1_pwr_gpio_en ? board_info->dp1_pwr_gpio_sel : NO_BL_PWM_SEL;
            info->ports[13].pwm_sel = board_info->dp1_pwm_en ? board_info->dp1_pwm_sel : NO_BL_PWM_SEL;
        }
        if (!board_info->dp1_aux_gpio_en && board_info->dp1_aux_gpio_sel == 1) {
            info->ports[13].disable_aux = true;
        }
    }

    if (!board_info->dp2_en && info->ports[14].type > 0) {
        info->ports_num--;
        info->ports[14].type = LJM_PORT_CLOSE;
    } else {
        info->ports[14].extra_pipe_index = board_info->dp2_aux_gpio_en ? board_info->dp2_aux_gpio_sel: NO_EXTRA_PIPE_INDEX;
        if (board_info->dp2_edp_en) {
            info->ports[14].is_edp = true;
            info->ports[14].connector_info.connector_type = DRM_MODE_CONNECTOR_eDP;
            info->ports[14].bl_pwr_sel = board_info->dp2_pwr_gpio_en ? board_info->dp2_pwr_gpio_sel : NO_BL_PWM_SEL;
            info->ports[14].pwm_sel = board_info->dp2_pwm_en ? board_info->dp2_pwm_sel : NO_BL_PWM_SEL;
        }
        if (!board_info->dp2_aux_gpio_en && board_info->dp2_aux_gpio_sel == 1) {
            info->ports[14].disable_aux = true;
        }
    }

    if (!board_info->dp3_en && info->ports[15].type > 0) {
        info->ports_num--;
        info->ports[15].type = LJM_PORT_CLOSE;
    } else {
        info->ports[15].extra_pipe_index = board_info->dp3_aux_gpio_en ? board_info->dp3_aux_gpio_sel: NO_EXTRA_PIPE_INDEX;
        if (board_info->dp3_edp_en) {
            info->ports[15].is_edp = true;
            info->ports[15].connector_info.connector_type = DRM_MODE_CONNECTOR_eDP;
            info->ports[15].bl_pwr_sel = board_info->dp3_pwr_gpio_en ? board_info->dp3_pwr_gpio_sel : NO_BL_PWM_SEL;
            info->ports[15].pwm_sel = board_info->dp3_pwm_en ? board_info->dp3_pwm_sel : NO_BL_PWM_SEL;
        }
        if (!board_info->dp3_aux_gpio_en && board_info->dp3_aux_gpio_sel == 1) {
            info->ports[15].disable_aux = true;
        }
    }
#endif

    return 0;
}

/*****************************************************************************
 *  函 数 名：drm_info_dump
 *  功能描述：打印drm_info的信息
 *  输入参数：struct ljm_drm_info *info
 *  输出参数：无
 *  返 回 值：void
 *
 *  修改历史：
 *         1. 日    期：2023年1月9日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
void ljm_drm_info_dump(struct ljm_drm_info *info)
{
    int i;

    if(!info)
        return;

    DEBUG_PRINT("ljm_drm_info_dump:\n");
    DEBUG_PRINT("gpu family: \t%d\n", info->gpu_family);
    DEBUG_PRINT("ports_num: \t%d\n", info->ports_num);

    for(i=0; i<LJM_MAX_PORTS; i++)
    {
        DEBUG_PRINT("-----------------------\n");
        if(info->ports[i].type == 0)
        {
            DEBUG_PRINT("port[%d]: \tclose\n", i);
            continue;
        }

        DEBUG_PRINT("port[%d]: \ttype-%d sub_id[%d]\n", i, info->ports[i].type, info->ports[i].sub_id);
        DEBUG_PRINT("hpd: %d \n", info->ports[i].hpd);
        DEBUG_PRINT("is_dvii: %d is_edp: %d extra_pipe_index:%d disable_aux:%d\n",
                    info->ports[i].is_dvii, info->ports[i].is_edp, info->ports[i].extra_pipe_index, info->ports[i].disable_aux);
        DEBUG_PRINT("encoder[%d]: type-%d\n",
                    info->ports[i].encoder_info.index, info->ports[i].encoder_info.hard_type);
        DEBUG_PRINT("possible_crtcs: 0x%X possible_clones: 0x%X\n",
                    info->ports[i].encoder_info.possible_crtcs, info->ports[i].encoder_info.possible_clones);
        DEBUG_PRINT("connector[%d]: type-%d\n",
                    info->ports[i].connector_info.index, info->ports[i].connector_info.connector_type);
        DEBUG_PRINT("polled: %d ddc_type: %d hpd_type: %d\n",
                    info->ports[i].connector_info.polled, info->ports[i].connector_info.ddc_type, info->ports[i].connector_info.hpd_type);
    }

    return;
}

/*****************************************************************************
 *  函 数 名：ljm_drm_info_duplicate
 *  功能描述：复制drm_info结构体。对*src指向的ljm_drm_info结构体做浅复制
 *  输入参数：const struct ljm_drm_info *src
 *  输出参数：无
 *  返 回 值：struct ljm_drm_info * 复制的新结构体的指针
 *
 *  修改历史：
 *         1. 日    期：2023年1月9日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
struct ljm_drm_info *ljm_drm_info_duplicate(const struct ljm_drm_info *src)
{
    struct ljm_drm_info *des;

    des = kmalloc(sizeof(struct ljm_drm_info) , GFP_KERNEL); //free at ljm_device_fini()
    if(!des)
        return NULL;

    *des = *src;    //copy struct

    return des;
}

/*****************************************************************************
 *  函 数 名：ljm_dev_info_init
 *  功能描述：ljm_dev->drm_info初始化, 若过程出错，则goto至force流程
 *  输入参数：struct ljm_device *ljm_dev
 *  输出参数：无
 *  返 回 值：int
 *
 *  修改历史：
 *         1. 日    期：2023年1月9日
 *            作    者：zhaiyf
 *            修改内容：新生成函数
 *
 *****************************************************************************/
int ljm_dev_info_init(struct ljm_device *ljm_dev)
{
    struct pci_dev *pdev;
    struct ljm_port_info port_tmp[2];
    char board_name_str[14];

    /* 如果是FPGA则直接使用 ljm_fpga_info*/
    if (ljm_dev->flags & LJMICRO_IS_FPGA)
    {
        ljm_dev->drm_info = ljm_drm_info_duplicate(&ljm_fpga_info);
        WARN_PRINT("[%s] Use fpga info.\n", __FUNCTION__);
        return 0;
    }

    if(ljm_dev->pdev)
        pdev = ljm_dev->pdev;
    else
        goto force;

    switch(pdev->device){
        case 0x001:
            ljm_dev->drm_info = ljm_drm_info_duplicate(&ljm_gp101_info);
            if(!ljm_dev->drm_info)
                goto bad;
            ljm_dev->drm_info->gpu_family = CHIP_GP100;
            break;
        case 0x201:
            /* 取board_name的前13个字节，便于后续异型卡的区分 */
            strncpy(board_name_str, g_board_info[ljm_dev->cur_card].board_name, 13);
            board_name_str[13] = '\0';
            if (strcmp(board_name_str, "GP201Y-MXM-11") == 0 || strcmp(board_name_str, "GP201A-MXM-40") == 0)
            {
                /* 调换 HDMI 接口向内核的注册顺序 */
                /**
                 * GP201Y-MXM-11与GP201A-MXM-40型号GPU的4路hdmi硬件接口顺序特殊。调整sub_id和encoder id，
                 * 使驱动依旧对DRM保持HDMI接口自然顺序。
                */
                port_tmp[0] = ljm_gp201_info.ports[4];
                port_tmp[1] = ljm_gp201_info.ports[5];
                ljm_gp201_info.ports[4] = ljm_gp201_info.ports[6];
                ljm_gp201_info.ports[5] = ljm_gp201_info.ports[7];
                ljm_gp201_info.ports[6] =  port_tmp[0];
                ljm_gp201_info.ports[7] =  port_tmp[1];

                ljm_dev->drm_info = ljm_drm_info_duplicate(&ljm_gp201_info);
            }
            else
                ljm_dev->drm_info = ljm_drm_info_duplicate(&ljm_gp201_info);

            if(!ljm_dev->drm_info)
                goto bad;
            ljm_dev->drm_info->gpu_family = CHIP_GP200;
            ljm_dev->once_edp = true;
            break;
        case 0x202:
            ljm_dev->drm_info = ljm_drm_info_duplicate(&ljm_gp202_info);
            if(!ljm_dev->drm_info)
                goto bad;
            ljm_dev->drm_info->gpu_family = CHIP_GP200;
            break;
        default:
            goto force;
            break;
    }

    ljm_board_info_port_enable(&g_board_info[ljm_dev->cur_card], ljm_dev->drm_info);
    ljm_drm_info_dump(ljm_dev->drm_info);
    return 0;

force:
    ljm_dev->drm_info = ljm_drm_info_duplicate(&ljm_gp201_info);
    ljm_dev->once_edp = true;
    WARN_PRINT("[%s]Unknown device_id. Use gp201 info.\n", __FUNCTION__);

    if(!ljm_dev->drm_info)
        goto bad;
    return 0;

bad:
    ERR_PRINT("[%s]drm_info init failed\n", __FUNCTION__);
    return -1;
}

/******************************************************************************
 *                          虚拟edid和虚拟dpcd                                 *
 ******************************************************************************/
int ljmSpiReadBuffer(void *mmio_base, char *buf, int addr, int size);

static uint16_t crc16_for_board_info(uint16_t *pv16, int count)
{
    int i;
    uint16_t sum = 0;

    for (i=0; i<count; i++)
        sum += *pv16++;

    return (0 - sum);
}

static bool ljm_board_info_image_header_valid(struct board_info_image_header *header)
{
    return header->vendor_id == IMAGE_HEAD_IDENTIFICATION;
}

static bool ljm_board_info_image_valid(struct board_info_image_header *header, void *data)
{
    if(!ljm_board_info_image_header_valid(header))
        return false;

    if(header->crc != crc16_for_board_info((uint16_t *)data, (header->data_size+1)/2))
        return false;

    return true;
}

static struct board_info_image_header *board_info_image_next(void *mmio_base, struct board_info_image_header *header)
{
    struct board_info_image_header *next = NULL;
    uint8_t *data = NULL;
    int offset;

    if(header)
    {
        if(header->image_offest == 0)
            return NULL;
        offset = BOARDINFO_IMAGE_ADDR + header->image_offest;
    }
    else
        offset = BOARDINFO_IMAGE_ADDR;

    next = kmalloc(sizeof(struct board_info_image_header), GFP_KERNEL);
    if(!next)
    {
        ERR_PRINT("Failed to malloc header.\n");
        return NULL;
    }
    if(ljmSpiReadBuffer(mmio_base, (char *)next, offset, (sizeof(struct board_info_image_header)-sizeof(void *))) < 0)
        goto header_null;

    if(!ljm_board_info_image_header_valid(next))
        goto header_null;

    data = kmalloc(next->data_size, GFP_KERNEL);
    if(!data)
    {
        ERR_PRINT("Failed to malloc data.\n");
        goto header_null;
    }
    if(ljmSpiReadBuffer(mmio_base, data, (offset + next->data_offest), next->data_size) < 0)
        goto data_null;

    if(!ljm_board_info_image_valid(next, data))
        goto data_null;

    next->data = data;
    goto success;

data_null:
    kfree(data);
header_null:
    kfree(next);
    next = NULL;
success:
    return next;
}

static void ljm_port_info_image_add(struct ljm_port_info *port_info, struct board_info_image_header *header)
{
    int i;

    for(i=0; i<PORT_INFO_IMAGE_MAX; i++)
    {
        if(!port_info->image[i])
        {
            port_info->image[i] = header;
            break;
        }
    }

    return;
}

int board_info_image_handle(struct ljm_device *ljm_dev)
{
    struct board_info_image_header *h;
    struct ljm_drm_info *drm_info = ljm_dev->drm_info;
    struct ljm_port_info *port_info;
    int i;

    for(h = board_info_image_next(ljm_dev->mmio_base, NULL); h != NULL;
        h = board_info_image_next(ljm_dev->mmio_base, h))
    {
        if(h->image_type == IMAGE_TYPE_DP)
        {
            for(i=0, port_info=&drm_info->ports[0]; i<LJM_MAX_PORTS; i++, port_info=&drm_info->ports[i])
            {
                if(h->image_num == port_info->sub_id && port_info->type == LJM_PORT_DP)
                    ljm_port_info_image_add(port_info, h);
            }
        }
    }

    return 0;
}