/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_INFO_H__
#define __LJMICRO_INFO_H__

#include "ljmicro_base.h"
#include "ljmicro_device.h"

/******************************************************************************
 *                          虚拟edid和虚拟dpcd                                 *
 ******************************************************************************/
#define BOARDINFO_IMAGE_ADDR       0x70200

#define IMAGE_HEAD_SIZE     (sizeof(struct board_info_image_header))

#define IMAGE_TYPE_DP   0x5044

#define IMAGE_DATA_TYPE_EDID 0
#define IMAGE_DATA_TYPE_DPCD 1

struct board_info_image_header
{
    uint16_t vendor_id;    /* 0x0: 0x0709 */
    uint16_t image_type;   /* 0x2: dp_image=0x5044 */
    uint16_t image_num;    /* 0x4: dp_inage_num=0,1,2... */
    uint16_t data_type;    /* 0x6: edid_data=0,dpcd_data=1 */
    uint16_t data_offest;  /* 0x8: offest of data from the image start address */
    uint16_t data_size;    /* 0xa: size of image data,exclude image header */
    uint16_t crc;          /* 0xc: 16-bit checksum for image data */
    uint16_t image_offest; /* 0xe: offest address of the next image,0:no next image */
    uint8_t reserved[48];     /* 0x10~0x3f: reserved=oxff */

    uint8_t *data; /* data pack pointer */
}__attribute__((packed));

int board_info_image_handle(struct ljm_device *ljm_dev);
/*************************************************************************/
/*
 * 用于保存encoder基本配置信息，在ljm_encoder_init()时将
 * 这些信息传递给ljm_encoder和drm_encoder.
 */
struct ljm_encoder_info
{
    int     index;

    enum ljm_encoder_hard_type hard_type;

    uint32_t    possible_crtcs;
    uint32_t    possible_clones;
};

struct ljm_connector_info
{
    int     index;

    int     connector_type; /* drm_connector type*/
    uint8_t polled;     /*Connector polling mode*/
    enum ljm_ddc_type ddc_type;     /* ddc类型，硬件初始化时设置 */
    enum ljm_hpd_type hpd_type;     /* 热插拔类型，硬件初始化时设置 */

};

struct ljm_port_info
{
    int index; /* PORT ID */
    enum ljm_port_type type;
    int     sub_id; /* Video device ID */

#define LJM_PORT_HPD_DISABLE 0
#define LJM_PORT_HPD_ENABLE 1
    bool hpd; /*If true, hotplug interrupt is supported */

#define LJM_PORT_NOT_DVI_I  0
#define LJM_PORT_IS_DVI_I   1
    bool is_dvii;

#define LJM_PORT_NOT_EDP 0
#define LJM_PORT_IS_EDP 1
    bool is_edp;
    /**
     * @bl_pwr_sel: edp背光电源控制所使用的GPIO组。
     * @pwm_sel: pwm亮度控制所使用的pwm通道。
     * 仅在@is_edp变量为真时，这两个变量才有效。
     * bl_pwr_sel或pwm_sel值为“-1”时，表示不支持使用gpio控制背光电源或调节亮度。
    */
#define NO_BL_PWM_SEL     (-1)
    int bl_pwr_sel;
    int pwm_sel;

#define LJM_PORT_NOT_DIS_AUX 0
#define LJM_PORT_IS_DIS_AUX 1
    bool disable_aux;

    /**
     * @extra_pipe_index: 额外通道编号
     * HDMI/VGA为DVI0I时，对应的VGA通道; DP为eDP时，对应GPIO通道。
     * 因为该变量0值有意义，因此使用-1表示无额外通道。
    */
#define NO_EXTRA_PIPE_INDEX     (-1)
    int extra_pipe_index;

    struct ljm_encoder_info encoder_info;
    struct ljm_connector_info connector_info;

#define PORT_INFO_IMAGE_MAX 4
    struct board_info_image_header *image[PORT_INFO_IMAGE_MAX];
};

struct ljm_drm_info
{
    enum ljm_family gpu_family;

    int    ports_num;
    struct ljm_port_info ports[LJM_MAX_PORTS];

};

#define LJM_ENCODER_INFO(id, t, pcrt, pclo) \
    .index = (id), .hard_type = (t), .possible_crtcs = (pcrt), \
    .possible_clones = (pclo)

#define LJM_CONNECTOR_INFO(id, t, po, ddct, hpdt) \
    .index = (id), .connector_type = (t), .polled = (po), \
    .ddc_type = (ddct), .hpd_type = (hpdt)

#define LJM_PORT_INFO(id, t, sub, h, dvi, edp, ex, enc, conn) \
    .index = (id), .type = (t), .sub_id = (sub), .hpd = (h), \
    .is_dvii = (dvi), .is_edp = (edp), .extra_pipe_index = (ex), \
    .encoder_info = {enc}, \
    .connector_info = {conn}, .image = {0}

#endif /*__LJMICRO_INFO_H__*/