/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，武汉凌久微电子有限公司
 *
 ******************************************************************************
 *
 *  文件名称：ljmicro_util_ioctl.h
 *  版 本 号：V1.0
 *  作    者：zhaiyf
 *  生成日期：2022年5月11日
 *
 *  功能描述：util通过ioctl控制驱动的头文件，主要包含io-
                   ctl相关数据结构和宏定义
 *
 *  修改历史：
 *         1. 日    期：2022年5月11日
 *            作    者：zhaiyf
 *            修改内容：创建文件
 *
 *****************************************************************************/
#ifndef __LJMICRO_UTIL_IOCTL_H__
#define __LJMICRO_UTIL_IOCTL_H__
/******************************************************************************
 *                                   标准库头文件                                   *
 ******************************************************************************/
#include <linux/ioctl.h>
/******************************************************************************
 *                                     宏定义                                     *
 ******************************************************************************/
#ifdef __cplusplus
extern "C"
{
#endif

/******************************************************************************
 *                                   数据类型定义                                   *
 ******************************************************************************/
#define LJM_IOCTL_MMIO_BAR_VRAM 0
#define LJM_IOCTL_MMIO_BAR_REG  2

#define LJM_IOCTL_MMIO_OP_READ  0
#define LJM_IOCTL_MMIO_OP_WRITE 1
#define LJM_IOCTL_MMIO_OP_COUNT 2

#define LJM_IOCTL_MMIO_WIDTH_8  1
#define LJM_IOCTL_MMIO_WIDTH_16 2
#define LJM_IOCTL_MMIO_WIDTH_32 4
#define LJM_IOCTL_MMIO_WIDTH_64 8

    struct ljm_ioctl_mmio
    {
        uint32_t card;   /* card index from 0 */
        uint32_t bar;    /* mmio bar number */
        uint32_t op;     /* operation type: 0: read, 1: write */
        uint32_t width;  /* word width. 0: 8-bit, 1: 16-bit, 2: 32-bit, 3: 64-bit */
        uint32_t offset; /* mmio offset */
        /* input or output */
        uint64_t value; /* value */
    };

#define LJM_IOCTL_DDC_CH_COUNT 16

#define LJM_IOCTL_DDC_OP_READ  0 /* DDC read */
#define LJM_IOCTL_DDC_OP_COUNT 1

#ifndef MAX_EDID
#define MAX_EDID 512
#endif

struct ljm_ioctl_ddc
{
    uint32_t card; /* card index from 0 */
    uint32_t ch;   /* ddc channel */
    uint32_t op;   /* operation type: 0: ddc read */
    /* input or output */
    uint32_t bytes;          /* edid size */
    uint8_t  edid[MAX_EDID]; /* edid */
    uint32_t valid;          /* valid flag */
};

struct ljm_ioctl_ver
{
    /* input or output */
    uint32_t major;
    uint32_t minor;
    uint32_t patch;
    uint32_t build;
};

#define LJM_IOCTL_PCI_OP_READ  0
#define LJM_IOCTL_PCI_OP_WRITE 1
#define LJM_IOCTL_PCI_OP_COUNT 2

struct ljm_ioctl_pci
{
    uint32_t card;   /* card index from 0 */
    uint32_t op;     /* operation type: 0: read, 1: write */
    uint32_t offset; /* pci offset */
    /* input or output */
    uint32_t value; /* value */
};

#define LJM_IOCTL_KMEM_OP_READ  0
#define LJM_IOCTL_KMEM_OP_WRITE 1
#define LJM_IOCTL_KMEM_OP_COUNT 2

#define LJM_IOCTL_KMEM_WIDTH_8  1
#define LJM_IOCTL_KMEM_WIDTH_16 2
#define LJM_IOCTL_KMEM_WIDTH_32 4
#define LJM_IOCTL_KMEM_WIDTH_64 8

struct ljm_ioctl_kmem
{
    uint64_t kvaddr; /* kernel virtual address */
    uint32_t op;     /* operation type: 0: read, 1: write */
    uint32_t width;  /* word width. 0: 8-bit, 1: 16-bit, 2: 32-bit, 3: 64-bit */
    uint64_t value;  /* value */
};

struct ljm_ioctl_info
{
    uint32_t card;
    uint32_t vram_used;
    uint32_t vram_free;
    uint32_t vram_total;
};

// HDMI ioctl cmd data
struct ljm_hdmi_cmd_phy_data
{
    uint8_t  addr;
    uint16_t data;
};
struct ljm_hdmi_cmd_vsi_data
{
    uint8_t vic_3d;
    uint8_t format;
};

struct ljm_hdmi_cmd_hdr_data
{
    int eotf;

    uint32_t prim_x0;
    uint32_t prim_y0;
    uint32_t prim_x1;
    uint32_t prim_y1;
    uint32_t prim_x2;
    uint32_t prim_y2;
    uint32_t white_px;
    uint32_t white_py;
    uint32_t max_dml;
    uint32_t min_dml;
    uint32_t max_cll;  /*max content light level*/
    uint32_t max_fall; /*max frame-average light level*/
};

struct ljm_hdmi_cmd_spd_data
{
    char    vendor_name[20];
    char    product_name[20];
    uint8_t deviceinfo;
};

struct ljm_hdmi_cmd_videomode
{
    uint32_t refresh; /* optional */
    uint32_t xres;
    uint32_t yres;
    uint32_t pixclock;
    uint32_t left_margin;
    uint32_t right_margin;
    uint32_t upper_margin;
    uint32_t lower_margin;
    uint32_t hsync_len;
    uint32_t vsync_len;
    uint32_t sync;
    uint32_t vmode;
    uint32_t flag;
};

struct ljm_hdmi_cmd_vpara_data
{
    unsigned int                  vic; /* Video information code*/
    struct ljm_hdmi_cmd_videomode mode;

    unsigned int color_input;        /* Input video color mode*/
    unsigned int color_output;       /* Output video color mode*/
    unsigned int color_output_depth; /* Output video Color Depth*/
    unsigned int colorimetry;        /* Output Colorimetry */
    unsigned int sink_hdmi;          /* Output signal is DVI or HDMI*/
    unsigned int format_3d;          /* Output 3D mode*/
    unsigned int eotf;               /* EOTF */
};

/*hdmi command type*/
#define LJM_HDMI_CMD_PHY_R       0 /* struct ljm_hdmi_cmd_phy_data */
#define LJM_HDMI_CMD_PHY_W       1
#define LJM_HDMI_CMD_VSI_R       2 /* struct ljm_hdmi_cmd_vsi_data */
#define LJM_HDMI_CMD_VSI_w       3
#define LJM_HDMI_CMD_HDR_R       4 /* struct ljm_hdmi_cmd_hdr_data */
#define LJM_HDMI_CMD_HDR_W       5
#define LJM_HDMI_CMD_SPD_R       6 /* struct ljm_hdmi_cmd_spd_data */
#define LJM_HDMI_CMD_SPD_W       7
#define LJM_HDMI_CMD_MODE_R      8 /* ljm_hdmi_cmd_vpara_data */
#define LJM_HDMI_CMD_MODE_W      9
#define LJM_HDMI_CMD_DPMS_R      10 /* data: (int) */
#define LJM_HDMI_CMD_DPMS_W      11
#define LJM_HDMI_CMD_HPDSENSE_R  12 /* data: (bool) */
#define LJM_HDMI_CMD_HPDCLEAR_W  13 /* data: (NULL) */
#define LJM_HDMI_CMD_HPDENABLE_R 14 /* data: (int) */
#define LJM_HDMI_CMD_HPDENABLE_W 15
#define LJM_HDMI_CMD_SCDC_R      16 /* data: (unsigned short) */

/* command type */
struct ljm_ioctl_3d_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t type;
    uint32_t size;
    uint64_t pdata;
};

struct ljm_ioctl_2d_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t type;
    uint32_t size;
    uint64_t pdata;
};
struct ljm_ioctl_dac_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t type;
    uint32_t size;
    uint64_t pdata;
};
struct ljm_ioctl_hdmi_common
{
    uint32_t card;  /* card index */
    uint32_t index; /* HDMI index */
    uint32_t cmd;   /* HDMI command */

    uint32_t size;  /* data size (byte) */
    uint64_t pdata; /* data pointer */
};

struct ljm_ioctl_dvo_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t type;
    uint32_t size;
    uint64_t pdata;
};

#define LJM_DP_CMD_DPCD_WRITE   0   /* read DPCD */
#define LJM_DP_CMD_DPCD_READ    1   /* write DPCD */
#define LJM_DP_CMD_EDID_GET     2   /* get EDID */

struct ljm_ioctl_dp_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t cmd;
    uint32_t offset;
    uint32_t size;
    uint8_t pdata[MAX_EDID];
};

struct ljm_ioctl_ddr_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t type;
    uint32_t size;
    uint64_t pdata;
};

struct ljm_ioctl_spi_command
{
    /* input or output */
    uint32_t card, index;
    uint32_t type;
    uint32_t size;
    uint64_t pdata;
};


#define LJM_IOCTL_I2C_OP_READ  0
#define LJM_IOCTL_I2C_OP_WRITE 1

struct ljm_ioctl_i2c_command
{
    /* input or output */
    uint32_t card;
    uint32_t ch;
    uint32_t target;
    uint32_t addr;
    uint32_t op;
    uint32_t size;
    uint8_t pdata[MAX_EDID];
    int flag;
};

typedef enum _QUERY_IOCTL_LIST{
    QUERY_DEVICE_ID,
    QUERY_FREQ_AXI,
    QUERY_FREQ_VPU,
    QUERY_FREQ_3D,
    QUERY_FREQ_2D,
    QUERY_FREQ_DDR,
    QUERY_NUM_3DCORE,
    QUERY_NUM_2DCORE,
    QUERY_NUM_CARD,
    QUERY_INFO_GOP,
    QUERY_CHANNEL_MEM,
    QUERY_LOAD_RATE,
    QUERY_INFO_FW,
} QUERY_IOCTL_LIST;

#define MAX_QUERY_INFO  128

struct ljm_ioctl_query
{
    uint32_t card;
    uint32_t info;
    uint8_t pdata[MAX_QUERY_INFO];
};

typedef enum _CMD_IOCTL_TYPE {
    CMD_TYPE_GET_PVT_TEMP,
    CMD_TYPE_GET_PVT_VOLT,
    CMD_TYPE_GET_SVT_PROCESS,
    CMD_TYPE_GET_LVT_PROCESS,
    CMD_TYPE_GET_ULVT_PROCESS,
    CMD_TYPE_DUMP_VRAM_FREE,
} CMD_IOCTL_TYPE;

#define MAX_CMD_DATA  64

struct ljm_ioctl_cmd
{
    uint32_t card;
    uint32_t type;
    uint32_t index;
    uint32_t size;
    uint8_t pdata[MAX_CMD_DATA];
};

#define LJM_IOCTL_IOPORT_OP_READ    0
#define LJM_IOCTL_IOPORT_OP_WRITE   1

struct ljm_ioctl_ioport
{
    uint8_t offset;
    uint8_t val;
    uint8_t op;
};

#define LJM_IOCTL_GAMMA_SIZE 256
struct ljm_ioctl_gamma
{
    int crtc_id;

    uint32_t gamma_size; //Default: LJM_IOCTL_GAMMA_SIZE=256

    /* pointers to arrays */
    uint64_t red;
    uint64_t green;
    uint64_t blue;
};

#define LJM_IOCTL_POWER_OP_GET 0
#define LJM_IOCTL_POWER_OP_SET 1

#define LJM_POWER_STATE_INVALID     (-1)
#define LJM_POWER_STATE_PERFORMANCE 0
#define LJM_POWER_STATE_SAVE        1
#define LJM_POWER_STATE_STANDBY     2
#define LJM_POWER_STATE_MAX         3

struct ljm_ioctl_power
{
    uint32_t op;
    int32_t  state;
};

#define LJM_IOCTL_MAGIC 'd'

#define LJM_IOCTL_MMIO   _IOWR(LJM_IOCTL_MAGIC, 1, struct ljm_ioctl_mmio)
#define LJM_IOCTL_DDC    _IOWR(LJM_IOCTL_MAGIC, 2, struct ljm_ioctl_ddc)
#define LJM_IOCTL_VER    _IOWR(LJM_IOCTL_MAGIC, 3, struct ljm_ioctl_ver)
#define LJM_IOCTL_PCI    _IOWR(LJM_IOCTL_MAGIC, 4, struct ljm_ioctl_pci)
#define LJM_IOCTL_INFO   _IOWR(LJM_IOCTL_MAGIC, 5, struct ljm_ioctl_info)
#define LJM_IOCTL_KMEM   _IOWR(LJM_IOCTL_MAGIC, 6, struct ljm_ioctl_kmem)
#define LJM_IOCTL_3D     _IOWR(LJM_IOCTL_MAGIC, 7, struct ljm_ioctl_3d_command)
#define LJM_IOCTL_2D     _IOWR(LJM_IOCTL_MAGIC, 8, struct ljm_ioctl_2d_command)
#define LJM_IOCTL_DAC    _IOWR(LJM_IOCTL_MAGIC, 9, struct ljm_ioctl_dac_command)
#define LJM_IOCTL_HDMI   _IOWR(LJM_IOCTL_MAGIC, 10, struct ljm_ioctl_hdmi_common)
#define LJM_IOCTL_DVO    _IOWR(LJM_IOCTL_MAGIC, 11, struct ljm_ioctl_dvo_command)
#define LJM_IOCTL_DP     _IOWR(LJM_IOCTL_MAGIC, 12, struct ljm_ioctl_dp_command)
#define LJM_IOCTL_DDR    _IOWR(LJM_IOCTL_MAGIC, 13, struct ljm_ioctl_ddr_command)
#define LJM_IOCTL_SPI    _IOWR(LJM_IOCTL_MAGIC, 14, struct ljm_ioctl_spi_command)
#define LJM_IOCTL_I2C    _IOWR(LJM_IOCTL_MAGIC, 15, struct ljm_ioctl_i2c_command)
#define LJM_IOCTL_QUERY  _IOWR(LJM_IOCTL_MAGIC, 16, struct ljm_ioctl_query)
#define LJM_IOCTL_CMD    _IOWR(LJM_IOCTL_MAGIC, 17, struct ljm_ioctl_cmd)
#define LJM_IOCTL_IOPORT _IOWR(LJM_IOCTL_MAGIC, 18, struct ljm_ioctl_ioport)
#define LJM_IOCTL_GAMMA  _IOWR(LJM_IOCTL_MAGIC, 19, struct ljm_ioctl_gamma)
#define LJM_IOCTL_POWER  _IOWR(LJM_IOCTL_MAGIC, 20, struct ljm_ioctl_power)
#define LJM_IOCTL_SET_FAN_MANUAL   _IOWR(LJM_IOCTL_MAGIC, 21, int)/* 开启风扇转速手动控制 */
#define LJM_IOCTL_CLEAR_MANUAL _IO(LJM_IOCTL_MAGIC, 22)/* 清除风扇转速手动控制 */

#ifdef __cplusplus
}
#endif

#endif /*__LJMICRO_UTIL_IOCTL_H__*/
