/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_hdmi.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

#ifdef UEFI_GOP
#define FB_VMODE_NONINTERLACED  0	/* non interlaced */
#define FB_VMODE_INTERLACED     1	/* interlaced	*/
#define FB_VMODE_DOUBLE         2	/* double scan */
#define FB_VMODE_ODD_FLD_FIRST  4	/* interlaced: top line first */
#define FB_VMODE_MASK           255

#define FB_VMODE_YWRAP          256	/* ywrap instead of panning     */
#define FB_VMODE_SMOOTH_XPAN    512	/* smooth xpan possible (internally used) */
#define FB_VMODE_CONUPDATE      512	/* don't update x/yoffset	*/
#endif

static const struct hdmi_video_timing hdmi_mode[] = {
    {
        .mode = {
            .name = "720x480i@60Hz",
            .refresh = 60,
            .xres = 720,
            .yres = 480,
            .pixclock = 27000000,
            .left_margin = 57,
            .right_margin = 19,
            .upper_margin = 15,
            .lower_margin = 4,
            .hsync_len = 62,
            .vsync_len = 3,
            .sync = 0,
            .vmode = FB_VMODE_INTERLACED,
            .flag = 0,
        },
        .vic = HDMI_720X480I_60HZ_4_3,
        .vic_2nd = HDMI_720X480I_60HZ_16_9,
        .pixelrepeat = 2,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "720x576i@50Hz",
            .refresh = 50,
            .xres = 720,
            .yres = 576,
            .pixclock = 27000000,
            .left_margin = 69,
            .right_margin = 12,
            .upper_margin = 19,
            .lower_margin = 2,
            .hsync_len = 63,
            .vsync_len = 3,
            .sync = 0,
            .vmode = FB_VMODE_INTERLACED,
            .flag = 0,
        },
        .vic = HDMI_720X576I_50HZ_4_3,
        .vic_2nd = HDMI_720X576I_50HZ_16_9,
        .pixelrepeat = 2,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "720x480p@60Hz",
            .refresh = 60,
            .xres = 720,
            .yres = 480,
            .pixclock = 27000000,
            .left_margin = 60,
            .right_margin = 16,
            .upper_margin = 30,
            .lower_margin = 9,
            .hsync_len = 62,
            .vsync_len = 6,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_720X480P_60HZ_4_3,
        .vic_2nd = HDMI_720X480P_60HZ_16_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "720x576p@50Hz",
            .refresh = 50,
            .xres = 720,
            .yres = 576,
            .pixclock = 27000000,
            .left_margin = 68,
            .right_margin = 12,
            .upper_margin = 39,
            .lower_margin = 5,
            .hsync_len = 64,
            .vsync_len = 5,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_720X576P_50HZ_4_3,
        .vic_2nd = HDMI_720X576P_50HZ_16_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x720p@24Hz",
            .refresh = 24,
            .xres = 1280,
            .yres = 720,
            .pixclock = 59400000,
            .left_margin = 220,
            .right_margin = 1760,
            .upper_margin = 20,
            .lower_margin = 5,
            .hsync_len = 40,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1280X720P_24HZ,
        .vic_2nd = HDMI_1280X720P_24HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x720p@25Hz",
            .refresh = 25,
            .xres = 1280,
            .yres = 720,
            .pixclock = 74250000,
            .left_margin = 220,
            .right_margin = 2420,
            .upper_margin = 20,
            .lower_margin = 5,
            .hsync_len = 40,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1280X720P_25HZ,
        .vic_2nd = HDMI_1280X720P_25HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x720p@30Hz",
            .refresh = 30,
            .xres = 1280,
            .yres = 720,
            .pixclock = 74250000,
            .left_margin = 220,
            .right_margin = 1760,
            .upper_margin = 20,
            .lower_margin = 5,
            .hsync_len = 40,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1280X720P_30HZ,
        .vic_2nd = HDMI_1280X720P_30HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x720p@50Hz",
            .refresh = 50,
            .xres = 1280,
            .yres = 720,
            .pixclock = 74250000,
            .left_margin = 220,
            .right_margin = 440,
            .upper_margin = 20,
            .lower_margin = 5,
            .hsync_len = 40,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1280X720P_50HZ,
        .vic_2nd = HDMI_1280X720P_50HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x720p@60Hz",
            .refresh = 60,
            .xres = 1280,
            .yres = 720,
            .pixclock = 74250000,
            .left_margin = 220,
            .right_margin = 110,
            .upper_margin = 20,
            .lower_margin = 5,
            .hsync_len = 40,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1280X720P_60HZ,
        .vic_2nd = HDMI_1280X720P_60HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080i@50Hz",
            .refresh = 50,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 74250000,
            .left_margin = 148,
            .right_margin = 528,
            .upper_margin = 15,
            .lower_margin = 2,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = FB_VMODE_INTERLACED,
            .flag = 0,
        },
        .vic = HDMI_1920X1080I_50HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080i@60Hz",
            .refresh = 60,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 74250000,
            .left_margin = 148,
            .right_margin = 88,
            .upper_margin = 15,
            .lower_margin = 2,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = FB_VMODE_INTERLACED,
            .flag = 0,
        },
        .vic = HDMI_1920X1080I_60HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080p@24Hz",
            .refresh = 24,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 74250000,
            .left_margin = 148,
            .right_margin = 638,
            .upper_margin = 36,
            .lower_margin = 4,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1920X1080P_24HZ,
        .vic_2nd = HDMI_1920X1080P_24HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080p@25Hz",
            .refresh = 25,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 74250000,
            .left_margin = 148,
            .right_margin = 528,
            .upper_margin = 36,
            .lower_margin = 4,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1920X1080P_25HZ,
        .vic_2nd = HDMI_1920X1080P_25HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080p@30Hz",
            .refresh = 30,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 74250000,
            .left_margin = 148,
            .right_margin = 88,
            .upper_margin = 36,
            .lower_margin = 4,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1920X1080P_30HZ,
        .vic_2nd = HDMI_1920X1080P_30HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080p@50Hz",
            .refresh = 50,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 148500000,
            .left_margin = 148,
            .right_margin = 528,
            .upper_margin = 36,
            .lower_margin = 4,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1920X1080P_50HZ,
        .vic_2nd = HDMI_1920X1080P_50HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1920x1080p@60Hz",
            .refresh = 60,
            .xres = 1920,
            .yres = 1080,
            .pixclock = 148500000,
            .left_margin = 148,
            .right_margin = 88,
            .upper_margin = 36,
            .lower_margin = 4,
            .hsync_len = 44,
            .vsync_len = 5,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_1920X1080P_60HZ,
        .vic_2nd = HDMI_1920X1080P_60HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "3840x2160p@24Hz",
            .refresh = 24,
            .xres = 3840,
            .yres = 2160,
            .pixclock = 297000000,
            .left_margin = 296,
            .right_margin = 1276,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_3840X2160P_24HZ,
        .vic_2nd = HDMI_3840X2160P_24HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "3840x2160p@25Hz",
            .refresh = 25,
            .xres = 3840,
            .yres = 2160,
            .pixclock = 297000000,
            .left_margin = 296,
            .right_margin = 1056,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_3840X2160P_25HZ,
        .vic_2nd = HDMI_3840X2160P_25HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "3840x2160p@30Hz",
            .refresh = 30,
            .xres = 3840,
            .yres = 2160,
            .pixclock = 297000000,
            .left_margin = 296,
            .right_margin = 176,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_3840X2160P_30HZ,
        .vic_2nd = HDMI_3840X2160P_30HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "4096x2160p@24Hz",
            .refresh = 24,
            .xres = 4096,
            .yres = 2160,
            .pixclock = 297000000,
            .left_margin = 296,
            .right_margin = 1020,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_4096X2160P_24HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "4096x2160p@25Hz",
            .refresh = 25,
            .xres = 4096,
            .yres = 2160,
            .pixclock = 297000000,
            .left_margin = 128,
            .right_margin = 968,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_4096X2160P_25HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "4096x2160p@30Hz",
            .refresh = 30,
            .xres = 4096,
            .yres = 2160,
            .pixclock = 297000000,
            .left_margin = 128,
            .right_margin = 88,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_4096X2160P_30HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "3840x2160p@50Hz",
            .refresh = 50,
            .xres = 3840,
            .yres = 2160,
            .pixclock = 594000000,
            .left_margin = 296,
            .right_margin = 1056,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_3840X2160P_50HZ,
        .vic_2nd = HDMI_3840X2160P_50HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "3840x2160p@60Hz",
            .refresh = 60,
            .xres = 3840,
            .yres = 2160,
            .pixclock = 594000000,
            .left_margin = 296,
            .right_margin = 176,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_3840X2160P_60HZ,
        .vic_2nd = HDMI_3840X2160P_60HZ_21_9,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "4096x2160p@50Hz",
            .refresh = 50,
            .xres = 4096,
            .yres = 2160,
            .pixclock = 594000000,
            .left_margin = 128,
            .right_margin = 968,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_4096X2160P_50HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "4096x2160p@60Hz",
            .refresh = 60,
            .xres = 4096,
            .yres = 2160,
            .pixclock = 594000000,
            .left_margin = 128,
            .right_margin = 88,
            .upper_margin = 72,
            .lower_margin = 8,
            .hsync_len = 88,
            .vsync_len = 10,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_4096X2160P_60HZ,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "800x600p@60Hz",
            .refresh = 60,
            .xres = 800,
            .yres = 600,
            .pixclock = 40000000,
            .left_margin = 88,
            .right_margin = 40,
            .upper_margin = 23,
            .lower_margin = 1,
            .hsync_len = 128,
            .vsync_len = 4,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 1,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1024x768p@60Hz",
            .refresh = 60,
            .xres = 1024,
            .yres = 768,
            .pixclock = 65000000,
            .left_margin = 160,
            .right_margin = 24,
            .upper_margin = 29,
            .lower_margin = 3,
            .hsync_len = 136,
            .vsync_len = 6,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 2,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x960p@60Hz",
            .refresh = 60,
            .xres = 1280,
            .yres = 960,
            .pixclock = 108000000,
            .left_margin = 312,
            .right_margin = 96,
            .upper_margin = 36,
            .lower_margin = 1,
            .hsync_len = 112,
            .vsync_len = 3,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 3,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1280x1024p@60Hz",
            .refresh = 60,
            .xres = 1280,
            .yres = 1024,
            .pixclock = 108000000,
            .left_margin = 248,
            .right_margin = 48,
            .upper_margin = 38,
            .lower_margin = 1,
            .hsync_len = 112,
            .vsync_len = 3,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 4,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1360x768p@60Hz",
            .refresh = 60,
            .xres = 1360,
            .yres = 768,
            .pixclock = 85500000,
            .left_margin = 256,
            .right_margin = 64,
            .upper_margin = 18,
            .lower_margin = 3,
            .hsync_len = 112,
            .vsync_len = 6,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 5,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1366x768p@60Hz",
            .refresh = 60,
            .xres = 1366,
            .yres = 768,
            .pixclock = 85500000,
            .left_margin = 213,
            .right_margin = 70,
            .upper_margin = 24,
            .lower_margin = 3,
            .hsync_len = 143,
            .vsync_len = 3,
            .sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 6,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1440x900p@60Hz",
            .refresh = 60,
            .xres = 1440,
            .yres = 900,
            .pixclock = 106500000,
            .left_margin = 232,
            .right_margin = 80,
            .upper_margin = 25,
            .lower_margin = 3,
            .hsync_len = 152,
            .vsync_len = 6,
            .sync = FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 7,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1600x900p@60Hz",
            .refresh = 60,
            .xres = 1600,
            .yres = 900,
            .pixclock = 108000000,
            .left_margin = 96,
            .right_margin = 24,
            .upper_margin = 96,
            .lower_margin = 1,
            .hsync_len = 80,
            .vsync_len = 3,
            .sync = FB_SYNC_VERT_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 8,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1680x1050@60Hz",
            .refresh = 60,
            .xres = 1680,
            .yres = 1050,
            .pixclock = 146250000,
            .left_margin = 280,
            .right_margin = 104,
            .upper_margin = 30,
            .lower_margin = 3,
            .hsync_len = 176,
            .vsync_len = 6,
            .sync = FB_SYNC_VERT_HIGH_ACT,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DMT | 9,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        .mode = {
            .name = "1440x1280@60Hz",
            .refresh = 60,
            .xres = 1440,
            .yres = 1280,
            .pixclock = 148500000,
            .left_margin = 84,
            .right_margin = 360,
            .upper_margin = 8,
            .lower_margin = 10,
            .hsync_len = 20,
            .vsync_len = 2,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DISCRETE_VR | 1,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        /* AUO 3.81 */
        .mode = {
            .name = "2160x1200@75Hz",
            .refresh = 75,
            .xres = 2160,
            .yres = 1200,
            .pixclock = 245000000,
            .left_margin = 100,
            .right_margin = 420,
            .upper_margin = 3,
            .lower_margin = 6,
            .hsync_len = 32,
            .vsync_len = 3,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DISCRETE_VR | 3,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        /* sharp 2.89 */
        .mode = {
            .name = "2880x1440@75Hz",
            .refresh = 75,
            .xres = 2880,
            .yres = 1440,
            .pixclock = 340000000,
            .left_margin = 100,
            .right_margin = 50,
            .upper_margin = 8,
            .lower_margin = 6,
            .hsync_len = 50,
            .vsync_len = 1,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DISCRETE_VR | 4,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        /* RAYKEN 5.46 */
        .mode = {
            .name = "1440x2560@60Hz",
            .refresh = 60,
            .xres = 1440,
            .yres = 2560,
            .pixclock = 268500000,
            .left_margin = 50,
            .right_margin = 200,
            .upper_margin = 20,
            .lower_margin = 20,
            .hsync_len = 20,
            .vsync_len = 10,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DISCRETE_VR | 5,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
    {
        /* samsung */
        .mode = {
            .name = "1440x2560@70Hz",
            .refresh = 70,
            .xres = 1440,
            .yres = 2560,
            .pixclock = 285000000,
            .left_margin = 40,
            .right_margin = 80,
            .upper_margin = 2,
            .lower_margin = 6,
            .hsync_len = 20,
            .vsync_len = 8,
            .sync = 0,
            .vmode = 0,
            .flag = 0,
        },
        .vic = HDMI_VIDEO_DISCRETE_VR | 6,
        .vic_2nd = 0,
        .pixelrepeat = 1,
        .interface = OUT_P888,
    },
};

/**
 * hdmi_videomode_to_vic: transverse video mode to vic
 * @vmode: videomode to transverse
 *
 */
int hdmi_videomode_to_vic(struct fb_videomode *vmode)
{
    struct fb_videomode *mode;
    unsigned int vic = 0;
    int i = 0;

    for (i = 0; i < ARRAY_SIZE(hdmi_mode); i++) {
        mode = (struct fb_videomode *)&hdmi_mode[i].mode;
        if (vmode->vmode == mode->vmode &&
            //vmode->refresh == mode->refresh &&    /* hym: as refresh is optional */
            vmode->pixclock == mode->pixclock &&    /* hym: replace refresh with pixclock */
            vmode->xres == mode->xres &&
            vmode->yres == mode->yres &&
            vmode->left_margin == mode->left_margin &&
            vmode->right_margin == mode->right_margin &&
            vmode->upper_margin == mode->upper_margin &&
            vmode->lower_margin == mode->lower_margin &&
            vmode->hsync_len == mode->hsync_len &&
            vmode->vsync_len == mode->vsync_len) {
            vic = hdmi_mode[i].vic;
            break;
        }
    }
    return vic;
}

/**
 * hdmi_vic2timing: transverse vic mode to video timing
 * @vmode: vic to transverse
 *
 */
const struct hdmi_video_timing *hdmi_vic2timing(int vic)
{
    int i;

    if (vic == 0)
        return NULL;

    for (i = 0; i < ARRAY_SIZE(hdmi_mode); i++) {
        if (hdmi_mode[i].vic == vic || hdmi_mode[i].vic_2nd == vic)
            return &hdmi_mode[i];
    }
    return NULL;
}

/**
 * hdmi_vic_to_videomode: transverse vic mode to video mode
 * @vmode: vic to transverse
 *
 */
const struct fb_videomode *hdmi_vic_to_videomode(int vic)
{
    int i, vid;

    if (vic == 0)
        return NULL;
    else if ((vic & HDMI_VIDEO_DMT) || (vic & HDMI_VIDEO_DISCRETE_VR))
        vid = vic;
    else
        vid = vic & HDMI_VIC_MASK;
    for (i = 0; i < ARRAY_SIZE(hdmi_mode); i++) {
        if (hdmi_mode[i].vic == vid ||
            hdmi_mode[i].vic_2nd == vid)
            return &hdmi_mode[i].mode;
    }
    return NULL;
}

/**
 * hdmi_displaymode_to_vic: transverse video mode to vic
 * @vmode: videomode to transverse
 *
 */
int hdmi_displaymode_to_vic(struct drm_display_mode *dmode)
{
    struct fb_videomode *mode;
    unsigned int vic = 0;
    int i = 0;
    u32 vmode = (dmode->flags & DRM_MODE_FLAG_INTERLACE) ? FB_VMODE_INTERLACED : FB_VMODE_NONINTERLACED;
    u32 xres = dmode->hdisplay;
    u32 yres = dmode->vdisplay;
    u32 left_margin = dmode->htotal - dmode->hsync_end;
    u32 right_margin = dmode->hsync_start - dmode->hdisplay;
    u32 hsync_len = dmode->hsync_end - dmode->hsync_start;
    u32 upper_margin = dmode->vtotal - dmode->vsync_end;
    u32 lower_margin = dmode->vsync_start - dmode->vdisplay;
    u32 vsync_len = dmode->vsync_end - dmode->vsync_start;
    u32 pixclock = dmode->clock * 1000;

    DRM_DEBUG_KMS("displaymode: %d(%d %d %d)x%d(%d %d %d)@%d_%d\n",
        dmode->hdisplay, dmode->hsync_start, dmode->hsync_end, dmode->htotal,
        dmode->vdisplay, dmode->vsync_start, dmode->vsync_end, dmode->vtotal,
        dmode->clock, dmode->flags);

    for (i = 0; i < ARRAY_SIZE(hdmi_mode); i++) {
        mode = (struct fb_videomode *)&hdmi_mode[i].mode;
        if (vmode == mode->vmode &&
            pixclock == mode->pixclock &&
            xres == mode->xres &&
            yres == mode->yres &&
            left_margin == mode->left_margin &&
            right_margin == mode->right_margin &&
            upper_margin == mode->upper_margin &&
            lower_margin == mode->lower_margin &&
            hsync_len == mode->hsync_len &&
            vsync_len == mode->vsync_len)
        {
            vic = hdmi_mode[i].vic;
            DRM_DEBUG_KMS("%dx%d@%d is found\n", xres, yres, pixclock);
            break;
        }
    }

    if (vic == 0) {
        /* 只根据显示宽、高和像素时钟查找VIC */
        for (i = 0; i < ARRAY_SIZE(hdmi_mode); i++) {
            mode = (struct fb_videomode *)&hdmi_mode[i].mode;
            if (xres == mode->xres &&
                yres == mode->yres &&
                pixclock == mode->pixclock)
            {
                vic = hdmi_mode[i].vic;
                DRM_DEBUG_KMS("%dx%d@%d is found except sync timing\n", xres, yres, pixclock);
                break;
            }
        }
    }

    return vic;
}

/**
 * hdmi_videomode_to_vic: transverse video mode to vic
 * @vmode: videomode to transverse
 *
 */
void hdmi_displaymode_to_videomode(struct drm_display_mode *dmode, struct fb_videomode *vmode)
{
    vmode->xres = dmode->hdisplay;
    vmode->yres = dmode->vdisplay;
    vmode->pixclock = dmode->clock * 1000;
    vmode->left_margin = dmode->htotal - dmode->hsync_end;
    vmode->right_margin = dmode->hsync_start - dmode->hdisplay;
    vmode->hsync_len = dmode->hsync_end - dmode->hsync_start;
    vmode->upper_margin = dmode->vtotal - dmode->vsync_end;
    vmode->lower_margin = dmode->vsync_start - dmode->vdisplay;
    vmode->vsync_len = dmode->vsync_end - dmode->vsync_start;
    vmode->sync = (dmode->flags & DRM_MODE_FLAG_PHSYNC) ? FB_SYNC_HOR_HIGH_ACT : 0;
    vmode->sync |= (dmode->flags & DRM_MODE_FLAG_PVSYNC) ? FB_SYNC_VERT_HIGH_ACT : 0;
    vmode->vmode = (dmode->flags & DRM_MODE_FLAG_INTERLACE) ? FB_VMODE_INTERLACED : FB_VMODE_NONINTERLACED;
    vmode->vmode |= (dmode->flags & DRM_MODE_FLAG_DBLSCAN) ? FB_VMODE_DOUBLE : 0;
    vmode->flag = 0;
    vmode->refresh = (vmode->pixclock + dmode->htotal * dmode->vtotal / 2) / (dmode->htotal * dmode->vtotal);
}

