/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *                                  标准库头文件                                *
 ******************************************************************************/
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
#include "drm/drm_dp_helper.h"
#else
#include "drm/display/drm_dp_helper.h"
#endif

/******************************************************************************
 *                                  非标准库头文件                              *
 ******************************************************************************/
#include "ljmicro_mcu.h"
#include "ljmicro_base.h"
#include "ljmicro_dp_hw.h"
#include "gp200.h"

/******************************************************************************
 *                              函数声明&全局变量                              *
 ******************************************************************************/
uint8_t ljm_crtc_reg_nolock_read(int card, uint8_t ind);

/******************************************************************************
 *                                  宏定义                                     *
 ******************************************************************************/
#define MCU_REG_R8(mcu, offset) RREG8_BASE((mcu)->regbase, offset)
#define MCU_REG_W8(mcu, offset, val) \
    do \
    { \
        WREG8_BASE((mcu)->regbase, (offset), (val)); \
    }while(0)

#define MCU_REG_R32(mcu, offset) RREG32_BASE((mcu)->regbase, offset)
#define MCU_REG_W32(mcu, offset, val) \
    do \
    { \
        WREG32_BASE((mcu)->regbase, (offset), (val)); \
    }while(0)

#define BIT_CRTC(nr) (0x1 << (nr))
#define ABMA_TIMEOUT_CNT 0x0F /* Read axi max times to wait axi down */
#define AXI_READ_TIMEOUT       (-2) /* wait read timeout */
#define GP201_AXI_READ_READY_MASK 0xF0
#define GP201_AXI_READ_READY 0xE0
#define GP202_AXI_READ_READY_MASK 0XF0
#define GP202_AXI_READ_READY 0XA0
#define AXI_READ_ADDR_MASK 0x00
#define AXI_READ_CONTROL_MASK 0x40
#define AXI_REG_CONTROL_MASK 0xB0
#define AXI_REG_ADDR_CTL 0xB1
#define AXI_REG_ADDR 0xB2
#define AXI_REG_READ_BYTEEN 0xB7
#define AXI_REG_READ_DATA 0xB8
#define AXI_REG_STATUS 0xB9 /* 在GP201中，AXI_REG_STATUS 寄存器为只读寄存器，不可写入 */
#define AXI_RD_WR_SIZE 0XBA


#define CR_PRE_RD(card, ind) ljm_crtc_reg_nolock_read((card), (ind))

#define CHECK_AXI_READ_COMPLETE(uc, card, ready_rm, ready_r) \
    do \
    { \
        int __timeout_cnt = ABMA_TIMEOUT_CNT;\
        CR_WR(card, AXI_REG_CONTROL_MASK, AXI_READ_CONTROL_MASK);\
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x00);\
        CR_WR(card, AXI_REG_STATUS, 0x00);\
        uc = CR_PRE_RD(card, AXI_REG_STATUS);\
        while (((uc & ready_rm)!=ready_r)  && __timeout_cnt > 0) {\
            uc = CR_PRE_RD(card, AXI_REG_STATUS); \
            usleep_range(50, 60); \
            __timeout_cnt--; \
        } \
        if (__timeout_cnt <= 0) \
        { \
            return AXI_READ_TIMEOUT; \
        } \
    } \
    while (0)

#define CHECK_AXI_OKEY(card)\
    do{ \
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x01);\
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x00);\
    } while(0);

#define CHECK_AXI_ADDRESS_CHOOSE(card)\
    do{ \
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x20);\
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x00);\
    } while(0);

#define CHECK_AXI_READ_CHOOSE(card)\
    do{ \
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x04);\
        CR_WR(card, AXI_REG_CONTROL_MASK, 0x00);\
    } while(0);

/******************************************************************************
 *                                 函数实现                                    *
 ******************************************************************************/
/**
 * ljm_mcu_firmware_detect - 探测mcu固件
 * @card: GPU序号；
 * @mcu_flash_offset: mcu固件在flash上的起始地址。
 *
 * 返回值：
 * 有mcu固件则返回0；无或者失败时则返回负值；
*/
int ljm_mcu_firmware_detect(int card, uint64_t mcu_flash_offset, unsigned long flags)
{
    uint8_t val = 0, uc, a_lowest, byte_en;
    uint16_t strob_en;
    uint64_t offset;
    int i;

    offset = mcu_flash_offset & 0xFFFFFFFFFFFFFFF0ULL;
    a_lowest  = (uint8_t)((offset & 0xff) >> 0);
    byte_en = 0x0f;
    strob_en = 0x01 << (a_lowest & byte_en);

    REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, 0x3);

    for (i = 0; i < 5; ++i)
    {
        CR_WR(card, AXI_REG_ADDR, (uint8_t)((offset >> (8 * i)) & 0xff)); //向B2寄存器写入address[7；0]
        CR_WR(card, AXI_REG_ADDR_CTL, AXI_READ_ADDR_MASK + BIT_CRTC(i));    //向B1寄存器写入address mask 01
        CHECK_AXI_ADDRESS_CHOOSE(card);
    }

    if ((flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
    {
        /* axi_rd_wr_size[3:0]:控制单次写入数据位宽,建议写入0x04。[7:4]:控制单次读取数据位宽。建议写入0x04。 */
        CR_WR(card, AXI_RD_WR_SIZE, 0x44);
        CHECK_AXI_READ_COMPLETE(uc, card, GP202_AXI_READ_READY_MASK, GP202_AXI_READ_READY);
    }
    else
        CHECK_AXI_READ_COMPLETE(uc, card, GP201_AXI_READ_READY_MASK, GP201_AXI_READ_READY);

    CR_WR(card, AXI_REG_READ_BYTEEN, (a_lowest & byte_en) | 0x10);
    CHECK_AXI_READ_CHOOSE(card);

    val = CR_PRE_RD(card, AXI_REG_READ_DATA);

    if ((flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
        CHECK_AXI_READ_COMPLETE(uc, card, GP202_AXI_READ_READY_MASK, GP202_AXI_READ_READY);
    else
        CHECK_AXI_READ_COMPLETE(uc, card, GP201_AXI_READ_READY_MASK, GP201_AXI_READ_READY);

    CHECK_AXI_OKEY(card);

    REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, 0x0);
    return (val && (val != 0xff)) ? 0 : -1;
}

bool ljm_mcu_ready(int card)
{
    return (MCU_IDENTIFICATION_CODE2 == REG_R32(card, MCU_IDENTIFICATION_ADDR));
}

/**
 * ljm_mcu_reset - 复位mcu。
 * 在复位完成后，将mcu标识码反转；
 *
 * @card: GPU序号。
 *
 * 返回值：
 * 复位成功则返回0；复位失败返回负值。
*/
int ljm_mcu_reset(int card)
{
    int count=0;

    //todo: mcu复位有条件限制，如：mcu固件、xip模式等。本函数缺少这些条件检查。

    LJM_CRU_W32(card, CRU_HWRST, 0x00);
    LJM_CRU_W32(card, CRU_HWRST, 0x08);
    msleep(2);
    while(!(MCU_IDENTIFICATION_CODE1 == REG_R32(card, MCU_IDENTIFICATION_ADDR)))
    {
        msleep(1);
        if(count++>=200)
            return -ENODEV;
    }
    //将该寄存器的值做位反转，清除标志位。避免在后面可能发生的复位过程中，检测到上一次的遗留值。
    REG_W32(card, MCU_IDENTIFICATION_ADDR, MCU_IDENTIFICATION_CODE2);
    return 0;
}

int ljm_mcu_pre_init(int card, uint64_t mcu_flash_offset, unsigned long flags)
{
    uint32_t val;
    int ret;

    val = LJM_MISC_R32(card, MISC_QSPI_MCU_DONE);
    if (val & 0x1)
    {
        DEBUG_PRINT("SPI has exited from XIP mode. Asumming mcu was inited in BIOS.\n");
        return 0;
    }

    if(ljm_mcu_firmware_detect(card, mcu_flash_offset, flags) < 0)
        return 0;

    DEBUG_PRINT("Detect mcu firmware.\n");
    ret = ljm_mcu_reset(card);
    if(ret)
    {
        ERR_PRINT("Mcu's resetting failed.\n");
        return ret;
    }
    return 0;
}

struct ljm_mcu *ljm_mcu_init(int card, void *base)
{
    struct ljm_mcu *mcu;

    if(!ljm_mcu_ready(card))
        return NULL;

    mcu = kmalloc(sizeof(struct ljm_mcu), GFP_KERNEL);
    if(!mcu)
        return NULL;

    mcu->regbase = base;
    mcu->uart = !!(MCU_REG_R8(mcu, CHIP_CAPACITY) & CHIP_CAPACITY_UART);
    DEBUG_PRINT("Mcu init.\n");

    return mcu;
}

static inline void ljm_mcu_aux_launch(struct ljm_mcu *mcu)
{
    MCU_REG_W8(mcu, MCU_AUX_STATUS_OFFSET, MCU_AUX_STATUS_LAUNCH);
    return;
}

static int ljm_dp_mcu_fifo_write(struct ljm_mcu *mcu, u8 *buff, int len)
{
    int i, count=0;
    for(i=0; i<len; i++)
    {
        MCU_REG_W8(mcu, MCU_AUX_DATA_OFFSET+i, buff[i]);
        count++;
    }

    return count;
}

static int ljm_dp_mcu_fifo_read(struct ljm_mcu *mcu, u8 *buff, int len)
{
    int i, count=0;
    for(i=0; i<len; i++)
    {
        buff[i] = MCU_REG_R8(mcu, MCU_AUX_DATA_OFFSET+i);
        count++;
    }

    return count;
}

void ljm_dp_mcu_reg_clear(struct ljm_mcu *mcu)
{
    int i;
    for(i=0; i<256 ;i++)
        MCU_REG_W8(mcu, _MCU_BASE+i, 0x0);
    return;
}

static int ljm_dp_mcu_dpcd_trans(struct ljm_mcu *mcu, struct mcu_arg_aux *args)
{
    int times=10, ret=0;
    u8 reply_code, reply_data_count;
    bool is_write = false;

    /* 检测mcu是否完成上次操作。若上次操作仍未结束，驱动认为mcu出现异常，清楚所有寄存器以复位mcu。 */
    if(MCU_AUX_STATUS_LAUNCH == MCU_REG_R8(mcu, MCU_AUX_STATUS_OFFSET))
    {
        ljm_dp_mcu_reg_clear(mcu);
        msleep(5);
    }

    switch (args->cmd & ~DP_AUX_I2C_MOT){
        case DP_AUX_NATIVE_WRITE:
            is_write = true;
            MCU_REG_W8(mcu, MCU_AUX_CMD_OFFSET, MCU_AUX_CMD_DPCD_WRITE);
            ret = ljm_dp_mcu_fifo_write(mcu, args->buff, args->len);
            break;
        //case DP_AUX_I2C_WRITE_STATUS_UPDATE:
        case DP_AUX_NATIVE_READ:
            MCU_REG_W8(mcu, MCU_AUX_CMD_OFFSET, MCU_AUX_CMD_DPCD_READ);
            break;
        case DP_AUX_I2C_MOT:
        case DP_AUX_I2C_WRITE:
            return 0;
        default:
            WARN_PRINT("mcu: invalid aux cmd %x\n", args->cmd);
            ret = -EINVAL;
            goto end;
    }

    MCU_REG_W32(mcu, MCU_AUX_ADDR_OFFSET, args->addr);
    MCU_REG_W8(mcu, MCU_AUX_LEN_OFFSET, args->len-1);
    MCU_REG_W8(mcu, MCU_AUX_DATA_NUM_OFFSET, args->len);
    MCU_REG_W8(mcu, MCU_AUX_GPIO_CHANNEL, (((args->gpio_direction & 0xf) << 4) | (args->gpio_data & 0xf)));
    ljm_mcu_aux_launch(mcu);

    while(times--)
    {
        if(MCU_END_FLAG == MCU_REG_R32(mcu, MCU_AUX_DATA_GET_FLAG))
        {
            reply_code = MCU_REG_R8(mcu, MCU_AUX_REPLY_OFFSET);
            args->reply = (reply_code & 0xf0) >> 4;
            if(reply_code & MCU_AUX_REPLY_CODE_ERROR)
            {
                ret = -EPIPE;
                goto end;
            }
            else
                break;
        }
        msleep(1);
    }

    if(times <= 0)
    {
        ret = -EBUSY;
        goto end;
    }
    if(!is_write)
    {
        reply_data_count = MCU_REG_R8(mcu, MCU_AUX_REPLY_DATA_NUM);
        if(reply_data_count > args->len)
            reply_data_count = args->len;
        ret = ljm_dp_mcu_fifo_read(mcu, args->buff, reply_data_count);
    }

end:
    MCU_REG_W32(mcu, MCU_AUX_DATA_GET_FLAG, 0);
    return ret;
}

int ljm_dp_mcu_get_edid(struct ljm_mcu *mcu, struct mcu_arg_edid *args)
{
    int times=100, ret=0;
    u8 reply_code;

    MCU_REG_W8(mcu, MCU_AUX_CMD_OFFSET, MCU_AUX_CMD_EDID_READ);
    MCU_REG_W8(mcu, MCU_AUX_SUBADDR_OFFSET, args->subaddr);
    MCU_REG_W32(mcu, MCU_AUX_ADDR_OFFSET, 0x50);
    MCU_REG_W8(mcu, MCU_AUX_DATA_NUM_OFFSET, args->len);
    MCU_REG_W8(mcu, MCU_AUX_GPIO_CHANNEL, (((args->gpio_direction & 0xf) << 4) | (args->gpio_data & 0xf)));
    ljm_mcu_aux_launch(mcu);

    while(times--)
    {
        if(MCU_END_FLAG == MCU_REG_R32(mcu, MCU_AUX_DATA_GET_FLAG))
        {
            reply_code = MCU_REG_R8(mcu, MCU_AUX_REPLY_OFFSET);
            if(reply_code & MCU_AUX_REPLY_CODE_ERROR)
            {
                ret = -EPIPE;
                goto end;
            }
            else
                break;
        }
        msleep(1);
    }

    if(times <= 0)
    {
        ret = -EBUSY;
        WARN_PRINT("[%s] get edid timeout: %d\n", __FUNCTION__,ret);
        goto end;
    }

    ret = ljm_dp_mcu_fifo_read(mcu, args->buff, args->len);

end:
    MCU_REG_W32(mcu, MCU_AUX_DATA_GET_FLAG, 0);
    return ret;
}

/**
 * 带有“投票”机制的edid检查，连续读几次edid并取其中出现频率高的数据，使用此方法提高稳定性。
 * 由于mcu中已有此处理，因此该功能没有在驱动中使用。
 * 在新增mcu串口功能时，ljm_dp_mcu_get_edid()函数发生改动，因此ljm_dp_mcu_get_edid_with_check()被注释。
*/
#if 0
static int majority_from_3bytes(uint8_t *buff, uint8_t *majority)
{
    int ret=1;

    if(buff[0] == buff[1] || buff[0] == buff[2])
    {
        *majority = buff[0];
        ret++;
    }

    if(buff[1] == buff[2])
    {
        *majority =  buff[1];
        ret++;
    }

    return ret;
}

int ljm_dp_mcu_get_edid_with_check(struct ljm_mcu *mcu, uint8_t *buff)
{
    uint8_t edid1[DDC_EDID_SIZE], edid2[DDC_EDID_SIZE];
    uint8_t val;
    uint8_t judge[5];
    int i, ret, retry_count;

    ljm_dp_mcu_get_edid(mcu, 0, DDC_EDID_SIZE, edid1);
    ljm_dp_mcu_get_edid(mcu, 0, DDC_EDID_SIZE, edid2);

    for(i=0; i<DDC_EDID_SIZE; i++)
    {
        if(edid1[i] != edid2[i])
        {
            retry_count = 0;
            while(retry_count < 3)
            {
                ljm_dp_mcu_get_edid(mcu, i, 1, &judge[0]);
                ljm_dp_mcu_get_edid(mcu, i, 1, &judge[1]);
                ljm_dp_mcu_get_edid(mcu, i, 1, &judge[2]);

                ret = majority_from_3bytes(judge, &val);
                if( ret == 3)
                {
                    edid1[i] = val;
                    break;
                }
                else if(ret == 2 && (val == edid1[1] || val == edid2[i]))
                {
                    edid1[i] = (val == edid1[i]) ? edid1[i] : edid2[i];
                    break;
                }
                retry_count++;
            }
        }
    }

    memcpy(buff, edid1, DDC_EDID_SIZE);
    return DDC_EDID_SIZE;
}
#endif

/**
 * ljm_mcu_uart_tx - MCU uart发送函数
 * @mcu: MCU设备指针；
 * @args: uart操作参数；
 *
 * 配置MCU寄存器，通过MCU串口实现发送数据功能。
 *
 * 返回值：成功返回0，失败返回负值。
*/
int ljm_mcu_uart_tx(struct ljm_mcu *mcu, struct mcu_arg_uart *args)
{
    int timeout = 100;
    char status = MCU_UART_TRANS_STATUS_FAILED;

    if(!mcu->uart)
        return -EPERM;

    MCU_REG_W8(mcu, MCU_UART_TRANS_STATUS, 0xff); //清空状态寄存器
    MCU_REG_W8(mcu, MCU_UART_DATA_LENGTH, args->len-1);
    MCU_REG_W8(mcu, MCU_UART_CMD, MCU_UART_CMD_TX);
    memset_io((mcu->regbase+MCU_UART_DATA_OFFSET), 0, args->len);
    memcpy_toio((mcu->regbase+MCU_UART_DATA_OFFSET), args->data, args->len);
    MCU_REG_W8(mcu, MCU_UART_OPTION, MCU_UART_LAUNCH);

    while(timeout--)
    {
        status = MCU_REG_R8(mcu, MCU_UART_TRANS_STATUS);
        if(status == MCU_UART_TRANS_STATUS_SUCCESS)
        {
            return 0;
        }
        msleep(2);
    }

    if(timeout == 0)
    {
        ERR_PRINT("MCU uart tx timeout.\n");
        return -EBUSY;
    }

    return 0;
}

/**
 * ljm_mcu_handle - MCU操作处理函数
 * @mcu: MCU设备指针；
 * @cmd: MCU操作指令；
 * @args: MCU指令参数。
 *
 * MCU操作处理统一接口，发起MCU操作需要指定MCU设备、明确操作命令，填充其所需的的参数。
 * 不同的cmd对应不同的args。详情可以查看ljmicro_mcu.h文件。
 *
 * 返回值：成功返回0，失败返回负值。
*/
int ljm_mcu_handle(struct ljm_mcu *mcu, enum mcu_cmd cmd, void *args)
{
    int ret;

    if(!ljm_mcu_ready(0))
        return -ENXIO;

    switch(cmd)
    {
        case MCU_CMD_DPCD:
            ret = ljm_dp_mcu_dpcd_trans(mcu, (struct mcu_arg_aux *)args);
            break;
        case MCU_CMD_EDID:
            ret = ljm_dp_mcu_get_edid(mcu, (struct mcu_arg_edid *)args);
            break;
        case MCU_CMD_UARTTX:
            ret = ljm_mcu_uart_tx(mcu, (struct mcu_arg_uart *)args);
            break;
        default:
            ERR_PRINT("Unknown mcu command.\n");
            return -EINVAL;
    }

    return ret;
}