/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_MCU_H__
#define __LJMICRO_MCU_H__

#include <linux/regmap.h>

/*设定SRAM1寄存器的高256个字节用于命令的传输与数据的接收*/
#define _MCU_BASE   0x3ff00
#define MCU_RESERVED_ADDRESS_1  (_MCU_BASE+0x0)
#define MCU_RESERVED_ADDRESS_2  (_MCU_BASE+0x4)

#define MCU_IDENTIFICATION_ADDR  (_MCU_BASE+0x8)
#define MCU_IDENTIFICATION_CODE1  (0x0709aa55) // 该值由mcu写入，表示mcu正常启动；
#define MCU_IDENTIFICATION_CODE2  (0xf8f655aa) // ~(0x0709aa55) 该值由驱动写入。清除0x0709aa55标志，同时表示驱动已知mcu运行中。

#define CHIP_TYPE_VERSION_NUMBER    (_MCU_BASE+0xc)
#define CHIP_CAPACITY               (_MCU_BASE+0x22)
#define CHIP_CAPACITY_UART  BIT(0)

#define MCU_AUX_LEN_OFFSET      (_MCU_BASE+0x10)

#define MCU_AUX_CMD_OFFSET      (_MCU_BASE+0x11)
#define MCU_AUX_CMD_EDID_READ   0x1
#define MCU_AUX_CMD_DPCD_WRITE  0x8
#define MCU_AUX_CMD_DPCD_READ   0x9

#define MCU_AUX_REPLY_OFFSET    (_MCU_BASE+0x12)
#define MCU_AUX_REPLY_CODE_ERROR    (0x4)
#define MCU_AUX_REPLY_CODE_REQUESTING   (0x2)
#define MCU_AUX_REPLY_CODE_REPLYING     (0x1)
#define MCU_AUX_REPLY_CODE_RECEIEVED    (0x0)
#define MCU_AUX_REPLY_AUX_I2C_ACK   (0 << 8)
#define MCU_AUX_REPLY_AUX_NACK      (1 << 8)
#define MCU_AUX_REPLY_AUX_DEFER     (2 << 8)
#define MCU_AUX_REPLY_I2C_NACK      (4 << 8)
#define MCU_AUX_REPLY_I2C_DEFER     (8 << 8)

#define MCU_AUX_STATUS_OFFSET   (_MCU_BASE+0x13)
#define MCU_AUX_STATUS_LAUNCH (0xaa)
#define MCU_AUX_STATUS_FINISH (0)

#define MCU_AUX_ADDR_OFFSET     (_MCU_BASE+0x14)
#define MCU_AUX_REPLY_DATA_NUM  (_MCU_BASE+0x18) //mcu实际读取到的data长度
#define MCU_AUX_DATA_NUM_OFFSET (_MCU_BASE+0x19) //aux请求读取的data长度
#define MCU_AUX_SUBADDR_OFFSET  (_MCU_BASE+0x1a)
#define MCU_AUX_DATA_GET_FLAG   (_MCU_BASE+0x1c)
#define MCU_END_FLAG    0x55AA55AA
#define MCU_AUX_GPIO_CHANNEL    (_MCU_BASE+0x20)
#define MCU_AUX_DATA_OFFSET     (_MCU_BASE+0x80)

enum mcu_cmd
{
    MCU_CMD_UNKNOWN=0,
    MCU_CMD_DPCD,   /* struct mcu_arg_aux */
    MCU_CMD_EDID,   /* struct mcu_arg_edid */
    MCU_CMD_UARTTX, /* struct ljm_mcu */
};

struct mcu_arg_aux
{
    uint8_t     gpio_direction;
    uint8_t     gpio_data;
    u8 cmd;
    u32 addr;
    unsigned char *buff;
    u8 len;
    uint8_t reply;
};

struct mcu_arg_edid
{
    uint8_t     gpio_direction;
    uint8_t     gpio_data;
    uint8_t subaddr;
    uint16_t len;
    uint8_t *buff;
};


#define MCU_UART_DATA_LENGTH     (_MCU_BASE+0x10)
#define MCU_UART_CMD             (_MCU_BASE+0x11)
#define MCU_UART_CMD_RX      (0x0)
#define MCU_UART_CMD_TX      (0x1)
#define MCU_UART_TRANS_STATUS    (_MCU_BASE+0x12)
#define MCU_UART_TRANS_STATUS_SUCCESS (0x0)
#define MCU_UART_TRANS_STATUS_RUNNING (0x1)
#define MCU_UART_TRANS_STATUS_FAILED  (0x2)
#define MCU_UART_OPTION         (MCU_AUX_STATUS_OFFSET)
#define MCU_UART_LAUNCH               (0xbb)
#define MCU_UART_DATA_OFFSET     (_MCU_BASE+0x80)
#define MCU_UART_DATA_MAX        (0x80)

struct mcu_arg_uart
{
    uint16_t len; /*长度范围 1~128*/
    uint8_t  *data;
};

struct ljm_mcu
{
    void __iomem *      regbase;

    /*Capacity*/
    bool uart;
};

struct ljm_mcu *ljm_mcu_init(int card, void *base);
int ljm_mcu_handle(struct ljm_mcu *mcu, enum mcu_cmd cmd, void *arg);
int ljm_mcu_pre_init(int card, uint64_t mcu_flash_offset, unsigned long flags);

#endif /*__LJMICRO_MCU_H__*/