/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_MODE_H__
#define __LJMICRO_MODE_H__
#include "ljmicro_device.h"
#include "ljmicro_object.h"
#include "ljmicro_dac.h"
#include "ljmicro_hdmi.h"
#include "ljmicro_i2c.h"

/* DVI/VGA CRTC for FPGA version */
typedef struct _crtc_b_mode_reg {
    char name[16];
    unsigned short w;
    unsigned short h;
    int pclk;               /* Pixel Clock Freq(kHz) */
    unsigned int regs[9];   /* Screen B registers, mmio 0x0-20 */
} crtc_b_mode_reg;

typedef struct _crtc_mod_reg {
    char name[16];
    unsigned short w;
    unsigned short h;
    int pclk;                   /* Pixel Clock Freq(kHz) */
    unsigned char cr60_6F[16];  /* CRTC horizontal and vertical timing registers, CR60-6F */
    unsigned char cr78;         /* CRTC horizontal back porch[11:8] and sync[11:8] register, CR78 */
    unsigned char cr81_87[7];   /* Pixel clock registers, CR81-87 */
} crtc_mod_reg;

extern int calcDviPllParamWithDiv(
    IN int Xin,     /* PLL input frequency, kHz */
    IN int ClkOut,  /* output frequency, kHz */
    OUT int *N,     /* input divider value */
    OUT int *M,     /* feedback divider value */
    OUT int *OD,    /* 2^OD represents output divider value */
    OUT int *Div    /* enable flag to divide PLL output clock */
    );

extern void ljm_crtc_init(struct drm_device *ddev, int index);
extern void ljm_crtc_destroy(struct drm_crtc *crtc);

extern int ljm_setup_enc_conn(struct drm_device *ddev);
extern void ljm_encoder_destroy(struct drm_encoder *encoder);
extern void ljm_connector_destroy(IN struct drm_connector *connector);
extern int ljm_plane_init(struct ljm_device *ljm_dev);
extern int ljm_plane_fini(struct ljm_device *ljm_dev);
extern int ljm_zoomvideo_init(struct ljm_device *ljm_dev);
extern void ljm_zoomvideo_fini(struct ljm_device *ljm_dev);

extern struct ljm_fbdev *ljm_fbdev_init(struct drm_device *ljm_dev);
extern void ljm_fbdev_fini(struct ljm_device *ljm_dev);
extern struct drm_framebuffer *
    ljm_user_framebuffer_create(struct drm_device *ddev, struct drm_file *file_priv, const struct drm_mode_fb_cmd2 *mode_cmd);
//extern void ljm_fb_clear(struct ljm_device *ljm_dev);

extern void ljm_legacy_init_crtc(struct drm_device *ddev,
                struct ljm_crtc *ljm_crtc);
extern int ljm_crtc_mode_set_base(struct drm_crtc *crtc, int x, int y,
                struct drm_framebuffer *old_fb);
extern int ljm_crtc_cursor_set2(struct drm_crtc *crtc,
                   struct drm_file *file_priv,
                   uint32_t handle,
                   uint32_t width,
                   uint32_t height,
                   int32_t hot_x,
                   int32_t hot_y);
extern int ljm_crtc_cursor_move(struct drm_crtc *crtc,
                   int x, int y);
extern void ljm_cursor_reset(struct drm_crtc *crtc);

extern void ljm_crtc_handle_vblank(struct ljm_device *ljm_dev, int crtc);
extern void ljm_crtc_handle_flip(struct ljm_device *ljm_dev, int crtc);
extern void ljm_irq_kms_pflip_irq_get(struct ljm_device *ljm_dev, int crtc);
extern void ljm_irq_kms_pflip_irq_put(struct ljm_device *ljm_dev, int crtc);

extern int ljm_get_crtc_scanoutpos(struct drm_device *ddev, unsigned int pipe,
                   int *vpos, int *hpos,
                   ktime_t *stime, ktime_t *etime,
                   const struct drm_display_mode *mode);

bool ljm_fbdev_bo_is_fb(struct ljm_device *ljm_dev, struct ljm_gem_object *ljm_obj);

#endif
