/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_device.h"
#include "ljmicro_compat.h"
#include "ljmicro_mode.h"
#include "ljmicro_base.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,6,0)
#include <drm/drm_modeset_helper_vtables.h>
#endif
#define _LJ_OBJ_ZONE    ljvZONE_KMS

struct dp_dev;
void ljm_backlight_dev_init(struct dp_dev *dp_dev, struct drm_connector *drm_connector);

void
ljm_encoder_destroy(
    struct drm_encoder *encoder
    )
{
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    struct ljm_port *port = to_ljm_port(ljm_encoder);

    HEADER_ARG("encoder=0x%p", encoder);

    drm_encoder_cleanup(encoder);
    kfree(port);

    FOOTER();
}

/* drm_encoder operation functions */
static const struct drm_encoder_funcs ljm_encoder_funcs = {
    .destroy = ljm_encoder_destroy,
};

static void ljm_encoder_dpms(struct drm_encoder *encoder, int mode)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(encoder->dev);

    HEADER_ARG("encoder%d=%p %d mode=%d", encoder->index, encoder, encoder->base.id, mode);

    encoder_dpms(ljm_dev, encoder, mode);

    FOOTER();
}

static bool
ljm_encoder_mode_fixup(
    struct drm_encoder *encoder,
    const struct drm_display_mode *mode,
    struct drm_display_mode *adjustedMode
    )
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(encoder->dev);
    int r = false;

//  HEADER_ARG("encoder%d=0x%llx %d mode=%llx adjustedMode", encoder->index, (u64)encoder, encoder->base.id, (u64)mode, (u64)adjustedMode);

    r = encoder_mode_fixup(ljm_dev, encoder, mode, adjustedMode);

//  FOOTER();
    return r;
}

static void
ljm_encoder_mode_set(
    struct drm_encoder *encoder,
    struct drm_display_mode *mode,
    struct drm_display_mode *adjustedMode
    )
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(encoder->dev);

    HEADER_ARG("encoder%d=%p %d mode=%lx adjustedMode=%lx",
        encoder->index, encoder, encoder->base.id, (unsigned long)mode, (unsigned long)adjustedMode);

    encoder_mode_set(ljm_dev, encoder, mode, adjustedMode);

    FOOTER();
}

static void ljm_encoder_prepare(struct drm_encoder *encoder)
{
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    HEADER_ARG("encoder%d=%p %d", encoder->index, encoder, encoder->base.id);

    if (ljm_encoder->hard_type != LJM_ENCODER_DP)
        ljm_encoder_dpms(encoder, DRM_MODE_DPMS_OFF);

    FOOTER();
}

static void ljm_encoder_commit(struct drm_encoder *encoder)
{
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    HEADER_ARG("encoder%d=%p %d", encoder->index, encoder, encoder->base.id);

    if (ljm_encoder->hard_type != LJM_ENCODER_DP)
        ljm_encoder_dpms(encoder, DRM_MODE_DPMS_ON);

    FOOTER();
}


/* drm_encoder legacy helper functions */
static const struct drm_encoder_helper_funcs ljm_encoder_helper_funcs = {
    .dpms = ljm_encoder_dpms,
    .mode_fixup = ljm_encoder_mode_fixup,
    .mode_set = ljm_encoder_mode_set,
    .prepare = ljm_encoder_prepare,
    .commit = ljm_encoder_commit,
};

inline void ljm_drm_encoder_init(struct drm_device *ddev, struct drm_encoder *encoder)
{
#if LINUX_VERSION_CODE == KERNEL_VERSION(4,4,32) && defined CONFIG_MIPS
    drm_encoder_init(ddev, encoder, &ljm_encoder_funcs,
                     DRM_MODE_ENCODER_TMDS);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,0) || (LINUX_VERSION_CODE == KERNEL_VERSION(3,10,0) && defined CONFIG_MIPS)
    drm_encoder_init(ddev, encoder, &ljm_encoder_funcs,
                     DRM_MODE_ENCODER_TMDS, NULL);
#else
    drm_encoder_init(ddev, encoder, &ljm_encoder_funcs,
                     DRM_MODE_ENCODER_TMDS);
#endif

    return;
}

int gp200_port_get_crtc_id(int card, uint32_t port_id);

static struct drm_encoder *ljm_encoder_init(IN struct drm_device *ddev, struct ljm_port_info *info, struct ljm_port *port)
{
    struct drm_encoder *encoder;
    struct ljm_encoder *ljm_encoder;
    int id = info->encoder_info.index;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    HEADER_ARG("ddev=0x%p id=%d", ddev, id);

    /* Check whether we have already added it. */
    list_for_each_entry(encoder, &ddev->mode_config.encoder_list, head) {
        //ljm_encoder = to_ljm_encoder(encoder);
        //if (ljm_encoder->id == id) {
        if (encoder->index == id) {
            FOOTER_ARG("old encoder(0x%p)", encoder);
            return encoder;
        }
    }

    ljm_encoder = &port->base;
    if (!ljm_encoder) {
        DRM_ERROR("Failed to init encoder: out of memory.\n");
        return NULL;
    }

    encoder = &ljm_encoder->base;
    ljm_encoder->id = info->encoder_info.index;
    encoder->possible_crtcs = info->encoder_info.possible_crtcs;
    encoder->possible_clones = info->encoder_info.possible_clones;
    ljm_encoder->hard_type = info->encoder_info.hard_type;

    /* 由于DVO接口没有实际的encoder硬件, 无法通过encoder关闭显示输出.
     * 为了解决只有DVO接口时的关屏问题, 暂将DVO[n]绑定CRTC[n]
     */
    if (ljm_encoder->id >= 8 && ljm_encoder->id < 12) {
        encoder->possible_crtcs = 0x1 << (ljm_encoder->id - 8);
    }

    if(ljm_dev->once_edp)
    {
        if (info->is_edp) {
            /* lock current attached crtc for edp port */
            ljm_encoder->once_edp_crtc = gp200_port_get_crtc_id(ljm_dev->cur_card, ljm_encoder->id);
            encoder->possible_crtcs = 0x1 << ljm_encoder->once_edp_crtc;
            DEBUG_PRINT("Port%d is eDP with fixed crtc %d\n", ljm_encoder->id, ljm_encoder->once_edp_crtc);
            ljm_encoder->once_edp_modeset = !!ljm_dev->once_edp_vga_inited;
        }
        else {
            ljm_encoder->once_edp_crtc = -1;
        }
    }

    ljm_drm_encoder_init(ddev, encoder);
    drm_encoder_helper_add(encoder, &ljm_encoder_helper_funcs);

    DRM_DEBUG_KMS("new encode(%lx): index=%d base.id=%d id=%d\n", (unsigned long)encoder, encoder->index, encoder->base.id, ljm_encoder->id);

    FOOTER_ARG("new encoder%d=%p(0x%lx) %d", encoder->index, encoder, (unsigned long)encoder, encoder->base.id);
    return encoder;
}

extern uint ddcEnable;

/* Assign to ljm_connector_funcs.detect. Detect the monitor supported resolution */
static enum drm_connector_status
ljm_connector_do_detect(
    IN struct drm_connector *connector,
    IN bool force
    )
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);

    return ljm_connector_detect(ljm_dev, connector);
}

/* assign to ljm_connector_funcs.destroy */
void
ljm_connector_destroy(
    IN struct drm_connector *connector
    )
{
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);

    HEADER_ARG("Connector=0x%p", connector);

    if (ljm_connector->edid) {
        kfree(ljm_connector->edid);
        ljm_connector->edid = NULL;
    }

    drm_connector_unregister(connector);
    drm_connector_cleanup(connector);
    kfree(ljm_connector);    /* fix bug: kfree(Connector) */

    FOOTER();
}

static int ljm_connector_set_property(struct drm_connector *connector, struct drm_property *property,
						  uint64_t val)
{
    return 0;
}

/* drm_connector operation functions */
struct drm_connector_funcs ljm_connector_funcs = {
    .dpms = drm_helper_connector_dpms,
    .detect = ljm_connector_do_detect,
    .fill_modes = drm_helper_probe_single_connector_modes,
    .destroy = ljm_connector_destroy,
    .set_property = ljm_connector_set_property,
};

static int
ljm_connector_mode_valid(
    struct drm_connector *connector,
    struct drm_display_mode *mode
    )
{
    int r;
    //bool found = false;
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);

//  HEADER_ARG("connector=0x%p mode=0x%p", connector, mode);

//  TRACE("valid mode %dx%d@%d(pclk)", mode->hdisplay, mode->vdisplay, mode->clock);

    r = ljm_connector_mode_set_valid(ljm_dev, mode, ljm_connector);

//  FOOTER_ARG("%d", r);
    return r;
}


static struct drm_encoder *
ljm_connector_best_encoder(
    struct drm_connector *connector
    )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    struct drm_encoder *encoder;

    HEADER_ARG("connector%d=0x%lx", connector->index, (unsigned long)connector);

    drm_connector_for_each_possible_encoder(connector, encoder) {
        return encoder;
    }

    FOOTER();
    return NULL;
#else
    int enc_id = connector->encoder_ids[0];
    struct drm_encoder *encoder;

    HEADER_ARG("connector%d=0x%lx enc_id=%d", connector->index, (unsigned long)connector, enc_id);

    /* pick the encoder ids */
    if (enc_id) {
        encoder = DRM_ENCODER_FIND(connector->dev, NULL, enc_id);

        FOOTER_ARG("best encoder%d=0x%lx", encoder->index, (unsigned long)encoder);
        return encoder;
    }

    FOOTER();
    return NULL;
#endif
}

int ljm_add_modes_noedid(struct drm_connector *connector,
    int hdisplay, int vdisplay)
{
    static const struct drm_display_mode ljm_drm_modes[] = {
#if !USE_IN_SDT
        /* 0x00 - 3840x2160@60Hz*/
        { DRM_MODE("3840x2160", DRM_MODE_TYPE_DRIVER, 594000, 3840, 4016,
            4104, 4400, 0, 2160, 2168, 2178, 2250, 0,
            DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },//部分显示器在4K60时需要scdc支持，可能出现强推4K60但无显示的情况。
#endif
        /* 0x01 - 3840x2160@30Hz */
        { DRM_MODE("3840x2160", DRM_MODE_TYPE_DRIVER, 297000, 3840, 4016,
           4104, 4400, 0, 2160, 2168, 2178, 2250, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x02 - 1920x1080@60Hz */
        { DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008,
           2052, 2200, 0, 1080, 1084, 1089, 1125, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC) },
        /* 0x03 - 1600x1200@60Hz */
        { DRM_MODE("1600x1200", DRM_MODE_TYPE_DRIVER, 162000, 1600, 1664,
           1856, 2160, 0, 1200, 1201, 1204, 1250, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x04 - 1680x1050@60Hz */
        { DRM_MODE("1680x1050", DRM_MODE_TYPE_DRIVER, 146250, 1680, 1784,
           1960, 2240, 0, 1050, 1053, 1059, 1089, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x05 - 1440x900@60Hz */
        { DRM_MODE("1440x900", DRM_MODE_TYPE_DRIVER, 106500, 1440, 1520,
           1672, 1904, 0, 900, 903, 909, 934, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x06 - 1280x1024@60Hz */
        { DRM_MODE("1280x1024", DRM_MODE_TYPE_DRIVER, 108000, 1280, 1328,
           1440, 1688, 0, 1024, 1025, 1028, 1066, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x07 - 1280x1024@75Hz */
        { DRM_MODE("1280x1024", DRM_MODE_TYPE_DRIVER, 135000, 1280, 1296,
           1440, 1688, 0, 1024, 1025, 1028, 1066, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x08 - 1280x720@60Hz */
        { DRM_MODE("1280x720", DRM_MODE_TYPE_DRIVER, 74250, 1280, 1390,
           1430, 1650, 0, 720, 725, 730, 750, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x09 - 1024x768@60Hz */
        { DRM_MODE("1024x768", DRM_MODE_TYPE_DRIVER, 65000, 1024, 1048,
           1184, 1344, 0, 768, 771, 777, 806, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC) },
         /* 0x51 - 1366x768@60Hz */
        { DRM_MODE("1366x768", DRM_MODE_TYPE_DRIVER, 85500, 1366, 1436,
           1579, 1792, 0, 768, 771, 774, 798, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC) },
        /* 0x0a - 1024x768@75Hz */
        { DRM_MODE("1024x768", DRM_MODE_TYPE_DRIVER, 78750, 1024, 1040,
           1136, 1312, 0, 768, 769, 772, 800, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x0a - 800x600@60Hz */
        { DRM_MODE("800x600", DRM_MODE_TYPE_DRIVER, 40000, 800, 840,
           968, 1056, 0, 600, 601, 605, 628, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC) },
        /* 0x03 - 720x400@85Hz */
        { DRM_MODE("720x400", DRM_MODE_TYPE_DRIVER, 35500, 720, 756,
           828, 936, 0, 400, 401, 404, 446, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_PVSYNC) },
    };

    struct drm_display_mode *mode;
    int i, count, num_modes = 0;
    struct drm_device *dev = connector->dev;


    HEADER();

    count = sizeof(ljm_drm_modes) / sizeof(struct drm_display_mode);
    if (hdisplay < 0)
        hdisplay = 0;
    if (vdisplay < 0)
        vdisplay = 0;

    for (i = 0; i < count; i++) {
        const struct drm_display_mode *ptr = &ljm_drm_modes[i];
        if (hdisplay && vdisplay) {
            /*
             * Only when two are valid, they will be used to check
             * whether the mode should be added to the mode list of
             * the connector.
             */
            if (ptr->hdisplay > hdisplay || ptr->vdisplay > vdisplay)
                continue;
        }
        if (drm_mode_vrefresh(ptr) > 120)
            continue;
        mode = drm_mode_duplicate(dev, ptr);
        if (mode) {
            drm_mode_probed_add(connector, mode);
            num_modes++;
        }
    }

    FOOTER();
    return num_modes;
}

extern uint hDisplay;
extern uint vDisplay;

/* Get optimal resolution through DDC.
 * Assign to ljm_connector_helper_funcs.get_modes.
 */
static int
ljm_connector_get_modes(
    struct drm_connector *connector
    )
{
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    struct ljm_port *ljm_port = to_ljm_port(ljm_connector->encoder);
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);
    int ret = 0;
    uint w, h;

    HEADER_ARG("connector=%p", connector);

    DRM_DEBUG_KMS("connector%d=0x%lx id=%d\n", connector->index, (unsigned long)connector, ljm_connector->id);

    if (hDisplay == 0) w = ljm_dev->disp_attr.default_hdisplay;
    else w = hDisplay;

    if (vDisplay == 0) h = ljm_dev->disp_attr.default_vdisplay;
    else h = vDisplay;

    if (ddcEnable == 0) {
        DRM_INFO("Add %d modes (%dx%d) for noedid monitor\n", ret, w, h);
        goto no_edid;
    }

    if (ljm_connector->edid_dirty) {
        if (ljm_connector->edid) {
            kfree(ljm_connector->edid);
            ljm_connector->edid = NULL;
        }
        ljm_connector->edid_dirty = false;
    }

    if (ljm_connector->edid == NULL)
        ljm_connector->edid = ljm_connector_get_edid(ljm_dev, connector, NULL);

    if (ljm_connector->edid) {
        ret = drm_add_edid_modes(connector, ljm_connector->edid);
        if(ret <=0 )
            goto no_edid;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0x13, 0x43)
        drm_connector_update_edid_property(connector, ljm_connector->edid);
#else
        drm_mode_connector_update_edid_property(connector, ljm_connector->edid);
#endif
    }
    else {
        DEBUG_PRINT("EDID is bad, use noedid mode.\n");
        goto no_edid;
    }

    FOOTER();
    return ret;

no_edid:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0x13, 0x43)
    drm_connector_update_edid_property(connector, NULL);
#else
    drm_mode_connector_update_edid_property(connector, NULL);
#endif
    ret = ljm_add_modes_noedid(connector, ljm_dev->disp_attr.max_hdisplay, ljm_dev->disp_attr.max_vdisplay);      //default hw init
    drm_set_preferred_mode(connector, w, h);
    if (ljm_connector->ddc_type == LJM_DDC_TYPE_HDMIV1 || ljm_connector->ddc_type == LJM_DDC_TYPE_HDMIV2)
    {
        ljm_port->vdev.hdmi.edid.sink_hdmi = 1;    /* default in hdmi mode */
        ljm_port->vdev.hdmi.edid.vendor = EDID_VENDOR_DEFAULT;
    }

    FOOTER();
    return ret;
}

/* drm_connector legacy helper functions */
struct drm_connector_helper_funcs ljm_connector_helper_funcs = {
    .get_modes = ljm_connector_get_modes,
    .mode_valid = ljm_connector_mode_valid,
    .best_encoder = ljm_connector_best_encoder,
};

bool ljm_edp_init_connector(struct ljm_port *ljm_port);

static struct drm_connector *ljm_connector_init(IN struct drm_device *ddev, struct ljm_port_info *info, struct ljm_port *port)
{
    struct drm_connector *connector;
    struct ljm_connector *ljm_connector;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct ljm_connector_info *connector_info;
    int id = info->connector_info.index;
#if USE_DP_CONNECTOR
    bool backlight_dev = false;
#endif
    HEADER_ARG("ddev=0x%p id=%d", ddev, id);

    /* Check whether we have already added it. */
    list_for_each_entry(connector, &ddev->mode_config.connector_list, head) {
        //ljm_connector = to_ljm_connector(connector);
        //if (ljm_connector->id == id) {
        if (connector->index == id) {
            //ljm_connector->devices |= supported_device;//just ignore it now

            FOOTER_ARG("old connector(0x%p)", connector);
            return connector;
        }
    }

    connector_info = &info->connector_info;

    ljm_connector = kzalloc(sizeof(struct ljm_connector), GFP_KERNEL);
    if (!ljm_connector) {
        DRM_ERROR("Failed to init connector: Out of memory.\n");
        FOOTER_ARG("ERR: %d", -ENOMEM);
        return NULL;
    }

    ljm_connector->id = connector_info->index;
    connector = &ljm_connector->base;
    connector->polled = connector_info->polled;
    ljm_connector->ddc_type = connector_info->ddc_type;
    ljm_connector->hpd_type = connector_info->hpd_type;
    port->vdev_connector = ljm_connector;

    drm_connector_init(ddev, connector, &ljm_connector_funcs, connector_info->connector_type);
    switch (ljm_dev->family) {
        case CHIP_GP100:
            /* HDMI A: 最广泛的HDMI线缆，19引脚；
             * HDMI B: 较少见，29引脚，相当于双链路；
             * HDMI C: Mini HDMI，19引脚，主要用于便携设备中；
             * HDMI D: Micro HDMI，19引脚，最新更小的尺寸；
             * HDMI E: 车载娱乐系统，具备抗震性、防潮、耐高强度、耐温差大
             */
            if (id == 0) { /* Connector 0: HDMI 0 */
                sema_init(&ljm_connector->edid_sema, 0);

            }
            else if (id == 1) { /* Connector 1: HDMI 1 */
                sema_init(&ljm_connector->edid_sema, 0);

            }
            break;
        case CHIP_GP200:
        case CHIP_GP300:
#if USE_DP_CONNECTOR
            if (port->info->type == LJM_PORT_DP) {  /* DP */
                            /* HDP polling for DP */
                if(!ljm_dev->dp_irq_is_extra && ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201))
                {
                    connector->polled = DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT;
                }
                if (port->vdev.dp.is_edp) {
                    ljm_edp_init_connector(port);
                    backlight_dev = true;
                }

                ljm_dp_add_properties(&port->vdev.dp, connector);
            }
#endif
            break;
        default:
            printk(KERN_ERR "FIXME: TODO in ljm_connector_init\n");
            break;
    }

    ljm_connector->hpd.hpd = (enum ljm_hpd_id)ljm_connector->id;
    ljm_connector->hpd.plugged_state = -1;

    drm_connector_helper_add(connector, &ljm_connector_helper_funcs);

    drm_connector_register(connector);
#if USE_DP_CONNECTOR
    if (backlight_dev)
        ljm_backlight_dev_init(&port->vdev.dp, connector);
#endif
    DRM_DEBUG_KMS("new connector(%lx): index=%d base.id=%d id=%d\n",
        (unsigned long)connector, connector->index, connector->base.id, ljm_connector->id);

    FOOTER_ARG("new connector%d=%p(%lx) %d", connector->index, connector, (unsigned long)connector, connector->base.id);
    return connector;
}

inline void ljm_connector_attach_encoder(struct drm_connector *connector, struct drm_encoder *encoder)
{
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    struct ljm_encoder *ljm_encoder = to_ljm_encoder(encoder);
    struct ljm_port *ljm_port = to_ljm_port(ljm_encoder);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0x13, 0x53)
        drm_connector_attach_encoder(connector, encoder);
#else
        drm_mode_connector_attach_encoder(connector, encoder);
#endif

    ljm_connector->encoder = ljm_encoder;
    ljm_port->vdev_connector = ljm_connector;

    return;
}

int ljm_setup_enc_conn(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct ljm_drm_info *info = ljm_dev->drm_info;
    struct drm_encoder *encoder;
    struct drm_connector *connector;
    int i, j;

    HEADER_ARG("ddev=%p ljm_dev->ports=%d", ddev, ljm_dev->num_ports);

    for(i=0; i<ljm_dev->num_ports; i++) {
        j = 0;
        do
        {
            if(ljm_dev->ports[i]->id == info->ports[j].index && info->ports[j].type != LJM_PORT_CLOSE)
                break;
        }while(++j < LJM_MAX_PORTS);

        if(j == LJM_MAX_PORTS)
        {
            ERR_PRINT("[PORT %d]Can't find the proper port_info.\n", ljm_dev->ports[i]->id);
            continue;
        }

        encoder = ljm_encoder_init(ddev, &info->ports[j], ljm_dev->ports[i]);
        if (!encoder) {
            ERR_PRINT("%s: Failed to init encoder\n", __FUNCTION__);
            FOOTER_ARG("%d", -ENOMEM);
            return -ENOMEM;
        }

        connector = ljm_connector_init(ddev, &info->ports[j], ljm_dev->ports[i]);
        if (!connector) {
            ljm_encoder_destroy(encoder);

            ERR_PRINT("%s: Failed to init connector\n", __FUNCTION__);
            FOOTER_ARG("%d", -ENOMEM);
            return -ENOMEM;
        }

        ljm_connector_attach_encoder(connector, encoder);
    }

    FOOTER();
    return 0;
}

void ljm_hdmiv2_scdc_work_stop(struct hdmi_dev *hdmi_dev);

void ljm_connector_hotplug(struct drm_connector *connector)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(connector->dev);
    struct ljm_connector *ljm_connector = to_ljm_connector(connector);
    struct ljm_port *ljm_port;

    //HEADER_ARG("connector%d=%lx", connector->index, (unsigned long)connector);

    ljm_port = to_ljm_port(ljm_connector->encoder);
    /* bail if the connector does not have hpd pin, e.g.,
     * VGA, TV, etc.
     */
    if (ljm_connector->hpd.hpd == LJM_HPD_NONE || ljm_connector->hpd_type == LJM_HPD_TYPE_DAC) {
        //FOOTER_ARG("connector%d does not support HPD", connector->index);
        return;
    }

    ljm_hpd_set_polarity(ljm_dev, ljm_connector);

    /* if the connector is already off, don't turn it back on */
    if (connector->dpms != DRM_MODE_DPMS_ON) {
        //FOOTER_ARG("connector%d is off", connector->index);
        ljm_connector->edid_dirty = true;
    }

    if (ljm_hpd_sense(ljm_dev, ljm_connector)) {
        if (ljm_connector->hpd.plugged_state != LJM_HPD_STATE_IN) {
            DRM_INFO("%s is plugged in\n", connector->name);
            ljm_connector->hpd.plugged_state = LJM_HPD_STATE_IN;
        }
    } else {
        if (ljm_connector->hpd.plugged_state != LJM_HPD_STATE_OUT) {
            DRM_INFO("%s is plugged out\n", connector->name);
            ljm_connector->hpd.plugged_state = LJM_HPD_STATE_OUT;
            ljm_connector->edid_dirty = true;
        }
    }
/* 如果同时接hdmi和dp，hdmi的热插拔会影响dp */
#if 0
    /* FIXME: deal with DP */
    /* just deal with DP (not eDP) here. */

    if (ljm_connector->hpd_type == LJM_HPD_TYPE_DP && ljm_connector->base.connector_type != DRM_MODE_CONNECTOR_eDP)
    {
        if (ljm_dp_hpd_sense(&ljm_port->vdev.dp))
        {
            ljm_dptx_mode_set(&ljm_port->vdev.dp, &ljm_port->vdev.dp.mode);
        }
    }
#endif

    if(ljm_port->type == LJM_PORT_HDMI &&  ljm_connector->hpd.plugged_state == LJM_HPD_STATE_OUT)
    {
        /*判断拔出前hdmi是否使用scdc，若有使用，则下次接入时仍然需要配置。*/
        if (ljm_port->vdev.hdmi.edid.scdc_present == 1) {
            if (ljm_port->vdev.hdmi.tmdsclk > 340000000 ||ljm_port->vdev.hdmi.edid.lte_340mcsc_scramble)
                ljm_port->vdev.hdmi.need_scdc = true;
        }
        ljm_hdmiv2_scdc_work_stop(&ljm_port->vdev.hdmi);
    }
    //FOOTER();
}

