/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_plane.h"

#include "lj_hal_drm.h"
#include "ljmicro_device.h"
#include "ljmicro_vgav2.h"

#define LJM_PLANE_REG_R8(p, addr)       REG_R8((p)->cur_card, (addr) + LAYER_REG_SIZE * ((p)->id))
#define LJM_PLANE_REG_W8(p, addr, val)  REG_W8((p)->cur_card, (addr) + LAYER_REG_SIZE * ((p)->id), (val))
#define LJM_PLANE_REG_R16(p, addr)      REG_R16((p)->cur_card, (addr) + LAYER_REG_SIZE * ((p)->id))
#define LJM_PLANE_REG_W16(p, addr, val) REG_W16((p)->cur_card, (addr) + LAYER_REG_SIZE * ((p)->id), (val))
#define LJM_PLANE_REG_R32(p, addr)      REG_R32((p)->cur_card, (addr) + LAYER_REG_SIZE * ((p)->id))
#define LJM_PLANE_REG_W32(p, addr, val) REG_W32((p)->cur_card, (addr) + LAYER_REG_SIZE * ((p)->id), (val))

/*图层堆叠顺序的结构体*/
static struct ljm_plane *plane_stack[LJM_MAX_LAYERS] = {NULL};
static void ljm_plane_stack_update(void);

/**
 * ljm_plane_alloc - 创建一个ljm_plane实例
 *
 * 为ljm_plane和ljm_plane_state分配内存，ljm_plane_state的指针由
 * ljm_plane->state成员保存。该函数不保存ljm_plane的指针，只把指针
 * 做函数返回值传出。
 *
 * Returns:
 * 成功，则返回ljm_plane指针；失败，则返回NULL。
 */
struct ljm_plane *ljm_plane_alloc(void)
{
    struct ljm_plane *      ljm_plane;
    struct ljm_plane_state *state;

    ljm_plane = kmalloc(sizeof(struct ljm_plane), GFP_KERNEL);
    if (!ljm_plane)
        goto plane_err;

    state = kzalloc(sizeof(struct ljm_plane_state), GFP_KERNEL);
    if (!state)
        goto state_err;

    ljm_plane->active = false;
    state->visible    = false;
    ljm_plane->state  = state;
    return ljm_plane;

state_err:
    kfree(ljm_plane);
plane_err:
    return NULL;
}

/**
 * ljm_plane_init - ljm_plane初始化
 * @ljm_dev: 驱动设备指针
 *
 * 在驱动加载时执行，初始化若干各ljm_plane实例，用于驱动管理图层以及
 * 与drm_plane交互。初始化后的ljm_plane指针保存在struct ljm_device的
 * planes成员中。
 *
 * Returns:
 * 成功，则返回0；失败，则返回一个负值。
 */
int ljm_plane_init(struct ljm_device *ljm_dev)
{
    int               i;
    struct ljm_plane *ljm_plane;

    for (i = 0; i < ljm_dev->num_planes; i++)
    {
        ljm_plane = ljm_plane_alloc();
        if (!ljm_plane)
            return -ENOMEM;

        //ljm_plane->ljm_dev  = ljm_dev;
        ljm_plane->id       = i;
        ljm_plane->cur_card = ljm_dev->cur_card;
        sprintf(ljm_plane->name, "layer-%d", i);
        ljm_dev->planes[i] = ljm_plane;
        LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, 0x10);   /**< enable layer, one factor alpha blending mode */
    }

    return 0;
}

int ljm_plane_fini(struct ljm_device *ljm_dev)
{
    // todo:卸载驱动时调用，释放内存，注销接口调用。
    return 0;
}

static int ljm_plane_visible(struct ljm_plane *ljm_plane, bool visible)
{
    uint16_t val;

    val = LJM_PLANE_REG_R16(ljm_plane, LAYER1_CTRL_ADDR);
    if (visible)
        val |= LAYER_CTRL_EN_MASK;
    else
        val &= ~LAYER_CTRL_EN_MASK;
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, val);

    ljm_plane->state->visible = visible;
    ljm_plane_stack_update();
    return 0;
}
#if 0
static void ljm_plane_set_crtc(struct ljm_plane *ljm_plane, struct ljm_crtc *ljm_crtc)
{
    uint16_t val;

    val = LJM_PLANE_REG_R16(ljm_plane, LAYER1_CTRL_ADDR);
    val &= 0xf3ff;
    val |= (ljm_crtc->id << 10);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, val);
    ljm_plane->state->crtc = ljm_crtc;

    return;
}
#endif
/**
 * 根据 port 值对应接口类型和 10024 寄存器的值，将layer, crtc 与接口进行映射
*/
static int ljm_plane_find_crtc_with_port(struct ljm_device *ljm_dev, struct ljm_plane *ljm_plane, uint32_t port)
{
    uint16_t val;
    uint32_t crtc_mux_old, crtc_mux;

    crtc_mux_old = REG_R32(ljm_dev->cur_card, HDMI_DAC_DVO_OUT_MUX);

    val = LJM_PLANE_REG_R16(ljm_plane, LAYER1_CTRL_ADDR);
    val &= 0xf3ff;

    switch (port)
    {
        //todo:不同的接口有可能映射到同一 CRTC 上，需要对接口和CRTC使用的状态进行一个详细区分
        case 0: /* VGA0 */
            crtc_mux = (crtc_mux_old & 0x3);

            break;
        case 1: /* VGA1 */
            crtc_mux_old = ((crtc_mux_old & 0x300) >> 8);
            if (crtc_mux_old == 0x0) /* crtc1 */
                crtc_mux = 0x1; /* 修改为 0x1，以写入 0x10100，与 layer1 建立映射 */
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 2: /* VGA2 */
            crtc_mux_old = ((crtc_mux_old & 0x30000) >> 16);
            if (crtc_mux_old == 0x0) /* crtc2 */
                crtc_mux = 0x2;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 3: /* VGA3 */
            crtc_mux_old = ((crtc_mux_old & 0x3000000) >> 24);
            if (crtc_mux_old == 0x0) /* crtc3 */
                crtc_mux = 0x3;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = 0x2;

            break;
        case 4: /* HDMI0 */
            crtc_mux = ((crtc_mux_old & 0xC) >> 2);

            break;
        case 5: /* HDMI1 */
            crtc_mux_old = ((crtc_mux_old & 0xC00) >> 10);
            if (crtc_mux_old == 0x0) /* crtc1 */
                crtc_mux = 0x1;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 6: /* HDMI2 */
            crtc_mux_old = ((crtc_mux_old & 0xC0000) >> 18);
            if (crtc_mux_old == 0x0) /* crtc2 */
                crtc_mux = 0x2;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 7: /* HDMI3 */
            crtc_mux_old = ((crtc_mux_old & 0xC000000) >> 26);
            if (crtc_mux_old == 0x0) /* crtc3 */
                crtc_mux = 0x3;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = 0x2;

            break;
        case 8: /* DVO0 */
            crtc_mux = ((crtc_mux_old & 0x30) >> 4);

            break;
        case 9: /* DVO1 */
            crtc_mux_old = ((crtc_mux_old & 0x3000) >> 12);
            if (crtc_mux_old == 0x0) /* crtc1 */
                crtc_mux = 0x1;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 10: /* DVO2 */
            crtc_mux_old = ((crtc_mux_old & 0x300000) >> 20);
            if (crtc_mux_old == 0x0) /* crtc2 */
                crtc_mux = 0x2;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 11: /* DVO3 */
            crtc_mux_old = ((crtc_mux_old & 0x30000000) >> 28);
            if (crtc_mux_old == 0x0) /* crtc3 */
                crtc_mux = 0x3;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = 0x2;

            break;
        case 12: /* DP0 */
            crtc_mux = ((crtc_mux_old & 0xC0) >> 6);

            break;
        case 13: /* DP1 */
            crtc_mux_old = ((crtc_mux_old & 0xC000) >> 14);
            if (crtc_mux_old == 0x0) /* crtc1 */
                crtc_mux = 0x1;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 14: /* DP2 */
            crtc_mux_old = ((crtc_mux_old & 0xC00000) >> 22);
            if (crtc_mux_old == 0x0) /* crtc2 */
                crtc_mux = 0x2;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = crtc_mux_old;

            break;
        case 15: /* DP3 */
            crtc_mux_old = ((crtc_mux_old & 0xC0000000) >> 30);
            if (crtc_mux_old == 0x0) /* crtc3 */
                crtc_mux = 0x3;
            else if (crtc_mux_old == 0x1) /* crtc0 */
                crtc_mux = 0x0;
            else if (crtc_mux_old == 0x2) /* crtc1 */
                crtc_mux = 0x1;
            else
                crtc_mux = 0x2;

            break;
        default:
            return -EINVAL;
    }

    val |= (crtc_mux << 10 );
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, val);

    return 0;
}

/**
 * ljm_plane_set_crtc_pos - 设置图层窗口的crtc坐标
 * @ljm_plane: plane指针
 * @x: 图层起始点在当前crtc上的位置横坐标
 * @y: 图层起始点在当前crtc上的位置纵坐标
 *
 * Return:
 */
static void ljm_plane_set_crtc_pos(struct ljm_plane *ljm_plane, int x, int y)
{
    struct ljm_plane_state *state = ljm_plane->state;

    /* 实际应用中发现， crtc->x/crtc->y并不能准确记录当前crtc在整幅图像上的位置。
     * 图层切换crtc的操作改由专门的ljm_plane_find_crtc_with_port()函数实现。
    */
    state->crtc_x = x;
    state->crtc_y = y;

    LJM_PLANE_REG_W16(ljm_plane, LAYER1_HPOSITION_ADDR, state->crtc_x);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_VPOSITION_ADDR, state->crtc_y);

    return;
}

static void ljm_plane_set_size(struct ljm_plane *ljm_plane, uint32_t w, uint32_t h)
{
    struct ljm_plane_state *state = ljm_plane->state;

    state->crtc_w = w;
    state->crtc_h = h;
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_WIDTH_ADDR, state->crtc_w);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_HEIGHT_ADDR, state->crtc_h);

    return;
}

static void ljm_plane_transparent_enable(struct ljm_plane *ljm_plane, bool enable)
{
    uint16_t val;
    val = LJM_PLANE_REG_R16(ljm_plane, LAYER1_CTRL_ADDR);
    val &= 0xfffd;
    val |= ((uint16_t)enable) << 1;
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, val);

    return;
}

/**
 * ljm_plane_set_pitches - 设置图层显示的每行偏移值
 * @ljm_plane: 图层指针
 * @pitches: 一行所占的64位地址（8字节对齐）
 *
 * 假如色深为24（即1个像素4个字节），则
 * pitches = 行的像素宽度 * 4 / 8
 *
 */
static int ljm_plane_set_pitches(struct ljm_plane *ljm_plane, uint32_t pitches)
{
    uint32_t val;
    struct ljm_plane_state *state = ljm_plane->state;

    state->pitches = pitches;
    /**
     * LAYER1_OFFSET_ADDR寄存器定义一行所占的64位地址，256位数据地址对齐。
    */
    val = pitches * 8 / 64;
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_OFFSET_ADDR, (uint16_t)val);
    return 0;
}

/* drm_crtc->x和drm_crtc->y不能准确反映crtc的位置信息。
 * 删除根据绝对坐标自动切换crtc的功能。
 * 日期：2024年12月18日
 * 作者：zhaiyf
*/
#if 0
/**
 * ljm_plane_find_crtc_with_pos - 根据图层位置的绝对坐标确定crtc
 * @ljm_dev: 驱动设备，用于遍历crtc
 * @x: 图层在显存图像上的绝对横坐标
 * @y: 图层在显存图像上的绝对纵坐标
 *
 * 参考ljm_plane_set_crtc_pos()函数，需要根据绝对坐标确定图层需要被映射到
 * 哪一个crtc上去。
 *
 * Return:
 * 成功，则返回需要映射的ljm_crtc指针；失败，则返回NULL。
 */
struct ljm_crtc *ljm_plane_find_crtc_with_pos(struct ljm_device *ljm_dev, uint32_t x, uint32_t y)
{
    int              i;
    int              x_start, y_start, x_end, y_end;
    struct ljm_crtc *ljm_crtc = NULL;

    for (i = 0; i < ljm_dev->num_crtcs; i++)
    {
        x_start = ljm_dev->crtcs[i]->x;
        x_end   = x_start + ljm_dev->crtcs[i]->mode.hdisplay;
        y_start = ljm_dev->crtcs[i]->y;
        y_end   = y_start + ljm_dev->crtcs[i]->mode.vdisplay;
        if ((x_start <= x && x < x_end) && (y_start <= y && y < y_end))
        {
            ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[i]);
            break;
        }
    }

    return ljm_crtc;
}

/**
 * ljm_plane_need_crtc_update - 根据绝对坐标判断是否需要更新crtc映射
 * @ljm_dev: 驱动设备，用于遍历crtc
 * @x: 图层在显存图像上的绝对横坐标
 * @y: 图层在显存图像上的绝对纵坐标
 *
 * 当图层位置变化时，使用该函数检查图层是否仍位于旧的crtc上。
 *
 * Return:
 * true - 图层不位于旧的crtc上，需要更新crtc映射
 * false - 图层仍位于旧crtc上，不需要更新crtc映射
 */
bool ljm_plane_need_crtc_update(struct ljm_plane *ljm_plane, uint32_t x, uint32_t y)
{
    struct ljm_plane_state *state = ljm_plane->state;
    struct drm_crtc *       crtc;

    /**
     * state->crtc变量在ljm_plane_set_crtc()中被设置。
     * 若ljm_plane是新的，则此值为空，需要设置crtc映射。
     */
    if (!state->crtc)
        return true;

    crtc = &state->crtc->base;

    if ((x >= crtc->x + state->crtc_w) || (x < crtc->x) || (y >= crtc->y + state->crtc_h) || (y < crtc->y))
        return true;

    return false;
}
#endif

static int ljm_plane_update_pos(struct ljm_device *ljm_dev, struct ljm_plane *ljm_plane, uint32_t x, uint32_t y)
{
#if 0
    struct ljm_crtc *ljm_crtc;

    if (!ljm_plane_need_crtc_update(ljm_plane, x, y))
        goto update_pos;

    ljm_crtc = ljm_plane_find_crtc_with_pos(ljm_dev, x, y);
    if (!ljm_crtc)
    {
        ERR_PRINT("[%s]Failed to find proper crtc.\n", ljm_plane->name);
        return -ENXIO;
    }

    ljm_plane_set_crtc(ljm_plane, ljm_crtc);

update_pos:
#endif
    ljm_plane_set_crtc_pos(ljm_plane, x, y);
    return 0;
}

static int ljm_plane_set_bpp(struct ljm_plane *ljm_plane, unsigned char bpp)
{
    uint8_t val;
    struct ljm_plane_state *state = ljm_plane->state;

    val = LJM_PLANE_REG_R8(ljm_plane, LAYER1_COLOR_MODE_ADDR);
    switch (bpp)
    {
        case 15:
            val = (val & 0xf8) | 0x1;
            break;
        case 16:
            val = (val & 0xf8) | 0x2;
            break;
        case 24:
        case 32:
            val = (val & 0xf8) | 0x4;
            break;
        default:
            return -EINVAL;
    }
    state->bpp = bpp;
    LJM_PLANE_REG_W8(ljm_plane, LAYER1_COLOR_MODE_ADDR, val);
    return 0;
}

static int ljm_plane_set_zpos(struct ljm_plane *ljm_plane, unsigned int hwId)
{
    uint16_t val;

    if(!(hwId >= 0 && hwId < LJM_MAX_LAYERS))
        return 0;

    val = LJM_PLANE_REG_R16(ljm_plane, LAYER1_CTRL_ADDR);
    val &= 0xfcff;
    switch(ljm_plane->id)
    {
        case 0:
            val |= hwId << 8;
            break;
        case 1:
            if(hwId == 0)
                val |= 1 << 8;
            else if(hwId == 1)
                val |= 0 << 8;
            else
                val |= hwId << 8;
            break;
        case 2:
            if(hwId == 0)
                val |= 1 << 8;
            else if(hwId == 1)
                val |= 2 << 8;
            else if(hwId == 2)
                val |= 0 << 8;
            else if(hwId == 3)
                val |= 3 << 8;
            break;
        case 3:
            if(hwId == 0)
                val |= 1 << 8;
            else if(hwId == 1)
                val |= 2 << 8;
            else if(hwId == 2)
                val |= 3 << 8;
            else if(hwId == 3)
                val |= 0 << 8;
            break;
    }
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, val);

    return 0;
}

/**
 * ljm_plane_stack_update - 更新图层优先级数组。
 *
 * 被激活的图层按id排序，未被激活的图层为NULL。根据plane_stack的顺序，
 * 依次设置图层优先级。
*/
static void ljm_plane_stack_update(void)
{
    struct ljm_plane *pa[LJM_MAX_LAYERS] = {NULL};
    int i=0, j=0;

    /* 按id正序排列所有plane.*/
    for(i=0; i<LJM_MAX_LAYERS; i++)
    {
        if(plane_stack[i])
            pa[plane_stack[i]->id] = plane_stack[i];
    }

    /* 按id正序依次配置各图层，详见硬件图层寄存器定义*/
    for(i=0; i<LJM_MAX_LAYERS; i++)
    {
        if(plane_stack[i] == NULL || !plane_stack[i]->state->visible)
            continue;

        while(pa[j] == NULL || (pa[j] && (!pa[j]->state->visible)))
            j++;

        ljm_plane_set_zpos(pa[j], plane_stack[i]->id);
        j++;
    }

    return;
}

/**
 * ljm_plane_stack_move - 在堆叠数组中移动plane
 * @plane：硬图层设备指针；
 * @zpos：图层要移动的位置，有效数值为1~4。
 *
 * 若需移动的plane不在stack中，则直接添加在指定位置。若此位置已有plane，则依次将冲突plane向前
 * 或向后移动，空出位置。若需移动的plane已在stack，则先将其从stack删除，再添加至新的位置。
 *
 * 返回值：
 * 成功返回0，失败返回负数。
*/
static int ljm_plane_stack_move(struct ljm_plane *plane, uint32_t zpos)
{
    uint32_t new_index, old_index = 0;
    int i;
    struct ljm_plane *p, *t;

    if(zpos < 1 || zpos > 4)
        return -EINVAL;

    new_index = zpos - 1;
    /*在堆叠顺序数组中查找需要移动的plane。*/
    for(i=0; i<LJM_MAX_LAYERS; i++)
    {
        if(plane_stack[i] == NULL)
            continue;

        if(plane->id == plane_stack[i]->id)
        {
            old_index = i; //记录图层原有位置。
            plane_stack[i] = NULL; //从数组中删除该图层。
        }
    }

    if(new_index > old_index)
    {
        i = new_index;
        p = plane;
        while(plane_stack[i] != NULL)
        {
            t = plane_stack[i];
            plane_stack[i] = p;
            p = t;
            i--;
        }
        plane_stack[i] = p;
    }
    else if(new_index < old_index)
    {
        i = new_index;
        p = plane;
        while(plane_stack[i] != NULL)
        {
            t = plane_stack[i];
            plane_stack[i] = p;
            p = t;
            i++;
        }
        plane_stack[i] = p;
    }

    ljm_plane_stack_update();

    return 0;
}

static int ljm_plane_stack_add(struct ljm_plane *plane)
{
    plane->state->visible = false;
    return ljm_plane_stack_move(plane, 4);
}

static void ljm_plane_stack_del(struct ljm_plane *plane)
{
    int i;
    for(i=0; i<LJM_MAX_LAYERS; i++)
    {
        if(plane_stack[i] == NULL)
            continue;
        if(plane->id == plane_stack[i]->id)
            plane_stack[i] = NULL;
    }

    ljm_plane_stack_update();
    return;
}

static void ljm_plane_hw_clear(struct ljm_plane *ljm_plane)
{
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_CTRL_ADDR, 0);
    LJM_PLANE_REG_W8(ljm_plane, LAYER1_COLOR_MODE_ADDR, 0);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_HPOSITION_ADDR, 0);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_VPOSITION_ADDR, 0);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_WIDTH_ADDR, 0);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_HEIGHT_ADDR, 0);
    LJM_PLANE_REG_W16(ljm_plane, LAYER1_OFFSET_ADDR, 0);
    LJM_PLANE_REG_W8(ljm_plane, LAYER1_ALPHA_ADDR, 0);
    LJM_PLANE_REG_W32(ljm_plane, LAYER1_TRANSPARENT_ADDR, 0);
    LJM_PLANE_REG_W32(ljm_plane, LAYER1_BUF0_START_ADDR, 0);
    LJM_PLANE_REG_W32(ljm_plane, LAYER1_BUF1_START_ADDR, 0);
    return;
}

bool ljm_plane_idle(struct ljm_plane *ljm_plane)
{
    return !ljm_plane->active;
}

/**
 * ljm_plane_obtain - 获取一个空闲的ljm_plane
 * @ljm_dev:驱动设备指针，用于遍历所有ljm_plane
 *
 * 获取一个未被使用的ljm_plane实例。这个函数不会更改ljm_plane的使用状态，
 * 它只返回一个空闲的ljm_plane，若决定使用该ljm_plane，请将成员active更改为true.
 *
 * Return:
 * 成功，则返回空闲的ljm_plane指针；失败，则返回BULL。
 */
struct ljm_plane *ljm_plane_obtain(struct ljm_device *ljm_dev)
{
    int i;
    for (i = 0; i < ljm_dev->num_planes; i++)
        if (ljm_plane_idle(ljm_dev->planes[i]))
            return ljm_dev->planes[i];

    return NULL;
}

struct ljm_plane *ljm_plane_find(struct ljm_device *ljm_dev, int plane_id)
{
    int i;

    for (i = 0; i < ljm_dev->num_planes; i++)
    {
        if (ljm_dev->planes[i]->id == plane_id)
            return ljm_dev->planes[i];
    }

    return NULL;
}

/**
 * ioctl funcs for ljm_plane.
 */
int ljm_ioctl_plane_request(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(dev);
    struct ljm_plane * ljm_plane;
    int *              arg = (int *)data;

    ljm_plane = ljm_plane_obtain(ljm_dev);
    if (!ljm_plane)
    {
        ERR_PRINT("Failed to request hardware layers.\n");
        return -EBUSY;
    }

    *arg              = ljm_plane->id;
    ljm_plane->active = true;
    ljm_plane_set_bpp(ljm_plane, 24);                     // 默认24位色深
    LJM_PLANE_REG_W8(ljm_plane, LAYER1_ALPHA_ADDR, 0xff); // 默认alpha透明度为255
    LJM_PLANE_REG_W8(ljm_plane, LAYER1_REG_OFFSET, 0x10); // 整个平面采用同一个alpha factor
    ljm_plane_stack_add(ljm_plane);
    return 0;
}

int ljm_ioctl_plane_free(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    int *              plane_id = (int *)data;
    struct ljm_device *ljm_dev  = GetLjmDevFrmDrm(dev);
    int                i;
    struct ljm_plane * plane;

    for (i = 0; i < ljm_dev->num_planes; i++)
    {
        if (ljm_dev->planes[i]->id == *plane_id)
        {
            plane         = ljm_dev->planes[i];
            plane->active = false;
            ljm_plane_visible(plane, false);
            ljm_plane_stack_del(plane);
            ljm_plane_hw_clear(plane);
            return 0;
        }
    }
    return -ENXIO;
}

int ljm_ioctl_plane_property_set(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    struct ljm_device *        ljm_dev = GetLjmDevFrmDrm(dev);
    struct ljm_plane_property *request = (struct ljm_plane_property *)data;
    struct ljm_plane *         ljm_plane;
    uint64_t *                 values;

    ljm_plane = ljm_plane_find(ljm_dev, request->plane_id);
    if (!ljm_plane)
    {
        ERR_PRINT("Invalid ljm_plane id for setting.\n");
        return -EINVAL;
    }

    values = kzalloc(sizeof(uint64_t) * request->num_values, GFP_KERNEL);
    if (copy_from_user(values, request->values, sizeof(uint64_t) * request->num_values))
        return -EFAULT;

    switch (request->name)
    {
        case PLANE_PROP_VISIBLE:
            ljm_plane_visible(ljm_plane, values[0]);
            break;
        case PLANE_PROP_COLOR_DEPTH:
            ljm_plane_set_bpp(ljm_plane, values[0]);
            break;
        case PLANE_PROP_Z_POS:
            ljm_plane_stack_move(ljm_plane, values[0]);
            break;
        case PLANE_PROP_X_POS:
            ljm_plane_update_pos(ljm_dev, ljm_plane, values[0], ljm_plane->state->crtc_y);
            break;
        case PLANE_PROP_Y_POS:
            ljm_plane_update_pos(ljm_dev, ljm_plane, ljm_plane->state->crtc_x, values[0]);
            break;
        case PLANE_PROP_WIDTH:
            ljm_plane_set_size(ljm_plane, values[0], ljm_plane->state->crtc_h);
            break;
        case PLANE_PROP_HEIGHT:
            ljm_plane_set_size(ljm_plane, ljm_plane->state->crtc_w, values[0]);
            break;
        case PLANE_PROP_ALPHA:
            LJM_PLANE_REG_W16(ljm_plane, LAYER1_ALPHA_ADDR, values[0]);
            break;
        case PLANE_PROP_TRANSPARENT_ENALBE:
            ljm_plane_transparent_enable(ljm_plane, !!values[0]);
            break;
        case PLANE_PROP_TRANSPARENT:
            LJM_PLANE_REG_W32(ljm_plane, LAYER1_TRANSPARENT_ADDR, values[0]);
            break;
        case PLANE_PROP_VRAM0:
            LJM_PLANE_REG_W32(ljm_plane, LAYER1_BUF0_START_ADDR, values[0] >> 2);
            break;
        case PLANE_PROP_VRAM1:
            LJM_PLANE_REG_W32(ljm_plane, LAYER1_BUF1_START_ADDR, values[0] >> 2);
            break;
        case PLANE_PROP_PITCHES:
            ljm_plane_set_pitches(ljm_plane, (uint32_t)values[0]);
            break;
        case PLANE_PROP_RESIZE:
            ljm_plane_update_pos(ljm_dev, ljm_plane, values[0], values[1]);
            ljm_plane_set_size(ljm_plane, values[2], values[3]);
            break;
        case PLANE_PROP_PORT:
            ljm_plane_find_crtc_with_port(ljm_dev, ljm_plane, (uint32_t)values[0]);
            break;
        default:
            ERR_PRINT("Unknown cmommand: %d\n", request->name);
            kfree(values);
            return -EINVAL;
    }

    kfree(values);
    return 0;
}

int ljm_ioctl_plane_property_get(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    struct ljm_device *        ljm_dev = GetLjmDevFrmDrm(dev);
    struct ljm_plane_property *request = (struct ljm_plane_property *)data;
    struct ljm_plane *         ljm_plane;
    uint64_t *                 values;

    ljm_plane = ljm_plane_find(ljm_dev, request->plane_id);
    if (!ljm_plane)
    {
        ERR_PRINT("Invalid ljm_plane id for setting.\n");
        return -EINVAL;
    }

    if(request->num_values <=0)
    {
        ERR_PRINT("No space for driver's writing.\n");
        return -EINVAL;
    }

    values = kzalloc(sizeof(uint64_t) * request->num_values, GFP_KERNEL);

    switch (request->name)
    {
        case PLANE_PROP_IDLE:
            values[0] = ljm_plane_idle(ljm_plane);
            break;
        default:
            ERR_PRINT("Unknown cmommand: %d\n", request->name);
            kfree(values);
            return -EINVAL;
    }

    if (copy_to_user(request->values, values, sizeof(uint64_t) * request->num_values))
        return -EFAULT;

    kfree(values);
    return 0;
}