/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __LJMICRO_PLANE_H__
#define __LJMICRO_PLANE_H__

#include "ljmicro_base.h"

struct ljm_plane_state
{
    struct ljm_crtc *crtc;
    /**
     * @visible:
     *
     * Visibility of the plane. This can be false even if fb!=NULL and
     * crtc!=NULL, due to clipping.
     */
    bool visible;

    /**
     * @crtc_x:
     *
     * Left position of visible portion of plane on crtc, signed dest
     * location allows it to be partially off screen.
     */

    int crtc_x;
    /**
     * @crtc_y:
     *
     * Upper position of visible portion of plane on crtc, signed dest
     * location allows it to be partially off screen.
     */
    int crtc_y;

    /** @crtc_w: width of visible portion of plane on crtc */
    /** @crtc_h: height of visible portion of plane on crtc */
    uint32_t crtc_w, crtc_h;

    u16 alpha;

    u8 bpp;
    u32 pitches;

    /**
     * @zpos:
     * 图层堆叠优先级.
     *
     * 数字大的图层在堆叠时会覆盖数字小的图层。当两个图层zpos相同，
     * 则当前修改的zpos会覆盖原有的图层。这一操作由驱动程序实现而非
     * 硬件。
     */
    unsigned int zpos;
};

struct ljm_plane
{
    // struct drm_plane base;
    //struct ljm_device *ljm_dev;
    int  id;
    char name[20];

    bool active;
    int  cur_card;

    struct ljm_plane_state *state;
};

int ljm_ioctl_plane_request(struct drm_device *dev, void *data, struct drm_file *file_priv);
int ljm_ioctl_plane_free(struct drm_device *dev, void *data, struct drm_file *file_priv);
int ljm_ioctl_plane_property_set(struct drm_device *dev, void *data, struct drm_file *file_priv);
int ljm_ioctl_plane_property_get(struct drm_device *dev, void *data, struct drm_file *file_priv);
#endif /* __LJMICRO_PLANE_H__ */