/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/fs.h>
#include <linux/err.h>
#include "ljmicro_device.h"
#include "ljmicro_pm.h"

/* get current power_mode */
static ssize_t ljm_power_mode_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);

    return sprintf(buf, "%d\n", ljm_dev->power_state);
}

/* set power_mode */
static ssize_t ljm_power_mode_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    int new_state;

    if (kstrtoint(buf, 10, &new_state))
    {
        return -EINVAL;
    }

    if (new_state < 0 || new_state > 2)
        return -EINVAL;

    if (ljm_power_set_state(ljm_dev, new_state) == 0)
    {
        ljm_dev->power_state = (uint32_t)new_state;
    }
    else
    {
        return -EIO;//如果更改失败，返回I/O错误
    }

    return count;
}
/* 产生 sys/class/drm/card0/device/ljm_power_mode 节点 */
static DEVICE_ATTR_RW(ljm_power_mode);

/* show power_manager level : performance/balance/save , readonly*/
static ssize_t ljm_pm_level_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);

    if (ljm_dev->power_state == 0)
    {
        return snprintf(buf, PAGE_SIZE, "%s\n", "performance");
    }
    else if (ljm_dev->power_state == 1)
    {
        return snprintf(buf, PAGE_SIZE, "%s\n", "save");
    }
    else if (ljm_dev->power_state == 2)
    {
        return snprintf(buf, PAGE_SIZE, "%s\n", "standby");
    }
    else
        return snprintf(buf, PAGE_SIZE, "%s\n", "unknown");
}
/* /sys/class/drm/card0/device/ljm_pm_level 节点只读，用于查看当前模式 */
static DEVICE_ATTR_RO(ljm_pm_level);

/* 创建 ljm_power_mode, ljm_pm_level 节点 */
static int ljm_sysfs_init(struct ljm_device *ljm_dev)
{
    int ret;
    ret = device_create_file(ljm_dev->dev, &dev_attr_ljm_power_mode);
    if (ret)
    {
        dev_err(ljm_dev->dev, "faile to create device file ljm_power_mode\n");
        return ret;
    }

    ret = device_create_file(ljm_dev->dev, &dev_attr_ljm_pm_level);
    if (ret)
    {
        dev_err(ljm_dev->dev, "faile to create device file ljm_pm_level\n");
        return ret;
    }

    return ret;
}

static void ljm_sysfs_remove(struct ljm_device *ljm_dev)
{
    device_remove_file(ljm_dev->dev, &dev_attr_ljm_power_mode);
}

/* 初始化 power_manager  */
int ljm_pm_init(struct ljm_device *ljm_dev)
{
    return ljm_sysfs_init(ljm_dev);
}

/* 销毁 power_manager  */
void ljm_pm_exit(struct ljm_device *ljm_dev)
{
    ljm_sysfs_remove(ljm_dev);
    pr_info("power control unload\n");
}