/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "ljmicro_proc.h"
#include "ljmicro_device.h"
#include "ljmicro_version.h"
#include "ljmicro_base.h"
#include "ljmicro_ioctl.h"
#include "ljmicro_object.h"

extern struct ljm_device * get_ljm_devs(int card);
extern int gp_meminfo_get(int card, struct ljm_ioctl_info *pgpu_info);
extern ljm_board_info *ljm_get_board_info(int card);
int ljmSpiReadBuffer(void *mmio_base, char *buf, int addr, int size);

static Gop_info g_gop_info = {0};

Gop_info *get_gop_info(void)
{
    return &g_gop_info;
}

static int gpuinfo_proc_show(struct seq_file *m, void *v)
{
    return 0;
}

static int gpuinfo_proc_open(struct inode *inode, struct file *file)
{
    return single_open(file, gpuinfo_proc_show, NULL);
}

static ssize_t gpuinfo_proc_write(struct file *file, const char __user *buffer, size_t count, loff_t *f_pos)
{
    return 0;
}

#ifdef LJMICRO_DRV_ONLY
int lj_load_get(int core)
{
    return 0;
}
#else
int lj_load_get(int core);
#endif

ssize_t gpuinfo_proc_read(struct file *file, char __user *buf, size_t len, loff_t *off)
{
    int                   len_copy, ret = 0, l, _name_card;
    char                  data[512+256];
#if defined(_DEBUG)
    uint32_t              freq_axi_proc;
    uint32_t              freq_vpu_proc;
#endif
    uint32_t              freq_3d_proc;
    uint32_t              freq_2d_proc;
    uint32_t              freq_ddr_proc;
#if defined(_DEBUG)
    uint32_t              coreNum_3d = 0;
    uint32_t              coreNum_2d = 0;
#endif
#if 0
    uint8_t               ddr_channel;
#endif
    struct ljm_device *   ljm_dev = NULL;
    struct ljm_ioctl_info gpu_info;
    char _card_name;
    char* vendor_name = "Wuhan Lingjiu Microelectronics Co., Ltd.";
    char* device_all[] = {"GP101/102", "GP201", "GP202", "GP301", "GP200S"};
    char* device_name = "";
    char firmware_info[FIRMWARE_INFO_SIZE + 1] = {0};
    int i;

    l =  strlen(file->f_path.dentry->d_iname);
    _card_name = file->f_path.dentry->d_iname[l - 1];
    _name_card = _card_name - '0';

    ljm_dev = get_ljm_devs(0); /* only a ljm_dev */
    gpu_info.card = _name_card;

    switch(ljm_dev->flags & LJMICRO_CHIP_MASK)
    {
    case CHIP_GP101:
        device_name = device_all[0];
        break;
    case CHIP_GP201:
        device_name = device_all[1];
        break;
    case CHIP_GP202:
        device_name = device_all[2];
        if (ljm_dev->pdev->subsystem_device == 0x200)
            device_name = device_all[4];
        break;
    case CHIP_GP301:
        device_name = device_all[3];
        break;
    }

#if defined(_DEBUG)
    freq_axi_proc = ljm_gp200_get_freq(_name_card, freq_axi);
    freq_vpu_proc = ljm_gp200_get_freq(_name_card, freq_vpu);
#endif
    freq_3d_proc  = ljm_gp200_get_freq(_name_card, freq_3d);
    /* 核频率在proc节点打印时与规格书上保持一致 */
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201)
        freq_3d_proc = freq_3d_proc * 1200 / 923;
    else if((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
        freq_3d_proc = freq_3d_proc * 1000 / 800;

    freq_2d_proc  = ljm_gp200_get_freq(_name_card, freq_2d);
    freq_ddr_proc = ljm_gp200_get_freq(_name_card, freq_ddr);
#if defined(_DEBUG)
    coreNum_3d = get_3Dcore_num(_name_card);
    coreNum_2d = get_2Dcore_num(_name_card);
#endif

    if (gp_meminfo_get(_name_card, &gpu_info))
    {
        printk(KERN_INFO "%s: get vram information failed\n", __FUNCTION__);
    }

    len_copy = sprintf(data,
                  "Vendor: %s\n"
                  "Device: %s\n"
                  "Vendor ID: %04x\n"
                  "Device ID: %04x\n"
                  "Driver Version: %s %s\n"
#if defined(_DEBUG)
                  "AXI Frequency: %dMHz\n"
                  "VPU Frequency: %dMHz\n"
#endif
                  "Core Frequency: %dMHz\n"
                  "2D Frequency: %dMHz\n"
#if defined(_DEBUG)
                  "3D Core Number: %d\n"
                  "2D Core Number: %d\n"
#endif
                  "Memory Size: %dGB\n"
                  "Memory Usage: %d(MB) free,   %d(MB) used\n"
                  "Memory Bandwidth: %dMbps\n"
                  "GPU Utilize Rate(3D): %d%%\n"
                  "GPU Utilize Rate(2D): %d%%\n",
                  vendor_name,
                  device_name,
                  LJM_VENDOR_ID,
                  ljm_dev->flags & LJMICRO_CHIP_MASK,
                  LJM_VERSION_STRING, LJM_VERSION_DATE,
#if defined(_DEBUG)
                  freq_axi_proc,
                  freq_vpu_proc,
#endif
                  freq_3d_proc,
                  freq_2d_proc,
#if defined(_DEBUG)
                  coreNum_3d,
                  coreNum_2d,
#endif
                  gpu_info.vram_total / 1024,
#if defined(_DEBUG)
                  gpu_info.vram_free,
#else
                  gpu_info.vram_total - gpu_info.vram_used,
#endif
                  gpu_info.vram_used,
                  freq_ddr_proc * 4,
                  lj_load_get(0),
                  lj_load_get(16)
                  );

    if (ljm_dev->family == CHIP_GP200) {
        ljm_board_info *boardinfo;
#if 0   /* 取消显存通道数的显示 */
        ddr_channel = ljm_ddrv2_get_channel_mask(_name_card);
        len_copy += sprintf(&data[len_copy], "video mem channel: %s\n",
            (ddr_channel == 0x1) ? "1" :
                ((ddr_channel == 0x2) ? "2" :
                    ((ddr_channel == 0x3) ? "1 2" :
                        ((ddr_channel == 0xc) ? "3 4" :
                            ((ddr_channel == 0xf) ? "1 2 3 4" : "1")))));
#endif
        boardinfo = ljm_get_board_info(_name_card);
        if (boardinfo->vendor_id == LJM_VENDOR_ID) {
            len_copy += sprintf(&data[len_copy], "Memory Chip: %s\n",
                boardinfo->ddr_chip_name);
        }
    }

    if (strlen(g_gop_info.fw_ver) > 0) {
        len_copy += sprintf(&data[len_copy], "GOP Version: %s\n", g_gop_info.fw_ver);
        //对获取的Gop_Info信息进行获取，若获取到才进行打印
    }

    ljmSpiReadBuffer(ljm_dev->mmio_base, firmware_info, FIRMWARE_INFO_OFFSET, FIRMWARE_INFO_SIZE);
    for (i=0; i<FIRMWARE_INFO_SIZE; i++) {
        if (!(firmware_info[i] >= ' ' && firmware_info[i] <= 0x7e)) {   /* 0x7e ('~') is the last printable ASCI */
            firmware_info[i] = '\0';
            break;
        }
    }

    if (strlen(firmware_info) > 0) {
        len_copy += sprintf(&data[len_copy], "FW: %s\n", firmware_info);
    }

    if(*off >= len_copy)
        return 0;

    if ((*off + len) > len_copy) //如果剩下没读的数据长度少于len,则只复制出剩下没读部分
        len_copy = strlen(data) - *off;

    ret = copy_to_user(buf, data + *off, len_copy);
    if(!ret)
        ret = len_copy;
    *off += ret; //在read/write函数里必须通过off来改变
    return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static struct proc_ops gpuinfo_proc_fops = {
    .proc_open = gpuinfo_proc_open,
    .proc_release = single_release,
    .proc_read = gpuinfo_proc_read,
    .proc_lseek = seq_lseek,
    .proc_write = gpuinfo_proc_write,
};
#else
static struct file_operations gpuinfo_proc_fops = {
    .owner  = THIS_MODULE,
    .open   = gpuinfo_proc_open,
    .release = single_release,
    .read   = gpuinfo_proc_read,
    .llseek = seq_lseek,
    .write  = gpuinfo_proc_write,
};
#endif

struct proc_dir_entry *gpuinfo_proc_dir = NULL;

int gpuinfo_proc_init(unsigned int card)
{
    struct proc_dir_entry *file;
    char proc_name[32];

    snprintf(proc_name, 32, "gpuinfo_%d", card);
    file = proc_create(proc_name, 0666, NULL, &gpuinfo_proc_fops);
    if (!file)
    {
        printk("%s proc_create failed!\n", __func__);
        return -EPERM;
    }

    return 0;
}

uint32_t ljm_gp201_get_freq(int card, int type);

uint32_t ljm_gp202_get_freq(int card, int type);

uint32_t ljm_gp200_get_freq(int card, int type)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201)
        return ljm_gp201_get_freq(card, type);
    else
        return ljm_gp202_get_freq(card, type);
}

int get_3Dcore_num(uint32_t card)
{
#ifdef LJMICRO_DRV_ONLY
    return 0;
#else
    int count_3D = 0;
    int i;
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    ljkGALDEVICE gal_dev = (ljkGALDEVICE)ljm_dev->ddev->dev_private;
    ljkDEVICE device = gal_dev->devices[0];

    for (i = ljvCORE_MAJOR; i < ljvCORE_3D_MAX; i++)
    {
        if (device->kernels[i] != NULL)
            count_3D++;
    }
    return count_3D;
#endif
}

int get_2Dcore_num(uint32_t card)
{
#ifdef LJMICRO_DRV_ONLY
    return 0;
#else
    int count_2D = 0;
    int i;
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    ljkGALDEVICE gal_dev = (ljkGALDEVICE)ljm_dev->ddev->dev_private;
    ljkDEVICE device = gal_dev->devices[0];

    for (i = ljvCORE_2D; i < ljvCORE_2D_MAX; i++)
    {
        if (device->kernels[i] != NULL)
            count_2D++;
    }
    return count_2D;
#endif
}

int ljm_gop_info_get(struct pci_dev *pdev)
{
    int i;
    resource_size_t varm_start = 0;
    size_t vram_len = 0;
    u8 *vram_base = NULL;
    char *raw_info = NULL;
    int r = 0;

    if (g_gop_info.sign_code == SIGN_CODE)
        return 0;

    varm_start = pci_resource_start(pdev, 0);   /* 获取显存起始地址 */
    vram_len = GOP_INFO_SIZE;      /* 分配的内存大小 */
    if (vram_len > pci_resource_len(pdev, 0))
        vram_len = pci_resource_len(pdev, 0);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    vram_base  = (u8 *)ioremap(varm_start + GOP_INFO_OFFSET, vram_len);  /* 将物理地址转化为虚拟地址 */
#else
    vram_base  = (u8 *)ioremap_nocache(varm_start + GOP_INFO_OFFSET, vram_len);  /* 将物理地址转化为虚拟地址 */
#endif

    //printk(KERN_INFO "varm_start=%llx, vram_len=%zx, vram_base=0x%lx", varm_start, vram_len, (unsigned long)vram_base);
    if (vram_base == NULL)
    {
        printk(KERN_ERR "map 0x%lx bytes @ 0x%lx for GOP table failed", vram_len,
               (unsigned long)(varm_start + GOP_INFO_OFFSET));
        return 0;
    }

    raw_info = (char *)&g_gop_info; /* 将结构体地址指针赋值给char *类型指针 */

    for (i = 0; i < sizeof(g_gop_info); i++)
    {
        raw_info[i] = readb(vram_base + i); /* 逐个字节读取Bar0区域显存数据 */
    }

    if (g_gop_info.sign_code == SIGN_CODE)  /* 验证特征码数据 */
    {
        //printk(KERN_INFO "GOP_INFO CORRECT");
        g_gop_info.fw_ver[63] = '\0';
    }
    else
    {
        WARN_PRINT("GOP_INFO ERROR");
        memset(&g_gop_info, 0, sizeof(g_gop_info));
        //g_gop_info.fw_ver[0] = 0;
        r = -ENOENT;
    }

    iounmap(vram_base); /* 将映射过来的物理地址还原,释放掉虚拟地址 */
    return r;
}


void ljmicro_proc_exit(unsigned int card)
{
    char proc_name[32];

    sprintf(proc_name, "gpuinfo_%d", card);
    remove_proc_entry(proc_name, NULL);
}
