/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/version.h>
#include "ljmicro_device.h"
#include "ljmicro_drv.h"
#include "ljmicro_version.h"
#include "ljmicro_ioctl.h"
#include "ljmicro_proc.h"
#include "ljmicro_device.h"
#include "ljmicro_mcu.h"
#include "gp200.h"

#define UTIL_DEVICE_NAME "ljmrw"
#define LJM_IOCTL_MAGIC  'd'
#define UTIL_CLASS_NAME  "ljmrw_class"

/* 插卡数量 */
extern int g_card_cnt;
/*gop_info*/
Gop_info *get_gop_info(void);

/* declare the function */
unsigned long get_mmio_base(int card);
unsigned long get_vram_start(int card);
unsigned long get_vram_len(int card);
unsigned long get_vram_base(int card);

#ifndef LJM_MAX_CARDS_NUM
#define LJM_MAX_CARDS_NUM 4 /* max cards number */
#endif

static void __iomem *vram_base[LJM_MAX_CARDS_NUM]  = {NULL};
static unsigned long vram_start[LJM_MAX_CARDS_NUM] = {0};
static unsigned long vram_len[LJM_MAX_CARDS_NUM]   = {0};

static long ljmicro_ioctl(struct file *filp, unsigned int cmd, unsigned long arg);
static int  ljmicro_open(struct inode *inode, struct file *filp);
static ssize_t ljmicro_write(struct file *f, const char __user *buf, size_t size, loff_t *pos);
static int  ljmicro_release(struct inode *inode, struct file *filp);
int lj_load_get(int core);

struct class *device_class = NULL;

static struct file_operations ljmicro_util_fops = {
    .owner          = THIS_MODULE,
    .unlocked_ioctl = ljmicro_ioctl,
    .open           = ljmicro_open,
    .write          = ljmicro_write,
    .release        = ljmicro_release,
};

struct ljmicro_dev
{
    dev_t       device_no;
    struct cdev cdev;
} ljmicro_util_dev;

static int ljmicro_open(struct inode *inode, struct file *filp)
{
    return 0;
}

static ssize_t ljmicro_write(struct file *f, const char __user *buf,
                     size_t size, loff_t *pos)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    struct mcu_arg_uart args;
    int ret;

    if(!ljm_dev->mcu)
        return -ENXIO;

    if(size > MCU_UART_DATA_MAX)
        return -ENOMEM;

    args.len = size;
    args.data = memdup_user(buf, size);
    if(!args.data)
        return -ENOMEM;

    ret = ljm_mcu_handle(ljm_dev->mcu, MCU_CMD_UARTTX, &args);
    if(ret)
        return ret;

    kfree(args.data);
    return args.len;
}

static int ljmicro_release(struct inode *inode, struct file *filp)
{
    return 0;
}

static long ljmicro_ioctl_mmio(unsigned long data)
{
    struct ljm_ioctl_mmio mmio_rw;
    void __iomem *         mmio_base;
    struct ljm_ioctl_mmio *args = &mmio_rw;

    /* copy args to kernel space */
    if (copy_from_user(&mmio_rw, (struct ljm_ioctl_mmio *)data, sizeof(struct ljm_ioctl_mmio)))
    {
        printk(KERN_ERR "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    /* check args */
    if (args->card < 0 || args->card >= g_card_cnt)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, args->card);
        return -EINVAL;
    }

    if (args->bar != LJM_IOCTL_MMIO_BAR_VRAM && args->bar != LJM_IOCTL_MMIO_BAR_REG)
    {
        printk(KERN_ERR "%s: bar (%d) error!\n", __FUNCTION__, args->bar);
        return -EINVAL;
    }

    if (args->op >= LJM_IOCTL_MMIO_OP_COUNT)
    {
        printk(KERN_ERR "%s: operation error!\n", __FUNCTION__);
        return -EINVAL;
    }

    if (args->bar == LJM_IOCTL_MMIO_BAR_VRAM)
    {
        if (!(vram_base[args->card] && args->offset >= vram_start[args->card]
              && (args->offset + args->width) <= (vram_start[args->card] + vram_len[args->card])))
        {
            if (vram_base[args->card])
                iounmap(vram_base[args->card]);
            vram_start[args->card] = args->offset & PAGE_MASK;
            vram_len[args->card]   = PAGE_SIZE;
            vram_base[args->card]  = ioremap(get_vram_start(args->card) + vram_start[args->card], vram_len[args->card]);
            if (vram_base[args->card] == NULL)
            {
                pr_warn("%s: cannot map video memory area %ld@0x%lx\n", __FUNCTION__, vram_len[args->card],
                        get_vram_start(args->card) + vram_start[args->card]);
                vram_start[args->card] = 0;
                vram_len[args->card]   = 0;
                return -ENOMEM;
            }
        }
        args->offset -= vram_start[args->card];
        mmio_base = vram_base[args->card];
    }
    else
        mmio_base = (void __iomem *)get_mmio_base(args->card);

    /* operate */
    if (args->op == LJM_IOCTL_MMIO_OP_READ)
    {
        switch (args->width)
        {
            case LJM_IOCTL_MMIO_WIDTH_8:
                *((__u8 *)&args->value) = readb(mmio_base + args->offset);
                break;
            case LJM_IOCTL_MMIO_WIDTH_16:
                *((__u16 *)&args->value) = readw(mmio_base + args->offset);
                break;
            case LJM_IOCTL_MMIO_WIDTH_64:
                args->value = readq(mmio_base + args->offset);
                break;
            case LJM_IOCTL_MMIO_WIDTH_32:
            default:
                *((__u32 *)&args->value) = readl(mmio_base + args->offset);
                break;
        }
    }
    else if (args->op == LJM_IOCTL_MMIO_OP_WRITE)
    {
        switch (args->width)
        {
            case LJM_IOCTL_MMIO_WIDTH_8:
                writeb(*((__u8 *)&args->value), mmio_base + args->offset);
                break;
            case LJM_IOCTL_MMIO_WIDTH_16:
                writew(*((__u16 *)&args->value), mmio_base + args->offset);
                break;
            case LJM_IOCTL_MMIO_WIDTH_64:
                writeq(args->value, mmio_base + args->offset);
                break;
            case LJM_IOCTL_MMIO_WIDTH_32:
            default:
                writel(*((__u32 *)&args->value), mmio_base + args->offset);
                break;
        }
    }

    /* copy data back to userspace */
    if (copy_to_user((struct ljm_ioctl_mmio *)data, &mmio_rw, sizeof(struct ljm_ioctl_mmio)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EPERM;
    }
    return 0;
}

struct edid;
struct ljm_device;
struct drm_connector;

extern struct ljm_device *   get_ljm_dev(void);
extern struct drm_connector *get_ljm_connector(int card, int index);
extern bool                  drm_edid_is_valid(struct edid *edid);

/* hdmi operation */
static long ljmicro_ioctl_ddc(unsigned long data)
{
    struct ljm_ioctl_ddc  ddc;
    struct edid *         edid;
    struct drm_connector *connector = NULL;
    struct ljm_device *   ljm_dev   = NULL;

    /* copy args to kernel space */
    if (copy_from_user(&ddc, (struct ljm_ioctl_ddc *)data, sizeof(struct ljm_ioctl_ddc)))
    {
        printk(KERN_ERR "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    if (ddc.card < 0 || ddc.card >= g_card_cnt)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, ddc.card);
        return -EINVAL;
    }

    connector = get_ljm_connector(ddc.card, ddc.ch);
    if (NULL == connector)
    {
        printk(KERN_ERR "%s: channel (%d) err!\n", __FUNCTION__, ddc.card);
        return -EINVAL;
    }

    ljm_dev = GetLjmDevFrmDrm(connector->dev);
    edid    = ljm_connector_get_edid(ljm_dev, connector, &ddc.bytes);
    if (edid)
    {
        ddc.valid = drm_edid_is_valid(edid) ? 1 : 0;
        memcpy((void *)ddc.edid, (void *)edid, (ddc.bytes <= MAX_EDID) ? ddc.bytes : MAX_EDID);
        kfree(edid);
        /* copy data back to userspace */
        if (copy_to_user((struct ljm_ioctl_ddc *)data, &ddc, sizeof(struct ljm_ioctl_ddc)))
        {
            printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
            return -EPERM;
        }
    }
    else
    {
        printk(KERN_ERR "%s: connector %d ddc read failed\n", __FUNCTION__, ddc.ch);
        return -EPERM;
    }
    return 0;
}

/* hdmi operation */
static long ljmicro_ioctl_ver(unsigned long data)
{
    struct ljm_ioctl_ver version;

#ifdef LJMICRO_DRV_ONLY
    version.major = DRIVER_MAJOR;
    version.minor = DRIVER_MINOR;
    version.patch = DRIVER_PATCHLEVEL;
#else
    version.major = LJM_VERSION_MAJOR;
    version.minor = LJM_VERSION_MINOR;
    version.patch = LJM_VERSION_PATCH;
    version.build = LJM_VERSION_BUILD;
#endif

    if (copy_to_user((struct ljm_ioctl_ver *)data, &version, sizeof(struct ljm_ioctl_ver)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EPERM;
    }

    return 0;
}

extern struct pci_dev *get_drm_pdev(int card);

static long ljmicro_ioctl_pci(unsigned long data)
{
    struct ljm_ioctl_pci  pci_rw;
    struct ljm_ioctl_pci *args = &pci_rw;
    struct pci_dev *      pdev;

    /* copy args to kernel space */
    if (copy_from_user(&pci_rw, (struct ljm_ioctl_pci *)data, sizeof(struct ljm_ioctl_pci)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    if (args->card < 0 || args->card >= g_card_cnt)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, args->card);
        return -EINVAL;
    }

    pdev = get_drm_pdev(args->card);

    /* check args */
    if (args->op >= LJM_IOCTL_PCI_OP_COUNT)
    {
        printk(KERN_INFO "%s: operation error!\n", __FUNCTION__);
        return -EINVAL;
    }

    /* operate */
    if (args->op == LJM_IOCTL_PCI_OP_READ)
    {
        pci_read_config_dword(pdev, args->offset, &args->value);

        /* copy data back to userspace */
        if (copy_to_user((struct ljm_ioctl_pci *)data, &pci_rw, sizeof(struct ljm_ioctl_pci)))
        {
            printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
            return -EPERM;
        }
    }
    else if (args->op == LJM_IOCTL_PCI_OP_WRITE)
    {
        pci_write_config_dword(pdev, args->offset, args->value);
    }

    return 0;
}

extern int gp_meminfo_get(int card, struct ljm_ioctl_info *pgpu_info);

static long ljmicro_ioctl_info(unsigned long data)
{
    struct ljm_ioctl_info gpu_info;

    /* copy args to kernel space */
    if (copy_from_user(&gpu_info, (struct ljm_ioctl_info *)data, sizeof(struct ljm_ioctl_info)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    if (gpu_info.card < 0 || gpu_info.card >= g_card_cnt)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, gpu_info.card);
        return -EINVAL;
    }

    if (gp_meminfo_get(gpu_info.card, &gpu_info))
    {
        printk(KERN_INFO "%s: get vram information failed\n", __FUNCTION__);
    }

    if (copy_to_user((struct ljm_ioctl_info *)data, &gpu_info, sizeof(struct ljm_ioctl_info)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EPERM;
    }
    return 0;
}

static long ljmicro_ioctl_kmem(unsigned long data)
{
    struct ljm_ioctl_kmem  kmem_rw;
    struct ljm_ioctl_kmem *args = &kmem_rw;

    /* copy args to kernel space */
    if (copy_from_user(&kmem_rw, (struct ljm_ioctl_kmem *)data, sizeof(struct ljm_ioctl_kmem)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    /* check args */
    if (args->op >= LJM_IOCTL_KMEM_OP_COUNT)
    {
        printk(KERN_INFO "%s: operation error!\n", __FUNCTION__);
        return -EINVAL;
    }

    /* operate */
    if (args->op == LJM_IOCTL_KMEM_OP_READ)
    {
        switch (args->width)
        {
            case LJM_IOCTL_KMEM_WIDTH_8:
                *((__u8 *)&args->value) = *((__u8 *)args->kvaddr);
                break;
            case LJM_IOCTL_KMEM_WIDTH_16:
                *((__u16 *)&args->value) = *((__u16 *)args->kvaddr);
                break;
            case LJM_IOCTL_KMEM_WIDTH_64:
                args->value = *((__u64 *)args->kvaddr);
                break;
            case LJM_IOCTL_KMEM_WIDTH_32:
            default:
                *((__u32 *)&args->value) = *((__u32 *)args->kvaddr);
                break;
        }
    }
    else if (args->op == LJM_IOCTL_KMEM_OP_WRITE)
    {
        switch (args->width)
        {
            case LJM_IOCTL_KMEM_WIDTH_8:
                *((__u8 *)args->kvaddr) = *((__u8 *)&args->value);
                break;
            case LJM_IOCTL_KMEM_WIDTH_16:
                *((__u16 *)args->kvaddr) = *((__u16 *)&args->value);
                break;
            case LJM_IOCTL_KMEM_WIDTH_64:
                *((__u64 *)args->kvaddr) = *((__u64 *)&args->value);
                break;
            case LJM_IOCTL_KMEM_WIDTH_32:
            default:
                *((__u32 *)args->kvaddr) = *((__u32 *)&args->value);
                break;
        }
    }

    if (copy_to_user((struct ljm_ioctl_kmem *)data, &kmem_rw, sizeof(struct ljm_ioctl_kmem)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EPERM;
    }
    return 0;
}

static long ljmicro_ioctl_hdmi(unsigned long data)
{
    struct ljm_ioctl_hdmi_common  hdmi_rw;
    struct ljm_ioctl_hdmi_common *args = &hdmi_rw;

    /* copy args to kernel space */
    if (copy_from_user(&hdmi_rw, (struct ljm_ioctl_hdmi_common *)data, sizeof(struct ljm_ioctl_hdmi_common)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    /* operate */
    switch (args->cmd)
    {
        case LJM_HDMI_CMD_PHY_R:
            {
                struct ljm_hdmi_cmd_phy_data phy_data;
                if (copy_from_user(&phy_data, (struct ljm_hdmi_cmd_phy_data *)args->pdata,
                                   sizeof(struct ljm_hdmi_cmd_phy_data)))
                {
                    printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
                    return -EPERM;
                }
                phy_data.data = phy_data.addr; //测试ioctl,简单的把addr赋值给data。后续补充具体操作
                if (copy_to_user((struct ljm_hdmi_cmd_phy_data *)args->pdata, &phy_data,
                                 sizeof(struct ljm_hdmi_cmd_phy_data)))
                {
                    printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
                    return -EPERM;
                }
                break;
            }
        default:
            printk(KERN_WARNING "%s:HDMI cmd is invalid!\n", __FUNCTION__);
            return -EINVAL;
    }

    return 0;
}

#if USE_DP_CONNECTOR
static long ljmicro_ioctl_dp(unsigned long data)
{
    struct ljm_ioctl_dp_command  dp_cmd;
    struct ljm_ioctl_dp_command *args = &dp_cmd;
    struct ljm_device *ljm_dev;
    struct ljm_port *ljm_port;
    int i, found;
    ssize_t size;
    struct edid *cur_edid;

    /* copy args to kernel space */
    if (copy_from_user(args, (struct ljm_ioctl_dp_command *)data, sizeof(struct ljm_ioctl_dp_command)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    if (args->card < 0 || args->card >= g_card_cnt) {
        printk(KERN_WARNING "card %d is not exist!\n", args->card);
        return -EINVAL;
    }

    ljm_dev = get_ljm_devs(args->card);

    /* find the DP port */
    for (i=0, found=0; i<ljm_dev->num_ports; i++) {
        ljm_port = ljm_dev->ports[i];
        if (ljm_port->type == LJM_PORT_DP) {
            if (args->index == ljm_port->vdev.dp.index) {
                found = 1;
                break;
            }
        }
    }

    if (found == 0) {
        printk(KERN_WARNING "dp %d is not exist!\n", args->index);
        return -EINVAL;
    }

    /* operate */
    switch (args->cmd)
    {
    case LJM_DP_CMD_DPCD_WRITE:
        size = drm_dp_dpcd_write(&ljm_port->vdev.dp.aux, args->offset, args->pdata, args->size);
        if (size > 0) return 0;
        else return (long)size;
    case LJM_DP_CMD_DPCD_READ:
        size = drm_dp_dpcd_read(&ljm_port->vdev.dp.aux, args->offset, args->pdata, args->size);
        if (size > 0 && size <= 16) {
            args->size = (uint32_t)size;
            if (copy_to_user((struct ljm_ioctl_dp_command *)data, args, sizeof(struct ljm_ioctl_dp_command))) {
                printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
                return -EPERM;
            }
        } else {
            return -EIO;
        }
        break;
    case LJM_DP_CMD_EDID_GET:
        cur_edid =  ljm_dp_get_edid(ljm_port->vdev_connector, &args->size);
        if (cur_edid) {
            args->size = DDC_EDID_SIZE + DDC_EDID_SIZE * cur_edid->extensions;
            memcpy(args->pdata, cur_edid, args->size);
            kfree(cur_edid);
            if (copy_to_user((struct ljm_ioctl_dp_command *)data, args, sizeof(struct ljm_ioctl_dp_command)))
            {
                printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
                return -EPERM;
            }
        } else {
            return -EIO;
        }
        break;
    default:
        printk(KERN_WARNING "%s:HDMI cmd is invalid!\n", __FUNCTION__);
        return -EINVAL;
    }

    return 0;
}
#endif

static long ljmicro_ioctl_i2c(unsigned long data)/*I2C的错误return主要与tool中i2c error相匹配*/
{
    struct ljm_ioctl_i2c_command  i2c_rw;
    struct ljm_ioctl_i2c_command *args = &i2c_rw;
    struct ljm_device *   ljm_dev   = NULL;
    u8 *pdata;

    /* copy args to kernel space */
    if (copy_from_user(&i2c_rw, (struct ljm_ioctl_i2c_command *)data, sizeof(struct ljm_ioctl_i2c_command)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EINTR;
    }

    if (i2c_rw.card < 0 || i2c_rw.card >= g_card_cnt)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, i2c_rw.card);
        return -EINTR;
    }

    ljm_dev = get_ljm_devs(i2c_rw.card);
    if (NULL == ljm_dev)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, i2c_rw.card);
        return -EINTR;
    }

    pdata = kmalloc((i2c_rw.size <= MAX_EDID) ? i2c_rw.size: MAX_EDID, GFP_KERNEL);
    if(args->op == LJM_IOCTL_I2C_OP_WRITE)
        memcpy((void *)pdata, (void *)i2c_rw.pdata,(i2c_rw.size <= MAX_EDID) ? i2c_rw.size: MAX_EDID);

    mutex_lock(&ljm_dev->i2c_mutex);
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
        if (args->ch <= 3 && args->ch >= 0) { /*i2c oc opearator*/
            if(args->op == LJM_IOCTL_I2C_OP_READ)
                i2c_rw.flag = ljm_i2c_oc_read_req(ljm_dev->i2c_devs[args->ch], args->target, args->addr, pdata, args->size);
            if(args->op == LJM_IOCTL_I2C_OP_WRITE)
                i2c_rw.flag = ljm_i2c_oc_write_req(ljm_dev->i2c_devs[args->ch], args->target, args->addr, pdata, args->size);
        }
        else if (args->ch >= 4 && args->ch <= 5) { /*i2c dw opearator*/
            if(args->op == LJM_IOCTL_I2C_OP_READ)
                i2c_rw.flag = ljm_i2c_dw_read_req(ljm_dev->i2c_devs[args->ch], args->target, args->addr, pdata, args->size);
            if(args->op == LJM_IOCTL_I2C_OP_WRITE)
                i2c_rw.flag = ljm_i2c_dw_write_req(ljm_dev->i2c_devs[args->ch], args->target, args->addr, pdata, args->size);
        }
    }
    else {   /* dw I2C for GP202 */
        if(args->op == LJM_IOCTL_I2C_OP_READ)
            i2c_rw.flag = ljm_i2c_dw_read_req(ljm_dev->i2c_devs[args->ch], args->target, args->addr, pdata, args->size);
        if(args->op == LJM_IOCTL_I2C_OP_WRITE)
            i2c_rw.flag = ljm_i2c_dw_write_req(ljm_dev->i2c_devs[args->ch], args->target, args->addr, pdata, args->size);
        usleep_range(100, 200);
    }
    mutex_unlock(&ljm_dev->i2c_mutex);

    if(args->op == LJM_IOCTL_I2C_OP_READ)
        memcpy((void *)i2c_rw.pdata, (void *)pdata,(i2c_rw.size <= MAX_EDID) ? i2c_rw.size: MAX_EDID);

    kfree(pdata);

    if (copy_to_user((struct ljm_ioctl_i2c_command *)data, &i2c_rw, sizeof(struct ljm_ioctl_i2c_command)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EINTR;
    }

    return 0;
}

inline u32 ljm_get_device_id(int card);
extern int ljmSpiReadBuffer(void *mmio_base, char *buf, int addr, int size);
static long ljmicro_ioctl_query(unsigned long data)
{
    struct ljm_ioctl_query query_info;
    struct ljm_device *   ljm_dev     = NULL;
    uint16_t              id_buf      = 0;
    uint32_t              freq_buf    = 0;
    uint8_t               raw_channel = 0;
    int                   core        = 0;
    int                   i           = 0;
    Gop_info *gop_info;

    gop_info = get_gop_info();
    /* copy args to kernel space */
    if (copy_from_user(&query_info, (struct ljm_ioctl_query *)data, sizeof(struct ljm_ioctl_query)))
    {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EINTR;
    }

    if (query_info.card < 0 || query_info.card >= g_card_cnt)
    {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, query_info.card);
        return -ENXIO;
    }

    ljm_dev = get_ljm_devs(query_info.card);
    switch (query_info.info)
    {
        case QUERY_DEVICE_ID:
            id_buf = ljm_dev->flags & LJMICRO_CHIP_MASK;

            query_info.pdata[1] = (uint8_t)((id_buf & 0xFF00) >> 8);
            query_info.pdata[0] = (uint8_t)(id_buf & 0xFF);
            break;
        case QUERY_FREQ_AXI:
            freq_buf = ljm_gp200_get_freq(query_info.card, freq_axi);

            query_info.pdata[3] = (uint8_t)((freq_buf & 0xFF000000) >> 24);
            query_info.pdata[2] = (uint8_t)((freq_buf & 0xFF0000) >> 16);
            query_info.pdata[1] = (uint8_t)((freq_buf & 0xFF00) >> 8);
            query_info.pdata[0] = (uint8_t)(freq_buf & 0xFF);

            break;
        case QUERY_FREQ_VPU:
            freq_buf = ljm_gp200_get_freq(query_info.card, freq_vpu);

            query_info.pdata[3] = (uint8_t)((freq_buf & 0xFF000000) >> 24);
            query_info.pdata[2] = (uint8_t)((freq_buf & 0xFF0000) >> 16);
            query_info.pdata[1] = (uint8_t)((freq_buf & 0xFF00) >> 8);
            query_info.pdata[0] = (uint8_t)(freq_buf & 0xFF);

            break;
        case QUERY_FREQ_3D:
            freq_buf = ljm_gp200_get_freq(query_info.card, freq_3d);

            query_info.pdata[3] = (uint8_t)((freq_buf & 0xFF000000) >> 24);
            query_info.pdata[2] = (uint8_t)((freq_buf & 0xFF0000) >> 16);
            query_info.pdata[1] = (uint8_t)((freq_buf & 0xFF00) >> 8);
            query_info.pdata[0] = (uint8_t)(freq_buf & 0xFF);
            break;
        case QUERY_FREQ_2D:
            freq_buf = ljm_gp200_get_freq(query_info.card, freq_2d);

            query_info.pdata[3] = (uint8_t)((freq_buf & 0xFF000000) >> 24);
            query_info.pdata[2] = (uint8_t)((freq_buf & 0xFF0000) >> 16);
            query_info.pdata[1] = (uint8_t)((freq_buf & 0xFF00) >> 8);
            query_info.pdata[0] = (uint8_t)(freq_buf & 0xFF);
            break;
        case QUERY_FREQ_DDR:
            freq_buf = ljm_gp200_get_freq(query_info.card, freq_ddr);

            query_info.pdata[3] = (uint8_t)((freq_buf & 0xFF000000) >> 24);
            query_info.pdata[2] = (uint8_t)((freq_buf & 0xFF0000) >> 16);
            query_info.pdata[1] = (uint8_t)((freq_buf & 0xFF00) >> 8);
            query_info.pdata[0] = (uint8_t)(freq_buf & 0xFF);
            break;
        case QUERY_NUM_3DCORE:
            query_info.pdata[0] = (uint8_t)get_3Dcore_num(query_info.card);
            break;
        case QUERY_NUM_2DCORE:
            query_info.pdata[0] = 1;
            break;
        case QUERY_NUM_CARD:
            query_info.pdata[0] = (uint8_t)g_card_cnt;
            break;
        case QUERY_INFO_GOP:
            memcpy(query_info.pdata, gop_info->fw_ver, sizeof(gop_info->fw_ver));
            break;
        case QUERY_CHANNEL_MEM:
            if (ljm_dev->family == CHIP_GP200) {
                switch (ljm_dev->flags & LJMICRO_CHIP_MASK) {
                case CHIP_GP201:
                    raw_channel = gp201_ddrv2_get_channel_mask(query_info.card);
                    break;
                case CHIP_GP202:
                    raw_channel = gp202_ddrv2_get_channel_mask(query_info.card);
                    break;
                default:
                    ERR_PRINT("device id (0x%x) is not supported!\n", ljm_dev->flags & LJMICRO_CHIP_MASK);
                    break;
                }
                query_info.pdata[0] = raw_channel;
            }
            else {
                query_info.pdata[0] = 1;
            }
            break;
        case QUERY_LOAD_RATE:
            core = (int)query_info.pdata[0];
            *((int *)(query_info.pdata)) = lj_load_get(core);
            break;
        case QUERY_INFO_FW:
            query_info.pdata[FIRMWARE_INFO_SIZE-1] = '\0';
            ljmSpiReadBuffer(ljm_dev->mmio_base, query_info.pdata, FIRMWARE_INFO_OFFSET, FIRMWARE_INFO_SIZE);
            for (; i<FIRMWARE_INFO_SIZE; i++)
            {
                if (!(query_info.pdata[i] >= ' ' && query_info.pdata[i] <= 0x7e))
                {   /* 0x7e ('~') is the last printable ASCI */
                    query_info.pdata[i] = '\0';
                    break;
                }
            }
            break;
        default:
            return -EINTR;
    }

    if (copy_to_user((struct ljm_ioctl_query *)data, &query_info,  sizeof(struct ljm_ioctl_query)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EINTR;
    }
    return 0;
}

#ifndef LJMICRO_DRV_ONLY
void gpu_dump_external_free(void);
void gpu_dump_exclusive_free(void);
#endif

static long ljmicro_ioctl_cmd(unsigned long data)
{
    struct ljm_ioctl_cmd cmd;
    struct ljm_device *ljm_dev = NULL;
    uint32_t tmp;

    /* copy args to kernel space */
    if (copy_from_user(&cmd, (struct ljm_ioctl_cmd *)data, sizeof(struct ljm_ioctl_cmd))) {
        printk(KERN_INFO "%s: copy from user failed\n", __FUNCTION__);
        return -EINTR;
    }

    if (cmd.card < 0 || cmd.card >= g_card_cnt) {
        printk(KERN_ERR "%s: card (%d) err!\n", __FUNCTION__, cmd.card);
        return -ENXIO;
    }

    ljm_dev = get_ljm_devs(cmd.card);
    switch (cmd.type) {
        case CMD_TYPE_GET_PVT_TEMP:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp201_pvt_get_temp(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp202_pvt_get_temp(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else {
                ERR_PRINT("device id (0x%x) is not supported!\n", ljm_dev->flags & LJMICRO_CHIP_MASK);
            }
            break;
        case CMD_TYPE_GET_PVT_VOLT:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp201_pvt_get_volt(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp202_pvt_get_volt(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else {
                ERR_PRINT("device id (0x%x) is not supported!\n", ljm_dev->flags & LJMICRO_CHIP_MASK);
            }
            break;
        case CMD_TYPE_GET_SVT_PROCESS:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp201_pvt_get_svt_process(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp202_pvt_get_svt_process(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else {
                ERR_PRINT("device id (0x%x) is not supported!\n", ljm_dev->flags & LJMICRO_CHIP_MASK);
            }
            break;
        case CMD_TYPE_GET_LVT_PROCESS:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp201_pvt_get_lvt_process(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp202_pvt_get_lvt_process(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else {
                ERR_PRINT("device id (0x%x) is not supported!\n", ljm_dev->flags & LJMICRO_CHIP_MASK);
            }
            break;
        case CMD_TYPE_GET_ULVT_PROCESS:
            if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp201_pvt_get_ulvt_process(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202) {
                mutex_lock(&ljm_dev->hw_mutex);
                tmp = gp202_pvt_get_ulvt_process(ljm_dev->cur_card, cmd.index);
                mutex_unlock(&ljm_dev->hw_mutex);

                *((uint32_t *)cmd.pdata) = tmp & m_PVT_DATA_MASK;
            } else {
                ERR_PRINT("device id (0x%x) is not supported!\n", ljm_dev->flags & LJMICRO_CHIP_MASK);
            }
            break;
        case CMD_TYPE_DUMP_VRAM_FREE:
#ifndef LJMICRO_DRV_ONLY
            gpu_dump_external_free();
            gpu_dump_exclusive_free();
#endif
            break;
        default:
            return -EINVAL;
    }

    if (copy_to_user((struct ljm_ioctl_cmd *)data, &cmd, sizeof(struct ljm_ioctl_cmd))) {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EINTR;
    }
    return 0;
}

#define LEGACY_INDEX_PORT   0x3d4
#define LEGACY_DATA_PORT    0x3d5

static long ljmicro_ioctl_ioport(unsigned long data)
{
    struct ljm_ioctl_ioport ioport_rw;
    struct ljm_ioctl_ioport *args = &ioport_rw;

    /* copy args to kernel space */
    if (copy_from_user(&ioport_rw, (struct ljm_ioctl_ioport *)data, sizeof(struct ljm_ioctl_ioport)))
    {
        printk(KERN_ERR "%s: copy from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    if (args->op == LJM_IOCTL_IOPORT_OP_READ)
    {
        outb(args->offset, LEGACY_INDEX_PORT);
        args->val =  inb(LEGACY_DATA_PORT);
        /* copy data back to userspace */
        if (copy_to_user((struct ljm_ioctl_ioport *)data, &ioport_rw, sizeof(struct ljm_ioctl_ioport)))
        {
            printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
            return -EPERM;
        }
    }
    else {
        outb(args->offset, LEGACY_INDEX_PORT);
        outb(args->val, LEGACY_DATA_PORT);
    }

    return 0;
}

static long ljmicro_ioctl_gamma(unsigned long data)
{
    struct ljm_ioctl_gamma gamma_rw;
    uint16_t *gamma_store;
    uint32_t size;
    void *r_base, *g_base, *b_base;
    int crtc_id;
    struct ljm_device *ljm_dev;
    struct ljm_crtc *ljm_crtc;

    ljm_dev = get_ljm_devs(0);

    /* copy args to kernel space */
    if (copy_from_user(&gamma_rw, (struct ljm_ioctl_gamma *)data, sizeof(struct ljm_ioctl_gamma)))
    {
        printk(KERN_ERR "%s: copy ioctl data from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    crtc_id = gamma_rw.crtc_id;
    size = gamma_rw.gamma_size * (sizeof(uint16_t));
    gamma_store = kcalloc(gamma_rw.gamma_size, sizeof(uint16_t) * 3, GFP_KERNEL);

    r_base = gamma_store;
    if (copy_from_user(r_base, (void __user *)(unsigned long)gamma_rw.red, size)) {
        printk(KERN_ERR "%s: copy r_base from user failed\n", __FUNCTION__);
        return -EFAULT;
    }

    g_base = r_base + size;
    if (copy_from_user(g_base, (void __user *)(unsigned long)gamma_rw.green, size)) {
        printk(KERN_ERR "%s: copy g_base from user failed\n", __FUNCTION__);
        return -EFAULT;
    }

    b_base = g_base + size;
    if (copy_from_user(b_base, (void __user *)(unsigned long)gamma_rw.blue, size)) {
        printk(KERN_ERR "%s: copy b_base from user failed\n", __FUNCTION__);
        return -EFAULT;
    }

    ljm_crtc = to_ljm_crtc(ljm_dev->crtcs[crtc_id]);
    return ljm_crtc_load_lut(ljm_dev, ljm_crtc, r_base, g_base, b_base, gamma_rw.gamma_size);
}

static long ljmicro_ioctl_power(unsigned long data)
{
    struct ljm_ioctl_power power;
    struct ljm_device *    ljm_dev;
    int32_t old_state;
    int r;

    ljm_dev = get_ljm_devs(0);

    /* copy args to kernel space */
    if (copy_from_user(&power, (struct ljm_ioctl_power *)data, sizeof(struct ljm_ioctl_power)))
    {
        printk(KERN_ERR "%s: copy ioctl data from user failed\n", __FUNCTION__);
        return -EPERM;
    }

    switch (power.op)
    {
        case LJM_IOCTL_POWER_OP_GET:
            power.state = ljm_dev->power_state;
            break;
        case LJM_IOCTL_POWER_OP_SET:
            if (ljm_dev->power_state == power.state)
                return 0;
            else
            {
                old_state = ljm_dev->power_state;
                r = ljm_set_power(ljm_dev, power.state);
                if (r)
                {
                    /* failed, restore old power state */
                    ljm_set_power(ljm_dev, old_state);
                    return -EPERM;
                }
                else
                {
                    ljm_dev->power_state = power.state;
                    power.state = old_state;
                }
            }
            break;
        default:
            return -EINVAL;
    }

    if (copy_to_user((struct ljm_ioctl_power *)data, &power, sizeof(struct ljm_ioctl_power)))
    {
        printk(KERN_INFO "%s: copy to user failed\n", __FUNCTION__);
        return -EPERM;
    }

    return 0;
}

static long ljmicro_ioctl(struct file *filp, unsigned int cmd, unsigned long data)
{
    long ret = 0;
    switch (cmd)
    {
        case LJM_IOCTL_MMIO:
            ret = ljmicro_ioctl_mmio(data);
            break;
        case LJM_IOCTL_DDC:
            ret = ljmicro_ioctl_ddc(data);
            break;
        case LJM_IOCTL_VER:
            ret = ljmicro_ioctl_ver(data);
            break;
        case LJM_IOCTL_PCI:
            ret = ljmicro_ioctl_pci(data);
            break;
        case LJM_IOCTL_INFO:
            ret = ljmicro_ioctl_info(data);
            break;
        case LJM_IOCTL_KMEM:
            ret = ljmicro_ioctl_kmem(data);
            break;
        case LJM_IOCTL_HDMI:
            ret = ljmicro_ioctl_hdmi(data);
            break;
#if USE_DP_CONNECTOR
        case LJM_IOCTL_DP:
            ret = ljmicro_ioctl_dp(data);
            break;
#endif
        case LJM_IOCTL_I2C:
            ret = ljmicro_ioctl_i2c(data);
            break;
        case LJM_IOCTL_QUERY:
            ret = ljmicro_ioctl_query(data);
            break;
        case LJM_IOCTL_CMD:
            ret = ljmicro_ioctl_cmd(data);
            break;
        case LJM_IOCTL_IOPORT:
            ret = ljmicro_ioctl_ioport(data);
            break;
        case LJM_IOCTL_GAMMA:
            ret = ljmicro_ioctl_gamma(data);
            break;
        case LJM_IOCTL_POWER:
            ret = ljmicro_ioctl_power(data);
            break;
        case LJM_IOCTL_SET_FAN_MANUAL:
            ret = ljmicro_ioctl_thermal(cmd, data);
            break;
        case LJM_IOCTL_CLEAR_MANUAL:
            ret = ljmicro_ioctl_thermal(cmd, data);
            break;
        default:
            printk(KERN_INFO "%s: command error\n", __FUNCTION__);
            return -EINTR;
    }

    return ret;
}

int ljmicro_rw_init(void)
{
    int r;

    r = alloc_chrdev_region(&ljmicro_util_dev.device_no, 0, 1, UTIL_DEVICE_NAME); /* alloc the device number */

    if (r < 0)
    {
        printk(KERN_WARNING "ljmrw: device number allocation failed\n");
        return r;
    }

    cdev_init(&ljmicro_util_dev.cdev, &ljmicro_util_fops);
    r = cdev_add(&ljmicro_util_dev.cdev, ljmicro_util_dev.device_no, 1); /* add cdev device */

    if (r < 0)
    {
        printk(KERN_WARNING "ljmrw: device is not created\n");
        return r;
    }

    //printk(KERN_INFO "ljmrw: major: %d, minor: %d\n", MAJOR(ljmicro_util_dev.device_no),
    //       MINOR(ljmicro_util_dev.device_no));

    /* Create the device class */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6,6,0)
    device_class = class_create(THIS_MODULE, UTIL_CLASS_NAME);
#else
    device_class = class_create(UTIL_CLASS_NAME);
#endif
    if (IS_ERR(device_class))
        printk(KERN_WARNING "ljmrw: failed to create the class\n");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
    device_create(device_class, NULL, ljmicro_util_dev.device_no, NULL, UTIL_DEVICE_NAME);
#else
    device_create(device_class, NULL, ljmicro_util_dev.device_no, UTIL_DEVICE_NAME);
#endif

    return 0;
}

void ljmicro_rw_exit(void)
{
    int i;

    for (i = 0; i < LJM_MAX_CARDS_NUM; i++)
    {
        if (vram_base[i])
        {
            iounmap(vram_base[i]);
            vram_base[i]  = NULL;
            vram_start[i] = 0;
            vram_len[i]   = 0;
        }
    }

    cdev_del(&ljmicro_util_dev.cdev); /* delete the device */
    device_destroy(device_class, ljmicro_util_dev.device_no);
    class_destroy(device_class);                             /* destroy the device class */
    unregister_chrdev_region(ljmicro_util_dev.device_no, 1); /* free the device number */
    return;
}
