/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_base.h"
#include "ljmicro_family.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

/**
 * SPI Flash
 */

#define GP201_SPI_CTL_FIFO_SIZE     1024 /* 1024 bytes for GP201 */
#define GP202_SPI_CTL_FIFO_SIZE     128  /* 128 byets for GP202 */

#define SPI_REG_CTRLR0        0x0
#define SPI_REG_CTRLR1        0x4
#define SPI_REG_SSIENR        0x8
#define SPI_REG_SER           0x10
#define SPI_REG_BAUDR         0x14
#define SPI_REG_TXFTLR        0x18
#define SPI_REG_RXFTLR        0x1c
#define SPI_REG_TXFLR         0x20
#define SPI_REG_RXFLR         0x24
#define SPI_REG_SR            0x28
#define SPI_REG_IMR           0x2c
#define SPI_REG_DR            0x60
#define SPI_REG_RX_SAMPLE_DLY 0xf0
#define SPI_REG_SPI_CTRLR0    0xf4

#define SPI_R32_BASE(BASE, ADDR)      readl((void __iomem *)(BASE) + ljm_gpu_hardware.reg_spi_offset + (ADDR))
#define SPI_W32_BASE(BASE, ADDR, VAL) writel((u32)(VAL), (void __iomem *)(BASE) + ljm_gpu_hardware.reg_spi_offset + (ADDR))

#define SPI_CHECK_VAL_TIMEOUT(addr, val, mask_bit, cmp_val, time) \
    do{ \
        typeof(time) __timeout = 20; \
        val = SPI_R32_BASE(mmio_base, addr); \
        while ((val & mask_bit) != cmp_val) \
        { \
            val = SPI_R32_BASE(mmio_base, addr); \
            usleep_range(10, 20); \
            --__timeout; \
            if (__timeout < 0) \
            { \
                goto err; \
            }\
        } \
    } while(0);

/*****************************************************************************/
/* spi operation                                                             */
/*****************************************************************************/

/*****************************************************************************
 *  函 数 名: ljmSpiExitCrmMode
 *  功能描述: 推出SPI Flash的XiP模式
 *  输入参数: void
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年6月5日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int Judge_Exit_Xip(uint32_t status) {
    return status & m_EXITXIP_FIN_IO_MMIO? 1 : 0;
}

int Judge_Spi_Mode(uint32_t status) {
    return status & m_SPI_MODE? 1 : 0;
}

void ljmSpiExitCrmMode(void *mmio_base)
{
    uint32_t val;

    HEADER_ARG("mmio_base=%p", mmio_base);

    val = LJM_MISC_R32_BASE(mmio_base, MISC_QSPI_MCU_DONE);
    TRACE("SPI: MISC_QSPI_MCU_DONE=0x%x\n", val);
    if (!(val & m_QSPI_MCU_DONE)) {
        uint32_t r01 = LJM_MISC_R32_BASE(mmio_base, MISC_CHIP_R01);
        uint32_t status;
        uint32_t Spi_Mode = 0;

        if ((r01 & 0xffff0000) == 0x02020000) {
            status = LJM_MISC_R32_BASE(mmio_base, MISC_QSPI_INIT_INTR_STATUS);
            Spi_Mode = Judge_Spi_Mode(status);
            if (Spi_Mode) {
                DRM_INFO("QSPI MODE \n");
            }
            else {
                DRM_INFO("SPI MODE \n");
            }
        }

        if ((r01 & 0xffff0000) == 0x02010000 || Spi_Mode) { /* GP201 and GP202(QSPI) should exit from XiP mode */
            TRACE("SPI: exit from XIP mode\n");
            SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000000);
            SPI_W32_BASE(mmio_base, SPI_REG_CTRLR0, 0x005F0300);
            SPI_W32_BASE(mmio_base, SPI_REG_CTRLR1, 0x00000003);
            SPI_W32_BASE(mmio_base, SPI_REG_BAUDR, 0x00000004);
            SPI_W32_BASE(mmio_base, SPI_REG_TXFTLR, 0x00000007);
            SPI_W32_BASE(mmio_base, SPI_REG_RXFTLR, 0x00000007);
            SPI_W32_BASE(mmio_base, SPI_REG_IMR, 0x0000003F);
            SPI_W32_BASE(mmio_base, SPI_REG_RX_SAMPLE_DLY, 0x00000001);
            SPI_W32_BASE(mmio_base, SPI_REG_SPI_CTRLR0, 0x00002021);
            SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000001);
            SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000001);

            LJM_MISC_W32_BASE(mmio_base, MISC_QSPI_XIP_EN_IO_MMIO, 0x1);
            SPI_R32_BASE(mmio_base, SPI_REG_CTRLR0); /* don't care the value */
            LJM_MISC_W32_BASE(mmio_base, MISC_QSPI_XIP_EN_IO_MMIO, 0x0);

            /* GP202Judge_Exit_Xip */
            if (Spi_Mode) {
                if (Judge_Exit_Xip(status)) {
                    DRM_INFO("EXIT XIP MODE SUCCESS");
                }
                else {
                    DRM_INFO("NOT EXIT XIP MODE");
                }
            }
        }

        /** attach QSPI with apb*_io/mmio bus */
        LJM_MISC_W32_BASE(mmio_base, MISC_QSPI_MCU_DONE, val | m_QSPI_MCU_DONE);
    }

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: ljmSpiReadBuffer
 *  功能描述: 从SPI Flash读取一块数据
 *  输入参数: char *buf  缓存指针
 *            int addr   SPI Flash偏移地址
 *            int size   读取缓存容量
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2021年9月24日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数,支持SPI Flash读数据
 *         2. 日    期: 2022年6月5日
 *            作    者: 胡艳明
 *            修改内容: 修改为QSPI Flash驱动函数
 *****************************************************************************/
int ljmSpiReadBuffer(void *mmio_base, char *buf, int addr, int size)
{
    uint32_t val;
    int left_size, cur_size;
    int i;
    int timeout = 60;
    uint32_t r01;
    int fifo_size;

    HEADER_ARG("mmio_base=%p buf=0x%p addr=0x%x size=0x%x", mmio_base, buf, addr, size);

    ljmSpiExitCrmMode(mmio_base);

    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR0, 0x00070100);
    SPI_W32_BASE(mmio_base, SPI_REG_BAUDR, 0x00000004);
    SPI_W32_BASE(mmio_base, SPI_REG_TXFTLR, 0x000000ff);
    SPI_W32_BASE(mmio_base, SPI_REG_RXFTLR, 0x000000ff);
    SPI_W32_BASE(mmio_base, SPI_REG_TXFLR, 0x000000ff);
    SPI_W32_BASE(mmio_base, SPI_REG_RXFLR, 0x000000ff);
    SPI_W32_BASE(mmio_base, SPI_REG_IMR, 0x0000003F);
    SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000001);
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000001);
    SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000001);

    /* wren */
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000006);
    SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0xff, 0x06, timeout);


    /* en wrsr */
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000050);
    SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0xff, 0x06, timeout);


    /* wrsr */
    SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000001);
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000002);
    SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000001);
    SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0xff, 0x06, timeout);


    /* rdsr05 */
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR0, 0x00070300);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR1, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000001);
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000005);
    SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0x08, 0x08, timeout);

    val = SPI_R32_BASE(mmio_base, SPI_REG_DR);
    while ((val & 0x1) != 0x0)
    {
        SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000005);
        SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0x08, 0x08, timeout);
        val = SPI_R32_BASE(mmio_base, SPI_REG_DR);
        --timeout;
        if (timeout < 0)
        {
            printk(KERN_INFO "ERROR: FLASH WRSR BUSY\n");
            goto err;
        }
    }

    /* rdsr35 */
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR0, 0x00070300);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR1, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000001);
    SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000035);
    SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0x08, 0x08, timeout);

    val = SPI_R32_BASE(mmio_base, SPI_REG_DR);
    while ((val & 0xff) != 0x2)
    {
        SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x00000035);
        SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0x08, 0x08, timeout);
        val = SPI_R32_BASE(mmio_base, SPI_REG_DR);
        --timeout;
        if (timeout <= 0)
        {
            printk(KERN_INFO "ERROR: QSPI MODE(FLASH) NOT ENABLED\n");
            goto err;
        }

    }

    /* read data */
    /* move into the following while block in kernel but libutil_ljm.so
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000000);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR0, 0x005F0200);
    SPI_W32_BASE(mmio_base, SPI_REG_CTRLR1, 0x000000FF);
    SPI_W32_BASE(mmio_base, SPI_REG_RX_SAMPLE_DLY, 0x00000001);
    SPI_W32_BASE(mmio_base, SPI_REG_SPI_CTRLR0, 0x00004218);
    SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000001);
    */

    left_size = size;
    r01 = LJM_MISC_R32_BASE(mmio_base, MISC_CHIP_R01);
    if ((r01 & 0xffff0000) == 0x02010000) /* GP201 */
        fifo_size = 1024;
    else
        fifo_size = 128;

    while (left_size > 0)
    {
        /* FIXME: libutil_ljm库只需要配置一次这6个寄存器，但是内核驱动必须每次读取时都配置一次，为什么? */
        SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000000);
        SPI_W32_BASE(mmio_base, SPI_REG_CTRLR0, 0x005F0200);
        SPI_W32_BASE(mmio_base, SPI_REG_CTRLR1, 0x000000FF);
        SPI_W32_BASE(mmio_base, SPI_REG_RX_SAMPLE_DLY, 0x00000001);
        SPI_W32_BASE(mmio_base, SPI_REG_SPI_CTRLR0, 0x00004218);
        SPI_W32_BASE(mmio_base, SPI_REG_SSIENR, 0x00000001);

        if (left_size > fifo_size)
        {
            cur_size = fifo_size;
            left_size -= fifo_size;
        }
        else
        {
            cur_size  = left_size;
            left_size = 0;
        }

        SPI_W32_BASE(mmio_base, SPI_REG_RXFTLR, (cur_size + 3) / 4 - 1);

        SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000000);
        SPI_W32_BASE(mmio_base, SPI_REG_DR, 0x0000006b);
        SPI_W32_BASE(mmio_base, SPI_REG_DR, addr);
        SPI_W32_BASE(mmio_base, SPI_REG_SER, 0x00000001);
        SPI_CHECK_VAL_TIMEOUT(SPI_REG_SR, val, 0x10, 0x10, timeout);

        for (i = 0; i < cur_size / 4; i++)
        {
            val = SPI_R32_BASE(mmio_base, SPI_REG_DR);

            *buf++ = (char)((val & 0xff000000) >> 24);
            *buf++ = (char)((val & 0x00ff0000) >> 16);
            *buf++ = (char)((val & 0x0000ff00) >> 8);
            *buf++ = (char)((val & 0x000000ff) >> 0);
        }

        if ((cur_size % 4) > 0)
        {
            val = SPI_R32_BASE(mmio_base, SPI_REG_DR);
            for (i = 0; i < (cur_size % 4); i++)
            {
                *buf++ = (char)((val & (0xff000000 >> 8 * i)) >> (24 - 8 * i));
            }
        }

        addr += cur_size;
    }

    FOOTER();
    return 0;
err:
    DRM_INFO("SPI READ READ ERROR\n");
    FOOTER();
    return -ENXIO;
}


