/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_vgav2.c
 *  版 本 号: V1.0
 *  作    者: 胡艳明，刘天玥
 *  生成日期: 2022年4月15日
 *
 *  功能描述: CRTC V1驱动
 *
 *  修改历史:
 *         1. 日    期: 2022年5月5日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 创建文件
 *
 ******************************************************************************/

#include "ljmicro_vgav2.h"
#include "ljmicro_device.h"
//#include "ljmicro_clk.h"
#include "gp200.h"
#include "ljmicro_device.h"
struct ljm_device *get_ljm_dev(void);

/* Zone used for header/footer. */
#define _LJ_OBJ_ZONE ljvZONE_KMS

/* 定序器寄存器组 */
static struct ljm_init_reg gp200_init_sr[] = {
    { 0x01, 0x01 }, /* [5]=0 => 正常屏幕操作, [5]=1 => 关闭该路显示, 将不会读显存 */
    { 0x02, 0x0f },
    { 0x03, 0x00 },
    { 0x04, 0x0e }
};

/* 图形控制器寄存器组 */
static struct ljm_init_reg gp200_init_gr[] = {
    { 0x00, 0x00 },
    { 0x01, 0x00 },
    { 0x02, 0x00 },
    { 0x03, 0x00 },
    { 0x04, 0x00 },
    { 0x05, 0x40 },
    { 0x06, 0x05 },
    { 0x07, 0x0f },
    { 0x08, 0xff }
};

/* 属性控制器寄存器组 */
static struct ljm_init_reg gp200_init_ar[] = {
    { 0x00, 0x00 },
    { 0x01, 0x01 },
    { 0x02, 0x02 },
    { 0x03, 0x03 },
    { 0x04, 0x04 },
    { 0x05, 0x05 },
    { 0x06, 0x06 },
    { 0x07, 0x07 },
    { 0x08, 0x08 },
    { 0x09, 0x09 },
    { 0x0a, 0x0a },
    { 0x0b, 0x0b },
    { 0x0c, 0x0c },
    { 0x0d, 0x0d },
    { 0x0e, 0x0e },
    { 0x0f, 0x0f },
    { 0x10, 0x01 },
    { 0x11, 0x00 },
    { 0x12, 0x0f },
    { 0x13, 0x00 },
    { 0x14, 0x00 }
};

/* FIXME: 调色板寄存器(Palette Register DAC接口) */

/* FIXME: init monitor 800 * 600 */
#define GP200_INIT_CR_00_18 { 0x00, 0x5f },\
                            { 0x01, 0x4f },\
                            { 0x02, 0x50 },\
                            { 0x03, 0x82 },\
                            { 0x04, 0x54 },\
                            { 0x05, 0x80 },\
                            { 0x06, 0xbf },\
                            { 0x07, 0x1f },\
                            { 0x08, 0x00 },\
                            { 0x09, 0x41 },\
                            { 0x0a, 0x00 },\
                            { 0x0b, 0x00 },\
/*                            { 0x0c, 0x00 },*/\
/*                            { 0x0d, 0x00 },*/\
                            { 0x0e, 0x00 },\
                            { 0x0f, 0x00 },\
                            { 0x10, 0x9c },\
                            { 0x11, 0x8e },\
                            { 0x12, 0x8f },\
/*                            { 0x13, 0x28 },*/\
                            { 0x14, 0x40 },\
                            { 0x15, 0x96 },\
                            { 0x16, 0xb9 },\
                            { 0x17, 0xa3 },\
                            { 0x18, 0xff },

/* set VBE mode, disable and clear interupts */
#define GP200_INIT_CR_70_80 { 0x73, 0x80 },\
                            { 0x74, 0x01 },\
                            { 0x80, 0x80 },\
                            { 0x79, 0x08 },\
                            { 0x89, 0x00 },\
                            { 0x88, 0x02 },\
                            { 0x8b, 0x00 },\
                            { 0x8a, 0x02 },\
                            { 0x98, 0x03 }


/* CRTC及其扩展寄存器组(GP201卡) */
static struct ljm_init_reg gp200_init_cr[] = {
    /* CRTC registers */
    GP200_INIT_CR_00_18

    /* CRTC extended registers */
    GP200_INIT_CR_70_80
};

typedef struct _fpga_pll_argument {
    unsigned short w;
    unsigned short h;
    unsigned int pclk;
    unsigned char reg_pll[3];
} fpga_pll_argument;

static const fpga_pll_argument fpga_mode_pll[] = {
#if FPGA_HAPS
    {800,  600,  40000,  {25,  10, 0x1}},
    {1024, 768,  65000,  {20,  13, 0x1}},
    {1920, 1080, 148500, { 5,  37, 0x5}}
#else
    {3840, 2160, 297000, { 5,  74, 0x5}},   /* 297M */
    {1920, 1080, 148500, { 5,  37, 0x5}},   /* 148M */
    {1280, 1024, 108000, {10,  11, 0x1}},   /* 110M */
    {1024, 768,  65000,  {20,  13, 0x1}},
    {800,  600,  40000,  {25,  10, 0x1}},
    {720,  400,  35500,  {20,   7, 0x1}}    /* 35M */
#endif
};
static unsigned int fpga_mode_pll_count = sizeof(fpga_mode_pll) / sizeof(fpga_pll_argument);

#define CRTCX_INIT(CARD, X) do {    \
        u8 data = REG_R8(CARD, CRTC1_BW_ERR_ST_ADR + CRTCx_REG_SIZE*(X-1)); \
        if (data)   \
            DEBUG_PRINT("CRTC %d: found bandwidth error(0x%02X)\n", X, data);   \
        /* disable vblank interrupt */  \
        REG_W8(CARD, CRTC1_DISP_CTRL + CRTCx_REG_SIZE*(X-1), RREG8(CRTC1_DISP_CTRL + CRTCx_REG_SIZE*(X-1)) & 0x1f); \
        /* clear vblank interrupt status */ \
        REG_W8(CARD, CRTC1_HV_ST_ADR + CRTCx_REG_SIZE*(X-1), 0x08); \
        /* disable hpd interrupt, disable bandwidth error, enabe frame counter */   \
        REG_W8(CARD, CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE*(X-1), 0x08);    \
        /* clear hpd interrupt status */    \
        REG_W8(CARD, CRTC1_HPD_ST_ADR + CRTCx_REG_SIZE*(X-1), 0x02);    \
        /* clear bandwidth error */ \
        REG_W8(CARD, CRTC1_BW_ERR_ST_ADR + CRTCx_REG_SIZE*(X-1), 0x03); \
    } while(0)

/*****************************************************************************
 *  函 数 名: ljm_vgav2_init
 *  功能描述: 初始化VGA控制器
 *  输入参数: int card  显卡序号
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月5日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_vgav2_init(int card, int count)
{
    int i;
    u8  data;
    struct ljm_device *ljm_dev = get_ljm_dev();

    HEADER_ARG("card=%d count=%d", card, count);

    data = REG_R8(card, VGAREG_READ_MISC_OUTPUT);
    /**
     * FIXME:
     * 使用`(data & 0x3) == 0x3`判断硬件是否在bios阶段初始化不够完备。
     * 在集显和GP201共存的情况下，集显的bios有可能改变该寄存器值。恰好使条件满足时
     * 程序错误地跳过vga初始化。
     * 使用更完善的判断条件以避免该问题。
     *
    */
    if (ljm_dev->once_edp)
    {
        if ((data & 0x3) == 0x3) //判断硬件是否在bios阶段已被初始化过
        {
            ljm_dev->once_edp_vga_inited = true;
            DEBUG_PRINT("eDP is initialized in VBIOS\n");
            FOOTER();
            return 0;
        }
        else
        {
            /**
             * REG[HDMI_DAC_DVO_OUT_MUX] = 0x16161401:
             * CRTC0 --> DP0 & others
             * CRTC1 --> DP1 VGA0 VGA1 VGA2 VGA3
             * CRTC2 --> DP2
             * CRTC3 --> DP3
             * bios_edp模式下，每个edp对应固定的crtc。
            */
            ljm_dev->once_edp_vga_inited = false;
            REG_W32(card, HDMI_DAC_DVO_OUT_MUX, 0x16161401);
            DEBUG_PRINT("eDP is not initialized in VBIOS\n");
        }
    }
    {
        INFO_PRINT("display controller init...");

        /* set miscellaneous output reg. [0]=1 => color mode, [1]=1 => enable */
        REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, 0x3);

        /* set attribute ctl */
        for (i = 0; i < ARRAY_SIZE(gp200_init_ar); ++i)
        {
            AR_WR(card, gp200_init_ar[i].addr, gp200_init_ar[i].data);
        }

        /* set sequencer ctl */
        SR_WR(card, 0, 0x1);
        for (i = 0; i < ARRAY_SIZE(gp200_init_sr); ++i)
        {
            SR_WR(card, gp200_init_sr[i].addr, gp200_init_sr[i].data);
        }
        SR_WR(card, 0, 0x3);

        /* set graphics ctl */
        for (i = 0; i < ARRAY_SIZE(gp200_init_gr); ++i)
        {
            GR_WR(card, gp200_init_gr[i].addr, gp200_init_gr[i].data);
        }

        /* disable crtc write protection */
        CR_WR(card, 0x11, 0x0);
        /* set VBE mode */
        CR_WR(card, CRTC_SEL_VGA_VBE, 0x1);

        data = CR_RD(card, CRTC_BW_ERR_ST_ADR);
        if (data)
            DEBUG_PRINT("CRTC 0: found bandwidth error(0x%02X)\n", data);

        /* set crtc regs */
        for (i = 0; i < ARRAY_SIZE(gp200_init_cr); ++i)
        {
            CR_WR(card, gp200_init_cr[i].addr, gp200_init_cr[i].data);
        }

        TRACE("CRTC 0: now bandwidth status is 0x%02X\n", CR_RD(card, CRTC_BW_ERR_ST_ADR));

#if 0
        /* CRTC间隔相同时间读显存 */
        REG_W32(card, CRTC_PROG_RESET_ADDR, 0x10100);
        udelay(5);
        REG_W32(card, CRTC_PROG_RESET_ADDR, 0x100);
#endif

        /**
         * CRTC1
         */
        if (count > 1)
            CRTCX_INIT(card, 1);

        /**
         * CRTC2
         */
        if (count > 2)
            CRTCX_INIT(card, 2);

        /**
         * CRTC3
         */
        if (count > 3)
            CRTCX_INIT(card, 3);

        INFO_PRINT("done\n");
    }

    FOOTER();
    return 0;
}

// depend on PLL IP spec
//Fref: 7~600MHz
#define REF_MAX         600
#define REF_MIN         7
//Fout: 23~3000MHz
#define OUT_MAX         3000
#define OUT_MIN         23
// Fvco:3000~6000MHz
#define VCO_MAX         6000
#define VCO_MIN         3000
// Fpfd:7~200MHz
#define PFD_MAX         200
#define PFD_MIN         7
// DIVF[8:0]: 1~512
#define M_MAX           512
#define M_MIN           1
// DIVR[5:0]: 1~64
#define N_MAX           64
#define N_MIN           1
// DIVQ[2:0]: 2,4,8,16,32,64,128
#define NO_MAX          7
#define NO_MIN          1
// DIVQ2[3:0]: 1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16
#define NO2_MAX         16
#define NO2_MIN         1
// RANGE[2:0]: 7~11, 11~18, 18-30, 30-50, 50-80, 80-130, 130-200
#define LPF_RANGE1      11
#define LPF_RANGE2      18
#define LPF_RANGE3      30
#define LPF_RANGE4      50
#define LPF_RANGE5      80
#define LPF_RANGE6      130

#define ABS_DELTA(X, Y) (((X) >= (Y)) ? ((X) - (Y)) : ((Y) - (X)))

/**
 * ljm_vgav2_pllcalc: gp201 calc pll para
 *
 * @input param - input_freq : 7000KHz~600000KHz   example:27000 (Khz)
 * @input param - no2_range : 1~16    example:16
 * @input param - output_freq : 23000~3000000 KHz ,example: 25172
 * @output param - pll_para: para_value of 32 bit for register
 * @return - 0: succeed, other: error
*/
int ljm_vgav2_pllcalc(unsigned int input_freq, unsigned int no2_range, unsigned int output_freq, unsigned int *pll_para)
{
    unsigned int best_para = 0;
    unsigned int m, n, range, no, no2, no_pow, para;
    unsigned int best_m = 0, best_n = 0, best_no = 0, best_no_pow = 0, best_no2 = NO2_MAX, best_range = 0;
    unsigned int pfd;
    unsigned int xin = input_freq ;
    unsigned int vco;
    unsigned int fout;
    unsigned int delta = 1000;
    unsigned int best_fout = 0;
    int found = 0;

    if (xin > (REF_MAX*1000) || xin < (PFD_MIN*1000))
    {
        ERR_PRINT("input frequency %d is out of range\n", xin);
        return -1;
    }
    if (output_freq >= 1000*1000) /*当输出频率大于1000Khz时，误差范围为2000*/
    {
        delta = 2000;
    }
    for (m=M_MIN; m<=M_MAX; m++)
    {
        for (n=N_MIN; n<=N_MAX; n++)
        {
            pfd = xin/n;
            if (pfd>=(PFD_MIN*1000) && pfd<=(PFD_MAX*1000) )
            {
                if (pfd <= LPF_RANGE1*1000)      range=1;
                else if (pfd <= LPF_RANGE2*1000) range=2;
                else if (pfd <= LPF_RANGE3*1000) range=3;
                else if (pfd <= LPF_RANGE4*1000) range=4;
                else if (pfd <= LPF_RANGE5*1000) range=5;
                else if (pfd <= LPF_RANGE6*1000) range=6;
                else                             range=7;
                vco = xin * m * 2 / n;
                if (vco >= (VCO_MIN*1000) && vco <= (VCO_MAX*1000))
                {
                    for (no=NO_MIN; no<=NO_MAX; no++)
                    {
                        for (no2=NO2_MIN; no2<=no2_range; no2++)
                        {
                            no_pow = 1<<no ;
                            fout = vco / no_pow / no2 ;
                            para = range + no*16 + (n-1)*256 + (m-1)*0x10000 +  (no2-1)*0x10000000;
                            if ((delta > ABS_DELTA(fout, output_freq)) || ((delta == ABS_DELTA(fout, output_freq)) && best_no2 > no2))
                            {
                                found = 1;
                                delta = ABS_DELTA(fout, output_freq);
                                best_m = m ;
                                best_n = n;
                                best_no = no;
                                best_no2 = no2;
                                best_no_pow = no_pow;
                                best_range = range;
                                best_fout = fout;
                                best_para = para;
                            }
                        }
                    }
                }
            }
        }
    }

    if (pll_para)
        *pll_para = best_para;

    if (found == 0) {
        return -2;
    }
    else {
        DEBUG_PRINT("m=%d n=%d no=%d no2=%d no_pow=%d range=%d fout=%d",
            best_m, best_n, best_no, best_no2, best_no_pow, best_range, best_fout);
        return 0;
    }
}

/*****************************************************************************
 *  函 数 名: ljm_vga_crtc0_set_black
 *  功能描述: 设置CRTC0是否输出黑色，输出黑色时此时不读显存
 *  输入参数: int card     显卡序号
 *            bool enable  是否进入输出黑色模式使能
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_vga_crtc0_set_black(int card, bool enable)
{
    if (enable)
        SR_WR(card, 0x1, 0x21); /* disable RGB output */
    else
        SR_WR(card, 0x1, 0x1); /* enable RGB output */
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtc0_mode_set
 *  功能描述: CRTC0显示模式设置
 *  输入参数: int card                             显卡序号
 *            const struct drm_display_mode *mode  显示模式
 *            is_fpga                              是否是FPGA
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月5日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int ljm_vgav2_crtc0_mode_set(int card, const struct drm_display_mode *mode, bool is_fpga)
{
    int w = mode->hdisplay;
    int h = mode->vdisplay;
    int pclk = mode->clock;
    int r = 0;
    int pllN = 0, pllM = 0, pllOD = 0, div = 0;
    unsigned char miscOutput;
    int chosen, i;
    u32 pll_state;
    bool found = false;
    u32 pll_para;

    HEADER_ARG("card=%d mode=0x%p", card, mode);

    if (is_fpga) {
        for (chosen = 0; chosen < fpga_mode_pll_count; chosen++) {
            if((fpga_mode_pll[chosen].w == mode->hdisplay) && (fpga_mode_pll[chosen].h == mode->vdisplay) && (fpga_mode_pll[chosen].pclk == mode->clock)) {
                pllN = fpga_mode_pll[chosen].reg_pll[2];
                pllM = fpga_mode_pll[chosen].reg_pll[1];
                pllOD = fpga_mode_pll[chosen].reg_pll[0];
                found = true;
                break;
            }
        }
        if (!found) {
            ERR_PRINT("CRTC 0: Resolution %dx%d@%dkHz is not supported!!!\n", w, h, pclk);
            goto out;
        }
    } else {
        r = ljm_vgav2_pllcalc(27000, 16, pclk, &pll_para);
        if (r) {
            ERR_PRINT("CRTC 0: Resolution %dx%d@%dkHz is not supported!!!\n", w, h, pclk);
            goto out;
        } else {
            DEBUG_PRINT("crtc 0: pll_para=0x%x\n", pll_para);
        }
    }

    DEBUG_PRINT("CRTC PLL: N=%d M=%d OD=%d div=%d\n", pllN, pllM, pllOD, div);
    TRACE("CRTC 0 set: (%dx%d@%d), PLL N=%d M=%d OD=%d div=%d", mode->hdisplay, mode->vdisplay, mode->clock, pllN, pllM, pllOD, div);

    CR_WR(card, CRTC_TIMING_HTOTAL_HIGH, ((mode->htotal & 0xff00) >> 8));
    CR_WR(card, CRTC_TIMING_HTOTAL_LOW,  (mode->htotal & 0xff));
    CR_WR(card, CRTC_TIMING_HDISPLAY_HIGH, ((mode->hdisplay & 0xff00) >> 8));
    CR_WR(card, CRTC_TIMING_HDISPLAY_LOW,  (mode->hdisplay & 0xff));
    CR_WR(card, CRTC_TIMING_HRIGHTBORDER, 0);  /* set 8 if 640x480 ? */
    CR_WR(card, CRTC_TIMING_HLEFTBORDER,  0);  /* set 8 if 640x480 ? */
    CR_WR(card, CRTC_TIMING_HSYNC_LOW, ((mode->hsync_end - mode->hsync_start) & 0xff));
    CR_WR(card, CRTC_TIMING_HBACKPORCH_LOW, ((mode->htotal - mode->hsync_end) & 0xff));
    CR_WR(card, CRTC_TIMING_HSYNC_BACK_HIGH, ((((mode->hsync_end - mode->hsync_start) & 0xf00)>>4)
                                      | (((mode->htotal - mode->hsync_end) & 0xf00)>>8)));

    CR_WR(card, CRTC_TIMING_VTOTAL_HIGH, ((mode->vtotal & 0xff00) >> 8));
    CR_WR(card, CRTC_TIMING_VTOTAL_LOW,  (mode->vtotal & 0xff));
    CR_WR(card, CRTC_TIMING_VDISPLAY_HIGH, ((mode->vdisplay & 0xff00) >> 8));
    CR_WR(card, CRTC_TIMING_VDISPLAY_LOW,  (mode->vdisplay & 0xff));
    CR_WR(card, CRTC_TIMING_VBOTTOMBORDER, 0);
    CR_WR(card, CRTC_TIMING_VTOPBORDER,  0);
    CR_WR(card, CRTC_TIMING_VSYNC_LOW, ((mode->vsync_end - mode->vsync_start) & 0xff));
    CR_WR(card, CRTC_TIMING_VBACKPORCH_LOW, ((mode->vtotal - mode->vsync_end) & 0xff));

    if (is_fpga) { /* FPGA board */
        CR_WR(card, CRTC_TIMING_PLL_OD_COUNT, (pllOD & 0xff));
        CR_WR(card, CRTC_TIMING_PLL_M_COUNT, (pllM & 0xff));
        CR_WR(card, CRTC_TIMING_PLL_N_COUNT, (pllN & 0xff));
        CR_WR(card, CRTC_TIMING_PLL_DIVISION, (div & 0x1));
    } else {    /* GP201 */
        CR_WR(card, CRTC_TIMING_PLL_OD_COUNT, (pll_para >> 0)  & 0xff);
        CR_WR(card, CRTC_TIMING_PLL_M_COUNT,  (pll_para >> 8)  & 0xff);
        CR_WR(card, CRTC_TIMING_PLL_N_COUNT,  (pll_para >> 16) & 0xff);
        CR_WR(card, CRTC_TIMING_PLL_DIVISION, (pll_para >> 24) & 0xff);
    }

    miscOutput = RREG8(VGAREG_READ_MISC_OUTPUT);
    if (mode->flags & DRM_MODE_FLAG_PHSYNC) miscOutput &= 0xBF;
    if (mode->flags & DRM_MODE_FLAG_NHSYNC) miscOutput |= 0x40;
    if (mode->flags & DRM_MODE_FLAG_PVSYNC) miscOutput &= 0x7F;
    if (mode->flags & DRM_MODE_FLAG_NVSYNC) miscOutput |= 0x80;
    REG_W8(card, VGAREG_WRITE_MISC_OUTPUT, miscOutput);

    i = 10;
    pll_state = LJM_CRU_R32(card, CRU_STATUS);
    while (!(pll_state & 0x800000))
    {
        usleep_range(10, 30);
        pll_state = LJM_CRU_R32(card, CRU_STATUS);
        i--;
        if (i < 0)
        {
            DRM_ERROR("CRTC0 PXLPLL lock error\n");
            break;
        }
    }

out:
    if (is_fpga) {
        FOOTER_ARG("found = %d", found);
    }
    else {
        FOOTER_ARG("r=%d", r);
    }
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtcx_mode_set
 *  功能描述: CRTC1~3显示模式设置
 *  输入参数: int card                             显卡序号
 *            int crtc_id                          CRTC编号
 *            const struct drm_display_mode *mode  显示模式
 *            boot is_fpga                         是否是FPGA板
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月6日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int ljm_vgav2_crtcx_mode_set(int card, int crtc_id, const struct drm_display_mode *mode, bool is_fpga)
{
    int r = 0;
    int w = mode->hdisplay;
    int h = mode->vdisplay;
    int pclk = mode->clock;
    int pllN = 0, pllM = 0, pllOD = 0, div = 0;
    int chosen, i;
    bool found = false;
    u32 pll_para, pll_state;
    u8 tmp8;

    HEADER_ARG("card=%d crtc_id=%d mode=0x%p is_fpag=%d", card, crtc_id, mode, is_fpga);

    if (crtc_id < 1 || crtc_id > 3)
        ERR_PRINT("%s: id %d is out of range\n", __FUNCTION__, crtc_id);

    if (is_fpga) {
        for (chosen = 0; chosen < fpga_mode_pll_count; chosen++) {
            if((fpga_mode_pll[chosen].w == mode->hdisplay) && (fpga_mode_pll[chosen].h == mode->vdisplay) && (fpga_mode_pll[chosen].pclk == mode->clock)) {
                pllN = fpga_mode_pll[chosen].reg_pll[2];
                pllM = fpga_mode_pll[chosen].reg_pll[1];
                pllOD = fpga_mode_pll[chosen].reg_pll[0];
                found = true;
                break;
            }
        }
        if (!found) {
            ERR_PRINT("CRTC 1: Resolution %dx%d@%dkHz is not supported!!!\n", w, h, pclk);
            goto out;
        }
    }
    else {
        r = ljm_vgav2_pllcalc(27000, 16, pclk, &pll_para);
        if (r) {
            ERR_PRINT("CRTC %d: Resolution %dx%d@%dkHz is not supported!!!\n", crtc_id, w, h, pclk);
            goto out;
        } else {
            DEBUG_PRINT("crtc %d: pll_para=0x%x\n", crtc_id, pll_para);
        }
    }

    TRACE("CRTC %d set: (%dx%d@%d), PLL N=%d M=%d OD=%d div=%d", crtc_id, mode->hdisplay, mode->vdisplay, mode->clock, pllN, pllM, pllOD, div);

    REG_W16(card, CRTC1_HTOTAL + CRTCx_REG_SIZE * (crtc_id-1), mode->htotal & 0xffff);
    REG_W16(card, CRTC1_HDISPALY + CRTCx_REG_SIZE * (crtc_id-1), mode->hdisplay & 0xffff);
    REG_W8(card, CRTC1_LEFT_BORDER + CRTCx_REG_SIZE * (crtc_id-1), 0);
    REG_W8(card, CRTC1_RIGHT_BORDER + CRTCx_REG_SIZE * (crtc_id-1), 0);
    REG_W16(card, CRTC1_HSYNC + CRTCx_REG_SIZE * (crtc_id-1), (mode->hsync_end - mode->hsync_start) & 0xffff);
    REG_W16(card, CRTC1_BACK_PORCH_H + CRTCx_REG_SIZE * (crtc_id-1), (mode->htotal - mode->hsync_end) & 0xffff);
    REG_W16(card, CRTC1_VTOTAL + CRTCx_REG_SIZE * (crtc_id-1), mode->vtotal & 0xffff);
    REG_W16(card, CRTC1_VDISPLAY + CRTCx_REG_SIZE * (crtc_id-1), mode->vdisplay & 0xffff);
    REG_W8(card, CRTC1_VSYNC + CRTCx_REG_SIZE * (crtc_id-1), (mode->vsync_end - mode->vsync_start) & 0xff);
    REG_W8(card, CRTC1_BACK_PORCH_V + CRTCx_REG_SIZE * (crtc_id-1), (mode->vtotal - mode->vsync_end) & 0xff);
    if (is_fpga) { /* FPGA board */
        REG_W32(card, CRTC1_PXL_CLK_SEL + CRTCx_REG_SIZE * (crtc_id-1), ((div & 0x1)<<24) | ((pllN & 0xff)<<16) | ((pllM & 0xff)<<8) | (pllOD & 0xff));
    } else {    /* GP201 */
        REG_W32(card, CRTC1_PXL_CLK_SEL + CRTCx_REG_SIZE * (crtc_id-1), pll_para);
    }

    tmp8 = REG_R8(card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id-1));
    if (mode->flags & DRM_MODE_FLAG_PHSYNC) tmp8 &= 0xf7;
    if (mode->flags & DRM_MODE_FLAG_NHSYNC) tmp8 |= 0x08;
    if (mode->flags & DRM_MODE_FLAG_PVSYNC) tmp8 &= 0xef;
    if (mode->flags & DRM_MODE_FLAG_NVSYNC) tmp8 |= 0x10;
    REG_W8(card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id-1), tmp8);

    i = 10;
    pll_state = LJM_CRU_R32(card, CRU_STATUS);
    while (!(pll_state & (0x800000 >> crtc_id)))
    {
        usleep_range(10, 30);
        pll_state = LJM_CRU_R32(card, CRU_STATUS);
        i--;
        if (i < 0)
        {
            DRM_ERROR("CRTC%d PXLPLL lock error\n", crtc_id);
            break;
        }
    }
out:
    if (is_fpga) {
        FOOTER_ARG("found = %d", found);
    }
    else {
        FOOTER_ARG("r=%d", r);
    }
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtc_mode_set
 *  功能描述: CRTC1显示模式设置
 *  输入参数: int card                             显卡序号
 *            int crtc_id                          crtc序号
 *            const struct drm_display_mode *mode  显示模式
 *            boot is_fpga                         是否是FPGA板
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_vgav2_crtc_mode_set(int card, int crtc_id, const struct drm_display_mode *mode, bool is_fpga)
{
    int r;

    HEADER_ARG("card=%d crtc_id=%d mode=%p is_fpga=%d", card, crtc_id, mode, is_fpga);

    /* choose a function to set mode according to crtc->id */
    switch (crtc_id) {
        case 0:
            r = ljm_vgav2_crtc0_mode_set(card, mode, is_fpga);
            break;
        case 1:
        case 2:
        case 3:
            r = ljm_vgav2_crtcx_mode_set(card, crtc_id, mode, is_fpga);
            break;
        default:
            ERR_PRINT("crtc %d is out of range\n", crtc_id);
            break;
    }

    if (r) {
        ERR_PRINT("ERR: _crtc_mode_set\n");
        FOOTER_ARG("r=%d", r);
        return r;
    }

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtc_enable
 *  功能描述: 使能CRTC显示
 *  输入参数: int card     显卡序号
 *            int crtc_id  CRTC序号
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年5月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_vgav2_crtc_enable(int card, int crtc_id)
{
    HEADER_ARG("card=%d crtc_id=%d", card, crtc_id);

    switch (crtc_id) {
        case 0:
            CR_WR(card, CRTC_SYNC_CTRL, (CR_RD(card, CRTC_SYNC_CTRL) | 0x80));
            ljm_vga_crtc0_set_black(card, false);
            CR_WR(card, CRTC_DISP_CTRL_ADR, CR_RD(card, CRTC_DISP_CTRL_ADR) & 0xfc);
            break;
        case 1:
        case 2:
        case 3:
            REG_W8(card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1),
                (RREG8(CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1)) & 0xf8) | (1<<0));
            break;
        default:
            ERR_PRINT("crtc id (%d) is out of range\n", crtc_id);
            break;
    }

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtc_disable
 *  功能描述: 禁用CRTC显示
 *  输入参数: int card     显卡序号
 *            int crtc_id  CRTC序号
 *  输出参数: 无
 *  返 回 值: void
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
void ljm_vgav2_crtc_disable(int card, int crtc_id)
{
    HEADER_ARG("card=%d crtc_id=%d", card, crtc_id);

    switch (crtc_id) {
        case 0:
            CR_WR(card, CRTC_DISP_CTRL_ADR, CR_RD(card, CRTC_DISP_CTRL_ADR) | (1<<0 | 1<<1));
            ljm_vga_crtc0_set_black(card, true);
            CR_WR(card, CRTC_SYNC_CTRL, (CR_RD(card, CRTC_SYNC_CTRL) & 0x7f));
            break;
        case 1:
        case 2:
        case 3:
            REG_W8(card, CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1),
                RREG8(CRTC1_DISP_CTRL + CRTCx_REG_SIZE * (crtc_id - 1)) & 0xFE);
            break;
        default:
            ERR_PRINT("crtc id (%d) is out of range\n", crtc_id);
            break;
    }

    FOOTER();
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtc_set_base
 *  功能描述: 设置CRTC显示的基地址
 *  输入参数: int card       卡序号
 *            int crtc_id    CRTC序号
 *            u64 crtc_base  显存偏移地址
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2022年5月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_vgav2_crtc_set_base(int card, int crtc_id, u64 crtc_base)
{
    u32 base;
    u8 val;
    struct ljm_device *ljm_dev = get_ljm_devs(card);

    //HEADER_ARG("card=%d crtc_id=%d crtc_base=0x%lx", card, crtc_id, (unsigned long)crtc_base);

    crtc_base /= 4; /* 32-bit address */
    base = (u32)crtc_base;
    if (crtc_id == 0) {
        if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
            val = REG_R8(card, CRTC_FRAME_BUFF_SEL);
            if (val & 0x1) {
                /* set base 0 address */
                CR_WR(card, CRTC_BASE_HIGHEST, (base & 0xff000000) >> 24);
                CR_WR(card, CRTC_BASE_HIGH, (base & 0x00ff0000) >> 16);
                CR_WR(card, CRTC_BASE_LOW, (base & 0xff00) >> 8);
                CR_WR(card, CRTC_BASE_LOWEST, base & 0xff); /* lowest address must be the last byte, by hym */
                val &= 0xfe;
            } else {
                /* set base 1 address */
                CR_WR(card, CRTC_BASE1_HIGHEST, (base & 0xff000000) >> 24);
                CR_WR(card, CRTC_BASE1_HIGH, (base & 0x00ff0000) >> 16);
                CR_WR(card, CRTC_BASE1_LOW, (base & 0xff00) >> 8);
                CR_WR(card, CRTC_BASE1_LOWEST, base & 0xff);
                val |= 0x1;
            }
            REG_W8(card, CRTC_FRAME_BUFF_SEL, val);
        } else {
            /* set base 0 address */
            CR_WR(card, CRTC_BASE_HIGHEST, (base & 0xff000000) >> 24);
            CR_WR(card, CRTC_BASE_HIGH, (base & 0x00ff0000) >> 16);
            CR_WR(card, CRTC_BASE_LOW, (base & 0xff00) >> 8);
            CR_WR(card, CRTC_BASE_LOWEST, base & 0xff); /* lowest address must be the last byte */
        }
    }
    else {
        /* set base address */
        REG_W32(card, CRTC1_START_ADDR + CRTCx_REG_SIZE * (crtc_id - 1), base);
    }

    //FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_vgav2_crtc_mode_set_base
 *  功能描述: 设置CRTC显示的基本设置
 *  输入参数: int card       显卡序号
 *            int crtc_id    CRTC序号
 *            int bpp        每像素比特数
 *            int pitches    每行字节数
 *            u64 crtc_base  显示区域显存偏移地址
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年5月6日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_vgav2_crtc_mode_set_base(int card, int crtc_id, int bpp, int pitches, u64 crtc_base)
{
    int r;

    HEADER_ARG("card=%d crtc_id=%d bpp=%d pitches=%d crtc_base=0x%lx",
        card, crtc_id, bpp, pitches, (unsigned long)crtc_base);

    if (bpp == 24) bpp = 32;    /* bpp 24 is realized to bpp 32 */

    if ((bpp != 16) && (bpp != 32) && (bpp != 15)) {   /* only bpp 16, 24, 32 are supported */
        ERR_PRINT("crtc %d BPP %d is not supported!!!\n", crtc_id, bpp);
        r = -EPERM;
    }

    if ((pitches & 0x7) || (crtc_base & 0x3)) {
        ERR_PRINT("crtc base (%llx) or pitches (%d) alignment is error!", crtc_base, pitches);
        r = -EPERM;
    }

    /* set crtc base address */
    ljm_vgav2_crtc_set_base(card, crtc_id, crtc_base);

    pitches /= 8; // 64 bits address

    switch (crtc_id) {
    case 0:
        /* set bpp */
        switch (bpp) {
            case 8:
                CR_WR(card, CRTC_COLOR_MODE, 1<<3);
                break;
            case 15:
                CR_WR(card, CRTC_COLOR_MODE, 1<<4);
                break;
            case 16:
                CR_WR(card, CRTC_COLOR_MODE, 1<<5);
                break;
            case 32:
                CR_WR(card, CRTC_COLOR_MODE, 1<<7);
                break;
            default:
                break;
        }

        /* set pitches */
        CR_WR(card, CRTC_PITCH_LOW, pitches & 0xff);
        CR_WR(card, CRTC_PITCH_HIGH, (pitches & 0xf00)>>8);
        break;
    case 1:
    case 2:
    case 3:
        /* set bpp, just support 15/16/32bit */
        switch (bpp) {
            case 15:
                REG_W8(card, CRTC1_COLOR_MODE + CRTCx_REG_SIZE * (crtc_id - 1), 1<<0);
                break;
            case 16:
                REG_W8(card, CRTC1_COLOR_MODE + CRTCx_REG_SIZE * (crtc_id - 1), 1<<1);
                break;
            case 32:
                REG_W8(card, CRTC1_COLOR_MODE + CRTCx_REG_SIZE * (crtc_id - 1), 1<<2);
                break;
            default:
                break;
        }

        /* set pitches */
        REG_W16(card, CRTC1_OFFSET_ADDR + CRTCx_REG_SIZE * (crtc_id - 1), pitches);
        break;
    default:
        ERR_PRINT("crtc id (%d) is out of range\n", crtc_id);
        break;
    }

    FOOTER();
    return 0;
}

int ljm_vgav2_crtc_prog_reset(int card, int reg_val)
{
    HEADER_ARG("card=%d reg_val=0x%X", card, reg_val);

    /* CRTC间隔相同时间读显存 */
    REG_W32(card, CRTC_PROG_RESET_ADDR, CRTC_PROG_RESET_BIT | CRTC_PROG_COUNT_MASK(reg_val));
    udelay(7);
    REG_W32(card, CRTC_PROG_RESET_ADDR, CRTC_PROG_COUNT_MASK(reg_val));

    FOOTER();
    return 0;
}
/*****************************************************************************
 *  函 数 名: ljm_vgav2_get_vblank_counter
 *  功能描述: 过去vblank计数值
 *  输入参数: int card
 *            int crtc_id
 *  输出参数: 无
 *  返 回 值: u32
 *
 *  修改历史:
 *         1. 日    期: 2022年5月6日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
u32 ljm_vgav2_get_vblank_counter(int card, int crtc_id)
{
    u32 data, count;

    if (0 == crtc_id) {
        count = CR_RD(card, CRTC_FRAME_CNT_BYTE0);
        data = CR_RD(card, CRTC_FRAME_CNT_BYTE1);
        count += data << 8;
        data = CR_RD(card, CRTC_FRAME_CNT_BYTE2);
        count += data << 16;
        data = CR_RD(card, CRTC_FRAME_CNT_BYTE3);
        count += data << 24;
    }
    else {
        count = REG_R32(card, CRTC1_FRAME_CNT_ADR + CRTCx_REG_SIZE * (crtc_id - 1));
    }
    //TRACE("crtc %d: frame count %u", crtc_id, count);
    return count;
}

bool ljm_vgav2_page_flip_pending(int card, int crtc_id)
{
    u8 data;

    /* CRTC_PFLIP_PENDING_MASK在vsync下降沿时清除 */

    if (0 == crtc_id)
        data = CR_RD(card, CRTC_STATUS_ADR) & CRTC_PFLIP_PENDING_MASK;
    else
        data = REG_R8(card, CRTC1_HV_ST_ADR + CRTCx_REG_SIZE * (crtc_id - 1)) & CRTC_PFLIP_PENDING_MASK;

    /* Return current update_pending status: */
    return !!data;

#if 0
    /* page_flip_pending判断在vsync的上升沿中断中执行，由于vsync的下降沿肯定会清除pflip_pending状态，
     * 故直接返回为false，通过驱动可以写base寄存器 */
    return false;
#endif
}

/*
 * Retrieve current video scanout position of crtc on a given gpu.
 *
 * \param ljm_dev ljm_device to query.
 * \param pipe Crtc index to query.
 * \param *vpos Location where vertical scanout position should be stored.
 * \param *hpos Location where horizontal scanout position should go.
 *
 * Returns vpos as a positive number from vgate start until end of vblank.
 * Returns hpos as a positive number from hgate start until end of hblank.
 *
 * \return Flags, or'ed together as follows:
 *
 * DRM_SCANOUTPOS_VALID = Query successful.
 * DRM_SCANOUTPOS_INVBL = Inside vblank.
 * DRM_SCANOUTPOS_ACCURATE = Returned position is accurate.
 */
int ljm_vgav2_get_crtc_scanout_pos(int card, unsigned int pipe, int *vpos, int *hpos)
{
    u32 pos;
    int v_pos = 0, h_pos = 0;
//  int r = DRM_SCANOUTPOS_ACCURATE | DRM_SCANOUTPOS_VALID;
    int r = 0;

    if (0 == pipe) {
        v_pos = CR_RD(card, CRTC_LINE_CNT_LOW) + (((int)CR_RD(card, CRTC_LINE_CNT_HIGH))<<8);
        h_pos = CR_RD(card, CRTC_PIXEL_CNT_LOW) + (((int)CR_RD(card, CRTC_PIXEL_CNT_HIGH))<<8);
//      if (RCR(CRTC_STATUS_ADR) & CRTC_STATUS_VBLANK_MASK)
//          r |= DRM_SCANOUTPOS_IN_VBLANK;
    }
    else if (pipe > 0 && pipe <= 3) {
        pos = REG_R32(card, CRTC1_SCANOUT_CNT + CRTCx_REG_SIZE * (pipe - 1));
        v_pos = pos & 0xffff;
        h_pos = (pos & 0xffff0000)>>16;
//      if (RREG8(CRTC1_HV_ST_ADR + CRTCx_REG_SIZE * (pipe-1)) & CRTC_STATUS_VBLANK_MASK)
//          r |= DRM_SCANOUTPOS_IN_VBLANK;
    }
    else
        r = 0;

    TRACE("scanout %d: vpos=%d hpos=%d r=%d", pipe, v_pos, h_pos, r);

    if (vpos)
        *vpos = v_pos;
    if (hpos)
        *hpos = h_pos;
    return r;
}

bool ljm_vgav2_crtc_mode_valid(const struct drm_display_mode *mode, bool is_fpga)
{
    int chosen;
    bool found = false;
    bool r = true;

//    HEADER_ARG("mode=0x%p is_fpga=%d", mode, is_fpga);

    if (is_fpga) {
        for (chosen = 0; chosen < fpga_mode_pll_count; chosen++) {
            if((fpga_mode_pll[chosen].w == mode->hdisplay) && (fpga_mode_pll[chosen].h == mode->vdisplay) && (fpga_mode_pll[chosen].pclk == mode->clock)) {
                found = true;
                break;
            }
        }
        if (!found) {
            r = false;
        }
    }

#if 0
    if (r) {
        TRACE("%dx%d@%d is valid", mode->hdisplay, mode->vdisplay, mode->clock);
    } else {
        TRACE("%dx%d@%d is invalid", mode->hdisplay, mode->vdisplay, mode->clock);
    }
#endif

//    FOOTER_ARG("%d", r);
    return r;
}

void ljm_dvo_hpd_enable(int card, int index, int enable)
{
    uint8_t data8;
    struct ljm_device *ljm_dev = get_ljm_dev();

    if(index == 5)    /* DVO8 */
    {
        data8 = CR_RD(card, CRTC_HPD_EN_ADR);
        if (enable)
            data8 |= CRTC_HPD_INT_EN_MASK;
        else
            data8 &= ~CRTC_HPD_INT_EN_MASK;
        CR_WR(card, CRTC_HPD_EN_ADR, data8);
    }
    else if (index == 6)    /* DVO9 */
    {
        data8 = REG_R8(card, CRTC1_HPD_EN_ADR);
        if (enable) {
            data8 |= CRTC_HPD_INT_EN_MASK;
        }
        else {
            data8 &= ~CRTC_HPD_INT_EN_MASK;
        }
        REG_W8(card, CRTC1_HPD_EN_ADR, data8);
    }
    else if (index == 7 && ljm_dev->flags & LJMICRO_IS_FPGA)    /* FPGA: DVO10 */
    {
        data8 = REG_R8(card, CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE);
        if (enable) {
            data8 |= CRTC_HPD_INT_EN_MASK;
        }
        else {
            data8 &= ~CRTC_HPD_INT_EN_MASK;
        }
        REG_W8(card, CRTC1_HPD_EN_ADR + CRTCx_REG_SIZE, data8);
    }

    return;
}

void ljm_dvo_dpms(int card, int index, int mode)
{
    u8 data;

    HEADER_ARG("card=%d index=%d mode=%d", card, index, mode);

    /* TODO: add dpms */
    switch (mode) {
    case DRM_MODE_DPMS_ON:
        /* enable dvo */
        if (index == 0) {
            data = CR_RD(card, CRTC_CRT_MODE_CTRL2) | 0x1;
            CR_WR(card, CRTC_CRT_MODE_CTRL2, data);
        }
        else {
            data = REG_R8(card, CRTC1_DISP_CTRL2 + CRTCx_REG_SIZE * (index - 1)) | 0x1;
            REG_W8(card, CRTC1_DISP_CTRL2 + CRTCx_REG_SIZE * (index - 1), data);
        }
        break;
    case DRM_MODE_DPMS_STANDBY:
    case DRM_MODE_DPMS_SUSPEND:
    case DRM_MODE_DPMS_OFF:
        /* disable dvo */
        if (index == 0) {
            data = CR_RD(card, CRTC_CRT_MODE_CTRL2) & 0xfe;
            CR_WR(card, CRTC_CRT_MODE_CTRL2, data);
        }
        else {
            data = REG_R8(card, CRTC1_DISP_CTRL2 + CRTCx_REG_SIZE * (index - 1)) & 0xfe;
            REG_W8(card, CRTC1_DISP_CTRL2 + CRTCx_REG_SIZE * (index - 1), data);
        }
        break;
    }

    FOOTER();
}