/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，武汉凌久微电子有限公司
 *
 ******************************************************************************
 *
 *  文件名称: ljmicro_vgav2.h
 *  版 本 号: V1.0
 *  作    者: 胡艳明，刘天玥
 *  生成日期: 2022年5月5日
 *
 *  功能描述: CRTCv2驱动
 *
 *  修改历史:
 *         1. 日    期: 2022年5月5日
 *            作    者: 胡艳明，刘天玥
 *            修改内容: 创建文件
 *
 *****************************************************************************/

#ifndef __LJMICRO_VGAV2_H__
#define __LJMICRO_VGAV2_H__

/*******************************************************************************
 *                                   非标准库头文件                            *
 *******************************************************************************/

#include "ljmicro_base.h"

/******************************************************************************
 *                                     宏定义                                 *
 ******************************************************************************/

/**********************************************
 * VGA registers
 ***********************************************/
#define CRTC_CRT_MODE_CTRL              0x17
#define CRTC_CRT_MODE_CTRL2             0x19
#define m_CRTC_CRT_MODE_CTRL2_DVO0      (0x1)
#define m_CRTC_CRT_MODE_CTRL2_ADDRLATCH (0x1 << 1)
#define m_CRTC_CRT_MODE_CTRL2_GAMMA     (0x1 << 2)
#define m_CRTC_CRT_MODE_CTRL2_2BUFFER   (0x1 << 3)

/* extended register of GP201's CRTC */
#define CRTC_TIMING_HTOTAL_HIGH         0x60
#define CRTC_TIMING_HTOTAL_LOW          0x61
#define CRTC_TIMING_HDISPLAY_HIGH       0x62
#define CRTC_TIMING_HDISPLAY_LOW        0x63
#define CRTC_TIMING_HRIGHTBORDER        0x64
#define CRTC_TIMING_HLEFTBORDER         0x65
#define CRTC_TIMING_HSYNC_LOW           0x66
#define CRTC_TIMING_HBACKPORCH_LOW      0x67
#define CRTC_TIMING_VTOTAL_HIGH         0x68
#define CRTC_TIMING_VTOTAL_LOW          0x69
#define CRTC_TIMING_VDISPLAY_HIGH       0x6A
#define CRTC_TIMING_VDISPLAY_LOW        0x6B
#define CRTC_TIMING_VBOTTOMBORDER       0x6C
#define CRTC_TIMING_VTOPBORDER          0x6D
#define CRTC_TIMING_VSYNC_LOW           0x6E
#define CRTC_TIMING_VBACKPORCH_LOW      0x6F
#define CRTC_COLOR_MODE                 0x73    /* D[3]=1: 8 bit color; D[4]=1: 15 bit color; D[5]=1: 16 bit color; D[7]=1: 32 bit color */
#define CRTC_SEL_VGA_VBE                0x74    /* select between VGA and VBE */
#define CRTC_BASE_HIGHEST               0x75    /* base address [31:24], 32-bit address */
#define CRTC_BASE_HIGH                  0x76    /* base address [23:16], 32-bit address */
#define CRTC_PITCH_HIGH                 0x77    /* pitch [11:8], 64 bit address */
#define CRTC_TIMING_HSYNC_BACK_HIGH     0x78
#define CRTC_STATUS_ADR                 0x79    /* [0]: vertical vblank valid, [1]: horizontal vblank valid,
                                                   [2]: vertical display valid, [3]: vertical sync interrupt valid,
                                                   [4]: buffer address is pending, [5]: pixel clock is locked */
#define CRTC_DISP_CTRL_ADR              0x80    /* [0]: disable horizontal synchronization, [1]: disable vertical synchronization,
                                                   [2]: enable vblank interrupt, [3]: enable stereoscopic mode
                                                   [4]: enable real resolution mode, [5]: enable bandwidth error interrupt
                                                   [6]: enable data beforehand mode (not realized yet), [7]: enable frame counter */
#define CRTC_TIMING_PLL_OD_COUNT        0x81
#define CRTC_TIMING_PLL_M_COUNT         0x83
#define CRTC_TIMING_PLL_N_COUNT         0x85
#define CRTC_TIMING_PLL_DIVISION        0x87    /* [0]: 1 => PLL add divider 8 */
#define CRTC_HPD_ST_ADR                 0x88    /* bit 0: connect status(0-disconnect);bit 1: int status(1-int occur, write 1 to clean int)*/
#define CRTC_HPD_EN_ADR                 0x89    /* bit 0: int enable (0-disable, 1-enable) */

#define CRTC_BASE1_LOWEST               0x90
#define CRTC_BASE1_LOW                  0x91
#define CRTC_BASE1_HIGH                 0x92
#define CRTC_BASE1_HIGHEST              0x93

#define CRTC_BW_ERR_ST_ADR              0x98    /* bandwidth error status, [0]: bandwidth error status, [1]: bandwidth error interrupt valid */
#define CRTC_LINE_CNT_LOW               0x9A    /* line counter, low byte */
#define CRTC_LINE_CNT_HIGH              0x9B    /* line counter, high byte */
#define CRTC_FRAME_CNT_BYTE0            0x9C    /* line counter[7:0] */
#define CRTC_FRAME_CNT_BYTE1            0x9D    /* line counter[15:8] */
#define CRTC_FRAME_CNT_BYTE2            0x9E    /* line counter[23:16] */
#define CRTC_FRAME_CNT_BYTE3            0x9F    /* line counter[31:24] */
#define CRTC_PIXEL_CNT_LOW              0xAA    /* pixel counter in a line, low byte */
#define CRTC_PIXEL_CNT_HIGH             0xAB    /* pixel counter in a line, high byte */
#define CRTC_PWM_FAN_FRE_DEVISION_ONE   0XD8    /* PWN FAN one-stage frequency division register 8bit*/
#define CRTC_PWM_FAN_EN                 0XD9    /* PWN FAN enable 8bit*/
#define CRTC_PWM_FAN_FRE_DEVISION_SEC_LOW 0xD4  /* PWN FAN two-stage frequency division register low 8bit*/
#define CRTC_PWM_FAN_FRE_DEVISION_SEC_HIGH 0xD5 /* PWN FAN two-stage frequency division register high 8bit*/
#define CRTC_PWM_FAN_DUTYCYCLE_LOW      0xD6    /* PWN FAN duty cycle register low 8bit */
#define CRTC_PWM_FAN_DUTYCYCLE_HIGH     0xD7    /* PWN FAN duty cycle register high 8bit */

/* hardware misc registers */
#define CRTC_FRAME_BUFF_SEL             (REG_VGA_OFFSET + 0x20)
#define HDMI_DAC_DVO_OUT_MUX            (REG_VGA_OFFSET + 0x24)     /* output port mux */
#define CRTC_PROG_RESET_ADDR            (REG_VGA_OFFSET + 0x28)
#define CRTC_PROG_RESET_BIT             (1 << 16) /* 1: reset   0: run */
#define CRTC_PROG_COUNT_MASK(x)         (0xffff & x)

#define CRTC1_REG_BASE                  REG_VGA_OFFSET
#define CRTCx_REG_SIZE                  0x40

#define CRTC1_HTOTAL                    (CRTC1_REG_BASE+0x040)  /* W(16bit): htotal */
#define CRTC1_HDISPALY                  (CRTC1_REG_BASE+0x042)  /* W */
#define CRTC1_LEFT_BORDER               (CRTC1_REG_BASE+0x044)  /* B(8bit) */
#define CRTC1_RIGHT_BORDER              (CRTC1_REG_BASE+0x045)  /* B */
#define CRTC1_HSYNC                     (CRTC1_REG_BASE+0x048)  /* W */
#define CRTC1_BACK_PORCH_H              (CRTC1_REG_BASE+0x04A)  /* W */
#define CRTC1_VTOTAL                    (CRTC1_REG_BASE+0x04C)  /* W */
#define CRTC1_VDISPLAY                  (CRTC1_REG_BASE+0x04E)  /* W */
#define CRTC1_TOP_BORDER                (CRTC1_REG_BASE+0x050)  /* B */
#define CRTC1_BOTTOM_BORDER             (CRTC1_REG_BASE+0x051)  /* B */
#define CRTC1_VSYNC                     (CRTC1_REG_BASE+0x052)  /* B */
#define CRTC1_BACK_PORCH_V              (CRTC1_REG_BASE+0x053)  /* B */
#define CRTC1_COLOR_MODE                (CRTC1_REG_BASE+0x054)  /* W: b[0]=1: 15 bit color; b[1]=1: 16 bit color; b[2]=1: 32 bit color */
#define CRTC1_OFFSET_ADDR               (CRTC1_REG_BASE+0x056)  /* W: pitch, 64-bit address */
#define CRTC1_START_ADDR                (CRTC1_REG_BASE+0x058)  /* DW(32-bit): start address, 32-bit address */
#define CRTC1_PXL_CLK_SEL               (CRTC1_REG_BASE+0x05C)  /* DW: b[7:0]: OD; b[15:8]: M; b[23:16]: N; b[24]: div8 */
#define CRTC1_DISP_CTRL2                (CRTC1_REG_BASE+0x060)  /* B: [0]:DVO enable; [1]: lock address at hsync start */
#define m_CRTC1_DISP_CTRL2_DVO0         (0x1)
#define m_CRTC1_DISP_CTRL2_ADDRLATCH    (0x1 << 1)
#define m_CRTC1_DISP_CTRL2_GAMMA        (0x1 << 2)
#define m_CRTC1_DISP_CTRL2_2BUFFER      (0x1 << 3)
#define CRTC1_DISP_CTRL                 (CRTC1_REG_BASE+0x064)  /* B: [0]:enable display; [1]: disable hblank, [2]: disable vblank, [3]: h polarity,
                                                                      [4]: v polarity, [5]: vblank int enable, [6]: enable actual mode, [7]: black enable */
#define CRTC1_HV_ST_ADR                 (CRTC1_REG_BASE+0x065)  /* B: [0]: vblank valid, [1]: hactive valid, [2]: hactive valid, [3]: vblank int valid
                                                                      [4]: buffer address is pending, [5]: pixel clock locked */
#define CRTC1_HPD_ST_ADR                (CRTC1_REG_BASE+0x066)  /* B: HPD INT status, [0]: 1=>linked, 0=>unlinked, [1]: 1=> hpd interrupt, write 1 to clear */
#define CRTC1_HPD_EN_ADR                (CRTC1_REG_BASE+0x067)  /* B: [0]: 1=> enable hpd interupt, [1]: 1=> enable bandwdith error interrupt,
                                                                      [2]: 1=> enable high bandwith mode, [3]: 1=> enable frame counter */
#define CRTC1_BW_ERR_ST_ADR             (CRTC1_REG_BASE+0x06C)  /* [0]: 1=> bandwidth error, write to clear, [1]: bandwidth error interrupt status, write to clear */
#define CRTC1_SCANOUT_CNT               (CRTC1_REG_BASE+0x070)  /* [15:0]: line count, [31:16]: pixel count in a line. 0~hgate_total/vgate_total is in active zoom */
#define CRTC1_FRAME_CNT_ADR             (CRTC1_REG_BASE+0x074)  /* frame counter */

#define CRTC_GAMMA_LUT_RWCTL            (CRTC1_REG_BASE+0x200)  /* Bit[0]: write, Bit[1] read*/
#define CRTC_GAMMA_LUT_OFFSET           (0x010)
#define CRTCX_GAMMA_LUTR_ADDR(x)        (CRTC1_REG_BASE+(CRTC_GAMMA_LUT_OFFSET*(x))+0x204)
#define CRTCX_GAMMA_LUTR_DATA(x)        (CRTC1_REG_BASE+(CRTC_GAMMA_LUT_OFFSET*(x))+0x206)
#define CRTCX_GAMMA_LUTG_ADDR(x)        (CRTC1_REG_BASE+(CRTC_GAMMA_LUT_OFFSET*(x))+0x208)
#define CRTCX_GAMMA_LUTG_DATA(x)        (CRTC1_REG_BASE+(CRTC_GAMMA_LUT_OFFSET*(x))+0x20A)
#define CRTCX_GAMMA_LUTB_ADDR(x)        (CRTC1_REG_BASE+(CRTC_GAMMA_LUT_OFFSET*(x))+0x20C)
#define CRTCX_GAMMA_LUTB_DATA(x)        (CRTC1_REG_BASE+(CRTC_GAMMA_LUT_OFFSET*(x))+0x20E)

/* CRTC HPD MASK */
#define CRTC_HPD_ST_MASK                (1u<<0)  /* mask fo hot plug pin status */

#define CRTC_HPD_INT_ST_MASK            (1u<<1)  /* mask for hot plug detect (HPD) interrupt status */
#define CRTC_HPD_INT_EN_MASK            (1u<<0)  /* mask for hot plug detect (HPD) interrupt enable */

#define CRTC_VSYNC_INT_MASK             (1u<<3)  /* mask for vsync/vblank interrupt status */

#define CRTC_BW_ERR_ST_MASK             (1u<<0)  /* mask for bandwidth error status */
#define CRTC_BW_ERR_ST_INT_MASK         (1u<<1)  /* mask for bandwidth error interrupt status */

#define CRTC_PFLIP_PENDING_MASK         (1u<<4)  /* mask for page flip pending status */
#define CRTC_STATUS_VBLANK_MASK         (1u<<0)  /* mask for vblank */

#define LAYER1_REG_OFFSET               (REG_VGA_OFFSET + 0x100)    /* layer1 controller register offset */
#define LAYER_REG_SIZE                  0x20
#define LAYER1_CTRL_ADDR                (LAYER1_REG_OFFSET + 0x0)
#   define LAYER_CTRL_EN_MASK               (1U<<0) 	/* layer enable */
#   define LAYER_CTRL_TRANS_COLOR_EN_MASK   (1U<<1) 	/* enable transparent color */
#   define LAYER_CTRL_BWE_INT_EN_MASK       (1U<<12)	/* bandwidth error interrupt enable */
#define LAYER1_COLOR_MODE_ADDR          (LAYER1_REG_OFFSET + 0x2)
#define LAYER1_HPOSITION_ADDR           (LAYER1_REG_OFFSET + 0x4)
#define LAYER1_VPOSITION_ADDR           (LAYER1_REG_OFFSET + 0x6)
#define LAYER1_WIDTH_ADDR               (LAYER1_REG_OFFSET + 0x8)
#define LAYER1_HEIGHT_ADDR              (LAYER1_REG_OFFSET + 0xA)
#define LAYER1_OFFSET_ADDR              (LAYER1_REG_OFFSET + 0xC)
#define LAYER1_ALPHA_ADDR               (LAYER1_REG_OFFSET + 0xE)
#define LAYER1_TRANSPARENT_ADDR         (LAYER1_REG_OFFSET + 0x10)
#define LAYER1_BUF0_START_ADDR          (LAYER1_REG_OFFSET + 0x14)
#define LAYER1_BUF1_START_ADDR          (LAYER1_REG_OFFSET + 0x18)
#define LAYER1_BW_ERR_ST_ADDR           (LAYER1_REG_OFFSET + 0x1C)

#define ZV0_REG_OFFSET                  (REG_VGA_OFFSET + 0x180)    /* zoom video 0 controller register offset */
#define ZV_REG_SIZE                     0x10

#define ZV0_CTRL_ADDR                   (ZV0_REG_OFFSET + 0x0)	/* [23:0] */
#define ZV0_INT_ST_ADDR                 (ZV0_REG_OFFSET + 0x3)
#define ZV0_BUF0_ADDR                   (ZV0_REG_OFFSET + 0x4)
#define ZV0_BUF1_ADDR                   (ZV0_REG_OFFSET + 0x8)
#define ZV0_WIDTH_ADDR                  (ZV0_REG_OFFSET + 0xC)
#define ZV0_HEIGHT_ADDR                 (ZV0_REG_OFFSET + 0xE)

    /* hardware cursor register */
#define HW_CURSOR_BASE                  REG_VGA_OFFSET
#define CURSOR_CTRL                     (0x000 + HW_CURSOR_BASE)
#define CURSOR_H                        (0x004 + HW_CURSOR_BASE)
#define CURSOR_V                        (0x006 + HW_CURSOR_BASE)
#define CURSOR_STARTADDR                (0x010 + HW_CURSOR_BASE)

/******************************************************************************
 *                                    函数声明                                *
 ******************************************************************************/

int  ljm_vgav2_init(int card, int count);
int  ljm_vgav2_crtc_mode_set(int card, int crtc_id, const struct drm_display_mode *mode, bool is_fpga);
void ljm_vgav2_crtc_enable(int card, int crtc_id);
void ljm_vgav2_crtc_disable(int card, int crtc_id);
void ljm_vgav2_set_hdmiclk(int card, int crtc_id);
int  ljm_vgav2_crtc_set_base(int card, int crtc_id, u64 crtc_base);
int  ljm_vgav2_crtc_mode_set_base(int card, int crtc_id, int bpp, int pitches, u64 crtc_base);
int ljm_vgav2_crtc_prog_reset(int card, int reg_val);
u32  ljm_vgav2_get_vblank_counter(int card, int crtc_id);
bool ljm_vgav2_page_flip_pending(int card, int crtc_id);
int  ljm_vgav2_get_crtc_scanout_pos(int card, unsigned int pipe, int *vpos, int *hpos);
bool ljm_vgav2_crtc_mode_valid(const struct drm_display_mode *mode, bool is_fpga);
int ljm_vgav2_pllcalc(unsigned int input_freq, unsigned int no2_range, unsigned int output_freq, unsigned int *pll_para);
void ljm_dvo_hpd_enable(int card, int index, int enable);
void ljm_dvo_dpms(int card, int index, int mode);

#endif /* ~__LJMICRO_VGAV2_H__ */
