/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/*******************************************************************************
 *                                   标准库头文件                               *
 ******************************************************************************/
#include <linux/spinlock.h>
#include <linux/list.h>
/*******************************************************************************
 *                                 非标准库头文件                               *
 ******************************************************************************/
#include "ljmicro_zoomvideo.h"
#include "ljmicro_device.h"
#include "ljmicro_vgav2.h"
#include "lj_hal_drm.h"
#include "linux/cdev.h"
#define LJM_ZV_REG_R8(p, addr)       REG_R8((p)->cur_card, (addr) + ZV_REG_SIZE * ((p)->id))
#define LJM_ZV_REG_W8(p, addr, val)  REG_W8((p)->cur_card, (addr) + ZV_REG_SIZE * ((p)->id), (val))
#define LJM_ZV_REG_R16(p, addr)      REG_R16((p)->cur_card, (addr) + ZV_REG_SIZE * ((p)->id))
#define LJM_ZV_REG_W16(p, addr, val) REG_W16((p)->cur_card, (addr) + ZV_REG_SIZE * ((p)->id), (val))
#define LJM_ZV_REG_R32(p, addr)      REG_R32((p)->cur_card, (addr) + ZV_REG_SIZE * ((p)->id))
#define LJM_ZV_REG_W32(p, addr, val) REG_W32((p)->cur_card, (addr) + ZV_REG_SIZE * ((p)->id), (val))

/*******************************************************************************
 *                                 数据结构                                     *
 ******************************************************************************/
struct ljm_zoomvideo_device
{
    struct class *class;
    struct device *device;

    dev_t       devid;
    struct cdev cdev;

    /** @filp: Pointer to the core file structure. */
    struct file *file;

    wait_queue_head_t event_wait;
    spinlock_t event_lock;

    /**
     * @pending_events:
     * 已经被挂起的帧完成事件，在zoomvideo_device_read()函数中清零，表示已被事件已被处理。
     * 按bitmap方式表示zv，bit0表示zv0，bit1表示zv1以此类推。
    */
    unsigned char pending_events;

};

/******************************************************************************
 *                                     全局变量                                *
 ******************************************************************************/
const static struct ljm_zoomvideo_config default_cfg =
{
    .enable = 0,
    .av_code = 1,
    .eav_repeat = 1,
    .vblank = 1,
};

struct ljm_zoomvideo_device ljm_zv_dev;
/******************************************************************************
 *                                     函数实现                                *
 ******************************************************************************/
static int zoomvideo_device_open(struct inode *inode, struct file *filp)
{
    init_waitqueue_head(&ljm_zv_dev.event_wait);
    filp->private_data = &ljm_zv_dev;
    ljm_zv_dev.file = filp;

    return 0;
}

static int zoomvideo_device_release(struct inode *inode, struct file *filp)
{
    ljm_zv_dev.file = NULL;
    return 0;
}

static ssize_t zoomvideo_device_read(struct file *filp, char __user *buffer,
         size_t count, loff_t *offset)
{
    struct ljm_zoomvideo_device *zv_dev = filp->private_data;
    int ret = 0;
    unsigned char bitmap;

    spin_lock_irq(&zv_dev->event_lock);
    bitmap = zv_dev->pending_events;
    spin_unlock_irq(&zv_dev->event_lock);
    if(bitmap)
    {
        if (copy_to_user(buffer, &bitmap, sizeof(bitmap)))
        {
            ret = -EFAULT;
            goto end;
        }
        ret = sizeof(bitmap);
    }

end:
    return ret;
}

static __poll_t zoomvideo_device_poll(struct file *filp, struct poll_table_struct *wait)
{
    struct ljm_zoomvideo_device *zv_dev = filp->private_data;
    __poll_t mask = 0;

    poll_wait(filp, &zv_dev->event_wait, wait);

    if (zv_dev->pending_events)
        mask |= EPOLLIN | EPOLLRDNORM;

    return mask;
}

static struct file_operations zoomvideo_fops = {
    .owner = THIS_MODULE,
    .open = zoomvideo_device_open,
    .release = zoomvideo_device_release,
    .read = zoomvideo_device_read,
    .poll = zoomvideo_device_poll,
};

static int ljm_zoomvideo_device_create(void)
{
    int ret;

    ret = alloc_chrdev_region(&ljm_zv_dev.devid, 0, 1, LJM_ZOOMVIDEO_DEVICE_NAME);
    if(ret < 0)
        return ret;

    cdev_init(&ljm_zv_dev.cdev, &zoomvideo_fops);
    cdev_add(&ljm_zv_dev.cdev, ljm_zv_dev.devid, 1);

#if LINUX_VERSION_CODE < KERNEL_VERSION(6,6,0)
    ljm_zv_dev.class = class_create(THIS_MODULE, "zoomvideo");
#else
    ljm_zv_dev.class = class_create("zoomvideo");
#endif
    ljm_zv_dev.device = device_create(ljm_zv_dev.class, NULL, ljm_zv_dev.devid, NULL, LJM_ZOOMVIDEO_DEVICE_NAME);

    spin_lock_init(&ljm_zv_dev.event_lock);

    return 0;
}

int ljm_zoomvideo_init(struct ljm_device *ljm_dev)
{
    int               i, ret;
    struct ljm_zoomvideo *ljm_zv;
    uint32_t *val = (uint32_t*)&default_cfg;

    for (i = 0; i < LJM_MAX_ZVIDEOS; i++)
    {
        ljm_zv = kmalloc(sizeof(struct ljm_zoomvideo), GFP_KERNEL);;
        if (!ljm_zv)
            return -ENOMEM;

        ljm_dev->num_zvideos++;
        ljm_zv->id       = i;
        ljm_zv->cur_card = ljm_dev->cur_card;
        sprintf(ljm_zv->name, "zoomvideo-%d", i);
        LJM_ZV_REG_W32(ljm_zv, ZV0_CTRL_ADDR, *val);
        ljm_zv->lock = false;
        ljm_zv->cfg = default_cfg;
        ljm_dev->zvideos[i] = ljm_zv;
    }

    ret = ljm_zoomvideo_device_create();
    if(ret)
        ERR_PRINT("Faild to create zoomvideo device. (%d)\n", ret);

    return 0;
}

void ljm_zoomvideo_fini(struct ljm_device *ljm_dev)
{
    return ;
}

bool ljm_zv_idle(struct ljm_zoomvideo *zv)
{
    return !zv->lock;
}

struct ljm_zoomvideo *ljm_zvideo_find(struct ljm_device *ljm_dev, int zv_id)
{
    int i;

    for (i = 0; i < ljm_dev->num_zvideos; i++)
    {
        if (ljm_dev->zvideos[i]->id == zv_id)
            return ljm_dev->zvideos[i];
    }

    return NULL;
}

int ljm_ioctl_zv_free(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    return 0;
}

static int ljm_zv_property_set_maplayer(struct ljm_zoomvideo *zv, int layer_id)
{
    uint32_t *cfg_val;

    switch(layer_id)
    {
        case 0: zv->cfg.tolayer1 = 1; break;
        case 1: zv->cfg.tolayer2 = 1; break;
        case 2: zv->cfg.tolayer3 = 1; break;
        case 3: zv->cfg.tolayer4 = 1; break;
        default: break;
    }

    cfg_val = (uint32_t*)&zv->cfg;
    LJM_ZV_REG_W32(zv, ZV0_CTRL_ADDR, *cfg_val);

    return 0;
}

int ljm_ioctl_zv_property_set(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    struct ljm_device *        ljm_dev = GetLjmDevFrmDrm(dev);
    struct ljm_zvideo_property *request = (struct ljm_zvideo_property *)data;
    struct ljm_zoomvideo        *zv;
    uint64_t *values;
    uint32_t *cfg_val;

    zv = ljm_zvideo_find(ljm_dev, request->zv_id);
    if (!zv)
    {
        ERR_PRINT("Invalid ljm_zoomvideo id for setting.\n");
        return -EINVAL;
    }

    if(request->num_values == 0)
    {
        switch (request->name)
        {
            case ZV_TRYLOCK:
                if(!zv->lock)
                    zv->lock = true;
                else
                    return -EBUSY;
                break;
            case ZV_UNLOCK:
                zv->lock = false;
                zv->cfg.enable = false;
                cfg_val = (uint32_t*)&zv->cfg;
                LJM_ZV_REG_W32(zv, ZV0_CTRL_ADDR, *cfg_val);
                break;
            default:
                ERR_PRINT("Unknown cmommand: %d\n", request->name);
                break;
        }
        return 0;
    }

    values = kzalloc(sizeof(uint64_t) * request->num_values, GFP_KERNEL);
    if (copy_from_user(values, request->values, sizeof(uint64_t) * request->num_values))
        return -EFAULT;

    switch (request->name)
    {
        case ZV_PROP_ENABLE:
            zv->cfg.enable = !!values[0];
            cfg_val = (uint32_t*)&zv->cfg;
            LJM_ZV_REG_W32(zv, ZV0_CTRL_ADDR, *cfg_val);
            break;
        case ZV_PROP_BUFF0:
            LJM_ZV_REG_W32(zv, ZV0_BUF0_ADDR, values[0] >> 2);
            break;
        case ZV_PROP_BUFF1:
            LJM_ZV_REG_W32(zv, ZV0_BUF1_ADDR, values[0] >> 2);
            break;
        case ZV_PROP_MAPLAYER:
            ljm_zv_property_set_maplayer(zv, (int)values[0]);
            break;
        default:
            ERR_PRINT("Unknown cmommand: %d\n", request->name);
            kfree(values);
            return -EINVAL;
    }

    kfree(values);
    return 0;
}

int ljm_ioctl_zv_property_get(struct drm_device *dev, void *data, struct drm_file *file_priv)
{
    struct ljm_device *        ljm_dev = GetLjmDevFrmDrm(dev);
    struct ljm_zvideo_property *request = (struct ljm_zvideo_property *)data;
    struct ljm_zoomvideo        *zv;
    uint64_t *values;

    zv = ljm_zvideo_find(ljm_dev, request->zv_id);
    if (!zv)
    {
        ERR_PRINT("Invalid ljm_zoomvideo id for setting.\n");
        return -EINVAL;
    }

    if(request->num_values <=0)
    {
        ERR_PRINT("No space for driver's writing.\n");
        return -EINVAL;
    }

    values = kzalloc(sizeof(uint64_t) * request->num_values, GFP_KERNEL);

    switch (request->name)
    {
        case ZV_PROP_WHITH:
            values[0] = LJM_ZV_REG_R16(zv, ZV0_WIDTH_ADDR);
            break;
        case ZV_PROP_HEIGHT:
            values[0] = LJM_ZV_REG_R16(zv, ZV0_HEIGHT_ADDR);
            break;
        default:
            ERR_PRINT("Unknown cmommand: %d\n", request->name);
            kfree(values);
            return -EINVAL;
    }

    if (copy_to_user(request->values, values, sizeof(uint64_t) * request->num_values))
        return -EFAULT;

    kfree(values);
    return 0;
}

int ljm_zoomvideo_irq_handle(int zoomvideo_id)
{
    /* 若zoomvideo设备文件没有被打开，则忽略中断 */
    if(!ljm_zv_dev.file)
        return 0;

    ljm_zv_dev.pending_events |= BIT(zoomvideo_id);
    wake_up(&ljm_zv_dev.event_wait);
    return 0;
}