/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __ZOOMVIDEO_H__
#define __ZOOMVIDEO_H__
#include "ljmicro_base.h"
#include <linux/list.h>

struct ljm_zoomvideo_config
{
    /**
     * @enalbe: 视频捕捉使能；0-禁止，1-使能。
    */
    uint32_t enable : 1;

    /**
     * @csc:
     * Color space conversion - 色彩空间转换.
     *
     * 0: zoomvideo源使用YUV模式；
     * 1: zoomvideo源使用RGB模式.
    */
    uint32_t csc : 1;

    /**
     * @href: 行同步信号极性, 0-正 1-负;
     * @vref: 场同步信号极性, 0-正 1-负;
    */
    uint32_t href : 1;
    uint32_t vref : 1;

    /**
     * @hcapture: 视频水平捕获方向， 0-从左到右， 1-从右到左;
     * @vsamp: 视频水平捕获方向， 0-从左到右， 1-从右到左;
    */
    uint32_t hcapture : 1;
    uint32_t vcapture : 1;

    /**
     * @y_uv: Y-UV数据交换, 0-正常，1-Y/UV数据交换;
     * @u_v: UV数据交换, 0-正常，1-U/V数据交换;
     *
     */
    uint32_t y_uv : 1;
    uint32_t u_v : 1;

    /**
     * @tolayer1: 将zoomvideo视频流映射到图层layer1, 0-禁止, 1-使能;
     * @tolayer2: 将zoomvideo视频流映射到图层layer2, 0-禁止, 1-使能;
    */
    uint32_t tolayer1 : 1;
    uint32_t tolayer2 : 1;

    /**
     * @interlaced: 隔行扫描, 0-逐行扫描模式，1-隔行扫描模式.
    */
    uint32_t interlaced : 1;

    uint32_t tolayer3 : 1;
    uint32_t tolayer4 : 1;

    /**
     * @parity: Zoom Video Port 1 field极性信号; 0(0----odd,1---even) ;1(0---even,1---odd)。
    */
    uint32_t parity : 1;

    /**
     * @av_code: 采用field/href/vref标识行场信号；1-BT模式下，采用AV Code标识行场。
    */
    uint32_t av_code : 1;

    /**
     * @eav_repeat:  0-16-Bit模式下EAV/SAV Code 分开，1-16-Bit模式下EAV/SAV Code重复.
    */
    uint32_t eav_repeat : 1;

    /**
     * @bt656_bits: 0-16-Bit BT1120/BT656模式，1-8-Bit BT1120/656模式。
     * @bt656_EAV_order: BT656/1120模式下，0-EAV/SAV Code正常顺序，4thByte；1-EAV/SAV Code交换，3rdByte。
    */
    uint32_t bt656_bits : 1;
    uint32_t bt656_EAV_order : 1;

    /**
     * @yuv_quantification: YUV数据量化情况，0-未量化，1-已量化。
    */
    uint32_t yuv_quantification : 1;

    /**
     * @encode_fmt: 转换参数选择，0-BT601,1-BT709,2-BT2020.
    */
    uint32_t encode_fmt : 2;

    /**
     * @valid_bits: 8-Bit Data模式，16位数据中低8位数据有效；1，8-Bit Data模式，16位数据中高8位数据有效。
    */
    uint32_t valid_bits : 1;

    uint32_t : 1;

    /**
     * @vblank: 帧中断使能，0，禁止中断，1，使能中断。
    */
    uint32_t vblank : 1;

    uint32_t reserved : 8;
};

struct ljm_zoomvideo
{
    int id;
    int cur_card;
    char name[20];

    /**
     * @lock: 当zoomvideo为lock状态时，表示该zoomvideo已被应用层使用，
     * 其他用户态进程无法获取该zoomvideo。
    */
    bool lock;

    struct ljm_zoomvideo_config cfg;
    int w, h;
};

int ljm_ioctl_zv_property_set(struct drm_device *dev, void *data, struct drm_file *file_priv);
int ljm_ioctl_zv_property_get(struct drm_device *dev, void *data, struct drm_file *file_priv);
int ljm_zoomvideo_irq_handle(int zoomvideo_id);

#endif /*__ZOOMVIDEO_H__*/