/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_profiler_shared_h_
#define __lj_hal_profiler_shared_h_

#ifdef __cplusplus
extern "C" {
#endif

#define ANDROID_PROFILER_COUNTERS   1
#define FPGA_INFO                   0
#define RECORD_COUNTER_ADDRESS      0

/* HW profile information. */
typedef struct _ljsPROFILER_COUNTERS_PART1 {
    ljtUINT32       gpuTotalRead64BytesPerFrame;
    ljtUINT32       gpuTotalWrite64BytesPerFrame;

    /* FE */
    ljtUINT32       fe_draw_count;
    ljtUINT32       fe_out_vertex_count;
    ljtUINT32       fe_cache_miss_count;
    ljtUINT32       fe_cache_lk_count;
    ljtUINT32       fe_stall_count;
    ljtUINT32       fe_starve_count;
    ljtUINT32       fe_process_count;

    /* PE */
    ljtUINT32       pe0_pixel_count_killed_by_color_pipe;
    ljtUINT32       pe0_pixel_count_killed_by_depth_pipe;
    ljtUINT32       pe0_pixel_count_drawn_by_color_pipe;
    ljtUINT32       pe0_pixel_count_drawn_by_depth_pipe;
    ljtUINT32       pe1_pixel_count_killed_by_color_pipe;
    ljtUINT32       pe1_pixel_count_killed_by_depth_pipe;
    ljtUINT32       pe1_pixel_count_drawn_by_color_pipe;
    ljtUINT32       pe1_pixel_count_drawn_by_depth_pipe;

    /* SH */
    ljtUINT32       shader_cycle_count;
    ljtUINT32       vs_shader_cycle_count;
    ljtUINT32       ps_shader_cycle_count;
    ljtUINT32       ps_inst_counter;
    ljtUINT32       ps_rendered_pixel_counter;
    ljtUINT32       vs_inst_counter;
    ljtUINT32       vs_rendered_vertice_counter;
    ljtUINT32       vs_branch_inst_counter;
    ljtUINT32       vs_texld_inst_counter;
    ljtUINT32       ps_branch_inst_counter;
    ljtUINT32       ps_texld_inst_counter;
    ljtUINT32       vs_non_idle_starve_count;
    ljtUINT32       vs_starve_count;
    ljtUINT32       vs_stall_count;
    ljtUINT32       vs_process_count;
    ljtUINT32       ps_non_idle_starve_count;
    ljtUINT32       ps_starve_count;
    ljtUINT32       ps_stall_count;
    ljtUINT32       ps_process_count;

    /* PA */
    ljtUINT32       pa_input_vtx_counter;
    ljtUINT32       pa_input_prim_counter;
    ljtUINT32       pa_output_prim_counter;
    ljtUINT32       pa_depth_clipped_counter;
    ljtUINT32       pa_trivial_rejected_counter;
    ljtUINT32       pa_culled_prim_counter;
    ljtUINT32       pa_droped_prim_counter;
    ljtUINT32       pa_frustum_clipped_prim_counter;
    ljtUINT32       pa_frustum_clipdroped_prim_counter;
    ljtUINT32       pa_non_idle_starve_count;
    ljtUINT32       pa_starve_count;
    ljtUINT32       pa_stall_count;
    ljtUINT32       pa_process_count;

    /* SE */
    ljtUINT32       se_culled_triangle_count;
    ljtUINT32       se_culled_lines_count;
    ljtUINT32       se_clipped_triangle_count;
    ljtUINT32       se_clipped_line_count;
    ljtUINT32       se_starve_count;
    ljtUINT32       se_stall_count;
    ljtUINT32       se_receive_triangle_count;
    ljtUINT32       se_send_triangle_count;
    ljtUINT32       se_receive_lines_count;
    ljtUINT32       se_send_lines_count;
    ljtUINT32       se_process_count;
    ljtUINT32       se_trivial_rejected_line_count;
    ljtUINT32       se_non_idle_starve_count;

    /* RA */
    ljtUINT32       ra_input_prim_count;
    ljtUINT32       ra_total_quad_count;
    ljtUINT32       ra_valid_quad_count_after_early_z;
    ljtUINT32       ra_valid_pixel_count_to_render;
    ljtUINT32       ra_output_valid_quad_count;
    ljtUINT32       ra_output_valid_pixel_count;
    ljtUINT32       ra_pipe_cache_miss_counter;
    ljtUINT32       ra_pipe_hz_cache_miss_counter;
    ljtUINT32       ra_prefetch_cache_miss_counter;
    ljtUINT32       ra_prefetch_hz_cache_miss_counter;
    ljtUINT32       ra_eez_culled_counter;
    ljtUINT32       ra_non_idle_starve_count;
    ljtUINT32       ra_starve_count;
    ljtUINT32       ra_stall_count;
    ljtUINT32       ra_process_count;

    /* TX */
    ljtUINT32       tx_total_bilinear_requests;
    ljtUINT32       tx_total_trilinear_requests;
    ljtUINT32       tx_total_discarded_texture_requests;
    ljtUINT32       tx_total_texture_requests;
    ljtUINT32       tx_mc0_miss_count;
    ljtUINT32       tx_mc0_request_byte_count;
    ljtUINT32       tx_mc1_miss_count;
    ljtUINT32       tx_mc1_request_byte_count;
    ljtUINT32       tx_non_idle_starve_count;
    ljtUINT32       tx_starve_count;
    ljtUINT32       tx_stall_count;
    ljtUINT32       tx_process_count;
} ljsPROFILER_COUNTERS_PART1;

typedef struct _ljsPROFILER_COUNTERS_PART2 {
    /* MCC */
    ljtUINT32       mcc_total_read_req_8B_from_colorpipe;
    ljtUINT32       mcc_total_read_req_8B_sentout_from_colorpipe;
    ljtUINT32       mcc_total_write_req_8B_from_colorpipe;
    ljtUINT32       mcc_total_read_req_sentout_from_colorpipe;
    ljtUINT32       mcc_total_write_req_from_colorpipe;
    ljtUINT32       mcc_total_read_req_8B_from_depthpipe;
    ljtUINT32       mcc_total_read_req_8B_sentout_from_depthpipe;
    ljtUINT32       mcc_total_write_req_8B_from_depthpipe;
    ljtUINT32       mcc_total_read_req_sentout_from_depthpipe;
    ljtUINT32       mcc_total_write_req_from_depthpipe;
    ljtUINT32       mcc_total_read_req_8B_from_others;
    ljtUINT32       mcc_total_write_req_8B_from_others;
    ljtUINT32       mcc_total_read_req_from_others;
    ljtUINT32       mcc_total_write_req_from_others;
    ljtUINT32       mcc_axi_total_latency;
    ljtUINT32       mcc_axi_sample_count;
    ljtUINT32       mcc_axi_max_latency;
    ljtUINT32       mcc_axi_min_latency;
    ljtUINT32       mc_fe_read_bandwidth;
    ljtUINT32       mc_mmu_read_bandwidth;
    ljtUINT32       mc_blt_read_bandwidth;
    ljtUINT32       mc_sh0_read_bandwidth;
    ljtUINT32       mc_sh1_read_bandwidth;
    ljtUINT32       mc_pe_write_bandwidth;
    ljtUINT32       mc_blt_write_bandwidth;
    ljtUINT32       mc_sh0_write_bandwidth;
    ljtUINT32       mc_sh1_write_bandwidth;

    /* MCZ */
    ljtUINT32       mcz_total_read_req_8B_from_colorpipe;
    ljtUINT32       mcz_total_read_req_8B_sentout_from_colorpipe;
    ljtUINT32       mcz_total_write_req_8B_from_colorpipe;
    ljtUINT32       mcz_total_read_req_sentout_from_colorpipe;
    ljtUINT32       mcz_total_write_req_from_colorpipe;
    ljtUINT32       mcz_total_read_req_8B_from_depthpipe;
    ljtUINT32       mcz_total_read_req_8B_sentout_from_depthpipe;
    ljtUINT32       mcz_total_write_req_8B_from_depthpipe;
    ljtUINT32       mcz_total_read_req_sentout_from_depthpipe;
    ljtUINT32       mcz_total_write_req_from_depthpipe;
    ljtUINT32       mcz_total_read_req_8B_from_others;
    ljtUINT32       mcz_total_write_req_8B_from_others;
    ljtUINT32       mcz_total_read_req_from_others;
    ljtUINT32       mcz_total_write_req_from_others;
    ljtUINT32       mcz_axi_total_latency;
    ljtUINT32       mcz_axi_sample_count;
    ljtUINT32       mcz_axi_max_latency;
    ljtUINT32       mcz_axi_min_latency;

    /* HI */
    ljtUINT32       hi0_total_read_8B_count;
    ljtUINT32       hi0_total_write_8B_count;
    ljtUINT32       hi0_total_read_request_count;
    ljtUINT32       hi0_total_write_request_count;
    ljtUINT32       hi0_axi_cycles_read_request_stalled;
    ljtUINT32       hi0_axi_cycles_write_request_stalled;
    ljtUINT32       hi0_axi_cycles_write_data_stalled;
    ljtUINT32       hi1_total_read_8B_count;
    ljtUINT32       hi1_total_write_8B_count;
    ljtUINT32       hi1_total_read_request_count;
    ljtUINT32       hi1_total_write_request_count;
    ljtUINT32       hi1_axi_cycles_read_request_stalled;
    ljtUINT32       hi1_axi_cycles_write_request_stalled;
    ljtUINT32       hi1_axi_cycles_write_data_stalled;
    ljtUINT32       hi_total_cycle_count;
    ljtUINT32       hi_total_idle_cycle_count;
    ljtUINT32       hi_total_read_8B_count;
    ljtUINT32       hi_total_write_8B_count;
    ljtUINT32       hi_total_readOCB_16B_count;
    ljtUINT32       hi_total_writeOCB_16B_count;

    /* L2 */
    ljtUINT32       l2_total_axi0_read_request_count;
    ljtUINT32       l2_total_axi1_read_request_count;
    ljtUINT32       l2_total_axi0_write_request_count;
    ljtUINT32       l2_total_axi1_write_request_count;
    ljtUINT32       l2_total_read_transactions_request_by_axi0;
    ljtUINT32       l2_total_read_transactions_request_by_axi1;
    ljtUINT32       l2_total_write_transactions_request_by_axi0;
    ljtUINT32       l2_total_write_transactions_request_by_axi1;
    ljtUINT32       l2_axi0_minmax_latency;
    ljtUINT32       l2_axi0_min_latency;
    ljtUINT32       l2_axi0_max_latency;
    ljtUINT32       l2_axi0_total_latency;
    ljtUINT32       l2_axi0_total_request_count;
    ljtUINT32       l2_axi1_minmax_latency;
    ljtUINT32       l2_axi1_min_latency;
    ljtUINT32       l2_axi1_max_latency;
    ljtUINT32       l2_axi1_total_latency;
    ljtUINT32       l2_axi1_total_request_count;
} ljsPROFILER_COUNTERS_PART2;

typedef struct _ljsPROFILER_COUNTERS {
    ljsPROFILER_COUNTERS_PART1 counters_part1;
    ljsPROFILER_COUNTERS_PART2 counters_part2;
} ljsPROFILER_COUNTERS;

typedef enum _ljeVIP_PROBE_COUNTER {
    ljvVIP_PROBE_COUNTER_NEURAL_NET,
    ljvVIP_PROBE_COUNTER_TENSOR_PROCESSOR,
    ljvVIP_PROBE_COUNTER_COUNT
} ljeVIP_PROBE_COUNTER;

/* Mask definations for overflow indicator of TP */
typedef enum _ljeTPCOUNTER_OVERFLOW {
    ljvTPCOUNTER_LAYER_ID_OVERFLOW                  = (1 << 0),
    ljvTPCOUNTER_TOTAL_BUSY_CYCLE_OVERFLOW          = (1 << 1),
    ljvTPCOUNTER_TOTAL_READ_BW_DDR_OVERFLOW         = (1 << 2),
    ljvTPCOUNTER_TOTAL_WRITE_BW_DDR_OVERFLOW        = (1 << 3),
    ljvTPCOUNTER_TOTAL_READ_BW_SRAM_OVERFLOW        = (1 << 4),
    ljvTPCOUNTER_TOTAL_WRITE_BW_SRAM_OVERFLOW       = (1 << 5),
    ljvTPCOUNTER_TOTAL_READ_BW_OCB_OVERFLOW         = (1 << 6),
    ljvTPCOUNTER_TOTAL_WRITE_BW_OCB_OVERFLOW        = (1 << 7),
    ljvTPCOUNTER_FC_PIX_CNT_OVERFLOW                = (1 << 8),
    ljvTPCOUNTER_FC_ZERO_SKIP_OVERFLOW              = (1 << 9),
    ljvTPCOUNTER_FC_COEF_CNT_OVERFLOW               = (1 << 10),
    ljvTPCOUNTER_FC_COEF_ZERO_CNT_OVERFLOW          = (1 << 11),
    ljvTPCOUNTER_TOTAL_IDLE_CYCLE_CORE0_OVERFLOW    = (1 << 0),
    ljvTPCOUNTER_TOTAL_IDLE_CYCLE_CORE1_OVERFLOW    = (1 << 1),
    ljvTPCOUNTER_TOTAL_IDLE_CYCLE_CORE2_OVERFLOW    = (1 << 2),
    ljvTPCOUNTER_TOTAL_IDLE_CYCLE_CORE3_OVERFLOW    = (1 << 3),
} _ljeTPCOUNTER_OVERFLOW;

/* Mask definations for overflow indicator of NN */
typedef enum _ljeNNCOUNTER_OVERFLOW {
    ljvNNCOUNTER_TOTAL_BUSY_CYCLE_OVERFLOW          = (1 << 0),
    ljvNNCOUNTER_TOTAL_READ_CYCLE_DDR_OVERFLOW      = (1 << 2),
    ljvNNCOUNTER_TOTAL_READ_BW_DDR_OVERFLOW         = (1 << 3),
    ljvNNCOUNTER_TOTAL_WRITE_CYCLE_DDR_OVERFLOW     = (1 << 4),
    ljvNNCOUNTER_TOTAL_WRITE_BW_DDR_OVERFLOW        = (1 << 5),
    ljvNNCOUNTER_TOTAL_READ_SYCLE_SRAM_OVERFLOW     = (1 << 6),
    ljvNNCOUNTER_TOTAL_WRITE_CYCLE_SRAM_OVERFLOW    = (1 << 7),
    ljvNNCOUNTER_TOTAL_MAC_CYCLE_OVERFLOW           = (1 << 8),
    ljvNNCOUNTER_TOTAL_MAC_COUNT_OVERFLOW           = (1 << 9),
    ljvNNCOUNTER_ZERO_COEF_SKIP_COUNT_OVERFLOW      = (1 << 10),
    ljvNNCOUNTER_NON_ZERO_COEF_COUNT_OVERFLOW       = (1 << 11),
} _ljeNNCOUNTER_OVERFLOW;

#define   MODULE_NN_RESERVED_COUNTER_NUM            0x9
typedef struct _ljsPROFILER_VIP_PROBE_COUNTERS {
    /* NN */
    ljtUINT32       nn_layer_id;
    ljtUINT32       nn_layer_id_overflow;
    ljtUINT32       nn_instr_info;
    ljtUINT32       nn_total_busy_cycle;
    ljtUINT32       nn_total_busy_cycle_overflow;
    ljtUINT32       nn_total_read_cycle_ddr;
    ljtUINT32       nn_total_read_cycle_ddr_overflow;
    ljtUINT32       nn_total_read_valid_bandwidth_ddr;
    ljtUINT32       nn_total_read_valid_bandwidth_ddr_overflow;
    ljtUINT32       nn_total_write_cycle_ddr;
    ljtUINT32       nn_total_write_cycle_ddr_overflow;
    ljtUINT32       nn_total_write_valid_bandwidth_ddr;
    ljtUINT32       nn_total_write_valid_bandwidth_ddr_overflow;
    ljtUINT32       nn_total_read_cycle_sram;
    ljtUINT32       nn_total_read_cycle_sram_overflow;
    ljtUINT32       nn_total_write_cycle_sram;
    ljtUINT32       nn_total_write_cycle_sram_overflow;
    ljtUINT32       nn_total_mac_cycle;
    ljtUINT32       nn_total_mac_cycle_overflow;
    ljtUINT32       nn_total_mac_count;
    ljtUINT32       nn_total_mac_count_overflow;
    ljtUINT32       nn_zero_coef_skip_count;
    ljtUINT32       nn_zero_coef_skip_count_overflow;
    ljtUINT32       nn_non_zero_coef_count;
    ljtUINT32       nn_non_zero_coef_count_overflow;

    ljtUINT32       nn_reserved_counter[4 * MODULE_NN_RESERVED_COUNTER_NUM];
    ljtUINT32       nn_total_idle_cycle_core_overflow[4];
    ljtUINT32       nn_total_idle_cycle_core[16];
    ljtUINT32       nn_wait_for_in_event_cycle;
    ljtUINT32       nn_wait_for_out_event_cycle;
    ljtUINT32       nn_wait_for_in_event_cycle_overflow;
    ljtUINT32       nn_wait_for_out_event_cycle_overflow;
    ljtUINT32       nn_wait_for_in0_event_cycle;
    ljtUINT32       nn_wait_for_in1_event_cycle;
    ljtUINT32       nn_wait_for_in0_event_cycle_overflow;
    ljtUINT32       nn_wait_for_in1_event_cycle_overflow;
    ljtUINT32       nn_wait_for_out0_event_cycle;
    ljtUINT32       nn_wait_for_out1_event_cycle;
    ljtUINT32       nn_wait_for_out2_event_cycle;
    ljtUINT32       nn_wait_for_out0_event_cycle_overflow;
    ljtUINT32       nn_wait_for_out1_event_cycle_overflow;
    ljtUINT32       nn_wait_for_out2_event_cycle_overflow;
    ljtUINT32       nn_idel_cycle;
    ljtUINT32       nn_idel_cycle_overflow;

    /* TP */
    ljtUINT32       tp_layer_id;
    ljtUINT32       tp_layer_id_overflow;
    ljtUINT32       tp_total_busy_cycle;
    ljtUINT32       tp_total_busy_cycle_overflow;

    ljtUINT32       tp_total_read_bandwidth_cache;
    ljtUINT32       tp_total_read_bandwidth_cache_overflow;
    ljtUINT32       tp_total_write_bandwidth_cache;
    ljtUINT32       tp_total_write_bandwidth_cache_overflow;

    ljtUINT32       tp_total_read_bandwidth_sram;
    ljtUINT32       tp_total_read_bandwidth_sram_overflow;
    ljtUINT32       tp_total_write_bandwidth_sram;
    ljtUINT32       tp_total_write_bandwidth_sram_overflow;

    ljtUINT32       tp_total_read_bandwidth_ocb;
    ljtUINT32       tp_total_read_bandwidth_ocb_overflow;
    ljtUINT32       tp_total_write_bandwidth_ocb;
    ljtUINT32       tp_total_write_bandwidth_ocb_overflow;

    ljtUINT32       tp_fc_pix_count;
    ljtUINT32       tp_fc_zero_skip_count;
    ljtUINT32       tp_fc_pix_count_overflow;
    ljtUINT32       tp_fc_zero_skip_count_overflow;

    ljtUINT32       tp_fc_coef_count;
    ljtUINT32       tp_fc_coef_zero_count;
    ljtUINT32       tp_fc_coef_count_overflow;
    ljtUINT32       tp_fc_coef_zero_count_overflow;

    ljtUINT32       tp_total_idle_cycle_core[16];
    ljtUINT32       tp_total_idle_cycle_core_overflows[16];
} ljsPROFILER_VIP_PROBE_COUNTERS;

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_profiler_shared_h_ */



